%    The contents of this file are subject to the Mozilla Public License
%    Version  1.1  (the "License"); you may not use this file except in
%    compliance with the License. You may obtain a copy of the License at:
%
%    http://www.mozilla.org/MPL/
%
%    Software  distributed  under  the License is distributed on an "AS
%    IS"  basis,  WITHOUT  WARRANTY  OF  ANY  KIND,  either  express or
%    implied.
%
% Purpose: Use to factorise the Size and NCol so that we have less facts
%          (called manually to generate part of tables.pl for the tables of a given combinatorial object)
% Author : Nicolas Beldiceanu, IMT Atlantique

% Each table has the following five arguments:
%  . the name of the table
%  . the value of MaxN (i.e. the maximum number of vertices) used to generate the table
%  . the number of columns of the table
%  . the number of parameters of the lower/upper bound
%  . the lower/upper bound considered

:- dynamic table/5.

top :-
    findall(Name, table(Name,_,_,_,_), TableNames),
    sort(TableNames, Names),
    gen_table_params(Names).

gen_table_params([]) :- !.
gen_table_params([Name|R]) :-
    findall(Size-NCol, table(Name, Size, NCol, _, _), Pairs),
    once(table(Name, _, _, NParams, Bound)),
    write('table(graph, '),
    write(Name),
    write(', '),
    write(Pairs),
    write(', '),
    write(NParams),
    write(', '),
    write(Bound),
    write(').'),
    nl,
    gen_table_params(R).

table(low_c_v, 2, 18, 1, low(v)).
table(low_minc_v, 2, 11, 1, low(v)).
table(low_maxc_v, 2, 11, 1, low(v)).
table(low_s_v, 2, 11, 1, low(v)).
table(low_mins_v, 2, 18, 1, low(v)).
table(low_maxs_v, 2, 18, 1, low(v)).
table(low_c_minc_v, 2, 11, 2, low(v)).
table(low_c_maxc_v, 2, 11, 2, low(v)).
table(low_c_s_v, 2, 18, 2, low(v)).
table(low_c_mins_v, 2, 18, 2, low(v)).
table(low_c_maxs_v, 2, 18, 2, low(v)).
table(low_minc_maxc_v, 2, 11, 2, low(v)).
table(low_minc_s_v, 2, 18, 2, low(v)).
table(low_minc_mins_v, 2, 18, 2, low(v)).
table(low_minc_maxs_v, 2, 18, 2, low(v)).
table(low_maxc_s_v, 2, 18, 2, low(v)).
table(low_maxc_mins_v, 2, 18, 2, low(v)).
table(low_maxc_maxs_v, 2, 18, 2, low(v)).
table(low_s_mins_v, 2, 11, 2, low(v)).
table(low_s_maxs_v, 2, 11, 2, low(v)).
table(low_mins_maxs_v, 2, 18, 2, low(v)).
table(low_c_minc_maxc_v, 2, 11, 3, low(v)).
table(low_c_minc_s_v, 2, 18, 3, low(v)).
table(low_c_minc_mins_v, 2, 18, 3, low(v)).
table(low_c_minc_maxs_v, 2, 18, 3, low(v)).
table(low_c_maxc_s_v, 2, 18, 3, low(v)).
table(low_c_maxc_mins_v, 2, 18, 3, low(v)).
table(low_c_maxc_maxs_v, 2, 18, 3, low(v)).
table(low_c_s_mins_v, 2, 18, 3, low(v)).
table(low_c_s_maxs_v, 2, 18, 3, low(v)).
table(low_c_mins_maxs_v, 2, 18, 3, low(v)).
table(low_minc_maxc_s_v, 2, 18, 3, low(v)).
table(low_minc_maxc_mins_v, 2, 18, 3, low(v)).
table(low_minc_maxc_maxs_v, 2, 18, 3, low(v)).
table(low_minc_s_mins_v, 2, 18, 3, low(v)).
table(low_minc_s_maxs_v, 2, 18, 3, low(v)).
table(low_minc_mins_maxs_v, 2, 18, 3, low(v)).
table(low_maxc_s_mins_v, 2, 18, 3, low(v)).
table(low_maxc_s_maxs_v, 2, 18, 3, low(v)).
table(low_maxc_mins_maxs_v, 2, 18, 3, low(v)).
table(low_s_mins_maxs_v, 2, 11, 3, low(v)).
table(low_c_minc_maxc_s_v, 2, 18, 4, low(v)).
table(low_c_minc_maxc_mins_v, 2, 18, 4, low(v)).
table(low_c_minc_maxc_maxs_v, 2, 18, 4, low(v)).
table(low_c_minc_s_mins_v, 2, 18, 4, low(v)).
table(low_c_minc_s_maxs_v, 2, 18, 4, low(v)).
table(low_c_minc_mins_maxs_v, 2, 18, 4, low(v)).
table(low_c_maxc_s_mins_v, 2, 18, 4, low(v)).
table(low_c_maxc_s_maxs_v, 2, 18, 4, low(v)).
table(low_c_maxc_mins_maxs_v, 2, 18, 4, low(v)).
table(low_c_s_mins_maxs_v, 2, 18, 4, low(v)).
table(low_minc_maxc_s_mins_v, 2, 18, 4, low(v)).
table(low_minc_maxc_s_maxs_v, 2, 18, 4, low(v)).
table(low_minc_maxc_mins_maxs_v, 2, 18, 4, low(v)).
table(low_minc_s_mins_maxs_v, 2, 18, 4, low(v)).
table(low_maxc_s_mins_maxs_v, 2, 18, 4, low(v)).
table(low_c_minc_maxc_s_mins_v, 2, 18, 5, low(v)).
table(low_c_minc_maxc_s_maxs_v, 2, 18, 5, low(v)).
table(low_c_minc_maxc_mins_maxs_v, 2, 18, 5, low(v)).
table(low_c_minc_s_mins_maxs_v, 2, 18, 5, low(v)).
table(low_c_maxc_s_mins_maxs_v, 2, 18, 5, low(v)).
table(low_minc_maxc_s_mins_maxs_v, 2, 18, 5, low(v)).
table(low_c_minc_maxc_s_mins_maxs_v, 2, 18, 6, low(v)).
table(up_c_maxc_v, 2, 11, 2, up(v)).
table(up_maxc_s_v, 2, 18, 2, up(v)).
table(up_s_maxs_v, 2, 11, 2, up(v)).
table(up_c_minc_maxc_v, 2, 11, 3, up(v)).
table(up_c_maxc_s_v, 2, 18, 3, up(v)).
table(up_c_maxc_mins_v, 2, 18, 3, up(v)).
table(up_c_maxc_maxs_v, 2, 18, 3, up(v)).
table(up_c_s_maxs_v, 2, 18, 3, up(v)).
table(up_minc_maxc_s_v, 2, 18, 3, up(v)).
table(up_minc_s_maxs_v, 2, 18, 3, up(v)).
table(up_maxc_s_mins_v, 2, 18, 3, up(v)).
table(up_maxc_s_maxs_v, 2, 18, 3, up(v)).
table(up_s_mins_maxs_v, 2, 11, 3, up(v)).
table(up_c_minc_maxc_s_v, 2, 18, 4, up(v)).
table(up_c_minc_maxc_mins_v, 2, 18, 4, up(v)).
table(up_c_minc_maxc_maxs_v, 2, 18, 4, up(v)).
table(up_c_minc_s_maxs_v, 2, 18, 4, up(v)).
table(up_c_maxc_s_mins_v, 2, 18, 4, up(v)).
table(up_c_maxc_s_maxs_v, 2, 18, 4, up(v)).
table(up_c_maxc_mins_maxs_v, 2, 18, 4, up(v)).
table(up_c_s_mins_maxs_v, 2, 18, 4, up(v)).
table(up_minc_maxc_s_mins_v, 2, 18, 4, up(v)).
table(up_minc_maxc_s_maxs_v, 2, 18, 4, up(v)).
table(up_minc_s_mins_maxs_v, 2, 18, 4, up(v)).
table(up_maxc_s_mins_maxs_v, 2, 18, 4, up(v)).
table(up_c_minc_maxc_s_mins_v, 2, 18, 5, up(v)).
table(up_c_minc_maxc_s_maxs_v, 2, 18, 5, up(v)).
table(up_c_minc_maxc_mins_maxs_v, 2, 18, 5, up(v)).
table(up_c_minc_s_mins_maxs_v, 2, 18, 5, up(v)).
table(up_c_maxc_s_mins_maxs_v, 2, 18, 5, up(v)).
table(up_minc_maxc_s_mins_maxs_v, 2, 18, 5, up(v)).
table(up_c_minc_maxc_s_mins_maxs_v, 2, 18, 6, up(v)).
table(low_v_c, 2, 11, 1, low(c)).
table(low_minc_c, 2, 11, 1, low(c)).
table(low_maxc_c, 2, 11, 1, low(c)).
table(low_s_c, 2, 11, 1, low(c)).
table(low_mins_c, 2, 10, 1, low(c)).
table(low_maxs_c, 2, 10, 1, low(c)).
table(low_v_minc_c, 2, 11, 2, low(c)).
table(low_v_maxc_c, 2, 11, 2, low(c)).
table(low_v_s_c, 2, 18, 2, low(c)).
table(low_v_mins_c, 2, 18, 2, low(c)).
table(low_v_maxs_c, 2, 18, 2, low(c)).
table(low_minc_maxc_c, 2, 11, 2, low(c)).
table(low_minc_s_c, 2, 18, 2, low(c)).
table(low_minc_mins_c, 2, 18, 2, low(c)).
table(low_minc_maxs_c, 2, 18, 2, low(c)).
table(low_maxc_s_c, 2, 18, 2, low(c)).
table(low_maxc_mins_c, 2, 18, 2, low(c)).
table(low_maxc_maxs_c, 2, 18, 2, low(c)).
table(low_s_mins_c, 2, 18, 2, low(c)).
table(low_s_maxs_c, 2, 18, 2, low(c)).
table(low_mins_maxs_c, 2, 10, 2, low(c)).
table(low_v_minc_maxc_c, 2, 11, 3, low(c)).
table(low_v_minc_s_c, 2, 18, 3, low(c)).
table(low_v_minc_mins_c, 2, 18, 3, low(c)).
table(low_v_minc_maxs_c, 2, 18, 3, low(c)).
table(low_v_maxc_s_c, 2, 18, 3, low(c)).
table(low_v_maxc_mins_c, 2, 18, 3, low(c)).
table(low_v_maxc_maxs_c, 2, 18, 3, low(c)).
table(low_v_s_mins_c, 2, 18, 3, low(c)).
table(low_v_s_maxs_c, 2, 18, 3, low(c)).
table(low_v_mins_maxs_c, 2, 18, 3, low(c)).
table(low_minc_maxc_s_c, 2, 18, 3, low(c)).
table(low_minc_maxc_mins_c, 2, 18, 3, low(c)).
table(low_minc_maxc_maxs_c, 2, 18, 3, low(c)).
table(low_minc_s_mins_c, 2, 18, 3, low(c)).
table(low_minc_s_maxs_c, 2, 18, 3, low(c)).
table(low_minc_mins_maxs_c, 2, 18, 3, low(c)).
table(low_maxc_s_mins_c, 2, 18, 3, low(c)).
table(low_maxc_s_maxs_c, 2, 18, 3, low(c)).
table(low_maxc_mins_maxs_c, 2, 18, 3, low(c)).
table(low_s_mins_maxs_c, 2, 18, 3, low(c)).
table(low_v_minc_maxc_s_c, 2, 18, 4, low(c)).
table(low_v_minc_maxc_mins_c, 2, 18, 4, low(c)).
table(low_v_minc_maxc_maxs_c, 2, 18, 4, low(c)).
table(low_v_minc_s_mins_c, 2, 18, 4, low(c)).
table(low_v_minc_s_maxs_c, 2, 18, 4, low(c)).
table(low_v_minc_mins_maxs_c, 2, 18, 4, low(c)).
table(low_v_maxc_s_mins_c, 2, 18, 4, low(c)).
table(low_v_maxc_s_maxs_c, 2, 18, 4, low(c)).
table(low_v_maxc_mins_maxs_c, 2, 18, 4, low(c)).
table(low_v_s_mins_maxs_c, 2, 18, 4, low(c)).
table(low_minc_maxc_s_mins_c, 2, 18, 4, low(c)).
table(low_minc_maxc_s_maxs_c, 2, 18, 4, low(c)).
table(low_minc_maxc_mins_maxs_c, 2, 18, 4, low(c)).
table(low_minc_s_mins_maxs_c, 2, 18, 4, low(c)).
table(low_maxc_s_mins_maxs_c, 2, 18, 4, low(c)).
table(low_v_minc_maxc_s_mins_c, 2, 18, 5, low(c)).
table(low_v_minc_maxc_s_maxs_c, 2, 18, 5, low(c)).
table(low_v_minc_maxc_mins_maxs_c, 2, 18, 5, low(c)).
table(low_v_minc_s_mins_maxs_c, 2, 18, 5, low(c)).
table(low_v_maxc_s_mins_maxs_c, 2, 18, 5, low(c)).
table(low_minc_maxc_s_mins_maxs_c, 2, 18, 5, low(c)).
table(low_v_minc_maxc_s_mins_maxs_c, 2, 18, 6, low(c)).
table(up_v_c, 2, 18, 1, up(c)).
table(up_s_c, 2, 11, 1, up(c)).
table(up_v_minc_c, 2, 11, 2, up(c)).
table(up_v_maxc_c, 2, 11, 2, up(c)).
table(up_v_s_c, 2, 18, 2, up(c)).
table(up_v_mins_c, 2, 18, 2, up(c)).
table(up_v_maxs_c, 2, 18, 2, up(c)).
table(up_minc_s_c, 2, 18, 2, up(c)).
table(up_maxc_s_c, 2, 18, 2, up(c)).
table(up_s_mins_c, 2, 18, 2, up(c)).
table(up_s_maxs_c, 2, 18, 2, up(c)).
table(up_v_minc_maxc_c, 2, 11, 3, up(c)).
table(up_v_minc_s_c, 2, 18, 3, up(c)).
table(up_v_minc_mins_c, 2, 18, 3, up(c)).
table(up_v_minc_maxs_c, 2, 18, 3, up(c)).
table(up_v_maxc_s_c, 2, 18, 3, up(c)).
table(up_v_maxc_mins_c, 2, 18, 3, up(c)).
table(up_v_maxc_maxs_c, 2, 18, 3, up(c)).
table(up_v_s_mins_c, 2, 18, 3, up(c)).
table(up_v_s_maxs_c, 2, 18, 3, up(c)).
table(up_v_mins_maxs_c, 2, 18, 3, up(c)).
table(up_minc_maxc_s_c, 2, 18, 3, up(c)).
table(up_minc_s_mins_c, 2, 18, 3, up(c)).
table(up_minc_s_maxs_c, 2, 18, 3, up(c)).
table(up_maxc_s_mins_c, 2, 18, 3, up(c)).
table(up_maxc_s_maxs_c, 2, 18, 3, up(c)).
table(up_s_mins_maxs_c, 2, 18, 3, up(c)).
table(up_v_minc_maxc_s_c, 2, 18, 4, up(c)).
table(up_v_minc_maxc_mins_c, 2, 18, 4, up(c)).
table(up_v_minc_maxc_maxs_c, 2, 18, 4, up(c)).
table(up_v_minc_s_mins_c, 2, 18, 4, up(c)).
table(up_v_minc_s_maxs_c, 2, 18, 4, up(c)).
table(up_v_minc_mins_maxs_c, 2, 18, 4, up(c)).
table(up_v_maxc_s_mins_c, 2, 18, 4, up(c)).
table(up_v_maxc_s_maxs_c, 2, 18, 4, up(c)).
table(up_v_maxc_mins_maxs_c, 2, 18, 4, up(c)).
table(up_v_s_mins_maxs_c, 2, 18, 4, up(c)).
table(up_minc_maxc_s_mins_c, 2, 18, 4, up(c)).
table(up_minc_maxc_s_maxs_c, 2, 18, 4, up(c)).
table(up_minc_s_mins_maxs_c, 2, 18, 4, up(c)).
table(up_maxc_s_mins_maxs_c, 2, 18, 4, up(c)).
table(up_v_minc_maxc_s_mins_c, 2, 18, 5, up(c)).
table(up_v_minc_maxc_s_maxs_c, 2, 18, 5, up(c)).
table(up_v_minc_maxc_mins_maxs_c, 2, 18, 5, up(c)).
table(up_v_minc_s_mins_maxs_c, 2, 18, 5, up(c)).
table(up_v_maxc_s_mins_maxs_c, 2, 18, 5, up(c)).
table(up_minc_maxc_s_mins_maxs_c, 2, 18, 5, up(c)).
table(up_v_minc_maxc_s_mins_maxs_c, 2, 18, 6, up(c)).
table(low_v_minc, 2, 18, 1, low(minc)).
table(low_c_minc, 2, 18, 1, low(minc)).
table(low_maxc_minc, 2, 7, 1, low(minc)).
table(low_s_minc, 2, 18, 1, low(minc)).
table(low_mins_minc, 2, 11, 1, low(minc)).
table(low_maxs_minc, 2, 11, 1, low(minc)).
table(low_v_c_minc, 2, 11, 2, low(minc)).
table(low_v_maxc_minc, 2, 11, 2, low(minc)).
table(low_v_s_minc, 2, 18, 2, low(minc)).
table(low_v_mins_minc, 2, 18, 2, low(minc)).
table(low_v_maxs_minc, 2, 18, 2, low(minc)).
table(low_c_maxc_minc, 2, 11, 2, low(minc)).
table(low_c_s_minc, 2, 18, 2, low(minc)).
table(low_c_mins_minc, 2, 18, 2, low(minc)).
table(low_c_maxs_minc, 2, 18, 2, low(minc)).
table(low_maxc_s_minc, 2, 18, 2, low(minc)).
table(low_maxc_mins_minc, 2, 11, 2, low(minc)).
table(low_maxc_maxs_minc, 2, 11, 2, low(minc)).
table(low_s_mins_minc, 2, 18, 2, low(minc)).
table(low_s_maxs_minc, 2, 18, 2, low(minc)).
table(low_mins_maxs_minc, 2, 11, 2, low(minc)).
table(low_v_c_maxc_minc, 2, 11, 3, low(minc)).
table(low_v_c_s_minc, 2, 18, 3, low(minc)).
table(low_v_c_mins_minc, 2, 18, 3, low(minc)).
table(low_v_c_maxs_minc, 2, 18, 3, low(minc)).
table(low_v_maxc_s_minc, 2, 18, 3, low(minc)).
table(low_v_maxc_mins_minc, 2, 18, 3, low(minc)).
table(low_v_maxc_maxs_minc, 2, 18, 3, low(minc)).
table(low_v_s_mins_minc, 2, 18, 3, low(minc)).
table(low_v_s_maxs_minc, 2, 18, 3, low(minc)).
table(low_v_mins_maxs_minc, 2, 18, 3, low(minc)).
table(low_c_maxc_s_minc, 2, 18, 3, low(minc)).
table(low_c_maxc_mins_minc, 2, 18, 3, low(minc)).
table(low_c_maxc_maxs_minc, 2, 18, 3, low(minc)).
table(low_c_s_mins_minc, 2, 18, 3, low(minc)).
table(low_c_s_maxs_minc, 2, 18, 3, low(minc)).
table(low_c_mins_maxs_minc, 2, 18, 3, low(minc)).
table(low_maxc_s_mins_minc, 2, 18, 3, low(minc)).
table(low_maxc_s_maxs_minc, 2, 18, 3, low(minc)).
table(low_maxc_mins_maxs_minc, 2, 11, 3, low(minc)).
table(low_s_mins_maxs_minc, 2, 18, 3, low(minc)).
table(low_v_c_maxc_s_minc, 2, 18, 4, low(minc)).
table(low_v_c_maxc_mins_minc, 2, 18, 4, low(minc)).
table(low_v_c_maxc_maxs_minc, 2, 18, 4, low(minc)).
table(low_v_c_s_mins_minc, 2, 18, 4, low(minc)).
table(low_v_c_s_maxs_minc, 2, 18, 4, low(minc)).
table(low_v_c_mins_maxs_minc, 2, 18, 4, low(minc)).
table(low_v_maxc_s_mins_minc, 2, 18, 4, low(minc)).
table(low_v_maxc_s_maxs_minc, 2, 18, 4, low(minc)).
table(low_v_maxc_mins_maxs_minc, 2, 18, 4, low(minc)).
table(low_v_s_mins_maxs_minc, 2, 18, 4, low(minc)).
table(low_c_maxc_s_mins_minc, 2, 18, 4, low(minc)).
table(low_c_maxc_s_maxs_minc, 2, 18, 4, low(minc)).
table(low_c_maxc_mins_maxs_minc, 2, 18, 4, low(minc)).
table(low_c_s_mins_maxs_minc, 2, 18, 4, low(minc)).
table(low_maxc_s_mins_maxs_minc, 2, 18, 4, low(minc)).
table(low_v_c_maxc_s_mins_minc, 2, 18, 5, low(minc)).
table(low_v_c_maxc_s_maxs_minc, 2, 18, 5, low(minc)).
table(low_v_c_maxc_mins_maxs_minc, 2, 18, 5, low(minc)).
table(low_v_c_s_mins_maxs_minc, 2, 18, 5, low(minc)).
table(low_v_maxc_s_mins_maxs_minc, 2, 18, 5, low(minc)).
table(low_c_maxc_s_mins_maxs_minc, 2, 18, 5, low(minc)).
table(low_v_c_maxc_s_mins_maxs_minc, 2, 18, 6, low(minc)).
table(up_v_minc, 2, 11, 1, up(minc)).
table(up_maxc_minc, 2, 7, 1, up(minc)).
table(up_v_c_minc, 2, 11, 2, up(minc)).
table(up_v_maxc_minc, 2, 11, 2, up(minc)).
table(up_v_s_minc, 2, 18, 2, up(minc)).
table(up_v_mins_minc, 2, 18, 2, up(minc)).
table(up_v_maxs_minc, 2, 18, 2, up(minc)).
table(up_c_maxc_minc, 2, 11, 2, up(minc)).
table(up_maxc_s_minc, 2, 18, 2, up(minc)).
table(up_maxc_mins_minc, 2, 11, 2, up(minc)).
table(up_maxc_maxs_minc, 2, 11, 2, up(minc)).
table(up_s_maxs_minc, 2, 18, 2, up(minc)).
table(up_v_c_maxc_minc, 2, 11, 3, up(minc)).
table(up_v_c_s_minc, 2, 18, 3, up(minc)).
table(up_v_c_mins_minc, 2, 18, 3, up(minc)).
table(up_v_c_maxs_minc, 2, 18, 3, up(minc)).
table(up_v_maxc_s_minc, 2, 18, 3, up(minc)).
table(up_v_maxc_mins_minc, 2, 18, 3, up(minc)).
table(up_v_maxc_maxs_minc, 2, 18, 3, up(minc)).
table(up_v_s_mins_minc, 2, 18, 3, up(minc)).
table(up_v_s_maxs_minc, 2, 18, 3, up(minc)).
table(up_v_mins_maxs_minc, 2, 18, 3, up(minc)).
table(up_c_maxc_s_minc, 2, 18, 3, up(minc)).
table(up_c_maxc_mins_minc, 2, 18, 3, up(minc)).
table(up_c_maxc_maxs_minc, 2, 18, 3, up(minc)).
table(up_c_s_maxs_minc, 2, 18, 3, up(minc)).
table(up_maxc_s_mins_minc, 2, 18, 3, up(minc)).
table(up_maxc_s_maxs_minc, 2, 18, 3, up(minc)).
table(up_maxc_mins_maxs_minc, 2, 11, 3, up(minc)).
table(up_s_mins_maxs_minc, 2, 18, 3, up(minc)).
table(up_v_c_maxc_s_minc, 2, 18, 4, up(minc)).
table(up_v_c_maxc_mins_minc, 2, 18, 4, up(minc)).
table(up_v_c_maxc_maxs_minc, 2, 18, 4, up(minc)).
table(up_v_c_s_mins_minc, 2, 18, 4, up(minc)).
table(up_v_c_s_maxs_minc, 2, 18, 4, up(minc)).
table(up_v_c_mins_maxs_minc, 2, 18, 4, up(minc)).
table(up_v_maxc_s_mins_minc, 2, 18, 4, up(minc)).
table(up_v_maxc_s_maxs_minc, 2, 18, 4, up(minc)).
table(up_v_maxc_mins_maxs_minc, 2, 18, 4, up(minc)).
table(up_v_s_mins_maxs_minc, 2, 18, 4, up(minc)).
table(up_c_maxc_s_mins_minc, 2, 18, 4, up(minc)).
table(up_c_maxc_s_maxs_minc, 2, 18, 4, up(minc)).
table(up_c_maxc_mins_maxs_minc, 2, 18, 4, up(minc)).
table(up_c_s_mins_maxs_minc, 2, 18, 4, up(minc)).
table(up_maxc_s_mins_maxs_minc, 2, 18, 4, up(minc)).
table(up_v_c_maxc_s_mins_minc, 2, 18, 5, up(minc)).
table(up_v_c_maxc_s_maxs_minc, 2, 18, 5, up(minc)).
table(up_v_c_maxc_mins_maxs_minc, 2, 18, 5, up(minc)).
table(up_v_c_s_mins_maxs_minc, 2, 18, 5, up(minc)).
table(up_v_maxc_s_mins_maxs_minc, 2, 18, 5, up(minc)).
table(up_c_maxc_s_mins_maxs_minc, 2, 18, 5, up(minc)).
table(up_v_c_maxc_s_mins_maxs_minc, 2, 18, 6, up(minc)).
table(low_v_maxc, 2, 18, 1, low(maxc)).
table(low_c_maxc, 2, 18, 1, low(maxc)).
table(low_minc_maxc, 2, 7, 1, low(maxc)).
table(low_s_maxc, 2, 18, 1, low(maxc)).
table(low_mins_maxc, 2, 11, 1, low(maxc)).
table(low_maxs_maxc, 2, 11, 1, low(maxc)).
table(low_v_c_maxc, 2, 11, 2, low(maxc)).
table(low_v_minc_maxc, 2, 11, 2, low(maxc)).
table(low_v_s_maxc, 2, 18, 2, low(maxc)).
table(low_v_mins_maxc, 2, 18, 2, low(maxc)).
table(low_v_maxs_maxc, 2, 18, 2, low(maxc)).
table(low_c_minc_maxc, 2, 11, 2, low(maxc)).
table(low_c_s_maxc, 2, 18, 2, low(maxc)).
table(low_c_mins_maxc, 2, 18, 2, low(maxc)).
table(low_c_maxs_maxc, 2, 18, 2, low(maxc)).
table(low_minc_s_maxc, 2, 18, 2, low(maxc)).
table(low_minc_mins_maxc, 2, 11, 2, low(maxc)).
table(low_minc_maxs_maxc, 2, 11, 2, low(maxc)).
table(low_s_mins_maxc, 2, 18, 2, low(maxc)).
table(low_s_maxs_maxc, 2, 18, 2, low(maxc)).
table(low_mins_maxs_maxc, 2, 11, 2, low(maxc)).
table(low_v_c_minc_maxc, 2, 11, 3, low(maxc)).
table(low_v_c_s_maxc, 2, 18, 3, low(maxc)).
table(low_v_c_mins_maxc, 2, 18, 3, low(maxc)).
table(low_v_c_maxs_maxc, 2, 18, 3, low(maxc)).
table(low_v_minc_s_maxc, 2, 18, 3, low(maxc)).
table(low_v_minc_mins_maxc, 2, 18, 3, low(maxc)).
table(low_v_minc_maxs_maxc, 2, 18, 3, low(maxc)).
table(low_v_s_mins_maxc, 2, 18, 3, low(maxc)).
table(low_v_s_maxs_maxc, 2, 18, 3, low(maxc)).
table(low_v_mins_maxs_maxc, 2, 18, 3, low(maxc)).
table(low_c_minc_s_maxc, 2, 18, 3, low(maxc)).
table(low_c_minc_mins_maxc, 2, 18, 3, low(maxc)).
table(low_c_minc_maxs_maxc, 2, 18, 3, low(maxc)).
table(low_c_s_mins_maxc, 2, 18, 3, low(maxc)).
table(low_c_s_maxs_maxc, 2, 18, 3, low(maxc)).
table(low_c_mins_maxs_maxc, 2, 18, 3, low(maxc)).
table(low_minc_s_mins_maxc, 2, 18, 3, low(maxc)).
table(low_minc_s_maxs_maxc, 2, 18, 3, low(maxc)).
table(low_minc_mins_maxs_maxc, 2, 11, 3, low(maxc)).
table(low_s_mins_maxs_maxc, 2, 18, 3, low(maxc)).
table(low_v_c_minc_s_maxc, 2, 18, 4, low(maxc)).
table(low_v_c_minc_mins_maxc, 2, 18, 4, low(maxc)).
table(low_v_c_minc_maxs_maxc, 2, 18, 4, low(maxc)).
table(low_v_c_s_mins_maxc, 2, 18, 4, low(maxc)).
table(low_v_c_s_maxs_maxc, 2, 18, 4, low(maxc)).
table(low_v_c_mins_maxs_maxc, 2, 18, 4, low(maxc)).
table(low_v_minc_s_mins_maxc, 2, 18, 4, low(maxc)).
table(low_v_minc_s_maxs_maxc, 2, 18, 4, low(maxc)).
table(low_v_minc_mins_maxs_maxc, 2, 18, 4, low(maxc)).
table(low_v_s_mins_maxs_maxc, 2, 18, 4, low(maxc)).
table(low_c_minc_s_mins_maxc, 2, 18, 4, low(maxc)).
table(low_c_minc_s_maxs_maxc, 2, 18, 4, low(maxc)).
table(low_c_minc_mins_maxs_maxc, 2, 18, 4, low(maxc)).
table(low_c_s_mins_maxs_maxc, 2, 18, 4, low(maxc)).
table(low_minc_s_mins_maxs_maxc, 2, 18, 4, low(maxc)).
table(low_v_c_minc_s_mins_maxc, 2, 18, 5, low(maxc)).
table(low_v_c_minc_s_maxs_maxc, 2, 18, 5, low(maxc)).
table(low_v_c_minc_mins_maxs_maxc, 2, 18, 5, low(maxc)).
table(low_v_c_s_mins_maxs_maxc, 2, 18, 5, low(maxc)).
table(low_v_minc_s_mins_maxs_maxc, 2, 18, 5, low(maxc)).
table(low_c_minc_s_mins_maxs_maxc, 2, 18, 5, low(maxc)).
table(low_v_c_minc_s_mins_maxs_maxc, 2, 18, 6, low(maxc)).
table(up_v_maxc, 2, 11, 1, up(maxc)).
table(up_v_c_maxc, 2, 11, 2, up(maxc)).
table(up_v_minc_maxc, 2, 11, 2, up(maxc)).
table(up_v_s_maxc, 2, 18, 2, up(maxc)).
table(up_v_mins_maxc, 2, 18, 2, up(maxc)).
table(up_v_maxs_maxc, 2, 18, 2, up(maxc)).
table(up_v_c_minc_maxc, 2, 11, 3, up(maxc)).
table(up_v_c_s_maxc, 2, 18, 3, up(maxc)).
table(up_v_c_mins_maxc, 2, 18, 3, up(maxc)).
table(up_v_c_maxs_maxc, 2, 18, 3, up(maxc)).
table(up_v_minc_s_maxc, 2, 18, 3, up(maxc)).
table(up_v_minc_mins_maxc, 2, 18, 3, up(maxc)).
table(up_v_minc_maxs_maxc, 2, 18, 3, up(maxc)).
table(up_v_s_mins_maxc, 2, 18, 3, up(maxc)).
table(up_v_s_maxs_maxc, 2, 18, 3, up(maxc)).
table(up_v_mins_maxs_maxc, 2, 18, 3, up(maxc)).
table(up_v_c_minc_s_maxc, 2, 18, 4, up(maxc)).
table(up_v_c_minc_mins_maxc, 2, 18, 4, up(maxc)).
table(up_v_c_minc_maxs_maxc, 2, 18, 4, up(maxc)).
table(up_v_c_s_mins_maxc, 2, 18, 4, up(maxc)).
table(up_v_c_s_maxs_maxc, 2, 18, 4, up(maxc)).
table(up_v_c_mins_maxs_maxc, 2, 18, 4, up(maxc)).
table(up_v_minc_s_mins_maxc, 2, 18, 4, up(maxc)).
table(up_v_minc_s_maxs_maxc, 2, 18, 4, up(maxc)).
table(up_v_minc_mins_maxs_maxc, 2, 18, 4, up(maxc)).
table(up_v_s_mins_maxs_maxc, 2, 18, 4, up(maxc)).
table(up_v_c_minc_s_mins_maxc, 2, 18, 5, up(maxc)).
table(up_v_c_minc_s_maxs_maxc, 2, 18, 5, up(maxc)).
table(up_v_c_minc_mins_maxs_maxc, 2, 18, 5, up(maxc)).
table(up_v_c_s_mins_maxs_maxc, 2, 18, 5, up(maxc)).
table(up_v_minc_s_mins_maxs_maxc, 2, 18, 5, up(maxc)).
table(up_v_c_minc_s_mins_maxs_maxc, 2, 18, 6, up(maxc)).
table(low_v_s, 2, 18, 1, low(s)).
table(low_c_s, 2, 11, 1, low(s)).
table(low_minc_s, 2, 18, 1, low(s)).
table(low_maxc_s, 2, 18, 1, low(s)).
table(low_mins_s, 2, 18, 1, low(s)).
table(low_maxs_s, 2, 18, 1, low(s)).
table(low_v_c_s, 2, 18, 2, low(s)).
table(low_v_minc_s, 2, 18, 2, low(s)).
table(low_v_maxc_s, 2, 18, 2, low(s)).
table(low_v_mins_s, 2, 11, 2, low(s)).
table(low_v_maxs_s, 2, 11, 2, low(s)).
table(low_c_minc_s, 2, 18, 2, low(s)).
table(low_c_maxc_s, 2, 18, 2, low(s)).
table(low_c_mins_s, 2, 18, 2, low(s)).
table(low_c_maxs_s, 2, 18, 2, low(s)).
table(low_minc_maxc_s, 2, 18, 2, low(s)).
table(low_minc_mins_s, 2, 18, 2, low(s)).
table(low_minc_maxs_s, 2, 18, 2, low(s)).
table(low_maxc_mins_s, 2, 18, 2, low(s)).
table(low_maxc_maxs_s, 2, 18, 2, low(s)).
table(low_mins_maxs_s, 2, 18, 2, low(s)).
table(low_v_c_minc_s, 2, 18, 3, low(s)).
table(low_v_c_maxc_s, 2, 18, 3, low(s)).
table(low_v_c_mins_s, 2, 18, 3, low(s)).
table(low_v_c_maxs_s, 2, 18, 3, low(s)).
table(low_v_minc_maxc_s, 2, 18, 3, low(s)).
table(low_v_minc_mins_s, 2, 18, 3, low(s)).
table(low_v_minc_maxs_s, 2, 18, 3, low(s)).
table(low_v_maxc_mins_s, 2, 18, 3, low(s)).
table(low_v_maxc_maxs_s, 2, 18, 3, low(s)).
table(low_v_mins_maxs_s, 2, 11, 3, low(s)).
table(low_c_minc_maxc_s, 2, 18, 3, low(s)).
table(low_c_minc_mins_s, 2, 18, 3, low(s)).
table(low_c_minc_maxs_s, 2, 18, 3, low(s)).
table(low_c_maxc_mins_s, 2, 18, 3, low(s)).
table(low_c_maxc_maxs_s, 2, 18, 3, low(s)).
table(low_c_mins_maxs_s, 2, 18, 3, low(s)).
table(low_minc_maxc_mins_s, 2, 18, 3, low(s)).
table(low_minc_maxc_maxs_s, 2, 18, 3, low(s)).
table(low_minc_mins_maxs_s, 2, 18, 3, low(s)).
table(low_maxc_mins_maxs_s, 2, 18, 3, low(s)).
table(low_v_c_minc_maxc_s, 2, 18, 4, low(s)).
table(low_v_c_minc_mins_s, 2, 18, 4, low(s)).
table(low_v_c_minc_maxs_s, 2, 18, 4, low(s)).
table(low_v_c_maxc_mins_s, 2, 18, 4, low(s)).
table(low_v_c_maxc_maxs_s, 2, 18, 4, low(s)).
table(low_v_c_mins_maxs_s, 2, 18, 4, low(s)).
table(low_v_minc_maxc_mins_s, 2, 18, 4, low(s)).
table(low_v_minc_maxc_maxs_s, 2, 18, 4, low(s)).
table(low_v_minc_mins_maxs_s, 2, 18, 4, low(s)).
table(low_v_maxc_mins_maxs_s, 2, 18, 4, low(s)).
table(low_c_minc_maxc_mins_s, 2, 18, 4, low(s)).
table(low_c_minc_maxc_maxs_s, 2, 18, 4, low(s)).
table(low_c_minc_mins_maxs_s, 2, 18, 4, low(s)).
table(low_c_maxc_mins_maxs_s, 2, 18, 4, low(s)).
table(low_minc_maxc_mins_maxs_s, 2, 18, 4, low(s)).
table(low_v_c_minc_maxc_mins_s, 2, 18, 5, low(s)).
table(low_v_c_minc_maxc_maxs_s, 2, 18, 5, low(s)).
table(low_v_c_minc_mins_maxs_s, 2, 18, 5, low(s)).
table(low_v_c_maxc_mins_maxs_s, 2, 18, 5, low(s)).
table(low_v_minc_maxc_mins_maxs_s, 2, 18, 5, low(s)).
table(low_c_minc_maxc_mins_maxs_s, 2, 18, 5, low(s)).
table(low_v_c_minc_maxc_mins_maxs_s, 2, 18, 6, low(s)).
table(up_v_s, 2, 11, 1, up(s)).
table(up_v_c_s, 2, 18, 2, up(s)).
table(up_v_minc_s, 2, 18, 2, up(s)).
table(up_v_maxc_s, 2, 18, 2, up(s)).
table(up_v_mins_s, 2, 11, 2, up(s)).
table(up_v_maxs_s, 2, 11, 2, up(s)).
table(up_c_maxc_s, 2, 18, 2, up(s)).
table(up_v_c_minc_s, 2, 18, 3, up(s)).
table(up_v_c_maxc_s, 2, 18, 3, up(s)).
table(up_v_c_mins_s, 2, 18, 3, up(s)).
table(up_v_c_maxs_s, 2, 18, 3, up(s)).
table(up_v_minc_maxc_s, 2, 18, 3, up(s)).
table(up_v_minc_mins_s, 2, 18, 3, up(s)).
table(up_v_minc_maxs_s, 2, 18, 3, up(s)).
table(up_v_maxc_mins_s, 2, 18, 3, up(s)).
table(up_v_maxc_maxs_s, 2, 18, 3, up(s)).
table(up_v_mins_maxs_s, 2, 11, 3, up(s)).
table(up_c_minc_maxc_s, 2, 18, 3, up(s)).
table(up_c_maxc_mins_s, 2, 18, 3, up(s)).
table(up_c_maxc_maxs_s, 2, 18, 3, up(s)).
table(up_v_c_minc_maxc_s, 2, 18, 4, up(s)).
table(up_v_c_minc_mins_s, 2, 18, 4, up(s)).
table(up_v_c_minc_maxs_s, 2, 18, 4, up(s)).
table(up_v_c_maxc_mins_s, 2, 18, 4, up(s)).
table(up_v_c_maxc_maxs_s, 2, 18, 4, up(s)).
table(up_v_c_mins_maxs_s, 2, 18, 4, up(s)).
table(up_v_minc_maxc_mins_s, 2, 18, 4, up(s)).
table(up_v_minc_maxc_maxs_s, 2, 18, 4, up(s)).
table(up_v_minc_mins_maxs_s, 2, 18, 4, up(s)).
table(up_v_maxc_mins_maxs_s, 2, 18, 4, up(s)).
table(up_c_minc_maxc_mins_s, 2, 18, 4, up(s)).
table(up_c_minc_maxc_maxs_s, 2, 18, 4, up(s)).
table(up_c_maxc_mins_maxs_s, 2, 18, 4, up(s)).
table(up_v_c_minc_maxc_mins_s, 2, 18, 5, up(s)).
table(up_v_c_minc_maxc_maxs_s, 2, 18, 5, up(s)).
table(up_v_c_minc_mins_maxs_s, 2, 18, 5, up(s)).
table(up_v_c_maxc_mins_maxs_s, 2, 18, 5, up(s)).
table(up_v_minc_maxc_mins_maxs_s, 2, 18, 5, up(s)).
table(up_c_minc_maxc_mins_maxs_s, 2, 18, 5, up(s)).
table(up_v_c_minc_maxc_mins_maxs_s, 2, 18, 6, up(s)).
table(low_v_mins, 2, 11, 1, low(mins)).
table(low_c_mins, 2, 10, 1, low(mins)).
table(low_minc_mins, 2, 11, 1, low(mins)).
table(low_maxc_mins, 2, 11, 1, low(mins)).
table(low_s_mins, 2, 11, 1, low(mins)).
table(low_maxs_mins, 2, 7, 1, low(mins)).
table(low_v_c_mins, 2, 18, 2, low(mins)).
table(low_v_minc_mins, 2, 18, 2, low(mins)).
table(low_v_maxc_mins, 2, 18, 2, low(mins)).
table(low_v_s_mins, 2, 11, 2, low(mins)).
table(low_v_maxs_mins, 2, 11, 2, low(mins)).
table(low_c_minc_mins, 2, 18, 2, low(mins)).
table(low_c_maxc_mins, 2, 18, 2, low(mins)).
table(low_c_s_mins, 2, 18, 2, low(mins)).
table(low_c_maxs_mins, 2, 10, 2, low(mins)).
table(low_minc_maxc_mins, 2, 11, 2, low(mins)).
table(low_minc_s_mins, 2, 18, 2, low(mins)).
table(low_minc_maxs_mins, 2, 11, 2, low(mins)).
table(low_maxc_s_mins, 2, 18, 2, low(mins)).
table(low_maxc_maxs_mins, 2, 11, 2, low(mins)).
table(low_s_maxs_mins, 2, 11, 2, low(mins)).
table(low_v_c_minc_mins, 2, 18, 3, low(mins)).
table(low_v_c_maxc_mins, 2, 18, 3, low(mins)).
table(low_v_c_s_mins, 2, 18, 3, low(mins)).
table(low_v_c_maxs_mins, 2, 18, 3, low(mins)).
table(low_v_minc_maxc_mins, 2, 18, 3, low(mins)).
table(low_v_minc_s_mins, 2, 18, 3, low(mins)).
table(low_v_minc_maxs_mins, 2, 18, 3, low(mins)).
table(low_v_maxc_s_mins, 2, 18, 3, low(mins)).
table(low_v_maxc_maxs_mins, 2, 18, 3, low(mins)).
table(low_v_s_maxs_mins, 2, 11, 3, low(mins)).
table(low_c_minc_maxc_mins, 2, 18, 3, low(mins)).
table(low_c_minc_s_mins, 2, 18, 3, low(mins)).
table(low_c_minc_maxs_mins, 2, 18, 3, low(mins)).
table(low_c_maxc_s_mins, 2, 18, 3, low(mins)).
table(low_c_maxc_maxs_mins, 2, 18, 3, low(mins)).
table(low_c_s_maxs_mins, 2, 18, 3, low(mins)).
table(low_minc_maxc_s_mins, 2, 18, 3, low(mins)).
table(low_minc_maxc_maxs_mins, 2, 11, 3, low(mins)).
table(low_minc_s_maxs_mins, 2, 18, 3, low(mins)).
table(low_maxc_s_maxs_mins, 2, 18, 3, low(mins)).
table(low_v_c_minc_maxc_mins, 2, 18, 4, low(mins)).
table(low_v_c_minc_s_mins, 2, 18, 4, low(mins)).
table(low_v_c_minc_maxs_mins, 2, 18, 4, low(mins)).
table(low_v_c_maxc_s_mins, 2, 18, 4, low(mins)).
table(low_v_c_maxc_maxs_mins, 2, 18, 4, low(mins)).
table(low_v_c_s_maxs_mins, 2, 18, 4, low(mins)).
table(low_v_minc_maxc_s_mins, 2, 18, 4, low(mins)).
table(low_v_minc_maxc_maxs_mins, 2, 18, 4, low(mins)).
table(low_v_minc_s_maxs_mins, 2, 18, 4, low(mins)).
table(low_v_maxc_s_maxs_mins, 2, 18, 4, low(mins)).
table(low_c_minc_maxc_s_mins, 2, 18, 4, low(mins)).
table(low_c_minc_maxc_maxs_mins, 2, 18, 4, low(mins)).
table(low_c_minc_s_maxs_mins, 2, 18, 4, low(mins)).
table(low_c_maxc_s_maxs_mins, 2, 18, 4, low(mins)).
table(low_minc_maxc_s_maxs_mins, 2, 18, 4, low(mins)).
table(low_v_c_minc_maxc_s_mins, 2, 18, 5, low(mins)).
table(low_v_c_minc_maxc_maxs_mins, 2, 18, 5, low(mins)).
table(low_v_c_minc_s_maxs_mins, 2, 18, 5, low(mins)).
table(low_v_c_maxc_s_maxs_mins, 2, 18, 5, low(mins)).
table(low_v_minc_maxc_s_maxs_mins, 2, 18, 5, low(mins)).
table(low_c_minc_maxc_s_maxs_mins, 2, 18, 5, low(mins)).
table(low_v_c_minc_maxc_s_maxs_mins, 2, 18, 6, low(mins)).
table(up_v_mins, 2, 18, 1, up(mins)).
table(up_maxc_mins, 2, 11, 1, up(mins)).
table(up_maxs_mins, 2, 7, 1, up(mins)).
table(up_v_c_mins, 2, 18, 2, up(mins)).
table(up_v_minc_mins, 2, 18, 2, up(mins)).
table(up_v_maxc_mins, 2, 18, 2, up(mins)).
table(up_v_s_mins, 2, 11, 2, up(mins)).
table(up_v_maxs_mins, 2, 11, 2, up(mins)).
table(up_c_maxc_mins, 2, 18, 2, up(mins)).
table(up_c_maxs_mins, 2, 10, 2, up(mins)).
table(up_minc_maxc_mins, 2, 11, 2, up(mins)).
table(up_minc_maxs_mins, 2, 11, 2, up(mins)).
table(up_maxc_s_mins, 2, 18, 2, up(mins)).
table(up_maxc_maxs_mins, 2, 11, 2, up(mins)).
table(up_s_maxs_mins, 2, 11, 2, up(mins)).
table(up_v_c_minc_mins, 2, 18, 3, up(mins)).
table(up_v_c_maxc_mins, 2, 18, 3, up(mins)).
table(up_v_c_s_mins, 2, 18, 3, up(mins)).
table(up_v_c_maxs_mins, 2, 18, 3, up(mins)).
table(up_v_minc_maxc_mins, 2, 18, 3, up(mins)).
table(up_v_minc_s_mins, 2, 18, 3, up(mins)).
table(up_v_minc_maxs_mins, 2, 18, 3, up(mins)).
table(up_v_maxc_s_mins, 2, 18, 3, up(mins)).
table(up_v_maxc_maxs_mins, 2, 18, 3, up(mins)).
table(up_v_s_maxs_mins, 2, 11, 3, up(mins)).
table(up_c_minc_maxc_mins, 2, 18, 3, up(mins)).
table(up_c_minc_maxs_mins, 2, 18, 3, up(mins)).
table(up_c_maxc_s_mins, 2, 18, 3, up(mins)).
table(up_c_maxc_maxs_mins, 2, 18, 3, up(mins)).
table(up_c_s_maxs_mins, 2, 18, 3, up(mins)).
table(up_minc_maxc_s_mins, 2, 18, 3, up(mins)).
table(up_minc_maxc_maxs_mins, 2, 11, 3, up(mins)).
table(up_minc_s_maxs_mins, 2, 18, 3, up(mins)).
table(up_maxc_s_maxs_mins, 2, 18, 3, up(mins)).
table(up_v_c_minc_maxc_mins, 2, 18, 4, up(mins)).
table(up_v_c_minc_s_mins, 2, 18, 4, up(mins)).
table(up_v_c_minc_maxs_mins, 2, 18, 4, up(mins)).
table(up_v_c_maxc_s_mins, 2, 18, 4, up(mins)).
table(up_v_c_maxc_maxs_mins, 2, 18, 4, up(mins)).
table(up_v_c_s_maxs_mins, 2, 18, 4, up(mins)).
table(up_v_minc_maxc_s_mins, 2, 18, 4, up(mins)).
table(up_v_minc_maxc_maxs_mins, 2, 18, 4, up(mins)).
table(up_v_minc_s_maxs_mins, 2, 18, 4, up(mins)).
table(up_v_maxc_s_maxs_mins, 2, 18, 4, up(mins)).
table(up_c_minc_maxc_s_mins, 2, 18, 4, up(mins)).
table(up_c_minc_maxc_maxs_mins, 2, 18, 4, up(mins)).
table(up_c_minc_s_maxs_mins, 2, 18, 4, up(mins)).
table(up_c_maxc_s_maxs_mins, 2, 18, 4, up(mins)).
table(up_minc_maxc_s_maxs_mins, 2, 18, 4, up(mins)).
table(up_v_c_minc_maxc_s_mins, 2, 18, 5, up(mins)).
table(up_v_c_minc_maxc_maxs_mins, 2, 18, 5, up(mins)).
table(up_v_c_minc_s_maxs_mins, 2, 18, 5, up(mins)).
table(up_v_c_maxc_s_maxs_mins, 2, 18, 5, up(mins)).
table(up_v_minc_maxc_s_maxs_mins, 2, 18, 5, up(mins)).
table(up_c_minc_maxc_s_maxs_mins, 2, 18, 5, up(mins)).
table(up_v_c_minc_maxc_s_maxs_mins, 2, 18, 6, up(mins)).
table(low_v_maxs, 2, 11, 1, low(maxs)).
table(low_c_maxs, 2, 10, 1, low(maxs)).
table(low_minc_maxs, 2, 11, 1, low(maxs)).
table(low_maxc_maxs, 2, 11, 1, low(maxs)).
table(low_s_maxs, 2, 11, 1, low(maxs)).
table(low_mins_maxs, 2, 7, 1, low(maxs)).
table(low_v_c_maxs, 2, 18, 2, low(maxs)).
table(low_v_minc_maxs, 2, 18, 2, low(maxs)).
table(low_v_maxc_maxs, 2, 18, 2, low(maxs)).
table(low_v_s_maxs, 2, 11, 2, low(maxs)).
table(low_v_mins_maxs, 2, 11, 2, low(maxs)).
table(low_c_minc_maxs, 2, 18, 2, low(maxs)).
table(low_c_maxc_maxs, 2, 18, 2, low(maxs)).
table(low_c_s_maxs, 2, 18, 2, low(maxs)).
table(low_c_mins_maxs, 2, 10, 2, low(maxs)).
table(low_minc_maxc_maxs, 2, 11, 2, low(maxs)).
table(low_minc_s_maxs, 2, 18, 2, low(maxs)).
table(low_minc_mins_maxs, 2, 11, 2, low(maxs)).
table(low_maxc_s_maxs, 2, 18, 2, low(maxs)).
table(low_maxc_mins_maxs, 2, 11, 2, low(maxs)).
table(low_s_mins_maxs, 2, 11, 2, low(maxs)).
table(low_v_c_minc_maxs, 2, 18, 3, low(maxs)).
table(low_v_c_maxc_maxs, 2, 18, 3, low(maxs)).
table(low_v_c_s_maxs, 2, 18, 3, low(maxs)).
table(low_v_c_mins_maxs, 2, 18, 3, low(maxs)).
table(low_v_minc_maxc_maxs, 2, 18, 3, low(maxs)).
table(low_v_minc_s_maxs, 2, 18, 3, low(maxs)).
table(low_v_minc_mins_maxs, 2, 18, 3, low(maxs)).
table(low_v_maxc_s_maxs, 2, 18, 3, low(maxs)).
table(low_v_maxc_mins_maxs, 2, 18, 3, low(maxs)).
table(low_v_s_mins_maxs, 2, 11, 3, low(maxs)).
table(low_c_minc_maxc_maxs, 2, 18, 3, low(maxs)).
table(low_c_minc_s_maxs, 2, 18, 3, low(maxs)).
table(low_c_minc_mins_maxs, 2, 18, 3, low(maxs)).
table(low_c_maxc_s_maxs, 2, 18, 3, low(maxs)).
table(low_c_maxc_mins_maxs, 2, 18, 3, low(maxs)).
table(low_c_s_mins_maxs, 2, 18, 3, low(maxs)).
table(low_minc_maxc_s_maxs, 2, 18, 3, low(maxs)).
table(low_minc_maxc_mins_maxs, 2, 11, 3, low(maxs)).
table(low_minc_s_mins_maxs, 2, 18, 3, low(maxs)).
table(low_maxc_s_mins_maxs, 2, 18, 3, low(maxs)).
table(low_v_c_minc_maxc_maxs, 2, 18, 4, low(maxs)).
table(low_v_c_minc_s_maxs, 2, 18, 4, low(maxs)).
table(low_v_c_minc_mins_maxs, 2, 18, 4, low(maxs)).
table(low_v_c_maxc_s_maxs, 2, 18, 4, low(maxs)).
table(low_v_c_maxc_mins_maxs, 2, 18, 4, low(maxs)).
table(low_v_c_s_mins_maxs, 2, 18, 4, low(maxs)).
table(low_v_minc_maxc_s_maxs, 2, 18, 4, low(maxs)).
table(low_v_minc_maxc_mins_maxs, 2, 18, 4, low(maxs)).
table(low_v_minc_s_mins_maxs, 2, 18, 4, low(maxs)).
table(low_v_maxc_s_mins_maxs, 2, 18, 4, low(maxs)).
table(low_c_minc_maxc_s_maxs, 2, 18, 4, low(maxs)).
table(low_c_minc_maxc_mins_maxs, 2, 18, 4, low(maxs)).
table(low_c_minc_s_mins_maxs, 2, 18, 4, low(maxs)).
table(low_c_maxc_s_mins_maxs, 2, 18, 4, low(maxs)).
table(low_minc_maxc_s_mins_maxs, 2, 18, 4, low(maxs)).
table(low_v_c_minc_maxc_s_maxs, 2, 18, 5, low(maxs)).
table(low_v_c_minc_maxc_mins_maxs, 2, 18, 5, low(maxs)).
table(low_v_c_minc_s_mins_maxs, 2, 18, 5, low(maxs)).
table(low_v_c_maxc_s_mins_maxs, 2, 18, 5, low(maxs)).
table(low_v_minc_maxc_s_mins_maxs, 2, 18, 5, low(maxs)).
table(low_c_minc_maxc_s_mins_maxs, 2, 18, 5, low(maxs)).
table(low_v_c_minc_maxc_s_mins_maxs, 2, 18, 6, low(maxs)).
table(up_v_maxs, 2, 18, 1, up(maxs)).
table(up_maxc_maxs, 2, 11, 1, up(maxs)).
table(up_v_c_maxs, 2, 18, 2, up(maxs)).
table(up_v_minc_maxs, 2, 18, 2, up(maxs)).
table(up_v_maxc_maxs, 2, 18, 2, up(maxs)).
table(up_v_s_maxs, 2, 11, 2, up(maxs)).
table(up_v_mins_maxs, 2, 11, 2, up(maxs)).
table(up_c_maxc_maxs, 2, 18, 2, up(maxs)).
table(up_minc_maxc_maxs, 2, 11, 2, up(maxs)).
table(up_maxc_s_maxs, 2, 18, 2, up(maxs)).
table(up_maxc_mins_maxs, 2, 11, 2, up(maxs)).
table(up_v_c_minc_maxs, 2, 18, 3, up(maxs)).
table(up_v_c_maxc_maxs, 2, 18, 3, up(maxs)).
table(up_v_c_s_maxs, 2, 18, 3, up(maxs)).
table(up_v_c_mins_maxs, 2, 18, 3, up(maxs)).
table(up_v_minc_maxc_maxs, 2, 18, 3, up(maxs)).
table(up_v_minc_s_maxs, 2, 18, 3, up(maxs)).
table(up_v_minc_mins_maxs, 2, 18, 3, up(maxs)).
table(up_v_maxc_s_maxs, 2, 18, 3, up(maxs)).
table(up_v_maxc_mins_maxs, 2, 18, 3, up(maxs)).
table(up_v_s_mins_maxs, 2, 11, 3, up(maxs)).
table(up_c_minc_maxc_maxs, 2, 18, 3, up(maxs)).
table(up_c_maxc_s_maxs, 2, 18, 3, up(maxs)).
table(up_c_maxc_mins_maxs, 2, 18, 3, up(maxs)).
table(up_minc_maxc_s_maxs, 2, 18, 3, up(maxs)).
table(up_minc_maxc_mins_maxs, 2, 11, 3, up(maxs)).
table(up_maxc_s_mins_maxs, 2, 18, 3, up(maxs)).
table(up_v_c_minc_maxc_maxs, 2, 18, 4, up(maxs)).
table(up_v_c_minc_s_maxs, 2, 18, 4, up(maxs)).
table(up_v_c_minc_mins_maxs, 2, 18, 4, up(maxs)).
table(up_v_c_maxc_s_maxs, 2, 18, 4, up(maxs)).
table(up_v_c_maxc_mins_maxs, 2, 18, 4, up(maxs)).
table(up_v_c_s_mins_maxs, 2, 18, 4, up(maxs)).
table(up_v_minc_maxc_s_maxs, 2, 18, 4, up(maxs)).
table(up_v_minc_maxc_mins_maxs, 2, 18, 4, up(maxs)).
table(up_v_minc_s_mins_maxs, 2, 18, 4, up(maxs)).
table(up_v_maxc_s_mins_maxs, 2, 18, 4, up(maxs)).
table(up_c_minc_maxc_s_maxs, 2, 18, 4, up(maxs)).
table(up_c_minc_maxc_mins_maxs, 2, 18, 4, up(maxs)).
table(up_c_maxc_s_mins_maxs, 2, 18, 4, up(maxs)).
table(up_minc_maxc_s_mins_maxs, 2, 18, 4, up(maxs)).
table(up_v_c_minc_maxc_s_maxs, 2, 18, 5, up(maxs)).
table(up_v_c_minc_maxc_mins_maxs, 2, 18, 5, up(maxs)).
table(up_v_c_minc_s_mins_maxs, 2, 18, 5, up(maxs)).
table(up_v_c_maxc_s_mins_maxs, 2, 18, 5, up(maxs)).
table(up_v_minc_maxc_s_mins_maxs, 2, 18, 5, up(maxs)).
table(up_c_minc_maxc_s_mins_maxs, 2, 18, 5, up(maxs)).
table(up_v_c_minc_maxc_s_mins_maxs, 2, 18, 6, up(maxs)).
table(low_v_mina, 2, 19, 1, low(mina)).
table(low_c_mina, 2, 11, 1, low(mina)).
table(low_minc_mina, 2, 19, 1, low(mina)).
table(low_maxc_mina, 2, 19, 1, low(mina)).
table(low_s_mina, 2, 19, 1, low(mina)).
table(low_mins_mina, 2, 11, 1, low(mina)).
table(low_maxs_mina, 2, 11, 1, low(mina)).
table(low_v_c_mina, 2, 19, 2, low(mina)).
table(low_v_minc_mina, 2, 19, 2, low(mina)).
table(low_v_maxc_mina, 2, 19, 2, low(mina)).
table(low_v_s_mina, 2, 19, 2, low(mina)).
table(low_v_mins_mina, 2, 19, 2, low(mina)).
table(low_v_maxs_mina, 2, 19, 2, low(mina)).
table(low_c_minc_mina, 2, 19, 2, low(mina)).
table(low_c_maxc_mina, 2, 19, 2, low(mina)).
table(low_c_s_mina, 2, 19, 2, low(mina)).
table(low_c_mins_mina, 2, 11, 2, low(mina)).
table(low_c_maxs_mina, 2, 11, 2, low(mina)).
table(low_minc_maxc_mina, 2, 19, 2, low(mina)).
table(low_minc_s_mina, 2, 19, 2, low(mina)).
table(low_minc_mins_mina, 2, 19, 2, low(mina)).
table(low_minc_maxs_mina, 2, 19, 2, low(mina)).
table(low_maxc_s_mina, 2, 19, 2, low(mina)).
table(low_maxc_mins_mina, 2, 19, 2, low(mina)).
table(low_maxc_maxs_mina, 2, 19, 2, low(mina)).
table(low_s_mins_mina, 2, 19, 2, low(mina)).
table(low_s_maxs_mina, 2, 19, 2, low(mina)).
table(low_mins_maxs_mina, 2, 11, 2, low(mina)).
table(low_v_c_minc_mina, 2, 19, 3, low(mina)).
table(low_v_c_maxc_mina, 2, 19, 3, low(mina)).
table(low_v_c_s_mina, 2, 19, 3, low(mina)).
table(low_v_c_mins_mina, 2, 19, 3, low(mina)).
table(low_v_c_maxs_mina, 2, 19, 3, low(mina)).
table(low_v_minc_maxc_mina, 2, 19, 3, low(mina)).
table(low_v_minc_s_mina, 2, 19, 3, low(mina)).
table(low_v_minc_mins_mina, 2, 19, 3, low(mina)).
table(low_v_minc_maxs_mina, 2, 19, 3, low(mina)).
table(low_v_maxc_s_mina, 2, 19, 3, low(mina)).
table(low_v_maxc_mins_mina, 2, 19, 3, low(mina)).
table(low_v_maxc_maxs_mina, 2, 19, 3, low(mina)).
table(low_v_s_mins_mina, 2, 19, 3, low(mina)).
table(low_v_s_maxs_mina, 2, 19, 3, low(mina)).
table(low_v_mins_maxs_mina, 2, 19, 3, low(mina)).
table(low_c_minc_maxc_mina, 2, 19, 3, low(mina)).
table(low_c_minc_s_mina, 2, 19, 3, low(mina)).
table(low_c_minc_mins_mina, 2, 19, 3, low(mina)).
table(low_c_minc_maxs_mina, 2, 19, 3, low(mina)).
table(low_c_maxc_s_mina, 2, 19, 3, low(mina)).
table(low_c_maxc_mins_mina, 2, 19, 3, low(mina)).
table(low_c_maxc_maxs_mina, 2, 19, 3, low(mina)).
table(low_c_s_mins_mina, 2, 19, 3, low(mina)).
table(low_c_s_maxs_mina, 2, 19, 3, low(mina)).
table(low_c_mins_maxs_mina, 2, 11, 3, low(mina)).
table(low_minc_maxc_s_mina, 2, 19, 3, low(mina)).
table(low_minc_maxc_mins_mina, 2, 19, 3, low(mina)).
table(low_minc_maxc_maxs_mina, 2, 19, 3, low(mina)).
table(low_minc_s_mins_mina, 2, 19, 3, low(mina)).
table(low_minc_s_maxs_mina, 2, 19, 3, low(mina)).
table(low_minc_mins_maxs_mina, 2, 19, 3, low(mina)).
table(low_maxc_s_mins_mina, 2, 19, 3, low(mina)).
table(low_maxc_s_maxs_mina, 2, 19, 3, low(mina)).
table(low_maxc_mins_maxs_mina, 2, 19, 3, low(mina)).
table(low_s_mins_maxs_mina, 2, 19, 3, low(mina)).
table(low_v_c_minc_maxc_mina, 2, 19, 4, low(mina)).
table(low_v_c_minc_s_mina, 2, 19, 4, low(mina)).
table(low_v_c_minc_mins_mina, 2, 19, 4, low(mina)).
table(low_v_c_minc_maxs_mina, 2, 19, 4, low(mina)).
table(low_v_c_maxc_s_mina, 2, 19, 4, low(mina)).
table(low_v_c_maxc_mins_mina, 2, 19, 4, low(mina)).
table(low_v_c_maxc_maxs_mina, 2, 19, 4, low(mina)).
table(low_v_c_s_mins_mina, 2, 19, 4, low(mina)).
table(low_v_c_s_maxs_mina, 2, 19, 4, low(mina)).
table(low_v_c_mins_maxs_mina, 2, 19, 4, low(mina)).
table(low_v_minc_maxc_s_mina, 2, 19, 4, low(mina)).
table(low_v_minc_maxc_mins_mina, 2, 19, 4, low(mina)).
table(low_v_minc_maxc_maxs_mina, 2, 19, 4, low(mina)).
table(low_v_minc_s_mins_mina, 2, 19, 4, low(mina)).
table(low_v_minc_s_maxs_mina, 2, 19, 4, low(mina)).
table(low_v_minc_mins_maxs_mina, 2, 19, 4, low(mina)).
table(low_v_maxc_s_mins_mina, 2, 19, 4, low(mina)).
table(low_v_maxc_s_maxs_mina, 2, 19, 4, low(mina)).
table(low_v_maxc_mins_maxs_mina, 2, 19, 4, low(mina)).
table(low_v_s_mins_maxs_mina, 2, 19, 4, low(mina)).
table(low_c_minc_maxc_s_mina, 2, 19, 4, low(mina)).
table(low_c_minc_maxc_mins_mina, 2, 19, 4, low(mina)).
table(low_c_minc_maxc_maxs_mina, 2, 19, 4, low(mina)).
table(low_c_minc_s_mins_mina, 2, 19, 4, low(mina)).
table(low_c_minc_s_maxs_mina, 2, 19, 4, low(mina)).
table(low_c_minc_mins_maxs_mina, 2, 19, 4, low(mina)).
table(low_c_maxc_s_mins_mina, 2, 19, 4, low(mina)).
table(low_c_maxc_s_maxs_mina, 2, 19, 4, low(mina)).
table(low_c_maxc_mins_maxs_mina, 2, 19, 4, low(mina)).
table(low_c_s_mins_maxs_mina, 2, 19, 4, low(mina)).
table(low_minc_maxc_s_mins_mina, 2, 19, 4, low(mina)).
table(low_minc_maxc_s_maxs_mina, 2, 19, 4, low(mina)).
table(low_minc_maxc_mins_maxs_mina, 2, 19, 4, low(mina)).
table(low_minc_s_mins_maxs_mina, 2, 19, 4, low(mina)).
table(low_maxc_s_mins_maxs_mina, 2, 19, 4, low(mina)).
table(low_v_c_minc_maxc_s_mina, 2, 19, 5, low(mina)).
table(low_v_c_minc_maxc_mins_mina, 2, 19, 5, low(mina)).
table(low_v_c_minc_maxc_maxs_mina, 2, 19, 5, low(mina)).
table(low_v_c_minc_s_mins_mina, 2, 19, 5, low(mina)).
table(low_v_c_minc_s_maxs_mina, 2, 19, 5, low(mina)).
table(low_v_c_minc_mins_maxs_mina, 2, 19, 5, low(mina)).
table(low_v_c_maxc_s_mins_mina, 2, 19, 5, low(mina)).
table(low_v_c_maxc_s_maxs_mina, 2, 19, 5, low(mina)).
table(low_v_c_maxc_mins_maxs_mina, 2, 19, 5, low(mina)).
table(low_v_c_s_mins_maxs_mina, 2, 19, 5, low(mina)).
table(low_v_minc_maxc_s_mins_mina, 2, 19, 5, low(mina)).
table(low_v_minc_maxc_s_maxs_mina, 2, 19, 5, low(mina)).
table(low_v_minc_maxc_mins_maxs_mina, 2, 19, 5, low(mina)).
table(low_v_minc_s_mins_maxs_mina, 2, 19, 5, low(mina)).
table(low_v_maxc_s_mins_maxs_mina, 2, 19, 5, low(mina)).
table(low_c_minc_maxc_s_mins_mina, 2, 19, 5, low(mina)).
table(low_c_minc_maxc_s_maxs_mina, 2, 19, 5, low(mina)).
table(low_c_minc_maxc_mins_maxs_mina, 2, 19, 5, low(mina)).
table(low_c_minc_s_mins_maxs_mina, 2, 19, 5, low(mina)).
table(low_c_maxc_s_mins_maxs_mina, 2, 19, 5, low(mina)).
table(low_minc_maxc_s_mins_maxs_mina, 2, 19, 5, low(mina)).
table(low_v_c_minc_maxc_s_mins_mina, 2, 19, 6, low(mina)).
table(low_v_c_minc_maxc_s_maxs_mina, 2, 19, 6, low(mina)).
table(low_v_c_minc_maxc_mins_maxs_mina, 2, 19, 6, low(mina)).
table(low_v_c_minc_s_mins_maxs_mina, 2, 19, 6, low(mina)).
table(low_v_c_maxc_s_mins_maxs_mina, 2, 19, 6, low(mina)).
table(low_v_minc_maxc_s_mins_maxs_mina, 2, 19, 6, low(mina)).
table(low_c_minc_maxc_s_mins_maxs_mina, 2, 19, 6, low(mina)).
table(low_v_c_minc_maxc_s_mins_maxs_mina, 2, 19, 7, low(mina)).
table(up_v_maxa, 2, 19, 1, up(maxa)).
table(up_v_c_maxa, 2, 19, 2, up(maxa)).
table(up_v_minc_maxa, 2, 19, 2, up(maxa)).
table(up_v_maxc_maxa, 2, 19, 2, up(maxa)).
table(up_v_s_maxa, 2, 19, 2, up(maxa)).
table(up_v_mins_maxa, 2, 19, 2, up(maxa)).
table(up_v_maxs_maxa, 2, 19, 2, up(maxa)).
table(up_c_maxc_maxa, 2, 19, 2, up(maxa)).
table(up_maxc_s_maxa, 2, 19, 2, up(maxa)).
table(up_s_maxs_maxa, 2, 19, 2, up(maxa)).
table(up_v_c_minc_maxa, 2, 19, 3, up(maxa)).
table(up_v_c_maxc_maxa, 2, 19, 3, up(maxa)).
table(up_v_c_s_maxa, 2, 19, 3, up(maxa)).
table(up_v_c_mins_maxa, 2, 19, 3, up(maxa)).
table(up_v_c_maxs_maxa, 2, 19, 3, up(maxa)).
table(up_v_minc_maxc_maxa, 2, 19, 3, up(maxa)).
table(up_v_minc_s_maxa, 2, 19, 3, up(maxa)).
table(up_v_minc_mins_maxa, 2, 19, 3, up(maxa)).
table(up_v_minc_maxs_maxa, 2, 19, 3, up(maxa)).
table(up_v_maxc_s_maxa, 2, 19, 3, up(maxa)).
table(up_v_maxc_mins_maxa, 2, 19, 3, up(maxa)).
table(up_v_maxc_maxs_maxa, 2, 19, 3, up(maxa)).
table(up_v_s_mins_maxa, 2, 19, 3, up(maxa)).
table(up_v_s_maxs_maxa, 2, 19, 3, up(maxa)).
table(up_v_mins_maxs_maxa, 2, 19, 3, up(maxa)).
table(up_c_minc_maxc_maxa, 2, 19, 3, up(maxa)).
table(up_c_maxc_s_maxa, 2, 19, 3, up(maxa)).
table(up_c_maxc_mins_maxa, 2, 19, 3, up(maxa)).
table(up_c_maxc_maxs_maxa, 2, 19, 3, up(maxa)).
table(up_c_s_maxs_maxa, 2, 19, 3, up(maxa)).
table(up_minc_maxc_s_maxa, 2, 19, 3, up(maxa)).
table(up_minc_s_maxs_maxa, 2, 19, 3, up(maxa)).
table(up_maxc_s_mins_maxa, 2, 19, 3, up(maxa)).
table(up_maxc_s_maxs_maxa, 2, 19, 3, up(maxa)).
table(up_s_mins_maxs_maxa, 2, 19, 3, up(maxa)).
table(up_v_c_minc_maxc_maxa, 2, 19, 4, up(maxa)).
table(up_v_c_minc_s_maxa, 2, 19, 4, up(maxa)).
table(up_v_c_minc_mins_maxa, 2, 19, 4, up(maxa)).
table(up_v_c_minc_maxs_maxa, 2, 19, 4, up(maxa)).
table(up_v_c_maxc_s_maxa, 2, 19, 4, up(maxa)).
table(up_v_c_maxc_mins_maxa, 2, 19, 4, up(maxa)).
table(up_v_c_maxc_maxs_maxa, 2, 19, 4, up(maxa)).
table(up_v_c_s_mins_maxa, 2, 19, 4, up(maxa)).
table(up_v_c_s_maxs_maxa, 2, 19, 4, up(maxa)).
table(up_v_c_mins_maxs_maxa, 2, 19, 4, up(maxa)).
table(up_v_minc_maxc_s_maxa, 2, 19, 4, up(maxa)).
table(up_v_minc_maxc_mins_maxa, 2, 19, 4, up(maxa)).
table(up_v_minc_maxc_maxs_maxa, 2, 19, 4, up(maxa)).
table(up_v_minc_s_mins_maxa, 2, 19, 4, up(maxa)).
table(up_v_minc_s_maxs_maxa, 2, 19, 4, up(maxa)).
table(up_v_minc_mins_maxs_maxa, 2, 19, 4, up(maxa)).
table(up_v_maxc_s_mins_maxa, 2, 19, 4, up(maxa)).
table(up_v_maxc_s_maxs_maxa, 2, 19, 4, up(maxa)).
table(up_v_maxc_mins_maxs_maxa, 2, 19, 4, up(maxa)).
table(up_v_s_mins_maxs_maxa, 2, 19, 4, up(maxa)).
table(up_c_minc_maxc_s_maxa, 2, 19, 4, up(maxa)).
table(up_c_minc_maxc_mins_maxa, 2, 19, 4, up(maxa)).
table(up_c_minc_maxc_maxs_maxa, 2, 19, 4, up(maxa)).
table(up_c_minc_s_maxs_maxa, 2, 19, 4, up(maxa)).
table(up_c_maxc_s_mins_maxa, 2, 19, 4, up(maxa)).
table(up_c_maxc_s_maxs_maxa, 2, 19, 4, up(maxa)).
table(up_c_maxc_mins_maxs_maxa, 2, 19, 4, up(maxa)).
table(up_c_s_mins_maxs_maxa, 2, 19, 4, up(maxa)).
table(up_minc_maxc_s_mins_maxa, 2, 19, 4, up(maxa)).
table(up_minc_maxc_s_maxs_maxa, 2, 19, 4, up(maxa)).
table(up_minc_s_mins_maxs_maxa, 2, 19, 4, up(maxa)).
table(up_maxc_s_mins_maxs_maxa, 2, 19, 4, up(maxa)).
table(up_v_c_minc_maxc_s_maxa, 2, 19, 5, up(maxa)).
table(up_v_c_minc_maxc_mins_maxa, 2, 19, 5, up(maxa)).
table(up_v_c_minc_maxc_maxs_maxa, 2, 19, 5, up(maxa)).
table(up_v_c_minc_s_mins_maxa, 2, 19, 5, up(maxa)).
table(up_v_c_minc_s_maxs_maxa, 2, 19, 5, up(maxa)).
table(up_v_c_minc_mins_maxs_maxa, 2, 19, 5, up(maxa)).
table(up_v_c_maxc_s_mins_maxa, 2, 19, 5, up(maxa)).
table(up_v_c_maxc_s_maxs_maxa, 2, 19, 5, up(maxa)).
table(up_v_c_maxc_mins_maxs_maxa, 2, 19, 5, up(maxa)).
table(up_v_c_s_mins_maxs_maxa, 2, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_mins_maxa, 2, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_maxs_maxa, 2, 19, 5, up(maxa)).
table(up_v_minc_maxc_mins_maxs_maxa, 2, 19, 5, up(maxa)).
table(up_v_minc_s_mins_maxs_maxa, 2, 19, 5, up(maxa)).
table(up_v_maxc_s_mins_maxs_maxa, 2, 19, 5, up(maxa)).
table(up_c_minc_maxc_s_mins_maxa, 2, 19, 5, up(maxa)).
table(up_c_minc_maxc_s_maxs_maxa, 2, 19, 5, up(maxa)).
table(up_c_minc_maxc_mins_maxs_maxa, 2, 19, 5, up(maxa)).
table(up_c_minc_s_mins_maxs_maxa, 2, 19, 5, up(maxa)).
table(up_c_maxc_s_mins_maxs_maxa, 2, 19, 5, up(maxa)).
table(up_minc_maxc_s_mins_maxs_maxa, 2, 19, 5, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxa, 2, 19, 6, up(maxa)).
table(up_v_c_minc_maxc_s_maxs_maxa, 2, 19, 6, up(maxa)).
table(up_v_c_minc_maxc_mins_maxs_maxa, 2, 19, 6, up(maxa)).
table(up_v_c_minc_s_mins_maxs_maxa, 2, 19, 6, up(maxa)).
table(up_v_c_maxc_s_mins_maxs_maxa, 2, 19, 6, up(maxa)).
table(up_v_minc_maxc_s_mins_maxs_maxa, 2, 19, 6, up(maxa)).
table(up_c_minc_maxc_s_mins_maxs_maxa, 2, 19, 6, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxs_maxa, 2, 19, 7, up(maxa)).
table(low_c_v, 3, 18, 1, low(v)).
table(low_minc_v, 3, 11, 1, low(v)).
table(low_maxc_v, 3, 11, 1, low(v)).
table(low_s_v, 3, 9, 1, low(v)).
table(low_mins_v, 3, 18, 1, low(v)).
table(low_maxs_v, 3, 18, 1, low(v)).
table(low_c_minc_v, 3, 11, 2, low(v)).
table(low_c_maxc_v, 3, 11, 2, low(v)).
table(low_c_s_v, 3, 18, 2, low(v)).
table(low_c_mins_v, 3, 18, 2, low(v)).
table(low_c_maxs_v, 3, 18, 2, low(v)).
table(low_minc_maxc_v, 3, 11, 2, low(v)).
table(low_minc_s_v, 3, 10, 2, low(v)).
table(low_minc_mins_v, 3, 12, 2, low(v)).
table(low_minc_maxs_v, 3, 18, 2, low(v)).
table(low_maxc_s_v, 3, 18, 2, low(v)).
table(low_maxc_mins_v, 3, 12, 2, low(v)).
table(low_maxc_maxs_v, 3, 18, 2, low(v)).
table(low_s_mins_v, 3, 9, 2, low(v)).
table(low_s_maxs_v, 3, 9, 2, low(v)).
table(low_mins_maxs_v, 3, 13, 2, low(v)).
table(low_c_minc_maxc_v, 3, 11, 3, low(v)).
table(low_c_minc_s_v, 3, 18, 3, low(v)).
table(low_c_minc_mins_v, 3, 12, 3, low(v)).
table(low_c_minc_maxs_v, 3, 18, 3, low(v)).
table(low_c_maxc_s_v, 3, 18, 3, low(v)).
table(low_c_maxc_mins_v, 3, 12, 3, low(v)).
table(low_c_maxc_maxs_v, 3, 18, 3, low(v)).
table(low_c_s_mins_v, 3, 18, 3, low(v)).
table(low_c_s_maxs_v, 3, 18, 3, low(v)).
table(low_c_mins_maxs_v, 3, 18, 3, low(v)).
table(low_minc_maxc_s_v, 3, 18, 3, low(v)).
table(low_minc_maxc_mins_v, 3, 12, 3, low(v)).
table(low_minc_maxc_maxs_v, 3, 18, 3, low(v)).
table(low_minc_s_mins_v, 3, 10, 3, low(v)).
table(low_minc_s_maxs_v, 3, 10, 3, low(v)).
table(low_minc_mins_maxs_v, 3, 18, 3, low(v)).
table(low_maxc_s_mins_v, 3, 18, 3, low(v)).
table(low_maxc_s_maxs_v, 3, 18, 3, low(v)).
table(low_maxc_mins_maxs_v, 3, 18, 3, low(v)).
table(low_s_mins_maxs_v, 3, 9, 3, low(v)).
table(low_c_minc_maxc_s_v, 3, 18, 4, low(v)).
table(low_c_minc_maxc_mins_v, 3, 12, 4, low(v)).
table(low_c_minc_maxc_maxs_v, 3, 18, 4, low(v)).
table(low_c_minc_s_mins_v, 3, 18, 4, low(v)).
table(low_c_minc_s_maxs_v, 3, 18, 4, low(v)).
table(low_c_minc_mins_maxs_v, 3, 18, 4, low(v)).
table(low_c_maxc_s_mins_v, 3, 18, 4, low(v)).
table(low_c_maxc_s_maxs_v, 3, 18, 4, low(v)).
table(low_c_maxc_mins_maxs_v, 3, 18, 4, low(v)).
table(low_c_s_mins_maxs_v, 3, 18, 4, low(v)).
table(low_minc_maxc_s_mins_v, 3, 18, 4, low(v)).
table(low_minc_maxc_s_maxs_v, 3, 18, 4, low(v)).
table(low_minc_maxc_mins_maxs_v, 3, 18, 4, low(v)).
table(low_minc_s_mins_maxs_v, 3, 10, 4, low(v)).
table(low_maxc_s_mins_maxs_v, 3, 18, 4, low(v)).
table(low_c_minc_maxc_s_mins_v, 3, 18, 5, low(v)).
table(low_c_minc_maxc_s_maxs_v, 3, 18, 5, low(v)).
table(low_c_minc_maxc_mins_maxs_v, 3, 18, 5, low(v)).
table(low_c_minc_s_mins_maxs_v, 3, 18, 5, low(v)).
table(low_c_maxc_s_mins_maxs_v, 3, 18, 5, low(v)).
table(low_minc_maxc_s_mins_maxs_v, 3, 18, 5, low(v)).
table(low_c_minc_maxc_s_mins_maxs_v, 3, 18, 6, low(v)).
table(up_c_maxc_v, 3, 11, 2, up(v)).
table(up_maxc_s_v, 3, 18, 2, up(v)).
table(up_s_maxs_v, 3, 9, 2, up(v)).
table(up_c_minc_maxc_v, 3, 11, 3, up(v)).
table(up_c_maxc_s_v, 3, 18, 3, up(v)).
table(up_c_maxc_mins_v, 3, 12, 3, up(v)).
table(up_c_maxc_maxs_v, 3, 18, 3, up(v)).
table(up_c_s_maxs_v, 3, 18, 3, up(v)).
table(up_minc_maxc_s_v, 3, 18, 3, up(v)).
table(up_minc_s_maxs_v, 3, 10, 3, up(v)).
table(up_maxc_s_mins_v, 3, 18, 3, up(v)).
table(up_maxc_s_maxs_v, 3, 18, 3, up(v)).
table(up_s_mins_maxs_v, 3, 9, 3, up(v)).
table(up_c_minc_maxc_s_v, 3, 18, 4, up(v)).
table(up_c_minc_maxc_mins_v, 3, 12, 4, up(v)).
table(up_c_minc_maxc_maxs_v, 3, 18, 4, up(v)).
table(up_c_minc_s_maxs_v, 3, 18, 4, up(v)).
table(up_c_maxc_s_mins_v, 3, 18, 4, up(v)).
table(up_c_maxc_s_maxs_v, 3, 18, 4, up(v)).
table(up_c_maxc_mins_maxs_v, 3, 18, 4, up(v)).
table(up_c_s_mins_maxs_v, 3, 18, 4, up(v)).
table(up_minc_maxc_s_mins_v, 3, 18, 4, up(v)).
table(up_minc_maxc_s_maxs_v, 3, 18, 4, up(v)).
table(up_minc_s_mins_maxs_v, 3, 10, 4, up(v)).
table(up_maxc_s_mins_maxs_v, 3, 18, 4, up(v)).
table(up_c_minc_maxc_s_mins_v, 3, 18, 5, up(v)).
table(up_c_minc_maxc_s_maxs_v, 3, 18, 5, up(v)).
table(up_c_minc_maxc_mins_maxs_v, 3, 18, 5, up(v)).
table(up_c_minc_s_mins_maxs_v, 3, 18, 5, up(v)).
table(up_c_maxc_s_mins_maxs_v, 3, 18, 5, up(v)).
table(up_minc_maxc_s_mins_maxs_v, 3, 18, 5, up(v)).
table(up_c_minc_maxc_s_mins_maxs_v, 3, 18, 6, up(v)).
table(low_v_c, 3, 11, 1, low(c)).
table(low_minc_c, 3, 11, 1, low(c)).
table(low_maxc_c, 3, 11, 1, low(c)).
table(low_s_c, 3, 8, 1, low(c)).
table(low_mins_c, 3, 8, 1, low(c)).
table(low_maxs_c, 3, 9, 1, low(c)).
table(low_v_minc_c, 3, 11, 2, low(c)).
table(low_v_maxc_c, 3, 11, 2, low(c)).
table(low_v_s_c, 3, 18, 2, low(c)).
table(low_v_mins_c, 3, 12, 2, low(c)).
table(low_v_maxs_c, 3, 18, 2, low(c)).
table(low_minc_maxc_c, 3, 11, 2, low(c)).
table(low_minc_s_c, 3, 7, 2, low(c)).
table(low_minc_mins_c, 3, 12, 2, low(c)).
table(low_minc_maxs_c, 3, 18, 2, low(c)).
table(low_maxc_s_c, 3, 18, 2, low(c)).
table(low_maxc_mins_c, 3, 12, 2, low(c)).
table(low_maxc_maxs_c, 3, 18, 2, low(c)).
table(low_s_mins_c, 3, 10, 2, low(c)).
table(low_s_maxs_c, 3, 18, 2, low(c)).
table(low_mins_maxs_c, 3, 10, 2, low(c)).
table(low_v_minc_maxc_c, 3, 11, 3, low(c)).
table(low_v_minc_s_c, 3, 18, 3, low(c)).
table(low_v_minc_mins_c, 3, 12, 3, low(c)).
table(low_v_minc_maxs_c, 3, 18, 3, low(c)).
table(low_v_maxc_s_c, 3, 18, 3, low(c)).
table(low_v_maxc_mins_c, 3, 12, 3, low(c)).
table(low_v_maxc_maxs_c, 3, 18, 3, low(c)).
table(low_v_s_mins_c, 3, 18, 3, low(c)).
table(low_v_s_maxs_c, 3, 18, 3, low(c)).
table(low_v_mins_maxs_c, 3, 18, 3, low(c)).
table(low_minc_maxc_s_c, 3, 18, 3, low(c)).
table(low_minc_maxc_mins_c, 3, 12, 3, low(c)).
table(low_minc_maxc_maxs_c, 3, 18, 3, low(c)).
table(low_minc_s_mins_c, 3, 7, 3, low(c)).
table(low_minc_s_maxs_c, 3, 18, 3, low(c)).
table(low_minc_mins_maxs_c, 3, 18, 3, low(c)).
table(low_maxc_s_mins_c, 3, 18, 3, low(c)).
table(low_maxc_s_maxs_c, 3, 18, 3, low(c)).
table(low_maxc_mins_maxs_c, 3, 18, 3, low(c)).
table(low_s_mins_maxs_c, 3, 18, 3, low(c)).
table(low_v_minc_maxc_s_c, 3, 18, 4, low(c)).
table(low_v_minc_maxc_mins_c, 3, 12, 4, low(c)).
table(low_v_minc_maxc_maxs_c, 3, 18, 4, low(c)).
table(low_v_minc_s_mins_c, 3, 18, 4, low(c)).
table(low_v_minc_s_maxs_c, 3, 18, 4, low(c)).
table(low_v_minc_mins_maxs_c, 3, 18, 4, low(c)).
table(low_v_maxc_s_mins_c, 3, 18, 4, low(c)).
table(low_v_maxc_s_maxs_c, 3, 18, 4, low(c)).
table(low_v_maxc_mins_maxs_c, 3, 18, 4, low(c)).
table(low_v_s_mins_maxs_c, 3, 18, 4, low(c)).
table(low_minc_maxc_s_mins_c, 3, 18, 4, low(c)).
table(low_minc_maxc_s_maxs_c, 3, 18, 4, low(c)).
table(low_minc_maxc_mins_maxs_c, 3, 18, 4, low(c)).
table(low_minc_s_mins_maxs_c, 3, 18, 4, low(c)).
table(low_maxc_s_mins_maxs_c, 3, 18, 4, low(c)).
table(low_v_minc_maxc_s_mins_c, 3, 18, 5, low(c)).
table(low_v_minc_maxc_s_maxs_c, 3, 18, 5, low(c)).
table(low_v_minc_maxc_mins_maxs_c, 3, 18, 5, low(c)).
table(low_v_minc_s_mins_maxs_c, 3, 18, 5, low(c)).
table(low_v_maxc_s_mins_maxs_c, 3, 18, 5, low(c)).
table(low_minc_maxc_s_mins_maxs_c, 3, 18, 5, low(c)).
table(low_v_minc_maxc_s_mins_maxs_c, 3, 18, 6, low(c)).
table(up_v_c, 3, 18, 1, up(c)).
table(up_s_c, 3, 5, 1, up(c)).
table(up_v_minc_c, 3, 11, 2, up(c)).
table(up_v_maxc_c, 3, 11, 2, up(c)).
table(up_v_s_c, 3, 18, 2, up(c)).
table(up_v_mins_c, 3, 18, 2, up(c)).
table(up_v_maxs_c, 3, 18, 2, up(c)).
table(up_minc_s_c, 3, 7, 2, up(c)).
table(up_maxc_s_c, 3, 18, 2, up(c)).
table(up_s_mins_c, 3, 7, 2, up(c)).
table(up_s_maxs_c, 3, 18, 2, up(c)).
table(up_v_minc_maxc_c, 3, 11, 3, up(c)).
table(up_v_minc_s_c, 3, 18, 3, up(c)).
table(up_v_minc_mins_c, 3, 12, 3, up(c)).
table(up_v_minc_maxs_c, 3, 18, 3, up(c)).
table(up_v_maxc_s_c, 3, 18, 3, up(c)).
table(up_v_maxc_mins_c, 3, 12, 3, up(c)).
table(up_v_maxc_maxs_c, 3, 18, 3, up(c)).
table(up_v_s_mins_c, 3, 18, 3, up(c)).
table(up_v_s_maxs_c, 3, 18, 3, up(c)).
table(up_v_mins_maxs_c, 3, 18, 3, up(c)).
table(up_minc_maxc_s_c, 3, 18, 3, up(c)).
table(up_minc_s_mins_c, 3, 7, 3, up(c)).
table(up_minc_s_maxs_c, 3, 18, 3, up(c)).
table(up_maxc_s_mins_c, 3, 18, 3, up(c)).
table(up_maxc_s_maxs_c, 3, 18, 3, up(c)).
table(up_s_mins_maxs_c, 3, 18, 3, up(c)).
table(up_v_minc_maxc_s_c, 3, 18, 4, up(c)).
table(up_v_minc_maxc_mins_c, 3, 12, 4, up(c)).
table(up_v_minc_maxc_maxs_c, 3, 18, 4, up(c)).
table(up_v_minc_s_mins_c, 3, 18, 4, up(c)).
table(up_v_minc_s_maxs_c, 3, 18, 4, up(c)).
table(up_v_minc_mins_maxs_c, 3, 18, 4, up(c)).
table(up_v_maxc_s_mins_c, 3, 18, 4, up(c)).
table(up_v_maxc_s_maxs_c, 3, 18, 4, up(c)).
table(up_v_maxc_mins_maxs_c, 3, 18, 4, up(c)).
table(up_v_s_mins_maxs_c, 3, 18, 4, up(c)).
table(up_minc_maxc_s_mins_c, 3, 18, 4, up(c)).
table(up_minc_maxc_s_maxs_c, 3, 18, 4, up(c)).
table(up_minc_s_mins_maxs_c, 3, 18, 4, up(c)).
table(up_maxc_s_mins_maxs_c, 3, 18, 4, up(c)).
table(up_v_minc_maxc_s_mins_c, 3, 18, 5, up(c)).
table(up_v_minc_maxc_s_maxs_c, 3, 18, 5, up(c)).
table(up_v_minc_maxc_mins_maxs_c, 3, 18, 5, up(c)).
table(up_v_minc_s_mins_maxs_c, 3, 18, 5, up(c)).
table(up_v_maxc_s_mins_maxs_c, 3, 18, 5, up(c)).
table(up_minc_maxc_s_mins_maxs_c, 3, 18, 5, up(c)).
table(up_v_minc_maxc_s_mins_maxs_c, 3, 18, 6, up(c)).
table(low_v_minc, 3, 5, 1, low(minc)).
table(low_c_minc, 3, 5, 1, low(minc)).
table(low_maxc_minc, 3, 7, 1, low(minc)).
table(low_s_minc, 3, 5, 1, low(minc)).
table(low_mins_minc, 3, 4, 1, low(minc)).
table(low_maxs_minc, 3, 6, 1, low(minc)).
table(low_v_c_minc, 3, 11, 2, low(minc)).
table(low_v_maxc_minc, 3, 11, 2, low(minc)).
table(low_v_s_minc, 3, 10, 2, low(minc)).
table(low_v_mins_minc, 3, 5, 2, low(minc)).
table(low_v_maxs_minc, 3, 10, 2, low(minc)).
table(low_c_maxc_minc, 3, 11, 2, low(minc)).
table(low_c_s_minc, 3, 7, 2, low(minc)).
table(low_c_mins_minc, 3, 5, 2, low(minc)).
table(low_c_maxs_minc, 3, 7, 2, low(minc)).
table(low_maxc_s_minc, 3, 18, 2, low(minc)).
table(low_maxc_mins_minc, 3, 8, 2, low(minc)).
table(low_maxc_maxs_minc, 3, 11, 2, low(minc)).
table(low_s_mins_minc, 3, 5, 2, low(minc)).
table(low_s_maxs_minc, 3, 10, 2, low(minc)).
table(low_mins_maxs_minc, 3, 6, 2, low(minc)).
table(low_v_c_maxc_minc, 3, 11, 3, low(minc)).
table(low_v_c_s_minc, 3, 18, 3, low(minc)).
table(low_v_c_mins_minc, 3, 12, 3, low(minc)).
table(low_v_c_maxs_minc, 3, 18, 3, low(minc)).
table(low_v_maxc_s_minc, 3, 18, 3, low(minc)).
table(low_v_maxc_mins_minc, 3, 12, 3, low(minc)).
table(low_v_maxc_maxs_minc, 3, 18, 3, low(minc)).
table(low_v_s_mins_minc, 3, 10, 3, low(minc)).
table(low_v_s_maxs_minc, 3, 10, 3, low(minc)).
table(low_v_mins_maxs_minc, 3, 10, 3, low(minc)).
table(low_c_maxc_s_minc, 3, 18, 3, low(minc)).
table(low_c_maxc_mins_minc, 3, 12, 3, low(minc)).
table(low_c_maxc_maxs_minc, 3, 18, 3, low(minc)).
table(low_c_s_mins_minc, 3, 7, 3, low(minc)).
table(low_c_s_maxs_minc, 3, 18, 3, low(minc)).
table(low_c_mins_maxs_minc, 3, 7, 3, low(minc)).
table(low_maxc_s_mins_minc, 3, 18, 3, low(minc)).
table(low_maxc_s_maxs_minc, 3, 18, 3, low(minc)).
table(low_maxc_mins_maxs_minc, 3, 11, 3, low(minc)).
table(low_s_mins_maxs_minc, 3, 10, 3, low(minc)).
table(low_v_c_maxc_s_minc, 3, 18, 4, low(minc)).
table(low_v_c_maxc_mins_minc, 3, 12, 4, low(minc)).
table(low_v_c_maxc_maxs_minc, 3, 18, 4, low(minc)).
table(low_v_c_s_mins_minc, 3, 18, 4, low(minc)).
table(low_v_c_s_maxs_minc, 3, 18, 4, low(minc)).
table(low_v_c_mins_maxs_minc, 3, 18, 4, low(minc)).
table(low_v_maxc_s_mins_minc, 3, 18, 4, low(minc)).
table(low_v_maxc_s_maxs_minc, 3, 18, 4, low(minc)).
table(low_v_maxc_mins_maxs_minc, 3, 18, 4, low(minc)).
table(low_v_s_mins_maxs_minc, 3, 10, 4, low(minc)).
table(low_c_maxc_s_mins_minc, 3, 18, 4, low(minc)).
table(low_c_maxc_s_maxs_minc, 3, 18, 4, low(minc)).
table(low_c_maxc_mins_maxs_minc, 3, 18, 4, low(minc)).
table(low_c_s_mins_maxs_minc, 3, 18, 4, low(minc)).
table(low_maxc_s_mins_maxs_minc, 3, 18, 4, low(minc)).
table(low_v_c_maxc_s_mins_minc, 3, 18, 5, low(minc)).
table(low_v_c_maxc_s_maxs_minc, 3, 18, 5, low(minc)).
table(low_v_c_maxc_mins_maxs_minc, 3, 18, 5, low(minc)).
table(low_v_c_s_mins_maxs_minc, 3, 18, 5, low(minc)).
table(low_v_maxc_s_mins_maxs_minc, 3, 18, 5, low(minc)).
table(low_c_maxc_s_mins_maxs_minc, 3, 18, 5, low(minc)).
table(low_v_c_maxc_s_mins_maxs_minc, 3, 18, 6, low(minc)).
table(up_v_minc, 3, 11, 1, up(minc)).
table(up_maxc_minc, 3, 7, 1, up(minc)).
table(up_v_c_minc, 3, 11, 2, up(minc)).
table(up_v_maxc_minc, 3, 11, 2, up(minc)).
table(up_v_s_minc, 3, 18, 2, up(minc)).
table(up_v_mins_minc, 3, 12, 2, up(minc)).
table(up_v_maxs_minc, 3, 18, 2, up(minc)).
table(up_c_maxc_minc, 3, 11, 2, up(minc)).
table(up_maxc_s_minc, 3, 18, 2, up(minc)).
table(up_maxc_mins_minc, 3, 8, 2, up(minc)).
table(up_maxc_maxs_minc, 3, 11, 2, up(minc)).
table(up_s_maxs_minc, 3, 18, 2, up(minc)).
table(up_v_c_maxc_minc, 3, 11, 3, up(minc)).
table(up_v_c_s_minc, 3, 18, 3, up(minc)).
table(up_v_c_mins_minc, 3, 12, 3, up(minc)).
table(up_v_c_maxs_minc, 3, 18, 3, up(minc)).
table(up_v_maxc_s_minc, 3, 18, 3, up(minc)).
table(up_v_maxc_mins_minc, 3, 12, 3, up(minc)).
table(up_v_maxc_maxs_minc, 3, 18, 3, up(minc)).
table(up_v_s_mins_minc, 3, 18, 3, up(minc)).
table(up_v_s_maxs_minc, 3, 18, 3, up(minc)).
table(up_v_mins_maxs_minc, 3, 18, 3, up(minc)).
table(up_c_maxc_s_minc, 3, 18, 3, up(minc)).
table(up_c_maxc_mins_minc, 3, 12, 3, up(minc)).
table(up_c_maxc_maxs_minc, 3, 18, 3, up(minc)).
table(up_c_s_maxs_minc, 3, 18, 3, up(minc)).
table(up_maxc_s_mins_minc, 3, 18, 3, up(minc)).
table(up_maxc_s_maxs_minc, 3, 18, 3, up(minc)).
table(up_maxc_mins_maxs_minc, 3, 11, 3, up(minc)).
table(up_s_mins_maxs_minc, 3, 18, 3, up(minc)).
table(up_v_c_maxc_s_minc, 3, 18, 4, up(minc)).
table(up_v_c_maxc_mins_minc, 3, 12, 4, up(minc)).
table(up_v_c_maxc_maxs_minc, 3, 18, 4, up(minc)).
table(up_v_c_s_mins_minc, 3, 18, 4, up(minc)).
table(up_v_c_s_maxs_minc, 3, 18, 4, up(minc)).
table(up_v_c_mins_maxs_minc, 3, 18, 4, up(minc)).
table(up_v_maxc_s_mins_minc, 3, 18, 4, up(minc)).
table(up_v_maxc_s_maxs_minc, 3, 18, 4, up(minc)).
table(up_v_maxc_mins_maxs_minc, 3, 18, 4, up(minc)).
table(up_v_s_mins_maxs_minc, 3, 18, 4, up(minc)).
table(up_c_maxc_s_mins_minc, 3, 18, 4, up(minc)).
table(up_c_maxc_s_maxs_minc, 3, 18, 4, up(minc)).
table(up_c_maxc_mins_maxs_minc, 3, 18, 4, up(minc)).
table(up_c_s_mins_maxs_minc, 3, 18, 4, up(minc)).
table(up_maxc_s_mins_maxs_minc, 3, 18, 4, up(minc)).
table(up_v_c_maxc_s_mins_minc, 3, 18, 5, up(minc)).
table(up_v_c_maxc_s_maxs_minc, 3, 18, 5, up(minc)).
table(up_v_c_maxc_mins_maxs_minc, 3, 18, 5, up(minc)).
table(up_v_c_s_mins_maxs_minc, 3, 18, 5, up(minc)).
table(up_v_maxc_s_mins_maxs_minc, 3, 18, 5, up(minc)).
table(up_c_maxc_s_mins_maxs_minc, 3, 18, 5, up(minc)).
table(up_v_c_maxc_s_mins_maxs_minc, 3, 18, 6, up(minc)).
table(low_v_maxc, 3, 18, 1, low(maxc)).
table(low_c_maxc, 3, 18, 1, low(maxc)).
table(low_minc_maxc, 3, 7, 1, low(maxc)).
table(low_s_maxc, 3, 18, 1, low(maxc)).
table(low_mins_maxc, 3, 11, 1, low(maxc)).
table(low_maxs_maxc, 3, 7, 1, low(maxc)).
table(low_v_c_maxc, 3, 11, 2, low(maxc)).
table(low_v_minc_maxc, 3, 11, 2, low(maxc)).
table(low_v_s_maxc, 3, 18, 2, low(maxc)).
table(low_v_mins_maxc, 3, 18, 2, low(maxc)).
table(low_v_maxs_maxc, 3, 18, 2, low(maxc)).
table(low_c_minc_maxc, 3, 11, 2, low(maxc)).
table(low_c_s_maxc, 3, 18, 2, low(maxc)).
table(low_c_mins_maxc, 3, 18, 2, low(maxc)).
table(low_c_maxs_maxc, 3, 18, 2, low(maxc)).
table(low_minc_s_maxc, 3, 18, 2, low(maxc)).
table(low_minc_mins_maxc, 3, 8, 2, low(maxc)).
table(low_minc_maxs_maxc, 3, 11, 2, low(maxc)).
table(low_s_mins_maxc, 3, 18, 2, low(maxc)).
table(low_s_maxs_maxc, 3, 18, 2, low(maxc)).
table(low_mins_maxs_maxc, 3, 11, 2, low(maxc)).
table(low_v_c_minc_maxc, 3, 11, 3, low(maxc)).
table(low_v_c_s_maxc, 3, 18, 3, low(maxc)).
table(low_v_c_mins_maxc, 3, 12, 3, low(maxc)).
table(low_v_c_maxs_maxc, 3, 18, 3, low(maxc)).
table(low_v_minc_s_maxc, 3, 18, 3, low(maxc)).
table(low_v_minc_mins_maxc, 3, 12, 3, low(maxc)).
table(low_v_minc_maxs_maxc, 3, 18, 3, low(maxc)).
table(low_v_s_mins_maxc, 3, 18, 3, low(maxc)).
table(low_v_s_maxs_maxc, 3, 18, 3, low(maxc)).
table(low_v_mins_maxs_maxc, 3, 18, 3, low(maxc)).
table(low_c_minc_s_maxc, 3, 18, 3, low(maxc)).
table(low_c_minc_mins_maxc, 3, 12, 3, low(maxc)).
table(low_c_minc_maxs_maxc, 3, 18, 3, low(maxc)).
table(low_c_s_mins_maxc, 3, 18, 3, low(maxc)).
table(low_c_s_maxs_maxc, 3, 18, 3, low(maxc)).
table(low_c_mins_maxs_maxc, 3, 18, 3, low(maxc)).
table(low_minc_s_mins_maxc, 3, 18, 3, low(maxc)).
table(low_minc_s_maxs_maxc, 3, 18, 3, low(maxc)).
table(low_minc_mins_maxs_maxc, 3, 11, 3, low(maxc)).
table(low_s_mins_maxs_maxc, 3, 18, 3, low(maxc)).
table(low_v_c_minc_s_maxc, 3, 18, 4, low(maxc)).
table(low_v_c_minc_mins_maxc, 3, 12, 4, low(maxc)).
table(low_v_c_minc_maxs_maxc, 3, 18, 4, low(maxc)).
table(low_v_c_s_mins_maxc, 3, 18, 4, low(maxc)).
table(low_v_c_s_maxs_maxc, 3, 18, 4, low(maxc)).
table(low_v_c_mins_maxs_maxc, 3, 18, 4, low(maxc)).
table(low_v_minc_s_mins_maxc, 3, 18, 4, low(maxc)).
table(low_v_minc_s_maxs_maxc, 3, 18, 4, low(maxc)).
table(low_v_minc_mins_maxs_maxc, 3, 18, 4, low(maxc)).
table(low_v_s_mins_maxs_maxc, 3, 18, 4, low(maxc)).
table(low_c_minc_s_mins_maxc, 3, 18, 4, low(maxc)).
table(low_c_minc_s_maxs_maxc, 3, 18, 4, low(maxc)).
table(low_c_minc_mins_maxs_maxc, 3, 18, 4, low(maxc)).
table(low_c_s_mins_maxs_maxc, 3, 18, 4, low(maxc)).
table(low_minc_s_mins_maxs_maxc, 3, 18, 4, low(maxc)).
table(low_v_c_minc_s_mins_maxc, 3, 18, 5, low(maxc)).
table(low_v_c_minc_s_maxs_maxc, 3, 18, 5, low(maxc)).
table(low_v_c_minc_mins_maxs_maxc, 3, 18, 5, low(maxc)).
table(low_v_c_s_mins_maxs_maxc, 3, 18, 5, low(maxc)).
table(low_v_minc_s_mins_maxs_maxc, 3, 18, 5, low(maxc)).
table(low_c_minc_s_mins_maxs_maxc, 3, 18, 5, low(maxc)).
table(low_v_c_minc_s_mins_maxs_maxc, 3, 18, 6, low(maxc)).
table(up_v_maxc, 3, 11, 1, up(maxc)).
table(up_v_c_maxc, 3, 11, 2, up(maxc)).
table(up_v_minc_maxc, 3, 11, 2, up(maxc)).
table(up_v_s_maxc, 3, 18, 2, up(maxc)).
table(up_v_mins_maxc, 3, 12, 2, up(maxc)).
table(up_v_maxs_maxc, 3, 18, 2, up(maxc)).
table(up_v_c_minc_maxc, 3, 11, 3, up(maxc)).
table(up_v_c_s_maxc, 3, 18, 3, up(maxc)).
table(up_v_c_mins_maxc, 3, 12, 3, up(maxc)).
table(up_v_c_maxs_maxc, 3, 18, 3, up(maxc)).
table(up_v_minc_s_maxc, 3, 18, 3, up(maxc)).
table(up_v_minc_mins_maxc, 3, 12, 3, up(maxc)).
table(up_v_minc_maxs_maxc, 3, 18, 3, up(maxc)).
table(up_v_s_mins_maxc, 3, 18, 3, up(maxc)).
table(up_v_s_maxs_maxc, 3, 18, 3, up(maxc)).
table(up_v_mins_maxs_maxc, 3, 18, 3, up(maxc)).
table(up_v_c_minc_s_maxc, 3, 18, 4, up(maxc)).
table(up_v_c_minc_mins_maxc, 3, 12, 4, up(maxc)).
table(up_v_c_minc_maxs_maxc, 3, 18, 4, up(maxc)).
table(up_v_c_s_mins_maxc, 3, 18, 4, up(maxc)).
table(up_v_c_s_maxs_maxc, 3, 18, 4, up(maxc)).
table(up_v_c_mins_maxs_maxc, 3, 18, 4, up(maxc)).
table(up_v_minc_s_mins_maxc, 3, 18, 4, up(maxc)).
table(up_v_minc_s_maxs_maxc, 3, 18, 4, up(maxc)).
table(up_v_minc_mins_maxs_maxc, 3, 18, 4, up(maxc)).
table(up_v_s_mins_maxs_maxc, 3, 18, 4, up(maxc)).
table(up_v_c_minc_s_mins_maxc, 3, 18, 5, up(maxc)).
table(up_v_c_minc_s_maxs_maxc, 3, 18, 5, up(maxc)).
table(up_v_c_minc_mins_maxs_maxc, 3, 18, 5, up(maxc)).
table(up_v_c_s_mins_maxs_maxc, 3, 18, 5, up(maxc)).
table(up_v_minc_s_mins_maxs_maxc, 3, 18, 5, up(maxc)).
table(up_v_c_minc_s_mins_maxs_maxc, 3, 18, 6, up(maxc)).
table(low_v_s, 3, 18, 1, low(s)).
table(low_c_s, 3, 5, 1, low(s)).
table(low_minc_s, 3, 18, 1, low(s)).
table(low_maxc_s, 3, 18, 1, low(s)).
table(low_mins_s, 3, 18, 1, low(s)).
table(low_maxs_s, 3, 18, 1, low(s)).
table(low_v_c_s, 3, 18, 2, low(s)).
table(low_v_minc_s, 3, 18, 2, low(s)).
table(low_v_maxc_s, 3, 18, 2, low(s)).
table(low_v_mins_s, 3, 9, 2, low(s)).
table(low_v_maxs_s, 3, 9, 2, low(s)).
table(low_c_minc_s, 3, 7, 2, low(s)).
table(low_c_maxc_s, 3, 18, 2, low(s)).
table(low_c_mins_s, 3, 7, 2, low(s)).
table(low_c_maxs_s, 3, 18, 2, low(s)).
table(low_minc_maxc_s, 3, 18, 2, low(s)).
table(low_minc_mins_s, 3, 18, 2, low(s)).
table(low_minc_maxs_s, 3, 18, 2, low(s)).
table(low_maxc_mins_s, 3, 8, 2, low(s)).
table(low_maxc_maxs_s, 3, 18, 2, low(s)).
table(low_mins_maxs_s, 3, 13, 2, low(s)).
table(low_v_c_minc_s, 3, 18, 3, low(s)).
table(low_v_c_maxc_s, 3, 18, 3, low(s)).
table(low_v_c_mins_s, 3, 18, 3, low(s)).
table(low_v_c_maxs_s, 3, 18, 3, low(s)).
table(low_v_minc_maxc_s, 3, 18, 3, low(s)).
table(low_v_minc_mins_s, 3, 18, 3, low(s)).
table(low_v_minc_maxs_s, 3, 10, 3, low(s)).
table(low_v_maxc_mins_s, 3, 18, 3, low(s)).
table(low_v_maxc_maxs_s, 3, 18, 3, low(s)).
table(low_v_mins_maxs_s, 3, 9, 3, low(s)).
table(low_c_minc_maxc_s, 3, 18, 3, low(s)).
table(low_c_minc_mins_s, 3, 7, 3, low(s)).
table(low_c_minc_maxs_s, 3, 18, 3, low(s)).
table(low_c_maxc_mins_s, 3, 18, 3, low(s)).
table(low_c_maxc_maxs_s, 3, 18, 3, low(s)).
table(low_c_mins_maxs_s, 3, 18, 3, low(s)).
table(low_minc_maxc_mins_s, 3, 18, 3, low(s)).
table(low_minc_maxc_maxs_s, 3, 18, 3, low(s)).
table(low_minc_mins_maxs_s, 3, 18, 3, low(s)).
table(low_maxc_mins_maxs_s, 3, 18, 3, low(s)).
table(low_v_c_minc_maxc_s, 3, 18, 4, low(s)).
table(low_v_c_minc_mins_s, 3, 18, 4, low(s)).
table(low_v_c_minc_maxs_s, 3, 18, 4, low(s)).
table(low_v_c_maxc_mins_s, 3, 18, 4, low(s)).
table(low_v_c_maxc_maxs_s, 3, 18, 4, low(s)).
table(low_v_c_mins_maxs_s, 3, 18, 4, low(s)).
table(low_v_minc_maxc_mins_s, 3, 18, 4, low(s)).
table(low_v_minc_maxc_maxs_s, 3, 18, 4, low(s)).
table(low_v_minc_mins_maxs_s, 3, 10, 4, low(s)).
table(low_v_maxc_mins_maxs_s, 3, 18, 4, low(s)).
table(low_c_minc_maxc_mins_s, 3, 18, 4, low(s)).
table(low_c_minc_maxc_maxs_s, 3, 18, 4, low(s)).
table(low_c_minc_mins_maxs_s, 3, 18, 4, low(s)).
table(low_c_maxc_mins_maxs_s, 3, 18, 4, low(s)).
table(low_minc_maxc_mins_maxs_s, 3, 18, 4, low(s)).
table(low_v_c_minc_maxc_mins_s, 3, 18, 5, low(s)).
table(low_v_c_minc_maxc_maxs_s, 3, 18, 5, low(s)).
table(low_v_c_minc_mins_maxs_s, 3, 18, 5, low(s)).
table(low_v_c_maxc_mins_maxs_s, 3, 18, 5, low(s)).
table(low_v_minc_maxc_mins_maxs_s, 3, 18, 5, low(s)).
table(low_c_minc_maxc_mins_maxs_s, 3, 18, 5, low(s)).
table(low_v_c_minc_maxc_mins_maxs_s, 3, 18, 6, low(s)).
table(up_v_s, 3, 9, 1, up(s)).
table(up_v_c_s, 3, 18, 2, up(s)).
table(up_v_minc_s, 3, 10, 2, up(s)).
table(up_v_maxc_s, 3, 18, 2, up(s)).
table(up_v_mins_s, 3, 9, 2, up(s)).
table(up_v_maxs_s, 3, 9, 2, up(s)).
table(up_c_maxc_s, 3, 18, 2, up(s)).
table(up_v_c_minc_s, 3, 18, 3, up(s)).
table(up_v_c_maxc_s, 3, 18, 3, up(s)).
table(up_v_c_mins_s, 3, 18, 3, up(s)).
table(up_v_c_maxs_s, 3, 18, 3, up(s)).
table(up_v_minc_maxc_s, 3, 18, 3, up(s)).
table(up_v_minc_mins_s, 3, 10, 3, up(s)).
table(up_v_minc_maxs_s, 3, 10, 3, up(s)).
table(up_v_maxc_mins_s, 3, 18, 3, up(s)).
table(up_v_maxc_maxs_s, 3, 18, 3, up(s)).
table(up_v_mins_maxs_s, 3, 9, 3, up(s)).
table(up_c_minc_maxc_s, 3, 18, 3, up(s)).
table(up_c_maxc_mins_s, 3, 18, 3, up(s)).
table(up_c_maxc_maxs_s, 3, 18, 3, up(s)).
table(up_v_c_minc_maxc_s, 3, 18, 4, up(s)).
table(up_v_c_minc_mins_s, 3, 18, 4, up(s)).
table(up_v_c_minc_maxs_s, 3, 18, 4, up(s)).
table(up_v_c_maxc_mins_s, 3, 18, 4, up(s)).
table(up_v_c_maxc_maxs_s, 3, 18, 4, up(s)).
table(up_v_c_mins_maxs_s, 3, 18, 4, up(s)).
table(up_v_minc_maxc_mins_s, 3, 18, 4, up(s)).
table(up_v_minc_maxc_maxs_s, 3, 18, 4, up(s)).
table(up_v_minc_mins_maxs_s, 3, 10, 4, up(s)).
table(up_v_maxc_mins_maxs_s, 3, 18, 4, up(s)).
table(up_c_minc_maxc_mins_s, 3, 18, 4, up(s)).
table(up_c_minc_maxc_maxs_s, 3, 18, 4, up(s)).
table(up_c_maxc_mins_maxs_s, 3, 18, 4, up(s)).
table(up_v_c_minc_maxc_mins_s, 3, 18, 5, up(s)).
table(up_v_c_minc_maxc_maxs_s, 3, 18, 5, up(s)).
table(up_v_c_minc_mins_maxs_s, 3, 18, 5, up(s)).
table(up_v_c_maxc_mins_maxs_s, 3, 18, 5, up(s)).
table(up_v_minc_maxc_mins_maxs_s, 3, 18, 5, up(s)).
table(up_c_minc_maxc_mins_maxs_s, 3, 18, 5, up(s)).
table(up_v_c_minc_maxc_mins_maxs_s, 3, 18, 6, up(s)).
table(low_v_mins, 3, 4, 1, low(mins)).
table(low_c_mins, 3, 4, 1, low(mins)).
table(low_minc_mins, 3, 4, 1, low(mins)).
table(low_maxc_mins, 3, 5, 1, low(mins)).
table(low_s_mins, 3, 4, 1, low(mins)).
table(low_maxs_mins, 3, 5, 1, low(mins)).
table(low_v_c_mins, 3, 12, 2, low(mins)).
table(low_v_minc_mins, 3, 5, 2, low(mins)).
table(low_v_maxc_mins, 3, 12, 2, low(mins)).
table(low_v_s_mins, 3, 9, 2, low(mins)).
table(low_v_maxs_mins, 3, 9, 2, low(mins)).
table(low_c_minc_mins, 3, 5, 2, low(mins)).
table(low_c_maxc_mins, 3, 12, 2, low(mins)).
table(low_c_s_mins, 3, 5, 2, low(mins)).
table(low_c_maxs_mins, 3, 6, 2, low(mins)).
table(low_minc_maxc_mins, 3, 8, 2, low(mins)).
table(low_minc_s_mins, 3, 5, 2, low(mins)).
table(low_minc_maxs_mins, 3, 6, 2, low(mins)).
table(low_maxc_s_mins, 3, 8, 2, low(mins)).
table(low_maxc_maxs_mins, 3, 8, 2, low(mins)).
table(low_s_maxs_mins, 3, 9, 2, low(mins)).
table(low_v_c_minc_mins, 3, 12, 3, low(mins)).
table(low_v_c_maxc_mins, 3, 12, 3, low(mins)).
table(low_v_c_s_mins, 3, 18, 3, low(mins)).
table(low_v_c_maxs_mins, 3, 18, 3, low(mins)).
table(low_v_minc_maxc_mins, 3, 12, 3, low(mins)).
table(low_v_minc_s_mins, 3, 10, 3, low(mins)).
table(low_v_minc_maxs_mins, 3, 10, 3, low(mins)).
table(low_v_maxc_s_mins, 3, 18, 3, low(mins)).
table(low_v_maxc_maxs_mins, 3, 18, 3, low(mins)).
table(low_v_s_maxs_mins, 3, 9, 3, low(mins)).
table(low_c_minc_maxc_mins, 3, 12, 3, low(mins)).
table(low_c_minc_s_mins, 3, 7, 3, low(mins)).
table(low_c_minc_maxs_mins, 3, 7, 3, low(mins)).
table(low_c_maxc_s_mins, 3, 18, 3, low(mins)).
table(low_c_maxc_maxs_mins, 3, 18, 3, low(mins)).
table(low_c_s_maxs_mins, 3, 18, 3, low(mins)).
table(low_minc_maxc_s_mins, 3, 18, 3, low(mins)).
table(low_minc_maxc_maxs_mins, 3, 11, 3, low(mins)).
table(low_minc_s_maxs_mins, 3, 10, 3, low(mins)).
table(low_maxc_s_maxs_mins, 3, 18, 3, low(mins)).
table(low_v_c_minc_maxc_mins, 3, 12, 4, low(mins)).
table(low_v_c_minc_s_mins, 3, 18, 4, low(mins)).
table(low_v_c_minc_maxs_mins, 3, 18, 4, low(mins)).
table(low_v_c_maxc_s_mins, 3, 18, 4, low(mins)).
table(low_v_c_maxc_maxs_mins, 3, 18, 4, low(mins)).
table(low_v_c_s_maxs_mins, 3, 18, 4, low(mins)).
table(low_v_minc_maxc_s_mins, 3, 18, 4, low(mins)).
table(low_v_minc_maxc_maxs_mins, 3, 18, 4, low(mins)).
table(low_v_minc_s_maxs_mins, 3, 10, 4, low(mins)).
table(low_v_maxc_s_maxs_mins, 3, 18, 4, low(mins)).
table(low_c_minc_maxc_s_mins, 3, 18, 4, low(mins)).
table(low_c_minc_maxc_maxs_mins, 3, 18, 4, low(mins)).
table(low_c_minc_s_maxs_mins, 3, 18, 4, low(mins)).
table(low_c_maxc_s_maxs_mins, 3, 18, 4, low(mins)).
table(low_minc_maxc_s_maxs_mins, 3, 18, 4, low(mins)).
table(low_v_c_minc_maxc_s_mins, 3, 18, 5, low(mins)).
table(low_v_c_minc_maxc_maxs_mins, 3, 18, 5, low(mins)).
table(low_v_c_minc_s_maxs_mins, 3, 18, 5, low(mins)).
table(low_v_c_maxc_s_maxs_mins, 3, 18, 5, low(mins)).
table(low_v_minc_maxc_s_maxs_mins, 3, 18, 5, low(mins)).
table(low_c_minc_maxc_s_maxs_mins, 3, 18, 5, low(mins)).
table(low_v_c_minc_maxc_s_maxs_mins, 3, 18, 6, low(mins)).
table(up_v_mins, 3, 18, 1, up(mins)).
table(up_maxc_mins, 3, 11, 1, up(mins)).
table(up_maxs_mins, 3, 5, 1, up(mins)).
table(up_v_c_mins, 3, 12, 2, up(mins)).
table(up_v_minc_mins, 3, 5, 2, up(mins)).
table(up_v_maxc_mins, 3, 12, 2, up(mins)).
table(up_v_s_mins, 3, 9, 2, up(mins)).
table(up_v_maxs_mins, 3, 9, 2, up(mins)).
table(up_c_maxc_mins, 3, 12, 2, up(mins)).
table(up_c_maxs_mins, 3, 6, 2, up(mins)).
table(up_minc_maxc_mins, 3, 8, 2, up(mins)).
table(up_minc_maxs_mins, 3, 6, 2, up(mins)).
table(up_maxc_s_mins, 3, 8, 2, up(mins)).
table(up_maxc_maxs_mins, 3, 8, 2, up(mins)).
table(up_s_maxs_mins, 3, 9, 2, up(mins)).
table(up_v_c_minc_mins, 3, 12, 3, up(mins)).
table(up_v_c_maxc_mins, 3, 12, 3, up(mins)).
table(up_v_c_s_mins, 3, 18, 3, up(mins)).
table(up_v_c_maxs_mins, 3, 18, 3, up(mins)).
table(up_v_minc_maxc_mins, 3, 12, 3, up(mins)).
table(up_v_minc_s_mins, 3, 10, 3, up(mins)).
table(up_v_minc_maxs_mins, 3, 10, 3, up(mins)).
table(up_v_maxc_s_mins, 3, 18, 3, up(mins)).
table(up_v_maxc_maxs_mins, 3, 18, 3, up(mins)).
table(up_v_s_maxs_mins, 3, 9, 3, up(mins)).
table(up_c_minc_maxc_mins, 3, 12, 3, up(mins)).
table(up_c_minc_maxs_mins, 3, 7, 3, up(mins)).
table(up_c_maxc_s_mins, 3, 18, 3, up(mins)).
table(up_c_maxc_maxs_mins, 3, 18, 3, up(mins)).
table(up_c_s_maxs_mins, 3, 18, 3, up(mins)).
table(up_minc_maxc_s_mins, 3, 18, 3, up(mins)).
table(up_minc_maxc_maxs_mins, 3, 11, 3, up(mins)).
table(up_minc_s_maxs_mins, 3, 10, 3, up(mins)).
table(up_maxc_s_maxs_mins, 3, 18, 3, up(mins)).
table(up_v_c_minc_maxc_mins, 3, 12, 4, up(mins)).
table(up_v_c_minc_s_mins, 3, 18, 4, up(mins)).
table(up_v_c_minc_maxs_mins, 3, 18, 4, up(mins)).
table(up_v_c_maxc_s_mins, 3, 18, 4, up(mins)).
table(up_v_c_maxc_maxs_mins, 3, 18, 4, up(mins)).
table(up_v_c_s_maxs_mins, 3, 18, 4, up(mins)).
table(up_v_minc_maxc_s_mins, 3, 18, 4, up(mins)).
table(up_v_minc_maxc_maxs_mins, 3, 18, 4, up(mins)).
table(up_v_minc_s_maxs_mins, 3, 10, 4, up(mins)).
table(up_v_maxc_s_maxs_mins, 3, 18, 4, up(mins)).
table(up_c_minc_maxc_s_mins, 3, 18, 4, up(mins)).
table(up_c_minc_maxc_maxs_mins, 3, 18, 4, up(mins)).
table(up_c_minc_s_maxs_mins, 3, 18, 4, up(mins)).
table(up_c_maxc_s_maxs_mins, 3, 18, 4, up(mins)).
table(up_minc_maxc_s_maxs_mins, 3, 18, 4, up(mins)).
table(up_v_c_minc_maxc_s_mins, 3, 18, 5, up(mins)).
table(up_v_c_minc_maxc_maxs_mins, 3, 18, 5, up(mins)).
table(up_v_c_minc_s_maxs_mins, 3, 18, 5, up(mins)).
table(up_v_c_maxc_s_maxs_mins, 3, 18, 5, up(mins)).
table(up_v_minc_maxc_s_maxs_mins, 3, 18, 5, up(mins)).
table(up_c_minc_maxc_s_maxs_mins, 3, 18, 5, up(mins)).
table(up_v_c_minc_maxc_s_maxs_mins, 3, 18, 6, up(mins)).
table(low_v_maxs, 3, 9, 1, low(maxs)).
table(low_c_maxs, 3, 6, 1, low(maxs)).
table(low_minc_maxs, 3, 6, 1, low(maxs)).
table(low_maxc_maxs, 3, 8, 1, low(maxs)).
table(low_s_maxs, 3, 9, 1, low(maxs)).
table(low_mins_maxs, 3, 5, 1, low(maxs)).
table(low_v_c_maxs, 3, 18, 2, low(maxs)).
table(low_v_minc_maxs, 3, 10, 2, low(maxs)).
table(low_v_maxc_maxs, 3, 18, 2, low(maxs)).
table(low_v_s_maxs, 3, 9, 2, low(maxs)).
table(low_v_mins_maxs, 3, 9, 2, low(maxs)).
table(low_c_minc_maxs, 3, 7, 2, low(maxs)).
table(low_c_maxc_maxs, 3, 18, 2, low(maxs)).
table(low_c_s_maxs, 3, 18, 2, low(maxs)).
table(low_c_mins_maxs, 3, 6, 2, low(maxs)).
table(low_minc_maxc_maxs, 3, 11, 2, low(maxs)).
table(low_minc_s_maxs, 3, 10, 2, low(maxs)).
table(low_minc_mins_maxs, 3, 6, 2, low(maxs)).
table(low_maxc_s_maxs, 3, 18, 2, low(maxs)).
table(low_maxc_mins_maxs, 3, 8, 2, low(maxs)).
table(low_s_mins_maxs, 3, 9, 2, low(maxs)).
table(low_v_c_minc_maxs, 3, 18, 3, low(maxs)).
table(low_v_c_maxc_maxs, 3, 18, 3, low(maxs)).
table(low_v_c_s_maxs, 3, 18, 3, low(maxs)).
table(low_v_c_mins_maxs, 3, 18, 3, low(maxs)).
table(low_v_minc_maxc_maxs, 3, 18, 3, low(maxs)).
table(low_v_minc_s_maxs, 3, 10, 3, low(maxs)).
table(low_v_minc_mins_maxs, 3, 10, 3, low(maxs)).
table(low_v_maxc_s_maxs, 3, 18, 3, low(maxs)).
table(low_v_maxc_mins_maxs, 3, 18, 3, low(maxs)).
table(low_v_s_mins_maxs, 3, 9, 3, low(maxs)).
table(low_c_minc_maxc_maxs, 3, 18, 3, low(maxs)).
table(low_c_minc_s_maxs, 3, 18, 3, low(maxs)).
table(low_c_minc_mins_maxs, 3, 7, 3, low(maxs)).
table(low_c_maxc_s_maxs, 3, 18, 3, low(maxs)).
table(low_c_maxc_mins_maxs, 3, 18, 3, low(maxs)).
table(low_c_s_mins_maxs, 3, 18, 3, low(maxs)).
table(low_minc_maxc_s_maxs, 3, 18, 3, low(maxs)).
table(low_minc_maxc_mins_maxs, 3, 11, 3, low(maxs)).
table(low_minc_s_mins_maxs, 3, 10, 3, low(maxs)).
table(low_maxc_s_mins_maxs, 3, 18, 3, low(maxs)).
table(low_v_c_minc_maxc_maxs, 3, 18, 4, low(maxs)).
table(low_v_c_minc_s_maxs, 3, 18, 4, low(maxs)).
table(low_v_c_minc_mins_maxs, 3, 18, 4, low(maxs)).
table(low_v_c_maxc_s_maxs, 3, 18, 4, low(maxs)).
table(low_v_c_maxc_mins_maxs, 3, 18, 4, low(maxs)).
table(low_v_c_s_mins_maxs, 3, 18, 4, low(maxs)).
table(low_v_minc_maxc_s_maxs, 3, 18, 4, low(maxs)).
table(low_v_minc_maxc_mins_maxs, 3, 18, 4, low(maxs)).
table(low_v_minc_s_mins_maxs, 3, 10, 4, low(maxs)).
table(low_v_maxc_s_mins_maxs, 3, 18, 4, low(maxs)).
table(low_c_minc_maxc_s_maxs, 3, 18, 4, low(maxs)).
table(low_c_minc_maxc_mins_maxs, 3, 18, 4, low(maxs)).
table(low_c_minc_s_mins_maxs, 3, 18, 4, low(maxs)).
table(low_c_maxc_s_mins_maxs, 3, 18, 4, low(maxs)).
table(low_minc_maxc_s_mins_maxs, 3, 18, 4, low(maxs)).
table(low_v_c_minc_maxc_s_maxs, 3, 18, 5, low(maxs)).
table(low_v_c_minc_maxc_mins_maxs, 3, 18, 5, low(maxs)).
table(low_v_c_minc_s_mins_maxs, 3, 18, 5, low(maxs)).
table(low_v_c_maxc_s_mins_maxs, 3, 18, 5, low(maxs)).
table(low_v_minc_maxc_s_mins_maxs, 3, 18, 5, low(maxs)).
table(low_c_minc_maxc_s_mins_maxs, 3, 18, 5, low(maxs)).
table(low_v_c_minc_maxc_s_mins_maxs, 3, 18, 6, low(maxs)).
table(up_v_maxs, 3, 18, 1, up(maxs)).
table(up_maxc_maxs, 3, 7, 1, up(maxs)).
table(up_v_c_maxs, 3, 18, 2, up(maxs)).
table(up_v_minc_maxs, 3, 18, 2, up(maxs)).
table(up_v_maxc_maxs, 3, 18, 2, up(maxs)).
table(up_v_s_maxs, 3, 9, 2, up(maxs)).
table(up_v_mins_maxs, 3, 9, 2, up(maxs)).
table(up_c_maxc_maxs, 3, 18, 2, up(maxs)).
table(up_minc_maxc_maxs, 3, 11, 2, up(maxs)).
table(up_maxc_s_maxs, 3, 18, 2, up(maxs)).
table(up_maxc_mins_maxs, 3, 11, 2, up(maxs)).
table(up_v_c_minc_maxs, 3, 18, 3, up(maxs)).
table(up_v_c_maxc_maxs, 3, 18, 3, up(maxs)).
table(up_v_c_s_maxs, 3, 18, 3, up(maxs)).
table(up_v_c_mins_maxs, 3, 18, 3, up(maxs)).
table(up_v_minc_maxc_maxs, 3, 18, 3, up(maxs)).
table(up_v_minc_s_maxs, 3, 10, 3, up(maxs)).
table(up_v_minc_mins_maxs, 3, 18, 3, up(maxs)).
table(up_v_maxc_s_maxs, 3, 18, 3, up(maxs)).
table(up_v_maxc_mins_maxs, 3, 18, 3, up(maxs)).
table(up_v_s_mins_maxs, 3, 9, 3, up(maxs)).
table(up_c_minc_maxc_maxs, 3, 18, 3, up(maxs)).
table(up_c_maxc_s_maxs, 3, 18, 3, up(maxs)).
table(up_c_maxc_mins_maxs, 3, 18, 3, up(maxs)).
table(up_minc_maxc_s_maxs, 3, 18, 3, up(maxs)).
table(up_minc_maxc_mins_maxs, 3, 11, 3, up(maxs)).
table(up_maxc_s_mins_maxs, 3, 18, 3, up(maxs)).
table(up_v_c_minc_maxc_maxs, 3, 18, 4, up(maxs)).
table(up_v_c_minc_s_maxs, 3, 18, 4, up(maxs)).
table(up_v_c_minc_mins_maxs, 3, 18, 4, up(maxs)).
table(up_v_c_maxc_s_maxs, 3, 18, 4, up(maxs)).
table(up_v_c_maxc_mins_maxs, 3, 18, 4, up(maxs)).
table(up_v_c_s_mins_maxs, 3, 18, 4, up(maxs)).
table(up_v_minc_maxc_s_maxs, 3, 18, 4, up(maxs)).
table(up_v_minc_maxc_mins_maxs, 3, 18, 4, up(maxs)).
table(up_v_minc_s_mins_maxs, 3, 10, 4, up(maxs)).
table(up_v_maxc_s_mins_maxs, 3, 18, 4, up(maxs)).
table(up_c_minc_maxc_s_maxs, 3, 18, 4, up(maxs)).
table(up_c_minc_maxc_mins_maxs, 3, 18, 4, up(maxs)).
table(up_c_maxc_s_mins_maxs, 3, 18, 4, up(maxs)).
table(up_minc_maxc_s_mins_maxs, 3, 18, 4, up(maxs)).
table(up_v_c_minc_maxc_s_maxs, 3, 18, 5, up(maxs)).
table(up_v_c_minc_maxc_mins_maxs, 3, 18, 5, up(maxs)).
table(up_v_c_minc_s_mins_maxs, 3, 18, 5, up(maxs)).
table(up_v_c_maxc_s_mins_maxs, 3, 18, 5, up(maxs)).
table(up_v_minc_maxc_s_mins_maxs, 3, 18, 5, up(maxs)).
table(up_c_minc_maxc_s_mins_maxs, 3, 18, 5, up(maxs)).
table(up_v_c_minc_maxc_s_mins_maxs, 3, 18, 6, up(maxs)).
table(low_v_mina, 3, 14, 1, low(mina)).
table(low_c_mina, 3, 7, 1, low(mina)).
table(low_minc_mina, 3, 19, 1, low(mina)).
table(low_maxc_mina, 3, 19, 1, low(mina)).
table(low_s_mina, 3, 14, 1, low(mina)).
table(low_mins_mina, 3, 11, 1, low(mina)).
table(low_maxs_mina, 3, 11, 1, low(mina)).
table(low_v_c_mina, 3, 19, 2, low(mina)).
table(low_v_minc_mina, 3, 19, 2, low(mina)).
table(low_v_maxc_mina, 3, 19, 2, low(mina)).
table(low_v_s_mina, 3, 14, 2, low(mina)).
table(low_v_mins_mina, 3, 14, 2, low(mina)).
table(low_v_maxs_mina, 3, 14, 2, low(mina)).
table(low_c_minc_mina, 3, 8, 2, low(mina)).
table(low_c_maxc_mina, 3, 19, 2, low(mina)).
table(low_c_s_mina, 3, 19, 2, low(mina)).
table(low_c_mins_mina, 3, 7, 2, low(mina)).
table(low_c_maxs_mina, 3, 7, 2, low(mina)).
table(low_minc_maxc_mina, 3, 19, 2, low(mina)).
table(low_minc_s_mina, 3, 19, 2, low(mina)).
table(low_minc_mins_mina, 3, 19, 2, low(mina)).
table(low_minc_maxs_mina, 3, 19, 2, low(mina)).
table(low_maxc_s_mina, 3, 19, 2, low(mina)).
table(low_maxc_mins_mina, 3, 19, 2, low(mina)).
table(low_maxc_maxs_mina, 3, 19, 2, low(mina)).
table(low_s_mins_mina, 3, 14, 2, low(mina)).
table(low_s_maxs_mina, 3, 14, 2, low(mina)).
table(low_mins_maxs_mina, 3, 8, 2, low(mina)).
table(low_v_c_minc_mina, 3, 19, 3, low(mina)).
table(low_v_c_maxc_mina, 3, 19, 3, low(mina)).
table(low_v_c_s_mina, 3, 19, 3, low(mina)).
table(low_v_c_mins_mina, 3, 19, 3, low(mina)).
table(low_v_c_maxs_mina, 3, 19, 3, low(mina)).
table(low_v_minc_maxc_mina, 3, 19, 3, low(mina)).
table(low_v_minc_s_mina, 3, 19, 3, low(mina)).
table(low_v_minc_mins_mina, 3, 19, 3, low(mina)).
table(low_v_minc_maxs_mina, 3, 19, 3, low(mina)).
table(low_v_maxc_s_mina, 3, 19, 3, low(mina)).
table(low_v_maxc_mins_mina, 3, 19, 3, low(mina)).
table(low_v_maxc_maxs_mina, 3, 19, 3, low(mina)).
table(low_v_s_mins_mina, 3, 14, 3, low(mina)).
table(low_v_s_maxs_mina, 3, 14, 3, low(mina)).
table(low_v_mins_maxs_mina, 3, 14, 3, low(mina)).
table(low_c_minc_maxc_mina, 3, 19, 3, low(mina)).
table(low_c_minc_s_mina, 3, 19, 3, low(mina)).
table(low_c_minc_mins_mina, 3, 8, 3, low(mina)).
table(low_c_minc_maxs_mina, 3, 8, 3, low(mina)).
table(low_c_maxc_s_mina, 3, 19, 3, low(mina)).
table(low_c_maxc_mins_mina, 3, 19, 3, low(mina)).
table(low_c_maxc_maxs_mina, 3, 19, 3, low(mina)).
table(low_c_s_mins_mina, 3, 19, 3, low(mina)).
table(low_c_s_maxs_mina, 3, 19, 3, low(mina)).
table(low_c_mins_maxs_mina, 3, 7, 3, low(mina)).
table(low_minc_maxc_s_mina, 3, 19, 3, low(mina)).
table(low_minc_maxc_mins_mina, 3, 19, 3, low(mina)).
table(low_minc_maxc_maxs_mina, 3, 19, 3, low(mina)).
table(low_minc_s_mins_mina, 3, 19, 3, low(mina)).
table(low_minc_s_maxs_mina, 3, 19, 3, low(mina)).
table(low_minc_mins_maxs_mina, 3, 19, 3, low(mina)).
table(low_maxc_s_mins_mina, 3, 19, 3, low(mina)).
table(low_maxc_s_maxs_mina, 3, 19, 3, low(mina)).
table(low_maxc_mins_maxs_mina, 3, 19, 3, low(mina)).
table(low_s_mins_maxs_mina, 3, 14, 3, low(mina)).
table(low_v_c_minc_maxc_mina, 3, 19, 4, low(mina)).
table(low_v_c_minc_s_mina, 3, 19, 4, low(mina)).
table(low_v_c_minc_mins_mina, 3, 19, 4, low(mina)).
table(low_v_c_minc_maxs_mina, 3, 19, 4, low(mina)).
table(low_v_c_maxc_s_mina, 3, 19, 4, low(mina)).
table(low_v_c_maxc_mins_mina, 3, 19, 4, low(mina)).
table(low_v_c_maxc_maxs_mina, 3, 19, 4, low(mina)).
table(low_v_c_s_mins_mina, 3, 19, 4, low(mina)).
table(low_v_c_s_maxs_mina, 3, 19, 4, low(mina)).
table(low_v_c_mins_maxs_mina, 3, 19, 4, low(mina)).
table(low_v_minc_maxc_s_mina, 3, 19, 4, low(mina)).
table(low_v_minc_maxc_mins_mina, 3, 19, 4, low(mina)).
table(low_v_minc_maxc_maxs_mina, 3, 19, 4, low(mina)).
table(low_v_minc_s_mins_mina, 3, 19, 4, low(mina)).
table(low_v_minc_s_maxs_mina, 3, 19, 4, low(mina)).
table(low_v_minc_mins_maxs_mina, 3, 19, 4, low(mina)).
table(low_v_maxc_s_mins_mina, 3, 19, 4, low(mina)).
table(low_v_maxc_s_maxs_mina, 3, 19, 4, low(mina)).
table(low_v_maxc_mins_maxs_mina, 3, 19, 4, low(mina)).
table(low_v_s_mins_maxs_mina, 3, 14, 4, low(mina)).
table(low_c_minc_maxc_s_mina, 3, 19, 4, low(mina)).
table(low_c_minc_maxc_mins_mina, 3, 19, 4, low(mina)).
table(low_c_minc_maxc_maxs_mina, 3, 19, 4, low(mina)).
table(low_c_minc_s_mins_mina, 3, 19, 4, low(mina)).
table(low_c_minc_s_maxs_mina, 3, 19, 4, low(mina)).
table(low_c_minc_mins_maxs_mina, 3, 8, 4, low(mina)).
table(low_c_maxc_s_mins_mina, 3, 19, 4, low(mina)).
table(low_c_maxc_s_maxs_mina, 3, 19, 4, low(mina)).
table(low_c_maxc_mins_maxs_mina, 3, 19, 4, low(mina)).
table(low_c_s_mins_maxs_mina, 3, 19, 4, low(mina)).
table(low_minc_maxc_s_mins_mina, 3, 19, 4, low(mina)).
table(low_minc_maxc_s_maxs_mina, 3, 19, 4, low(mina)).
table(low_minc_maxc_mins_maxs_mina, 3, 19, 4, low(mina)).
table(low_minc_s_mins_maxs_mina, 3, 19, 4, low(mina)).
table(low_maxc_s_mins_maxs_mina, 3, 19, 4, low(mina)).
table(low_v_c_minc_maxc_s_mina, 3, 19, 5, low(mina)).
table(low_v_c_minc_maxc_mins_mina, 3, 19, 5, low(mina)).
table(low_v_c_minc_maxc_maxs_mina, 3, 19, 5, low(mina)).
table(low_v_c_minc_s_mins_mina, 3, 19, 5, low(mina)).
table(low_v_c_minc_s_maxs_mina, 3, 19, 5, low(mina)).
table(low_v_c_minc_mins_maxs_mina, 3, 19, 5, low(mina)).
table(low_v_c_maxc_s_mins_mina, 3, 19, 5, low(mina)).
table(low_v_c_maxc_s_maxs_mina, 3, 19, 5, low(mina)).
table(low_v_c_maxc_mins_maxs_mina, 3, 19, 5, low(mina)).
table(low_v_c_s_mins_maxs_mina, 3, 19, 5, low(mina)).
table(low_v_minc_maxc_s_mins_mina, 3, 19, 5, low(mina)).
table(low_v_minc_maxc_s_maxs_mina, 3, 19, 5, low(mina)).
table(low_v_minc_maxc_mins_maxs_mina, 3, 19, 5, low(mina)).
table(low_v_minc_s_mins_maxs_mina, 3, 19, 5, low(mina)).
table(low_v_maxc_s_mins_maxs_mina, 3, 19, 5, low(mina)).
table(low_c_minc_maxc_s_mins_mina, 3, 19, 5, low(mina)).
table(low_c_minc_maxc_s_maxs_mina, 3, 19, 5, low(mina)).
table(low_c_minc_maxc_mins_maxs_mina, 3, 19, 5, low(mina)).
table(low_c_minc_s_mins_maxs_mina, 3, 19, 5, low(mina)).
table(low_c_maxc_s_mins_maxs_mina, 3, 19, 5, low(mina)).
table(low_minc_maxc_s_mins_maxs_mina, 3, 19, 5, low(mina)).
table(low_v_c_minc_maxc_s_mins_mina, 3, 19, 6, low(mina)).
table(low_v_c_minc_maxc_s_maxs_mina, 3, 19, 6, low(mina)).
table(low_v_c_minc_maxc_mins_maxs_mina, 3, 19, 6, low(mina)).
table(low_v_c_minc_s_mins_maxs_mina, 3, 19, 6, low(mina)).
table(low_v_c_maxc_s_mins_maxs_mina, 3, 19, 6, low(mina)).
table(low_v_minc_maxc_s_mins_maxs_mina, 3, 19, 6, low(mina)).
table(low_c_minc_maxc_s_mins_maxs_mina, 3, 19, 6, low(mina)).
table(low_v_c_minc_maxc_s_mins_maxs_mina, 3, 19, 7, low(mina)).
table(up_v_maxa, 3, 19, 1, up(maxa)).
table(up_v_c_maxa, 3, 19, 2, up(maxa)).
table(up_v_minc_maxa, 3, 19, 2, up(maxa)).
table(up_v_maxc_maxa, 3, 19, 2, up(maxa)).
table(up_v_s_maxa, 3, 19, 2, up(maxa)).
table(up_v_mins_maxa, 3, 19, 2, up(maxa)).
table(up_v_maxs_maxa, 3, 19, 2, up(maxa)).
table(up_c_maxc_maxa, 3, 19, 2, up(maxa)).
table(up_maxc_s_maxa, 3, 19, 2, up(maxa)).
table(up_s_maxs_maxa, 3, 19, 2, up(maxa)).
table(up_v_c_minc_maxa, 3, 19, 3, up(maxa)).
table(up_v_c_maxc_maxa, 3, 19, 3, up(maxa)).
table(up_v_c_s_maxa, 3, 19, 3, up(maxa)).
table(up_v_c_mins_maxa, 3, 19, 3, up(maxa)).
table(up_v_c_maxs_maxa, 3, 19, 3, up(maxa)).
table(up_v_minc_maxc_maxa, 3, 19, 3, up(maxa)).
table(up_v_minc_s_maxa, 3, 19, 3, up(maxa)).
table(up_v_minc_mins_maxa, 3, 19, 3, up(maxa)).
table(up_v_minc_maxs_maxa, 3, 19, 3, up(maxa)).
table(up_v_maxc_s_maxa, 3, 19, 3, up(maxa)).
table(up_v_maxc_mins_maxa, 3, 19, 3, up(maxa)).
table(up_v_maxc_maxs_maxa, 3, 19, 3, up(maxa)).
table(up_v_s_mins_maxa, 3, 19, 3, up(maxa)).
table(up_v_s_maxs_maxa, 3, 19, 3, up(maxa)).
table(up_v_mins_maxs_maxa, 3, 19, 3, up(maxa)).
table(up_c_minc_maxc_maxa, 3, 19, 3, up(maxa)).
table(up_c_maxc_s_maxa, 3, 19, 3, up(maxa)).
table(up_c_maxc_mins_maxa, 3, 19, 3, up(maxa)).
table(up_c_maxc_maxs_maxa, 3, 19, 3, up(maxa)).
table(up_c_s_maxs_maxa, 3, 19, 3, up(maxa)).
table(up_minc_maxc_s_maxa, 3, 19, 3, up(maxa)).
table(up_minc_s_maxs_maxa, 3, 19, 3, up(maxa)).
table(up_maxc_s_mins_maxa, 3, 19, 3, up(maxa)).
table(up_maxc_s_maxs_maxa, 3, 19, 3, up(maxa)).
table(up_s_mins_maxs_maxa, 3, 19, 3, up(maxa)).
table(up_v_c_minc_maxc_maxa, 3, 19, 4, up(maxa)).
table(up_v_c_minc_s_maxa, 3, 19, 4, up(maxa)).
table(up_v_c_minc_mins_maxa, 3, 19, 4, up(maxa)).
table(up_v_c_minc_maxs_maxa, 3, 19, 4, up(maxa)).
table(up_v_c_maxc_s_maxa, 3, 19, 4, up(maxa)).
table(up_v_c_maxc_mins_maxa, 3, 19, 4, up(maxa)).
table(up_v_c_maxc_maxs_maxa, 3, 19, 4, up(maxa)).
table(up_v_c_s_mins_maxa, 3, 19, 4, up(maxa)).
table(up_v_c_s_maxs_maxa, 3, 19, 4, up(maxa)).
table(up_v_c_mins_maxs_maxa, 3, 19, 4, up(maxa)).
table(up_v_minc_maxc_s_maxa, 3, 19, 4, up(maxa)).
table(up_v_minc_maxc_mins_maxa, 3, 19, 4, up(maxa)).
table(up_v_minc_maxc_maxs_maxa, 3, 19, 4, up(maxa)).
table(up_v_minc_s_mins_maxa, 3, 19, 4, up(maxa)).
table(up_v_minc_s_maxs_maxa, 3, 19, 4, up(maxa)).
table(up_v_minc_mins_maxs_maxa, 3, 19, 4, up(maxa)).
table(up_v_maxc_s_mins_maxa, 3, 19, 4, up(maxa)).
table(up_v_maxc_s_maxs_maxa, 3, 19, 4, up(maxa)).
table(up_v_maxc_mins_maxs_maxa, 3, 19, 4, up(maxa)).
table(up_v_s_mins_maxs_maxa, 3, 19, 4, up(maxa)).
table(up_c_minc_maxc_s_maxa, 3, 19, 4, up(maxa)).
table(up_c_minc_maxc_mins_maxa, 3, 19, 4, up(maxa)).
table(up_c_minc_maxc_maxs_maxa, 3, 19, 4, up(maxa)).
table(up_c_minc_s_maxs_maxa, 3, 19, 4, up(maxa)).
table(up_c_maxc_s_mins_maxa, 3, 19, 4, up(maxa)).
table(up_c_maxc_s_maxs_maxa, 3, 19, 4, up(maxa)).
table(up_c_maxc_mins_maxs_maxa, 3, 19, 4, up(maxa)).
table(up_c_s_mins_maxs_maxa, 3, 19, 4, up(maxa)).
table(up_minc_maxc_s_mins_maxa, 3, 19, 4, up(maxa)).
table(up_minc_maxc_s_maxs_maxa, 3, 19, 4, up(maxa)).
table(up_minc_s_mins_maxs_maxa, 3, 19, 4, up(maxa)).
table(up_maxc_s_mins_maxs_maxa, 3, 19, 4, up(maxa)).
table(up_v_c_minc_maxc_s_maxa, 3, 19, 5, up(maxa)).
table(up_v_c_minc_maxc_mins_maxa, 3, 19, 5, up(maxa)).
table(up_v_c_minc_maxc_maxs_maxa, 3, 19, 5, up(maxa)).
table(up_v_c_minc_s_mins_maxa, 3, 19, 5, up(maxa)).
table(up_v_c_minc_s_maxs_maxa, 3, 19, 5, up(maxa)).
table(up_v_c_minc_mins_maxs_maxa, 3, 19, 5, up(maxa)).
table(up_v_c_maxc_s_mins_maxa, 3, 19, 5, up(maxa)).
table(up_v_c_maxc_s_maxs_maxa, 3, 19, 5, up(maxa)).
table(up_v_c_maxc_mins_maxs_maxa, 3, 19, 5, up(maxa)).
table(up_v_c_s_mins_maxs_maxa, 3, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_mins_maxa, 3, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_maxs_maxa, 3, 19, 5, up(maxa)).
table(up_v_minc_maxc_mins_maxs_maxa, 3, 19, 5, up(maxa)).
table(up_v_minc_s_mins_maxs_maxa, 3, 19, 5, up(maxa)).
table(up_v_maxc_s_mins_maxs_maxa, 3, 19, 5, up(maxa)).
table(up_c_minc_maxc_s_mins_maxa, 3, 19, 5, up(maxa)).
table(up_c_minc_maxc_s_maxs_maxa, 3, 19, 5, up(maxa)).
table(up_c_minc_maxc_mins_maxs_maxa, 3, 19, 5, up(maxa)).
table(up_c_minc_s_mins_maxs_maxa, 3, 19, 5, up(maxa)).
table(up_c_maxc_s_mins_maxs_maxa, 3, 19, 5, up(maxa)).
table(up_minc_maxc_s_mins_maxs_maxa, 3, 19, 5, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxa, 3, 19, 6, up(maxa)).
table(up_v_c_minc_maxc_s_maxs_maxa, 3, 19, 6, up(maxa)).
table(up_v_c_minc_maxc_mins_maxs_maxa, 3, 19, 6, up(maxa)).
table(up_v_c_minc_s_mins_maxs_maxa, 3, 19, 6, up(maxa)).
table(up_v_c_maxc_s_mins_maxs_maxa, 3, 19, 6, up(maxa)).
table(up_v_minc_maxc_s_mins_maxs_maxa, 3, 19, 6, up(maxa)).
table(up_c_minc_maxc_s_mins_maxs_maxa, 3, 19, 6, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxs_maxa, 3, 19, 7, up(maxa)).
table(low_c_v, 4, 18, 1, low(v)).
table(low_minc_v, 4, 11, 1, low(v)).
table(low_maxc_v, 4, 11, 1, low(v)).
table(low_s_v, 4, 9, 1, low(v)).
table(low_mins_v, 4, 18, 1, low(v)).
table(low_maxs_v, 4, 18, 1, low(v)).
table(low_c_minc_v, 4, 11, 2, low(v)).
table(low_c_maxc_v, 4, 11, 2, low(v)).
table(low_c_s_v, 4, 10, 2, low(v)).
table(low_c_mins_v, 4, 18, 2, low(v)).
table(low_c_maxs_v, 4, 18, 2, low(v)).
table(low_minc_maxc_v, 4, 11, 2, low(v)).
table(low_minc_s_v, 4, 5, 2, low(v)).
table(low_minc_mins_v, 4, 12, 2, low(v)).
table(low_minc_maxs_v, 4, 14, 2, low(v)).
table(low_maxc_s_v, 4, 6, 2, low(v)).
table(low_maxc_mins_v, 4, 12, 2, low(v)).
table(low_maxc_maxs_v, 4, 14, 2, low(v)).
table(low_s_mins_v, 4, 9, 2, low(v)).
table(low_s_maxs_v, 4, 9, 2, low(v)).
table(low_mins_maxs_v, 4, 13, 2, low(v)).
table(low_c_minc_maxc_v, 4, 11, 3, low(v)).
table(low_c_minc_s_v, 4, 13, 3, low(v)).
table(low_c_minc_mins_v, 4, 12, 3, low(v)).
table(low_c_minc_maxs_v, 4, 13, 3, low(v)).
table(low_c_maxc_s_v, 4, 13, 3, low(v)).
table(low_c_maxc_mins_v, 4, 12, 3, low(v)).
table(low_c_maxc_maxs_v, 4, 14, 3, low(v)).
table(low_c_s_mins_v, 4, 10, 3, low(v)).
table(low_c_s_maxs_v, 4, 10, 3, low(v)).
table(low_c_mins_maxs_v, 4, 18, 3, low(v)).
table(low_minc_maxc_s_v, 4, 13, 3, low(v)).
table(low_minc_maxc_mins_v, 4, 12, 3, low(v)).
table(low_minc_maxc_maxs_v, 4, 14, 3, low(v)).
table(low_minc_s_mins_v, 4, 10, 3, low(v)).
table(low_minc_s_maxs_v, 4, 10, 3, low(v)).
table(low_minc_mins_maxs_v, 4, 18, 3, low(v)).
table(low_maxc_s_mins_v, 4, 11, 3, low(v)).
table(low_maxc_s_maxs_v, 4, 11, 3, low(v)).
table(low_maxc_mins_maxs_v, 4, 18, 3, low(v)).
table(low_s_mins_maxs_v, 4, 9, 3, low(v)).
table(low_c_minc_maxc_s_v, 4, 13, 4, low(v)).
table(low_c_minc_maxc_mins_v, 4, 12, 4, low(v)).
table(low_c_minc_maxc_maxs_v, 4, 13, 4, low(v)).
table(low_c_minc_s_mins_v, 4, 18, 4, low(v)).
table(low_c_minc_s_maxs_v, 4, 18, 4, low(v)).
table(low_c_minc_mins_maxs_v, 4, 18, 4, low(v)).
table(low_c_maxc_s_mins_v, 4, 18, 4, low(v)).
table(low_c_maxc_s_maxs_v, 4, 18, 4, low(v)).
table(low_c_maxc_mins_maxs_v, 4, 18, 4, low(v)).
table(low_c_s_mins_maxs_v, 4, 10, 4, low(v)).
table(low_minc_maxc_s_mins_v, 4, 18, 4, low(v)).
table(low_minc_maxc_s_maxs_v, 4, 18, 4, low(v)).
table(low_minc_maxc_mins_maxs_v, 4, 18, 4, low(v)).
table(low_minc_s_mins_maxs_v, 4, 10, 4, low(v)).
table(low_maxc_s_mins_maxs_v, 4, 11, 4, low(v)).
table(low_c_minc_maxc_s_mins_v, 4, 18, 5, low(v)).
table(low_c_minc_maxc_s_maxs_v, 4, 18, 5, low(v)).
table(low_c_minc_maxc_mins_maxs_v, 4, 18, 5, low(v)).
table(low_c_minc_s_mins_maxs_v, 4, 18, 5, low(v)).
table(low_c_maxc_s_mins_maxs_v, 4, 18, 5, low(v)).
table(low_minc_maxc_s_mins_maxs_v, 4, 18, 5, low(v)).
table(low_c_minc_maxc_s_mins_maxs_v, 4, 18, 6, low(v)).
table(up_c_maxc_v, 4, 11, 2, up(v)).
table(up_maxc_s_v, 4, 6, 2, up(v)).
table(up_s_maxs_v, 4, 9, 2, up(v)).
table(up_c_minc_maxc_v, 4, 11, 3, up(v)).
table(up_c_maxc_s_v, 4, 13, 3, up(v)).
table(up_c_maxc_mins_v, 4, 12, 3, up(v)).
table(up_c_maxc_maxs_v, 4, 13, 3, up(v)).
table(up_c_s_maxs_v, 4, 10, 3, up(v)).
table(up_minc_maxc_s_v, 4, 13, 3, up(v)).
table(up_minc_s_maxs_v, 4, 10, 3, up(v)).
table(up_maxc_s_mins_v, 4, 11, 3, up(v)).
table(up_maxc_s_maxs_v, 4, 11, 3, up(v)).
table(up_s_mins_maxs_v, 4, 9, 3, up(v)).
table(up_c_minc_maxc_s_v, 4, 13, 4, up(v)).
table(up_c_minc_maxc_mins_v, 4, 12, 4, up(v)).
table(up_c_minc_maxc_maxs_v, 4, 13, 4, up(v)).
table(up_c_minc_s_maxs_v, 4, 18, 4, up(v)).
table(up_c_maxc_s_mins_v, 4, 18, 4, up(v)).
table(up_c_maxc_s_maxs_v, 4, 18, 4, up(v)).
table(up_c_maxc_mins_maxs_v, 4, 18, 4, up(v)).
table(up_c_s_mins_maxs_v, 4, 10, 4, up(v)).
table(up_minc_maxc_s_mins_v, 4, 18, 4, up(v)).
table(up_minc_maxc_s_maxs_v, 4, 18, 4, up(v)).
table(up_minc_s_mins_maxs_v, 4, 10, 4, up(v)).
table(up_maxc_s_mins_maxs_v, 4, 11, 4, up(v)).
table(up_c_minc_maxc_s_mins_v, 4, 18, 5, up(v)).
table(up_c_minc_maxc_s_maxs_v, 4, 18, 5, up(v)).
table(up_c_minc_maxc_mins_maxs_v, 4, 18, 5, up(v)).
table(up_c_minc_s_mins_maxs_v, 4, 18, 5, up(v)).
table(up_c_maxc_s_mins_maxs_v, 4, 18, 5, up(v)).
table(up_minc_maxc_s_mins_maxs_v, 4, 18, 5, up(v)).
table(up_c_minc_maxc_s_mins_maxs_v, 4, 18, 6, up(v)).
table(low_v_c, 4, 11, 1, low(c)).
table(low_minc_c, 4, 11, 1, low(c)).
table(low_maxc_c, 4, 11, 1, low(c)).
table(low_s_c, 4, 8, 1, low(c)).
table(low_mins_c, 4, 8, 1, low(c)).
table(low_maxs_c, 4, 8, 1, low(c)).
table(low_v_minc_c, 4, 11, 2, low(c)).
table(low_v_maxc_c, 4, 11, 2, low(c)).
table(low_v_s_c, 4, 13, 2, low(c)).
table(low_v_mins_c, 4, 12, 2, low(c)).
table(low_v_maxs_c, 4, 14, 2, low(c)).
table(low_minc_maxc_c, 4, 11, 2, low(c)).
table(low_minc_s_c, 4, 5, 2, low(c)).
table(low_minc_mins_c, 4, 12, 2, low(c)).
table(low_minc_maxs_c, 4, 10, 2, low(c)).
table(low_maxc_s_c, 4, 6, 2, low(c)).
table(low_maxc_mins_c, 4, 12, 2, low(c)).
table(low_maxc_maxs_c, 4, 14, 2, low(c)).
table(low_s_mins_c, 4, 10, 2, low(c)).
table(low_s_maxs_c, 4, 11, 2, low(c)).
table(low_mins_maxs_c, 4, 9, 2, low(c)).
table(low_v_minc_maxc_c, 4, 11, 3, low(c)).
table(low_v_minc_s_c, 4, 13, 3, low(c)).
table(low_v_minc_mins_c, 4, 12, 3, low(c)).
table(low_v_minc_maxs_c, 4, 13, 3, low(c)).
table(low_v_maxc_s_c, 4, 13, 3, low(c)).
table(low_v_maxc_mins_c, 4, 12, 3, low(c)).
table(low_v_maxc_maxs_c, 4, 13, 3, low(c)).
table(low_v_s_mins_c, 4, 18, 3, low(c)).
table(low_v_s_maxs_c, 4, 18, 3, low(c)).
table(low_v_mins_maxs_c, 4, 18, 3, low(c)).
table(low_minc_maxc_s_c, 4, 13, 3, low(c)).
table(low_minc_maxc_mins_c, 4, 12, 3, low(c)).
table(low_minc_maxc_maxs_c, 4, 14, 3, low(c)).
table(low_minc_s_mins_c, 4, 6, 3, low(c)).
table(low_minc_s_maxs_c, 4, 18, 3, low(c)).
table(low_minc_mins_maxs_c, 4, 13, 3, low(c)).
table(low_maxc_s_mins_c, 4, 7, 3, low(c)).
table(low_maxc_s_maxs_c, 4, 7, 3, low(c)).
table(low_maxc_mins_maxs_c, 4, 8, 3, low(c)).
table(low_s_mins_maxs_c, 4, 18, 3, low(c)).
table(low_v_minc_maxc_s_c, 4, 13, 4, low(c)).
table(low_v_minc_maxc_mins_c, 4, 12, 4, low(c)).
table(low_v_minc_maxc_maxs_c, 4, 13, 4, low(c)).
table(low_v_minc_s_mins_c, 4, 18, 4, low(c)).
table(low_v_minc_s_maxs_c, 4, 18, 4, low(c)).
table(low_v_minc_mins_maxs_c, 4, 18, 4, low(c)).
table(low_v_maxc_s_mins_c, 4, 18, 4, low(c)).
table(low_v_maxc_s_maxs_c, 4, 18, 4, low(c)).
table(low_v_maxc_mins_maxs_c, 4, 18, 4, low(c)).
table(low_v_s_mins_maxs_c, 4, 18, 4, low(c)).
table(low_minc_maxc_s_mins_c, 4, 18, 4, low(c)).
table(low_minc_maxc_s_maxs_c, 4, 18, 4, low(c)).
table(low_minc_maxc_mins_maxs_c, 4, 18, 4, low(c)).
table(low_minc_s_mins_maxs_c, 4, 18, 4, low(c)).
table(low_maxc_s_mins_maxs_c, 4, 18, 4, low(c)).
table(low_v_minc_maxc_s_mins_c, 4, 18, 5, low(c)).
table(low_v_minc_maxc_s_maxs_c, 4, 18, 5, low(c)).
table(low_v_minc_maxc_mins_maxs_c, 4, 18, 5, low(c)).
table(low_v_minc_s_mins_maxs_c, 4, 18, 5, low(c)).
table(low_v_maxc_s_mins_maxs_c, 4, 18, 5, low(c)).
table(low_minc_maxc_s_mins_maxs_c, 4, 18, 5, low(c)).
table(low_v_minc_maxc_s_mins_maxs_c, 4, 18, 6, low(c)).
table(up_v_c, 4, 18, 1, up(c)).
table(up_s_c, 4, 5, 1, up(c)).
table(up_v_minc_c, 4, 11, 2, up(c)).
table(up_v_maxc_c, 4, 11, 2, up(c)).
table(up_v_s_c, 4, 6, 2, up(c)).
table(up_v_mins_c, 4, 18, 2, up(c)).
table(up_v_maxs_c, 4, 18, 2, up(c)).
table(up_minc_s_c, 4, 6, 2, up(c)).
table(up_maxc_s_c, 4, 6, 2, up(c)).
table(up_s_mins_c, 4, 7, 2, up(c)).
table(up_s_maxs_c, 4, 7, 2, up(c)).
table(up_v_minc_maxc_c, 4, 11, 3, up(c)).
table(up_v_minc_s_c, 4, 13, 3, up(c)).
table(up_v_minc_mins_c, 4, 12, 3, up(c)).
table(up_v_minc_maxs_c, 4, 13, 3, up(c)).
table(up_v_maxc_s_c, 4, 13, 3, up(c)).
table(up_v_maxc_mins_c, 4, 12, 3, up(c)).
table(up_v_maxc_maxs_c, 4, 14, 3, up(c)).
table(up_v_s_mins_c, 4, 18, 3, up(c)).
table(up_v_s_maxs_c, 4, 18, 3, up(c)).
table(up_v_mins_maxs_c, 4, 18, 3, up(c)).
table(up_minc_maxc_s_c, 4, 13, 3, up(c)).
table(up_minc_s_mins_c, 4, 7, 3, up(c)).
table(up_minc_s_maxs_c, 4, 18, 3, up(c)).
table(up_maxc_s_mins_c, 4, 18, 3, up(c)).
table(up_maxc_s_maxs_c, 4, 7, 3, up(c)).
table(up_s_mins_maxs_c, 4, 18, 3, up(c)).
table(up_v_minc_maxc_s_c, 4, 13, 4, up(c)).
table(up_v_minc_maxc_mins_c, 4, 12, 4, up(c)).
table(up_v_minc_maxc_maxs_c, 4, 13, 4, up(c)).
table(up_v_minc_s_mins_c, 4, 18, 4, up(c)).
table(up_v_minc_s_maxs_c, 4, 18, 4, up(c)).
table(up_v_minc_mins_maxs_c, 4, 18, 4, up(c)).
table(up_v_maxc_s_mins_c, 4, 18, 4, up(c)).
table(up_v_maxc_s_maxs_c, 4, 18, 4, up(c)).
table(up_v_maxc_mins_maxs_c, 4, 18, 4, up(c)).
table(up_v_s_mins_maxs_c, 4, 18, 4, up(c)).
table(up_minc_maxc_s_mins_c, 4, 18, 4, up(c)).
table(up_minc_maxc_s_maxs_c, 4, 18, 4, up(c)).
table(up_minc_s_mins_maxs_c, 4, 18, 4, up(c)).
table(up_maxc_s_mins_maxs_c, 4, 18, 4, up(c)).
table(up_v_minc_maxc_s_mins_c, 4, 18, 5, up(c)).
table(up_v_minc_maxc_s_maxs_c, 4, 18, 5, up(c)).
table(up_v_minc_maxc_mins_maxs_c, 4, 18, 5, up(c)).
table(up_v_minc_s_mins_maxs_c, 4, 18, 5, up(c)).
table(up_v_maxc_s_mins_maxs_c, 4, 18, 5, up(c)).
table(up_minc_maxc_s_mins_maxs_c, 4, 18, 5, up(c)).
table(up_v_minc_maxc_s_mins_maxs_c, 4, 18, 6, up(c)).
table(low_v_minc, 4, 5, 1, low(minc)).
table(low_c_minc, 4, 5, 1, low(minc)).
table(low_maxc_minc, 4, 7, 1, low(minc)).
table(low_s_minc, 4, 5, 1, low(minc)).
table(low_mins_minc, 4, 4, 1, low(minc)).
table(low_maxs_minc, 4, 6, 1, low(minc)).
table(low_v_c_minc, 4, 11, 2, low(minc)).
table(low_v_maxc_minc, 4, 11, 2, low(minc)).
table(low_v_s_minc, 4, 10, 2, low(minc)).
table(low_v_mins_minc, 4, 5, 2, low(minc)).
table(low_v_maxs_minc, 4, 10, 2, low(minc)).
table(low_c_maxc_minc, 4, 11, 2, low(minc)).
table(low_c_s_minc, 4, 6, 2, low(minc)).
table(low_c_mins_minc, 4, 5, 2, low(minc)).
table(low_c_maxs_minc, 4, 7, 2, low(minc)).
table(low_maxc_s_minc, 4, 9, 2, low(minc)).
table(low_maxc_mins_minc, 4, 7, 2, low(minc)).
table(low_maxc_maxs_minc, 4, 9, 2, low(minc)).
table(low_s_mins_minc, 4, 5, 2, low(minc)).
table(low_s_maxs_minc, 4, 10, 2, low(minc)).
table(low_mins_maxs_minc, 4, 5, 2, low(minc)).
table(low_v_c_maxc_minc, 4, 11, 3, low(minc)).
table(low_v_c_s_minc, 4, 13, 3, low(minc)).
table(low_v_c_mins_minc, 4, 12, 3, low(minc)).
table(low_v_c_maxs_minc, 4, 14, 3, low(minc)).
table(low_v_maxc_s_minc, 4, 13, 3, low(minc)).
table(low_v_maxc_mins_minc, 4, 12, 3, low(minc)).
table(low_v_maxc_maxs_minc, 4, 14, 3, low(minc)).
table(low_v_s_mins_minc, 4, 10, 3, low(minc)).
table(low_v_s_maxs_minc, 4, 10, 3, low(minc)).
table(low_v_mins_maxs_minc, 4, 10, 3, low(minc)).
table(low_c_maxc_s_minc, 4, 13, 3, low(minc)).
table(low_c_maxc_mins_minc, 4, 12, 3, low(minc)).
table(low_c_maxc_maxs_minc, 4, 14, 3, low(minc)).
table(low_c_s_mins_minc, 4, 6, 3, low(minc)).
table(low_c_s_maxs_minc, 4, 18, 3, low(minc)).
table(low_c_mins_maxs_minc, 4, 7, 3, low(minc)).
table(low_maxc_s_mins_minc, 4, 10, 3, low(minc)).
table(low_maxc_s_maxs_minc, 4, 18, 3, low(minc)).
table(low_maxc_mins_maxs_minc, 4, 9, 3, low(minc)).
table(low_s_mins_maxs_minc, 4, 10, 3, low(minc)).
table(low_v_c_maxc_s_minc, 4, 13, 4, low(minc)).
table(low_v_c_maxc_mins_minc, 4, 12, 4, low(minc)).
table(low_v_c_maxc_maxs_minc, 4, 13, 4, low(minc)).
table(low_v_c_s_mins_minc, 4, 18, 4, low(minc)).
table(low_v_c_s_maxs_minc, 4, 18, 4, low(minc)).
table(low_v_c_mins_maxs_minc, 4, 18, 4, low(minc)).
table(low_v_maxc_s_mins_minc, 4, 18, 4, low(minc)).
table(low_v_maxc_s_maxs_minc, 4, 18, 4, low(minc)).
table(low_v_maxc_mins_maxs_minc, 4, 18, 4, low(minc)).
table(low_v_s_mins_maxs_minc, 4, 10, 4, low(minc)).
table(low_c_maxc_s_mins_minc, 4, 18, 4, low(minc)).
table(low_c_maxc_s_maxs_minc, 4, 18, 4, low(minc)).
table(low_c_maxc_mins_maxs_minc, 4, 18, 4, low(minc)).
table(low_c_s_mins_maxs_minc, 4, 18, 4, low(minc)).
table(low_maxc_s_mins_maxs_minc, 4, 18, 4, low(minc)).
table(low_v_c_maxc_s_mins_minc, 4, 18, 5, low(minc)).
table(low_v_c_maxc_s_maxs_minc, 4, 18, 5, low(minc)).
table(low_v_c_maxc_mins_maxs_minc, 4, 18, 5, low(minc)).
table(low_v_c_s_mins_maxs_minc, 4, 18, 5, low(minc)).
table(low_v_maxc_s_mins_maxs_minc, 4, 18, 5, low(minc)).
table(low_c_maxc_s_mins_maxs_minc, 4, 18, 5, low(minc)).
table(low_v_c_maxc_s_mins_maxs_minc, 4, 18, 6, low(minc)).
table(up_v_minc, 4, 11, 1, up(minc)).
table(up_maxc_minc, 4, 6, 1, up(minc)).
table(up_v_c_minc, 4, 11, 2, up(minc)).
table(up_v_maxc_minc, 4, 11, 2, up(minc)).
table(up_v_s_minc, 4, 13, 2, up(minc)).
table(up_v_mins_minc, 4, 12, 2, up(minc)).
table(up_v_maxs_minc, 4, 14, 2, up(minc)).
table(up_c_maxc_minc, 4, 11, 2, up(minc)).
table(up_maxc_s_minc, 4, 7, 2, up(minc)).
table(up_maxc_mins_minc, 4, 7, 2, up(minc)).
table(up_maxc_maxs_minc, 4, 7, 2, up(minc)).
table(up_s_maxs_minc, 4, 18, 2, up(minc)).
table(up_v_c_maxc_minc, 4, 11, 3, up(minc)).
table(up_v_c_s_minc, 4, 13, 3, up(minc)).
table(up_v_c_mins_minc, 4, 12, 3, up(minc)).
table(up_v_c_maxs_minc, 4, 13, 3, up(minc)).
table(up_v_maxc_s_minc, 4, 13, 3, up(minc)).
table(up_v_maxc_mins_minc, 4, 12, 3, up(minc)).
table(up_v_maxc_maxs_minc, 4, 13, 3, up(minc)).
table(up_v_s_mins_minc, 4, 18, 3, up(minc)).
table(up_v_s_maxs_minc, 4, 18, 3, up(minc)).
table(up_v_mins_maxs_minc, 4, 18, 3, up(minc)).
table(up_c_maxc_s_minc, 4, 13, 3, up(minc)).
table(up_c_maxc_mins_minc, 4, 12, 3, up(minc)).
table(up_c_maxc_maxs_minc, 4, 13, 3, up(minc)).
table(up_c_s_maxs_minc, 4, 18, 3, up(minc)).
table(up_maxc_s_mins_minc, 4, 18, 3, up(minc)).
table(up_maxc_s_maxs_minc, 4, 18, 3, up(minc)).
table(up_maxc_mins_maxs_minc, 4, 8, 3, up(minc)).
table(up_s_mins_maxs_minc, 4, 18, 3, up(minc)).
table(up_v_c_maxc_s_minc, 4, 13, 4, up(minc)).
table(up_v_c_maxc_mins_minc, 4, 12, 4, up(minc)).
table(up_v_c_maxc_maxs_minc, 4, 13, 4, up(minc)).
table(up_v_c_s_mins_minc, 4, 18, 4, up(minc)).
table(up_v_c_s_maxs_minc, 4, 18, 4, up(minc)).
table(up_v_c_mins_maxs_minc, 4, 18, 4, up(minc)).
table(up_v_maxc_s_mins_minc, 4, 18, 4, up(minc)).
table(up_v_maxc_s_maxs_minc, 4, 18, 4, up(minc)).
table(up_v_maxc_mins_maxs_minc, 4, 18, 4, up(minc)).
table(up_v_s_mins_maxs_minc, 4, 18, 4, up(minc)).
table(up_c_maxc_s_mins_minc, 4, 18, 4, up(minc)).
table(up_c_maxc_s_maxs_minc, 4, 18, 4, up(minc)).
table(up_c_maxc_mins_maxs_minc, 4, 18, 4, up(minc)).
table(up_c_s_mins_maxs_minc, 4, 18, 4, up(minc)).
table(up_maxc_s_mins_maxs_minc, 4, 18, 4, up(minc)).
table(up_v_c_maxc_s_mins_minc, 4, 18, 5, up(minc)).
table(up_v_c_maxc_s_maxs_minc, 4, 18, 5, up(minc)).
table(up_v_c_maxc_mins_maxs_minc, 4, 18, 5, up(minc)).
table(up_v_c_s_mins_maxs_minc, 4, 18, 5, up(minc)).
table(up_v_maxc_s_mins_maxs_minc, 4, 18, 5, up(minc)).
table(up_c_maxc_s_mins_maxs_minc, 4, 18, 5, up(minc)).
table(up_v_c_maxc_s_mins_maxs_minc, 4, 18, 6, up(minc)).
table(low_v_maxc, 4, 18, 1, low(maxc)).
table(low_c_maxc, 4, 18, 1, low(maxc)).
table(low_minc_maxc, 4, 6, 1, low(maxc)).
table(low_s_maxc, 4, 18, 1, low(maxc)).
table(low_mins_maxc, 4, 9, 1, low(maxc)).
table(low_maxs_maxc, 4, 4, 1, low(maxc)).
table(low_v_c_maxc, 4, 11, 2, low(maxc)).
table(low_v_minc_maxc, 4, 11, 2, low(maxc)).
table(low_v_s_maxc, 4, 11, 2, low(maxc)).
table(low_v_mins_maxc, 4, 18, 2, low(maxc)).
table(low_v_maxs_maxc, 4, 7, 2, low(maxc)).
table(low_c_minc_maxc, 4, 11, 2, low(maxc)).
table(low_c_s_maxc, 4, 7, 2, low(maxc)).
table(low_c_mins_maxc, 4, 18, 2, low(maxc)).
table(low_c_maxs_maxc, 4, 5, 2, low(maxc)).
table(low_minc_s_maxc, 4, 7, 2, low(maxc)).
table(low_minc_mins_maxc, 4, 7, 2, low(maxc)).
table(low_minc_maxs_maxc, 4, 7, 2, low(maxc)).
table(low_s_mins_maxc, 4, 18, 2, low(maxc)).
table(low_s_maxs_maxc, 4, 5, 2, low(maxc)).
table(low_mins_maxs_maxc, 4, 5, 2, low(maxc)).
table(low_v_c_minc_maxc, 4, 11, 3, low(maxc)).
table(low_v_c_s_maxc, 4, 13, 3, low(maxc)).
table(low_v_c_mins_maxc, 4, 12, 3, low(maxc)).
table(low_v_c_maxs_maxc, 4, 13, 3, low(maxc)).
table(low_v_minc_s_maxc, 4, 13, 3, low(maxc)).
table(low_v_minc_mins_maxc, 4, 12, 3, low(maxc)).
table(low_v_minc_maxs_maxc, 4, 13, 3, low(maxc)).
table(low_v_s_mins_maxc, 4, 11, 3, low(maxc)).
table(low_v_s_maxs_maxc, 4, 11, 3, low(maxc)).
table(low_v_mins_maxs_maxc, 4, 11, 3, low(maxc)).
table(low_c_minc_s_maxc, 4, 13, 3, low(maxc)).
table(low_c_minc_mins_maxc, 4, 12, 3, low(maxc)).
table(low_c_minc_maxs_maxc, 4, 13, 3, low(maxc)).
table(low_c_s_mins_maxc, 4, 7, 3, low(maxc)).
table(low_c_s_maxs_maxc, 4, 7, 3, low(maxc)).
table(low_c_mins_maxs_maxc, 4, 8, 3, low(maxc)).
table(low_minc_s_mins_maxc, 4, 18, 3, low(maxc)).
table(low_minc_s_maxs_maxc, 4, 18, 3, low(maxc)).
table(low_minc_mins_maxs_maxc, 4, 8, 3, low(maxc)).
table(low_s_mins_maxs_maxc, 4, 11, 3, low(maxc)).
table(low_v_c_minc_s_maxc, 4, 13, 4, low(maxc)).
table(low_v_c_minc_mins_maxc, 4, 12, 4, low(maxc)).
table(low_v_c_minc_maxs_maxc, 4, 13, 4, low(maxc)).
table(low_v_c_s_mins_maxc, 4, 18, 4, low(maxc)).
table(low_v_c_s_maxs_maxc, 4, 18, 4, low(maxc)).
table(low_v_c_mins_maxs_maxc, 4, 18, 4, low(maxc)).
table(low_v_minc_s_mins_maxc, 4, 18, 4, low(maxc)).
table(low_v_minc_s_maxs_maxc, 4, 18, 4, low(maxc)).
table(low_v_minc_mins_maxs_maxc, 4, 18, 4, low(maxc)).
table(low_v_s_mins_maxs_maxc, 4, 11, 4, low(maxc)).
table(low_c_minc_s_mins_maxc, 4, 18, 4, low(maxc)).
table(low_c_minc_s_maxs_maxc, 4, 18, 4, low(maxc)).
table(low_c_minc_mins_maxs_maxc, 4, 18, 4, low(maxc)).
table(low_c_s_mins_maxs_maxc, 4, 18, 4, low(maxc)).
table(low_minc_s_mins_maxs_maxc, 4, 18, 4, low(maxc)).
table(low_v_c_minc_s_mins_maxc, 4, 18, 5, low(maxc)).
table(low_v_c_minc_s_maxs_maxc, 4, 18, 5, low(maxc)).
table(low_v_c_minc_mins_maxs_maxc, 4, 18, 5, low(maxc)).
table(low_v_c_s_mins_maxs_maxc, 4, 18, 5, low(maxc)).
table(low_v_minc_s_mins_maxs_maxc, 4, 18, 5, low(maxc)).
table(low_c_minc_s_mins_maxs_maxc, 4, 18, 5, low(maxc)).
table(low_v_c_minc_s_mins_maxs_maxc, 4, 18, 6, low(maxc)).
table(up_v_maxc, 4, 11, 1, up(maxc)).
table(up_v_c_maxc, 4, 11, 2, up(maxc)).
table(up_v_minc_maxc, 4, 11, 2, up(maxc)).
table(up_v_s_maxc, 4, 13, 2, up(maxc)).
table(up_v_mins_maxc, 4, 12, 2, up(maxc)).
table(up_v_maxs_maxc, 4, 14, 2, up(maxc)).
table(up_v_c_minc_maxc, 4, 11, 3, up(maxc)).
table(up_v_c_s_maxc, 4, 13, 3, up(maxc)).
table(up_v_c_mins_maxc, 4, 12, 3, up(maxc)).
table(up_v_c_maxs_maxc, 4, 14, 3, up(maxc)).
table(up_v_minc_s_maxc, 4, 13, 3, up(maxc)).
table(up_v_minc_mins_maxc, 4, 12, 3, up(maxc)).
table(up_v_minc_maxs_maxc, 4, 13, 3, up(maxc)).
table(up_v_s_mins_maxc, 4, 18, 3, up(maxc)).
table(up_v_s_maxs_maxc, 4, 18, 3, up(maxc)).
table(up_v_mins_maxs_maxc, 4, 18, 3, up(maxc)).
table(up_v_c_minc_s_maxc, 4, 13, 4, up(maxc)).
table(up_v_c_minc_mins_maxc, 4, 12, 4, up(maxc)).
table(up_v_c_minc_maxs_maxc, 4, 13, 4, up(maxc)).
table(up_v_c_s_mins_maxc, 4, 18, 4, up(maxc)).
table(up_v_c_s_maxs_maxc, 4, 18, 4, up(maxc)).
table(up_v_c_mins_maxs_maxc, 4, 18, 4, up(maxc)).
table(up_v_minc_s_mins_maxc, 4, 18, 4, up(maxc)).
table(up_v_minc_s_maxs_maxc, 4, 18, 4, up(maxc)).
table(up_v_minc_mins_maxs_maxc, 4, 18, 4, up(maxc)).
table(up_v_s_mins_maxs_maxc, 4, 18, 4, up(maxc)).
table(up_v_c_minc_s_mins_maxc, 4, 18, 5, up(maxc)).
table(up_v_c_minc_s_maxs_maxc, 4, 18, 5, up(maxc)).
table(up_v_c_minc_mins_maxs_maxc, 4, 18, 5, up(maxc)).
table(up_v_c_s_mins_maxs_maxc, 4, 18, 5, up(maxc)).
table(up_v_minc_s_mins_maxs_maxc, 4, 18, 5, up(maxc)).
table(up_v_c_minc_s_mins_maxs_maxc, 4, 18, 6, up(maxc)).
table(low_v_s, 4, 18, 1, low(s)).
table(low_c_s, 4, 5, 1, low(s)).
table(low_minc_s, 4, 18, 1, low(s)).
table(low_maxc_s, 4, 18, 1, low(s)).
table(low_mins_s, 4, 18, 1, low(s)).
table(low_maxs_s, 4, 18, 1, low(s)).
table(low_v_c_s, 4, 6, 2, low(s)).
table(low_v_minc_s, 4, 18, 2, low(s)).
table(low_v_maxc_s, 4, 18, 2, low(s)).
table(low_v_mins_s, 4, 9, 2, low(s)).
table(low_v_maxs_s, 4, 9, 2, low(s)).
table(low_c_minc_s, 4, 7, 2, low(s)).
table(low_c_maxc_s, 4, 7, 2, low(s)).
table(low_c_mins_s, 4, 7, 2, low(s)).
table(low_c_maxs_s, 4, 7, 2, low(s)).
table(low_minc_maxc_s, 4, 18, 2, low(s)).
table(low_minc_mins_s, 4, 18, 2, low(s)).
table(low_minc_maxs_s, 4, 18, 2, low(s)).
table(low_maxc_mins_s, 4, 8, 2, low(s)).
table(low_maxc_maxs_s, 4, 18, 2, low(s)).
table(low_mins_maxs_s, 4, 13, 2, low(s)).
table(low_v_c_minc_s, 4, 18, 3, low(s)).
table(low_v_c_maxc_s, 4, 18, 3, low(s)).
table(low_v_c_mins_s, 4, 18, 3, low(s)).
table(low_v_c_maxs_s, 4, 10, 3, low(s)).
table(low_v_minc_maxc_s, 4, 18, 3, low(s)).
table(low_v_minc_mins_s, 4, 18, 3, low(s)).
table(low_v_minc_maxs_s, 4, 10, 3, low(s)).
table(low_v_maxc_mins_s, 4, 11, 3, low(s)).
table(low_v_maxc_maxs_s, 4, 11, 3, low(s)).
table(low_v_mins_maxs_s, 4, 9, 3, low(s)).
table(low_c_minc_maxc_s, 4, 18, 3, low(s)).
table(low_c_minc_mins_s, 4, 7, 3, low(s)).
table(low_c_minc_maxs_s, 4, 18, 3, low(s)).
table(low_c_maxc_mins_s, 4, 18, 3, low(s)).
table(low_c_maxc_maxs_s, 4, 7, 3, low(s)).
table(low_c_mins_maxs_s, 4, 18, 3, low(s)).
table(low_minc_maxc_mins_s, 4, 18, 3, low(s)).
table(low_minc_maxc_maxs_s, 4, 18, 3, low(s)).
table(low_minc_mins_maxs_s, 4, 18, 3, low(s)).
table(low_maxc_mins_maxs_s, 4, 18, 3, low(s)).
table(low_v_c_minc_maxc_s, 4, 18, 4, low(s)).
table(low_v_c_minc_mins_s, 4, 18, 4, low(s)).
table(low_v_c_minc_maxs_s, 4, 18, 4, low(s)).
table(low_v_c_maxc_mins_s, 4, 18, 4, low(s)).
table(low_v_c_maxc_maxs_s, 4, 18, 4, low(s)).
table(low_v_c_mins_maxs_s, 4, 10, 4, low(s)).
table(low_v_minc_maxc_mins_s, 4, 18, 4, low(s)).
table(low_v_minc_maxc_maxs_s, 4, 18, 4, low(s)).
table(low_v_minc_mins_maxs_s, 4, 10, 4, low(s)).
table(low_v_maxc_mins_maxs_s, 4, 11, 4, low(s)).
table(low_c_minc_maxc_mins_s, 4, 18, 4, low(s)).
table(low_c_minc_maxc_maxs_s, 4, 18, 4, low(s)).
table(low_c_minc_mins_maxs_s, 4, 18, 4, low(s)).
table(low_c_maxc_mins_maxs_s, 4, 18, 4, low(s)).
table(low_minc_maxc_mins_maxs_s, 4, 18, 4, low(s)).
table(low_v_c_minc_maxc_mins_s, 4, 18, 5, low(s)).
table(low_v_c_minc_maxc_maxs_s, 4, 18, 5, low(s)).
table(low_v_c_minc_mins_maxs_s, 4, 18, 5, low(s)).
table(low_v_c_maxc_mins_maxs_s, 4, 18, 5, low(s)).
table(low_v_minc_maxc_mins_maxs_s, 4, 18, 5, low(s)).
table(low_c_minc_maxc_mins_maxs_s, 4, 18, 5, low(s)).
table(low_v_c_minc_maxc_mins_maxs_s, 4, 18, 6, low(s)).
table(up_v_s, 4, 9, 1, up(s)).
table(up_v_c_s, 4, 10, 2, up(s)).
table(up_v_minc_s, 4, 10, 2, up(s)).
table(up_v_maxc_s, 4, 11, 2, up(s)).
table(up_v_mins_s, 4, 9, 2, up(s)).
table(up_v_maxs_s, 4, 9, 2, up(s)).
table(up_c_maxc_s, 4, 18, 2, up(s)).
table(up_v_c_minc_s, 4, 18, 3, up(s)).
table(up_v_c_maxc_s, 4, 18, 3, up(s)).
table(up_v_c_mins_s, 4, 10, 3, up(s)).
table(up_v_c_maxs_s, 4, 10, 3, up(s)).
table(up_v_minc_maxc_s, 4, 18, 3, up(s)).
table(up_v_minc_mins_s, 4, 10, 3, up(s)).
table(up_v_minc_maxs_s, 4, 10, 3, up(s)).
table(up_v_maxc_mins_s, 4, 11, 3, up(s)).
table(up_v_maxc_maxs_s, 4, 11, 3, up(s)).
table(up_v_mins_maxs_s, 4, 9, 3, up(s)).
table(up_c_minc_maxc_s, 4, 18, 3, up(s)).
table(up_c_maxc_mins_s, 4, 18, 3, up(s)).
table(up_c_maxc_maxs_s, 4, 18, 3, up(s)).
table(up_v_c_minc_maxc_s, 4, 18, 4, up(s)).
table(up_v_c_minc_mins_s, 4, 18, 4, up(s)).
table(up_v_c_minc_maxs_s, 4, 18, 4, up(s)).
table(up_v_c_maxc_mins_s, 4, 18, 4, up(s)).
table(up_v_c_maxc_maxs_s, 4, 18, 4, up(s)).
table(up_v_c_mins_maxs_s, 4, 10, 4, up(s)).
table(up_v_minc_maxc_mins_s, 4, 18, 4, up(s)).
table(up_v_minc_maxc_maxs_s, 4, 18, 4, up(s)).
table(up_v_minc_mins_maxs_s, 4, 10, 4, up(s)).
table(up_v_maxc_mins_maxs_s, 4, 11, 4, up(s)).
table(up_c_minc_maxc_mins_s, 4, 18, 4, up(s)).
table(up_c_minc_maxc_maxs_s, 4, 18, 4, up(s)).
table(up_c_maxc_mins_maxs_s, 4, 18, 4, up(s)).
table(up_v_c_minc_maxc_mins_s, 4, 18, 5, up(s)).
table(up_v_c_minc_maxc_maxs_s, 4, 18, 5, up(s)).
table(up_v_c_minc_mins_maxs_s, 4, 18, 5, up(s)).
table(up_v_c_maxc_mins_maxs_s, 4, 18, 5, up(s)).
table(up_v_minc_maxc_mins_maxs_s, 4, 18, 5, up(s)).
table(up_c_minc_maxc_mins_maxs_s, 4, 18, 5, up(s)).
table(up_v_c_minc_maxc_mins_maxs_s, 4, 18, 6, up(s)).
table(low_v_mins, 4, 4, 1, low(mins)).
table(low_c_mins, 4, 4, 1, low(mins)).
table(low_minc_mins, 4, 4, 1, low(mins)).
table(low_maxc_mins, 4, 4, 1, low(mins)).
table(low_s_mins, 4, 4, 1, low(mins)).
table(low_maxs_mins, 4, 5, 1, low(mins)).
table(low_v_c_mins, 4, 5, 2, low(mins)).
table(low_v_minc_mins, 4, 5, 2, low(mins)).
table(low_v_maxc_mins, 4, 6, 2, low(mins)).
table(low_v_s_mins, 4, 9, 2, low(mins)).
table(low_v_maxs_mins, 4, 9, 2, low(mins)).
table(low_c_minc_mins, 4, 5, 2, low(mins)).
table(low_c_maxc_mins, 4, 5, 2, low(mins)).
table(low_c_s_mins, 4, 5, 2, low(mins)).
table(low_c_maxs_mins, 4, 6, 2, low(mins)).
table(low_minc_maxc_mins, 4, 7, 2, low(mins)).
table(low_minc_s_mins, 4, 5, 2, low(mins)).
table(low_minc_maxs_mins, 4, 6, 2, low(mins)).
table(low_maxc_s_mins, 4, 5, 2, low(mins)).
table(low_maxc_maxs_mins, 4, 6, 2, low(mins)).
table(low_s_maxs_mins, 4, 9, 2, low(mins)).
table(low_v_c_minc_mins, 4, 12, 3, low(mins)).
table(low_v_c_maxc_mins, 4, 12, 3, low(mins)).
table(low_v_c_s_mins, 4, 10, 3, low(mins)).
table(low_v_c_maxs_mins, 4, 10, 3, low(mins)).
table(low_v_minc_maxc_mins, 4, 12, 3, low(mins)).
table(low_v_minc_s_mins, 4, 10, 3, low(mins)).
table(low_v_minc_maxs_mins, 4, 10, 3, low(mins)).
table(low_v_maxc_s_mins, 4, 11, 3, low(mins)).
table(low_v_maxc_maxs_mins, 4, 11, 3, low(mins)).
table(low_v_s_maxs_mins, 4, 9, 3, low(mins)).
table(low_c_minc_maxc_mins, 4, 12, 3, low(mins)).
table(low_c_minc_s_mins, 4, 6, 3, low(mins)).
table(low_c_minc_maxs_mins, 4, 7, 3, low(mins)).
table(low_c_maxc_s_mins, 4, 7, 3, low(mins)).
table(low_c_maxc_maxs_mins, 4, 7, 3, low(mins)).
table(low_c_s_maxs_mins, 4, 10, 3, low(mins)).
table(low_minc_maxc_s_mins, 4, 10, 3, low(mins)).
table(low_minc_maxc_maxs_mins, 4, 9, 3, low(mins)).
table(low_minc_s_maxs_mins, 4, 10, 3, low(mins)).
table(low_maxc_s_maxs_mins, 4, 11, 3, low(mins)).
table(low_v_c_minc_maxc_mins, 4, 12, 4, low(mins)).
table(low_v_c_minc_s_mins, 4, 18, 4, low(mins)).
table(low_v_c_minc_maxs_mins, 4, 18, 4, low(mins)).
table(low_v_c_maxc_s_mins, 4, 18, 4, low(mins)).
table(low_v_c_maxc_maxs_mins, 4, 18, 4, low(mins)).
table(low_v_c_s_maxs_mins, 4, 10, 4, low(mins)).
table(low_v_minc_maxc_s_mins, 4, 18, 4, low(mins)).
table(low_v_minc_maxc_maxs_mins, 4, 18, 4, low(mins)).
table(low_v_minc_s_maxs_mins, 4, 10, 4, low(mins)).
table(low_v_maxc_s_maxs_mins, 4, 11, 4, low(mins)).
table(low_c_minc_maxc_s_mins, 4, 18, 4, low(mins)).
table(low_c_minc_maxc_maxs_mins, 4, 18, 4, low(mins)).
table(low_c_minc_s_maxs_mins, 4, 18, 4, low(mins)).
table(low_c_maxc_s_maxs_mins, 4, 18, 4, low(mins)).
table(low_minc_maxc_s_maxs_mins, 4, 18, 4, low(mins)).
table(low_v_c_minc_maxc_s_mins, 4, 18, 5, low(mins)).
table(low_v_c_minc_maxc_maxs_mins, 4, 18, 5, low(mins)).
table(low_v_c_minc_s_maxs_mins, 4, 18, 5, low(mins)).
table(low_v_c_maxc_s_maxs_mins, 4, 18, 5, low(mins)).
table(low_v_minc_maxc_s_maxs_mins, 4, 18, 5, low(mins)).
table(low_c_minc_maxc_s_maxs_mins, 4, 18, 5, low(mins)).
table(low_v_c_minc_maxc_s_maxs_mins, 4, 18, 6, low(mins)).
table(up_v_mins, 4, 18, 1, up(mins)).
table(up_maxc_mins, 4, 9, 1, up(mins)).
table(up_maxs_mins, 4, 4, 1, up(mins)).
table(up_v_c_mins, 4, 12, 2, up(mins)).
table(up_v_minc_mins, 4, 5, 2, up(mins)).
table(up_v_maxc_mins, 4, 12, 2, up(mins)).
table(up_v_s_mins, 4, 9, 2, up(mins)).
table(up_v_maxs_mins, 4, 9, 2, up(mins)).
table(up_c_maxc_mins, 4, 12, 2, up(mins)).
table(up_c_maxs_mins, 4, 5, 2, up(mins)).
table(up_minc_maxc_mins, 4, 7, 2, up(mins)).
table(up_minc_maxs_mins, 4, 5, 2, up(mins)).
table(up_maxc_s_mins, 4, 5, 2, up(mins)).
table(up_maxc_maxs_mins, 4, 5, 2, up(mins)).
table(up_s_maxs_mins, 4, 9, 2, up(mins)).
table(up_v_c_minc_mins, 4, 12, 3, up(mins)).
table(up_v_c_maxc_mins, 4, 12, 3, up(mins)).
table(up_v_c_s_mins, 4, 10, 3, up(mins)).
table(up_v_c_maxs_mins, 4, 10, 3, up(mins)).
table(up_v_minc_maxc_mins, 4, 12, 3, up(mins)).
table(up_v_minc_s_mins, 4, 10, 3, up(mins)).
table(up_v_minc_maxs_mins, 4, 10, 3, up(mins)).
table(up_v_maxc_s_mins, 4, 11, 3, up(mins)).
table(up_v_maxc_maxs_mins, 4, 11, 3, up(mins)).
table(up_v_s_maxs_mins, 4, 9, 3, up(mins)).
table(up_c_minc_maxc_mins, 4, 12, 3, up(mins)).
table(up_c_minc_maxs_mins, 4, 7, 3, up(mins)).
table(up_c_maxc_s_mins, 4, 7, 3, up(mins)).
table(up_c_maxc_maxs_mins, 4, 7, 3, up(mins)).
table(up_c_s_maxs_mins, 4, 10, 3, up(mins)).
table(up_minc_maxc_s_mins, 4, 10, 3, up(mins)).
table(up_minc_maxc_maxs_mins, 4, 8, 3, up(mins)).
table(up_minc_s_maxs_mins, 4, 10, 3, up(mins)).
table(up_maxc_s_maxs_mins, 4, 11, 3, up(mins)).
table(up_v_c_minc_maxc_mins, 4, 12, 4, up(mins)).
table(up_v_c_minc_s_mins, 4, 18, 4, up(mins)).
table(up_v_c_minc_maxs_mins, 4, 18, 4, up(mins)).
table(up_v_c_maxc_s_mins, 4, 18, 4, up(mins)).
table(up_v_c_maxc_maxs_mins, 4, 18, 4, up(mins)).
table(up_v_c_s_maxs_mins, 4, 10, 4, up(mins)).
table(up_v_minc_maxc_s_mins, 4, 18, 4, up(mins)).
table(up_v_minc_maxc_maxs_mins, 4, 18, 4, up(mins)).
table(up_v_minc_s_maxs_mins, 4, 10, 4, up(mins)).
table(up_v_maxc_s_maxs_mins, 4, 11, 4, up(mins)).
table(up_c_minc_maxc_s_mins, 4, 18, 4, up(mins)).
table(up_c_minc_maxc_maxs_mins, 4, 18, 4, up(mins)).
table(up_c_minc_s_maxs_mins, 4, 18, 4, up(mins)).
table(up_c_maxc_s_maxs_mins, 4, 18, 4, up(mins)).
table(up_minc_maxc_s_maxs_mins, 4, 18, 4, up(mins)).
table(up_v_c_minc_maxc_s_mins, 4, 18, 5, up(mins)).
table(up_v_c_minc_maxc_maxs_mins, 4, 18, 5, up(mins)).
table(up_v_c_minc_s_maxs_mins, 4, 18, 5, up(mins)).
table(up_v_c_maxc_s_maxs_mins, 4, 18, 5, up(mins)).
table(up_v_minc_maxc_s_maxs_mins, 4, 18, 5, up(mins)).
table(up_c_minc_maxc_s_maxs_mins, 4, 18, 5, up(mins)).
table(up_v_c_minc_maxc_s_maxs_mins, 4, 18, 6, up(mins)).
table(low_v_maxs, 4, 9, 1, low(maxs)).
table(low_c_maxs, 4, 6, 1, low(maxs)).
table(low_minc_maxs, 4, 6, 1, low(maxs)).
table(low_maxc_maxs, 4, 6, 1, low(maxs)).
table(low_s_maxs, 4, 9, 1, low(maxs)).
table(low_mins_maxs, 4, 4, 1, low(maxs)).
table(low_v_c_maxs, 4, 10, 2, low(maxs)).
table(low_v_minc_maxs, 4, 10, 2, low(maxs)).
table(low_v_maxc_maxs, 4, 11, 2, low(maxs)).
table(low_v_s_maxs, 4, 9, 2, low(maxs)).
table(low_v_mins_maxs, 4, 9, 2, low(maxs)).
table(low_c_minc_maxs, 4, 7, 2, low(maxs)).
table(low_c_maxc_maxs, 4, 7, 2, low(maxs)).
table(low_c_s_maxs, 4, 10, 2, low(maxs)).
table(low_c_mins_maxs, 4, 5, 2, low(maxs)).
table(low_minc_maxc_maxs, 4, 9, 2, low(maxs)).
table(low_minc_s_maxs, 4, 10, 2, low(maxs)).
table(low_minc_mins_maxs, 4, 5, 2, low(maxs)).
table(low_maxc_s_maxs, 4, 11, 2, low(maxs)).
table(low_maxc_mins_maxs, 4, 5, 2, low(maxs)).
table(low_s_mins_maxs, 4, 9, 2, low(maxs)).
table(low_v_c_minc_maxs, 4, 18, 3, low(maxs)).
table(low_v_c_maxc_maxs, 4, 18, 3, low(maxs)).
table(low_v_c_s_maxs, 4, 10, 3, low(maxs)).
table(low_v_c_mins_maxs, 4, 10, 3, low(maxs)).
table(low_v_minc_maxc_maxs, 4, 18, 3, low(maxs)).
table(low_v_minc_s_maxs, 4, 10, 3, low(maxs)).
table(low_v_minc_mins_maxs, 4, 10, 3, low(maxs)).
table(low_v_maxc_s_maxs, 4, 11, 3, low(maxs)).
table(low_v_maxc_mins_maxs, 4, 11, 3, low(maxs)).
table(low_v_s_mins_maxs, 4, 9, 3, low(maxs)).
table(low_c_minc_maxc_maxs, 4, 18, 3, low(maxs)).
table(low_c_minc_s_maxs, 4, 18, 3, low(maxs)).
table(low_c_minc_mins_maxs, 4, 7, 3, low(maxs)).
table(low_c_maxc_s_maxs, 4, 7, 3, low(maxs)).
table(low_c_maxc_mins_maxs, 4, 7, 3, low(maxs)).
table(low_c_s_mins_maxs, 4, 10, 3, low(maxs)).
table(low_minc_maxc_s_maxs, 4, 18, 3, low(maxs)).
table(low_minc_maxc_mins_maxs, 4, 8, 3, low(maxs)).
table(low_minc_s_mins_maxs, 4, 10, 3, low(maxs)).
table(low_maxc_s_mins_maxs, 4, 11, 3, low(maxs)).
table(low_v_c_minc_maxc_maxs, 4, 18, 4, low(maxs)).
table(low_v_c_minc_s_maxs, 4, 18, 4, low(maxs)).
table(low_v_c_minc_mins_maxs, 4, 18, 4, low(maxs)).
table(low_v_c_maxc_s_maxs, 4, 18, 4, low(maxs)).
table(low_v_c_maxc_mins_maxs, 4, 18, 4, low(maxs)).
table(low_v_c_s_mins_maxs, 4, 10, 4, low(maxs)).
table(low_v_minc_maxc_s_maxs, 4, 18, 4, low(maxs)).
table(low_v_minc_maxc_mins_maxs, 4, 18, 4, low(maxs)).
table(low_v_minc_s_mins_maxs, 4, 10, 4, low(maxs)).
table(low_v_maxc_s_mins_maxs, 4, 11, 4, low(maxs)).
table(low_c_minc_maxc_s_maxs, 4, 18, 4, low(maxs)).
table(low_c_minc_maxc_mins_maxs, 4, 18, 4, low(maxs)).
table(low_c_minc_s_mins_maxs, 4, 18, 4, low(maxs)).
table(low_c_maxc_s_mins_maxs, 4, 18, 4, low(maxs)).
table(low_minc_maxc_s_mins_maxs, 4, 18, 4, low(maxs)).
table(low_v_c_minc_maxc_s_maxs, 4, 18, 5, low(maxs)).
table(low_v_c_minc_maxc_mins_maxs, 4, 18, 5, low(maxs)).
table(low_v_c_minc_s_mins_maxs, 4, 18, 5, low(maxs)).
table(low_v_c_maxc_s_mins_maxs, 4, 18, 5, low(maxs)).
table(low_v_minc_maxc_s_mins_maxs, 4, 18, 5, low(maxs)).
table(low_c_minc_maxc_s_mins_maxs, 4, 18, 5, low(maxs)).
table(low_v_c_minc_maxc_s_mins_maxs, 4, 18, 6, low(maxs)).
table(up_v_maxs, 4, 18, 1, up(maxs)).
table(up_maxc_maxs, 4, 4, 1, up(maxs)).
table(up_v_c_maxs, 4, 18, 2, up(maxs)).
table(up_v_minc_maxs, 4, 13, 2, up(maxs)).
table(up_v_maxc_maxs, 4, 7, 2, up(maxs)).
table(up_v_s_maxs, 4, 9, 2, up(maxs)).
table(up_v_mins_maxs, 4, 9, 2, up(maxs)).
table(up_c_maxc_maxs, 4, 5, 2, up(maxs)).
table(up_minc_maxc_maxs, 4, 7, 2, up(maxs)).
table(up_maxc_s_maxs, 4, 5, 2, up(maxs)).
table(up_maxc_mins_maxs, 4, 5, 2, up(maxs)).
table(up_v_c_minc_maxs, 4, 13, 3, up(maxs)).
table(up_v_c_maxc_maxs, 4, 13, 3, up(maxs)).
table(up_v_c_s_maxs, 4, 10, 3, up(maxs)).
table(up_v_c_mins_maxs, 4, 18, 3, up(maxs)).
table(up_v_minc_maxc_maxs, 4, 13, 3, up(maxs)).
table(up_v_minc_s_maxs, 4, 10, 3, up(maxs)).
table(up_v_minc_mins_maxs, 4, 18, 3, up(maxs)).
table(up_v_maxc_s_maxs, 4, 11, 3, up(maxs)).
table(up_v_maxc_mins_maxs, 4, 11, 3, up(maxs)).
table(up_v_s_mins_maxs, 4, 9, 3, up(maxs)).
table(up_c_minc_maxc_maxs, 4, 13, 3, up(maxs)).
table(up_c_maxc_s_maxs, 4, 7, 3, up(maxs)).
table(up_c_maxc_mins_maxs, 4, 8, 3, up(maxs)).
table(up_minc_maxc_s_maxs, 4, 18, 3, up(maxs)).
table(up_minc_maxc_mins_maxs, 4, 9, 3, up(maxs)).
table(up_maxc_s_mins_maxs, 4, 11, 3, up(maxs)).
table(up_v_c_minc_maxc_maxs, 4, 13, 4, up(maxs)).
table(up_v_c_minc_s_maxs, 4, 18, 4, up(maxs)).
table(up_v_c_minc_mins_maxs, 4, 18, 4, up(maxs)).
table(up_v_c_maxc_s_maxs, 4, 18, 4, up(maxs)).
table(up_v_c_maxc_mins_maxs, 4, 18, 4, up(maxs)).
table(up_v_c_s_mins_maxs, 4, 10, 4, up(maxs)).
table(up_v_minc_maxc_s_maxs, 4, 18, 4, up(maxs)).
table(up_v_minc_maxc_mins_maxs, 4, 18, 4, up(maxs)).
table(up_v_minc_s_mins_maxs, 4, 10, 4, up(maxs)).
table(up_v_maxc_s_mins_maxs, 4, 11, 4, up(maxs)).
table(up_c_minc_maxc_s_maxs, 4, 18, 4, up(maxs)).
table(up_c_minc_maxc_mins_maxs, 4, 18, 4, up(maxs)).
table(up_c_maxc_s_mins_maxs, 4, 18, 4, up(maxs)).
table(up_minc_maxc_s_mins_maxs, 4, 18, 4, up(maxs)).
table(up_v_c_minc_maxc_s_maxs, 4, 18, 5, up(maxs)).
table(up_v_c_minc_maxc_mins_maxs, 4, 18, 5, up(maxs)).
table(up_v_c_minc_s_mins_maxs, 4, 18, 5, up(maxs)).
table(up_v_c_maxc_s_mins_maxs, 4, 18, 5, up(maxs)).
table(up_v_minc_maxc_s_mins_maxs, 4, 18, 5, up(maxs)).
table(up_c_minc_maxc_s_mins_maxs, 4, 18, 5, up(maxs)).
table(up_v_c_minc_maxc_s_mins_maxs, 4, 18, 6, up(maxs)).
table(low_v_mina, 4, 14, 1, low(mina)).
table(low_c_mina, 4, 7, 1, low(mina)).
table(low_minc_mina, 4, 19, 1, low(mina)).
table(low_maxc_mina, 4, 19, 1, low(mina)).
table(low_s_mina, 4, 14, 1, low(mina)).
table(low_mins_mina, 4, 11, 1, low(mina)).
table(low_maxs_mina, 4, 11, 1, low(mina)).
table(low_v_c_mina, 4, 19, 2, low(mina)).
table(low_v_minc_mina, 4, 15, 2, low(mina)).
table(low_v_maxc_mina, 4, 19, 2, low(mina)).
table(low_v_s_mina, 4, 6, 2, low(mina)).
table(low_v_mins_mina, 4, 14, 2, low(mina)).
table(low_v_maxs_mina, 4, 14, 2, low(mina)).
table(low_c_minc_mina, 4, 8, 2, low(mina)).
table(low_c_maxc_mina, 4, 8, 2, low(mina)).
table(low_c_s_mina, 4, 19, 2, low(mina)).
table(low_c_mins_mina, 4, 7, 2, low(mina)).
table(low_c_maxs_mina, 4, 7, 2, low(mina)).
table(low_minc_maxc_mina, 4, 19, 2, low(mina)).
table(low_minc_s_mina, 4, 15, 2, low(mina)).
table(low_minc_mins_mina, 4, 19, 2, low(mina)).
table(low_minc_maxs_mina, 4, 19, 2, low(mina)).
table(low_maxc_s_mina, 4, 19, 2, low(mina)).
table(low_maxc_mins_mina, 4, 19, 2, low(mina)).
table(low_maxc_maxs_mina, 4, 19, 2, low(mina)).
table(low_s_mins_mina, 4, 14, 2, low(mina)).
table(low_s_maxs_mina, 4, 14, 2, low(mina)).
table(low_mins_maxs_mina, 4, 6, 2, low(mina)).
table(low_v_c_minc_mina, 4, 19, 3, low(mina)).
table(low_v_c_maxc_mina, 4, 19, 3, low(mina)).
table(low_v_c_s_mina, 4, 7, 3, low(mina)).
table(low_v_c_mins_mina, 4, 19, 3, low(mina)).
table(low_v_c_maxs_mina, 4, 19, 3, low(mina)).
table(low_v_minc_maxc_mina, 4, 19, 3, low(mina)).
table(low_v_minc_s_mina, 4, 15, 3, low(mina)).
table(low_v_minc_mins_mina, 4, 15, 3, low(mina)).
table(low_v_minc_maxs_mina, 4, 15, 3, low(mina)).
table(low_v_maxc_s_mina, 4, 19, 3, low(mina)).
table(low_v_maxc_mins_mina, 4, 19, 3, low(mina)).
table(low_v_maxc_maxs_mina, 4, 19, 3, low(mina)).
table(low_v_s_mins_mina, 4, 14, 3, low(mina)).
table(low_v_s_maxs_mina, 4, 14, 3, low(mina)).
table(low_v_mins_maxs_mina, 4, 14, 3, low(mina)).
table(low_c_minc_maxc_mina, 4, 19, 3, low(mina)).
table(low_c_minc_s_mina, 4, 19, 3, low(mina)).
table(low_c_minc_mins_mina, 4, 8, 3, low(mina)).
table(low_c_minc_maxs_mina, 4, 8, 3, low(mina)).
table(low_c_maxc_s_mina, 4, 19, 3, low(mina)).
table(low_c_maxc_mins_mina, 4, 8, 3, low(mina)).
table(low_c_maxc_maxs_mina, 4, 8, 3, low(mina)).
table(low_c_s_mins_mina, 4, 19, 3, low(mina)).
table(low_c_s_maxs_mina, 4, 19, 3, low(mina)).
table(low_c_mins_maxs_mina, 4, 7, 3, low(mina)).
table(low_minc_maxc_s_mina, 4, 19, 3, low(mina)).
table(low_minc_maxc_mins_mina, 4, 19, 3, low(mina)).
table(low_minc_maxc_maxs_mina, 4, 19, 3, low(mina)).
table(low_minc_s_mins_mina, 4, 15, 3, low(mina)).
table(low_minc_s_maxs_mina, 4, 15, 3, low(mina)).
table(low_minc_mins_maxs_mina, 4, 19, 3, low(mina)).
table(low_maxc_s_mins_mina, 4, 19, 3, low(mina)).
table(low_maxc_s_maxs_mina, 4, 19, 3, low(mina)).
table(low_maxc_mins_maxs_mina, 4, 9, 3, low(mina)).
table(low_s_mins_maxs_mina, 4, 14, 3, low(mina)).
table(low_v_c_minc_maxc_mina, 4, 19, 4, low(mina)).
table(low_v_c_minc_s_mina, 4, 19, 4, low(mina)).
table(low_v_c_minc_mins_mina, 4, 19, 4, low(mina)).
table(low_v_c_minc_maxs_mina, 4, 19, 4, low(mina)).
table(low_v_c_maxc_s_mina, 4, 19, 4, low(mina)).
table(low_v_c_maxc_mins_mina, 4, 19, 4, low(mina)).
table(low_v_c_maxc_maxs_mina, 4, 19, 4, low(mina)).
table(low_v_c_s_mins_mina, 4, 19, 4, low(mina)).
table(low_v_c_s_maxs_mina, 4, 19, 4, low(mina)).
table(low_v_c_mins_maxs_mina, 4, 19, 4, low(mina)).
table(low_v_minc_maxc_s_mina, 4, 19, 4, low(mina)).
table(low_v_minc_maxc_mins_mina, 4, 19, 4, low(mina)).
table(low_v_minc_maxc_maxs_mina, 4, 19, 4, low(mina)).
table(low_v_minc_s_mins_mina, 4, 15, 4, low(mina)).
table(low_v_minc_s_maxs_mina, 4, 15, 4, low(mina)).
table(low_v_minc_mins_maxs_mina, 4, 15, 4, low(mina)).
table(low_v_maxc_s_mins_mina, 4, 19, 4, low(mina)).
table(low_v_maxc_s_maxs_mina, 4, 19, 4, low(mina)).
table(low_v_maxc_mins_maxs_mina, 4, 19, 4, low(mina)).
table(low_v_s_mins_maxs_mina, 4, 14, 4, low(mina)).
table(low_c_minc_maxc_s_mina, 4, 19, 4, low(mina)).
table(low_c_minc_maxc_mins_mina, 4, 19, 4, low(mina)).
table(low_c_minc_maxc_maxs_mina, 4, 19, 4, low(mina)).
table(low_c_minc_s_mins_mina, 4, 19, 4, low(mina)).
table(low_c_minc_s_maxs_mina, 4, 19, 4, low(mina)).
table(low_c_minc_mins_maxs_mina, 4, 8, 4, low(mina)).
table(low_c_maxc_s_mins_mina, 4, 19, 4, low(mina)).
table(low_c_maxc_s_maxs_mina, 4, 19, 4, low(mina)).
table(low_c_maxc_mins_maxs_mina, 4, 8, 4, low(mina)).
table(low_c_s_mins_maxs_mina, 4, 19, 4, low(mina)).
table(low_minc_maxc_s_mins_mina, 4, 19, 4, low(mina)).
table(low_minc_maxc_s_maxs_mina, 4, 19, 4, low(mina)).
table(low_minc_maxc_mins_maxs_mina, 4, 19, 4, low(mina)).
table(low_minc_s_mins_maxs_mina, 4, 15, 4, low(mina)).
table(low_maxc_s_mins_maxs_mina, 4, 19, 4, low(mina)).
table(low_v_c_minc_maxc_s_mina, 4, 19, 5, low(mina)).
table(low_v_c_minc_maxc_mins_mina, 4, 19, 5, low(mina)).
table(low_v_c_minc_maxc_maxs_mina, 4, 19, 5, low(mina)).
table(low_v_c_minc_s_mins_mina, 4, 19, 5, low(mina)).
table(low_v_c_minc_s_maxs_mina, 4, 19, 5, low(mina)).
table(low_v_c_minc_mins_maxs_mina, 4, 19, 5, low(mina)).
table(low_v_c_maxc_s_mins_mina, 4, 19, 5, low(mina)).
table(low_v_c_maxc_s_maxs_mina, 4, 19, 5, low(mina)).
table(low_v_c_maxc_mins_maxs_mina, 4, 19, 5, low(mina)).
table(low_v_c_s_mins_maxs_mina, 4, 19, 5, low(mina)).
table(low_v_minc_maxc_s_mins_mina, 4, 19, 5, low(mina)).
table(low_v_minc_maxc_s_maxs_mina, 4, 19, 5, low(mina)).
table(low_v_minc_maxc_mins_maxs_mina, 4, 19, 5, low(mina)).
table(low_v_minc_s_mins_maxs_mina, 4, 15, 5, low(mina)).
table(low_v_maxc_s_mins_maxs_mina, 4, 19, 5, low(mina)).
table(low_c_minc_maxc_s_mins_mina, 4, 19, 5, low(mina)).
table(low_c_minc_maxc_s_maxs_mina, 4, 19, 5, low(mina)).
table(low_c_minc_maxc_mins_maxs_mina, 4, 19, 5, low(mina)).
table(low_c_minc_s_mins_maxs_mina, 4, 19, 5, low(mina)).
table(low_c_maxc_s_mins_maxs_mina, 4, 19, 5, low(mina)).
table(low_minc_maxc_s_mins_maxs_mina, 4, 19, 5, low(mina)).
table(low_v_c_minc_maxc_s_mins_mina, 4, 19, 6, low(mina)).
table(low_v_c_minc_maxc_s_maxs_mina, 4, 19, 6, low(mina)).
table(low_v_c_minc_maxc_mins_maxs_mina, 4, 19, 6, low(mina)).
table(low_v_c_minc_s_mins_maxs_mina, 4, 19, 6, low(mina)).
table(low_v_c_maxc_s_mins_maxs_mina, 4, 19, 6, low(mina)).
table(low_v_minc_maxc_s_mins_maxs_mina, 4, 19, 6, low(mina)).
table(low_c_minc_maxc_s_mins_maxs_mina, 4, 19, 6, low(mina)).
table(low_v_c_minc_maxc_s_mins_maxs_mina, 4, 19, 7, low(mina)).
table(up_v_maxa, 4, 19, 1, up(maxa)).
table(up_v_c_maxa, 4, 19, 2, up(maxa)).
table(up_v_minc_maxa, 4, 19, 2, up(maxa)).
table(up_v_maxc_maxa, 4, 19, 2, up(maxa)).
table(up_v_s_maxa, 4, 19, 2, up(maxa)).
table(up_v_mins_maxa, 4, 19, 2, up(maxa)).
table(up_v_maxs_maxa, 4, 19, 2, up(maxa)).
table(up_c_maxc_maxa, 4, 19, 2, up(maxa)).
table(up_maxc_s_maxa, 4, 19, 2, up(maxa)).
table(up_s_maxs_maxa, 4, 19, 2, up(maxa)).
table(up_v_c_minc_maxa, 4, 19, 3, up(maxa)).
table(up_v_c_maxc_maxa, 4, 19, 3, up(maxa)).
table(up_v_c_s_maxa, 4, 19, 3, up(maxa)).
table(up_v_c_mins_maxa, 4, 19, 3, up(maxa)).
table(up_v_c_maxs_maxa, 4, 8, 3, up(maxa)).
table(up_v_minc_maxc_maxa, 4, 19, 3, up(maxa)).
table(up_v_minc_s_maxa, 4, 19, 3, up(maxa)).
table(up_v_minc_mins_maxa, 4, 19, 3, up(maxa)).
table(up_v_minc_maxs_maxa, 4, 19, 3, up(maxa)).
table(up_v_maxc_s_maxa, 4, 19, 3, up(maxa)).
table(up_v_maxc_mins_maxa, 4, 19, 3, up(maxa)).
table(up_v_maxc_maxs_maxa, 4, 19, 3, up(maxa)).
table(up_v_s_mins_maxa, 4, 19, 3, up(maxa)).
table(up_v_s_maxs_maxa, 4, 19, 3, up(maxa)).
table(up_v_mins_maxs_maxa, 4, 19, 3, up(maxa)).
table(up_c_minc_maxc_maxa, 4, 19, 3, up(maxa)).
table(up_c_maxc_s_maxa, 4, 19, 3, up(maxa)).
table(up_c_maxc_mins_maxa, 4, 19, 3, up(maxa)).
table(up_c_maxc_maxs_maxa, 4, 19, 3, up(maxa)).
table(up_c_s_maxs_maxa, 4, 19, 3, up(maxa)).
table(up_minc_maxc_s_maxa, 4, 19, 3, up(maxa)).
table(up_minc_s_maxs_maxa, 4, 19, 3, up(maxa)).
table(up_maxc_s_mins_maxa, 4, 19, 3, up(maxa)).
table(up_maxc_s_maxs_maxa, 4, 19, 3, up(maxa)).
table(up_s_mins_maxs_maxa, 4, 19, 3, up(maxa)).
table(up_v_c_minc_maxc_maxa, 4, 19, 4, up(maxa)).
table(up_v_c_minc_s_maxa, 4, 19, 4, up(maxa)).
table(up_v_c_minc_mins_maxa, 4, 19, 4, up(maxa)).
table(up_v_c_minc_maxs_maxa, 4, 19, 4, up(maxa)).
table(up_v_c_maxc_s_maxa, 4, 19, 4, up(maxa)).
table(up_v_c_maxc_mins_maxa, 4, 19, 4, up(maxa)).
table(up_v_c_maxc_maxs_maxa, 4, 19, 4, up(maxa)).
table(up_v_c_s_mins_maxa, 4, 19, 4, up(maxa)).
table(up_v_c_s_maxs_maxa, 4, 19, 4, up(maxa)).
table(up_v_c_mins_maxs_maxa, 4, 19, 4, up(maxa)).
table(up_v_minc_maxc_s_maxa, 4, 19, 4, up(maxa)).
table(up_v_minc_maxc_mins_maxa, 4, 19, 4, up(maxa)).
table(up_v_minc_maxc_maxs_maxa, 4, 19, 4, up(maxa)).
table(up_v_minc_s_mins_maxa, 4, 19, 4, up(maxa)).
table(up_v_minc_s_maxs_maxa, 4, 19, 4, up(maxa)).
table(up_v_minc_mins_maxs_maxa, 4, 19, 4, up(maxa)).
table(up_v_maxc_s_mins_maxa, 4, 19, 4, up(maxa)).
table(up_v_maxc_s_maxs_maxa, 4, 19, 4, up(maxa)).
table(up_v_maxc_mins_maxs_maxa, 4, 19, 4, up(maxa)).
table(up_v_s_mins_maxs_maxa, 4, 19, 4, up(maxa)).
table(up_c_minc_maxc_s_maxa, 4, 19, 4, up(maxa)).
table(up_c_minc_maxc_mins_maxa, 4, 19, 4, up(maxa)).
table(up_c_minc_maxc_maxs_maxa, 4, 19, 4, up(maxa)).
table(up_c_minc_s_maxs_maxa, 4, 19, 4, up(maxa)).
table(up_c_maxc_s_mins_maxa, 4, 19, 4, up(maxa)).
table(up_c_maxc_s_maxs_maxa, 4, 19, 4, up(maxa)).
table(up_c_maxc_mins_maxs_maxa, 4, 19, 4, up(maxa)).
table(up_c_s_mins_maxs_maxa, 4, 19, 4, up(maxa)).
table(up_minc_maxc_s_mins_maxa, 4, 19, 4, up(maxa)).
table(up_minc_maxc_s_maxs_maxa, 4, 19, 4, up(maxa)).
table(up_minc_s_mins_maxs_maxa, 4, 19, 4, up(maxa)).
table(up_maxc_s_mins_maxs_maxa, 4, 19, 4, up(maxa)).
table(up_v_c_minc_maxc_s_maxa, 4, 19, 5, up(maxa)).
table(up_v_c_minc_maxc_mins_maxa, 4, 19, 5, up(maxa)).
table(up_v_c_minc_maxc_maxs_maxa, 4, 19, 5, up(maxa)).
table(up_v_c_minc_s_mins_maxa, 4, 19, 5, up(maxa)).
table(up_v_c_minc_s_maxs_maxa, 4, 19, 5, up(maxa)).
table(up_v_c_minc_mins_maxs_maxa, 4, 19, 5, up(maxa)).
table(up_v_c_maxc_s_mins_maxa, 4, 19, 5, up(maxa)).
table(up_v_c_maxc_s_maxs_maxa, 4, 19, 5, up(maxa)).
table(up_v_c_maxc_mins_maxs_maxa, 4, 19, 5, up(maxa)).
table(up_v_c_s_mins_maxs_maxa, 4, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_mins_maxa, 4, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_maxs_maxa, 4, 19, 5, up(maxa)).
table(up_v_minc_maxc_mins_maxs_maxa, 4, 19, 5, up(maxa)).
table(up_v_minc_s_mins_maxs_maxa, 4, 19, 5, up(maxa)).
table(up_v_maxc_s_mins_maxs_maxa, 4, 19, 5, up(maxa)).
table(up_c_minc_maxc_s_mins_maxa, 4, 19, 5, up(maxa)).
table(up_c_minc_maxc_s_maxs_maxa, 4, 19, 5, up(maxa)).
table(up_c_minc_maxc_mins_maxs_maxa, 4, 19, 5, up(maxa)).
table(up_c_minc_s_mins_maxs_maxa, 4, 19, 5, up(maxa)).
table(up_c_maxc_s_mins_maxs_maxa, 4, 19, 5, up(maxa)).
table(up_minc_maxc_s_mins_maxs_maxa, 4, 19, 5, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxa, 4, 19, 6, up(maxa)).
table(up_v_c_minc_maxc_s_maxs_maxa, 4, 19, 6, up(maxa)).
table(up_v_c_minc_maxc_mins_maxs_maxa, 4, 19, 6, up(maxa)).
table(up_v_c_minc_s_mins_maxs_maxa, 4, 19, 6, up(maxa)).
table(up_v_c_maxc_s_mins_maxs_maxa, 4, 19, 6, up(maxa)).
table(up_v_minc_maxc_s_mins_maxs_maxa, 4, 19, 6, up(maxa)).
table(up_c_minc_maxc_s_mins_maxs_maxa, 4, 19, 6, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxs_maxa, 4, 19, 7, up(maxa)).
table(low_c_v, 5, 18, 1, low(v)).
table(low_minc_v, 5, 11, 1, low(v)).
table(low_maxc_v, 5, 11, 1, low(v)).
table(low_s_v, 5, 9, 1, low(v)).
table(low_mins_v, 5, 18, 1, low(v)).
table(low_maxs_v, 5, 18, 1, low(v)).
table(low_c_minc_v, 5, 11, 2, low(v)).
table(low_c_maxc_v, 5, 11, 2, low(v)).
table(low_c_s_v, 5, 10, 2, low(v)).
table(low_c_mins_v, 5, 18, 2, low(v)).
table(low_c_maxs_v, 5, 18, 2, low(v)).
table(low_minc_maxc_v, 5, 11, 2, low(v)).
table(low_minc_s_v, 5, 5, 2, low(v)).
table(low_minc_mins_v, 5, 12, 2, low(v)).
table(low_minc_maxs_v, 5, 12, 2, low(v)).
table(low_maxc_s_v, 5, 5, 2, low(v)).
table(low_maxc_mins_v, 5, 12, 2, low(v)).
table(low_maxc_maxs_v, 5, 13, 2, low(v)).
table(low_s_mins_v, 5, 9, 2, low(v)).
table(low_s_maxs_v, 5, 9, 2, low(v)).
table(low_mins_maxs_v, 5, 12, 2, low(v)).
table(low_c_minc_maxc_v, 5, 11, 3, low(v)).
table(low_c_minc_s_v, 5, 8, 3, low(v)).
table(low_c_minc_mins_v, 5, 12, 3, low(v)).
table(low_c_minc_maxs_v, 5, 12, 3, low(v)).
table(low_c_maxc_s_v, 5, 13, 3, low(v)).
table(low_c_maxc_mins_v, 5, 12, 3, low(v)).
table(low_c_maxc_maxs_v, 5, 13, 3, low(v)).
table(low_c_s_mins_v, 5, 10, 3, low(v)).
table(low_c_s_maxs_v, 5, 10, 3, low(v)).
table(low_c_mins_maxs_v, 5, 18, 3, low(v)).
table(low_minc_maxc_s_v, 5, 7, 3, low(v)).
table(low_minc_maxc_mins_v, 5, 12, 3, low(v)).
table(low_minc_maxc_maxs_v, 5, 12, 3, low(v)).
table(low_minc_s_mins_v, 5, 6, 3, low(v)).
table(low_minc_s_maxs_v, 5, 10, 3, low(v)).
table(low_minc_mins_maxs_v, 5, 14, 3, low(v)).
table(low_maxc_s_mins_v, 5, 6, 3, low(v)).
table(low_maxc_s_maxs_v, 5, 10, 3, low(v)).
table(low_maxc_mins_maxs_v, 5, 14, 3, low(v)).
table(low_s_mins_maxs_v, 5, 9, 3, low(v)).
table(low_c_minc_maxc_s_v, 5, 12, 4, low(v)).
table(low_c_minc_maxc_mins_v, 5, 12, 4, low(v)).
table(low_c_minc_maxc_maxs_v, 5, 12, 4, low(v)).
table(low_c_minc_s_mins_v, 5, 9, 4, low(v)).
table(low_c_minc_s_maxs_v, 5, 13, 4, low(v)).
table(low_c_minc_mins_maxs_v, 5, 14, 4, low(v)).
table(low_c_maxc_s_mins_v, 5, 14, 4, low(v)).
table(low_c_maxc_s_maxs_v, 5, 18, 4, low(v)).
table(low_c_maxc_mins_maxs_v, 5, 14, 4, low(v)).
table(low_c_s_mins_maxs_v, 5, 10, 4, low(v)).
table(low_minc_maxc_s_mins_v, 5, 8, 4, low(v)).
table(low_minc_maxc_s_maxs_v, 5, 12, 4, low(v)).
table(low_minc_maxc_mins_maxs_v, 5, 13, 4, low(v)).
table(low_minc_s_mins_maxs_v, 5, 10, 4, low(v)).
table(low_maxc_s_mins_maxs_v, 5, 10, 4, low(v)).
table(low_c_minc_maxc_s_mins_v, 5, 13, 5, low(v)).
table(low_c_minc_maxc_s_maxs_v, 5, 18, 5, low(v)).
table(low_c_minc_maxc_mins_maxs_v, 5, 13, 5, low(v)).
table(low_c_minc_s_mins_maxs_v, 5, 13, 5, low(v)).
table(low_c_maxc_s_mins_maxs_v, 5, 18, 5, low(v)).
table(low_minc_maxc_s_mins_maxs_v, 5, 12, 5, low(v)).
table(low_c_minc_maxc_s_mins_maxs_v, 5, 18, 6, low(v)).
table(up_c_maxc_v, 5, 11, 2, up(v)).
table(up_maxc_s_v, 5, 5, 2, up(v)).
table(up_s_maxs_v, 5, 9, 2, up(v)).
table(up_c_minc_maxc_v, 5, 11, 3, up(v)).
table(up_c_maxc_s_v, 5, 12, 3, up(v)).
table(up_c_maxc_mins_v, 5, 12, 3, up(v)).
table(up_c_maxc_maxs_v, 5, 12, 3, up(v)).
table(up_c_s_maxs_v, 5, 10, 3, up(v)).
table(up_minc_maxc_s_v, 5, 7, 3, up(v)).
table(up_minc_s_maxs_v, 5, 10, 3, up(v)).
table(up_maxc_s_mins_v, 5, 6, 3, up(v)).
table(up_maxc_s_maxs_v, 5, 10, 3, up(v)).
table(up_s_mins_maxs_v, 5, 9, 3, up(v)).
table(up_c_minc_maxc_s_v, 5, 12, 4, up(v)).
table(up_c_minc_maxc_mins_v, 5, 12, 4, up(v)).
table(up_c_minc_maxc_maxs_v, 5, 12, 4, up(v)).
table(up_c_minc_s_maxs_v, 5, 13, 4, up(v)).
table(up_c_maxc_s_mins_v, 5, 13, 4, up(v)).
table(up_c_maxc_s_maxs_v, 5, 18, 4, up(v)).
table(up_c_maxc_mins_maxs_v, 5, 13, 4, up(v)).
table(up_c_s_mins_maxs_v, 5, 10, 4, up(v)).
table(up_minc_maxc_s_mins_v, 5, 8, 4, up(v)).
table(up_minc_maxc_s_maxs_v, 5, 12, 4, up(v)).
table(up_minc_s_mins_maxs_v, 5, 10, 4, up(v)).
table(up_maxc_s_mins_maxs_v, 5, 10, 4, up(v)).
table(up_c_minc_maxc_s_mins_v, 5, 13, 5, up(v)).
table(up_c_minc_maxc_s_maxs_v, 5, 18, 5, up(v)).
table(up_c_minc_maxc_mins_maxs_v, 5, 13, 5, up(v)).
table(up_c_minc_s_mins_maxs_v, 5, 13, 5, up(v)).
table(up_c_maxc_s_mins_maxs_v, 5, 18, 5, up(v)).
table(up_minc_maxc_s_mins_maxs_v, 5, 12, 5, up(v)).
table(up_c_minc_maxc_s_mins_maxs_v, 5, 18, 6, up(v)).
table(low_v_c, 5, 11, 1, low(c)).
table(low_minc_c, 5, 11, 1, low(c)).
table(low_maxc_c, 5, 11, 1, low(c)).
table(low_s_c, 5, 8, 1, low(c)).
table(low_mins_c, 5, 8, 1, low(c)).
table(low_maxs_c, 5, 8, 1, low(c)).
table(low_v_minc_c, 5, 11, 2, low(c)).
table(low_v_maxc_c, 5, 11, 2, low(c)).
table(low_v_s_c, 5, 13, 2, low(c)).
table(low_v_mins_c, 5, 12, 2, low(c)).
table(low_v_maxs_c, 5, 13, 2, low(c)).
table(low_minc_maxc_c, 5, 11, 2, low(c)).
table(low_minc_s_c, 5, 5, 2, low(c)).
table(low_minc_mins_c, 5, 12, 2, low(c)).
table(low_minc_maxs_c, 5, 9, 2, low(c)).
table(low_maxc_s_c, 5, 6, 2, low(c)).
table(low_maxc_mins_c, 5, 12, 2, low(c)).
table(low_maxc_maxs_c, 5, 13, 2, low(c)).
table(low_s_mins_c, 5, 10, 2, low(c)).
table(low_s_maxs_c, 5, 11, 2, low(c)).
table(low_mins_maxs_c, 5, 9, 2, low(c)).
table(low_v_minc_maxc_c, 5, 11, 3, low(c)).
table(low_v_minc_s_c, 5, 12, 3, low(c)).
table(low_v_minc_mins_c, 5, 12, 3, low(c)).
table(low_v_minc_maxs_c, 5, 12, 3, low(c)).
table(low_v_maxc_s_c, 5, 12, 3, low(c)).
table(low_v_maxc_mins_c, 5, 12, 3, low(c)).
table(low_v_maxc_maxs_c, 5, 12, 3, low(c)).
table(low_v_s_mins_c, 5, 14, 3, low(c)).
table(low_v_s_maxs_c, 5, 18, 3, low(c)).
table(low_v_mins_maxs_c, 5, 14, 3, low(c)).
table(low_minc_maxc_s_c, 5, 7, 3, low(c)).
table(low_minc_maxc_mins_c, 5, 12, 3, low(c)).
table(low_minc_maxc_maxs_c, 5, 12, 3, low(c)).
table(low_minc_s_mins_c, 5, 6, 3, low(c)).
table(low_minc_s_maxs_c, 5, 8, 3, low(c)).
table(low_minc_mins_maxs_c, 5, 7, 3, low(c)).
table(low_maxc_s_mins_c, 5, 7, 3, low(c)).
table(low_maxc_s_maxs_c, 5, 7, 3, low(c)).
table(low_maxc_mins_maxs_c, 5, 6, 3, low(c)).
table(low_s_mins_maxs_c, 5, 12, 3, low(c)).
table(low_v_minc_maxc_s_c, 5, 12, 4, low(c)).
table(low_v_minc_maxc_mins_c, 5, 12, 4, low(c)).
table(low_v_minc_maxc_maxs_c, 5, 12, 4, low(c)).
table(low_v_minc_s_mins_c, 5, 13, 4, low(c)).
table(low_v_minc_s_maxs_c, 5, 18, 4, low(c)).
table(low_v_minc_mins_maxs_c, 5, 13, 4, low(c)).
table(low_v_maxc_s_mins_c, 5, 13, 4, low(c)).
table(low_v_maxc_s_maxs_c, 5, 18, 4, low(c)).
table(low_v_maxc_mins_maxs_c, 5, 13, 4, low(c)).
table(low_v_s_mins_maxs_c, 5, 18, 4, low(c)).
table(low_minc_maxc_s_mins_c, 5, 8, 4, low(c)).
table(low_minc_maxc_s_maxs_c, 5, 10, 4, low(c)).
table(low_minc_maxc_mins_maxs_c, 5, 13, 4, low(c)).
table(low_minc_s_mins_maxs_c, 5, 8, 4, low(c)).
table(low_maxc_s_mins_maxs_c, 5, 10, 4, low(c)).
table(low_v_minc_maxc_s_mins_c, 5, 13, 5, low(c)).
table(low_v_minc_maxc_s_maxs_c, 5, 18, 5, low(c)).
table(low_v_minc_maxc_mins_maxs_c, 5, 13, 5, low(c)).
table(low_v_minc_s_mins_maxs_c, 5, 18, 5, low(c)).
table(low_v_maxc_s_mins_maxs_c, 5, 18, 5, low(c)).
table(low_minc_maxc_s_mins_maxs_c, 5, 10, 5, low(c)).
table(low_v_minc_maxc_s_mins_maxs_c, 5, 18, 6, low(c)).
table(up_v_c, 5, 18, 1, up(c)).
table(up_s_c, 5, 5, 1, up(c)).
table(up_v_minc_c, 5, 11, 2, up(c)).
table(up_v_maxc_c, 5, 11, 2, up(c)).
table(up_v_s_c, 5, 6, 2, up(c)).
table(up_v_mins_c, 5, 18, 2, up(c)).
table(up_v_maxs_c, 5, 18, 2, up(c)).
table(up_minc_s_c, 5, 5, 2, up(c)).
table(up_maxc_s_c, 5, 6, 2, up(c)).
table(up_s_mins_c, 5, 7, 2, up(c)).
table(up_s_maxs_c, 5, 7, 2, up(c)).
table(up_v_minc_maxc_c, 5, 11, 3, up(c)).
table(up_v_minc_s_c, 5, 8, 3, up(c)).
table(up_v_minc_mins_c, 5, 12, 3, up(c)).
table(up_v_minc_maxs_c, 5, 12, 3, up(c)).
table(up_v_maxc_s_c, 5, 13, 3, up(c)).
table(up_v_maxc_mins_c, 5, 12, 3, up(c)).
table(up_v_maxc_maxs_c, 5, 13, 3, up(c)).
table(up_v_s_mins_c, 5, 10, 3, up(c)).
table(up_v_s_maxs_c, 5, 18, 3, up(c)).
table(up_v_mins_maxs_c, 5, 18, 3, up(c)).
table(up_minc_maxc_s_c, 5, 7, 3, up(c)).
table(up_minc_s_mins_c, 5, 6, 3, up(c)).
table(up_minc_s_maxs_c, 5, 7, 3, up(c)).
table(up_maxc_s_mins_c, 5, 9, 3, up(c)).
table(up_maxc_s_maxs_c, 5, 7, 3, up(c)).
table(up_s_mins_maxs_c, 5, 10, 3, up(c)).
table(up_v_minc_maxc_s_c, 5, 12, 4, up(c)).
table(up_v_minc_maxc_mins_c, 5, 12, 4, up(c)).
table(up_v_minc_maxc_maxs_c, 5, 12, 4, up(c)).
table(up_v_minc_s_mins_c, 5, 9, 4, up(c)).
table(up_v_minc_s_maxs_c, 5, 18, 4, up(c)).
table(up_v_minc_mins_maxs_c, 5, 13, 4, up(c)).
table(up_v_maxc_s_mins_c, 5, 14, 4, up(c)).
table(up_v_maxc_s_maxs_c, 5, 18, 4, up(c)).
table(up_v_maxc_mins_maxs_c, 5, 14, 4, up(c)).
table(up_v_s_mins_maxs_c, 5, 18, 4, up(c)).
table(up_minc_maxc_s_mins_c, 5, 8, 4, up(c)).
table(up_minc_maxc_s_maxs_c, 5, 10, 4, up(c)).
table(up_minc_s_mins_maxs_c, 5, 7, 4, up(c)).
table(up_maxc_s_mins_maxs_c, 5, 8, 4, up(c)).
table(up_v_minc_maxc_s_mins_c, 5, 13, 5, up(c)).
table(up_v_minc_maxc_s_maxs_c, 5, 18, 5, up(c)).
table(up_v_minc_maxc_mins_maxs_c, 5, 13, 5, up(c)).
table(up_v_minc_s_mins_maxs_c, 5, 18, 5, up(c)).
table(up_v_maxc_s_mins_maxs_c, 5, 18, 5, up(c)).
table(up_minc_maxc_s_mins_maxs_c, 5, 10, 5, up(c)).
table(up_v_minc_maxc_s_mins_maxs_c, 5, 18, 6, up(c)).
table(low_v_minc, 5, 5, 1, low(minc)).
table(low_c_minc, 5, 5, 1, low(minc)).
table(low_maxc_minc, 5, 5, 1, low(minc)).
table(low_s_minc, 5, 5, 1, low(minc)).
table(low_mins_minc, 5, 4, 1, low(minc)).
table(low_maxs_minc, 5, 5, 1, low(minc)).
table(low_v_c_minc, 5, 7, 2, low(minc)).
table(low_v_maxc_minc, 5, 6, 2, low(minc)).
table(low_v_s_minc, 5, 6, 2, low(minc)).
table(low_v_mins_minc, 5, 5, 2, low(minc)).
table(low_v_maxs_minc, 5, 6, 2, low(minc)).
table(low_c_maxc_minc, 5, 6, 2, low(minc)).
table(low_c_s_minc, 5, 6, 2, low(minc)).
table(low_c_mins_minc, 5, 5, 2, low(minc)).
table(low_c_maxs_minc, 5, 6, 2, low(minc)).
table(low_maxc_s_minc, 5, 6, 2, low(minc)).
table(low_maxc_mins_minc, 5, 6, 2, low(minc)).
table(low_maxc_maxs_minc, 5, 6, 2, low(minc)).
table(low_s_mins_minc, 5, 5, 2, low(minc)).
table(low_s_maxs_minc, 5, 6, 2, low(minc)).
table(low_mins_maxs_minc, 5, 5, 2, low(minc)).
table(low_v_c_maxc_minc, 5, 11, 3, low(minc)).
table(low_v_c_s_minc, 5, 8, 3, low(minc)).
table(low_v_c_mins_minc, 5, 8, 3, low(minc)).
table(low_v_c_maxs_minc, 5, 8, 3, low(minc)).
table(low_v_maxc_s_minc, 5, 7, 3, low(minc)).
table(low_v_maxc_mins_minc, 5, 7, 3, low(minc)).
table(low_v_maxc_maxs_minc, 5, 7, 3, low(minc)).
table(low_v_s_mins_minc, 5, 6, 3, low(minc)).
table(low_v_s_maxs_minc, 5, 10, 3, low(minc)).
table(low_v_mins_maxs_minc, 5, 6, 3, low(minc)).
table(low_c_maxc_s_minc, 5, 8, 3, low(minc)).
table(low_c_maxc_mins_minc, 5, 7, 3, low(minc)).
table(low_c_maxc_maxs_minc, 5, 7, 3, low(minc)).
table(low_c_s_mins_minc, 5, 6, 3, low(minc)).
table(low_c_s_maxs_minc, 5, 8, 3, low(minc)).
table(low_c_mins_maxs_minc, 5, 6, 3, low(minc)).
table(low_maxc_s_mins_minc, 5, 7, 3, low(minc)).
table(low_maxc_s_maxs_minc, 5, 8, 3, low(minc)).
table(low_maxc_mins_maxs_minc, 5, 7, 3, low(minc)).
table(low_s_mins_maxs_minc, 5, 6, 3, low(minc)).
table(low_v_c_maxc_s_minc, 5, 12, 4, low(minc)).
table(low_v_c_maxc_mins_minc, 5, 12, 4, low(minc)).
table(low_v_c_maxc_maxs_minc, 5, 12, 4, low(minc)).
table(low_v_c_s_mins_minc, 5, 9, 4, low(minc)).
table(low_v_c_s_maxs_minc, 5, 13, 4, low(minc)).
table(low_v_c_mins_maxs_minc, 5, 9, 4, low(minc)).
table(low_v_maxc_s_mins_minc, 5, 8, 4, low(minc)).
table(low_v_maxc_s_maxs_minc, 5, 12, 4, low(minc)).
table(low_v_maxc_mins_maxs_minc, 5, 8, 4, low(minc)).
table(low_v_s_mins_maxs_minc, 5, 10, 4, low(minc)).
table(low_c_maxc_s_mins_minc, 5, 9, 4, low(minc)).
table(low_c_maxc_s_maxs_minc, 5, 10, 4, low(minc)).
table(low_c_maxc_mins_maxs_minc, 5, 8, 4, low(minc)).
table(low_c_s_mins_maxs_minc, 5, 8, 4, low(minc)).
table(low_maxc_s_mins_maxs_minc, 5, 8, 4, low(minc)).
table(low_v_c_maxc_s_mins_minc, 5, 13, 5, low(minc)).
table(low_v_c_maxc_s_maxs_minc, 5, 18, 5, low(minc)).
table(low_v_c_maxc_mins_maxs_minc, 5, 13, 5, low(minc)).
table(low_v_c_s_mins_maxs_minc, 5, 13, 5, low(minc)).
table(low_v_maxc_s_mins_maxs_minc, 5, 12, 5, low(minc)).
table(low_c_maxc_s_mins_maxs_minc, 5, 10, 5, low(minc)).
table(low_v_c_maxc_s_mins_maxs_minc, 5, 18, 6, low(minc)).
table(up_v_minc, 5, 11, 1, up(minc)).
table(up_maxc_minc, 5, 6, 1, up(minc)).
table(up_v_c_minc, 5, 7, 2, up(minc)).
table(up_v_maxc_minc, 5, 6, 2, up(minc)).
table(up_v_s_minc, 5, 13, 2, up(minc)).
table(up_v_mins_minc, 5, 12, 2, up(minc)).
table(up_v_maxs_minc, 5, 13, 2, up(minc)).
table(up_c_maxc_minc, 5, 6, 2, up(minc)).
table(up_maxc_s_minc, 5, 6, 2, up(minc)).
table(up_maxc_mins_minc, 5, 7, 2, up(minc)).
table(up_maxc_maxs_minc, 5, 7, 2, up(minc)).
table(up_s_maxs_minc, 5, 18, 2, up(minc)).
table(up_v_c_maxc_minc, 5, 11, 3, up(minc)).
table(up_v_c_s_minc, 5, 8, 3, up(minc)).
table(up_v_c_mins_minc, 5, 8, 3, up(minc)).
table(up_v_c_maxs_minc, 5, 8, 3, up(minc)).
table(up_v_maxc_s_minc, 5, 7, 3, up(minc)).
table(up_v_maxc_mins_minc, 5, 7, 3, up(minc)).
table(up_v_maxc_maxs_minc, 5, 7, 3, up(minc)).
table(up_v_s_mins_minc, 5, 14, 3, up(minc)).
table(up_v_s_maxs_minc, 5, 18, 3, up(minc)).
table(up_v_mins_maxs_minc, 5, 14, 3, up(minc)).
table(up_c_maxc_s_minc, 5, 7, 3, up(minc)).
table(up_c_maxc_mins_minc, 5, 7, 3, up(minc)).
table(up_c_maxc_maxs_minc, 5, 7, 3, up(minc)).
table(up_c_s_maxs_minc, 5, 7, 3, up(minc)).
table(up_maxc_s_mins_minc, 5, 8, 3, up(minc)).
table(up_maxc_s_maxs_minc, 5, 12, 3, up(minc)).
table(up_maxc_mins_maxs_minc, 5, 8, 3, up(minc)).
table(up_s_mins_maxs_minc, 5, 18, 3, up(minc)).
table(up_v_c_maxc_s_minc, 5, 12, 4, up(minc)).
table(up_v_c_maxc_mins_minc, 5, 12, 4, up(minc)).
table(up_v_c_maxc_maxs_minc, 5, 12, 4, up(minc)).
table(up_v_c_s_mins_minc, 5, 9, 4, up(minc)).
table(up_v_c_s_maxs_minc, 5, 13, 4, up(minc)).
table(up_v_c_mins_maxs_minc, 5, 9, 4, up(minc)).
table(up_v_maxc_s_mins_minc, 5, 8, 4, up(minc)).
table(up_v_maxc_s_maxs_minc, 5, 12, 4, up(minc)).
table(up_v_maxc_mins_maxs_minc, 5, 8, 4, up(minc)).
table(up_v_s_mins_maxs_minc, 5, 18, 4, up(minc)).
table(up_c_maxc_s_mins_minc, 5, 8, 4, up(minc)).
table(up_c_maxc_s_maxs_minc, 5, 10, 4, up(minc)).
table(up_c_maxc_mins_maxs_minc, 5, 8, 4, up(minc)).
table(up_c_s_mins_maxs_minc, 5, 7, 4, up(minc)).
table(up_maxc_s_mins_maxs_minc, 5, 12, 4, up(minc)).
table(up_v_c_maxc_s_mins_minc, 5, 13, 5, up(minc)).
table(up_v_c_maxc_s_maxs_minc, 5, 18, 5, up(minc)).
table(up_v_c_maxc_mins_maxs_minc, 5, 13, 5, up(minc)).
table(up_v_c_s_mins_maxs_minc, 5, 13, 5, up(minc)).
table(up_v_maxc_s_mins_maxs_minc, 5, 12, 5, up(minc)).
table(up_c_maxc_s_mins_maxs_minc, 5, 10, 5, up(minc)).
table(up_v_c_maxc_s_mins_maxs_minc, 5, 18, 6, up(minc)).
table(low_v_maxc, 5, 18, 1, low(maxc)).
table(low_c_maxc, 5, 18, 1, low(maxc)).
table(low_minc_maxc, 5, 6, 1, low(maxc)).
table(low_s_maxc, 5, 18, 1, low(maxc)).
table(low_mins_maxc, 5, 9, 1, low(maxc)).
table(low_maxs_maxc, 5, 4, 1, low(maxc)).
table(low_v_c_maxc, 5, 11, 2, low(maxc)).
table(low_v_minc_maxc, 5, 11, 2, low(maxc)).
table(low_v_s_maxc, 5, 10, 2, low(maxc)).
table(low_v_mins_maxc, 5, 18, 2, low(maxc)).
table(low_v_maxs_maxc, 5, 5, 2, low(maxc)).
table(low_c_minc_maxc, 5, 11, 2, low(maxc)).
table(low_c_s_maxc, 5, 7, 2, low(maxc)).
table(low_c_mins_maxc, 5, 18, 2, low(maxc)).
table(low_c_maxs_maxc, 5, 5, 2, low(maxc)).
table(low_minc_s_maxc, 5, 7, 2, low(maxc)).
table(low_minc_mins_maxc, 5, 7, 2, low(maxc)).
table(low_minc_maxs_maxc, 5, 6, 2, low(maxc)).
table(low_s_mins_maxc, 5, 18, 2, low(maxc)).
table(low_s_maxs_maxc, 5, 5, 2, low(maxc)).
table(low_mins_maxs_maxc, 5, 5, 2, low(maxc)).
table(low_v_c_minc_maxc, 5, 11, 3, low(maxc)).
table(low_v_c_s_maxc, 5, 12, 3, low(maxc)).
table(low_v_c_mins_maxc, 5, 12, 3, low(maxc)).
table(low_v_c_maxs_maxc, 5, 12, 3, low(maxc)).
table(low_v_minc_s_maxc, 5, 7, 3, low(maxc)).
table(low_v_minc_mins_maxc, 5, 12, 3, low(maxc)).
table(low_v_minc_maxs_maxc, 5, 7, 3, low(maxc)).
table(low_v_s_mins_maxc, 5, 10, 3, low(maxc)).
table(low_v_s_maxs_maxc, 5, 10, 3, low(maxc)).
table(low_v_mins_maxs_maxc, 5, 6, 3, low(maxc)).
table(low_c_minc_s_maxc, 5, 8, 3, low(maxc)).
table(low_c_minc_mins_maxc, 5, 12, 3, low(maxc)).
table(low_c_minc_maxs_maxc, 5, 7, 3, low(maxc)).
table(low_c_s_mins_maxc, 5, 7, 3, low(maxc)).
table(low_c_s_maxs_maxc, 5, 7, 3, low(maxc)).
table(low_c_mins_maxs_maxc, 5, 6, 3, low(maxc)).
table(low_minc_s_mins_maxc, 5, 14, 3, low(maxc)).
table(low_minc_s_maxs_maxc, 5, 8, 3, low(maxc)).
table(low_minc_mins_maxs_maxc, 5, 7, 3, low(maxc)).
table(low_s_mins_maxs_maxc, 5, 6, 3, low(maxc)).
table(low_v_c_minc_s_maxc, 5, 12, 4, low(maxc)).
table(low_v_c_minc_mins_maxc, 5, 12, 4, low(maxc)).
table(low_v_c_minc_maxs_maxc, 5, 12, 4, low(maxc)).
table(low_v_c_s_mins_maxc, 5, 13, 4, low(maxc)).
table(low_v_c_s_maxs_maxc, 5, 18, 4, low(maxc)).
table(low_v_c_mins_maxs_maxc, 5, 13, 4, low(maxc)).
table(low_v_minc_s_mins_maxc, 5, 8, 4, low(maxc)).
table(low_v_minc_s_maxs_maxc, 5, 12, 4, low(maxc)).
table(low_v_minc_mins_maxs_maxc, 5, 8, 4, low(maxc)).
table(low_v_s_mins_maxs_maxc, 5, 10, 4, low(maxc)).
table(low_c_minc_s_mins_maxc, 5, 9, 4, low(maxc)).
table(low_c_minc_s_maxs_maxc, 5, 10, 4, low(maxc)).
table(low_c_minc_mins_maxs_maxc, 5, 8, 4, low(maxc)).
table(low_c_s_mins_maxs_maxc, 5, 10, 4, low(maxc)).
table(low_minc_s_mins_maxs_maxc, 5, 8, 4, low(maxc)).
table(low_v_c_minc_s_mins_maxc, 5, 13, 5, low(maxc)).
table(low_v_c_minc_s_maxs_maxc, 5, 18, 5, low(maxc)).
table(low_v_c_minc_mins_maxs_maxc, 5, 13, 5, low(maxc)).
table(low_v_c_s_mins_maxs_maxc, 5, 18, 5, low(maxc)).
table(low_v_minc_s_mins_maxs_maxc, 5, 12, 5, low(maxc)).
table(low_c_minc_s_mins_maxs_maxc, 5, 10, 5, low(maxc)).
table(low_v_c_minc_s_mins_maxs_maxc, 5, 18, 6, low(maxc)).
table(up_v_maxc, 5, 11, 1, up(maxc)).
table(up_v_c_maxc, 5, 11, 2, up(maxc)).
table(up_v_minc_maxc, 5, 11, 2, up(maxc)).
table(up_v_s_maxc, 5, 13, 2, up(maxc)).
table(up_v_mins_maxc, 5, 12, 2, up(maxc)).
table(up_v_maxs_maxc, 5, 13, 2, up(maxc)).
table(up_v_c_minc_maxc, 5, 11, 3, up(maxc)).
table(up_v_c_s_maxc, 5, 13, 3, up(maxc)).
table(up_v_c_mins_maxc, 5, 12, 3, up(maxc)).
table(up_v_c_maxs_maxc, 5, 13, 3, up(maxc)).
table(up_v_minc_s_maxc, 5, 12, 3, up(maxc)).
table(up_v_minc_mins_maxc, 5, 12, 3, up(maxc)).
table(up_v_minc_maxs_maxc, 5, 12, 3, up(maxc)).
table(up_v_s_mins_maxc, 5, 14, 3, up(maxc)).
table(up_v_s_maxs_maxc, 5, 18, 3, up(maxc)).
table(up_v_mins_maxs_maxc, 5, 14, 3, up(maxc)).
table(up_v_c_minc_s_maxc, 5, 12, 4, up(maxc)).
table(up_v_c_minc_mins_maxc, 5, 12, 4, up(maxc)).
table(up_v_c_minc_maxs_maxc, 5, 12, 4, up(maxc)).
table(up_v_c_s_mins_maxc, 5, 14, 4, up(maxc)).
table(up_v_c_s_maxs_maxc, 5, 18, 4, up(maxc)).
table(up_v_c_mins_maxs_maxc, 5, 14, 4, up(maxc)).
table(up_v_minc_s_mins_maxc, 5, 13, 4, up(maxc)).
table(up_v_minc_s_maxs_maxc, 5, 18, 4, up(maxc)).
table(up_v_minc_mins_maxs_maxc, 5, 13, 4, up(maxc)).
table(up_v_s_mins_maxs_maxc, 5, 18, 4, up(maxc)).
table(up_v_c_minc_s_mins_maxc, 5, 13, 5, up(maxc)).
table(up_v_c_minc_s_maxs_maxc, 5, 18, 5, up(maxc)).
table(up_v_c_minc_mins_maxs_maxc, 5, 13, 5, up(maxc)).
table(up_v_c_s_mins_maxs_maxc, 5, 18, 5, up(maxc)).
table(up_v_minc_s_mins_maxs_maxc, 5, 18, 5, up(maxc)).
table(up_v_c_minc_s_mins_maxs_maxc, 5, 18, 6, up(maxc)).
table(low_v_s, 5, 18, 1, low(s)).
table(low_c_s, 5, 5, 1, low(s)).
table(low_minc_s, 5, 18, 1, low(s)).
table(low_maxc_s, 5, 18, 1, low(s)).
table(low_mins_s, 5, 18, 1, low(s)).
table(low_maxs_s, 5, 18, 1, low(s)).
table(low_v_c_s, 5, 6, 2, low(s)).
table(low_v_minc_s, 5, 18, 2, low(s)).
table(low_v_maxc_s, 5, 18, 2, low(s)).
table(low_v_mins_s, 5, 9, 2, low(s)).
table(low_v_maxs_s, 5, 9, 2, low(s)).
table(low_c_minc_s, 5, 7, 2, low(s)).
table(low_c_maxc_s, 5, 7, 2, low(s)).
table(low_c_mins_s, 5, 7, 2, low(s)).
table(low_c_maxs_s, 5, 7, 2, low(s)).
table(low_minc_maxc_s, 5, 18, 2, low(s)).
table(low_minc_mins_s, 5, 18, 2, low(s)).
table(low_minc_maxs_s, 5, 18, 2, low(s)).
table(low_maxc_mins_s, 5, 8, 2, low(s)).
table(low_maxc_maxs_s, 5, 18, 2, low(s)).
table(low_mins_maxs_s, 5, 12, 2, low(s)).
table(low_v_c_minc_s, 5, 10, 3, low(s)).
table(low_v_c_maxc_s, 5, 18, 3, low(s)).
table(low_v_c_mins_s, 5, 10, 3, low(s)).
table(low_v_c_maxs_s, 5, 10, 3, low(s)).
table(low_v_minc_maxc_s, 5, 18, 3, low(s)).
table(low_v_minc_mins_s, 5, 13, 3, low(s)).
table(low_v_minc_maxs_s, 5, 10, 3, low(s)).
table(low_v_maxc_mins_s, 5, 6, 3, low(s)).
table(low_v_maxc_maxs_s, 5, 10, 3, low(s)).
table(low_v_mins_maxs_s, 5, 9, 3, low(s)).
table(low_c_minc_maxc_s, 5, 10, 3, low(s)).
table(low_c_minc_mins_s, 5, 6, 3, low(s)).
table(low_c_minc_maxs_s, 5, 10, 3, low(s)).
table(low_c_maxc_mins_s, 5, 10, 3, low(s)).
table(low_c_maxc_maxs_s, 5, 7, 3, low(s)).
table(low_c_mins_maxs_s, 5, 10, 3, low(s)).
table(low_minc_maxc_mins_s, 5, 13, 3, low(s)).
table(low_minc_maxc_maxs_s, 5, 18, 3, low(s)).
table(low_minc_mins_maxs_s, 5, 8, 3, low(s)).
table(low_maxc_mins_maxs_s, 5, 10, 3, low(s)).
table(low_v_c_minc_maxc_s, 5, 18, 4, low(s)).
table(low_v_c_minc_mins_s, 5, 9, 4, low(s)).
table(low_v_c_minc_maxs_s, 5, 13, 4, low(s)).
table(low_v_c_maxc_mins_s, 5, 13, 4, low(s)).
table(low_v_c_maxc_maxs_s, 5, 18, 4, low(s)).
table(low_v_c_mins_maxs_s, 5, 10, 4, low(s)).
table(low_v_minc_maxc_mins_s, 5, 13, 4, low(s)).
table(low_v_minc_maxc_maxs_s, 5, 12, 4, low(s)).
table(low_v_minc_mins_maxs_s, 5, 10, 4, low(s)).
table(low_v_maxc_mins_maxs_s, 5, 10, 4, low(s)).
table(low_c_minc_maxc_mins_s, 5, 8, 4, low(s)).
table(low_c_minc_maxc_maxs_s, 5, 10, 4, low(s)).
table(low_c_minc_mins_maxs_s, 5, 7, 4, low(s)).
table(low_c_maxc_mins_maxs_s, 5, 8, 4, low(s)).
table(low_minc_maxc_mins_maxs_s, 5, 18, 4, low(s)).
table(low_v_c_minc_maxc_mins_s, 5, 13, 5, low(s)).
table(low_v_c_minc_maxc_maxs_s, 5, 18, 5, low(s)).
table(low_v_c_minc_mins_maxs_s, 5, 13, 5, low(s)).
table(low_v_c_maxc_mins_maxs_s, 5, 18, 5, low(s)).
table(low_v_minc_maxc_mins_maxs_s, 5, 12, 5, low(s)).
table(low_c_minc_maxc_mins_maxs_s, 5, 10, 5, low(s)).
table(low_v_c_minc_maxc_mins_maxs_s, 5, 18, 6, low(s)).
table(up_v_s, 5, 9, 1, up(s)).
table(up_v_c_s, 5, 10, 2, up(s)).
table(up_v_minc_s, 5, 10, 2, up(s)).
table(up_v_maxc_s, 5, 10, 2, up(s)).
table(up_v_mins_s, 5, 9, 2, up(s)).
table(up_v_maxs_s, 5, 9, 2, up(s)).
table(up_c_maxc_s, 5, 18, 2, up(s)).
table(up_v_c_minc_s, 5, 13, 3, up(s)).
table(up_v_c_maxc_s, 5, 18, 3, up(s)).
table(up_v_c_mins_s, 5, 10, 3, up(s)).
table(up_v_c_maxs_s, 5, 10, 3, up(s)).
table(up_v_minc_maxc_s, 5, 12, 3, up(s)).
table(up_v_minc_mins_s, 5, 10, 3, up(s)).
table(up_v_minc_maxs_s, 5, 10, 3, up(s)).
table(up_v_maxc_mins_s, 5, 10, 3, up(s)).
table(up_v_maxc_maxs_s, 5, 10, 3, up(s)).
table(up_v_mins_maxs_s, 5, 9, 3, up(s)).
table(up_c_minc_maxc_s, 5, 18, 3, up(s)).
table(up_c_maxc_mins_s, 5, 18, 3, up(s)).
table(up_c_maxc_maxs_s, 5, 18, 3, up(s)).
table(up_v_c_minc_maxc_s, 5, 18, 4, up(s)).
table(up_v_c_minc_mins_s, 5, 13, 4, up(s)).
table(up_v_c_minc_maxs_s, 5, 13, 4, up(s)).
table(up_v_c_maxc_mins_s, 5, 18, 4, up(s)).
table(up_v_c_maxc_maxs_s, 5, 18, 4, up(s)).
table(up_v_c_mins_maxs_s, 5, 10, 4, up(s)).
table(up_v_minc_maxc_mins_s, 5, 12, 4, up(s)).
table(up_v_minc_maxc_maxs_s, 5, 12, 4, up(s)).
table(up_v_minc_mins_maxs_s, 5, 10, 4, up(s)).
table(up_v_maxc_mins_maxs_s, 5, 10, 4, up(s)).
table(up_c_minc_maxc_mins_s, 5, 18, 4, up(s)).
table(up_c_minc_maxc_maxs_s, 5, 18, 4, up(s)).
table(up_c_maxc_mins_maxs_s, 5, 18, 4, up(s)).
table(up_v_c_minc_maxc_mins_s, 5, 18, 5, up(s)).
table(up_v_c_minc_maxc_maxs_s, 5, 18, 5, up(s)).
table(up_v_c_minc_mins_maxs_s, 5, 13, 5, up(s)).
table(up_v_c_maxc_mins_maxs_s, 5, 18, 5, up(s)).
table(up_v_minc_maxc_mins_maxs_s, 5, 12, 5, up(s)).
table(up_c_minc_maxc_mins_maxs_s, 5, 18, 5, up(s)).
table(up_v_c_minc_maxc_mins_maxs_s, 5, 18, 6, up(s)).
table(low_v_mins, 5, 4, 1, low(mins)).
table(low_c_mins, 5, 4, 1, low(mins)).
table(low_minc_mins, 5, 4, 1, low(mins)).
table(low_maxc_mins, 5, 4, 1, low(mins)).
table(low_s_mins, 5, 4, 1, low(mins)).
table(low_maxs_mins, 5, 4, 1, low(mins)).
table(low_v_c_mins, 5, 5, 2, low(mins)).
table(low_v_minc_mins, 5, 5, 2, low(mins)).
table(low_v_maxc_mins, 5, 5, 2, low(mins)).
table(low_v_s_mins, 5, 5, 2, low(mins)).
table(low_v_maxs_mins, 5, 5, 2, low(mins)).
table(low_c_minc_mins, 5, 5, 2, low(mins)).
table(low_c_maxc_mins, 5, 5, 2, low(mins)).
table(low_c_s_mins, 5, 5, 2, low(mins)).
table(low_c_maxs_mins, 5, 5, 2, low(mins)).
table(low_minc_maxc_mins, 5, 6, 2, low(mins)).
table(low_minc_s_mins, 5, 5, 2, low(mins)).
table(low_minc_maxs_mins, 5, 5, 2, low(mins)).
table(low_maxc_s_mins, 5, 5, 2, low(mins)).
table(low_maxc_maxs_mins, 5, 5, 2, low(mins)).
table(low_s_maxs_mins, 5, 5, 2, low(mins)).
table(low_v_c_minc_mins, 5, 8, 3, low(mins)).
table(low_v_c_maxc_mins, 5, 12, 3, low(mins)).
table(low_v_c_s_mins, 5, 6, 3, low(mins)).
table(low_v_c_maxs_mins, 5, 6, 3, low(mins)).
table(low_v_minc_maxc_mins, 5, 7, 3, low(mins)).
table(low_v_minc_s_mins, 5, 6, 3, low(mins)).
table(low_v_minc_maxs_mins, 5, 6, 3, low(mins)).
table(low_v_maxc_s_mins, 5, 6, 3, low(mins)).
table(low_v_maxc_maxs_mins, 5, 6, 3, low(mins)).
table(low_v_s_maxs_mins, 5, 9, 3, low(mins)).
table(low_c_minc_maxc_mins, 5, 7, 3, low(mins)).
table(low_c_minc_s_mins, 5, 6, 3, low(mins)).
table(low_c_minc_maxs_mins, 5, 6, 3, low(mins)).
table(low_c_maxc_s_mins, 5, 6, 3, low(mins)).
table(low_c_maxc_maxs_mins, 5, 6, 3, low(mins)).
table(low_c_s_maxs_mins, 5, 6, 3, low(mins)).
table(low_minc_maxc_s_mins, 5, 7, 3, low(mins)).
table(low_minc_maxc_maxs_mins, 5, 7, 3, low(mins)).
table(low_minc_s_maxs_mins, 5, 6, 3, low(mins)).
table(low_maxc_s_maxs_mins, 5, 6, 3, low(mins)).
table(low_v_c_minc_maxc_mins, 5, 12, 4, low(mins)).
table(low_v_c_minc_s_mins, 5, 9, 4, low(mins)).
table(low_v_c_minc_maxs_mins, 5, 9, 4, low(mins)).
table(low_v_c_maxc_s_mins, 5, 13, 4, low(mins)).
table(low_v_c_maxc_maxs_mins, 5, 13, 4, low(mins)).
table(low_v_c_s_maxs_mins, 5, 10, 4, low(mins)).
table(low_v_minc_maxc_s_mins, 5, 8, 4, low(mins)).
table(low_v_minc_maxc_maxs_mins, 5, 8, 4, low(mins)).
table(low_v_minc_s_maxs_mins, 5, 10, 4, low(mins)).
table(low_v_maxc_s_maxs_mins, 5, 10, 4, low(mins)).
table(low_c_minc_maxc_s_mins, 5, 8, 4, low(mins)).
table(low_c_minc_maxc_maxs_mins, 5, 8, 4, low(mins)).
table(low_c_minc_s_maxs_mins, 5, 7, 4, low(mins)).
table(low_c_maxc_s_maxs_mins, 5, 8, 4, low(mins)).
table(low_minc_maxc_s_maxs_mins, 5, 8, 4, low(mins)).
table(low_v_c_minc_maxc_s_mins, 5, 13, 5, low(mins)).
table(low_v_c_minc_maxc_maxs_mins, 5, 13, 5, low(mins)).
table(low_v_c_minc_s_maxs_mins, 5, 13, 5, low(mins)).
table(low_v_c_maxc_s_maxs_mins, 5, 18, 5, low(mins)).
table(low_v_minc_maxc_s_maxs_mins, 5, 12, 5, low(mins)).
table(low_c_minc_maxc_s_maxs_mins, 5, 10, 5, low(mins)).
table(low_v_c_minc_maxc_s_maxs_mins, 5, 18, 6, low(mins)).
table(up_v_mins, 5, 18, 1, up(mins)).
table(up_maxc_mins, 5, 9, 1, up(mins)).
table(up_maxs_mins, 5, 4, 1, up(mins)).
table(up_v_c_mins, 5, 8, 2, up(mins)).
table(up_v_minc_mins, 5, 5, 2, up(mins)).
table(up_v_maxc_mins, 5, 7, 2, up(mins)).
table(up_v_s_mins, 5, 5, 2, up(mins)).
table(up_v_maxs_mins, 5, 5, 2, up(mins)).
table(up_c_maxc_mins, 5, 7, 2, up(mins)).
table(up_c_maxs_mins, 5, 5, 2, up(mins)).
table(up_minc_maxc_mins, 5, 6, 2, up(mins)).
table(up_minc_maxs_mins, 5, 5, 2, up(mins)).
table(up_maxc_s_mins, 5, 5, 2, up(mins)).
table(up_maxc_maxs_mins, 5, 5, 2, up(mins)).
table(up_s_maxs_mins, 5, 5, 2, up(mins)).
table(up_v_c_minc_mins, 5, 8, 3, up(mins)).
table(up_v_c_maxc_mins, 5, 12, 3, up(mins)).
table(up_v_c_s_mins, 5, 6, 3, up(mins)).
table(up_v_c_maxs_mins, 5, 6, 3, up(mins)).
table(up_v_minc_maxc_mins, 5, 7, 3, up(mins)).
table(up_v_minc_s_mins, 5, 6, 3, up(mins)).
table(up_v_minc_maxs_mins, 5, 6, 3, up(mins)).
table(up_v_maxc_s_mins, 5, 6, 3, up(mins)).
table(up_v_maxc_maxs_mins, 5, 6, 3, up(mins)).
table(up_v_s_maxs_mins, 5, 9, 3, up(mins)).
table(up_c_minc_maxc_mins, 5, 7, 3, up(mins)).
table(up_c_minc_maxs_mins, 5, 6, 3, up(mins)).
table(up_c_maxc_s_mins, 5, 6, 3, up(mins)).
table(up_c_maxc_maxs_mins, 5, 6, 3, up(mins)).
table(up_c_s_maxs_mins, 5, 6, 3, up(mins)).
table(up_minc_maxc_s_mins, 5, 7, 3, up(mins)).
table(up_minc_maxc_maxs_mins, 5, 7, 3, up(mins)).
table(up_minc_s_maxs_mins, 5, 6, 3, up(mins)).
table(up_maxc_s_maxs_mins, 5, 6, 3, up(mins)).
table(up_v_c_minc_maxc_mins, 5, 12, 4, up(mins)).
table(up_v_c_minc_s_mins, 5, 9, 4, up(mins)).
table(up_v_c_minc_maxs_mins, 5, 9, 4, up(mins)).
table(up_v_c_maxc_s_mins, 5, 13, 4, up(mins)).
table(up_v_c_maxc_maxs_mins, 5, 13, 4, up(mins)).
table(up_v_c_s_maxs_mins, 5, 10, 4, up(mins)).
table(up_v_minc_maxc_s_mins, 5, 8, 4, up(mins)).
table(up_v_minc_maxc_maxs_mins, 5, 8, 4, up(mins)).
table(up_v_minc_s_maxs_mins, 5, 10, 4, up(mins)).
table(up_v_maxc_s_maxs_mins, 5, 10, 4, up(mins)).
table(up_c_minc_maxc_s_mins, 5, 8, 4, up(mins)).
table(up_c_minc_maxc_maxs_mins, 5, 8, 4, up(mins)).
table(up_c_minc_s_maxs_mins, 5, 7, 4, up(mins)).
table(up_c_maxc_s_maxs_mins, 5, 8, 4, up(mins)).
table(up_minc_maxc_s_maxs_mins, 5, 8, 4, up(mins)).
table(up_v_c_minc_maxc_s_mins, 5, 13, 5, up(mins)).
table(up_v_c_minc_maxc_maxs_mins, 5, 13, 5, up(mins)).
table(up_v_c_minc_s_maxs_mins, 5, 13, 5, up(mins)).
table(up_v_c_maxc_s_maxs_mins, 5, 18, 5, up(mins)).
table(up_v_minc_maxc_s_maxs_mins, 5, 12, 5, up(mins)).
table(up_c_minc_maxc_s_maxs_mins, 5, 10, 5, up(mins)).
table(up_v_c_minc_maxc_s_maxs_mins, 5, 18, 6, up(mins)).
table(low_v_maxs, 5, 9, 1, low(maxs)).
table(low_c_maxs, 5, 6, 1, low(maxs)).
table(low_minc_maxs, 5, 6, 1, low(maxs)).
table(low_maxc_maxs, 5, 6, 1, low(maxs)).
table(low_s_maxs, 5, 9, 1, low(maxs)).
table(low_mins_maxs, 5, 4, 1, low(maxs)).
table(low_v_c_maxs, 5, 10, 2, low(maxs)).
table(low_v_minc_maxs, 5, 10, 2, low(maxs)).
table(low_v_maxc_maxs, 5, 10, 2, low(maxs)).
table(low_v_s_maxs, 5, 9, 2, low(maxs)).
table(low_v_mins_maxs, 5, 9, 2, low(maxs)).
table(low_c_minc_maxs, 5, 7, 2, low(maxs)).
table(low_c_maxc_maxs, 5, 7, 2, low(maxs)).
table(low_c_s_maxs, 5, 10, 2, low(maxs)).
table(low_c_mins_maxs, 5, 5, 2, low(maxs)).
table(low_minc_maxc_maxs, 5, 8, 2, low(maxs)).
table(low_minc_s_maxs, 5, 6, 2, low(maxs)).
table(low_minc_mins_maxs, 5, 5, 2, low(maxs)).
table(low_maxc_s_maxs, 5, 6, 2, low(maxs)).
table(low_maxc_mins_maxs, 5, 5, 2, low(maxs)).
table(low_s_mins_maxs, 5, 9, 2, low(maxs)).
table(low_v_c_minc_maxs, 5, 13, 3, low(maxs)).
table(low_v_c_maxc_maxs, 5, 18, 3, low(maxs)).
table(low_v_c_s_maxs, 5, 10, 3, low(maxs)).
table(low_v_c_mins_maxs, 5, 10, 3, low(maxs)).
table(low_v_minc_maxc_maxs, 5, 12, 3, low(maxs)).
table(low_v_minc_s_maxs, 5, 10, 3, low(maxs)).
table(low_v_minc_mins_maxs, 5, 10, 3, low(maxs)).
table(low_v_maxc_s_maxs, 5, 10, 3, low(maxs)).
table(low_v_maxc_mins_maxs, 5, 10, 3, low(maxs)).
table(low_v_s_mins_maxs, 5, 9, 3, low(maxs)).
table(low_c_minc_maxc_maxs, 5, 9, 3, low(maxs)).
table(low_c_minc_s_maxs, 5, 7, 3, low(maxs)).
table(low_c_minc_mins_maxs, 5, 7, 3, low(maxs)).
table(low_c_maxc_s_maxs, 5, 7, 3, low(maxs)).
table(low_c_maxc_mins_maxs, 5, 7, 3, low(maxs)).
table(low_c_s_mins_maxs, 5, 10, 3, low(maxs)).
table(low_minc_maxc_s_maxs, 5, 12, 3, low(maxs)).
table(low_minc_maxc_mins_maxs, 5, 7, 3, low(maxs)).
table(low_minc_s_mins_maxs, 5, 6, 3, low(maxs)).
table(low_maxc_s_mins_maxs, 5, 6, 3, low(maxs)).
table(low_v_c_minc_maxc_maxs, 5, 18, 4, low(maxs)).
table(low_v_c_minc_s_maxs, 5, 13, 4, low(maxs)).
table(low_v_c_minc_mins_maxs, 5, 13, 4, low(maxs)).
table(low_v_c_maxc_s_maxs, 5, 18, 4, low(maxs)).
table(low_v_c_maxc_mins_maxs, 5, 18, 4, low(maxs)).
table(low_v_c_s_mins_maxs, 5, 10, 4, low(maxs)).
table(low_v_minc_maxc_s_maxs, 5, 12, 4, low(maxs)).
table(low_v_minc_maxc_mins_maxs, 5, 12, 4, low(maxs)).
table(low_v_minc_s_mins_maxs, 5, 10, 4, low(maxs)).
table(low_v_maxc_s_mins_maxs, 5, 10, 4, low(maxs)).
table(low_c_minc_maxc_s_maxs, 5, 10, 4, low(maxs)).
table(low_c_minc_maxc_mins_maxs, 5, 9, 4, low(maxs)).
table(low_c_minc_s_mins_maxs, 5, 7, 4, low(maxs)).
table(low_c_maxc_s_mins_maxs, 5, 8, 4, low(maxs)).
table(low_minc_maxc_s_mins_maxs, 5, 12, 4, low(maxs)).
table(low_v_c_minc_maxc_s_maxs, 5, 18, 5, low(maxs)).
table(low_v_c_minc_maxc_mins_maxs, 5, 18, 5, low(maxs)).
table(low_v_c_minc_s_mins_maxs, 5, 13, 5, low(maxs)).
table(low_v_c_maxc_s_mins_maxs, 5, 18, 5, low(maxs)).
table(low_v_minc_maxc_s_mins_maxs, 5, 12, 5, low(maxs)).
table(low_c_minc_maxc_s_mins_maxs, 5, 10, 5, low(maxs)).
table(low_v_c_minc_maxc_s_mins_maxs, 5, 18, 6, low(maxs)).
table(up_v_maxs, 5, 18, 1, up(maxs)).
table(up_maxc_maxs, 5, 4, 1, up(maxs)).
table(up_v_c_maxs, 5, 18, 2, up(maxs)).
table(up_v_minc_maxs, 5, 12, 2, up(maxs)).
table(up_v_maxc_maxs, 5, 5, 2, up(maxs)).
table(up_v_s_maxs, 5, 9, 2, up(maxs)).
table(up_v_mins_maxs, 5, 9, 2, up(maxs)).
table(up_c_maxc_maxs, 5, 5, 2, up(maxs)).
table(up_minc_maxc_maxs, 5, 6, 2, up(maxs)).
table(up_maxc_s_maxs, 5, 5, 2, up(maxs)).
table(up_maxc_mins_maxs, 5, 5, 2, up(maxs)).
table(up_v_c_minc_maxs, 5, 12, 3, up(maxs)).
table(up_v_c_maxc_maxs, 5, 12, 3, up(maxs)).
table(up_v_c_s_maxs, 5, 10, 3, up(maxs)).
table(up_v_c_mins_maxs, 5, 18, 3, up(maxs)).
table(up_v_minc_maxc_maxs, 5, 7, 3, up(maxs)).
table(up_v_minc_s_maxs, 5, 10, 3, up(maxs)).
table(up_v_minc_mins_maxs, 5, 18, 3, up(maxs)).
table(up_v_maxc_s_maxs, 5, 10, 3, up(maxs)).
table(up_v_maxc_mins_maxs, 5, 6, 3, up(maxs)).
table(up_v_s_mins_maxs, 5, 9, 3, up(maxs)).
table(up_c_minc_maxc_maxs, 5, 7, 3, up(maxs)).
table(up_c_maxc_s_maxs, 5, 7, 3, up(maxs)).
table(up_c_maxc_mins_maxs, 5, 6, 3, up(maxs)).
table(up_minc_maxc_s_maxs, 5, 8, 3, up(maxs)).
table(up_minc_maxc_mins_maxs, 5, 7, 3, up(maxs)).
table(up_maxc_s_mins_maxs, 5, 6, 3, up(maxs)).
table(up_v_c_minc_maxc_maxs, 5, 12, 4, up(maxs)).
table(up_v_c_minc_s_maxs, 5, 13, 4, up(maxs)).
table(up_v_c_minc_mins_maxs, 5, 18, 4, up(maxs)).
table(up_v_c_maxc_s_maxs, 5, 18, 4, up(maxs)).
table(up_v_c_maxc_mins_maxs, 5, 13, 4, up(maxs)).
table(up_v_c_s_mins_maxs, 5, 10, 4, up(maxs)).
table(up_v_minc_maxc_s_maxs, 5, 12, 4, up(maxs)).
table(up_v_minc_maxc_mins_maxs, 5, 8, 4, up(maxs)).
table(up_v_minc_s_mins_maxs, 5, 10, 4, up(maxs)).
table(up_v_maxc_s_mins_maxs, 5, 10, 4, up(maxs)).
table(up_c_minc_maxc_s_maxs, 5, 10, 4, up(maxs)).
table(up_c_minc_maxc_mins_maxs, 5, 8, 4, up(maxs)).
table(up_c_maxc_s_mins_maxs, 5, 10, 4, up(maxs)).
table(up_minc_maxc_s_mins_maxs, 5, 8, 4, up(maxs)).
table(up_v_c_minc_maxc_s_maxs, 5, 18, 5, up(maxs)).
table(up_v_c_minc_maxc_mins_maxs, 5, 13, 5, up(maxs)).
table(up_v_c_minc_s_mins_maxs, 5, 13, 5, up(maxs)).
table(up_v_c_maxc_s_mins_maxs, 5, 18, 5, up(maxs)).
table(up_v_minc_maxc_s_mins_maxs, 5, 12, 5, up(maxs)).
table(up_c_minc_maxc_s_mins_maxs, 5, 10, 5, up(maxs)).
table(up_v_c_minc_maxc_s_mins_maxs, 5, 18, 6, up(maxs)).
table(low_v_mina, 5, 13, 1, low(mina)).
table(low_c_mina, 5, 7, 1, low(mina)).
table(low_minc_mina, 5, 19, 1, low(mina)).
table(low_maxc_mina, 5, 19, 1, low(mina)).
table(low_s_mina, 5, 13, 1, low(mina)).
table(low_mins_mina, 5, 11, 1, low(mina)).
table(low_maxs_mina, 5, 11, 1, low(mina)).
table(low_v_c_mina, 5, 19, 2, low(mina)).
table(low_v_minc_mina, 5, 15, 2, low(mina)).
table(low_v_maxc_mina, 5, 19, 2, low(mina)).
table(low_v_s_mina, 5, 6, 2, low(mina)).
table(low_v_mins_mina, 5, 13, 2, low(mina)).
table(low_v_maxs_mina, 5, 13, 2, low(mina)).
table(low_c_minc_mina, 5, 8, 2, low(mina)).
table(low_c_maxc_mina, 5, 8, 2, low(mina)).
table(low_c_s_mina, 5, 19, 2, low(mina)).
table(low_c_mins_mina, 5, 7, 2, low(mina)).
table(low_c_maxs_mina, 5, 7, 2, low(mina)).
table(low_minc_maxc_mina, 5, 19, 2, low(mina)).
table(low_minc_s_mina, 5, 15, 2, low(mina)).
table(low_minc_mins_mina, 5, 19, 2, low(mina)).
table(low_minc_maxs_mina, 5, 19, 2, low(mina)).
table(low_maxc_s_mina, 5, 19, 2, low(mina)).
table(low_maxc_mins_mina, 5, 19, 2, low(mina)).
table(low_maxc_maxs_mina, 5, 19, 2, low(mina)).
table(low_s_mins_mina, 5, 13, 2, low(mina)).
table(low_s_maxs_mina, 5, 13, 2, low(mina)).
table(low_mins_maxs_mina, 5, 6, 2, low(mina)).
table(low_v_c_minc_mina, 5, 19, 3, low(mina)).
table(low_v_c_maxc_mina, 5, 19, 3, low(mina)).
table(low_v_c_s_mina, 5, 7, 3, low(mina)).
table(low_v_c_mins_mina, 5, 19, 3, low(mina)).
table(low_v_c_maxs_mina, 5, 19, 3, low(mina)).
table(low_v_minc_maxc_mina, 5, 19, 3, low(mina)).
table(low_v_minc_s_mina, 5, 9, 3, low(mina)).
table(low_v_minc_mins_mina, 5, 15, 3, low(mina)).
table(low_v_minc_maxs_mina, 5, 15, 3, low(mina)).
table(low_v_maxc_s_mina, 5, 19, 3, low(mina)).
table(low_v_maxc_mins_mina, 5, 19, 3, low(mina)).
table(low_v_maxc_maxs_mina, 5, 19, 3, low(mina)).
table(low_v_s_mins_mina, 5, 13, 3, low(mina)).
table(low_v_s_maxs_mina, 5, 13, 3, low(mina)).
table(low_v_mins_maxs_mina, 5, 13, 3, low(mina)).
table(low_c_minc_maxc_mina, 5, 10, 3, low(mina)).
table(low_c_minc_s_mina, 5, 19, 3, low(mina)).
table(low_c_minc_mins_mina, 5, 8, 3, low(mina)).
table(low_c_minc_maxs_mina, 5, 8, 3, low(mina)).
table(low_c_maxc_s_mina, 5, 12, 3, low(mina)).
table(low_c_maxc_mins_mina, 5, 8, 3, low(mina)).
table(low_c_maxc_maxs_mina, 5, 8, 3, low(mina)).
table(low_c_s_mins_mina, 5, 19, 3, low(mina)).
table(low_c_s_maxs_mina, 5, 19, 3, low(mina)).
table(low_c_mins_maxs_mina, 5, 7, 3, low(mina)).
table(low_minc_maxc_s_mina, 5, 19, 3, low(mina)).
table(low_minc_maxc_mins_mina, 5, 19, 3, low(mina)).
table(low_minc_maxc_maxs_mina, 5, 19, 3, low(mina)).
table(low_minc_s_mins_mina, 5, 15, 3, low(mina)).
table(low_minc_s_maxs_mina, 5, 15, 3, low(mina)).
table(low_minc_mins_maxs_mina, 5, 19, 3, low(mina)).
table(low_maxc_s_mins_mina, 5, 19, 3, low(mina)).
table(low_maxc_s_maxs_mina, 5, 19, 3, low(mina)).
table(low_maxc_mins_maxs_mina, 5, 9, 3, low(mina)).
table(low_s_mins_maxs_mina, 5, 13, 3, low(mina)).
table(low_v_c_minc_maxc_mina, 5, 19, 4, low(mina)).
table(low_v_c_minc_s_mina, 5, 11, 4, low(mina)).
table(low_v_c_minc_mins_mina, 5, 19, 4, low(mina)).
table(low_v_c_minc_maxs_mina, 5, 19, 4, low(mina)).
table(low_v_c_maxc_s_mina, 5, 19, 4, low(mina)).
table(low_v_c_maxc_mins_mina, 5, 19, 4, low(mina)).
table(low_v_c_maxc_maxs_mina, 5, 19, 4, low(mina)).
table(low_v_c_s_mins_mina, 5, 11, 4, low(mina)).
table(low_v_c_s_maxs_mina, 5, 19, 4, low(mina)).
table(low_v_c_mins_maxs_mina, 5, 19, 4, low(mina)).
table(low_v_minc_maxc_s_mina, 5, 19, 4, low(mina)).
table(low_v_minc_maxc_mins_mina, 5, 19, 4, low(mina)).
table(low_v_minc_maxc_maxs_mina, 5, 19, 4, low(mina)).
table(low_v_minc_s_mins_mina, 5, 9, 4, low(mina)).
table(low_v_minc_s_maxs_mina, 5, 15, 4, low(mina)).
table(low_v_minc_mins_maxs_mina, 5, 15, 4, low(mina)).
table(low_v_maxc_s_mins_mina, 5, 19, 4, low(mina)).
table(low_v_maxc_s_maxs_mina, 5, 19, 4, low(mina)).
table(low_v_maxc_mins_maxs_mina, 5, 19, 4, low(mina)).
table(low_v_s_mins_maxs_mina, 5, 13, 4, low(mina)).
table(low_c_minc_maxc_s_mina, 5, 19, 4, low(mina)).
table(low_c_minc_maxc_mins_mina, 5, 10, 4, low(mina)).
table(low_c_minc_maxc_maxs_mina, 5, 10, 4, low(mina)).
table(low_c_minc_s_mins_mina, 5, 19, 4, low(mina)).
table(low_c_minc_s_maxs_mina, 5, 19, 4, low(mina)).
table(low_c_minc_mins_maxs_mina, 5, 8, 4, low(mina)).
table(low_c_maxc_s_mins_mina, 5, 12, 4, low(mina)).
table(low_c_maxc_s_maxs_mina, 5, 19, 4, low(mina)).
table(low_c_maxc_mins_maxs_mina, 5, 8, 4, low(mina)).
table(low_c_s_mins_maxs_mina, 5, 19, 4, low(mina)).
table(low_minc_maxc_s_mins_mina, 5, 19, 4, low(mina)).
table(low_minc_maxc_s_maxs_mina, 5, 19, 4, low(mina)).
table(low_minc_maxc_mins_maxs_mina, 5, 19, 4, low(mina)).
table(low_minc_s_mins_maxs_mina, 5, 15, 4, low(mina)).
table(low_maxc_s_mins_maxs_mina, 5, 19, 4, low(mina)).
table(low_v_c_minc_maxc_s_mina, 5, 19, 5, low(mina)).
table(low_v_c_minc_maxc_mins_mina, 5, 19, 5, low(mina)).
table(low_v_c_minc_maxc_maxs_mina, 5, 19, 5, low(mina)).
table(low_v_c_minc_s_mins_mina, 5, 11, 5, low(mina)).
table(low_v_c_minc_s_maxs_mina, 5, 19, 5, low(mina)).
table(low_v_c_minc_mins_maxs_mina, 5, 19, 5, low(mina)).
table(low_v_c_maxc_s_mins_mina, 5, 19, 5, low(mina)).
table(low_v_c_maxc_s_maxs_mina, 5, 19, 5, low(mina)).
table(low_v_c_maxc_mins_maxs_mina, 5, 19, 5, low(mina)).
table(low_v_c_s_mins_maxs_mina, 5, 19, 5, low(mina)).
table(low_v_minc_maxc_s_mins_mina, 5, 19, 5, low(mina)).
table(low_v_minc_maxc_s_maxs_mina, 5, 19, 5, low(mina)).
table(low_v_minc_maxc_mins_maxs_mina, 5, 19, 5, low(mina)).
table(low_v_minc_s_mins_maxs_mina, 5, 15, 5, low(mina)).
table(low_v_maxc_s_mins_maxs_mina, 5, 19, 5, low(mina)).
table(low_c_minc_maxc_s_mins_mina, 5, 19, 5, low(mina)).
table(low_c_minc_maxc_s_maxs_mina, 5, 19, 5, low(mina)).
table(low_c_minc_maxc_mins_maxs_mina, 5, 10, 5, low(mina)).
table(low_c_minc_s_mins_maxs_mina, 5, 19, 5, low(mina)).
table(low_c_maxc_s_mins_maxs_mina, 5, 19, 5, low(mina)).
table(low_minc_maxc_s_mins_maxs_mina, 5, 19, 5, low(mina)).
table(low_v_c_minc_maxc_s_mins_mina, 5, 19, 6, low(mina)).
table(low_v_c_minc_maxc_s_maxs_mina, 5, 19, 6, low(mina)).
table(low_v_c_minc_maxc_mins_maxs_mina, 5, 19, 6, low(mina)).
table(low_v_c_minc_s_mins_maxs_mina, 5, 19, 6, low(mina)).
table(low_v_c_maxc_s_mins_maxs_mina, 5, 19, 6, low(mina)).
table(low_v_minc_maxc_s_mins_maxs_mina, 5, 19, 6, low(mina)).
table(low_c_minc_maxc_s_mins_maxs_mina, 5, 19, 6, low(mina)).
table(low_v_c_minc_maxc_s_mins_maxs_mina, 5, 19, 7, low(mina)).
table(up_v_maxa, 5, 19, 1, up(maxa)).
table(up_v_c_maxa, 5, 19, 2, up(maxa)).
table(up_v_minc_maxa, 5, 19, 2, up(maxa)).
table(up_v_maxc_maxa, 5, 19, 2, up(maxa)).
table(up_v_s_maxa, 5, 19, 2, up(maxa)).
table(up_v_mins_maxa, 5, 19, 2, up(maxa)).
table(up_v_maxs_maxa, 5, 19, 2, up(maxa)).
table(up_c_maxc_maxa, 5, 19, 2, up(maxa)).
table(up_maxc_s_maxa, 5, 19, 2, up(maxa)).
table(up_s_maxs_maxa, 5, 19, 2, up(maxa)).
table(up_v_c_minc_maxa, 5, 19, 3, up(maxa)).
table(up_v_c_maxc_maxa, 5, 19, 3, up(maxa)).
table(up_v_c_s_maxa, 5, 19, 3, up(maxa)).
table(up_v_c_mins_maxa, 5, 19, 3, up(maxa)).
table(up_v_c_maxs_maxa, 5, 7, 3, up(maxa)).
table(up_v_minc_maxc_maxa, 5, 19, 3, up(maxa)).
table(up_v_minc_s_maxa, 5, 19, 3, up(maxa)).
table(up_v_minc_mins_maxa, 5, 19, 3, up(maxa)).
table(up_v_minc_maxs_maxa, 5, 19, 3, up(maxa)).
table(up_v_maxc_s_maxa, 5, 19, 3, up(maxa)).
table(up_v_maxc_mins_maxa, 5, 19, 3, up(maxa)).
table(up_v_maxc_maxs_maxa, 5, 19, 3, up(maxa)).
table(up_v_s_mins_maxa, 5, 19, 3, up(maxa)).
table(up_v_s_maxs_maxa, 5, 19, 3, up(maxa)).
table(up_v_mins_maxs_maxa, 5, 19, 3, up(maxa)).
table(up_c_minc_maxc_maxa, 5, 19, 3, up(maxa)).
table(up_c_maxc_s_maxa, 5, 19, 3, up(maxa)).
table(up_c_maxc_mins_maxa, 5, 19, 3, up(maxa)).
table(up_c_maxc_maxs_maxa, 5, 19, 3, up(maxa)).
table(up_c_s_maxs_maxa, 5, 19, 3, up(maxa)).
table(up_minc_maxc_s_maxa, 5, 19, 3, up(maxa)).
table(up_minc_s_maxs_maxa, 5, 19, 3, up(maxa)).
table(up_maxc_s_mins_maxa, 5, 19, 3, up(maxa)).
table(up_maxc_s_maxs_maxa, 5, 19, 3, up(maxa)).
table(up_s_mins_maxs_maxa, 5, 19, 3, up(maxa)).
table(up_v_c_minc_maxc_maxa, 5, 19, 4, up(maxa)).
table(up_v_c_minc_s_maxa, 5, 19, 4, up(maxa)).
table(up_v_c_minc_mins_maxa, 5, 19, 4, up(maxa)).
table(up_v_c_minc_maxs_maxa, 5, 11, 4, up(maxa)).
table(up_v_c_maxc_s_maxa, 5, 19, 4, up(maxa)).
table(up_v_c_maxc_mins_maxa, 5, 19, 4, up(maxa)).
table(up_v_c_maxc_maxs_maxa, 5, 19, 4, up(maxa)).
table(up_v_c_s_mins_maxa, 5, 19, 4, up(maxa)).
table(up_v_c_s_maxs_maxa, 5, 19, 4, up(maxa)).
table(up_v_c_mins_maxs_maxa, 5, 11, 4, up(maxa)).
table(up_v_minc_maxc_s_maxa, 5, 19, 4, up(maxa)).
table(up_v_minc_maxc_mins_maxa, 5, 19, 4, up(maxa)).
table(up_v_minc_maxc_maxs_maxa, 5, 19, 4, up(maxa)).
table(up_v_minc_s_mins_maxa, 5, 19, 4, up(maxa)).
table(up_v_minc_s_maxs_maxa, 5, 19, 4, up(maxa)).
table(up_v_minc_mins_maxs_maxa, 5, 19, 4, up(maxa)).
table(up_v_maxc_s_mins_maxa, 5, 19, 4, up(maxa)).
table(up_v_maxc_s_maxs_maxa, 5, 19, 4, up(maxa)).
table(up_v_maxc_mins_maxs_maxa, 5, 19, 4, up(maxa)).
table(up_v_s_mins_maxs_maxa, 5, 19, 4, up(maxa)).
table(up_c_minc_maxc_s_maxa, 5, 19, 4, up(maxa)).
table(up_c_minc_maxc_mins_maxa, 5, 19, 4, up(maxa)).
table(up_c_minc_maxc_maxs_maxa, 5, 19, 4, up(maxa)).
table(up_c_minc_s_maxs_maxa, 5, 19, 4, up(maxa)).
table(up_c_maxc_s_mins_maxa, 5, 19, 4, up(maxa)).
table(up_c_maxc_s_maxs_maxa, 5, 19, 4, up(maxa)).
table(up_c_maxc_mins_maxs_maxa, 5, 19, 4, up(maxa)).
table(up_c_s_mins_maxs_maxa, 5, 19, 4, up(maxa)).
table(up_minc_maxc_s_mins_maxa, 5, 19, 4, up(maxa)).
table(up_minc_maxc_s_maxs_maxa, 5, 19, 4, up(maxa)).
table(up_minc_s_mins_maxs_maxa, 5, 19, 4, up(maxa)).
table(up_maxc_s_mins_maxs_maxa, 5, 19, 4, up(maxa)).
table(up_v_c_minc_maxc_s_maxa, 5, 19, 5, up(maxa)).
table(up_v_c_minc_maxc_mins_maxa, 5, 19, 5, up(maxa)).
table(up_v_c_minc_maxc_maxs_maxa, 5, 19, 5, up(maxa)).
table(up_v_c_minc_s_mins_maxa, 5, 19, 5, up(maxa)).
table(up_v_c_minc_s_maxs_maxa, 5, 19, 5, up(maxa)).
table(up_v_c_minc_mins_maxs_maxa, 5, 11, 5, up(maxa)).
table(up_v_c_maxc_s_mins_maxa, 5, 19, 5, up(maxa)).
table(up_v_c_maxc_s_maxs_maxa, 5, 19, 5, up(maxa)).
table(up_v_c_maxc_mins_maxs_maxa, 5, 19, 5, up(maxa)).
table(up_v_c_s_mins_maxs_maxa, 5, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_mins_maxa, 5, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_maxs_maxa, 5, 19, 5, up(maxa)).
table(up_v_minc_maxc_mins_maxs_maxa, 5, 19, 5, up(maxa)).
table(up_v_minc_s_mins_maxs_maxa, 5, 19, 5, up(maxa)).
table(up_v_maxc_s_mins_maxs_maxa, 5, 19, 5, up(maxa)).
table(up_c_minc_maxc_s_mins_maxa, 5, 19, 5, up(maxa)).
table(up_c_minc_maxc_s_maxs_maxa, 5, 19, 5, up(maxa)).
table(up_c_minc_maxc_mins_maxs_maxa, 5, 19, 5, up(maxa)).
table(up_c_minc_s_mins_maxs_maxa, 5, 19, 5, up(maxa)).
table(up_c_maxc_s_mins_maxs_maxa, 5, 19, 5, up(maxa)).
table(up_minc_maxc_s_mins_maxs_maxa, 5, 19, 5, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxa, 5, 19, 6, up(maxa)).
table(up_v_c_minc_maxc_s_maxs_maxa, 5, 19, 6, up(maxa)).
table(up_v_c_minc_maxc_mins_maxs_maxa, 5, 19, 6, up(maxa)).
table(up_v_c_minc_s_mins_maxs_maxa, 5, 19, 6, up(maxa)).
table(up_v_c_maxc_s_mins_maxs_maxa, 5, 19, 6, up(maxa)).
table(up_v_minc_maxc_s_mins_maxs_maxa, 5, 19, 6, up(maxa)).
table(up_c_minc_maxc_s_mins_maxs_maxa, 5, 19, 6, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxs_maxa, 5, 19, 7, up(maxa)).
table(low_c_v, 6, 18, 1, low(v)).
table(low_minc_v, 6, 11, 1, low(v)).
table(low_maxc_v, 6, 11, 1, low(v)).
table(low_s_v, 6, 7, 1, low(v)).
table(low_mins_v, 6, 18, 1, low(v)).
table(low_maxs_v, 6, 18, 1, low(v)).
table(low_c_minc_v, 6, 11, 2, low(v)).
table(low_c_maxc_v, 6, 11, 2, low(v)).
table(low_c_s_v, 6, 8, 2, low(v)).
table(low_c_mins_v, 6, 18, 2, low(v)).
table(low_c_maxs_v, 6, 18, 2, low(v)).
table(low_minc_maxc_v, 6, 11, 2, low(v)).
table(low_minc_s_v, 6, 3, 2, low(v)).
table(low_minc_mins_v, 6, 12, 2, low(v)).
table(low_minc_maxs_v, 6, 12, 2, low(v)).
table(low_maxc_s_v, 6, 3, 2, low(v)).
table(low_maxc_mins_v, 6, 12, 2, low(v)).
table(low_maxc_maxs_v, 6, 13, 2, low(v)).
table(low_s_mins_v, 6, 7, 2, low(v)).
table(low_s_maxs_v, 6, 7, 2, low(v)).
table(low_mins_maxs_v, 6, 12, 2, low(v)).
table(low_c_minc_maxc_v, 6, 11, 3, low(v)).
table(low_c_minc_s_v, 6, 6, 3, low(v)).
table(low_c_minc_mins_v, 6, 12, 3, low(v)).
table(low_c_minc_maxs_v, 6, 12, 3, low(v)).
table(low_c_maxc_s_v, 6, 13, 3, low(v)).
table(low_c_maxc_mins_v, 6, 12, 3, low(v)).
table(low_c_maxc_maxs_v, 6, 13, 3, low(v)).
table(low_c_s_mins_v, 6, 8, 3, low(v)).
table(low_c_s_maxs_v, 6, 8, 3, low(v)).
table(low_c_mins_maxs_v, 6, 18, 3, low(v)).
table(low_minc_maxc_s_v, 6, 5, 3, low(v)).
table(low_minc_maxc_mins_v, 6, 12, 3, low(v)).
table(low_minc_maxc_maxs_v, 6, 12, 3, low(v)).
table(low_minc_s_mins_v, 6, 4, 3, low(v)).
table(low_minc_s_maxs_v, 6, 8, 3, low(v)).
table(low_minc_mins_maxs_v, 6, 13, 3, low(v)).
table(low_maxc_s_mins_v, 6, 4, 3, low(v)).
table(low_maxc_s_maxs_v, 6, 8, 3, low(v)).
table(low_maxc_mins_maxs_v, 6, 14, 3, low(v)).
table(low_s_mins_maxs_v, 6, 7, 3, low(v)).
table(low_c_minc_maxc_s_v, 6, 12, 4, low(v)).
table(low_c_minc_maxc_mins_v, 6, 12, 4, low(v)).
table(low_c_minc_maxc_maxs_v, 6, 12, 4, low(v)).
table(low_c_minc_s_mins_v, 6, 7, 4, low(v)).
table(low_c_minc_s_maxs_v, 6, 11, 4, low(v)).
table(low_c_minc_mins_maxs_v, 6, 13, 4, low(v)).
table(low_c_maxc_s_mins_v, 6, 14, 4, low(v)).
table(low_c_maxc_s_maxs_v, 6, 17, 4, low(v)).
table(low_c_maxc_mins_maxs_v, 6, 14, 4, low(v)).
table(low_c_s_mins_maxs_v, 6, 8, 4, low(v)).
table(low_minc_maxc_s_mins_v, 6, 6, 4, low(v)).
table(low_minc_maxc_s_maxs_v, 6, 10, 4, low(v)).
table(low_minc_maxc_mins_maxs_v, 6, 13, 4, low(v)).
table(low_minc_s_mins_maxs_v, 6, 8, 4, low(v)).
table(low_maxc_s_mins_maxs_v, 6, 8, 4, low(v)).
table(low_c_minc_maxc_s_mins_v, 6, 13, 5, low(v)).
table(low_c_minc_maxc_s_maxs_v, 6, 17, 5, low(v)).
table(low_c_minc_maxc_mins_maxs_v, 6, 13, 5, low(v)).
table(low_c_minc_s_mins_maxs_v, 6, 11, 5, low(v)).
table(low_c_maxc_s_mins_maxs_v, 6, 17, 5, low(v)).
table(low_minc_maxc_s_mins_maxs_v, 6, 10, 5, low(v)).
table(low_c_minc_maxc_s_mins_maxs_v, 6, 17, 6, low(v)).
table(up_c_maxc_v, 6, 11, 2, up(v)).
table(up_maxc_s_v, 6, 3, 2, up(v)).
table(up_s_maxs_v, 6, 7, 2, up(v)).
table(up_c_minc_maxc_v, 6, 11, 3, up(v)).
table(up_c_maxc_s_v, 6, 12, 3, up(v)).
table(up_c_maxc_mins_v, 6, 12, 3, up(v)).
table(up_c_maxc_maxs_v, 6, 12, 3, up(v)).
table(up_c_s_maxs_v, 6, 8, 3, up(v)).
table(up_minc_maxc_s_v, 6, 5, 3, up(v)).
table(up_minc_s_maxs_v, 6, 8, 3, up(v)).
table(up_maxc_s_mins_v, 6, 4, 3, up(v)).
table(up_maxc_s_maxs_v, 6, 8, 3, up(v)).
table(up_s_mins_maxs_v, 6, 7, 3, up(v)).
table(up_c_minc_maxc_s_v, 6, 12, 4, up(v)).
table(up_c_minc_maxc_mins_v, 6, 12, 4, up(v)).
table(up_c_minc_maxc_maxs_v, 6, 12, 4, up(v)).
table(up_c_minc_s_maxs_v, 6, 11, 4, up(v)).
table(up_c_maxc_s_mins_v, 6, 13, 4, up(v)).
table(up_c_maxc_s_maxs_v, 6, 17, 4, up(v)).
table(up_c_maxc_mins_maxs_v, 6, 13, 4, up(v)).
table(up_c_s_mins_maxs_v, 6, 8, 4, up(v)).
table(up_minc_maxc_s_mins_v, 6, 6, 4, up(v)).
table(up_minc_maxc_s_maxs_v, 6, 10, 4, up(v)).
table(up_minc_s_mins_maxs_v, 6, 8, 4, up(v)).
table(up_maxc_s_mins_maxs_v, 6, 8, 4, up(v)).
table(up_c_minc_maxc_s_mins_v, 6, 13, 5, up(v)).
table(up_c_minc_maxc_s_maxs_v, 6, 17, 5, up(v)).
table(up_c_minc_maxc_mins_maxs_v, 6, 13, 5, up(v)).
table(up_c_minc_s_mins_maxs_v, 6, 11, 5, up(v)).
table(up_c_maxc_s_mins_maxs_v, 6, 17, 5, up(v)).
table(up_minc_maxc_s_mins_maxs_v, 6, 10, 5, up(v)).
table(up_c_minc_maxc_s_mins_maxs_v, 6, 17, 6, up(v)).
table(low_v_c, 6, 11, 1, low(c)).
table(low_minc_c, 6, 11, 1, low(c)).
table(low_maxc_c, 6, 11, 1, low(c)).
table(low_s_c, 6, 8, 1, low(c)).
table(low_mins_c, 6, 8, 1, low(c)).
table(low_maxs_c, 6, 8, 1, low(c)).
table(low_v_minc_c, 6, 11, 2, low(c)).
table(low_v_maxc_c, 6, 11, 2, low(c)).
table(low_v_s_c, 6, 13, 2, low(c)).
table(low_v_mins_c, 6, 12, 2, low(c)).
table(low_v_maxs_c, 6, 13, 2, low(c)).
table(low_minc_maxc_c, 6, 11, 2, low(c)).
table(low_minc_s_c, 6, 5, 2, low(c)).
table(low_minc_mins_c, 6, 12, 2, low(c)).
table(low_minc_maxs_c, 6, 9, 2, low(c)).
table(low_maxc_s_c, 6, 5, 2, low(c)).
table(low_maxc_mins_c, 6, 12, 2, low(c)).
table(low_maxc_maxs_c, 6, 13, 2, low(c)).
table(low_s_mins_c, 6, 10, 2, low(c)).
table(low_s_maxs_c, 6, 11, 2, low(c)).
table(low_mins_maxs_c, 6, 9, 2, low(c)).
table(low_v_minc_maxc_c, 6, 11, 3, low(c)).
table(low_v_minc_s_c, 6, 12, 3, low(c)).
table(low_v_minc_mins_c, 6, 12, 3, low(c)).
table(low_v_minc_maxs_c, 6, 12, 3, low(c)).
table(low_v_maxc_s_c, 6, 12, 3, low(c)).
table(low_v_maxc_mins_c, 6, 12, 3, low(c)).
table(low_v_maxc_maxs_c, 6, 12, 3, low(c)).
table(low_v_s_mins_c, 6, 14, 3, low(c)).
table(low_v_s_maxs_c, 6, 18, 3, low(c)).
table(low_v_mins_maxs_c, 6, 14, 3, low(c)).
table(low_minc_maxc_s_c, 6, 5, 3, low(c)).
table(low_minc_maxc_mins_c, 6, 12, 3, low(c)).
table(low_minc_maxc_maxs_c, 6, 12, 3, low(c)).
table(low_minc_s_mins_c, 6, 6, 3, low(c)).
table(low_minc_s_maxs_c, 6, 7, 3, low(c)).
table(low_minc_mins_maxs_c, 6, 7, 3, low(c)).
table(low_maxc_s_mins_c, 6, 6, 3, low(c)).
table(low_maxc_s_maxs_c, 6, 6, 3, low(c)).
table(low_maxc_mins_maxs_c, 6, 6, 3, low(c)).
table(low_s_mins_maxs_c, 6, 12, 3, low(c)).
table(low_v_minc_maxc_s_c, 6, 12, 4, low(c)).
table(low_v_minc_maxc_mins_c, 6, 12, 4, low(c)).
table(low_v_minc_maxc_maxs_c, 6, 12, 4, low(c)).
table(low_v_minc_s_mins_c, 6, 13, 4, low(c)).
table(low_v_minc_s_maxs_c, 6, 17, 4, low(c)).
table(low_v_minc_mins_maxs_c, 6, 13, 4, low(c)).
table(low_v_maxc_s_mins_c, 6, 13, 4, low(c)).
table(low_v_maxc_s_maxs_c, 6, 17, 4, low(c)).
table(low_v_maxc_mins_maxs_c, 6, 13, 4, low(c)).
table(low_v_s_mins_maxs_c, 6, 18, 4, low(c)).
table(low_minc_maxc_s_mins_c, 6, 6, 4, low(c)).
table(low_minc_maxc_s_maxs_c, 6, 7, 4, low(c)).
table(low_minc_maxc_mins_maxs_c, 6, 13, 4, low(c)).
table(low_minc_s_mins_maxs_c, 6, 8, 4, low(c)).
table(low_maxc_s_mins_maxs_c, 6, 7, 4, low(c)).
table(low_v_minc_maxc_s_mins_c, 6, 13, 5, low(c)).
table(low_v_minc_maxc_s_maxs_c, 6, 17, 5, low(c)).
table(low_v_minc_maxc_mins_maxs_c, 6, 13, 5, low(c)).
table(low_v_minc_s_mins_maxs_c, 6, 17, 5, low(c)).
table(low_v_maxc_s_mins_maxs_c, 6, 17, 5, low(c)).
table(low_minc_maxc_s_mins_maxs_c, 6, 7, 5, low(c)).
table(low_v_minc_maxc_s_mins_maxs_c, 6, 17, 6, low(c)).
table(up_v_c, 6, 18, 1, up(c)).
table(up_s_c, 6, 3, 1, up(c)).
table(up_v_minc_c, 6, 11, 2, up(c)).
table(up_v_maxc_c, 6, 11, 2, up(c)).
table(up_v_s_c, 6, 4, 2, up(c)).
table(up_v_mins_c, 6, 18, 2, up(c)).
table(up_v_maxs_c, 6, 18, 2, up(c)).
table(up_minc_s_c, 6, 3, 2, up(c)).
table(up_maxc_s_c, 6, 4, 2, up(c)).
table(up_s_mins_c, 6, 5, 2, up(c)).
table(up_s_maxs_c, 6, 5, 2, up(c)).
table(up_v_minc_maxc_c, 6, 11, 3, up(c)).
table(up_v_minc_s_c, 6, 6, 3, up(c)).
table(up_v_minc_mins_c, 6, 12, 3, up(c)).
table(up_v_minc_maxs_c, 6, 12, 3, up(c)).
table(up_v_maxc_s_c, 6, 13, 3, up(c)).
table(up_v_maxc_mins_c, 6, 12, 3, up(c)).
table(up_v_maxc_maxs_c, 6, 13, 3, up(c)).
table(up_v_s_mins_c, 6, 8, 3, up(c)).
table(up_v_s_maxs_c, 6, 18, 3, up(c)).
table(up_v_mins_maxs_c, 6, 18, 3, up(c)).
table(up_minc_maxc_s_c, 6, 5, 3, up(c)).
table(up_minc_s_mins_c, 6, 4, 3, up(c)).
table(up_minc_s_maxs_c, 6, 6, 3, up(c)).
table(up_maxc_s_mins_c, 6, 7, 3, up(c)).
table(up_maxc_s_maxs_c, 6, 5, 3, up(c)).
table(up_s_mins_maxs_c, 6, 8, 3, up(c)).
table(up_v_minc_maxc_s_c, 6, 12, 4, up(c)).
table(up_v_minc_maxc_mins_c, 6, 12, 4, up(c)).
table(up_v_minc_maxc_maxs_c, 6, 12, 4, up(c)).
table(up_v_minc_s_mins_c, 6, 7, 4, up(c)).
table(up_v_minc_s_maxs_c, 6, 18, 4, up(c)).
table(up_v_minc_mins_maxs_c, 6, 13, 4, up(c)).
table(up_v_maxc_s_mins_c, 6, 14, 4, up(c)).
table(up_v_maxc_s_maxs_c, 6, 18, 4, up(c)).
table(up_v_maxc_mins_maxs_c, 6, 14, 4, up(c)).
table(up_v_s_mins_maxs_c, 6, 18, 4, up(c)).
table(up_minc_maxc_s_mins_c, 6, 6, 4, up(c)).
table(up_minc_maxc_s_maxs_c, 6, 7, 4, up(c)).
table(up_minc_s_mins_maxs_c, 6, 5, 4, up(c)).
table(up_maxc_s_mins_maxs_c, 6, 6, 4, up(c)).
table(up_v_minc_maxc_s_mins_c, 6, 13, 5, up(c)).
table(up_v_minc_maxc_s_maxs_c, 6, 17, 5, up(c)).
table(up_v_minc_maxc_mins_maxs_c, 6, 13, 5, up(c)).
table(up_v_minc_s_mins_maxs_c, 6, 18, 5, up(c)).
table(up_v_maxc_s_mins_maxs_c, 6, 18, 5, up(c)).
table(up_minc_maxc_s_mins_maxs_c, 6, 7, 5, up(c)).
table(up_v_minc_maxc_s_mins_maxs_c, 6, 17, 6, up(c)).
table(low_v_minc, 6, 3, 1, low(minc)).
table(low_c_minc, 6, 3, 1, low(minc)).
table(low_maxc_minc, 6, 3, 1, low(minc)).
table(low_s_minc, 6, 3, 1, low(minc)).
table(low_mins_minc, 6, 2, 1, low(minc)).
table(low_maxs_minc, 6, 3, 1, low(minc)).
table(low_v_c_minc, 6, 5, 2, low(minc)).
table(low_v_maxc_minc, 6, 4, 2, low(minc)).
table(low_v_s_minc, 6, 4, 2, low(minc)).
table(low_v_mins_minc, 6, 3, 2, low(minc)).
table(low_v_maxs_minc, 6, 4, 2, low(minc)).
table(low_c_maxc_minc, 6, 4, 2, low(minc)).
table(low_c_s_minc, 6, 4, 2, low(minc)).
table(low_c_mins_minc, 6, 3, 2, low(minc)).
table(low_c_maxs_minc, 6, 4, 2, low(minc)).
table(low_maxc_s_minc, 6, 4, 2, low(minc)).
table(low_maxc_mins_minc, 6, 4, 2, low(minc)).
table(low_maxc_maxs_minc, 6, 4, 2, low(minc)).
table(low_s_mins_minc, 6, 3, 2, low(minc)).
table(low_s_maxs_minc, 6, 4, 2, low(minc)).
table(low_mins_maxs_minc, 6, 3, 2, low(minc)).
table(low_v_c_maxc_minc, 6, 11, 3, low(minc)).
table(low_v_c_s_minc, 6, 6, 3, low(minc)).
table(low_v_c_mins_minc, 6, 6, 3, low(minc)).
table(low_v_c_maxs_minc, 6, 6, 3, low(minc)).
table(low_v_maxc_s_minc, 6, 5, 3, low(minc)).
table(low_v_maxc_mins_minc, 6, 5, 3, low(minc)).
table(low_v_maxc_maxs_minc, 6, 5, 3, low(minc)).
table(low_v_s_mins_minc, 6, 4, 3, low(minc)).
table(low_v_s_maxs_minc, 6, 8, 3, low(minc)).
table(low_v_mins_maxs_minc, 6, 4, 3, low(minc)).
table(low_c_maxc_s_minc, 6, 5, 3, low(minc)).
table(low_c_maxc_mins_minc, 6, 5, 3, low(minc)).
table(low_c_maxc_maxs_minc, 6, 5, 3, low(minc)).
table(low_c_s_mins_minc, 6, 4, 3, low(minc)).
table(low_c_s_maxs_minc, 6, 6, 3, low(minc)).
table(low_c_mins_maxs_minc, 6, 4, 3, low(minc)).
table(low_maxc_s_mins_minc, 6, 5, 3, low(minc)).
table(low_maxc_s_maxs_minc, 6, 6, 3, low(minc)).
table(low_maxc_mins_maxs_minc, 6, 5, 3, low(minc)).
table(low_s_mins_maxs_minc, 6, 4, 3, low(minc)).
table(low_v_c_maxc_s_minc, 6, 12, 4, low(minc)).
table(low_v_c_maxc_mins_minc, 6, 12, 4, low(minc)).
table(low_v_c_maxc_maxs_minc, 6, 12, 4, low(minc)).
table(low_v_c_s_mins_minc, 6, 7, 4, low(minc)).
table(low_v_c_s_maxs_minc, 6, 11, 4, low(minc)).
table(low_v_c_mins_maxs_minc, 6, 7, 4, low(minc)).
table(low_v_maxc_s_mins_minc, 6, 6, 4, low(minc)).
table(low_v_maxc_s_maxs_minc, 6, 10, 4, low(minc)).
table(low_v_maxc_mins_maxs_minc, 6, 6, 4, low(minc)).
table(low_v_s_mins_maxs_minc, 6, 8, 4, low(minc)).
table(low_c_maxc_s_mins_minc, 6, 6, 4, low(minc)).
table(low_c_maxc_s_maxs_minc, 6, 7, 4, low(minc)).
table(low_c_maxc_mins_maxs_minc, 6, 6, 4, low(minc)).
table(low_c_s_mins_maxs_minc, 6, 6, 4, low(minc)).
table(low_maxc_s_mins_maxs_minc, 6, 6, 4, low(minc)).
table(low_v_c_maxc_s_mins_minc, 6, 13, 5, low(minc)).
table(low_v_c_maxc_s_maxs_minc, 6, 17, 5, low(minc)).
table(low_v_c_maxc_mins_maxs_minc, 6, 13, 5, low(minc)).
table(low_v_c_s_mins_maxs_minc, 6, 11, 5, low(minc)).
table(low_v_maxc_s_mins_maxs_minc, 6, 10, 5, low(minc)).
table(low_c_maxc_s_mins_maxs_minc, 6, 7, 5, low(minc)).
table(low_v_c_maxc_s_mins_maxs_minc, 6, 17, 6, low(minc)).
table(up_v_minc, 6, 11, 1, up(minc)).
table(up_maxc_minc, 6, 6, 1, up(minc)).
table(up_v_c_minc, 6, 7, 2, up(minc)).
table(up_v_maxc_minc, 6, 6, 2, up(minc)).
table(up_v_s_minc, 6, 13, 2, up(minc)).
table(up_v_mins_minc, 6, 12, 2, up(minc)).
table(up_v_maxs_minc, 6, 13, 2, up(minc)).
table(up_c_maxc_minc, 6, 4, 2, up(minc)).
table(up_maxc_s_minc, 6, 7, 2, up(minc)).
table(up_maxc_mins_minc, 6, 7, 2, up(minc)).
table(up_maxc_maxs_minc, 6, 7, 2, up(minc)).
table(up_s_maxs_minc, 6, 18, 2, up(minc)).
table(up_v_c_maxc_minc, 6, 11, 3, up(minc)).
table(up_v_c_s_minc, 6, 8, 3, up(minc)).
table(up_v_c_mins_minc, 6, 8, 3, up(minc)).
table(up_v_c_maxs_minc, 6, 6, 3, up(minc)).
table(up_v_maxc_s_minc, 6, 7, 3, up(minc)).
table(up_v_maxc_mins_minc, 6, 7, 3, up(minc)).
table(up_v_maxc_maxs_minc, 6, 7, 3, up(minc)).
table(up_v_s_mins_minc, 6, 14, 3, up(minc)).
table(up_v_s_maxs_minc, 6, 18, 3, up(minc)).
table(up_v_mins_maxs_minc, 6, 14, 3, up(minc)).
table(up_c_maxc_s_minc, 6, 5, 3, up(minc)).
table(up_c_maxc_mins_minc, 6, 5, 3, up(minc)).
table(up_c_maxc_maxs_minc, 6, 5, 3, up(minc)).
table(up_c_s_maxs_minc, 6, 4, 3, up(minc)).
table(up_maxc_s_mins_minc, 6, 8, 3, up(minc)).
table(up_maxc_s_maxs_minc, 6, 7, 3, up(minc)).
table(up_maxc_mins_maxs_minc, 6, 8, 3, up(minc)).
table(up_s_mins_maxs_minc, 6, 18, 3, up(minc)).
table(up_v_c_maxc_s_minc, 6, 12, 4, up(minc)).
table(up_v_c_maxc_mins_minc, 6, 12, 4, up(minc)).
table(up_v_c_maxc_maxs_minc, 6, 12, 4, up(minc)).
table(up_v_c_s_mins_minc, 6, 7, 4, up(minc)).
table(up_v_c_s_maxs_minc, 6, 11, 4, up(minc)).
table(up_v_c_mins_maxs_minc, 6, 7, 4, up(minc)).
table(up_v_maxc_s_mins_minc, 6, 8, 4, up(minc)).
table(up_v_maxc_s_maxs_minc, 6, 12, 4, up(minc)).
table(up_v_maxc_mins_maxs_minc, 6, 8, 4, up(minc)).
table(up_v_s_mins_maxs_minc, 6, 18, 4, up(minc)).
table(up_c_maxc_s_mins_minc, 6, 6, 4, up(minc)).
table(up_c_maxc_s_maxs_minc, 6, 7, 4, up(minc)).
table(up_c_maxc_mins_maxs_minc, 6, 6, 4, up(minc)).
table(up_c_s_mins_maxs_minc, 6, 5, 4, up(minc)).
table(up_maxc_s_mins_maxs_minc, 6, 12, 4, up(minc)).
table(up_v_c_maxc_s_mins_minc, 6, 13, 5, up(minc)).
table(up_v_c_maxc_s_maxs_minc, 6, 17, 5, up(minc)).
table(up_v_c_maxc_mins_maxs_minc, 6, 13, 5, up(minc)).
table(up_v_c_s_mins_maxs_minc, 6, 11, 5, up(minc)).
table(up_v_maxc_s_mins_maxs_minc, 6, 12, 5, up(minc)).
table(up_c_maxc_s_mins_maxs_minc, 6, 7, 5, up(minc)).
table(up_v_c_maxc_s_mins_maxs_minc, 6, 17, 6, up(minc)).
table(low_v_maxc, 6, 18, 1, low(maxc)).
table(low_c_maxc, 6, 18, 1, low(maxc)).
table(low_minc_maxc, 6, 6, 1, low(maxc)).
table(low_s_maxc, 6, 18, 1, low(maxc)).
table(low_mins_maxc, 6, 9, 1, low(maxc)).
table(low_maxs_maxc, 6, 2, 1, low(maxc)).
table(low_v_c_maxc, 6, 11, 2, low(maxc)).
table(low_v_minc_maxc, 6, 11, 2, low(maxc)).
table(low_v_s_maxc, 6, 10, 2, low(maxc)).
table(low_v_mins_maxc, 6, 18, 2, low(maxc)).
table(low_v_maxs_maxc, 6, 3, 2, low(maxc)).
table(low_c_minc_maxc, 6, 11, 2, low(maxc)).
table(low_c_s_maxc, 6, 6, 2, low(maxc)).
table(low_c_mins_maxc, 6, 18, 2, low(maxc)).
table(low_c_maxs_maxc, 6, 3, 2, low(maxc)).
table(low_minc_s_maxc, 6, 7, 2, low(maxc)).
table(low_minc_mins_maxc, 6, 7, 2, low(maxc)).
table(low_minc_maxs_maxc, 6, 4, 2, low(maxc)).
table(low_s_mins_maxc, 6, 18, 2, low(maxc)).
table(low_s_maxs_maxc, 6, 3, 2, low(maxc)).
table(low_mins_maxs_maxc, 6, 3, 2, low(maxc)).
table(low_v_c_minc_maxc, 6, 11, 3, low(maxc)).
table(low_v_c_s_maxc, 6, 12, 3, low(maxc)).
table(low_v_c_mins_maxc, 6, 12, 3, low(maxc)).
table(low_v_c_maxs_maxc, 6, 12, 3, low(maxc)).
table(low_v_minc_s_maxc, 6, 7, 3, low(maxc)).
table(low_v_minc_mins_maxc, 6, 12, 3, low(maxc)).
table(low_v_minc_maxs_maxc, 6, 5, 3, low(maxc)).
table(low_v_s_mins_maxc, 6, 8, 3, low(maxc)).
table(low_v_s_maxs_maxc, 6, 8, 3, low(maxc)).
table(low_v_mins_maxs_maxc, 6, 4, 3, low(maxc)).
table(low_c_minc_s_maxc, 6, 7, 3, low(maxc)).
table(low_c_minc_mins_maxc, 6, 12, 3, low(maxc)).
table(low_c_minc_maxs_maxc, 6, 5, 3, low(maxc)).
table(low_c_s_mins_maxc, 6, 6, 3, low(maxc)).
table(low_c_s_maxs_maxc, 6, 4, 3, low(maxc)).
table(low_c_mins_maxs_maxc, 6, 4, 3, low(maxc)).
table(low_minc_s_mins_maxc, 6, 8, 3, low(maxc)).
table(low_minc_s_maxs_maxc, 6, 5, 3, low(maxc)).
table(low_minc_mins_maxs_maxc, 6, 5, 3, low(maxc)).
table(low_s_mins_maxs_maxc, 6, 4, 3, low(maxc)).
table(low_v_c_minc_s_maxc, 6, 12, 4, low(maxc)).
table(low_v_c_minc_mins_maxc, 6, 12, 4, low(maxc)).
table(low_v_c_minc_maxs_maxc, 6, 12, 4, low(maxc)).
table(low_v_c_s_mins_maxc, 6, 13, 4, low(maxc)).
table(low_v_c_s_maxs_maxc, 6, 18, 4, low(maxc)).
table(low_v_c_mins_maxs_maxc, 6, 13, 4, low(maxc)).
table(low_v_minc_s_mins_maxc, 6, 8, 4, low(maxc)).
table(low_v_minc_s_maxs_maxc, 6, 10, 4, low(maxc)).
table(low_v_minc_mins_maxs_maxc, 6, 6, 4, low(maxc)).
table(low_v_s_mins_maxs_maxc, 6, 8, 4, low(maxc)).
table(low_c_minc_s_mins_maxc, 6, 8, 4, low(maxc)).
table(low_c_minc_s_maxs_maxc, 6, 7, 4, low(maxc)).
table(low_c_minc_mins_maxs_maxc, 6, 6, 4, low(maxc)).
table(low_c_s_mins_maxs_maxc, 6, 5, 4, low(maxc)).
table(low_minc_s_mins_maxs_maxc, 6, 6, 4, low(maxc)).
table(low_v_c_minc_s_mins_maxc, 6, 13, 5, low(maxc)).
table(low_v_c_minc_s_maxs_maxc, 6, 17, 5, low(maxc)).
table(low_v_c_minc_mins_maxs_maxc, 6, 13, 5, low(maxc)).
table(low_v_c_s_mins_maxs_maxc, 6, 18, 5, low(maxc)).
table(low_v_minc_s_mins_maxs_maxc, 6, 10, 5, low(maxc)).
table(low_c_minc_s_mins_maxs_maxc, 6, 7, 5, low(maxc)).
table(low_v_c_minc_s_mins_maxs_maxc, 6, 17, 6, low(maxc)).
table(up_v_maxc, 6, 11, 1, up(maxc)).
table(up_v_c_maxc, 6, 11, 2, up(maxc)).
table(up_v_minc_maxc, 6, 11, 2, up(maxc)).
table(up_v_s_maxc, 6, 13, 2, up(maxc)).
table(up_v_mins_maxc, 6, 12, 2, up(maxc)).
table(up_v_maxs_maxc, 6, 13, 2, up(maxc)).
table(up_v_c_minc_maxc, 6, 11, 3, up(maxc)).
table(up_v_c_s_maxc, 6, 13, 3, up(maxc)).
table(up_v_c_mins_maxc, 6, 12, 3, up(maxc)).
table(up_v_c_maxs_maxc, 6, 13, 3, up(maxc)).
table(up_v_minc_s_maxc, 6, 12, 3, up(maxc)).
table(up_v_minc_mins_maxc, 6, 12, 3, up(maxc)).
table(up_v_minc_maxs_maxc, 6, 12, 3, up(maxc)).
table(up_v_s_mins_maxc, 6, 14, 3, up(maxc)).
table(up_v_s_maxs_maxc, 6, 18, 3, up(maxc)).
table(up_v_mins_maxs_maxc, 6, 14, 3, up(maxc)).
table(up_v_c_minc_s_maxc, 6, 12, 4, up(maxc)).
table(up_v_c_minc_mins_maxc, 6, 12, 4, up(maxc)).
table(up_v_c_minc_maxs_maxc, 6, 12, 4, up(maxc)).
table(up_v_c_s_mins_maxc, 6, 14, 4, up(maxc)).
table(up_v_c_s_maxs_maxc, 6, 18, 4, up(maxc)).
table(up_v_c_mins_maxs_maxc, 6, 14, 4, up(maxc)).
table(up_v_minc_s_mins_maxc, 6, 13, 4, up(maxc)).
table(up_v_minc_s_maxs_maxc, 6, 17, 4, up(maxc)).
table(up_v_minc_mins_maxs_maxc, 6, 13, 4, up(maxc)).
table(up_v_s_mins_maxs_maxc, 6, 18, 4, up(maxc)).
table(up_v_c_minc_s_mins_maxc, 6, 13, 5, up(maxc)).
table(up_v_c_minc_s_maxs_maxc, 6, 17, 5, up(maxc)).
table(up_v_c_minc_mins_maxs_maxc, 6, 13, 5, up(maxc)).
table(up_v_c_s_mins_maxs_maxc, 6, 18, 5, up(maxc)).
table(up_v_minc_s_mins_maxs_maxc, 6, 17, 5, up(maxc)).
table(up_v_c_minc_s_mins_maxs_maxc, 6, 17, 6, up(maxc)).
table(low_v_s, 6, 18, 1, low(s)).
table(low_c_s, 6, 3, 1, low(s)).
table(low_minc_s, 6, 18, 1, low(s)).
table(low_maxc_s, 6, 18, 1, low(s)).
table(low_mins_s, 6, 18, 1, low(s)).
table(low_maxs_s, 6, 18, 1, low(s)).
table(low_v_c_s, 6, 4, 2, low(s)).
table(low_v_minc_s, 6, 18, 2, low(s)).
table(low_v_maxc_s, 6, 18, 2, low(s)).
table(low_v_mins_s, 6, 9, 2, low(s)).
table(low_v_maxs_s, 6, 7, 2, low(s)).
table(low_c_minc_s, 6, 5, 2, low(s)).
table(low_c_maxc_s, 6, 5, 2, low(s)).
table(low_c_mins_s, 6, 5, 2, low(s)).
table(low_c_maxs_s, 6, 5, 2, low(s)).
table(low_minc_maxc_s, 6, 18, 2, low(s)).
table(low_minc_mins_s, 6, 18, 2, low(s)).
table(low_minc_maxs_s, 6, 18, 2, low(s)).
table(low_maxc_mins_s, 6, 7, 2, low(s)).
table(low_maxc_maxs_s, 6, 18, 2, low(s)).
table(low_mins_maxs_s, 6, 12, 2, low(s)).
table(low_v_c_minc_s, 6, 8, 3, low(s)).
table(low_v_c_maxc_s, 6, 18, 3, low(s)).
table(low_v_c_mins_s, 6, 8, 3, low(s)).
table(low_v_c_maxs_s, 6, 8, 3, low(s)).
table(low_v_minc_maxc_s, 6, 18, 3, low(s)).
table(low_v_minc_mins_s, 6, 13, 3, low(s)).
table(low_v_minc_maxs_s, 6, 8, 3, low(s)).
table(low_v_maxc_mins_s, 6, 4, 3, low(s)).
table(low_v_maxc_maxs_s, 6, 8, 3, low(s)).
table(low_v_mins_maxs_s, 6, 7, 3, low(s)).
table(low_c_minc_maxc_s, 6, 8, 3, low(s)).
table(low_c_minc_mins_s, 6, 4, 3, low(s)).
table(low_c_minc_maxs_s, 6, 8, 3, low(s)).
table(low_c_maxc_mins_s, 6, 8, 3, low(s)).
table(low_c_maxc_maxs_s, 6, 5, 3, low(s)).
table(low_c_mins_maxs_s, 6, 8, 3, low(s)).
table(low_minc_maxc_mins_s, 6, 13, 3, low(s)).
table(low_minc_maxc_maxs_s, 6, 18, 3, low(s)).
table(low_minc_mins_maxs_s, 6, 8, 3, low(s)).
table(low_maxc_mins_maxs_s, 6, 10, 3, low(s)).
table(low_v_c_minc_maxc_s, 6, 18, 4, low(s)).
table(low_v_c_minc_mins_s, 6, 7, 4, low(s)).
table(low_v_c_minc_maxs_s, 6, 11, 4, low(s)).
table(low_v_c_maxc_mins_s, 6, 13, 4, low(s)).
table(low_v_c_maxc_maxs_s, 6, 18, 4, low(s)).
table(low_v_c_mins_maxs_s, 6, 8, 4, low(s)).
table(low_v_minc_maxc_mins_s, 6, 13, 4, low(s)).
table(low_v_minc_maxc_maxs_s, 6, 10, 4, low(s)).
table(low_v_minc_mins_maxs_s, 6, 8, 4, low(s)).
table(low_v_maxc_mins_maxs_s, 6, 8, 4, low(s)).
table(low_c_minc_maxc_mins_s, 6, 6, 4, low(s)).
table(low_c_minc_maxc_maxs_s, 6, 8, 4, low(s)).
table(low_c_minc_mins_maxs_s, 6, 5, 4, low(s)).
table(low_c_maxc_mins_maxs_s, 6, 6, 4, low(s)).
table(low_minc_maxc_mins_maxs_s, 6, 17, 4, low(s)).
table(low_v_c_minc_maxc_mins_s, 6, 13, 5, low(s)).
table(low_v_c_minc_maxc_maxs_s, 6, 18, 5, low(s)).
table(low_v_c_minc_mins_maxs_s, 6, 11, 5, low(s)).
table(low_v_c_maxc_mins_maxs_s, 6, 17, 5, low(s)).
table(low_v_minc_maxc_mins_maxs_s, 6, 10, 5, low(s)).
table(low_c_minc_maxc_mins_maxs_s, 6, 7, 5, low(s)).
table(low_v_c_minc_maxc_mins_maxs_s, 6, 17, 6, low(s)).
table(up_v_s, 6, 7, 1, up(s)).
table(up_v_c_s, 6, 8, 2, up(s)).
table(up_v_minc_s, 6, 8, 2, up(s)).
table(up_v_maxc_s, 6, 8, 2, up(s)).
table(up_v_mins_s, 6, 7, 2, up(s)).
table(up_v_maxs_s, 6, 7, 2, up(s)).
table(up_c_maxc_s, 6, 18, 2, up(s)).
table(up_v_c_minc_s, 6, 11, 3, up(s)).
table(up_v_c_maxc_s, 6, 18, 3, up(s)).
table(up_v_c_mins_s, 6, 8, 3, up(s)).
table(up_v_c_maxs_s, 6, 8, 3, up(s)).
table(up_v_minc_maxc_s, 6, 10, 3, up(s)).
table(up_v_minc_mins_s, 6, 8, 3, up(s)).
table(up_v_minc_maxs_s, 6, 8, 3, up(s)).
table(up_v_maxc_mins_s, 6, 8, 3, up(s)).
table(up_v_maxc_maxs_s, 6, 8, 3, up(s)).
table(up_v_mins_maxs_s, 6, 7, 3, up(s)).
table(up_c_minc_maxc_s, 6, 18, 3, up(s)).
table(up_c_maxc_mins_s, 6, 18, 3, up(s)).
table(up_c_maxc_maxs_s, 6, 17, 3, up(s)).
table(up_v_c_minc_maxc_s, 6, 18, 4, up(s)).
table(up_v_c_minc_mins_s, 6, 11, 4, up(s)).
table(up_v_c_minc_maxs_s, 6, 11, 4, up(s)).
table(up_v_c_maxc_mins_s, 6, 18, 4, up(s)).
table(up_v_c_maxc_maxs_s, 6, 17, 4, up(s)).
table(up_v_c_mins_maxs_s, 6, 8, 4, up(s)).
table(up_v_minc_maxc_mins_s, 6, 10, 4, up(s)).
table(up_v_minc_maxc_maxs_s, 6, 10, 4, up(s)).
table(up_v_minc_mins_maxs_s, 6, 8, 4, up(s)).
table(up_v_maxc_mins_maxs_s, 6, 8, 4, up(s)).
table(up_c_minc_maxc_mins_s, 6, 18, 4, up(s)).
table(up_c_minc_maxc_maxs_s, 6, 17, 4, up(s)).
table(up_c_maxc_mins_maxs_s, 6, 17, 4, up(s)).
table(up_v_c_minc_maxc_mins_s, 6, 18, 5, up(s)).
table(up_v_c_minc_maxc_maxs_s, 6, 17, 5, up(s)).
table(up_v_c_minc_mins_maxs_s, 6, 11, 5, up(s)).
table(up_v_c_maxc_mins_maxs_s, 6, 17, 5, up(s)).
table(up_v_minc_maxc_mins_maxs_s, 6, 10, 5, up(s)).
table(up_c_minc_maxc_mins_maxs_s, 6, 17, 5, up(s)).
table(up_v_c_minc_maxc_mins_maxs_s, 6, 17, 6, up(s)).
table(low_v_mins, 6, 2, 1, low(mins)).
table(low_c_mins, 6, 2, 1, low(mins)).
table(low_minc_mins, 6, 2, 1, low(mins)).
table(low_maxc_mins, 6, 2, 1, low(mins)).
table(low_s_mins, 6, 2, 1, low(mins)).
table(low_maxs_mins, 6, 2, 1, low(mins)).
table(low_v_c_mins, 6, 3, 2, low(mins)).
table(low_v_minc_mins, 6, 3, 2, low(mins)).
table(low_v_maxc_mins, 6, 3, 2, low(mins)).
table(low_v_s_mins, 6, 3, 2, low(mins)).
table(low_v_maxs_mins, 6, 3, 2, low(mins)).
table(low_c_minc_mins, 6, 3, 2, low(mins)).
table(low_c_maxc_mins, 6, 3, 2, low(mins)).
table(low_c_s_mins, 6, 3, 2, low(mins)).
table(low_c_maxs_mins, 6, 3, 2, low(mins)).
table(low_minc_maxc_mins, 6, 4, 2, low(mins)).
table(low_minc_s_mins, 6, 3, 2, low(mins)).
table(low_minc_maxs_mins, 6, 3, 2, low(mins)).
table(low_maxc_s_mins, 6, 3, 2, low(mins)).
table(low_maxc_maxs_mins, 6, 3, 2, low(mins)).
table(low_s_maxs_mins, 6, 3, 2, low(mins)).
table(low_v_c_minc_mins, 6, 6, 3, low(mins)).
table(low_v_c_maxc_mins, 6, 12, 3, low(mins)).
table(low_v_c_s_mins, 6, 4, 3, low(mins)).
table(low_v_c_maxs_mins, 6, 4, 3, low(mins)).
table(low_v_minc_maxc_mins, 6, 5, 3, low(mins)).
table(low_v_minc_s_mins, 6, 4, 3, low(mins)).
table(low_v_minc_maxs_mins, 6, 4, 3, low(mins)).
table(low_v_maxc_s_mins, 6, 4, 3, low(mins)).
table(low_v_maxc_maxs_mins, 6, 4, 3, low(mins)).
table(low_v_s_maxs_mins, 6, 7, 3, low(mins)).
table(low_c_minc_maxc_mins, 6, 5, 3, low(mins)).
table(low_c_minc_s_mins, 6, 4, 3, low(mins)).
table(low_c_minc_maxs_mins, 6, 4, 3, low(mins)).
table(low_c_maxc_s_mins, 6, 4, 3, low(mins)).
table(low_c_maxc_maxs_mins, 6, 4, 3, low(mins)).
table(low_c_s_maxs_mins, 6, 4, 3, low(mins)).
table(low_minc_maxc_s_mins, 6, 5, 3, low(mins)).
table(low_minc_maxc_maxs_mins, 6, 5, 3, low(mins)).
table(low_minc_s_maxs_mins, 6, 4, 3, low(mins)).
table(low_maxc_s_maxs_mins, 6, 4, 3, low(mins)).
table(low_v_c_minc_maxc_mins, 6, 12, 4, low(mins)).
table(low_v_c_minc_s_mins, 6, 7, 4, low(mins)).
table(low_v_c_minc_maxs_mins, 6, 7, 4, low(mins)).
table(low_v_c_maxc_s_mins, 6, 13, 4, low(mins)).
table(low_v_c_maxc_maxs_mins, 6, 13, 4, low(mins)).
table(low_v_c_s_maxs_mins, 6, 8, 4, low(mins)).
table(low_v_minc_maxc_s_mins, 6, 6, 4, low(mins)).
table(low_v_minc_maxc_maxs_mins, 6, 6, 4, low(mins)).
table(low_v_minc_s_maxs_mins, 6, 8, 4, low(mins)).
table(low_v_maxc_s_maxs_mins, 6, 8, 4, low(mins)).
table(low_c_minc_maxc_s_mins, 6, 6, 4, low(mins)).
table(low_c_minc_maxc_maxs_mins, 6, 6, 4, low(mins)).
table(low_c_minc_s_maxs_mins, 6, 5, 4, low(mins)).
table(low_c_maxc_s_maxs_mins, 6, 5, 4, low(mins)).
table(low_minc_maxc_s_maxs_mins, 6, 6, 4, low(mins)).
table(low_v_c_minc_maxc_s_mins, 6, 13, 5, low(mins)).
table(low_v_c_minc_maxc_maxs_mins, 6, 13, 5, low(mins)).
table(low_v_c_minc_s_maxs_mins, 6, 11, 5, low(mins)).
table(low_v_c_maxc_s_maxs_mins, 6, 17, 5, low(mins)).
table(low_v_minc_maxc_s_maxs_mins, 6, 10, 5, low(mins)).
table(low_c_minc_maxc_s_maxs_mins, 6, 7, 5, low(mins)).
table(low_v_c_minc_maxc_s_maxs_mins, 6, 17, 6, low(mins)).
table(up_v_mins, 6, 18, 1, up(mins)).
table(up_maxc_mins, 6, 9, 1, up(mins)).
table(up_maxs_mins, 6, 2, 1, up(mins)).
table(up_v_c_mins, 6, 8, 2, up(mins)).
table(up_v_minc_mins, 6, 3, 2, up(mins)).
table(up_v_maxc_mins, 6, 7, 2, up(mins)).
table(up_v_s_mins, 6, 3, 2, up(mins)).
table(up_v_maxs_mins, 6, 3, 2, up(mins)).
table(up_c_maxc_mins, 6, 5, 2, up(mins)).
table(up_c_maxs_mins, 6, 3, 2, up(mins)).
table(up_minc_maxc_mins, 6, 4, 2, up(mins)).
table(up_minc_maxs_mins, 6, 3, 2, up(mins)).
table(up_maxc_s_mins, 6, 3, 2, up(mins)).
table(up_maxc_maxs_mins, 6, 3, 2, up(mins)).
table(up_s_maxs_mins, 6, 3, 2, up(mins)).
table(up_v_c_minc_mins, 6, 6, 3, up(mins)).
table(up_v_c_maxc_mins, 6, 12, 3, up(mins)).
table(up_v_c_s_mins, 6, 4, 3, up(mins)).
table(up_v_c_maxs_mins, 6, 4, 3, up(mins)).
table(up_v_minc_maxc_mins, 6, 5, 3, up(mins)).
table(up_v_minc_s_mins, 6, 4, 3, up(mins)).
table(up_v_minc_maxs_mins, 6, 4, 3, up(mins)).
table(up_v_maxc_s_mins, 6, 4, 3, up(mins)).
table(up_v_maxc_maxs_mins, 6, 4, 3, up(mins)).
table(up_v_s_maxs_mins, 6, 7, 3, up(mins)).
table(up_c_minc_maxc_mins, 6, 5, 3, up(mins)).
table(up_c_minc_maxs_mins, 6, 4, 3, up(mins)).
table(up_c_maxc_s_mins, 6, 4, 3, up(mins)).
table(up_c_maxc_maxs_mins, 6, 4, 3, up(mins)).
table(up_c_s_maxs_mins, 6, 4, 3, up(mins)).
table(up_minc_maxc_s_mins, 6, 5, 3, up(mins)).
table(up_minc_maxc_maxs_mins, 6, 5, 3, up(mins)).
table(up_minc_s_maxs_mins, 6, 4, 3, up(mins)).
table(up_maxc_s_maxs_mins, 6, 4, 3, up(mins)).
table(up_v_c_minc_maxc_mins, 6, 12, 4, up(mins)).
table(up_v_c_minc_s_mins, 6, 7, 4, up(mins)).
table(up_v_c_minc_maxs_mins, 6, 7, 4, up(mins)).
table(up_v_c_maxc_s_mins, 6, 13, 4, up(mins)).
table(up_v_c_maxc_maxs_mins, 6, 13, 4, up(mins)).
table(up_v_c_s_maxs_mins, 6, 8, 4, up(mins)).
table(up_v_minc_maxc_s_mins, 6, 6, 4, up(mins)).
table(up_v_minc_maxc_maxs_mins, 6, 6, 4, up(mins)).
table(up_v_minc_s_maxs_mins, 6, 8, 4, up(mins)).
table(up_v_maxc_s_maxs_mins, 6, 8, 4, up(mins)).
table(up_c_minc_maxc_s_mins, 6, 6, 4, up(mins)).
table(up_c_minc_maxc_maxs_mins, 6, 6, 4, up(mins)).
table(up_c_minc_s_maxs_mins, 6, 6, 4, up(mins)).
table(up_c_maxc_s_maxs_mins, 6, 5, 4, up(mins)).
table(up_minc_maxc_s_maxs_mins, 6, 6, 4, up(mins)).
table(up_v_c_minc_maxc_s_mins, 6, 13, 5, up(mins)).
table(up_v_c_minc_maxc_maxs_mins, 6, 13, 5, up(mins)).
table(up_v_c_minc_s_maxs_mins, 6, 11, 5, up(mins)).
table(up_v_c_maxc_s_maxs_mins, 6, 17, 5, up(mins)).
table(up_v_minc_maxc_s_maxs_mins, 6, 10, 5, up(mins)).
table(up_c_minc_maxc_s_maxs_mins, 6, 7, 5, up(mins)).
table(up_v_c_minc_maxc_s_maxs_mins, 6, 17, 6, up(mins)).
table(low_v_maxs, 6, 7, 1, low(maxs)).
table(low_c_maxs, 6, 4, 1, low(maxs)).
table(low_minc_maxs, 6, 4, 1, low(maxs)).
table(low_maxc_maxs, 6, 4, 1, low(maxs)).
table(low_s_maxs, 6, 7, 1, low(maxs)).
table(low_mins_maxs, 6, 2, 1, low(maxs)).
table(low_v_c_maxs, 6, 8, 2, low(maxs)).
table(low_v_minc_maxs, 6, 8, 2, low(maxs)).
table(low_v_maxc_maxs, 6, 8, 2, low(maxs)).
table(low_v_s_maxs, 6, 7, 2, low(maxs)).
table(low_v_mins_maxs, 6, 7, 2, low(maxs)).
table(low_c_minc_maxs, 6, 5, 2, low(maxs)).
table(low_c_maxc_maxs, 6, 5, 2, low(maxs)).
table(low_c_s_maxs, 6, 8, 2, low(maxs)).
table(low_c_mins_maxs, 6, 3, 2, low(maxs)).
table(low_minc_maxc_maxs, 6, 6, 2, low(maxs)).
table(low_minc_s_maxs, 6, 3, 2, low(maxs)).
table(low_minc_mins_maxs, 6, 3, 2, low(maxs)).
table(low_maxc_s_maxs, 6, 3, 2, low(maxs)).
table(low_maxc_mins_maxs, 6, 3, 2, low(maxs)).
table(low_s_mins_maxs, 6, 7, 2, low(maxs)).
table(low_v_c_minc_maxs, 6, 11, 3, low(maxs)).
table(low_v_c_maxc_maxs, 6, 18, 3, low(maxs)).
table(low_v_c_s_maxs, 6, 8, 3, low(maxs)).
table(low_v_c_mins_maxs, 6, 8, 3, low(maxs)).
table(low_v_minc_maxc_maxs, 6, 10, 3, low(maxs)).
table(low_v_minc_s_maxs, 6, 8, 3, low(maxs)).
table(low_v_minc_mins_maxs, 6, 8, 3, low(maxs)).
table(low_v_maxc_s_maxs, 6, 8, 3, low(maxs)).
table(low_v_maxc_mins_maxs, 6, 8, 3, low(maxs)).
table(low_v_s_mins_maxs, 6, 7, 3, low(maxs)).
table(low_c_minc_maxc_maxs, 6, 7, 3, low(maxs)).
table(low_c_minc_s_maxs, 6, 4, 3, low(maxs)).
table(low_c_minc_mins_maxs, 6, 4, 3, low(maxs)).
table(low_c_maxc_s_maxs, 6, 4, 3, low(maxs)).
table(low_c_maxc_mins_maxs, 6, 5, 3, low(maxs)).
table(low_c_s_mins_maxs, 6, 8, 3, low(maxs)).
table(low_minc_maxc_s_maxs, 6, 5, 3, low(maxs)).
table(low_minc_maxc_mins_maxs, 6, 5, 3, low(maxs)).
table(low_minc_s_mins_maxs, 6, 4, 3, low(maxs)).
table(low_maxc_s_mins_maxs, 6, 4, 3, low(maxs)).
table(low_v_c_minc_maxc_maxs, 6, 18, 4, low(maxs)).
table(low_v_c_minc_s_maxs, 6, 11, 4, low(maxs)).
table(low_v_c_minc_mins_maxs, 6, 11, 4, low(maxs)).
table(low_v_c_maxc_s_maxs, 6, 17, 4, low(maxs)).
table(low_v_c_maxc_mins_maxs, 6, 18, 4, low(maxs)).
table(low_v_c_s_mins_maxs, 6, 8, 4, low(maxs)).
table(low_v_minc_maxc_s_maxs, 6, 10, 4, low(maxs)).
table(low_v_minc_maxc_mins_maxs, 6, 10, 4, low(maxs)).
table(low_v_minc_s_mins_maxs, 6, 8, 4, low(maxs)).
table(low_v_maxc_s_mins_maxs, 6, 8, 4, low(maxs)).
table(low_c_minc_maxc_s_maxs, 6, 7, 4, low(maxs)).
table(low_c_minc_maxc_mins_maxs, 6, 7, 4, low(maxs)).
table(low_c_minc_s_mins_maxs, 6, 5, 4, low(maxs)).
table(low_c_maxc_s_mins_maxs, 6, 5, 4, low(maxs)).
table(low_minc_maxc_s_mins_maxs, 6, 6, 4, low(maxs)).
table(low_v_c_minc_maxc_s_maxs, 6, 17, 5, low(maxs)).
table(low_v_c_minc_maxc_mins_maxs, 6, 18, 5, low(maxs)).
table(low_v_c_minc_s_mins_maxs, 6, 11, 5, low(maxs)).
table(low_v_c_maxc_s_mins_maxs, 6, 17, 5, low(maxs)).
table(low_v_minc_maxc_s_mins_maxs, 6, 10, 5, low(maxs)).
table(low_c_minc_maxc_s_mins_maxs, 6, 7, 5, low(maxs)).
table(low_v_c_minc_maxc_s_mins_maxs, 6, 17, 6, low(maxs)).
table(up_v_maxs, 6, 18, 1, up(maxs)).
table(up_maxc_maxs, 6, 2, 1, up(maxs)).
table(up_v_c_maxs, 6, 18, 2, up(maxs)).
table(up_v_minc_maxs, 6, 12, 2, up(maxs)).
table(up_v_maxc_maxs, 6, 3, 2, up(maxs)).
table(up_v_s_maxs, 6, 7, 2, up(maxs)).
table(up_v_mins_maxs, 6, 9, 2, up(maxs)).
table(up_c_maxc_maxs, 6, 3, 2, up(maxs)).
table(up_minc_maxc_maxs, 6, 4, 2, up(maxs)).
table(up_maxc_s_maxs, 6, 3, 2, up(maxs)).
table(up_maxc_mins_maxs, 6, 3, 2, up(maxs)).
table(up_v_c_minc_maxs, 6, 12, 3, up(maxs)).
table(up_v_c_maxc_maxs, 6, 12, 3, up(maxs)).
table(up_v_c_s_maxs, 6, 8, 3, up(maxs)).
table(up_v_c_mins_maxs, 6, 18, 3, up(maxs)).
table(up_v_minc_maxc_maxs, 6, 5, 3, up(maxs)).
table(up_v_minc_s_maxs, 6, 8, 3, up(maxs)).
table(up_v_minc_mins_maxs, 6, 14, 3, up(maxs)).
table(up_v_maxc_s_maxs, 6, 8, 3, up(maxs)).
table(up_v_maxc_mins_maxs, 6, 4, 3, up(maxs)).
table(up_v_s_mins_maxs, 6, 7, 3, up(maxs)).
table(up_c_minc_maxc_maxs, 6, 5, 3, up(maxs)).
table(up_c_maxc_s_maxs, 6, 4, 3, up(maxs)).
table(up_c_maxc_mins_maxs, 6, 4, 3, up(maxs)).
table(up_minc_maxc_s_maxs, 6, 5, 3, up(maxs)).
table(up_minc_maxc_mins_maxs, 6, 5, 3, up(maxs)).
table(up_maxc_s_mins_maxs, 6, 4, 3, up(maxs)).
table(up_v_c_minc_maxc_maxs, 6, 12, 4, up(maxs)).
table(up_v_c_minc_s_maxs, 6, 11, 4, up(maxs)).
table(up_v_c_minc_mins_maxs, 6, 13, 4, up(maxs)).
table(up_v_c_maxc_s_maxs, 6, 17, 4, up(maxs)).
table(up_v_c_maxc_mins_maxs, 6, 13, 4, up(maxs)).
table(up_v_c_s_mins_maxs, 6, 8, 4, up(maxs)).
table(up_v_minc_maxc_s_maxs, 6, 10, 4, up(maxs)).
table(up_v_minc_maxc_mins_maxs, 6, 6, 4, up(maxs)).
table(up_v_minc_s_mins_maxs, 6, 8, 4, up(maxs)).
table(up_v_maxc_s_mins_maxs, 6, 8, 4, up(maxs)).
table(up_c_minc_maxc_s_maxs, 6, 7, 4, up(maxs)).
table(up_c_minc_maxc_mins_maxs, 6, 6, 4, up(maxs)).
table(up_c_maxc_s_mins_maxs, 6, 5, 4, up(maxs)).
table(up_minc_maxc_s_mins_maxs, 6, 6, 4, up(maxs)).
table(up_v_c_minc_maxc_s_maxs, 6, 17, 5, up(maxs)).
table(up_v_c_minc_maxc_mins_maxs, 6, 13, 5, up(maxs)).
table(up_v_c_minc_s_mins_maxs, 6, 11, 5, up(maxs)).
table(up_v_c_maxc_s_mins_maxs, 6, 17, 5, up(maxs)).
table(up_v_minc_maxc_s_mins_maxs, 6, 10, 5, up(maxs)).
table(up_c_minc_maxc_s_mins_maxs, 6, 7, 5, up(maxs)).
table(up_v_c_minc_maxc_s_mins_maxs, 6, 17, 6, up(maxs)).
table(low_v_mina, 6, 13, 1, low(mina)).
table(low_c_mina, 6, 7, 1, low(mina)).
table(low_minc_mina, 6, 19, 1, low(mina)).
table(low_maxc_mina, 6, 19, 1, low(mina)).
table(low_s_mina, 6, 13, 1, low(mina)).
table(low_mins_mina, 6, 11, 1, low(mina)).
table(low_maxs_mina, 6, 11, 1, low(mina)).
table(low_v_c_mina, 6, 12, 2, low(mina)).
table(low_v_minc_mina, 6, 12, 2, low(mina)).
table(low_v_maxc_mina, 6, 11, 2, low(mina)).
table(low_v_s_mina, 6, 4, 2, low(mina)).
table(low_v_mins_mina, 6, 6, 2, low(mina)).
table(low_v_maxs_mina, 6, 10, 2, low(mina)).
table(low_c_minc_mina, 6, 8, 2, low(mina)).
table(low_c_maxc_mina, 6, 6, 2, low(mina)).
table(low_c_s_mina, 6, 12, 2, low(mina)).
table(low_c_mins_mina, 6, 7, 2, low(mina)).
table(low_c_maxs_mina, 6, 5, 2, low(mina)).
table(low_minc_maxc_mina, 6, 19, 2, low(mina)).
table(low_minc_s_mina, 6, 12, 2, low(mina)).
table(low_minc_mins_mina, 6, 19, 2, low(mina)).
table(low_minc_maxs_mina, 6, 19, 2, low(mina)).
table(low_maxc_s_mina, 6, 11, 2, low(mina)).
table(low_maxc_mins_mina, 6, 19, 2, low(mina)).
table(low_maxc_maxs_mina, 6, 19, 2, low(mina)).
table(low_s_mins_mina, 6, 13, 2, low(mina)).
table(low_s_maxs_mina, 6, 10, 2, low(mina)).
table(low_mins_maxs_mina, 6, 6, 2, low(mina)).
table(low_v_c_minc_mina, 6, 19, 3, low(mina)).
table(low_v_c_maxc_mina, 6, 19, 3, low(mina)).
table(low_v_c_s_mina, 6, 4, 3, low(mina)).
table(low_v_c_mins_mina, 6, 12, 3, low(mina)).
table(low_v_c_maxs_mina, 6, 12, 3, low(mina)).
table(low_v_minc_maxc_mina, 6, 19, 3, low(mina)).
table(low_v_minc_s_mina, 6, 6, 3, low(mina)).
table(low_v_minc_mins_mina, 6, 5, 3, low(mina)).
table(low_v_minc_maxs_mina, 6, 12, 3, low(mina)).
table(low_v_maxc_s_mina, 6, 10, 3, low(mina)).
table(low_v_maxc_mins_mina, 6, 11, 3, low(mina)).
table(low_v_maxc_maxs_mina, 6, 10, 3, low(mina)).
table(low_v_s_mins_mina, 6, 5, 3, low(mina)).
table(low_v_s_maxs_mina, 6, 10, 3, low(mina)).
table(low_v_mins_maxs_mina, 6, 10, 3, low(mina)).
table(low_c_minc_maxc_mina, 6, 8, 3, low(mina)).
table(low_c_minc_s_mina, 6, 19, 3, low(mina)).
table(low_c_minc_mins_mina, 6, 8, 3, low(mina)).
table(low_c_minc_maxs_mina, 6, 6, 3, low(mina)).
table(low_c_maxc_s_mina, 6, 9, 3, low(mina)).
table(low_c_maxc_mins_mina, 6, 6, 3, low(mina)).
table(low_c_maxc_maxs_mina, 6, 6, 3, low(mina)).
table(low_c_s_mins_mina, 6, 12, 3, low(mina)).
table(low_c_s_maxs_mina, 6, 12, 3, low(mina)).
table(low_c_mins_maxs_mina, 6, 5, 3, low(mina)).
table(low_minc_maxc_s_mina, 6, 18, 3, low(mina)).
table(low_minc_maxc_mins_mina, 6, 19, 3, low(mina)).
table(low_minc_maxc_maxs_mina, 6, 18, 3, low(mina)).
table(low_minc_s_mins_mina, 6, 12, 3, low(mina)).
table(low_minc_s_maxs_mina, 6, 12, 3, low(mina)).
table(low_minc_mins_maxs_mina, 6, 19, 3, low(mina)).
table(low_maxc_s_mins_mina, 6, 11, 3, low(mina)).
table(low_maxc_s_maxs_mina, 6, 10, 3, low(mina)).
table(low_maxc_mins_maxs_mina, 6, 9, 3, low(mina)).
table(low_s_mins_maxs_mina, 6, 10, 3, low(mina)).
table(low_v_c_minc_maxc_mina, 6, 19, 4, low(mina)).
table(low_v_c_minc_s_mina, 6, 8, 4, low(mina)).
table(low_v_c_minc_mins_mina, 6, 19, 4, low(mina)).
table(low_v_c_minc_maxs_mina, 6, 18, 4, low(mina)).
table(low_v_c_maxc_s_mina, 6, 18, 4, low(mina)).
table(low_v_c_maxc_mins_mina, 6, 19, 4, low(mina)).
table(low_v_c_maxc_maxs_mina, 6, 18, 4, low(mina)).
table(low_v_c_s_mins_mina, 6, 5, 4, low(mina)).
table(low_v_c_s_maxs_mina, 6, 12, 4, low(mina)).
table(low_v_c_mins_maxs_mina, 6, 12, 4, low(mina)).
table(low_v_minc_maxc_s_mina, 6, 18, 4, low(mina)).
table(low_v_minc_maxc_mins_mina, 6, 19, 4, low(mina)).
table(low_v_minc_maxc_maxs_mina, 6, 18, 4, low(mina)).
table(low_v_minc_s_mins_mina, 6, 6, 4, low(mina)).
table(low_v_minc_s_maxs_mina, 6, 12, 4, low(mina)).
table(low_v_minc_mins_maxs_mina, 6, 12, 4, low(mina)).
table(low_v_maxc_s_mins_mina, 6, 10, 4, low(mina)).
table(low_v_maxc_s_maxs_mina, 6, 10, 4, low(mina)).
table(low_v_maxc_mins_maxs_mina, 6, 10, 4, low(mina)).
table(low_v_s_mins_maxs_mina, 6, 10, 4, low(mina)).
table(low_c_minc_maxc_s_mina, 6, 11, 4, low(mina)).
table(low_c_minc_maxc_mins_mina, 6, 8, 4, low(mina)).
table(low_c_minc_maxc_maxs_mina, 6, 8, 4, low(mina)).
table(low_c_minc_s_mins_mina, 6, 19, 4, low(mina)).
table(low_c_minc_s_maxs_mina, 6, 18, 4, low(mina)).
table(low_c_minc_mins_maxs_mina, 6, 6, 4, low(mina)).
table(low_c_maxc_s_mins_mina, 6, 9, 4, low(mina)).
table(low_c_maxc_s_maxs_mina, 6, 18, 4, low(mina)).
table(low_c_maxc_mins_maxs_mina, 6, 6, 4, low(mina)).
table(low_c_s_mins_maxs_mina, 6, 12, 4, low(mina)).
table(low_minc_maxc_s_mins_mina, 6, 18, 4, low(mina)).
table(low_minc_maxc_s_maxs_mina, 6, 18, 4, low(mina)).
table(low_minc_maxc_mins_maxs_mina, 6, 18, 4, low(mina)).
table(low_minc_s_mins_maxs_mina, 6, 12, 4, low(mina)).
table(low_maxc_s_mins_maxs_mina, 6, 10, 4, low(mina)).
table(low_v_c_minc_maxc_s_mina, 6, 18, 5, low(mina)).
table(low_v_c_minc_maxc_mins_mina, 6, 19, 5, low(mina)).
table(low_v_c_minc_maxc_maxs_mina, 6, 18, 5, low(mina)).
table(low_v_c_minc_s_mins_mina, 6, 8, 5, low(mina)).
table(low_v_c_minc_s_maxs_mina, 6, 18, 5, low(mina)).
table(low_v_c_minc_mins_maxs_mina, 6, 18, 5, low(mina)).
table(low_v_c_maxc_s_mins_mina, 6, 18, 5, low(mina)).
table(low_v_c_maxc_s_maxs_mina, 6, 18, 5, low(mina)).
table(low_v_c_maxc_mins_maxs_mina, 6, 18, 5, low(mina)).
table(low_v_c_s_mins_maxs_mina, 6, 12, 5, low(mina)).
table(low_v_minc_maxc_s_mins_mina, 6, 18, 5, low(mina)).
table(low_v_minc_maxc_s_maxs_mina, 6, 18, 5, low(mina)).
table(low_v_minc_maxc_mins_maxs_mina, 6, 18, 5, low(mina)).
table(low_v_minc_s_mins_maxs_mina, 6, 12, 5, low(mina)).
table(low_v_maxc_s_mins_maxs_mina, 6, 10, 5, low(mina)).
table(low_c_minc_maxc_s_mins_mina, 6, 11, 5, low(mina)).
table(low_c_minc_maxc_s_maxs_mina, 6, 18, 5, low(mina)).
table(low_c_minc_maxc_mins_maxs_mina, 6, 8, 5, low(mina)).
table(low_c_minc_s_mins_maxs_mina, 6, 18, 5, low(mina)).
table(low_c_maxc_s_mins_maxs_mina, 6, 18, 5, low(mina)).
table(low_minc_maxc_s_mins_maxs_mina, 6, 18, 5, low(mina)).
table(low_v_c_minc_maxc_s_mins_mina, 6, 18, 6, low(mina)).
table(low_v_c_minc_maxc_s_maxs_mina, 6, 18, 6, low(mina)).
table(low_v_c_minc_maxc_mins_maxs_mina, 6, 18, 6, low(mina)).
table(low_v_c_minc_s_mins_maxs_mina, 6, 18, 6, low(mina)).
table(low_v_c_maxc_s_mins_maxs_mina, 6, 18, 6, low(mina)).
table(low_v_minc_maxc_s_mins_maxs_mina, 6, 18, 6, low(mina)).
table(low_c_minc_maxc_s_mins_maxs_mina, 6, 18, 6, low(mina)).
table(low_v_c_minc_maxc_s_mins_maxs_mina, 6, 18, 7, low(mina)).
table(up_v_maxa, 6, 19, 1, up(maxa)).
table(up_v_c_maxa, 6, 19, 2, up(maxa)).
table(up_v_minc_maxa, 6, 19, 2, up(maxa)).
table(up_v_maxc_maxa, 6, 19, 2, up(maxa)).
table(up_v_s_maxa, 6, 19, 2, up(maxa)).
table(up_v_mins_maxa, 6, 19, 2, up(maxa)).
table(up_v_maxs_maxa, 6, 19, 2, up(maxa)).
table(up_c_maxc_maxa, 6, 19, 2, up(maxa)).
table(up_maxc_s_maxa, 6, 18, 2, up(maxa)).
table(up_s_maxs_maxa, 6, 19, 2, up(maxa)).
table(up_v_c_minc_maxa, 6, 19, 3, up(maxa)).
table(up_v_c_maxc_maxa, 6, 19, 3, up(maxa)).
table(up_v_c_s_maxa, 6, 19, 3, up(maxa)).
table(up_v_c_mins_maxa, 6, 19, 3, up(maxa)).
table(up_v_c_maxs_maxa, 6, 7, 3, up(maxa)).
table(up_v_minc_maxc_maxa, 6, 19, 3, up(maxa)).
table(up_v_minc_s_maxa, 6, 18, 3, up(maxa)).
table(up_v_minc_mins_maxa, 6, 19, 3, up(maxa)).
table(up_v_minc_maxs_maxa, 6, 19, 3, up(maxa)).
table(up_v_maxc_s_maxa, 6, 18, 3, up(maxa)).
table(up_v_maxc_mins_maxa, 6, 19, 3, up(maxa)).
table(up_v_maxc_maxs_maxa, 6, 19, 3, up(maxa)).
table(up_v_s_mins_maxa, 6, 19, 3, up(maxa)).
table(up_v_s_maxs_maxa, 6, 19, 3, up(maxa)).
table(up_v_mins_maxs_maxa, 6, 19, 3, up(maxa)).
table(up_c_minc_maxc_maxa, 6, 19, 3, up(maxa)).
table(up_c_maxc_s_maxa, 6, 18, 3, up(maxa)).
table(up_c_maxc_mins_maxa, 6, 19, 3, up(maxa)).
table(up_c_maxc_maxs_maxa, 6, 19, 3, up(maxa)).
table(up_c_s_maxs_maxa, 6, 19, 3, up(maxa)).
table(up_minc_maxc_s_maxa, 6, 18, 3, up(maxa)).
table(up_minc_s_maxs_maxa, 6, 18, 3, up(maxa)).
table(up_maxc_s_mins_maxa, 6, 18, 3, up(maxa)).
table(up_maxc_s_maxs_maxa, 6, 18, 3, up(maxa)).
table(up_s_mins_maxs_maxa, 6, 19, 3, up(maxa)).
table(up_v_c_minc_maxc_maxa, 6, 19, 4, up(maxa)).
table(up_v_c_minc_s_maxa, 6, 18, 4, up(maxa)).
table(up_v_c_minc_mins_maxa, 6, 19, 4, up(maxa)).
table(up_v_c_minc_maxs_maxa, 6, 11, 4, up(maxa)).
table(up_v_c_maxc_s_maxa, 6, 18, 4, up(maxa)).
table(up_v_c_maxc_mins_maxa, 6, 19, 4, up(maxa)).
table(up_v_c_maxc_maxs_maxa, 6, 19, 4, up(maxa)).
table(up_v_c_s_mins_maxa, 6, 19, 4, up(maxa)).
table(up_v_c_s_maxs_maxa, 6, 19, 4, up(maxa)).
table(up_v_c_mins_maxs_maxa, 6, 11, 4, up(maxa)).
table(up_v_minc_maxc_s_maxa, 6, 18, 4, up(maxa)).
table(up_v_minc_maxc_mins_maxa, 6, 19, 4, up(maxa)).
table(up_v_minc_maxc_maxs_maxa, 6, 19, 4, up(maxa)).
table(up_v_minc_s_mins_maxa, 6, 18, 4, up(maxa)).
table(up_v_minc_s_maxs_maxa, 6, 18, 4, up(maxa)).
table(up_v_minc_mins_maxs_maxa, 6, 18, 4, up(maxa)).
table(up_v_maxc_s_mins_maxa, 6, 18, 4, up(maxa)).
table(up_v_maxc_s_maxs_maxa, 6, 18, 4, up(maxa)).
table(up_v_maxc_mins_maxs_maxa, 6, 18, 4, up(maxa)).
table(up_v_s_mins_maxs_maxa, 6, 19, 4, up(maxa)).
table(up_c_minc_maxc_s_maxa, 6, 18, 4, up(maxa)).
table(up_c_minc_maxc_mins_maxa, 6, 19, 4, up(maxa)).
table(up_c_minc_maxc_maxs_maxa, 6, 19, 4, up(maxa)).
table(up_c_minc_s_maxs_maxa, 6, 18, 4, up(maxa)).
table(up_c_maxc_s_mins_maxa, 6, 18, 4, up(maxa)).
table(up_c_maxc_s_maxs_maxa, 6, 18, 4, up(maxa)).
table(up_c_maxc_mins_maxs_maxa, 6, 18, 4, up(maxa)).
table(up_c_s_mins_maxs_maxa, 6, 19, 4, up(maxa)).
table(up_minc_maxc_s_mins_maxa, 6, 18, 4, up(maxa)).
table(up_minc_maxc_s_maxs_maxa, 6, 18, 4, up(maxa)).
table(up_minc_s_mins_maxs_maxa, 6, 18, 4, up(maxa)).
table(up_maxc_s_mins_maxs_maxa, 6, 18, 4, up(maxa)).
table(up_v_c_minc_maxc_s_maxa, 6, 18, 5, up(maxa)).
table(up_v_c_minc_maxc_mins_maxa, 6, 19, 5, up(maxa)).
table(up_v_c_minc_maxc_maxs_maxa, 6, 19, 5, up(maxa)).
table(up_v_c_minc_s_mins_maxa, 6, 18, 5, up(maxa)).
table(up_v_c_minc_s_maxs_maxa, 6, 18, 5, up(maxa)).
table(up_v_c_minc_mins_maxs_maxa, 6, 10, 5, up(maxa)).
table(up_v_c_maxc_s_mins_maxa, 6, 18, 5, up(maxa)).
table(up_v_c_maxc_s_maxs_maxa, 6, 18, 5, up(maxa)).
table(up_v_c_maxc_mins_maxs_maxa, 6, 18, 5, up(maxa)).
table(up_v_c_s_mins_maxs_maxa, 6, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_mins_maxa, 6, 18, 5, up(maxa)).
table(up_v_minc_maxc_s_maxs_maxa, 6, 18, 5, up(maxa)).
table(up_v_minc_maxc_mins_maxs_maxa, 6, 18, 5, up(maxa)).
table(up_v_minc_s_mins_maxs_maxa, 6, 18, 5, up(maxa)).
table(up_v_maxc_s_mins_maxs_maxa, 6, 18, 5, up(maxa)).
table(up_c_minc_maxc_s_mins_maxa, 6, 18, 5, up(maxa)).
table(up_c_minc_maxc_s_maxs_maxa, 6, 18, 5, up(maxa)).
table(up_c_minc_maxc_mins_maxs_maxa, 6, 18, 5, up(maxa)).
table(up_c_minc_s_mins_maxs_maxa, 6, 18, 5, up(maxa)).
table(up_c_maxc_s_mins_maxs_maxa, 6, 18, 5, up(maxa)).
table(up_minc_maxc_s_mins_maxs_maxa, 6, 18, 5, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxa, 6, 18, 6, up(maxa)).
table(up_v_c_minc_maxc_s_maxs_maxa, 6, 18, 6, up(maxa)).
table(up_v_c_minc_maxc_mins_maxs_maxa, 6, 18, 6, up(maxa)).
table(up_v_c_minc_s_mins_maxs_maxa, 6, 18, 6, up(maxa)).
table(up_v_c_maxc_s_mins_maxs_maxa, 6, 18, 6, up(maxa)).
table(up_v_minc_maxc_s_mins_maxs_maxa, 6, 18, 6, up(maxa)).
table(up_c_minc_maxc_s_mins_maxs_maxa, 6, 18, 6, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxs_maxa, 6, 18, 7, up(maxa)).
table(low_c_v, 7, 18, 1, low(v)).
table(low_minc_v, 7, 11, 1, low(v)).
table(low_maxc_v, 7, 11, 1, low(v)).
table(low_s_v, 7, 7, 1, low(v)).
table(low_mins_v, 7, 18, 1, low(v)).
table(low_maxs_v, 7, 18, 1, low(v)).
table(low_c_minc_v, 7, 11, 2, low(v)).
table(low_c_maxc_v, 7, 11, 2, low(v)).
table(low_c_s_v, 7, 8, 2, low(v)).
table(low_c_mins_v, 7, 18, 2, low(v)).
table(low_c_maxs_v, 7, 18, 2, low(v)).
table(low_minc_maxc_v, 7, 11, 2, low(v)).
table(low_minc_s_v, 7, 3, 2, low(v)).
table(low_minc_mins_v, 7, 12, 2, low(v)).
table(low_minc_maxs_v, 7, 12, 2, low(v)).
table(low_maxc_s_v, 7, 3, 2, low(v)).
table(low_maxc_mins_v, 7, 12, 2, low(v)).
table(low_maxc_maxs_v, 7, 13, 2, low(v)).
table(low_s_mins_v, 7, 7, 2, low(v)).
table(low_s_maxs_v, 7, 7, 2, low(v)).
table(low_mins_maxs_v, 7, 12, 2, low(v)).
table(low_c_minc_maxc_v, 7, 11, 3, low(v)).
table(low_c_minc_s_v, 7, 6, 3, low(v)).
table(low_c_minc_mins_v, 7, 12, 3, low(v)).
table(low_c_minc_maxs_v, 7, 12, 3, low(v)).
table(low_c_maxc_s_v, 7, 6, 3, low(v)).
table(low_c_maxc_mins_v, 7, 12, 3, low(v)).
table(low_c_maxc_maxs_v, 7, 13, 3, low(v)).
table(low_c_s_mins_v, 7, 8, 3, low(v)).
table(low_c_s_maxs_v, 7, 8, 3, low(v)).
table(low_c_mins_maxs_v, 7, 18, 3, low(v)).
table(low_minc_maxc_s_v, 7, 5, 3, low(v)).
table(low_minc_maxc_mins_v, 7, 12, 3, low(v)).
table(low_minc_maxc_maxs_v, 7, 12, 3, low(v)).
table(low_minc_s_mins_v, 7, 4, 3, low(v)).
table(low_minc_s_maxs_v, 7, 4, 3, low(v)).
table(low_minc_mins_maxs_v, 7, 13, 3, low(v)).
table(low_maxc_s_mins_v, 7, 4, 3, low(v)).
table(low_maxc_s_maxs_v, 7, 4, 3, low(v)).
table(low_maxc_mins_maxs_v, 7, 14, 3, low(v)).
table(low_s_mins_maxs_v, 7, 7, 3, low(v)).
table(low_c_minc_maxc_s_v, 7, 12, 4, low(v)).
table(low_c_minc_maxc_mins_v, 7, 12, 4, low(v)).
table(low_c_minc_maxc_maxs_v, 7, 12, 4, low(v)).
table(low_c_minc_s_mins_v, 7, 7, 4, low(v)).
table(low_c_minc_s_maxs_v, 7, 7, 4, low(v)).
table(low_c_minc_mins_maxs_v, 7, 13, 4, low(v)).
table(low_c_maxc_s_mins_v, 7, 7, 4, low(v)).
table(low_c_maxc_s_maxs_v, 7, 7, 4, low(v)).
table(low_c_maxc_mins_maxs_v, 7, 14, 4, low(v)).
table(low_c_s_mins_maxs_v, 7, 8, 4, low(v)).
table(low_minc_maxc_s_mins_v, 7, 6, 4, low(v)).
table(low_minc_maxc_s_maxs_v, 7, 6, 4, low(v)).
table(low_minc_maxc_mins_maxs_v, 7, 13, 4, low(v)).
table(low_minc_s_mins_maxs_v, 7, 8, 4, low(v)).
table(low_maxc_s_mins_maxs_v, 7, 8, 4, low(v)).
table(low_c_minc_maxc_s_mins_v, 7, 13, 5, low(v)).
table(low_c_minc_maxc_s_maxs_v, 7, 13, 5, low(v)).
table(low_c_minc_maxc_mins_maxs_v, 7, 13, 5, low(v)).
table(low_c_minc_s_mins_maxs_v, 7, 11, 5, low(v)).
table(low_c_maxc_s_mins_maxs_v, 7, 11, 5, low(v)).
table(low_minc_maxc_s_mins_maxs_v, 7, 10, 5, low(v)).
table(low_c_minc_maxc_s_mins_maxs_v, 7, 17, 6, low(v)).
table(up_c_maxc_v, 7, 5, 2, up(v)).
table(up_maxc_s_v, 7, 3, 2, up(v)).
table(up_s_maxs_v, 7, 3, 2, up(v)).
table(up_c_minc_maxc_v, 7, 11, 3, up(v)).
table(up_c_maxc_s_v, 7, 6, 3, up(v)).
table(up_c_maxc_mins_v, 7, 6, 3, up(v)).
table(up_c_maxc_maxs_v, 7, 6, 3, up(v)).
table(up_c_s_maxs_v, 7, 4, 3, up(v)).
table(up_minc_maxc_s_v, 7, 5, 3, up(v)).
table(up_minc_s_maxs_v, 7, 4, 3, up(v)).
table(up_maxc_s_mins_v, 7, 4, 3, up(v)).
table(up_maxc_s_maxs_v, 7, 4, 3, up(v)).
table(up_s_mins_maxs_v, 7, 7, 3, up(v)).
table(up_c_minc_maxc_s_v, 7, 12, 4, up(v)).
table(up_c_minc_maxc_mins_v, 7, 12, 4, up(v)).
table(up_c_minc_maxc_maxs_v, 7, 12, 4, up(v)).
table(up_c_minc_s_maxs_v, 7, 7, 4, up(v)).
table(up_c_maxc_s_mins_v, 7, 7, 4, up(v)).
table(up_c_maxc_s_maxs_v, 7, 7, 4, up(v)).
table(up_c_maxc_mins_maxs_v, 7, 7, 4, up(v)).
table(up_c_s_mins_maxs_v, 7, 8, 4, up(v)).
table(up_minc_maxc_s_mins_v, 7, 6, 4, up(v)).
table(up_minc_maxc_s_maxs_v, 7, 6, 4, up(v)).
table(up_minc_s_mins_maxs_v, 7, 8, 4, up(v)).
table(up_maxc_s_mins_maxs_v, 7, 8, 4, up(v)).
table(up_c_minc_maxc_s_mins_v, 7, 13, 5, up(v)).
table(up_c_minc_maxc_s_maxs_v, 7, 13, 5, up(v)).
table(up_c_minc_maxc_mins_maxs_v, 7, 13, 5, up(v)).
table(up_c_minc_s_mins_maxs_v, 7, 11, 5, up(v)).
table(up_c_maxc_s_mins_maxs_v, 7, 11, 5, up(v)).
table(up_minc_maxc_s_mins_maxs_v, 7, 10, 5, up(v)).
table(up_c_minc_maxc_s_mins_maxs_v, 7, 17, 6, up(v)).
table(low_v_c, 7, 11, 1, low(c)).
table(low_minc_c, 7, 11, 1, low(c)).
table(low_maxc_c, 7, 11, 1, low(c)).
table(low_s_c, 7, 8, 1, low(c)).
table(low_mins_c, 7, 8, 1, low(c)).
table(low_maxs_c, 7, 8, 1, low(c)).
table(low_v_minc_c, 7, 11, 2, low(c)).
table(low_v_maxc_c, 7, 5, 2, low(c)).
table(low_v_s_c, 7, 13, 2, low(c)).
table(low_v_mins_c, 7, 12, 2, low(c)).
table(low_v_maxs_c, 7, 13, 2, low(c)).
table(low_minc_maxc_c, 7, 11, 2, low(c)).
table(low_minc_s_c, 7, 5, 2, low(c)).
table(low_minc_mins_c, 7, 12, 2, low(c)).
table(low_minc_maxs_c, 7, 9, 2, low(c)).
table(low_maxc_s_c, 7, 5, 2, low(c)).
table(low_maxc_mins_c, 7, 12, 2, low(c)).
table(low_maxc_maxs_c, 7, 13, 2, low(c)).
table(low_s_mins_c, 7, 10, 2, low(c)).
table(low_s_maxs_c, 7, 11, 2, low(c)).
table(low_mins_maxs_c, 7, 9, 2, low(c)).
table(low_v_minc_maxc_c, 7, 11, 3, low(c)).
table(low_v_minc_s_c, 7, 12, 3, low(c)).
table(low_v_minc_mins_c, 7, 12, 3, low(c)).
table(low_v_minc_maxs_c, 7, 12, 3, low(c)).
table(low_v_maxc_s_c, 7, 6, 3, low(c)).
table(low_v_maxc_mins_c, 7, 6, 3, low(c)).
table(low_v_maxc_maxs_c, 7, 6, 3, low(c)).
table(low_v_s_mins_c, 7, 14, 3, low(c)).
table(low_v_s_maxs_c, 7, 14, 3, low(c)).
table(low_v_mins_maxs_c, 7, 14, 3, low(c)).
table(low_minc_maxc_s_c, 7, 5, 3, low(c)).
table(low_minc_maxc_mins_c, 7, 12, 3, low(c)).
table(low_minc_maxc_maxs_c, 7, 12, 3, low(c)).
table(low_minc_s_mins_c, 7, 6, 3, low(c)).
table(low_minc_s_maxs_c, 7, 7, 3, low(c)).
table(low_minc_mins_maxs_c, 7, 7, 3, low(c)).
table(low_maxc_s_mins_c, 7, 6, 3, low(c)).
table(low_maxc_s_maxs_c, 7, 6, 3, low(c)).
table(low_maxc_mins_maxs_c, 7, 6, 3, low(c)).
table(low_s_mins_maxs_c, 7, 12, 3, low(c)).
table(low_v_minc_maxc_s_c, 7, 12, 4, low(c)).
table(low_v_minc_maxc_mins_c, 7, 12, 4, low(c)).
table(low_v_minc_maxc_maxs_c, 7, 12, 4, low(c)).
table(low_v_minc_s_mins_c, 7, 13, 4, low(c)).
table(low_v_minc_s_maxs_c, 7, 13, 4, low(c)).
table(low_v_minc_mins_maxs_c, 7, 13, 4, low(c)).
table(low_v_maxc_s_mins_c, 7, 7, 4, low(c)).
table(low_v_maxc_s_maxs_c, 7, 7, 4, low(c)).
table(low_v_maxc_mins_maxs_c, 7, 7, 4, low(c)).
table(low_v_s_mins_maxs_c, 7, 18, 4, low(c)).
table(low_minc_maxc_s_mins_c, 7, 6, 4, low(c)).
table(low_minc_maxc_s_maxs_c, 7, 6, 4, low(c)).
table(low_minc_maxc_mins_maxs_c, 7, 13, 4, low(c)).
table(low_minc_s_mins_maxs_c, 7, 8, 4, low(c)).
table(low_maxc_s_mins_maxs_c, 7, 7, 4, low(c)).
table(low_v_minc_maxc_s_mins_c, 7, 13, 5, low(c)).
table(low_v_minc_maxc_s_maxs_c, 7, 13, 5, low(c)).
table(low_v_minc_maxc_mins_maxs_c, 7, 13, 5, low(c)).
table(low_v_minc_s_mins_maxs_c, 7, 17, 5, low(c)).
table(low_v_maxc_s_mins_maxs_c, 7, 11, 5, low(c)).
table(low_minc_maxc_s_mins_maxs_c, 7, 7, 5, low(c)).
table(low_v_minc_maxc_s_mins_maxs_c, 7, 17, 6, low(c)).
table(up_v_c, 7, 18, 1, up(c)).
table(up_s_c, 7, 3, 1, up(c)).
table(up_v_minc_c, 7, 11, 2, up(c)).
table(up_v_maxc_c, 7, 11, 2, up(c)).
table(up_v_s_c, 7, 4, 2, up(c)).
table(up_v_mins_c, 7, 18, 2, up(c)).
table(up_v_maxs_c, 7, 18, 2, up(c)).
table(up_minc_s_c, 7, 3, 2, up(c)).
table(up_maxc_s_c, 7, 4, 2, up(c)).
table(up_s_mins_c, 7, 5, 2, up(c)).
table(up_s_maxs_c, 7, 5, 2, up(c)).
table(up_v_minc_maxc_c, 7, 11, 3, up(c)).
table(up_v_minc_s_c, 7, 6, 3, up(c)).
table(up_v_minc_mins_c, 7, 12, 3, up(c)).
table(up_v_minc_maxs_c, 7, 12, 3, up(c)).
table(up_v_maxc_s_c, 7, 7, 3, up(c)).
table(up_v_maxc_mins_c, 7, 12, 3, up(c)).
table(up_v_maxc_maxs_c, 7, 13, 3, up(c)).
table(up_v_s_mins_c, 7, 8, 3, up(c)).
table(up_v_s_maxs_c, 7, 8, 3, up(c)).
table(up_v_mins_maxs_c, 7, 18, 3, up(c)).
table(up_minc_maxc_s_c, 7, 5, 3, up(c)).
table(up_minc_s_mins_c, 7, 4, 3, up(c)).
table(up_minc_s_maxs_c, 7, 4, 3, up(c)).
table(up_maxc_s_mins_c, 7, 5, 3, up(c)).
table(up_maxc_s_maxs_c, 7, 5, 3, up(c)).
table(up_s_mins_maxs_c, 7, 8, 3, up(c)).
table(up_v_minc_maxc_s_c, 7, 12, 4, up(c)).
table(up_v_minc_maxc_mins_c, 7, 12, 4, up(c)).
table(up_v_minc_maxc_maxs_c, 7, 12, 4, up(c)).
table(up_v_minc_s_mins_c, 7, 7, 4, up(c)).
table(up_v_minc_s_maxs_c, 7, 13, 4, up(c)).
table(up_v_minc_mins_maxs_c, 7, 13, 4, up(c)).
table(up_v_maxc_s_mins_c, 7, 14, 4, up(c)).
table(up_v_maxc_s_maxs_c, 7, 8, 4, up(c)).
table(up_v_maxc_mins_maxs_c, 7, 14, 4, up(c)).
table(up_v_s_mins_maxs_c, 7, 18, 4, up(c)).
table(up_minc_maxc_s_mins_c, 7, 6, 4, up(c)).
table(up_minc_maxc_s_maxs_c, 7, 6, 4, up(c)).
table(up_minc_s_mins_maxs_c, 7, 5, 4, up(c)).
table(up_maxc_s_mins_maxs_c, 7, 6, 4, up(c)).
table(up_v_minc_maxc_s_mins_c, 7, 13, 5, up(c)).
table(up_v_minc_maxc_s_maxs_c, 7, 13, 5, up(c)).
table(up_v_minc_maxc_mins_maxs_c, 7, 13, 5, up(c)).
table(up_v_minc_s_mins_maxs_c, 7, 17, 5, up(c)).
table(up_v_maxc_s_mins_maxs_c, 7, 18, 5, up(c)).
table(up_minc_maxc_s_mins_maxs_c, 7, 7, 5, up(c)).
table(up_v_minc_maxc_s_mins_maxs_c, 7, 17, 6, up(c)).
table(low_v_minc, 7, 3, 1, low(minc)).
table(low_c_minc, 7, 3, 1, low(minc)).
table(low_maxc_minc, 7, 3, 1, low(minc)).
table(low_s_minc, 7, 3, 1, low(minc)).
table(low_mins_minc, 7, 2, 1, low(minc)).
table(low_maxs_minc, 7, 3, 1, low(minc)).
table(low_v_c_minc, 7, 5, 2, low(minc)).
table(low_v_maxc_minc, 7, 4, 2, low(minc)).
table(low_v_s_minc, 7, 4, 2, low(minc)).
table(low_v_mins_minc, 7, 3, 2, low(minc)).
table(low_v_maxs_minc, 7, 4, 2, low(minc)).
table(low_c_maxc_minc, 7, 4, 2, low(minc)).
table(low_c_s_minc, 7, 4, 2, low(minc)).
table(low_c_mins_minc, 7, 3, 2, low(minc)).
table(low_c_maxs_minc, 7, 4, 2, low(minc)).
table(low_maxc_s_minc, 7, 4, 2, low(minc)).
table(low_maxc_mins_minc, 7, 4, 2, low(minc)).
table(low_maxc_maxs_minc, 7, 4, 2, low(minc)).
table(low_s_mins_minc, 7, 3, 2, low(minc)).
table(low_s_maxs_minc, 7, 4, 2, low(minc)).
table(low_mins_maxs_minc, 7, 3, 2, low(minc)).
table(low_v_c_maxc_minc, 7, 11, 3, low(minc)).
table(low_v_c_s_minc, 7, 6, 3, low(minc)).
table(low_v_c_mins_minc, 7, 6, 3, low(minc)).
table(low_v_c_maxs_minc, 7, 6, 3, low(minc)).
table(low_v_maxc_s_minc, 7, 5, 3, low(minc)).
table(low_v_maxc_mins_minc, 7, 5, 3, low(minc)).
table(low_v_maxc_maxs_minc, 7, 5, 3, low(minc)).
table(low_v_s_mins_minc, 7, 4, 3, low(minc)).
table(low_v_s_maxs_minc, 7, 8, 3, low(minc)).
table(low_v_mins_maxs_minc, 7, 4, 3, low(minc)).
table(low_c_maxc_s_minc, 7, 5, 3, low(minc)).
table(low_c_maxc_mins_minc, 7, 5, 3, low(minc)).
table(low_c_maxc_maxs_minc, 7, 5, 3, low(minc)).
table(low_c_s_mins_minc, 7, 4, 3, low(minc)).
table(low_c_s_maxs_minc, 7, 6, 3, low(minc)).
table(low_c_mins_maxs_minc, 7, 4, 3, low(minc)).
table(low_maxc_s_mins_minc, 7, 5, 3, low(minc)).
table(low_maxc_s_maxs_minc, 7, 5, 3, low(minc)).
table(low_maxc_mins_maxs_minc, 7, 5, 3, low(minc)).
table(low_s_mins_maxs_minc, 7, 4, 3, low(minc)).
table(low_v_c_maxc_s_minc, 7, 12, 4, low(minc)).
table(low_v_c_maxc_mins_minc, 7, 12, 4, low(minc)).
table(low_v_c_maxc_maxs_minc, 7, 12, 4, low(minc)).
table(low_v_c_s_mins_minc, 7, 7, 4, low(minc)).
table(low_v_c_s_maxs_minc, 7, 7, 4, low(minc)).
table(low_v_c_mins_maxs_minc, 7, 7, 4, low(minc)).
table(low_v_maxc_s_mins_minc, 7, 6, 4, low(minc)).
table(low_v_maxc_s_maxs_minc, 7, 6, 4, low(minc)).
table(low_v_maxc_mins_maxs_minc, 7, 6, 4, low(minc)).
table(low_v_s_mins_maxs_minc, 7, 8, 4, low(minc)).
table(low_c_maxc_s_mins_minc, 7, 6, 4, low(minc)).
table(low_c_maxc_s_maxs_minc, 7, 6, 4, low(minc)).
table(low_c_maxc_mins_maxs_minc, 7, 6, 4, low(minc)).
table(low_c_s_mins_maxs_minc, 7, 6, 4, low(minc)).
table(low_maxc_s_mins_maxs_minc, 7, 6, 4, low(minc)).
table(low_v_c_maxc_s_mins_minc, 7, 13, 5, low(minc)).
table(low_v_c_maxc_s_maxs_minc, 7, 13, 5, low(minc)).
table(low_v_c_maxc_mins_maxs_minc, 7, 13, 5, low(minc)).
table(low_v_c_s_mins_maxs_minc, 7, 11, 5, low(minc)).
table(low_v_maxc_s_mins_maxs_minc, 7, 10, 5, low(minc)).
table(low_c_maxc_s_mins_maxs_minc, 7, 7, 5, low(minc)).
table(low_v_c_maxc_s_mins_maxs_minc, 7, 17, 6, low(minc)).
table(up_v_minc, 7, 11, 1, up(minc)).
table(up_maxc_minc, 7, 6, 1, up(minc)).
table(up_v_c_minc, 7, 5, 2, up(minc)).
table(up_v_maxc_minc, 7, 6, 2, up(minc)).
table(up_v_s_minc, 7, 13, 2, up(minc)).
table(up_v_mins_minc, 7, 12, 2, up(minc)).
table(up_v_maxs_minc, 7, 13, 2, up(minc)).
table(up_c_maxc_minc, 7, 4, 2, up(minc)).
table(up_maxc_s_minc, 7, 6, 2, up(minc)).
table(up_maxc_mins_minc, 7, 7, 2, up(minc)).
table(up_maxc_maxs_minc, 7, 7, 2, up(minc)).
table(up_s_maxs_minc, 7, 14, 2, up(minc)).
table(up_v_c_maxc_minc, 7, 11, 3, up(minc)).
table(up_v_c_s_minc, 7, 6, 3, up(minc)).
table(up_v_c_mins_minc, 7, 6, 3, up(minc)).
table(up_v_c_maxs_minc, 7, 6, 3, up(minc)).
table(up_v_maxc_s_minc, 7, 7, 3, up(minc)).
table(up_v_maxc_mins_minc, 7, 7, 3, up(minc)).
table(up_v_maxc_maxs_minc, 7, 7, 3, up(minc)).
table(up_v_s_mins_minc, 7, 14, 3, up(minc)).
table(up_v_s_maxs_minc, 7, 14, 3, up(minc)).
table(up_v_mins_maxs_minc, 7, 14, 3, up(minc)).
table(up_c_maxc_s_minc, 7, 5, 3, up(minc)).
table(up_c_maxc_mins_minc, 7, 5, 3, up(minc)).
table(up_c_maxc_maxs_minc, 7, 5, 3, up(minc)).
table(up_c_s_maxs_minc, 7, 4, 3, up(minc)).
table(up_maxc_s_mins_minc, 7, 7, 3, up(minc)).
table(up_maxc_s_maxs_minc, 7, 7, 3, up(minc)).
table(up_maxc_mins_maxs_minc, 7, 8, 3, up(minc)).
table(up_s_mins_maxs_minc, 7, 18, 3, up(minc)).
table(up_v_c_maxc_s_minc, 7, 12, 4, up(minc)).
table(up_v_c_maxc_mins_minc, 7, 12, 4, up(minc)).
table(up_v_c_maxc_maxs_minc, 7, 12, 4, up(minc)).
table(up_v_c_s_mins_minc, 7, 7, 4, up(minc)).
table(up_v_c_s_maxs_minc, 7, 7, 4, up(minc)).
table(up_v_c_mins_maxs_minc, 7, 7, 4, up(minc)).
table(up_v_maxc_s_mins_minc, 7, 8, 4, up(minc)).
table(up_v_maxc_s_maxs_minc, 7, 8, 4, up(minc)).
table(up_v_maxc_mins_maxs_minc, 7, 8, 4, up(minc)).
table(up_v_s_mins_maxs_minc, 7, 18, 4, up(minc)).
table(up_c_maxc_s_mins_minc, 7, 6, 4, up(minc)).
table(up_c_maxc_s_maxs_minc, 7, 6, 4, up(minc)).
table(up_c_maxc_mins_maxs_minc, 7, 6, 4, up(minc)).
table(up_c_s_mins_maxs_minc, 7, 5, 4, up(minc)).
table(up_maxc_s_mins_maxs_minc, 7, 12, 4, up(minc)).
table(up_v_c_maxc_s_mins_minc, 7, 13, 5, up(minc)).
table(up_v_c_maxc_s_maxs_minc, 7, 13, 5, up(minc)).
table(up_v_c_maxc_mins_maxs_minc, 7, 13, 5, up(minc)).
table(up_v_c_s_mins_maxs_minc, 7, 11, 5, up(minc)).
table(up_v_maxc_s_mins_maxs_minc, 7, 12, 5, up(minc)).
table(up_c_maxc_s_mins_maxs_minc, 7, 7, 5, up(minc)).
table(up_v_c_maxc_s_mins_maxs_minc, 7, 17, 6, up(minc)).
table(low_v_maxc, 7, 18, 1, low(maxc)).
table(low_c_maxc, 7, 18, 1, low(maxc)).
table(low_minc_maxc, 7, 6, 1, low(maxc)).
table(low_s_maxc, 7, 18, 1, low(maxc)).
table(low_mins_maxc, 7, 9, 1, low(maxc)).
table(low_maxs_maxc, 7, 2, 1, low(maxc)).
table(low_v_c_maxc, 7, 5, 2, low(maxc)).
table(low_v_minc_maxc, 7, 11, 2, low(maxc)).
table(low_v_s_maxc, 7, 6, 2, low(maxc)).
table(low_v_mins_maxc, 7, 18, 2, low(maxc)).
table(low_v_maxs_maxc, 7, 3, 2, low(maxc)).
table(low_c_minc_maxc, 7, 11, 2, low(maxc)).
table(low_c_s_maxc, 7, 6, 2, low(maxc)).
table(low_c_mins_maxc, 7, 18, 2, low(maxc)).
table(low_c_maxs_maxc, 7, 3, 2, low(maxc)).
table(low_minc_s_maxc, 7, 7, 2, low(maxc)).
table(low_minc_mins_maxc, 7, 7, 2, low(maxc)).
table(low_minc_maxs_maxc, 7, 4, 2, low(maxc)).
table(low_s_mins_maxc, 7, 18, 2, low(maxc)).
table(low_s_maxs_maxc, 7, 3, 2, low(maxc)).
table(low_mins_maxs_maxc, 7, 3, 2, low(maxc)).
table(low_v_c_minc_maxc, 7, 11, 3, low(maxc)).
table(low_v_c_s_maxc, 7, 6, 3, low(maxc)).
table(low_v_c_mins_maxc, 7, 6, 3, low(maxc)).
table(low_v_c_maxs_maxc, 7, 6, 3, low(maxc)).
table(low_v_minc_s_maxc, 7, 7, 3, low(maxc)).
table(low_v_minc_mins_maxc, 7, 12, 3, low(maxc)).
table(low_v_minc_maxs_maxc, 7, 5, 3, low(maxc)).
table(low_v_s_mins_maxc, 7, 8, 3, low(maxc)).
table(low_v_s_maxs_maxc, 7, 4, 3, low(maxc)).
table(low_v_mins_maxs_maxc, 7, 4, 3, low(maxc)).
table(low_c_minc_s_maxc, 7, 5, 3, low(maxc)).
table(low_c_minc_mins_maxc, 7, 12, 3, low(maxc)).
table(low_c_minc_maxs_maxc, 7, 5, 3, low(maxc)).
table(low_c_s_mins_maxc, 7, 6, 3, low(maxc)).
table(low_c_s_maxs_maxc, 7, 4, 3, low(maxc)).
table(low_c_mins_maxs_maxc, 7, 4, 3, low(maxc)).
table(low_minc_s_mins_maxc, 7, 8, 3, low(maxc)).
table(low_minc_s_maxs_maxc, 7, 5, 3, low(maxc)).
table(low_minc_mins_maxs_maxc, 7, 5, 3, low(maxc)).
table(low_s_mins_maxs_maxc, 7, 4, 3, low(maxc)).
table(low_v_c_minc_s_maxc, 7, 12, 4, low(maxc)).
table(low_v_c_minc_mins_maxc, 7, 12, 4, low(maxc)).
table(low_v_c_minc_maxs_maxc, 7, 12, 4, low(maxc)).
table(low_v_c_s_mins_maxc, 7, 7, 4, low(maxc)).
table(low_v_c_s_maxs_maxc, 7, 7, 4, low(maxc)).
table(low_v_c_mins_maxs_maxc, 7, 7, 4, low(maxc)).
table(low_v_minc_s_mins_maxc, 7, 8, 4, low(maxc)).
table(low_v_minc_s_maxs_maxc, 7, 6, 4, low(maxc)).
table(low_v_minc_mins_maxs_maxc, 7, 6, 4, low(maxc)).
table(low_v_s_mins_maxs_maxc, 7, 8, 4, low(maxc)).
table(low_c_minc_s_mins_maxc, 7, 6, 4, low(maxc)).
table(low_c_minc_s_maxs_maxc, 7, 6, 4, low(maxc)).
table(low_c_minc_mins_maxs_maxc, 7, 6, 4, low(maxc)).
table(low_c_s_mins_maxs_maxc, 7, 5, 4, low(maxc)).
table(low_minc_s_mins_maxs_maxc, 7, 6, 4, low(maxc)).
table(low_v_c_minc_s_mins_maxc, 7, 13, 5, low(maxc)).
table(low_v_c_minc_s_maxs_maxc, 7, 13, 5, low(maxc)).
table(low_v_c_minc_mins_maxs_maxc, 7, 13, 5, low(maxc)).
table(low_v_c_s_mins_maxs_maxc, 7, 11, 5, low(maxc)).
table(low_v_minc_s_mins_maxs_maxc, 7, 10, 5, low(maxc)).
table(low_c_minc_s_mins_maxs_maxc, 7, 7, 5, low(maxc)).
table(low_v_c_minc_s_mins_maxs_maxc, 7, 17, 6, low(maxc)).
table(up_v_maxc, 7, 11, 1, up(maxc)).
table(up_v_c_maxc, 7, 11, 2, up(maxc)).
table(up_v_minc_maxc, 7, 11, 2, up(maxc)).
table(up_v_s_maxc, 7, 13, 2, up(maxc)).
table(up_v_mins_maxc, 7, 12, 2, up(maxc)).
table(up_v_maxs_maxc, 7, 13, 2, up(maxc)).
table(up_v_c_minc_maxc, 7, 11, 3, up(maxc)).
table(up_v_c_s_maxc, 7, 13, 3, up(maxc)).
table(up_v_c_mins_maxc, 7, 12, 3, up(maxc)).
table(up_v_c_maxs_maxc, 7, 13, 3, up(maxc)).
table(up_v_minc_s_maxc, 7, 12, 3, up(maxc)).
table(up_v_minc_mins_maxc, 7, 12, 3, up(maxc)).
table(up_v_minc_maxs_maxc, 7, 12, 3, up(maxc)).
table(up_v_s_mins_maxc, 7, 14, 3, up(maxc)).
table(up_v_s_maxs_maxc, 7, 14, 3, up(maxc)).
table(up_v_mins_maxs_maxc, 7, 14, 3, up(maxc)).
table(up_v_c_minc_s_maxc, 7, 12, 4, up(maxc)).
table(up_v_c_minc_mins_maxc, 7, 12, 4, up(maxc)).
table(up_v_c_minc_maxs_maxc, 7, 12, 4, up(maxc)).
table(up_v_c_s_mins_maxc, 7, 14, 4, up(maxc)).
table(up_v_c_s_maxs_maxc, 7, 8, 4, up(maxc)).
table(up_v_c_mins_maxs_maxc, 7, 14, 4, up(maxc)).
table(up_v_minc_s_mins_maxc, 7, 13, 4, up(maxc)).
table(up_v_minc_s_maxs_maxc, 7, 13, 4, up(maxc)).
table(up_v_minc_mins_maxs_maxc, 7, 13, 4, up(maxc)).
table(up_v_s_mins_maxs_maxc, 7, 18, 4, up(maxc)).
table(up_v_c_minc_s_mins_maxc, 7, 13, 5, up(maxc)).
table(up_v_c_minc_s_maxs_maxc, 7, 13, 5, up(maxc)).
table(up_v_c_minc_mins_maxs_maxc, 7, 13, 5, up(maxc)).
table(up_v_c_s_mins_maxs_maxc, 7, 18, 5, up(maxc)).
table(up_v_minc_s_mins_maxs_maxc, 7, 17, 5, up(maxc)).
table(up_v_c_minc_s_mins_maxs_maxc, 7, 17, 6, up(maxc)).
table(low_v_s, 7, 18, 1, low(s)).
table(low_c_s, 7, 3, 1, low(s)).
table(low_minc_s, 7, 18, 1, low(s)).
table(low_maxc_s, 7, 18, 1, low(s)).
table(low_mins_s, 7, 18, 1, low(s)).
table(low_maxs_s, 7, 18, 1, low(s)).
table(low_v_c_s, 7, 4, 2, low(s)).
table(low_v_minc_s, 7, 18, 2, low(s)).
table(low_v_maxc_s, 7, 8, 2, low(s)).
table(low_v_mins_s, 7, 9, 2, low(s)).
table(low_v_maxs_s, 7, 3, 2, low(s)).
table(low_c_minc_s, 7, 5, 2, low(s)).
table(low_c_maxc_s, 7, 5, 2, low(s)).
table(low_c_mins_s, 7, 5, 2, low(s)).
table(low_c_maxs_s, 7, 5, 2, low(s)).
table(low_minc_maxc_s, 7, 18, 2, low(s)).
table(low_minc_mins_s, 7, 18, 2, low(s)).
table(low_minc_maxs_s, 7, 14, 2, low(s)).
table(low_maxc_mins_s, 7, 7, 2, low(s)).
table(low_maxc_maxs_s, 7, 14, 2, low(s)).
table(low_mins_maxs_s, 7, 12, 2, low(s)).
table(low_v_c_minc_s, 7, 8, 3, low(s)).
table(low_v_c_maxc_s, 7, 8, 3, low(s)).
table(low_v_c_mins_s, 7, 8, 3, low(s)).
table(low_v_c_maxs_s, 7, 4, 3, low(s)).
table(low_v_minc_maxc_s, 7, 18, 3, low(s)).
table(low_v_minc_mins_s, 7, 13, 3, low(s)).
table(low_v_minc_maxs_s, 7, 4, 3, low(s)).
table(low_v_maxc_mins_s, 7, 4, 3, low(s)).
table(low_v_maxc_maxs_s, 7, 4, 3, low(s)).
table(low_v_mins_maxs_s, 7, 7, 3, low(s)).
table(low_c_minc_maxc_s, 7, 8, 3, low(s)).
table(low_c_minc_mins_s, 7, 4, 3, low(s)).
table(low_c_minc_maxs_s, 7, 5, 3, low(s)).
table(low_c_maxc_mins_s, 7, 8, 3, low(s)).
table(low_c_maxc_maxs_s, 7, 5, 3, low(s)).
table(low_c_mins_maxs_s, 7, 8, 3, low(s)).
table(low_minc_maxc_mins_s, 7, 13, 3, low(s)).
table(low_minc_maxc_maxs_s, 7, 14, 3, low(s)).
table(low_minc_mins_maxs_s, 7, 8, 3, low(s)).
table(low_maxc_mins_maxs_s, 7, 10, 3, low(s)).
table(low_v_c_minc_maxc_s, 7, 18, 4, low(s)).
table(low_v_c_minc_mins_s, 7, 7, 4, low(s)).
table(low_v_c_minc_maxs_s, 7, 7, 4, low(s)).
table(low_v_c_maxc_mins_s, 7, 13, 4, low(s)).
table(low_v_c_maxc_maxs_s, 7, 7, 4, low(s)).
table(low_v_c_mins_maxs_s, 7, 8, 4, low(s)).
table(low_v_minc_maxc_mins_s, 7, 13, 4, low(s)).
table(low_v_minc_maxc_maxs_s, 7, 6, 4, low(s)).
table(low_v_minc_mins_maxs_s, 7, 8, 4, low(s)).
table(low_v_maxc_mins_maxs_s, 7, 8, 4, low(s)).
table(low_c_minc_maxc_mins_s, 7, 6, 4, low(s)).
table(low_c_minc_maxc_maxs_s, 7, 7, 4, low(s)).
table(low_c_minc_mins_maxs_s, 7, 5, 4, low(s)).
table(low_c_maxc_mins_maxs_s, 7, 6, 4, low(s)).
table(low_minc_maxc_mins_maxs_s, 7, 17, 4, low(s)).
table(low_v_c_minc_maxc_mins_s, 7, 13, 5, low(s)).
table(low_v_c_minc_maxc_maxs_s, 7, 14, 5, low(s)).
table(low_v_c_minc_mins_maxs_s, 7, 11, 5, low(s)).
table(low_v_c_maxc_mins_maxs_s, 7, 11, 5, low(s)).
table(low_v_minc_maxc_mins_maxs_s, 7, 10, 5, low(s)).
table(low_c_minc_maxc_mins_maxs_s, 7, 7, 5, low(s)).
table(low_v_c_minc_maxc_mins_maxs_s, 7, 17, 6, low(s)).
table(up_v_s, 7, 7, 1, up(s)).
table(up_v_c_s, 7, 8, 2, up(s)).
table(up_v_minc_s, 7, 8, 2, up(s)).
table(up_v_maxc_s, 7, 8, 2, up(s)).
table(up_v_mins_s, 7, 7, 2, up(s)).
table(up_v_maxs_s, 7, 7, 2, up(s)).
table(up_c_maxc_s, 7, 11, 2, up(s)).
table(up_v_c_minc_s, 7, 11, 3, up(s)).
table(up_v_c_maxc_s, 7, 11, 3, up(s)).
table(up_v_c_mins_s, 7, 8, 3, up(s)).
table(up_v_c_maxs_s, 7, 8, 3, up(s)).
table(up_v_minc_maxc_s, 7, 10, 3, up(s)).
table(up_v_minc_mins_s, 7, 8, 3, up(s)).
table(up_v_minc_maxs_s, 7, 8, 3, up(s)).
table(up_v_maxc_mins_s, 7, 8, 3, up(s)).
table(up_v_maxc_maxs_s, 7, 8, 3, up(s)).
table(up_v_mins_maxs_s, 7, 7, 3, up(s)).
table(up_c_minc_maxc_s, 7, 18, 3, up(s)).
table(up_c_maxc_mins_s, 7, 7, 3, up(s)).
table(up_c_maxc_maxs_s, 7, 11, 3, up(s)).
table(up_v_c_minc_maxc_s, 7, 18, 4, up(s)).
table(up_v_c_minc_mins_s, 7, 11, 4, up(s)).
table(up_v_c_minc_maxs_s, 7, 11, 4, up(s)).
table(up_v_c_maxc_mins_s, 7, 11, 4, up(s)).
table(up_v_c_maxc_maxs_s, 7, 11, 4, up(s)).
table(up_v_c_mins_maxs_s, 7, 8, 4, up(s)).
table(up_v_minc_maxc_mins_s, 7, 10, 4, up(s)).
table(up_v_minc_maxc_maxs_s, 7, 10, 4, up(s)).
table(up_v_minc_mins_maxs_s, 7, 8, 4, up(s)).
table(up_v_maxc_mins_maxs_s, 7, 8, 4, up(s)).
table(up_c_minc_maxc_mins_s, 7, 18, 4, up(s)).
table(up_c_minc_maxc_maxs_s, 7, 17, 4, up(s)).
table(up_c_maxc_mins_maxs_s, 7, 11, 4, up(s)).
table(up_v_c_minc_maxc_mins_s, 7, 18, 5, up(s)).
table(up_v_c_minc_maxc_maxs_s, 7, 17, 5, up(s)).
table(up_v_c_minc_mins_maxs_s, 7, 11, 5, up(s)).
table(up_v_c_maxc_mins_maxs_s, 7, 11, 5, up(s)).
table(up_v_minc_maxc_mins_maxs_s, 7, 10, 5, up(s)).
table(up_c_minc_maxc_mins_maxs_s, 7, 17, 5, up(s)).
table(up_v_c_minc_maxc_mins_maxs_s, 7, 17, 6, up(s)).
table(low_v_mins, 7, 2, 1, low(mins)).
table(low_c_mins, 7, 2, 1, low(mins)).
table(low_minc_mins, 7, 2, 1, low(mins)).
table(low_maxc_mins, 7, 2, 1, low(mins)).
table(low_s_mins, 7, 2, 1, low(mins)).
table(low_maxs_mins, 7, 2, 1, low(mins)).
table(low_v_c_mins, 7, 3, 2, low(mins)).
table(low_v_minc_mins, 7, 3, 2, low(mins)).
table(low_v_maxc_mins, 7, 3, 2, low(mins)).
table(low_v_s_mins, 7, 3, 2, low(mins)).
table(low_v_maxs_mins, 7, 3, 2, low(mins)).
table(low_c_minc_mins, 7, 3, 2, low(mins)).
table(low_c_maxc_mins, 7, 3, 2, low(mins)).
table(low_c_s_mins, 7, 3, 2, low(mins)).
table(low_c_maxs_mins, 7, 3, 2, low(mins)).
table(low_minc_maxc_mins, 7, 4, 2, low(mins)).
table(low_minc_s_mins, 7, 3, 2, low(mins)).
table(low_minc_maxs_mins, 7, 3, 2, low(mins)).
table(low_maxc_s_mins, 7, 3, 2, low(mins)).
table(low_maxc_maxs_mins, 7, 3, 2, low(mins)).
table(low_s_maxs_mins, 7, 3, 2, low(mins)).
table(low_v_c_minc_mins, 7, 6, 3, low(mins)).
table(low_v_c_maxc_mins, 7, 6, 3, low(mins)).
table(low_v_c_s_mins, 7, 4, 3, low(mins)).
table(low_v_c_maxs_mins, 7, 4, 3, low(mins)).
table(low_v_minc_maxc_mins, 7, 5, 3, low(mins)).
table(low_v_minc_s_mins, 7, 4, 3, low(mins)).
table(low_v_minc_maxs_mins, 7, 4, 3, low(mins)).
table(low_v_maxc_s_mins, 7, 4, 3, low(mins)).
table(low_v_maxc_maxs_mins, 7, 4, 3, low(mins)).
table(low_v_s_maxs_mins, 7, 7, 3, low(mins)).
table(low_c_minc_maxc_mins, 7, 5, 3, low(mins)).
table(low_c_minc_s_mins, 7, 4, 3, low(mins)).
table(low_c_minc_maxs_mins, 7, 4, 3, low(mins)).
table(low_c_maxc_s_mins, 7, 4, 3, low(mins)).
table(low_c_maxc_maxs_mins, 7, 4, 3, low(mins)).
table(low_c_s_maxs_mins, 7, 4, 3, low(mins)).
table(low_minc_maxc_s_mins, 7, 5, 3, low(mins)).
table(low_minc_maxc_maxs_mins, 7, 5, 3, low(mins)).
table(low_minc_s_maxs_mins, 7, 4, 3, low(mins)).
table(low_maxc_s_maxs_mins, 7, 4, 3, low(mins)).
table(low_v_c_minc_maxc_mins, 7, 12, 4, low(mins)).
table(low_v_c_minc_s_mins, 7, 7, 4, low(mins)).
table(low_v_c_minc_maxs_mins, 7, 7, 4, low(mins)).
table(low_v_c_maxc_s_mins, 7, 7, 4, low(mins)).
table(low_v_c_maxc_maxs_mins, 7, 7, 4, low(mins)).
table(low_v_c_s_maxs_mins, 7, 8, 4, low(mins)).
table(low_v_minc_maxc_s_mins, 7, 6, 4, low(mins)).
table(low_v_minc_maxc_maxs_mins, 7, 6, 4, low(mins)).
table(low_v_minc_s_maxs_mins, 7, 8, 4, low(mins)).
table(low_v_maxc_s_maxs_mins, 7, 8, 4, low(mins)).
table(low_c_minc_maxc_s_mins, 7, 6, 4, low(mins)).
table(low_c_minc_maxc_maxs_mins, 7, 6, 4, low(mins)).
table(low_c_minc_s_maxs_mins, 7, 5, 4, low(mins)).
table(low_c_maxc_s_maxs_mins, 7, 5, 4, low(mins)).
table(low_minc_maxc_s_maxs_mins, 7, 6, 4, low(mins)).
table(low_v_c_minc_maxc_s_mins, 7, 13, 5, low(mins)).
table(low_v_c_minc_maxc_maxs_mins, 7, 13, 5, low(mins)).
table(low_v_c_minc_s_maxs_mins, 7, 11, 5, low(mins)).
table(low_v_c_maxc_s_maxs_mins, 7, 11, 5, low(mins)).
table(low_v_minc_maxc_s_maxs_mins, 7, 10, 5, low(mins)).
table(low_c_minc_maxc_s_maxs_mins, 7, 7, 5, low(mins)).
table(low_v_c_minc_maxc_s_maxs_mins, 7, 17, 6, low(mins)).
table(up_v_mins, 7, 18, 1, up(mins)).
table(up_maxc_mins, 7, 9, 1, up(mins)).
table(up_maxs_mins, 7, 2, 1, up(mins)).
table(up_v_c_mins, 7, 6, 2, up(mins)).
table(up_v_minc_mins, 7, 3, 2, up(mins)).
table(up_v_maxc_mins, 7, 7, 2, up(mins)).
table(up_v_s_mins, 7, 3, 2, up(mins)).
table(up_v_maxs_mins, 7, 3, 2, up(mins)).
table(up_c_maxc_mins, 7, 5, 2, up(mins)).
table(up_c_maxs_mins, 7, 3, 2, up(mins)).
table(up_minc_maxc_mins, 7, 4, 2, up(mins)).
table(up_minc_maxs_mins, 7, 3, 2, up(mins)).
table(up_maxc_s_mins, 7, 3, 2, up(mins)).
table(up_maxc_maxs_mins, 7, 3, 2, up(mins)).
table(up_s_maxs_mins, 7, 3, 2, up(mins)).
table(up_v_c_minc_mins, 7, 6, 3, up(mins)).
table(up_v_c_maxc_mins, 7, 12, 3, up(mins)).
table(up_v_c_s_mins, 7, 4, 3, up(mins)).
table(up_v_c_maxs_mins, 7, 4, 3, up(mins)).
table(up_v_minc_maxc_mins, 7, 5, 3, up(mins)).
table(up_v_minc_s_mins, 7, 4, 3, up(mins)).
table(up_v_minc_maxs_mins, 7, 4, 3, up(mins)).
table(up_v_maxc_s_mins, 7, 4, 3, up(mins)).
table(up_v_maxc_maxs_mins, 7, 4, 3, up(mins)).
table(up_v_s_maxs_mins, 7, 7, 3, up(mins)).
table(up_c_minc_maxc_mins, 7, 5, 3, up(mins)).
table(up_c_minc_maxs_mins, 7, 4, 3, up(mins)).
table(up_c_maxc_s_mins, 7, 4, 3, up(mins)).
table(up_c_maxc_maxs_mins, 7, 4, 3, up(mins)).
table(up_c_s_maxs_mins, 7, 4, 3, up(mins)).
table(up_minc_maxc_s_mins, 7, 5, 3, up(mins)).
table(up_minc_maxc_maxs_mins, 7, 5, 3, up(mins)).
table(up_minc_s_maxs_mins, 7, 4, 3, up(mins)).
table(up_maxc_s_maxs_mins, 7, 4, 3, up(mins)).
table(up_v_c_minc_maxc_mins, 7, 12, 4, up(mins)).
table(up_v_c_minc_s_mins, 7, 7, 4, up(mins)).
table(up_v_c_minc_maxs_mins, 7, 7, 4, up(mins)).
table(up_v_c_maxc_s_mins, 7, 7, 4, up(mins)).
table(up_v_c_maxc_maxs_mins, 7, 7, 4, up(mins)).
table(up_v_c_s_maxs_mins, 7, 8, 4, up(mins)).
table(up_v_minc_maxc_s_mins, 7, 6, 4, up(mins)).
table(up_v_minc_maxc_maxs_mins, 7, 6, 4, up(mins)).
table(up_v_minc_s_maxs_mins, 7, 8, 4, up(mins)).
table(up_v_maxc_s_maxs_mins, 7, 8, 4, up(mins)).
table(up_c_minc_maxc_s_mins, 7, 6, 4, up(mins)).
table(up_c_minc_maxc_maxs_mins, 7, 6, 4, up(mins)).
table(up_c_minc_s_maxs_mins, 7, 5, 4, up(mins)).
table(up_c_maxc_s_maxs_mins, 7, 5, 4, up(mins)).
table(up_minc_maxc_s_maxs_mins, 7, 6, 4, up(mins)).
table(up_v_c_minc_maxc_s_mins, 7, 13, 5, up(mins)).
table(up_v_c_minc_maxc_maxs_mins, 7, 13, 5, up(mins)).
table(up_v_c_minc_s_maxs_mins, 7, 11, 5, up(mins)).
table(up_v_c_maxc_s_maxs_mins, 7, 11, 5, up(mins)).
table(up_v_minc_maxc_s_maxs_mins, 7, 10, 5, up(mins)).
table(up_c_minc_maxc_s_maxs_mins, 7, 7, 5, up(mins)).
table(up_v_c_minc_maxc_s_maxs_mins, 7, 17, 6, up(mins)).
table(low_v_maxs, 7, 7, 1, low(maxs)).
table(low_c_maxs, 7, 4, 1, low(maxs)).
table(low_minc_maxs, 7, 4, 1, low(maxs)).
table(low_maxc_maxs, 7, 4, 1, low(maxs)).
table(low_s_maxs, 7, 7, 1, low(maxs)).
table(low_mins_maxs, 7, 2, 1, low(maxs)).
table(low_v_c_maxs, 7, 8, 2, low(maxs)).
table(low_v_minc_maxs, 7, 8, 2, low(maxs)).
table(low_v_maxc_maxs, 7, 8, 2, low(maxs)).
table(low_v_s_maxs, 7, 3, 2, low(maxs)).
table(low_v_mins_maxs, 7, 7, 2, low(maxs)).
table(low_c_minc_maxs, 7, 5, 2, low(maxs)).
table(low_c_maxc_maxs, 7, 5, 2, low(maxs)).
table(low_c_s_maxs, 7, 8, 2, low(maxs)).
table(low_c_mins_maxs, 7, 3, 2, low(maxs)).
table(low_minc_maxc_maxs, 7, 6, 2, low(maxs)).
table(low_minc_s_maxs, 7, 3, 2, low(maxs)).
table(low_minc_mins_maxs, 7, 3, 2, low(maxs)).
table(low_maxc_s_maxs, 7, 3, 2, low(maxs)).
table(low_maxc_mins_maxs, 7, 3, 2, low(maxs)).
table(low_s_mins_maxs, 7, 7, 2, low(maxs)).
table(low_v_c_minc_maxs, 7, 11, 3, low(maxs)).
table(low_v_c_maxc_maxs, 7, 11, 3, low(maxs)).
table(low_v_c_s_maxs, 7, 4, 3, low(maxs)).
table(low_v_c_mins_maxs, 7, 8, 3, low(maxs)).
table(low_v_minc_maxc_maxs, 7, 10, 3, low(maxs)).
table(low_v_minc_s_maxs, 7, 4, 3, low(maxs)).
table(low_v_minc_mins_maxs, 7, 8, 3, low(maxs)).
table(low_v_maxc_s_maxs, 7, 4, 3, low(maxs)).
table(low_v_maxc_mins_maxs, 7, 8, 3, low(maxs)).
table(low_v_s_mins_maxs, 7, 7, 3, low(maxs)).
table(low_c_minc_maxc_maxs, 7, 7, 3, low(maxs)).
table(low_c_minc_s_maxs, 7, 4, 3, low(maxs)).
table(low_c_minc_mins_maxs, 7, 4, 3, low(maxs)).
table(low_c_maxc_s_maxs, 7, 4, 3, low(maxs)).
table(low_c_maxc_mins_maxs, 7, 5, 3, low(maxs)).
table(low_c_s_mins_maxs, 7, 8, 3, low(maxs)).
table(low_minc_maxc_s_maxs, 7, 5, 3, low(maxs)).
table(low_minc_maxc_mins_maxs, 7, 5, 3, low(maxs)).
table(low_minc_s_mins_maxs, 7, 4, 3, low(maxs)).
table(low_maxc_s_mins_maxs, 7, 4, 3, low(maxs)).
table(low_v_c_minc_maxc_maxs, 7, 18, 4, low(maxs)).
table(low_v_c_minc_s_maxs, 7, 7, 4, low(maxs)).
table(low_v_c_minc_mins_maxs, 7, 11, 4, low(maxs)).
table(low_v_c_maxc_s_maxs, 7, 7, 4, low(maxs)).
table(low_v_c_maxc_mins_maxs, 7, 11, 4, low(maxs)).
table(low_v_c_s_mins_maxs, 7, 8, 4, low(maxs)).
table(low_v_minc_maxc_s_maxs, 7, 6, 4, low(maxs)).
table(low_v_minc_maxc_mins_maxs, 7, 10, 4, low(maxs)).
table(low_v_minc_s_mins_maxs, 7, 8, 4, low(maxs)).
table(low_v_maxc_s_mins_maxs, 7, 8, 4, low(maxs)).
table(low_c_minc_maxc_s_maxs, 7, 6, 4, low(maxs)).
table(low_c_minc_maxc_mins_maxs, 7, 7, 4, low(maxs)).
table(low_c_minc_s_mins_maxs, 7, 5, 4, low(maxs)).
table(low_c_maxc_s_mins_maxs, 7, 5, 4, low(maxs)).
table(low_minc_maxc_s_mins_maxs, 7, 6, 4, low(maxs)).
table(low_v_c_minc_maxc_s_maxs, 7, 13, 5, low(maxs)).
table(low_v_c_minc_maxc_mins_maxs, 7, 18, 5, low(maxs)).
table(low_v_c_minc_s_mins_maxs, 7, 11, 5, low(maxs)).
table(low_v_c_maxc_s_mins_maxs, 7, 11, 5, low(maxs)).
table(low_v_minc_maxc_s_mins_maxs, 7, 10, 5, low(maxs)).
table(low_c_minc_maxc_s_mins_maxs, 7, 7, 5, low(maxs)).
table(low_v_c_minc_maxc_s_mins_maxs, 7, 17, 6, low(maxs)).
table(up_v_maxs, 7, 18, 1, up(maxs)).
table(up_maxc_maxs, 7, 2, 1, up(maxs)).
table(up_v_c_maxs, 7, 18, 2, up(maxs)).
table(up_v_minc_maxs, 7, 12, 2, up(maxs)).
table(up_v_maxc_maxs, 7, 3, 2, up(maxs)).
table(up_v_s_maxs, 7, 7, 2, up(maxs)).
table(up_v_mins_maxs, 7, 9, 2, up(maxs)).
table(up_c_maxc_maxs, 7, 3, 2, up(maxs)).
table(up_minc_maxc_maxs, 7, 4, 2, up(maxs)).
table(up_maxc_s_maxs, 7, 3, 2, up(maxs)).
table(up_maxc_mins_maxs, 7, 3, 2, up(maxs)).
table(up_v_c_minc_maxs, 7, 12, 3, up(maxs)).
table(up_v_c_maxc_maxs, 7, 6, 3, up(maxs)).
table(up_v_c_s_maxs, 7, 8, 3, up(maxs)).
table(up_v_c_mins_maxs, 7, 18, 3, up(maxs)).
table(up_v_minc_maxc_maxs, 7, 5, 3, up(maxs)).
table(up_v_minc_s_maxs, 7, 8, 3, up(maxs)).
table(up_v_minc_mins_maxs, 7, 14, 3, up(maxs)).
table(up_v_maxc_s_maxs, 7, 4, 3, up(maxs)).
table(up_v_maxc_mins_maxs, 7, 4, 3, up(maxs)).
table(up_v_s_mins_maxs, 7, 7, 3, up(maxs)).
table(up_c_minc_maxc_maxs, 7, 5, 3, up(maxs)).
table(up_c_maxc_s_maxs, 7, 4, 3, up(maxs)).
table(up_c_maxc_mins_maxs, 7, 4, 3, up(maxs)).
table(up_minc_maxc_s_maxs, 7, 5, 3, up(maxs)).
table(up_minc_maxc_mins_maxs, 7, 5, 3, up(maxs)).
table(up_maxc_s_mins_maxs, 7, 4, 3, up(maxs)).
table(up_v_c_minc_maxc_maxs, 7, 12, 4, up(maxs)).
table(up_v_c_minc_s_maxs, 7, 11, 4, up(maxs)).
table(up_v_c_minc_mins_maxs, 7, 13, 4, up(maxs)).
table(up_v_c_maxc_s_maxs, 7, 7, 4, up(maxs)).
table(up_v_c_maxc_mins_maxs, 7, 7, 4, up(maxs)).
table(up_v_c_s_mins_maxs, 7, 8, 4, up(maxs)).
table(up_v_minc_maxc_s_maxs, 7, 10, 4, up(maxs)).
table(up_v_minc_maxc_mins_maxs, 7, 6, 4, up(maxs)).
table(up_v_minc_s_mins_maxs, 7, 8, 4, up(maxs)).
table(up_v_maxc_s_mins_maxs, 7, 8, 4, up(maxs)).
table(up_c_minc_maxc_s_maxs, 7, 7, 4, up(maxs)).
table(up_c_minc_maxc_mins_maxs, 7, 6, 4, up(maxs)).
table(up_c_maxc_s_mins_maxs, 7, 5, 4, up(maxs)).
table(up_minc_maxc_s_mins_maxs, 7, 6, 4, up(maxs)).
table(up_v_c_minc_maxc_s_maxs, 7, 17, 5, up(maxs)).
table(up_v_c_minc_maxc_mins_maxs, 7, 13, 5, up(maxs)).
table(up_v_c_minc_s_mins_maxs, 7, 11, 5, up(maxs)).
table(up_v_c_maxc_s_mins_maxs, 7, 11, 5, up(maxs)).
table(up_v_minc_maxc_s_mins_maxs, 7, 10, 5, up(maxs)).
table(up_c_minc_maxc_s_mins_maxs, 7, 7, 5, up(maxs)).
table(up_v_c_minc_maxc_s_mins_maxs, 7, 17, 6, up(maxs)).
table(low_v_mina, 7, 12, 1, low(mina)).
table(low_c_mina, 7, 7, 1, low(mina)).
table(low_minc_mina, 7, 19, 1, low(mina)).
table(low_maxc_mina, 7, 19, 1, low(mina)).
table(low_s_mina, 7, 12, 1, low(mina)).
table(low_mins_mina, 7, 11, 1, low(mina)).
table(low_maxs_mina, 7, 11, 1, low(mina)).
table(low_v_c_mina, 7, 12, 2, low(mina)).
table(low_v_minc_mina, 7, 12, 2, low(mina)).
table(low_v_maxc_mina, 7, 11, 2, low(mina)).
table(low_v_s_mina, 7, 4, 2, low(mina)).
table(low_v_mins_mina, 7, 6, 2, low(mina)).
table(low_v_maxs_mina, 7, 10, 2, low(mina)).
table(low_c_minc_mina, 7, 8, 2, low(mina)).
table(low_c_maxc_mina, 7, 6, 2, low(mina)).
table(low_c_s_mina, 7, 12, 2, low(mina)).
table(low_c_mins_mina, 7, 7, 2, low(mina)).
table(low_c_maxs_mina, 7, 5, 2, low(mina)).
table(low_minc_maxc_mina, 7, 19, 2, low(mina)).
table(low_minc_s_mina, 7, 12, 2, low(mina)).
table(low_minc_mins_mina, 7, 19, 2, low(mina)).
table(low_minc_maxs_mina, 7, 19, 2, low(mina)).
table(low_maxc_s_mina, 7, 11, 2, low(mina)).
table(low_maxc_mins_mina, 7, 19, 2, low(mina)).
table(low_maxc_maxs_mina, 7, 19, 2, low(mina)).
table(low_s_mins_mina, 7, 12, 2, low(mina)).
table(low_s_maxs_mina, 7, 10, 2, low(mina)).
table(low_mins_maxs_mina, 7, 6, 2, low(mina)).
table(low_v_c_minc_mina, 7, 14, 3, low(mina)).
table(low_v_c_maxc_mina, 7, 19, 3, low(mina)).
table(low_v_c_s_mina, 7, 4, 3, low(mina)).
table(low_v_c_mins_mina, 7, 12, 3, low(mina)).
table(low_v_c_maxs_mina, 7, 12, 3, low(mina)).
table(low_v_minc_maxc_mina, 7, 13, 3, low(mina)).
table(low_v_minc_s_mina, 7, 5, 3, low(mina)).
table(low_v_minc_mins_mina, 7, 5, 3, low(mina)).
table(low_v_minc_maxs_mina, 7, 11, 3, low(mina)).
table(low_v_maxc_s_mina, 7, 4, 3, low(mina)).
table(low_v_maxc_mins_mina, 7, 11, 3, low(mina)).
table(low_v_maxc_maxs_mina, 7, 10, 3, low(mina)).
table(low_v_s_mins_mina, 7, 5, 3, low(mina)).
table(low_v_s_maxs_mina, 7, 5, 3, low(mina)).
table(low_v_mins_maxs_mina, 7, 10, 3, low(mina)).
table(low_c_minc_maxc_mina, 7, 8, 3, low(mina)).
table(low_c_minc_s_mina, 7, 14, 3, low(mina)).
table(low_c_minc_mins_mina, 7, 8, 3, low(mina)).
table(low_c_minc_maxs_mina, 7, 6, 3, low(mina)).
table(low_c_maxc_s_mina, 7, 9, 3, low(mina)).
table(low_c_maxc_mins_mina, 7, 6, 3, low(mina)).
table(low_c_maxc_maxs_mina, 7, 6, 3, low(mina)).
table(low_c_s_mins_mina, 7, 12, 3, low(mina)).
table(low_c_s_maxs_mina, 7, 12, 3, low(mina)).
table(low_c_mins_maxs_mina, 7, 5, 3, low(mina)).
table(low_minc_maxc_s_mina, 7, 10, 3, low(mina)).
table(low_minc_maxc_mins_mina, 7, 19, 3, low(mina)).
table(low_minc_maxc_maxs_mina, 7, 18, 3, low(mina)).
table(low_minc_s_mins_mina, 7, 12, 3, low(mina)).
table(low_minc_s_maxs_mina, 7, 11, 3, low(mina)).
table(low_minc_mins_maxs_mina, 7, 19, 3, low(mina)).
table(low_maxc_s_mins_mina, 7, 11, 3, low(mina)).
table(low_maxc_s_maxs_mina, 7, 10, 3, low(mina)).
table(low_maxc_mins_maxs_mina, 7, 9, 3, low(mina)).
table(low_s_mins_maxs_mina, 7, 10, 3, low(mina)).
table(low_v_c_minc_maxc_mina, 7, 19, 4, low(mina)).
table(low_v_c_minc_s_mina, 7, 8, 4, low(mina)).
table(low_v_c_minc_mins_mina, 7, 14, 4, low(mina)).
table(low_v_c_minc_maxs_mina, 7, 14, 4, low(mina)).
table(low_v_c_maxc_s_mina, 7, 7, 4, low(mina)).
table(low_v_c_maxc_mins_mina, 7, 19, 4, low(mina)).
table(low_v_c_maxc_maxs_mina, 7, 18, 4, low(mina)).
table(low_v_c_s_mins_mina, 7, 5, 4, low(mina)).
table(low_v_c_s_maxs_mina, 7, 7, 4, low(mina)).
table(low_v_c_mins_maxs_mina, 7, 12, 4, low(mina)).
table(low_v_minc_maxc_s_mina, 7, 10, 4, low(mina)).
table(low_v_minc_maxc_mins_mina, 7, 13, 4, low(mina)).
table(low_v_minc_maxc_maxs_mina, 7, 12, 4, low(mina)).
table(low_v_minc_s_mins_mina, 7, 5, 4, low(mina)).
table(low_v_minc_s_maxs_mina, 7, 11, 4, low(mina)).
table(low_v_minc_mins_maxs_mina, 7, 11, 4, low(mina)).
table(low_v_maxc_s_mins_mina, 7, 8, 4, low(mina)).
table(low_v_maxc_s_maxs_mina, 7, 5, 4, low(mina)).
table(low_v_maxc_mins_maxs_mina, 7, 10, 4, low(mina)).
table(low_v_s_mins_maxs_mina, 7, 10, 4, low(mina)).
table(low_c_minc_maxc_s_mina, 7, 10, 4, low(mina)).
table(low_c_minc_maxc_mins_mina, 7, 8, 4, low(mina)).
table(low_c_minc_maxc_maxs_mina, 7, 8, 4, low(mina)).
table(low_c_minc_s_mins_mina, 7, 14, 4, low(mina)).
table(low_c_minc_s_maxs_mina, 7, 14, 4, low(mina)).
table(low_c_minc_mins_maxs_mina, 7, 6, 4, low(mina)).
table(low_c_maxc_s_mins_mina, 7, 9, 4, low(mina)).
table(low_c_maxc_s_maxs_mina, 7, 18, 4, low(mina)).
table(low_c_maxc_mins_maxs_mina, 7, 6, 4, low(mina)).
table(low_c_s_mins_maxs_mina, 7, 12, 4, low(mina)).
table(low_minc_maxc_s_mins_mina, 7, 10, 4, low(mina)).
table(low_minc_maxc_s_maxs_mina, 7, 12, 4, low(mina)).
table(low_minc_maxc_mins_maxs_mina, 7, 18, 4, low(mina)).
table(low_minc_s_mins_maxs_mina, 7, 11, 4, low(mina)).
table(low_maxc_s_mins_maxs_mina, 7, 10, 4, low(mina)).
table(low_v_c_minc_maxc_s_mina, 7, 16, 5, low(mina)).
table(low_v_c_minc_maxc_mins_mina, 7, 19, 5, low(mina)).
table(low_v_c_minc_maxc_maxs_mina, 7, 18, 5, low(mina)).
table(low_v_c_minc_s_mins_mina, 7, 8, 5, low(mina)).
table(low_v_c_minc_s_maxs_mina, 7, 14, 5, low(mina)).
table(low_v_c_minc_mins_maxs_mina, 7, 14, 5, low(mina)).
table(low_v_c_maxc_s_mins_mina, 7, 16, 5, low(mina)).
table(low_v_c_maxc_s_maxs_mina, 7, 8, 5, low(mina)).
table(low_v_c_maxc_mins_maxs_mina, 7, 18, 5, low(mina)).
table(low_v_c_s_mins_maxs_mina, 7, 12, 5, low(mina)).
table(low_v_minc_maxc_s_mins_mina, 7, 10, 5, low(mina)).
table(low_v_minc_maxc_s_maxs_mina, 7, 12, 5, low(mina)).
table(low_v_minc_maxc_mins_maxs_mina, 7, 12, 5, low(mina)).
table(low_v_minc_s_mins_maxs_mina, 7, 11, 5, low(mina)).
table(low_v_maxc_s_mins_maxs_mina, 7, 10, 5, low(mina)).
table(low_c_minc_maxc_s_mins_mina, 7, 10, 5, low(mina)).
table(low_c_minc_maxc_s_maxs_mina, 7, 18, 5, low(mina)).
table(low_c_minc_maxc_mins_maxs_mina, 7, 8, 5, low(mina)).
table(low_c_minc_s_mins_maxs_mina, 7, 14, 5, low(mina)).
table(low_c_maxc_s_mins_maxs_mina, 7, 18, 5, low(mina)).
table(low_minc_maxc_s_mins_maxs_mina, 7, 12, 5, low(mina)).
table(low_v_c_minc_maxc_s_mins_mina, 7, 16, 6, low(mina)).
table(low_v_c_minc_maxc_s_maxs_mina, 7, 18, 6, low(mina)).
table(low_v_c_minc_maxc_mins_maxs_mina, 7, 18, 6, low(mina)).
table(low_v_c_minc_s_mins_maxs_mina, 7, 14, 6, low(mina)).
table(low_v_c_maxc_s_mins_maxs_mina, 7, 18, 6, low(mina)).
table(low_v_minc_maxc_s_mins_maxs_mina, 7, 12, 6, low(mina)).
table(low_c_minc_maxc_s_mins_maxs_mina, 7, 18, 6, low(mina)).
table(low_v_c_minc_maxc_s_mins_maxs_mina, 7, 18, 7, low(mina)).
table(up_v_maxa, 7, 19, 1, up(maxa)).
table(up_v_c_maxa, 7, 19, 2, up(maxa)).
table(up_v_minc_maxa, 7, 19, 2, up(maxa)).
table(up_v_maxc_maxa, 7, 19, 2, up(maxa)).
table(up_v_s_maxa, 7, 19, 2, up(maxa)).
table(up_v_mins_maxa, 7, 19, 2, up(maxa)).
table(up_v_maxs_maxa, 7, 19, 2, up(maxa)).
table(up_c_maxc_maxa, 7, 19, 2, up(maxa)).
table(up_maxc_s_maxa, 7, 18, 2, up(maxa)).
table(up_s_maxs_maxa, 7, 19, 2, up(maxa)).
table(up_v_c_minc_maxa, 7, 19, 3, up(maxa)).
table(up_v_c_maxc_maxa, 7, 19, 3, up(maxa)).
table(up_v_c_s_maxa, 7, 19, 3, up(maxa)).
table(up_v_c_mins_maxa, 7, 19, 3, up(maxa)).
table(up_v_c_maxs_maxa, 7, 7, 3, up(maxa)).
table(up_v_minc_maxc_maxa, 7, 19, 3, up(maxa)).
table(up_v_minc_s_maxa, 7, 18, 3, up(maxa)).
table(up_v_minc_mins_maxa, 7, 19, 3, up(maxa)).
table(up_v_minc_maxs_maxa, 7, 19, 3, up(maxa)).
table(up_v_maxc_s_maxa, 7, 18, 3, up(maxa)).
table(up_v_maxc_mins_maxa, 7, 19, 3, up(maxa)).
table(up_v_maxc_maxs_maxa, 7, 9, 3, up(maxa)).
table(up_v_s_mins_maxa, 7, 19, 3, up(maxa)).
table(up_v_s_maxs_maxa, 7, 19, 3, up(maxa)).
table(up_v_mins_maxs_maxa, 7, 19, 3, up(maxa)).
table(up_c_minc_maxc_maxa, 7, 19, 3, up(maxa)).
table(up_c_maxc_s_maxa, 7, 18, 3, up(maxa)).
table(up_c_maxc_mins_maxa, 7, 19, 3, up(maxa)).
table(up_c_maxc_maxs_maxa, 7, 9, 3, up(maxa)).
table(up_c_s_maxs_maxa, 7, 19, 3, up(maxa)).
table(up_minc_maxc_s_maxa, 7, 18, 3, up(maxa)).
table(up_minc_s_maxs_maxa, 7, 18, 3, up(maxa)).
table(up_maxc_s_mins_maxa, 7, 18, 3, up(maxa)).
table(up_maxc_s_maxs_maxa, 7, 18, 3, up(maxa)).
table(up_s_mins_maxs_maxa, 7, 19, 3, up(maxa)).
table(up_v_c_minc_maxc_maxa, 7, 19, 4, up(maxa)).
table(up_v_c_minc_s_maxa, 7, 18, 4, up(maxa)).
table(up_v_c_minc_mins_maxa, 7, 19, 4, up(maxa)).
table(up_v_c_minc_maxs_maxa, 7, 11, 4, up(maxa)).
table(up_v_c_maxc_s_maxa, 7, 18, 4, up(maxa)).
table(up_v_c_maxc_mins_maxa, 7, 19, 4, up(maxa)).
table(up_v_c_maxc_maxs_maxa, 7, 9, 4, up(maxa)).
table(up_v_c_s_mins_maxa, 7, 19, 4, up(maxa)).
table(up_v_c_s_maxs_maxa, 7, 19, 4, up(maxa)).
table(up_v_c_mins_maxs_maxa, 7, 11, 4, up(maxa)).
table(up_v_minc_maxc_s_maxa, 7, 18, 4, up(maxa)).
table(up_v_minc_maxc_mins_maxa, 7, 19, 4, up(maxa)).
table(up_v_minc_maxc_maxs_maxa, 7, 19, 4, up(maxa)).
table(up_v_minc_s_mins_maxa, 7, 18, 4, up(maxa)).
table(up_v_minc_s_maxs_maxa, 7, 18, 4, up(maxa)).
table(up_v_minc_mins_maxs_maxa, 7, 18, 4, up(maxa)).
table(up_v_maxc_s_mins_maxa, 7, 18, 4, up(maxa)).
table(up_v_maxc_s_maxs_maxa, 7, 18, 4, up(maxa)).
table(up_v_maxc_mins_maxs_maxa, 7, 8, 4, up(maxa)).
table(up_v_s_mins_maxs_maxa, 7, 19, 4, up(maxa)).
table(up_c_minc_maxc_s_maxa, 7, 18, 4, up(maxa)).
table(up_c_minc_maxc_mins_maxa, 7, 19, 4, up(maxa)).
table(up_c_minc_maxc_maxs_maxa, 7, 19, 4, up(maxa)).
table(up_c_minc_s_maxs_maxa, 7, 18, 4, up(maxa)).
table(up_c_maxc_s_mins_maxa, 7, 18, 4, up(maxa)).
table(up_c_maxc_s_maxs_maxa, 7, 18, 4, up(maxa)).
table(up_c_maxc_mins_maxs_maxa, 7, 8, 4, up(maxa)).
table(up_c_s_mins_maxs_maxa, 7, 19, 4, up(maxa)).
table(up_minc_maxc_s_mins_maxa, 7, 18, 4, up(maxa)).
table(up_minc_maxc_s_maxs_maxa, 7, 18, 4, up(maxa)).
table(up_minc_s_mins_maxs_maxa, 7, 18, 4, up(maxa)).
table(up_maxc_s_mins_maxs_maxa, 7, 18, 4, up(maxa)).
table(up_v_c_minc_maxc_s_maxa, 7, 18, 5, up(maxa)).
table(up_v_c_minc_maxc_mins_maxa, 7, 19, 5, up(maxa)).
table(up_v_c_minc_maxc_maxs_maxa, 7, 19, 5, up(maxa)).
table(up_v_c_minc_s_mins_maxa, 7, 18, 5, up(maxa)).
table(up_v_c_minc_s_maxs_maxa, 7, 18, 5, up(maxa)).
table(up_v_c_minc_mins_maxs_maxa, 7, 10, 5, up(maxa)).
table(up_v_c_maxc_s_mins_maxa, 7, 18, 5, up(maxa)).
table(up_v_c_maxc_s_maxs_maxa, 7, 18, 5, up(maxa)).
table(up_v_c_maxc_mins_maxs_maxa, 7, 8, 5, up(maxa)).
table(up_v_c_s_mins_maxs_maxa, 7, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_mins_maxa, 7, 18, 5, up(maxa)).
table(up_v_minc_maxc_s_maxs_maxa, 7, 18, 5, up(maxa)).
table(up_v_minc_maxc_mins_maxs_maxa, 7, 18, 5, up(maxa)).
table(up_v_minc_s_mins_maxs_maxa, 7, 18, 5, up(maxa)).
table(up_v_maxc_s_mins_maxs_maxa, 7, 18, 5, up(maxa)).
table(up_c_minc_maxc_s_mins_maxa, 7, 18, 5, up(maxa)).
table(up_c_minc_maxc_s_maxs_maxa, 7, 18, 5, up(maxa)).
table(up_c_minc_maxc_mins_maxs_maxa, 7, 18, 5, up(maxa)).
table(up_c_minc_s_mins_maxs_maxa, 7, 18, 5, up(maxa)).
table(up_c_maxc_s_mins_maxs_maxa, 7, 18, 5, up(maxa)).
table(up_minc_maxc_s_mins_maxs_maxa, 7, 18, 5, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxa, 7, 18, 6, up(maxa)).
table(up_v_c_minc_maxc_s_maxs_maxa, 7, 18, 6, up(maxa)).
table(up_v_c_minc_maxc_mins_maxs_maxa, 7, 18, 6, up(maxa)).
table(up_v_c_minc_s_mins_maxs_maxa, 7, 18, 6, up(maxa)).
table(up_v_c_maxc_s_mins_maxs_maxa, 7, 18, 6, up(maxa)).
table(up_v_minc_maxc_s_mins_maxs_maxa, 7, 18, 6, up(maxa)).
table(up_c_minc_maxc_s_mins_maxs_maxa, 7, 18, 6, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxs_maxa, 7, 18, 7, up(maxa)).
table(low_c_v, 8, 18, 1, low(v)).
table(low_minc_v, 8, 11, 1, low(v)).
table(low_maxc_v, 8, 11, 1, low(v)).
table(low_s_v, 8, 7, 1, low(v)).
table(low_mins_v, 8, 18, 1, low(v)).
table(low_maxs_v, 8, 18, 1, low(v)).
table(low_c_minc_v, 8, 11, 2, low(v)).
table(low_c_maxc_v, 8, 11, 2, low(v)).
table(low_c_s_v, 8, 8, 2, low(v)).
table(low_c_mins_v, 8, 18, 2, low(v)).
table(low_c_maxs_v, 8, 18, 2, low(v)).
table(low_minc_maxc_v, 8, 11, 2, low(v)).
table(low_minc_s_v, 8, 3, 2, low(v)).
table(low_minc_mins_v, 8, 12, 2, low(v)).
table(low_minc_maxs_v, 8, 12, 2, low(v)).
table(low_maxc_s_v, 8, 3, 2, low(v)).
table(low_maxc_mins_v, 8, 12, 2, low(v)).
table(low_maxc_maxs_v, 8, 13, 2, low(v)).
table(low_s_mins_v, 8, 7, 2, low(v)).
table(low_s_maxs_v, 8, 7, 2, low(v)).
table(low_mins_maxs_v, 8, 12, 2, low(v)).
table(low_c_minc_maxc_v, 8, 11, 3, low(v)).
table(low_c_minc_s_v, 8, 6, 3, low(v)).
table(low_c_minc_mins_v, 8, 12, 3, low(v)).
table(low_c_minc_maxs_v, 8, 12, 3, low(v)).
table(low_c_maxc_s_v, 8, 4, 3, low(v)).
table(low_c_maxc_mins_v, 8, 12, 3, low(v)).
table(low_c_maxc_maxs_v, 8, 13, 3, low(v)).
table(low_c_s_mins_v, 8, 8, 3, low(v)).
table(low_c_s_maxs_v, 8, 8, 3, low(v)).
table(low_c_mins_maxs_v, 8, 18, 3, low(v)).
table(low_minc_maxc_s_v, 8, 5, 3, low(v)).
table(low_minc_maxc_mins_v, 8, 12, 3, low(v)).
table(low_minc_maxc_maxs_v, 8, 12, 3, low(v)).
table(low_minc_s_mins_v, 8, 4, 3, low(v)).
table(low_minc_s_maxs_v, 8, 4, 3, low(v)).
table(low_minc_mins_maxs_v, 8, 13, 3, low(v)).
table(low_maxc_s_mins_v, 8, 4, 3, low(v)).
table(low_maxc_s_maxs_v, 8, 4, 3, low(v)).
table(low_maxc_mins_maxs_v, 8, 14, 3, low(v)).
table(low_s_mins_maxs_v, 8, 7, 3, low(v)).
table(low_c_minc_maxc_s_v, 8, 7, 4, low(v)).
table(low_c_minc_maxc_mins_v, 8, 12, 4, low(v)).
table(low_c_minc_maxc_maxs_v, 8, 12, 4, low(v)).
table(low_c_minc_s_mins_v, 8, 7, 4, low(v)).
table(low_c_minc_s_maxs_v, 8, 7, 4, low(v)).
table(low_c_minc_mins_maxs_v, 8, 13, 4, low(v)).
table(low_c_maxc_s_mins_v, 8, 5, 4, low(v)).
table(low_c_maxc_s_maxs_v, 8, 5, 4, low(v)).
table(low_c_maxc_mins_maxs_v, 8, 14, 4, low(v)).
table(low_c_s_mins_maxs_v, 8, 8, 4, low(v)).
table(low_minc_maxc_s_mins_v, 8, 6, 4, low(v)).
table(low_minc_maxc_s_maxs_v, 8, 6, 4, low(v)).
table(low_minc_maxc_mins_maxs_v, 8, 13, 4, low(v)).
table(low_minc_s_mins_maxs_v, 8, 5, 4, low(v)).
table(low_maxc_s_mins_maxs_v, 8, 5, 4, low(v)).
table(low_c_minc_maxc_s_mins_v, 8, 8, 5, low(v)).
table(low_c_minc_maxc_s_maxs_v, 8, 8, 5, low(v)).
table(low_c_minc_maxc_mins_maxs_v, 8, 13, 5, low(v)).
table(low_c_minc_s_mins_maxs_v, 8, 8, 5, low(v)).
table(low_c_maxc_s_mins_maxs_v, 8, 6, 5, low(v)).
table(low_minc_maxc_s_mins_maxs_v, 8, 7, 5, low(v)).
table(low_c_minc_maxc_s_mins_maxs_v, 8, 9, 6, low(v)).
table(up_c_maxc_v, 8, 3, 2, up(v)).
table(up_maxc_s_v, 8, 3, 2, up(v)).
table(up_s_maxs_v, 8, 3, 2, up(v)).
table(up_c_minc_maxc_v, 8, 6, 3, up(v)).
table(up_c_maxc_s_v, 8, 4, 3, up(v)).
table(up_c_maxc_mins_v, 8, 4, 3, up(v)).
table(up_c_maxc_maxs_v, 8, 4, 3, up(v)).
table(up_c_s_maxs_v, 8, 4, 3, up(v)).
table(up_minc_maxc_s_v, 8, 5, 3, up(v)).
table(up_minc_s_maxs_v, 8, 4, 3, up(v)).
table(up_maxc_s_mins_v, 8, 4, 3, up(v)).
table(up_maxc_s_maxs_v, 8, 4, 3, up(v)).
table(up_s_mins_maxs_v, 8, 4, 3, up(v)).
table(up_c_minc_maxc_s_v, 8, 7, 4, up(v)).
table(up_c_minc_maxc_mins_v, 8, 7, 4, up(v)).
table(up_c_minc_maxc_maxs_v, 8, 7, 4, up(v)).
table(up_c_minc_s_maxs_v, 8, 7, 4, up(v)).
table(up_c_maxc_s_mins_v, 8, 5, 4, up(v)).
table(up_c_maxc_s_maxs_v, 8, 5, 4, up(v)).
table(up_c_maxc_mins_maxs_v, 8, 5, 4, up(v)).
table(up_c_s_mins_maxs_v, 8, 5, 4, up(v)).
table(up_minc_maxc_s_mins_v, 8, 6, 4, up(v)).
table(up_minc_maxc_s_maxs_v, 8, 6, 4, up(v)).
table(up_minc_s_mins_maxs_v, 8, 5, 4, up(v)).
table(up_maxc_s_mins_maxs_v, 8, 5, 4, up(v)).
table(up_c_minc_maxc_s_mins_v, 8, 8, 5, up(v)).
table(up_c_minc_maxc_s_maxs_v, 8, 8, 5, up(v)).
table(up_c_minc_maxc_mins_maxs_v, 8, 8, 5, up(v)).
table(up_c_minc_s_mins_maxs_v, 8, 8, 5, up(v)).
table(up_c_maxc_s_mins_maxs_v, 8, 6, 5, up(v)).
table(up_minc_maxc_s_mins_maxs_v, 8, 7, 5, up(v)).
table(up_c_minc_maxc_s_mins_maxs_v, 8, 9, 6, up(v)).
table(low_v_c, 8, 11, 1, low(c)).
table(low_minc_c, 8, 11, 1, low(c)).
table(low_maxc_c, 8, 11, 1, low(c)).
table(low_s_c, 8, 8, 1, low(c)).
table(low_mins_c, 8, 8, 1, low(c)).
table(low_maxs_c, 8, 8, 1, low(c)).
table(low_v_minc_c, 8, 11, 2, low(c)).
table(low_v_maxc_c, 8, 5, 2, low(c)).
table(low_v_s_c, 8, 13, 2, low(c)).
table(low_v_mins_c, 8, 12, 2, low(c)).
table(low_v_maxs_c, 8, 13, 2, low(c)).
table(low_minc_maxc_c, 8, 11, 2, low(c)).
table(low_minc_s_c, 8, 5, 2, low(c)).
table(low_minc_mins_c, 8, 9, 2, low(c)).
table(low_minc_maxs_c, 8, 9, 2, low(c)).
table(low_maxc_s_c, 8, 5, 2, low(c)).
table(low_maxc_mins_c, 8, 12, 2, low(c)).
table(low_maxc_maxs_c, 8, 13, 2, low(c)).
table(low_s_mins_c, 8, 10, 2, low(c)).
table(low_s_maxs_c, 8, 11, 2, low(c)).
table(low_mins_maxs_c, 8, 9, 2, low(c)).
table(low_v_minc_maxc_c, 8, 6, 3, low(c)).
table(low_v_minc_s_c, 8, 12, 3, low(c)).
table(low_v_minc_mins_c, 8, 12, 3, low(c)).
table(low_v_minc_maxs_c, 8, 12, 3, low(c)).
table(low_v_maxc_s_c, 8, 6, 3, low(c)).
table(low_v_maxc_mins_c, 8, 6, 3, low(c)).
table(low_v_maxc_maxs_c, 8, 6, 3, low(c)).
table(low_v_s_mins_c, 8, 14, 3, low(c)).
table(low_v_s_maxs_c, 8, 14, 3, low(c)).
table(low_v_mins_maxs_c, 8, 14, 3, low(c)).
table(low_minc_maxc_s_c, 8, 5, 3, low(c)).
table(low_minc_maxc_mins_c, 8, 12, 3, low(c)).
table(low_minc_maxc_maxs_c, 8, 12, 3, low(c)).
table(low_minc_s_mins_c, 8, 6, 3, low(c)).
table(low_minc_s_maxs_c, 8, 7, 3, low(c)).
table(low_minc_mins_maxs_c, 8, 7, 3, low(c)).
table(low_maxc_s_mins_c, 8, 6, 3, low(c)).
table(low_maxc_s_maxs_c, 8, 6, 3, low(c)).
table(low_maxc_mins_maxs_c, 8, 6, 3, low(c)).
table(low_s_mins_maxs_c, 8, 12, 3, low(c)).
table(low_v_minc_maxc_s_c, 8, 7, 4, low(c)).
table(low_v_minc_maxc_mins_c, 8, 7, 4, low(c)).
table(low_v_minc_maxc_maxs_c, 8, 7, 4, low(c)).
table(low_v_minc_s_mins_c, 8, 13, 4, low(c)).
table(low_v_minc_s_maxs_c, 8, 13, 4, low(c)).
table(low_v_minc_mins_maxs_c, 8, 13, 4, low(c)).
table(low_v_maxc_s_mins_c, 8, 7, 4, low(c)).
table(low_v_maxc_s_maxs_c, 8, 7, 4, low(c)).
table(low_v_maxc_mins_maxs_c, 8, 7, 4, low(c)).
table(low_v_s_mins_maxs_c, 8, 15, 4, low(c)).
table(low_minc_maxc_s_mins_c, 8, 6, 4, low(c)).
table(low_minc_maxc_s_maxs_c, 8, 6, 4, low(c)).
table(low_minc_maxc_mins_maxs_c, 8, 13, 4, low(c)).
table(low_minc_s_mins_maxs_c, 8, 8, 4, low(c)).
table(low_maxc_s_mins_maxs_c, 8, 7, 4, low(c)).
table(low_v_minc_maxc_s_mins_c, 8, 8, 5, low(c)).
table(low_v_minc_maxc_s_maxs_c, 8, 8, 5, low(c)).
table(low_v_minc_maxc_mins_maxs_c, 8, 8, 5, low(c)).
table(low_v_minc_s_mins_maxs_c, 8, 14, 5, low(c)).
table(low_v_maxc_s_mins_maxs_c, 8, 8, 5, low(c)).
table(low_minc_maxc_s_mins_maxs_c, 8, 7, 5, low(c)).
table(low_v_minc_maxc_s_mins_maxs_c, 8, 9, 6, low(c)).
table(up_v_c, 8, 18, 1, up(c)).
table(up_s_c, 8, 3, 1, up(c)).
table(up_v_minc_c, 8, 11, 2, up(c)).
table(up_v_maxc_c, 8, 11, 2, up(c)).
table(up_v_s_c, 8, 4, 2, up(c)).
table(up_v_mins_c, 8, 18, 2, up(c)).
table(up_v_maxs_c, 8, 18, 2, up(c)).
table(up_minc_s_c, 8, 3, 2, up(c)).
table(up_maxc_s_c, 8, 4, 2, up(c)).
table(up_s_mins_c, 8, 5, 2, up(c)).
table(up_s_maxs_c, 8, 5, 2, up(c)).
table(up_v_minc_maxc_c, 8, 11, 3, up(c)).
table(up_v_minc_s_c, 8, 6, 3, up(c)).
table(up_v_minc_mins_c, 8, 12, 3, up(c)).
table(up_v_minc_maxs_c, 8, 9, 3, up(c)).
table(up_v_maxc_s_c, 8, 5, 3, up(c)).
table(up_v_maxc_mins_c, 8, 12, 3, up(c)).
table(up_v_maxc_maxs_c, 8, 13, 3, up(c)).
table(up_v_s_mins_c, 8, 8, 3, up(c)).
table(up_v_s_maxs_c, 8, 6, 3, up(c)).
table(up_v_mins_maxs_c, 8, 18, 3, up(c)).
table(up_minc_maxc_s_c, 8, 5, 3, up(c)).
table(up_minc_s_mins_c, 8, 4, 3, up(c)).
table(up_minc_s_maxs_c, 8, 4, 3, up(c)).
table(up_maxc_s_mins_c, 8, 5, 3, up(c)).
table(up_maxc_s_maxs_c, 8, 5, 3, up(c)).
table(up_s_mins_maxs_c, 8, 8, 3, up(c)).
table(up_v_minc_maxc_s_c, 8, 7, 4, up(c)).
table(up_v_minc_maxc_mins_c, 8, 12, 4, up(c)).
table(up_v_minc_maxc_maxs_c, 8, 12, 4, up(c)).
table(up_v_minc_s_mins_c, 8, 7, 4, up(c)).
table(up_v_minc_s_maxs_c, 8, 7, 4, up(c)).
table(up_v_minc_mins_maxs_c, 8, 10, 4, up(c)).
table(up_v_maxc_s_mins_c, 8, 9, 4, up(c)).
table(up_v_maxc_s_maxs_c, 8, 6, 4, up(c)).
table(up_v_maxc_mins_maxs_c, 8, 14, 4, up(c)).
table(up_v_s_mins_maxs_c, 8, 10, 4, up(c)).
table(up_minc_maxc_s_mins_c, 8, 6, 4, up(c)).
table(up_minc_maxc_s_maxs_c, 8, 6, 4, up(c)).
table(up_minc_s_mins_maxs_c, 8, 5, 4, up(c)).
table(up_maxc_s_mins_maxs_c, 8, 6, 4, up(c)).
table(up_v_minc_maxc_s_mins_c, 8, 8, 5, up(c)).
table(up_v_minc_maxc_s_maxs_c, 8, 8, 5, up(c)).
table(up_v_minc_maxc_mins_maxs_c, 8, 13, 5, up(c)).
table(up_v_minc_s_mins_maxs_c, 8, 8, 5, up(c)).
table(up_v_maxc_s_mins_maxs_c, 8, 7, 5, up(c)).
table(up_minc_maxc_s_mins_maxs_c, 8, 7, 5, up(c)).
table(up_v_minc_maxc_s_mins_maxs_c, 8, 9, 6, up(c)).
table(low_v_minc, 8, 3, 1, low(minc)).
table(low_c_minc, 8, 3, 1, low(minc)).
table(low_maxc_minc, 8, 3, 1, low(minc)).
table(low_s_minc, 8, 3, 1, low(minc)).
table(low_mins_minc, 8, 2, 1, low(minc)).
table(low_maxs_minc, 8, 3, 1, low(minc)).
table(low_v_c_minc, 8, 5, 2, low(minc)).
table(low_v_maxc_minc, 8, 4, 2, low(minc)).
table(low_v_s_minc, 8, 4, 2, low(minc)).
table(low_v_mins_minc, 8, 3, 2, low(minc)).
table(low_v_maxs_minc, 8, 4, 2, low(minc)).
table(low_c_maxc_minc, 8, 4, 2, low(minc)).
table(low_c_s_minc, 8, 4, 2, low(minc)).
table(low_c_mins_minc, 8, 3, 2, low(minc)).
table(low_c_maxs_minc, 8, 4, 2, low(minc)).
table(low_maxc_s_minc, 8, 4, 2, low(minc)).
table(low_maxc_mins_minc, 8, 4, 2, low(minc)).
table(low_maxc_maxs_minc, 8, 4, 2, low(minc)).
table(low_s_mins_minc, 8, 3, 2, low(minc)).
table(low_s_maxs_minc, 8, 4, 2, low(minc)).
table(low_mins_maxs_minc, 8, 3, 2, low(minc)).
table(low_v_c_maxc_minc, 8, 6, 3, low(minc)).
table(low_v_c_s_minc, 8, 6, 3, low(minc)).
table(low_v_c_mins_minc, 8, 6, 3, low(minc)).
table(low_v_c_maxs_minc, 8, 6, 3, low(minc)).
table(low_v_maxc_s_minc, 8, 5, 3, low(minc)).
table(low_v_maxc_mins_minc, 8, 5, 3, low(minc)).
table(low_v_maxc_maxs_minc, 8, 5, 3, low(minc)).
table(low_v_s_mins_minc, 8, 4, 3, low(minc)).
table(low_v_s_maxs_minc, 8, 5, 3, low(minc)).
table(low_v_mins_maxs_minc, 8, 4, 3, low(minc)).
table(low_c_maxc_s_minc, 8, 5, 3, low(minc)).
table(low_c_maxc_mins_minc, 8, 5, 3, low(minc)).
table(low_c_maxc_maxs_minc, 8, 5, 3, low(minc)).
table(low_c_s_mins_minc, 8, 4, 3, low(minc)).
table(low_c_s_maxs_minc, 8, 6, 3, low(minc)).
table(low_c_mins_maxs_minc, 8, 4, 3, low(minc)).
table(low_maxc_s_mins_minc, 8, 5, 3, low(minc)).
table(low_maxc_s_maxs_minc, 8, 5, 3, low(minc)).
table(low_maxc_mins_maxs_minc, 8, 5, 3, low(minc)).
table(low_s_mins_maxs_minc, 8, 4, 3, low(minc)).
table(low_v_c_maxc_s_minc, 8, 7, 4, low(minc)).
table(low_v_c_maxc_mins_minc, 8, 7, 4, low(minc)).
table(low_v_c_maxc_maxs_minc, 8, 7, 4, low(minc)).
table(low_v_c_s_mins_minc, 8, 7, 4, low(minc)).
table(low_v_c_s_maxs_minc, 8, 7, 4, low(minc)).
table(low_v_c_mins_maxs_minc, 8, 7, 4, low(minc)).
table(low_v_maxc_s_mins_minc, 8, 6, 4, low(minc)).
table(low_v_maxc_s_maxs_minc, 8, 6, 4, low(minc)).
table(low_v_maxc_mins_maxs_minc, 8, 6, 4, low(minc)).
table(low_v_s_mins_maxs_minc, 8, 5, 4, low(minc)).
table(low_c_maxc_s_mins_minc, 8, 6, 4, low(minc)).
table(low_c_maxc_s_maxs_minc, 8, 6, 4, low(minc)).
table(low_c_maxc_mins_maxs_minc, 8, 6, 4, low(minc)).
table(low_c_s_mins_maxs_minc, 8, 6, 4, low(minc)).
table(low_maxc_s_mins_maxs_minc, 8, 6, 4, low(minc)).
table(low_v_c_maxc_s_mins_minc, 8, 8, 5, low(minc)).
table(low_v_c_maxc_s_maxs_minc, 8, 8, 5, low(minc)).
table(low_v_c_maxc_mins_maxs_minc, 8, 8, 5, low(minc)).
table(low_v_c_s_mins_maxs_minc, 8, 8, 5, low(minc)).
table(low_v_maxc_s_mins_maxs_minc, 8, 7, 5, low(minc)).
table(low_c_maxc_s_mins_maxs_minc, 8, 7, 5, low(minc)).
table(low_v_c_maxc_s_mins_maxs_minc, 8, 9, 6, low(minc)).
table(up_v_minc, 8, 11, 1, up(minc)).
table(up_maxc_minc, 8, 6, 1, up(minc)).
table(up_v_c_minc, 8, 5, 2, up(minc)).
table(up_v_maxc_minc, 8, 6, 2, up(minc)).
table(up_v_s_minc, 8, 13, 2, up(minc)).
table(up_v_mins_minc, 8, 12, 2, up(minc)).
table(up_v_maxs_minc, 8, 13, 2, up(minc)).
table(up_c_maxc_minc, 8, 4, 2, up(minc)).
table(up_maxc_s_minc, 8, 6, 2, up(minc)).
table(up_maxc_mins_minc, 8, 6, 2, up(minc)).
table(up_maxc_maxs_minc, 8, 7, 2, up(minc)).
table(up_s_maxs_minc, 8, 14, 2, up(minc)).
table(up_v_c_maxc_minc, 8, 6, 3, up(minc)).
table(up_v_c_s_minc, 8, 6, 3, up(minc)).
table(up_v_c_mins_minc, 8, 6, 3, up(minc)).
table(up_v_c_maxs_minc, 8, 6, 3, up(minc)).
table(up_v_maxc_s_minc, 8, 7, 3, up(minc)).
table(up_v_maxc_mins_minc, 8, 7, 3, up(minc)).
table(up_v_maxc_maxs_minc, 8, 7, 3, up(minc)).
table(up_v_s_mins_minc, 8, 14, 3, up(minc)).
table(up_v_s_maxs_minc, 8, 14, 3, up(minc)).
table(up_v_mins_maxs_minc, 8, 14, 3, up(minc)).
table(up_c_maxc_s_minc, 8, 5, 3, up(minc)).
table(up_c_maxc_mins_minc, 8, 5, 3, up(minc)).
table(up_c_maxc_maxs_minc, 8, 5, 3, up(minc)).
table(up_c_s_maxs_minc, 8, 4, 3, up(minc)).
table(up_maxc_s_mins_minc, 8, 7, 3, up(minc)).
table(up_maxc_s_maxs_minc, 8, 7, 3, up(minc)).
table(up_maxc_mins_maxs_minc, 8, 7, 3, up(minc)).
table(up_s_mins_maxs_minc, 8, 15, 3, up(minc)).
table(up_v_c_maxc_s_minc, 8, 7, 4, up(minc)).
table(up_v_c_maxc_mins_minc, 8, 7, 4, up(minc)).
table(up_v_c_maxc_maxs_minc, 8, 7, 4, up(minc)).
table(up_v_c_s_mins_minc, 8, 7, 4, up(minc)).
table(up_v_c_s_maxs_minc, 8, 7, 4, up(minc)).
table(up_v_c_mins_maxs_minc, 8, 7, 4, up(minc)).
table(up_v_maxc_s_mins_minc, 8, 8, 4, up(minc)).
table(up_v_maxc_s_maxs_minc, 8, 8, 4, up(minc)).
table(up_v_maxc_mins_maxs_minc, 8, 8, 4, up(minc)).
table(up_v_s_mins_maxs_minc, 8, 15, 4, up(minc)).
table(up_c_maxc_s_mins_minc, 8, 6, 4, up(minc)).
table(up_c_maxc_s_maxs_minc, 8, 6, 4, up(minc)).
table(up_c_maxc_mins_maxs_minc, 8, 6, 4, up(minc)).
table(up_c_s_mins_maxs_minc, 8, 5, 4, up(minc)).
table(up_maxc_s_mins_maxs_minc, 8, 8, 4, up(minc)).
table(up_v_c_maxc_s_mins_minc, 8, 8, 5, up(minc)).
table(up_v_c_maxc_s_maxs_minc, 8, 8, 5, up(minc)).
table(up_v_c_maxc_mins_maxs_minc, 8, 8, 5, up(minc)).
table(up_v_c_s_mins_maxs_minc, 8, 8, 5, up(minc)).
table(up_v_maxc_s_mins_maxs_minc, 8, 9, 5, up(minc)).
table(up_c_maxc_s_mins_maxs_minc, 8, 7, 5, up(minc)).
table(up_v_c_maxc_s_mins_maxs_minc, 8, 9, 6, up(minc)).
table(low_v_maxc, 8, 18, 1, low(maxc)).
table(low_c_maxc, 8, 18, 1, low(maxc)).
table(low_minc_maxc, 8, 6, 1, low(maxc)).
table(low_s_maxc, 8, 18, 1, low(maxc)).
table(low_mins_maxc, 8, 9, 1, low(maxc)).
table(low_maxs_maxc, 8, 2, 1, low(maxc)).
table(low_v_c_maxc, 8, 5, 2, low(maxc)).
table(low_v_minc_maxc, 8, 11, 2, low(maxc)).
table(low_v_s_maxc, 8, 6, 2, low(maxc)).
table(low_v_mins_maxc, 8, 18, 2, low(maxc)).
table(low_v_maxs_maxc, 8, 3, 2, low(maxc)).
table(low_c_minc_maxc, 8, 11, 2, low(maxc)).
table(low_c_s_maxc, 8, 6, 2, low(maxc)).
table(low_c_mins_maxc, 8, 18, 2, low(maxc)).
table(low_c_maxs_maxc, 8, 3, 2, low(maxc)).
table(low_minc_s_maxc, 8, 7, 2, low(maxc)).
table(low_minc_mins_maxc, 8, 7, 2, low(maxc)).
table(low_minc_maxs_maxc, 8, 4, 2, low(maxc)).
table(low_s_mins_maxc, 8, 18, 2, low(maxc)).
table(low_s_maxs_maxc, 8, 3, 2, low(maxc)).
table(low_mins_maxs_maxc, 8, 3, 2, low(maxc)).
table(low_v_c_minc_maxc, 8, 6, 3, low(maxc)).
table(low_v_c_s_maxc, 8, 6, 3, low(maxc)).
table(low_v_c_mins_maxc, 8, 6, 3, low(maxc)).
table(low_v_c_maxs_maxc, 8, 4, 3, low(maxc)).
table(low_v_minc_s_maxc, 8, 5, 3, low(maxc)).
table(low_v_minc_mins_maxc, 8, 12, 3, low(maxc)).
table(low_v_minc_maxs_maxc, 8, 5, 3, low(maxc)).
table(low_v_s_mins_maxc, 8, 5, 3, low(maxc)).
table(low_v_s_maxs_maxc, 8, 4, 3, low(maxc)).
table(low_v_mins_maxs_maxc, 8, 4, 3, low(maxc)).
table(low_c_minc_s_maxc, 8, 5, 3, low(maxc)).
table(low_c_minc_mins_maxc, 8, 12, 3, low(maxc)).
table(low_c_minc_maxs_maxc, 8, 5, 3, low(maxc)).
table(low_c_s_mins_maxc, 8, 6, 3, low(maxc)).
table(low_c_s_maxs_maxc, 8, 4, 3, low(maxc)).
table(low_c_mins_maxs_maxc, 8, 4, 3, low(maxc)).
table(low_minc_s_mins_maxc, 8, 8, 3, low(maxc)).
table(low_minc_s_maxs_maxc, 8, 5, 3, low(maxc)).
table(low_minc_mins_maxs_maxc, 8, 5, 3, low(maxc)).
table(low_s_mins_maxs_maxc, 8, 4, 3, low(maxc)).
table(low_v_c_minc_s_maxc, 8, 7, 4, low(maxc)).
table(low_v_c_minc_mins_maxc, 8, 7, 4, low(maxc)).
table(low_v_c_minc_maxs_maxc, 8, 7, 4, low(maxc)).
table(low_v_c_s_mins_maxc, 8, 5, 4, low(maxc)).
table(low_v_c_s_maxs_maxc, 8, 5, 4, low(maxc)).
table(low_v_c_mins_maxs_maxc, 8, 5, 4, low(maxc)).
table(low_v_minc_s_mins_maxc, 8, 6, 4, low(maxc)).
table(low_v_minc_s_maxs_maxc, 8, 6, 4, low(maxc)).
table(low_v_minc_mins_maxs_maxc, 8, 6, 4, low(maxc)).
table(low_v_s_mins_maxs_maxc, 8, 5, 4, low(maxc)).
table(low_c_minc_s_mins_maxc, 8, 6, 4, low(maxc)).
table(low_c_minc_s_maxs_maxc, 8, 6, 4, low(maxc)).
table(low_c_minc_mins_maxs_maxc, 8, 6, 4, low(maxc)).
table(low_c_s_mins_maxs_maxc, 8, 5, 4, low(maxc)).
table(low_minc_s_mins_maxs_maxc, 8, 6, 4, low(maxc)).
table(low_v_c_minc_s_mins_maxc, 8, 8, 5, low(maxc)).
table(low_v_c_minc_s_maxs_maxc, 8, 8, 5, low(maxc)).
table(low_v_c_minc_mins_maxs_maxc, 8, 8, 5, low(maxc)).
table(low_v_c_s_mins_maxs_maxc, 8, 6, 5, low(maxc)).
table(low_v_minc_s_mins_maxs_maxc, 8, 7, 5, low(maxc)).
table(low_c_minc_s_mins_maxs_maxc, 8, 7, 5, low(maxc)).
table(low_v_c_minc_s_mins_maxs_maxc, 8, 9, 6, low(maxc)).
table(up_v_maxc, 8, 11, 1, up(maxc)).
table(up_v_c_maxc, 8, 11, 2, up(maxc)).
table(up_v_minc_maxc, 8, 11, 2, up(maxc)).
table(up_v_s_maxc, 8, 13, 2, up(maxc)).
table(up_v_mins_maxc, 8, 12, 2, up(maxc)).
table(up_v_maxs_maxc, 8, 13, 2, up(maxc)).
table(up_v_c_minc_maxc, 8, 11, 3, up(maxc)).
table(up_v_c_s_maxc, 8, 13, 3, up(maxc)).
table(up_v_c_mins_maxc, 8, 12, 3, up(maxc)).
table(up_v_c_maxs_maxc, 8, 13, 3, up(maxc)).
table(up_v_minc_s_maxc, 8, 12, 3, up(maxc)).
table(up_v_minc_mins_maxc, 8, 12, 3, up(maxc)).
table(up_v_minc_maxs_maxc, 8, 12, 3, up(maxc)).
table(up_v_s_mins_maxc, 8, 14, 3, up(maxc)).
table(up_v_s_maxs_maxc, 8, 14, 3, up(maxc)).
table(up_v_mins_maxs_maxc, 8, 14, 3, up(maxc)).
table(up_v_c_minc_s_maxc, 8, 12, 4, up(maxc)).
table(up_v_c_minc_mins_maxc, 8, 12, 4, up(maxc)).
table(up_v_c_minc_maxs_maxc, 8, 12, 4, up(maxc)).
table(up_v_c_s_mins_maxc, 8, 14, 4, up(maxc)).
table(up_v_c_s_maxs_maxc, 8, 6, 4, up(maxc)).
table(up_v_c_mins_maxs_maxc, 8, 14, 4, up(maxc)).
table(up_v_minc_s_mins_maxc, 8, 13, 4, up(maxc)).
table(up_v_minc_s_maxs_maxc, 8, 13, 4, up(maxc)).
table(up_v_minc_mins_maxs_maxc, 8, 13, 4, up(maxc)).
table(up_v_s_mins_maxs_maxc, 8, 15, 4, up(maxc)).
table(up_v_c_minc_s_mins_maxc, 8, 13, 5, up(maxc)).
table(up_v_c_minc_s_maxs_maxc, 8, 8, 5, up(maxc)).
table(up_v_c_minc_mins_maxs_maxc, 8, 13, 5, up(maxc)).
table(up_v_c_s_mins_maxs_maxc, 8, 10, 5, up(maxc)).
table(up_v_minc_s_mins_maxs_maxc, 8, 14, 5, up(maxc)).
table(up_v_c_minc_s_mins_maxs_maxc, 8, 9, 6, up(maxc)).
table(low_v_s, 8, 18, 1, low(s)).
table(low_c_s, 8, 3, 1, low(s)).
table(low_minc_s, 8, 18, 1, low(s)).
table(low_maxc_s, 8, 18, 1, low(s)).
table(low_mins_s, 8, 18, 1, low(s)).
table(low_maxs_s, 8, 18, 1, low(s)).
table(low_v_c_s, 8, 4, 2, low(s)).
table(low_v_minc_s, 8, 18, 2, low(s)).
table(low_v_maxc_s, 8, 8, 2, low(s)).
table(low_v_mins_s, 8, 9, 2, low(s)).
table(low_v_maxs_s, 8, 3, 2, low(s)).
table(low_c_minc_s, 8, 5, 2, low(s)).
table(low_c_maxc_s, 8, 5, 2, low(s)).
table(low_c_mins_s, 8, 5, 2, low(s)).
table(low_c_maxs_s, 8, 5, 2, low(s)).
table(low_minc_maxc_s, 8, 18, 2, low(s)).
table(low_minc_mins_s, 8, 13, 2, low(s)).
table(low_minc_maxs_s, 8, 14, 2, low(s)).
table(low_maxc_mins_s, 8, 7, 2, low(s)).
table(low_maxc_maxs_s, 8, 14, 2, low(s)).
table(low_mins_maxs_s, 8, 12, 2, low(s)).
table(low_v_c_minc_s, 8, 8, 3, low(s)).
table(low_v_c_maxc_s, 8, 6, 3, low(s)).
table(low_v_c_mins_s, 8, 8, 3, low(s)).
table(low_v_c_maxs_s, 8, 4, 3, low(s)).
table(low_v_minc_maxc_s, 8, 10, 3, low(s)).
table(low_v_minc_mins_s, 8, 13, 3, low(s)).
table(low_v_minc_maxs_s, 8, 4, 3, low(s)).
table(low_v_maxc_mins_s, 8, 4, 3, low(s)).
table(low_v_maxc_maxs_s, 8, 4, 3, low(s)).
table(low_v_mins_maxs_s, 8, 4, 3, low(s)).
table(low_c_minc_maxc_s, 8, 8, 3, low(s)).
table(low_c_minc_mins_s, 8, 4, 3, low(s)).
table(low_c_minc_maxs_s, 8, 5, 3, low(s)).
table(low_c_maxc_mins_s, 8, 8, 3, low(s)).
table(low_c_maxc_maxs_s, 8, 5, 3, low(s)).
table(low_c_mins_maxs_s, 8, 8, 3, low(s)).
table(low_minc_maxc_mins_s, 8, 13, 3, low(s)).
table(low_minc_maxc_maxs_s, 8, 14, 3, low(s)).
table(low_minc_mins_maxs_s, 8, 8, 3, low(s)).
table(low_maxc_mins_maxs_s, 8, 10, 3, low(s)).
table(low_v_c_minc_maxc_s, 8, 10, 4, low(s)).
table(low_v_c_minc_mins_s, 8, 7, 4, low(s)).
table(low_v_c_minc_maxs_s, 8, 7, 4, low(s)).
table(low_v_c_maxc_mins_s, 8, 8, 4, low(s)).
table(low_v_c_maxc_maxs_s, 8, 5, 4, low(s)).
table(low_v_c_mins_maxs_s, 8, 5, 4, low(s)).
table(low_v_minc_maxc_mins_s, 8, 8, 4, low(s)).
table(low_v_minc_maxc_maxs_s, 8, 6, 4, low(s)).
table(low_v_minc_mins_maxs_s, 8, 5, 4, low(s)).
table(low_v_maxc_mins_maxs_s, 8, 5, 4, low(s)).
table(low_c_minc_maxc_mins_s, 8, 6, 4, low(s)).
table(low_c_minc_maxc_maxs_s, 8, 7, 4, low(s)).
table(low_c_minc_mins_maxs_s, 8, 5, 4, low(s)).
table(low_c_maxc_mins_maxs_s, 8, 6, 4, low(s)).
table(low_minc_maxc_mins_maxs_s, 8, 14, 4, low(s)).
table(low_v_c_minc_maxc_mins_s, 8, 8, 5, low(s)).
table(low_v_c_minc_maxc_maxs_s, 8, 8, 5, low(s)).
table(low_v_c_minc_mins_maxs_s, 8, 8, 5, low(s)).
table(low_v_c_maxc_mins_maxs_s, 8, 6, 5, low(s)).
table(low_v_minc_maxc_mins_maxs_s, 8, 7, 5, low(s)).
table(low_c_minc_maxc_mins_maxs_s, 8, 7, 5, low(s)).
table(low_v_c_minc_maxc_mins_maxs_s, 8, 9, 6, low(s)).
table(up_v_s, 8, 7, 1, up(s)).
table(up_v_c_s, 8, 8, 2, up(s)).
table(up_v_minc_s, 8, 8, 2, up(s)).
table(up_v_maxc_s, 8, 8, 2, up(s)).
table(up_v_mins_s, 8, 7, 2, up(s)).
table(up_v_maxs_s, 8, 7, 2, up(s)).
table(up_c_maxc_s, 8, 9, 2, up(s)).
table(up_v_c_minc_s, 8, 11, 3, up(s)).
table(up_v_c_maxc_s, 8, 9, 3, up(s)).
table(up_v_c_mins_s, 8, 8, 3, up(s)).
table(up_v_c_maxs_s, 8, 8, 3, up(s)).
table(up_v_minc_maxc_s, 8, 10, 3, up(s)).
table(up_v_minc_mins_s, 8, 8, 3, up(s)).
table(up_v_minc_maxs_s, 8, 8, 3, up(s)).
table(up_v_maxc_mins_s, 8, 8, 3, up(s)).
table(up_v_maxc_maxs_s, 8, 8, 3, up(s)).
table(up_v_mins_maxs_s, 8, 7, 3, up(s)).
table(up_c_minc_maxc_s, 8, 12, 3, up(s)).
table(up_c_maxc_mins_s, 8, 6, 3, up(s)).
table(up_c_maxc_maxs_s, 8, 9, 3, up(s)).
table(up_v_c_minc_maxc_s, 8, 12, 4, up(s)).
table(up_v_c_minc_mins_s, 8, 11, 4, up(s)).
table(up_v_c_minc_maxs_s, 8, 11, 4, up(s)).
table(up_v_c_maxc_mins_s, 8, 9, 4, up(s)).
table(up_v_c_maxc_maxs_s, 8, 9, 4, up(s)).
table(up_v_c_mins_maxs_s, 8, 8, 4, up(s)).
table(up_v_minc_maxc_mins_s, 8, 10, 4, up(s)).
table(up_v_minc_maxc_maxs_s, 8, 10, 4, up(s)).
table(up_v_minc_mins_maxs_s, 8, 8, 4, up(s)).
table(up_v_maxc_mins_maxs_s, 8, 8, 4, up(s)).
table(up_c_minc_maxc_mins_s, 8, 8, 4, up(s)).
table(up_c_minc_maxc_maxs_s, 8, 12, 4, up(s)).
table(up_c_maxc_mins_maxs_s, 8, 6, 4, up(s)).
table(up_v_c_minc_maxc_mins_s, 8, 12, 5, up(s)).
table(up_v_c_minc_maxc_maxs_s, 8, 12, 5, up(s)).
table(up_v_c_minc_mins_maxs_s, 8, 11, 5, up(s)).
table(up_v_c_maxc_mins_maxs_s, 8, 9, 5, up(s)).
table(up_v_minc_maxc_mins_maxs_s, 8, 10, 5, up(s)).
table(up_c_minc_maxc_mins_maxs_s, 8, 8, 5, up(s)).
table(up_v_c_minc_maxc_mins_maxs_s, 8, 12, 6, up(s)).
table(low_v_mins, 8, 2, 1, low(mins)).
table(low_c_mins, 8, 2, 1, low(mins)).
table(low_minc_mins, 8, 2, 1, low(mins)).
table(low_maxc_mins, 8, 2, 1, low(mins)).
table(low_s_mins, 8, 2, 1, low(mins)).
table(low_maxs_mins, 8, 2, 1, low(mins)).
table(low_v_c_mins, 8, 3, 2, low(mins)).
table(low_v_minc_mins, 8, 3, 2, low(mins)).
table(low_v_maxc_mins, 8, 3, 2, low(mins)).
table(low_v_s_mins, 8, 3, 2, low(mins)).
table(low_v_maxs_mins, 8, 3, 2, low(mins)).
table(low_c_minc_mins, 8, 3, 2, low(mins)).
table(low_c_maxc_mins, 8, 3, 2, low(mins)).
table(low_c_s_mins, 8, 3, 2, low(mins)).
table(low_c_maxs_mins, 8, 3, 2, low(mins)).
table(low_minc_maxc_mins, 8, 4, 2, low(mins)).
table(low_minc_s_mins, 8, 3, 2, low(mins)).
table(low_minc_maxs_mins, 8, 3, 2, low(mins)).
table(low_maxc_s_mins, 8, 3, 2, low(mins)).
table(low_maxc_maxs_mins, 8, 3, 2, low(mins)).
table(low_s_maxs_mins, 8, 3, 2, low(mins)).
table(low_v_c_minc_mins, 8, 6, 3, low(mins)).
table(low_v_c_maxc_mins, 8, 4, 3, low(mins)).
table(low_v_c_s_mins, 8, 4, 3, low(mins)).
table(low_v_c_maxs_mins, 8, 4, 3, low(mins)).
table(low_v_minc_maxc_mins, 8, 5, 3, low(mins)).
table(low_v_minc_s_mins, 8, 4, 3, low(mins)).
table(low_v_minc_maxs_mins, 8, 4, 3, low(mins)).
table(low_v_maxc_s_mins, 8, 4, 3, low(mins)).
table(low_v_maxc_maxs_mins, 8, 4, 3, low(mins)).
table(low_v_s_maxs_mins, 8, 4, 3, low(mins)).
table(low_c_minc_maxc_mins, 8, 5, 3, low(mins)).
table(low_c_minc_s_mins, 8, 4, 3, low(mins)).
table(low_c_minc_maxs_mins, 8, 4, 3, low(mins)).
table(low_c_maxc_s_mins, 8, 4, 3, low(mins)).
table(low_c_maxc_maxs_mins, 8, 4, 3, low(mins)).
table(low_c_s_maxs_mins, 8, 4, 3, low(mins)).
table(low_minc_maxc_s_mins, 8, 5, 3, low(mins)).
table(low_minc_maxc_maxs_mins, 8, 5, 3, low(mins)).
table(low_minc_s_maxs_mins, 8, 4, 3, low(mins)).
table(low_maxc_s_maxs_mins, 8, 4, 3, low(mins)).
table(low_v_c_minc_maxc_mins, 8, 7, 4, low(mins)).
table(low_v_c_minc_s_mins, 8, 7, 4, low(mins)).
table(low_v_c_minc_maxs_mins, 8, 7, 4, low(mins)).
table(low_v_c_maxc_s_mins, 8, 5, 4, low(mins)).
table(low_v_c_maxc_maxs_mins, 8, 5, 4, low(mins)).
table(low_v_c_s_maxs_mins, 8, 5, 4, low(mins)).
table(low_v_minc_maxc_s_mins, 8, 6, 4, low(mins)).
table(low_v_minc_maxc_maxs_mins, 8, 6, 4, low(mins)).
table(low_v_minc_s_maxs_mins, 8, 5, 4, low(mins)).
table(low_v_maxc_s_maxs_mins, 8, 5, 4, low(mins)).
table(low_c_minc_maxc_s_mins, 8, 6, 4, low(mins)).
table(low_c_minc_maxc_maxs_mins, 8, 6, 4, low(mins)).
table(low_c_minc_s_maxs_mins, 8, 5, 4, low(mins)).
table(low_c_maxc_s_maxs_mins, 8, 5, 4, low(mins)).
table(low_minc_maxc_s_maxs_mins, 8, 6, 4, low(mins)).
table(low_v_c_minc_maxc_s_mins, 8, 8, 5, low(mins)).
table(low_v_c_minc_maxc_maxs_mins, 8, 8, 5, low(mins)).
table(low_v_c_minc_s_maxs_mins, 8, 8, 5, low(mins)).
table(low_v_c_maxc_s_maxs_mins, 8, 6, 5, low(mins)).
table(low_v_minc_maxc_s_maxs_mins, 8, 7, 5, low(mins)).
table(low_c_minc_maxc_s_maxs_mins, 8, 7, 5, low(mins)).
table(low_v_c_minc_maxc_s_maxs_mins, 8, 9, 6, low(mins)).
table(up_v_mins, 8, 18, 1, up(mins)).
table(up_maxc_mins, 8, 9, 1, up(mins)).
table(up_maxs_mins, 8, 2, 1, up(mins)).
table(up_v_c_mins, 8, 6, 2, up(mins)).
table(up_v_minc_mins, 8, 3, 2, up(mins)).
table(up_v_maxc_mins, 8, 7, 2, up(mins)).
table(up_v_s_mins, 8, 3, 2, up(mins)).
table(up_v_maxs_mins, 8, 3, 2, up(mins)).
table(up_c_maxc_mins, 8, 5, 2, up(mins)).
table(up_c_maxs_mins, 8, 3, 2, up(mins)).
table(up_minc_maxc_mins, 8, 4, 2, up(mins)).
table(up_minc_maxs_mins, 8, 3, 2, up(mins)).
table(up_maxc_s_mins, 8, 3, 2, up(mins)).
table(up_maxc_maxs_mins, 8, 3, 2, up(mins)).
table(up_s_maxs_mins, 8, 3, 2, up(mins)).
table(up_v_c_minc_mins, 8, 6, 3, up(mins)).
table(up_v_c_maxc_mins, 8, 7, 3, up(mins)).
table(up_v_c_s_mins, 8, 4, 3, up(mins)).
table(up_v_c_maxs_mins, 8, 4, 3, up(mins)).
table(up_v_minc_maxc_mins, 8, 5, 3, up(mins)).
table(up_v_minc_s_mins, 8, 4, 3, up(mins)).
table(up_v_minc_maxs_mins, 8, 4, 3, up(mins)).
table(up_v_maxc_s_mins, 8, 4, 3, up(mins)).
table(up_v_maxc_maxs_mins, 8, 4, 3, up(mins)).
table(up_v_s_maxs_mins, 8, 4, 3, up(mins)).
table(up_c_minc_maxc_mins, 8, 5, 3, up(mins)).
table(up_c_minc_maxs_mins, 8, 4, 3, up(mins)).
table(up_c_maxc_s_mins, 8, 4, 3, up(mins)).
table(up_c_maxc_maxs_mins, 8, 4, 3, up(mins)).
table(up_c_s_maxs_mins, 8, 4, 3, up(mins)).
table(up_minc_maxc_s_mins, 8, 5, 3, up(mins)).
table(up_minc_maxc_maxs_mins, 8, 5, 3, up(mins)).
table(up_minc_s_maxs_mins, 8, 4, 3, up(mins)).
table(up_maxc_s_maxs_mins, 8, 4, 3, up(mins)).
table(up_v_c_minc_maxc_mins, 8, 7, 4, up(mins)).
table(up_v_c_minc_s_mins, 8, 7, 4, up(mins)).
table(up_v_c_minc_maxs_mins, 8, 7, 4, up(mins)).
table(up_v_c_maxc_s_mins, 8, 5, 4, up(mins)).
table(up_v_c_maxc_maxs_mins, 8, 5, 4, up(mins)).
table(up_v_c_s_maxs_mins, 8, 5, 4, up(mins)).
table(up_v_minc_maxc_s_mins, 8, 6, 4, up(mins)).
table(up_v_minc_maxc_maxs_mins, 8, 6, 4, up(mins)).
table(up_v_minc_s_maxs_mins, 8, 5, 4, up(mins)).
table(up_v_maxc_s_maxs_mins, 8, 5, 4, up(mins)).
table(up_c_minc_maxc_s_mins, 8, 6, 4, up(mins)).
table(up_c_minc_maxc_maxs_mins, 8, 6, 4, up(mins)).
table(up_c_minc_s_maxs_mins, 8, 5, 4, up(mins)).
table(up_c_maxc_s_maxs_mins, 8, 5, 4, up(mins)).
table(up_minc_maxc_s_maxs_mins, 8, 6, 4, up(mins)).
table(up_v_c_minc_maxc_s_mins, 8, 8, 5, up(mins)).
table(up_v_c_minc_maxc_maxs_mins, 8, 8, 5, up(mins)).
table(up_v_c_minc_s_maxs_mins, 8, 8, 5, up(mins)).
table(up_v_c_maxc_s_maxs_mins, 8, 6, 5, up(mins)).
table(up_v_minc_maxc_s_maxs_mins, 8, 7, 5, up(mins)).
table(up_c_minc_maxc_s_maxs_mins, 8, 7, 5, up(mins)).
table(up_v_c_minc_maxc_s_maxs_mins, 8, 9, 6, up(mins)).
table(low_v_maxs, 8, 7, 1, low(maxs)).
table(low_c_maxs, 8, 4, 1, low(maxs)).
table(low_minc_maxs, 8, 4, 1, low(maxs)).
table(low_maxc_maxs, 8, 4, 1, low(maxs)).
table(low_s_maxs, 8, 7, 1, low(maxs)).
table(low_mins_maxs, 8, 2, 1, low(maxs)).
table(low_v_c_maxs, 8, 8, 2, low(maxs)).
table(low_v_minc_maxs, 8, 8, 2, low(maxs)).
table(low_v_maxc_maxs, 8, 8, 2, low(maxs)).
table(low_v_s_maxs, 8, 3, 2, low(maxs)).
table(low_v_mins_maxs, 8, 7, 2, low(maxs)).
table(low_c_minc_maxs, 8, 5, 2, low(maxs)).
table(low_c_maxc_maxs, 8, 5, 2, low(maxs)).
table(low_c_s_maxs, 8, 8, 2, low(maxs)).
table(low_c_mins_maxs, 8, 3, 2, low(maxs)).
table(low_minc_maxc_maxs, 8, 6, 2, low(maxs)).
table(low_minc_s_maxs, 8, 3, 2, low(maxs)).
table(low_minc_mins_maxs, 8, 3, 2, low(maxs)).
table(low_maxc_s_maxs, 8, 3, 2, low(maxs)).
table(low_maxc_mins_maxs, 8, 3, 2, low(maxs)).
table(low_s_mins_maxs, 8, 7, 2, low(maxs)).
table(low_v_c_minc_maxs, 8, 11, 3, low(maxs)).
table(low_v_c_maxc_maxs, 8, 9, 3, low(maxs)).
table(low_v_c_s_maxs, 8, 4, 3, low(maxs)).
table(low_v_c_mins_maxs, 8, 8, 3, low(maxs)).
table(low_v_minc_maxc_maxs, 8, 10, 3, low(maxs)).
table(low_v_minc_s_maxs, 8, 4, 3, low(maxs)).
table(low_v_minc_mins_maxs, 8, 8, 3, low(maxs)).
table(low_v_maxc_s_maxs, 8, 4, 3, low(maxs)).
table(low_v_maxc_mins_maxs, 8, 8, 3, low(maxs)).
table(low_v_s_mins_maxs, 8, 4, 3, low(maxs)).
table(low_c_minc_maxc_maxs, 8, 7, 3, low(maxs)).
table(low_c_minc_s_maxs, 8, 4, 3, low(maxs)).
table(low_c_minc_mins_maxs, 8, 4, 3, low(maxs)).
table(low_c_maxc_s_maxs, 8, 4, 3, low(maxs)).
table(low_c_maxc_mins_maxs, 8, 4, 3, low(maxs)).
table(low_c_s_mins_maxs, 8, 8, 3, low(maxs)).
table(low_minc_maxc_s_maxs, 8, 5, 3, low(maxs)).
table(low_minc_maxc_mins_maxs, 8, 5, 3, low(maxs)).
table(low_minc_s_mins_maxs, 8, 4, 3, low(maxs)).
table(low_maxc_s_mins_maxs, 8, 4, 3, low(maxs)).
table(low_v_c_minc_maxc_maxs, 8, 12, 4, low(maxs)).
table(low_v_c_minc_s_maxs, 8, 7, 4, low(maxs)).
table(low_v_c_minc_mins_maxs, 8, 11, 4, low(maxs)).
table(low_v_c_maxc_s_maxs, 8, 5, 4, low(maxs)).
table(low_v_c_maxc_mins_maxs, 8, 9, 4, low(maxs)).
table(low_v_c_s_mins_maxs, 8, 5, 4, low(maxs)).
table(low_v_minc_maxc_s_maxs, 8, 6, 4, low(maxs)).
table(low_v_minc_maxc_mins_maxs, 8, 10, 4, low(maxs)).
table(low_v_minc_s_mins_maxs, 8, 5, 4, low(maxs)).
table(low_v_maxc_s_mins_maxs, 8, 5, 4, low(maxs)).
table(low_c_minc_maxc_s_maxs, 8, 6, 4, low(maxs)).
table(low_c_minc_maxc_mins_maxs, 8, 7, 4, low(maxs)).
table(low_c_minc_s_mins_maxs, 8, 5, 4, low(maxs)).
table(low_c_maxc_s_mins_maxs, 8, 5, 4, low(maxs)).
table(low_minc_maxc_s_mins_maxs, 8, 6, 4, low(maxs)).
table(low_v_c_minc_maxc_s_maxs, 8, 8, 5, low(maxs)).
table(low_v_c_minc_maxc_mins_maxs, 8, 12, 5, low(maxs)).
table(low_v_c_minc_s_mins_maxs, 8, 8, 5, low(maxs)).
table(low_v_c_maxc_s_mins_maxs, 8, 6, 5, low(maxs)).
table(low_v_minc_maxc_s_mins_maxs, 8, 7, 5, low(maxs)).
table(low_c_minc_maxc_s_mins_maxs, 8, 7, 5, low(maxs)).
table(low_v_c_minc_maxc_s_mins_maxs, 8, 9, 6, low(maxs)).
table(up_v_maxs, 8, 18, 1, up(maxs)).
table(up_maxc_maxs, 8, 2, 1, up(maxs)).
table(up_v_c_maxs, 8, 18, 2, up(maxs)).
table(up_v_minc_maxs, 8, 12, 2, up(maxs)).
table(up_v_maxc_maxs, 8, 3, 2, up(maxs)).
table(up_v_s_maxs, 8, 7, 2, up(maxs)).
table(up_v_mins_maxs, 8, 9, 2, up(maxs)).
table(up_c_maxc_maxs, 8, 3, 2, up(maxs)).
table(up_minc_maxc_maxs, 8, 4, 2, up(maxs)).
table(up_maxc_s_maxs, 8, 3, 2, up(maxs)).
table(up_maxc_mins_maxs, 8, 3, 2, up(maxs)).
table(up_v_c_minc_maxs, 8, 12, 3, up(maxs)).
table(up_v_c_maxc_maxs, 8, 4, 3, up(maxs)).
table(up_v_c_s_maxs, 8, 8, 3, up(maxs)).
table(up_v_c_mins_maxs, 8, 18, 3, up(maxs)).
table(up_v_minc_maxc_maxs, 8, 5, 3, up(maxs)).
table(up_v_minc_s_maxs, 8, 5, 3, up(maxs)).
table(up_v_minc_mins_maxs, 8, 14, 3, up(maxs)).
table(up_v_maxc_s_maxs, 8, 4, 3, up(maxs)).
table(up_v_maxc_mins_maxs, 8, 4, 3, up(maxs)).
table(up_v_s_mins_maxs, 8, 7, 3, up(maxs)).
table(up_c_minc_maxc_maxs, 8, 5, 3, up(maxs)).
table(up_c_maxc_s_maxs, 8, 4, 3, up(maxs)).
table(up_c_maxc_mins_maxs, 8, 4, 3, up(maxs)).
table(up_minc_maxc_s_maxs, 8, 5, 3, up(maxs)).
table(up_minc_maxc_mins_maxs, 8, 5, 3, up(maxs)).
table(up_maxc_s_mins_maxs, 8, 4, 3, up(maxs)).
table(up_v_c_minc_maxc_maxs, 8, 7, 4, up(maxs)).
table(up_v_c_minc_s_maxs, 8, 8, 4, up(maxs)).
table(up_v_c_minc_mins_maxs, 8, 13, 4, up(maxs)).
table(up_v_c_maxc_s_maxs, 8, 5, 4, up(maxs)).
table(up_v_c_maxc_mins_maxs, 8, 5, 4, up(maxs)).
table(up_v_c_s_mins_maxs, 8, 8, 4, up(maxs)).
table(up_v_minc_maxc_s_maxs, 8, 6, 4, up(maxs)).
table(up_v_minc_maxc_mins_maxs, 8, 6, 4, up(maxs)).
table(up_v_minc_s_mins_maxs, 8, 8, 4, up(maxs)).
table(up_v_maxc_s_mins_maxs, 8, 5, 4, up(maxs)).
table(up_c_minc_maxc_s_maxs, 8, 6, 4, up(maxs)).
table(up_c_minc_maxc_mins_maxs, 8, 6, 4, up(maxs)).
table(up_c_maxc_s_mins_maxs, 8, 5, 4, up(maxs)).
table(up_minc_maxc_s_mins_maxs, 8, 6, 4, up(maxs)).
table(up_v_c_minc_maxc_s_maxs, 8, 8, 5, up(maxs)).
table(up_v_c_minc_maxc_mins_maxs, 8, 8, 5, up(maxs)).
table(up_v_c_minc_s_mins_maxs, 8, 11, 5, up(maxs)).
table(up_v_c_maxc_s_mins_maxs, 8, 6, 5, up(maxs)).
table(up_v_minc_maxc_s_mins_maxs, 8, 7, 5, up(maxs)).
table(up_c_minc_maxc_s_mins_maxs, 8, 7, 5, up(maxs)).
table(up_v_c_minc_maxc_s_mins_maxs, 8, 9, 6, up(maxs)).
table(low_v_mina, 8, 12, 1, low(mina)).
table(low_c_mina, 8, 7, 1, low(mina)).
table(low_minc_mina, 8, 19, 1, low(mina)).
table(low_maxc_mina, 8, 19, 1, low(mina)).
table(low_s_mina, 8, 12, 1, low(mina)).
table(low_mins_mina, 8, 11, 1, low(mina)).
table(low_maxs_mina, 8, 11, 1, low(mina)).
table(low_v_c_mina, 8, 12, 2, low(mina)).
table(low_v_minc_mina, 8, 12, 2, low(mina)).
table(low_v_maxc_mina, 8, 11, 2, low(mina)).
table(low_v_s_mina, 8, 4, 2, low(mina)).
table(low_v_mins_mina, 8, 6, 2, low(mina)).
table(low_v_maxs_mina, 8, 10, 2, low(mina)).
table(low_c_minc_mina, 8, 8, 2, low(mina)).
table(low_c_maxc_mina, 8, 6, 2, low(mina)).
table(low_c_s_mina, 8, 12, 2, low(mina)).
table(low_c_mins_mina, 8, 7, 2, low(mina)).
table(low_c_maxs_mina, 8, 5, 2, low(mina)).
table(low_minc_maxc_mina, 8, 19, 2, low(mina)).
table(low_minc_s_mina, 8, 12, 2, low(mina)).
table(low_minc_mins_mina, 8, 19, 2, low(mina)).
table(low_minc_maxs_mina, 8, 19, 2, low(mina)).
table(low_maxc_s_mina, 8, 11, 2, low(mina)).
table(low_maxc_mins_mina, 8, 19, 2, low(mina)).
table(low_maxc_maxs_mina, 8, 19, 2, low(mina)).
table(low_s_mins_mina, 8, 12, 2, low(mina)).
table(low_s_maxs_mina, 8, 10, 2, low(mina)).
table(low_mins_maxs_mina, 8, 6, 2, low(mina)).
table(low_v_c_minc_mina, 8, 13, 3, low(mina)).
table(low_v_c_maxc_mina, 8, 19, 3, low(mina)).
table(low_v_c_s_mina, 8, 4, 3, low(mina)).
table(low_v_c_mins_mina, 8, 9, 3, low(mina)).
table(low_v_c_maxs_mina, 8, 12, 3, low(mina)).
table(low_v_minc_maxc_mina, 8, 13, 3, low(mina)).
table(low_v_minc_s_mina, 8, 5, 3, low(mina)).
table(low_v_minc_mins_mina, 8, 5, 3, low(mina)).
table(low_v_minc_maxs_mina, 8, 11, 3, low(mina)).
table(low_v_maxc_s_mina, 8, 4, 3, low(mina)).
table(low_v_maxc_mins_mina, 8, 11, 3, low(mina)).
table(low_v_maxc_maxs_mina, 8, 10, 3, low(mina)).
table(low_v_s_mins_mina, 8, 5, 3, low(mina)).
table(low_v_s_maxs_mina, 8, 5, 3, low(mina)).
table(low_v_mins_maxs_mina, 8, 10, 3, low(mina)).
table(low_c_minc_maxc_mina, 8, 8, 3, low(mina)).
table(low_c_minc_s_mina, 8, 11, 3, low(mina)).
table(low_c_minc_mins_mina, 8, 8, 3, low(mina)).
table(low_c_minc_maxs_mina, 8, 6, 3, low(mina)).
table(low_c_maxc_s_mina, 8, 9, 3, low(mina)).
table(low_c_maxc_mins_mina, 8, 6, 3, low(mina)).
table(low_c_maxc_maxs_mina, 8, 6, 3, low(mina)).
table(low_c_s_mins_mina, 8, 12, 3, low(mina)).
table(low_c_s_maxs_mina, 8, 12, 3, low(mina)).
table(low_c_mins_maxs_mina, 8, 5, 3, low(mina)).
table(low_minc_maxc_s_mina, 8, 10, 3, low(mina)).
table(low_minc_maxc_mins_mina, 8, 19, 3, low(mina)).
table(low_minc_maxc_maxs_mina, 8, 18, 3, low(mina)).
table(low_minc_s_mins_mina, 8, 9, 3, low(mina)).
table(low_minc_s_maxs_mina, 8, 11, 3, low(mina)).
table(low_minc_mins_maxs_mina, 8, 19, 3, low(mina)).
table(low_maxc_s_mins_mina, 8, 8, 3, low(mina)).
table(low_maxc_s_maxs_mina, 8, 10, 3, low(mina)).
table(low_maxc_mins_maxs_mina, 8, 9, 3, low(mina)).
table(low_s_mins_maxs_mina, 8, 10, 3, low(mina)).
table(low_v_c_minc_maxc_mina, 8, 19, 4, low(mina)).
table(low_v_c_minc_s_mina, 8, 8, 4, low(mina)).
table(low_v_c_minc_mins_mina, 8, 10, 4, low(mina)).
table(low_v_c_minc_maxs_mina, 8, 13, 4, low(mina)).
table(low_v_c_maxc_s_mina, 8, 5, 4, low(mina)).
table(low_v_c_maxc_mins_mina, 8, 19, 4, low(mina)).
table(low_v_c_maxc_maxs_mina, 8, 18, 4, low(mina)).
table(low_v_c_s_mins_mina, 8, 5, 4, low(mina)).
table(low_v_c_s_maxs_mina, 8, 5, 4, low(mina)).
table(low_v_c_mins_maxs_mina, 8, 12, 4, low(mina)).
table(low_v_minc_maxc_s_mina, 8, 7, 4, low(mina)).
table(low_v_minc_maxc_mins_mina, 8, 13, 4, low(mina)).
table(low_v_minc_maxc_maxs_mina, 8, 12, 4, low(mina)).
table(low_v_minc_s_mins_mina, 8, 5, 4, low(mina)).
table(low_v_minc_s_maxs_mina, 8, 7, 4, low(mina)).
table(low_v_minc_mins_maxs_mina, 8, 11, 4, low(mina)).
table(low_v_maxc_s_mins_mina, 8, 7, 4, low(mina)).
table(low_v_maxc_s_maxs_mina, 8, 5, 4, low(mina)).
table(low_v_maxc_mins_maxs_mina, 8, 10, 4, low(mina)).
table(low_v_s_mins_maxs_mina, 8, 10, 4, low(mina)).
table(low_c_minc_maxc_s_mina, 8, 10, 4, low(mina)).
table(low_c_minc_maxc_mins_mina, 8, 8, 4, low(mina)).
table(low_c_minc_maxc_maxs_mina, 8, 8, 4, low(mina)).
table(low_c_minc_s_mins_mina, 8, 10, 4, low(mina)).
table(low_c_minc_s_maxs_mina, 8, 13, 4, low(mina)).
table(low_c_minc_mins_maxs_mina, 8, 6, 4, low(mina)).
table(low_c_maxc_s_mins_mina, 8, 7, 4, low(mina)).
table(low_c_maxc_s_maxs_mina, 8, 12, 4, low(mina)).
table(low_c_maxc_mins_maxs_mina, 8, 6, 4, low(mina)).
table(low_c_s_mins_maxs_mina, 8, 12, 4, low(mina)).
table(low_minc_maxc_s_mins_mina, 8, 9, 4, low(mina)).
table(low_minc_maxc_s_maxs_mina, 8, 12, 4, low(mina)).
table(low_minc_maxc_mins_maxs_mina, 8, 18, 4, low(mina)).
table(low_minc_s_mins_maxs_mina, 8, 11, 4, low(mina)).
table(low_maxc_s_mins_maxs_mina, 8, 10, 4, low(mina)).
table(low_v_c_minc_maxc_s_mina, 8, 9, 5, low(mina)).
table(low_v_c_minc_maxc_mins_mina, 8, 19, 5, low(mina)).
table(low_v_c_minc_maxc_maxs_mina, 8, 18, 5, low(mina)).
table(low_v_c_minc_s_mins_mina, 8, 8, 5, low(mina)).
table(low_v_c_minc_s_maxs_mina, 8, 9, 5, low(mina)).
table(low_v_c_minc_mins_maxs_mina, 8, 13, 5, low(mina)).
table(low_v_c_maxc_s_mins_mina, 8, 9, 5, low(mina)).
table(low_v_c_maxc_s_maxs_mina, 8, 6, 5, low(mina)).
table(low_v_c_maxc_mins_maxs_mina, 8, 18, 5, low(mina)).
table(low_v_c_s_mins_maxs_mina, 8, 6, 5, low(mina)).
table(low_v_minc_maxc_s_mins_mina, 8, 7, 5, low(mina)).
table(low_v_minc_maxc_s_maxs_mina, 8, 8, 5, low(mina)).
table(low_v_minc_maxc_mins_maxs_mina, 8, 12, 5, low(mina)).
table(low_v_minc_s_mins_maxs_mina, 8, 7, 5, low(mina)).
table(low_v_maxc_s_mins_maxs_mina, 8, 6, 5, low(mina)).
table(low_c_minc_maxc_s_mins_mina, 8, 9, 5, low(mina)).
table(low_c_minc_maxc_s_maxs_mina, 8, 18, 5, low(mina)).
table(low_c_minc_maxc_mins_maxs_mina, 8, 8, 5, low(mina)).
table(low_c_minc_s_mins_maxs_mina, 8, 13, 5, low(mina)).
table(low_c_maxc_s_mins_maxs_mina, 8, 12, 5, low(mina)).
table(low_minc_maxc_s_mins_maxs_mina, 8, 12, 5, low(mina)).
table(low_v_c_minc_maxc_s_mins_mina, 8, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_maxs_mina, 8, 10, 6, low(mina)).
table(low_v_c_minc_maxc_mins_maxs_mina, 8, 18, 6, low(mina)).
table(low_v_c_minc_s_mins_maxs_mina, 8, 9, 6, low(mina)).
table(low_v_c_maxc_s_mins_maxs_mina, 8, 7, 6, low(mina)).
table(low_v_minc_maxc_s_mins_maxs_mina, 8, 8, 6, low(mina)).
table(low_c_minc_maxc_s_mins_maxs_mina, 8, 18, 6, low(mina)).
table(low_v_c_minc_maxc_s_mins_maxs_mina, 8, 10, 7, low(mina)).
table(up_v_maxa, 8, 19, 1, up(maxa)).
table(up_v_c_maxa, 8, 19, 2, up(maxa)).
table(up_v_minc_maxa, 8, 19, 2, up(maxa)).
table(up_v_maxc_maxa, 8, 19, 2, up(maxa)).
table(up_v_s_maxa, 8, 19, 2, up(maxa)).
table(up_v_mins_maxa, 8, 19, 2, up(maxa)).
table(up_v_maxs_maxa, 8, 19, 2, up(maxa)).
table(up_c_maxc_maxa, 8, 19, 2, up(maxa)).
table(up_maxc_s_maxa, 8, 18, 2, up(maxa)).
table(up_s_maxs_maxa, 8, 19, 2, up(maxa)).
table(up_v_c_minc_maxa, 8, 19, 3, up(maxa)).
table(up_v_c_maxc_maxa, 8, 19, 3, up(maxa)).
table(up_v_c_s_maxa, 8, 19, 3, up(maxa)).
table(up_v_c_mins_maxa, 8, 19, 3, up(maxa)).
table(up_v_c_maxs_maxa, 8, 7, 3, up(maxa)).
table(up_v_minc_maxc_maxa, 8, 19, 3, up(maxa)).
table(up_v_minc_s_maxa, 8, 18, 3, up(maxa)).
table(up_v_minc_mins_maxa, 8, 19, 3, up(maxa)).
table(up_v_minc_maxs_maxa, 8, 19, 3, up(maxa)).
table(up_v_maxc_s_maxa, 8, 18, 3, up(maxa)).
table(up_v_maxc_mins_maxa, 8, 19, 3, up(maxa)).
table(up_v_maxc_maxs_maxa, 8, 9, 3, up(maxa)).
table(up_v_s_mins_maxa, 8, 19, 3, up(maxa)).
table(up_v_s_maxs_maxa, 8, 19, 3, up(maxa)).
table(up_v_mins_maxs_maxa, 8, 19, 3, up(maxa)).
table(up_c_minc_maxc_maxa, 8, 19, 3, up(maxa)).
table(up_c_maxc_s_maxa, 8, 18, 3, up(maxa)).
table(up_c_maxc_mins_maxa, 8, 19, 3, up(maxa)).
table(up_c_maxc_maxs_maxa, 8, 6, 3, up(maxa)).
table(up_c_s_maxs_maxa, 8, 19, 3, up(maxa)).
table(up_minc_maxc_s_maxa, 8, 18, 3, up(maxa)).
table(up_minc_s_maxs_maxa, 8, 18, 3, up(maxa)).
table(up_maxc_s_mins_maxa, 8, 18, 3, up(maxa)).
table(up_maxc_s_maxs_maxa, 8, 18, 3, up(maxa)).
table(up_s_mins_maxs_maxa, 8, 19, 3, up(maxa)).
table(up_v_c_minc_maxc_maxa, 8, 19, 4, up(maxa)).
table(up_v_c_minc_s_maxa, 8, 18, 4, up(maxa)).
table(up_v_c_minc_mins_maxa, 8, 19, 4, up(maxa)).
table(up_v_c_minc_maxs_maxa, 8, 11, 4, up(maxa)).
table(up_v_c_maxc_s_maxa, 8, 18, 4, up(maxa)).
table(up_v_c_maxc_mins_maxa, 8, 19, 4, up(maxa)).
table(up_v_c_maxc_maxs_maxa, 8, 6, 4, up(maxa)).
table(up_v_c_s_mins_maxa, 8, 19, 4, up(maxa)).
table(up_v_c_s_maxs_maxa, 8, 19, 4, up(maxa)).
table(up_v_c_mins_maxs_maxa, 8, 11, 4, up(maxa)).
table(up_v_minc_maxc_s_maxa, 8, 18, 4, up(maxa)).
table(up_v_minc_maxc_mins_maxa, 8, 19, 4, up(maxa)).
table(up_v_minc_maxc_maxs_maxa, 8, 10, 4, up(maxa)).
table(up_v_minc_s_mins_maxa, 8, 18, 4, up(maxa)).
table(up_v_minc_s_maxs_maxa, 8, 18, 4, up(maxa)).
table(up_v_minc_mins_maxs_maxa, 8, 18, 4, up(maxa)).
table(up_v_maxc_s_mins_maxa, 8, 18, 4, up(maxa)).
table(up_v_maxc_s_maxs_maxa, 8, 18, 4, up(maxa)).
table(up_v_maxc_mins_maxs_maxa, 8, 8, 4, up(maxa)).
table(up_v_s_mins_maxs_maxa, 8, 19, 4, up(maxa)).
table(up_c_minc_maxc_s_maxa, 8, 18, 4, up(maxa)).
table(up_c_minc_maxc_mins_maxa, 8, 19, 4, up(maxa)).
table(up_c_minc_maxc_maxs_maxa, 8, 10, 4, up(maxa)).
table(up_c_minc_s_maxs_maxa, 8, 12, 4, up(maxa)).
table(up_c_maxc_s_mins_maxa, 8, 18, 4, up(maxa)).
table(up_c_maxc_s_maxs_maxa, 8, 18, 4, up(maxa)).
table(up_c_maxc_mins_maxs_maxa, 8, 9, 4, up(maxa)).
table(up_c_s_mins_maxs_maxa, 8, 19, 4, up(maxa)).
table(up_minc_maxc_s_mins_maxa, 8, 18, 4, up(maxa)).
table(up_minc_maxc_s_maxs_maxa, 8, 18, 4, up(maxa)).
table(up_minc_s_mins_maxs_maxa, 8, 18, 4, up(maxa)).
table(up_maxc_s_mins_maxs_maxa, 8, 18, 4, up(maxa)).
table(up_v_c_minc_maxc_s_maxa, 8, 18, 5, up(maxa)).
table(up_v_c_minc_maxc_mins_maxa, 8, 19, 5, up(maxa)).
table(up_v_c_minc_maxc_maxs_maxa, 8, 10, 5, up(maxa)).
table(up_v_c_minc_s_mins_maxa, 8, 18, 5, up(maxa)).
table(up_v_c_minc_s_maxs_maxa, 8, 12, 5, up(maxa)).
table(up_v_c_minc_mins_maxs_maxa, 8, 10, 5, up(maxa)).
table(up_v_c_maxc_s_mins_maxa, 8, 18, 5, up(maxa)).
table(up_v_c_maxc_s_maxs_maxa, 8, 18, 5, up(maxa)).
table(up_v_c_maxc_mins_maxs_maxa, 8, 6, 5, up(maxa)).
table(up_v_c_s_mins_maxs_maxa, 8, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_mins_maxa, 8, 18, 5, up(maxa)).
table(up_v_minc_maxc_s_maxs_maxa, 8, 18, 5, up(maxa)).
table(up_v_minc_maxc_mins_maxs_maxa, 8, 9, 5, up(maxa)).
table(up_v_minc_s_mins_maxs_maxa, 8, 18, 5, up(maxa)).
table(up_v_maxc_s_mins_maxs_maxa, 8, 18, 5, up(maxa)).
table(up_c_minc_maxc_s_mins_maxa, 8, 18, 5, up(maxa)).
table(up_c_minc_maxc_s_maxs_maxa, 8, 18, 5, up(maxa)).
table(up_c_minc_maxc_mins_maxs_maxa, 8, 9, 5, up(maxa)).
table(up_c_minc_s_mins_maxs_maxa, 8, 12, 5, up(maxa)).
table(up_c_maxc_s_mins_maxs_maxa, 8, 18, 5, up(maxa)).
table(up_minc_maxc_s_mins_maxs_maxa, 8, 18, 5, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxa, 8, 18, 6, up(maxa)).
table(up_v_c_minc_maxc_s_maxs_maxa, 8, 18, 6, up(maxa)).
table(up_v_c_minc_maxc_mins_maxs_maxa, 8, 9, 6, up(maxa)).
table(up_v_c_minc_s_mins_maxs_maxa, 8, 12, 6, up(maxa)).
table(up_v_c_maxc_s_mins_maxs_maxa, 8, 18, 6, up(maxa)).
table(up_v_minc_maxc_s_mins_maxs_maxa, 8, 18, 6, up(maxa)).
table(up_c_minc_maxc_s_mins_maxs_maxa, 8, 18, 6, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxs_maxa, 8, 18, 7, up(maxa)).
table(low_c_v, 9, 18, 1, low(v)).
table(low_minc_v, 9, 11, 1, low(v)).
table(low_maxc_v, 9, 11, 1, low(v)).
table(low_s_v, 9, 7, 1, low(v)).
table(low_mins_v, 9, 18, 1, low(v)).
table(low_maxs_v, 9, 18, 1, low(v)).
table(low_c_minc_v, 9, 11, 2, low(v)).
table(low_c_maxc_v, 9, 11, 2, low(v)).
table(low_c_s_v, 9, 8, 2, low(v)).
table(low_c_mins_v, 9, 18, 2, low(v)).
table(low_c_maxs_v, 9, 18, 2, low(v)).
table(low_minc_maxc_v, 9, 11, 2, low(v)).
table(low_minc_s_v, 9, 3, 2, low(v)).
table(low_minc_mins_v, 9, 12, 2, low(v)).
table(low_minc_maxs_v, 9, 12, 2, low(v)).
table(low_maxc_s_v, 9, 3, 2, low(v)).
table(low_maxc_mins_v, 9, 12, 2, low(v)).
table(low_maxc_maxs_v, 9, 13, 2, low(v)).
table(low_s_mins_v, 9, 7, 2, low(v)).
table(low_s_maxs_v, 9, 7, 2, low(v)).
table(low_mins_maxs_v, 9, 12, 2, low(v)).
table(low_c_minc_maxc_v, 9, 11, 3, low(v)).
table(low_c_minc_s_v, 9, 6, 3, low(v)).
table(low_c_minc_mins_v, 9, 12, 3, low(v)).
table(low_c_minc_maxs_v, 9, 12, 3, low(v)).
table(low_c_maxc_s_v, 9, 4, 3, low(v)).
table(low_c_maxc_mins_v, 9, 12, 3, low(v)).
table(low_c_maxc_maxs_v, 9, 13, 3, low(v)).
table(low_c_s_mins_v, 9, 8, 3, low(v)).
table(low_c_s_maxs_v, 9, 8, 3, low(v)).
table(low_c_mins_maxs_v, 9, 18, 3, low(v)).
table(low_minc_maxc_s_v, 9, 5, 3, low(v)).
table(low_minc_maxc_mins_v, 9, 12, 3, low(v)).
table(low_minc_maxc_maxs_v, 9, 12, 3, low(v)).
table(low_minc_s_mins_v, 9, 4, 3, low(v)).
table(low_minc_s_maxs_v, 9, 4, 3, low(v)).
table(low_minc_mins_maxs_v, 9, 13, 3, low(v)).
table(low_maxc_s_mins_v, 9, 4, 3, low(v)).
table(low_maxc_s_maxs_v, 9, 4, 3, low(v)).
table(low_maxc_mins_maxs_v, 9, 14, 3, low(v)).
table(low_s_mins_maxs_v, 9, 7, 3, low(v)).
table(low_c_minc_maxc_s_v, 9, 7, 4, low(v)).
table(low_c_minc_maxc_mins_v, 9, 12, 4, low(v)).
table(low_c_minc_maxc_maxs_v, 9, 12, 4, low(v)).
table(low_c_minc_s_mins_v, 9, 7, 4, low(v)).
table(low_c_minc_s_maxs_v, 9, 7, 4, low(v)).
table(low_c_minc_mins_maxs_v, 9, 13, 4, low(v)).
table(low_c_maxc_s_mins_v, 9, 5, 4, low(v)).
table(low_c_maxc_s_maxs_v, 9, 5, 4, low(v)).
table(low_c_maxc_mins_maxs_v, 9, 14, 4, low(v)).
table(low_c_s_mins_maxs_v, 9, 8, 4, low(v)).
table(low_minc_maxc_s_mins_v, 9, 6, 4, low(v)).
table(low_minc_maxc_s_maxs_v, 9, 6, 4, low(v)).
table(low_minc_maxc_mins_maxs_v, 9, 13, 4, low(v)).
table(low_minc_s_mins_maxs_v, 9, 5, 4, low(v)).
table(low_maxc_s_mins_maxs_v, 9, 5, 4, low(v)).
table(low_c_minc_maxc_s_mins_v, 9, 8, 5, low(v)).
table(low_c_minc_maxc_s_maxs_v, 9, 8, 5, low(v)).
table(low_c_minc_maxc_mins_maxs_v, 9, 13, 5, low(v)).
table(low_c_minc_s_mins_maxs_v, 9, 8, 5, low(v)).
table(low_c_maxc_s_mins_maxs_v, 9, 6, 5, low(v)).
table(low_minc_maxc_s_mins_maxs_v, 9, 7, 5, low(v)).
table(low_c_minc_maxc_s_mins_maxs_v, 9, 9, 6, low(v)).
table(up_c_maxc_v, 9, 3, 2, up(v)).
table(up_maxc_s_v, 9, 3, 2, up(v)).
table(up_s_maxs_v, 9, 3, 2, up(v)).
table(up_c_minc_maxc_v, 9, 6, 3, up(v)).
table(up_c_maxc_s_v, 9, 4, 3, up(v)).
table(up_c_maxc_mins_v, 9, 4, 3, up(v)).
table(up_c_maxc_maxs_v, 9, 4, 3, up(v)).
table(up_c_s_maxs_v, 9, 4, 3, up(v)).
table(up_minc_maxc_s_v, 9, 5, 3, up(v)).
table(up_minc_s_maxs_v, 9, 4, 3, up(v)).
table(up_maxc_s_mins_v, 9, 4, 3, up(v)).
table(up_maxc_s_maxs_v, 9, 4, 3, up(v)).
table(up_s_mins_maxs_v, 9, 4, 3, up(v)).
table(up_c_minc_maxc_s_v, 9, 7, 4, up(v)).
table(up_c_minc_maxc_mins_v, 9, 7, 4, up(v)).
table(up_c_minc_maxc_maxs_v, 9, 7, 4, up(v)).
table(up_c_minc_s_maxs_v, 9, 7, 4, up(v)).
table(up_c_maxc_s_mins_v, 9, 5, 4, up(v)).
table(up_c_maxc_s_maxs_v, 9, 5, 4, up(v)).
table(up_c_maxc_mins_maxs_v, 9, 5, 4, up(v)).
table(up_c_s_mins_maxs_v, 9, 5, 4, up(v)).
table(up_minc_maxc_s_mins_v, 9, 6, 4, up(v)).
table(up_minc_maxc_s_maxs_v, 9, 6, 4, up(v)).
table(up_minc_s_mins_maxs_v, 9, 5, 4, up(v)).
table(up_maxc_s_mins_maxs_v, 9, 5, 4, up(v)).
table(up_c_minc_maxc_s_mins_v, 9, 8, 5, up(v)).
table(up_c_minc_maxc_s_maxs_v, 9, 8, 5, up(v)).
table(up_c_minc_maxc_mins_maxs_v, 9, 8, 5, up(v)).
table(up_c_minc_s_mins_maxs_v, 9, 8, 5, up(v)).
table(up_c_maxc_s_mins_maxs_v, 9, 6, 5, up(v)).
table(up_minc_maxc_s_mins_maxs_v, 9, 7, 5, up(v)).
table(up_c_minc_maxc_s_mins_maxs_v, 9, 9, 6, up(v)).
table(low_v_c, 9, 11, 1, low(c)).
table(low_minc_c, 9, 11, 1, low(c)).
table(low_maxc_c, 9, 11, 1, low(c)).
table(low_s_c, 9, 8, 1, low(c)).
table(low_mins_c, 9, 8, 1, low(c)).
table(low_maxs_c, 9, 8, 1, low(c)).
table(low_v_minc_c, 9, 11, 2, low(c)).
table(low_v_maxc_c, 9, 3, 2, low(c)).
table(low_v_s_c, 9, 13, 2, low(c)).
table(low_v_mins_c, 9, 12, 2, low(c)).
table(low_v_maxs_c, 9, 13, 2, low(c)).
table(low_minc_maxc_c, 9, 11, 2, low(c)).
table(low_minc_s_c, 9, 5, 2, low(c)).
table(low_minc_mins_c, 9, 9, 2, low(c)).
table(low_minc_maxs_c, 9, 9, 2, low(c)).
table(low_maxc_s_c, 9, 3, 2, low(c)).
table(low_maxc_mins_c, 9, 12, 2, low(c)).
table(low_maxc_maxs_c, 9, 13, 2, low(c)).
table(low_s_mins_c, 9, 10, 2, low(c)).
table(low_s_maxs_c, 9, 11, 2, low(c)).
table(low_mins_maxs_c, 9, 9, 2, low(c)).
table(low_v_minc_maxc_c, 9, 6, 3, low(c)).
table(low_v_minc_s_c, 9, 12, 3, low(c)).
table(low_v_minc_mins_c, 9, 12, 3, low(c)).
table(low_v_minc_maxs_c, 9, 12, 3, low(c)).
table(low_v_maxc_s_c, 9, 4, 3, low(c)).
table(low_v_maxc_mins_c, 9, 4, 3, low(c)).
table(low_v_maxc_maxs_c, 9, 4, 3, low(c)).
table(low_v_s_mins_c, 9, 14, 3, low(c)).
table(low_v_s_maxs_c, 9, 14, 3, low(c)).
table(low_v_mins_maxs_c, 9, 14, 3, low(c)).
table(low_minc_maxc_s_c, 9, 5, 3, low(c)).
table(low_minc_maxc_mins_c, 9, 12, 3, low(c)).
table(low_minc_maxc_maxs_c, 9, 12, 3, low(c)).
table(low_minc_s_mins_c, 9, 6, 3, low(c)).
table(low_minc_s_maxs_c, 9, 7, 3, low(c)).
table(low_minc_mins_maxs_c, 9, 7, 3, low(c)).
table(low_maxc_s_mins_c, 9, 4, 3, low(c)).
table(low_maxc_s_maxs_c, 9, 4, 3, low(c)).
table(low_maxc_mins_maxs_c, 9, 6, 3, low(c)).
table(low_s_mins_maxs_c, 9, 12, 3, low(c)).
table(low_v_minc_maxc_s_c, 9, 7, 4, low(c)).
table(low_v_minc_maxc_mins_c, 9, 7, 4, low(c)).
table(low_v_minc_maxc_maxs_c, 9, 7, 4, low(c)).
table(low_v_minc_s_mins_c, 9, 13, 4, low(c)).
table(low_v_minc_s_maxs_c, 9, 13, 4, low(c)).
table(low_v_minc_mins_maxs_c, 9, 13, 4, low(c)).
table(low_v_maxc_s_mins_c, 9, 5, 4, low(c)).
table(low_v_maxc_s_maxs_c, 9, 5, 4, low(c)).
table(low_v_maxc_mins_maxs_c, 9, 5, 4, low(c)).
table(low_v_s_mins_maxs_c, 9, 15, 4, low(c)).
table(low_minc_maxc_s_mins_c, 9, 6, 4, low(c)).
table(low_minc_maxc_s_maxs_c, 9, 6, 4, low(c)).
table(low_minc_maxc_mins_maxs_c, 9, 13, 4, low(c)).
table(low_minc_s_mins_maxs_c, 9, 8, 4, low(c)).
table(low_maxc_s_mins_maxs_c, 9, 5, 4, low(c)).
table(low_v_minc_maxc_s_mins_c, 9, 8, 5, low(c)).
table(low_v_minc_maxc_s_maxs_c, 9, 8, 5, low(c)).
table(low_v_minc_maxc_mins_maxs_c, 9, 8, 5, low(c)).
table(low_v_minc_s_mins_maxs_c, 9, 14, 5, low(c)).
table(low_v_maxc_s_mins_maxs_c, 9, 6, 5, low(c)).
table(low_minc_maxc_s_mins_maxs_c, 9, 7, 5, low(c)).
table(low_v_minc_maxc_s_mins_maxs_c, 9, 9, 6, low(c)).
table(up_v_c, 9, 18, 1, up(c)).
table(up_s_c, 9, 3, 1, up(c)).
table(up_v_minc_c, 9, 11, 2, up(c)).
table(up_v_maxc_c, 9, 11, 2, up(c)).
table(up_v_s_c, 9, 4, 2, up(c)).
table(up_v_mins_c, 9, 18, 2, up(c)).
table(up_v_maxs_c, 9, 18, 2, up(c)).
table(up_minc_s_c, 9, 3, 2, up(c)).
table(up_maxc_s_c, 9, 4, 2, up(c)).
table(up_s_mins_c, 9, 5, 2, up(c)).
table(up_s_maxs_c, 9, 5, 2, up(c)).
table(up_v_minc_maxc_c, 9, 11, 3, up(c)).
table(up_v_minc_s_c, 9, 6, 3, up(c)).
table(up_v_minc_mins_c, 9, 12, 3, up(c)).
table(up_v_minc_maxs_c, 9, 7, 3, up(c)).
table(up_v_maxc_s_c, 9, 5, 3, up(c)).
table(up_v_maxc_mins_c, 9, 12, 3, up(c)).
table(up_v_maxc_maxs_c, 9, 13, 3, up(c)).
table(up_v_s_mins_c, 9, 8, 3, up(c)).
table(up_v_s_maxs_c, 9, 6, 3, up(c)).
table(up_v_mins_maxs_c, 9, 18, 3, up(c)).
table(up_minc_maxc_s_c, 9, 5, 3, up(c)).
table(up_minc_s_mins_c, 9, 4, 3, up(c)).
table(up_minc_s_maxs_c, 9, 4, 3, up(c)).
table(up_maxc_s_mins_c, 9, 5, 3, up(c)).
table(up_maxc_s_maxs_c, 9, 5, 3, up(c)).
table(up_s_mins_maxs_c, 9, 8, 3, up(c)).
table(up_v_minc_maxc_s_c, 9, 7, 4, up(c)).
table(up_v_minc_maxc_mins_c, 9, 12, 4, up(c)).
table(up_v_minc_maxc_maxs_c, 9, 12, 4, up(c)).
table(up_v_minc_s_mins_c, 9, 7, 4, up(c)).
table(up_v_minc_s_maxs_c, 9, 7, 4, up(c)).
table(up_v_minc_mins_maxs_c, 9, 8, 4, up(c)).
table(up_v_maxc_s_mins_c, 9, 9, 4, up(c)).
table(up_v_maxc_s_maxs_c, 9, 6, 4, up(c)).
table(up_v_maxc_mins_maxs_c, 9, 14, 4, up(c)).
table(up_v_s_mins_maxs_c, 9, 10, 4, up(c)).
table(up_minc_maxc_s_mins_c, 9, 6, 4, up(c)).
table(up_minc_maxc_s_maxs_c, 9, 6, 4, up(c)).
table(up_minc_s_mins_maxs_c, 9, 5, 4, up(c)).
table(up_maxc_s_mins_maxs_c, 9, 6, 4, up(c)).
table(up_v_minc_maxc_s_mins_c, 9, 8, 5, up(c)).
table(up_v_minc_maxc_s_maxs_c, 9, 8, 5, up(c)).
table(up_v_minc_maxc_mins_maxs_c, 9, 13, 5, up(c)).
table(up_v_minc_s_mins_maxs_c, 9, 8, 5, up(c)).
table(up_v_maxc_s_mins_maxs_c, 9, 7, 5, up(c)).
table(up_minc_maxc_s_mins_maxs_c, 9, 7, 5, up(c)).
table(up_v_minc_maxc_s_mins_maxs_c, 9, 9, 6, up(c)).
table(low_v_minc, 9, 3, 1, low(minc)).
table(low_c_minc, 9, 3, 1, low(minc)).
table(low_maxc_minc, 9, 3, 1, low(minc)).
table(low_s_minc, 9, 3, 1, low(minc)).
table(low_mins_minc, 9, 2, 1, low(minc)).
table(low_maxs_minc, 9, 3, 1, low(minc)).
table(low_v_c_minc, 9, 5, 2, low(minc)).
table(low_v_maxc_minc, 9, 4, 2, low(minc)).
table(low_v_s_minc, 9, 4, 2, low(minc)).
table(low_v_mins_minc, 9, 3, 2, low(minc)).
table(low_v_maxs_minc, 9, 4, 2, low(minc)).
table(low_c_maxc_minc, 9, 4, 2, low(minc)).
table(low_c_s_minc, 9, 4, 2, low(minc)).
table(low_c_mins_minc, 9, 3, 2, low(minc)).
table(low_c_maxs_minc, 9, 4, 2, low(minc)).
table(low_maxc_s_minc, 9, 4, 2, low(minc)).
table(low_maxc_mins_minc, 9, 4, 2, low(minc)).
table(low_maxc_maxs_minc, 9, 4, 2, low(minc)).
table(low_s_mins_minc, 9, 3, 2, low(minc)).
table(low_s_maxs_minc, 9, 4, 2, low(minc)).
table(low_mins_maxs_minc, 9, 3, 2, low(minc)).
table(low_v_c_maxc_minc, 9, 6, 3, low(minc)).
table(low_v_c_s_minc, 9, 6, 3, low(minc)).
table(low_v_c_mins_minc, 9, 6, 3, low(minc)).
table(low_v_c_maxs_minc, 9, 6, 3, low(minc)).
table(low_v_maxc_s_minc, 9, 5, 3, low(minc)).
table(low_v_maxc_mins_minc, 9, 5, 3, low(minc)).
table(low_v_maxc_maxs_minc, 9, 5, 3, low(minc)).
table(low_v_s_mins_minc, 9, 4, 3, low(minc)).
table(low_v_s_maxs_minc, 9, 5, 3, low(minc)).
table(low_v_mins_maxs_minc, 9, 4, 3, low(minc)).
table(low_c_maxc_s_minc, 9, 5, 3, low(minc)).
table(low_c_maxc_mins_minc, 9, 5, 3, low(minc)).
table(low_c_maxc_maxs_minc, 9, 5, 3, low(minc)).
table(low_c_s_mins_minc, 9, 4, 3, low(minc)).
table(low_c_s_maxs_minc, 9, 6, 3, low(minc)).
table(low_c_mins_maxs_minc, 9, 4, 3, low(minc)).
table(low_maxc_s_mins_minc, 9, 5, 3, low(minc)).
table(low_maxc_s_maxs_minc, 9, 5, 3, low(minc)).
table(low_maxc_mins_maxs_minc, 9, 5, 3, low(minc)).
table(low_s_mins_maxs_minc, 9, 4, 3, low(minc)).
table(low_v_c_maxc_s_minc, 9, 7, 4, low(minc)).
table(low_v_c_maxc_mins_minc, 9, 7, 4, low(minc)).
table(low_v_c_maxc_maxs_minc, 9, 7, 4, low(minc)).
table(low_v_c_s_mins_minc, 9, 7, 4, low(minc)).
table(low_v_c_s_maxs_minc, 9, 7, 4, low(minc)).
table(low_v_c_mins_maxs_minc, 9, 7, 4, low(minc)).
table(low_v_maxc_s_mins_minc, 9, 6, 4, low(minc)).
table(low_v_maxc_s_maxs_minc, 9, 6, 4, low(minc)).
table(low_v_maxc_mins_maxs_minc, 9, 6, 4, low(minc)).
table(low_v_s_mins_maxs_minc, 9, 5, 4, low(minc)).
table(low_c_maxc_s_mins_minc, 9, 6, 4, low(minc)).
table(low_c_maxc_s_maxs_minc, 9, 6, 4, low(minc)).
table(low_c_maxc_mins_maxs_minc, 9, 6, 4, low(minc)).
table(low_c_s_mins_maxs_minc, 9, 6, 4, low(minc)).
table(low_maxc_s_mins_maxs_minc, 9, 6, 4, low(minc)).
table(low_v_c_maxc_s_mins_minc, 9, 8, 5, low(minc)).
table(low_v_c_maxc_s_maxs_minc, 9, 8, 5, low(minc)).
table(low_v_c_maxc_mins_maxs_minc, 9, 8, 5, low(minc)).
table(low_v_c_s_mins_maxs_minc, 9, 8, 5, low(minc)).
table(low_v_maxc_s_mins_maxs_minc, 9, 7, 5, low(minc)).
table(low_c_maxc_s_mins_maxs_minc, 9, 7, 5, low(minc)).
table(low_v_c_maxc_s_mins_maxs_minc, 9, 9, 6, low(minc)).
table(up_v_minc, 9, 11, 1, up(minc)).
table(up_maxc_minc, 9, 6, 1, up(minc)).
table(up_v_c_minc, 9, 5, 2, up(minc)).
table(up_v_maxc_minc, 9, 6, 2, up(minc)).
table(up_v_s_minc, 9, 13, 2, up(minc)).
table(up_v_mins_minc, 9, 12, 2, up(minc)).
table(up_v_maxs_minc, 9, 13, 2, up(minc)).
table(up_c_maxc_minc, 9, 4, 2, up(minc)).
table(up_maxc_s_minc, 9, 6, 2, up(minc)).
table(up_maxc_mins_minc, 9, 6, 2, up(minc)).
table(up_maxc_maxs_minc, 9, 7, 2, up(minc)).
table(up_s_maxs_minc, 9, 14, 2, up(minc)).
table(up_v_c_maxc_minc, 9, 6, 3, up(minc)).
table(up_v_c_s_minc, 9, 6, 3, up(minc)).
table(up_v_c_mins_minc, 9, 6, 3, up(minc)).
table(up_v_c_maxs_minc, 9, 6, 3, up(minc)).
table(up_v_maxc_s_minc, 9, 7, 3, up(minc)).
table(up_v_maxc_mins_minc, 9, 7, 3, up(minc)).
table(up_v_maxc_maxs_minc, 9, 7, 3, up(minc)).
table(up_v_s_mins_minc, 9, 14, 3, up(minc)).
table(up_v_s_maxs_minc, 9, 14, 3, up(minc)).
table(up_v_mins_maxs_minc, 9, 14, 3, up(minc)).
table(up_c_maxc_s_minc, 9, 5, 3, up(minc)).
table(up_c_maxc_mins_minc, 9, 5, 3, up(minc)).
table(up_c_maxc_maxs_minc, 9, 5, 3, up(minc)).
table(up_c_s_maxs_minc, 9, 4, 3, up(minc)).
table(up_maxc_s_mins_minc, 9, 7, 3, up(minc)).
table(up_maxc_s_maxs_minc, 9, 7, 3, up(minc)).
table(up_maxc_mins_maxs_minc, 9, 7, 3, up(minc)).
table(up_s_mins_maxs_minc, 9, 15, 3, up(minc)).
table(up_v_c_maxc_s_minc, 9, 7, 4, up(minc)).
table(up_v_c_maxc_mins_minc, 9, 7, 4, up(minc)).
table(up_v_c_maxc_maxs_minc, 9, 7, 4, up(minc)).
table(up_v_c_s_mins_minc, 9, 7, 4, up(minc)).
table(up_v_c_s_maxs_minc, 9, 7, 4, up(minc)).
table(up_v_c_mins_maxs_minc, 9, 7, 4, up(minc)).
table(up_v_maxc_s_mins_minc, 9, 8, 4, up(minc)).
table(up_v_maxc_s_maxs_minc, 9, 8, 4, up(minc)).
table(up_v_maxc_mins_maxs_minc, 9, 8, 4, up(minc)).
table(up_v_s_mins_maxs_minc, 9, 15, 4, up(minc)).
table(up_c_maxc_s_mins_minc, 9, 6, 4, up(minc)).
table(up_c_maxc_s_maxs_minc, 9, 6, 4, up(minc)).
table(up_c_maxc_mins_maxs_minc, 9, 6, 4, up(minc)).
table(up_c_s_mins_maxs_minc, 9, 5, 4, up(minc)).
table(up_maxc_s_mins_maxs_minc, 9, 8, 4, up(minc)).
table(up_v_c_maxc_s_mins_minc, 9, 8, 5, up(minc)).
table(up_v_c_maxc_s_maxs_minc, 9, 8, 5, up(minc)).
table(up_v_c_maxc_mins_maxs_minc, 9, 8, 5, up(minc)).
table(up_v_c_s_mins_maxs_minc, 9, 8, 5, up(minc)).
table(up_v_maxc_s_mins_maxs_minc, 9, 9, 5, up(minc)).
table(up_c_maxc_s_mins_maxs_minc, 9, 7, 5, up(minc)).
table(up_v_c_maxc_s_mins_maxs_minc, 9, 9, 6, up(minc)).
table(low_v_maxc, 9, 18, 1, low(maxc)).
table(low_c_maxc, 9, 18, 1, low(maxc)).
table(low_minc_maxc, 9, 6, 1, low(maxc)).
table(low_s_maxc, 9, 18, 1, low(maxc)).
table(low_mins_maxc, 9, 9, 1, low(maxc)).
table(low_maxs_maxc, 9, 2, 1, low(maxc)).
table(low_v_c_maxc, 9, 3, 2, low(maxc)).
table(low_v_minc_maxc, 9, 11, 2, low(maxc)).
table(low_v_s_maxc, 9, 4, 2, low(maxc)).
table(low_v_mins_maxc, 9, 18, 2, low(maxc)).
table(low_v_maxs_maxc, 9, 3, 2, low(maxc)).
table(low_c_minc_maxc, 9, 11, 2, low(maxc)).
table(low_c_s_maxc, 9, 4, 2, low(maxc)).
table(low_c_mins_maxc, 9, 18, 2, low(maxc)).
table(low_c_maxs_maxc, 9, 3, 2, low(maxc)).
table(low_minc_s_maxc, 9, 7, 2, low(maxc)).
table(low_minc_mins_maxc, 9, 7, 2, low(maxc)).
table(low_minc_maxs_maxc, 9, 4, 2, low(maxc)).
table(low_s_mins_maxc, 9, 18, 2, low(maxc)).
table(low_s_maxs_maxc, 9, 3, 2, low(maxc)).
table(low_mins_maxs_maxc, 9, 3, 2, low(maxc)).
table(low_v_c_minc_maxc, 9, 6, 3, low(maxc)).
table(low_v_c_s_maxc, 9, 4, 3, low(maxc)).
table(low_v_c_mins_maxc, 9, 4, 3, low(maxc)).
table(low_v_c_maxs_maxc, 9, 4, 3, low(maxc)).
table(low_v_minc_s_maxc, 9, 5, 3, low(maxc)).
table(low_v_minc_mins_maxc, 9, 12, 3, low(maxc)).
table(low_v_minc_maxs_maxc, 9, 5, 3, low(maxc)).
table(low_v_s_mins_maxc, 9, 5, 3, low(maxc)).
table(low_v_s_maxs_maxc, 9, 4, 3, low(maxc)).
table(low_v_mins_maxs_maxc, 9, 4, 3, low(maxc)).
table(low_c_minc_s_maxc, 9, 5, 3, low(maxc)).
table(low_c_minc_mins_maxc, 9, 12, 3, low(maxc)).
table(low_c_minc_maxs_maxc, 9, 5, 3, low(maxc)).
table(low_c_s_mins_maxc, 9, 4, 3, low(maxc)).
table(low_c_s_maxs_maxc, 9, 4, 3, low(maxc)).
table(low_c_mins_maxs_maxc, 9, 4, 3, low(maxc)).
table(low_minc_s_mins_maxc, 9, 8, 3, low(maxc)).
table(low_minc_s_maxs_maxc, 9, 5, 3, low(maxc)).
table(low_minc_mins_maxs_maxc, 9, 5, 3, low(maxc)).
table(low_s_mins_maxs_maxc, 9, 4, 3, low(maxc)).
table(low_v_c_minc_s_maxc, 9, 7, 4, low(maxc)).
table(low_v_c_minc_mins_maxc, 9, 7, 4, low(maxc)).
table(low_v_c_minc_maxs_maxc, 9, 7, 4, low(maxc)).
table(low_v_c_s_mins_maxc, 9, 5, 4, low(maxc)).
table(low_v_c_s_maxs_maxc, 9, 5, 4, low(maxc)).
table(low_v_c_mins_maxs_maxc, 9, 5, 4, low(maxc)).
table(low_v_minc_s_mins_maxc, 9, 6, 4, low(maxc)).
table(low_v_minc_s_maxs_maxc, 9, 6, 4, low(maxc)).
table(low_v_minc_mins_maxs_maxc, 9, 6, 4, low(maxc)).
table(low_v_s_mins_maxs_maxc, 9, 5, 4, low(maxc)).
table(low_c_minc_s_mins_maxc, 9, 6, 4, low(maxc)).
table(low_c_minc_s_maxs_maxc, 9, 6, 4, low(maxc)).
table(low_c_minc_mins_maxs_maxc, 9, 6, 4, low(maxc)).
table(low_c_s_mins_maxs_maxc, 9, 5, 4, low(maxc)).
table(low_minc_s_mins_maxs_maxc, 9, 6, 4, low(maxc)).
table(low_v_c_minc_s_mins_maxc, 9, 8, 5, low(maxc)).
table(low_v_c_minc_s_maxs_maxc, 9, 8, 5, low(maxc)).
table(low_v_c_minc_mins_maxs_maxc, 9, 8, 5, low(maxc)).
table(low_v_c_s_mins_maxs_maxc, 9, 6, 5, low(maxc)).
table(low_v_minc_s_mins_maxs_maxc, 9, 7, 5, low(maxc)).
table(low_c_minc_s_mins_maxs_maxc, 9, 7, 5, low(maxc)).
table(low_v_c_minc_s_mins_maxs_maxc, 9, 9, 6, low(maxc)).
table(up_v_maxc, 9, 11, 1, up(maxc)).
table(up_v_c_maxc, 9, 11, 2, up(maxc)).
table(up_v_minc_maxc, 9, 11, 2, up(maxc)).
table(up_v_s_maxc, 9, 13, 2, up(maxc)).
table(up_v_mins_maxc, 9, 12, 2, up(maxc)).
table(up_v_maxs_maxc, 9, 13, 2, up(maxc)).
table(up_v_c_minc_maxc, 9, 11, 3, up(maxc)).
table(up_v_c_s_maxc, 9, 13, 3, up(maxc)).
table(up_v_c_mins_maxc, 9, 12, 3, up(maxc)).
table(up_v_c_maxs_maxc, 9, 13, 3, up(maxc)).
table(up_v_minc_s_maxc, 9, 12, 3, up(maxc)).
table(up_v_minc_mins_maxc, 9, 12, 3, up(maxc)).
table(up_v_minc_maxs_maxc, 9, 12, 3, up(maxc)).
table(up_v_s_mins_maxc, 9, 14, 3, up(maxc)).
table(up_v_s_maxs_maxc, 9, 14, 3, up(maxc)).
table(up_v_mins_maxs_maxc, 9, 14, 3, up(maxc)).
table(up_v_c_minc_s_maxc, 9, 12, 4, up(maxc)).
table(up_v_c_minc_mins_maxc, 9, 12, 4, up(maxc)).
table(up_v_c_minc_maxs_maxc, 9, 12, 4, up(maxc)).
table(up_v_c_s_mins_maxc, 9, 14, 4, up(maxc)).
table(up_v_c_s_maxs_maxc, 9, 6, 4, up(maxc)).
table(up_v_c_mins_maxs_maxc, 9, 14, 4, up(maxc)).
table(up_v_minc_s_mins_maxc, 9, 13, 4, up(maxc)).
table(up_v_minc_s_maxs_maxc, 9, 13, 4, up(maxc)).
table(up_v_minc_mins_maxs_maxc, 9, 13, 4, up(maxc)).
table(up_v_s_mins_maxs_maxc, 9, 15, 4, up(maxc)).
table(up_v_c_minc_s_mins_maxc, 9, 13, 5, up(maxc)).
table(up_v_c_minc_s_maxs_maxc, 9, 8, 5, up(maxc)).
table(up_v_c_minc_mins_maxs_maxc, 9, 13, 5, up(maxc)).
table(up_v_c_s_mins_maxs_maxc, 9, 10, 5, up(maxc)).
table(up_v_minc_s_mins_maxs_maxc, 9, 14, 5, up(maxc)).
table(up_v_c_minc_s_mins_maxs_maxc, 9, 9, 6, up(maxc)).
table(low_v_s, 9, 18, 1, low(s)).
table(low_c_s, 9, 3, 1, low(s)).
table(low_minc_s, 9, 18, 1, low(s)).
table(low_maxc_s, 9, 18, 1, low(s)).
table(low_mins_s, 9, 18, 1, low(s)).
table(low_maxs_s, 9, 18, 1, low(s)).
table(low_v_c_s, 9, 4, 2, low(s)).
table(low_v_minc_s, 9, 18, 2, low(s)).
table(low_v_maxc_s, 9, 6, 2, low(s)).
table(low_v_mins_s, 9, 9, 2, low(s)).
table(low_v_maxs_s, 9, 3, 2, low(s)).
table(low_c_minc_s, 9, 5, 2, low(s)).
table(low_c_maxc_s, 9, 5, 2, low(s)).
table(low_c_mins_s, 9, 5, 2, low(s)).
table(low_c_maxs_s, 9, 5, 2, low(s)).
table(low_minc_maxc_s, 9, 18, 2, low(s)).
table(low_minc_mins_s, 9, 12, 2, low(s)).
table(low_minc_maxs_s, 9, 14, 2, low(s)).
table(low_maxc_mins_s, 9, 7, 2, low(s)).
table(low_maxc_maxs_s, 9, 14, 2, low(s)).
table(low_mins_maxs_s, 9, 12, 2, low(s)).
table(low_v_c_minc_s, 9, 8, 3, low(s)).
table(low_v_c_maxc_s, 9, 6, 3, low(s)).
table(low_v_c_mins_s, 9, 8, 3, low(s)).
table(low_v_c_maxs_s, 9, 4, 3, low(s)).
table(low_v_minc_maxc_s, 9, 10, 3, low(s)).
table(low_v_minc_mins_s, 9, 13, 3, low(s)).
table(low_v_minc_maxs_s, 9, 4, 3, low(s)).
table(low_v_maxc_mins_s, 9, 4, 3, low(s)).
table(low_v_maxc_maxs_s, 9, 4, 3, low(s)).
table(low_v_mins_maxs_s, 9, 4, 3, low(s)).
table(low_c_minc_maxc_s, 9, 8, 3, low(s)).
table(low_c_minc_mins_s, 9, 4, 3, low(s)).
table(low_c_minc_maxs_s, 9, 5, 3, low(s)).
table(low_c_maxc_mins_s, 9, 8, 3, low(s)).
table(low_c_maxc_maxs_s, 9, 5, 3, low(s)).
table(low_c_mins_maxs_s, 9, 8, 3, low(s)).
table(low_minc_maxc_mins_s, 9, 13, 3, low(s)).
table(low_minc_maxc_maxs_s, 9, 14, 3, low(s)).
table(low_minc_mins_maxs_s, 9, 8, 3, low(s)).
table(low_maxc_mins_maxs_s, 9, 10, 3, low(s)).
table(low_v_c_minc_maxc_s, 9, 10, 4, low(s)).
table(low_v_c_minc_mins_s, 9, 7, 4, low(s)).
table(low_v_c_minc_maxs_s, 9, 7, 4, low(s)).
table(low_v_c_maxc_mins_s, 9, 8, 4, low(s)).
table(low_v_c_maxc_maxs_s, 9, 5, 4, low(s)).
table(low_v_c_mins_maxs_s, 9, 5, 4, low(s)).
table(low_v_minc_maxc_mins_s, 9, 8, 4, low(s)).
table(low_v_minc_maxc_maxs_s, 9, 6, 4, low(s)).
table(low_v_minc_mins_maxs_s, 9, 5, 4, low(s)).
table(low_v_maxc_mins_maxs_s, 9, 5, 4, low(s)).
table(low_c_minc_maxc_mins_s, 9, 6, 4, low(s)).
table(low_c_minc_maxc_maxs_s, 9, 7, 4, low(s)).
table(low_c_minc_mins_maxs_s, 9, 5, 4, low(s)).
table(low_c_maxc_mins_maxs_s, 9, 6, 4, low(s)).
table(low_minc_maxc_mins_maxs_s, 9, 14, 4, low(s)).
table(low_v_c_minc_maxc_mins_s, 9, 8, 5, low(s)).
table(low_v_c_minc_maxc_maxs_s, 9, 8, 5, low(s)).
table(low_v_c_minc_mins_maxs_s, 9, 8, 5, low(s)).
table(low_v_c_maxc_mins_maxs_s, 9, 6, 5, low(s)).
table(low_v_minc_maxc_mins_maxs_s, 9, 7, 5, low(s)).
table(low_c_minc_maxc_mins_maxs_s, 9, 7, 5, low(s)).
table(low_v_c_minc_maxc_mins_maxs_s, 9, 9, 6, low(s)).
table(up_v_s, 9, 7, 1, up(s)).
table(up_v_c_s, 9, 8, 2, up(s)).
table(up_v_minc_s, 9, 8, 2, up(s)).
table(up_v_maxc_s, 9, 8, 2, up(s)).
table(up_v_mins_s, 9, 7, 2, up(s)).
table(up_v_maxs_s, 9, 7, 2, up(s)).
table(up_c_maxc_s, 9, 9, 2, up(s)).
table(up_v_c_minc_s, 9, 11, 3, up(s)).
table(up_v_c_maxc_s, 9, 9, 3, up(s)).
table(up_v_c_mins_s, 9, 8, 3, up(s)).
table(up_v_c_maxs_s, 9, 8, 3, up(s)).
table(up_v_minc_maxc_s, 9, 10, 3, up(s)).
table(up_v_minc_mins_s, 9, 8, 3, up(s)).
table(up_v_minc_maxs_s, 9, 8, 3, up(s)).
table(up_v_maxc_mins_s, 9, 8, 3, up(s)).
table(up_v_maxc_maxs_s, 9, 8, 3, up(s)).
table(up_v_mins_maxs_s, 9, 7, 3, up(s)).
table(up_c_minc_maxc_s, 9, 12, 3, up(s)).
table(up_c_maxc_mins_s, 9, 5, 3, up(s)).
table(up_c_maxc_maxs_s, 9, 9, 3, up(s)).
table(up_v_c_minc_maxc_s, 9, 12, 4, up(s)).
table(up_v_c_minc_mins_s, 9, 11, 4, up(s)).
table(up_v_c_minc_maxs_s, 9, 11, 4, up(s)).
table(up_v_c_maxc_mins_s, 9, 9, 4, up(s)).
table(up_v_c_maxc_maxs_s, 9, 9, 4, up(s)).
table(up_v_c_mins_maxs_s, 9, 8, 4, up(s)).
table(up_v_minc_maxc_mins_s, 9, 10, 4, up(s)).
table(up_v_minc_maxc_maxs_s, 9, 10, 4, up(s)).
table(up_v_minc_mins_maxs_s, 9, 8, 4, up(s)).
table(up_v_maxc_mins_maxs_s, 9, 8, 4, up(s)).
table(up_c_minc_maxc_mins_s, 9, 7, 4, up(s)).
table(up_c_minc_maxc_maxs_s, 9, 12, 4, up(s)).
table(up_c_maxc_mins_maxs_s, 9, 6, 4, up(s)).
table(up_v_c_minc_maxc_mins_s, 9, 12, 5, up(s)).
table(up_v_c_minc_maxc_maxs_s, 9, 12, 5, up(s)).
table(up_v_c_minc_mins_maxs_s, 9, 11, 5, up(s)).
table(up_v_c_maxc_mins_maxs_s, 9, 9, 5, up(s)).
table(up_v_minc_maxc_mins_maxs_s, 9, 10, 5, up(s)).
table(up_c_minc_maxc_mins_maxs_s, 9, 8, 5, up(s)).
table(up_v_c_minc_maxc_mins_maxs_s, 9, 12, 6, up(s)).
table(low_v_mins, 9, 2, 1, low(mins)).
table(low_c_mins, 9, 2, 1, low(mins)).
table(low_minc_mins, 9, 2, 1, low(mins)).
table(low_maxc_mins, 9, 2, 1, low(mins)).
table(low_s_mins, 9, 2, 1, low(mins)).
table(low_maxs_mins, 9, 2, 1, low(mins)).
table(low_v_c_mins, 9, 3, 2, low(mins)).
table(low_v_minc_mins, 9, 3, 2, low(mins)).
table(low_v_maxc_mins, 9, 3, 2, low(mins)).
table(low_v_s_mins, 9, 3, 2, low(mins)).
table(low_v_maxs_mins, 9, 3, 2, low(mins)).
table(low_c_minc_mins, 9, 3, 2, low(mins)).
table(low_c_maxc_mins, 9, 3, 2, low(mins)).
table(low_c_s_mins, 9, 3, 2, low(mins)).
table(low_c_maxs_mins, 9, 3, 2, low(mins)).
table(low_minc_maxc_mins, 9, 4, 2, low(mins)).
table(low_minc_s_mins, 9, 3, 2, low(mins)).
table(low_minc_maxs_mins, 9, 3, 2, low(mins)).
table(low_maxc_s_mins, 9, 3, 2, low(mins)).
table(low_maxc_maxs_mins, 9, 3, 2, low(mins)).
table(low_s_maxs_mins, 9, 3, 2, low(mins)).
table(low_v_c_minc_mins, 9, 6, 3, low(mins)).
table(low_v_c_maxc_mins, 9, 4, 3, low(mins)).
table(low_v_c_s_mins, 9, 4, 3, low(mins)).
table(low_v_c_maxs_mins, 9, 4, 3, low(mins)).
table(low_v_minc_maxc_mins, 9, 5, 3, low(mins)).
table(low_v_minc_s_mins, 9, 4, 3, low(mins)).
table(low_v_minc_maxs_mins, 9, 4, 3, low(mins)).
table(low_v_maxc_s_mins, 9, 4, 3, low(mins)).
table(low_v_maxc_maxs_mins, 9, 4, 3, low(mins)).
table(low_v_s_maxs_mins, 9, 4, 3, low(mins)).
table(low_c_minc_maxc_mins, 9, 5, 3, low(mins)).
table(low_c_minc_s_mins, 9, 4, 3, low(mins)).
table(low_c_minc_maxs_mins, 9, 4, 3, low(mins)).
table(low_c_maxc_s_mins, 9, 4, 3, low(mins)).
table(low_c_maxc_maxs_mins, 9, 4, 3, low(mins)).
table(low_c_s_maxs_mins, 9, 4, 3, low(mins)).
table(low_minc_maxc_s_mins, 9, 5, 3, low(mins)).
table(low_minc_maxc_maxs_mins, 9, 5, 3, low(mins)).
table(low_minc_s_maxs_mins, 9, 4, 3, low(mins)).
table(low_maxc_s_maxs_mins, 9, 4, 3, low(mins)).
table(low_v_c_minc_maxc_mins, 9, 7, 4, low(mins)).
table(low_v_c_minc_s_mins, 9, 7, 4, low(mins)).
table(low_v_c_minc_maxs_mins, 9, 7, 4, low(mins)).
table(low_v_c_maxc_s_mins, 9, 5, 4, low(mins)).
table(low_v_c_maxc_maxs_mins, 9, 5, 4, low(mins)).
table(low_v_c_s_maxs_mins, 9, 5, 4, low(mins)).
table(low_v_minc_maxc_s_mins, 9, 6, 4, low(mins)).
table(low_v_minc_maxc_maxs_mins, 9, 6, 4, low(mins)).
table(low_v_minc_s_maxs_mins, 9, 5, 4, low(mins)).
table(low_v_maxc_s_maxs_mins, 9, 5, 4, low(mins)).
table(low_c_minc_maxc_s_mins, 9, 6, 4, low(mins)).
table(low_c_minc_maxc_maxs_mins, 9, 6, 4, low(mins)).
table(low_c_minc_s_maxs_mins, 9, 5, 4, low(mins)).
table(low_c_maxc_s_maxs_mins, 9, 5, 4, low(mins)).
table(low_minc_maxc_s_maxs_mins, 9, 6, 4, low(mins)).
table(low_v_c_minc_maxc_s_mins, 9, 8, 5, low(mins)).
table(low_v_c_minc_maxc_maxs_mins, 9, 8, 5, low(mins)).
table(low_v_c_minc_s_maxs_mins, 9, 8, 5, low(mins)).
table(low_v_c_maxc_s_maxs_mins, 9, 6, 5, low(mins)).
table(low_v_minc_maxc_s_maxs_mins, 9, 7, 5, low(mins)).
table(low_c_minc_maxc_s_maxs_mins, 9, 7, 5, low(mins)).
table(low_v_c_minc_maxc_s_maxs_mins, 9, 9, 6, low(mins)).
table(up_v_mins, 9, 18, 1, up(mins)).
table(up_maxc_mins, 9, 9, 1, up(mins)).
table(up_maxs_mins, 9, 2, 1, up(mins)).
table(up_v_c_mins, 9, 6, 2, up(mins)).
table(up_v_minc_mins, 9, 3, 2, up(mins)).
table(up_v_maxc_mins, 9, 7, 2, up(mins)).
table(up_v_s_mins, 9, 3, 2, up(mins)).
table(up_v_maxs_mins, 9, 3, 2, up(mins)).
table(up_c_maxc_mins, 9, 5, 2, up(mins)).
table(up_c_maxs_mins, 9, 3, 2, up(mins)).
table(up_minc_maxc_mins, 9, 4, 2, up(mins)).
table(up_minc_maxs_mins, 9, 3, 2, up(mins)).
table(up_maxc_s_mins, 9, 3, 2, up(mins)).
table(up_maxc_maxs_mins, 9, 3, 2, up(mins)).
table(up_s_maxs_mins, 9, 3, 2, up(mins)).
table(up_v_c_minc_mins, 9, 6, 3, up(mins)).
table(up_v_c_maxc_mins, 9, 7, 3, up(mins)).
table(up_v_c_s_mins, 9, 4, 3, up(mins)).
table(up_v_c_maxs_mins, 9, 4, 3, up(mins)).
table(up_v_minc_maxc_mins, 9, 5, 3, up(mins)).
table(up_v_minc_s_mins, 9, 4, 3, up(mins)).
table(up_v_minc_maxs_mins, 9, 4, 3, up(mins)).
table(up_v_maxc_s_mins, 9, 4, 3, up(mins)).
table(up_v_maxc_maxs_mins, 9, 4, 3, up(mins)).
table(up_v_s_maxs_mins, 9, 4, 3, up(mins)).
table(up_c_minc_maxc_mins, 9, 5, 3, up(mins)).
table(up_c_minc_maxs_mins, 9, 4, 3, up(mins)).
table(up_c_maxc_s_mins, 9, 4, 3, up(mins)).
table(up_c_maxc_maxs_mins, 9, 4, 3, up(mins)).
table(up_c_s_maxs_mins, 9, 4, 3, up(mins)).
table(up_minc_maxc_s_mins, 9, 5, 3, up(mins)).
table(up_minc_maxc_maxs_mins, 9, 5, 3, up(mins)).
table(up_minc_s_maxs_mins, 9, 4, 3, up(mins)).
table(up_maxc_s_maxs_mins, 9, 4, 3, up(mins)).
table(up_v_c_minc_maxc_mins, 9, 7, 4, up(mins)).
table(up_v_c_minc_s_mins, 9, 7, 4, up(mins)).
table(up_v_c_minc_maxs_mins, 9, 7, 4, up(mins)).
table(up_v_c_maxc_s_mins, 9, 5, 4, up(mins)).
table(up_v_c_maxc_maxs_mins, 9, 5, 4, up(mins)).
table(up_v_c_s_maxs_mins, 9, 5, 4, up(mins)).
table(up_v_minc_maxc_s_mins, 9, 6, 4, up(mins)).
table(up_v_minc_maxc_maxs_mins, 9, 6, 4, up(mins)).
table(up_v_minc_s_maxs_mins, 9, 5, 4, up(mins)).
table(up_v_maxc_s_maxs_mins, 9, 5, 4, up(mins)).
table(up_c_minc_maxc_s_mins, 9, 6, 4, up(mins)).
table(up_c_minc_maxc_maxs_mins, 9, 6, 4, up(mins)).
table(up_c_minc_s_maxs_mins, 9, 5, 4, up(mins)).
table(up_c_maxc_s_maxs_mins, 9, 5, 4, up(mins)).
table(up_minc_maxc_s_maxs_mins, 9, 6, 4, up(mins)).
table(up_v_c_minc_maxc_s_mins, 9, 8, 5, up(mins)).
table(up_v_c_minc_maxc_maxs_mins, 9, 8, 5, up(mins)).
table(up_v_c_minc_s_maxs_mins, 9, 8, 5, up(mins)).
table(up_v_c_maxc_s_maxs_mins, 9, 6, 5, up(mins)).
table(up_v_minc_maxc_s_maxs_mins, 9, 7, 5, up(mins)).
table(up_c_minc_maxc_s_maxs_mins, 9, 7, 5, up(mins)).
table(up_v_c_minc_maxc_s_maxs_mins, 9, 9, 6, up(mins)).
table(low_v_maxs, 9, 7, 1, low(maxs)).
table(low_c_maxs, 9, 4, 1, low(maxs)).
table(low_minc_maxs, 9, 4, 1, low(maxs)).
table(low_maxc_maxs, 9, 4, 1, low(maxs)).
table(low_s_maxs, 9, 7, 1, low(maxs)).
table(low_mins_maxs, 9, 2, 1, low(maxs)).
table(low_v_c_maxs, 9, 8, 2, low(maxs)).
table(low_v_minc_maxs, 9, 8, 2, low(maxs)).
table(low_v_maxc_maxs, 9, 8, 2, low(maxs)).
table(low_v_s_maxs, 9, 3, 2, low(maxs)).
table(low_v_mins_maxs, 9, 7, 2, low(maxs)).
table(low_c_minc_maxs, 9, 5, 2, low(maxs)).
table(low_c_maxc_maxs, 9, 5, 2, low(maxs)).
table(low_c_s_maxs, 9, 8, 2, low(maxs)).
table(low_c_mins_maxs, 9, 3, 2, low(maxs)).
table(low_minc_maxc_maxs, 9, 6, 2, low(maxs)).
table(low_minc_s_maxs, 9, 3, 2, low(maxs)).
table(low_minc_mins_maxs, 9, 3, 2, low(maxs)).
table(low_maxc_s_maxs, 9, 3, 2, low(maxs)).
table(low_maxc_mins_maxs, 9, 3, 2, low(maxs)).
table(low_s_mins_maxs, 9, 7, 2, low(maxs)).
table(low_v_c_minc_maxs, 9, 11, 3, low(maxs)).
table(low_v_c_maxc_maxs, 9, 9, 3, low(maxs)).
table(low_v_c_s_maxs, 9, 4, 3, low(maxs)).
table(low_v_c_mins_maxs, 9, 8, 3, low(maxs)).
table(low_v_minc_maxc_maxs, 9, 10, 3, low(maxs)).
table(low_v_minc_s_maxs, 9, 4, 3, low(maxs)).
table(low_v_minc_mins_maxs, 9, 8, 3, low(maxs)).
table(low_v_maxc_s_maxs, 9, 4, 3, low(maxs)).
table(low_v_maxc_mins_maxs, 9, 8, 3, low(maxs)).
table(low_v_s_mins_maxs, 9, 4, 3, low(maxs)).
table(low_c_minc_maxc_maxs, 9, 7, 3, low(maxs)).
table(low_c_minc_s_maxs, 9, 4, 3, low(maxs)).
table(low_c_minc_mins_maxs, 9, 4, 3, low(maxs)).
table(low_c_maxc_s_maxs, 9, 4, 3, low(maxs)).
table(low_c_maxc_mins_maxs, 9, 4, 3, low(maxs)).
table(low_c_s_mins_maxs, 9, 8, 3, low(maxs)).
table(low_minc_maxc_s_maxs, 9, 5, 3, low(maxs)).
table(low_minc_maxc_mins_maxs, 9, 5, 3, low(maxs)).
table(low_minc_s_mins_maxs, 9, 4, 3, low(maxs)).
table(low_maxc_s_mins_maxs, 9, 4, 3, low(maxs)).
table(low_v_c_minc_maxc_maxs, 9, 12, 4, low(maxs)).
table(low_v_c_minc_s_maxs, 9, 7, 4, low(maxs)).
table(low_v_c_minc_mins_maxs, 9, 11, 4, low(maxs)).
table(low_v_c_maxc_s_maxs, 9, 5, 4, low(maxs)).
table(low_v_c_maxc_mins_maxs, 9, 9, 4, low(maxs)).
table(low_v_c_s_mins_maxs, 9, 5, 4, low(maxs)).
table(low_v_minc_maxc_s_maxs, 9, 6, 4, low(maxs)).
table(low_v_minc_maxc_mins_maxs, 9, 10, 4, low(maxs)).
table(low_v_minc_s_mins_maxs, 9, 5, 4, low(maxs)).
table(low_v_maxc_s_mins_maxs, 9, 5, 4, low(maxs)).
table(low_c_minc_maxc_s_maxs, 9, 6, 4, low(maxs)).
table(low_c_minc_maxc_mins_maxs, 9, 7, 4, low(maxs)).
table(low_c_minc_s_mins_maxs, 9, 5, 4, low(maxs)).
table(low_c_maxc_s_mins_maxs, 9, 5, 4, low(maxs)).
table(low_minc_maxc_s_mins_maxs, 9, 6, 4, low(maxs)).
table(low_v_c_minc_maxc_s_maxs, 9, 8, 5, low(maxs)).
table(low_v_c_minc_maxc_mins_maxs, 9, 12, 5, low(maxs)).
table(low_v_c_minc_s_mins_maxs, 9, 8, 5, low(maxs)).
table(low_v_c_maxc_s_mins_maxs, 9, 6, 5, low(maxs)).
table(low_v_minc_maxc_s_mins_maxs, 9, 7, 5, low(maxs)).
table(low_c_minc_maxc_s_mins_maxs, 9, 7, 5, low(maxs)).
table(low_v_c_minc_maxc_s_mins_maxs, 9, 9, 6, low(maxs)).
table(up_v_maxs, 9, 18, 1, up(maxs)).
table(up_maxc_maxs, 9, 2, 1, up(maxs)).
table(up_v_c_maxs, 9, 18, 2, up(maxs)).
table(up_v_minc_maxs, 9, 12, 2, up(maxs)).
table(up_v_maxc_maxs, 9, 3, 2, up(maxs)).
table(up_v_s_maxs, 9, 7, 2, up(maxs)).
table(up_v_mins_maxs, 9, 9, 2, up(maxs)).
table(up_c_maxc_maxs, 9, 3, 2, up(maxs)).
table(up_minc_maxc_maxs, 9, 4, 2, up(maxs)).
table(up_maxc_s_maxs, 9, 3, 2, up(maxs)).
table(up_maxc_mins_maxs, 9, 3, 2, up(maxs)).
table(up_v_c_minc_maxs, 9, 12, 3, up(maxs)).
table(up_v_c_maxc_maxs, 9, 4, 3, up(maxs)).
table(up_v_c_s_maxs, 9, 8, 3, up(maxs)).
table(up_v_c_mins_maxs, 9, 18, 3, up(maxs)).
table(up_v_minc_maxc_maxs, 9, 5, 3, up(maxs)).
table(up_v_minc_s_maxs, 9, 5, 3, up(maxs)).
table(up_v_minc_mins_maxs, 9, 14, 3, up(maxs)).
table(up_v_maxc_s_maxs, 9, 4, 3, up(maxs)).
table(up_v_maxc_mins_maxs, 9, 4, 3, up(maxs)).
table(up_v_s_mins_maxs, 9, 7, 3, up(maxs)).
table(up_c_minc_maxc_maxs, 9, 5, 3, up(maxs)).
table(up_c_maxc_s_maxs, 9, 4, 3, up(maxs)).
table(up_c_maxc_mins_maxs, 9, 4, 3, up(maxs)).
table(up_minc_maxc_s_maxs, 9, 5, 3, up(maxs)).
table(up_minc_maxc_mins_maxs, 9, 5, 3, up(maxs)).
table(up_maxc_s_mins_maxs, 9, 4, 3, up(maxs)).
table(up_v_c_minc_maxc_maxs, 9, 7, 4, up(maxs)).
table(up_v_c_minc_s_maxs, 9, 7, 4, up(maxs)).
table(up_v_c_minc_mins_maxs, 9, 13, 4, up(maxs)).
table(up_v_c_maxc_s_maxs, 9, 5, 4, up(maxs)).
table(up_v_c_maxc_mins_maxs, 9, 5, 4, up(maxs)).
table(up_v_c_s_mins_maxs, 9, 8, 4, up(maxs)).
table(up_v_minc_maxc_s_maxs, 9, 6, 4, up(maxs)).
table(up_v_minc_maxc_mins_maxs, 9, 6, 4, up(maxs)).
table(up_v_minc_s_mins_maxs, 9, 8, 4, up(maxs)).
table(up_v_maxc_s_mins_maxs, 9, 5, 4, up(maxs)).
table(up_c_minc_maxc_s_maxs, 9, 6, 4, up(maxs)).
table(up_c_minc_maxc_mins_maxs, 9, 6, 4, up(maxs)).
table(up_c_maxc_s_mins_maxs, 9, 5, 4, up(maxs)).
table(up_minc_maxc_s_mins_maxs, 9, 6, 4, up(maxs)).
table(up_v_c_minc_maxc_s_maxs, 9, 8, 5, up(maxs)).
table(up_v_c_minc_maxc_mins_maxs, 9, 8, 5, up(maxs)).
table(up_v_c_minc_s_mins_maxs, 9, 8, 5, up(maxs)).
table(up_v_c_maxc_s_mins_maxs, 9, 6, 5, up(maxs)).
table(up_v_minc_maxc_s_mins_maxs, 9, 7, 5, up(maxs)).
table(up_c_minc_maxc_s_mins_maxs, 9, 7, 5, up(maxs)).
table(up_v_c_minc_maxc_s_mins_maxs, 9, 9, 6, up(maxs)).
table(low_v_mina, 9, 12, 1, low(mina)).
table(low_c_mina, 9, 7, 1, low(mina)).
table(low_minc_mina, 9, 19, 1, low(mina)).
table(low_maxc_mina, 9, 19, 1, low(mina)).
table(low_s_mina, 9, 12, 1, low(mina)).
table(low_mins_mina, 9, 11, 1, low(mina)).
table(low_maxs_mina, 9, 11, 1, low(mina)).
table(low_v_c_mina, 9, 10, 2, low(mina)).
table(low_v_minc_mina, 9, 12, 2, low(mina)).
table(low_v_maxc_mina, 9, 11, 2, low(mina)).
table(low_v_s_mina, 9, 4, 2, low(mina)).
table(low_v_mins_mina, 9, 4, 2, low(mina)).
table(low_v_maxs_mina, 9, 10, 2, low(mina)).
table(low_c_minc_mina, 9, 8, 2, low(mina)).
table(low_c_maxc_mina, 9, 6, 2, low(mina)).
table(low_c_s_mina, 9, 10, 2, low(mina)).
table(low_c_mins_mina, 9, 7, 2, low(mina)).
table(low_c_maxs_mina, 9, 5, 2, low(mina)).
table(low_minc_maxc_mina, 9, 19, 2, low(mina)).
table(low_minc_s_mina, 9, 12, 2, low(mina)).
table(low_minc_mins_mina, 9, 19, 2, low(mina)).
table(low_minc_maxs_mina, 9, 19, 2, low(mina)).
table(low_maxc_s_mina, 9, 11, 2, low(mina)).
table(low_maxc_mins_mina, 9, 19, 2, low(mina)).
table(low_maxc_maxs_mina, 9, 19, 2, low(mina)).
table(low_s_mins_mina, 9, 12, 2, low(mina)).
table(low_s_maxs_mina, 9, 10, 2, low(mina)).
table(low_mins_maxs_mina, 9, 6, 2, low(mina)).
table(low_v_c_minc_mina, 9, 13, 3, low(mina)).
table(low_v_c_maxc_mina, 9, 19, 3, low(mina)).
table(low_v_c_s_mina, 9, 4, 3, low(mina)).
table(low_v_c_mins_mina, 9, 7, 3, low(mina)).
table(low_v_c_maxs_mina, 9, 10, 3, low(mina)).
table(low_v_minc_maxc_mina, 9, 13, 3, low(mina)).
table(low_v_minc_s_mina, 9, 4, 3, low(mina)).
table(low_v_minc_mins_mina, 9, 5, 3, low(mina)).
table(low_v_minc_maxs_mina, 9, 11, 3, low(mina)).
table(low_v_maxc_s_mina, 9, 4, 3, low(mina)).
table(low_v_maxc_mins_mina, 9, 11, 3, low(mina)).
table(low_v_maxc_maxs_mina, 9, 10, 3, low(mina)).
table(low_v_s_mins_mina, 9, 5, 3, low(mina)).
table(low_v_s_maxs_mina, 9, 5, 3, low(mina)).
table(low_v_mins_maxs_mina, 9, 10, 3, low(mina)).
table(low_c_minc_maxc_mina, 9, 8, 3, low(mina)).
table(low_c_minc_s_mina, 9, 11, 3, low(mina)).
table(low_c_minc_mins_mina, 9, 8, 3, low(mina)).
table(low_c_minc_maxs_mina, 9, 6, 3, low(mina)).
table(low_c_maxc_s_mina, 9, 9, 3, low(mina)).
table(low_c_maxc_mins_mina, 9, 6, 3, low(mina)).
table(low_c_maxc_maxs_mina, 9, 6, 3, low(mina)).
table(low_c_s_mins_mina, 9, 10, 3, low(mina)).
table(low_c_s_maxs_mina, 9, 10, 3, low(mina)).
table(low_c_mins_maxs_mina, 9, 5, 3, low(mina)).
table(low_minc_maxc_s_mina, 9, 10, 3, low(mina)).
table(low_minc_maxc_mins_mina, 9, 17, 3, low(mina)).
table(low_minc_maxc_maxs_mina, 9, 18, 3, low(mina)).
table(low_minc_s_mins_mina, 9, 9, 3, low(mina)).
table(low_minc_s_maxs_mina, 9, 11, 3, low(mina)).
table(low_minc_mins_maxs_mina, 9, 19, 3, low(mina)).
table(low_maxc_s_mins_mina, 9, 8, 3, low(mina)).
table(low_maxc_s_maxs_mina, 9, 10, 3, low(mina)).
table(low_maxc_mins_maxs_mina, 9, 9, 3, low(mina)).
table(low_s_mins_maxs_mina, 9, 10, 3, low(mina)).
table(low_v_c_minc_maxc_mina, 9, 19, 4, low(mina)).
table(low_v_c_minc_s_mina, 9, 8, 4, low(mina)).
table(low_v_c_minc_mins_mina, 9, 10, 4, low(mina)).
table(low_v_c_minc_maxs_mina, 9, 13, 4, low(mina)).
table(low_v_c_maxc_s_mina, 9, 5, 4, low(mina)).
table(low_v_c_maxc_mins_mina, 9, 17, 4, low(mina)).
table(low_v_c_maxc_maxs_mina, 9, 18, 4, low(mina)).
table(low_v_c_s_mins_mina, 9, 5, 4, low(mina)).
table(low_v_c_s_maxs_mina, 9, 5, 4, low(mina)).
table(low_v_c_mins_maxs_mina, 9, 10, 4, low(mina)).
table(low_v_minc_maxc_s_mina, 9, 7, 4, low(mina)).
table(low_v_minc_maxc_mins_mina, 9, 11, 4, low(mina)).
table(low_v_minc_maxc_maxs_mina, 9, 12, 4, low(mina)).
table(low_v_minc_s_mins_mina, 9, 5, 4, low(mina)).
table(low_v_minc_s_maxs_mina, 9, 6, 4, low(mina)).
table(low_v_minc_mins_maxs_mina, 9, 11, 4, low(mina)).
table(low_v_maxc_s_mins_mina, 9, 5, 4, low(mina)).
table(low_v_maxc_s_maxs_mina, 9, 5, 4, low(mina)).
table(low_v_maxc_mins_maxs_mina, 9, 10, 4, low(mina)).
table(low_v_s_mins_maxs_mina, 9, 6, 4, low(mina)).
table(low_c_minc_maxc_s_mina, 9, 10, 4, low(mina)).
table(low_c_minc_maxc_mins_mina, 9, 7, 4, low(mina)).
table(low_c_minc_maxc_maxs_mina, 9, 8, 4, low(mina)).
table(low_c_minc_s_mins_mina, 9, 10, 4, low(mina)).
table(low_c_minc_s_maxs_mina, 9, 13, 4, low(mina)).
table(low_c_minc_mins_maxs_mina, 9, 6, 4, low(mina)).
table(low_c_maxc_s_mins_mina, 9, 7, 4, low(mina)).
table(low_c_maxc_s_maxs_mina, 9, 9, 4, low(mina)).
table(low_c_maxc_mins_maxs_mina, 9, 6, 4, low(mina)).
table(low_c_s_mins_maxs_mina, 9, 10, 4, low(mina)).
table(low_minc_maxc_s_mins_mina, 9, 9, 4, low(mina)).
table(low_minc_maxc_s_maxs_mina, 9, 12, 4, low(mina)).
table(low_minc_maxc_mins_maxs_mina, 9, 18, 4, low(mina)).
table(low_minc_s_mins_maxs_mina, 9, 11, 4, low(mina)).
table(low_maxc_s_mins_maxs_mina, 9, 10, 4, low(mina)).
table(low_v_c_minc_maxc_s_mina, 9, 9, 5, low(mina)).
table(low_v_c_minc_maxc_mins_mina, 9, 17, 5, low(mina)).
table(low_v_c_minc_maxc_maxs_mina, 9, 18, 5, low(mina)).
table(low_v_c_minc_s_mins_mina, 9, 8, 5, low(mina)).
table(low_v_c_minc_s_maxs_mina, 9, 9, 5, low(mina)).
table(low_v_c_minc_mins_maxs_mina, 9, 13, 5, low(mina)).
table(low_v_c_maxc_s_mins_mina, 9, 6, 5, low(mina)).
table(low_v_c_maxc_s_maxs_mina, 9, 6, 5, low(mina)).
table(low_v_c_maxc_mins_maxs_mina, 9, 18, 5, low(mina)).
table(low_v_c_s_mins_maxs_mina, 9, 6, 5, low(mina)).
table(low_v_minc_maxc_s_mins_mina, 9, 7, 5, low(mina)).
table(low_v_minc_maxc_s_maxs_mina, 9, 8, 5, low(mina)).
table(low_v_minc_maxc_mins_maxs_mina, 9, 12, 5, low(mina)).
table(low_v_minc_s_mins_maxs_mina, 9, 7, 5, low(mina)).
table(low_v_maxc_s_mins_maxs_mina, 9, 6, 5, low(mina)).
table(low_c_minc_maxc_s_mins_mina, 9, 9, 5, low(mina)).
table(low_c_minc_maxc_s_maxs_mina, 9, 11, 5, low(mina)).
table(low_c_minc_maxc_mins_maxs_mina, 9, 8, 5, low(mina)).
table(low_c_minc_s_mins_maxs_mina, 9, 13, 5, low(mina)).
table(low_c_maxc_s_mins_maxs_mina, 9, 9, 5, low(mina)).
table(low_minc_maxc_s_mins_maxs_mina, 9, 12, 5, low(mina)).
table(low_v_c_minc_maxc_s_mins_mina, 9, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_maxs_mina, 9, 10, 6, low(mina)).
table(low_v_c_minc_maxc_mins_maxs_mina, 9, 18, 6, low(mina)).
table(low_v_c_minc_s_mins_maxs_mina, 9, 9, 6, low(mina)).
table(low_v_c_maxc_s_mins_maxs_mina, 9, 7, 6, low(mina)).
table(low_v_minc_maxc_s_mins_maxs_mina, 9, 8, 6, low(mina)).
table(low_c_minc_maxc_s_mins_maxs_mina, 9, 11, 6, low(mina)).
table(low_v_c_minc_maxc_s_mins_maxs_mina, 9, 10, 7, low(mina)).
table(up_v_maxa, 9, 19, 1, up(maxa)).
table(up_v_c_maxa, 9, 19, 2, up(maxa)).
table(up_v_minc_maxa, 9, 19, 2, up(maxa)).
table(up_v_maxc_maxa, 9, 19, 2, up(maxa)).
table(up_v_s_maxa, 9, 19, 2, up(maxa)).
table(up_v_mins_maxa, 9, 19, 2, up(maxa)).
table(up_v_maxs_maxa, 9, 19, 2, up(maxa)).
table(up_c_maxc_maxa, 9, 19, 2, up(maxa)).
table(up_maxc_s_maxa, 9, 18, 2, up(maxa)).
table(up_s_maxs_maxa, 9, 19, 2, up(maxa)).
table(up_v_c_minc_maxa, 9, 19, 3, up(maxa)).
table(up_v_c_maxc_maxa, 9, 19, 3, up(maxa)).
table(up_v_c_s_maxa, 9, 19, 3, up(maxa)).
table(up_v_c_mins_maxa, 9, 19, 3, up(maxa)).
table(up_v_c_maxs_maxa, 9, 7, 3, up(maxa)).
table(up_v_minc_maxc_maxa, 9, 19, 3, up(maxa)).
table(up_v_minc_s_maxa, 9, 18, 3, up(maxa)).
table(up_v_minc_mins_maxa, 9, 19, 3, up(maxa)).
table(up_v_minc_maxs_maxa, 9, 19, 3, up(maxa)).
table(up_v_maxc_s_maxa, 9, 18, 3, up(maxa)).
table(up_v_maxc_mins_maxa, 9, 19, 3, up(maxa)).
table(up_v_maxc_maxs_maxa, 9, 9, 3, up(maxa)).
table(up_v_s_mins_maxa, 9, 19, 3, up(maxa)).
table(up_v_s_maxs_maxa, 9, 19, 3, up(maxa)).
table(up_v_mins_maxs_maxa, 9, 19, 3, up(maxa)).
table(up_c_minc_maxc_maxa, 9, 19, 3, up(maxa)).
table(up_c_maxc_s_maxa, 9, 18, 3, up(maxa)).
table(up_c_maxc_mins_maxa, 9, 19, 3, up(maxa)).
table(up_c_maxc_maxs_maxa, 9, 7, 3, up(maxa)).
table(up_c_s_maxs_maxa, 9, 19, 3, up(maxa)).
table(up_minc_maxc_s_maxa, 9, 8, 3, up(maxa)).
table(up_minc_s_maxs_maxa, 9, 18, 3, up(maxa)).
table(up_maxc_s_mins_maxa, 9, 18, 3, up(maxa)).
table(up_maxc_s_maxs_maxa, 9, 6, 3, up(maxa)).
table(up_s_mins_maxs_maxa, 9, 19, 3, up(maxa)).
table(up_v_c_minc_maxc_maxa, 9, 19, 4, up(maxa)).
table(up_v_c_minc_s_maxa, 9, 18, 4, up(maxa)).
table(up_v_c_minc_mins_maxa, 9, 19, 4, up(maxa)).
table(up_v_c_minc_maxs_maxa, 9, 8, 4, up(maxa)).
table(up_v_c_maxc_s_maxa, 9, 18, 4, up(maxa)).
table(up_v_c_maxc_mins_maxa, 9, 19, 4, up(maxa)).
table(up_v_c_maxc_maxs_maxa, 9, 6, 4, up(maxa)).
table(up_v_c_s_mins_maxa, 9, 19, 4, up(maxa)).
table(up_v_c_s_maxs_maxa, 9, 19, 4, up(maxa)).
table(up_v_c_mins_maxs_maxa, 9, 11, 4, up(maxa)).
table(up_v_minc_maxc_s_maxa, 9, 18, 4, up(maxa)).
table(up_v_minc_maxc_mins_maxa, 9, 19, 4, up(maxa)).
table(up_v_minc_maxc_maxs_maxa, 9, 10, 4, up(maxa)).
table(up_v_minc_s_mins_maxa, 9, 18, 4, up(maxa)).
table(up_v_minc_s_maxs_maxa, 9, 18, 4, up(maxa)).
table(up_v_minc_mins_maxs_maxa, 9, 18, 4, up(maxa)).
table(up_v_maxc_s_mins_maxa, 9, 18, 4, up(maxa)).
table(up_v_maxc_s_maxs_maxa, 9, 18, 4, up(maxa)).
table(up_v_maxc_mins_maxs_maxa, 9, 8, 4, up(maxa)).
table(up_v_s_mins_maxs_maxa, 9, 19, 4, up(maxa)).
table(up_c_minc_maxc_s_maxa, 9, 18, 4, up(maxa)).
table(up_c_minc_maxc_mins_maxa, 9, 19, 4, up(maxa)).
table(up_c_minc_maxc_maxs_maxa, 9, 10, 4, up(maxa)).
table(up_c_minc_s_maxs_maxa, 9, 18, 4, up(maxa)).
table(up_c_maxc_s_mins_maxa, 9, 18, 4, up(maxa)).
table(up_c_maxc_s_maxs_maxa, 9, 18, 4, up(maxa)).
table(up_c_maxc_mins_maxs_maxa, 9, 6, 4, up(maxa)).
table(up_c_s_mins_maxs_maxa, 9, 19, 4, up(maxa)).
table(up_minc_maxc_s_mins_maxa, 9, 18, 4, up(maxa)).
table(up_minc_maxc_s_maxs_maxa, 9, 8, 4, up(maxa)).
table(up_minc_s_mins_maxs_maxa, 9, 18, 4, up(maxa)).
table(up_maxc_s_mins_maxs_maxa, 9, 6, 4, up(maxa)).
table(up_v_c_minc_maxc_s_maxa, 9, 18, 5, up(maxa)).
table(up_v_c_minc_maxc_mins_maxa, 9, 19, 5, up(maxa)).
table(up_v_c_minc_maxc_maxs_maxa, 9, 10, 5, up(maxa)).
table(up_v_c_minc_s_mins_maxa, 9, 18, 5, up(maxa)).
table(up_v_c_minc_s_maxs_maxa, 9, 12, 5, up(maxa)).
table(up_v_c_minc_mins_maxs_maxa, 9, 8, 5, up(maxa)).
table(up_v_c_maxc_s_mins_maxa, 9, 18, 5, up(maxa)).
table(up_v_c_maxc_s_maxs_maxa, 9, 18, 5, up(maxa)).
table(up_v_c_maxc_mins_maxs_maxa, 9, 6, 5, up(maxa)).
table(up_v_c_s_mins_maxs_maxa, 9, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_mins_maxa, 9, 18, 5, up(maxa)).
table(up_v_minc_maxc_s_maxs_maxa, 9, 18, 5, up(maxa)).
table(up_v_minc_maxc_mins_maxs_maxa, 9, 9, 5, up(maxa)).
table(up_v_minc_s_mins_maxs_maxa, 9, 18, 5, up(maxa)).
table(up_v_maxc_s_mins_maxs_maxa, 9, 18, 5, up(maxa)).
table(up_c_minc_maxc_s_mins_maxa, 9, 18, 5, up(maxa)).
table(up_c_minc_maxc_s_maxs_maxa, 9, 18, 5, up(maxa)).
table(up_c_minc_maxc_mins_maxs_maxa, 9, 9, 5, up(maxa)).
table(up_c_minc_s_mins_maxs_maxa, 9, 18, 5, up(maxa)).
table(up_c_maxc_s_mins_maxs_maxa, 9, 18, 5, up(maxa)).
table(up_minc_maxc_s_mins_maxs_maxa, 9, 18, 5, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxa, 9, 18, 6, up(maxa)).
table(up_v_c_minc_maxc_s_maxs_maxa, 9, 18, 6, up(maxa)).
table(up_v_c_minc_maxc_mins_maxs_maxa, 9, 9, 6, up(maxa)).
table(up_v_c_minc_s_mins_maxs_maxa, 9, 12, 6, up(maxa)).
table(up_v_c_maxc_s_mins_maxs_maxa, 9, 18, 6, up(maxa)).
table(up_v_minc_maxc_s_mins_maxs_maxa, 9, 18, 6, up(maxa)).
table(up_c_minc_maxc_s_mins_maxs_maxa, 9, 18, 6, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxs_maxa, 9, 18, 7, up(maxa)).
table(low_c_v, 10, 18, 1, low(v)).
table(low_minc_v, 10, 11, 1, low(v)).
table(low_maxc_v, 10, 11, 1, low(v)).
table(low_s_v, 10, 7, 1, low(v)).
table(low_mins_v, 10, 18, 1, low(v)).
table(low_maxs_v, 10, 18, 1, low(v)).
table(low_c_minc_v, 10, 11, 2, low(v)).
table(low_c_maxc_v, 10, 11, 2, low(v)).
table(low_c_s_v, 10, 8, 2, low(v)).
table(low_c_mins_v, 10, 18, 2, low(v)).
table(low_c_maxs_v, 10, 18, 2, low(v)).
table(low_minc_maxc_v, 10, 11, 2, low(v)).
table(low_minc_s_v, 10, 3, 2, low(v)).
table(low_minc_mins_v, 10, 12, 2, low(v)).
table(low_minc_maxs_v, 10, 12, 2, low(v)).
table(low_maxc_s_v, 10, 3, 2, low(v)).
table(low_maxc_mins_v, 10, 12, 2, low(v)).
table(low_maxc_maxs_v, 10, 13, 2, low(v)).
table(low_s_mins_v, 10, 7, 2, low(v)).
table(low_s_maxs_v, 10, 7, 2, low(v)).
table(low_mins_maxs_v, 10, 12, 2, low(v)).
table(low_c_minc_maxc_v, 10, 11, 3, low(v)).
table(low_c_minc_s_v, 10, 6, 3, low(v)).
table(low_c_minc_mins_v, 10, 12, 3, low(v)).
table(low_c_minc_maxs_v, 10, 12, 3, low(v)).
table(low_c_maxc_s_v, 10, 4, 3, low(v)).
table(low_c_maxc_mins_v, 10, 12, 3, low(v)).
table(low_c_maxc_maxs_v, 10, 13, 3, low(v)).
table(low_c_s_mins_v, 10, 8, 3, low(v)).
table(low_c_s_maxs_v, 10, 8, 3, low(v)).
table(low_c_mins_maxs_v, 10, 18, 3, low(v)).
table(low_minc_maxc_s_v, 10, 5, 3, low(v)).
table(low_minc_maxc_mins_v, 10, 12, 3, low(v)).
table(low_minc_maxc_maxs_v, 10, 12, 3, low(v)).
table(low_minc_s_mins_v, 10, 4, 3, low(v)).
table(low_minc_s_maxs_v, 10, 4, 3, low(v)).
table(low_minc_mins_maxs_v, 10, 13, 3, low(v)).
table(low_maxc_s_mins_v, 10, 4, 3, low(v)).
table(low_maxc_s_maxs_v, 10, 4, 3, low(v)).
table(low_maxc_mins_maxs_v, 10, 14, 3, low(v)).
table(low_s_mins_maxs_v, 10, 7, 3, low(v)).
table(low_c_minc_maxc_s_v, 10, 7, 4, low(v)).
table(low_c_minc_maxc_mins_v, 10, 12, 4, low(v)).
table(low_c_minc_maxc_maxs_v, 10, 12, 4, low(v)).
table(low_c_minc_s_mins_v, 10, 7, 4, low(v)).
table(low_c_minc_s_maxs_v, 10, 7, 4, low(v)).
table(low_c_minc_mins_maxs_v, 10, 13, 4, low(v)).
table(low_c_maxc_s_mins_v, 10, 5, 4, low(v)).
table(low_c_maxc_s_maxs_v, 10, 5, 4, low(v)).
table(low_c_maxc_mins_maxs_v, 10, 14, 4, low(v)).
table(low_c_s_mins_maxs_v, 10, 8, 4, low(v)).
table(low_minc_maxc_s_mins_v, 10, 6, 4, low(v)).
table(low_minc_maxc_s_maxs_v, 10, 6, 4, low(v)).
table(low_minc_maxc_mins_maxs_v, 10, 13, 4, low(v)).
table(low_minc_s_mins_maxs_v, 10, 5, 4, low(v)).
table(low_maxc_s_mins_maxs_v, 10, 5, 4, low(v)).
table(low_c_minc_maxc_s_mins_v, 10, 8, 5, low(v)).
table(low_c_minc_maxc_s_maxs_v, 10, 8, 5, low(v)).
table(low_c_minc_maxc_mins_maxs_v, 10, 13, 5, low(v)).
table(low_c_minc_s_mins_maxs_v, 10, 8, 5, low(v)).
table(low_c_maxc_s_mins_maxs_v, 10, 6, 5, low(v)).
table(low_minc_maxc_s_mins_maxs_v, 10, 7, 5, low(v)).
table(low_c_minc_maxc_s_mins_maxs_v, 10, 9, 6, low(v)).
table(up_c_maxc_v, 10, 3, 2, up(v)).
table(up_maxc_s_v, 10, 3, 2, up(v)).
table(up_s_maxs_v, 10, 3, 2, up(v)).
table(up_c_minc_maxc_v, 10, 6, 3, up(v)).
table(up_c_maxc_s_v, 10, 4, 3, up(v)).
table(up_c_maxc_mins_v, 10, 4, 3, up(v)).
table(up_c_maxc_maxs_v, 10, 4, 3, up(v)).
table(up_c_s_maxs_v, 10, 4, 3, up(v)).
table(up_minc_maxc_s_v, 10, 5, 3, up(v)).
table(up_minc_s_maxs_v, 10, 4, 3, up(v)).
table(up_maxc_s_mins_v, 10, 4, 3, up(v)).
table(up_maxc_s_maxs_v, 10, 4, 3, up(v)).
table(up_s_mins_maxs_v, 10, 4, 3, up(v)).
table(up_c_minc_maxc_s_v, 10, 7, 4, up(v)).
table(up_c_minc_maxc_mins_v, 10, 7, 4, up(v)).
table(up_c_minc_maxc_maxs_v, 10, 7, 4, up(v)).
table(up_c_minc_s_maxs_v, 10, 7, 4, up(v)).
table(up_c_maxc_s_mins_v, 10, 5, 4, up(v)).
table(up_c_maxc_s_maxs_v, 10, 5, 4, up(v)).
table(up_c_maxc_mins_maxs_v, 10, 5, 4, up(v)).
table(up_c_s_mins_maxs_v, 10, 5, 4, up(v)).
table(up_minc_maxc_s_mins_v, 10, 6, 4, up(v)).
table(up_minc_maxc_s_maxs_v, 10, 6, 4, up(v)).
table(up_minc_s_mins_maxs_v, 10, 5, 4, up(v)).
table(up_maxc_s_mins_maxs_v, 10, 5, 4, up(v)).
table(up_c_minc_maxc_s_mins_v, 10, 8, 5, up(v)).
table(up_c_minc_maxc_s_maxs_v, 10, 8, 5, up(v)).
table(up_c_minc_maxc_mins_maxs_v, 10, 8, 5, up(v)).
table(up_c_minc_s_mins_maxs_v, 10, 8, 5, up(v)).
table(up_c_maxc_s_mins_maxs_v, 10, 6, 5, up(v)).
table(up_minc_maxc_s_mins_maxs_v, 10, 7, 5, up(v)).
table(up_c_minc_maxc_s_mins_maxs_v, 10, 9, 6, up(v)).
table(low_v_c, 10, 11, 1, low(c)).
table(low_minc_c, 10, 11, 1, low(c)).
table(low_maxc_c, 10, 11, 1, low(c)).
table(low_s_c, 10, 8, 1, low(c)).
table(low_mins_c, 10, 8, 1, low(c)).
table(low_maxs_c, 10, 8, 1, low(c)).
table(low_v_minc_c, 10, 11, 2, low(c)).
table(low_v_maxc_c, 10, 3, 2, low(c)).
table(low_v_s_c, 10, 13, 2, low(c)).
table(low_v_mins_c, 10, 12, 2, low(c)).
table(low_v_maxs_c, 10, 13, 2, low(c)).
table(low_minc_maxc_c, 10, 11, 2, low(c)).
table(low_minc_s_c, 10, 5, 2, low(c)).
table(low_minc_mins_c, 10, 9, 2, low(c)).
table(low_minc_maxs_c, 10, 9, 2, low(c)).
table(low_maxc_s_c, 10, 3, 2, low(c)).
table(low_maxc_mins_c, 10, 12, 2, low(c)).
table(low_maxc_maxs_c, 10, 13, 2, low(c)).
table(low_s_mins_c, 10, 10, 2, low(c)).
table(low_s_maxs_c, 10, 11, 2, low(c)).
table(low_mins_maxs_c, 10, 9, 2, low(c)).
table(low_v_minc_maxc_c, 10, 6, 3, low(c)).
table(low_v_minc_s_c, 10, 12, 3, low(c)).
table(low_v_minc_mins_c, 10, 12, 3, low(c)).
table(low_v_minc_maxs_c, 10, 12, 3, low(c)).
table(low_v_maxc_s_c, 10, 4, 3, low(c)).
table(low_v_maxc_mins_c, 10, 4, 3, low(c)).
table(low_v_maxc_maxs_c, 10, 4, 3, low(c)).
table(low_v_s_mins_c, 10, 14, 3, low(c)).
table(low_v_s_maxs_c, 10, 14, 3, low(c)).
table(low_v_mins_maxs_c, 10, 14, 3, low(c)).
table(low_minc_maxc_s_c, 10, 5, 3, low(c)).
table(low_minc_maxc_mins_c, 10, 12, 3, low(c)).
table(low_minc_maxc_maxs_c, 10, 12, 3, low(c)).
table(low_minc_s_mins_c, 10, 6, 3, low(c)).
table(low_minc_s_maxs_c, 10, 7, 3, low(c)).
table(low_minc_mins_maxs_c, 10, 7, 3, low(c)).
table(low_maxc_s_mins_c, 10, 4, 3, low(c)).
table(low_maxc_s_maxs_c, 10, 4, 3, low(c)).
table(low_maxc_mins_maxs_c, 10, 6, 3, low(c)).
table(low_s_mins_maxs_c, 10, 12, 3, low(c)).
table(low_v_minc_maxc_s_c, 10, 7, 4, low(c)).
table(low_v_minc_maxc_mins_c, 10, 7, 4, low(c)).
table(low_v_minc_maxc_maxs_c, 10, 7, 4, low(c)).
table(low_v_minc_s_mins_c, 10, 13, 4, low(c)).
table(low_v_minc_s_maxs_c, 10, 13, 4, low(c)).
table(low_v_minc_mins_maxs_c, 10, 13, 4, low(c)).
table(low_v_maxc_s_mins_c, 10, 5, 4, low(c)).
table(low_v_maxc_s_maxs_c, 10, 5, 4, low(c)).
table(low_v_maxc_mins_maxs_c, 10, 5, 4, low(c)).
table(low_v_s_mins_maxs_c, 10, 15, 4, low(c)).
table(low_minc_maxc_s_mins_c, 10, 6, 4, low(c)).
table(low_minc_maxc_s_maxs_c, 10, 6, 4, low(c)).
table(low_minc_maxc_mins_maxs_c, 10, 13, 4, low(c)).
table(low_minc_s_mins_maxs_c, 10, 8, 4, low(c)).
table(low_maxc_s_mins_maxs_c, 10, 5, 4, low(c)).
table(low_v_minc_maxc_s_mins_c, 10, 8, 5, low(c)).
table(low_v_minc_maxc_s_maxs_c, 10, 8, 5, low(c)).
table(low_v_minc_maxc_mins_maxs_c, 10, 8, 5, low(c)).
table(low_v_minc_s_mins_maxs_c, 10, 14, 5, low(c)).
table(low_v_maxc_s_mins_maxs_c, 10, 6, 5, low(c)).
table(low_minc_maxc_s_mins_maxs_c, 10, 7, 5, low(c)).
table(low_v_minc_maxc_s_mins_maxs_c, 10, 9, 6, low(c)).
table(up_v_c, 10, 18, 1, up(c)).
table(up_s_c, 10, 3, 1, up(c)).
table(up_v_minc_c, 10, 11, 2, up(c)).
table(up_v_maxc_c, 10, 11, 2, up(c)).
table(up_v_s_c, 10, 4, 2, up(c)).
table(up_v_mins_c, 10, 18, 2, up(c)).
table(up_v_maxs_c, 10, 18, 2, up(c)).
table(up_minc_s_c, 10, 3, 2, up(c)).
table(up_maxc_s_c, 10, 4, 2, up(c)).
table(up_s_mins_c, 10, 5, 2, up(c)).
table(up_s_maxs_c, 10, 5, 2, up(c)).
table(up_v_minc_maxc_c, 10, 11, 3, up(c)).
table(up_v_minc_s_c, 10, 6, 3, up(c)).
table(up_v_minc_mins_c, 10, 12, 3, up(c)).
table(up_v_minc_maxs_c, 10, 7, 3, up(c)).
table(up_v_maxc_s_c, 10, 5, 3, up(c)).
table(up_v_maxc_mins_c, 10, 12, 3, up(c)).
table(up_v_maxc_maxs_c, 10, 13, 3, up(c)).
table(up_v_s_mins_c, 10, 8, 3, up(c)).
table(up_v_s_maxs_c, 10, 6, 3, up(c)).
table(up_v_mins_maxs_c, 10, 18, 3, up(c)).
table(up_minc_maxc_s_c, 10, 5, 3, up(c)).
table(up_minc_s_mins_c, 10, 4, 3, up(c)).
table(up_minc_s_maxs_c, 10, 4, 3, up(c)).
table(up_maxc_s_mins_c, 10, 5, 3, up(c)).
table(up_maxc_s_maxs_c, 10, 5, 3, up(c)).
table(up_s_mins_maxs_c, 10, 8, 3, up(c)).
table(up_v_minc_maxc_s_c, 10, 7, 4, up(c)).
table(up_v_minc_maxc_mins_c, 10, 12, 4, up(c)).
table(up_v_minc_maxc_maxs_c, 10, 12, 4, up(c)).
table(up_v_minc_s_mins_c, 10, 7, 4, up(c)).
table(up_v_minc_s_maxs_c, 10, 7, 4, up(c)).
table(up_v_minc_mins_maxs_c, 10, 8, 4, up(c)).
table(up_v_maxc_s_mins_c, 10, 9, 4, up(c)).
table(up_v_maxc_s_maxs_c, 10, 6, 4, up(c)).
table(up_v_maxc_mins_maxs_c, 10, 14, 4, up(c)).
table(up_v_s_mins_maxs_c, 10, 10, 4, up(c)).
table(up_minc_maxc_s_mins_c, 10, 6, 4, up(c)).
table(up_minc_maxc_s_maxs_c, 10, 6, 4, up(c)).
table(up_minc_s_mins_maxs_c, 10, 5, 4, up(c)).
table(up_maxc_s_mins_maxs_c, 10, 6, 4, up(c)).
table(up_v_minc_maxc_s_mins_c, 10, 8, 5, up(c)).
table(up_v_minc_maxc_s_maxs_c, 10, 8, 5, up(c)).
table(up_v_minc_maxc_mins_maxs_c, 10, 13, 5, up(c)).
table(up_v_minc_s_mins_maxs_c, 10, 8, 5, up(c)).
table(up_v_maxc_s_mins_maxs_c, 10, 7, 5, up(c)).
table(up_minc_maxc_s_mins_maxs_c, 10, 7, 5, up(c)).
table(up_v_minc_maxc_s_mins_maxs_c, 10, 9, 6, up(c)).
table(low_v_minc, 10, 3, 1, low(minc)).
table(low_c_minc, 10, 3, 1, low(minc)).
table(low_maxc_minc, 10, 3, 1, low(minc)).
table(low_s_minc, 10, 3, 1, low(minc)).
table(low_mins_minc, 10, 2, 1, low(minc)).
table(low_maxs_minc, 10, 3, 1, low(minc)).
table(low_v_c_minc, 10, 5, 2, low(minc)).
table(low_v_maxc_minc, 10, 4, 2, low(minc)).
table(low_v_s_minc, 10, 4, 2, low(minc)).
table(low_v_mins_minc, 10, 3, 2, low(minc)).
table(low_v_maxs_minc, 10, 4, 2, low(minc)).
table(low_c_maxc_minc, 10, 4, 2, low(minc)).
table(low_c_s_minc, 10, 4, 2, low(minc)).
table(low_c_mins_minc, 10, 3, 2, low(minc)).
table(low_c_maxs_minc, 10, 4, 2, low(minc)).
table(low_maxc_s_minc, 10, 4, 2, low(minc)).
table(low_maxc_mins_minc, 10, 4, 2, low(minc)).
table(low_maxc_maxs_minc, 10, 4, 2, low(minc)).
table(low_s_mins_minc, 10, 3, 2, low(minc)).
table(low_s_maxs_minc, 10, 4, 2, low(minc)).
table(low_mins_maxs_minc, 10, 3, 2, low(minc)).
table(low_v_c_maxc_minc, 10, 6, 3, low(minc)).
table(low_v_c_s_minc, 10, 6, 3, low(minc)).
table(low_v_c_mins_minc, 10, 6, 3, low(minc)).
table(low_v_c_maxs_minc, 10, 6, 3, low(minc)).
table(low_v_maxc_s_minc, 10, 5, 3, low(minc)).
table(low_v_maxc_mins_minc, 10, 5, 3, low(minc)).
table(low_v_maxc_maxs_minc, 10, 5, 3, low(minc)).
table(low_v_s_mins_minc, 10, 4, 3, low(minc)).
table(low_v_s_maxs_minc, 10, 5, 3, low(minc)).
table(low_v_mins_maxs_minc, 10, 4, 3, low(minc)).
table(low_c_maxc_s_minc, 10, 5, 3, low(minc)).
table(low_c_maxc_mins_minc, 10, 5, 3, low(minc)).
table(low_c_maxc_maxs_minc, 10, 5, 3, low(minc)).
table(low_c_s_mins_minc, 10, 4, 3, low(minc)).
table(low_c_s_maxs_minc, 10, 6, 3, low(minc)).
table(low_c_mins_maxs_minc, 10, 4, 3, low(minc)).
table(low_maxc_s_mins_minc, 10, 5, 3, low(minc)).
table(low_maxc_s_maxs_minc, 10, 5, 3, low(minc)).
table(low_maxc_mins_maxs_minc, 10, 5, 3, low(minc)).
table(low_s_mins_maxs_minc, 10, 4, 3, low(minc)).
table(low_v_c_maxc_s_minc, 10, 7, 4, low(minc)).
table(low_v_c_maxc_mins_minc, 10, 7, 4, low(minc)).
table(low_v_c_maxc_maxs_minc, 10, 7, 4, low(minc)).
table(low_v_c_s_mins_minc, 10, 7, 4, low(minc)).
table(low_v_c_s_maxs_minc, 10, 7, 4, low(minc)).
table(low_v_c_mins_maxs_minc, 10, 7, 4, low(minc)).
table(low_v_maxc_s_mins_minc, 10, 6, 4, low(minc)).
table(low_v_maxc_s_maxs_minc, 10, 6, 4, low(minc)).
table(low_v_maxc_mins_maxs_minc, 10, 6, 4, low(minc)).
table(low_v_s_mins_maxs_minc, 10, 5, 4, low(minc)).
table(low_c_maxc_s_mins_minc, 10, 6, 4, low(minc)).
table(low_c_maxc_s_maxs_minc, 10, 6, 4, low(minc)).
table(low_c_maxc_mins_maxs_minc, 10, 6, 4, low(minc)).
table(low_c_s_mins_maxs_minc, 10, 6, 4, low(minc)).
table(low_maxc_s_mins_maxs_minc, 10, 6, 4, low(minc)).
table(low_v_c_maxc_s_mins_minc, 10, 8, 5, low(minc)).
table(low_v_c_maxc_s_maxs_minc, 10, 8, 5, low(minc)).
table(low_v_c_maxc_mins_maxs_minc, 10, 8, 5, low(minc)).
table(low_v_c_s_mins_maxs_minc, 10, 8, 5, low(minc)).
table(low_v_maxc_s_mins_maxs_minc, 10, 7, 5, low(minc)).
table(low_c_maxc_s_mins_maxs_minc, 10, 7, 5, low(minc)).
table(low_v_c_maxc_s_mins_maxs_minc, 10, 9, 6, low(minc)).
table(up_v_minc, 10, 11, 1, up(minc)).
table(up_maxc_minc, 10, 6, 1, up(minc)).
table(up_v_c_minc, 10, 5, 2, up(minc)).
table(up_v_maxc_minc, 10, 6, 2, up(minc)).
table(up_v_s_minc, 10, 13, 2, up(minc)).
table(up_v_mins_minc, 10, 12, 2, up(minc)).
table(up_v_maxs_minc, 10, 13, 2, up(minc)).
table(up_c_maxc_minc, 10, 4, 2, up(minc)).
table(up_maxc_s_minc, 10, 7, 2, up(minc)).
table(up_maxc_mins_minc, 10, 6, 2, up(minc)).
table(up_maxc_maxs_minc, 10, 7, 2, up(minc)).
table(up_s_maxs_minc, 10, 14, 2, up(minc)).
table(up_v_c_maxc_minc, 10, 6, 3, up(minc)).
table(up_v_c_s_minc, 10, 6, 3, up(minc)).
table(up_v_c_mins_minc, 10, 6, 3, up(minc)).
table(up_v_c_maxs_minc, 10, 6, 3, up(minc)).
table(up_v_maxc_s_minc, 10, 7, 3, up(minc)).
table(up_v_maxc_mins_minc, 10, 7, 3, up(minc)).
table(up_v_maxc_maxs_minc, 10, 7, 3, up(minc)).
table(up_v_s_mins_minc, 10, 14, 3, up(minc)).
table(up_v_s_maxs_minc, 10, 14, 3, up(minc)).
table(up_v_mins_maxs_minc, 10, 14, 3, up(minc)).
table(up_c_maxc_s_minc, 10, 5, 3, up(minc)).
table(up_c_maxc_mins_minc, 10, 5, 3, up(minc)).
table(up_c_maxc_maxs_minc, 10, 5, 3, up(minc)).
table(up_c_s_maxs_minc, 10, 4, 3, up(minc)).
table(up_maxc_s_mins_minc, 10, 7, 3, up(minc)).
table(up_maxc_s_maxs_minc, 10, 7, 3, up(minc)).
table(up_maxc_mins_maxs_minc, 10, 7, 3, up(minc)).
table(up_s_mins_maxs_minc, 10, 15, 3, up(minc)).
table(up_v_c_maxc_s_minc, 10, 7, 4, up(minc)).
table(up_v_c_maxc_mins_minc, 10, 7, 4, up(minc)).
table(up_v_c_maxc_maxs_minc, 10, 7, 4, up(minc)).
table(up_v_c_s_mins_minc, 10, 7, 4, up(minc)).
table(up_v_c_s_maxs_minc, 10, 7, 4, up(minc)).
table(up_v_c_mins_maxs_minc, 10, 7, 4, up(minc)).
table(up_v_maxc_s_mins_minc, 10, 8, 4, up(minc)).
table(up_v_maxc_s_maxs_minc, 10, 8, 4, up(minc)).
table(up_v_maxc_mins_maxs_minc, 10, 8, 4, up(minc)).
table(up_v_s_mins_maxs_minc, 10, 15, 4, up(minc)).
table(up_c_maxc_s_mins_minc, 10, 6, 4, up(minc)).
table(up_c_maxc_s_maxs_minc, 10, 6, 4, up(minc)).
table(up_c_maxc_mins_maxs_minc, 10, 6, 4, up(minc)).
table(up_c_s_mins_maxs_minc, 10, 5, 4, up(minc)).
table(up_maxc_s_mins_maxs_minc, 10, 8, 4, up(minc)).
table(up_v_c_maxc_s_mins_minc, 10, 8, 5, up(minc)).
table(up_v_c_maxc_s_maxs_minc, 10, 8, 5, up(minc)).
table(up_v_c_maxc_mins_maxs_minc, 10, 8, 5, up(minc)).
table(up_v_c_s_mins_maxs_minc, 10, 8, 5, up(minc)).
table(up_v_maxc_s_mins_maxs_minc, 10, 9, 5, up(minc)).
table(up_c_maxc_s_mins_maxs_minc, 10, 7, 5, up(minc)).
table(up_v_c_maxc_s_mins_maxs_minc, 10, 9, 6, up(minc)).
table(low_v_maxc, 10, 18, 1, low(maxc)).
table(low_c_maxc, 10, 18, 1, low(maxc)).
table(low_minc_maxc, 10, 6, 1, low(maxc)).
table(low_s_maxc, 10, 18, 1, low(maxc)).
table(low_mins_maxc, 10, 9, 1, low(maxc)).
table(low_maxs_maxc, 10, 2, 1, low(maxc)).
table(low_v_c_maxc, 10, 3, 2, low(maxc)).
table(low_v_minc_maxc, 10, 11, 2, low(maxc)).
table(low_v_s_maxc, 10, 4, 2, low(maxc)).
table(low_v_mins_maxc, 10, 18, 2, low(maxc)).
table(low_v_maxs_maxc, 10, 3, 2, low(maxc)).
table(low_c_minc_maxc, 10, 11, 2, low(maxc)).
table(low_c_s_maxc, 10, 4, 2, low(maxc)).
table(low_c_mins_maxc, 10, 18, 2, low(maxc)).
table(low_c_maxs_maxc, 10, 3, 2, low(maxc)).
table(low_minc_s_maxc, 10, 7, 2, low(maxc)).
table(low_minc_mins_maxc, 10, 7, 2, low(maxc)).
table(low_minc_maxs_maxc, 10, 4, 2, low(maxc)).
table(low_s_mins_maxc, 10, 18, 2, low(maxc)).
table(low_s_maxs_maxc, 10, 3, 2, low(maxc)).
table(low_mins_maxs_maxc, 10, 3, 2, low(maxc)).
table(low_v_c_minc_maxc, 10, 6, 3, low(maxc)).
table(low_v_c_s_maxc, 10, 4, 3, low(maxc)).
table(low_v_c_mins_maxc, 10, 4, 3, low(maxc)).
table(low_v_c_maxs_maxc, 10, 4, 3, low(maxc)).
table(low_v_minc_s_maxc, 10, 5, 3, low(maxc)).
table(low_v_minc_mins_maxc, 10, 12, 3, low(maxc)).
table(low_v_minc_maxs_maxc, 10, 5, 3, low(maxc)).
table(low_v_s_mins_maxc, 10, 5, 3, low(maxc)).
table(low_v_s_maxs_maxc, 10, 4, 3, low(maxc)).
table(low_v_mins_maxs_maxc, 10, 4, 3, low(maxc)).
table(low_c_minc_s_maxc, 10, 5, 3, low(maxc)).
table(low_c_minc_mins_maxc, 10, 12, 3, low(maxc)).
table(low_c_minc_maxs_maxc, 10, 5, 3, low(maxc)).
table(low_c_s_mins_maxc, 10, 4, 3, low(maxc)).
table(low_c_s_maxs_maxc, 10, 4, 3, low(maxc)).
table(low_c_mins_maxs_maxc, 10, 4, 3, low(maxc)).
table(low_minc_s_mins_maxc, 10, 8, 3, low(maxc)).
table(low_minc_s_maxs_maxc, 10, 5, 3, low(maxc)).
table(low_minc_mins_maxs_maxc, 10, 5, 3, low(maxc)).
table(low_s_mins_maxs_maxc, 10, 4, 3, low(maxc)).
table(low_v_c_minc_s_maxc, 10, 7, 4, low(maxc)).
table(low_v_c_minc_mins_maxc, 10, 7, 4, low(maxc)).
table(low_v_c_minc_maxs_maxc, 10, 7, 4, low(maxc)).
table(low_v_c_s_mins_maxc, 10, 5, 4, low(maxc)).
table(low_v_c_s_maxs_maxc, 10, 5, 4, low(maxc)).
table(low_v_c_mins_maxs_maxc, 10, 5, 4, low(maxc)).
table(low_v_minc_s_mins_maxc, 10, 6, 4, low(maxc)).
table(low_v_minc_s_maxs_maxc, 10, 6, 4, low(maxc)).
table(low_v_minc_mins_maxs_maxc, 10, 6, 4, low(maxc)).
table(low_v_s_mins_maxs_maxc, 10, 5, 4, low(maxc)).
table(low_c_minc_s_mins_maxc, 10, 6, 4, low(maxc)).
table(low_c_minc_s_maxs_maxc, 10, 6, 4, low(maxc)).
table(low_c_minc_mins_maxs_maxc, 10, 6, 4, low(maxc)).
table(low_c_s_mins_maxs_maxc, 10, 5, 4, low(maxc)).
table(low_minc_s_mins_maxs_maxc, 10, 6, 4, low(maxc)).
table(low_v_c_minc_s_mins_maxc, 10, 8, 5, low(maxc)).
table(low_v_c_minc_s_maxs_maxc, 10, 8, 5, low(maxc)).
table(low_v_c_minc_mins_maxs_maxc, 10, 8, 5, low(maxc)).
table(low_v_c_s_mins_maxs_maxc, 10, 6, 5, low(maxc)).
table(low_v_minc_s_mins_maxs_maxc, 10, 7, 5, low(maxc)).
table(low_c_minc_s_mins_maxs_maxc, 10, 7, 5, low(maxc)).
table(low_v_c_minc_s_mins_maxs_maxc, 10, 9, 6, low(maxc)).
table(up_v_maxc, 10, 11, 1, up(maxc)).
table(up_v_c_maxc, 10, 11, 2, up(maxc)).
table(up_v_minc_maxc, 10, 11, 2, up(maxc)).
table(up_v_s_maxc, 10, 13, 2, up(maxc)).
table(up_v_mins_maxc, 10, 12, 2, up(maxc)).
table(up_v_maxs_maxc, 10, 13, 2, up(maxc)).
table(up_v_c_minc_maxc, 10, 11, 3, up(maxc)).
table(up_v_c_s_maxc, 10, 13, 3, up(maxc)).
table(up_v_c_mins_maxc, 10, 12, 3, up(maxc)).
table(up_v_c_maxs_maxc, 10, 13, 3, up(maxc)).
table(up_v_minc_s_maxc, 10, 12, 3, up(maxc)).
table(up_v_minc_mins_maxc, 10, 12, 3, up(maxc)).
table(up_v_minc_maxs_maxc, 10, 12, 3, up(maxc)).
table(up_v_s_mins_maxc, 10, 14, 3, up(maxc)).
table(up_v_s_maxs_maxc, 10, 14, 3, up(maxc)).
table(up_v_mins_maxs_maxc, 10, 14, 3, up(maxc)).
table(up_v_c_minc_s_maxc, 10, 12, 4, up(maxc)).
table(up_v_c_minc_mins_maxc, 10, 12, 4, up(maxc)).
table(up_v_c_minc_maxs_maxc, 10, 12, 4, up(maxc)).
table(up_v_c_s_mins_maxc, 10, 14, 4, up(maxc)).
table(up_v_c_s_maxs_maxc, 10, 6, 4, up(maxc)).
table(up_v_c_mins_maxs_maxc, 10, 14, 4, up(maxc)).
table(up_v_minc_s_mins_maxc, 10, 13, 4, up(maxc)).
table(up_v_minc_s_maxs_maxc, 10, 13, 4, up(maxc)).
table(up_v_minc_mins_maxs_maxc, 10, 13, 4, up(maxc)).
table(up_v_s_mins_maxs_maxc, 10, 15, 4, up(maxc)).
table(up_v_c_minc_s_mins_maxc, 10, 13, 5, up(maxc)).
table(up_v_c_minc_s_maxs_maxc, 10, 8, 5, up(maxc)).
table(up_v_c_minc_mins_maxs_maxc, 10, 13, 5, up(maxc)).
table(up_v_c_s_mins_maxs_maxc, 10, 10, 5, up(maxc)).
table(up_v_minc_s_mins_maxs_maxc, 10, 14, 5, up(maxc)).
table(up_v_c_minc_s_mins_maxs_maxc, 10, 9, 6, up(maxc)).
table(low_v_s, 10, 18, 1, low(s)).
table(low_c_s, 10, 3, 1, low(s)).
table(low_minc_s, 10, 18, 1, low(s)).
table(low_maxc_s, 10, 18, 1, low(s)).
table(low_mins_s, 10, 18, 1, low(s)).
table(low_maxs_s, 10, 18, 1, low(s)).
table(low_v_c_s, 10, 4, 2, low(s)).
table(low_v_minc_s, 10, 18, 2, low(s)).
table(low_v_maxc_s, 10, 6, 2, low(s)).
table(low_v_mins_s, 10, 9, 2, low(s)).
table(low_v_maxs_s, 10, 3, 2, low(s)).
table(low_c_minc_s, 10, 5, 2, low(s)).
table(low_c_maxc_s, 10, 5, 2, low(s)).
table(low_c_mins_s, 10, 5, 2, low(s)).
table(low_c_maxs_s, 10, 5, 2, low(s)).
table(low_minc_maxc_s, 10, 18, 2, low(s)).
table(low_minc_mins_s, 10, 12, 2, low(s)).
table(low_minc_maxs_s, 10, 14, 2, low(s)).
table(low_maxc_mins_s, 10, 7, 2, low(s)).
table(low_maxc_maxs_s, 10, 14, 2, low(s)).
table(low_mins_maxs_s, 10, 12, 2, low(s)).
table(low_v_c_minc_s, 10, 8, 3, low(s)).
table(low_v_c_maxc_s, 10, 6, 3, low(s)).
table(low_v_c_mins_s, 10, 8, 3, low(s)).
table(low_v_c_maxs_s, 10, 4, 3, low(s)).
table(low_v_minc_maxc_s, 10, 10, 3, low(s)).
table(low_v_minc_mins_s, 10, 13, 3, low(s)).
table(low_v_minc_maxs_s, 10, 4, 3, low(s)).
table(low_v_maxc_mins_s, 10, 4, 3, low(s)).
table(low_v_maxc_maxs_s, 10, 4, 3, low(s)).
table(low_v_mins_maxs_s, 10, 4, 3, low(s)).
table(low_c_minc_maxc_s, 10, 8, 3, low(s)).
table(low_c_minc_mins_s, 10, 4, 3, low(s)).
table(low_c_minc_maxs_s, 10, 5, 3, low(s)).
table(low_c_maxc_mins_s, 10, 8, 3, low(s)).
table(low_c_maxc_maxs_s, 10, 5, 3, low(s)).
table(low_c_mins_maxs_s, 10, 8, 3, low(s)).
table(low_minc_maxc_mins_s, 10, 13, 3, low(s)).
table(low_minc_maxc_maxs_s, 10, 14, 3, low(s)).
table(low_minc_mins_maxs_s, 10, 8, 3, low(s)).
table(low_maxc_mins_maxs_s, 10, 9, 3, low(s)).
table(low_v_c_minc_maxc_s, 10, 10, 4, low(s)).
table(low_v_c_minc_mins_s, 10, 7, 4, low(s)).
table(low_v_c_minc_maxs_s, 10, 7, 4, low(s)).
table(low_v_c_maxc_mins_s, 10, 5, 4, low(s)).
table(low_v_c_maxc_maxs_s, 10, 5, 4, low(s)).
table(low_v_c_mins_maxs_s, 10, 5, 4, low(s)).
table(low_v_minc_maxc_mins_s, 10, 8, 4, low(s)).
table(low_v_minc_maxc_maxs_s, 10, 6, 4, low(s)).
table(low_v_minc_mins_maxs_s, 10, 5, 4, low(s)).
table(low_v_maxc_mins_maxs_s, 10, 5, 4, low(s)).
table(low_c_minc_maxc_mins_s, 10, 6, 4, low(s)).
table(low_c_minc_maxc_maxs_s, 10, 7, 4, low(s)).
table(low_c_minc_mins_maxs_s, 10, 5, 4, low(s)).
table(low_c_maxc_mins_maxs_s, 10, 6, 4, low(s)).
table(low_minc_maxc_mins_maxs_s, 10, 14, 4, low(s)).
table(low_v_c_minc_maxc_mins_s, 10, 8, 5, low(s)).
table(low_v_c_minc_maxc_maxs_s, 10, 8, 5, low(s)).
table(low_v_c_minc_mins_maxs_s, 10, 8, 5, low(s)).
table(low_v_c_maxc_mins_maxs_s, 10, 6, 5, low(s)).
table(low_v_minc_maxc_mins_maxs_s, 10, 7, 5, low(s)).
table(low_c_minc_maxc_mins_maxs_s, 10, 7, 5, low(s)).
table(low_v_c_minc_maxc_mins_maxs_s, 10, 9, 6, low(s)).
table(up_v_s, 10, 7, 1, up(s)).
table(up_v_c_s, 10, 8, 2, up(s)).
table(up_v_minc_s, 10, 8, 2, up(s)).
table(up_v_maxc_s, 10, 8, 2, up(s)).
table(up_v_mins_s, 10, 7, 2, up(s)).
table(up_v_maxs_s, 10, 7, 2, up(s)).
table(up_c_maxc_s, 10, 9, 2, up(s)).
table(up_v_c_minc_s, 10, 11, 3, up(s)).
table(up_v_c_maxc_s, 10, 9, 3, up(s)).
table(up_v_c_mins_s, 10, 8, 3, up(s)).
table(up_v_c_maxs_s, 10, 8, 3, up(s)).
table(up_v_minc_maxc_s, 10, 10, 3, up(s)).
table(up_v_minc_mins_s, 10, 8, 3, up(s)).
table(up_v_minc_maxs_s, 10, 8, 3, up(s)).
table(up_v_maxc_mins_s, 10, 8, 3, up(s)).
table(up_v_maxc_maxs_s, 10, 8, 3, up(s)).
table(up_v_mins_maxs_s, 10, 7, 3, up(s)).
table(up_c_minc_maxc_s, 10, 12, 3, up(s)).
table(up_c_maxc_mins_s, 10, 5, 3, up(s)).
table(up_c_maxc_maxs_s, 10, 9, 3, up(s)).
table(up_v_c_minc_maxc_s, 10, 12, 4, up(s)).
table(up_v_c_minc_mins_s, 10, 11, 4, up(s)).
table(up_v_c_minc_maxs_s, 10, 11, 4, up(s)).
table(up_v_c_maxc_mins_s, 10, 9, 4, up(s)).
table(up_v_c_maxc_maxs_s, 10, 9, 4, up(s)).
table(up_v_c_mins_maxs_s, 10, 8, 4, up(s)).
table(up_v_minc_maxc_mins_s, 10, 10, 4, up(s)).
table(up_v_minc_maxc_maxs_s, 10, 10, 4, up(s)).
table(up_v_minc_mins_maxs_s, 10, 8, 4, up(s)).
table(up_v_maxc_mins_maxs_s, 10, 8, 4, up(s)).
table(up_c_minc_maxc_mins_s, 10, 8, 4, up(s)).
table(up_c_minc_maxc_maxs_s, 10, 12, 4, up(s)).
table(up_c_maxc_mins_maxs_s, 10, 6, 4, up(s)).
table(up_v_c_minc_maxc_mins_s, 10, 12, 5, up(s)).
table(up_v_c_minc_maxc_maxs_s, 10, 12, 5, up(s)).
table(up_v_c_minc_mins_maxs_s, 10, 11, 5, up(s)).
table(up_v_c_maxc_mins_maxs_s, 10, 9, 5, up(s)).
table(up_v_minc_maxc_mins_maxs_s, 10, 10, 5, up(s)).
table(up_c_minc_maxc_mins_maxs_s, 10, 8, 5, up(s)).
table(up_v_c_minc_maxc_mins_maxs_s, 10, 12, 6, up(s)).
table(low_v_mins, 10, 2, 1, low(mins)).
table(low_c_mins, 10, 2, 1, low(mins)).
table(low_minc_mins, 10, 2, 1, low(mins)).
table(low_maxc_mins, 10, 2, 1, low(mins)).
table(low_s_mins, 10, 2, 1, low(mins)).
table(low_maxs_mins, 10, 2, 1, low(mins)).
table(low_v_c_mins, 10, 3, 2, low(mins)).
table(low_v_minc_mins, 10, 3, 2, low(mins)).
table(low_v_maxc_mins, 10, 3, 2, low(mins)).
table(low_v_s_mins, 10, 3, 2, low(mins)).
table(low_v_maxs_mins, 10, 3, 2, low(mins)).
table(low_c_minc_mins, 10, 3, 2, low(mins)).
table(low_c_maxc_mins, 10, 3, 2, low(mins)).
table(low_c_s_mins, 10, 3, 2, low(mins)).
table(low_c_maxs_mins, 10, 3, 2, low(mins)).
table(low_minc_maxc_mins, 10, 4, 2, low(mins)).
table(low_minc_s_mins, 10, 3, 2, low(mins)).
table(low_minc_maxs_mins, 10, 3, 2, low(mins)).
table(low_maxc_s_mins, 10, 3, 2, low(mins)).
table(low_maxc_maxs_mins, 10, 3, 2, low(mins)).
table(low_s_maxs_mins, 10, 3, 2, low(mins)).
table(low_v_c_minc_mins, 10, 6, 3, low(mins)).
table(low_v_c_maxc_mins, 10, 4, 3, low(mins)).
table(low_v_c_s_mins, 10, 4, 3, low(mins)).
table(low_v_c_maxs_mins, 10, 4, 3, low(mins)).
table(low_v_minc_maxc_mins, 10, 5, 3, low(mins)).
table(low_v_minc_s_mins, 10, 4, 3, low(mins)).
table(low_v_minc_maxs_mins, 10, 4, 3, low(mins)).
table(low_v_maxc_s_mins, 10, 4, 3, low(mins)).
table(low_v_maxc_maxs_mins, 10, 4, 3, low(mins)).
table(low_v_s_maxs_mins, 10, 4, 3, low(mins)).
table(low_c_minc_maxc_mins, 10, 5, 3, low(mins)).
table(low_c_minc_s_mins, 10, 4, 3, low(mins)).
table(low_c_minc_maxs_mins, 10, 4, 3, low(mins)).
table(low_c_maxc_s_mins, 10, 4, 3, low(mins)).
table(low_c_maxc_maxs_mins, 10, 4, 3, low(mins)).
table(low_c_s_maxs_mins, 10, 4, 3, low(mins)).
table(low_minc_maxc_s_mins, 10, 5, 3, low(mins)).
table(low_minc_maxc_maxs_mins, 10, 5, 3, low(mins)).
table(low_minc_s_maxs_mins, 10, 4, 3, low(mins)).
table(low_maxc_s_maxs_mins, 10, 4, 3, low(mins)).
table(low_v_c_minc_maxc_mins, 10, 7, 4, low(mins)).
table(low_v_c_minc_s_mins, 10, 7, 4, low(mins)).
table(low_v_c_minc_maxs_mins, 10, 7, 4, low(mins)).
table(low_v_c_maxc_s_mins, 10, 5, 4, low(mins)).
table(low_v_c_maxc_maxs_mins, 10, 5, 4, low(mins)).
table(low_v_c_s_maxs_mins, 10, 5, 4, low(mins)).
table(low_v_minc_maxc_s_mins, 10, 6, 4, low(mins)).
table(low_v_minc_maxc_maxs_mins, 10, 6, 4, low(mins)).
table(low_v_minc_s_maxs_mins, 10, 5, 4, low(mins)).
table(low_v_maxc_s_maxs_mins, 10, 5, 4, low(mins)).
table(low_c_minc_maxc_s_mins, 10, 6, 4, low(mins)).
table(low_c_minc_maxc_maxs_mins, 10, 6, 4, low(mins)).
table(low_c_minc_s_maxs_mins, 10, 5, 4, low(mins)).
table(low_c_maxc_s_maxs_mins, 10, 5, 4, low(mins)).
table(low_minc_maxc_s_maxs_mins, 10, 6, 4, low(mins)).
table(low_v_c_minc_maxc_s_mins, 10, 8, 5, low(mins)).
table(low_v_c_minc_maxc_maxs_mins, 10, 8, 5, low(mins)).
table(low_v_c_minc_s_maxs_mins, 10, 8, 5, low(mins)).
table(low_v_c_maxc_s_maxs_mins, 10, 6, 5, low(mins)).
table(low_v_minc_maxc_s_maxs_mins, 10, 7, 5, low(mins)).
table(low_c_minc_maxc_s_maxs_mins, 10, 7, 5, low(mins)).
table(low_v_c_minc_maxc_s_maxs_mins, 10, 9, 6, low(mins)).
table(up_v_mins, 10, 18, 1, up(mins)).
table(up_maxc_mins, 10, 9, 1, up(mins)).
table(up_maxs_mins, 10, 2, 1, up(mins)).
table(up_v_c_mins, 10, 6, 2, up(mins)).
table(up_v_minc_mins, 10, 3, 2, up(mins)).
table(up_v_maxc_mins, 10, 7, 2, up(mins)).
table(up_v_s_mins, 10, 3, 2, up(mins)).
table(up_v_maxs_mins, 10, 3, 2, up(mins)).
table(up_c_maxc_mins, 10, 5, 2, up(mins)).
table(up_c_maxs_mins, 10, 3, 2, up(mins)).
table(up_minc_maxc_mins, 10, 4, 2, up(mins)).
table(up_minc_maxs_mins, 10, 3, 2, up(mins)).
table(up_maxc_s_mins, 10, 3, 2, up(mins)).
table(up_maxc_maxs_mins, 10, 3, 2, up(mins)).
table(up_s_maxs_mins, 10, 3, 2, up(mins)).
table(up_v_c_minc_mins, 10, 6, 3, up(mins)).
table(up_v_c_maxc_mins, 10, 7, 3, up(mins)).
table(up_v_c_s_mins, 10, 4, 3, up(mins)).
table(up_v_c_maxs_mins, 10, 4, 3, up(mins)).
table(up_v_minc_maxc_mins, 10, 5, 3, up(mins)).
table(up_v_minc_s_mins, 10, 4, 3, up(mins)).
table(up_v_minc_maxs_mins, 10, 4, 3, up(mins)).
table(up_v_maxc_s_mins, 10, 4, 3, up(mins)).
table(up_v_maxc_maxs_mins, 10, 4, 3, up(mins)).
table(up_v_s_maxs_mins, 10, 4, 3, up(mins)).
table(up_c_minc_maxc_mins, 10, 5, 3, up(mins)).
table(up_c_minc_maxs_mins, 10, 4, 3, up(mins)).
table(up_c_maxc_s_mins, 10, 4, 3, up(mins)).
table(up_c_maxc_maxs_mins, 10, 4, 3, up(mins)).
table(up_c_s_maxs_mins, 10, 4, 3, up(mins)).
table(up_minc_maxc_s_mins, 10, 5, 3, up(mins)).
table(up_minc_maxc_maxs_mins, 10, 5, 3, up(mins)).
table(up_minc_s_maxs_mins, 10, 4, 3, up(mins)).
table(up_maxc_s_maxs_mins, 10, 4, 3, up(mins)).
table(up_v_c_minc_maxc_mins, 10, 7, 4, up(mins)).
table(up_v_c_minc_s_mins, 10, 7, 4, up(mins)).
table(up_v_c_minc_maxs_mins, 10, 7, 4, up(mins)).
table(up_v_c_maxc_s_mins, 10, 5, 4, up(mins)).
table(up_v_c_maxc_maxs_mins, 10, 5, 4, up(mins)).
table(up_v_c_s_maxs_mins, 10, 5, 4, up(mins)).
table(up_v_minc_maxc_s_mins, 10, 6, 4, up(mins)).
table(up_v_minc_maxc_maxs_mins, 10, 6, 4, up(mins)).
table(up_v_minc_s_maxs_mins, 10, 5, 4, up(mins)).
table(up_v_maxc_s_maxs_mins, 10, 5, 4, up(mins)).
table(up_c_minc_maxc_s_mins, 10, 6, 4, up(mins)).
table(up_c_minc_maxc_maxs_mins, 10, 6, 4, up(mins)).
table(up_c_minc_s_maxs_mins, 10, 5, 4, up(mins)).
table(up_c_maxc_s_maxs_mins, 10, 5, 4, up(mins)).
table(up_minc_maxc_s_maxs_mins, 10, 6, 4, up(mins)).
table(up_v_c_minc_maxc_s_mins, 10, 8, 5, up(mins)).
table(up_v_c_minc_maxc_maxs_mins, 10, 8, 5, up(mins)).
table(up_v_c_minc_s_maxs_mins, 10, 8, 5, up(mins)).
table(up_v_c_maxc_s_maxs_mins, 10, 6, 5, up(mins)).
table(up_v_minc_maxc_s_maxs_mins, 10, 7, 5, up(mins)).
table(up_c_minc_maxc_s_maxs_mins, 10, 7, 5, up(mins)).
table(up_v_c_minc_maxc_s_maxs_mins, 10, 9, 6, up(mins)).
table(low_v_maxs, 10, 7, 1, low(maxs)).
table(low_c_maxs, 10, 4, 1, low(maxs)).
table(low_minc_maxs, 10, 4, 1, low(maxs)).
table(low_maxc_maxs, 10, 4, 1, low(maxs)).
table(low_s_maxs, 10, 7, 1, low(maxs)).
table(low_mins_maxs, 10, 2, 1, low(maxs)).
table(low_v_c_maxs, 10, 8, 2, low(maxs)).
table(low_v_minc_maxs, 10, 8, 2, low(maxs)).
table(low_v_maxc_maxs, 10, 8, 2, low(maxs)).
table(low_v_s_maxs, 10, 3, 2, low(maxs)).
table(low_v_mins_maxs, 10, 7, 2, low(maxs)).
table(low_c_minc_maxs, 10, 5, 2, low(maxs)).
table(low_c_maxc_maxs, 10, 5, 2, low(maxs)).
table(low_c_s_maxs, 10, 8, 2, low(maxs)).
table(low_c_mins_maxs, 10, 3, 2, low(maxs)).
table(low_minc_maxc_maxs, 10, 6, 2, low(maxs)).
table(low_minc_s_maxs, 10, 3, 2, low(maxs)).
table(low_minc_mins_maxs, 10, 3, 2, low(maxs)).
table(low_maxc_s_maxs, 10, 3, 2, low(maxs)).
table(low_maxc_mins_maxs, 10, 3, 2, low(maxs)).
table(low_s_mins_maxs, 10, 7, 2, low(maxs)).
table(low_v_c_minc_maxs, 10, 11, 3, low(maxs)).
table(low_v_c_maxc_maxs, 10, 9, 3, low(maxs)).
table(low_v_c_s_maxs, 10, 4, 3, low(maxs)).
table(low_v_c_mins_maxs, 10, 8, 3, low(maxs)).
table(low_v_minc_maxc_maxs, 10, 10, 3, low(maxs)).
table(low_v_minc_s_maxs, 10, 4, 3, low(maxs)).
table(low_v_minc_mins_maxs, 10, 8, 3, low(maxs)).
table(low_v_maxc_s_maxs, 10, 4, 3, low(maxs)).
table(low_v_maxc_mins_maxs, 10, 8, 3, low(maxs)).
table(low_v_s_mins_maxs, 10, 4, 3, low(maxs)).
table(low_c_minc_maxc_maxs, 10, 7, 3, low(maxs)).
table(low_c_minc_s_maxs, 10, 4, 3, low(maxs)).
table(low_c_minc_mins_maxs, 10, 4, 3, low(maxs)).
table(low_c_maxc_s_maxs, 10, 4, 3, low(maxs)).
table(low_c_maxc_mins_maxs, 10, 4, 3, low(maxs)).
table(low_c_s_mins_maxs, 10, 8, 3, low(maxs)).
table(low_minc_maxc_s_maxs, 10, 5, 3, low(maxs)).
table(low_minc_maxc_mins_maxs, 10, 5, 3, low(maxs)).
table(low_minc_s_mins_maxs, 10, 4, 3, low(maxs)).
table(low_maxc_s_mins_maxs, 10, 4, 3, low(maxs)).
table(low_v_c_minc_maxc_maxs, 10, 12, 4, low(maxs)).
table(low_v_c_minc_s_maxs, 10, 7, 4, low(maxs)).
table(low_v_c_minc_mins_maxs, 10, 11, 4, low(maxs)).
table(low_v_c_maxc_s_maxs, 10, 5, 4, low(maxs)).
table(low_v_c_maxc_mins_maxs, 10, 9, 4, low(maxs)).
table(low_v_c_s_mins_maxs, 10, 5, 4, low(maxs)).
table(low_v_minc_maxc_s_maxs, 10, 6, 4, low(maxs)).
table(low_v_minc_maxc_mins_maxs, 10, 10, 4, low(maxs)).
table(low_v_minc_s_mins_maxs, 10, 5, 4, low(maxs)).
table(low_v_maxc_s_mins_maxs, 10, 5, 4, low(maxs)).
table(low_c_minc_maxc_s_maxs, 10, 6, 4, low(maxs)).
table(low_c_minc_maxc_mins_maxs, 10, 6, 4, low(maxs)).
table(low_c_minc_s_mins_maxs, 10, 5, 4, low(maxs)).
table(low_c_maxc_s_mins_maxs, 10, 5, 4, low(maxs)).
table(low_minc_maxc_s_mins_maxs, 10, 6, 4, low(maxs)).
table(low_v_c_minc_maxc_s_maxs, 10, 8, 5, low(maxs)).
table(low_v_c_minc_maxc_mins_maxs, 10, 12, 5, low(maxs)).
table(low_v_c_minc_s_mins_maxs, 10, 8, 5, low(maxs)).
table(low_v_c_maxc_s_mins_maxs, 10, 6, 5, low(maxs)).
table(low_v_minc_maxc_s_mins_maxs, 10, 7, 5, low(maxs)).
table(low_c_minc_maxc_s_mins_maxs, 10, 7, 5, low(maxs)).
table(low_v_c_minc_maxc_s_mins_maxs, 10, 9, 6, low(maxs)).
table(up_v_maxs, 10, 18, 1, up(maxs)).
table(up_maxc_maxs, 10, 2, 1, up(maxs)).
table(up_v_c_maxs, 10, 18, 2, up(maxs)).
table(up_v_minc_maxs, 10, 12, 2, up(maxs)).
table(up_v_maxc_maxs, 10, 3, 2, up(maxs)).
table(up_v_s_maxs, 10, 7, 2, up(maxs)).
table(up_v_mins_maxs, 10, 9, 2, up(maxs)).
table(up_c_maxc_maxs, 10, 3, 2, up(maxs)).
table(up_minc_maxc_maxs, 10, 4, 2, up(maxs)).
table(up_maxc_s_maxs, 10, 3, 2, up(maxs)).
table(up_maxc_mins_maxs, 10, 3, 2, up(maxs)).
table(up_v_c_minc_maxs, 10, 12, 3, up(maxs)).
table(up_v_c_maxc_maxs, 10, 4, 3, up(maxs)).
table(up_v_c_s_maxs, 10, 8, 3, up(maxs)).
table(up_v_c_mins_maxs, 10, 18, 3, up(maxs)).
table(up_v_minc_maxc_maxs, 10, 5, 3, up(maxs)).
table(up_v_minc_s_maxs, 10, 5, 3, up(maxs)).
table(up_v_minc_mins_maxs, 10, 14, 3, up(maxs)).
table(up_v_maxc_s_maxs, 10, 4, 3, up(maxs)).
table(up_v_maxc_mins_maxs, 10, 4, 3, up(maxs)).
table(up_v_s_mins_maxs, 10, 7, 3, up(maxs)).
table(up_c_minc_maxc_maxs, 10, 5, 3, up(maxs)).
table(up_c_maxc_s_maxs, 10, 4, 3, up(maxs)).
table(up_c_maxc_mins_maxs, 10, 4, 3, up(maxs)).
table(up_minc_maxc_s_maxs, 10, 5, 3, up(maxs)).
table(up_minc_maxc_mins_maxs, 10, 5, 3, up(maxs)).
table(up_maxc_s_mins_maxs, 10, 4, 3, up(maxs)).
table(up_v_c_minc_maxc_maxs, 10, 7, 4, up(maxs)).
table(up_v_c_minc_s_maxs, 10, 7, 4, up(maxs)).
table(up_v_c_minc_mins_maxs, 10, 13, 4, up(maxs)).
table(up_v_c_maxc_s_maxs, 10, 5, 4, up(maxs)).
table(up_v_c_maxc_mins_maxs, 10, 5, 4, up(maxs)).
table(up_v_c_s_mins_maxs, 10, 8, 4, up(maxs)).
table(up_v_minc_maxc_s_maxs, 10, 6, 4, up(maxs)).
table(up_v_minc_maxc_mins_maxs, 10, 6, 4, up(maxs)).
table(up_v_minc_s_mins_maxs, 10, 6, 4, up(maxs)).
table(up_v_maxc_s_mins_maxs, 10, 5, 4, up(maxs)).
table(up_c_minc_maxc_s_maxs, 10, 6, 4, up(maxs)).
table(up_c_minc_maxc_mins_maxs, 10, 6, 4, up(maxs)).
table(up_c_maxc_s_mins_maxs, 10, 5, 4, up(maxs)).
table(up_minc_maxc_s_mins_maxs, 10, 6, 4, up(maxs)).
table(up_v_c_minc_maxc_s_maxs, 10, 8, 5, up(maxs)).
table(up_v_c_minc_maxc_mins_maxs, 10, 8, 5, up(maxs)).
table(up_v_c_minc_s_mins_maxs, 10, 8, 5, up(maxs)).
table(up_v_c_maxc_s_mins_maxs, 10, 6, 5, up(maxs)).
table(up_v_minc_maxc_s_mins_maxs, 10, 7, 5, up(maxs)).
table(up_c_minc_maxc_s_mins_maxs, 10, 7, 5, up(maxs)).
table(up_v_c_minc_maxc_s_mins_maxs, 10, 9, 6, up(maxs)).
table(low_v_mina, 10, 12, 1, low(mina)).
table(low_c_mina, 10, 7, 1, low(mina)).
table(low_minc_mina, 10, 19, 1, low(mina)).
table(low_maxc_mina, 10, 19, 1, low(mina)).
table(low_s_mina, 10, 12, 1, low(mina)).
table(low_mins_mina, 10, 11, 1, low(mina)).
table(low_maxs_mina, 10, 11, 1, low(mina)).
table(low_v_c_mina, 10, 10, 2, low(mina)).
table(low_v_minc_mina, 10, 12, 2, low(mina)).
table(low_v_maxc_mina, 10, 11, 2, low(mina)).
table(low_v_s_mina, 10, 4, 2, low(mina)).
table(low_v_mins_mina, 10, 4, 2, low(mina)).
table(low_v_maxs_mina, 10, 10, 2, low(mina)).
table(low_c_minc_mina, 10, 8, 2, low(mina)).
table(low_c_maxc_mina, 10, 6, 2, low(mina)).
table(low_c_s_mina, 10, 10, 2, low(mina)).
table(low_c_mins_mina, 10, 7, 2, low(mina)).
table(low_c_maxs_mina, 10, 5, 2, low(mina)).
table(low_minc_maxc_mina, 10, 19, 2, low(mina)).
table(low_minc_s_mina, 10, 12, 2, low(mina)).
table(low_minc_mins_mina, 10, 19, 2, low(mina)).
table(low_minc_maxs_mina, 10, 19, 2, low(mina)).
table(low_maxc_s_mina, 10, 11, 2, low(mina)).
table(low_maxc_mins_mina, 10, 19, 2, low(mina)).
table(low_maxc_maxs_mina, 10, 19, 2, low(mina)).
table(low_s_mins_mina, 10, 12, 2, low(mina)).
table(low_s_maxs_mina, 10, 10, 2, low(mina)).
table(low_mins_maxs_mina, 10, 6, 2, low(mina)).
table(low_v_c_minc_mina, 10, 13, 3, low(mina)).
table(low_v_c_maxc_mina, 10, 13, 3, low(mina)).
table(low_v_c_s_mina, 10, 4, 3, low(mina)).
table(low_v_c_mins_mina, 10, 7, 3, low(mina)).
table(low_v_c_maxs_mina, 10, 10, 3, low(mina)).
table(low_v_minc_maxc_mina, 10, 13, 3, low(mina)).
table(low_v_minc_s_mina, 10, 4, 3, low(mina)).
table(low_v_minc_mins_mina, 10, 5, 3, low(mina)).
table(low_v_minc_maxs_mina, 10, 11, 3, low(mina)).
table(low_v_maxc_s_mina, 10, 4, 3, low(mina)).
table(low_v_maxc_mins_mina, 10, 6, 3, low(mina)).
table(low_v_maxc_maxs_mina, 10, 10, 3, low(mina)).
table(low_v_s_mins_mina, 10, 5, 3, low(mina)).
table(low_v_s_maxs_mina, 10, 5, 3, low(mina)).
table(low_v_mins_maxs_mina, 10, 5, 3, low(mina)).
table(low_c_minc_maxc_mina, 10, 8, 3, low(mina)).
table(low_c_minc_s_mina, 10, 11, 3, low(mina)).
table(low_c_minc_mins_mina, 10, 8, 3, low(mina)).
table(low_c_minc_maxs_mina, 10, 6, 3, low(mina)).
table(low_c_maxc_s_mina, 10, 8, 3, low(mina)).
table(low_c_maxc_mins_mina, 10, 6, 3, low(mina)).
table(low_c_maxc_maxs_mina, 10, 6, 3, low(mina)).
table(low_c_s_mins_mina, 10, 10, 3, low(mina)).
table(low_c_s_maxs_mina, 10, 10, 3, low(mina)).
table(low_c_mins_maxs_mina, 10, 5, 3, low(mina)).
table(low_minc_maxc_s_mina, 10, 10, 3, low(mina)).
table(low_minc_maxc_mins_mina, 10, 16, 3, low(mina)).
table(low_minc_maxc_maxs_mina, 10, 18, 3, low(mina)).
table(low_minc_s_mins_mina, 10, 7, 3, low(mina)).
table(low_minc_s_maxs_mina, 10, 8, 3, low(mina)).
table(low_minc_mins_maxs_mina, 10, 19, 3, low(mina)).
table(low_maxc_s_mins_mina, 10, 8, 3, low(mina)).
table(low_maxc_s_maxs_mina, 10, 7, 3, low(mina)).
table(low_maxc_mins_maxs_mina, 10, 9, 3, low(mina)).
table(low_s_mins_maxs_mina, 10, 10, 3, low(mina)).
table(low_v_c_minc_maxc_mina, 10, 19, 4, low(mina)).
table(low_v_c_minc_s_mina, 10, 8, 4, low(mina)).
table(low_v_c_minc_mins_mina, 10, 10, 4, low(mina)).
table(low_v_c_minc_maxs_mina, 10, 13, 4, low(mina)).
table(low_v_c_maxc_s_mina, 10, 5, 4, low(mina)).
table(low_v_c_maxc_mins_mina, 10, 10, 4, low(mina)).
table(low_v_c_maxc_maxs_mina, 10, 13, 4, low(mina)).
table(low_v_c_s_mins_mina, 10, 5, 4, low(mina)).
table(low_v_c_s_maxs_mina, 10, 5, 4, low(mina)).
table(low_v_c_mins_maxs_mina, 10, 10, 4, low(mina)).
table(low_v_minc_maxc_s_mina, 10, 7, 4, low(mina)).
table(low_v_minc_maxc_mins_mina, 10, 7, 4, low(mina)).
table(low_v_minc_maxc_maxs_mina, 10, 12, 4, low(mina)).
table(low_v_minc_s_mins_mina, 10, 5, 4, low(mina)).
table(low_v_minc_s_maxs_mina, 10, 6, 4, low(mina)).
table(low_v_minc_mins_maxs_mina, 10, 6, 4, low(mina)).
table(low_v_maxc_s_mins_mina, 10, 5, 4, low(mina)).
table(low_v_maxc_s_maxs_mina, 10, 5, 4, low(mina)).
table(low_v_maxc_mins_maxs_mina, 10, 5, 4, low(mina)).
table(low_v_s_mins_maxs_mina, 10, 6, 4, low(mina)).
table(low_c_minc_maxc_s_mina, 10, 10, 4, low(mina)).
table(low_c_minc_maxc_mins_mina, 10, 7, 4, low(mina)).
table(low_c_minc_maxc_maxs_mina, 10, 8, 4, low(mina)).
table(low_c_minc_s_mins_mina, 10, 10, 4, low(mina)).
table(low_c_minc_s_maxs_mina, 10, 10, 4, low(mina)).
table(low_c_minc_mins_maxs_mina, 10, 6, 4, low(mina)).
table(low_c_maxc_s_mins_mina, 10, 6, 4, low(mina)).
table(low_c_maxc_s_maxs_mina, 10, 6, 4, low(mina)).
table(low_c_maxc_mins_maxs_mina, 10, 6, 4, low(mina)).
table(low_c_s_mins_maxs_mina, 10, 10, 4, low(mina)).
table(low_minc_maxc_s_mins_mina, 10, 9, 4, low(mina)).
table(low_minc_maxc_s_maxs_mina, 10, 9, 4, low(mina)).
table(low_minc_maxc_mins_maxs_mina, 10, 18, 4, low(mina)).
table(low_minc_s_mins_maxs_mina, 10, 7, 4, low(mina)).
table(low_maxc_s_mins_maxs_mina, 10, 6, 4, low(mina)).
table(low_v_c_minc_maxc_s_mina, 10, 9, 5, low(mina)).
table(low_v_c_minc_maxc_mins_mina, 10, 16, 5, low(mina)).
table(low_v_c_minc_maxc_maxs_mina, 10, 18, 5, low(mina)).
table(low_v_c_minc_s_mins_mina, 10, 8, 5, low(mina)).
table(low_v_c_minc_s_maxs_mina, 10, 8, 5, low(mina)).
table(low_v_c_minc_mins_maxs_mina, 10, 13, 5, low(mina)).
table(low_v_c_maxc_s_mins_mina, 10, 6, 5, low(mina)).
table(low_v_c_maxc_s_maxs_mina, 10, 6, 5, low(mina)).
table(low_v_c_maxc_mins_maxs_mina, 10, 13, 5, low(mina)).
table(low_v_c_s_mins_maxs_mina, 10, 6, 5, low(mina)).
table(low_v_minc_maxc_s_mins_mina, 10, 7, 5, low(mina)).
table(low_v_minc_maxc_s_maxs_mina, 10, 7, 5, low(mina)).
table(low_v_minc_maxc_mins_maxs_mina, 10, 7, 5, low(mina)).
table(low_v_minc_s_mins_maxs_mina, 10, 7, 5, low(mina)).
table(low_v_maxc_s_mins_maxs_mina, 10, 6, 5, low(mina)).
table(low_c_minc_maxc_s_mins_mina, 10, 9, 5, low(mina)).
table(low_c_minc_maxc_s_maxs_mina, 10, 9, 5, low(mina)).
table(low_c_minc_maxc_mins_maxs_mina, 10, 8, 5, low(mina)).
table(low_c_minc_s_mins_maxs_mina, 10, 13, 5, low(mina)).
table(low_c_maxc_s_mins_maxs_mina, 10, 8, 5, low(mina)).
table(low_minc_maxc_s_mins_maxs_mina, 10, 12, 5, low(mina)).
table(low_v_c_minc_maxc_s_mins_mina, 10, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_maxs_mina, 10, 9, 6, low(mina)).
table(low_v_c_minc_maxc_mins_maxs_mina, 10, 18, 6, low(mina)).
table(low_v_c_minc_s_mins_maxs_mina, 10, 9, 6, low(mina)).
table(low_v_c_maxc_s_mins_maxs_mina, 10, 7, 6, low(mina)).
table(low_v_minc_maxc_s_mins_maxs_mina, 10, 8, 6, low(mina)).
table(low_c_minc_maxc_s_mins_maxs_mina, 10, 11, 6, low(mina)).
table(low_v_c_minc_maxc_s_mins_maxs_mina, 10, 10, 7, low(mina)).
table(up_v_maxa, 10, 19, 1, up(maxa)).
table(up_v_c_maxa, 10, 19, 2, up(maxa)).
table(up_v_minc_maxa, 10, 19, 2, up(maxa)).
table(up_v_maxc_maxa, 10, 19, 2, up(maxa)).
table(up_v_s_maxa, 10, 19, 2, up(maxa)).
table(up_v_mins_maxa, 10, 19, 2, up(maxa)).
table(up_v_maxs_maxa, 10, 19, 2, up(maxa)).
table(up_c_maxc_maxa, 10, 19, 2, up(maxa)).
table(up_maxc_s_maxa, 10, 18, 2, up(maxa)).
table(up_s_maxs_maxa, 10, 19, 2, up(maxa)).
table(up_v_c_minc_maxa, 10, 19, 3, up(maxa)).
table(up_v_c_maxc_maxa, 10, 19, 3, up(maxa)).
table(up_v_c_s_maxa, 10, 19, 3, up(maxa)).
table(up_v_c_mins_maxa, 10, 19, 3, up(maxa)).
table(up_v_c_maxs_maxa, 10, 7, 3, up(maxa)).
table(up_v_minc_maxc_maxa, 10, 19, 3, up(maxa)).
table(up_v_minc_s_maxa, 10, 18, 3, up(maxa)).
table(up_v_minc_mins_maxa, 10, 19, 3, up(maxa)).
table(up_v_minc_maxs_maxa, 10, 19, 3, up(maxa)).
table(up_v_maxc_s_maxa, 10, 18, 3, up(maxa)).
table(up_v_maxc_mins_maxa, 10, 19, 3, up(maxa)).
table(up_v_maxc_maxs_maxa, 10, 9, 3, up(maxa)).
table(up_v_s_mins_maxa, 10, 19, 3, up(maxa)).
table(up_v_s_maxs_maxa, 10, 19, 3, up(maxa)).
table(up_v_mins_maxs_maxa, 10, 19, 3, up(maxa)).
table(up_c_minc_maxc_maxa, 10, 19, 3, up(maxa)).
table(up_c_maxc_s_maxa, 10, 18, 3, up(maxa)).
table(up_c_maxc_mins_maxa, 10, 19, 3, up(maxa)).
table(up_c_maxc_maxs_maxa, 10, 6, 3, up(maxa)).
table(up_c_s_maxs_maxa, 10, 19, 3, up(maxa)).
table(up_minc_maxc_s_maxa, 10, 18, 3, up(maxa)).
table(up_minc_s_maxs_maxa, 10, 18, 3, up(maxa)).
table(up_maxc_s_mins_maxa, 10, 18, 3, up(maxa)).
table(up_maxc_s_maxs_maxa, 10, 18, 3, up(maxa)).
table(up_s_mins_maxs_maxa, 10, 19, 3, up(maxa)).
table(up_v_c_minc_maxc_maxa, 10, 19, 4, up(maxa)).
table(up_v_c_minc_s_maxa, 10, 18, 4, up(maxa)).
table(up_v_c_minc_mins_maxa, 10, 19, 4, up(maxa)).
table(up_v_c_minc_maxs_maxa, 10, 8, 4, up(maxa)).
table(up_v_c_maxc_s_maxa, 10, 18, 4, up(maxa)).
table(up_v_c_maxc_mins_maxa, 10, 19, 4, up(maxa)).
table(up_v_c_maxc_maxs_maxa, 10, 6, 4, up(maxa)).
table(up_v_c_s_mins_maxa, 10, 19, 4, up(maxa)).
table(up_v_c_s_maxs_maxa, 10, 19, 4, up(maxa)).
table(up_v_c_mins_maxs_maxa, 10, 11, 4, up(maxa)).
table(up_v_minc_maxc_s_maxa, 10, 18, 4, up(maxa)).
table(up_v_minc_maxc_mins_maxa, 10, 19, 4, up(maxa)).
table(up_v_minc_maxc_maxs_maxa, 10, 10, 4, up(maxa)).
table(up_v_minc_s_mins_maxa, 10, 18, 4, up(maxa)).
table(up_v_minc_s_maxs_maxa, 10, 18, 4, up(maxa)).
table(up_v_minc_mins_maxs_maxa, 10, 18, 4, up(maxa)).
table(up_v_maxc_s_mins_maxa, 10, 18, 4, up(maxa)).
table(up_v_maxc_s_maxs_maxa, 10, 18, 4, up(maxa)).
table(up_v_maxc_mins_maxs_maxa, 10, 8, 4, up(maxa)).
table(up_v_s_mins_maxs_maxa, 10, 19, 4, up(maxa)).
table(up_c_minc_maxc_s_maxa, 10, 18, 4, up(maxa)).
table(up_c_minc_maxc_mins_maxa, 10, 19, 4, up(maxa)).
table(up_c_minc_maxc_maxs_maxa, 10, 10, 4, up(maxa)).
table(up_c_minc_s_maxs_maxa, 10, 12, 4, up(maxa)).
table(up_c_maxc_s_mins_maxa, 10, 18, 4, up(maxa)).
table(up_c_maxc_s_maxs_maxa, 10, 18, 4, up(maxa)).
table(up_c_maxc_mins_maxs_maxa, 10, 6, 4, up(maxa)).
table(up_c_s_mins_maxs_maxa, 10, 19, 4, up(maxa)).
table(up_minc_maxc_s_mins_maxa, 10, 18, 4, up(maxa)).
table(up_minc_maxc_s_maxs_maxa, 10, 10, 4, up(maxa)).
table(up_minc_s_mins_maxs_maxa, 10, 18, 4, up(maxa)).
table(up_maxc_s_mins_maxs_maxa, 10, 18, 4, up(maxa)).
table(up_v_c_minc_maxc_s_maxa, 10, 18, 5, up(maxa)).
table(up_v_c_minc_maxc_mins_maxa, 10, 19, 5, up(maxa)).
table(up_v_c_minc_maxc_maxs_maxa, 10, 10, 5, up(maxa)).
table(up_v_c_minc_s_mins_maxa, 10, 18, 5, up(maxa)).
table(up_v_c_minc_s_maxs_maxa, 10, 12, 5, up(maxa)).
table(up_v_c_minc_mins_maxs_maxa, 10, 8, 5, up(maxa)).
table(up_v_c_maxc_s_mins_maxa, 10, 18, 5, up(maxa)).
table(up_v_c_maxc_s_maxs_maxa, 10, 18, 5, up(maxa)).
table(up_v_c_maxc_mins_maxs_maxa, 10, 6, 5, up(maxa)).
table(up_v_c_s_mins_maxs_maxa, 10, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_mins_maxa, 10, 18, 5, up(maxa)).
table(up_v_minc_maxc_s_maxs_maxa, 10, 18, 5, up(maxa)).
table(up_v_minc_maxc_mins_maxs_maxa, 10, 9, 5, up(maxa)).
table(up_v_minc_s_mins_maxs_maxa, 10, 18, 5, up(maxa)).
table(up_v_maxc_s_mins_maxs_maxa, 10, 18, 5, up(maxa)).
table(up_c_minc_maxc_s_mins_maxa, 10, 18, 5, up(maxa)).
table(up_c_minc_maxc_s_maxs_maxa, 10, 18, 5, up(maxa)).
table(up_c_minc_maxc_mins_maxs_maxa, 10, 9, 5, up(maxa)).
table(up_c_minc_s_mins_maxs_maxa, 10, 12, 5, up(maxa)).
table(up_c_maxc_s_mins_maxs_maxa, 10, 18, 5, up(maxa)).
table(up_minc_maxc_s_mins_maxs_maxa, 10, 10, 5, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxa, 10, 18, 6, up(maxa)).
table(up_v_c_minc_maxc_s_maxs_maxa, 10, 18, 6, up(maxa)).
table(up_v_c_minc_maxc_mins_maxs_maxa, 10, 9, 6, up(maxa)).
table(up_v_c_minc_s_mins_maxs_maxa, 10, 12, 6, up(maxa)).
table(up_v_c_maxc_s_mins_maxs_maxa, 10, 18, 6, up(maxa)).
table(up_v_minc_maxc_s_mins_maxs_maxa, 10, 18, 6, up(maxa)).
table(up_c_minc_maxc_s_mins_maxs_maxa, 10, 18, 6, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxs_maxa, 10, 18, 7, up(maxa)).
table(low_c_v, 11, 18, 1, low(v)).
table(low_minc_v, 11, 11, 1, low(v)).
table(low_maxc_v, 11, 11, 1, low(v)).
table(low_s_v, 11, 7, 1, low(v)).
table(low_mins_v, 11, 18, 1, low(v)).
table(low_maxs_v, 11, 18, 1, low(v)).
table(low_c_minc_v, 11, 11, 2, low(v)).
table(low_c_maxc_v, 11, 11, 2, low(v)).
table(low_c_s_v, 11, 8, 2, low(v)).
table(low_c_mins_v, 11, 18, 2, low(v)).
table(low_c_maxs_v, 11, 18, 2, low(v)).
table(low_minc_maxc_v, 11, 11, 2, low(v)).
table(low_minc_s_v, 11, 3, 2, low(v)).
table(low_minc_mins_v, 11, 12, 2, low(v)).
table(low_minc_maxs_v, 11, 12, 2, low(v)).
table(low_maxc_s_v, 11, 3, 2, low(v)).
table(low_maxc_mins_v, 11, 12, 2, low(v)).
table(low_maxc_maxs_v, 11, 13, 2, low(v)).
table(low_s_mins_v, 11, 7, 2, low(v)).
table(low_s_maxs_v, 11, 7, 2, low(v)).
table(low_mins_maxs_v, 11, 12, 2, low(v)).
table(low_c_minc_maxc_v, 11, 11, 3, low(v)).
table(low_c_minc_s_v, 11, 6, 3, low(v)).
table(low_c_minc_mins_v, 11, 12, 3, low(v)).
table(low_c_minc_maxs_v, 11, 12, 3, low(v)).
table(low_c_maxc_s_v, 11, 4, 3, low(v)).
table(low_c_maxc_mins_v, 11, 12, 3, low(v)).
table(low_c_maxc_maxs_v, 11, 13, 3, low(v)).
table(low_c_s_mins_v, 11, 8, 3, low(v)).
table(low_c_s_maxs_v, 11, 8, 3, low(v)).
table(low_c_mins_maxs_v, 11, 18, 3, low(v)).
table(low_minc_maxc_s_v, 11, 5, 3, low(v)).
table(low_minc_maxc_mins_v, 11, 12, 3, low(v)).
table(low_minc_maxc_maxs_v, 11, 12, 3, low(v)).
table(low_minc_s_mins_v, 11, 4, 3, low(v)).
table(low_minc_s_maxs_v, 11, 4, 3, low(v)).
table(low_minc_mins_maxs_v, 11, 13, 3, low(v)).
table(low_maxc_s_mins_v, 11, 4, 3, low(v)).
table(low_maxc_s_maxs_v, 11, 4, 3, low(v)).
table(low_maxc_mins_maxs_v, 11, 14, 3, low(v)).
table(low_s_mins_maxs_v, 11, 7, 3, low(v)).
table(low_c_minc_maxc_s_v, 11, 7, 4, low(v)).
table(low_c_minc_maxc_mins_v, 11, 12, 4, low(v)).
table(low_c_minc_maxc_maxs_v, 11, 12, 4, low(v)).
table(low_c_minc_s_mins_v, 11, 7, 4, low(v)).
table(low_c_minc_s_maxs_v, 11, 7, 4, low(v)).
table(low_c_minc_mins_maxs_v, 11, 13, 4, low(v)).
table(low_c_maxc_s_mins_v, 11, 5, 4, low(v)).
table(low_c_maxc_s_maxs_v, 11, 5, 4, low(v)).
table(low_c_maxc_mins_maxs_v, 11, 14, 4, low(v)).
table(low_c_s_mins_maxs_v, 11, 8, 4, low(v)).
table(low_minc_maxc_s_mins_v, 11, 6, 4, low(v)).
table(low_minc_maxc_s_maxs_v, 11, 6, 4, low(v)).
table(low_minc_maxc_mins_maxs_v, 11, 13, 4, low(v)).
table(low_minc_s_mins_maxs_v, 11, 5, 4, low(v)).
table(low_maxc_s_mins_maxs_v, 11, 5, 4, low(v)).
table(low_c_minc_maxc_s_mins_v, 11, 8, 5, low(v)).
table(low_c_minc_maxc_s_maxs_v, 11, 8, 5, low(v)).
table(low_c_minc_maxc_mins_maxs_v, 11, 13, 5, low(v)).
table(low_c_minc_s_mins_maxs_v, 11, 8, 5, low(v)).
table(low_c_maxc_s_mins_maxs_v, 11, 6, 5, low(v)).
table(low_minc_maxc_s_mins_maxs_v, 11, 7, 5, low(v)).
table(low_c_minc_maxc_s_mins_maxs_v, 11, 9, 6, low(v)).
table(up_c_maxc_v, 11, 3, 2, up(v)).
table(up_maxc_s_v, 11, 3, 2, up(v)).
table(up_s_maxs_v, 11, 3, 2, up(v)).
table(up_c_minc_maxc_v, 11, 6, 3, up(v)).
table(up_c_maxc_s_v, 11, 4, 3, up(v)).
table(up_c_maxc_mins_v, 11, 4, 3, up(v)).
table(up_c_maxc_maxs_v, 11, 4, 3, up(v)).
table(up_c_s_maxs_v, 11, 4, 3, up(v)).
table(up_minc_maxc_s_v, 11, 5, 3, up(v)).
table(up_minc_s_maxs_v, 11, 4, 3, up(v)).
table(up_maxc_s_mins_v, 11, 4, 3, up(v)).
table(up_maxc_s_maxs_v, 11, 4, 3, up(v)).
table(up_s_mins_maxs_v, 11, 4, 3, up(v)).
table(up_c_minc_maxc_s_v, 11, 7, 4, up(v)).
table(up_c_minc_maxc_mins_v, 11, 7, 4, up(v)).
table(up_c_minc_maxc_maxs_v, 11, 7, 4, up(v)).
table(up_c_minc_s_maxs_v, 11, 7, 4, up(v)).
table(up_c_maxc_s_mins_v, 11, 5, 4, up(v)).
table(up_c_maxc_s_maxs_v, 11, 5, 4, up(v)).
table(up_c_maxc_mins_maxs_v, 11, 5, 4, up(v)).
table(up_c_s_mins_maxs_v, 11, 5, 4, up(v)).
table(up_minc_maxc_s_mins_v, 11, 6, 4, up(v)).
table(up_minc_maxc_s_maxs_v, 11, 6, 4, up(v)).
table(up_minc_s_mins_maxs_v, 11, 5, 4, up(v)).
table(up_maxc_s_mins_maxs_v, 11, 5, 4, up(v)).
table(up_c_minc_maxc_s_mins_v, 11, 8, 5, up(v)).
table(up_c_minc_maxc_s_maxs_v, 11, 8, 5, up(v)).
table(up_c_minc_maxc_mins_maxs_v, 11, 8, 5, up(v)).
table(up_c_minc_s_mins_maxs_v, 11, 8, 5, up(v)).
table(up_c_maxc_s_mins_maxs_v, 11, 6, 5, up(v)).
table(up_minc_maxc_s_mins_maxs_v, 11, 7, 5, up(v)).
table(up_c_minc_maxc_s_mins_maxs_v, 11, 9, 6, up(v)).
table(low_v_c, 11, 11, 1, low(c)).
table(low_minc_c, 11, 11, 1, low(c)).
table(low_maxc_c, 11, 11, 1, low(c)).
table(low_s_c, 11, 8, 1, low(c)).
table(low_mins_c, 11, 8, 1, low(c)).
table(low_maxs_c, 11, 8, 1, low(c)).
table(low_v_minc_c, 11, 11, 2, low(c)).
table(low_v_maxc_c, 11, 3, 2, low(c)).
table(low_v_s_c, 11, 13, 2, low(c)).
table(low_v_mins_c, 11, 12, 2, low(c)).
table(low_v_maxs_c, 11, 13, 2, low(c)).
table(low_minc_maxc_c, 11, 11, 2, low(c)).
table(low_minc_s_c, 11, 5, 2, low(c)).
table(low_minc_mins_c, 11, 9, 2, low(c)).
table(low_minc_maxs_c, 11, 9, 2, low(c)).
table(low_maxc_s_c, 11, 3, 2, low(c)).
table(low_maxc_mins_c, 11, 12, 2, low(c)).
table(low_maxc_maxs_c, 11, 13, 2, low(c)).
table(low_s_mins_c, 11, 10, 2, low(c)).
table(low_s_maxs_c, 11, 11, 2, low(c)).
table(low_mins_maxs_c, 11, 9, 2, low(c)).
table(low_v_minc_maxc_c, 11, 6, 3, low(c)).
table(low_v_minc_s_c, 11, 12, 3, low(c)).
table(low_v_minc_mins_c, 11, 12, 3, low(c)).
table(low_v_minc_maxs_c, 11, 12, 3, low(c)).
table(low_v_maxc_s_c, 11, 4, 3, low(c)).
table(low_v_maxc_mins_c, 11, 4, 3, low(c)).
table(low_v_maxc_maxs_c, 11, 4, 3, low(c)).
table(low_v_s_mins_c, 11, 14, 3, low(c)).
table(low_v_s_maxs_c, 11, 14, 3, low(c)).
table(low_v_mins_maxs_c, 11, 14, 3, low(c)).
table(low_minc_maxc_s_c, 11, 5, 3, low(c)).
table(low_minc_maxc_mins_c, 11, 12, 3, low(c)).
table(low_minc_maxc_maxs_c, 11, 12, 3, low(c)).
table(low_minc_s_mins_c, 11, 6, 3, low(c)).
table(low_minc_s_maxs_c, 11, 7, 3, low(c)).
table(low_minc_mins_maxs_c, 11, 7, 3, low(c)).
table(low_maxc_s_mins_c, 11, 4, 3, low(c)).
table(low_maxc_s_maxs_c, 11, 4, 3, low(c)).
table(low_maxc_mins_maxs_c, 11, 6, 3, low(c)).
table(low_s_mins_maxs_c, 11, 12, 3, low(c)).
table(low_v_minc_maxc_s_c, 11, 7, 4, low(c)).
table(low_v_minc_maxc_mins_c, 11, 7, 4, low(c)).
table(low_v_minc_maxc_maxs_c, 11, 7, 4, low(c)).
table(low_v_minc_s_mins_c, 11, 13, 4, low(c)).
table(low_v_minc_s_maxs_c, 11, 13, 4, low(c)).
table(low_v_minc_mins_maxs_c, 11, 13, 4, low(c)).
table(low_v_maxc_s_mins_c, 11, 5, 4, low(c)).
table(low_v_maxc_s_maxs_c, 11, 5, 4, low(c)).
table(low_v_maxc_mins_maxs_c, 11, 5, 4, low(c)).
table(low_v_s_mins_maxs_c, 11, 15, 4, low(c)).
table(low_minc_maxc_s_mins_c, 11, 6, 4, low(c)).
table(low_minc_maxc_s_maxs_c, 11, 6, 4, low(c)).
table(low_minc_maxc_mins_maxs_c, 11, 13, 4, low(c)).
table(low_minc_s_mins_maxs_c, 11, 8, 4, low(c)).
table(low_maxc_s_mins_maxs_c, 11, 5, 4, low(c)).
table(low_v_minc_maxc_s_mins_c, 11, 8, 5, low(c)).
table(low_v_minc_maxc_s_maxs_c, 11, 8, 5, low(c)).
table(low_v_minc_maxc_mins_maxs_c, 11, 8, 5, low(c)).
table(low_v_minc_s_mins_maxs_c, 11, 14, 5, low(c)).
table(low_v_maxc_s_mins_maxs_c, 11, 6, 5, low(c)).
table(low_minc_maxc_s_mins_maxs_c, 11, 7, 5, low(c)).
table(low_v_minc_maxc_s_mins_maxs_c, 11, 9, 6, low(c)).
table(up_v_c, 11, 18, 1, up(c)).
table(up_s_c, 11, 3, 1, up(c)).
table(up_v_minc_c, 11, 8, 2, up(c)).
table(up_v_maxc_c, 11, 11, 2, up(c)).
table(up_v_s_c, 11, 4, 2, up(c)).
table(up_v_mins_c, 11, 12, 2, up(c)).
table(up_v_maxs_c, 11, 18, 2, up(c)).
table(up_minc_s_c, 11, 3, 2, up(c)).
table(up_maxc_s_c, 11, 4, 2, up(c)).
table(up_s_mins_c, 11, 5, 2, up(c)).
table(up_s_maxs_c, 11, 5, 2, up(c)).
table(up_v_minc_maxc_c, 11, 11, 3, up(c)).
table(up_v_minc_s_c, 11, 6, 3, up(c)).
table(up_v_minc_mins_c, 11, 9, 3, up(c)).
table(up_v_minc_maxs_c, 11, 7, 3, up(c)).
table(up_v_maxc_s_c, 11, 5, 3, up(c)).
table(up_v_maxc_mins_c, 11, 12, 3, up(c)).
table(up_v_maxc_maxs_c, 11, 13, 3, up(c)).
table(up_v_s_mins_c, 11, 8, 3, up(c)).
table(up_v_s_maxs_c, 11, 6, 3, up(c)).
table(up_v_mins_maxs_c, 11, 18, 3, up(c)).
table(up_minc_maxc_s_c, 11, 5, 3, up(c)).
table(up_minc_s_mins_c, 11, 4, 3, up(c)).
table(up_minc_s_maxs_c, 11, 4, 3, up(c)).
table(up_maxc_s_mins_c, 11, 5, 3, up(c)).
table(up_maxc_s_maxs_c, 11, 5, 3, up(c)).
table(up_s_mins_maxs_c, 11, 8, 3, up(c)).
table(up_v_minc_maxc_s_c, 11, 7, 4, up(c)).
table(up_v_minc_maxc_mins_c, 11, 12, 4, up(c)).
table(up_v_minc_maxc_maxs_c, 11, 12, 4, up(c)).
table(up_v_minc_s_mins_c, 11, 7, 4, up(c)).
table(up_v_minc_s_maxs_c, 11, 7, 4, up(c)).
table(up_v_minc_mins_maxs_c, 11, 8, 4, up(c)).
table(up_v_maxc_s_mins_c, 11, 9, 4, up(c)).
table(up_v_maxc_s_maxs_c, 11, 6, 4, up(c)).
table(up_v_maxc_mins_maxs_c, 11, 14, 4, up(c)).
table(up_v_s_mins_maxs_c, 11, 10, 4, up(c)).
table(up_minc_maxc_s_mins_c, 11, 6, 4, up(c)).
table(up_minc_maxc_s_maxs_c, 11, 6, 4, up(c)).
table(up_minc_s_mins_maxs_c, 11, 5, 4, up(c)).
table(up_maxc_s_mins_maxs_c, 11, 6, 4, up(c)).
table(up_v_minc_maxc_s_mins_c, 11, 8, 5, up(c)).
table(up_v_minc_maxc_s_maxs_c, 11, 8, 5, up(c)).
table(up_v_minc_maxc_mins_maxs_c, 11, 13, 5, up(c)).
table(up_v_minc_s_mins_maxs_c, 11, 8, 5, up(c)).
table(up_v_maxc_s_mins_maxs_c, 11, 7, 5, up(c)).
table(up_minc_maxc_s_mins_maxs_c, 11, 7, 5, up(c)).
table(up_v_minc_maxc_s_mins_maxs_c, 11, 9, 6, up(c)).
table(low_v_minc, 11, 3, 1, low(minc)).
table(low_c_minc, 11, 3, 1, low(minc)).
table(low_maxc_minc, 11, 3, 1, low(minc)).
table(low_s_minc, 11, 3, 1, low(minc)).
table(low_mins_minc, 11, 2, 1, low(minc)).
table(low_maxs_minc, 11, 3, 1, low(minc)).
table(low_v_c_minc, 11, 5, 2, low(minc)).
table(low_v_maxc_minc, 11, 4, 2, low(minc)).
table(low_v_s_minc, 11, 4, 2, low(minc)).
table(low_v_mins_minc, 11, 3, 2, low(minc)).
table(low_v_maxs_minc, 11, 4, 2, low(minc)).
table(low_c_maxc_minc, 11, 4, 2, low(minc)).
table(low_c_s_minc, 11, 4, 2, low(minc)).
table(low_c_mins_minc, 11, 3, 2, low(minc)).
table(low_c_maxs_minc, 11, 4, 2, low(minc)).
table(low_maxc_s_minc, 11, 4, 2, low(minc)).
table(low_maxc_mins_minc, 11, 4, 2, low(minc)).
table(low_maxc_maxs_minc, 11, 4, 2, low(minc)).
table(low_s_mins_minc, 11, 3, 2, low(minc)).
table(low_s_maxs_minc, 11, 4, 2, low(minc)).
table(low_mins_maxs_minc, 11, 3, 2, low(minc)).
table(low_v_c_maxc_minc, 11, 6, 3, low(minc)).
table(low_v_c_s_minc, 11, 6, 3, low(minc)).
table(low_v_c_mins_minc, 11, 6, 3, low(minc)).
table(low_v_c_maxs_minc, 11, 6, 3, low(minc)).
table(low_v_maxc_s_minc, 11, 5, 3, low(minc)).
table(low_v_maxc_mins_minc, 11, 5, 3, low(minc)).
table(low_v_maxc_maxs_minc, 11, 5, 3, low(minc)).
table(low_v_s_mins_minc, 11, 4, 3, low(minc)).
table(low_v_s_maxs_minc, 11, 5, 3, low(minc)).
table(low_v_mins_maxs_minc, 11, 4, 3, low(minc)).
table(low_c_maxc_s_minc, 11, 5, 3, low(minc)).
table(low_c_maxc_mins_minc, 11, 5, 3, low(minc)).
table(low_c_maxc_maxs_minc, 11, 5, 3, low(minc)).
table(low_c_s_mins_minc, 11, 4, 3, low(minc)).
table(low_c_s_maxs_minc, 11, 6, 3, low(minc)).
table(low_c_mins_maxs_minc, 11, 4, 3, low(minc)).
table(low_maxc_s_mins_minc, 11, 5, 3, low(minc)).
table(low_maxc_s_maxs_minc, 11, 5, 3, low(minc)).
table(low_maxc_mins_maxs_minc, 11, 5, 3, low(minc)).
table(low_s_mins_maxs_minc, 11, 4, 3, low(minc)).
table(low_v_c_maxc_s_minc, 11, 7, 4, low(minc)).
table(low_v_c_maxc_mins_minc, 11, 7, 4, low(minc)).
table(low_v_c_maxc_maxs_minc, 11, 7, 4, low(minc)).
table(low_v_c_s_mins_minc, 11, 7, 4, low(minc)).
table(low_v_c_s_maxs_minc, 11, 7, 4, low(minc)).
table(low_v_c_mins_maxs_minc, 11, 7, 4, low(minc)).
table(low_v_maxc_s_mins_minc, 11, 6, 4, low(minc)).
table(low_v_maxc_s_maxs_minc, 11, 6, 4, low(minc)).
table(low_v_maxc_mins_maxs_minc, 11, 6, 4, low(minc)).
table(low_v_s_mins_maxs_minc, 11, 5, 4, low(minc)).
table(low_c_maxc_s_mins_minc, 11, 6, 4, low(minc)).
table(low_c_maxc_s_maxs_minc, 11, 6, 4, low(minc)).
table(low_c_maxc_mins_maxs_minc, 11, 6, 4, low(minc)).
table(low_c_s_mins_maxs_minc, 11, 6, 4, low(minc)).
table(low_maxc_s_mins_maxs_minc, 11, 6, 4, low(minc)).
table(low_v_c_maxc_s_mins_minc, 11, 8, 5, low(minc)).
table(low_v_c_maxc_s_maxs_minc, 11, 8, 5, low(minc)).
table(low_v_c_maxc_mins_maxs_minc, 11, 8, 5, low(minc)).
table(low_v_c_s_mins_maxs_minc, 11, 8, 5, low(minc)).
table(low_v_maxc_s_mins_maxs_minc, 11, 7, 5, low(minc)).
table(low_c_maxc_s_mins_maxs_minc, 11, 7, 5, low(minc)).
table(low_v_c_maxc_s_mins_maxs_minc, 11, 9, 6, low(minc)).
table(up_v_minc, 11, 11, 1, up(minc)).
table(up_maxc_minc, 11, 6, 1, up(minc)).
table(up_v_c_minc, 11, 5, 2, up(minc)).
table(up_v_maxc_minc, 11, 6, 2, up(minc)).
table(up_v_s_minc, 11, 13, 2, up(minc)).
table(up_v_mins_minc, 11, 12, 2, up(minc)).
table(up_v_maxs_minc, 11, 13, 2, up(minc)).
table(up_c_maxc_minc, 11, 4, 2, up(minc)).
table(up_maxc_s_minc, 11, 6, 2, up(minc)).
table(up_maxc_mins_minc, 11, 6, 2, up(minc)).
table(up_maxc_maxs_minc, 11, 7, 2, up(minc)).
table(up_s_maxs_minc, 11, 14, 2, up(minc)).
table(up_v_c_maxc_minc, 11, 6, 3, up(minc)).
table(up_v_c_s_minc, 11, 6, 3, up(minc)).
table(up_v_c_mins_minc, 11, 6, 3, up(minc)).
table(up_v_c_maxs_minc, 11, 6, 3, up(minc)).
table(up_v_maxc_s_minc, 11, 7, 3, up(minc)).
table(up_v_maxc_mins_minc, 11, 7, 3, up(minc)).
table(up_v_maxc_maxs_minc, 11, 7, 3, up(minc)).
table(up_v_s_mins_minc, 11, 14, 3, up(minc)).
table(up_v_s_maxs_minc, 11, 14, 3, up(minc)).
table(up_v_mins_maxs_minc, 11, 14, 3, up(minc)).
table(up_c_maxc_s_minc, 11, 5, 3, up(minc)).
table(up_c_maxc_mins_minc, 11, 5, 3, up(minc)).
table(up_c_maxc_maxs_minc, 11, 5, 3, up(minc)).
table(up_c_s_maxs_minc, 11, 4, 3, up(minc)).
table(up_maxc_s_mins_minc, 11, 5, 3, up(minc)).
table(up_maxc_s_maxs_minc, 11, 7, 3, up(minc)).
table(up_maxc_mins_maxs_minc, 11, 7, 3, up(minc)).
table(up_s_mins_maxs_minc, 11, 15, 3, up(minc)).
table(up_v_c_maxc_s_minc, 11, 7, 4, up(minc)).
table(up_v_c_maxc_mins_minc, 11, 7, 4, up(minc)).
table(up_v_c_maxc_maxs_minc, 11, 7, 4, up(minc)).
table(up_v_c_s_mins_minc, 11, 7, 4, up(minc)).
table(up_v_c_s_maxs_minc, 11, 7, 4, up(minc)).
table(up_v_c_mins_maxs_minc, 11, 7, 4, up(minc)).
table(up_v_maxc_s_mins_minc, 11, 6, 4, up(minc)).
table(up_v_maxc_s_maxs_minc, 11, 6, 4, up(minc)).
table(up_v_maxc_mins_maxs_minc, 11, 6, 4, up(minc)).
table(up_v_s_mins_maxs_minc, 11, 15, 4, up(minc)).
table(up_c_maxc_s_mins_minc, 11, 6, 4, up(minc)).
table(up_c_maxc_s_maxs_minc, 11, 6, 4, up(minc)).
table(up_c_maxc_mins_maxs_minc, 11, 6, 4, up(minc)).
table(up_c_s_mins_maxs_minc, 11, 5, 4, up(minc)).
table(up_maxc_s_mins_maxs_minc, 11, 6, 4, up(minc)).
table(up_v_c_maxc_s_mins_minc, 11, 8, 5, up(minc)).
table(up_v_c_maxc_s_maxs_minc, 11, 8, 5, up(minc)).
table(up_v_c_maxc_mins_maxs_minc, 11, 8, 5, up(minc)).
table(up_v_c_s_mins_maxs_minc, 11, 8, 5, up(minc)).
table(up_v_maxc_s_mins_maxs_minc, 11, 7, 5, up(minc)).
table(up_c_maxc_s_mins_maxs_minc, 11, 7, 5, up(minc)).
table(up_v_c_maxc_s_mins_maxs_minc, 11, 9, 6, up(minc)).
table(low_v_maxc, 11, 18, 1, low(maxc)).
table(low_c_maxc, 11, 18, 1, low(maxc)).
table(low_minc_maxc, 11, 6, 1, low(maxc)).
table(low_s_maxc, 11, 18, 1, low(maxc)).
table(low_mins_maxc, 11, 9, 1, low(maxc)).
table(low_maxs_maxc, 11, 2, 1, low(maxc)).
table(low_v_c_maxc, 11, 3, 2, low(maxc)).
table(low_v_minc_maxc, 11, 6, 2, low(maxc)).
table(low_v_s_maxc, 11, 4, 2, low(maxc)).
table(low_v_mins_maxc, 11, 9, 2, low(maxc)).
table(low_v_maxs_maxc, 11, 3, 2, low(maxc)).
table(low_c_minc_maxc, 11, 11, 2, low(maxc)).
table(low_c_s_maxc, 11, 4, 2, low(maxc)).
table(low_c_mins_maxc, 11, 18, 2, low(maxc)).
table(low_c_maxs_maxc, 11, 3, 2, low(maxc)).
table(low_minc_s_maxc, 11, 6, 2, low(maxc)).
table(low_minc_mins_maxc, 11, 6, 2, low(maxc)).
table(low_minc_maxs_maxc, 11, 4, 2, low(maxc)).
table(low_s_mins_maxc, 11, 18, 2, low(maxc)).
table(low_s_maxs_maxc, 11, 3, 2, low(maxc)).
table(low_mins_maxs_maxc, 11, 3, 2, low(maxc)).
table(low_v_c_minc_maxc, 11, 6, 3, low(maxc)).
table(low_v_c_s_maxc, 11, 4, 3, low(maxc)).
table(low_v_c_mins_maxc, 11, 4, 3, low(maxc)).
table(low_v_c_maxs_maxc, 11, 4, 3, low(maxc)).
table(low_v_minc_s_maxc, 11, 5, 3, low(maxc)).
table(low_v_minc_mins_maxc, 11, 7, 3, low(maxc)).
table(low_v_minc_maxs_maxc, 11, 5, 3, low(maxc)).
table(low_v_s_mins_maxc, 11, 5, 3, low(maxc)).
table(low_v_s_maxs_maxc, 11, 4, 3, low(maxc)).
table(low_v_mins_maxs_maxc, 11, 4, 3, low(maxc)).
table(low_c_minc_s_maxc, 11, 5, 3, low(maxc)).
table(low_c_minc_mins_maxc, 11, 8, 3, low(maxc)).
table(low_c_minc_maxs_maxc, 11, 5, 3, low(maxc)).
table(low_c_s_mins_maxc, 11, 4, 3, low(maxc)).
table(low_c_s_maxs_maxc, 11, 4, 3, low(maxc)).
table(low_c_mins_maxs_maxc, 11, 4, 3, low(maxc)).
table(low_minc_s_mins_maxc, 11, 7, 3, low(maxc)).
table(low_minc_s_maxs_maxc, 11, 5, 3, low(maxc)).
table(low_minc_mins_maxs_maxc, 11, 5, 3, low(maxc)).
table(low_s_mins_maxs_maxc, 11, 4, 3, low(maxc)).
table(low_v_c_minc_s_maxc, 11, 7, 4, low(maxc)).
table(low_v_c_minc_mins_maxc, 11, 7, 4, low(maxc)).
table(low_v_c_minc_maxs_maxc, 11, 7, 4, low(maxc)).
table(low_v_c_s_mins_maxc, 11, 5, 4, low(maxc)).
table(low_v_c_s_maxs_maxc, 11, 5, 4, low(maxc)).
table(low_v_c_mins_maxs_maxc, 11, 5, 4, low(maxc)).
table(low_v_minc_s_mins_maxc, 11, 6, 4, low(maxc)).
table(low_v_minc_s_maxs_maxc, 11, 6, 4, low(maxc)).
table(low_v_minc_mins_maxs_maxc, 11, 6, 4, low(maxc)).
table(low_v_s_mins_maxs_maxc, 11, 5, 4, low(maxc)).
table(low_c_minc_s_mins_maxc, 11, 6, 4, low(maxc)).
table(low_c_minc_s_maxs_maxc, 11, 6, 4, low(maxc)).
table(low_c_minc_mins_maxs_maxc, 11, 6, 4, low(maxc)).
table(low_c_s_mins_maxs_maxc, 11, 5, 4, low(maxc)).
table(low_minc_s_mins_maxs_maxc, 11, 6, 4, low(maxc)).
table(low_v_c_minc_s_mins_maxc, 11, 8, 5, low(maxc)).
table(low_v_c_minc_s_maxs_maxc, 11, 8, 5, low(maxc)).
table(low_v_c_minc_mins_maxs_maxc, 11, 8, 5, low(maxc)).
table(low_v_c_s_mins_maxs_maxc, 11, 6, 5, low(maxc)).
table(low_v_minc_s_mins_maxs_maxc, 11, 7, 5, low(maxc)).
table(low_c_minc_s_mins_maxs_maxc, 11, 7, 5, low(maxc)).
table(low_v_c_minc_s_mins_maxs_maxc, 11, 9, 6, low(maxc)).
table(up_v_maxc, 11, 11, 1, up(maxc)).
table(up_v_c_maxc, 11, 11, 2, up(maxc)).
table(up_v_minc_maxc, 11, 11, 2, up(maxc)).
table(up_v_s_maxc, 11, 13, 2, up(maxc)).
table(up_v_mins_maxc, 11, 12, 2, up(maxc)).
table(up_v_maxs_maxc, 11, 13, 2, up(maxc)).
table(up_v_c_minc_maxc, 11, 11, 3, up(maxc)).
table(up_v_c_s_maxc, 11, 13, 3, up(maxc)).
table(up_v_c_mins_maxc, 11, 12, 3, up(maxc)).
table(up_v_c_maxs_maxc, 11, 13, 3, up(maxc)).
table(up_v_minc_s_maxc, 11, 12, 3, up(maxc)).
table(up_v_minc_mins_maxc, 11, 12, 3, up(maxc)).
table(up_v_minc_maxs_maxc, 11, 12, 3, up(maxc)).
table(up_v_s_mins_maxc, 11, 14, 3, up(maxc)).
table(up_v_s_maxs_maxc, 11, 14, 3, up(maxc)).
table(up_v_mins_maxs_maxc, 11, 14, 3, up(maxc)).
table(up_v_c_minc_s_maxc, 11, 12, 4, up(maxc)).
table(up_v_c_minc_mins_maxc, 11, 12, 4, up(maxc)).
table(up_v_c_minc_maxs_maxc, 11, 12, 4, up(maxc)).
table(up_v_c_s_mins_maxc, 11, 14, 4, up(maxc)).
table(up_v_c_s_maxs_maxc, 11, 6, 4, up(maxc)).
table(up_v_c_mins_maxs_maxc, 11, 14, 4, up(maxc)).
table(up_v_minc_s_mins_maxc, 11, 13, 4, up(maxc)).
table(up_v_minc_s_maxs_maxc, 11, 13, 4, up(maxc)).
table(up_v_minc_mins_maxs_maxc, 11, 13, 4, up(maxc)).
table(up_v_s_mins_maxs_maxc, 11, 15, 4, up(maxc)).
table(up_v_c_minc_s_mins_maxc, 11, 13, 5, up(maxc)).
table(up_v_c_minc_s_maxs_maxc, 11, 8, 5, up(maxc)).
table(up_v_c_minc_mins_maxs_maxc, 11, 13, 5, up(maxc)).
table(up_v_c_s_mins_maxs_maxc, 11, 10, 5, up(maxc)).
table(up_v_minc_s_mins_maxs_maxc, 11, 14, 5, up(maxc)).
table(up_v_c_minc_s_mins_maxs_maxc, 11, 9, 6, up(maxc)).
table(low_v_s, 11, 18, 1, low(s)).
table(low_c_s, 11, 3, 1, low(s)).
table(low_minc_s, 11, 18, 1, low(s)).
table(low_maxc_s, 11, 18, 1, low(s)).
table(low_mins_s, 11, 18, 1, low(s)).
table(low_maxs_s, 11, 18, 1, low(s)).
table(low_v_c_s, 11, 4, 2, low(s)).
table(low_v_minc_s, 11, 18, 2, low(s)).
table(low_v_maxc_s, 11, 6, 2, low(s)).
table(low_v_mins_s, 11, 9, 2, low(s)).
table(low_v_maxs_s, 11, 3, 2, low(s)).
table(low_c_minc_s, 11, 5, 2, low(s)).
table(low_c_maxc_s, 11, 5, 2, low(s)).
table(low_c_mins_s, 11, 5, 2, low(s)).
table(low_c_maxs_s, 11, 5, 2, low(s)).
table(low_minc_maxc_s, 11, 18, 2, low(s)).
table(low_minc_mins_s, 11, 12, 2, low(s)).
table(low_minc_maxs_s, 11, 14, 2, low(s)).
table(low_maxc_mins_s, 11, 7, 2, low(s)).
table(low_maxc_maxs_s, 11, 14, 2, low(s)).
table(low_mins_maxs_s, 11, 12, 2, low(s)).
table(low_v_c_minc_s, 11, 8, 3, low(s)).
table(low_v_c_maxc_s, 11, 6, 3, low(s)).
table(low_v_c_mins_s, 11, 8, 3, low(s)).
table(low_v_c_maxs_s, 11, 4, 3, low(s)).
table(low_v_minc_maxc_s, 11, 10, 3, low(s)).
table(low_v_minc_mins_s, 11, 13, 3, low(s)).
table(low_v_minc_maxs_s, 11, 4, 3, low(s)).
table(low_v_maxc_mins_s, 11, 4, 3, low(s)).
table(low_v_maxc_maxs_s, 11, 4, 3, low(s)).
table(low_v_mins_maxs_s, 11, 4, 3, low(s)).
table(low_c_minc_maxc_s, 11, 8, 3, low(s)).
table(low_c_minc_mins_s, 11, 4, 3, low(s)).
table(low_c_minc_maxs_s, 11, 5, 3, low(s)).
table(low_c_maxc_mins_s, 11, 8, 3, low(s)).
table(low_c_maxc_maxs_s, 11, 5, 3, low(s)).
table(low_c_mins_maxs_s, 11, 8, 3, low(s)).
table(low_minc_maxc_mins_s, 11, 13, 3, low(s)).
table(low_minc_maxc_maxs_s, 11, 14, 3, low(s)).
table(low_minc_mins_maxs_s, 11, 8, 3, low(s)).
table(low_maxc_mins_maxs_s, 11, 9, 3, low(s)).
table(low_v_c_minc_maxc_s, 11, 10, 4, low(s)).
table(low_v_c_minc_mins_s, 11, 7, 4, low(s)).
table(low_v_c_minc_maxs_s, 11, 7, 4, low(s)).
table(low_v_c_maxc_mins_s, 11, 5, 4, low(s)).
table(low_v_c_maxc_maxs_s, 11, 5, 4, low(s)).
table(low_v_c_mins_maxs_s, 11, 5, 4, low(s)).
table(low_v_minc_maxc_mins_s, 11, 8, 4, low(s)).
table(low_v_minc_maxc_maxs_s, 11, 6, 4, low(s)).
table(low_v_minc_mins_maxs_s, 11, 5, 4, low(s)).
table(low_v_maxc_mins_maxs_s, 11, 5, 4, low(s)).
table(low_c_minc_maxc_mins_s, 11, 6, 4, low(s)).
table(low_c_minc_maxc_maxs_s, 11, 7, 4, low(s)).
table(low_c_minc_mins_maxs_s, 11, 5, 4, low(s)).
table(low_c_maxc_mins_maxs_s, 11, 6, 4, low(s)).
table(low_minc_maxc_mins_maxs_s, 11, 14, 4, low(s)).
table(low_v_c_minc_maxc_mins_s, 11, 8, 5, low(s)).
table(low_v_c_minc_maxc_maxs_s, 11, 8, 5, low(s)).
table(low_v_c_minc_mins_maxs_s, 11, 8, 5, low(s)).
table(low_v_c_maxc_mins_maxs_s, 11, 6, 5, low(s)).
table(low_v_minc_maxc_mins_maxs_s, 11, 7, 5, low(s)).
table(low_c_minc_maxc_mins_maxs_s, 11, 7, 5, low(s)).
table(low_v_c_minc_maxc_mins_maxs_s, 11, 9, 6, low(s)).
table(up_v_s, 11, 7, 1, up(s)).
table(up_v_c_s, 11, 8, 2, up(s)).
table(up_v_minc_s, 11, 8, 2, up(s)).
table(up_v_maxc_s, 11, 8, 2, up(s)).
table(up_v_mins_s, 11, 4, 2, up(s)).
table(up_v_maxs_s, 11, 7, 2, up(s)).
table(up_c_maxc_s, 11, 9, 2, up(s)).
table(up_v_c_minc_s, 11, 11, 3, up(s)).
table(up_v_c_maxc_s, 11, 9, 3, up(s)).
table(up_v_c_mins_s, 11, 5, 3, up(s)).
table(up_v_c_maxs_s, 11, 8, 3, up(s)).
table(up_v_minc_maxc_s, 11, 10, 3, up(s)).
table(up_v_minc_mins_s, 11, 5, 3, up(s)).
table(up_v_minc_maxs_s, 11, 8, 3, up(s)).
table(up_v_maxc_mins_s, 11, 5, 3, up(s)).
table(up_v_maxc_maxs_s, 11, 8, 3, up(s)).
table(up_v_mins_maxs_s, 11, 7, 3, up(s)).
table(up_c_minc_maxc_s, 11, 12, 3, up(s)).
table(up_c_maxc_mins_s, 11, 5, 3, up(s)).
table(up_c_maxc_maxs_s, 11, 9, 3, up(s)).
table(up_v_c_minc_maxc_s, 11, 12, 4, up(s)).
table(up_v_c_minc_mins_s, 11, 8, 4, up(s)).
table(up_v_c_minc_maxs_s, 11, 11, 4, up(s)).
table(up_v_c_maxc_mins_s, 11, 6, 4, up(s)).
table(up_v_c_maxc_maxs_s, 11, 9, 4, up(s)).
table(up_v_c_mins_maxs_s, 11, 8, 4, up(s)).
table(up_v_minc_maxc_mins_s, 11, 7, 4, up(s)).
table(up_v_minc_maxc_maxs_s, 11, 10, 4, up(s)).
table(up_v_minc_mins_maxs_s, 11, 8, 4, up(s)).
table(up_v_maxc_mins_maxs_s, 11, 8, 4, up(s)).
table(up_c_minc_maxc_mins_s, 11, 7, 4, up(s)).
table(up_c_minc_maxc_maxs_s, 11, 12, 4, up(s)).
table(up_c_maxc_mins_maxs_s, 11, 6, 4, up(s)).
table(up_v_c_minc_maxc_mins_s, 11, 9, 5, up(s)).
table(up_v_c_minc_maxc_maxs_s, 11, 12, 5, up(s)).
table(up_v_c_minc_mins_maxs_s, 11, 11, 5, up(s)).
table(up_v_c_maxc_mins_maxs_s, 11, 9, 5, up(s)).
table(up_v_minc_maxc_mins_maxs_s, 11, 10, 5, up(s)).
table(up_c_minc_maxc_mins_maxs_s, 11, 8, 5, up(s)).
table(up_v_c_minc_maxc_mins_maxs_s, 11, 12, 6, up(s)).
table(low_v_mins, 11, 2, 1, low(mins)).
table(low_c_mins, 11, 2, 1, low(mins)).
table(low_minc_mins, 11, 2, 1, low(mins)).
table(low_maxc_mins, 11, 2, 1, low(mins)).
table(low_s_mins, 11, 2, 1, low(mins)).
table(low_maxs_mins, 11, 2, 1, low(mins)).
table(low_v_c_mins, 11, 3, 2, low(mins)).
table(low_v_minc_mins, 11, 3, 2, low(mins)).
table(low_v_maxc_mins, 11, 3, 2, low(mins)).
table(low_v_s_mins, 11, 3, 2, low(mins)).
table(low_v_maxs_mins, 11, 3, 2, low(mins)).
table(low_c_minc_mins, 11, 3, 2, low(mins)).
table(low_c_maxc_mins, 11, 3, 2, low(mins)).
table(low_c_s_mins, 11, 3, 2, low(mins)).
table(low_c_maxs_mins, 11, 3, 2, low(mins)).
table(low_minc_maxc_mins, 11, 4, 2, low(mins)).
table(low_minc_s_mins, 11, 3, 2, low(mins)).
table(low_minc_maxs_mins, 11, 3, 2, low(mins)).
table(low_maxc_s_mins, 11, 3, 2, low(mins)).
table(low_maxc_maxs_mins, 11, 3, 2, low(mins)).
table(low_s_maxs_mins, 11, 3, 2, low(mins)).
table(low_v_c_minc_mins, 11, 6, 3, low(mins)).
table(low_v_c_maxc_mins, 11, 4, 3, low(mins)).
table(low_v_c_s_mins, 11, 4, 3, low(mins)).
table(low_v_c_maxs_mins, 11, 4, 3, low(mins)).
table(low_v_minc_maxc_mins, 11, 5, 3, low(mins)).
table(low_v_minc_s_mins, 11, 4, 3, low(mins)).
table(low_v_minc_maxs_mins, 11, 4, 3, low(mins)).
table(low_v_maxc_s_mins, 11, 4, 3, low(mins)).
table(low_v_maxc_maxs_mins, 11, 4, 3, low(mins)).
table(low_v_s_maxs_mins, 11, 4, 3, low(mins)).
table(low_c_minc_maxc_mins, 11, 5, 3, low(mins)).
table(low_c_minc_s_mins, 11, 4, 3, low(mins)).
table(low_c_minc_maxs_mins, 11, 4, 3, low(mins)).
table(low_c_maxc_s_mins, 11, 4, 3, low(mins)).
table(low_c_maxc_maxs_mins, 11, 4, 3, low(mins)).
table(low_c_s_maxs_mins, 11, 4, 3, low(mins)).
table(low_minc_maxc_s_mins, 11, 5, 3, low(mins)).
table(low_minc_maxc_maxs_mins, 11, 5, 3, low(mins)).
table(low_minc_s_maxs_mins, 11, 4, 3, low(mins)).
table(low_maxc_s_maxs_mins, 11, 4, 3, low(mins)).
table(low_v_c_minc_maxc_mins, 11, 7, 4, low(mins)).
table(low_v_c_minc_s_mins, 11, 7, 4, low(mins)).
table(low_v_c_minc_maxs_mins, 11, 7, 4, low(mins)).
table(low_v_c_maxc_s_mins, 11, 5, 4, low(mins)).
table(low_v_c_maxc_maxs_mins, 11, 5, 4, low(mins)).
table(low_v_c_s_maxs_mins, 11, 5, 4, low(mins)).
table(low_v_minc_maxc_s_mins, 11, 6, 4, low(mins)).
table(low_v_minc_maxc_maxs_mins, 11, 6, 4, low(mins)).
table(low_v_minc_s_maxs_mins, 11, 5, 4, low(mins)).
table(low_v_maxc_s_maxs_mins, 11, 5, 4, low(mins)).
table(low_c_minc_maxc_s_mins, 11, 6, 4, low(mins)).
table(low_c_minc_maxc_maxs_mins, 11, 6, 4, low(mins)).
table(low_c_minc_s_maxs_mins, 11, 5, 4, low(mins)).
table(low_c_maxc_s_maxs_mins, 11, 5, 4, low(mins)).
table(low_minc_maxc_s_maxs_mins, 11, 6, 4, low(mins)).
table(low_v_c_minc_maxc_s_mins, 11, 8, 5, low(mins)).
table(low_v_c_minc_maxc_maxs_mins, 11, 8, 5, low(mins)).
table(low_v_c_minc_s_maxs_mins, 11, 8, 5, low(mins)).
table(low_v_c_maxc_s_maxs_mins, 11, 6, 5, low(mins)).
table(low_v_minc_maxc_s_maxs_mins, 11, 7, 5, low(mins)).
table(low_c_minc_maxc_s_maxs_mins, 11, 7, 5, low(mins)).
table(low_v_c_minc_maxc_s_maxs_mins, 11, 9, 6, low(mins)).
table(up_v_mins, 11, 18, 1, up(mins)).
table(up_maxc_mins, 11, 9, 1, up(mins)).
table(up_maxs_mins, 11, 2, 1, up(mins)).
table(up_v_c_mins, 11, 6, 2, up(mins)).
table(up_v_minc_mins, 11, 3, 2, up(mins)).
table(up_v_maxc_mins, 11, 7, 2, up(mins)).
table(up_v_s_mins, 11, 3, 2, up(mins)).
table(up_v_maxs_mins, 11, 3, 2, up(mins)).
table(up_c_maxc_mins, 11, 5, 2, up(mins)).
table(up_c_maxs_mins, 11, 3, 2, up(mins)).
table(up_minc_maxc_mins, 11, 4, 2, up(mins)).
table(up_minc_maxs_mins, 11, 3, 2, up(mins)).
table(up_maxc_s_mins, 11, 3, 2, up(mins)).
table(up_maxc_maxs_mins, 11, 3, 2, up(mins)).
table(up_s_maxs_mins, 11, 3, 2, up(mins)).
table(up_v_c_minc_mins, 11, 6, 3, up(mins)).
table(up_v_c_maxc_mins, 11, 7, 3, up(mins)).
table(up_v_c_s_mins, 11, 4, 3, up(mins)).
table(up_v_c_maxs_mins, 11, 4, 3, up(mins)).
table(up_v_minc_maxc_mins, 11, 5, 3, up(mins)).
table(up_v_minc_s_mins, 11, 4, 3, up(mins)).
table(up_v_minc_maxs_mins, 11, 4, 3, up(mins)).
table(up_v_maxc_s_mins, 11, 4, 3, up(mins)).
table(up_v_maxc_maxs_mins, 11, 4, 3, up(mins)).
table(up_v_s_maxs_mins, 11, 4, 3, up(mins)).
table(up_c_minc_maxc_mins, 11, 5, 3, up(mins)).
table(up_c_minc_maxs_mins, 11, 4, 3, up(mins)).
table(up_c_maxc_s_mins, 11, 4, 3, up(mins)).
table(up_c_maxc_maxs_mins, 11, 4, 3, up(mins)).
table(up_c_s_maxs_mins, 11, 4, 3, up(mins)).
table(up_minc_maxc_s_mins, 11, 5, 3, up(mins)).
table(up_minc_maxc_maxs_mins, 11, 5, 3, up(mins)).
table(up_minc_s_maxs_mins, 11, 4, 3, up(mins)).
table(up_maxc_s_maxs_mins, 11, 4, 3, up(mins)).
table(up_v_c_minc_maxc_mins, 11, 7, 4, up(mins)).
table(up_v_c_minc_s_mins, 11, 7, 4, up(mins)).
table(up_v_c_minc_maxs_mins, 11, 7, 4, up(mins)).
table(up_v_c_maxc_s_mins, 11, 5, 4, up(mins)).
table(up_v_c_maxc_maxs_mins, 11, 5, 4, up(mins)).
table(up_v_c_s_maxs_mins, 11, 5, 4, up(mins)).
table(up_v_minc_maxc_s_mins, 11, 6, 4, up(mins)).
table(up_v_minc_maxc_maxs_mins, 11, 6, 4, up(mins)).
table(up_v_minc_s_maxs_mins, 11, 5, 4, up(mins)).
table(up_v_maxc_s_maxs_mins, 11, 5, 4, up(mins)).
table(up_c_minc_maxc_s_mins, 11, 6, 4, up(mins)).
table(up_c_minc_maxc_maxs_mins, 11, 6, 4, up(mins)).
table(up_c_minc_s_maxs_mins, 11, 5, 4, up(mins)).
table(up_c_maxc_s_maxs_mins, 11, 5, 4, up(mins)).
table(up_minc_maxc_s_maxs_mins, 11, 6, 4, up(mins)).
table(up_v_c_minc_maxc_s_mins, 11, 8, 5, up(mins)).
table(up_v_c_minc_maxc_maxs_mins, 11, 8, 5, up(mins)).
table(up_v_c_minc_s_maxs_mins, 11, 8, 5, up(mins)).
table(up_v_c_maxc_s_maxs_mins, 11, 6, 5, up(mins)).
table(up_v_minc_maxc_s_maxs_mins, 11, 7, 5, up(mins)).
table(up_c_minc_maxc_s_maxs_mins, 11, 7, 5, up(mins)).
table(up_v_c_minc_maxc_s_maxs_mins, 11, 9, 6, up(mins)).
table(low_v_maxs, 11, 7, 1, low(maxs)).
table(low_c_maxs, 11, 4, 1, low(maxs)).
table(low_minc_maxs, 11, 4, 1, low(maxs)).
table(low_maxc_maxs, 11, 4, 1, low(maxs)).
table(low_s_maxs, 11, 7, 1, low(maxs)).
table(low_mins_maxs, 11, 2, 1, low(maxs)).
table(low_v_c_maxs, 11, 8, 2, low(maxs)).
table(low_v_minc_maxs, 11, 8, 2, low(maxs)).
table(low_v_maxc_maxs, 11, 8, 2, low(maxs)).
table(low_v_s_maxs, 11, 3, 2, low(maxs)).
table(low_v_mins_maxs, 11, 3, 2, low(maxs)).
table(low_c_minc_maxs, 11, 5, 2, low(maxs)).
table(low_c_maxc_maxs, 11, 5, 2, low(maxs)).
table(low_c_s_maxs, 11, 8, 2, low(maxs)).
table(low_c_mins_maxs, 11, 3, 2, low(maxs)).
table(low_minc_maxc_maxs, 11, 6, 2, low(maxs)).
table(low_minc_s_maxs, 11, 3, 2, low(maxs)).
table(low_minc_mins_maxs, 11, 3, 2, low(maxs)).
table(low_maxc_s_maxs, 11, 3, 2, low(maxs)).
table(low_maxc_mins_maxs, 11, 3, 2, low(maxs)).
table(low_s_mins_maxs, 11, 7, 2, low(maxs)).
table(low_v_c_minc_maxs, 11, 11, 3, low(maxs)).
table(low_v_c_maxc_maxs, 11, 9, 3, low(maxs)).
table(low_v_c_s_maxs, 11, 4, 3, low(maxs)).
table(low_v_c_mins_maxs, 11, 4, 3, low(maxs)).
table(low_v_minc_maxc_maxs, 11, 10, 3, low(maxs)).
table(low_v_minc_s_maxs, 11, 4, 3, low(maxs)).
table(low_v_minc_mins_maxs, 11, 4, 3, low(maxs)).
table(low_v_maxc_s_maxs, 11, 4, 3, low(maxs)).
table(low_v_maxc_mins_maxs, 11, 4, 3, low(maxs)).
table(low_v_s_mins_maxs, 11, 4, 3, low(maxs)).
table(low_c_minc_maxc_maxs, 11, 7, 3, low(maxs)).
table(low_c_minc_s_maxs, 11, 4, 3, low(maxs)).
table(low_c_minc_mins_maxs, 11, 4, 3, low(maxs)).
table(low_c_maxc_s_maxs, 11, 4, 3, low(maxs)).
table(low_c_maxc_mins_maxs, 11, 4, 3, low(maxs)).
table(low_c_s_mins_maxs, 11, 8, 3, low(maxs)).
table(low_minc_maxc_s_maxs, 11, 5, 3, low(maxs)).
table(low_minc_maxc_mins_maxs, 11, 5, 3, low(maxs)).
table(low_minc_s_mins_maxs, 11, 4, 3, low(maxs)).
table(low_maxc_s_mins_maxs, 11, 4, 3, low(maxs)).
table(low_v_c_minc_maxc_maxs, 11, 12, 4, low(maxs)).
table(low_v_c_minc_s_maxs, 11, 7, 4, low(maxs)).
table(low_v_c_minc_mins_maxs, 11, 7, 4, low(maxs)).
table(low_v_c_maxc_s_maxs, 11, 5, 4, low(maxs)).
table(low_v_c_maxc_mins_maxs, 11, 5, 4, low(maxs)).
table(low_v_c_s_mins_maxs, 11, 5, 4, low(maxs)).
table(low_v_minc_maxc_s_maxs, 11, 6, 4, low(maxs)).
table(low_v_minc_maxc_mins_maxs, 11, 6, 4, low(maxs)).
table(low_v_minc_s_mins_maxs, 11, 5, 4, low(maxs)).
table(low_v_maxc_s_mins_maxs, 11, 5, 4, low(maxs)).
table(low_c_minc_maxc_s_maxs, 11, 6, 4, low(maxs)).
table(low_c_minc_maxc_mins_maxs, 11, 6, 4, low(maxs)).
table(low_c_minc_s_mins_maxs, 11, 5, 4, low(maxs)).
table(low_c_maxc_s_mins_maxs, 11, 5, 4, low(maxs)).
table(low_minc_maxc_s_mins_maxs, 11, 6, 4, low(maxs)).
table(low_v_c_minc_maxc_s_maxs, 11, 8, 5, low(maxs)).
table(low_v_c_minc_maxc_mins_maxs, 11, 8, 5, low(maxs)).
table(low_v_c_minc_s_mins_maxs, 11, 8, 5, low(maxs)).
table(low_v_c_maxc_s_mins_maxs, 11, 6, 5, low(maxs)).
table(low_v_minc_maxc_s_mins_maxs, 11, 7, 5, low(maxs)).
table(low_c_minc_maxc_s_mins_maxs, 11, 7, 5, low(maxs)).
table(low_v_c_minc_maxc_s_mins_maxs, 11, 9, 6, low(maxs)).
table(up_v_maxs, 11, 18, 1, up(maxs)).
table(up_maxc_maxs, 11, 2, 1, up(maxs)).
table(up_v_c_maxs, 11, 18, 2, up(maxs)).
table(up_v_minc_maxs, 11, 12, 2, up(maxs)).
table(up_v_maxc_maxs, 11, 3, 2, up(maxs)).
table(up_v_s_maxs, 11, 7, 2, up(maxs)).
table(up_v_mins_maxs, 11, 9, 2, up(maxs)).
table(up_c_maxc_maxs, 11, 3, 2, up(maxs)).
table(up_minc_maxc_maxs, 11, 4, 2, up(maxs)).
table(up_maxc_s_maxs, 11, 3, 2, up(maxs)).
table(up_maxc_mins_maxs, 11, 3, 2, up(maxs)).
table(up_v_c_minc_maxs, 11, 12, 3, up(maxs)).
table(up_v_c_maxc_maxs, 11, 4, 3, up(maxs)).
table(up_v_c_s_maxs, 11, 8, 3, up(maxs)).
table(up_v_c_mins_maxs, 11, 18, 3, up(maxs)).
table(up_v_minc_maxc_maxs, 11, 5, 3, up(maxs)).
table(up_v_minc_s_maxs, 11, 5, 3, up(maxs)).
table(up_v_minc_mins_maxs, 11, 14, 3, up(maxs)).
table(up_v_maxc_s_maxs, 11, 4, 3, up(maxs)).
table(up_v_maxc_mins_maxs, 11, 4, 3, up(maxs)).
table(up_v_s_mins_maxs, 11, 7, 3, up(maxs)).
table(up_c_minc_maxc_maxs, 11, 5, 3, up(maxs)).
table(up_c_maxc_s_maxs, 11, 4, 3, up(maxs)).
table(up_c_maxc_mins_maxs, 11, 4, 3, up(maxs)).
table(up_minc_maxc_s_maxs, 11, 5, 3, up(maxs)).
table(up_minc_maxc_mins_maxs, 11, 5, 3, up(maxs)).
table(up_maxc_s_mins_maxs, 11, 4, 3, up(maxs)).
table(up_v_c_minc_maxc_maxs, 11, 7, 4, up(maxs)).
table(up_v_c_minc_s_maxs, 11, 7, 4, up(maxs)).
table(up_v_c_minc_mins_maxs, 11, 13, 4, up(maxs)).
table(up_v_c_maxc_s_maxs, 11, 5, 4, up(maxs)).
table(up_v_c_maxc_mins_maxs, 11, 5, 4, up(maxs)).
table(up_v_c_s_mins_maxs, 11, 8, 4, up(maxs)).
table(up_v_minc_maxc_s_maxs, 11, 6, 4, up(maxs)).
table(up_v_minc_maxc_mins_maxs, 11, 6, 4, up(maxs)).
table(up_v_minc_s_mins_maxs, 11, 6, 4, up(maxs)).
table(up_v_maxc_s_mins_maxs, 11, 5, 4, up(maxs)).
table(up_c_minc_maxc_s_maxs, 11, 6, 4, up(maxs)).
table(up_c_minc_maxc_mins_maxs, 11, 6, 4, up(maxs)).
table(up_c_maxc_s_mins_maxs, 11, 5, 4, up(maxs)).
table(up_minc_maxc_s_mins_maxs, 11, 6, 4, up(maxs)).
table(up_v_c_minc_maxc_s_maxs, 11, 8, 5, up(maxs)).
table(up_v_c_minc_maxc_mins_maxs, 11, 8, 5, up(maxs)).
table(up_v_c_minc_s_mins_maxs, 11, 8, 5, up(maxs)).
table(up_v_c_maxc_s_mins_maxs, 11, 6, 5, up(maxs)).
table(up_v_minc_maxc_s_mins_maxs, 11, 7, 5, up(maxs)).
table(up_c_minc_maxc_s_mins_maxs, 11, 7, 5, up(maxs)).
table(up_v_c_minc_maxc_s_mins_maxs, 11, 9, 6, up(maxs)).
table(low_v_mina, 11, 12, 1, low(mina)).
table(low_c_mina, 11, 7, 1, low(mina)).
table(low_minc_mina, 11, 19, 1, low(mina)).
table(low_maxc_mina, 11, 19, 1, low(mina)).
table(low_s_mina, 11, 12, 1, low(mina)).
table(low_mins_mina, 11, 11, 1, low(mina)).
table(low_maxs_mina, 11, 11, 1, low(mina)).
table(low_v_c_mina, 11, 10, 2, low(mina)).
table(low_v_minc_mina, 11, 11, 2, low(mina)).
table(low_v_maxc_mina, 11, 11, 2, low(mina)).
table(low_v_s_mina, 11, 4, 2, low(mina)).
table(low_v_mins_mina, 11, 4, 2, low(mina)).
table(low_v_maxs_mina, 11, 10, 2, low(mina)).
table(low_c_minc_mina, 11, 8, 2, low(mina)).
table(low_c_maxc_mina, 11, 6, 2, low(mina)).
table(low_c_s_mina, 11, 10, 2, low(mina)).
table(low_c_mins_mina, 11, 7, 2, low(mina)).
table(low_c_maxs_mina, 11, 5, 2, low(mina)).
table(low_minc_maxc_mina, 11, 19, 2, low(mina)).
table(low_minc_s_mina, 11, 11, 2, low(mina)).
table(low_minc_mins_mina, 11, 19, 2, low(mina)).
table(low_minc_maxs_mina, 11, 19, 2, low(mina)).
table(low_maxc_s_mina, 11, 11, 2, low(mina)).
table(low_maxc_mins_mina, 11, 19, 2, low(mina)).
table(low_maxc_maxs_mina, 11, 19, 2, low(mina)).
table(low_s_mins_mina, 11, 12, 2, low(mina)).
table(low_s_maxs_mina, 11, 10, 2, low(mina)).
table(low_mins_maxs_mina, 11, 6, 2, low(mina)).
table(low_v_c_minc_mina, 11, 13, 3, low(mina)).
table(low_v_c_maxc_mina, 11, 11, 3, low(mina)).
table(low_v_c_s_mina, 11, 4, 3, low(mina)).
table(low_v_c_mins_mina, 11, 7, 3, low(mina)).
table(low_v_c_maxs_mina, 11, 10, 3, low(mina)).
table(low_v_minc_maxc_mina, 11, 13, 3, low(mina)).
table(low_v_minc_s_mina, 11, 4, 3, low(mina)).
table(low_v_minc_mins_mina, 11, 4, 3, low(mina)).
table(low_v_minc_maxs_mina, 11, 11, 3, low(mina)).
table(low_v_maxc_s_mina, 11, 4, 3, low(mina)).
table(low_v_maxc_mins_mina, 11, 6, 3, low(mina)).
table(low_v_maxc_maxs_mina, 11, 10, 3, low(mina)).
table(low_v_s_mins_mina, 11, 5, 3, low(mina)).
table(low_v_s_maxs_mina, 11, 5, 3, low(mina)).
table(low_v_mins_maxs_mina, 11, 5, 3, low(mina)).
table(low_c_minc_maxc_mina, 11, 8, 3, low(mina)).
table(low_c_minc_s_mina, 11, 11, 3, low(mina)).
table(low_c_minc_mins_mina, 11, 8, 3, low(mina)).
table(low_c_minc_maxs_mina, 11, 6, 3, low(mina)).
table(low_c_maxc_s_mina, 11, 8, 3, low(mina)).
table(low_c_maxc_mins_mina, 11, 6, 3, low(mina)).
table(low_c_maxc_maxs_mina, 11, 6, 3, low(mina)).
table(low_c_s_mins_mina, 11, 10, 3, low(mina)).
table(low_c_s_maxs_mina, 11, 10, 3, low(mina)).
table(low_c_mins_maxs_mina, 11, 5, 3, low(mina)).
table(low_minc_maxc_s_mina, 11, 10, 3, low(mina)).
table(low_minc_maxc_mins_mina, 11, 16, 3, low(mina)).
table(low_minc_maxc_maxs_mina, 11, 18, 3, low(mina)).
table(low_minc_s_mins_mina, 11, 6, 3, low(mina)).
table(low_minc_s_maxs_mina, 11, 8, 3, low(mina)).
table(low_minc_mins_maxs_mina, 11, 19, 3, low(mina)).
table(low_maxc_s_mins_mina, 11, 8, 3, low(mina)).
table(low_maxc_s_maxs_mina, 11, 7, 3, low(mina)).
table(low_maxc_mins_maxs_mina, 11, 9, 3, low(mina)).
table(low_s_mins_maxs_mina, 11, 10, 3, low(mina)).
table(low_v_c_minc_maxc_mina, 11, 15, 4, low(mina)).
table(low_v_c_minc_s_mina, 11, 8, 4, low(mina)).
table(low_v_c_minc_mins_mina, 11, 10, 4, low(mina)).
table(low_v_c_minc_maxs_mina, 11, 13, 4, low(mina)).
table(low_v_c_maxc_s_mina, 11, 5, 4, low(mina)).
table(low_v_c_maxc_mins_mina, 11, 8, 4, low(mina)).
table(low_v_c_maxc_maxs_mina, 11, 11, 4, low(mina)).
table(low_v_c_s_mins_mina, 11, 5, 4, low(mina)).
table(low_v_c_s_maxs_mina, 11, 5, 4, low(mina)).
table(low_v_c_mins_maxs_mina, 11, 10, 4, low(mina)).
table(low_v_minc_maxc_s_mina, 11, 7, 4, low(mina)).
table(low_v_minc_maxc_mins_mina, 11, 7, 4, low(mina)).
table(low_v_minc_maxc_maxs_mina, 11, 12, 4, low(mina)).
table(low_v_minc_s_mins_mina, 11, 5, 4, low(mina)).
table(low_v_minc_s_maxs_mina, 11, 6, 4, low(mina)).
table(low_v_minc_mins_maxs_mina, 11, 6, 4, low(mina)).
table(low_v_maxc_s_mins_mina, 11, 5, 4, low(mina)).
table(low_v_maxc_s_maxs_mina, 11, 5, 4, low(mina)).
table(low_v_maxc_mins_maxs_mina, 11, 5, 4, low(mina)).
table(low_v_s_mins_maxs_mina, 11, 6, 4, low(mina)).
table(low_c_minc_maxc_s_mina, 11, 9, 4, low(mina)).
table(low_c_minc_maxc_mins_mina, 11, 7, 4, low(mina)).
table(low_c_minc_maxc_maxs_mina, 11, 8, 4, low(mina)).
table(low_c_minc_s_mins_mina, 11, 10, 4, low(mina)).
table(low_c_minc_s_maxs_mina, 11, 10, 4, low(mina)).
table(low_c_minc_mins_maxs_mina, 11, 6, 4, low(mina)).
table(low_c_maxc_s_mins_mina, 11, 6, 4, low(mina)).
table(low_c_maxc_s_maxs_mina, 11, 6, 4, low(mina)).
table(low_c_maxc_mins_maxs_mina, 11, 6, 4, low(mina)).
table(low_c_s_mins_maxs_mina, 11, 10, 4, low(mina)).
table(low_minc_maxc_s_mins_mina, 11, 9, 4, low(mina)).
table(low_minc_maxc_s_maxs_mina, 11, 9, 4, low(mina)).
table(low_minc_maxc_mins_maxs_mina, 11, 18, 4, low(mina)).
table(low_minc_s_mins_maxs_mina, 11, 7, 4, low(mina)).
table(low_maxc_s_mins_maxs_mina, 11, 6, 4, low(mina)).
table(low_v_c_minc_maxc_s_mina, 11, 9, 5, low(mina)).
table(low_v_c_minc_maxc_mins_mina, 11, 12, 5, low(mina)).
table(low_v_c_minc_maxc_maxs_mina, 11, 15, 5, low(mina)).
table(low_v_c_minc_s_mins_mina, 11, 8, 5, low(mina)).
table(low_v_c_minc_s_maxs_mina, 11, 8, 5, low(mina)).
table(low_v_c_minc_mins_maxs_mina, 11, 13, 5, low(mina)).
table(low_v_c_maxc_s_mins_mina, 11, 6, 5, low(mina)).
table(low_v_c_maxc_s_maxs_mina, 11, 6, 5, low(mina)).
table(low_v_c_maxc_mins_maxs_mina, 11, 11, 5, low(mina)).
table(low_v_c_s_mins_maxs_mina, 11, 6, 5, low(mina)).
table(low_v_minc_maxc_s_mins_mina, 11, 7, 5, low(mina)).
table(low_v_minc_maxc_s_maxs_mina, 11, 7, 5, low(mina)).
table(low_v_minc_maxc_mins_maxs_mina, 11, 7, 5, low(mina)).
table(low_v_minc_s_mins_maxs_mina, 11, 7, 5, low(mina)).
table(low_v_maxc_s_mins_maxs_mina, 11, 6, 5, low(mina)).
table(low_c_minc_maxc_s_mins_mina, 11, 8, 5, low(mina)).
table(low_c_minc_maxc_s_maxs_mina, 11, 8, 5, low(mina)).
table(low_c_minc_maxc_mins_maxs_mina, 11, 8, 5, low(mina)).
table(low_c_minc_s_mins_maxs_mina, 11, 12, 5, low(mina)).
table(low_c_maxc_s_mins_maxs_mina, 11, 8, 5, low(mina)).
table(low_minc_maxc_s_mins_maxs_mina, 11, 11, 5, low(mina)).
table(low_v_c_minc_maxc_s_mins_mina, 11, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_maxs_mina, 11, 9, 6, low(mina)).
table(low_v_c_minc_maxc_mins_maxs_mina, 11, 15, 6, low(mina)).
table(low_v_c_minc_s_mins_maxs_mina, 11, 9, 6, low(mina)).
table(low_v_c_maxc_s_mins_maxs_mina, 11, 7, 6, low(mina)).
table(low_v_minc_maxc_s_mins_maxs_mina, 11, 8, 6, low(mina)).
table(low_c_minc_maxc_s_mins_maxs_mina, 11, 10, 6, low(mina)).
table(low_v_c_minc_maxc_s_mins_maxs_mina, 11, 10, 7, low(mina)).
table(up_v_maxa, 11, 19, 1, up(maxa)).
table(up_v_c_maxa, 11, 19, 2, up(maxa)).
table(up_v_minc_maxa, 11, 19, 2, up(maxa)).
table(up_v_maxc_maxa, 11, 19, 2, up(maxa)).
table(up_v_s_maxa, 11, 19, 2, up(maxa)).
table(up_v_mins_maxa, 11, 19, 2, up(maxa)).
table(up_v_maxs_maxa, 11, 19, 2, up(maxa)).
table(up_c_maxc_maxa, 11, 19, 2, up(maxa)).
table(up_maxc_s_maxa, 11, 18, 2, up(maxa)).
table(up_s_maxs_maxa, 11, 19, 2, up(maxa)).
table(up_v_c_minc_maxa, 11, 19, 3, up(maxa)).
table(up_v_c_maxc_maxa, 11, 19, 3, up(maxa)).
table(up_v_c_s_maxa, 11, 19, 3, up(maxa)).
table(up_v_c_mins_maxa, 11, 19, 3, up(maxa)).
table(up_v_c_maxs_maxa, 11, 7, 3, up(maxa)).
table(up_v_minc_maxc_maxa, 11, 19, 3, up(maxa)).
table(up_v_minc_s_maxa, 11, 18, 3, up(maxa)).
table(up_v_minc_mins_maxa, 11, 19, 3, up(maxa)).
table(up_v_minc_maxs_maxa, 11, 19, 3, up(maxa)).
table(up_v_maxc_s_maxa, 11, 18, 3, up(maxa)).
table(up_v_maxc_mins_maxa, 11, 19, 3, up(maxa)).
table(up_v_maxc_maxs_maxa, 11, 9, 3, up(maxa)).
table(up_v_s_mins_maxa, 11, 19, 3, up(maxa)).
table(up_v_s_maxs_maxa, 11, 19, 3, up(maxa)).
table(up_v_mins_maxs_maxa, 11, 19, 3, up(maxa)).
table(up_c_minc_maxc_maxa, 11, 19, 3, up(maxa)).
table(up_c_maxc_s_maxa, 11, 18, 3, up(maxa)).
table(up_c_maxc_mins_maxa, 11, 19, 3, up(maxa)).
table(up_c_maxc_maxs_maxa, 11, 7, 3, up(maxa)).
table(up_c_s_maxs_maxa, 11, 19, 3, up(maxa)).
table(up_minc_maxc_s_maxa, 11, 6, 3, up(maxa)).
table(up_minc_s_maxs_maxa, 11, 18, 3, up(maxa)).
table(up_maxc_s_mins_maxa, 11, 18, 3, up(maxa)).
table(up_maxc_s_maxs_maxa, 11, 18, 3, up(maxa)).
table(up_s_mins_maxs_maxa, 11, 19, 3, up(maxa)).
table(up_v_c_minc_maxc_maxa, 11, 19, 4, up(maxa)).
table(up_v_c_minc_s_maxa, 11, 18, 4, up(maxa)).
table(up_v_c_minc_mins_maxa, 11, 13, 4, up(maxa)).
table(up_v_c_minc_maxs_maxa, 11, 8, 4, up(maxa)).
table(up_v_c_maxc_s_maxa, 11, 18, 4, up(maxa)).
table(up_v_c_maxc_mins_maxa, 11, 19, 4, up(maxa)).
table(up_v_c_maxc_maxs_maxa, 11, 6, 4, up(maxa)).
table(up_v_c_s_mins_maxa, 11, 19, 4, up(maxa)).
table(up_v_c_s_maxs_maxa, 11, 19, 4, up(maxa)).
table(up_v_c_mins_maxs_maxa, 11, 11, 4, up(maxa)).
table(up_v_minc_maxc_s_maxa, 11, 18, 4, up(maxa)).
table(up_v_minc_maxc_mins_maxa, 11, 19, 4, up(maxa)).
table(up_v_minc_maxc_maxs_maxa, 11, 10, 4, up(maxa)).
table(up_v_minc_s_mins_maxa, 11, 18, 4, up(maxa)).
table(up_v_minc_s_maxs_maxa, 11, 18, 4, up(maxa)).
table(up_v_minc_mins_maxs_maxa, 11, 18, 4, up(maxa)).
table(up_v_maxc_s_mins_maxa, 11, 18, 4, up(maxa)).
table(up_v_maxc_s_maxs_maxa, 11, 18, 4, up(maxa)).
table(up_v_maxc_mins_maxs_maxa, 11, 8, 4, up(maxa)).
table(up_v_s_mins_maxs_maxa, 11, 19, 4, up(maxa)).
table(up_c_minc_maxc_s_maxa, 11, 18, 4, up(maxa)).
table(up_c_minc_maxc_mins_maxa, 11, 19, 4, up(maxa)).
table(up_c_minc_maxc_maxs_maxa, 11, 10, 4, up(maxa)).
table(up_c_minc_s_maxs_maxa, 11, 14, 4, up(maxa)).
table(up_c_maxc_s_mins_maxa, 11, 18, 4, up(maxa)).
table(up_c_maxc_s_maxs_maxa, 11, 18, 4, up(maxa)).
table(up_c_maxc_mins_maxs_maxa, 11, 6, 4, up(maxa)).
table(up_c_s_mins_maxs_maxa, 11, 19, 4, up(maxa)).
table(up_minc_maxc_s_mins_maxa, 11, 6, 4, up(maxa)).
table(up_minc_maxc_s_maxs_maxa, 11, 7, 4, up(maxa)).
table(up_minc_s_mins_maxs_maxa, 11, 18, 4, up(maxa)).
table(up_maxc_s_mins_maxs_maxa, 11, 18, 4, up(maxa)).
table(up_v_c_minc_maxc_s_maxa, 11, 18, 5, up(maxa)).
table(up_v_c_minc_maxc_mins_maxa, 11, 19, 5, up(maxa)).
table(up_v_c_minc_maxc_maxs_maxa, 11, 10, 5, up(maxa)).
table(up_v_c_minc_s_mins_maxa, 11, 12, 5, up(maxa)).
table(up_v_c_minc_s_maxs_maxa, 11, 12, 5, up(maxa)).
table(up_v_c_minc_mins_maxs_maxa, 11, 8, 5, up(maxa)).
table(up_v_c_maxc_s_mins_maxa, 11, 18, 5, up(maxa)).
table(up_v_c_maxc_s_maxs_maxa, 11, 18, 5, up(maxa)).
table(up_v_c_maxc_mins_maxs_maxa, 11, 6, 5, up(maxa)).
table(up_v_c_s_mins_maxs_maxa, 11, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_mins_maxa, 11, 18, 5, up(maxa)).
table(up_v_minc_maxc_s_maxs_maxa, 11, 18, 5, up(maxa)).
table(up_v_minc_maxc_mins_maxs_maxa, 11, 9, 5, up(maxa)).
table(up_v_minc_s_mins_maxs_maxa, 11, 18, 5, up(maxa)).
table(up_v_maxc_s_mins_maxs_maxa, 11, 18, 5, up(maxa)).
table(up_c_minc_maxc_s_mins_maxa, 11, 18, 5, up(maxa)).
table(up_c_minc_maxc_s_maxs_maxa, 11, 18, 5, up(maxa)).
table(up_c_minc_maxc_mins_maxs_maxa, 11, 9, 5, up(maxa)).
table(up_c_minc_s_mins_maxs_maxa, 11, 10, 5, up(maxa)).
table(up_c_maxc_s_mins_maxs_maxa, 11, 18, 5, up(maxa)).
table(up_minc_maxc_s_mins_maxs_maxa, 11, 7, 5, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxa, 11, 18, 6, up(maxa)).
table(up_v_c_minc_maxc_s_maxs_maxa, 11, 18, 6, up(maxa)).
table(up_v_c_minc_maxc_mins_maxs_maxa, 11, 9, 6, up(maxa)).
table(up_v_c_minc_s_mins_maxs_maxa, 11, 10, 6, up(maxa)).
table(up_v_c_maxc_s_mins_maxs_maxa, 11, 18, 6, up(maxa)).
table(up_v_minc_maxc_s_mins_maxs_maxa, 11, 18, 6, up(maxa)).
table(up_c_minc_maxc_s_mins_maxs_maxa, 11, 18, 6, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxs_maxa, 11, 18, 7, up(maxa)).
table(low_c_v, 12, 18, 1, low(v)).
table(low_minc_v, 12, 11, 1, low(v)).
table(low_maxc_v, 12, 11, 1, low(v)).
table(low_s_v, 12, 7, 1, low(v)).
table(low_mins_v, 12, 18, 1, low(v)).
table(low_maxs_v, 12, 18, 1, low(v)).
table(low_c_minc_v, 12, 11, 2, low(v)).
table(low_c_maxc_v, 12, 11, 2, low(v)).
table(low_c_s_v, 12, 8, 2, low(v)).
table(low_c_mins_v, 12, 18, 2, low(v)).
table(low_c_maxs_v, 12, 18, 2, low(v)).
table(low_minc_maxc_v, 12, 11, 2, low(v)).
table(low_minc_s_v, 12, 3, 2, low(v)).
table(low_minc_mins_v, 12, 12, 2, low(v)).
table(low_minc_maxs_v, 12, 12, 2, low(v)).
table(low_maxc_s_v, 12, 3, 2, low(v)).
table(low_maxc_mins_v, 12, 12, 2, low(v)).
table(low_maxc_maxs_v, 12, 13, 2, low(v)).
table(low_s_mins_v, 12, 7, 2, low(v)).
table(low_s_maxs_v, 12, 7, 2, low(v)).
table(low_mins_maxs_v, 12, 12, 2, low(v)).
table(low_c_minc_maxc_v, 12, 11, 3, low(v)).
table(low_c_minc_s_v, 12, 6, 3, low(v)).
table(low_c_minc_mins_v, 12, 12, 3, low(v)).
table(low_c_minc_maxs_v, 12, 12, 3, low(v)).
table(low_c_maxc_s_v, 12, 4, 3, low(v)).
table(low_c_maxc_mins_v, 12, 12, 3, low(v)).
table(low_c_maxc_maxs_v, 12, 13, 3, low(v)).
table(low_c_s_mins_v, 12, 8, 3, low(v)).
table(low_c_s_maxs_v, 12, 8, 3, low(v)).
table(low_c_mins_maxs_v, 12, 18, 3, low(v)).
table(low_minc_maxc_s_v, 12, 5, 3, low(v)).
table(low_minc_maxc_mins_v, 12, 12, 3, low(v)).
table(low_minc_maxc_maxs_v, 12, 12, 3, low(v)).
table(low_minc_s_mins_v, 12, 4, 3, low(v)).
table(low_minc_s_maxs_v, 12, 4, 3, low(v)).
table(low_minc_mins_maxs_v, 12, 13, 3, low(v)).
table(low_maxc_s_mins_v, 12, 4, 3, low(v)).
table(low_maxc_s_maxs_v, 12, 4, 3, low(v)).
table(low_maxc_mins_maxs_v, 12, 14, 3, low(v)).
table(low_s_mins_maxs_v, 12, 7, 3, low(v)).
table(low_c_minc_maxc_s_v, 12, 7, 4, low(v)).
table(low_c_minc_maxc_mins_v, 12, 12, 4, low(v)).
table(low_c_minc_maxc_maxs_v, 12, 12, 4, low(v)).
table(low_c_minc_s_mins_v, 12, 7, 4, low(v)).
table(low_c_minc_s_maxs_v, 12, 7, 4, low(v)).
table(low_c_minc_mins_maxs_v, 12, 13, 4, low(v)).
table(low_c_maxc_s_mins_v, 12, 5, 4, low(v)).
table(low_c_maxc_s_maxs_v, 12, 5, 4, low(v)).
table(low_c_maxc_mins_maxs_v, 12, 14, 4, low(v)).
table(low_c_s_mins_maxs_v, 12, 8, 4, low(v)).
table(low_minc_maxc_s_mins_v, 12, 6, 4, low(v)).
table(low_minc_maxc_s_maxs_v, 12, 6, 4, low(v)).
table(low_minc_maxc_mins_maxs_v, 12, 13, 4, low(v)).
table(low_minc_s_mins_maxs_v, 12, 5, 4, low(v)).
table(low_maxc_s_mins_maxs_v, 12, 5, 4, low(v)).
table(low_c_minc_maxc_s_mins_v, 12, 8, 5, low(v)).
table(low_c_minc_maxc_s_maxs_v, 12, 8, 5, low(v)).
table(low_c_minc_maxc_mins_maxs_v, 12, 13, 5, low(v)).
table(low_c_minc_s_mins_maxs_v, 12, 8, 5, low(v)).
table(low_c_maxc_s_mins_maxs_v, 12, 6, 5, low(v)).
table(low_minc_maxc_s_mins_maxs_v, 12, 7, 5, low(v)).
table(low_c_minc_maxc_s_mins_maxs_v, 12, 9, 6, low(v)).
table(up_c_maxc_v, 12, 3, 2, up(v)).
table(up_maxc_s_v, 12, 3, 2, up(v)).
table(up_s_maxs_v, 12, 3, 2, up(v)).
table(up_c_minc_maxc_v, 12, 6, 3, up(v)).
table(up_c_maxc_s_v, 12, 4, 3, up(v)).
table(up_c_maxc_mins_v, 12, 4, 3, up(v)).
table(up_c_maxc_maxs_v, 12, 4, 3, up(v)).
table(up_c_s_maxs_v, 12, 4, 3, up(v)).
table(up_minc_maxc_s_v, 12, 5, 3, up(v)).
table(up_minc_s_maxs_v, 12, 4, 3, up(v)).
table(up_maxc_s_mins_v, 12, 4, 3, up(v)).
table(up_maxc_s_maxs_v, 12, 4, 3, up(v)).
table(up_s_mins_maxs_v, 12, 4, 3, up(v)).
table(up_c_minc_maxc_s_v, 12, 7, 4, up(v)).
table(up_c_minc_maxc_mins_v, 12, 7, 4, up(v)).
table(up_c_minc_maxc_maxs_v, 12, 7, 4, up(v)).
table(up_c_minc_s_maxs_v, 12, 7, 4, up(v)).
table(up_c_maxc_s_mins_v, 12, 5, 4, up(v)).
table(up_c_maxc_s_maxs_v, 12, 5, 4, up(v)).
table(up_c_maxc_mins_maxs_v, 12, 5, 4, up(v)).
table(up_c_s_mins_maxs_v, 12, 5, 4, up(v)).
table(up_minc_maxc_s_mins_v, 12, 6, 4, up(v)).
table(up_minc_maxc_s_maxs_v, 12, 6, 4, up(v)).
table(up_minc_s_mins_maxs_v, 12, 5, 4, up(v)).
table(up_maxc_s_mins_maxs_v, 12, 5, 4, up(v)).
table(up_c_minc_maxc_s_mins_v, 12, 8, 5, up(v)).
table(up_c_minc_maxc_s_maxs_v, 12, 8, 5, up(v)).
table(up_c_minc_maxc_mins_maxs_v, 12, 8, 5, up(v)).
table(up_c_minc_s_mins_maxs_v, 12, 8, 5, up(v)).
table(up_c_maxc_s_mins_maxs_v, 12, 6, 5, up(v)).
table(up_minc_maxc_s_mins_maxs_v, 12, 7, 5, up(v)).
table(up_c_minc_maxc_s_mins_maxs_v, 12, 9, 6, up(v)).
table(low_v_c, 12, 11, 1, low(c)).
table(low_minc_c, 12, 11, 1, low(c)).
table(low_maxc_c, 12, 11, 1, low(c)).
table(low_s_c, 12, 8, 1, low(c)).
table(low_mins_c, 12, 8, 1, low(c)).
table(low_maxs_c, 12, 8, 1, low(c)).
table(low_v_minc_c, 12, 11, 2, low(c)).
table(low_v_maxc_c, 12, 3, 2, low(c)).
table(low_v_s_c, 12, 13, 2, low(c)).
table(low_v_mins_c, 12, 12, 2, low(c)).
table(low_v_maxs_c, 12, 13, 2, low(c)).
table(low_minc_maxc_c, 12, 11, 2, low(c)).
table(low_minc_s_c, 12, 5, 2, low(c)).
table(low_minc_mins_c, 12, 9, 2, low(c)).
table(low_minc_maxs_c, 12, 9, 2, low(c)).
table(low_maxc_s_c, 12, 3, 2, low(c)).
table(low_maxc_mins_c, 12, 12, 2, low(c)).
table(low_maxc_maxs_c, 12, 13, 2, low(c)).
table(low_s_mins_c, 12, 10, 2, low(c)).
table(low_s_maxs_c, 12, 11, 2, low(c)).
table(low_mins_maxs_c, 12, 9, 2, low(c)).
table(low_v_minc_maxc_c, 12, 6, 3, low(c)).
table(low_v_minc_s_c, 12, 12, 3, low(c)).
table(low_v_minc_mins_c, 12, 12, 3, low(c)).
table(low_v_minc_maxs_c, 12, 12, 3, low(c)).
table(low_v_maxc_s_c, 12, 4, 3, low(c)).
table(low_v_maxc_mins_c, 12, 4, 3, low(c)).
table(low_v_maxc_maxs_c, 12, 4, 3, low(c)).
table(low_v_s_mins_c, 12, 14, 3, low(c)).
table(low_v_s_maxs_c, 12, 14, 3, low(c)).
table(low_v_mins_maxs_c, 12, 14, 3, low(c)).
table(low_minc_maxc_s_c, 12, 5, 3, low(c)).
table(low_minc_maxc_mins_c, 12, 12, 3, low(c)).
table(low_minc_maxc_maxs_c, 12, 12, 3, low(c)).
table(low_minc_s_mins_c, 12, 6, 3, low(c)).
table(low_minc_s_maxs_c, 12, 7, 3, low(c)).
table(low_minc_mins_maxs_c, 12, 7, 3, low(c)).
table(low_maxc_s_mins_c, 12, 4, 3, low(c)).
table(low_maxc_s_maxs_c, 12, 4, 3, low(c)).
table(low_maxc_mins_maxs_c, 12, 6, 3, low(c)).
table(low_s_mins_maxs_c, 12, 12, 3, low(c)).
table(low_v_minc_maxc_s_c, 12, 7, 4, low(c)).
table(low_v_minc_maxc_mins_c, 12, 7, 4, low(c)).
table(low_v_minc_maxc_maxs_c, 12, 7, 4, low(c)).
table(low_v_minc_s_mins_c, 12, 13, 4, low(c)).
table(low_v_minc_s_maxs_c, 12, 13, 4, low(c)).
table(low_v_minc_mins_maxs_c, 12, 13, 4, low(c)).
table(low_v_maxc_s_mins_c, 12, 5, 4, low(c)).
table(low_v_maxc_s_maxs_c, 12, 5, 4, low(c)).
table(low_v_maxc_mins_maxs_c, 12, 5, 4, low(c)).
table(low_v_s_mins_maxs_c, 12, 15, 4, low(c)).
table(low_minc_maxc_s_mins_c, 12, 6, 4, low(c)).
table(low_minc_maxc_s_maxs_c, 12, 6, 4, low(c)).
table(low_minc_maxc_mins_maxs_c, 12, 13, 4, low(c)).
table(low_minc_s_mins_maxs_c, 12, 8, 4, low(c)).
table(low_maxc_s_mins_maxs_c, 12, 5, 4, low(c)).
table(low_v_minc_maxc_s_mins_c, 12, 8, 5, low(c)).
table(low_v_minc_maxc_s_maxs_c, 12, 8, 5, low(c)).
table(low_v_minc_maxc_mins_maxs_c, 12, 8, 5, low(c)).
table(low_v_minc_s_mins_maxs_c, 12, 14, 5, low(c)).
table(low_v_maxc_s_mins_maxs_c, 12, 6, 5, low(c)).
table(low_minc_maxc_s_mins_maxs_c, 12, 7, 5, low(c)).
table(low_v_minc_maxc_s_mins_maxs_c, 12, 9, 6, low(c)).
table(up_v_c, 12, 18, 1, up(c)).
table(up_s_c, 12, 3, 1, up(c)).
table(up_v_minc_c, 12, 8, 2, up(c)).
table(up_v_maxc_c, 12, 11, 2, up(c)).
table(up_v_s_c, 12, 4, 2, up(c)).
table(up_v_mins_c, 12, 12, 2, up(c)).
table(up_v_maxs_c, 12, 18, 2, up(c)).
table(up_minc_s_c, 12, 3, 2, up(c)).
table(up_maxc_s_c, 12, 4, 2, up(c)).
table(up_s_mins_c, 12, 5, 2, up(c)).
table(up_s_maxs_c, 12, 5, 2, up(c)).
table(up_v_minc_maxc_c, 12, 11, 3, up(c)).
table(up_v_minc_s_c, 12, 6, 3, up(c)).
table(up_v_minc_mins_c, 12, 7, 3, up(c)).
table(up_v_minc_maxs_c, 12, 7, 3, up(c)).
table(up_v_maxc_s_c, 12, 5, 3, up(c)).
table(up_v_maxc_mins_c, 12, 12, 3, up(c)).
table(up_v_maxc_maxs_c, 12, 13, 3, up(c)).
table(up_v_s_mins_c, 12, 8, 3, up(c)).
table(up_v_s_maxs_c, 12, 6, 3, up(c)).
table(up_v_mins_maxs_c, 12, 18, 3, up(c)).
table(up_minc_maxc_s_c, 12, 5, 3, up(c)).
table(up_minc_s_mins_c, 12, 4, 3, up(c)).
table(up_minc_s_maxs_c, 12, 4, 3, up(c)).
table(up_maxc_s_mins_c, 12, 5, 3, up(c)).
table(up_maxc_s_maxs_c, 12, 5, 3, up(c)).
table(up_s_mins_maxs_c, 12, 8, 3, up(c)).
table(up_v_minc_maxc_s_c, 12, 7, 4, up(c)).
table(up_v_minc_maxc_mins_c, 12, 12, 4, up(c)).
table(up_v_minc_maxc_maxs_c, 12, 12, 4, up(c)).
table(up_v_minc_s_mins_c, 12, 7, 4, up(c)).
table(up_v_minc_s_maxs_c, 12, 7, 4, up(c)).
table(up_v_minc_mins_maxs_c, 12, 8, 4, up(c)).
table(up_v_maxc_s_mins_c, 12, 9, 4, up(c)).
table(up_v_maxc_s_maxs_c, 12, 6, 4, up(c)).
table(up_v_maxc_mins_maxs_c, 12, 14, 4, up(c)).
table(up_v_s_mins_maxs_c, 12, 10, 4, up(c)).
table(up_minc_maxc_s_mins_c, 12, 6, 4, up(c)).
table(up_minc_maxc_s_maxs_c, 12, 6, 4, up(c)).
table(up_minc_s_mins_maxs_c, 12, 5, 4, up(c)).
table(up_maxc_s_mins_maxs_c, 12, 6, 4, up(c)).
table(up_v_minc_maxc_s_mins_c, 12, 8, 5, up(c)).
table(up_v_minc_maxc_s_maxs_c, 12, 8, 5, up(c)).
table(up_v_minc_maxc_mins_maxs_c, 12, 13, 5, up(c)).
table(up_v_minc_s_mins_maxs_c, 12, 8, 5, up(c)).
table(up_v_maxc_s_mins_maxs_c, 12, 7, 5, up(c)).
table(up_minc_maxc_s_mins_maxs_c, 12, 7, 5, up(c)).
table(up_v_minc_maxc_s_mins_maxs_c, 12, 9, 6, up(c)).
table(low_v_minc, 12, 3, 1, low(minc)).
table(low_c_minc, 12, 3, 1, low(minc)).
table(low_maxc_minc, 12, 3, 1, low(minc)).
table(low_s_minc, 12, 3, 1, low(minc)).
table(low_mins_minc, 12, 2, 1, low(minc)).
table(low_maxs_minc, 12, 3, 1, low(minc)).
table(low_v_c_minc, 12, 5, 2, low(minc)).
table(low_v_maxc_minc, 12, 4, 2, low(minc)).
table(low_v_s_minc, 12, 4, 2, low(minc)).
table(low_v_mins_minc, 12, 3, 2, low(minc)).
table(low_v_maxs_minc, 12, 4, 2, low(minc)).
table(low_c_maxc_minc, 12, 4, 2, low(minc)).
table(low_c_s_minc, 12, 4, 2, low(minc)).
table(low_c_mins_minc, 12, 3, 2, low(minc)).
table(low_c_maxs_minc, 12, 4, 2, low(minc)).
table(low_maxc_s_minc, 12, 4, 2, low(minc)).
table(low_maxc_mins_minc, 12, 4, 2, low(minc)).
table(low_maxc_maxs_minc, 12, 4, 2, low(minc)).
table(low_s_mins_minc, 12, 3, 2, low(minc)).
table(low_s_maxs_minc, 12, 4, 2, low(minc)).
table(low_mins_maxs_minc, 12, 3, 2, low(minc)).
table(low_v_c_maxc_minc, 12, 6, 3, low(minc)).
table(low_v_c_s_minc, 12, 6, 3, low(minc)).
table(low_v_c_mins_minc, 12, 6, 3, low(minc)).
table(low_v_c_maxs_minc, 12, 6, 3, low(minc)).
table(low_v_maxc_s_minc, 12, 5, 3, low(minc)).
table(low_v_maxc_mins_minc, 12, 5, 3, low(minc)).
table(low_v_maxc_maxs_minc, 12, 5, 3, low(minc)).
table(low_v_s_mins_minc, 12, 4, 3, low(minc)).
table(low_v_s_maxs_minc, 12, 5, 3, low(minc)).
table(low_v_mins_maxs_minc, 12, 4, 3, low(minc)).
table(low_c_maxc_s_minc, 12, 5, 3, low(minc)).
table(low_c_maxc_mins_minc, 12, 5, 3, low(minc)).
table(low_c_maxc_maxs_minc, 12, 5, 3, low(minc)).
table(low_c_s_mins_minc, 12, 4, 3, low(minc)).
table(low_c_s_maxs_minc, 12, 6, 3, low(minc)).
table(low_c_mins_maxs_minc, 12, 4, 3, low(minc)).
table(low_maxc_s_mins_minc, 12, 5, 3, low(minc)).
table(low_maxc_s_maxs_minc, 12, 5, 3, low(minc)).
table(low_maxc_mins_maxs_minc, 12, 5, 3, low(minc)).
table(low_s_mins_maxs_minc, 12, 4, 3, low(minc)).
table(low_v_c_maxc_s_minc, 12, 7, 4, low(minc)).
table(low_v_c_maxc_mins_minc, 12, 7, 4, low(minc)).
table(low_v_c_maxc_maxs_minc, 12, 7, 4, low(minc)).
table(low_v_c_s_mins_minc, 12, 7, 4, low(minc)).
table(low_v_c_s_maxs_minc, 12, 7, 4, low(minc)).
table(low_v_c_mins_maxs_minc, 12, 7, 4, low(minc)).
table(low_v_maxc_s_mins_minc, 12, 6, 4, low(minc)).
table(low_v_maxc_s_maxs_minc, 12, 6, 4, low(minc)).
table(low_v_maxc_mins_maxs_minc, 12, 6, 4, low(minc)).
table(low_v_s_mins_maxs_minc, 12, 5, 4, low(minc)).
table(low_c_maxc_s_mins_minc, 12, 6, 4, low(minc)).
table(low_c_maxc_s_maxs_minc, 12, 6, 4, low(minc)).
table(low_c_maxc_mins_maxs_minc, 12, 6, 4, low(minc)).
table(low_c_s_mins_maxs_minc, 12, 6, 4, low(minc)).
table(low_maxc_s_mins_maxs_minc, 12, 6, 4, low(minc)).
table(low_v_c_maxc_s_mins_minc, 12, 8, 5, low(minc)).
table(low_v_c_maxc_s_maxs_minc, 12, 8, 5, low(minc)).
table(low_v_c_maxc_mins_maxs_minc, 12, 8, 5, low(minc)).
table(low_v_c_s_mins_maxs_minc, 12, 8, 5, low(minc)).
table(low_v_maxc_s_mins_maxs_minc, 12, 7, 5, low(minc)).
table(low_c_maxc_s_mins_maxs_minc, 12, 7, 5, low(minc)).
table(low_v_c_maxc_s_mins_maxs_minc, 12, 9, 6, low(minc)).
table(up_v_minc, 12, 11, 1, up(minc)).
table(up_maxc_minc, 12, 6, 1, up(minc)).
table(up_v_c_minc, 12, 5, 2, up(minc)).
table(up_v_maxc_minc, 12, 6, 2, up(minc)).
table(up_v_s_minc, 12, 13, 2, up(minc)).
table(up_v_mins_minc, 12, 12, 2, up(minc)).
table(up_v_maxs_minc, 12, 13, 2, up(minc)).
table(up_c_maxc_minc, 12, 4, 2, up(minc)).
table(up_maxc_s_minc, 12, 4, 2, up(minc)).
table(up_maxc_mins_minc, 12, 4, 2, up(minc)).
table(up_maxc_maxs_minc, 12, 7, 2, up(minc)).
table(up_s_maxs_minc, 12, 14, 2, up(minc)).
table(up_v_c_maxc_minc, 12, 6, 3, up(minc)).
table(up_v_c_s_minc, 12, 6, 3, up(minc)).
table(up_v_c_mins_minc, 12, 6, 3, up(minc)).
table(up_v_c_maxs_minc, 12, 6, 3, up(minc)).
table(up_v_maxc_s_minc, 12, 7, 3, up(minc)).
table(up_v_maxc_mins_minc, 12, 7, 3, up(minc)).
table(up_v_maxc_maxs_minc, 12, 7, 3, up(minc)).
table(up_v_s_mins_minc, 12, 14, 3, up(minc)).
table(up_v_s_maxs_minc, 12, 14, 3, up(minc)).
table(up_v_mins_maxs_minc, 12, 14, 3, up(minc)).
table(up_c_maxc_s_minc, 12, 5, 3, up(minc)).
table(up_c_maxc_mins_minc, 12, 5, 3, up(minc)).
table(up_c_maxc_maxs_minc, 12, 5, 3, up(minc)).
table(up_c_s_maxs_minc, 12, 4, 3, up(minc)).
table(up_maxc_s_mins_minc, 12, 5, 3, up(minc)).
table(up_maxc_s_maxs_minc, 12, 5, 3, up(minc)).
table(up_maxc_mins_maxs_minc, 12, 5, 3, up(minc)).
table(up_s_mins_maxs_minc, 12, 15, 3, up(minc)).
table(up_v_c_maxc_s_minc, 12, 7, 4, up(minc)).
table(up_v_c_maxc_mins_minc, 12, 7, 4, up(minc)).
table(up_v_c_maxc_maxs_minc, 12, 7, 4, up(minc)).
table(up_v_c_s_mins_minc, 12, 7, 4, up(minc)).
table(up_v_c_s_maxs_minc, 12, 7, 4, up(minc)).
table(up_v_c_mins_maxs_minc, 12, 7, 4, up(minc)).
table(up_v_maxc_s_mins_minc, 12, 6, 4, up(minc)).
table(up_v_maxc_s_maxs_minc, 12, 6, 4, up(minc)).
table(up_v_maxc_mins_maxs_minc, 12, 6, 4, up(minc)).
table(up_v_s_mins_maxs_minc, 12, 15, 4, up(minc)).
table(up_c_maxc_s_mins_minc, 12, 6, 4, up(minc)).
table(up_c_maxc_s_maxs_minc, 12, 6, 4, up(minc)).
table(up_c_maxc_mins_maxs_minc, 12, 6, 4, up(minc)).
table(up_c_s_mins_maxs_minc, 12, 5, 4, up(minc)).
table(up_maxc_s_mins_maxs_minc, 12, 6, 4, up(minc)).
table(up_v_c_maxc_s_mins_minc, 12, 8, 5, up(minc)).
table(up_v_c_maxc_s_maxs_minc, 12, 8, 5, up(minc)).
table(up_v_c_maxc_mins_maxs_minc, 12, 8, 5, up(minc)).
table(up_v_c_s_mins_maxs_minc, 12, 8, 5, up(minc)).
table(up_v_maxc_s_mins_maxs_minc, 12, 7, 5, up(minc)).
table(up_c_maxc_s_mins_maxs_minc, 12, 7, 5, up(minc)).
table(up_v_c_maxc_s_mins_maxs_minc, 12, 9, 6, up(minc)).
table(low_v_maxc, 12, 18, 1, low(maxc)).
table(low_c_maxc, 12, 18, 1, low(maxc)).
table(low_minc_maxc, 12, 6, 1, low(maxc)).
table(low_s_maxc, 12, 18, 1, low(maxc)).
table(low_mins_maxc, 12, 9, 1, low(maxc)).
table(low_maxs_maxc, 12, 2, 1, low(maxc)).
table(low_v_c_maxc, 12, 3, 2, low(maxc)).
table(low_v_minc_maxc, 12, 6, 2, low(maxc)).
table(low_v_s_maxc, 12, 4, 2, low(maxc)).
table(low_v_mins_maxc, 12, 9, 2, low(maxc)).
table(low_v_maxs_maxc, 12, 3, 2, low(maxc)).
table(low_c_minc_maxc, 12, 11, 2, low(maxc)).
table(low_c_s_maxc, 12, 4, 2, low(maxc)).
table(low_c_mins_maxc, 12, 18, 2, low(maxc)).
table(low_c_maxs_maxc, 12, 3, 2, low(maxc)).
table(low_minc_s_maxc, 12, 7, 2, low(maxc)).
table(low_minc_mins_maxc, 12, 6, 2, low(maxc)).
table(low_minc_maxs_maxc, 12, 4, 2, low(maxc)).
table(low_s_mins_maxc, 12, 18, 2, low(maxc)).
table(low_s_maxs_maxc, 12, 3, 2, low(maxc)).
table(low_mins_maxs_maxc, 12, 3, 2, low(maxc)).
table(low_v_c_minc_maxc, 12, 6, 3, low(maxc)).
table(low_v_c_s_maxc, 12, 4, 3, low(maxc)).
table(low_v_c_mins_maxc, 12, 4, 3, low(maxc)).
table(low_v_c_maxs_maxc, 12, 4, 3, low(maxc)).
table(low_v_minc_s_maxc, 12, 5, 3, low(maxc)).
table(low_v_minc_mins_maxc, 12, 7, 3, low(maxc)).
table(low_v_minc_maxs_maxc, 12, 5, 3, low(maxc)).
table(low_v_s_mins_maxc, 12, 5, 3, low(maxc)).
table(low_v_s_maxs_maxc, 12, 4, 3, low(maxc)).
table(low_v_mins_maxs_maxc, 12, 4, 3, low(maxc)).
table(low_c_minc_s_maxc, 12, 5, 3, low(maxc)).
table(low_c_minc_mins_maxc, 12, 8, 3, low(maxc)).
table(low_c_minc_maxs_maxc, 12, 5, 3, low(maxc)).
table(low_c_s_mins_maxc, 12, 4, 3, low(maxc)).
table(low_c_s_maxs_maxc, 12, 4, 3, low(maxc)).
table(low_c_mins_maxs_maxc, 12, 4, 3, low(maxc)).
table(low_minc_s_mins_maxc, 12, 7, 3, low(maxc)).
table(low_minc_s_maxs_maxc, 12, 5, 3, low(maxc)).
table(low_minc_mins_maxs_maxc, 12, 5, 3, low(maxc)).
table(low_s_mins_maxs_maxc, 12, 4, 3, low(maxc)).
table(low_v_c_minc_s_maxc, 12, 7, 4, low(maxc)).
table(low_v_c_minc_mins_maxc, 12, 7, 4, low(maxc)).
table(low_v_c_minc_maxs_maxc, 12, 7, 4, low(maxc)).
table(low_v_c_s_mins_maxc, 12, 5, 4, low(maxc)).
table(low_v_c_s_maxs_maxc, 12, 5, 4, low(maxc)).
table(low_v_c_mins_maxs_maxc, 12, 5, 4, low(maxc)).
table(low_v_minc_s_mins_maxc, 12, 6, 4, low(maxc)).
table(low_v_minc_s_maxs_maxc, 12, 6, 4, low(maxc)).
table(low_v_minc_mins_maxs_maxc, 12, 6, 4, low(maxc)).
table(low_v_s_mins_maxs_maxc, 12, 5, 4, low(maxc)).
table(low_c_minc_s_mins_maxc, 12, 6, 4, low(maxc)).
table(low_c_minc_s_maxs_maxc, 12, 6, 4, low(maxc)).
table(low_c_minc_mins_maxs_maxc, 12, 6, 4, low(maxc)).
table(low_c_s_mins_maxs_maxc, 12, 5, 4, low(maxc)).
table(low_minc_s_mins_maxs_maxc, 12, 6, 4, low(maxc)).
table(low_v_c_minc_s_mins_maxc, 12, 8, 5, low(maxc)).
table(low_v_c_minc_s_maxs_maxc, 12, 8, 5, low(maxc)).
table(low_v_c_minc_mins_maxs_maxc, 12, 8, 5, low(maxc)).
table(low_v_c_s_mins_maxs_maxc, 12, 6, 5, low(maxc)).
table(low_v_minc_s_mins_maxs_maxc, 12, 7, 5, low(maxc)).
table(low_c_minc_s_mins_maxs_maxc, 12, 7, 5, low(maxc)).
table(low_v_c_minc_s_mins_maxs_maxc, 12, 9, 6, low(maxc)).
table(up_v_maxc, 12, 11, 1, up(maxc)).
table(up_v_c_maxc, 12, 11, 2, up(maxc)).
table(up_v_minc_maxc, 12, 11, 2, up(maxc)).
table(up_v_s_maxc, 12, 13, 2, up(maxc)).
table(up_v_mins_maxc, 12, 12, 2, up(maxc)).
table(up_v_maxs_maxc, 12, 13, 2, up(maxc)).
table(up_v_c_minc_maxc, 12, 11, 3, up(maxc)).
table(up_v_c_s_maxc, 12, 13, 3, up(maxc)).
table(up_v_c_mins_maxc, 12, 12, 3, up(maxc)).
table(up_v_c_maxs_maxc, 12, 13, 3, up(maxc)).
table(up_v_minc_s_maxc, 12, 12, 3, up(maxc)).
table(up_v_minc_mins_maxc, 12, 12, 3, up(maxc)).
table(up_v_minc_maxs_maxc, 12, 12, 3, up(maxc)).
table(up_v_s_mins_maxc, 12, 14, 3, up(maxc)).
table(up_v_s_maxs_maxc, 12, 14, 3, up(maxc)).
table(up_v_mins_maxs_maxc, 12, 14, 3, up(maxc)).
table(up_v_c_minc_s_maxc, 12, 12, 4, up(maxc)).
table(up_v_c_minc_mins_maxc, 12, 12, 4, up(maxc)).
table(up_v_c_minc_maxs_maxc, 12, 12, 4, up(maxc)).
table(up_v_c_s_mins_maxc, 12, 14, 4, up(maxc)).
table(up_v_c_s_maxs_maxc, 12, 6, 4, up(maxc)).
table(up_v_c_mins_maxs_maxc, 12, 14, 4, up(maxc)).
table(up_v_minc_s_mins_maxc, 12, 13, 4, up(maxc)).
table(up_v_minc_s_maxs_maxc, 12, 13, 4, up(maxc)).
table(up_v_minc_mins_maxs_maxc, 12, 13, 4, up(maxc)).
table(up_v_s_mins_maxs_maxc, 12, 15, 4, up(maxc)).
table(up_v_c_minc_s_mins_maxc, 12, 13, 5, up(maxc)).
table(up_v_c_minc_s_maxs_maxc, 12, 8, 5, up(maxc)).
table(up_v_c_minc_mins_maxs_maxc, 12, 13, 5, up(maxc)).
table(up_v_c_s_mins_maxs_maxc, 12, 10, 5, up(maxc)).
table(up_v_minc_s_mins_maxs_maxc, 12, 14, 5, up(maxc)).
table(up_v_c_minc_s_mins_maxs_maxc, 12, 9, 6, up(maxc)).
table(low_v_s, 12, 18, 1, low(s)).
table(low_c_s, 12, 3, 1, low(s)).
table(low_minc_s, 12, 18, 1, low(s)).
table(low_maxc_s, 12, 18, 1, low(s)).
table(low_mins_s, 12, 18, 1, low(s)).
table(low_maxs_s, 12, 18, 1, low(s)).
table(low_v_c_s, 12, 4, 2, low(s)).
table(low_v_minc_s, 12, 18, 2, low(s)).
table(low_v_maxc_s, 12, 6, 2, low(s)).
table(low_v_mins_s, 12, 9, 2, low(s)).
table(low_v_maxs_s, 12, 3, 2, low(s)).
table(low_c_minc_s, 12, 5, 2, low(s)).
table(low_c_maxc_s, 12, 5, 2, low(s)).
table(low_c_mins_s, 12, 5, 2, low(s)).
table(low_c_maxs_s, 12, 5, 2, low(s)).
table(low_minc_maxc_s, 12, 18, 2, low(s)).
table(low_minc_mins_s, 12, 12, 2, low(s)).
table(low_minc_maxs_s, 12, 14, 2, low(s)).
table(low_maxc_mins_s, 12, 7, 2, low(s)).
table(low_maxc_maxs_s, 12, 14, 2, low(s)).
table(low_mins_maxs_s, 12, 12, 2, low(s)).
table(low_v_c_minc_s, 12, 8, 3, low(s)).
table(low_v_c_maxc_s, 12, 6, 3, low(s)).
table(low_v_c_mins_s, 12, 8, 3, low(s)).
table(low_v_c_maxs_s, 12, 4, 3, low(s)).
table(low_v_minc_maxc_s, 12, 10, 3, low(s)).
table(low_v_minc_mins_s, 12, 13, 3, low(s)).
table(low_v_minc_maxs_s, 12, 4, 3, low(s)).
table(low_v_maxc_mins_s, 12, 4, 3, low(s)).
table(low_v_maxc_maxs_s, 12, 4, 3, low(s)).
table(low_v_mins_maxs_s, 12, 4, 3, low(s)).
table(low_c_minc_maxc_s, 12, 8, 3, low(s)).
table(low_c_minc_mins_s, 12, 4, 3, low(s)).
table(low_c_minc_maxs_s, 12, 5, 3, low(s)).
table(low_c_maxc_mins_s, 12, 8, 3, low(s)).
table(low_c_maxc_maxs_s, 12, 5, 3, low(s)).
table(low_c_mins_maxs_s, 12, 8, 3, low(s)).
table(low_minc_maxc_mins_s, 12, 13, 3, low(s)).
table(low_minc_maxc_maxs_s, 12, 14, 3, low(s)).
table(low_minc_mins_maxs_s, 12, 8, 3, low(s)).
table(low_maxc_mins_maxs_s, 12, 9, 3, low(s)).
table(low_v_c_minc_maxc_s, 12, 10, 4, low(s)).
table(low_v_c_minc_mins_s, 12, 7, 4, low(s)).
table(low_v_c_minc_maxs_s, 12, 7, 4, low(s)).
table(low_v_c_maxc_mins_s, 12, 5, 4, low(s)).
table(low_v_c_maxc_maxs_s, 12, 5, 4, low(s)).
table(low_v_c_mins_maxs_s, 12, 5, 4, low(s)).
table(low_v_minc_maxc_mins_s, 12, 8, 4, low(s)).
table(low_v_minc_maxc_maxs_s, 12, 6, 4, low(s)).
table(low_v_minc_mins_maxs_s, 12, 5, 4, low(s)).
table(low_v_maxc_mins_maxs_s, 12, 5, 4, low(s)).
table(low_c_minc_maxc_mins_s, 12, 6, 4, low(s)).
table(low_c_minc_maxc_maxs_s, 12, 7, 4, low(s)).
table(low_c_minc_mins_maxs_s, 12, 5, 4, low(s)).
table(low_c_maxc_mins_maxs_s, 12, 6, 4, low(s)).
table(low_minc_maxc_mins_maxs_s, 12, 14, 4, low(s)).
table(low_v_c_minc_maxc_mins_s, 12, 8, 5, low(s)).
table(low_v_c_minc_maxc_maxs_s, 12, 8, 5, low(s)).
table(low_v_c_minc_mins_maxs_s, 12, 8, 5, low(s)).
table(low_v_c_maxc_mins_maxs_s, 12, 6, 5, low(s)).
table(low_v_minc_maxc_mins_maxs_s, 12, 7, 5, low(s)).
table(low_c_minc_maxc_mins_maxs_s, 12, 7, 5, low(s)).
table(low_v_c_minc_maxc_mins_maxs_s, 12, 9, 6, low(s)).
table(up_v_s, 12, 7, 1, up(s)).
table(up_v_c_s, 12, 8, 2, up(s)).
table(up_v_minc_s, 12, 8, 2, up(s)).
table(up_v_maxc_s, 12, 8, 2, up(s)).
table(up_v_mins_s, 12, 4, 2, up(s)).
table(up_v_maxs_s, 12, 7, 2, up(s)).
table(up_c_maxc_s, 12, 9, 2, up(s)).
table(up_v_c_minc_s, 12, 11, 3, up(s)).
table(up_v_c_maxc_s, 12, 9, 3, up(s)).
table(up_v_c_mins_s, 12, 5, 3, up(s)).
table(up_v_c_maxs_s, 12, 8, 3, up(s)).
table(up_v_minc_maxc_s, 12, 10, 3, up(s)).
table(up_v_minc_mins_s, 12, 5, 3, up(s)).
table(up_v_minc_maxs_s, 12, 8, 3, up(s)).
table(up_v_maxc_mins_s, 12, 5, 3, up(s)).
table(up_v_maxc_maxs_s, 12, 8, 3, up(s)).
table(up_v_mins_maxs_s, 12, 7, 3, up(s)).
table(up_c_minc_maxc_s, 12, 12, 3, up(s)).
table(up_c_maxc_mins_s, 12, 5, 3, up(s)).
table(up_c_maxc_maxs_s, 12, 9, 3, up(s)).
table(up_v_c_minc_maxc_s, 12, 12, 4, up(s)).
table(up_v_c_minc_mins_s, 12, 8, 4, up(s)).
table(up_v_c_minc_maxs_s, 12, 11, 4, up(s)).
table(up_v_c_maxc_mins_s, 12, 6, 4, up(s)).
table(up_v_c_maxc_maxs_s, 12, 9, 4, up(s)).
table(up_v_c_mins_maxs_s, 12, 8, 4, up(s)).
table(up_v_minc_maxc_mins_s, 12, 7, 4, up(s)).
table(up_v_minc_maxc_maxs_s, 12, 10, 4, up(s)).
table(up_v_minc_mins_maxs_s, 12, 8, 4, up(s)).
table(up_v_maxc_mins_maxs_s, 12, 8, 4, up(s)).
table(up_c_minc_maxc_mins_s, 12, 7, 4, up(s)).
table(up_c_minc_maxc_maxs_s, 12, 12, 4, up(s)).
table(up_c_maxc_mins_maxs_s, 12, 6, 4, up(s)).
table(up_v_c_minc_maxc_mins_s, 12, 9, 5, up(s)).
table(up_v_c_minc_maxc_maxs_s, 12, 12, 5, up(s)).
table(up_v_c_minc_mins_maxs_s, 12, 11, 5, up(s)).
table(up_v_c_maxc_mins_maxs_s, 12, 9, 5, up(s)).
table(up_v_minc_maxc_mins_maxs_s, 12, 10, 5, up(s)).
table(up_c_minc_maxc_mins_maxs_s, 12, 8, 5, up(s)).
table(up_v_c_minc_maxc_mins_maxs_s, 12, 12, 6, up(s)).
table(low_v_mins, 12, 2, 1, low(mins)).
table(low_c_mins, 12, 2, 1, low(mins)).
table(low_minc_mins, 12, 2, 1, low(mins)).
table(low_maxc_mins, 12, 2, 1, low(mins)).
table(low_s_mins, 12, 2, 1, low(mins)).
table(low_maxs_mins, 12, 2, 1, low(mins)).
table(low_v_c_mins, 12, 3, 2, low(mins)).
table(low_v_minc_mins, 12, 3, 2, low(mins)).
table(low_v_maxc_mins, 12, 3, 2, low(mins)).
table(low_v_s_mins, 12, 3, 2, low(mins)).
table(low_v_maxs_mins, 12, 3, 2, low(mins)).
table(low_c_minc_mins, 12, 3, 2, low(mins)).
table(low_c_maxc_mins, 12, 3, 2, low(mins)).
table(low_c_s_mins, 12, 3, 2, low(mins)).
table(low_c_maxs_mins, 12, 3, 2, low(mins)).
table(low_minc_maxc_mins, 12, 4, 2, low(mins)).
table(low_minc_s_mins, 12, 3, 2, low(mins)).
table(low_minc_maxs_mins, 12, 3, 2, low(mins)).
table(low_maxc_s_mins, 12, 3, 2, low(mins)).
table(low_maxc_maxs_mins, 12, 3, 2, low(mins)).
table(low_s_maxs_mins, 12, 3, 2, low(mins)).
table(low_v_c_minc_mins, 12, 6, 3, low(mins)).
table(low_v_c_maxc_mins, 12, 4, 3, low(mins)).
table(low_v_c_s_mins, 12, 4, 3, low(mins)).
table(low_v_c_maxs_mins, 12, 4, 3, low(mins)).
table(low_v_minc_maxc_mins, 12, 5, 3, low(mins)).
table(low_v_minc_s_mins, 12, 4, 3, low(mins)).
table(low_v_minc_maxs_mins, 12, 4, 3, low(mins)).
table(low_v_maxc_s_mins, 12, 4, 3, low(mins)).
table(low_v_maxc_maxs_mins, 12, 4, 3, low(mins)).
table(low_v_s_maxs_mins, 12, 4, 3, low(mins)).
table(low_c_minc_maxc_mins, 12, 5, 3, low(mins)).
table(low_c_minc_s_mins, 12, 4, 3, low(mins)).
table(low_c_minc_maxs_mins, 12, 4, 3, low(mins)).
table(low_c_maxc_s_mins, 12, 4, 3, low(mins)).
table(low_c_maxc_maxs_mins, 12, 4, 3, low(mins)).
table(low_c_s_maxs_mins, 12, 4, 3, low(mins)).
table(low_minc_maxc_s_mins, 12, 5, 3, low(mins)).
table(low_minc_maxc_maxs_mins, 12, 5, 3, low(mins)).
table(low_minc_s_maxs_mins, 12, 4, 3, low(mins)).
table(low_maxc_s_maxs_mins, 12, 4, 3, low(mins)).
table(low_v_c_minc_maxc_mins, 12, 7, 4, low(mins)).
table(low_v_c_minc_s_mins, 12, 7, 4, low(mins)).
table(low_v_c_minc_maxs_mins, 12, 7, 4, low(mins)).
table(low_v_c_maxc_s_mins, 12, 5, 4, low(mins)).
table(low_v_c_maxc_maxs_mins, 12, 5, 4, low(mins)).
table(low_v_c_s_maxs_mins, 12, 5, 4, low(mins)).
table(low_v_minc_maxc_s_mins, 12, 6, 4, low(mins)).
table(low_v_minc_maxc_maxs_mins, 12, 6, 4, low(mins)).
table(low_v_minc_s_maxs_mins, 12, 5, 4, low(mins)).
table(low_v_maxc_s_maxs_mins, 12, 5, 4, low(mins)).
table(low_c_minc_maxc_s_mins, 12, 6, 4, low(mins)).
table(low_c_minc_maxc_maxs_mins, 12, 6, 4, low(mins)).
table(low_c_minc_s_maxs_mins, 12, 5, 4, low(mins)).
table(low_c_maxc_s_maxs_mins, 12, 5, 4, low(mins)).
table(low_minc_maxc_s_maxs_mins, 12, 6, 4, low(mins)).
table(low_v_c_minc_maxc_s_mins, 12, 8, 5, low(mins)).
table(low_v_c_minc_maxc_maxs_mins, 12, 8, 5, low(mins)).
table(low_v_c_minc_s_maxs_mins, 12, 8, 5, low(mins)).
table(low_v_c_maxc_s_maxs_mins, 12, 6, 5, low(mins)).
table(low_v_minc_maxc_s_maxs_mins, 12, 7, 5, low(mins)).
table(low_c_minc_maxc_s_maxs_mins, 12, 7, 5, low(mins)).
table(low_v_c_minc_maxc_s_maxs_mins, 12, 9, 6, low(mins)).
table(up_v_mins, 12, 18, 1, up(mins)).
table(up_maxc_mins, 12, 9, 1, up(mins)).
table(up_maxs_mins, 12, 2, 1, up(mins)).
table(up_v_c_mins, 12, 6, 2, up(mins)).
table(up_v_minc_mins, 12, 3, 2, up(mins)).
table(up_v_maxc_mins, 12, 7, 2, up(mins)).
table(up_v_s_mins, 12, 3, 2, up(mins)).
table(up_v_maxs_mins, 12, 3, 2, up(mins)).
table(up_c_maxc_mins, 12, 5, 2, up(mins)).
table(up_c_maxs_mins, 12, 3, 2, up(mins)).
table(up_minc_maxc_mins, 12, 4, 2, up(mins)).
table(up_minc_maxs_mins, 12, 3, 2, up(mins)).
table(up_maxc_s_mins, 12, 3, 2, up(mins)).
table(up_maxc_maxs_mins, 12, 3, 2, up(mins)).
table(up_s_maxs_mins, 12, 3, 2, up(mins)).
table(up_v_c_minc_mins, 12, 6, 3, up(mins)).
table(up_v_c_maxc_mins, 12, 7, 3, up(mins)).
table(up_v_c_s_mins, 12, 4, 3, up(mins)).
table(up_v_c_maxs_mins, 12, 4, 3, up(mins)).
table(up_v_minc_maxc_mins, 12, 5, 3, up(mins)).
table(up_v_minc_s_mins, 12, 4, 3, up(mins)).
table(up_v_minc_maxs_mins, 12, 4, 3, up(mins)).
table(up_v_maxc_s_mins, 12, 4, 3, up(mins)).
table(up_v_maxc_maxs_mins, 12, 4, 3, up(mins)).
table(up_v_s_maxs_mins, 12, 4, 3, up(mins)).
table(up_c_minc_maxc_mins, 12, 5, 3, up(mins)).
table(up_c_minc_maxs_mins, 12, 4, 3, up(mins)).
table(up_c_maxc_s_mins, 12, 4, 3, up(mins)).
table(up_c_maxc_maxs_mins, 12, 4, 3, up(mins)).
table(up_c_s_maxs_mins, 12, 4, 3, up(mins)).
table(up_minc_maxc_s_mins, 12, 5, 3, up(mins)).
table(up_minc_maxc_maxs_mins, 12, 5, 3, up(mins)).
table(up_minc_s_maxs_mins, 12, 4, 3, up(mins)).
table(up_maxc_s_maxs_mins, 12, 4, 3, up(mins)).
table(up_v_c_minc_maxc_mins, 12, 7, 4, up(mins)).
table(up_v_c_minc_s_mins, 12, 7, 4, up(mins)).
table(up_v_c_minc_maxs_mins, 12, 7, 4, up(mins)).
table(up_v_c_maxc_s_mins, 12, 5, 4, up(mins)).
table(up_v_c_maxc_maxs_mins, 12, 5, 4, up(mins)).
table(up_v_c_s_maxs_mins, 12, 5, 4, up(mins)).
table(up_v_minc_maxc_s_mins, 12, 6, 4, up(mins)).
table(up_v_minc_maxc_maxs_mins, 12, 6, 4, up(mins)).
table(up_v_minc_s_maxs_mins, 12, 5, 4, up(mins)).
table(up_v_maxc_s_maxs_mins, 12, 5, 4, up(mins)).
table(up_c_minc_maxc_s_mins, 12, 6, 4, up(mins)).
table(up_c_minc_maxc_maxs_mins, 12, 6, 4, up(mins)).
table(up_c_minc_s_maxs_mins, 12, 5, 4, up(mins)).
table(up_c_maxc_s_maxs_mins, 12, 5, 4, up(mins)).
table(up_minc_maxc_s_maxs_mins, 12, 6, 4, up(mins)).
table(up_v_c_minc_maxc_s_mins, 12, 8, 5, up(mins)).
table(up_v_c_minc_maxc_maxs_mins, 12, 8, 5, up(mins)).
table(up_v_c_minc_s_maxs_mins, 12, 8, 5, up(mins)).
table(up_v_c_maxc_s_maxs_mins, 12, 6, 5, up(mins)).
table(up_v_minc_maxc_s_maxs_mins, 12, 7, 5, up(mins)).
table(up_c_minc_maxc_s_maxs_mins, 12, 7, 5, up(mins)).
table(up_v_c_minc_maxc_s_maxs_mins, 12, 9, 6, up(mins)).
table(low_v_maxs, 12, 7, 1, low(maxs)).
table(low_c_maxs, 12, 4, 1, low(maxs)).
table(low_minc_maxs, 12, 4, 1, low(maxs)).
table(low_maxc_maxs, 12, 4, 1, low(maxs)).
table(low_s_maxs, 12, 7, 1, low(maxs)).
table(low_mins_maxs, 12, 2, 1, low(maxs)).
table(low_v_c_maxs, 12, 8, 2, low(maxs)).
table(low_v_minc_maxs, 12, 8, 2, low(maxs)).
table(low_v_maxc_maxs, 12, 8, 2, low(maxs)).
table(low_v_s_maxs, 12, 3, 2, low(maxs)).
table(low_v_mins_maxs, 12, 3, 2, low(maxs)).
table(low_c_minc_maxs, 12, 5, 2, low(maxs)).
table(low_c_maxc_maxs, 12, 5, 2, low(maxs)).
table(low_c_s_maxs, 12, 8, 2, low(maxs)).
table(low_c_mins_maxs, 12, 3, 2, low(maxs)).
table(low_minc_maxc_maxs, 12, 6, 2, low(maxs)).
table(low_minc_s_maxs, 12, 3, 2, low(maxs)).
table(low_minc_mins_maxs, 12, 3, 2, low(maxs)).
table(low_maxc_s_maxs, 12, 3, 2, low(maxs)).
table(low_maxc_mins_maxs, 12, 3, 2, low(maxs)).
table(low_s_mins_maxs, 12, 7, 2, low(maxs)).
table(low_v_c_minc_maxs, 12, 11, 3, low(maxs)).
table(low_v_c_maxc_maxs, 12, 9, 3, low(maxs)).
table(low_v_c_s_maxs, 12, 4, 3, low(maxs)).
table(low_v_c_mins_maxs, 12, 4, 3, low(maxs)).
table(low_v_minc_maxc_maxs, 12, 10, 3, low(maxs)).
table(low_v_minc_s_maxs, 12, 4, 3, low(maxs)).
table(low_v_minc_mins_maxs, 12, 4, 3, low(maxs)).
table(low_v_maxc_s_maxs, 12, 4, 3, low(maxs)).
table(low_v_maxc_mins_maxs, 12, 4, 3, low(maxs)).
table(low_v_s_mins_maxs, 12, 4, 3, low(maxs)).
table(low_c_minc_maxc_maxs, 12, 7, 3, low(maxs)).
table(low_c_minc_s_maxs, 12, 4, 3, low(maxs)).
table(low_c_minc_mins_maxs, 12, 4, 3, low(maxs)).
table(low_c_maxc_s_maxs, 12, 4, 3, low(maxs)).
table(low_c_maxc_mins_maxs, 12, 4, 3, low(maxs)).
table(low_c_s_mins_maxs, 12, 8, 3, low(maxs)).
table(low_minc_maxc_s_maxs, 12, 5, 3, low(maxs)).
table(low_minc_maxc_mins_maxs, 12, 5, 3, low(maxs)).
table(low_minc_s_mins_maxs, 12, 4, 3, low(maxs)).
table(low_maxc_s_mins_maxs, 12, 4, 3, low(maxs)).
table(low_v_c_minc_maxc_maxs, 12, 12, 4, low(maxs)).
table(low_v_c_minc_s_maxs, 12, 7, 4, low(maxs)).
table(low_v_c_minc_mins_maxs, 12, 7, 4, low(maxs)).
table(low_v_c_maxc_s_maxs, 12, 5, 4, low(maxs)).
table(low_v_c_maxc_mins_maxs, 12, 5, 4, low(maxs)).
table(low_v_c_s_mins_maxs, 12, 5, 4, low(maxs)).
table(low_v_minc_maxc_s_maxs, 12, 6, 4, low(maxs)).
table(low_v_minc_maxc_mins_maxs, 12, 6, 4, low(maxs)).
table(low_v_minc_s_mins_maxs, 12, 5, 4, low(maxs)).
table(low_v_maxc_s_mins_maxs, 12, 5, 4, low(maxs)).
table(low_c_minc_maxc_s_maxs, 12, 6, 4, low(maxs)).
table(low_c_minc_maxc_mins_maxs, 12, 6, 4, low(maxs)).
table(low_c_minc_s_mins_maxs, 12, 5, 4, low(maxs)).
table(low_c_maxc_s_mins_maxs, 12, 5, 4, low(maxs)).
table(low_minc_maxc_s_mins_maxs, 12, 6, 4, low(maxs)).
table(low_v_c_minc_maxc_s_maxs, 12, 8, 5, low(maxs)).
table(low_v_c_minc_maxc_mins_maxs, 12, 8, 5, low(maxs)).
table(low_v_c_minc_s_mins_maxs, 12, 8, 5, low(maxs)).
table(low_v_c_maxc_s_mins_maxs, 12, 6, 5, low(maxs)).
table(low_v_minc_maxc_s_mins_maxs, 12, 7, 5, low(maxs)).
table(low_c_minc_maxc_s_mins_maxs, 12, 7, 5, low(maxs)).
table(low_v_c_minc_maxc_s_mins_maxs, 12, 9, 6, low(maxs)).
table(up_v_maxs, 12, 18, 1, up(maxs)).
table(up_maxc_maxs, 12, 2, 1, up(maxs)).
table(up_v_c_maxs, 12, 18, 2, up(maxs)).
table(up_v_minc_maxs, 12, 12, 2, up(maxs)).
table(up_v_maxc_maxs, 12, 3, 2, up(maxs)).
table(up_v_s_maxs, 12, 7, 2, up(maxs)).
table(up_v_mins_maxs, 12, 9, 2, up(maxs)).
table(up_c_maxc_maxs, 12, 3, 2, up(maxs)).
table(up_minc_maxc_maxs, 12, 4, 2, up(maxs)).
table(up_maxc_s_maxs, 12, 3, 2, up(maxs)).
table(up_maxc_mins_maxs, 12, 3, 2, up(maxs)).
table(up_v_c_minc_maxs, 12, 12, 3, up(maxs)).
table(up_v_c_maxc_maxs, 12, 4, 3, up(maxs)).
table(up_v_c_s_maxs, 12, 8, 3, up(maxs)).
table(up_v_c_mins_maxs, 12, 18, 3, up(maxs)).
table(up_v_minc_maxc_maxs, 12, 5, 3, up(maxs)).
table(up_v_minc_s_maxs, 12, 5, 3, up(maxs)).
table(up_v_minc_mins_maxs, 12, 14, 3, up(maxs)).
table(up_v_maxc_s_maxs, 12, 4, 3, up(maxs)).
table(up_v_maxc_mins_maxs, 12, 4, 3, up(maxs)).
table(up_v_s_mins_maxs, 12, 7, 3, up(maxs)).
table(up_c_minc_maxc_maxs, 12, 5, 3, up(maxs)).
table(up_c_maxc_s_maxs, 12, 4, 3, up(maxs)).
table(up_c_maxc_mins_maxs, 12, 4, 3, up(maxs)).
table(up_minc_maxc_s_maxs, 12, 5, 3, up(maxs)).
table(up_minc_maxc_mins_maxs, 12, 5, 3, up(maxs)).
table(up_maxc_s_mins_maxs, 12, 4, 3, up(maxs)).
table(up_v_c_minc_maxc_maxs, 12, 7, 4, up(maxs)).
table(up_v_c_minc_s_maxs, 12, 7, 4, up(maxs)).
table(up_v_c_minc_mins_maxs, 12, 13, 4, up(maxs)).
table(up_v_c_maxc_s_maxs, 12, 5, 4, up(maxs)).
table(up_v_c_maxc_mins_maxs, 12, 5, 4, up(maxs)).
table(up_v_c_s_mins_maxs, 12, 8, 4, up(maxs)).
table(up_v_minc_maxc_s_maxs, 12, 6, 4, up(maxs)).
table(up_v_minc_maxc_mins_maxs, 12, 6, 4, up(maxs)).
table(up_v_minc_s_mins_maxs, 12, 6, 4, up(maxs)).
table(up_v_maxc_s_mins_maxs, 12, 5, 4, up(maxs)).
table(up_c_minc_maxc_s_maxs, 12, 6, 4, up(maxs)).
table(up_c_minc_maxc_mins_maxs, 12, 6, 4, up(maxs)).
table(up_c_maxc_s_mins_maxs, 12, 5, 4, up(maxs)).
table(up_minc_maxc_s_mins_maxs, 12, 6, 4, up(maxs)).
table(up_v_c_minc_maxc_s_maxs, 12, 8, 5, up(maxs)).
table(up_v_c_minc_maxc_mins_maxs, 12, 8, 5, up(maxs)).
table(up_v_c_minc_s_mins_maxs, 12, 8, 5, up(maxs)).
table(up_v_c_maxc_s_mins_maxs, 12, 6, 5, up(maxs)).
table(up_v_minc_maxc_s_mins_maxs, 12, 7, 5, up(maxs)).
table(up_c_minc_maxc_s_mins_maxs, 12, 7, 5, up(maxs)).
table(up_v_c_minc_maxc_s_mins_maxs, 12, 9, 6, up(maxs)).
table(low_v_mina, 12, 12, 1, low(mina)).
table(low_c_mina, 12, 7, 1, low(mina)).
table(low_minc_mina, 12, 19, 1, low(mina)).
table(low_maxc_mina, 12, 19, 1, low(mina)).
table(low_s_mina, 12, 12, 1, low(mina)).
table(low_mins_mina, 12, 11, 1, low(mina)).
table(low_maxs_mina, 12, 11, 1, low(mina)).
table(low_v_c_mina, 12, 10, 2, low(mina)).
table(low_v_minc_mina, 12, 11, 2, low(mina)).
table(low_v_maxc_mina, 12, 11, 2, low(mina)).
table(low_v_s_mina, 12, 4, 2, low(mina)).
table(low_v_mins_mina, 12, 4, 2, low(mina)).
table(low_v_maxs_mina, 12, 10, 2, low(mina)).
table(low_c_minc_mina, 12, 8, 2, low(mina)).
table(low_c_maxc_mina, 12, 6, 2, low(mina)).
table(low_c_s_mina, 12, 10, 2, low(mina)).
table(low_c_mins_mina, 12, 7, 2, low(mina)).
table(low_c_maxs_mina, 12, 5, 2, low(mina)).
table(low_minc_maxc_mina, 12, 19, 2, low(mina)).
table(low_minc_s_mina, 12, 11, 2, low(mina)).
table(low_minc_mins_mina, 12, 19, 2, low(mina)).
table(low_minc_maxs_mina, 12, 19, 2, low(mina)).
table(low_maxc_s_mina, 12, 11, 2, low(mina)).
table(low_maxc_mins_mina, 12, 19, 2, low(mina)).
table(low_maxc_maxs_mina, 12, 19, 2, low(mina)).
table(low_s_mins_mina, 12, 12, 2, low(mina)).
table(low_s_maxs_mina, 12, 10, 2, low(mina)).
table(low_mins_maxs_mina, 12, 6, 2, low(mina)).
table(low_v_c_minc_mina, 12, 13, 3, low(mina)).
table(low_v_c_maxc_mina, 12, 11, 3, low(mina)).
table(low_v_c_s_mina, 12, 4, 3, low(mina)).
table(low_v_c_mins_mina, 12, 7, 3, low(mina)).
table(low_v_c_maxs_mina, 12, 10, 3, low(mina)).
table(low_v_minc_maxc_mina, 12, 13, 3, low(mina)).
table(low_v_minc_s_mina, 12, 4, 3, low(mina)).
table(low_v_minc_mins_mina, 12, 4, 3, low(mina)).
table(low_v_minc_maxs_mina, 12, 11, 3, low(mina)).
table(low_v_maxc_s_mina, 12, 4, 3, low(mina)).
table(low_v_maxc_mins_mina, 12, 6, 3, low(mina)).
table(low_v_maxc_maxs_mina, 12, 10, 3, low(mina)).
table(low_v_s_mins_mina, 12, 5, 3, low(mina)).
table(low_v_s_maxs_mina, 12, 5, 3, low(mina)).
table(low_v_mins_maxs_mina, 12, 5, 3, low(mina)).
table(low_c_minc_maxc_mina, 12, 8, 3, low(mina)).
table(low_c_minc_s_mina, 12, 11, 3, low(mina)).
table(low_c_minc_mins_mina, 12, 8, 3, low(mina)).
table(low_c_minc_maxs_mina, 12, 6, 3, low(mina)).
table(low_c_maxc_s_mina, 12, 8, 3, low(mina)).
table(low_c_maxc_mins_mina, 12, 6, 3, low(mina)).
table(low_c_maxc_maxs_mina, 12, 6, 3, low(mina)).
table(low_c_s_mins_mina, 12, 10, 3, low(mina)).
table(low_c_s_maxs_mina, 12, 10, 3, low(mina)).
table(low_c_mins_maxs_mina, 12, 5, 3, low(mina)).
table(low_minc_maxc_s_mina, 12, 10, 3, low(mina)).
table(low_minc_maxc_mins_mina, 12, 16, 3, low(mina)).
table(low_minc_maxc_maxs_mina, 12, 18, 3, low(mina)).
table(low_minc_s_mins_mina, 12, 6, 3, low(mina)).
table(low_minc_s_maxs_mina, 12, 8, 3, low(mina)).
table(low_minc_mins_maxs_mina, 12, 17, 3, low(mina)).
table(low_maxc_s_mins_mina, 12, 8, 3, low(mina)).
table(low_maxc_s_maxs_mina, 12, 7, 3, low(mina)).
table(low_maxc_mins_maxs_mina, 12, 8, 3, low(mina)).
table(low_s_mins_maxs_mina, 12, 10, 3, low(mina)).
table(low_v_c_minc_maxc_mina, 12, 14, 4, low(mina)).
table(low_v_c_minc_s_mina, 12, 8, 4, low(mina)).
table(low_v_c_minc_mins_mina, 12, 10, 4, low(mina)).
table(low_v_c_minc_maxs_mina, 12, 13, 4, low(mina)).
table(low_v_c_maxc_s_mina, 12, 5, 4, low(mina)).
table(low_v_c_maxc_mins_mina, 12, 8, 4, low(mina)).
table(low_v_c_maxc_maxs_mina, 12, 11, 4, low(mina)).
table(low_v_c_s_mins_mina, 12, 5, 4, low(mina)).
table(low_v_c_s_maxs_mina, 12, 5, 4, low(mina)).
table(low_v_c_mins_maxs_mina, 12, 8, 4, low(mina)).
table(low_v_minc_maxc_s_mina, 12, 7, 4, low(mina)).
table(low_v_minc_maxc_mins_mina, 12, 7, 4, low(mina)).
table(low_v_minc_maxc_maxs_mina, 12, 12, 4, low(mina)).
table(low_v_minc_s_mins_mina, 12, 5, 4, low(mina)).
table(low_v_minc_s_maxs_mina, 12, 6, 4, low(mina)).
table(low_v_minc_mins_maxs_mina, 12, 6, 4, low(mina)).
table(low_v_maxc_s_mins_mina, 12, 5, 4, low(mina)).
table(low_v_maxc_s_maxs_mina, 12, 5, 4, low(mina)).
table(low_v_maxc_mins_maxs_mina, 12, 5, 4, low(mina)).
table(low_v_s_mins_maxs_mina, 12, 6, 4, low(mina)).
table(low_c_minc_maxc_s_mina, 12, 9, 4, low(mina)).
table(low_c_minc_maxc_mins_mina, 12, 7, 4, low(mina)).
table(low_c_minc_maxc_maxs_mina, 12, 8, 4, low(mina)).
table(low_c_minc_s_mins_mina, 12, 10, 4, low(mina)).
table(low_c_minc_s_maxs_mina, 12, 10, 4, low(mina)).
table(low_c_minc_mins_maxs_mina, 12, 6, 4, low(mina)).
table(low_c_maxc_s_mins_mina, 12, 6, 4, low(mina)).
table(low_c_maxc_s_maxs_mina, 12, 6, 4, low(mina)).
table(low_c_maxc_mins_maxs_mina, 12, 6, 4, low(mina)).
table(low_c_s_mins_maxs_mina, 12, 10, 4, low(mina)).
table(low_minc_maxc_s_mins_mina, 12, 9, 4, low(mina)).
table(low_minc_maxc_s_maxs_mina, 12, 9, 4, low(mina)).
table(low_minc_maxc_mins_maxs_mina, 12, 16, 4, low(mina)).
table(low_minc_s_mins_maxs_mina, 12, 7, 4, low(mina)).
table(low_maxc_s_mins_maxs_mina, 12, 6, 4, low(mina)).
table(low_v_c_minc_maxc_s_mina, 12, 9, 5, low(mina)).
table(low_v_c_minc_maxc_mins_mina, 12, 11, 5, low(mina)).
table(low_v_c_minc_maxc_maxs_mina, 12, 14, 5, low(mina)).
table(low_v_c_minc_s_mins_mina, 12, 8, 5, low(mina)).
table(low_v_c_minc_s_maxs_mina, 12, 8, 5, low(mina)).
table(low_v_c_minc_mins_maxs_mina, 12, 11, 5, low(mina)).
table(low_v_c_maxc_s_mins_mina, 12, 6, 5, low(mina)).
table(low_v_c_maxc_s_maxs_mina, 12, 6, 5, low(mina)).
table(low_v_c_maxc_mins_maxs_mina, 12, 9, 5, low(mina)).
table(low_v_c_s_mins_maxs_mina, 12, 6, 5, low(mina)).
table(low_v_minc_maxc_s_mins_mina, 12, 7, 5, low(mina)).
table(low_v_minc_maxc_s_maxs_mina, 12, 7, 5, low(mina)).
table(low_v_minc_maxc_mins_maxs_mina, 12, 7, 5, low(mina)).
table(low_v_minc_s_mins_maxs_mina, 12, 7, 5, low(mina)).
table(low_v_maxc_s_mins_maxs_mina, 12, 6, 5, low(mina)).
table(low_c_minc_maxc_s_mins_mina, 12, 8, 5, low(mina)).
table(low_c_minc_maxc_s_maxs_mina, 12, 8, 5, low(mina)).
table(low_c_minc_maxc_mins_maxs_mina, 12, 8, 5, low(mina)).
table(low_c_minc_s_mins_maxs_mina, 12, 11, 5, low(mina)).
table(low_c_maxc_s_mins_maxs_mina, 12, 7, 5, low(mina)).
table(low_minc_maxc_s_mins_maxs_mina, 12, 8, 5, low(mina)).
table(low_v_c_minc_maxc_s_mins_mina, 12, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_maxs_mina, 12, 9, 6, low(mina)).
table(low_v_c_minc_maxc_mins_maxs_mina, 12, 12, 6, low(mina)).
table(low_v_c_minc_s_mins_maxs_mina, 12, 9, 6, low(mina)).
table(low_v_c_maxc_s_mins_maxs_mina, 12, 7, 6, low(mina)).
table(low_v_minc_maxc_s_mins_maxs_mina, 12, 8, 6, low(mina)).
table(low_c_minc_maxc_s_mins_maxs_mina, 12, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_mins_maxs_mina, 12, 10, 7, low(mina)).
table(up_v_maxa, 12, 19, 1, up(maxa)).
table(up_v_c_maxa, 12, 19, 2, up(maxa)).
table(up_v_minc_maxa, 12, 19, 2, up(maxa)).
table(up_v_maxc_maxa, 12, 19, 2, up(maxa)).
table(up_v_s_maxa, 12, 19, 2, up(maxa)).
table(up_v_mins_maxa, 12, 19, 2, up(maxa)).
table(up_v_maxs_maxa, 12, 19, 2, up(maxa)).
table(up_c_maxc_maxa, 12, 19, 2, up(maxa)).
table(up_maxc_s_maxa, 12, 18, 2, up(maxa)).
table(up_s_maxs_maxa, 12, 19, 2, up(maxa)).
table(up_v_c_minc_maxa, 12, 19, 3, up(maxa)).
table(up_v_c_maxc_maxa, 12, 19, 3, up(maxa)).
table(up_v_c_s_maxa, 12, 19, 3, up(maxa)).
table(up_v_c_mins_maxa, 12, 19, 3, up(maxa)).
table(up_v_c_maxs_maxa, 12, 7, 3, up(maxa)).
table(up_v_minc_maxc_maxa, 12, 19, 3, up(maxa)).
table(up_v_minc_s_maxa, 12, 18, 3, up(maxa)).
table(up_v_minc_mins_maxa, 12, 19, 3, up(maxa)).
table(up_v_minc_maxs_maxa, 12, 19, 3, up(maxa)).
table(up_v_maxc_s_maxa, 12, 18, 3, up(maxa)).
table(up_v_maxc_mins_maxa, 12, 19, 3, up(maxa)).
table(up_v_maxc_maxs_maxa, 12, 7, 3, up(maxa)).
table(up_v_s_mins_maxa, 12, 19, 3, up(maxa)).
table(up_v_s_maxs_maxa, 12, 19, 3, up(maxa)).
table(up_v_mins_maxs_maxa, 12, 19, 3, up(maxa)).
table(up_c_minc_maxc_maxa, 12, 19, 3, up(maxa)).
table(up_c_maxc_s_maxa, 12, 18, 3, up(maxa)).
table(up_c_maxc_mins_maxa, 12, 19, 3, up(maxa)).
table(up_c_maxc_maxs_maxa, 12, 6, 3, up(maxa)).
table(up_c_s_maxs_maxa, 12, 19, 3, up(maxa)).
table(up_minc_maxc_s_maxa, 12, 8, 3, up(maxa)).
table(up_minc_s_maxs_maxa, 12, 18, 3, up(maxa)).
table(up_maxc_s_mins_maxa, 12, 18, 3, up(maxa)).
table(up_maxc_s_maxs_maxa, 12, 7, 3, up(maxa)).
table(up_s_mins_maxs_maxa, 12, 19, 3, up(maxa)).
table(up_v_c_minc_maxc_maxa, 12, 19, 4, up(maxa)).
table(up_v_c_minc_s_maxa, 12, 18, 4, up(maxa)).
table(up_v_c_minc_mins_maxa, 12, 11, 4, up(maxa)).
table(up_v_c_minc_maxs_maxa, 12, 8, 4, up(maxa)).
table(up_v_c_maxc_s_maxa, 12, 18, 4, up(maxa)).
table(up_v_c_maxc_mins_maxa, 12, 19, 4, up(maxa)).
table(up_v_c_maxc_maxs_maxa, 12, 6, 4, up(maxa)).
table(up_v_c_s_mins_maxa, 12, 19, 4, up(maxa)).
table(up_v_c_s_maxs_maxa, 12, 19, 4, up(maxa)).
table(up_v_c_mins_maxs_maxa, 12, 11, 4, up(maxa)).
table(up_v_minc_maxc_s_maxa, 12, 18, 4, up(maxa)).
table(up_v_minc_maxc_mins_maxa, 12, 19, 4, up(maxa)).
table(up_v_minc_maxc_maxs_maxa, 12, 10, 4, up(maxa)).
table(up_v_minc_s_mins_maxa, 12, 18, 4, up(maxa)).
table(up_v_minc_s_maxs_maxa, 12, 18, 4, up(maxa)).
table(up_v_minc_mins_maxs_maxa, 12, 18, 4, up(maxa)).
table(up_v_maxc_s_mins_maxa, 12, 18, 4, up(maxa)).
table(up_v_maxc_s_maxs_maxa, 12, 18, 4, up(maxa)).
table(up_v_maxc_mins_maxs_maxa, 12, 6, 4, up(maxa)).
table(up_v_s_mins_maxs_maxa, 12, 19, 4, up(maxa)).
table(up_c_minc_maxc_s_maxa, 12, 18, 4, up(maxa)).
table(up_c_minc_maxc_mins_maxa, 12, 19, 4, up(maxa)).
table(up_c_minc_maxc_maxs_maxa, 12, 10, 4, up(maxa)).
table(up_c_minc_s_maxs_maxa, 12, 9, 4, up(maxa)).
table(up_c_maxc_s_mins_maxa, 12, 18, 4, up(maxa)).
table(up_c_maxc_s_maxs_maxa, 12, 18, 4, up(maxa)).
table(up_c_maxc_mins_maxs_maxa, 12, 6, 4, up(maxa)).
table(up_c_s_mins_maxs_maxa, 12, 19, 4, up(maxa)).
table(up_minc_maxc_s_mins_maxa, 12, 18, 4, up(maxa)).
table(up_minc_maxc_s_maxs_maxa, 12, 6, 4, up(maxa)).
table(up_minc_s_mins_maxs_maxa, 12, 18, 4, up(maxa)).
table(up_maxc_s_mins_maxs_maxa, 12, 7, 4, up(maxa)).
table(up_v_c_minc_maxc_s_maxa, 12, 18, 5, up(maxa)).
table(up_v_c_minc_maxc_mins_maxa, 12, 19, 5, up(maxa)).
table(up_v_c_minc_maxc_maxs_maxa, 12, 10, 5, up(maxa)).
table(up_v_c_minc_s_mins_maxa, 12, 10, 5, up(maxa)).
table(up_v_c_minc_s_maxs_maxa, 12, 9, 5, up(maxa)).
table(up_v_c_minc_mins_maxs_maxa, 12, 8, 5, up(maxa)).
table(up_v_c_maxc_s_mins_maxa, 12, 18, 5, up(maxa)).
table(up_v_c_maxc_s_maxs_maxa, 12, 18, 5, up(maxa)).
table(up_v_c_maxc_mins_maxs_maxa, 12, 6, 5, up(maxa)).
table(up_v_c_s_mins_maxs_maxa, 12, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_mins_maxa, 12, 18, 5, up(maxa)).
table(up_v_minc_maxc_s_maxs_maxa, 12, 11, 5, up(maxa)).
table(up_v_minc_maxc_mins_maxs_maxa, 12, 9, 5, up(maxa)).
table(up_v_minc_s_mins_maxs_maxa, 12, 18, 5, up(maxa)).
table(up_v_maxc_s_mins_maxs_maxa, 12, 18, 5, up(maxa)).
table(up_c_minc_maxc_s_mins_maxa, 12, 18, 5, up(maxa)).
table(up_c_minc_maxc_s_maxs_maxa, 12, 11, 5, up(maxa)).
table(up_c_minc_maxc_mins_maxs_maxa, 12, 9, 5, up(maxa)).
table(up_c_minc_s_mins_maxs_maxa, 12, 10, 5, up(maxa)).
table(up_c_maxc_s_mins_maxs_maxa, 12, 18, 5, up(maxa)).
table(up_minc_maxc_s_mins_maxs_maxa, 12, 7, 5, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxa, 12, 18, 6, up(maxa)).
table(up_v_c_minc_maxc_s_maxs_maxa, 12, 11, 6, up(maxa)).
table(up_v_c_minc_maxc_mins_maxs_maxa, 12, 9, 6, up(maxa)).
table(up_v_c_minc_s_mins_maxs_maxa, 12, 10, 6, up(maxa)).
table(up_v_c_maxc_s_mins_maxs_maxa, 12, 18, 6, up(maxa)).
table(up_v_minc_maxc_s_mins_maxs_maxa, 12, 11, 6, up(maxa)).
table(up_c_minc_maxc_s_mins_maxs_maxa, 12, 11, 6, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxs_maxa, 12, 11, 7, up(maxa)).
table(low_c_v, 13, 18, 1, low(v)).
table(low_minc_v, 13, 11, 1, low(v)).
table(low_maxc_v, 13, 11, 1, low(v)).
table(low_s_v, 13, 7, 1, low(v)).
table(low_mins_v, 13, 18, 1, low(v)).
table(low_maxs_v, 13, 18, 1, low(v)).
table(low_c_minc_v, 13, 11, 2, low(v)).
table(low_c_maxc_v, 13, 11, 2, low(v)).
table(low_c_s_v, 13, 8, 2, low(v)).
table(low_c_mins_v, 13, 18, 2, low(v)).
table(low_c_maxs_v, 13, 18, 2, low(v)).
table(low_minc_maxc_v, 13, 11, 2, low(v)).
table(low_minc_s_v, 13, 3, 2, low(v)).
table(low_minc_mins_v, 13, 12, 2, low(v)).
table(low_minc_maxs_v, 13, 12, 2, low(v)).
table(low_maxc_s_v, 13, 3, 2, low(v)).
table(low_maxc_mins_v, 13, 12, 2, low(v)).
table(low_maxc_maxs_v, 13, 13, 2, low(v)).
table(low_s_mins_v, 13, 7, 2, low(v)).
table(low_s_maxs_v, 13, 7, 2, low(v)).
table(low_mins_maxs_v, 13, 12, 2, low(v)).
table(low_c_minc_maxc_v, 13, 11, 3, low(v)).
table(low_c_minc_s_v, 13, 6, 3, low(v)).
table(low_c_minc_mins_v, 13, 12, 3, low(v)).
table(low_c_minc_maxs_v, 13, 12, 3, low(v)).
table(low_c_maxc_s_v, 13, 4, 3, low(v)).
table(low_c_maxc_mins_v, 13, 12, 3, low(v)).
table(low_c_maxc_maxs_v, 13, 13, 3, low(v)).
table(low_c_s_mins_v, 13, 8, 3, low(v)).
table(low_c_s_maxs_v, 13, 8, 3, low(v)).
table(low_c_mins_maxs_v, 13, 18, 3, low(v)).
table(low_minc_maxc_s_v, 13, 5, 3, low(v)).
table(low_minc_maxc_mins_v, 13, 12, 3, low(v)).
table(low_minc_maxc_maxs_v, 13, 12, 3, low(v)).
table(low_minc_s_mins_v, 13, 4, 3, low(v)).
table(low_minc_s_maxs_v, 13, 4, 3, low(v)).
table(low_minc_mins_maxs_v, 13, 13, 3, low(v)).
table(low_maxc_s_mins_v, 13, 4, 3, low(v)).
table(low_maxc_s_maxs_v, 13, 4, 3, low(v)).
table(low_maxc_mins_maxs_v, 13, 14, 3, low(v)).
table(low_s_mins_maxs_v, 13, 7, 3, low(v)).
table(low_c_minc_maxc_s_v, 13, 7, 4, low(v)).
table(low_c_minc_maxc_mins_v, 13, 12, 4, low(v)).
table(low_c_minc_maxc_maxs_v, 13, 12, 4, low(v)).
table(low_c_minc_s_mins_v, 13, 7, 4, low(v)).
table(low_c_minc_s_maxs_v, 13, 7, 4, low(v)).
table(low_c_minc_mins_maxs_v, 13, 13, 4, low(v)).
table(low_c_maxc_s_mins_v, 13, 5, 4, low(v)).
table(low_c_maxc_s_maxs_v, 13, 5, 4, low(v)).
table(low_c_maxc_mins_maxs_v, 13, 14, 4, low(v)).
table(low_c_s_mins_maxs_v, 13, 8, 4, low(v)).
table(low_minc_maxc_s_mins_v, 13, 6, 4, low(v)).
table(low_minc_maxc_s_maxs_v, 13, 6, 4, low(v)).
table(low_minc_maxc_mins_maxs_v, 13, 13, 4, low(v)).
table(low_minc_s_mins_maxs_v, 13, 5, 4, low(v)).
table(low_maxc_s_mins_maxs_v, 13, 5, 4, low(v)).
table(low_c_minc_maxc_s_mins_v, 13, 8, 5, low(v)).
table(low_c_minc_maxc_s_maxs_v, 13, 8, 5, low(v)).
table(low_c_minc_maxc_mins_maxs_v, 13, 13, 5, low(v)).
table(low_c_minc_s_mins_maxs_v, 13, 8, 5, low(v)).
table(low_c_maxc_s_mins_maxs_v, 13, 6, 5, low(v)).
table(low_minc_maxc_s_mins_maxs_v, 13, 7, 5, low(v)).
table(low_c_minc_maxc_s_mins_maxs_v, 13, 9, 6, low(v)).
table(up_c_maxc_v, 13, 3, 2, up(v)).
table(up_maxc_s_v, 13, 3, 2, up(v)).
table(up_s_maxs_v, 13, 3, 2, up(v)).
table(up_c_minc_maxc_v, 13, 6, 3, up(v)).
table(up_c_maxc_s_v, 13, 4, 3, up(v)).
table(up_c_maxc_mins_v, 13, 4, 3, up(v)).
table(up_c_maxc_maxs_v, 13, 4, 3, up(v)).
table(up_c_s_maxs_v, 13, 4, 3, up(v)).
table(up_minc_maxc_s_v, 13, 5, 3, up(v)).
table(up_minc_s_maxs_v, 13, 4, 3, up(v)).
table(up_maxc_s_mins_v, 13, 4, 3, up(v)).
table(up_maxc_s_maxs_v, 13, 4, 3, up(v)).
table(up_s_mins_maxs_v, 13, 4, 3, up(v)).
table(up_c_minc_maxc_s_v, 13, 7, 4, up(v)).
table(up_c_minc_maxc_mins_v, 13, 7, 4, up(v)).
table(up_c_minc_maxc_maxs_v, 13, 7, 4, up(v)).
table(up_c_minc_s_maxs_v, 13, 7, 4, up(v)).
table(up_c_maxc_s_mins_v, 13, 5, 4, up(v)).
table(up_c_maxc_s_maxs_v, 13, 5, 4, up(v)).
table(up_c_maxc_mins_maxs_v, 13, 5, 4, up(v)).
table(up_c_s_mins_maxs_v, 13, 5, 4, up(v)).
table(up_minc_maxc_s_mins_v, 13, 6, 4, up(v)).
table(up_minc_maxc_s_maxs_v, 13, 6, 4, up(v)).
table(up_minc_s_mins_maxs_v, 13, 5, 4, up(v)).
table(up_maxc_s_mins_maxs_v, 13, 5, 4, up(v)).
table(up_c_minc_maxc_s_mins_v, 13, 8, 5, up(v)).
table(up_c_minc_maxc_s_maxs_v, 13, 8, 5, up(v)).
table(up_c_minc_maxc_mins_maxs_v, 13, 8, 5, up(v)).
table(up_c_minc_s_mins_maxs_v, 13, 8, 5, up(v)).
table(up_c_maxc_s_mins_maxs_v, 13, 6, 5, up(v)).
table(up_minc_maxc_s_mins_maxs_v, 13, 7, 5, up(v)).
table(up_c_minc_maxc_s_mins_maxs_v, 13, 9, 6, up(v)).
table(low_v_c, 13, 11, 1, low(c)).
table(low_minc_c, 13, 11, 1, low(c)).
table(low_maxc_c, 13, 11, 1, low(c)).
table(low_s_c, 13, 8, 1, low(c)).
table(low_mins_c, 13, 8, 1, low(c)).
table(low_maxs_c, 13, 8, 1, low(c)).
table(low_v_minc_c, 13, 11, 2, low(c)).
table(low_v_maxc_c, 13, 3, 2, low(c)).
table(low_v_s_c, 13, 13, 2, low(c)).
table(low_v_mins_c, 13, 12, 2, low(c)).
table(low_v_maxs_c, 13, 13, 2, low(c)).
table(low_minc_maxc_c, 13, 11, 2, low(c)).
table(low_minc_s_c, 13, 5, 2, low(c)).
table(low_minc_mins_c, 13, 9, 2, low(c)).
table(low_minc_maxs_c, 13, 9, 2, low(c)).
table(low_maxc_s_c, 13, 3, 2, low(c)).
table(low_maxc_mins_c, 13, 12, 2, low(c)).
table(low_maxc_maxs_c, 13, 13, 2, low(c)).
table(low_s_mins_c, 13, 10, 2, low(c)).
table(low_s_maxs_c, 13, 11, 2, low(c)).
table(low_mins_maxs_c, 13, 9, 2, low(c)).
table(low_v_minc_maxc_c, 13, 6, 3, low(c)).
table(low_v_minc_s_c, 13, 12, 3, low(c)).
table(low_v_minc_mins_c, 13, 12, 3, low(c)).
table(low_v_minc_maxs_c, 13, 12, 3, low(c)).
table(low_v_maxc_s_c, 13, 4, 3, low(c)).
table(low_v_maxc_mins_c, 13, 4, 3, low(c)).
table(low_v_maxc_maxs_c, 13, 4, 3, low(c)).
table(low_v_s_mins_c, 13, 14, 3, low(c)).
table(low_v_s_maxs_c, 13, 14, 3, low(c)).
table(low_v_mins_maxs_c, 13, 14, 3, low(c)).
table(low_minc_maxc_s_c, 13, 5, 3, low(c)).
table(low_minc_maxc_mins_c, 13, 12, 3, low(c)).
table(low_minc_maxc_maxs_c, 13, 12, 3, low(c)).
table(low_minc_s_mins_c, 13, 6, 3, low(c)).
table(low_minc_s_maxs_c, 13, 7, 3, low(c)).
table(low_minc_mins_maxs_c, 13, 7, 3, low(c)).
table(low_maxc_s_mins_c, 13, 4, 3, low(c)).
table(low_maxc_s_maxs_c, 13, 4, 3, low(c)).
table(low_maxc_mins_maxs_c, 13, 6, 3, low(c)).
table(low_s_mins_maxs_c, 13, 12, 3, low(c)).
table(low_v_minc_maxc_s_c, 13, 7, 4, low(c)).
table(low_v_minc_maxc_mins_c, 13, 7, 4, low(c)).
table(low_v_minc_maxc_maxs_c, 13, 7, 4, low(c)).
table(low_v_minc_s_mins_c, 13, 13, 4, low(c)).
table(low_v_minc_s_maxs_c, 13, 13, 4, low(c)).
table(low_v_minc_mins_maxs_c, 13, 13, 4, low(c)).
table(low_v_maxc_s_mins_c, 13, 5, 4, low(c)).
table(low_v_maxc_s_maxs_c, 13, 5, 4, low(c)).
table(low_v_maxc_mins_maxs_c, 13, 5, 4, low(c)).
table(low_v_s_mins_maxs_c, 13, 15, 4, low(c)).
table(low_minc_maxc_s_mins_c, 13, 6, 4, low(c)).
table(low_minc_maxc_s_maxs_c, 13, 6, 4, low(c)).
table(low_minc_maxc_mins_maxs_c, 13, 8, 4, low(c)).
table(low_minc_s_mins_maxs_c, 13, 8, 4, low(c)).
table(low_maxc_s_mins_maxs_c, 13, 5, 4, low(c)).
table(low_v_minc_maxc_s_mins_c, 13, 8, 5, low(c)).
table(low_v_minc_maxc_s_maxs_c, 13, 8, 5, low(c)).
table(low_v_minc_maxc_mins_maxs_c, 13, 8, 5, low(c)).
table(low_v_minc_s_mins_maxs_c, 13, 14, 5, low(c)).
table(low_v_maxc_s_mins_maxs_c, 13, 6, 5, low(c)).
table(low_minc_maxc_s_mins_maxs_c, 13, 7, 5, low(c)).
table(low_v_minc_maxc_s_mins_maxs_c, 13, 9, 6, low(c)).
table(up_v_c, 13, 18, 1, up(c)).
table(up_s_c, 13, 3, 1, up(c)).
table(up_v_minc_c, 13, 8, 2, up(c)).
table(up_v_maxc_c, 13, 11, 2, up(c)).
table(up_v_s_c, 13, 4, 2, up(c)).
table(up_v_mins_c, 13, 12, 2, up(c)).
table(up_v_maxs_c, 13, 18, 2, up(c)).
table(up_minc_s_c, 13, 3, 2, up(c)).
table(up_maxc_s_c, 13, 4, 2, up(c)).
table(up_s_mins_c, 13, 5, 2, up(c)).
table(up_s_maxs_c, 13, 5, 2, up(c)).
table(up_v_minc_maxc_c, 13, 11, 3, up(c)).
table(up_v_minc_s_c, 13, 6, 3, up(c)).
table(up_v_minc_mins_c, 13, 7, 3, up(c)).
table(up_v_minc_maxs_c, 13, 7, 3, up(c)).
table(up_v_maxc_s_c, 13, 5, 3, up(c)).
table(up_v_maxc_mins_c, 13, 12, 3, up(c)).
table(up_v_maxc_maxs_c, 13, 13, 3, up(c)).
table(up_v_s_mins_c, 13, 8, 3, up(c)).
table(up_v_s_maxs_c, 13, 6, 3, up(c)).
table(up_v_mins_maxs_c, 13, 18, 3, up(c)).
table(up_minc_maxc_s_c, 13, 5, 3, up(c)).
table(up_minc_s_mins_c, 13, 4, 3, up(c)).
table(up_minc_s_maxs_c, 13, 4, 3, up(c)).
table(up_maxc_s_mins_c, 13, 5, 3, up(c)).
table(up_maxc_s_maxs_c, 13, 5, 3, up(c)).
table(up_s_mins_maxs_c, 13, 8, 3, up(c)).
table(up_v_minc_maxc_s_c, 13, 7, 4, up(c)).
table(up_v_minc_maxc_mins_c, 13, 12, 4, up(c)).
table(up_v_minc_maxc_maxs_c, 13, 12, 4, up(c)).
table(up_v_minc_s_mins_c, 13, 7, 4, up(c)).
table(up_v_minc_s_maxs_c, 13, 7, 4, up(c)).
table(up_v_minc_mins_maxs_c, 13, 8, 4, up(c)).
table(up_v_maxc_s_mins_c, 13, 9, 4, up(c)).
table(up_v_maxc_s_maxs_c, 13, 6, 4, up(c)).
table(up_v_maxc_mins_maxs_c, 13, 14, 4, up(c)).
table(up_v_s_mins_maxs_c, 13, 10, 4, up(c)).
table(up_minc_maxc_s_mins_c, 13, 6, 4, up(c)).
table(up_minc_maxc_s_maxs_c, 13, 6, 4, up(c)).
table(up_minc_s_mins_maxs_c, 13, 5, 4, up(c)).
table(up_maxc_s_mins_maxs_c, 13, 6, 4, up(c)).
table(up_v_minc_maxc_s_mins_c, 13, 8, 5, up(c)).
table(up_v_minc_maxc_s_maxs_c, 13, 8, 5, up(c)).
table(up_v_minc_maxc_mins_maxs_c, 13, 13, 5, up(c)).
table(up_v_minc_s_mins_maxs_c, 13, 8, 5, up(c)).
table(up_v_maxc_s_mins_maxs_c, 13, 7, 5, up(c)).
table(up_minc_maxc_s_mins_maxs_c, 13, 7, 5, up(c)).
table(up_v_minc_maxc_s_mins_maxs_c, 13, 9, 6, up(c)).
table(low_v_minc, 13, 3, 1, low(minc)).
table(low_c_minc, 13, 3, 1, low(minc)).
table(low_maxc_minc, 13, 3, 1, low(minc)).
table(low_s_minc, 13, 3, 1, low(minc)).
table(low_mins_minc, 13, 2, 1, low(minc)).
table(low_maxs_minc, 13, 3, 1, low(minc)).
table(low_v_c_minc, 13, 5, 2, low(minc)).
table(low_v_maxc_minc, 13, 4, 2, low(minc)).
table(low_v_s_minc, 13, 4, 2, low(minc)).
table(low_v_mins_minc, 13, 3, 2, low(minc)).
table(low_v_maxs_minc, 13, 4, 2, low(minc)).
table(low_c_maxc_minc, 13, 4, 2, low(minc)).
table(low_c_s_minc, 13, 4, 2, low(minc)).
table(low_c_mins_minc, 13, 3, 2, low(minc)).
table(low_c_maxs_minc, 13, 4, 2, low(minc)).
table(low_maxc_s_minc, 13, 4, 2, low(minc)).
table(low_maxc_mins_minc, 13, 4, 2, low(minc)).
table(low_maxc_maxs_minc, 13, 4, 2, low(minc)).
table(low_s_mins_minc, 13, 3, 2, low(minc)).
table(low_s_maxs_minc, 13, 4, 2, low(minc)).
table(low_mins_maxs_minc, 13, 3, 2, low(minc)).
table(low_v_c_maxc_minc, 13, 6, 3, low(minc)).
table(low_v_c_s_minc, 13, 6, 3, low(minc)).
table(low_v_c_mins_minc, 13, 6, 3, low(minc)).
table(low_v_c_maxs_minc, 13, 6, 3, low(minc)).
table(low_v_maxc_s_minc, 13, 5, 3, low(minc)).
table(low_v_maxc_mins_minc, 13, 5, 3, low(minc)).
table(low_v_maxc_maxs_minc, 13, 5, 3, low(minc)).
table(low_v_s_mins_minc, 13, 4, 3, low(minc)).
table(low_v_s_maxs_minc, 13, 5, 3, low(minc)).
table(low_v_mins_maxs_minc, 13, 4, 3, low(minc)).
table(low_c_maxc_s_minc, 13, 5, 3, low(minc)).
table(low_c_maxc_mins_minc, 13, 5, 3, low(minc)).
table(low_c_maxc_maxs_minc, 13, 5, 3, low(minc)).
table(low_c_s_mins_minc, 13, 4, 3, low(minc)).
table(low_c_s_maxs_minc, 13, 6, 3, low(minc)).
table(low_c_mins_maxs_minc, 13, 4, 3, low(minc)).
table(low_maxc_s_mins_minc, 13, 5, 3, low(minc)).
table(low_maxc_s_maxs_minc, 13, 5, 3, low(minc)).
table(low_maxc_mins_maxs_minc, 13, 5, 3, low(minc)).
table(low_s_mins_maxs_minc, 13, 4, 3, low(minc)).
table(low_v_c_maxc_s_minc, 13, 7, 4, low(minc)).
table(low_v_c_maxc_mins_minc, 13, 7, 4, low(minc)).
table(low_v_c_maxc_maxs_minc, 13, 7, 4, low(minc)).
table(low_v_c_s_mins_minc, 13, 7, 4, low(minc)).
table(low_v_c_s_maxs_minc, 13, 7, 4, low(minc)).
table(low_v_c_mins_maxs_minc, 13, 7, 4, low(minc)).
table(low_v_maxc_s_mins_minc, 13, 6, 4, low(minc)).
table(low_v_maxc_s_maxs_minc, 13, 6, 4, low(minc)).
table(low_v_maxc_mins_maxs_minc, 13, 6, 4, low(minc)).
table(low_v_s_mins_maxs_minc, 13, 5, 4, low(minc)).
table(low_c_maxc_s_mins_minc, 13, 6, 4, low(minc)).
table(low_c_maxc_s_maxs_minc, 13, 6, 4, low(minc)).
table(low_c_maxc_mins_maxs_minc, 13, 6, 4, low(minc)).
table(low_c_s_mins_maxs_minc, 13, 6, 4, low(minc)).
table(low_maxc_s_mins_maxs_minc, 13, 6, 4, low(minc)).
table(low_v_c_maxc_s_mins_minc, 13, 8, 5, low(minc)).
table(low_v_c_maxc_s_maxs_minc, 13, 8, 5, low(minc)).
table(low_v_c_maxc_mins_maxs_minc, 13, 8, 5, low(minc)).
table(low_v_c_s_mins_maxs_minc, 13, 8, 5, low(minc)).
table(low_v_maxc_s_mins_maxs_minc, 13, 7, 5, low(minc)).
table(low_c_maxc_s_mins_maxs_minc, 13, 7, 5, low(minc)).
table(low_v_c_maxc_s_mins_maxs_minc, 13, 9, 6, low(minc)).
table(up_v_minc, 13, 11, 1, up(minc)).
table(up_maxc_minc, 13, 6, 1, up(minc)).
table(up_v_c_minc, 13, 5, 2, up(minc)).
table(up_v_maxc_minc, 13, 6, 2, up(minc)).
table(up_v_s_minc, 13, 13, 2, up(minc)).
table(up_v_mins_minc, 13, 12, 2, up(minc)).
table(up_v_maxs_minc, 13, 13, 2, up(minc)).
table(up_c_maxc_minc, 13, 4, 2, up(minc)).
table(up_maxc_s_minc, 13, 6, 2, up(minc)).
table(up_maxc_mins_minc, 13, 4, 2, up(minc)).
table(up_maxc_maxs_minc, 13, 7, 2, up(minc)).
table(up_s_maxs_minc, 13, 14, 2, up(minc)).
table(up_v_c_maxc_minc, 13, 6, 3, up(minc)).
table(up_v_c_s_minc, 13, 6, 3, up(minc)).
table(up_v_c_mins_minc, 13, 6, 3, up(minc)).
table(up_v_c_maxs_minc, 13, 6, 3, up(minc)).
table(up_v_maxc_s_minc, 13, 7, 3, up(minc)).
table(up_v_maxc_mins_minc, 13, 7, 3, up(minc)).
table(up_v_maxc_maxs_minc, 13, 7, 3, up(minc)).
table(up_v_s_mins_minc, 13, 14, 3, up(minc)).
table(up_v_s_maxs_minc, 13, 14, 3, up(minc)).
table(up_v_mins_maxs_minc, 13, 14, 3, up(minc)).
table(up_c_maxc_s_minc, 13, 5, 3, up(minc)).
table(up_c_maxc_mins_minc, 13, 5, 3, up(minc)).
table(up_c_maxc_maxs_minc, 13, 5, 3, up(minc)).
table(up_c_s_maxs_minc, 13, 4, 3, up(minc)).
table(up_maxc_s_mins_minc, 13, 5, 3, up(minc)).
table(up_maxc_s_maxs_minc, 13, 7, 3, up(minc)).
table(up_maxc_mins_maxs_minc, 13, 5, 3, up(minc)).
table(up_s_mins_maxs_minc, 13, 15, 3, up(minc)).
table(up_v_c_maxc_s_minc, 13, 7, 4, up(minc)).
table(up_v_c_maxc_mins_minc, 13, 7, 4, up(minc)).
table(up_v_c_maxc_maxs_minc, 13, 7, 4, up(minc)).
table(up_v_c_s_mins_minc, 13, 7, 4, up(minc)).
table(up_v_c_s_maxs_minc, 13, 7, 4, up(minc)).
table(up_v_c_mins_maxs_minc, 13, 7, 4, up(minc)).
table(up_v_maxc_s_mins_minc, 13, 6, 4, up(minc)).
table(up_v_maxc_s_maxs_minc, 13, 6, 4, up(minc)).
table(up_v_maxc_mins_maxs_minc, 13, 6, 4, up(minc)).
table(up_v_s_mins_maxs_minc, 13, 15, 4, up(minc)).
table(up_c_maxc_s_mins_minc, 13, 6, 4, up(minc)).
table(up_c_maxc_s_maxs_minc, 13, 6, 4, up(minc)).
table(up_c_maxc_mins_maxs_minc, 13, 6, 4, up(minc)).
table(up_c_s_mins_maxs_minc, 13, 5, 4, up(minc)).
table(up_maxc_s_mins_maxs_minc, 13, 6, 4, up(minc)).
table(up_v_c_maxc_s_mins_minc, 13, 8, 5, up(minc)).
table(up_v_c_maxc_s_maxs_minc, 13, 8, 5, up(minc)).
table(up_v_c_maxc_mins_maxs_minc, 13, 8, 5, up(minc)).
table(up_v_c_s_mins_maxs_minc, 13, 8, 5, up(minc)).
table(up_v_maxc_s_mins_maxs_minc, 13, 7, 5, up(minc)).
table(up_c_maxc_s_mins_maxs_minc, 13, 7, 5, up(minc)).
table(up_v_c_maxc_s_mins_maxs_minc, 13, 9, 6, up(minc)).
table(low_v_maxc, 13, 18, 1, low(maxc)).
table(low_c_maxc, 13, 18, 1, low(maxc)).
table(low_minc_maxc, 13, 6, 1, low(maxc)).
table(low_s_maxc, 13, 18, 1, low(maxc)).
table(low_mins_maxc, 13, 9, 1, low(maxc)).
table(low_maxs_maxc, 13, 2, 1, low(maxc)).
table(low_v_c_maxc, 13, 3, 2, low(maxc)).
table(low_v_minc_maxc, 13, 6, 2, low(maxc)).
table(low_v_s_maxc, 13, 4, 2, low(maxc)).
table(low_v_mins_maxc, 13, 9, 2, low(maxc)).
table(low_v_maxs_maxc, 13, 3, 2, low(maxc)).
table(low_c_minc_maxc, 13, 11, 2, low(maxc)).
table(low_c_s_maxc, 13, 4, 2, low(maxc)).
table(low_c_mins_maxc, 13, 18, 2, low(maxc)).
table(low_c_maxs_maxc, 13, 3, 2, low(maxc)).
table(low_minc_s_maxc, 13, 6, 2, low(maxc)).
table(low_minc_mins_maxc, 13, 6, 2, low(maxc)).
table(low_minc_maxs_maxc, 13, 4, 2, low(maxc)).
table(low_s_mins_maxc, 13, 18, 2, low(maxc)).
table(low_s_maxs_maxc, 13, 3, 2, low(maxc)).
table(low_mins_maxs_maxc, 13, 3, 2, low(maxc)).
table(low_v_c_minc_maxc, 13, 6, 3, low(maxc)).
table(low_v_c_s_maxc, 13, 4, 3, low(maxc)).
table(low_v_c_mins_maxc, 13, 4, 3, low(maxc)).
table(low_v_c_maxs_maxc, 13, 4, 3, low(maxc)).
table(low_v_minc_s_maxc, 13, 5, 3, low(maxc)).
table(low_v_minc_mins_maxc, 13, 7, 3, low(maxc)).
table(low_v_minc_maxs_maxc, 13, 5, 3, low(maxc)).
table(low_v_s_mins_maxc, 13, 5, 3, low(maxc)).
table(low_v_s_maxs_maxc, 13, 4, 3, low(maxc)).
table(low_v_mins_maxs_maxc, 13, 4, 3, low(maxc)).
table(low_c_minc_s_maxc, 13, 5, 3, low(maxc)).
table(low_c_minc_mins_maxc, 13, 8, 3, low(maxc)).
table(low_c_minc_maxs_maxc, 13, 5, 3, low(maxc)).
table(low_c_s_mins_maxc, 13, 4, 3, low(maxc)).
table(low_c_s_maxs_maxc, 13, 4, 3, low(maxc)).
table(low_c_mins_maxs_maxc, 13, 4, 3, low(maxc)).
table(low_minc_s_mins_maxc, 13, 7, 3, low(maxc)).
table(low_minc_s_maxs_maxc, 13, 5, 3, low(maxc)).
table(low_minc_mins_maxs_maxc, 13, 5, 3, low(maxc)).
table(low_s_mins_maxs_maxc, 13, 4, 3, low(maxc)).
table(low_v_c_minc_s_maxc, 13, 7, 4, low(maxc)).
table(low_v_c_minc_mins_maxc, 13, 7, 4, low(maxc)).
table(low_v_c_minc_maxs_maxc, 13, 7, 4, low(maxc)).
table(low_v_c_s_mins_maxc, 13, 5, 4, low(maxc)).
table(low_v_c_s_maxs_maxc, 13, 5, 4, low(maxc)).
table(low_v_c_mins_maxs_maxc, 13, 5, 4, low(maxc)).
table(low_v_minc_s_mins_maxc, 13, 6, 4, low(maxc)).
table(low_v_minc_s_maxs_maxc, 13, 6, 4, low(maxc)).
table(low_v_minc_mins_maxs_maxc, 13, 6, 4, low(maxc)).
table(low_v_s_mins_maxs_maxc, 13, 5, 4, low(maxc)).
table(low_c_minc_s_mins_maxc, 13, 6, 4, low(maxc)).
table(low_c_minc_s_maxs_maxc, 13, 6, 4, low(maxc)).
table(low_c_minc_mins_maxs_maxc, 13, 6, 4, low(maxc)).
table(low_c_s_mins_maxs_maxc, 13, 5, 4, low(maxc)).
table(low_minc_s_mins_maxs_maxc, 13, 6, 4, low(maxc)).
table(low_v_c_minc_s_mins_maxc, 13, 8, 5, low(maxc)).
table(low_v_c_minc_s_maxs_maxc, 13, 8, 5, low(maxc)).
table(low_v_c_minc_mins_maxs_maxc, 13, 8, 5, low(maxc)).
table(low_v_c_s_mins_maxs_maxc, 13, 6, 5, low(maxc)).
table(low_v_minc_s_mins_maxs_maxc, 13, 7, 5, low(maxc)).
table(low_c_minc_s_mins_maxs_maxc, 13, 7, 5, low(maxc)).
table(low_v_c_minc_s_mins_maxs_maxc, 13, 9, 6, low(maxc)).
table(up_v_maxc, 13, 11, 1, up(maxc)).
table(up_v_c_maxc, 13, 11, 2, up(maxc)).
table(up_v_minc_maxc, 13, 11, 2, up(maxc)).
table(up_v_s_maxc, 13, 13, 2, up(maxc)).
table(up_v_mins_maxc, 13, 12, 2, up(maxc)).
table(up_v_maxs_maxc, 13, 13, 2, up(maxc)).
table(up_v_c_minc_maxc, 13, 11, 3, up(maxc)).
table(up_v_c_s_maxc, 13, 13, 3, up(maxc)).
table(up_v_c_mins_maxc, 13, 12, 3, up(maxc)).
table(up_v_c_maxs_maxc, 13, 13, 3, up(maxc)).
table(up_v_minc_s_maxc, 13, 12, 3, up(maxc)).
table(up_v_minc_mins_maxc, 13, 12, 3, up(maxc)).
table(up_v_minc_maxs_maxc, 13, 12, 3, up(maxc)).
table(up_v_s_mins_maxc, 13, 14, 3, up(maxc)).
table(up_v_s_maxs_maxc, 13, 14, 3, up(maxc)).
table(up_v_mins_maxs_maxc, 13, 14, 3, up(maxc)).
table(up_v_c_minc_s_maxc, 13, 12, 4, up(maxc)).
table(up_v_c_minc_mins_maxc, 13, 12, 4, up(maxc)).
table(up_v_c_minc_maxs_maxc, 13, 12, 4, up(maxc)).
table(up_v_c_s_mins_maxc, 13, 14, 4, up(maxc)).
table(up_v_c_s_maxs_maxc, 13, 6, 4, up(maxc)).
table(up_v_c_mins_maxs_maxc, 13, 14, 4, up(maxc)).
table(up_v_minc_s_mins_maxc, 13, 13, 4, up(maxc)).
table(up_v_minc_s_maxs_maxc, 13, 13, 4, up(maxc)).
table(up_v_minc_mins_maxs_maxc, 13, 13, 4, up(maxc)).
table(up_v_s_mins_maxs_maxc, 13, 15, 4, up(maxc)).
table(up_v_c_minc_s_mins_maxc, 13, 13, 5, up(maxc)).
table(up_v_c_minc_s_maxs_maxc, 13, 8, 5, up(maxc)).
table(up_v_c_minc_mins_maxs_maxc, 13, 13, 5, up(maxc)).
table(up_v_c_s_mins_maxs_maxc, 13, 10, 5, up(maxc)).
table(up_v_minc_s_mins_maxs_maxc, 13, 14, 5, up(maxc)).
table(up_v_c_minc_s_mins_maxs_maxc, 13, 9, 6, up(maxc)).
table(low_v_s, 13, 18, 1, low(s)).
table(low_c_s, 13, 3, 1, low(s)).
table(low_minc_s, 13, 18, 1, low(s)).
table(low_maxc_s, 13, 18, 1, low(s)).
table(low_mins_s, 13, 18, 1, low(s)).
table(low_maxs_s, 13, 18, 1, low(s)).
table(low_v_c_s, 13, 4, 2, low(s)).
table(low_v_minc_s, 13, 18, 2, low(s)).
table(low_v_maxc_s, 13, 6, 2, low(s)).
table(low_v_mins_s, 13, 9, 2, low(s)).
table(low_v_maxs_s, 13, 3, 2, low(s)).
table(low_c_minc_s, 13, 5, 2, low(s)).
table(low_c_maxc_s, 13, 5, 2, low(s)).
table(low_c_mins_s, 13, 5, 2, low(s)).
table(low_c_maxs_s, 13, 5, 2, low(s)).
table(low_minc_maxc_s, 13, 18, 2, low(s)).
table(low_minc_mins_s, 13, 12, 2, low(s)).
table(low_minc_maxs_s, 13, 14, 2, low(s)).
table(low_maxc_mins_s, 13, 7, 2, low(s)).
table(low_maxc_maxs_s, 13, 14, 2, low(s)).
table(low_mins_maxs_s, 13, 12, 2, low(s)).
table(low_v_c_minc_s, 13, 8, 3, low(s)).
table(low_v_c_maxc_s, 13, 6, 3, low(s)).
table(low_v_c_mins_s, 13, 8, 3, low(s)).
table(low_v_c_maxs_s, 13, 4, 3, low(s)).
table(low_v_minc_maxc_s, 13, 10, 3, low(s)).
table(low_v_minc_mins_s, 13, 13, 3, low(s)).
table(low_v_minc_maxs_s, 13, 4, 3, low(s)).
table(low_v_maxc_mins_s, 13, 4, 3, low(s)).
table(low_v_maxc_maxs_s, 13, 4, 3, low(s)).
table(low_v_mins_maxs_s, 13, 4, 3, low(s)).
table(low_c_minc_maxc_s, 13, 8, 3, low(s)).
table(low_c_minc_mins_s, 13, 4, 3, low(s)).
table(low_c_minc_maxs_s, 13, 5, 3, low(s)).
table(low_c_maxc_mins_s, 13, 8, 3, low(s)).
table(low_c_maxc_maxs_s, 13, 5, 3, low(s)).
table(low_c_mins_maxs_s, 13, 8, 3, low(s)).
table(low_minc_maxc_mins_s, 13, 13, 3, low(s)).
table(low_minc_maxc_maxs_s, 13, 14, 3, low(s)).
table(low_minc_mins_maxs_s, 13, 8, 3, low(s)).
table(low_maxc_mins_maxs_s, 13, 9, 3, low(s)).
table(low_v_c_minc_maxc_s, 13, 10, 4, low(s)).
table(low_v_c_minc_mins_s, 13, 7, 4, low(s)).
table(low_v_c_minc_maxs_s, 13, 7, 4, low(s)).
table(low_v_c_maxc_mins_s, 13, 5, 4, low(s)).
table(low_v_c_maxc_maxs_s, 13, 5, 4, low(s)).
table(low_v_c_mins_maxs_s, 13, 5, 4, low(s)).
table(low_v_minc_maxc_mins_s, 13, 8, 4, low(s)).
table(low_v_minc_maxc_maxs_s, 13, 6, 4, low(s)).
table(low_v_minc_mins_maxs_s, 13, 5, 4, low(s)).
table(low_v_maxc_mins_maxs_s, 13, 5, 4, low(s)).
table(low_c_minc_maxc_mins_s, 13, 6, 4, low(s)).
table(low_c_minc_maxc_maxs_s, 13, 7, 4, low(s)).
table(low_c_minc_mins_maxs_s, 13, 5, 4, low(s)).
table(low_c_maxc_mins_maxs_s, 13, 6, 4, low(s)).
table(low_minc_maxc_mins_maxs_s, 13, 9, 4, low(s)).
table(low_v_c_minc_maxc_mins_s, 13, 8, 5, low(s)).
table(low_v_c_minc_maxc_maxs_s, 13, 8, 5, low(s)).
table(low_v_c_minc_mins_maxs_s, 13, 8, 5, low(s)).
table(low_v_c_maxc_mins_maxs_s, 13, 6, 5, low(s)).
table(low_v_minc_maxc_mins_maxs_s, 13, 7, 5, low(s)).
table(low_c_minc_maxc_mins_maxs_s, 13, 7, 5, low(s)).
table(low_v_c_minc_maxc_mins_maxs_s, 13, 9, 6, low(s)).
table(up_v_s, 13, 7, 1, up(s)).
table(up_v_c_s, 13, 8, 2, up(s)).
table(up_v_minc_s, 13, 8, 2, up(s)).
table(up_v_maxc_s, 13, 8, 2, up(s)).
table(up_v_mins_s, 13, 4, 2, up(s)).
table(up_v_maxs_s, 13, 7, 2, up(s)).
table(up_c_maxc_s, 13, 9, 2, up(s)).
table(up_v_c_minc_s, 13, 11, 3, up(s)).
table(up_v_c_maxc_s, 13, 9, 3, up(s)).
table(up_v_c_mins_s, 13, 5, 3, up(s)).
table(up_v_c_maxs_s, 13, 8, 3, up(s)).
table(up_v_minc_maxc_s, 13, 10, 3, up(s)).
table(up_v_minc_mins_s, 13, 5, 3, up(s)).
table(up_v_minc_maxs_s, 13, 8, 3, up(s)).
table(up_v_maxc_mins_s, 13, 5, 3, up(s)).
table(up_v_maxc_maxs_s, 13, 8, 3, up(s)).
table(up_v_mins_maxs_s, 13, 7, 3, up(s)).
table(up_c_minc_maxc_s, 13, 12, 3, up(s)).
table(up_c_maxc_mins_s, 13, 5, 3, up(s)).
table(up_c_maxc_maxs_s, 13, 9, 3, up(s)).
table(up_v_c_minc_maxc_s, 13, 12, 4, up(s)).
table(up_v_c_minc_mins_s, 13, 8, 4, up(s)).
table(up_v_c_minc_maxs_s, 13, 11, 4, up(s)).
table(up_v_c_maxc_mins_s, 13, 6, 4, up(s)).
table(up_v_c_maxc_maxs_s, 13, 9, 4, up(s)).
table(up_v_c_mins_maxs_s, 13, 8, 4, up(s)).
table(up_v_minc_maxc_mins_s, 13, 7, 4, up(s)).
table(up_v_minc_maxc_maxs_s, 13, 10, 4, up(s)).
table(up_v_minc_mins_maxs_s, 13, 8, 4, up(s)).
table(up_v_maxc_mins_maxs_s, 13, 8, 4, up(s)).
table(up_c_minc_maxc_mins_s, 13, 7, 4, up(s)).
table(up_c_minc_maxc_maxs_s, 13, 12, 4, up(s)).
table(up_c_maxc_mins_maxs_s, 13, 6, 4, up(s)).
table(up_v_c_minc_maxc_mins_s, 13, 9, 5, up(s)).
table(up_v_c_minc_maxc_maxs_s, 13, 12, 5, up(s)).
table(up_v_c_minc_mins_maxs_s, 13, 11, 5, up(s)).
table(up_v_c_maxc_mins_maxs_s, 13, 9, 5, up(s)).
table(up_v_minc_maxc_mins_maxs_s, 13, 10, 5, up(s)).
table(up_c_minc_maxc_mins_maxs_s, 13, 8, 5, up(s)).
table(up_v_c_minc_maxc_mins_maxs_s, 13, 12, 6, up(s)).
table(low_v_mins, 13, 2, 1, low(mins)).
table(low_c_mins, 13, 2, 1, low(mins)).
table(low_minc_mins, 13, 2, 1, low(mins)).
table(low_maxc_mins, 13, 2, 1, low(mins)).
table(low_s_mins, 13, 2, 1, low(mins)).
table(low_maxs_mins, 13, 2, 1, low(mins)).
table(low_v_c_mins, 13, 3, 2, low(mins)).
table(low_v_minc_mins, 13, 3, 2, low(mins)).
table(low_v_maxc_mins, 13, 3, 2, low(mins)).
table(low_v_s_mins, 13, 3, 2, low(mins)).
table(low_v_maxs_mins, 13, 3, 2, low(mins)).
table(low_c_minc_mins, 13, 3, 2, low(mins)).
table(low_c_maxc_mins, 13, 3, 2, low(mins)).
table(low_c_s_mins, 13, 3, 2, low(mins)).
table(low_c_maxs_mins, 13, 3, 2, low(mins)).
table(low_minc_maxc_mins, 13, 4, 2, low(mins)).
table(low_minc_s_mins, 13, 3, 2, low(mins)).
table(low_minc_maxs_mins, 13, 3, 2, low(mins)).
table(low_maxc_s_mins, 13, 3, 2, low(mins)).
table(low_maxc_maxs_mins, 13, 3, 2, low(mins)).
table(low_s_maxs_mins, 13, 3, 2, low(mins)).
table(low_v_c_minc_mins, 13, 6, 3, low(mins)).
table(low_v_c_maxc_mins, 13, 4, 3, low(mins)).
table(low_v_c_s_mins, 13, 4, 3, low(mins)).
table(low_v_c_maxs_mins, 13, 4, 3, low(mins)).
table(low_v_minc_maxc_mins, 13, 5, 3, low(mins)).
table(low_v_minc_s_mins, 13, 4, 3, low(mins)).
table(low_v_minc_maxs_mins, 13, 4, 3, low(mins)).
table(low_v_maxc_s_mins, 13, 4, 3, low(mins)).
table(low_v_maxc_maxs_mins, 13, 4, 3, low(mins)).
table(low_v_s_maxs_mins, 13, 4, 3, low(mins)).
table(low_c_minc_maxc_mins, 13, 5, 3, low(mins)).
table(low_c_minc_s_mins, 13, 4, 3, low(mins)).
table(low_c_minc_maxs_mins, 13, 4, 3, low(mins)).
table(low_c_maxc_s_mins, 13, 4, 3, low(mins)).
table(low_c_maxc_maxs_mins, 13, 4, 3, low(mins)).
table(low_c_s_maxs_mins, 13, 4, 3, low(mins)).
table(low_minc_maxc_s_mins, 13, 5, 3, low(mins)).
table(low_minc_maxc_maxs_mins, 13, 5, 3, low(mins)).
table(low_minc_s_maxs_mins, 13, 4, 3, low(mins)).
table(low_maxc_s_maxs_mins, 13, 4, 3, low(mins)).
table(low_v_c_minc_maxc_mins, 13, 7, 4, low(mins)).
table(low_v_c_minc_s_mins, 13, 7, 4, low(mins)).
table(low_v_c_minc_maxs_mins, 13, 7, 4, low(mins)).
table(low_v_c_maxc_s_mins, 13, 5, 4, low(mins)).
table(low_v_c_maxc_maxs_mins, 13, 5, 4, low(mins)).
table(low_v_c_s_maxs_mins, 13, 5, 4, low(mins)).
table(low_v_minc_maxc_s_mins, 13, 6, 4, low(mins)).
table(low_v_minc_maxc_maxs_mins, 13, 6, 4, low(mins)).
table(low_v_minc_s_maxs_mins, 13, 5, 4, low(mins)).
table(low_v_maxc_s_maxs_mins, 13, 5, 4, low(mins)).
table(low_c_minc_maxc_s_mins, 13, 6, 4, low(mins)).
table(low_c_minc_maxc_maxs_mins, 13, 6, 4, low(mins)).
table(low_c_minc_s_maxs_mins, 13, 5, 4, low(mins)).
table(low_c_maxc_s_maxs_mins, 13, 5, 4, low(mins)).
table(low_minc_maxc_s_maxs_mins, 13, 6, 4, low(mins)).
table(low_v_c_minc_maxc_s_mins, 13, 8, 5, low(mins)).
table(low_v_c_minc_maxc_maxs_mins, 13, 8, 5, low(mins)).
table(low_v_c_minc_s_maxs_mins, 13, 8, 5, low(mins)).
table(low_v_c_maxc_s_maxs_mins, 13, 6, 5, low(mins)).
table(low_v_minc_maxc_s_maxs_mins, 13, 7, 5, low(mins)).
table(low_c_minc_maxc_s_maxs_mins, 13, 7, 5, low(mins)).
table(low_v_c_minc_maxc_s_maxs_mins, 13, 9, 6, low(mins)).
table(up_v_mins, 13, 18, 1, up(mins)).
table(up_maxc_mins, 13, 9, 1, up(mins)).
table(up_maxs_mins, 13, 2, 1, up(mins)).
table(up_v_c_mins, 13, 6, 2, up(mins)).
table(up_v_minc_mins, 13, 3, 2, up(mins)).
table(up_v_maxc_mins, 13, 7, 2, up(mins)).
table(up_v_s_mins, 13, 3, 2, up(mins)).
table(up_v_maxs_mins, 13, 3, 2, up(mins)).
table(up_c_maxc_mins, 13, 5, 2, up(mins)).
table(up_c_maxs_mins, 13, 3, 2, up(mins)).
table(up_minc_maxc_mins, 13, 4, 2, up(mins)).
table(up_minc_maxs_mins, 13, 3, 2, up(mins)).
table(up_maxc_s_mins, 13, 3, 2, up(mins)).
table(up_maxc_maxs_mins, 13, 3, 2, up(mins)).
table(up_s_maxs_mins, 13, 3, 2, up(mins)).
table(up_v_c_minc_mins, 13, 6, 3, up(mins)).
table(up_v_c_maxc_mins, 13, 7, 3, up(mins)).
table(up_v_c_s_mins, 13, 4, 3, up(mins)).
table(up_v_c_maxs_mins, 13, 4, 3, up(mins)).
table(up_v_minc_maxc_mins, 13, 5, 3, up(mins)).
table(up_v_minc_s_mins, 13, 4, 3, up(mins)).
table(up_v_minc_maxs_mins, 13, 4, 3, up(mins)).
table(up_v_maxc_s_mins, 13, 4, 3, up(mins)).
table(up_v_maxc_maxs_mins, 13, 4, 3, up(mins)).
table(up_v_s_maxs_mins, 13, 4, 3, up(mins)).
table(up_c_minc_maxc_mins, 13, 5, 3, up(mins)).
table(up_c_minc_maxs_mins, 13, 4, 3, up(mins)).
table(up_c_maxc_s_mins, 13, 4, 3, up(mins)).
table(up_c_maxc_maxs_mins, 13, 4, 3, up(mins)).
table(up_c_s_maxs_mins, 13, 4, 3, up(mins)).
table(up_minc_maxc_s_mins, 13, 5, 3, up(mins)).
table(up_minc_maxc_maxs_mins, 13, 5, 3, up(mins)).
table(up_minc_s_maxs_mins, 13, 4, 3, up(mins)).
table(up_maxc_s_maxs_mins, 13, 4, 3, up(mins)).
table(up_v_c_minc_maxc_mins, 13, 7, 4, up(mins)).
table(up_v_c_minc_s_mins, 13, 7, 4, up(mins)).
table(up_v_c_minc_maxs_mins, 13, 7, 4, up(mins)).
table(up_v_c_maxc_s_mins, 13, 5, 4, up(mins)).
table(up_v_c_maxc_maxs_mins, 13, 5, 4, up(mins)).
table(up_v_c_s_maxs_mins, 13, 5, 4, up(mins)).
table(up_v_minc_maxc_s_mins, 13, 6, 4, up(mins)).
table(up_v_minc_maxc_maxs_mins, 13, 6, 4, up(mins)).
table(up_v_minc_s_maxs_mins, 13, 5, 4, up(mins)).
table(up_v_maxc_s_maxs_mins, 13, 5, 4, up(mins)).
table(up_c_minc_maxc_s_mins, 13, 6, 4, up(mins)).
table(up_c_minc_maxc_maxs_mins, 13, 6, 4, up(mins)).
table(up_c_minc_s_maxs_mins, 13, 5, 4, up(mins)).
table(up_c_maxc_s_maxs_mins, 13, 5, 4, up(mins)).
table(up_minc_maxc_s_maxs_mins, 13, 6, 4, up(mins)).
table(up_v_c_minc_maxc_s_mins, 13, 8, 5, up(mins)).
table(up_v_c_minc_maxc_maxs_mins, 13, 8, 5, up(mins)).
table(up_v_c_minc_s_maxs_mins, 13, 8, 5, up(mins)).
table(up_v_c_maxc_s_maxs_mins, 13, 6, 5, up(mins)).
table(up_v_minc_maxc_s_maxs_mins, 13, 7, 5, up(mins)).
table(up_c_minc_maxc_s_maxs_mins, 13, 7, 5, up(mins)).
table(up_v_c_minc_maxc_s_maxs_mins, 13, 9, 6, up(mins)).
table(low_v_maxs, 13, 7, 1, low(maxs)).
table(low_c_maxs, 13, 4, 1, low(maxs)).
table(low_minc_maxs, 13, 4, 1, low(maxs)).
table(low_maxc_maxs, 13, 4, 1, low(maxs)).
table(low_s_maxs, 13, 7, 1, low(maxs)).
table(low_mins_maxs, 13, 2, 1, low(maxs)).
table(low_v_c_maxs, 13, 8, 2, low(maxs)).
table(low_v_minc_maxs, 13, 8, 2, low(maxs)).
table(low_v_maxc_maxs, 13, 8, 2, low(maxs)).
table(low_v_s_maxs, 13, 3, 2, low(maxs)).
table(low_v_mins_maxs, 13, 3, 2, low(maxs)).
table(low_c_minc_maxs, 13, 5, 2, low(maxs)).
table(low_c_maxc_maxs, 13, 5, 2, low(maxs)).
table(low_c_s_maxs, 13, 8, 2, low(maxs)).
table(low_c_mins_maxs, 13, 3, 2, low(maxs)).
table(low_minc_maxc_maxs, 13, 6, 2, low(maxs)).
table(low_minc_s_maxs, 13, 3, 2, low(maxs)).
table(low_minc_mins_maxs, 13, 3, 2, low(maxs)).
table(low_maxc_s_maxs, 13, 3, 2, low(maxs)).
table(low_maxc_mins_maxs, 13, 3, 2, low(maxs)).
table(low_s_mins_maxs, 13, 7, 2, low(maxs)).
table(low_v_c_minc_maxs, 13, 11, 3, low(maxs)).
table(low_v_c_maxc_maxs, 13, 9, 3, low(maxs)).
table(low_v_c_s_maxs, 13, 4, 3, low(maxs)).
table(low_v_c_mins_maxs, 13, 4, 3, low(maxs)).
table(low_v_minc_maxc_maxs, 13, 10, 3, low(maxs)).
table(low_v_minc_s_maxs, 13, 4, 3, low(maxs)).
table(low_v_minc_mins_maxs, 13, 4, 3, low(maxs)).
table(low_v_maxc_s_maxs, 13, 4, 3, low(maxs)).
table(low_v_maxc_mins_maxs, 13, 4, 3, low(maxs)).
table(low_v_s_mins_maxs, 13, 4, 3, low(maxs)).
table(low_c_minc_maxc_maxs, 13, 7, 3, low(maxs)).
table(low_c_minc_s_maxs, 13, 4, 3, low(maxs)).
table(low_c_minc_mins_maxs, 13, 4, 3, low(maxs)).
table(low_c_maxc_s_maxs, 13, 4, 3, low(maxs)).
table(low_c_maxc_mins_maxs, 13, 4, 3, low(maxs)).
table(low_c_s_mins_maxs, 13, 8, 3, low(maxs)).
table(low_minc_maxc_s_maxs, 13, 5, 3, low(maxs)).
table(low_minc_maxc_mins_maxs, 13, 5, 3, low(maxs)).
table(low_minc_s_mins_maxs, 13, 4, 3, low(maxs)).
table(low_maxc_s_mins_maxs, 13, 4, 3, low(maxs)).
table(low_v_c_minc_maxc_maxs, 13, 12, 4, low(maxs)).
table(low_v_c_minc_s_maxs, 13, 7, 4, low(maxs)).
table(low_v_c_minc_mins_maxs, 13, 7, 4, low(maxs)).
table(low_v_c_maxc_s_maxs, 13, 5, 4, low(maxs)).
table(low_v_c_maxc_mins_maxs, 13, 5, 4, low(maxs)).
table(low_v_c_s_mins_maxs, 13, 5, 4, low(maxs)).
table(low_v_minc_maxc_s_maxs, 13, 6, 4, low(maxs)).
table(low_v_minc_maxc_mins_maxs, 13, 6, 4, low(maxs)).
table(low_v_minc_s_mins_maxs, 13, 5, 4, low(maxs)).
table(low_v_maxc_s_mins_maxs, 13, 5, 4, low(maxs)).
table(low_c_minc_maxc_s_maxs, 13, 6, 4, low(maxs)).
table(low_c_minc_maxc_mins_maxs, 13, 6, 4, low(maxs)).
table(low_c_minc_s_mins_maxs, 13, 5, 4, low(maxs)).
table(low_c_maxc_s_mins_maxs, 13, 5, 4, low(maxs)).
table(low_minc_maxc_s_mins_maxs, 13, 6, 4, low(maxs)).
table(low_v_c_minc_maxc_s_maxs, 13, 8, 5, low(maxs)).
table(low_v_c_minc_maxc_mins_maxs, 13, 8, 5, low(maxs)).
table(low_v_c_minc_s_mins_maxs, 13, 8, 5, low(maxs)).
table(low_v_c_maxc_s_mins_maxs, 13, 6, 5, low(maxs)).
table(low_v_minc_maxc_s_mins_maxs, 13, 7, 5, low(maxs)).
table(low_c_minc_maxc_s_mins_maxs, 13, 7, 5, low(maxs)).
table(low_v_c_minc_maxc_s_mins_maxs, 13, 9, 6, low(maxs)).
table(up_v_maxs, 13, 18, 1, up(maxs)).
table(up_maxc_maxs, 13, 2, 1, up(maxs)).
table(up_v_c_maxs, 13, 18, 2, up(maxs)).
table(up_v_minc_maxs, 13, 12, 2, up(maxs)).
table(up_v_maxc_maxs, 13, 3, 2, up(maxs)).
table(up_v_s_maxs, 13, 7, 2, up(maxs)).
table(up_v_mins_maxs, 13, 9, 2, up(maxs)).
table(up_c_maxc_maxs, 13, 3, 2, up(maxs)).
table(up_minc_maxc_maxs, 13, 4, 2, up(maxs)).
table(up_maxc_s_maxs, 13, 3, 2, up(maxs)).
table(up_maxc_mins_maxs, 13, 3, 2, up(maxs)).
table(up_v_c_minc_maxs, 13, 12, 3, up(maxs)).
table(up_v_c_maxc_maxs, 13, 4, 3, up(maxs)).
table(up_v_c_s_maxs, 13, 8, 3, up(maxs)).
table(up_v_c_mins_maxs, 13, 18, 3, up(maxs)).
table(up_v_minc_maxc_maxs, 13, 5, 3, up(maxs)).
table(up_v_minc_s_maxs, 13, 5, 3, up(maxs)).
table(up_v_minc_mins_maxs, 13, 14, 3, up(maxs)).
table(up_v_maxc_s_maxs, 13, 4, 3, up(maxs)).
table(up_v_maxc_mins_maxs, 13, 4, 3, up(maxs)).
table(up_v_s_mins_maxs, 13, 7, 3, up(maxs)).
table(up_c_minc_maxc_maxs, 13, 5, 3, up(maxs)).
table(up_c_maxc_s_maxs, 13, 4, 3, up(maxs)).
table(up_c_maxc_mins_maxs, 13, 4, 3, up(maxs)).
table(up_minc_maxc_s_maxs, 13, 5, 3, up(maxs)).
table(up_minc_maxc_mins_maxs, 13, 5, 3, up(maxs)).
table(up_maxc_s_mins_maxs, 13, 4, 3, up(maxs)).
table(up_v_c_minc_maxc_maxs, 13, 7, 4, up(maxs)).
table(up_v_c_minc_s_maxs, 13, 7, 4, up(maxs)).
table(up_v_c_minc_mins_maxs, 13, 13, 4, up(maxs)).
table(up_v_c_maxc_s_maxs, 13, 5, 4, up(maxs)).
table(up_v_c_maxc_mins_maxs, 13, 5, 4, up(maxs)).
table(up_v_c_s_mins_maxs, 13, 8, 4, up(maxs)).
table(up_v_minc_maxc_s_maxs, 13, 6, 4, up(maxs)).
table(up_v_minc_maxc_mins_maxs, 13, 6, 4, up(maxs)).
table(up_v_minc_s_mins_maxs, 13, 6, 4, up(maxs)).
table(up_v_maxc_s_mins_maxs, 13, 5, 4, up(maxs)).
table(up_c_minc_maxc_s_maxs, 13, 6, 4, up(maxs)).
table(up_c_minc_maxc_mins_maxs, 13, 6, 4, up(maxs)).
table(up_c_maxc_s_mins_maxs, 13, 5, 4, up(maxs)).
table(up_minc_maxc_s_mins_maxs, 13, 6, 4, up(maxs)).
table(up_v_c_minc_maxc_s_maxs, 13, 8, 5, up(maxs)).
table(up_v_c_minc_maxc_mins_maxs, 13, 8, 5, up(maxs)).
table(up_v_c_minc_s_mins_maxs, 13, 8, 5, up(maxs)).
table(up_v_c_maxc_s_mins_maxs, 13, 6, 5, up(maxs)).
table(up_v_minc_maxc_s_mins_maxs, 13, 7, 5, up(maxs)).
table(up_c_minc_maxc_s_mins_maxs, 13, 7, 5, up(maxs)).
table(up_v_c_minc_maxc_s_mins_maxs, 13, 9, 6, up(maxs)).
table(low_v_mina, 13, 12, 1, low(mina)).
table(low_c_mina, 13, 7, 1, low(mina)).
table(low_minc_mina, 13, 19, 1, low(mina)).
table(low_maxc_mina, 13, 19, 1, low(mina)).
table(low_s_mina, 13, 12, 1, low(mina)).
table(low_mins_mina, 13, 11, 1, low(mina)).
table(low_maxs_mina, 13, 11, 1, low(mina)).
table(low_v_c_mina, 13, 10, 2, low(mina)).
table(low_v_minc_mina, 13, 11, 2, low(mina)).
table(low_v_maxc_mina, 13, 11, 2, low(mina)).
table(low_v_s_mina, 13, 4, 2, low(mina)).
table(low_v_mins_mina, 13, 4, 2, low(mina)).
table(low_v_maxs_mina, 13, 10, 2, low(mina)).
table(low_c_minc_mina, 13, 8, 2, low(mina)).
table(low_c_maxc_mina, 13, 6, 2, low(mina)).
table(low_c_s_mina, 13, 10, 2, low(mina)).
table(low_c_mins_mina, 13, 7, 2, low(mina)).
table(low_c_maxs_mina, 13, 5, 2, low(mina)).
table(low_minc_maxc_mina, 13, 19, 2, low(mina)).
table(low_minc_s_mina, 13, 11, 2, low(mina)).
table(low_minc_mins_mina, 13, 19, 2, low(mina)).
table(low_minc_maxs_mina, 13, 19, 2, low(mina)).
table(low_maxc_s_mina, 13, 11, 2, low(mina)).
table(low_maxc_mins_mina, 13, 19, 2, low(mina)).
table(low_maxc_maxs_mina, 13, 19, 2, low(mina)).
table(low_s_mins_mina, 13, 12, 2, low(mina)).
table(low_s_maxs_mina, 13, 10, 2, low(mina)).
table(low_mins_maxs_mina, 13, 6, 2, low(mina)).
table(low_v_c_minc_mina, 13, 13, 3, low(mina)).
table(low_v_c_maxc_mina, 13, 11, 3, low(mina)).
table(low_v_c_s_mina, 13, 4, 3, low(mina)).
table(low_v_c_mins_mina, 13, 7, 3, low(mina)).
table(low_v_c_maxs_mina, 13, 10, 3, low(mina)).
table(low_v_minc_maxc_mina, 13, 13, 3, low(mina)).
table(low_v_minc_s_mina, 13, 4, 3, low(mina)).
table(low_v_minc_mins_mina, 13, 4, 3, low(mina)).
table(low_v_minc_maxs_mina, 13, 11, 3, low(mina)).
table(low_v_maxc_s_mina, 13, 4, 3, low(mina)).
table(low_v_maxc_mins_mina, 13, 6, 3, low(mina)).
table(low_v_maxc_maxs_mina, 13, 10, 3, low(mina)).
table(low_v_s_mins_mina, 13, 5, 3, low(mina)).
table(low_v_s_maxs_mina, 13, 5, 3, low(mina)).
table(low_v_mins_maxs_mina, 13, 5, 3, low(mina)).
table(low_c_minc_maxc_mina, 13, 8, 3, low(mina)).
table(low_c_minc_s_mina, 13, 11, 3, low(mina)).
table(low_c_minc_mins_mina, 13, 8, 3, low(mina)).
table(low_c_minc_maxs_mina, 13, 6, 3, low(mina)).
table(low_c_maxc_s_mina, 13, 8, 3, low(mina)).
table(low_c_maxc_mins_mina, 13, 6, 3, low(mina)).
table(low_c_maxc_maxs_mina, 13, 6, 3, low(mina)).
table(low_c_s_mins_mina, 13, 10, 3, low(mina)).
table(low_c_s_maxs_mina, 13, 10, 3, low(mina)).
table(low_c_mins_maxs_mina, 13, 5, 3, low(mina)).
table(low_minc_maxc_s_mina, 13, 10, 3, low(mina)).
table(low_minc_maxc_mins_mina, 13, 16, 3, low(mina)).
table(low_minc_maxc_maxs_mina, 13, 18, 3, low(mina)).
table(low_minc_s_mins_mina, 13, 6, 3, low(mina)).
table(low_minc_s_maxs_mina, 13, 8, 3, low(mina)).
table(low_minc_mins_maxs_mina, 13, 17, 3, low(mina)).
table(low_maxc_s_mins_mina, 13, 8, 3, low(mina)).
table(low_maxc_s_maxs_mina, 13, 7, 3, low(mina)).
table(low_maxc_mins_maxs_mina, 13, 8, 3, low(mina)).
table(low_s_mins_maxs_mina, 13, 10, 3, low(mina)).
table(low_v_c_minc_maxc_mina, 13, 14, 4, low(mina)).
table(low_v_c_minc_s_mina, 13, 8, 4, low(mina)).
table(low_v_c_minc_mins_mina, 13, 10, 4, low(mina)).
table(low_v_c_minc_maxs_mina, 13, 13, 4, low(mina)).
table(low_v_c_maxc_s_mina, 13, 5, 4, low(mina)).
table(low_v_c_maxc_mins_mina, 13, 8, 4, low(mina)).
table(low_v_c_maxc_maxs_mina, 13, 11, 4, low(mina)).
table(low_v_c_s_mins_mina, 13, 5, 4, low(mina)).
table(low_v_c_s_maxs_mina, 13, 5, 4, low(mina)).
table(low_v_c_mins_maxs_mina, 13, 8, 4, low(mina)).
table(low_v_minc_maxc_s_mina, 13, 7, 4, low(mina)).
table(low_v_minc_maxc_mins_mina, 13, 7, 4, low(mina)).
table(low_v_minc_maxc_maxs_mina, 13, 12, 4, low(mina)).
table(low_v_minc_s_mins_mina, 13, 5, 4, low(mina)).
table(low_v_minc_s_maxs_mina, 13, 6, 4, low(mina)).
table(low_v_minc_mins_maxs_mina, 13, 6, 4, low(mina)).
table(low_v_maxc_s_mins_mina, 13, 5, 4, low(mina)).
table(low_v_maxc_s_maxs_mina, 13, 5, 4, low(mina)).
table(low_v_maxc_mins_maxs_mina, 13, 5, 4, low(mina)).
table(low_v_s_mins_maxs_mina, 13, 6, 4, low(mina)).
table(low_c_minc_maxc_s_mina, 13, 9, 4, low(mina)).
table(low_c_minc_maxc_mins_mina, 13, 7, 4, low(mina)).
table(low_c_minc_maxc_maxs_mina, 13, 8, 4, low(mina)).
table(low_c_minc_s_mins_mina, 13, 10, 4, low(mina)).
table(low_c_minc_s_maxs_mina, 13, 10, 4, low(mina)).
table(low_c_minc_mins_maxs_mina, 13, 6, 4, low(mina)).
table(low_c_maxc_s_mins_mina, 13, 6, 4, low(mina)).
table(low_c_maxc_s_maxs_mina, 13, 6, 4, low(mina)).
table(low_c_maxc_mins_maxs_mina, 13, 6, 4, low(mina)).
table(low_c_s_mins_maxs_mina, 13, 10, 4, low(mina)).
table(low_minc_maxc_s_mins_mina, 13, 9, 4, low(mina)).
table(low_minc_maxc_s_maxs_mina, 13, 9, 4, low(mina)).
table(low_minc_maxc_mins_maxs_mina, 13, 16, 4, low(mina)).
table(low_minc_s_mins_maxs_mina, 13, 7, 4, low(mina)).
table(low_maxc_s_mins_maxs_mina, 13, 6, 4, low(mina)).
table(low_v_c_minc_maxc_s_mina, 13, 9, 5, low(mina)).
table(low_v_c_minc_maxc_mins_mina, 13, 11, 5, low(mina)).
table(low_v_c_minc_maxc_maxs_mina, 13, 14, 5, low(mina)).
table(low_v_c_minc_s_mins_mina, 13, 8, 5, low(mina)).
table(low_v_c_minc_s_maxs_mina, 13, 8, 5, low(mina)).
table(low_v_c_minc_mins_maxs_mina, 13, 11, 5, low(mina)).
table(low_v_c_maxc_s_mins_mina, 13, 6, 5, low(mina)).
table(low_v_c_maxc_s_maxs_mina, 13, 6, 5, low(mina)).
table(low_v_c_maxc_mins_maxs_mina, 13, 9, 5, low(mina)).
table(low_v_c_s_mins_maxs_mina, 13, 6, 5, low(mina)).
table(low_v_minc_maxc_s_mins_mina, 13, 7, 5, low(mina)).
table(low_v_minc_maxc_s_maxs_mina, 13, 7, 5, low(mina)).
table(low_v_minc_maxc_mins_maxs_mina, 13, 7, 5, low(mina)).
table(low_v_minc_s_mins_maxs_mina, 13, 7, 5, low(mina)).
table(low_v_maxc_s_mins_maxs_mina, 13, 6, 5, low(mina)).
table(low_c_minc_maxc_s_mins_mina, 13, 8, 5, low(mina)).
table(low_c_minc_maxc_s_maxs_mina, 13, 8, 5, low(mina)).
table(low_c_minc_maxc_mins_maxs_mina, 13, 8, 5, low(mina)).
table(low_c_minc_s_mins_maxs_mina, 13, 11, 5, low(mina)).
table(low_c_maxc_s_mins_maxs_mina, 13, 7, 5, low(mina)).
table(low_minc_maxc_s_mins_maxs_mina, 13, 8, 5, low(mina)).
table(low_v_c_minc_maxc_s_mins_mina, 13, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_maxs_mina, 13, 9, 6, low(mina)).
table(low_v_c_minc_maxc_mins_maxs_mina, 13, 12, 6, low(mina)).
table(low_v_c_minc_s_mins_maxs_mina, 13, 9, 6, low(mina)).
table(low_v_c_maxc_s_mins_maxs_mina, 13, 7, 6, low(mina)).
table(low_v_minc_maxc_s_mins_maxs_mina, 13, 8, 6, low(mina)).
table(low_c_minc_maxc_s_mins_maxs_mina, 13, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_mins_maxs_mina, 13, 10, 7, low(mina)).
table(up_v_maxa, 13, 19, 1, up(maxa)).
table(up_v_c_maxa, 13, 19, 2, up(maxa)).
table(up_v_minc_maxa, 13, 19, 2, up(maxa)).
table(up_v_maxc_maxa, 13, 19, 2, up(maxa)).
table(up_v_s_maxa, 13, 19, 2, up(maxa)).
table(up_v_mins_maxa, 13, 19, 2, up(maxa)).
table(up_v_maxs_maxa, 13, 19, 2, up(maxa)).
table(up_c_maxc_maxa, 13, 19, 2, up(maxa)).
table(up_maxc_s_maxa, 13, 18, 2, up(maxa)).
table(up_s_maxs_maxa, 13, 19, 2, up(maxa)).
table(up_v_c_minc_maxa, 13, 19, 3, up(maxa)).
table(up_v_c_maxc_maxa, 13, 19, 3, up(maxa)).
table(up_v_c_s_maxa, 13, 19, 3, up(maxa)).
table(up_v_c_mins_maxa, 13, 19, 3, up(maxa)).
table(up_v_c_maxs_maxa, 13, 7, 3, up(maxa)).
table(up_v_minc_maxc_maxa, 13, 19, 3, up(maxa)).
table(up_v_minc_s_maxa, 13, 18, 3, up(maxa)).
table(up_v_minc_mins_maxa, 13, 19, 3, up(maxa)).
table(up_v_minc_maxs_maxa, 13, 19, 3, up(maxa)).
table(up_v_maxc_s_maxa, 13, 18, 3, up(maxa)).
table(up_v_maxc_mins_maxa, 13, 13, 3, up(maxa)).
table(up_v_maxc_maxs_maxa, 13, 7, 3, up(maxa)).
table(up_v_s_mins_maxa, 13, 19, 3, up(maxa)).
table(up_v_s_maxs_maxa, 13, 19, 3, up(maxa)).
table(up_v_mins_maxs_maxa, 13, 19, 3, up(maxa)).
table(up_c_minc_maxc_maxa, 13, 19, 3, up(maxa)).
table(up_c_maxc_s_maxa, 13, 18, 3, up(maxa)).
table(up_c_maxc_mins_maxa, 13, 19, 3, up(maxa)).
table(up_c_maxc_maxs_maxa, 13, 7, 3, up(maxa)).
table(up_c_s_maxs_maxa, 13, 19, 3, up(maxa)).
table(up_minc_maxc_s_maxa, 13, 9, 3, up(maxa)).
table(up_minc_s_maxs_maxa, 13, 18, 3, up(maxa)).
table(up_maxc_s_mins_maxa, 13, 12, 3, up(maxa)).
table(up_maxc_s_maxs_maxa, 13, 6, 3, up(maxa)).
table(up_s_mins_maxs_maxa, 13, 19, 3, up(maxa)).
table(up_v_c_minc_maxc_maxa, 13, 19, 4, up(maxa)).
table(up_v_c_minc_s_maxa, 13, 18, 4, up(maxa)).
table(up_v_c_minc_mins_maxa, 13, 11, 4, up(maxa)).
table(up_v_c_minc_maxs_maxa, 13, 8, 4, up(maxa)).
table(up_v_c_maxc_s_maxa, 13, 18, 4, up(maxa)).
table(up_v_c_maxc_mins_maxa, 13, 13, 4, up(maxa)).
table(up_v_c_maxc_maxs_maxa, 13, 6, 4, up(maxa)).
table(up_v_c_s_mins_maxa, 13, 19, 4, up(maxa)).
table(up_v_c_s_maxs_maxa, 13, 19, 4, up(maxa)).
table(up_v_c_mins_maxs_maxa, 13, 11, 4, up(maxa)).
table(up_v_minc_maxc_s_maxa, 13, 18, 4, up(maxa)).
table(up_v_minc_maxc_mins_maxa, 13, 19, 4, up(maxa)).
table(up_v_minc_maxc_maxs_maxa, 13, 10, 4, up(maxa)).
table(up_v_minc_s_mins_maxa, 13, 18, 4, up(maxa)).
table(up_v_minc_s_maxs_maxa, 13, 18, 4, up(maxa)).
table(up_v_minc_mins_maxs_maxa, 13, 18, 4, up(maxa)).
table(up_v_maxc_s_mins_maxa, 13, 12, 4, up(maxa)).
table(up_v_maxc_s_maxs_maxa, 13, 18, 4, up(maxa)).
table(up_v_maxc_mins_maxs_maxa, 13, 6, 4, up(maxa)).
table(up_v_s_mins_maxs_maxa, 13, 19, 4, up(maxa)).
table(up_c_minc_maxc_s_maxa, 13, 18, 4, up(maxa)).
table(up_c_minc_maxc_mins_maxa, 13, 19, 4, up(maxa)).
table(up_c_minc_maxc_maxs_maxa, 13, 10, 4, up(maxa)).
table(up_c_minc_s_maxs_maxa, 13, 11, 4, up(maxa)).
table(up_c_maxc_s_mins_maxa, 13, 12, 4, up(maxa)).
table(up_c_maxc_s_maxs_maxa, 13, 18, 4, up(maxa)).
table(up_c_maxc_mins_maxs_maxa, 13, 6, 4, up(maxa)).
table(up_c_s_mins_maxs_maxa, 13, 19, 4, up(maxa)).
table(up_minc_maxc_s_mins_maxa, 13, 9, 4, up(maxa)).
table(up_minc_maxc_s_maxs_maxa, 13, 7, 4, up(maxa)).
table(up_minc_s_mins_maxs_maxa, 13, 18, 4, up(maxa)).
table(up_maxc_s_mins_maxs_maxa, 13, 6, 4, up(maxa)).
table(up_v_c_minc_maxc_s_maxa, 13, 18, 5, up(maxa)).
table(up_v_c_minc_maxc_mins_maxa, 13, 19, 5, up(maxa)).
table(up_v_c_minc_maxc_maxs_maxa, 13, 10, 5, up(maxa)).
table(up_v_c_minc_s_mins_maxa, 13, 10, 5, up(maxa)).
table(up_v_c_minc_s_maxs_maxa, 13, 9, 5, up(maxa)).
table(up_v_c_minc_mins_maxs_maxa, 13, 8, 5, up(maxa)).
table(up_v_c_maxc_s_mins_maxa, 13, 12, 5, up(maxa)).
table(up_v_c_maxc_s_maxs_maxa, 13, 18, 5, up(maxa)).
table(up_v_c_maxc_mins_maxs_maxa, 13, 6, 5, up(maxa)).
table(up_v_c_s_mins_maxs_maxa, 13, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_mins_maxa, 13, 18, 5, up(maxa)).
table(up_v_minc_maxc_s_maxs_maxa, 13, 11, 5, up(maxa)).
table(up_v_minc_maxc_mins_maxs_maxa, 13, 9, 5, up(maxa)).
table(up_v_minc_s_mins_maxs_maxa, 13, 18, 5, up(maxa)).
table(up_v_maxc_s_mins_maxs_maxa, 13, 12, 5, up(maxa)).
table(up_c_minc_maxc_s_mins_maxa, 13, 18, 5, up(maxa)).
table(up_c_minc_maxc_s_maxs_maxa, 13, 12, 5, up(maxa)).
table(up_c_minc_maxc_mins_maxs_maxa, 13, 9, 5, up(maxa)).
table(up_c_minc_s_mins_maxs_maxa, 13, 10, 5, up(maxa)).
table(up_c_maxc_s_mins_maxs_maxa, 13, 12, 5, up(maxa)).
table(up_minc_maxc_s_mins_maxs_maxa, 13, 7, 5, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxa, 13, 18, 6, up(maxa)).
table(up_v_c_minc_maxc_s_maxs_maxa, 13, 11, 6, up(maxa)).
table(up_v_c_minc_maxc_mins_maxs_maxa, 13, 9, 6, up(maxa)).
table(up_v_c_minc_s_mins_maxs_maxa, 13, 10, 6, up(maxa)).
table(up_v_c_maxc_s_mins_maxs_maxa, 13, 12, 6, up(maxa)).
table(up_v_minc_maxc_s_mins_maxs_maxa, 13, 11, 6, up(maxa)).
table(up_c_minc_maxc_s_mins_maxs_maxa, 13, 12, 6, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxs_maxa, 13, 11, 7, up(maxa)).
table(low_c_v, 14, 18, 1, low(v)).
table(low_minc_v, 14, 11, 1, low(v)).
table(low_maxc_v, 14, 11, 1, low(v)).
table(low_s_v, 14, 7, 1, low(v)).
table(low_mins_v, 14, 18, 1, low(v)).
table(low_maxs_v, 14, 18, 1, low(v)).
table(low_c_minc_v, 14, 11, 2, low(v)).
table(low_c_maxc_v, 14, 11, 2, low(v)).
table(low_c_s_v, 14, 8, 2, low(v)).
table(low_c_mins_v, 14, 18, 2, low(v)).
table(low_c_maxs_v, 14, 18, 2, low(v)).
table(low_minc_maxc_v, 14, 11, 2, low(v)).
table(low_minc_s_v, 14, 3, 2, low(v)).
table(low_minc_mins_v, 14, 12, 2, low(v)).
table(low_minc_maxs_v, 14, 12, 2, low(v)).
table(low_maxc_s_v, 14, 3, 2, low(v)).
table(low_maxc_mins_v, 14, 12, 2, low(v)).
table(low_maxc_maxs_v, 14, 13, 2, low(v)).
table(low_s_mins_v, 14, 7, 2, low(v)).
table(low_s_maxs_v, 14, 7, 2, low(v)).
table(low_mins_maxs_v, 14, 12, 2, low(v)).
table(low_c_minc_maxc_v, 14, 11, 3, low(v)).
table(low_c_minc_s_v, 14, 6, 3, low(v)).
table(low_c_minc_mins_v, 14, 12, 3, low(v)).
table(low_c_minc_maxs_v, 14, 12, 3, low(v)).
table(low_c_maxc_s_v, 14, 4, 3, low(v)).
table(low_c_maxc_mins_v, 14, 12, 3, low(v)).
table(low_c_maxc_maxs_v, 14, 13, 3, low(v)).
table(low_c_s_mins_v, 14, 8, 3, low(v)).
table(low_c_s_maxs_v, 14, 8, 3, low(v)).
table(low_c_mins_maxs_v, 14, 18, 3, low(v)).
table(low_minc_maxc_s_v, 14, 5, 3, low(v)).
table(low_minc_maxc_mins_v, 14, 12, 3, low(v)).
table(low_minc_maxc_maxs_v, 14, 12, 3, low(v)).
table(low_minc_s_mins_v, 14, 4, 3, low(v)).
table(low_minc_s_maxs_v, 14, 4, 3, low(v)).
table(low_minc_mins_maxs_v, 14, 13, 3, low(v)).
table(low_maxc_s_mins_v, 14, 4, 3, low(v)).
table(low_maxc_s_maxs_v, 14, 4, 3, low(v)).
table(low_maxc_mins_maxs_v, 14, 14, 3, low(v)).
table(low_s_mins_maxs_v, 14, 7, 3, low(v)).
table(low_c_minc_maxc_s_v, 14, 7, 4, low(v)).
table(low_c_minc_maxc_mins_v, 14, 12, 4, low(v)).
table(low_c_minc_maxc_maxs_v, 14, 12, 4, low(v)).
table(low_c_minc_s_mins_v, 14, 7, 4, low(v)).
table(low_c_minc_s_maxs_v, 14, 7, 4, low(v)).
table(low_c_minc_mins_maxs_v, 14, 13, 4, low(v)).
table(low_c_maxc_s_mins_v, 14, 5, 4, low(v)).
table(low_c_maxc_s_maxs_v, 14, 5, 4, low(v)).
table(low_c_maxc_mins_maxs_v, 14, 14, 4, low(v)).
table(low_c_s_mins_maxs_v, 14, 8, 4, low(v)).
table(low_minc_maxc_s_mins_v, 14, 6, 4, low(v)).
table(low_minc_maxc_s_maxs_v, 14, 6, 4, low(v)).
table(low_minc_maxc_mins_maxs_v, 14, 13, 4, low(v)).
table(low_minc_s_mins_maxs_v, 14, 5, 4, low(v)).
table(low_maxc_s_mins_maxs_v, 14, 5, 4, low(v)).
table(low_c_minc_maxc_s_mins_v, 14, 8, 5, low(v)).
table(low_c_minc_maxc_s_maxs_v, 14, 8, 5, low(v)).
table(low_c_minc_maxc_mins_maxs_v, 14, 13, 5, low(v)).
table(low_c_minc_s_mins_maxs_v, 14, 8, 5, low(v)).
table(low_c_maxc_s_mins_maxs_v, 14, 6, 5, low(v)).
table(low_minc_maxc_s_mins_maxs_v, 14, 7, 5, low(v)).
table(low_c_minc_maxc_s_mins_maxs_v, 14, 9, 6, low(v)).
table(up_c_maxc_v, 14, 3, 2, up(v)).
table(up_maxc_s_v, 14, 3, 2, up(v)).
table(up_s_maxs_v, 14, 3, 2, up(v)).
table(up_c_minc_maxc_v, 14, 6, 3, up(v)).
table(up_c_maxc_s_v, 14, 4, 3, up(v)).
table(up_c_maxc_mins_v, 14, 4, 3, up(v)).
table(up_c_maxc_maxs_v, 14, 4, 3, up(v)).
table(up_c_s_maxs_v, 14, 4, 3, up(v)).
table(up_minc_maxc_s_v, 14, 5, 3, up(v)).
table(up_minc_s_maxs_v, 14, 4, 3, up(v)).
table(up_maxc_s_mins_v, 14, 4, 3, up(v)).
table(up_maxc_s_maxs_v, 14, 4, 3, up(v)).
table(up_s_mins_maxs_v, 14, 4, 3, up(v)).
table(up_c_minc_maxc_s_v, 14, 7, 4, up(v)).
table(up_c_minc_maxc_mins_v, 14, 7, 4, up(v)).
table(up_c_minc_maxc_maxs_v, 14, 7, 4, up(v)).
table(up_c_minc_s_maxs_v, 14, 7, 4, up(v)).
table(up_c_maxc_s_mins_v, 14, 5, 4, up(v)).
table(up_c_maxc_s_maxs_v, 14, 5, 4, up(v)).
table(up_c_maxc_mins_maxs_v, 14, 5, 4, up(v)).
table(up_c_s_mins_maxs_v, 14, 5, 4, up(v)).
table(up_minc_maxc_s_mins_v, 14, 6, 4, up(v)).
table(up_minc_maxc_s_maxs_v, 14, 6, 4, up(v)).
table(up_minc_s_mins_maxs_v, 14, 5, 4, up(v)).
table(up_maxc_s_mins_maxs_v, 14, 5, 4, up(v)).
table(up_c_minc_maxc_s_mins_v, 14, 8, 5, up(v)).
table(up_c_minc_maxc_s_maxs_v, 14, 8, 5, up(v)).
table(up_c_minc_maxc_mins_maxs_v, 14, 8, 5, up(v)).
table(up_c_minc_s_mins_maxs_v, 14, 8, 5, up(v)).
table(up_c_maxc_s_mins_maxs_v, 14, 6, 5, up(v)).
table(up_minc_maxc_s_mins_maxs_v, 14, 7, 5, up(v)).
table(up_c_minc_maxc_s_mins_maxs_v, 14, 9, 6, up(v)).
table(low_v_c, 14, 11, 1, low(c)).
table(low_minc_c, 14, 11, 1, low(c)).
table(low_maxc_c, 14, 11, 1, low(c)).
table(low_s_c, 14, 8, 1, low(c)).
table(low_mins_c, 14, 8, 1, low(c)).
table(low_maxs_c, 14, 8, 1, low(c)).
table(low_v_minc_c, 14, 11, 2, low(c)).
table(low_v_maxc_c, 14, 3, 2, low(c)).
table(low_v_s_c, 14, 13, 2, low(c)).
table(low_v_mins_c, 14, 12, 2, low(c)).
table(low_v_maxs_c, 14, 13, 2, low(c)).
table(low_minc_maxc_c, 14, 11, 2, low(c)).
table(low_minc_s_c, 14, 5, 2, low(c)).
table(low_minc_mins_c, 14, 9, 2, low(c)).
table(low_minc_maxs_c, 14, 9, 2, low(c)).
table(low_maxc_s_c, 14, 3, 2, low(c)).
table(low_maxc_mins_c, 14, 12, 2, low(c)).
table(low_maxc_maxs_c, 14, 13, 2, low(c)).
table(low_s_mins_c, 14, 10, 2, low(c)).
table(low_s_maxs_c, 14, 11, 2, low(c)).
table(low_mins_maxs_c, 14, 9, 2, low(c)).
table(low_v_minc_maxc_c, 14, 6, 3, low(c)).
table(low_v_minc_s_c, 14, 12, 3, low(c)).
table(low_v_minc_mins_c, 14, 12, 3, low(c)).
table(low_v_minc_maxs_c, 14, 12, 3, low(c)).
table(low_v_maxc_s_c, 14, 4, 3, low(c)).
table(low_v_maxc_mins_c, 14, 4, 3, low(c)).
table(low_v_maxc_maxs_c, 14, 4, 3, low(c)).
table(low_v_s_mins_c, 14, 14, 3, low(c)).
table(low_v_s_maxs_c, 14, 14, 3, low(c)).
table(low_v_mins_maxs_c, 14, 14, 3, low(c)).
table(low_minc_maxc_s_c, 14, 5, 3, low(c)).
table(low_minc_maxc_mins_c, 14, 12, 3, low(c)).
table(low_minc_maxc_maxs_c, 14, 12, 3, low(c)).
table(low_minc_s_mins_c, 14, 6, 3, low(c)).
table(low_minc_s_maxs_c, 14, 7, 3, low(c)).
table(low_minc_mins_maxs_c, 14, 7, 3, low(c)).
table(low_maxc_s_mins_c, 14, 4, 3, low(c)).
table(low_maxc_s_maxs_c, 14, 4, 3, low(c)).
table(low_maxc_mins_maxs_c, 14, 6, 3, low(c)).
table(low_s_mins_maxs_c, 14, 12, 3, low(c)).
table(low_v_minc_maxc_s_c, 14, 7, 4, low(c)).
table(low_v_minc_maxc_mins_c, 14, 7, 4, low(c)).
table(low_v_minc_maxc_maxs_c, 14, 7, 4, low(c)).
table(low_v_minc_s_mins_c, 14, 13, 4, low(c)).
table(low_v_minc_s_maxs_c, 14, 13, 4, low(c)).
table(low_v_minc_mins_maxs_c, 14, 13, 4, low(c)).
table(low_v_maxc_s_mins_c, 14, 5, 4, low(c)).
table(low_v_maxc_s_maxs_c, 14, 5, 4, low(c)).
table(low_v_maxc_mins_maxs_c, 14, 5, 4, low(c)).
table(low_v_s_mins_maxs_c, 14, 15, 4, low(c)).
table(low_minc_maxc_s_mins_c, 14, 6, 4, low(c)).
table(low_minc_maxc_s_maxs_c, 14, 6, 4, low(c)).
table(low_minc_maxc_mins_maxs_c, 14, 8, 4, low(c)).
table(low_minc_s_mins_maxs_c, 14, 8, 4, low(c)).
table(low_maxc_s_mins_maxs_c, 14, 5, 4, low(c)).
table(low_v_minc_maxc_s_mins_c, 14, 8, 5, low(c)).
table(low_v_minc_maxc_s_maxs_c, 14, 8, 5, low(c)).
table(low_v_minc_maxc_mins_maxs_c, 14, 8, 5, low(c)).
table(low_v_minc_s_mins_maxs_c, 14, 14, 5, low(c)).
table(low_v_maxc_s_mins_maxs_c, 14, 6, 5, low(c)).
table(low_minc_maxc_s_mins_maxs_c, 14, 7, 5, low(c)).
table(low_v_minc_maxc_s_mins_maxs_c, 14, 9, 6, low(c)).
table(up_v_c, 14, 18, 1, up(c)).
table(up_s_c, 14, 3, 1, up(c)).
table(up_v_minc_c, 14, 8, 2, up(c)).
table(up_v_maxc_c, 14, 11, 2, up(c)).
table(up_v_s_c, 14, 4, 2, up(c)).
table(up_v_mins_c, 14, 12, 2, up(c)).
table(up_v_maxs_c, 14, 18, 2, up(c)).
table(up_minc_s_c, 14, 3, 2, up(c)).
table(up_maxc_s_c, 14, 4, 2, up(c)).
table(up_s_mins_c, 14, 5, 2, up(c)).
table(up_s_maxs_c, 14, 5, 2, up(c)).
table(up_v_minc_maxc_c, 14, 11, 3, up(c)).
table(up_v_minc_s_c, 14, 6, 3, up(c)).
table(up_v_minc_mins_c, 14, 7, 3, up(c)).
table(up_v_minc_maxs_c, 14, 7, 3, up(c)).
table(up_v_maxc_s_c, 14, 5, 3, up(c)).
table(up_v_maxc_mins_c, 14, 6, 3, up(c)).
table(up_v_maxc_maxs_c, 14, 13, 3, up(c)).
table(up_v_s_mins_c, 14, 8, 3, up(c)).
table(up_v_s_maxs_c, 14, 6, 3, up(c)).
table(up_v_mins_maxs_c, 14, 18, 3, up(c)).
table(up_minc_maxc_s_c, 14, 5, 3, up(c)).
table(up_minc_s_mins_c, 14, 4, 3, up(c)).
table(up_minc_s_maxs_c, 14, 4, 3, up(c)).
table(up_maxc_s_mins_c, 14, 5, 3, up(c)).
table(up_maxc_s_maxs_c, 14, 5, 3, up(c)).
table(up_s_mins_maxs_c, 14, 8, 3, up(c)).
table(up_v_minc_maxc_s_c, 14, 7, 4, up(c)).
table(up_v_minc_maxc_mins_c, 14, 12, 4, up(c)).
table(up_v_minc_maxc_maxs_c, 14, 12, 4, up(c)).
table(up_v_minc_s_mins_c, 14, 7, 4, up(c)).
table(up_v_minc_s_maxs_c, 14, 7, 4, up(c)).
table(up_v_minc_mins_maxs_c, 14, 8, 4, up(c)).
table(up_v_maxc_s_mins_c, 14, 6, 4, up(c)).
table(up_v_maxc_s_maxs_c, 14, 6, 4, up(c)).
table(up_v_maxc_mins_maxs_c, 14, 14, 4, up(c)).
table(up_v_s_mins_maxs_c, 14, 10, 4, up(c)).
table(up_minc_maxc_s_mins_c, 14, 6, 4, up(c)).
table(up_minc_maxc_s_maxs_c, 14, 6, 4, up(c)).
table(up_minc_s_mins_maxs_c, 14, 5, 4, up(c)).
table(up_maxc_s_mins_maxs_c, 14, 6, 4, up(c)).
table(up_v_minc_maxc_s_mins_c, 14, 8, 5, up(c)).
table(up_v_minc_maxc_s_maxs_c, 14, 8, 5, up(c)).
table(up_v_minc_maxc_mins_maxs_c, 14, 13, 5, up(c)).
table(up_v_minc_s_mins_maxs_c, 14, 8, 5, up(c)).
table(up_v_maxc_s_mins_maxs_c, 14, 7, 5, up(c)).
table(up_minc_maxc_s_mins_maxs_c, 14, 7, 5, up(c)).
table(up_v_minc_maxc_s_mins_maxs_c, 14, 9, 6, up(c)).
table(low_v_minc, 14, 3, 1, low(minc)).
table(low_c_minc, 14, 3, 1, low(minc)).
table(low_maxc_minc, 14, 3, 1, low(minc)).
table(low_s_minc, 14, 3, 1, low(minc)).
table(low_mins_minc, 14, 2, 1, low(minc)).
table(low_maxs_minc, 14, 3, 1, low(minc)).
table(low_v_c_minc, 14, 5, 2, low(minc)).
table(low_v_maxc_minc, 14, 4, 2, low(minc)).
table(low_v_s_minc, 14, 4, 2, low(minc)).
table(low_v_mins_minc, 14, 3, 2, low(minc)).
table(low_v_maxs_minc, 14, 4, 2, low(minc)).
table(low_c_maxc_minc, 14, 4, 2, low(minc)).
table(low_c_s_minc, 14, 4, 2, low(minc)).
table(low_c_mins_minc, 14, 3, 2, low(minc)).
table(low_c_maxs_minc, 14, 4, 2, low(minc)).
table(low_maxc_s_minc, 14, 4, 2, low(minc)).
table(low_maxc_mins_minc, 14, 4, 2, low(minc)).
table(low_maxc_maxs_minc, 14, 4, 2, low(minc)).
table(low_s_mins_minc, 14, 3, 2, low(minc)).
table(low_s_maxs_minc, 14, 4, 2, low(minc)).
table(low_mins_maxs_minc, 14, 3, 2, low(minc)).
table(low_v_c_maxc_minc, 14, 6, 3, low(minc)).
table(low_v_c_s_minc, 14, 6, 3, low(minc)).
table(low_v_c_mins_minc, 14, 6, 3, low(minc)).
table(low_v_c_maxs_minc, 14, 6, 3, low(minc)).
table(low_v_maxc_s_minc, 14, 5, 3, low(minc)).
table(low_v_maxc_mins_minc, 14, 5, 3, low(minc)).
table(low_v_maxc_maxs_minc, 14, 5, 3, low(minc)).
table(low_v_s_mins_minc, 14, 4, 3, low(minc)).
table(low_v_s_maxs_minc, 14, 5, 3, low(minc)).
table(low_v_mins_maxs_minc, 14, 4, 3, low(minc)).
table(low_c_maxc_s_minc, 14, 5, 3, low(minc)).
table(low_c_maxc_mins_minc, 14, 5, 3, low(minc)).
table(low_c_maxc_maxs_minc, 14, 5, 3, low(minc)).
table(low_c_s_mins_minc, 14, 4, 3, low(minc)).
table(low_c_s_maxs_minc, 14, 6, 3, low(minc)).
table(low_c_mins_maxs_minc, 14, 4, 3, low(minc)).
table(low_maxc_s_mins_minc, 14, 5, 3, low(minc)).
table(low_maxc_s_maxs_minc, 14, 5, 3, low(minc)).
table(low_maxc_mins_maxs_minc, 14, 5, 3, low(minc)).
table(low_s_mins_maxs_minc, 14, 4, 3, low(minc)).
table(low_v_c_maxc_s_minc, 14, 7, 4, low(minc)).
table(low_v_c_maxc_mins_minc, 14, 7, 4, low(minc)).
table(low_v_c_maxc_maxs_minc, 14, 7, 4, low(minc)).
table(low_v_c_s_mins_minc, 14, 7, 4, low(minc)).
table(low_v_c_s_maxs_minc, 14, 7, 4, low(minc)).
table(low_v_c_mins_maxs_minc, 14, 7, 4, low(minc)).
table(low_v_maxc_s_mins_minc, 14, 6, 4, low(minc)).
table(low_v_maxc_s_maxs_minc, 14, 6, 4, low(minc)).
table(low_v_maxc_mins_maxs_minc, 14, 6, 4, low(minc)).
table(low_v_s_mins_maxs_minc, 14, 5, 4, low(minc)).
table(low_c_maxc_s_mins_minc, 14, 6, 4, low(minc)).
table(low_c_maxc_s_maxs_minc, 14, 6, 4, low(minc)).
table(low_c_maxc_mins_maxs_minc, 14, 6, 4, low(minc)).
table(low_c_s_mins_maxs_minc, 14, 6, 4, low(minc)).
table(low_maxc_s_mins_maxs_minc, 14, 6, 4, low(minc)).
table(low_v_c_maxc_s_mins_minc, 14, 8, 5, low(minc)).
table(low_v_c_maxc_s_maxs_minc, 14, 8, 5, low(minc)).
table(low_v_c_maxc_mins_maxs_minc, 14, 8, 5, low(minc)).
table(low_v_c_s_mins_maxs_minc, 14, 8, 5, low(minc)).
table(low_v_maxc_s_mins_maxs_minc, 14, 7, 5, low(minc)).
table(low_c_maxc_s_mins_maxs_minc, 14, 7, 5, low(minc)).
table(low_v_c_maxc_s_mins_maxs_minc, 14, 9, 6, low(minc)).
table(up_v_minc, 14, 11, 1, up(minc)).
table(up_maxc_minc, 14, 6, 1, up(minc)).
table(up_v_c_minc, 14, 5, 2, up(minc)).
table(up_v_maxc_minc, 14, 6, 2, up(minc)).
table(up_v_s_minc, 14, 13, 2, up(minc)).
table(up_v_mins_minc, 14, 12, 2, up(minc)).
table(up_v_maxs_minc, 14, 13, 2, up(minc)).
table(up_c_maxc_minc, 14, 4, 2, up(minc)).
table(up_maxc_s_minc, 14, 6, 2, up(minc)).
table(up_maxc_mins_minc, 14, 4, 2, up(minc)).
table(up_maxc_maxs_minc, 14, 7, 2, up(minc)).
table(up_s_maxs_minc, 14, 14, 2, up(minc)).
table(up_v_c_maxc_minc, 14, 6, 3, up(minc)).
table(up_v_c_s_minc, 14, 6, 3, up(minc)).
table(up_v_c_mins_minc, 14, 6, 3, up(minc)).
table(up_v_c_maxs_minc, 14, 6, 3, up(minc)).
table(up_v_maxc_s_minc, 14, 7, 3, up(minc)).
table(up_v_maxc_mins_minc, 14, 7, 3, up(minc)).
table(up_v_maxc_maxs_minc, 14, 7, 3, up(minc)).
table(up_v_s_mins_minc, 14, 14, 3, up(minc)).
table(up_v_s_maxs_minc, 14, 14, 3, up(minc)).
table(up_v_mins_maxs_minc, 14, 14, 3, up(minc)).
table(up_c_maxc_s_minc, 14, 5, 3, up(minc)).
table(up_c_maxc_mins_minc, 14, 5, 3, up(minc)).
table(up_c_maxc_maxs_minc, 14, 5, 3, up(minc)).
table(up_c_s_maxs_minc, 14, 4, 3, up(minc)).
table(up_maxc_s_mins_minc, 14, 5, 3, up(minc)).
table(up_maxc_s_maxs_minc, 14, 7, 3, up(minc)).
table(up_maxc_mins_maxs_minc, 14, 5, 3, up(minc)).
table(up_s_mins_maxs_minc, 14, 15, 3, up(minc)).
table(up_v_c_maxc_s_minc, 14, 7, 4, up(minc)).
table(up_v_c_maxc_mins_minc, 14, 7, 4, up(minc)).
table(up_v_c_maxc_maxs_minc, 14, 7, 4, up(minc)).
table(up_v_c_s_mins_minc, 14, 7, 4, up(minc)).
table(up_v_c_s_maxs_minc, 14, 7, 4, up(minc)).
table(up_v_c_mins_maxs_minc, 14, 7, 4, up(minc)).
table(up_v_maxc_s_mins_minc, 14, 6, 4, up(minc)).
table(up_v_maxc_s_maxs_minc, 14, 6, 4, up(minc)).
table(up_v_maxc_mins_maxs_minc, 14, 6, 4, up(minc)).
table(up_v_s_mins_maxs_minc, 14, 15, 4, up(minc)).
table(up_c_maxc_s_mins_minc, 14, 6, 4, up(minc)).
table(up_c_maxc_s_maxs_minc, 14, 6, 4, up(minc)).
table(up_c_maxc_mins_maxs_minc, 14, 6, 4, up(minc)).
table(up_c_s_mins_maxs_minc, 14, 5, 4, up(minc)).
table(up_maxc_s_mins_maxs_minc, 14, 6, 4, up(minc)).
table(up_v_c_maxc_s_mins_minc, 14, 8, 5, up(minc)).
table(up_v_c_maxc_s_maxs_minc, 14, 8, 5, up(minc)).
table(up_v_c_maxc_mins_maxs_minc, 14, 8, 5, up(minc)).
table(up_v_c_s_mins_maxs_minc, 14, 8, 5, up(minc)).
table(up_v_maxc_s_mins_maxs_minc, 14, 7, 5, up(minc)).
table(up_c_maxc_s_mins_maxs_minc, 14, 7, 5, up(minc)).
table(up_v_c_maxc_s_mins_maxs_minc, 14, 9, 6, up(minc)).
table(low_v_maxc, 14, 18, 1, low(maxc)).
table(low_c_maxc, 14, 18, 1, low(maxc)).
table(low_minc_maxc, 14, 6, 1, low(maxc)).
table(low_s_maxc, 14, 18, 1, low(maxc)).
table(low_mins_maxc, 14, 9, 1, low(maxc)).
table(low_maxs_maxc, 14, 2, 1, low(maxc)).
table(low_v_c_maxc, 14, 3, 2, low(maxc)).
table(low_v_minc_maxc, 14, 6, 2, low(maxc)).
table(low_v_s_maxc, 14, 4, 2, low(maxc)).
table(low_v_mins_maxc, 14, 9, 2, low(maxc)).
table(low_v_maxs_maxc, 14, 3, 2, low(maxc)).
table(low_c_minc_maxc, 14, 11, 2, low(maxc)).
table(low_c_s_maxc, 14, 4, 2, low(maxc)).
table(low_c_mins_maxc, 14, 18, 2, low(maxc)).
table(low_c_maxs_maxc, 14, 3, 2, low(maxc)).
table(low_minc_s_maxc, 14, 6, 2, low(maxc)).
table(low_minc_mins_maxc, 14, 6, 2, low(maxc)).
table(low_minc_maxs_maxc, 14, 4, 2, low(maxc)).
table(low_s_mins_maxc, 14, 18, 2, low(maxc)).
table(low_s_maxs_maxc, 14, 3, 2, low(maxc)).
table(low_mins_maxs_maxc, 14, 3, 2, low(maxc)).
table(low_v_c_minc_maxc, 14, 6, 3, low(maxc)).
table(low_v_c_s_maxc, 14, 4, 3, low(maxc)).
table(low_v_c_mins_maxc, 14, 4, 3, low(maxc)).
table(low_v_c_maxs_maxc, 14, 4, 3, low(maxc)).
table(low_v_minc_s_maxc, 14, 5, 3, low(maxc)).
table(low_v_minc_mins_maxc, 14, 7, 3, low(maxc)).
table(low_v_minc_maxs_maxc, 14, 5, 3, low(maxc)).
table(low_v_s_mins_maxc, 14, 5, 3, low(maxc)).
table(low_v_s_maxs_maxc, 14, 4, 3, low(maxc)).
table(low_v_mins_maxs_maxc, 14, 4, 3, low(maxc)).
table(low_c_minc_s_maxc, 14, 5, 3, low(maxc)).
table(low_c_minc_mins_maxc, 14, 8, 3, low(maxc)).
table(low_c_minc_maxs_maxc, 14, 5, 3, low(maxc)).
table(low_c_s_mins_maxc, 14, 4, 3, low(maxc)).
table(low_c_s_maxs_maxc, 14, 4, 3, low(maxc)).
table(low_c_mins_maxs_maxc, 14, 4, 3, low(maxc)).
table(low_minc_s_mins_maxc, 14, 7, 3, low(maxc)).
table(low_minc_s_maxs_maxc, 14, 5, 3, low(maxc)).
table(low_minc_mins_maxs_maxc, 14, 5, 3, low(maxc)).
table(low_s_mins_maxs_maxc, 14, 4, 3, low(maxc)).
table(low_v_c_minc_s_maxc, 14, 7, 4, low(maxc)).
table(low_v_c_minc_mins_maxc, 14, 7, 4, low(maxc)).
table(low_v_c_minc_maxs_maxc, 14, 7, 4, low(maxc)).
table(low_v_c_s_mins_maxc, 14, 5, 4, low(maxc)).
table(low_v_c_s_maxs_maxc, 14, 5, 4, low(maxc)).
table(low_v_c_mins_maxs_maxc, 14, 5, 4, low(maxc)).
table(low_v_minc_s_mins_maxc, 14, 6, 4, low(maxc)).
table(low_v_minc_s_maxs_maxc, 14, 6, 4, low(maxc)).
table(low_v_minc_mins_maxs_maxc, 14, 6, 4, low(maxc)).
table(low_v_s_mins_maxs_maxc, 14, 5, 4, low(maxc)).
table(low_c_minc_s_mins_maxc, 14, 6, 4, low(maxc)).
table(low_c_minc_s_maxs_maxc, 14, 6, 4, low(maxc)).
table(low_c_minc_mins_maxs_maxc, 14, 6, 4, low(maxc)).
table(low_c_s_mins_maxs_maxc, 14, 5, 4, low(maxc)).
table(low_minc_s_mins_maxs_maxc, 14, 6, 4, low(maxc)).
table(low_v_c_minc_s_mins_maxc, 14, 8, 5, low(maxc)).
table(low_v_c_minc_s_maxs_maxc, 14, 8, 5, low(maxc)).
table(low_v_c_minc_mins_maxs_maxc, 14, 8, 5, low(maxc)).
table(low_v_c_s_mins_maxs_maxc, 14, 6, 5, low(maxc)).
table(low_v_minc_s_mins_maxs_maxc, 14, 7, 5, low(maxc)).
table(low_c_minc_s_mins_maxs_maxc, 14, 7, 5, low(maxc)).
table(low_v_c_minc_s_mins_maxs_maxc, 14, 9, 6, low(maxc)).
table(up_v_maxc, 14, 11, 1, up(maxc)).
table(up_v_c_maxc, 14, 11, 2, up(maxc)).
table(up_v_minc_maxc, 14, 11, 2, up(maxc)).
table(up_v_s_maxc, 14, 13, 2, up(maxc)).
table(up_v_mins_maxc, 14, 12, 2, up(maxc)).
table(up_v_maxs_maxc, 14, 13, 2, up(maxc)).
table(up_v_c_minc_maxc, 14, 11, 3, up(maxc)).
table(up_v_c_s_maxc, 14, 13, 3, up(maxc)).
table(up_v_c_mins_maxc, 14, 12, 3, up(maxc)).
table(up_v_c_maxs_maxc, 14, 13, 3, up(maxc)).
table(up_v_minc_s_maxc, 14, 12, 3, up(maxc)).
table(up_v_minc_mins_maxc, 14, 12, 3, up(maxc)).
table(up_v_minc_maxs_maxc, 14, 12, 3, up(maxc)).
table(up_v_s_mins_maxc, 14, 14, 3, up(maxc)).
table(up_v_s_maxs_maxc, 14, 14, 3, up(maxc)).
table(up_v_mins_maxs_maxc, 14, 14, 3, up(maxc)).
table(up_v_c_minc_s_maxc, 14, 12, 4, up(maxc)).
table(up_v_c_minc_mins_maxc, 14, 12, 4, up(maxc)).
table(up_v_c_minc_maxs_maxc, 14, 12, 4, up(maxc)).
table(up_v_c_s_mins_maxc, 14, 14, 4, up(maxc)).
table(up_v_c_s_maxs_maxc, 14, 6, 4, up(maxc)).
table(up_v_c_mins_maxs_maxc, 14, 14, 4, up(maxc)).
table(up_v_minc_s_mins_maxc, 14, 13, 4, up(maxc)).
table(up_v_minc_s_maxs_maxc, 14, 13, 4, up(maxc)).
table(up_v_minc_mins_maxs_maxc, 14, 13, 4, up(maxc)).
table(up_v_s_mins_maxs_maxc, 14, 15, 4, up(maxc)).
table(up_v_c_minc_s_mins_maxc, 14, 13, 5, up(maxc)).
table(up_v_c_minc_s_maxs_maxc, 14, 8, 5, up(maxc)).
table(up_v_c_minc_mins_maxs_maxc, 14, 13, 5, up(maxc)).
table(up_v_c_s_mins_maxs_maxc, 14, 10, 5, up(maxc)).
table(up_v_minc_s_mins_maxs_maxc, 14, 14, 5, up(maxc)).
table(up_v_c_minc_s_mins_maxs_maxc, 14, 9, 6, up(maxc)).
table(low_v_s, 14, 18, 1, low(s)).
table(low_c_s, 14, 3, 1, low(s)).
table(low_minc_s, 14, 18, 1, low(s)).
table(low_maxc_s, 14, 18, 1, low(s)).
table(low_mins_s, 14, 18, 1, low(s)).
table(low_maxs_s, 14, 18, 1, low(s)).
table(low_v_c_s, 14, 4, 2, low(s)).
table(low_v_minc_s, 14, 18, 2, low(s)).
table(low_v_maxc_s, 14, 6, 2, low(s)).
table(low_v_mins_s, 14, 9, 2, low(s)).
table(low_v_maxs_s, 14, 3, 2, low(s)).
table(low_c_minc_s, 14, 5, 2, low(s)).
table(low_c_maxc_s, 14, 5, 2, low(s)).
table(low_c_mins_s, 14, 5, 2, low(s)).
table(low_c_maxs_s, 14, 5, 2, low(s)).
table(low_minc_maxc_s, 14, 18, 2, low(s)).
table(low_minc_mins_s, 14, 12, 2, low(s)).
table(low_minc_maxs_s, 14, 14, 2, low(s)).
table(low_maxc_mins_s, 14, 7, 2, low(s)).
table(low_maxc_maxs_s, 14, 14, 2, low(s)).
table(low_mins_maxs_s, 14, 12, 2, low(s)).
table(low_v_c_minc_s, 14, 8, 3, low(s)).
table(low_v_c_maxc_s, 14, 6, 3, low(s)).
table(low_v_c_mins_s, 14, 8, 3, low(s)).
table(low_v_c_maxs_s, 14, 4, 3, low(s)).
table(low_v_minc_maxc_s, 14, 10, 3, low(s)).
table(low_v_minc_mins_s, 14, 13, 3, low(s)).
table(low_v_minc_maxs_s, 14, 4, 3, low(s)).
table(low_v_maxc_mins_s, 14, 4, 3, low(s)).
table(low_v_maxc_maxs_s, 14, 4, 3, low(s)).
table(low_v_mins_maxs_s, 14, 4, 3, low(s)).
table(low_c_minc_maxc_s, 14, 8, 3, low(s)).
table(low_c_minc_mins_s, 14, 4, 3, low(s)).
table(low_c_minc_maxs_s, 14, 5, 3, low(s)).
table(low_c_maxc_mins_s, 14, 8, 3, low(s)).
table(low_c_maxc_maxs_s, 14, 5, 3, low(s)).
table(low_c_mins_maxs_s, 14, 8, 3, low(s)).
table(low_minc_maxc_mins_s, 14, 13, 3, low(s)).
table(low_minc_maxc_maxs_s, 14, 14, 3, low(s)).
table(low_minc_mins_maxs_s, 14, 8, 3, low(s)).
table(low_maxc_mins_maxs_s, 14, 9, 3, low(s)).
table(low_v_c_minc_maxc_s, 14, 10, 4, low(s)).
table(low_v_c_minc_mins_s, 14, 7, 4, low(s)).
table(low_v_c_minc_maxs_s, 14, 7, 4, low(s)).
table(low_v_c_maxc_mins_s, 14, 5, 4, low(s)).
table(low_v_c_maxc_maxs_s, 14, 5, 4, low(s)).
table(low_v_c_mins_maxs_s, 14, 5, 4, low(s)).
table(low_v_minc_maxc_mins_s, 14, 8, 4, low(s)).
table(low_v_minc_maxc_maxs_s, 14, 6, 4, low(s)).
table(low_v_minc_mins_maxs_s, 14, 5, 4, low(s)).
table(low_v_maxc_mins_maxs_s, 14, 5, 4, low(s)).
table(low_c_minc_maxc_mins_s, 14, 6, 4, low(s)).
table(low_c_minc_maxc_maxs_s, 14, 7, 4, low(s)).
table(low_c_minc_mins_maxs_s, 14, 5, 4, low(s)).
table(low_c_maxc_mins_maxs_s, 14, 6, 4, low(s)).
table(low_minc_maxc_mins_maxs_s, 14, 9, 4, low(s)).
table(low_v_c_minc_maxc_mins_s, 14, 8, 5, low(s)).
table(low_v_c_minc_maxc_maxs_s, 14, 8, 5, low(s)).
table(low_v_c_minc_mins_maxs_s, 14, 8, 5, low(s)).
table(low_v_c_maxc_mins_maxs_s, 14, 6, 5, low(s)).
table(low_v_minc_maxc_mins_maxs_s, 14, 7, 5, low(s)).
table(low_c_minc_maxc_mins_maxs_s, 14, 7, 5, low(s)).
table(low_v_c_minc_maxc_mins_maxs_s, 14, 9, 6, low(s)).
table(up_v_s, 14, 7, 1, up(s)).
table(up_v_c_s, 14, 8, 2, up(s)).
table(up_v_minc_s, 14, 8, 2, up(s)).
table(up_v_maxc_s, 14, 8, 2, up(s)).
table(up_v_mins_s, 14, 4, 2, up(s)).
table(up_v_maxs_s, 14, 7, 2, up(s)).
table(up_c_maxc_s, 14, 9, 2, up(s)).
table(up_v_c_minc_s, 14, 11, 3, up(s)).
table(up_v_c_maxc_s, 14, 9, 3, up(s)).
table(up_v_c_mins_s, 14, 5, 3, up(s)).
table(up_v_c_maxs_s, 14, 8, 3, up(s)).
table(up_v_minc_maxc_s, 14, 10, 3, up(s)).
table(up_v_minc_mins_s, 14, 5, 3, up(s)).
table(up_v_minc_maxs_s, 14, 8, 3, up(s)).
table(up_v_maxc_mins_s, 14, 5, 3, up(s)).
table(up_v_maxc_maxs_s, 14, 8, 3, up(s)).
table(up_v_mins_maxs_s, 14, 7, 3, up(s)).
table(up_c_minc_maxc_s, 14, 12, 3, up(s)).
table(up_c_maxc_mins_s, 14, 5, 3, up(s)).
table(up_c_maxc_maxs_s, 14, 9, 3, up(s)).
table(up_v_c_minc_maxc_s, 14, 12, 4, up(s)).
table(up_v_c_minc_mins_s, 14, 8, 4, up(s)).
table(up_v_c_minc_maxs_s, 14, 11, 4, up(s)).
table(up_v_c_maxc_mins_s, 14, 6, 4, up(s)).
table(up_v_c_maxc_maxs_s, 14, 9, 4, up(s)).
table(up_v_c_mins_maxs_s, 14, 8, 4, up(s)).
table(up_v_minc_maxc_mins_s, 14, 7, 4, up(s)).
table(up_v_minc_maxc_maxs_s, 14, 10, 4, up(s)).
table(up_v_minc_mins_maxs_s, 14, 8, 4, up(s)).
table(up_v_maxc_mins_maxs_s, 14, 8, 4, up(s)).
table(up_c_minc_maxc_mins_s, 14, 7, 4, up(s)).
table(up_c_minc_maxc_maxs_s, 14, 12, 4, up(s)).
table(up_c_maxc_mins_maxs_s, 14, 6, 4, up(s)).
table(up_v_c_minc_maxc_mins_s, 14, 9, 5, up(s)).
table(up_v_c_minc_maxc_maxs_s, 14, 12, 5, up(s)).
table(up_v_c_minc_mins_maxs_s, 14, 11, 5, up(s)).
table(up_v_c_maxc_mins_maxs_s, 14, 9, 5, up(s)).
table(up_v_minc_maxc_mins_maxs_s, 14, 10, 5, up(s)).
table(up_c_minc_maxc_mins_maxs_s, 14, 8, 5, up(s)).
table(up_v_c_minc_maxc_mins_maxs_s, 14, 12, 6, up(s)).
table(low_v_mins, 14, 2, 1, low(mins)).
table(low_c_mins, 14, 2, 1, low(mins)).
table(low_minc_mins, 14, 2, 1, low(mins)).
table(low_maxc_mins, 14, 2, 1, low(mins)).
table(low_s_mins, 14, 2, 1, low(mins)).
table(low_maxs_mins, 14, 2, 1, low(mins)).
table(low_v_c_mins, 14, 3, 2, low(mins)).
table(low_v_minc_mins, 14, 3, 2, low(mins)).
table(low_v_maxc_mins, 14, 3, 2, low(mins)).
table(low_v_s_mins, 14, 3, 2, low(mins)).
table(low_v_maxs_mins, 14, 3, 2, low(mins)).
table(low_c_minc_mins, 14, 3, 2, low(mins)).
table(low_c_maxc_mins, 14, 3, 2, low(mins)).
table(low_c_s_mins, 14, 3, 2, low(mins)).
table(low_c_maxs_mins, 14, 3, 2, low(mins)).
table(low_minc_maxc_mins, 14, 4, 2, low(mins)).
table(low_minc_s_mins, 14, 3, 2, low(mins)).
table(low_minc_maxs_mins, 14, 3, 2, low(mins)).
table(low_maxc_s_mins, 14, 3, 2, low(mins)).
table(low_maxc_maxs_mins, 14, 3, 2, low(mins)).
table(low_s_maxs_mins, 14, 3, 2, low(mins)).
table(low_v_c_minc_mins, 14, 6, 3, low(mins)).
table(low_v_c_maxc_mins, 14, 4, 3, low(mins)).
table(low_v_c_s_mins, 14, 4, 3, low(mins)).
table(low_v_c_maxs_mins, 14, 4, 3, low(mins)).
table(low_v_minc_maxc_mins, 14, 5, 3, low(mins)).
table(low_v_minc_s_mins, 14, 4, 3, low(mins)).
table(low_v_minc_maxs_mins, 14, 4, 3, low(mins)).
table(low_v_maxc_s_mins, 14, 4, 3, low(mins)).
table(low_v_maxc_maxs_mins, 14, 4, 3, low(mins)).
table(low_v_s_maxs_mins, 14, 4, 3, low(mins)).
table(low_c_minc_maxc_mins, 14, 5, 3, low(mins)).
table(low_c_minc_s_mins, 14, 4, 3, low(mins)).
table(low_c_minc_maxs_mins, 14, 4, 3, low(mins)).
table(low_c_maxc_s_mins, 14, 4, 3, low(mins)).
table(low_c_maxc_maxs_mins, 14, 4, 3, low(mins)).
table(low_c_s_maxs_mins, 14, 4, 3, low(mins)).
table(low_minc_maxc_s_mins, 14, 5, 3, low(mins)).
table(low_minc_maxc_maxs_mins, 14, 5, 3, low(mins)).
table(low_minc_s_maxs_mins, 14, 4, 3, low(mins)).
table(low_maxc_s_maxs_mins, 14, 4, 3, low(mins)).
table(low_v_c_minc_maxc_mins, 14, 7, 4, low(mins)).
table(low_v_c_minc_s_mins, 14, 7, 4, low(mins)).
table(low_v_c_minc_maxs_mins, 14, 7, 4, low(mins)).
table(low_v_c_maxc_s_mins, 14, 5, 4, low(mins)).
table(low_v_c_maxc_maxs_mins, 14, 5, 4, low(mins)).
table(low_v_c_s_maxs_mins, 14, 5, 4, low(mins)).
table(low_v_minc_maxc_s_mins, 14, 6, 4, low(mins)).
table(low_v_minc_maxc_maxs_mins, 14, 6, 4, low(mins)).
table(low_v_minc_s_maxs_mins, 14, 5, 4, low(mins)).
table(low_v_maxc_s_maxs_mins, 14, 5, 4, low(mins)).
table(low_c_minc_maxc_s_mins, 14, 6, 4, low(mins)).
table(low_c_minc_maxc_maxs_mins, 14, 6, 4, low(mins)).
table(low_c_minc_s_maxs_mins, 14, 5, 4, low(mins)).
table(low_c_maxc_s_maxs_mins, 14, 5, 4, low(mins)).
table(low_minc_maxc_s_maxs_mins, 14, 6, 4, low(mins)).
table(low_v_c_minc_maxc_s_mins, 14, 8, 5, low(mins)).
table(low_v_c_minc_maxc_maxs_mins, 14, 8, 5, low(mins)).
table(low_v_c_minc_s_maxs_mins, 14, 8, 5, low(mins)).
table(low_v_c_maxc_s_maxs_mins, 14, 6, 5, low(mins)).
table(low_v_minc_maxc_s_maxs_mins, 14, 7, 5, low(mins)).
table(low_c_minc_maxc_s_maxs_mins, 14, 7, 5, low(mins)).
table(low_v_c_minc_maxc_s_maxs_mins, 14, 9, 6, low(mins)).
table(up_v_mins, 14, 18, 1, up(mins)).
table(up_maxc_mins, 14, 9, 1, up(mins)).
table(up_maxs_mins, 14, 2, 1, up(mins)).
table(up_v_c_mins, 14, 6, 2, up(mins)).
table(up_v_minc_mins, 14, 3, 2, up(mins)).
table(up_v_maxc_mins, 14, 7, 2, up(mins)).
table(up_v_s_mins, 14, 3, 2, up(mins)).
table(up_v_maxs_mins, 14, 3, 2, up(mins)).
table(up_c_maxc_mins, 14, 5, 2, up(mins)).
table(up_c_maxs_mins, 14, 3, 2, up(mins)).
table(up_minc_maxc_mins, 14, 4, 2, up(mins)).
table(up_minc_maxs_mins, 14, 3, 2, up(mins)).
table(up_maxc_s_mins, 14, 3, 2, up(mins)).
table(up_maxc_maxs_mins, 14, 3, 2, up(mins)).
table(up_s_maxs_mins, 14, 3, 2, up(mins)).
table(up_v_c_minc_mins, 14, 6, 3, up(mins)).
table(up_v_c_maxc_mins, 14, 7, 3, up(mins)).
table(up_v_c_s_mins, 14, 4, 3, up(mins)).
table(up_v_c_maxs_mins, 14, 4, 3, up(mins)).
table(up_v_minc_maxc_mins, 14, 5, 3, up(mins)).
table(up_v_minc_s_mins, 14, 4, 3, up(mins)).
table(up_v_minc_maxs_mins, 14, 4, 3, up(mins)).
table(up_v_maxc_s_mins, 14, 4, 3, up(mins)).
table(up_v_maxc_maxs_mins, 14, 4, 3, up(mins)).
table(up_v_s_maxs_mins, 14, 4, 3, up(mins)).
table(up_c_minc_maxc_mins, 14, 5, 3, up(mins)).
table(up_c_minc_maxs_mins, 14, 4, 3, up(mins)).
table(up_c_maxc_s_mins, 14, 4, 3, up(mins)).
table(up_c_maxc_maxs_mins, 14, 4, 3, up(mins)).
table(up_c_s_maxs_mins, 14, 4, 3, up(mins)).
table(up_minc_maxc_s_mins, 14, 5, 3, up(mins)).
table(up_minc_maxc_maxs_mins, 14, 5, 3, up(mins)).
table(up_minc_s_maxs_mins, 14, 4, 3, up(mins)).
table(up_maxc_s_maxs_mins, 14, 4, 3, up(mins)).
table(up_v_c_minc_maxc_mins, 14, 7, 4, up(mins)).
table(up_v_c_minc_s_mins, 14, 7, 4, up(mins)).
table(up_v_c_minc_maxs_mins, 14, 7, 4, up(mins)).
table(up_v_c_maxc_s_mins, 14, 5, 4, up(mins)).
table(up_v_c_maxc_maxs_mins, 14, 5, 4, up(mins)).
table(up_v_c_s_maxs_mins, 14, 5, 4, up(mins)).
table(up_v_minc_maxc_s_mins, 14, 6, 4, up(mins)).
table(up_v_minc_maxc_maxs_mins, 14, 6, 4, up(mins)).
table(up_v_minc_s_maxs_mins, 14, 5, 4, up(mins)).
table(up_v_maxc_s_maxs_mins, 14, 5, 4, up(mins)).
table(up_c_minc_maxc_s_mins, 14, 6, 4, up(mins)).
table(up_c_minc_maxc_maxs_mins, 14, 6, 4, up(mins)).
table(up_c_minc_s_maxs_mins, 14, 5, 4, up(mins)).
table(up_c_maxc_s_maxs_mins, 14, 5, 4, up(mins)).
table(up_minc_maxc_s_maxs_mins, 14, 6, 4, up(mins)).
table(up_v_c_minc_maxc_s_mins, 14, 8, 5, up(mins)).
table(up_v_c_minc_maxc_maxs_mins, 14, 8, 5, up(mins)).
table(up_v_c_minc_s_maxs_mins, 14, 8, 5, up(mins)).
table(up_v_c_maxc_s_maxs_mins, 14, 6, 5, up(mins)).
table(up_v_minc_maxc_s_maxs_mins, 14, 7, 5, up(mins)).
table(up_c_minc_maxc_s_maxs_mins, 14, 7, 5, up(mins)).
table(up_v_c_minc_maxc_s_maxs_mins, 14, 9, 6, up(mins)).
table(low_v_maxs, 14, 7, 1, low(maxs)).
table(low_c_maxs, 14, 4, 1, low(maxs)).
table(low_minc_maxs, 14, 4, 1, low(maxs)).
table(low_maxc_maxs, 14, 4, 1, low(maxs)).
table(low_s_maxs, 14, 7, 1, low(maxs)).
table(low_mins_maxs, 14, 2, 1, low(maxs)).
table(low_v_c_maxs, 14, 8, 2, low(maxs)).
table(low_v_minc_maxs, 14, 8, 2, low(maxs)).
table(low_v_maxc_maxs, 14, 8, 2, low(maxs)).
table(low_v_s_maxs, 14, 3, 2, low(maxs)).
table(low_v_mins_maxs, 14, 3, 2, low(maxs)).
table(low_c_minc_maxs, 14, 5, 2, low(maxs)).
table(low_c_maxc_maxs, 14, 5, 2, low(maxs)).
table(low_c_s_maxs, 14, 8, 2, low(maxs)).
table(low_c_mins_maxs, 14, 3, 2, low(maxs)).
table(low_minc_maxc_maxs, 14, 6, 2, low(maxs)).
table(low_minc_s_maxs, 14, 3, 2, low(maxs)).
table(low_minc_mins_maxs, 14, 3, 2, low(maxs)).
table(low_maxc_s_maxs, 14, 3, 2, low(maxs)).
table(low_maxc_mins_maxs, 14, 3, 2, low(maxs)).
table(low_s_mins_maxs, 14, 7, 2, low(maxs)).
table(low_v_c_minc_maxs, 14, 11, 3, low(maxs)).
table(low_v_c_maxc_maxs, 14, 9, 3, low(maxs)).
table(low_v_c_s_maxs, 14, 4, 3, low(maxs)).
table(low_v_c_mins_maxs, 14, 4, 3, low(maxs)).
table(low_v_minc_maxc_maxs, 14, 10, 3, low(maxs)).
table(low_v_minc_s_maxs, 14, 4, 3, low(maxs)).
table(low_v_minc_mins_maxs, 14, 4, 3, low(maxs)).
table(low_v_maxc_s_maxs, 14, 4, 3, low(maxs)).
table(low_v_maxc_mins_maxs, 14, 4, 3, low(maxs)).
table(low_v_s_mins_maxs, 14, 4, 3, low(maxs)).
table(low_c_minc_maxc_maxs, 14, 7, 3, low(maxs)).
table(low_c_minc_s_maxs, 14, 4, 3, low(maxs)).
table(low_c_minc_mins_maxs, 14, 4, 3, low(maxs)).
table(low_c_maxc_s_maxs, 14, 4, 3, low(maxs)).
table(low_c_maxc_mins_maxs, 14, 4, 3, low(maxs)).
table(low_c_s_mins_maxs, 14, 8, 3, low(maxs)).
table(low_minc_maxc_s_maxs, 14, 5, 3, low(maxs)).
table(low_minc_maxc_mins_maxs, 14, 5, 3, low(maxs)).
table(low_minc_s_mins_maxs, 14, 4, 3, low(maxs)).
table(low_maxc_s_mins_maxs, 14, 4, 3, low(maxs)).
table(low_v_c_minc_maxc_maxs, 14, 12, 4, low(maxs)).
table(low_v_c_minc_s_maxs, 14, 7, 4, low(maxs)).
table(low_v_c_minc_mins_maxs, 14, 7, 4, low(maxs)).
table(low_v_c_maxc_s_maxs, 14, 5, 4, low(maxs)).
table(low_v_c_maxc_mins_maxs, 14, 5, 4, low(maxs)).
table(low_v_c_s_mins_maxs, 14, 5, 4, low(maxs)).
table(low_v_minc_maxc_s_maxs, 14, 6, 4, low(maxs)).
table(low_v_minc_maxc_mins_maxs, 14, 6, 4, low(maxs)).
table(low_v_minc_s_mins_maxs, 14, 5, 4, low(maxs)).
table(low_v_maxc_s_mins_maxs, 14, 5, 4, low(maxs)).
table(low_c_minc_maxc_s_maxs, 14, 6, 4, low(maxs)).
table(low_c_minc_maxc_mins_maxs, 14, 6, 4, low(maxs)).
table(low_c_minc_s_mins_maxs, 14, 5, 4, low(maxs)).
table(low_c_maxc_s_mins_maxs, 14, 5, 4, low(maxs)).
table(low_minc_maxc_s_mins_maxs, 14, 6, 4, low(maxs)).
table(low_v_c_minc_maxc_s_maxs, 14, 8, 5, low(maxs)).
table(low_v_c_minc_maxc_mins_maxs, 14, 8, 5, low(maxs)).
table(low_v_c_minc_s_mins_maxs, 14, 8, 5, low(maxs)).
table(low_v_c_maxc_s_mins_maxs, 14, 6, 5, low(maxs)).
table(low_v_minc_maxc_s_mins_maxs, 14, 7, 5, low(maxs)).
table(low_c_minc_maxc_s_mins_maxs, 14, 7, 5, low(maxs)).
table(low_v_c_minc_maxc_s_mins_maxs, 14, 9, 6, low(maxs)).
table(up_v_maxs, 14, 18, 1, up(maxs)).
table(up_maxc_maxs, 14, 2, 1, up(maxs)).
table(up_v_c_maxs, 14, 18, 2, up(maxs)).
table(up_v_minc_maxs, 14, 12, 2, up(maxs)).
table(up_v_maxc_maxs, 14, 3, 2, up(maxs)).
table(up_v_s_maxs, 14, 7, 2, up(maxs)).
table(up_v_mins_maxs, 14, 9, 2, up(maxs)).
table(up_c_maxc_maxs, 14, 3, 2, up(maxs)).
table(up_minc_maxc_maxs, 14, 4, 2, up(maxs)).
table(up_maxc_s_maxs, 14, 3, 2, up(maxs)).
table(up_maxc_mins_maxs, 14, 3, 2, up(maxs)).
table(up_v_c_minc_maxs, 14, 12, 3, up(maxs)).
table(up_v_c_maxc_maxs, 14, 4, 3, up(maxs)).
table(up_v_c_s_maxs, 14, 8, 3, up(maxs)).
table(up_v_c_mins_maxs, 14, 18, 3, up(maxs)).
table(up_v_minc_maxc_maxs, 14, 5, 3, up(maxs)).
table(up_v_minc_s_maxs, 14, 5, 3, up(maxs)).
table(up_v_minc_mins_maxs, 14, 14, 3, up(maxs)).
table(up_v_maxc_s_maxs, 14, 4, 3, up(maxs)).
table(up_v_maxc_mins_maxs, 14, 4, 3, up(maxs)).
table(up_v_s_mins_maxs, 14, 7, 3, up(maxs)).
table(up_c_minc_maxc_maxs, 14, 5, 3, up(maxs)).
table(up_c_maxc_s_maxs, 14, 4, 3, up(maxs)).
table(up_c_maxc_mins_maxs, 14, 4, 3, up(maxs)).
table(up_minc_maxc_s_maxs, 14, 5, 3, up(maxs)).
table(up_minc_maxc_mins_maxs, 14, 5, 3, up(maxs)).
table(up_maxc_s_mins_maxs, 14, 4, 3, up(maxs)).
table(up_v_c_minc_maxc_maxs, 14, 7, 4, up(maxs)).
table(up_v_c_minc_s_maxs, 14, 7, 4, up(maxs)).
table(up_v_c_minc_mins_maxs, 14, 13, 4, up(maxs)).
table(up_v_c_maxc_s_maxs, 14, 5, 4, up(maxs)).
table(up_v_c_maxc_mins_maxs, 14, 5, 4, up(maxs)).
table(up_v_c_s_mins_maxs, 14, 8, 4, up(maxs)).
table(up_v_minc_maxc_s_maxs, 14, 6, 4, up(maxs)).
table(up_v_minc_maxc_mins_maxs, 14, 6, 4, up(maxs)).
table(up_v_minc_s_mins_maxs, 14, 6, 4, up(maxs)).
table(up_v_maxc_s_mins_maxs, 14, 5, 4, up(maxs)).
table(up_c_minc_maxc_s_maxs, 14, 6, 4, up(maxs)).
table(up_c_minc_maxc_mins_maxs, 14, 6, 4, up(maxs)).
table(up_c_maxc_s_mins_maxs, 14, 5, 4, up(maxs)).
table(up_minc_maxc_s_mins_maxs, 14, 6, 4, up(maxs)).
table(up_v_c_minc_maxc_s_maxs, 14, 8, 5, up(maxs)).
table(up_v_c_minc_maxc_mins_maxs, 14, 8, 5, up(maxs)).
table(up_v_c_minc_s_mins_maxs, 14, 8, 5, up(maxs)).
table(up_v_c_maxc_s_mins_maxs, 14, 6, 5, up(maxs)).
table(up_v_minc_maxc_s_mins_maxs, 14, 7, 5, up(maxs)).
table(up_c_minc_maxc_s_mins_maxs, 14, 7, 5, up(maxs)).
table(up_v_c_minc_maxc_s_mins_maxs, 14, 9, 6, up(maxs)).
table(low_v_mina, 14, 12, 1, low(mina)).
table(low_c_mina, 14, 7, 1, low(mina)).
table(low_minc_mina, 14, 19, 1, low(mina)).
table(low_maxc_mina, 14, 19, 1, low(mina)).
table(low_s_mina, 14, 12, 1, low(mina)).
table(low_mins_mina, 14, 11, 1, low(mina)).
table(low_maxs_mina, 14, 11, 1, low(mina)).
table(low_v_c_mina, 14, 10, 2, low(mina)).
table(low_v_minc_mina, 14, 11, 2, low(mina)).
table(low_v_maxc_mina, 14, 11, 2, low(mina)).
table(low_v_s_mina, 14, 4, 2, low(mina)).
table(low_v_mins_mina, 14, 4, 2, low(mina)).
table(low_v_maxs_mina, 14, 10, 2, low(mina)).
table(low_c_minc_mina, 14, 8, 2, low(mina)).
table(low_c_maxc_mina, 14, 6, 2, low(mina)).
table(low_c_s_mina, 14, 10, 2, low(mina)).
table(low_c_mins_mina, 14, 7, 2, low(mina)).
table(low_c_maxs_mina, 14, 5, 2, low(mina)).
table(low_minc_maxc_mina, 14, 19, 2, low(mina)).
table(low_minc_s_mina, 14, 11, 2, low(mina)).
table(low_minc_mins_mina, 14, 19, 2, low(mina)).
table(low_minc_maxs_mina, 14, 19, 2, low(mina)).
table(low_maxc_s_mina, 14, 11, 2, low(mina)).
table(low_maxc_mins_mina, 14, 19, 2, low(mina)).
table(low_maxc_maxs_mina, 14, 19, 2, low(mina)).
table(low_s_mins_mina, 14, 12, 2, low(mina)).
table(low_s_maxs_mina, 14, 10, 2, low(mina)).
table(low_mins_maxs_mina, 14, 6, 2, low(mina)).
table(low_v_c_minc_mina, 14, 13, 3, low(mina)).
table(low_v_c_maxc_mina, 14, 11, 3, low(mina)).
table(low_v_c_s_mina, 14, 4, 3, low(mina)).
table(low_v_c_mins_mina, 14, 7, 3, low(mina)).
table(low_v_c_maxs_mina, 14, 10, 3, low(mina)).
table(low_v_minc_maxc_mina, 14, 13, 3, low(mina)).
table(low_v_minc_s_mina, 14, 4, 3, low(mina)).
table(low_v_minc_mins_mina, 14, 4, 3, low(mina)).
table(low_v_minc_maxs_mina, 14, 11, 3, low(mina)).
table(low_v_maxc_s_mina, 14, 4, 3, low(mina)).
table(low_v_maxc_mins_mina, 14, 6, 3, low(mina)).
table(low_v_maxc_maxs_mina, 14, 10, 3, low(mina)).
table(low_v_s_mins_mina, 14, 5, 3, low(mina)).
table(low_v_s_maxs_mina, 14, 5, 3, low(mina)).
table(low_v_mins_maxs_mina, 14, 5, 3, low(mina)).
table(low_c_minc_maxc_mina, 14, 8, 3, low(mina)).
table(low_c_minc_s_mina, 14, 11, 3, low(mina)).
table(low_c_minc_mins_mina, 14, 8, 3, low(mina)).
table(low_c_minc_maxs_mina, 14, 6, 3, low(mina)).
table(low_c_maxc_s_mina, 14, 8, 3, low(mina)).
table(low_c_maxc_mins_mina, 14, 6, 3, low(mina)).
table(low_c_maxc_maxs_mina, 14, 6, 3, low(mina)).
table(low_c_s_mins_mina, 14, 10, 3, low(mina)).
table(low_c_s_maxs_mina, 14, 10, 3, low(mina)).
table(low_c_mins_maxs_mina, 14, 5, 3, low(mina)).
table(low_minc_maxc_s_mina, 14, 10, 3, low(mina)).
table(low_minc_maxc_mins_mina, 14, 16, 3, low(mina)).
table(low_minc_maxc_maxs_mina, 14, 18, 3, low(mina)).
table(low_minc_s_mins_mina, 14, 6, 3, low(mina)).
table(low_minc_s_maxs_mina, 14, 8, 3, low(mina)).
table(low_minc_mins_maxs_mina, 14, 17, 3, low(mina)).
table(low_maxc_s_mins_mina, 14, 8, 3, low(mina)).
table(low_maxc_s_maxs_mina, 14, 7, 3, low(mina)).
table(low_maxc_mins_maxs_mina, 14, 8, 3, low(mina)).
table(low_s_mins_maxs_mina, 14, 10, 3, low(mina)).
table(low_v_c_minc_maxc_mina, 14, 14, 4, low(mina)).
table(low_v_c_minc_s_mina, 14, 8, 4, low(mina)).
table(low_v_c_minc_mins_mina, 14, 10, 4, low(mina)).
table(low_v_c_minc_maxs_mina, 14, 13, 4, low(mina)).
table(low_v_c_maxc_s_mina, 14, 5, 4, low(mina)).
table(low_v_c_maxc_mins_mina, 14, 8, 4, low(mina)).
table(low_v_c_maxc_maxs_mina, 14, 11, 4, low(mina)).
table(low_v_c_s_mins_mina, 14, 5, 4, low(mina)).
table(low_v_c_s_maxs_mina, 14, 5, 4, low(mina)).
table(low_v_c_mins_maxs_mina, 14, 8, 4, low(mina)).
table(low_v_minc_maxc_s_mina, 14, 7, 4, low(mina)).
table(low_v_minc_maxc_mins_mina, 14, 7, 4, low(mina)).
table(low_v_minc_maxc_maxs_mina, 14, 12, 4, low(mina)).
table(low_v_minc_s_mins_mina, 14, 5, 4, low(mina)).
table(low_v_minc_s_maxs_mina, 14, 6, 4, low(mina)).
table(low_v_minc_mins_maxs_mina, 14, 6, 4, low(mina)).
table(low_v_maxc_s_mins_mina, 14, 5, 4, low(mina)).
table(low_v_maxc_s_maxs_mina, 14, 5, 4, low(mina)).
table(low_v_maxc_mins_maxs_mina, 14, 5, 4, low(mina)).
table(low_v_s_mins_maxs_mina, 14, 6, 4, low(mina)).
table(low_c_minc_maxc_s_mina, 14, 9, 4, low(mina)).
table(low_c_minc_maxc_mins_mina, 14, 7, 4, low(mina)).
table(low_c_minc_maxc_maxs_mina, 14, 8, 4, low(mina)).
table(low_c_minc_s_mins_mina, 14, 10, 4, low(mina)).
table(low_c_minc_s_maxs_mina, 14, 10, 4, low(mina)).
table(low_c_minc_mins_maxs_mina, 14, 6, 4, low(mina)).
table(low_c_maxc_s_mins_mina, 14, 6, 4, low(mina)).
table(low_c_maxc_s_maxs_mina, 14, 6, 4, low(mina)).
table(low_c_maxc_mins_maxs_mina, 14, 6, 4, low(mina)).
table(low_c_s_mins_maxs_mina, 14, 10, 4, low(mina)).
table(low_minc_maxc_s_mins_mina, 14, 9, 4, low(mina)).
table(low_minc_maxc_s_maxs_mina, 14, 9, 4, low(mina)).
table(low_minc_maxc_mins_maxs_mina, 14, 16, 4, low(mina)).
table(low_minc_s_mins_maxs_mina, 14, 7, 4, low(mina)).
table(low_maxc_s_mins_maxs_mina, 14, 6, 4, low(mina)).
table(low_v_c_minc_maxc_s_mina, 14, 9, 5, low(mina)).
table(low_v_c_minc_maxc_mins_mina, 14, 11, 5, low(mina)).
table(low_v_c_minc_maxc_maxs_mina, 14, 14, 5, low(mina)).
table(low_v_c_minc_s_mins_mina, 14, 8, 5, low(mina)).
table(low_v_c_minc_s_maxs_mina, 14, 8, 5, low(mina)).
table(low_v_c_minc_mins_maxs_mina, 14, 11, 5, low(mina)).
table(low_v_c_maxc_s_mins_mina, 14, 6, 5, low(mina)).
table(low_v_c_maxc_s_maxs_mina, 14, 6, 5, low(mina)).
table(low_v_c_maxc_mins_maxs_mina, 14, 9, 5, low(mina)).
table(low_v_c_s_mins_maxs_mina, 14, 6, 5, low(mina)).
table(low_v_minc_maxc_s_mins_mina, 14, 7, 5, low(mina)).
table(low_v_minc_maxc_s_maxs_mina, 14, 7, 5, low(mina)).
table(low_v_minc_maxc_mins_maxs_mina, 14, 7, 5, low(mina)).
table(low_v_minc_s_mins_maxs_mina, 14, 7, 5, low(mina)).
table(low_v_maxc_s_mins_maxs_mina, 14, 6, 5, low(mina)).
table(low_c_minc_maxc_s_mins_mina, 14, 8, 5, low(mina)).
table(low_c_minc_maxc_s_maxs_mina, 14, 8, 5, low(mina)).
table(low_c_minc_maxc_mins_maxs_mina, 14, 8, 5, low(mina)).
table(low_c_minc_s_mins_maxs_mina, 14, 11, 5, low(mina)).
table(low_c_maxc_s_mins_maxs_mina, 14, 7, 5, low(mina)).
table(low_minc_maxc_s_mins_maxs_mina, 14, 8, 5, low(mina)).
table(low_v_c_minc_maxc_s_mins_mina, 14, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_maxs_mina, 14, 9, 6, low(mina)).
table(low_v_c_minc_maxc_mins_maxs_mina, 14, 12, 6, low(mina)).
table(low_v_c_minc_s_mins_maxs_mina, 14, 9, 6, low(mina)).
table(low_v_c_maxc_s_mins_maxs_mina, 14, 7, 6, low(mina)).
table(low_v_minc_maxc_s_mins_maxs_mina, 14, 8, 6, low(mina)).
table(low_c_minc_maxc_s_mins_maxs_mina, 14, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_mins_maxs_mina, 14, 10, 7, low(mina)).
table(up_v_maxa, 14, 19, 1, up(maxa)).
table(up_v_c_maxa, 14, 19, 2, up(maxa)).
table(up_v_minc_maxa, 14, 19, 2, up(maxa)).
table(up_v_maxc_maxa, 14, 19, 2, up(maxa)).
table(up_v_s_maxa, 14, 19, 2, up(maxa)).
table(up_v_mins_maxa, 14, 19, 2, up(maxa)).
table(up_v_maxs_maxa, 14, 19, 2, up(maxa)).
table(up_c_maxc_maxa, 14, 19, 2, up(maxa)).
table(up_maxc_s_maxa, 14, 18, 2, up(maxa)).
table(up_s_maxs_maxa, 14, 19, 2, up(maxa)).
table(up_v_c_minc_maxa, 14, 19, 3, up(maxa)).
table(up_v_c_maxc_maxa, 14, 19, 3, up(maxa)).
table(up_v_c_s_maxa, 14, 19, 3, up(maxa)).
table(up_v_c_mins_maxa, 14, 19, 3, up(maxa)).
table(up_v_c_maxs_maxa, 14, 7, 3, up(maxa)).
table(up_v_minc_maxc_maxa, 14, 19, 3, up(maxa)).
table(up_v_minc_s_maxa, 14, 18, 3, up(maxa)).
table(up_v_minc_mins_maxa, 14, 19, 3, up(maxa)).
table(up_v_minc_maxs_maxa, 14, 19, 3, up(maxa)).
table(up_v_maxc_s_maxa, 14, 18, 3, up(maxa)).
table(up_v_maxc_mins_maxa, 14, 13, 3, up(maxa)).
table(up_v_maxc_maxs_maxa, 14, 7, 3, up(maxa)).
table(up_v_s_mins_maxa, 14, 19, 3, up(maxa)).
table(up_v_s_maxs_maxa, 14, 19, 3, up(maxa)).
table(up_v_mins_maxs_maxa, 14, 19, 3, up(maxa)).
table(up_c_minc_maxc_maxa, 14, 19, 3, up(maxa)).
table(up_c_maxc_s_maxa, 14, 18, 3, up(maxa)).
table(up_c_maxc_mins_maxa, 14, 19, 3, up(maxa)).
table(up_c_maxc_maxs_maxa, 14, 6, 3, up(maxa)).
table(up_c_s_maxs_maxa, 14, 19, 3, up(maxa)).
table(up_minc_maxc_s_maxa, 14, 11, 3, up(maxa)).
table(up_minc_s_maxs_maxa, 14, 18, 3, up(maxa)).
table(up_maxc_s_mins_maxa, 14, 18, 3, up(maxa)).
table(up_maxc_s_maxs_maxa, 14, 18, 3, up(maxa)).
table(up_s_mins_maxs_maxa, 14, 19, 3, up(maxa)).
table(up_v_c_minc_maxc_maxa, 14, 19, 4, up(maxa)).
table(up_v_c_minc_s_maxa, 14, 18, 4, up(maxa)).
table(up_v_c_minc_mins_maxa, 14, 11, 4, up(maxa)).
table(up_v_c_minc_maxs_maxa, 14, 8, 4, up(maxa)).
table(up_v_c_maxc_s_maxa, 14, 18, 4, up(maxa)).
table(up_v_c_maxc_mins_maxa, 14, 13, 4, up(maxa)).
table(up_v_c_maxc_maxs_maxa, 14, 6, 4, up(maxa)).
table(up_v_c_s_mins_maxa, 14, 19, 4, up(maxa)).
table(up_v_c_s_maxs_maxa, 14, 19, 4, up(maxa)).
table(up_v_c_mins_maxs_maxa, 14, 11, 4, up(maxa)).
table(up_v_minc_maxc_s_maxa, 14, 18, 4, up(maxa)).
table(up_v_minc_maxc_mins_maxa, 14, 19, 4, up(maxa)).
table(up_v_minc_maxc_maxs_maxa, 14, 10, 4, up(maxa)).
table(up_v_minc_s_mins_maxa, 14, 18, 4, up(maxa)).
table(up_v_minc_s_maxs_maxa, 14, 18, 4, up(maxa)).
table(up_v_minc_mins_maxs_maxa, 14, 18, 4, up(maxa)).
table(up_v_maxc_s_mins_maxa, 14, 12, 4, up(maxa)).
table(up_v_maxc_s_maxs_maxa, 14, 18, 4, up(maxa)).
table(up_v_maxc_mins_maxs_maxa, 14, 6, 4, up(maxa)).
table(up_v_s_mins_maxs_maxa, 14, 19, 4, up(maxa)).
table(up_c_minc_maxc_s_maxa, 14, 18, 4, up(maxa)).
table(up_c_minc_maxc_mins_maxa, 14, 19, 4, up(maxa)).
table(up_c_minc_maxc_maxs_maxa, 14, 10, 4, up(maxa)).
table(up_c_minc_s_maxs_maxa, 14, 12, 4, up(maxa)).
table(up_c_maxc_s_mins_maxa, 14, 18, 4, up(maxa)).
table(up_c_maxc_s_maxs_maxa, 14, 10, 4, up(maxa)).
table(up_c_maxc_mins_maxs_maxa, 14, 6, 4, up(maxa)).
table(up_c_s_mins_maxs_maxa, 14, 19, 4, up(maxa)).
table(up_minc_maxc_s_mins_maxa, 14, 11, 4, up(maxa)).
table(up_minc_maxc_s_maxs_maxa, 14, 8, 4, up(maxa)).
table(up_minc_s_mins_maxs_maxa, 14, 18, 4, up(maxa)).
table(up_maxc_s_mins_maxs_maxa, 14, 18, 4, up(maxa)).
table(up_v_c_minc_maxc_s_maxa, 14, 18, 5, up(maxa)).
table(up_v_c_minc_maxc_mins_maxa, 14, 19, 5, up(maxa)).
table(up_v_c_minc_maxc_maxs_maxa, 14, 10, 5, up(maxa)).
table(up_v_c_minc_s_mins_maxa, 14, 10, 5, up(maxa)).
table(up_v_c_minc_s_maxs_maxa, 14, 9, 5, up(maxa)).
table(up_v_c_minc_mins_maxs_maxa, 14, 8, 5, up(maxa)).
table(up_v_c_maxc_s_mins_maxa, 14, 12, 5, up(maxa)).
table(up_v_c_maxc_s_maxs_maxa, 14, 10, 5, up(maxa)).
table(up_v_c_maxc_mins_maxs_maxa, 14, 6, 5, up(maxa)).
table(up_v_c_s_mins_maxs_maxa, 14, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_mins_maxa, 14, 18, 5, up(maxa)).
table(up_v_minc_maxc_s_maxs_maxa, 14, 11, 5, up(maxa)).
table(up_v_minc_maxc_mins_maxs_maxa, 14, 9, 5, up(maxa)).
table(up_v_minc_s_mins_maxs_maxa, 14, 18, 5, up(maxa)).
table(up_v_maxc_s_mins_maxs_maxa, 14, 12, 5, up(maxa)).
table(up_c_minc_maxc_s_mins_maxa, 14, 18, 5, up(maxa)).
table(up_c_minc_maxc_s_maxs_maxa, 14, 11, 5, up(maxa)).
table(up_c_minc_maxc_mins_maxs_maxa, 14, 9, 5, up(maxa)).
table(up_c_minc_s_mins_maxs_maxa, 14, 10, 5, up(maxa)).
table(up_c_maxc_s_mins_maxs_maxa, 14, 10, 5, up(maxa)).
table(up_minc_maxc_s_mins_maxs_maxa, 14, 7, 5, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxa, 14, 18, 6, up(maxa)).
table(up_v_c_minc_maxc_s_maxs_maxa, 14, 11, 6, up(maxa)).
table(up_v_c_minc_maxc_mins_maxs_maxa, 14, 9, 6, up(maxa)).
table(up_v_c_minc_s_mins_maxs_maxa, 14, 10, 6, up(maxa)).
table(up_v_c_maxc_s_mins_maxs_maxa, 14, 9, 6, up(maxa)).
table(up_v_minc_maxc_s_mins_maxs_maxa, 14, 11, 6, up(maxa)).
table(up_c_minc_maxc_s_mins_maxs_maxa, 14, 11, 6, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxs_maxa, 14, 11, 7, up(maxa)).
table(low_c_v, 15, 18, 1, low(v)).
table(low_minc_v, 15, 11, 1, low(v)).
table(low_maxc_v, 15, 11, 1, low(v)).
table(low_s_v, 15, 7, 1, low(v)).
table(low_mins_v, 15, 18, 1, low(v)).
table(low_maxs_v, 15, 18, 1, low(v)).
table(low_c_minc_v, 15, 11, 2, low(v)).
table(low_c_maxc_v, 15, 11, 2, low(v)).
table(low_c_s_v, 15, 8, 2, low(v)).
table(low_c_mins_v, 15, 18, 2, low(v)).
table(low_c_maxs_v, 15, 18, 2, low(v)).
table(low_minc_maxc_v, 15, 11, 2, low(v)).
table(low_minc_s_v, 15, 3, 2, low(v)).
table(low_minc_mins_v, 15, 12, 2, low(v)).
table(low_minc_maxs_v, 15, 12, 2, low(v)).
table(low_maxc_s_v, 15, 3, 2, low(v)).
table(low_maxc_mins_v, 15, 12, 2, low(v)).
table(low_maxc_maxs_v, 15, 13, 2, low(v)).
table(low_s_mins_v, 15, 7, 2, low(v)).
table(low_s_maxs_v, 15, 7, 2, low(v)).
table(low_mins_maxs_v, 15, 12, 2, low(v)).
table(low_c_minc_maxc_v, 15, 11, 3, low(v)).
table(low_c_minc_s_v, 15, 6, 3, low(v)).
table(low_c_minc_mins_v, 15, 12, 3, low(v)).
table(low_c_minc_maxs_v, 15, 12, 3, low(v)).
table(low_c_maxc_s_v, 15, 4, 3, low(v)).
table(low_c_maxc_mins_v, 15, 12, 3, low(v)).
table(low_c_maxc_maxs_v, 15, 13, 3, low(v)).
table(low_c_s_mins_v, 15, 8, 3, low(v)).
table(low_c_s_maxs_v, 15, 8, 3, low(v)).
table(low_c_mins_maxs_v, 15, 18, 3, low(v)).
table(low_minc_maxc_s_v, 15, 5, 3, low(v)).
table(low_minc_maxc_mins_v, 15, 12, 3, low(v)).
table(low_minc_maxc_maxs_v, 15, 12, 3, low(v)).
table(low_minc_s_mins_v, 15, 4, 3, low(v)).
table(low_minc_s_maxs_v, 15, 4, 3, low(v)).
table(low_minc_mins_maxs_v, 15, 13, 3, low(v)).
table(low_maxc_s_mins_v, 15, 4, 3, low(v)).
table(low_maxc_s_maxs_v, 15, 4, 3, low(v)).
table(low_maxc_mins_maxs_v, 15, 14, 3, low(v)).
table(low_s_mins_maxs_v, 15, 7, 3, low(v)).
table(low_c_minc_maxc_s_v, 15, 7, 4, low(v)).
table(low_c_minc_maxc_mins_v, 15, 12, 4, low(v)).
table(low_c_minc_maxc_maxs_v, 15, 12, 4, low(v)).
table(low_c_minc_s_mins_v, 15, 7, 4, low(v)).
table(low_c_minc_s_maxs_v, 15, 7, 4, low(v)).
table(low_c_minc_mins_maxs_v, 15, 13, 4, low(v)).
table(low_c_maxc_s_mins_v, 15, 5, 4, low(v)).
table(low_c_maxc_s_maxs_v, 15, 5, 4, low(v)).
table(low_c_maxc_mins_maxs_v, 15, 14, 4, low(v)).
table(low_c_s_mins_maxs_v, 15, 8, 4, low(v)).
table(low_minc_maxc_s_mins_v, 15, 6, 4, low(v)).
table(low_minc_maxc_s_maxs_v, 15, 6, 4, low(v)).
table(low_minc_maxc_mins_maxs_v, 15, 13, 4, low(v)).
table(low_minc_s_mins_maxs_v, 15, 5, 4, low(v)).
table(low_maxc_s_mins_maxs_v, 15, 5, 4, low(v)).
table(low_c_minc_maxc_s_mins_v, 15, 8, 5, low(v)).
table(low_c_minc_maxc_s_maxs_v, 15, 8, 5, low(v)).
table(low_c_minc_maxc_mins_maxs_v, 15, 13, 5, low(v)).
table(low_c_minc_s_mins_maxs_v, 15, 8, 5, low(v)).
table(low_c_maxc_s_mins_maxs_v, 15, 6, 5, low(v)).
table(low_minc_maxc_s_mins_maxs_v, 15, 7, 5, low(v)).
table(low_c_minc_maxc_s_mins_maxs_v, 15, 9, 6, low(v)).
table(up_c_maxc_v, 15, 3, 2, up(v)).
table(up_maxc_s_v, 15, 3, 2, up(v)).
table(up_s_maxs_v, 15, 3, 2, up(v)).
table(up_c_minc_maxc_v, 15, 6, 3, up(v)).
table(up_c_maxc_s_v, 15, 4, 3, up(v)).
table(up_c_maxc_mins_v, 15, 4, 3, up(v)).
table(up_c_maxc_maxs_v, 15, 4, 3, up(v)).
table(up_c_s_maxs_v, 15, 4, 3, up(v)).
table(up_minc_maxc_s_v, 15, 5, 3, up(v)).
table(up_minc_s_maxs_v, 15, 4, 3, up(v)).
table(up_maxc_s_mins_v, 15, 4, 3, up(v)).
table(up_maxc_s_maxs_v, 15, 4, 3, up(v)).
table(up_s_mins_maxs_v, 15, 4, 3, up(v)).
table(up_c_minc_maxc_s_v, 15, 7, 4, up(v)).
table(up_c_minc_maxc_mins_v, 15, 7, 4, up(v)).
table(up_c_minc_maxc_maxs_v, 15, 7, 4, up(v)).
table(up_c_minc_s_maxs_v, 15, 7, 4, up(v)).
table(up_c_maxc_s_mins_v, 15, 5, 4, up(v)).
table(up_c_maxc_s_maxs_v, 15, 5, 4, up(v)).
table(up_c_maxc_mins_maxs_v, 15, 5, 4, up(v)).
table(up_c_s_mins_maxs_v, 15, 5, 4, up(v)).
table(up_minc_maxc_s_mins_v, 15, 6, 4, up(v)).
table(up_minc_maxc_s_maxs_v, 15, 6, 4, up(v)).
table(up_minc_s_mins_maxs_v, 15, 5, 4, up(v)).
table(up_maxc_s_mins_maxs_v, 15, 5, 4, up(v)).
table(up_c_minc_maxc_s_mins_v, 15, 8, 5, up(v)).
table(up_c_minc_maxc_s_maxs_v, 15, 8, 5, up(v)).
table(up_c_minc_maxc_mins_maxs_v, 15, 8, 5, up(v)).
table(up_c_minc_s_mins_maxs_v, 15, 8, 5, up(v)).
table(up_c_maxc_s_mins_maxs_v, 15, 6, 5, up(v)).
table(up_minc_maxc_s_mins_maxs_v, 15, 7, 5, up(v)).
table(up_c_minc_maxc_s_mins_maxs_v, 15, 9, 6, up(v)).
table(low_v_c, 15, 11, 1, low(c)).
table(low_minc_c, 15, 11, 1, low(c)).
table(low_maxc_c, 15, 11, 1, low(c)).
table(low_s_c, 15, 8, 1, low(c)).
table(low_mins_c, 15, 8, 1, low(c)).
table(low_maxs_c, 15, 8, 1, low(c)).
table(low_v_minc_c, 15, 11, 2, low(c)).
table(low_v_maxc_c, 15, 3, 2, low(c)).
table(low_v_s_c, 15, 13, 2, low(c)).
table(low_v_mins_c, 15, 12, 2, low(c)).
table(low_v_maxs_c, 15, 13, 2, low(c)).
table(low_minc_maxc_c, 15, 11, 2, low(c)).
table(low_minc_s_c, 15, 5, 2, low(c)).
table(low_minc_mins_c, 15, 9, 2, low(c)).
table(low_minc_maxs_c, 15, 9, 2, low(c)).
table(low_maxc_s_c, 15, 3, 2, low(c)).
table(low_maxc_mins_c, 15, 12, 2, low(c)).
table(low_maxc_maxs_c, 15, 13, 2, low(c)).
table(low_s_mins_c, 15, 10, 2, low(c)).
table(low_s_maxs_c, 15, 11, 2, low(c)).
table(low_mins_maxs_c, 15, 9, 2, low(c)).
table(low_v_minc_maxc_c, 15, 6, 3, low(c)).
table(low_v_minc_s_c, 15, 12, 3, low(c)).
table(low_v_minc_mins_c, 15, 12, 3, low(c)).
table(low_v_minc_maxs_c, 15, 12, 3, low(c)).
table(low_v_maxc_s_c, 15, 4, 3, low(c)).
table(low_v_maxc_mins_c, 15, 4, 3, low(c)).
table(low_v_maxc_maxs_c, 15, 4, 3, low(c)).
table(low_v_s_mins_c, 15, 14, 3, low(c)).
table(low_v_s_maxs_c, 15, 14, 3, low(c)).
table(low_v_mins_maxs_c, 15, 14, 3, low(c)).
table(low_minc_maxc_s_c, 15, 5, 3, low(c)).
table(low_minc_maxc_mins_c, 15, 12, 3, low(c)).
table(low_minc_maxc_maxs_c, 15, 12, 3, low(c)).
table(low_minc_s_mins_c, 15, 6, 3, low(c)).
table(low_minc_s_maxs_c, 15, 7, 3, low(c)).
table(low_minc_mins_maxs_c, 15, 7, 3, low(c)).
table(low_maxc_s_mins_c, 15, 4, 3, low(c)).
table(low_maxc_s_maxs_c, 15, 4, 3, low(c)).
table(low_maxc_mins_maxs_c, 15, 6, 3, low(c)).
table(low_s_mins_maxs_c, 15, 12, 3, low(c)).
table(low_v_minc_maxc_s_c, 15, 7, 4, low(c)).
table(low_v_minc_maxc_mins_c, 15, 7, 4, low(c)).
table(low_v_minc_maxc_maxs_c, 15, 7, 4, low(c)).
table(low_v_minc_s_mins_c, 15, 13, 4, low(c)).
table(low_v_minc_s_maxs_c, 15, 13, 4, low(c)).
table(low_v_minc_mins_maxs_c, 15, 13, 4, low(c)).
table(low_v_maxc_s_mins_c, 15, 5, 4, low(c)).
table(low_v_maxc_s_maxs_c, 15, 5, 4, low(c)).
table(low_v_maxc_mins_maxs_c, 15, 5, 4, low(c)).
table(low_v_s_mins_maxs_c, 15, 15, 4, low(c)).
table(low_minc_maxc_s_mins_c, 15, 6, 4, low(c)).
table(low_minc_maxc_s_maxs_c, 15, 6, 4, low(c)).
table(low_minc_maxc_mins_maxs_c, 15, 8, 4, low(c)).
table(low_minc_s_mins_maxs_c, 15, 8, 4, low(c)).
table(low_maxc_s_mins_maxs_c, 15, 5, 4, low(c)).
table(low_v_minc_maxc_s_mins_c, 15, 8, 5, low(c)).
table(low_v_minc_maxc_s_maxs_c, 15, 8, 5, low(c)).
table(low_v_minc_maxc_mins_maxs_c, 15, 8, 5, low(c)).
table(low_v_minc_s_mins_maxs_c, 15, 14, 5, low(c)).
table(low_v_maxc_s_mins_maxs_c, 15, 6, 5, low(c)).
table(low_minc_maxc_s_mins_maxs_c, 15, 7, 5, low(c)).
table(low_v_minc_maxc_s_mins_maxs_c, 15, 9, 6, low(c)).
table(up_v_c, 15, 18, 1, up(c)).
table(up_s_c, 15, 3, 1, up(c)).
table(up_v_minc_c, 15, 6, 2, up(c)).
table(up_v_maxc_c, 15, 11, 2, up(c)).
table(up_v_s_c, 15, 4, 2, up(c)).
table(up_v_mins_c, 15, 10, 2, up(c)).
table(up_v_maxs_c, 15, 18, 2, up(c)).
table(up_minc_s_c, 15, 3, 2, up(c)).
table(up_maxc_s_c, 15, 4, 2, up(c)).
table(up_s_mins_c, 15, 5, 2, up(c)).
table(up_s_maxs_c, 15, 5, 2, up(c)).
table(up_v_minc_maxc_c, 15, 11, 3, up(c)).
table(up_v_minc_s_c, 15, 6, 3, up(c)).
table(up_v_minc_mins_c, 15, 7, 3, up(c)).
table(up_v_minc_maxs_c, 15, 7, 3, up(c)).
table(up_v_maxc_s_c, 15, 5, 3, up(c)).
table(up_v_maxc_mins_c, 15, 6, 3, up(c)).
table(up_v_maxc_maxs_c, 15, 13, 3, up(c)).
table(up_v_s_mins_c, 15, 8, 3, up(c)).
table(up_v_s_maxs_c, 15, 6, 3, up(c)).
table(up_v_mins_maxs_c, 15, 18, 3, up(c)).
table(up_minc_maxc_s_c, 15, 5, 3, up(c)).
table(up_minc_s_mins_c, 15, 4, 3, up(c)).
table(up_minc_s_maxs_c, 15, 4, 3, up(c)).
table(up_maxc_s_mins_c, 15, 5, 3, up(c)).
table(up_maxc_s_maxs_c, 15, 5, 3, up(c)).
table(up_s_mins_maxs_c, 15, 8, 3, up(c)).
table(up_v_minc_maxc_s_c, 15, 7, 4, up(c)).
table(up_v_minc_maxc_mins_c, 15, 12, 4, up(c)).
table(up_v_minc_maxc_maxs_c, 15, 12, 4, up(c)).
table(up_v_minc_s_mins_c, 15, 7, 4, up(c)).
table(up_v_minc_s_maxs_c, 15, 7, 4, up(c)).
table(up_v_minc_mins_maxs_c, 15, 8, 4, up(c)).
table(up_v_maxc_s_mins_c, 15, 6, 4, up(c)).
table(up_v_maxc_s_maxs_c, 15, 6, 4, up(c)).
table(up_v_maxc_mins_maxs_c, 15, 14, 4, up(c)).
table(up_v_s_mins_maxs_c, 15, 10, 4, up(c)).
table(up_minc_maxc_s_mins_c, 15, 6, 4, up(c)).
table(up_minc_maxc_s_maxs_c, 15, 6, 4, up(c)).
table(up_minc_s_mins_maxs_c, 15, 5, 4, up(c)).
table(up_maxc_s_mins_maxs_c, 15, 6, 4, up(c)).
table(up_v_minc_maxc_s_mins_c, 15, 8, 5, up(c)).
table(up_v_minc_maxc_s_maxs_c, 15, 8, 5, up(c)).
table(up_v_minc_maxc_mins_maxs_c, 15, 13, 5, up(c)).
table(up_v_minc_s_mins_maxs_c, 15, 8, 5, up(c)).
table(up_v_maxc_s_mins_maxs_c, 15, 7, 5, up(c)).
table(up_minc_maxc_s_mins_maxs_c, 15, 7, 5, up(c)).
table(up_v_minc_maxc_s_mins_maxs_c, 15, 9, 6, up(c)).
table(low_v_minc, 15, 3, 1, low(minc)).
table(low_c_minc, 15, 3, 1, low(minc)).
table(low_maxc_minc, 15, 3, 1, low(minc)).
table(low_s_minc, 15, 3, 1, low(minc)).
table(low_mins_minc, 15, 2, 1, low(minc)).
table(low_maxs_minc, 15, 3, 1, low(minc)).
table(low_v_c_minc, 15, 5, 2, low(minc)).
table(low_v_maxc_minc, 15, 4, 2, low(minc)).
table(low_v_s_minc, 15, 4, 2, low(minc)).
table(low_v_mins_minc, 15, 3, 2, low(minc)).
table(low_v_maxs_minc, 15, 4, 2, low(minc)).
table(low_c_maxc_minc, 15, 4, 2, low(minc)).
table(low_c_s_minc, 15, 4, 2, low(minc)).
table(low_c_mins_minc, 15, 3, 2, low(minc)).
table(low_c_maxs_minc, 15, 4, 2, low(minc)).
table(low_maxc_s_minc, 15, 4, 2, low(minc)).
table(low_maxc_mins_minc, 15, 4, 2, low(minc)).
table(low_maxc_maxs_minc, 15, 4, 2, low(minc)).
table(low_s_mins_minc, 15, 3, 2, low(minc)).
table(low_s_maxs_minc, 15, 4, 2, low(minc)).
table(low_mins_maxs_minc, 15, 3, 2, low(minc)).
table(low_v_c_maxc_minc, 15, 6, 3, low(minc)).
table(low_v_c_s_minc, 15, 6, 3, low(minc)).
table(low_v_c_mins_minc, 15, 6, 3, low(minc)).
table(low_v_c_maxs_minc, 15, 6, 3, low(minc)).
table(low_v_maxc_s_minc, 15, 5, 3, low(minc)).
table(low_v_maxc_mins_minc, 15, 5, 3, low(minc)).
table(low_v_maxc_maxs_minc, 15, 5, 3, low(minc)).
table(low_v_s_mins_minc, 15, 4, 3, low(minc)).
table(low_v_s_maxs_minc, 15, 5, 3, low(minc)).
table(low_v_mins_maxs_minc, 15, 4, 3, low(minc)).
table(low_c_maxc_s_minc, 15, 5, 3, low(minc)).
table(low_c_maxc_mins_minc, 15, 5, 3, low(minc)).
table(low_c_maxc_maxs_minc, 15, 5, 3, low(minc)).
table(low_c_s_mins_minc, 15, 4, 3, low(minc)).
table(low_c_s_maxs_minc, 15, 6, 3, low(minc)).
table(low_c_mins_maxs_minc, 15, 4, 3, low(minc)).
table(low_maxc_s_mins_minc, 15, 5, 3, low(minc)).
table(low_maxc_s_maxs_minc, 15, 5, 3, low(minc)).
table(low_maxc_mins_maxs_minc, 15, 5, 3, low(minc)).
table(low_s_mins_maxs_minc, 15, 4, 3, low(minc)).
table(low_v_c_maxc_s_minc, 15, 7, 4, low(minc)).
table(low_v_c_maxc_mins_minc, 15, 7, 4, low(minc)).
table(low_v_c_maxc_maxs_minc, 15, 7, 4, low(minc)).
table(low_v_c_s_mins_minc, 15, 7, 4, low(minc)).
table(low_v_c_s_maxs_minc, 15, 7, 4, low(minc)).
table(low_v_c_mins_maxs_minc, 15, 7, 4, low(minc)).
table(low_v_maxc_s_mins_minc, 15, 6, 4, low(minc)).
table(low_v_maxc_s_maxs_minc, 15, 6, 4, low(minc)).
table(low_v_maxc_mins_maxs_minc, 15, 6, 4, low(minc)).
table(low_v_s_mins_maxs_minc, 15, 5, 4, low(minc)).
table(low_c_maxc_s_mins_minc, 15, 6, 4, low(minc)).
table(low_c_maxc_s_maxs_minc, 15, 6, 4, low(minc)).
table(low_c_maxc_mins_maxs_minc, 15, 6, 4, low(minc)).
table(low_c_s_mins_maxs_minc, 15, 6, 4, low(minc)).
table(low_maxc_s_mins_maxs_minc, 15, 6, 4, low(minc)).
table(low_v_c_maxc_s_mins_minc, 15, 8, 5, low(minc)).
table(low_v_c_maxc_s_maxs_minc, 15, 8, 5, low(minc)).
table(low_v_c_maxc_mins_maxs_minc, 15, 8, 5, low(minc)).
table(low_v_c_s_mins_maxs_minc, 15, 8, 5, low(minc)).
table(low_v_maxc_s_mins_maxs_minc, 15, 7, 5, low(minc)).
table(low_c_maxc_s_mins_maxs_minc, 15, 7, 5, low(minc)).
table(low_v_c_maxc_s_mins_maxs_minc, 15, 9, 6, low(minc)).
table(up_v_minc, 15, 11, 1, up(minc)).
table(up_maxc_minc, 15, 6, 1, up(minc)).
table(up_v_c_minc, 15, 5, 2, up(minc)).
table(up_v_maxc_minc, 15, 6, 2, up(minc)).
table(up_v_s_minc, 15, 13, 2, up(minc)).
table(up_v_mins_minc, 15, 12, 2, up(minc)).
table(up_v_maxs_minc, 15, 13, 2, up(minc)).
table(up_c_maxc_minc, 15, 4, 2, up(minc)).
table(up_maxc_s_minc, 15, 4, 2, up(minc)).
table(up_maxc_mins_minc, 15, 4, 2, up(minc)).
table(up_maxc_maxs_minc, 15, 7, 2, up(minc)).
table(up_s_maxs_minc, 15, 14, 2, up(minc)).
table(up_v_c_maxc_minc, 15, 6, 3, up(minc)).
table(up_v_c_s_minc, 15, 6, 3, up(minc)).
table(up_v_c_mins_minc, 15, 6, 3, up(minc)).
table(up_v_c_maxs_minc, 15, 6, 3, up(minc)).
table(up_v_maxc_s_minc, 15, 7, 3, up(minc)).
table(up_v_maxc_mins_minc, 15, 7, 3, up(minc)).
table(up_v_maxc_maxs_minc, 15, 7, 3, up(minc)).
table(up_v_s_mins_minc, 15, 14, 3, up(minc)).
table(up_v_s_maxs_minc, 15, 14, 3, up(minc)).
table(up_v_mins_maxs_minc, 15, 14, 3, up(minc)).
table(up_c_maxc_s_minc, 15, 5, 3, up(minc)).
table(up_c_maxc_mins_minc, 15, 5, 3, up(minc)).
table(up_c_maxc_maxs_minc, 15, 5, 3, up(minc)).
table(up_c_s_maxs_minc, 15, 4, 3, up(minc)).
table(up_maxc_s_mins_minc, 15, 5, 3, up(minc)).
table(up_maxc_s_maxs_minc, 15, 5, 3, up(minc)).
table(up_maxc_mins_maxs_minc, 15, 5, 3, up(minc)).
table(up_s_mins_maxs_minc, 15, 15, 3, up(minc)).
table(up_v_c_maxc_s_minc, 15, 7, 4, up(minc)).
table(up_v_c_maxc_mins_minc, 15, 7, 4, up(minc)).
table(up_v_c_maxc_maxs_minc, 15, 7, 4, up(minc)).
table(up_v_c_s_mins_minc, 15, 7, 4, up(minc)).
table(up_v_c_s_maxs_minc, 15, 7, 4, up(minc)).
table(up_v_c_mins_maxs_minc, 15, 7, 4, up(minc)).
table(up_v_maxc_s_mins_minc, 15, 6, 4, up(minc)).
table(up_v_maxc_s_maxs_minc, 15, 6, 4, up(minc)).
table(up_v_maxc_mins_maxs_minc, 15, 6, 4, up(minc)).
table(up_v_s_mins_maxs_minc, 15, 15, 4, up(minc)).
table(up_c_maxc_s_mins_minc, 15, 6, 4, up(minc)).
table(up_c_maxc_s_maxs_minc, 15, 6, 4, up(minc)).
table(up_c_maxc_mins_maxs_minc, 15, 6, 4, up(minc)).
table(up_c_s_mins_maxs_minc, 15, 5, 4, up(minc)).
table(up_maxc_s_mins_maxs_minc, 15, 6, 4, up(minc)).
table(up_v_c_maxc_s_mins_minc, 15, 8, 5, up(minc)).
table(up_v_c_maxc_s_maxs_minc, 15, 8, 5, up(minc)).
table(up_v_c_maxc_mins_maxs_minc, 15, 8, 5, up(minc)).
table(up_v_c_s_mins_maxs_minc, 15, 8, 5, up(minc)).
table(up_v_maxc_s_mins_maxs_minc, 15, 7, 5, up(minc)).
table(up_c_maxc_s_mins_maxs_minc, 15, 7, 5, up(minc)).
table(up_v_c_maxc_s_mins_maxs_minc, 15, 9, 6, up(minc)).
table(low_v_maxc, 15, 18, 1, low(maxc)).
table(low_c_maxc, 15, 18, 1, low(maxc)).
table(low_minc_maxc, 15, 6, 1, low(maxc)).
table(low_s_maxc, 15, 18, 1, low(maxc)).
table(low_mins_maxc, 15, 9, 1, low(maxc)).
table(low_maxs_maxc, 15, 2, 1, low(maxc)).
table(low_v_c_maxc, 15, 3, 2, low(maxc)).
table(low_v_minc_maxc, 15, 6, 2, low(maxc)).
table(low_v_s_maxc, 15, 4, 2, low(maxc)).
table(low_v_mins_maxc, 15, 9, 2, low(maxc)).
table(low_v_maxs_maxc, 15, 3, 2, low(maxc)).
table(low_c_minc_maxc, 15, 11, 2, low(maxc)).
table(low_c_s_maxc, 15, 4, 2, low(maxc)).
table(low_c_mins_maxc, 15, 18, 2, low(maxc)).
table(low_c_maxs_maxc, 15, 3, 2, low(maxc)).
table(low_minc_s_maxc, 15, 6, 2, low(maxc)).
table(low_minc_mins_maxc, 15, 4, 2, low(maxc)).
table(low_minc_maxs_maxc, 15, 4, 2, low(maxc)).
table(low_s_mins_maxc, 15, 18, 2, low(maxc)).
table(low_s_maxs_maxc, 15, 3, 2, low(maxc)).
table(low_mins_maxs_maxc, 15, 3, 2, low(maxc)).
table(low_v_c_minc_maxc, 15, 6, 3, low(maxc)).
table(low_v_c_s_maxc, 15, 4, 3, low(maxc)).
table(low_v_c_mins_maxc, 15, 4, 3, low(maxc)).
table(low_v_c_maxs_maxc, 15, 4, 3, low(maxc)).
table(low_v_minc_s_maxc, 15, 5, 3, low(maxc)).
table(low_v_minc_mins_maxc, 15, 7, 3, low(maxc)).
table(low_v_minc_maxs_maxc, 15, 5, 3, low(maxc)).
table(low_v_s_mins_maxc, 15, 5, 3, low(maxc)).
table(low_v_s_maxs_maxc, 15, 4, 3, low(maxc)).
table(low_v_mins_maxs_maxc, 15, 4, 3, low(maxc)).
table(low_c_minc_s_maxc, 15, 5, 3, low(maxc)).
table(low_c_minc_mins_maxc, 15, 6, 3, low(maxc)).
table(low_c_minc_maxs_maxc, 15, 5, 3, low(maxc)).
table(low_c_s_mins_maxc, 15, 4, 3, low(maxc)).
table(low_c_s_maxs_maxc, 15, 4, 3, low(maxc)).
table(low_c_mins_maxs_maxc, 15, 4, 3, low(maxc)).
table(low_minc_s_mins_maxc, 15, 5, 3, low(maxc)).
table(low_minc_s_maxs_maxc, 15, 5, 3, low(maxc)).
table(low_minc_mins_maxs_maxc, 15, 5, 3, low(maxc)).
table(low_s_mins_maxs_maxc, 15, 4, 3, low(maxc)).
table(low_v_c_minc_s_maxc, 15, 7, 4, low(maxc)).
table(low_v_c_minc_mins_maxc, 15, 7, 4, low(maxc)).
table(low_v_c_minc_maxs_maxc, 15, 7, 4, low(maxc)).
table(low_v_c_s_mins_maxc, 15, 5, 4, low(maxc)).
table(low_v_c_s_maxs_maxc, 15, 5, 4, low(maxc)).
table(low_v_c_mins_maxs_maxc, 15, 5, 4, low(maxc)).
table(low_v_minc_s_mins_maxc, 15, 6, 4, low(maxc)).
table(low_v_minc_s_maxs_maxc, 15, 6, 4, low(maxc)).
table(low_v_minc_mins_maxs_maxc, 15, 6, 4, low(maxc)).
table(low_v_s_mins_maxs_maxc, 15, 5, 4, low(maxc)).
table(low_c_minc_s_mins_maxc, 15, 6, 4, low(maxc)).
table(low_c_minc_s_maxs_maxc, 15, 6, 4, low(maxc)).
table(low_c_minc_mins_maxs_maxc, 15, 6, 4, low(maxc)).
table(low_c_s_mins_maxs_maxc, 15, 5, 4, low(maxc)).
table(low_minc_s_mins_maxs_maxc, 15, 6, 4, low(maxc)).
table(low_v_c_minc_s_mins_maxc, 15, 8, 5, low(maxc)).
table(low_v_c_minc_s_maxs_maxc, 15, 8, 5, low(maxc)).
table(low_v_c_minc_mins_maxs_maxc, 15, 8, 5, low(maxc)).
table(low_v_c_s_mins_maxs_maxc, 15, 6, 5, low(maxc)).
table(low_v_minc_s_mins_maxs_maxc, 15, 7, 5, low(maxc)).
table(low_c_minc_s_mins_maxs_maxc, 15, 7, 5, low(maxc)).
table(low_v_c_minc_s_mins_maxs_maxc, 15, 9, 6, low(maxc)).
table(up_v_maxc, 15, 11, 1, up(maxc)).
table(up_v_c_maxc, 15, 11, 2, up(maxc)).
table(up_v_minc_maxc, 15, 11, 2, up(maxc)).
table(up_v_s_maxc, 15, 13, 2, up(maxc)).
table(up_v_mins_maxc, 15, 12, 2, up(maxc)).
table(up_v_maxs_maxc, 15, 13, 2, up(maxc)).
table(up_v_c_minc_maxc, 15, 11, 3, up(maxc)).
table(up_v_c_s_maxc, 15, 13, 3, up(maxc)).
table(up_v_c_mins_maxc, 15, 12, 3, up(maxc)).
table(up_v_c_maxs_maxc, 15, 13, 3, up(maxc)).
table(up_v_minc_s_maxc, 15, 12, 3, up(maxc)).
table(up_v_minc_mins_maxc, 15, 12, 3, up(maxc)).
table(up_v_minc_maxs_maxc, 15, 12, 3, up(maxc)).
table(up_v_s_mins_maxc, 15, 14, 3, up(maxc)).
table(up_v_s_maxs_maxc, 15, 14, 3, up(maxc)).
table(up_v_mins_maxs_maxc, 15, 14, 3, up(maxc)).
table(up_v_c_minc_s_maxc, 15, 12, 4, up(maxc)).
table(up_v_c_minc_mins_maxc, 15, 12, 4, up(maxc)).
table(up_v_c_minc_maxs_maxc, 15, 12, 4, up(maxc)).
table(up_v_c_s_mins_maxc, 15, 14, 4, up(maxc)).
table(up_v_c_s_maxs_maxc, 15, 6, 4, up(maxc)).
table(up_v_c_mins_maxs_maxc, 15, 14, 4, up(maxc)).
table(up_v_minc_s_mins_maxc, 15, 13, 4, up(maxc)).
table(up_v_minc_s_maxs_maxc, 15, 13, 4, up(maxc)).
table(up_v_minc_mins_maxs_maxc, 15, 13, 4, up(maxc)).
table(up_v_s_mins_maxs_maxc, 15, 15, 4, up(maxc)).
table(up_v_c_minc_s_mins_maxc, 15, 13, 5, up(maxc)).
table(up_v_c_minc_s_maxs_maxc, 15, 8, 5, up(maxc)).
table(up_v_c_minc_mins_maxs_maxc, 15, 13, 5, up(maxc)).
table(up_v_c_s_mins_maxs_maxc, 15, 10, 5, up(maxc)).
table(up_v_minc_s_mins_maxs_maxc, 15, 14, 5, up(maxc)).
table(up_v_c_minc_s_mins_maxs_maxc, 15, 9, 6, up(maxc)).
table(low_v_s, 15, 18, 1, low(s)).
table(low_c_s, 15, 3, 1, low(s)).
table(low_minc_s, 15, 18, 1, low(s)).
table(low_maxc_s, 15, 18, 1, low(s)).
table(low_mins_s, 15, 18, 1, low(s)).
table(low_maxs_s, 15, 18, 1, low(s)).
table(low_v_c_s, 15, 4, 2, low(s)).
table(low_v_minc_s, 15, 18, 2, low(s)).
table(low_v_maxc_s, 15, 6, 2, low(s)).
table(low_v_mins_s, 15, 9, 2, low(s)).
table(low_v_maxs_s, 15, 3, 2, low(s)).
table(low_c_minc_s, 15, 5, 2, low(s)).
table(low_c_maxc_s, 15, 5, 2, low(s)).
table(low_c_mins_s, 15, 5, 2, low(s)).
table(low_c_maxs_s, 15, 5, 2, low(s)).
table(low_minc_maxc_s, 15, 18, 2, low(s)).
table(low_minc_mins_s, 15, 12, 2, low(s)).
table(low_minc_maxs_s, 15, 14, 2, low(s)).
table(low_maxc_mins_s, 15, 7, 2, low(s)).
table(low_maxc_maxs_s, 15, 14, 2, low(s)).
table(low_mins_maxs_s, 15, 12, 2, low(s)).
table(low_v_c_minc_s, 15, 8, 3, low(s)).
table(low_v_c_maxc_s, 15, 6, 3, low(s)).
table(low_v_c_mins_s, 15, 8, 3, low(s)).
table(low_v_c_maxs_s, 15, 4, 3, low(s)).
table(low_v_minc_maxc_s, 15, 10, 3, low(s)).
table(low_v_minc_mins_s, 15, 13, 3, low(s)).
table(low_v_minc_maxs_s, 15, 4, 3, low(s)).
table(low_v_maxc_mins_s, 15, 4, 3, low(s)).
table(low_v_maxc_maxs_s, 15, 4, 3, low(s)).
table(low_v_mins_maxs_s, 15, 4, 3, low(s)).
table(low_c_minc_maxc_s, 15, 8, 3, low(s)).
table(low_c_minc_mins_s, 15, 4, 3, low(s)).
table(low_c_minc_maxs_s, 15, 5, 3, low(s)).
table(low_c_maxc_mins_s, 15, 8, 3, low(s)).
table(low_c_maxc_maxs_s, 15, 5, 3, low(s)).
table(low_c_mins_maxs_s, 15, 8, 3, low(s)).
table(low_minc_maxc_mins_s, 15, 13, 3, low(s)).
table(low_minc_maxc_maxs_s, 15, 14, 3, low(s)).
table(low_minc_mins_maxs_s, 15, 8, 3, low(s)).
table(low_maxc_mins_maxs_s, 15, 9, 3, low(s)).
table(low_v_c_minc_maxc_s, 15, 10, 4, low(s)).
table(low_v_c_minc_mins_s, 15, 7, 4, low(s)).
table(low_v_c_minc_maxs_s, 15, 7, 4, low(s)).
table(low_v_c_maxc_mins_s, 15, 5, 4, low(s)).
table(low_v_c_maxc_maxs_s, 15, 5, 4, low(s)).
table(low_v_c_mins_maxs_s, 15, 5, 4, low(s)).
table(low_v_minc_maxc_mins_s, 15, 8, 4, low(s)).
table(low_v_minc_maxc_maxs_s, 15, 6, 4, low(s)).
table(low_v_minc_mins_maxs_s, 15, 5, 4, low(s)).
table(low_v_maxc_mins_maxs_s, 15, 5, 4, low(s)).
table(low_c_minc_maxc_mins_s, 15, 6, 4, low(s)).
table(low_c_minc_maxc_maxs_s, 15, 7, 4, low(s)).
table(low_c_minc_mins_maxs_s, 15, 5, 4, low(s)).
table(low_c_maxc_mins_maxs_s, 15, 6, 4, low(s)).
table(low_minc_maxc_mins_maxs_s, 15, 9, 4, low(s)).
table(low_v_c_minc_maxc_mins_s, 15, 8, 5, low(s)).
table(low_v_c_minc_maxc_maxs_s, 15, 8, 5, low(s)).
table(low_v_c_minc_mins_maxs_s, 15, 8, 5, low(s)).
table(low_v_c_maxc_mins_maxs_s, 15, 6, 5, low(s)).
table(low_v_minc_maxc_mins_maxs_s, 15, 7, 5, low(s)).
table(low_c_minc_maxc_mins_maxs_s, 15, 7, 5, low(s)).
table(low_v_c_minc_maxc_mins_maxs_s, 15, 9, 6, low(s)).
table(up_v_s, 15, 7, 1, up(s)).
table(up_v_c_s, 15, 8, 2, up(s)).
table(up_v_minc_s, 15, 8, 2, up(s)).
table(up_v_maxc_s, 15, 8, 2, up(s)).
table(up_v_mins_s, 15, 4, 2, up(s)).
table(up_v_maxs_s, 15, 7, 2, up(s)).
table(up_c_maxc_s, 15, 9, 2, up(s)).
table(up_v_c_minc_s, 15, 11, 3, up(s)).
table(up_v_c_maxc_s, 15, 9, 3, up(s)).
table(up_v_c_mins_s, 15, 5, 3, up(s)).
table(up_v_c_maxs_s, 15, 8, 3, up(s)).
table(up_v_minc_maxc_s, 15, 10, 3, up(s)).
table(up_v_minc_mins_s, 15, 5, 3, up(s)).
table(up_v_minc_maxs_s, 15, 8, 3, up(s)).
table(up_v_maxc_mins_s, 15, 5, 3, up(s)).
table(up_v_maxc_maxs_s, 15, 8, 3, up(s)).
table(up_v_mins_maxs_s, 15, 7, 3, up(s)).
table(up_c_minc_maxc_s, 15, 12, 3, up(s)).
table(up_c_maxc_mins_s, 15, 5, 3, up(s)).
table(up_c_maxc_maxs_s, 15, 9, 3, up(s)).
table(up_v_c_minc_maxc_s, 15, 12, 4, up(s)).
table(up_v_c_minc_mins_s, 15, 8, 4, up(s)).
table(up_v_c_minc_maxs_s, 15, 11, 4, up(s)).
table(up_v_c_maxc_mins_s, 15, 6, 4, up(s)).
table(up_v_c_maxc_maxs_s, 15, 9, 4, up(s)).
table(up_v_c_mins_maxs_s, 15, 8, 4, up(s)).
table(up_v_minc_maxc_mins_s, 15, 7, 4, up(s)).
table(up_v_minc_maxc_maxs_s, 15, 10, 4, up(s)).
table(up_v_minc_mins_maxs_s, 15, 8, 4, up(s)).
table(up_v_maxc_mins_maxs_s, 15, 8, 4, up(s)).
table(up_c_minc_maxc_mins_s, 15, 7, 4, up(s)).
table(up_c_minc_maxc_maxs_s, 15, 12, 4, up(s)).
table(up_c_maxc_mins_maxs_s, 15, 6, 4, up(s)).
table(up_v_c_minc_maxc_mins_s, 15, 9, 5, up(s)).
table(up_v_c_minc_maxc_maxs_s, 15, 12, 5, up(s)).
table(up_v_c_minc_mins_maxs_s, 15, 11, 5, up(s)).
table(up_v_c_maxc_mins_maxs_s, 15, 9, 5, up(s)).
table(up_v_minc_maxc_mins_maxs_s, 15, 10, 5, up(s)).
table(up_c_minc_maxc_mins_maxs_s, 15, 8, 5, up(s)).
table(up_v_c_minc_maxc_mins_maxs_s, 15, 12, 6, up(s)).
table(low_v_mins, 15, 2, 1, low(mins)).
table(low_c_mins, 15, 2, 1, low(mins)).
table(low_minc_mins, 15, 2, 1, low(mins)).
table(low_maxc_mins, 15, 2, 1, low(mins)).
table(low_s_mins, 15, 2, 1, low(mins)).
table(low_maxs_mins, 15, 2, 1, low(mins)).
table(low_v_c_mins, 15, 3, 2, low(mins)).
table(low_v_minc_mins, 15, 3, 2, low(mins)).
table(low_v_maxc_mins, 15, 3, 2, low(mins)).
table(low_v_s_mins, 15, 3, 2, low(mins)).
table(low_v_maxs_mins, 15, 3, 2, low(mins)).
table(low_c_minc_mins, 15, 3, 2, low(mins)).
table(low_c_maxc_mins, 15, 3, 2, low(mins)).
table(low_c_s_mins, 15, 3, 2, low(mins)).
table(low_c_maxs_mins, 15, 3, 2, low(mins)).
table(low_minc_maxc_mins, 15, 4, 2, low(mins)).
table(low_minc_s_mins, 15, 3, 2, low(mins)).
table(low_minc_maxs_mins, 15, 3, 2, low(mins)).
table(low_maxc_s_mins, 15, 3, 2, low(mins)).
table(low_maxc_maxs_mins, 15, 3, 2, low(mins)).
table(low_s_maxs_mins, 15, 3, 2, low(mins)).
table(low_v_c_minc_mins, 15, 6, 3, low(mins)).
table(low_v_c_maxc_mins, 15, 4, 3, low(mins)).
table(low_v_c_s_mins, 15, 4, 3, low(mins)).
table(low_v_c_maxs_mins, 15, 4, 3, low(mins)).
table(low_v_minc_maxc_mins, 15, 5, 3, low(mins)).
table(low_v_minc_s_mins, 15, 4, 3, low(mins)).
table(low_v_minc_maxs_mins, 15, 4, 3, low(mins)).
table(low_v_maxc_s_mins, 15, 4, 3, low(mins)).
table(low_v_maxc_maxs_mins, 15, 4, 3, low(mins)).
table(low_v_s_maxs_mins, 15, 4, 3, low(mins)).
table(low_c_minc_maxc_mins, 15, 5, 3, low(mins)).
table(low_c_minc_s_mins, 15, 4, 3, low(mins)).
table(low_c_minc_maxs_mins, 15, 4, 3, low(mins)).
table(low_c_maxc_s_mins, 15, 4, 3, low(mins)).
table(low_c_maxc_maxs_mins, 15, 4, 3, low(mins)).
table(low_c_s_maxs_mins, 15, 4, 3, low(mins)).
table(low_minc_maxc_s_mins, 15, 5, 3, low(mins)).
table(low_minc_maxc_maxs_mins, 15, 5, 3, low(mins)).
table(low_minc_s_maxs_mins, 15, 4, 3, low(mins)).
table(low_maxc_s_maxs_mins, 15, 4, 3, low(mins)).
table(low_v_c_minc_maxc_mins, 15, 7, 4, low(mins)).
table(low_v_c_minc_s_mins, 15, 7, 4, low(mins)).
table(low_v_c_minc_maxs_mins, 15, 7, 4, low(mins)).
table(low_v_c_maxc_s_mins, 15, 5, 4, low(mins)).
table(low_v_c_maxc_maxs_mins, 15, 5, 4, low(mins)).
table(low_v_c_s_maxs_mins, 15, 5, 4, low(mins)).
table(low_v_minc_maxc_s_mins, 15, 6, 4, low(mins)).
table(low_v_minc_maxc_maxs_mins, 15, 6, 4, low(mins)).
table(low_v_minc_s_maxs_mins, 15, 5, 4, low(mins)).
table(low_v_maxc_s_maxs_mins, 15, 5, 4, low(mins)).
table(low_c_minc_maxc_s_mins, 15, 6, 4, low(mins)).
table(low_c_minc_maxc_maxs_mins, 15, 6, 4, low(mins)).
table(low_c_minc_s_maxs_mins, 15, 5, 4, low(mins)).
table(low_c_maxc_s_maxs_mins, 15, 5, 4, low(mins)).
table(low_minc_maxc_s_maxs_mins, 15, 6, 4, low(mins)).
table(low_v_c_minc_maxc_s_mins, 15, 8, 5, low(mins)).
table(low_v_c_minc_maxc_maxs_mins, 15, 8, 5, low(mins)).
table(low_v_c_minc_s_maxs_mins, 15, 8, 5, low(mins)).
table(low_v_c_maxc_s_maxs_mins, 15, 6, 5, low(mins)).
table(low_v_minc_maxc_s_maxs_mins, 15, 7, 5, low(mins)).
table(low_c_minc_maxc_s_maxs_mins, 15, 7, 5, low(mins)).
table(low_v_c_minc_maxc_s_maxs_mins, 15, 9, 6, low(mins)).
table(up_v_mins, 15, 18, 1, up(mins)).
table(up_maxc_mins, 15, 9, 1, up(mins)).
table(up_maxs_mins, 15, 2, 1, up(mins)).
table(up_v_c_mins, 15, 6, 2, up(mins)).
table(up_v_minc_mins, 15, 3, 2, up(mins)).
table(up_v_maxc_mins, 15, 7, 2, up(mins)).
table(up_v_s_mins, 15, 3, 2, up(mins)).
table(up_v_maxs_mins, 15, 3, 2, up(mins)).
table(up_c_maxc_mins, 15, 5, 2, up(mins)).
table(up_c_maxs_mins, 15, 3, 2, up(mins)).
table(up_minc_maxc_mins, 15, 4, 2, up(mins)).
table(up_minc_maxs_mins, 15, 3, 2, up(mins)).
table(up_maxc_s_mins, 15, 3, 2, up(mins)).
table(up_maxc_maxs_mins, 15, 3, 2, up(mins)).
table(up_s_maxs_mins, 15, 3, 2, up(mins)).
table(up_v_c_minc_mins, 15, 6, 3, up(mins)).
table(up_v_c_maxc_mins, 15, 7, 3, up(mins)).
table(up_v_c_s_mins, 15, 4, 3, up(mins)).
table(up_v_c_maxs_mins, 15, 4, 3, up(mins)).
table(up_v_minc_maxc_mins, 15, 5, 3, up(mins)).
table(up_v_minc_s_mins, 15, 4, 3, up(mins)).
table(up_v_minc_maxs_mins, 15, 4, 3, up(mins)).
table(up_v_maxc_s_mins, 15, 4, 3, up(mins)).
table(up_v_maxc_maxs_mins, 15, 4, 3, up(mins)).
table(up_v_s_maxs_mins, 15, 4, 3, up(mins)).
table(up_c_minc_maxc_mins, 15, 5, 3, up(mins)).
table(up_c_minc_maxs_mins, 15, 4, 3, up(mins)).
table(up_c_maxc_s_mins, 15, 4, 3, up(mins)).
table(up_c_maxc_maxs_mins, 15, 4, 3, up(mins)).
table(up_c_s_maxs_mins, 15, 4, 3, up(mins)).
table(up_minc_maxc_s_mins, 15, 5, 3, up(mins)).
table(up_minc_maxc_maxs_mins, 15, 5, 3, up(mins)).
table(up_minc_s_maxs_mins, 15, 4, 3, up(mins)).
table(up_maxc_s_maxs_mins, 15, 4, 3, up(mins)).
table(up_v_c_minc_maxc_mins, 15, 7, 4, up(mins)).
table(up_v_c_minc_s_mins, 15, 7, 4, up(mins)).
table(up_v_c_minc_maxs_mins, 15, 7, 4, up(mins)).
table(up_v_c_maxc_s_mins, 15, 5, 4, up(mins)).
table(up_v_c_maxc_maxs_mins, 15, 5, 4, up(mins)).
table(up_v_c_s_maxs_mins, 15, 5, 4, up(mins)).
table(up_v_minc_maxc_s_mins, 15, 6, 4, up(mins)).
table(up_v_minc_maxc_maxs_mins, 15, 6, 4, up(mins)).
table(up_v_minc_s_maxs_mins, 15, 5, 4, up(mins)).
table(up_v_maxc_s_maxs_mins, 15, 5, 4, up(mins)).
table(up_c_minc_maxc_s_mins, 15, 6, 4, up(mins)).
table(up_c_minc_maxc_maxs_mins, 15, 6, 4, up(mins)).
table(up_c_minc_s_maxs_mins, 15, 5, 4, up(mins)).
table(up_c_maxc_s_maxs_mins, 15, 5, 4, up(mins)).
table(up_minc_maxc_s_maxs_mins, 15, 6, 4, up(mins)).
table(up_v_c_minc_maxc_s_mins, 15, 8, 5, up(mins)).
table(up_v_c_minc_maxc_maxs_mins, 15, 8, 5, up(mins)).
table(up_v_c_minc_s_maxs_mins, 15, 8, 5, up(mins)).
table(up_v_c_maxc_s_maxs_mins, 15, 6, 5, up(mins)).
table(up_v_minc_maxc_s_maxs_mins, 15, 7, 5, up(mins)).
table(up_c_minc_maxc_s_maxs_mins, 15, 7, 5, up(mins)).
table(up_v_c_minc_maxc_s_maxs_mins, 15, 9, 6, up(mins)).
table(low_v_maxs, 15, 7, 1, low(maxs)).
table(low_c_maxs, 15, 4, 1, low(maxs)).
table(low_minc_maxs, 15, 4, 1, low(maxs)).
table(low_maxc_maxs, 15, 4, 1, low(maxs)).
table(low_s_maxs, 15, 7, 1, low(maxs)).
table(low_mins_maxs, 15, 2, 1, low(maxs)).
table(low_v_c_maxs, 15, 8, 2, low(maxs)).
table(low_v_minc_maxs, 15, 8, 2, low(maxs)).
table(low_v_maxc_maxs, 15, 8, 2, low(maxs)).
table(low_v_s_maxs, 15, 3, 2, low(maxs)).
table(low_v_mins_maxs, 15, 3, 2, low(maxs)).
table(low_c_minc_maxs, 15, 5, 2, low(maxs)).
table(low_c_maxc_maxs, 15, 5, 2, low(maxs)).
table(low_c_s_maxs, 15, 8, 2, low(maxs)).
table(low_c_mins_maxs, 15, 3, 2, low(maxs)).
table(low_minc_maxc_maxs, 15, 6, 2, low(maxs)).
table(low_minc_s_maxs, 15, 3, 2, low(maxs)).
table(low_minc_mins_maxs, 15, 3, 2, low(maxs)).
table(low_maxc_s_maxs, 15, 3, 2, low(maxs)).
table(low_maxc_mins_maxs, 15, 3, 2, low(maxs)).
table(low_s_mins_maxs, 15, 7, 2, low(maxs)).
table(low_v_c_minc_maxs, 15, 11, 3, low(maxs)).
table(low_v_c_maxc_maxs, 15, 9, 3, low(maxs)).
table(low_v_c_s_maxs, 15, 4, 3, low(maxs)).
table(low_v_c_mins_maxs, 15, 4, 3, low(maxs)).
table(low_v_minc_maxc_maxs, 15, 10, 3, low(maxs)).
table(low_v_minc_s_maxs, 15, 4, 3, low(maxs)).
table(low_v_minc_mins_maxs, 15, 4, 3, low(maxs)).
table(low_v_maxc_s_maxs, 15, 4, 3, low(maxs)).
table(low_v_maxc_mins_maxs, 15, 4, 3, low(maxs)).
table(low_v_s_mins_maxs, 15, 4, 3, low(maxs)).
table(low_c_minc_maxc_maxs, 15, 7, 3, low(maxs)).
table(low_c_minc_s_maxs, 15, 4, 3, low(maxs)).
table(low_c_minc_mins_maxs, 15, 4, 3, low(maxs)).
table(low_c_maxc_s_maxs, 15, 4, 3, low(maxs)).
table(low_c_maxc_mins_maxs, 15, 4, 3, low(maxs)).
table(low_c_s_mins_maxs, 15, 8, 3, low(maxs)).
table(low_minc_maxc_s_maxs, 15, 5, 3, low(maxs)).
table(low_minc_maxc_mins_maxs, 15, 5, 3, low(maxs)).
table(low_minc_s_mins_maxs, 15, 4, 3, low(maxs)).
table(low_maxc_s_mins_maxs, 15, 4, 3, low(maxs)).
table(low_v_c_minc_maxc_maxs, 15, 12, 4, low(maxs)).
table(low_v_c_minc_s_maxs, 15, 7, 4, low(maxs)).
table(low_v_c_minc_mins_maxs, 15, 7, 4, low(maxs)).
table(low_v_c_maxc_s_maxs, 15, 5, 4, low(maxs)).
table(low_v_c_maxc_mins_maxs, 15, 5, 4, low(maxs)).
table(low_v_c_s_mins_maxs, 15, 5, 4, low(maxs)).
table(low_v_minc_maxc_s_maxs, 15, 6, 4, low(maxs)).
table(low_v_minc_maxc_mins_maxs, 15, 6, 4, low(maxs)).
table(low_v_minc_s_mins_maxs, 15, 5, 4, low(maxs)).
table(low_v_maxc_s_mins_maxs, 15, 5, 4, low(maxs)).
table(low_c_minc_maxc_s_maxs, 15, 6, 4, low(maxs)).
table(low_c_minc_maxc_mins_maxs, 15, 6, 4, low(maxs)).
table(low_c_minc_s_mins_maxs, 15, 5, 4, low(maxs)).
table(low_c_maxc_s_mins_maxs, 15, 5, 4, low(maxs)).
table(low_minc_maxc_s_mins_maxs, 15, 6, 4, low(maxs)).
table(low_v_c_minc_maxc_s_maxs, 15, 8, 5, low(maxs)).
table(low_v_c_minc_maxc_mins_maxs, 15, 8, 5, low(maxs)).
table(low_v_c_minc_s_mins_maxs, 15, 8, 5, low(maxs)).
table(low_v_c_maxc_s_mins_maxs, 15, 6, 5, low(maxs)).
table(low_v_minc_maxc_s_mins_maxs, 15, 7, 5, low(maxs)).
table(low_c_minc_maxc_s_mins_maxs, 15, 7, 5, low(maxs)).
table(low_v_c_minc_maxc_s_mins_maxs, 15, 9, 6, low(maxs)).
table(up_v_maxs, 15, 18, 1, up(maxs)).
table(up_maxc_maxs, 15, 2, 1, up(maxs)).
table(up_v_c_maxs, 15, 18, 2, up(maxs)).
table(up_v_minc_maxs, 15, 12, 2, up(maxs)).
table(up_v_maxc_maxs, 15, 3, 2, up(maxs)).
table(up_v_s_maxs, 15, 7, 2, up(maxs)).
table(up_v_mins_maxs, 15, 9, 2, up(maxs)).
table(up_c_maxc_maxs, 15, 3, 2, up(maxs)).
table(up_minc_maxc_maxs, 15, 4, 2, up(maxs)).
table(up_maxc_s_maxs, 15, 3, 2, up(maxs)).
table(up_maxc_mins_maxs, 15, 3, 2, up(maxs)).
table(up_v_c_minc_maxs, 15, 12, 3, up(maxs)).
table(up_v_c_maxc_maxs, 15, 4, 3, up(maxs)).
table(up_v_c_s_maxs, 15, 8, 3, up(maxs)).
table(up_v_c_mins_maxs, 15, 18, 3, up(maxs)).
table(up_v_minc_maxc_maxs, 15, 5, 3, up(maxs)).
table(up_v_minc_s_maxs, 15, 5, 3, up(maxs)).
table(up_v_minc_mins_maxs, 15, 14, 3, up(maxs)).
table(up_v_maxc_s_maxs, 15, 4, 3, up(maxs)).
table(up_v_maxc_mins_maxs, 15, 4, 3, up(maxs)).
table(up_v_s_mins_maxs, 15, 7, 3, up(maxs)).
table(up_c_minc_maxc_maxs, 15, 5, 3, up(maxs)).
table(up_c_maxc_s_maxs, 15, 4, 3, up(maxs)).
table(up_c_maxc_mins_maxs, 15, 4, 3, up(maxs)).
table(up_minc_maxc_s_maxs, 15, 5, 3, up(maxs)).
table(up_minc_maxc_mins_maxs, 15, 5, 3, up(maxs)).
table(up_maxc_s_mins_maxs, 15, 4, 3, up(maxs)).
table(up_v_c_minc_maxc_maxs, 15, 7, 4, up(maxs)).
table(up_v_c_minc_s_maxs, 15, 7, 4, up(maxs)).
table(up_v_c_minc_mins_maxs, 15, 13, 4, up(maxs)).
table(up_v_c_maxc_s_maxs, 15, 5, 4, up(maxs)).
table(up_v_c_maxc_mins_maxs, 15, 5, 4, up(maxs)).
table(up_v_c_s_mins_maxs, 15, 8, 4, up(maxs)).
table(up_v_minc_maxc_s_maxs, 15, 6, 4, up(maxs)).
table(up_v_minc_maxc_mins_maxs, 15, 6, 4, up(maxs)).
table(up_v_minc_s_mins_maxs, 15, 6, 4, up(maxs)).
table(up_v_maxc_s_mins_maxs, 15, 5, 4, up(maxs)).
table(up_c_minc_maxc_s_maxs, 15, 6, 4, up(maxs)).
table(up_c_minc_maxc_mins_maxs, 15, 6, 4, up(maxs)).
table(up_c_maxc_s_mins_maxs, 15, 5, 4, up(maxs)).
table(up_minc_maxc_s_mins_maxs, 15, 6, 4, up(maxs)).
table(up_v_c_minc_maxc_s_maxs, 15, 8, 5, up(maxs)).
table(up_v_c_minc_maxc_mins_maxs, 15, 8, 5, up(maxs)).
table(up_v_c_minc_s_mins_maxs, 15, 8, 5, up(maxs)).
table(up_v_c_maxc_s_mins_maxs, 15, 6, 5, up(maxs)).
table(up_v_minc_maxc_s_mins_maxs, 15, 7, 5, up(maxs)).
table(up_c_minc_maxc_s_mins_maxs, 15, 7, 5, up(maxs)).
table(up_v_c_minc_maxc_s_mins_maxs, 15, 9, 6, up(maxs)).
table(low_v_mina, 15, 12, 1, low(mina)).
table(low_c_mina, 15, 7, 1, low(mina)).
table(low_minc_mina, 15, 19, 1, low(mina)).
table(low_maxc_mina, 15, 19, 1, low(mina)).
table(low_s_mina, 15, 12, 1, low(mina)).
table(low_mins_mina, 15, 11, 1, low(mina)).
table(low_maxs_mina, 15, 11, 1, low(mina)).
table(low_v_c_mina, 15, 10, 2, low(mina)).
table(low_v_minc_mina, 15, 11, 2, low(mina)).
table(low_v_maxc_mina, 15, 11, 2, low(mina)).
table(low_v_s_mina, 15, 4, 2, low(mina)).
table(low_v_mins_mina, 15, 4, 2, low(mina)).
table(low_v_maxs_mina, 15, 10, 2, low(mina)).
table(low_c_minc_mina, 15, 8, 2, low(mina)).
table(low_c_maxc_mina, 15, 6, 2, low(mina)).
table(low_c_s_mina, 15, 10, 2, low(mina)).
table(low_c_mins_mina, 15, 7, 2, low(mina)).
table(low_c_maxs_mina, 15, 5, 2, low(mina)).
table(low_minc_maxc_mina, 15, 19, 2, low(mina)).
table(low_minc_s_mina, 15, 11, 2, low(mina)).
table(low_minc_mins_mina, 15, 19, 2, low(mina)).
table(low_minc_maxs_mina, 15, 19, 2, low(mina)).
table(low_maxc_s_mina, 15, 11, 2, low(mina)).
table(low_maxc_mins_mina, 15, 19, 2, low(mina)).
table(low_maxc_maxs_mina, 15, 19, 2, low(mina)).
table(low_s_mins_mina, 15, 12, 2, low(mina)).
table(low_s_maxs_mina, 15, 10, 2, low(mina)).
table(low_mins_maxs_mina, 15, 6, 2, low(mina)).
table(low_v_c_minc_mina, 15, 13, 3, low(mina)).
table(low_v_c_maxc_mina, 15, 11, 3, low(mina)).
table(low_v_c_s_mina, 15, 4, 3, low(mina)).
table(low_v_c_mins_mina, 15, 7, 3, low(mina)).
table(low_v_c_maxs_mina, 15, 10, 3, low(mina)).
table(low_v_minc_maxc_mina, 15, 13, 3, low(mina)).
table(low_v_minc_s_mina, 15, 4, 3, low(mina)).
table(low_v_minc_mins_mina, 15, 4, 3, low(mina)).
table(low_v_minc_maxs_mina, 15, 11, 3, low(mina)).
table(low_v_maxc_s_mina, 15, 4, 3, low(mina)).
table(low_v_maxc_mins_mina, 15, 6, 3, low(mina)).
table(low_v_maxc_maxs_mina, 15, 10, 3, low(mina)).
table(low_v_s_mins_mina, 15, 5, 3, low(mina)).
table(low_v_s_maxs_mina, 15, 5, 3, low(mina)).
table(low_v_mins_maxs_mina, 15, 5, 3, low(mina)).
table(low_c_minc_maxc_mina, 15, 8, 3, low(mina)).
table(low_c_minc_s_mina, 15, 11, 3, low(mina)).
table(low_c_minc_mins_mina, 15, 8, 3, low(mina)).
table(low_c_minc_maxs_mina, 15, 6, 3, low(mina)).
table(low_c_maxc_s_mina, 15, 8, 3, low(mina)).
table(low_c_maxc_mins_mina, 15, 6, 3, low(mina)).
table(low_c_maxc_maxs_mina, 15, 6, 3, low(mina)).
table(low_c_s_mins_mina, 15, 10, 3, low(mina)).
table(low_c_s_maxs_mina, 15, 10, 3, low(mina)).
table(low_c_mins_maxs_mina, 15, 5, 3, low(mina)).
table(low_minc_maxc_s_mina, 15, 10, 3, low(mina)).
table(low_minc_maxc_mins_mina, 15, 16, 3, low(mina)).
table(low_minc_maxc_maxs_mina, 15, 18, 3, low(mina)).
table(low_minc_s_mins_mina, 15, 6, 3, low(mina)).
table(low_minc_s_maxs_mina, 15, 8, 3, low(mina)).
table(low_minc_mins_maxs_mina, 15, 17, 3, low(mina)).
table(low_maxc_s_mins_mina, 15, 8, 3, low(mina)).
table(low_maxc_s_maxs_mina, 15, 7, 3, low(mina)).
table(low_maxc_mins_maxs_mina, 15, 8, 3, low(mina)).
table(low_s_mins_maxs_mina, 15, 10, 3, low(mina)).
table(low_v_c_minc_maxc_mina, 15, 14, 4, low(mina)).
table(low_v_c_minc_s_mina, 15, 8, 4, low(mina)).
table(low_v_c_minc_mins_mina, 15, 10, 4, low(mina)).
table(low_v_c_minc_maxs_mina, 15, 13, 4, low(mina)).
table(low_v_c_maxc_s_mina, 15, 5, 4, low(mina)).
table(low_v_c_maxc_mins_mina, 15, 8, 4, low(mina)).
table(low_v_c_maxc_maxs_mina, 15, 11, 4, low(mina)).
table(low_v_c_s_mins_mina, 15, 5, 4, low(mina)).
table(low_v_c_s_maxs_mina, 15, 5, 4, low(mina)).
table(low_v_c_mins_maxs_mina, 15, 8, 4, low(mina)).
table(low_v_minc_maxc_s_mina, 15, 6, 4, low(mina)).
table(low_v_minc_maxc_mins_mina, 15, 7, 4, low(mina)).
table(low_v_minc_maxc_maxs_mina, 15, 12, 4, low(mina)).
table(low_v_minc_s_mins_mina, 15, 5, 4, low(mina)).
table(low_v_minc_s_maxs_mina, 15, 5, 4, low(mina)).
table(low_v_minc_mins_maxs_mina, 15, 6, 4, low(mina)).
table(low_v_maxc_s_mins_mina, 15, 5, 4, low(mina)).
table(low_v_maxc_s_maxs_mina, 15, 5, 4, low(mina)).
table(low_v_maxc_mins_maxs_mina, 15, 5, 4, low(mina)).
table(low_v_s_mins_maxs_mina, 15, 6, 4, low(mina)).
table(low_c_minc_maxc_s_mina, 15, 9, 4, low(mina)).
table(low_c_minc_maxc_mins_mina, 15, 7, 4, low(mina)).
table(low_c_minc_maxc_maxs_mina, 15, 8, 4, low(mina)).
table(low_c_minc_s_mins_mina, 15, 10, 4, low(mina)).
table(low_c_minc_s_maxs_mina, 15, 10, 4, low(mina)).
table(low_c_minc_mins_maxs_mina, 15, 6, 4, low(mina)).
table(low_c_maxc_s_mins_mina, 15, 6, 4, low(mina)).
table(low_c_maxc_s_maxs_mina, 15, 6, 4, low(mina)).
table(low_c_maxc_mins_maxs_mina, 15, 6, 4, low(mina)).
table(low_c_s_mins_maxs_mina, 15, 10, 4, low(mina)).
table(low_minc_maxc_s_mins_mina, 15, 9, 4, low(mina)).
table(low_minc_maxc_s_maxs_mina, 15, 9, 4, low(mina)).
table(low_minc_maxc_mins_maxs_mina, 15, 16, 4, low(mina)).
table(low_minc_s_mins_maxs_mina, 15, 7, 4, low(mina)).
table(low_maxc_s_mins_maxs_mina, 15, 6, 4, low(mina)).
table(low_v_c_minc_maxc_s_mina, 15, 9, 5, low(mina)).
table(low_v_c_minc_maxc_mins_mina, 15, 11, 5, low(mina)).
table(low_v_c_minc_maxc_maxs_mina, 15, 14, 5, low(mina)).
table(low_v_c_minc_s_mins_mina, 15, 8, 5, low(mina)).
table(low_v_c_minc_s_maxs_mina, 15, 8, 5, low(mina)).
table(low_v_c_minc_mins_maxs_mina, 15, 11, 5, low(mina)).
table(low_v_c_maxc_s_mins_mina, 15, 6, 5, low(mina)).
table(low_v_c_maxc_s_maxs_mina, 15, 6, 5, low(mina)).
table(low_v_c_maxc_mins_maxs_mina, 15, 9, 5, low(mina)).
table(low_v_c_s_mins_maxs_mina, 15, 6, 5, low(mina)).
table(low_v_minc_maxc_s_mins_mina, 15, 7, 5, low(mina)).
table(low_v_minc_maxc_s_maxs_mina, 15, 7, 5, low(mina)).
table(low_v_minc_maxc_mins_maxs_mina, 15, 7, 5, low(mina)).
table(low_v_minc_s_mins_maxs_mina, 15, 6, 5, low(mina)).
table(low_v_maxc_s_mins_maxs_mina, 15, 6, 5, low(mina)).
table(low_c_minc_maxc_s_mins_mina, 15, 8, 5, low(mina)).
table(low_c_minc_maxc_s_maxs_mina, 15, 8, 5, low(mina)).
table(low_c_minc_maxc_mins_maxs_mina, 15, 8, 5, low(mina)).
table(low_c_minc_s_mins_maxs_mina, 15, 11, 5, low(mina)).
table(low_c_maxc_s_mins_maxs_mina, 15, 7, 5, low(mina)).
table(low_minc_maxc_s_mins_maxs_mina, 15, 8, 5, low(mina)).
table(low_v_c_minc_maxc_s_mins_mina, 15, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_maxs_mina, 15, 9, 6, low(mina)).
table(low_v_c_minc_maxc_mins_maxs_mina, 15, 12, 6, low(mina)).
table(low_v_c_minc_s_mins_maxs_mina, 15, 9, 6, low(mina)).
table(low_v_c_maxc_s_mins_maxs_mina, 15, 7, 6, low(mina)).
table(low_v_minc_maxc_s_mins_maxs_mina, 15, 8, 6, low(mina)).
table(low_c_minc_maxc_s_mins_maxs_mina, 15, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_mins_maxs_mina, 15, 10, 7, low(mina)).
table(up_v_maxa, 15, 19, 1, up(maxa)).
table(up_v_c_maxa, 15, 19, 2, up(maxa)).
table(up_v_minc_maxa, 15, 19, 2, up(maxa)).
table(up_v_maxc_maxa, 15, 19, 2, up(maxa)).
table(up_v_s_maxa, 15, 19, 2, up(maxa)).
table(up_v_mins_maxa, 15, 19, 2, up(maxa)).
table(up_v_maxs_maxa, 15, 19, 2, up(maxa)).
table(up_c_maxc_maxa, 15, 19, 2, up(maxa)).
table(up_maxc_s_maxa, 15, 18, 2, up(maxa)).
table(up_s_maxs_maxa, 15, 19, 2, up(maxa)).
table(up_v_c_minc_maxa, 15, 19, 3, up(maxa)).
table(up_v_c_maxc_maxa, 15, 19, 3, up(maxa)).
table(up_v_c_s_maxa, 15, 19, 3, up(maxa)).
table(up_v_c_mins_maxa, 15, 19, 3, up(maxa)).
table(up_v_c_maxs_maxa, 15, 7, 3, up(maxa)).
table(up_v_minc_maxc_maxa, 15, 19, 3, up(maxa)).
table(up_v_minc_s_maxa, 15, 18, 3, up(maxa)).
table(up_v_minc_mins_maxa, 15, 19, 3, up(maxa)).
table(up_v_minc_maxs_maxa, 15, 19, 3, up(maxa)).
table(up_v_maxc_s_maxa, 15, 18, 3, up(maxa)).
table(up_v_maxc_mins_maxa, 15, 11, 3, up(maxa)).
table(up_v_maxc_maxs_maxa, 15, 6, 3, up(maxa)).
table(up_v_s_mins_maxa, 15, 19, 3, up(maxa)).
table(up_v_s_maxs_maxa, 15, 19, 3, up(maxa)).
table(up_v_mins_maxs_maxa, 15, 19, 3, up(maxa)).
table(up_c_minc_maxc_maxa, 15, 19, 3, up(maxa)).
table(up_c_maxc_s_maxa, 15, 18, 3, up(maxa)).
table(up_c_maxc_mins_maxa, 15, 12, 3, up(maxa)).
table(up_c_maxc_maxs_maxa, 15, 6, 3, up(maxa)).
table(up_c_s_maxs_maxa, 15, 19, 3, up(maxa)).
table(up_minc_maxc_s_maxa, 15, 5, 3, up(maxa)).
table(up_minc_s_maxs_maxa, 15, 18, 3, up(maxa)).
table(up_maxc_s_mins_maxa, 15, 10, 3, up(maxa)).
table(up_maxc_s_maxs_maxa, 15, 4, 3, up(maxa)).
table(up_s_mins_maxs_maxa, 15, 19, 3, up(maxa)).
table(up_v_c_minc_maxc_maxa, 15, 19, 4, up(maxa)).
table(up_v_c_minc_s_maxa, 15, 18, 4, up(maxa)).
table(up_v_c_minc_mins_maxa, 15, 11, 4, up(maxa)).
table(up_v_c_minc_maxs_maxa, 15, 8, 4, up(maxa)).
table(up_v_c_maxc_s_maxa, 15, 18, 4, up(maxa)).
table(up_v_c_maxc_mins_maxa, 15, 11, 4, up(maxa)).
table(up_v_c_maxc_maxs_maxa, 15, 6, 4, up(maxa)).
table(up_v_c_s_mins_maxa, 15, 19, 4, up(maxa)).
table(up_v_c_s_maxs_maxa, 15, 19, 4, up(maxa)).
table(up_v_c_mins_maxs_maxa, 15, 11, 4, up(maxa)).
table(up_v_minc_maxc_s_maxa, 15, 18, 4, up(maxa)).
table(up_v_minc_maxc_mins_maxa, 15, 19, 4, up(maxa)).
table(up_v_minc_maxc_maxs_maxa, 15, 9, 4, up(maxa)).
table(up_v_minc_s_mins_maxa, 15, 18, 4, up(maxa)).
table(up_v_minc_s_maxs_maxa, 15, 18, 4, up(maxa)).
table(up_v_minc_mins_maxs_maxa, 15, 18, 4, up(maxa)).
table(up_v_maxc_s_mins_maxa, 15, 9, 4, up(maxa)).
table(up_v_maxc_s_maxs_maxa, 15, 8, 4, up(maxa)).
table(up_v_maxc_mins_maxs_maxa, 15, 6, 4, up(maxa)).
table(up_v_s_mins_maxs_maxa, 15, 19, 4, up(maxa)).
table(up_c_minc_maxc_s_maxa, 15, 18, 4, up(maxa)).
table(up_c_minc_maxc_mins_maxa, 15, 19, 4, up(maxa)).
table(up_c_minc_maxc_maxs_maxa, 15, 9, 4, up(maxa)).
table(up_c_minc_s_maxs_maxa, 15, 9, 4, up(maxa)).
table(up_c_maxc_s_mins_maxa, 15, 10, 4, up(maxa)).
table(up_c_maxc_s_maxs_maxa, 15, 6, 4, up(maxa)).
table(up_c_maxc_mins_maxs_maxa, 15, 6, 4, up(maxa)).
table(up_c_s_mins_maxs_maxa, 15, 19, 4, up(maxa)).
table(up_minc_maxc_s_mins_maxa, 15, 6, 4, up(maxa)).
table(up_minc_maxc_s_maxs_maxa, 15, 6, 4, up(maxa)).
table(up_minc_s_mins_maxs_maxa, 15, 18, 4, up(maxa)).
table(up_maxc_s_mins_maxs_maxa, 15, 5, 4, up(maxa)).
table(up_v_c_minc_maxc_s_maxa, 15, 18, 5, up(maxa)).
table(up_v_c_minc_maxc_mins_maxa, 15, 19, 5, up(maxa)).
table(up_v_c_minc_maxc_maxs_maxa, 15, 9, 5, up(maxa)).
table(up_v_c_minc_s_mins_maxa, 15, 10, 5, up(maxa)).
table(up_v_c_minc_s_maxs_maxa, 15, 9, 5, up(maxa)).
table(up_v_c_minc_mins_maxs_maxa, 15, 8, 5, up(maxa)).
table(up_v_c_maxc_s_mins_maxa, 15, 9, 5, up(maxa)).
table(up_v_c_maxc_s_maxs_maxa, 15, 6, 5, up(maxa)).
table(up_v_c_maxc_mins_maxs_maxa, 15, 6, 5, up(maxa)).
table(up_v_c_s_mins_maxs_maxa, 15, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_mins_maxa, 15, 14, 5, up(maxa)).
table(up_v_minc_maxc_s_maxs_maxa, 15, 11, 5, up(maxa)).
table(up_v_minc_maxc_mins_maxs_maxa, 15, 9, 5, up(maxa)).
table(up_v_minc_s_mins_maxs_maxa, 15, 18, 5, up(maxa)).
table(up_v_maxc_s_mins_maxs_maxa, 15, 8, 5, up(maxa)).
table(up_c_minc_maxc_s_mins_maxa, 15, 14, 5, up(maxa)).
table(up_c_minc_maxc_s_maxs_maxa, 15, 10, 5, up(maxa)).
table(up_c_minc_maxc_mins_maxs_maxa, 15, 9, 5, up(maxa)).
table(up_c_minc_s_mins_maxs_maxa, 15, 10, 5, up(maxa)).
table(up_c_maxc_s_mins_maxs_maxa, 15, 7, 5, up(maxa)).
table(up_minc_maxc_s_mins_maxs_maxa, 15, 7, 5, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxa, 15, 14, 6, up(maxa)).
table(up_v_c_minc_maxc_s_maxs_maxa, 15, 10, 6, up(maxa)).
table(up_v_c_minc_maxc_mins_maxs_maxa, 15, 9, 6, up(maxa)).
table(up_v_c_minc_s_mins_maxs_maxa, 15, 10, 6, up(maxa)).
table(up_v_c_maxc_s_mins_maxs_maxa, 15, 7, 6, up(maxa)).
table(up_v_minc_maxc_s_mins_maxs_maxa, 15, 11, 6, up(maxa)).
table(up_c_minc_maxc_s_mins_maxs_maxa, 15, 10, 6, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxs_maxa, 15, 10, 7, up(maxa)).
table(low_c_v, 16, 18, 1, low(v)).
table(low_minc_v, 16, 11, 1, low(v)).
table(low_maxc_v, 16, 11, 1, low(v)).
table(low_s_v, 16, 7, 1, low(v)).
table(low_mins_v, 16, 18, 1, low(v)).
table(low_maxs_v, 16, 18, 1, low(v)).
table(low_c_minc_v, 16, 11, 2, low(v)).
table(low_c_maxc_v, 16, 11, 2, low(v)).
table(low_c_s_v, 16, 8, 2, low(v)).
table(low_c_mins_v, 16, 18, 2, low(v)).
table(low_c_maxs_v, 16, 18, 2, low(v)).
table(low_minc_maxc_v, 16, 11, 2, low(v)).
table(low_minc_s_v, 16, 3, 2, low(v)).
table(low_minc_mins_v, 16, 12, 2, low(v)).
table(low_minc_maxs_v, 16, 12, 2, low(v)).
table(low_maxc_s_v, 16, 3, 2, low(v)).
table(low_maxc_mins_v, 16, 12, 2, low(v)).
table(low_maxc_maxs_v, 16, 13, 2, low(v)).
table(low_s_mins_v, 16, 7, 2, low(v)).
table(low_s_maxs_v, 16, 7, 2, low(v)).
table(low_mins_maxs_v, 16, 12, 2, low(v)).
table(low_c_minc_maxc_v, 16, 11, 3, low(v)).
table(low_c_minc_s_v, 16, 6, 3, low(v)).
table(low_c_minc_mins_v, 16, 12, 3, low(v)).
table(low_c_minc_maxs_v, 16, 12, 3, low(v)).
table(low_c_maxc_s_v, 16, 4, 3, low(v)).
table(low_c_maxc_mins_v, 16, 12, 3, low(v)).
table(low_c_maxc_maxs_v, 16, 13, 3, low(v)).
table(low_c_s_mins_v, 16, 8, 3, low(v)).
table(low_c_s_maxs_v, 16, 8, 3, low(v)).
table(low_c_mins_maxs_v, 16, 18, 3, low(v)).
table(low_minc_maxc_s_v, 16, 5, 3, low(v)).
table(low_minc_maxc_mins_v, 16, 12, 3, low(v)).
table(low_minc_maxc_maxs_v, 16, 12, 3, low(v)).
table(low_minc_s_mins_v, 16, 4, 3, low(v)).
table(low_minc_s_maxs_v, 16, 4, 3, low(v)).
table(low_minc_mins_maxs_v, 16, 13, 3, low(v)).
table(low_maxc_s_mins_v, 16, 4, 3, low(v)).
table(low_maxc_s_maxs_v, 16, 4, 3, low(v)).
table(low_maxc_mins_maxs_v, 16, 14, 3, low(v)).
table(low_s_mins_maxs_v, 16, 7, 3, low(v)).
table(low_c_minc_maxc_s_v, 16, 7, 4, low(v)).
table(low_c_minc_maxc_mins_v, 16, 12, 4, low(v)).
table(low_c_minc_maxc_maxs_v, 16, 12, 4, low(v)).
table(low_c_minc_s_mins_v, 16, 7, 4, low(v)).
table(low_c_minc_s_maxs_v, 16, 7, 4, low(v)).
table(low_c_minc_mins_maxs_v, 16, 13, 4, low(v)).
table(low_c_maxc_s_mins_v, 16, 5, 4, low(v)).
table(low_c_maxc_s_maxs_v, 16, 5, 4, low(v)).
table(low_c_maxc_mins_maxs_v, 16, 14, 4, low(v)).
table(low_c_s_mins_maxs_v, 16, 8, 4, low(v)).
table(low_minc_maxc_s_mins_v, 16, 6, 4, low(v)).
table(low_minc_maxc_s_maxs_v, 16, 6, 4, low(v)).
table(low_minc_maxc_mins_maxs_v, 16, 13, 4, low(v)).
table(low_minc_s_mins_maxs_v, 16, 5, 4, low(v)).
table(low_maxc_s_mins_maxs_v, 16, 5, 4, low(v)).
table(low_c_minc_maxc_s_mins_v, 16, 8, 5, low(v)).
table(low_c_minc_maxc_s_maxs_v, 16, 8, 5, low(v)).
table(low_c_minc_maxc_mins_maxs_v, 16, 13, 5, low(v)).
table(low_c_minc_s_mins_maxs_v, 16, 8, 5, low(v)).
table(low_c_maxc_s_mins_maxs_v, 16, 6, 5, low(v)).
table(low_minc_maxc_s_mins_maxs_v, 16, 7, 5, low(v)).
table(low_c_minc_maxc_s_mins_maxs_v, 16, 9, 6, low(v)).
table(up_c_maxc_v, 16, 3, 2, up(v)).
table(up_maxc_s_v, 16, 3, 2, up(v)).
table(up_s_maxs_v, 16, 3, 2, up(v)).
table(up_c_minc_maxc_v, 16, 6, 3, up(v)).
table(up_c_maxc_s_v, 16, 4, 3, up(v)).
table(up_c_maxc_mins_v, 16, 4, 3, up(v)).
table(up_c_maxc_maxs_v, 16, 4, 3, up(v)).
table(up_c_s_maxs_v, 16, 4, 3, up(v)).
table(up_minc_maxc_s_v, 16, 5, 3, up(v)).
table(up_minc_s_maxs_v, 16, 4, 3, up(v)).
table(up_maxc_s_mins_v, 16, 4, 3, up(v)).
table(up_maxc_s_maxs_v, 16, 4, 3, up(v)).
table(up_s_mins_maxs_v, 16, 4, 3, up(v)).
table(up_c_minc_maxc_s_v, 16, 7, 4, up(v)).
table(up_c_minc_maxc_mins_v, 16, 7, 4, up(v)).
table(up_c_minc_maxc_maxs_v, 16, 7, 4, up(v)).
table(up_c_minc_s_maxs_v, 16, 7, 4, up(v)).
table(up_c_maxc_s_mins_v, 16, 5, 4, up(v)).
table(up_c_maxc_s_maxs_v, 16, 5, 4, up(v)).
table(up_c_maxc_mins_maxs_v, 16, 5, 4, up(v)).
table(up_c_s_mins_maxs_v, 16, 5, 4, up(v)).
table(up_minc_maxc_s_mins_v, 16, 6, 4, up(v)).
table(up_minc_maxc_s_maxs_v, 16, 6, 4, up(v)).
table(up_minc_s_mins_maxs_v, 16, 5, 4, up(v)).
table(up_maxc_s_mins_maxs_v, 16, 5, 4, up(v)).
table(up_c_minc_maxc_s_mins_v, 16, 8, 5, up(v)).
table(up_c_minc_maxc_s_maxs_v, 16, 8, 5, up(v)).
table(up_c_minc_maxc_mins_maxs_v, 16, 8, 5, up(v)).
table(up_c_minc_s_mins_maxs_v, 16, 8, 5, up(v)).
table(up_c_maxc_s_mins_maxs_v, 16, 6, 5, up(v)).
table(up_minc_maxc_s_mins_maxs_v, 16, 7, 5, up(v)).
table(up_c_minc_maxc_s_mins_maxs_v, 16, 9, 6, up(v)).
table(low_v_c, 16, 11, 1, low(c)).
table(low_minc_c, 16, 11, 1, low(c)).
table(low_maxc_c, 16, 11, 1, low(c)).
table(low_s_c, 16, 8, 1, low(c)).
table(low_mins_c, 16, 8, 1, low(c)).
table(low_maxs_c, 16, 8, 1, low(c)).
table(low_v_minc_c, 16, 11, 2, low(c)).
table(low_v_maxc_c, 16, 3, 2, low(c)).
table(low_v_s_c, 16, 13, 2, low(c)).
table(low_v_mins_c, 16, 12, 2, low(c)).
table(low_v_maxs_c, 16, 13, 2, low(c)).
table(low_minc_maxc_c, 16, 11, 2, low(c)).
table(low_minc_s_c, 16, 5, 2, low(c)).
table(low_minc_mins_c, 16, 9, 2, low(c)).
table(low_minc_maxs_c, 16, 9, 2, low(c)).
table(low_maxc_s_c, 16, 3, 2, low(c)).
table(low_maxc_mins_c, 16, 12, 2, low(c)).
table(low_maxc_maxs_c, 16, 13, 2, low(c)).
table(low_s_mins_c, 16, 10, 2, low(c)).
table(low_s_maxs_c, 16, 11, 2, low(c)).
table(low_mins_maxs_c, 16, 9, 2, low(c)).
table(low_v_minc_maxc_c, 16, 6, 3, low(c)).
table(low_v_minc_s_c, 16, 12, 3, low(c)).
table(low_v_minc_mins_c, 16, 12, 3, low(c)).
table(low_v_minc_maxs_c, 16, 12, 3, low(c)).
table(low_v_maxc_s_c, 16, 4, 3, low(c)).
table(low_v_maxc_mins_c, 16, 4, 3, low(c)).
table(low_v_maxc_maxs_c, 16, 4, 3, low(c)).
table(low_v_s_mins_c, 16, 14, 3, low(c)).
table(low_v_s_maxs_c, 16, 14, 3, low(c)).
table(low_v_mins_maxs_c, 16, 14, 3, low(c)).
table(low_minc_maxc_s_c, 16, 5, 3, low(c)).
table(low_minc_maxc_mins_c, 16, 12, 3, low(c)).
table(low_minc_maxc_maxs_c, 16, 12, 3, low(c)).
table(low_minc_s_mins_c, 16, 6, 3, low(c)).
table(low_minc_s_maxs_c, 16, 7, 3, low(c)).
table(low_minc_mins_maxs_c, 16, 7, 3, low(c)).
table(low_maxc_s_mins_c, 16, 4, 3, low(c)).
table(low_maxc_s_maxs_c, 16, 4, 3, low(c)).
table(low_maxc_mins_maxs_c, 16, 6, 3, low(c)).
table(low_s_mins_maxs_c, 16, 12, 3, low(c)).
table(low_v_minc_maxc_s_c, 16, 7, 4, low(c)).
table(low_v_minc_maxc_mins_c, 16, 7, 4, low(c)).
table(low_v_minc_maxc_maxs_c, 16, 7, 4, low(c)).
table(low_v_minc_s_mins_c, 16, 13, 4, low(c)).
table(low_v_minc_s_maxs_c, 16, 13, 4, low(c)).
table(low_v_minc_mins_maxs_c, 16, 13, 4, low(c)).
table(low_v_maxc_s_mins_c, 16, 5, 4, low(c)).
table(low_v_maxc_s_maxs_c, 16, 5, 4, low(c)).
table(low_v_maxc_mins_maxs_c, 16, 5, 4, low(c)).
table(low_v_s_mins_maxs_c, 16, 15, 4, low(c)).
table(low_minc_maxc_s_mins_c, 16, 6, 4, low(c)).
table(low_minc_maxc_s_maxs_c, 16, 6, 4, low(c)).
table(low_minc_maxc_mins_maxs_c, 16, 8, 4, low(c)).
table(low_minc_s_mins_maxs_c, 16, 8, 4, low(c)).
table(low_maxc_s_mins_maxs_c, 16, 5, 4, low(c)).
table(low_v_minc_maxc_s_mins_c, 16, 8, 5, low(c)).
table(low_v_minc_maxc_s_maxs_c, 16, 8, 5, low(c)).
table(low_v_minc_maxc_mins_maxs_c, 16, 8, 5, low(c)).
table(low_v_minc_s_mins_maxs_c, 16, 14, 5, low(c)).
table(low_v_maxc_s_mins_maxs_c, 16, 6, 5, low(c)).
table(low_minc_maxc_s_mins_maxs_c, 16, 7, 5, low(c)).
table(low_v_minc_maxc_s_mins_maxs_c, 16, 9, 6, low(c)).
table(up_v_c, 16, 18, 1, up(c)).
table(up_s_c, 16, 3, 1, up(c)).
table(up_v_minc_c, 16, 6, 2, up(c)).
table(up_v_maxc_c, 16, 11, 2, up(c)).
table(up_v_s_c, 16, 4, 2, up(c)).
table(up_v_mins_c, 16, 10, 2, up(c)).
table(up_v_maxs_c, 16, 18, 2, up(c)).
table(up_minc_s_c, 16, 3, 2, up(c)).
table(up_maxc_s_c, 16, 4, 2, up(c)).
table(up_s_mins_c, 16, 5, 2, up(c)).
table(up_s_maxs_c, 16, 5, 2, up(c)).
table(up_v_minc_maxc_c, 16, 7, 3, up(c)).
table(up_v_minc_s_c, 16, 6, 3, up(c)).
table(up_v_minc_mins_c, 16, 7, 3, up(c)).
table(up_v_minc_maxs_c, 16, 7, 3, up(c)).
table(up_v_maxc_s_c, 16, 5, 3, up(c)).
table(up_v_maxc_mins_c, 16, 5, 3, up(c)).
table(up_v_maxc_maxs_c, 16, 13, 3, up(c)).
table(up_v_s_mins_c, 16, 8, 3, up(c)).
table(up_v_s_maxs_c, 16, 6, 3, up(c)).
table(up_v_mins_maxs_c, 16, 12, 3, up(c)).
table(up_minc_maxc_s_c, 16, 5, 3, up(c)).
table(up_minc_s_mins_c, 16, 4, 3, up(c)).
table(up_minc_s_maxs_c, 16, 4, 3, up(c)).
table(up_maxc_s_mins_c, 16, 5, 3, up(c)).
table(up_maxc_s_maxs_c, 16, 5, 3, up(c)).
table(up_s_mins_maxs_c, 16, 8, 3, up(c)).
table(up_v_minc_maxc_s_c, 16, 7, 4, up(c)).
table(up_v_minc_maxc_mins_c, 16, 8, 4, up(c)).
table(up_v_minc_maxc_maxs_c, 16, 8, 4, up(c)).
table(up_v_minc_s_mins_c, 16, 7, 4, up(c)).
table(up_v_minc_s_maxs_c, 16, 7, 4, up(c)).
table(up_v_minc_mins_maxs_c, 16, 8, 4, up(c)).
table(up_v_maxc_s_mins_c, 16, 6, 4, up(c)).
table(up_v_maxc_s_maxs_c, 16, 6, 4, up(c)).
table(up_v_maxc_mins_maxs_c, 16, 7, 4, up(c)).
table(up_v_s_mins_maxs_c, 16, 10, 4, up(c)).
table(up_minc_maxc_s_mins_c, 16, 6, 4, up(c)).
table(up_minc_maxc_s_maxs_c, 16, 6, 4, up(c)).
table(up_minc_s_mins_maxs_c, 16, 5, 4, up(c)).
table(up_maxc_s_mins_maxs_c, 16, 6, 4, up(c)).
table(up_v_minc_maxc_s_mins_c, 16, 8, 5, up(c)).
table(up_v_minc_maxc_s_maxs_c, 16, 8, 5, up(c)).
table(up_v_minc_maxc_mins_maxs_c, 16, 9, 5, up(c)).
table(up_v_minc_s_mins_maxs_c, 16, 8, 5, up(c)).
table(up_v_maxc_s_mins_maxs_c, 16, 7, 5, up(c)).
table(up_minc_maxc_s_mins_maxs_c, 16, 7, 5, up(c)).
table(up_v_minc_maxc_s_mins_maxs_c, 16, 9, 6, up(c)).
table(low_v_minc, 16, 3, 1, low(minc)).
table(low_c_minc, 16, 3, 1, low(minc)).
table(low_maxc_minc, 16, 3, 1, low(minc)).
table(low_s_minc, 16, 3, 1, low(minc)).
table(low_mins_minc, 16, 2, 1, low(minc)).
table(low_maxs_minc, 16, 3, 1, low(minc)).
table(low_v_c_minc, 16, 5, 2, low(minc)).
table(low_v_maxc_minc, 16, 4, 2, low(minc)).
table(low_v_s_minc, 16, 4, 2, low(minc)).
table(low_v_mins_minc, 16, 3, 2, low(minc)).
table(low_v_maxs_minc, 16, 4, 2, low(minc)).
table(low_c_maxc_minc, 16, 4, 2, low(minc)).
table(low_c_s_minc, 16, 4, 2, low(minc)).
table(low_c_mins_minc, 16, 3, 2, low(minc)).
table(low_c_maxs_minc, 16, 4, 2, low(minc)).
table(low_maxc_s_minc, 16, 4, 2, low(minc)).
table(low_maxc_mins_minc, 16, 4, 2, low(minc)).
table(low_maxc_maxs_minc, 16, 4, 2, low(minc)).
table(low_s_mins_minc, 16, 3, 2, low(minc)).
table(low_s_maxs_minc, 16, 4, 2, low(minc)).
table(low_mins_maxs_minc, 16, 3, 2, low(minc)).
table(low_v_c_maxc_minc, 16, 6, 3, low(minc)).
table(low_v_c_s_minc, 16, 6, 3, low(minc)).
table(low_v_c_mins_minc, 16, 6, 3, low(minc)).
table(low_v_c_maxs_minc, 16, 6, 3, low(minc)).
table(low_v_maxc_s_minc, 16, 5, 3, low(minc)).
table(low_v_maxc_mins_minc, 16, 5, 3, low(minc)).
table(low_v_maxc_maxs_minc, 16, 5, 3, low(minc)).
table(low_v_s_mins_minc, 16, 4, 3, low(minc)).
table(low_v_s_maxs_minc, 16, 5, 3, low(minc)).
table(low_v_mins_maxs_minc, 16, 4, 3, low(minc)).
table(low_c_maxc_s_minc, 16, 5, 3, low(minc)).
table(low_c_maxc_mins_minc, 16, 5, 3, low(minc)).
table(low_c_maxc_maxs_minc, 16, 5, 3, low(minc)).
table(low_c_s_mins_minc, 16, 4, 3, low(minc)).
table(low_c_s_maxs_minc, 16, 6, 3, low(minc)).
table(low_c_mins_maxs_minc, 16, 4, 3, low(minc)).
table(low_maxc_s_mins_minc, 16, 5, 3, low(minc)).
table(low_maxc_s_maxs_minc, 16, 5, 3, low(minc)).
table(low_maxc_mins_maxs_minc, 16, 5, 3, low(minc)).
table(low_s_mins_maxs_minc, 16, 4, 3, low(minc)).
table(low_v_c_maxc_s_minc, 16, 7, 4, low(minc)).
table(low_v_c_maxc_mins_minc, 16, 7, 4, low(minc)).
table(low_v_c_maxc_maxs_minc, 16, 7, 4, low(minc)).
table(low_v_c_s_mins_minc, 16, 7, 4, low(minc)).
table(low_v_c_s_maxs_minc, 16, 7, 4, low(minc)).
table(low_v_c_mins_maxs_minc, 16, 7, 4, low(minc)).
table(low_v_maxc_s_mins_minc, 16, 6, 4, low(minc)).
table(low_v_maxc_s_maxs_minc, 16, 6, 4, low(minc)).
table(low_v_maxc_mins_maxs_minc, 16, 6, 4, low(minc)).
table(low_v_s_mins_maxs_minc, 16, 5, 4, low(minc)).
table(low_c_maxc_s_mins_minc, 16, 6, 4, low(minc)).
table(low_c_maxc_s_maxs_minc, 16, 6, 4, low(minc)).
table(low_c_maxc_mins_maxs_minc, 16, 6, 4, low(minc)).
table(low_c_s_mins_maxs_minc, 16, 6, 4, low(minc)).
table(low_maxc_s_mins_maxs_minc, 16, 6, 4, low(minc)).
table(low_v_c_maxc_s_mins_minc, 16, 8, 5, low(minc)).
table(low_v_c_maxc_s_maxs_minc, 16, 8, 5, low(minc)).
table(low_v_c_maxc_mins_maxs_minc, 16, 8, 5, low(minc)).
table(low_v_c_s_mins_maxs_minc, 16, 8, 5, low(minc)).
table(low_v_maxc_s_mins_maxs_minc, 16, 7, 5, low(minc)).
table(low_c_maxc_s_mins_maxs_minc, 16, 7, 5, low(minc)).
table(low_v_c_maxc_s_mins_maxs_minc, 16, 9, 6, low(minc)).
table(up_v_minc, 16, 11, 1, up(minc)).
table(up_maxc_minc, 16, 6, 1, up(minc)).
table(up_v_c_minc, 16, 5, 2, up(minc)).
table(up_v_maxc_minc, 16, 4, 2, up(minc)).
table(up_v_s_minc, 16, 13, 2, up(minc)).
table(up_v_mins_minc, 16, 12, 2, up(minc)).
table(up_v_maxs_minc, 16, 13, 2, up(minc)).
table(up_c_maxc_minc, 16, 4, 2, up(minc)).
table(up_maxc_s_minc, 16, 4, 2, up(minc)).
table(up_maxc_mins_minc, 16, 4, 2, up(minc)).
table(up_maxc_maxs_minc, 16, 7, 2, up(minc)).
table(up_s_maxs_minc, 16, 14, 2, up(minc)).
table(up_v_c_maxc_minc, 16, 6, 3, up(minc)).
table(up_v_c_s_minc, 16, 6, 3, up(minc)).
table(up_v_c_mins_minc, 16, 6, 3, up(minc)).
table(up_v_c_maxs_minc, 16, 6, 3, up(minc)).
table(up_v_maxc_s_minc, 16, 5, 3, up(minc)).
table(up_v_maxc_mins_minc, 16, 5, 3, up(minc)).
table(up_v_maxc_maxs_minc, 16, 5, 3, up(minc)).
table(up_v_s_mins_minc, 16, 14, 3, up(minc)).
table(up_v_s_maxs_minc, 16, 14, 3, up(minc)).
table(up_v_mins_maxs_minc, 16, 14, 3, up(minc)).
table(up_c_maxc_s_minc, 16, 5, 3, up(minc)).
table(up_c_maxc_mins_minc, 16, 5, 3, up(minc)).
table(up_c_maxc_maxs_minc, 16, 5, 3, up(minc)).
table(up_c_s_maxs_minc, 16, 4, 3, up(minc)).
table(up_maxc_s_mins_minc, 16, 5, 3, up(minc)).
table(up_maxc_s_maxs_minc, 16, 5, 3, up(minc)).
table(up_maxc_mins_maxs_minc, 16, 5, 3, up(minc)).
table(up_s_mins_maxs_minc, 16, 15, 3, up(minc)).
table(up_v_c_maxc_s_minc, 16, 7, 4, up(minc)).
table(up_v_c_maxc_mins_minc, 16, 7, 4, up(minc)).
table(up_v_c_maxc_maxs_minc, 16, 7, 4, up(minc)).
table(up_v_c_s_mins_minc, 16, 7, 4, up(minc)).
table(up_v_c_s_maxs_minc, 16, 7, 4, up(minc)).
table(up_v_c_mins_maxs_minc, 16, 7, 4, up(minc)).
table(up_v_maxc_s_mins_minc, 16, 6, 4, up(minc)).
table(up_v_maxc_s_maxs_minc, 16, 6, 4, up(minc)).
table(up_v_maxc_mins_maxs_minc, 16, 6, 4, up(minc)).
table(up_v_s_mins_maxs_minc, 16, 15, 4, up(minc)).
table(up_c_maxc_s_mins_minc, 16, 6, 4, up(minc)).
table(up_c_maxc_s_maxs_minc, 16, 6, 4, up(minc)).
table(up_c_maxc_mins_maxs_minc, 16, 6, 4, up(minc)).
table(up_c_s_mins_maxs_minc, 16, 5, 4, up(minc)).
table(up_maxc_s_mins_maxs_minc, 16, 6, 4, up(minc)).
table(up_v_c_maxc_s_mins_minc, 16, 8, 5, up(minc)).
table(up_v_c_maxc_s_maxs_minc, 16, 8, 5, up(minc)).
table(up_v_c_maxc_mins_maxs_minc, 16, 8, 5, up(minc)).
table(up_v_c_s_mins_maxs_minc, 16, 8, 5, up(minc)).
table(up_v_maxc_s_mins_maxs_minc, 16, 7, 5, up(minc)).
table(up_c_maxc_s_mins_maxs_minc, 16, 7, 5, up(minc)).
table(up_v_c_maxc_s_mins_maxs_minc, 16, 9, 6, up(minc)).
table(low_v_maxc, 16, 18, 1, low(maxc)).
table(low_c_maxc, 16, 18, 1, low(maxc)).
table(low_minc_maxc, 16, 6, 1, low(maxc)).
table(low_s_maxc, 16, 18, 1, low(maxc)).
table(low_mins_maxc, 16, 9, 1, low(maxc)).
table(low_maxs_maxc, 16, 2, 1, low(maxc)).
table(low_v_c_maxc, 16, 3, 2, low(maxc)).
table(low_v_minc_maxc, 16, 6, 2, low(maxc)).
table(low_v_s_maxc, 16, 4, 2, low(maxc)).
table(low_v_mins_maxc, 16, 9, 2, low(maxc)).
table(low_v_maxs_maxc, 16, 3, 2, low(maxc)).
table(low_c_minc_maxc, 16, 11, 2, low(maxc)).
table(low_c_s_maxc, 16, 4, 2, low(maxc)).
table(low_c_mins_maxc, 16, 18, 2, low(maxc)).
table(low_c_maxs_maxc, 16, 3, 2, low(maxc)).
table(low_minc_s_maxc, 16, 7, 2, low(maxc)).
table(low_minc_mins_maxc, 16, 4, 2, low(maxc)).
table(low_minc_maxs_maxc, 16, 4, 2, low(maxc)).
table(low_s_mins_maxc, 16, 18, 2, low(maxc)).
table(low_s_maxs_maxc, 16, 3, 2, low(maxc)).
table(low_mins_maxs_maxc, 16, 3, 2, low(maxc)).
table(low_v_c_minc_maxc, 16, 6, 3, low(maxc)).
table(low_v_c_s_maxc, 16, 4, 3, low(maxc)).
table(low_v_c_mins_maxc, 16, 4, 3, low(maxc)).
table(low_v_c_maxs_maxc, 16, 4, 3, low(maxc)).
table(low_v_minc_s_maxc, 16, 5, 3, low(maxc)).
table(low_v_minc_mins_maxc, 16, 7, 3, low(maxc)).
table(low_v_minc_maxs_maxc, 16, 5, 3, low(maxc)).
table(low_v_s_mins_maxc, 16, 5, 3, low(maxc)).
table(low_v_s_maxs_maxc, 16, 4, 3, low(maxc)).
table(low_v_mins_maxs_maxc, 16, 4, 3, low(maxc)).
table(low_c_minc_s_maxc, 16, 5, 3, low(maxc)).
table(low_c_minc_mins_maxc, 16, 6, 3, low(maxc)).
table(low_c_minc_maxs_maxc, 16, 5, 3, low(maxc)).
table(low_c_s_mins_maxc, 16, 4, 3, low(maxc)).
table(low_c_s_maxs_maxc, 16, 4, 3, low(maxc)).
table(low_c_mins_maxs_maxc, 16, 4, 3, low(maxc)).
table(low_minc_s_mins_maxc, 16, 5, 3, low(maxc)).
table(low_minc_s_maxs_maxc, 16, 5, 3, low(maxc)).
table(low_minc_mins_maxs_maxc, 16, 5, 3, low(maxc)).
table(low_s_mins_maxs_maxc, 16, 4, 3, low(maxc)).
table(low_v_c_minc_s_maxc, 16, 7, 4, low(maxc)).
table(low_v_c_minc_mins_maxc, 16, 7, 4, low(maxc)).
table(low_v_c_minc_maxs_maxc, 16, 7, 4, low(maxc)).
table(low_v_c_s_mins_maxc, 16, 5, 4, low(maxc)).
table(low_v_c_s_maxs_maxc, 16, 5, 4, low(maxc)).
table(low_v_c_mins_maxs_maxc, 16, 5, 4, low(maxc)).
table(low_v_minc_s_mins_maxc, 16, 6, 4, low(maxc)).
table(low_v_minc_s_maxs_maxc, 16, 6, 4, low(maxc)).
table(low_v_minc_mins_maxs_maxc, 16, 6, 4, low(maxc)).
table(low_v_s_mins_maxs_maxc, 16, 5, 4, low(maxc)).
table(low_c_minc_s_mins_maxc, 16, 6, 4, low(maxc)).
table(low_c_minc_s_maxs_maxc, 16, 6, 4, low(maxc)).
table(low_c_minc_mins_maxs_maxc, 16, 6, 4, low(maxc)).
table(low_c_s_mins_maxs_maxc, 16, 5, 4, low(maxc)).
table(low_minc_s_mins_maxs_maxc, 16, 6, 4, low(maxc)).
table(low_v_c_minc_s_mins_maxc, 16, 8, 5, low(maxc)).
table(low_v_c_minc_s_maxs_maxc, 16, 8, 5, low(maxc)).
table(low_v_c_minc_mins_maxs_maxc, 16, 8, 5, low(maxc)).
table(low_v_c_s_mins_maxs_maxc, 16, 6, 5, low(maxc)).
table(low_v_minc_s_mins_maxs_maxc, 16, 7, 5, low(maxc)).
table(low_c_minc_s_mins_maxs_maxc, 16, 7, 5, low(maxc)).
table(low_v_c_minc_s_mins_maxs_maxc, 16, 9, 6, low(maxc)).
table(up_v_maxc, 16, 11, 1, up(maxc)).
table(up_v_c_maxc, 16, 11, 2, up(maxc)).
table(up_v_minc_maxc, 16, 11, 2, up(maxc)).
table(up_v_s_maxc, 16, 13, 2, up(maxc)).
table(up_v_mins_maxc, 16, 12, 2, up(maxc)).
table(up_v_maxs_maxc, 16, 13, 2, up(maxc)).
table(up_v_c_minc_maxc, 16, 11, 3, up(maxc)).
table(up_v_c_s_maxc, 16, 13, 3, up(maxc)).
table(up_v_c_mins_maxc, 16, 12, 3, up(maxc)).
table(up_v_c_maxs_maxc, 16, 13, 3, up(maxc)).
table(up_v_minc_s_maxc, 16, 12, 3, up(maxc)).
table(up_v_minc_mins_maxc, 16, 12, 3, up(maxc)).
table(up_v_minc_maxs_maxc, 16, 12, 3, up(maxc)).
table(up_v_s_mins_maxc, 16, 14, 3, up(maxc)).
table(up_v_s_maxs_maxc, 16, 14, 3, up(maxc)).
table(up_v_mins_maxs_maxc, 16, 14, 3, up(maxc)).
table(up_v_c_minc_s_maxc, 16, 12, 4, up(maxc)).
table(up_v_c_minc_mins_maxc, 16, 12, 4, up(maxc)).
table(up_v_c_minc_maxs_maxc, 16, 12, 4, up(maxc)).
table(up_v_c_s_mins_maxc, 16, 14, 4, up(maxc)).
table(up_v_c_s_maxs_maxc, 16, 6, 4, up(maxc)).
table(up_v_c_mins_maxs_maxc, 16, 10, 4, up(maxc)).
table(up_v_minc_s_mins_maxc, 16, 13, 4, up(maxc)).
table(up_v_minc_s_maxs_maxc, 16, 13, 4, up(maxc)).
table(up_v_minc_mins_maxs_maxc, 16, 13, 4, up(maxc)).
table(up_v_s_mins_maxs_maxc, 16, 15, 4, up(maxc)).
table(up_v_c_minc_s_mins_maxc, 16, 13, 5, up(maxc)).
table(up_v_c_minc_s_maxs_maxc, 16, 8, 5, up(maxc)).
table(up_v_c_minc_mins_maxs_maxc, 16, 9, 5, up(maxc)).
table(up_v_c_s_mins_maxs_maxc, 16, 10, 5, up(maxc)).
table(up_v_minc_s_mins_maxs_maxc, 16, 14, 5, up(maxc)).
table(up_v_c_minc_s_mins_maxs_maxc, 16, 9, 6, up(maxc)).
table(low_v_s, 16, 18, 1, low(s)).
table(low_c_s, 16, 3, 1, low(s)).
table(low_minc_s, 16, 18, 1, low(s)).
table(low_maxc_s, 16, 18, 1, low(s)).
table(low_mins_s, 16, 18, 1, low(s)).
table(low_maxs_s, 16, 18, 1, low(s)).
table(low_v_c_s, 16, 4, 2, low(s)).
table(low_v_minc_s, 16, 18, 2, low(s)).
table(low_v_maxc_s, 16, 6, 2, low(s)).
table(low_v_mins_s, 16, 9, 2, low(s)).
table(low_v_maxs_s, 16, 3, 2, low(s)).
table(low_c_minc_s, 16, 5, 2, low(s)).
table(low_c_maxc_s, 16, 5, 2, low(s)).
table(low_c_mins_s, 16, 5, 2, low(s)).
table(low_c_maxs_s, 16, 5, 2, low(s)).
table(low_minc_maxc_s, 16, 18, 2, low(s)).
table(low_minc_mins_s, 16, 12, 2, low(s)).
table(low_minc_maxs_s, 16, 14, 2, low(s)).
table(low_maxc_mins_s, 16, 7, 2, low(s)).
table(low_maxc_maxs_s, 16, 14, 2, low(s)).
table(low_mins_maxs_s, 16, 12, 2, low(s)).
table(low_v_c_minc_s, 16, 8, 3, low(s)).
table(low_v_c_maxc_s, 16, 6, 3, low(s)).
table(low_v_c_mins_s, 16, 8, 3, low(s)).
table(low_v_c_maxs_s, 16, 4, 3, low(s)).
table(low_v_minc_maxc_s, 16, 10, 3, low(s)).
table(low_v_minc_mins_s, 16, 13, 3, low(s)).
table(low_v_minc_maxs_s, 16, 4, 3, low(s)).
table(low_v_maxc_mins_s, 16, 4, 3, low(s)).
table(low_v_maxc_maxs_s, 16, 4, 3, low(s)).
table(low_v_mins_maxs_s, 16, 4, 3, low(s)).
table(low_c_minc_maxc_s, 16, 8, 3, low(s)).
table(low_c_minc_mins_s, 16, 4, 3, low(s)).
table(low_c_minc_maxs_s, 16, 5, 3, low(s)).
table(low_c_maxc_mins_s, 16, 8, 3, low(s)).
table(low_c_maxc_maxs_s, 16, 5, 3, low(s)).
table(low_c_mins_maxs_s, 16, 8, 3, low(s)).
table(low_minc_maxc_mins_s, 16, 13, 3, low(s)).
table(low_minc_maxc_maxs_s, 16, 14, 3, low(s)).
table(low_minc_mins_maxs_s, 16, 8, 3, low(s)).
table(low_maxc_mins_maxs_s, 16, 9, 3, low(s)).
table(low_v_c_minc_maxc_s, 16, 10, 4, low(s)).
table(low_v_c_minc_mins_s, 16, 7, 4, low(s)).
table(low_v_c_minc_maxs_s, 16, 7, 4, low(s)).
table(low_v_c_maxc_mins_s, 16, 5, 4, low(s)).
table(low_v_c_maxc_maxs_s, 16, 5, 4, low(s)).
table(low_v_c_mins_maxs_s, 16, 5, 4, low(s)).
table(low_v_minc_maxc_mins_s, 16, 8, 4, low(s)).
table(low_v_minc_maxc_maxs_s, 16, 6, 4, low(s)).
table(low_v_minc_mins_maxs_s, 16, 5, 4, low(s)).
table(low_v_maxc_mins_maxs_s, 16, 5, 4, low(s)).
table(low_c_minc_maxc_mins_s, 16, 6, 4, low(s)).
table(low_c_minc_maxc_maxs_s, 16, 7, 4, low(s)).
table(low_c_minc_mins_maxs_s, 16, 5, 4, low(s)).
table(low_c_maxc_mins_maxs_s, 16, 6, 4, low(s)).
table(low_minc_maxc_mins_maxs_s, 16, 9, 4, low(s)).
table(low_v_c_minc_maxc_mins_s, 16, 8, 5, low(s)).
table(low_v_c_minc_maxc_maxs_s, 16, 8, 5, low(s)).
table(low_v_c_minc_mins_maxs_s, 16, 8, 5, low(s)).
table(low_v_c_maxc_mins_maxs_s, 16, 6, 5, low(s)).
table(low_v_minc_maxc_mins_maxs_s, 16, 7, 5, low(s)).
table(low_c_minc_maxc_mins_maxs_s, 16, 7, 5, low(s)).
table(low_v_c_minc_maxc_mins_maxs_s, 16, 9, 6, low(s)).
table(up_v_s, 16, 7, 1, up(s)).
table(up_v_c_s, 16, 8, 2, up(s)).
table(up_v_minc_s, 16, 8, 2, up(s)).
table(up_v_maxc_s, 16, 8, 2, up(s)).
table(up_v_mins_s, 16, 4, 2, up(s)).
table(up_v_maxs_s, 16, 7, 2, up(s)).
table(up_c_maxc_s, 16, 9, 2, up(s)).
table(up_v_c_minc_s, 16, 11, 3, up(s)).
table(up_v_c_maxc_s, 16, 9, 3, up(s)).
table(up_v_c_mins_s, 16, 5, 3, up(s)).
table(up_v_c_maxs_s, 16, 8, 3, up(s)).
table(up_v_minc_maxc_s, 16, 10, 3, up(s)).
table(up_v_minc_mins_s, 16, 5, 3, up(s)).
table(up_v_minc_maxs_s, 16, 8, 3, up(s)).
table(up_v_maxc_mins_s, 16, 5, 3, up(s)).
table(up_v_maxc_maxs_s, 16, 8, 3, up(s)).
table(up_v_mins_maxs_s, 16, 5, 3, up(s)).
table(up_c_minc_maxc_s, 16, 12, 3, up(s)).
table(up_c_maxc_mins_s, 16, 5, 3, up(s)).
table(up_c_maxc_maxs_s, 16, 9, 3, up(s)).
table(up_v_c_minc_maxc_s, 16, 12, 4, up(s)).
table(up_v_c_minc_mins_s, 16, 8, 4, up(s)).
table(up_v_c_minc_maxs_s, 16, 11, 4, up(s)).
table(up_v_c_maxc_mins_s, 16, 6, 4, up(s)).
table(up_v_c_maxc_maxs_s, 16, 9, 4, up(s)).
table(up_v_c_mins_maxs_s, 16, 6, 4, up(s)).
table(up_v_minc_maxc_mins_s, 16, 7, 4, up(s)).
table(up_v_minc_maxc_maxs_s, 16, 10, 4, up(s)).
table(up_v_minc_mins_maxs_s, 16, 6, 4, up(s)).
table(up_v_maxc_mins_maxs_s, 16, 6, 4, up(s)).
table(up_c_minc_maxc_mins_s, 16, 7, 4, up(s)).
table(up_c_minc_maxc_maxs_s, 16, 12, 4, up(s)).
table(up_c_maxc_mins_maxs_s, 16, 6, 4, up(s)).
table(up_v_c_minc_maxc_mins_s, 16, 9, 5, up(s)).
table(up_v_c_minc_maxc_maxs_s, 16, 12, 5, up(s)).
table(up_v_c_minc_mins_maxs_s, 16, 9, 5, up(s)).
table(up_v_c_maxc_mins_maxs_s, 16, 7, 5, up(s)).
table(up_v_minc_maxc_mins_maxs_s, 16, 8, 5, up(s)).
table(up_c_minc_maxc_mins_maxs_s, 16, 8, 5, up(s)).
table(up_v_c_minc_maxc_mins_maxs_s, 16, 10, 6, up(s)).
table(low_v_mins, 16, 2, 1, low(mins)).
table(low_c_mins, 16, 2, 1, low(mins)).
table(low_minc_mins, 16, 2, 1, low(mins)).
table(low_maxc_mins, 16, 2, 1, low(mins)).
table(low_s_mins, 16, 2, 1, low(mins)).
table(low_maxs_mins, 16, 2, 1, low(mins)).
table(low_v_c_mins, 16, 3, 2, low(mins)).
table(low_v_minc_mins, 16, 3, 2, low(mins)).
table(low_v_maxc_mins, 16, 3, 2, low(mins)).
table(low_v_s_mins, 16, 3, 2, low(mins)).
table(low_v_maxs_mins, 16, 3, 2, low(mins)).
table(low_c_minc_mins, 16, 3, 2, low(mins)).
table(low_c_maxc_mins, 16, 3, 2, low(mins)).
table(low_c_s_mins, 16, 3, 2, low(mins)).
table(low_c_maxs_mins, 16, 3, 2, low(mins)).
table(low_minc_maxc_mins, 16, 4, 2, low(mins)).
table(low_minc_s_mins, 16, 3, 2, low(mins)).
table(low_minc_maxs_mins, 16, 3, 2, low(mins)).
table(low_maxc_s_mins, 16, 3, 2, low(mins)).
table(low_maxc_maxs_mins, 16, 3, 2, low(mins)).
table(low_s_maxs_mins, 16, 3, 2, low(mins)).
table(low_v_c_minc_mins, 16, 6, 3, low(mins)).
table(low_v_c_maxc_mins, 16, 4, 3, low(mins)).
table(low_v_c_s_mins, 16, 4, 3, low(mins)).
table(low_v_c_maxs_mins, 16, 4, 3, low(mins)).
table(low_v_minc_maxc_mins, 16, 5, 3, low(mins)).
table(low_v_minc_s_mins, 16, 4, 3, low(mins)).
table(low_v_minc_maxs_mins, 16, 4, 3, low(mins)).
table(low_v_maxc_s_mins, 16, 4, 3, low(mins)).
table(low_v_maxc_maxs_mins, 16, 4, 3, low(mins)).
table(low_v_s_maxs_mins, 16, 4, 3, low(mins)).
table(low_c_minc_maxc_mins, 16, 5, 3, low(mins)).
table(low_c_minc_s_mins, 16, 4, 3, low(mins)).
table(low_c_minc_maxs_mins, 16, 4, 3, low(mins)).
table(low_c_maxc_s_mins, 16, 4, 3, low(mins)).
table(low_c_maxc_maxs_mins, 16, 4, 3, low(mins)).
table(low_c_s_maxs_mins, 16, 4, 3, low(mins)).
table(low_minc_maxc_s_mins, 16, 5, 3, low(mins)).
table(low_minc_maxc_maxs_mins, 16, 5, 3, low(mins)).
table(low_minc_s_maxs_mins, 16, 4, 3, low(mins)).
table(low_maxc_s_maxs_mins, 16, 4, 3, low(mins)).
table(low_v_c_minc_maxc_mins, 16, 7, 4, low(mins)).
table(low_v_c_minc_s_mins, 16, 7, 4, low(mins)).
table(low_v_c_minc_maxs_mins, 16, 7, 4, low(mins)).
table(low_v_c_maxc_s_mins, 16, 5, 4, low(mins)).
table(low_v_c_maxc_maxs_mins, 16, 5, 4, low(mins)).
table(low_v_c_s_maxs_mins, 16, 5, 4, low(mins)).
table(low_v_minc_maxc_s_mins, 16, 6, 4, low(mins)).
table(low_v_minc_maxc_maxs_mins, 16, 6, 4, low(mins)).
table(low_v_minc_s_maxs_mins, 16, 5, 4, low(mins)).
table(low_v_maxc_s_maxs_mins, 16, 5, 4, low(mins)).
table(low_c_minc_maxc_s_mins, 16, 6, 4, low(mins)).
table(low_c_minc_maxc_maxs_mins, 16, 6, 4, low(mins)).
table(low_c_minc_s_maxs_mins, 16, 5, 4, low(mins)).
table(low_c_maxc_s_maxs_mins, 16, 5, 4, low(mins)).
table(low_minc_maxc_s_maxs_mins, 16, 6, 4, low(mins)).
table(low_v_c_minc_maxc_s_mins, 16, 8, 5, low(mins)).
table(low_v_c_minc_maxc_maxs_mins, 16, 8, 5, low(mins)).
table(low_v_c_minc_s_maxs_mins, 16, 8, 5, low(mins)).
table(low_v_c_maxc_s_maxs_mins, 16, 6, 5, low(mins)).
table(low_v_minc_maxc_s_maxs_mins, 16, 7, 5, low(mins)).
table(low_c_minc_maxc_s_maxs_mins, 16, 7, 5, low(mins)).
table(low_v_c_minc_maxc_s_maxs_mins, 16, 9, 6, low(mins)).
table(up_v_mins, 16, 18, 1, up(mins)).
table(up_maxc_mins, 16, 9, 1, up(mins)).
table(up_maxs_mins, 16, 2, 1, up(mins)).
table(up_v_c_mins, 16, 6, 2, up(mins)).
table(up_v_minc_mins, 16, 3, 2, up(mins)).
table(up_v_maxc_mins, 16, 5, 2, up(mins)).
table(up_v_s_mins, 16, 3, 2, up(mins)).
table(up_v_maxs_mins, 16, 3, 2, up(mins)).
table(up_c_maxc_mins, 16, 5, 2, up(mins)).
table(up_c_maxs_mins, 16, 3, 2, up(mins)).
table(up_minc_maxc_mins, 16, 4, 2, up(mins)).
table(up_minc_maxs_mins, 16, 3, 2, up(mins)).
table(up_maxc_s_mins, 16, 3, 2, up(mins)).
table(up_maxc_maxs_mins, 16, 3, 2, up(mins)).
table(up_s_maxs_mins, 16, 3, 2, up(mins)).
table(up_v_c_minc_mins, 16, 6, 3, up(mins)).
table(up_v_c_maxc_mins, 16, 7, 3, up(mins)).
table(up_v_c_s_mins, 16, 4, 3, up(mins)).
table(up_v_c_maxs_mins, 16, 4, 3, up(mins)).
table(up_v_minc_maxc_mins, 16, 5, 3, up(mins)).
table(up_v_minc_s_mins, 16, 4, 3, up(mins)).
table(up_v_minc_maxs_mins, 16, 4, 3, up(mins)).
table(up_v_maxc_s_mins, 16, 4, 3, up(mins)).
table(up_v_maxc_maxs_mins, 16, 4, 3, up(mins)).
table(up_v_s_maxs_mins, 16, 4, 3, up(mins)).
table(up_c_minc_maxc_mins, 16, 5, 3, up(mins)).
table(up_c_minc_maxs_mins, 16, 4, 3, up(mins)).
table(up_c_maxc_s_mins, 16, 4, 3, up(mins)).
table(up_c_maxc_maxs_mins, 16, 4, 3, up(mins)).
table(up_c_s_maxs_mins, 16, 4, 3, up(mins)).
table(up_minc_maxc_s_mins, 16, 5, 3, up(mins)).
table(up_minc_maxc_maxs_mins, 16, 5, 3, up(mins)).
table(up_minc_s_maxs_mins, 16, 4, 3, up(mins)).
table(up_maxc_s_maxs_mins, 16, 4, 3, up(mins)).
table(up_v_c_minc_maxc_mins, 16, 7, 4, up(mins)).
table(up_v_c_minc_s_mins, 16, 7, 4, up(mins)).
table(up_v_c_minc_maxs_mins, 16, 7, 4, up(mins)).
table(up_v_c_maxc_s_mins, 16, 5, 4, up(mins)).
table(up_v_c_maxc_maxs_mins, 16, 5, 4, up(mins)).
table(up_v_c_s_maxs_mins, 16, 5, 4, up(mins)).
table(up_v_minc_maxc_s_mins, 16, 6, 4, up(mins)).
table(up_v_minc_maxc_maxs_mins, 16, 6, 4, up(mins)).
table(up_v_minc_s_maxs_mins, 16, 5, 4, up(mins)).
table(up_v_maxc_s_maxs_mins, 16, 5, 4, up(mins)).
table(up_c_minc_maxc_s_mins, 16, 6, 4, up(mins)).
table(up_c_minc_maxc_maxs_mins, 16, 6, 4, up(mins)).
table(up_c_minc_s_maxs_mins, 16, 5, 4, up(mins)).
table(up_c_maxc_s_maxs_mins, 16, 5, 4, up(mins)).
table(up_minc_maxc_s_maxs_mins, 16, 6, 4, up(mins)).
table(up_v_c_minc_maxc_s_mins, 16, 8, 5, up(mins)).
table(up_v_c_minc_maxc_maxs_mins, 16, 8, 5, up(mins)).
table(up_v_c_minc_s_maxs_mins, 16, 8, 5, up(mins)).
table(up_v_c_maxc_s_maxs_mins, 16, 6, 5, up(mins)).
table(up_v_minc_maxc_s_maxs_mins, 16, 7, 5, up(mins)).
table(up_c_minc_maxc_s_maxs_mins, 16, 7, 5, up(mins)).
table(up_v_c_minc_maxc_s_maxs_mins, 16, 9, 6, up(mins)).
table(low_v_maxs, 16, 7, 1, low(maxs)).
table(low_c_maxs, 16, 4, 1, low(maxs)).
table(low_minc_maxs, 16, 4, 1, low(maxs)).
table(low_maxc_maxs, 16, 4, 1, low(maxs)).
table(low_s_maxs, 16, 7, 1, low(maxs)).
table(low_mins_maxs, 16, 2, 1, low(maxs)).
table(low_v_c_maxs, 16, 8, 2, low(maxs)).
table(low_v_minc_maxs, 16, 8, 2, low(maxs)).
table(low_v_maxc_maxs, 16, 8, 2, low(maxs)).
table(low_v_s_maxs, 16, 3, 2, low(maxs)).
table(low_v_mins_maxs, 16, 3, 2, low(maxs)).
table(low_c_minc_maxs, 16, 5, 2, low(maxs)).
table(low_c_maxc_maxs, 16, 5, 2, low(maxs)).
table(low_c_s_maxs, 16, 8, 2, low(maxs)).
table(low_c_mins_maxs, 16, 3, 2, low(maxs)).
table(low_minc_maxc_maxs, 16, 6, 2, low(maxs)).
table(low_minc_s_maxs, 16, 3, 2, low(maxs)).
table(low_minc_mins_maxs, 16, 3, 2, low(maxs)).
table(low_maxc_s_maxs, 16, 3, 2, low(maxs)).
table(low_maxc_mins_maxs, 16, 3, 2, low(maxs)).
table(low_s_mins_maxs, 16, 7, 2, low(maxs)).
table(low_v_c_minc_maxs, 16, 11, 3, low(maxs)).
table(low_v_c_maxc_maxs, 16, 9, 3, low(maxs)).
table(low_v_c_s_maxs, 16, 4, 3, low(maxs)).
table(low_v_c_mins_maxs, 16, 4, 3, low(maxs)).
table(low_v_minc_maxc_maxs, 16, 10, 3, low(maxs)).
table(low_v_minc_s_maxs, 16, 4, 3, low(maxs)).
table(low_v_minc_mins_maxs, 16, 4, 3, low(maxs)).
table(low_v_maxc_s_maxs, 16, 4, 3, low(maxs)).
table(low_v_maxc_mins_maxs, 16, 4, 3, low(maxs)).
table(low_v_s_mins_maxs, 16, 4, 3, low(maxs)).
table(low_c_minc_maxc_maxs, 16, 7, 3, low(maxs)).
table(low_c_minc_s_maxs, 16, 4, 3, low(maxs)).
table(low_c_minc_mins_maxs, 16, 4, 3, low(maxs)).
table(low_c_maxc_s_maxs, 16, 4, 3, low(maxs)).
table(low_c_maxc_mins_maxs, 16, 4, 3, low(maxs)).
table(low_c_s_mins_maxs, 16, 8, 3, low(maxs)).
table(low_minc_maxc_s_maxs, 16, 5, 3, low(maxs)).
table(low_minc_maxc_mins_maxs, 16, 5, 3, low(maxs)).
table(low_minc_s_mins_maxs, 16, 4, 3, low(maxs)).
table(low_maxc_s_mins_maxs, 16, 4, 3, low(maxs)).
table(low_v_c_minc_maxc_maxs, 16, 12, 4, low(maxs)).
table(low_v_c_minc_s_maxs, 16, 7, 4, low(maxs)).
table(low_v_c_minc_mins_maxs, 16, 7, 4, low(maxs)).
table(low_v_c_maxc_s_maxs, 16, 5, 4, low(maxs)).
table(low_v_c_maxc_mins_maxs, 16, 5, 4, low(maxs)).
table(low_v_c_s_mins_maxs, 16, 5, 4, low(maxs)).
table(low_v_minc_maxc_s_maxs, 16, 6, 4, low(maxs)).
table(low_v_minc_maxc_mins_maxs, 16, 6, 4, low(maxs)).
table(low_v_minc_s_mins_maxs, 16, 5, 4, low(maxs)).
table(low_v_maxc_s_mins_maxs, 16, 5, 4, low(maxs)).
table(low_c_minc_maxc_s_maxs, 16, 6, 4, low(maxs)).
table(low_c_minc_maxc_mins_maxs, 16, 6, 4, low(maxs)).
table(low_c_minc_s_mins_maxs, 16, 5, 4, low(maxs)).
table(low_c_maxc_s_mins_maxs, 16, 5, 4, low(maxs)).
table(low_minc_maxc_s_mins_maxs, 16, 6, 4, low(maxs)).
table(low_v_c_minc_maxc_s_maxs, 16, 8, 5, low(maxs)).
table(low_v_c_minc_maxc_mins_maxs, 16, 8, 5, low(maxs)).
table(low_v_c_minc_s_mins_maxs, 16, 8, 5, low(maxs)).
table(low_v_c_maxc_s_mins_maxs, 16, 6, 5, low(maxs)).
table(low_v_minc_maxc_s_mins_maxs, 16, 7, 5, low(maxs)).
table(low_c_minc_maxc_s_mins_maxs, 16, 7, 5, low(maxs)).
table(low_v_c_minc_maxc_s_mins_maxs, 16, 9, 6, low(maxs)).
table(up_v_maxs, 16, 18, 1, up(maxs)).
table(up_maxc_maxs, 16, 2, 1, up(maxs)).
table(up_v_c_maxs, 16, 18, 2, up(maxs)).
table(up_v_minc_maxs, 16, 12, 2, up(maxs)).
table(up_v_maxc_maxs, 16, 3, 2, up(maxs)).
table(up_v_s_maxs, 16, 7, 2, up(maxs)).
table(up_v_mins_maxs, 16, 9, 2, up(maxs)).
table(up_c_maxc_maxs, 16, 3, 2, up(maxs)).
table(up_minc_maxc_maxs, 16, 4, 2, up(maxs)).
table(up_maxc_s_maxs, 16, 3, 2, up(maxs)).
table(up_maxc_mins_maxs, 16, 3, 2, up(maxs)).
table(up_v_c_minc_maxs, 16, 12, 3, up(maxs)).
table(up_v_c_maxc_maxs, 16, 4, 3, up(maxs)).
table(up_v_c_s_maxs, 16, 8, 3, up(maxs)).
table(up_v_c_mins_maxs, 16, 18, 3, up(maxs)).
table(up_v_minc_maxc_maxs, 16, 5, 3, up(maxs)).
table(up_v_minc_s_maxs, 16, 5, 3, up(maxs)).
table(up_v_minc_mins_maxs, 16, 14, 3, up(maxs)).
table(up_v_maxc_s_maxs, 16, 4, 3, up(maxs)).
table(up_v_maxc_mins_maxs, 16, 4, 3, up(maxs)).
table(up_v_s_mins_maxs, 16, 7, 3, up(maxs)).
table(up_c_minc_maxc_maxs, 16, 5, 3, up(maxs)).
table(up_c_maxc_s_maxs, 16, 4, 3, up(maxs)).
table(up_c_maxc_mins_maxs, 16, 4, 3, up(maxs)).
table(up_minc_maxc_s_maxs, 16, 5, 3, up(maxs)).
table(up_minc_maxc_mins_maxs, 16, 5, 3, up(maxs)).
table(up_maxc_s_mins_maxs, 16, 4, 3, up(maxs)).
table(up_v_c_minc_maxc_maxs, 16, 7, 4, up(maxs)).
table(up_v_c_minc_s_maxs, 16, 7, 4, up(maxs)).
table(up_v_c_minc_mins_maxs, 16, 13, 4, up(maxs)).
table(up_v_c_maxc_s_maxs, 16, 5, 4, up(maxs)).
table(up_v_c_maxc_mins_maxs, 16, 5, 4, up(maxs)).
table(up_v_c_s_mins_maxs, 16, 8, 4, up(maxs)).
table(up_v_minc_maxc_s_maxs, 16, 6, 4, up(maxs)).
table(up_v_minc_maxc_mins_maxs, 16, 6, 4, up(maxs)).
table(up_v_minc_s_mins_maxs, 16, 5, 4, up(maxs)).
table(up_v_maxc_s_mins_maxs, 16, 5, 4, up(maxs)).
table(up_c_minc_maxc_s_maxs, 16, 6, 4, up(maxs)).
table(up_c_minc_maxc_mins_maxs, 16, 6, 4, up(maxs)).
table(up_c_maxc_s_mins_maxs, 16, 5, 4, up(maxs)).
table(up_minc_maxc_s_mins_maxs, 16, 6, 4, up(maxs)).
table(up_v_c_minc_maxc_s_maxs, 16, 8, 5, up(maxs)).
table(up_v_c_minc_maxc_mins_maxs, 16, 8, 5, up(maxs)).
table(up_v_c_minc_s_mins_maxs, 16, 8, 5, up(maxs)).
table(up_v_c_maxc_s_mins_maxs, 16, 6, 5, up(maxs)).
table(up_v_minc_maxc_s_mins_maxs, 16, 7, 5, up(maxs)).
table(up_c_minc_maxc_s_mins_maxs, 16, 7, 5, up(maxs)).
table(up_v_c_minc_maxc_s_mins_maxs, 16, 9, 6, up(maxs)).
table(low_v_mina, 16, 12, 1, low(mina)).
table(low_c_mina, 16, 7, 1, low(mina)).
table(low_minc_mina, 16, 19, 1, low(mina)).
table(low_maxc_mina, 16, 19, 1, low(mina)).
table(low_s_mina, 16, 12, 1, low(mina)).
table(low_mins_mina, 16, 11, 1, low(mina)).
table(low_maxs_mina, 16, 11, 1, low(mina)).
table(low_v_c_mina, 16, 10, 2, low(mina)).
table(low_v_minc_mina, 16, 11, 2, low(mina)).
table(low_v_maxc_mina, 16, 11, 2, low(mina)).
table(low_v_s_mina, 16, 4, 2, low(mina)).
table(low_v_mins_mina, 16, 4, 2, low(mina)).
table(low_v_maxs_mina, 16, 10, 2, low(mina)).
table(low_c_minc_mina, 16, 8, 2, low(mina)).
table(low_c_maxc_mina, 16, 6, 2, low(mina)).
table(low_c_s_mina, 16, 10, 2, low(mina)).
table(low_c_mins_mina, 16, 7, 2, low(mina)).
table(low_c_maxs_mina, 16, 5, 2, low(mina)).
table(low_minc_maxc_mina, 16, 19, 2, low(mina)).
table(low_minc_s_mina, 16, 11, 2, low(mina)).
table(low_minc_mins_mina, 16, 19, 2, low(mina)).
table(low_minc_maxs_mina, 16, 19, 2, low(mina)).
table(low_maxc_s_mina, 16, 11, 2, low(mina)).
table(low_maxc_mins_mina, 16, 19, 2, low(mina)).
table(low_maxc_maxs_mina, 16, 19, 2, low(mina)).
table(low_s_mins_mina, 16, 12, 2, low(mina)).
table(low_s_maxs_mina, 16, 10, 2, low(mina)).
table(low_mins_maxs_mina, 16, 6, 2, low(mina)).
table(low_v_c_minc_mina, 16, 13, 3, low(mina)).
table(low_v_c_maxc_mina, 16, 11, 3, low(mina)).
table(low_v_c_s_mina, 16, 4, 3, low(mina)).
table(low_v_c_mins_mina, 16, 7, 3, low(mina)).
table(low_v_c_maxs_mina, 16, 10, 3, low(mina)).
table(low_v_minc_maxc_mina, 16, 12, 3, low(mina)).
table(low_v_minc_s_mina, 16, 4, 3, low(mina)).
table(low_v_minc_mins_mina, 16, 4, 3, low(mina)).
table(low_v_minc_maxs_mina, 16, 11, 3, low(mina)).
table(low_v_maxc_s_mina, 16, 4, 3, low(mina)).
table(low_v_maxc_mins_mina, 16, 6, 3, low(mina)).
table(low_v_maxc_maxs_mina, 16, 10, 3, low(mina)).
table(low_v_s_mins_mina, 16, 5, 3, low(mina)).
table(low_v_s_maxs_mina, 16, 5, 3, low(mina)).
table(low_v_mins_maxs_mina, 16, 5, 3, low(mina)).
table(low_c_minc_maxc_mina, 16, 8, 3, low(mina)).
table(low_c_minc_s_mina, 16, 11, 3, low(mina)).
table(low_c_minc_mins_mina, 16, 8, 3, low(mina)).
table(low_c_minc_maxs_mina, 16, 6, 3, low(mina)).
table(low_c_maxc_s_mina, 16, 8, 3, low(mina)).
table(low_c_maxc_mins_mina, 16, 6, 3, low(mina)).
table(low_c_maxc_maxs_mina, 16, 6, 3, low(mina)).
table(low_c_s_mins_mina, 16, 10, 3, low(mina)).
table(low_c_s_maxs_mina, 16, 10, 3, low(mina)).
table(low_c_mins_maxs_mina, 16, 5, 3, low(mina)).
table(low_minc_maxc_s_mina, 16, 10, 3, low(mina)).
table(low_minc_maxc_mins_mina, 16, 16, 3, low(mina)).
table(low_minc_maxc_maxs_mina, 16, 18, 3, low(mina)).
table(low_minc_s_mins_mina, 16, 6, 3, low(mina)).
table(low_minc_s_maxs_mina, 16, 8, 3, low(mina)).
table(low_minc_mins_maxs_mina, 16, 17, 3, low(mina)).
table(low_maxc_s_mins_mina, 16, 8, 3, low(mina)).
table(low_maxc_s_maxs_mina, 16, 7, 3, low(mina)).
table(low_maxc_mins_maxs_mina, 16, 8, 3, low(mina)).
table(low_s_mins_maxs_mina, 16, 10, 3, low(mina)).
table(low_v_c_minc_maxc_mina, 16, 14, 4, low(mina)).
table(low_v_c_minc_s_mina, 16, 8, 4, low(mina)).
table(low_v_c_minc_mins_mina, 16, 10, 4, low(mina)).
table(low_v_c_minc_maxs_mina, 16, 13, 4, low(mina)).
table(low_v_c_maxc_s_mina, 16, 5, 4, low(mina)).
table(low_v_c_maxc_mins_mina, 16, 8, 4, low(mina)).
table(low_v_c_maxc_maxs_mina, 16, 11, 4, low(mina)).
table(low_v_c_s_mins_mina, 16, 5, 4, low(mina)).
table(low_v_c_s_maxs_mina, 16, 5, 4, low(mina)).
table(low_v_c_mins_maxs_mina, 16, 8, 4, low(mina)).
table(low_v_minc_maxc_s_mina, 16, 6, 4, low(mina)).
table(low_v_minc_maxc_mins_mina, 16, 6, 4, low(mina)).
table(low_v_minc_maxc_maxs_mina, 16, 12, 4, low(mina)).
table(low_v_minc_s_mins_mina, 16, 5, 4, low(mina)).
table(low_v_minc_s_maxs_mina, 16, 5, 4, low(mina)).
table(low_v_minc_mins_maxs_mina, 16, 6, 4, low(mina)).
table(low_v_maxc_s_mins_mina, 16, 5, 4, low(mina)).
table(low_v_maxc_s_maxs_mina, 16, 5, 4, low(mina)).
table(low_v_maxc_mins_maxs_mina, 16, 5, 4, low(mina)).
table(low_v_s_mins_maxs_mina, 16, 6, 4, low(mina)).
table(low_c_minc_maxc_s_mina, 16, 9, 4, low(mina)).
table(low_c_minc_maxc_mins_mina, 16, 7, 4, low(mina)).
table(low_c_minc_maxc_maxs_mina, 16, 8, 4, low(mina)).
table(low_c_minc_s_mins_mina, 16, 10, 4, low(mina)).
table(low_c_minc_s_maxs_mina, 16, 10, 4, low(mina)).
table(low_c_minc_mins_maxs_mina, 16, 6, 4, low(mina)).
table(low_c_maxc_s_mins_mina, 16, 6, 4, low(mina)).
table(low_c_maxc_s_maxs_mina, 16, 6, 4, low(mina)).
table(low_c_maxc_mins_maxs_mina, 16, 6, 4, low(mina)).
table(low_c_s_mins_maxs_mina, 16, 10, 4, low(mina)).
table(low_minc_maxc_s_mins_mina, 16, 7, 4, low(mina)).
table(low_minc_maxc_s_maxs_mina, 16, 9, 4, low(mina)).
table(low_minc_maxc_mins_maxs_mina, 16, 16, 4, low(mina)).
table(low_minc_s_mins_maxs_mina, 16, 7, 4, low(mina)).
table(low_maxc_s_mins_maxs_mina, 16, 6, 4, low(mina)).
table(low_v_c_minc_maxc_s_mina, 16, 9, 5, low(mina)).
table(low_v_c_minc_maxc_mins_mina, 16, 11, 5, low(mina)).
table(low_v_c_minc_maxc_maxs_mina, 16, 14, 5, low(mina)).
table(low_v_c_minc_s_mins_mina, 16, 8, 5, low(mina)).
table(low_v_c_minc_s_maxs_mina, 16, 8, 5, low(mina)).
table(low_v_c_minc_mins_maxs_mina, 16, 11, 5, low(mina)).
table(low_v_c_maxc_s_mins_mina, 16, 6, 5, low(mina)).
table(low_v_c_maxc_s_maxs_mina, 16, 6, 5, low(mina)).
table(low_v_c_maxc_mins_maxs_mina, 16, 9, 5, low(mina)).
table(low_v_c_s_mins_maxs_mina, 16, 6, 5, low(mina)).
table(low_v_minc_maxc_s_mins_mina, 16, 7, 5, low(mina)).
table(low_v_minc_maxc_s_maxs_mina, 16, 7, 5, low(mina)).
table(low_v_minc_maxc_mins_maxs_mina, 16, 7, 5, low(mina)).
table(low_v_minc_s_mins_maxs_mina, 16, 6, 5, low(mina)).
table(low_v_maxc_s_mins_maxs_mina, 16, 6, 5, low(mina)).
table(low_c_minc_maxc_s_mins_mina, 16, 8, 5, low(mina)).
table(low_c_minc_maxc_s_maxs_mina, 16, 8, 5, low(mina)).
table(low_c_minc_maxc_mins_maxs_mina, 16, 8, 5, low(mina)).
table(low_c_minc_s_mins_maxs_mina, 16, 11, 5, low(mina)).
table(low_c_maxc_s_mins_maxs_mina, 16, 7, 5, low(mina)).
table(low_minc_maxc_s_mins_maxs_mina, 16, 8, 5, low(mina)).
table(low_v_c_minc_maxc_s_mins_mina, 16, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_maxs_mina, 16, 9, 6, low(mina)).
table(low_v_c_minc_maxc_mins_maxs_mina, 16, 12, 6, low(mina)).
table(low_v_c_minc_s_mins_maxs_mina, 16, 9, 6, low(mina)).
table(low_v_c_maxc_s_mins_maxs_mina, 16, 7, 6, low(mina)).
table(low_v_minc_maxc_s_mins_maxs_mina, 16, 8, 6, low(mina)).
table(low_c_minc_maxc_s_mins_maxs_mina, 16, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_mins_maxs_mina, 16, 10, 7, low(mina)).
table(up_v_maxa, 16, 19, 1, up(maxa)).
table(up_v_c_maxa, 16, 19, 2, up(maxa)).
table(up_v_minc_maxa, 16, 19, 2, up(maxa)).
table(up_v_maxc_maxa, 16, 19, 2, up(maxa)).
table(up_v_s_maxa, 16, 19, 2, up(maxa)).
table(up_v_mins_maxa, 16, 19, 2, up(maxa)).
table(up_v_maxs_maxa, 16, 19, 2, up(maxa)).
table(up_c_maxc_maxa, 16, 19, 2, up(maxa)).
table(up_maxc_s_maxa, 16, 18, 2, up(maxa)).
table(up_s_maxs_maxa, 16, 19, 2, up(maxa)).
table(up_v_c_minc_maxa, 16, 19, 3, up(maxa)).
table(up_v_c_maxc_maxa, 16, 19, 3, up(maxa)).
table(up_v_c_s_maxa, 16, 19, 3, up(maxa)).
table(up_v_c_mins_maxa, 16, 19, 3, up(maxa)).
table(up_v_c_maxs_maxa, 16, 7, 3, up(maxa)).
table(up_v_minc_maxc_maxa, 16, 19, 3, up(maxa)).
table(up_v_minc_s_maxa, 16, 18, 3, up(maxa)).
table(up_v_minc_mins_maxa, 16, 19, 3, up(maxa)).
table(up_v_minc_maxs_maxa, 16, 19, 3, up(maxa)).
table(up_v_maxc_s_maxa, 16, 18, 3, up(maxa)).
table(up_v_maxc_mins_maxa, 16, 11, 3, up(maxa)).
table(up_v_maxc_maxs_maxa, 16, 6, 3, up(maxa)).
table(up_v_s_mins_maxa, 16, 19, 3, up(maxa)).
table(up_v_s_maxs_maxa, 16, 19, 3, up(maxa)).
table(up_v_mins_maxs_maxa, 16, 19, 3, up(maxa)).
table(up_c_minc_maxc_maxa, 16, 19, 3, up(maxa)).
table(up_c_maxc_s_maxa, 16, 18, 3, up(maxa)).
table(up_c_maxc_mins_maxa, 16, 19, 3, up(maxa)).
table(up_c_maxc_maxs_maxa, 16, 6, 3, up(maxa)).
table(up_c_s_maxs_maxa, 16, 19, 3, up(maxa)).
table(up_minc_maxc_s_maxa, 16, 9, 3, up(maxa)).
table(up_minc_s_maxs_maxa, 16, 18, 3, up(maxa)).
table(up_maxc_s_mins_maxa, 16, 18, 3, up(maxa)).
table(up_maxc_s_maxs_maxa, 16, 7, 3, up(maxa)).
table(up_s_mins_maxs_maxa, 16, 19, 3, up(maxa)).
table(up_v_c_minc_maxc_maxa, 16, 19, 4, up(maxa)).
table(up_v_c_minc_s_maxa, 16, 18, 4, up(maxa)).
table(up_v_c_minc_mins_maxa, 16, 11, 4, up(maxa)).
table(up_v_c_minc_maxs_maxa, 16, 8, 4, up(maxa)).
table(up_v_c_maxc_s_maxa, 16, 18, 4, up(maxa)).
table(up_v_c_maxc_mins_maxa, 16, 11, 4, up(maxa)).
table(up_v_c_maxc_maxs_maxa, 16, 6, 4, up(maxa)).
table(up_v_c_s_mins_maxa, 16, 19, 4, up(maxa)).
table(up_v_c_s_maxs_maxa, 16, 19, 4, up(maxa)).
table(up_v_c_mins_maxs_maxa, 16, 11, 4, up(maxa)).
table(up_v_minc_maxc_s_maxa, 16, 18, 4, up(maxa)).
table(up_v_minc_maxc_mins_maxa, 16, 14, 4, up(maxa)).
table(up_v_minc_maxc_maxs_maxa, 16, 9, 4, up(maxa)).
table(up_v_minc_s_mins_maxa, 16, 18, 4, up(maxa)).
table(up_v_minc_s_maxs_maxa, 16, 18, 4, up(maxa)).
table(up_v_minc_mins_maxs_maxa, 16, 18, 4, up(maxa)).
table(up_v_maxc_s_mins_maxa, 16, 9, 4, up(maxa)).
table(up_v_maxc_s_maxs_maxa, 16, 8, 4, up(maxa)).
table(up_v_maxc_mins_maxs_maxa, 16, 6, 4, up(maxa)).
table(up_v_s_mins_maxs_maxa, 16, 19, 4, up(maxa)).
table(up_c_minc_maxc_s_maxa, 16, 18, 4, up(maxa)).
table(up_c_minc_maxc_mins_maxa, 16, 14, 4, up(maxa)).
table(up_c_minc_maxc_maxs_maxa, 16, 9, 4, up(maxa)).
table(up_c_minc_s_maxs_maxa, 16, 9, 4, up(maxa)).
table(up_c_maxc_s_mins_maxa, 16, 18, 4, up(maxa)).
table(up_c_maxc_s_maxs_maxa, 16, 7, 4, up(maxa)).
table(up_c_maxc_mins_maxs_maxa, 16, 5, 4, up(maxa)).
table(up_c_s_mins_maxs_maxa, 16, 19, 4, up(maxa)).
table(up_minc_maxc_s_mins_maxa, 16, 7, 4, up(maxa)).
table(up_minc_maxc_s_maxs_maxa, 16, 7, 4, up(maxa)).
table(up_minc_s_mins_maxs_maxa, 16, 18, 4, up(maxa)).
table(up_maxc_s_mins_maxs_maxa, 16, 7, 4, up(maxa)).
table(up_v_c_minc_maxc_s_maxa, 16, 18, 5, up(maxa)).
table(up_v_c_minc_maxc_mins_maxa, 16, 14, 5, up(maxa)).
table(up_v_c_minc_maxc_maxs_maxa, 16, 9, 5, up(maxa)).
table(up_v_c_minc_s_mins_maxa, 16, 10, 5, up(maxa)).
table(up_v_c_minc_s_maxs_maxa, 16, 9, 5, up(maxa)).
table(up_v_c_minc_mins_maxs_maxa, 16, 8, 5, up(maxa)).
table(up_v_c_maxc_s_mins_maxa, 16, 9, 5, up(maxa)).
table(up_v_c_maxc_s_maxs_maxa, 16, 6, 5, up(maxa)).
table(up_v_c_maxc_mins_maxs_maxa, 16, 6, 5, up(maxa)).
table(up_v_c_s_mins_maxs_maxa, 16, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_mins_maxa, 16, 11, 5, up(maxa)).
table(up_v_minc_maxc_s_maxs_maxa, 16, 11, 5, up(maxa)).
table(up_v_minc_maxc_mins_maxs_maxa, 16, 9, 5, up(maxa)).
table(up_v_minc_s_mins_maxs_maxa, 16, 18, 5, up(maxa)).
table(up_v_maxc_s_mins_maxs_maxa, 16, 8, 5, up(maxa)).
table(up_c_minc_maxc_s_mins_maxa, 16, 11, 5, up(maxa)).
table(up_c_minc_maxc_s_maxs_maxa, 16, 10, 5, up(maxa)).
table(up_c_minc_maxc_mins_maxs_maxa, 16, 9, 5, up(maxa)).
table(up_c_minc_s_mins_maxs_maxa, 16, 10, 5, up(maxa)).
table(up_c_maxc_s_mins_maxs_maxa, 16, 6, 5, up(maxa)).
table(up_minc_maxc_s_mins_maxs_maxa, 16, 7, 5, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxa, 16, 11, 6, up(maxa)).
table(up_v_c_minc_maxc_s_maxs_maxa, 16, 10, 6, up(maxa)).
table(up_v_c_minc_maxc_mins_maxs_maxa, 16, 9, 6, up(maxa)).
table(up_v_c_minc_s_mins_maxs_maxa, 16, 10, 6, up(maxa)).
table(up_v_c_maxc_s_mins_maxs_maxa, 16, 7, 6, up(maxa)).
table(up_v_minc_maxc_s_mins_maxs_maxa, 16, 11, 6, up(maxa)).
table(up_c_minc_maxc_s_mins_maxs_maxa, 16, 10, 6, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxs_maxa, 16, 10, 7, up(maxa)).
table(low_c_v, 17, 18, 1, low(v)).
table(low_minc_v, 17, 11, 1, low(v)).
table(low_maxc_v, 17, 11, 1, low(v)).
table(low_s_v, 17, 7, 1, low(v)).
table(low_mins_v, 17, 18, 1, low(v)).
table(low_maxs_v, 17, 18, 1, low(v)).
table(low_c_minc_v, 17, 11, 2, low(v)).
table(low_c_maxc_v, 17, 11, 2, low(v)).
table(low_c_s_v, 17, 8, 2, low(v)).
table(low_c_mins_v, 17, 18, 2, low(v)).
table(low_c_maxs_v, 17, 18, 2, low(v)).
table(low_minc_maxc_v, 17, 11, 2, low(v)).
table(low_minc_s_v, 17, 3, 2, low(v)).
table(low_minc_mins_v, 17, 12, 2, low(v)).
table(low_minc_maxs_v, 17, 12, 2, low(v)).
table(low_maxc_s_v, 17, 3, 2, low(v)).
table(low_maxc_mins_v, 17, 12, 2, low(v)).
table(low_maxc_maxs_v, 17, 13, 2, low(v)).
table(low_s_mins_v, 17, 7, 2, low(v)).
table(low_s_maxs_v, 17, 7, 2, low(v)).
table(low_mins_maxs_v, 17, 12, 2, low(v)).
table(low_c_minc_maxc_v, 17, 11, 3, low(v)).
table(low_c_minc_s_v, 17, 6, 3, low(v)).
table(low_c_minc_mins_v, 17, 12, 3, low(v)).
table(low_c_minc_maxs_v, 17, 12, 3, low(v)).
table(low_c_maxc_s_v, 17, 4, 3, low(v)).
table(low_c_maxc_mins_v, 17, 12, 3, low(v)).
table(low_c_maxc_maxs_v, 17, 13, 3, low(v)).
table(low_c_s_mins_v, 17, 8, 3, low(v)).
table(low_c_s_maxs_v, 17, 8, 3, low(v)).
table(low_c_mins_maxs_v, 17, 18, 3, low(v)).
table(low_minc_maxc_s_v, 17, 5, 3, low(v)).
table(low_minc_maxc_mins_v, 17, 12, 3, low(v)).
table(low_minc_maxc_maxs_v, 17, 12, 3, low(v)).
table(low_minc_s_mins_v, 17, 4, 3, low(v)).
table(low_minc_s_maxs_v, 17, 4, 3, low(v)).
table(low_minc_mins_maxs_v, 17, 13, 3, low(v)).
table(low_maxc_s_mins_v, 17, 4, 3, low(v)).
table(low_maxc_s_maxs_v, 17, 4, 3, low(v)).
table(low_maxc_mins_maxs_v, 17, 14, 3, low(v)).
table(low_s_mins_maxs_v, 17, 7, 3, low(v)).
table(low_c_minc_maxc_s_v, 17, 7, 4, low(v)).
table(low_c_minc_maxc_mins_v, 17, 12, 4, low(v)).
table(low_c_minc_maxc_maxs_v, 17, 12, 4, low(v)).
table(low_c_minc_s_mins_v, 17, 7, 4, low(v)).
table(low_c_minc_s_maxs_v, 17, 7, 4, low(v)).
table(low_c_minc_mins_maxs_v, 17, 13, 4, low(v)).
table(low_c_maxc_s_mins_v, 17, 5, 4, low(v)).
table(low_c_maxc_s_maxs_v, 17, 5, 4, low(v)).
table(low_c_maxc_mins_maxs_v, 17, 14, 4, low(v)).
table(low_c_s_mins_maxs_v, 17, 8, 4, low(v)).
table(low_minc_maxc_s_mins_v, 17, 6, 4, low(v)).
table(low_minc_maxc_s_maxs_v, 17, 6, 4, low(v)).
table(low_minc_maxc_mins_maxs_v, 17, 13, 4, low(v)).
table(low_minc_s_mins_maxs_v, 17, 5, 4, low(v)).
table(low_maxc_s_mins_maxs_v, 17, 5, 4, low(v)).
table(low_c_minc_maxc_s_mins_v, 17, 8, 5, low(v)).
table(low_c_minc_maxc_s_maxs_v, 17, 8, 5, low(v)).
table(low_c_minc_maxc_mins_maxs_v, 17, 13, 5, low(v)).
table(low_c_minc_s_mins_maxs_v, 17, 8, 5, low(v)).
table(low_c_maxc_s_mins_maxs_v, 17, 6, 5, low(v)).
table(low_minc_maxc_s_mins_maxs_v, 17, 7, 5, low(v)).
table(low_c_minc_maxc_s_mins_maxs_v, 17, 9, 6, low(v)).
table(up_c_maxc_v, 17, 3, 2, up(v)).
table(up_maxc_s_v, 17, 3, 2, up(v)).
table(up_s_maxs_v, 17, 3, 2, up(v)).
table(up_c_minc_maxc_v, 17, 6, 3, up(v)).
table(up_c_maxc_s_v, 17, 4, 3, up(v)).
table(up_c_maxc_mins_v, 17, 4, 3, up(v)).
table(up_c_maxc_maxs_v, 17, 4, 3, up(v)).
table(up_c_s_maxs_v, 17, 4, 3, up(v)).
table(up_minc_maxc_s_v, 17, 5, 3, up(v)).
table(up_minc_s_maxs_v, 17, 4, 3, up(v)).
table(up_maxc_s_mins_v, 17, 4, 3, up(v)).
table(up_maxc_s_maxs_v, 17, 4, 3, up(v)).
table(up_s_mins_maxs_v, 17, 4, 3, up(v)).
table(up_c_minc_maxc_s_v, 17, 7, 4, up(v)).
table(up_c_minc_maxc_mins_v, 17, 7, 4, up(v)).
table(up_c_minc_maxc_maxs_v, 17, 7, 4, up(v)).
table(up_c_minc_s_maxs_v, 17, 7, 4, up(v)).
table(up_c_maxc_s_mins_v, 17, 5, 4, up(v)).
table(up_c_maxc_s_maxs_v, 17, 5, 4, up(v)).
table(up_c_maxc_mins_maxs_v, 17, 5, 4, up(v)).
table(up_c_s_mins_maxs_v, 17, 5, 4, up(v)).
table(up_minc_maxc_s_mins_v, 17, 6, 4, up(v)).
table(up_minc_maxc_s_maxs_v, 17, 6, 4, up(v)).
table(up_minc_s_mins_maxs_v, 17, 5, 4, up(v)).
table(up_maxc_s_mins_maxs_v, 17, 5, 4, up(v)).
table(up_c_minc_maxc_s_mins_v, 17, 8, 5, up(v)).
table(up_c_minc_maxc_s_maxs_v, 17, 8, 5, up(v)).
table(up_c_minc_maxc_mins_maxs_v, 17, 8, 5, up(v)).
table(up_c_minc_s_mins_maxs_v, 17, 8, 5, up(v)).
table(up_c_maxc_s_mins_maxs_v, 17, 6, 5, up(v)).
table(up_minc_maxc_s_mins_maxs_v, 17, 7, 5, up(v)).
table(up_c_minc_maxc_s_mins_maxs_v, 17, 9, 6, up(v)).
table(low_v_c, 17, 11, 1, low(c)).
table(low_minc_c, 17, 11, 1, low(c)).
table(low_maxc_c, 17, 11, 1, low(c)).
table(low_s_c, 17, 8, 1, low(c)).
table(low_mins_c, 17, 8, 1, low(c)).
table(low_maxs_c, 17, 8, 1, low(c)).
table(low_v_minc_c, 17, 11, 2, low(c)).
table(low_v_maxc_c, 17, 3, 2, low(c)).
table(low_v_s_c, 17, 13, 2, low(c)).
table(low_v_mins_c, 17, 12, 2, low(c)).
table(low_v_maxs_c, 17, 13, 2, low(c)).
table(low_minc_maxc_c, 17, 11, 2, low(c)).
table(low_minc_s_c, 17, 5, 2, low(c)).
table(low_minc_mins_c, 17, 9, 2, low(c)).
table(low_minc_maxs_c, 17, 9, 2, low(c)).
table(low_maxc_s_c, 17, 3, 2, low(c)).
table(low_maxc_mins_c, 17, 12, 2, low(c)).
table(low_maxc_maxs_c, 17, 13, 2, low(c)).
table(low_s_mins_c, 17, 10, 2, low(c)).
table(low_s_maxs_c, 17, 11, 2, low(c)).
table(low_mins_maxs_c, 17, 9, 2, low(c)).
table(low_v_minc_maxc_c, 17, 6, 3, low(c)).
table(low_v_minc_s_c, 17, 12, 3, low(c)).
table(low_v_minc_mins_c, 17, 12, 3, low(c)).
table(low_v_minc_maxs_c, 17, 12, 3, low(c)).
table(low_v_maxc_s_c, 17, 4, 3, low(c)).
table(low_v_maxc_mins_c, 17, 4, 3, low(c)).
table(low_v_maxc_maxs_c, 17, 4, 3, low(c)).
table(low_v_s_mins_c, 17, 14, 3, low(c)).
table(low_v_s_maxs_c, 17, 14, 3, low(c)).
table(low_v_mins_maxs_c, 17, 14, 3, low(c)).
table(low_minc_maxc_s_c, 17, 5, 3, low(c)).
table(low_minc_maxc_mins_c, 17, 12, 3, low(c)).
table(low_minc_maxc_maxs_c, 17, 12, 3, low(c)).
table(low_minc_s_mins_c, 17, 6, 3, low(c)).
table(low_minc_s_maxs_c, 17, 7, 3, low(c)).
table(low_minc_mins_maxs_c, 17, 7, 3, low(c)).
table(low_maxc_s_mins_c, 17, 4, 3, low(c)).
table(low_maxc_s_maxs_c, 17, 4, 3, low(c)).
table(low_maxc_mins_maxs_c, 17, 6, 3, low(c)).
table(low_s_mins_maxs_c, 17, 12, 3, low(c)).
table(low_v_minc_maxc_s_c, 17, 7, 4, low(c)).
table(low_v_minc_maxc_mins_c, 17, 7, 4, low(c)).
table(low_v_minc_maxc_maxs_c, 17, 7, 4, low(c)).
table(low_v_minc_s_mins_c, 17, 13, 4, low(c)).
table(low_v_minc_s_maxs_c, 17, 13, 4, low(c)).
table(low_v_minc_mins_maxs_c, 17, 13, 4, low(c)).
table(low_v_maxc_s_mins_c, 17, 5, 4, low(c)).
table(low_v_maxc_s_maxs_c, 17, 5, 4, low(c)).
table(low_v_maxc_mins_maxs_c, 17, 5, 4, low(c)).
table(low_v_s_mins_maxs_c, 17, 15, 4, low(c)).
table(low_minc_maxc_s_mins_c, 17, 6, 4, low(c)).
table(low_minc_maxc_s_maxs_c, 17, 6, 4, low(c)).
table(low_minc_maxc_mins_maxs_c, 17, 8, 4, low(c)).
table(low_minc_s_mins_maxs_c, 17, 8, 4, low(c)).
table(low_maxc_s_mins_maxs_c, 17, 5, 4, low(c)).
table(low_v_minc_maxc_s_mins_c, 17, 8, 5, low(c)).
table(low_v_minc_maxc_s_maxs_c, 17, 8, 5, low(c)).
table(low_v_minc_maxc_mins_maxs_c, 17, 8, 5, low(c)).
table(low_v_minc_s_mins_maxs_c, 17, 14, 5, low(c)).
table(low_v_maxc_s_mins_maxs_c, 17, 6, 5, low(c)).
table(low_minc_maxc_s_mins_maxs_c, 17, 7, 5, low(c)).
table(low_v_minc_maxc_s_mins_maxs_c, 17, 9, 6, low(c)).
table(up_v_c, 17, 18, 1, up(c)).
table(up_s_c, 17, 3, 1, up(c)).
table(up_v_minc_c, 17, 6, 2, up(c)).
table(up_v_maxc_c, 17, 11, 2, up(c)).
table(up_v_s_c, 17, 4, 2, up(c)).
table(up_v_mins_c, 17, 10, 2, up(c)).
table(up_v_maxs_c, 17, 18, 2, up(c)).
table(up_minc_s_c, 17, 3, 2, up(c)).
table(up_maxc_s_c, 17, 4, 2, up(c)).
table(up_s_mins_c, 17, 5, 2, up(c)).
table(up_s_maxs_c, 17, 5, 2, up(c)).
table(up_v_minc_maxc_c, 17, 7, 3, up(c)).
table(up_v_minc_s_c, 17, 6, 3, up(c)).
table(up_v_minc_mins_c, 17, 7, 3, up(c)).
table(up_v_minc_maxs_c, 17, 7, 3, up(c)).
table(up_v_maxc_s_c, 17, 5, 3, up(c)).
table(up_v_maxc_mins_c, 17, 5, 3, up(c)).
table(up_v_maxc_maxs_c, 17, 13, 3, up(c)).
table(up_v_s_mins_c, 17, 8, 3, up(c)).
table(up_v_s_maxs_c, 17, 6, 3, up(c)).
table(up_v_mins_maxs_c, 17, 12, 3, up(c)).
table(up_minc_maxc_s_c, 17, 5, 3, up(c)).
table(up_minc_s_mins_c, 17, 4, 3, up(c)).
table(up_minc_s_maxs_c, 17, 4, 3, up(c)).
table(up_maxc_s_mins_c, 17, 5, 3, up(c)).
table(up_maxc_s_maxs_c, 17, 5, 3, up(c)).
table(up_s_mins_maxs_c, 17, 8, 3, up(c)).
table(up_v_minc_maxc_s_c, 17, 7, 4, up(c)).
table(up_v_minc_maxc_mins_c, 17, 8, 4, up(c)).
table(up_v_minc_maxc_maxs_c, 17, 8, 4, up(c)).
table(up_v_minc_s_mins_c, 17, 7, 4, up(c)).
table(up_v_minc_s_maxs_c, 17, 7, 4, up(c)).
table(up_v_minc_mins_maxs_c, 17, 8, 4, up(c)).
table(up_v_maxc_s_mins_c, 17, 6, 4, up(c)).
table(up_v_maxc_s_maxs_c, 17, 6, 4, up(c)).
table(up_v_maxc_mins_maxs_c, 17, 7, 4, up(c)).
table(up_v_s_mins_maxs_c, 17, 10, 4, up(c)).
table(up_minc_maxc_s_mins_c, 17, 6, 4, up(c)).
table(up_minc_maxc_s_maxs_c, 17, 6, 4, up(c)).
table(up_minc_s_mins_maxs_c, 17, 5, 4, up(c)).
table(up_maxc_s_mins_maxs_c, 17, 6, 4, up(c)).
table(up_v_minc_maxc_s_mins_c, 17, 8, 5, up(c)).
table(up_v_minc_maxc_s_maxs_c, 17, 8, 5, up(c)).
table(up_v_minc_maxc_mins_maxs_c, 17, 9, 5, up(c)).
table(up_v_minc_s_mins_maxs_c, 17, 8, 5, up(c)).
table(up_v_maxc_s_mins_maxs_c, 17, 7, 5, up(c)).
table(up_minc_maxc_s_mins_maxs_c, 17, 7, 5, up(c)).
table(up_v_minc_maxc_s_mins_maxs_c, 17, 9, 6, up(c)).
table(low_v_minc, 17, 3, 1, low(minc)).
table(low_c_minc, 17, 3, 1, low(minc)).
table(low_maxc_minc, 17, 3, 1, low(minc)).
table(low_s_minc, 17, 3, 1, low(minc)).
table(low_mins_minc, 17, 2, 1, low(minc)).
table(low_maxs_minc, 17, 3, 1, low(minc)).
table(low_v_c_minc, 17, 5, 2, low(minc)).
table(low_v_maxc_minc, 17, 4, 2, low(minc)).
table(low_v_s_minc, 17, 4, 2, low(minc)).
table(low_v_mins_minc, 17, 3, 2, low(minc)).
table(low_v_maxs_minc, 17, 4, 2, low(minc)).
table(low_c_maxc_minc, 17, 4, 2, low(minc)).
table(low_c_s_minc, 17, 4, 2, low(minc)).
table(low_c_mins_minc, 17, 3, 2, low(minc)).
table(low_c_maxs_minc, 17, 4, 2, low(minc)).
table(low_maxc_s_minc, 17, 4, 2, low(minc)).
table(low_maxc_mins_minc, 17, 4, 2, low(minc)).
table(low_maxc_maxs_minc, 17, 4, 2, low(minc)).
table(low_s_mins_minc, 17, 3, 2, low(minc)).
table(low_s_maxs_minc, 17, 4, 2, low(minc)).
table(low_mins_maxs_minc, 17, 3, 2, low(minc)).
table(low_v_c_maxc_minc, 17, 6, 3, low(minc)).
table(low_v_c_s_minc, 17, 6, 3, low(minc)).
table(low_v_c_mins_minc, 17, 6, 3, low(minc)).
table(low_v_c_maxs_minc, 17, 6, 3, low(minc)).
table(low_v_maxc_s_minc, 17, 5, 3, low(minc)).
table(low_v_maxc_mins_minc, 17, 5, 3, low(minc)).
table(low_v_maxc_maxs_minc, 17, 5, 3, low(minc)).
table(low_v_s_mins_minc, 17, 4, 3, low(minc)).
table(low_v_s_maxs_minc, 17, 5, 3, low(minc)).
table(low_v_mins_maxs_minc, 17, 4, 3, low(minc)).
table(low_c_maxc_s_minc, 17, 5, 3, low(minc)).
table(low_c_maxc_mins_minc, 17, 5, 3, low(minc)).
table(low_c_maxc_maxs_minc, 17, 5, 3, low(minc)).
table(low_c_s_mins_minc, 17, 4, 3, low(minc)).
table(low_c_s_maxs_minc, 17, 6, 3, low(minc)).
table(low_c_mins_maxs_minc, 17, 4, 3, low(minc)).
table(low_maxc_s_mins_minc, 17, 5, 3, low(minc)).
table(low_maxc_s_maxs_minc, 17, 5, 3, low(minc)).
table(low_maxc_mins_maxs_minc, 17, 5, 3, low(minc)).
table(low_s_mins_maxs_minc, 17, 4, 3, low(minc)).
table(low_v_c_maxc_s_minc, 17, 7, 4, low(minc)).
table(low_v_c_maxc_mins_minc, 17, 7, 4, low(minc)).
table(low_v_c_maxc_maxs_minc, 17, 7, 4, low(minc)).
table(low_v_c_s_mins_minc, 17, 7, 4, low(minc)).
table(low_v_c_s_maxs_minc, 17, 7, 4, low(minc)).
table(low_v_c_mins_maxs_minc, 17, 7, 4, low(minc)).
table(low_v_maxc_s_mins_minc, 17, 6, 4, low(minc)).
table(low_v_maxc_s_maxs_minc, 17, 6, 4, low(minc)).
table(low_v_maxc_mins_maxs_minc, 17, 6, 4, low(minc)).
table(low_v_s_mins_maxs_minc, 17, 5, 4, low(minc)).
table(low_c_maxc_s_mins_minc, 17, 6, 4, low(minc)).
table(low_c_maxc_s_maxs_minc, 17, 6, 4, low(minc)).
table(low_c_maxc_mins_maxs_minc, 17, 6, 4, low(minc)).
table(low_c_s_mins_maxs_minc, 17, 6, 4, low(minc)).
table(low_maxc_s_mins_maxs_minc, 17, 6, 4, low(minc)).
table(low_v_c_maxc_s_mins_minc, 17, 8, 5, low(minc)).
table(low_v_c_maxc_s_maxs_minc, 17, 8, 5, low(minc)).
table(low_v_c_maxc_mins_maxs_minc, 17, 8, 5, low(minc)).
table(low_v_c_s_mins_maxs_minc, 17, 8, 5, low(minc)).
table(low_v_maxc_s_mins_maxs_minc, 17, 7, 5, low(minc)).
table(low_c_maxc_s_mins_maxs_minc, 17, 7, 5, low(minc)).
table(low_v_c_maxc_s_mins_maxs_minc, 17, 9, 6, low(minc)).
table(up_v_minc, 17, 11, 1, up(minc)).
table(up_maxc_minc, 17, 6, 1, up(minc)).
table(up_v_c_minc, 17, 5, 2, up(minc)).
table(up_v_maxc_minc, 17, 4, 2, up(minc)).
table(up_v_s_minc, 17, 13, 2, up(minc)).
table(up_v_mins_minc, 17, 12, 2, up(minc)).
table(up_v_maxs_minc, 17, 13, 2, up(minc)).
table(up_c_maxc_minc, 17, 4, 2, up(minc)).
table(up_maxc_s_minc, 17, 6, 2, up(minc)).
table(up_maxc_mins_minc, 17, 4, 2, up(minc)).
table(up_maxc_maxs_minc, 17, 7, 2, up(minc)).
table(up_s_maxs_minc, 17, 14, 2, up(minc)).
table(up_v_c_maxc_minc, 17, 6, 3, up(minc)).
table(up_v_c_s_minc, 17, 6, 3, up(minc)).
table(up_v_c_mins_minc, 17, 6, 3, up(minc)).
table(up_v_c_maxs_minc, 17, 6, 3, up(minc)).
table(up_v_maxc_s_minc, 17, 5, 3, up(minc)).
table(up_v_maxc_mins_minc, 17, 5, 3, up(minc)).
table(up_v_maxc_maxs_minc, 17, 5, 3, up(minc)).
table(up_v_s_mins_minc, 17, 14, 3, up(minc)).
table(up_v_s_maxs_minc, 17, 14, 3, up(minc)).
table(up_v_mins_maxs_minc, 17, 14, 3, up(minc)).
table(up_c_maxc_s_minc, 17, 5, 3, up(minc)).
table(up_c_maxc_mins_minc, 17, 5, 3, up(minc)).
table(up_c_maxc_maxs_minc, 17, 5, 3, up(minc)).
table(up_c_s_maxs_minc, 17, 4, 3, up(minc)).
table(up_maxc_s_mins_minc, 17, 5, 3, up(minc)).
table(up_maxc_s_maxs_minc, 17, 5, 3, up(minc)).
table(up_maxc_mins_maxs_minc, 17, 5, 3, up(minc)).
table(up_s_mins_maxs_minc, 17, 15, 3, up(minc)).
table(up_v_c_maxc_s_minc, 17, 7, 4, up(minc)).
table(up_v_c_maxc_mins_minc, 17, 7, 4, up(minc)).
table(up_v_c_maxc_maxs_minc, 17, 7, 4, up(minc)).
table(up_v_c_s_mins_minc, 17, 7, 4, up(minc)).
table(up_v_c_s_maxs_minc, 17, 7, 4, up(minc)).
table(up_v_c_mins_maxs_minc, 17, 7, 4, up(minc)).
table(up_v_maxc_s_mins_minc, 17, 6, 4, up(minc)).
table(up_v_maxc_s_maxs_minc, 17, 6, 4, up(minc)).
table(up_v_maxc_mins_maxs_minc, 17, 6, 4, up(minc)).
table(up_v_s_mins_maxs_minc, 17, 15, 4, up(minc)).
table(up_c_maxc_s_mins_minc, 17, 6, 4, up(minc)).
table(up_c_maxc_s_maxs_minc, 17, 6, 4, up(minc)).
table(up_c_maxc_mins_maxs_minc, 17, 6, 4, up(minc)).
table(up_c_s_mins_maxs_minc, 17, 5, 4, up(minc)).
table(up_maxc_s_mins_maxs_minc, 17, 6, 4, up(minc)).
table(up_v_c_maxc_s_mins_minc, 17, 8, 5, up(minc)).
table(up_v_c_maxc_s_maxs_minc, 17, 8, 5, up(minc)).
table(up_v_c_maxc_mins_maxs_minc, 17, 8, 5, up(minc)).
table(up_v_c_s_mins_maxs_minc, 17, 8, 5, up(minc)).
table(up_v_maxc_s_mins_maxs_minc, 17, 7, 5, up(minc)).
table(up_c_maxc_s_mins_maxs_minc, 17, 7, 5, up(minc)).
table(up_v_c_maxc_s_mins_maxs_minc, 17, 9, 6, up(minc)).
table(low_v_maxc, 17, 18, 1, low(maxc)).
table(low_c_maxc, 17, 18, 1, low(maxc)).
table(low_minc_maxc, 17, 6, 1, low(maxc)).
table(low_s_maxc, 17, 18, 1, low(maxc)).
table(low_mins_maxc, 17, 9, 1, low(maxc)).
table(low_maxs_maxc, 17, 2, 1, low(maxc)).
table(low_v_c_maxc, 17, 3, 2, low(maxc)).
table(low_v_minc_maxc, 17, 6, 2, low(maxc)).
table(low_v_s_maxc, 17, 4, 2, low(maxc)).
table(low_v_mins_maxc, 17, 9, 2, low(maxc)).
table(low_v_maxs_maxc, 17, 3, 2, low(maxc)).
table(low_c_minc_maxc, 17, 11, 2, low(maxc)).
table(low_c_s_maxc, 17, 4, 2, low(maxc)).
table(low_c_mins_maxc, 17, 18, 2, low(maxc)).
table(low_c_maxs_maxc, 17, 3, 2, low(maxc)).
table(low_minc_s_maxc, 17, 6, 2, low(maxc)).
table(low_minc_mins_maxc, 17, 4, 2, low(maxc)).
table(low_minc_maxs_maxc, 17, 4, 2, low(maxc)).
table(low_s_mins_maxc, 17, 18, 2, low(maxc)).
table(low_s_maxs_maxc, 17, 3, 2, low(maxc)).
table(low_mins_maxs_maxc, 17, 3, 2, low(maxc)).
table(low_v_c_minc_maxc, 17, 6, 3, low(maxc)).
table(low_v_c_s_maxc, 17, 4, 3, low(maxc)).
table(low_v_c_mins_maxc, 17, 4, 3, low(maxc)).
table(low_v_c_maxs_maxc, 17, 4, 3, low(maxc)).
table(low_v_minc_s_maxc, 17, 5, 3, low(maxc)).
table(low_v_minc_mins_maxc, 17, 7, 3, low(maxc)).
table(low_v_minc_maxs_maxc, 17, 5, 3, low(maxc)).
table(low_v_s_mins_maxc, 17, 5, 3, low(maxc)).
table(low_v_s_maxs_maxc, 17, 4, 3, low(maxc)).
table(low_v_mins_maxs_maxc, 17, 4, 3, low(maxc)).
table(low_c_minc_s_maxc, 17, 5, 3, low(maxc)).
table(low_c_minc_mins_maxc, 17, 6, 3, low(maxc)).
table(low_c_minc_maxs_maxc, 17, 5, 3, low(maxc)).
table(low_c_s_mins_maxc, 17, 4, 3, low(maxc)).
table(low_c_s_maxs_maxc, 17, 4, 3, low(maxc)).
table(low_c_mins_maxs_maxc, 17, 4, 3, low(maxc)).
table(low_minc_s_mins_maxc, 17, 5, 3, low(maxc)).
table(low_minc_s_maxs_maxc, 17, 5, 3, low(maxc)).
table(low_minc_mins_maxs_maxc, 17, 5, 3, low(maxc)).
table(low_s_mins_maxs_maxc, 17, 4, 3, low(maxc)).
table(low_v_c_minc_s_maxc, 17, 7, 4, low(maxc)).
table(low_v_c_minc_mins_maxc, 17, 7, 4, low(maxc)).
table(low_v_c_minc_maxs_maxc, 17, 7, 4, low(maxc)).
table(low_v_c_s_mins_maxc, 17, 5, 4, low(maxc)).
table(low_v_c_s_maxs_maxc, 17, 5, 4, low(maxc)).
table(low_v_c_mins_maxs_maxc, 17, 5, 4, low(maxc)).
table(low_v_minc_s_mins_maxc, 17, 6, 4, low(maxc)).
table(low_v_minc_s_maxs_maxc, 17, 6, 4, low(maxc)).
table(low_v_minc_mins_maxs_maxc, 17, 6, 4, low(maxc)).
table(low_v_s_mins_maxs_maxc, 17, 5, 4, low(maxc)).
table(low_c_minc_s_mins_maxc, 17, 6, 4, low(maxc)).
table(low_c_minc_s_maxs_maxc, 17, 6, 4, low(maxc)).
table(low_c_minc_mins_maxs_maxc, 17, 6, 4, low(maxc)).
table(low_c_s_mins_maxs_maxc, 17, 5, 4, low(maxc)).
table(low_minc_s_mins_maxs_maxc, 17, 6, 4, low(maxc)).
table(low_v_c_minc_s_mins_maxc, 17, 8, 5, low(maxc)).
table(low_v_c_minc_s_maxs_maxc, 17, 8, 5, low(maxc)).
table(low_v_c_minc_mins_maxs_maxc, 17, 8, 5, low(maxc)).
table(low_v_c_s_mins_maxs_maxc, 17, 6, 5, low(maxc)).
table(low_v_minc_s_mins_maxs_maxc, 17, 7, 5, low(maxc)).
table(low_c_minc_s_mins_maxs_maxc, 17, 7, 5, low(maxc)).
table(low_v_c_minc_s_mins_maxs_maxc, 17, 9, 6, low(maxc)).
table(up_v_maxc, 17, 11, 1, up(maxc)).
table(up_v_c_maxc, 17, 11, 2, up(maxc)).
table(up_v_minc_maxc, 17, 11, 2, up(maxc)).
table(up_v_s_maxc, 17, 13, 2, up(maxc)).
table(up_v_mins_maxc, 17, 12, 2, up(maxc)).
table(up_v_maxs_maxc, 17, 13, 2, up(maxc)).
table(up_v_c_minc_maxc, 17, 11, 3, up(maxc)).
table(up_v_c_s_maxc, 17, 13, 3, up(maxc)).
table(up_v_c_mins_maxc, 17, 12, 3, up(maxc)).
table(up_v_c_maxs_maxc, 17, 13, 3, up(maxc)).
table(up_v_minc_s_maxc, 17, 12, 3, up(maxc)).
table(up_v_minc_mins_maxc, 17, 12, 3, up(maxc)).
table(up_v_minc_maxs_maxc, 17, 12, 3, up(maxc)).
table(up_v_s_mins_maxc, 17, 14, 3, up(maxc)).
table(up_v_s_maxs_maxc, 17, 14, 3, up(maxc)).
table(up_v_mins_maxs_maxc, 17, 14, 3, up(maxc)).
table(up_v_c_minc_s_maxc, 17, 12, 4, up(maxc)).
table(up_v_c_minc_mins_maxc, 17, 12, 4, up(maxc)).
table(up_v_c_minc_maxs_maxc, 17, 12, 4, up(maxc)).
table(up_v_c_s_mins_maxc, 17, 14, 4, up(maxc)).
table(up_v_c_s_maxs_maxc, 17, 6, 4, up(maxc)).
table(up_v_c_mins_maxs_maxc, 17, 10, 4, up(maxc)).
table(up_v_minc_s_mins_maxc, 17, 13, 4, up(maxc)).
table(up_v_minc_s_maxs_maxc, 17, 13, 4, up(maxc)).
table(up_v_minc_mins_maxs_maxc, 17, 13, 4, up(maxc)).
table(up_v_s_mins_maxs_maxc, 17, 15, 4, up(maxc)).
table(up_v_c_minc_s_mins_maxc, 17, 13, 5, up(maxc)).
table(up_v_c_minc_s_maxs_maxc, 17, 8, 5, up(maxc)).
table(up_v_c_minc_mins_maxs_maxc, 17, 9, 5, up(maxc)).
table(up_v_c_s_mins_maxs_maxc, 17, 10, 5, up(maxc)).
table(up_v_minc_s_mins_maxs_maxc, 17, 14, 5, up(maxc)).
table(up_v_c_minc_s_mins_maxs_maxc, 17, 9, 6, up(maxc)).
table(low_v_s, 17, 18, 1, low(s)).
table(low_c_s, 17, 3, 1, low(s)).
table(low_minc_s, 17, 18, 1, low(s)).
table(low_maxc_s, 17, 18, 1, low(s)).
table(low_mins_s, 17, 18, 1, low(s)).
table(low_maxs_s, 17, 18, 1, low(s)).
table(low_v_c_s, 17, 4, 2, low(s)).
table(low_v_minc_s, 17, 18, 2, low(s)).
table(low_v_maxc_s, 17, 6, 2, low(s)).
table(low_v_mins_s, 17, 9, 2, low(s)).
table(low_v_maxs_s, 17, 3, 2, low(s)).
table(low_c_minc_s, 17, 5, 2, low(s)).
table(low_c_maxc_s, 17, 5, 2, low(s)).
table(low_c_mins_s, 17, 5, 2, low(s)).
table(low_c_maxs_s, 17, 5, 2, low(s)).
table(low_minc_maxc_s, 17, 18, 2, low(s)).
table(low_minc_mins_s, 17, 12, 2, low(s)).
table(low_minc_maxs_s, 17, 14, 2, low(s)).
table(low_maxc_mins_s, 17, 7, 2, low(s)).
table(low_maxc_maxs_s, 17, 14, 2, low(s)).
table(low_mins_maxs_s, 17, 12, 2, low(s)).
table(low_v_c_minc_s, 17, 8, 3, low(s)).
table(low_v_c_maxc_s, 17, 6, 3, low(s)).
table(low_v_c_mins_s, 17, 8, 3, low(s)).
table(low_v_c_maxs_s, 17, 4, 3, low(s)).
table(low_v_minc_maxc_s, 17, 10, 3, low(s)).
table(low_v_minc_mins_s, 17, 13, 3, low(s)).
table(low_v_minc_maxs_s, 17, 4, 3, low(s)).
table(low_v_maxc_mins_s, 17, 4, 3, low(s)).
table(low_v_maxc_maxs_s, 17, 4, 3, low(s)).
table(low_v_mins_maxs_s, 17, 4, 3, low(s)).
table(low_c_minc_maxc_s, 17, 8, 3, low(s)).
table(low_c_minc_mins_s, 17, 4, 3, low(s)).
table(low_c_minc_maxs_s, 17, 5, 3, low(s)).
table(low_c_maxc_mins_s, 17, 8, 3, low(s)).
table(low_c_maxc_maxs_s, 17, 5, 3, low(s)).
table(low_c_mins_maxs_s, 17, 8, 3, low(s)).
table(low_minc_maxc_mins_s, 17, 13, 3, low(s)).
table(low_minc_maxc_maxs_s, 17, 14, 3, low(s)).
table(low_minc_mins_maxs_s, 17, 8, 3, low(s)).
table(low_maxc_mins_maxs_s, 17, 9, 3, low(s)).
table(low_v_c_minc_maxc_s, 17, 10, 4, low(s)).
table(low_v_c_minc_mins_s, 17, 7, 4, low(s)).
table(low_v_c_minc_maxs_s, 17, 7, 4, low(s)).
table(low_v_c_maxc_mins_s, 17, 5, 4, low(s)).
table(low_v_c_maxc_maxs_s, 17, 5, 4, low(s)).
table(low_v_c_mins_maxs_s, 17, 5, 4, low(s)).
table(low_v_minc_maxc_mins_s, 17, 8, 4, low(s)).
table(low_v_minc_maxc_maxs_s, 17, 6, 4, low(s)).
table(low_v_minc_mins_maxs_s, 17, 5, 4, low(s)).
table(low_v_maxc_mins_maxs_s, 17, 5, 4, low(s)).
table(low_c_minc_maxc_mins_s, 17, 6, 4, low(s)).
table(low_c_minc_maxc_maxs_s, 17, 7, 4, low(s)).
table(low_c_minc_mins_maxs_s, 17, 5, 4, low(s)).
table(low_c_maxc_mins_maxs_s, 17, 6, 4, low(s)).
table(low_minc_maxc_mins_maxs_s, 17, 9, 4, low(s)).
table(low_v_c_minc_maxc_mins_s, 17, 8, 5, low(s)).
table(low_v_c_minc_maxc_maxs_s, 17, 8, 5, low(s)).
table(low_v_c_minc_mins_maxs_s, 17, 8, 5, low(s)).
table(low_v_c_maxc_mins_maxs_s, 17, 6, 5, low(s)).
table(low_v_minc_maxc_mins_maxs_s, 17, 7, 5, low(s)).
table(low_c_minc_maxc_mins_maxs_s, 17, 7, 5, low(s)).
table(low_v_c_minc_maxc_mins_maxs_s, 17, 9, 6, low(s)).
table(up_v_s, 17, 7, 1, up(s)).
table(up_v_c_s, 17, 8, 2, up(s)).
table(up_v_minc_s, 17, 8, 2, up(s)).
table(up_v_maxc_s, 17, 8, 2, up(s)).
table(up_v_mins_s, 17, 4, 2, up(s)).
table(up_v_maxs_s, 17, 7, 2, up(s)).
table(up_c_maxc_s, 17, 9, 2, up(s)).
table(up_v_c_minc_s, 17, 11, 3, up(s)).
table(up_v_c_maxc_s, 17, 9, 3, up(s)).
table(up_v_c_mins_s, 17, 5, 3, up(s)).
table(up_v_c_maxs_s, 17, 8, 3, up(s)).
table(up_v_minc_maxc_s, 17, 10, 3, up(s)).
table(up_v_minc_mins_s, 17, 5, 3, up(s)).
table(up_v_minc_maxs_s, 17, 8, 3, up(s)).
table(up_v_maxc_mins_s, 17, 5, 3, up(s)).
table(up_v_maxc_maxs_s, 17, 8, 3, up(s)).
table(up_v_mins_maxs_s, 17, 5, 3, up(s)).
table(up_c_minc_maxc_s, 17, 12, 3, up(s)).
table(up_c_maxc_mins_s, 17, 5, 3, up(s)).
table(up_c_maxc_maxs_s, 17, 9, 3, up(s)).
table(up_v_c_minc_maxc_s, 17, 12, 4, up(s)).
table(up_v_c_minc_mins_s, 17, 8, 4, up(s)).
table(up_v_c_minc_maxs_s, 17, 11, 4, up(s)).
table(up_v_c_maxc_mins_s, 17, 6, 4, up(s)).
table(up_v_c_maxc_maxs_s, 17, 9, 4, up(s)).
table(up_v_c_mins_maxs_s, 17, 6, 4, up(s)).
table(up_v_minc_maxc_mins_s, 17, 7, 4, up(s)).
table(up_v_minc_maxc_maxs_s, 17, 10, 4, up(s)).
table(up_v_minc_mins_maxs_s, 17, 6, 4, up(s)).
table(up_v_maxc_mins_maxs_s, 17, 6, 4, up(s)).
table(up_c_minc_maxc_mins_s, 17, 7, 4, up(s)).
table(up_c_minc_maxc_maxs_s, 17, 12, 4, up(s)).
table(up_c_maxc_mins_maxs_s, 17, 6, 4, up(s)).
table(up_v_c_minc_maxc_mins_s, 17, 9, 5, up(s)).
table(up_v_c_minc_maxc_maxs_s, 17, 12, 5, up(s)).
table(up_v_c_minc_mins_maxs_s, 17, 9, 5, up(s)).
table(up_v_c_maxc_mins_maxs_s, 17, 7, 5, up(s)).
table(up_v_minc_maxc_mins_maxs_s, 17, 8, 5, up(s)).
table(up_c_minc_maxc_mins_maxs_s, 17, 8, 5, up(s)).
table(up_v_c_minc_maxc_mins_maxs_s, 17, 10, 6, up(s)).
table(low_v_mins, 17, 2, 1, low(mins)).
table(low_c_mins, 17, 2, 1, low(mins)).
table(low_minc_mins, 17, 2, 1, low(mins)).
table(low_maxc_mins, 17, 2, 1, low(mins)).
table(low_s_mins, 17, 2, 1, low(mins)).
table(low_maxs_mins, 17, 2, 1, low(mins)).
table(low_v_c_mins, 17, 3, 2, low(mins)).
table(low_v_minc_mins, 17, 3, 2, low(mins)).
table(low_v_maxc_mins, 17, 3, 2, low(mins)).
table(low_v_s_mins, 17, 3, 2, low(mins)).
table(low_v_maxs_mins, 17, 3, 2, low(mins)).
table(low_c_minc_mins, 17, 3, 2, low(mins)).
table(low_c_maxc_mins, 17, 3, 2, low(mins)).
table(low_c_s_mins, 17, 3, 2, low(mins)).
table(low_c_maxs_mins, 17, 3, 2, low(mins)).
table(low_minc_maxc_mins, 17, 4, 2, low(mins)).
table(low_minc_s_mins, 17, 3, 2, low(mins)).
table(low_minc_maxs_mins, 17, 3, 2, low(mins)).
table(low_maxc_s_mins, 17, 3, 2, low(mins)).
table(low_maxc_maxs_mins, 17, 3, 2, low(mins)).
table(low_s_maxs_mins, 17, 3, 2, low(mins)).
table(low_v_c_minc_mins, 17, 6, 3, low(mins)).
table(low_v_c_maxc_mins, 17, 4, 3, low(mins)).
table(low_v_c_s_mins, 17, 4, 3, low(mins)).
table(low_v_c_maxs_mins, 17, 4, 3, low(mins)).
table(low_v_minc_maxc_mins, 17, 5, 3, low(mins)).
table(low_v_minc_s_mins, 17, 4, 3, low(mins)).
table(low_v_minc_maxs_mins, 17, 4, 3, low(mins)).
table(low_v_maxc_s_mins, 17, 4, 3, low(mins)).
table(low_v_maxc_maxs_mins, 17, 4, 3, low(mins)).
table(low_v_s_maxs_mins, 17, 4, 3, low(mins)).
table(low_c_minc_maxc_mins, 17, 5, 3, low(mins)).
table(low_c_minc_s_mins, 17, 4, 3, low(mins)).
table(low_c_minc_maxs_mins, 17, 4, 3, low(mins)).
table(low_c_maxc_s_mins, 17, 4, 3, low(mins)).
table(low_c_maxc_maxs_mins, 17, 4, 3, low(mins)).
table(low_c_s_maxs_mins, 17, 4, 3, low(mins)).
table(low_minc_maxc_s_mins, 17, 5, 3, low(mins)).
table(low_minc_maxc_maxs_mins, 17, 5, 3, low(mins)).
table(low_minc_s_maxs_mins, 17, 4, 3, low(mins)).
table(low_maxc_s_maxs_mins, 17, 4, 3, low(mins)).
table(low_v_c_minc_maxc_mins, 17, 7, 4, low(mins)).
table(low_v_c_minc_s_mins, 17, 7, 4, low(mins)).
table(low_v_c_minc_maxs_mins, 17, 7, 4, low(mins)).
table(low_v_c_maxc_s_mins, 17, 5, 4, low(mins)).
table(low_v_c_maxc_maxs_mins, 17, 5, 4, low(mins)).
table(low_v_c_s_maxs_mins, 17, 5, 4, low(mins)).
table(low_v_minc_maxc_s_mins, 17, 6, 4, low(mins)).
table(low_v_minc_maxc_maxs_mins, 17, 6, 4, low(mins)).
table(low_v_minc_s_maxs_mins, 17, 5, 4, low(mins)).
table(low_v_maxc_s_maxs_mins, 17, 5, 4, low(mins)).
table(low_c_minc_maxc_s_mins, 17, 6, 4, low(mins)).
table(low_c_minc_maxc_maxs_mins, 17, 6, 4, low(mins)).
table(low_c_minc_s_maxs_mins, 17, 5, 4, low(mins)).
table(low_c_maxc_s_maxs_mins, 17, 5, 4, low(mins)).
table(low_minc_maxc_s_maxs_mins, 17, 6, 4, low(mins)).
table(low_v_c_minc_maxc_s_mins, 17, 8, 5, low(mins)).
table(low_v_c_minc_maxc_maxs_mins, 17, 8, 5, low(mins)).
table(low_v_c_minc_s_maxs_mins, 17, 8, 5, low(mins)).
table(low_v_c_maxc_s_maxs_mins, 17, 6, 5, low(mins)).
table(low_v_minc_maxc_s_maxs_mins, 17, 7, 5, low(mins)).
table(low_c_minc_maxc_s_maxs_mins, 17, 7, 5, low(mins)).
table(low_v_c_minc_maxc_s_maxs_mins, 17, 9, 6, low(mins)).
table(up_v_mins, 17, 18, 1, up(mins)).
table(up_maxc_mins, 17, 9, 1, up(mins)).
table(up_maxs_mins, 17, 2, 1, up(mins)).
table(up_v_c_mins, 17, 6, 2, up(mins)).
table(up_v_minc_mins, 17, 3, 2, up(mins)).
table(up_v_maxc_mins, 17, 5, 2, up(mins)).
table(up_v_s_mins, 17, 3, 2, up(mins)).
table(up_v_maxs_mins, 17, 3, 2, up(mins)).
table(up_c_maxc_mins, 17, 5, 2, up(mins)).
table(up_c_maxs_mins, 17, 3, 2, up(mins)).
table(up_minc_maxc_mins, 17, 4, 2, up(mins)).
table(up_minc_maxs_mins, 17, 3, 2, up(mins)).
table(up_maxc_s_mins, 17, 3, 2, up(mins)).
table(up_maxc_maxs_mins, 17, 3, 2, up(mins)).
table(up_s_maxs_mins, 17, 3, 2, up(mins)).
table(up_v_c_minc_mins, 17, 6, 3, up(mins)).
table(up_v_c_maxc_mins, 17, 7, 3, up(mins)).
table(up_v_c_s_mins, 17, 4, 3, up(mins)).
table(up_v_c_maxs_mins, 17, 4, 3, up(mins)).
table(up_v_minc_maxc_mins, 17, 5, 3, up(mins)).
table(up_v_minc_s_mins, 17, 4, 3, up(mins)).
table(up_v_minc_maxs_mins, 17, 4, 3, up(mins)).
table(up_v_maxc_s_mins, 17, 4, 3, up(mins)).
table(up_v_maxc_maxs_mins, 17, 4, 3, up(mins)).
table(up_v_s_maxs_mins, 17, 4, 3, up(mins)).
table(up_c_minc_maxc_mins, 17, 5, 3, up(mins)).
table(up_c_minc_maxs_mins, 17, 4, 3, up(mins)).
table(up_c_maxc_s_mins, 17, 4, 3, up(mins)).
table(up_c_maxc_maxs_mins, 17, 4, 3, up(mins)).
table(up_c_s_maxs_mins, 17, 4, 3, up(mins)).
table(up_minc_maxc_s_mins, 17, 5, 3, up(mins)).
table(up_minc_maxc_maxs_mins, 17, 5, 3, up(mins)).
table(up_minc_s_maxs_mins, 17, 4, 3, up(mins)).
table(up_maxc_s_maxs_mins, 17, 4, 3, up(mins)).
table(up_v_c_minc_maxc_mins, 17, 7, 4, up(mins)).
table(up_v_c_minc_s_mins, 17, 7, 4, up(mins)).
table(up_v_c_minc_maxs_mins, 17, 7, 4, up(mins)).
table(up_v_c_maxc_s_mins, 17, 5, 4, up(mins)).
table(up_v_c_maxc_maxs_mins, 17, 5, 4, up(mins)).
table(up_v_c_s_maxs_mins, 17, 5, 4, up(mins)).
table(up_v_minc_maxc_s_mins, 17, 6, 4, up(mins)).
table(up_v_minc_maxc_maxs_mins, 17, 6, 4, up(mins)).
table(up_v_minc_s_maxs_mins, 17, 5, 4, up(mins)).
table(up_v_maxc_s_maxs_mins, 17, 5, 4, up(mins)).
table(up_c_minc_maxc_s_mins, 17, 6, 4, up(mins)).
table(up_c_minc_maxc_maxs_mins, 17, 6, 4, up(mins)).
table(up_c_minc_s_maxs_mins, 17, 5, 4, up(mins)).
table(up_c_maxc_s_maxs_mins, 17, 5, 4, up(mins)).
table(up_minc_maxc_s_maxs_mins, 17, 6, 4, up(mins)).
table(up_v_c_minc_maxc_s_mins, 17, 8, 5, up(mins)).
table(up_v_c_minc_maxc_maxs_mins, 17, 8, 5, up(mins)).
table(up_v_c_minc_s_maxs_mins, 17, 8, 5, up(mins)).
table(up_v_c_maxc_s_maxs_mins, 17, 6, 5, up(mins)).
table(up_v_minc_maxc_s_maxs_mins, 17, 7, 5, up(mins)).
table(up_c_minc_maxc_s_maxs_mins, 17, 7, 5, up(mins)).
table(up_v_c_minc_maxc_s_maxs_mins, 17, 9, 6, up(mins)).
table(low_v_maxs, 17, 7, 1, low(maxs)).
table(low_c_maxs, 17, 4, 1, low(maxs)).
table(low_minc_maxs, 17, 4, 1, low(maxs)).
table(low_maxc_maxs, 17, 4, 1, low(maxs)).
table(low_s_maxs, 17, 7, 1, low(maxs)).
table(low_mins_maxs, 17, 2, 1, low(maxs)).
table(low_v_c_maxs, 17, 8, 2, low(maxs)).
table(low_v_minc_maxs, 17, 8, 2, low(maxs)).
table(low_v_maxc_maxs, 17, 8, 2, low(maxs)).
table(low_v_s_maxs, 17, 3, 2, low(maxs)).
table(low_v_mins_maxs, 17, 3, 2, low(maxs)).
table(low_c_minc_maxs, 17, 5, 2, low(maxs)).
table(low_c_maxc_maxs, 17, 5, 2, low(maxs)).
table(low_c_s_maxs, 17, 8, 2, low(maxs)).
table(low_c_mins_maxs, 17, 3, 2, low(maxs)).
table(low_minc_maxc_maxs, 17, 6, 2, low(maxs)).
table(low_minc_s_maxs, 17, 3, 2, low(maxs)).
table(low_minc_mins_maxs, 17, 3, 2, low(maxs)).
table(low_maxc_s_maxs, 17, 3, 2, low(maxs)).
table(low_maxc_mins_maxs, 17, 3, 2, low(maxs)).
table(low_s_mins_maxs, 17, 7, 2, low(maxs)).
table(low_v_c_minc_maxs, 17, 11, 3, low(maxs)).
table(low_v_c_maxc_maxs, 17, 9, 3, low(maxs)).
table(low_v_c_s_maxs, 17, 4, 3, low(maxs)).
table(low_v_c_mins_maxs, 17, 4, 3, low(maxs)).
table(low_v_minc_maxc_maxs, 17, 10, 3, low(maxs)).
table(low_v_minc_s_maxs, 17, 4, 3, low(maxs)).
table(low_v_minc_mins_maxs, 17, 4, 3, low(maxs)).
table(low_v_maxc_s_maxs, 17, 4, 3, low(maxs)).
table(low_v_maxc_mins_maxs, 17, 4, 3, low(maxs)).
table(low_v_s_mins_maxs, 17, 4, 3, low(maxs)).
table(low_c_minc_maxc_maxs, 17, 7, 3, low(maxs)).
table(low_c_minc_s_maxs, 17, 4, 3, low(maxs)).
table(low_c_minc_mins_maxs, 17, 4, 3, low(maxs)).
table(low_c_maxc_s_maxs, 17, 4, 3, low(maxs)).
table(low_c_maxc_mins_maxs, 17, 4, 3, low(maxs)).
table(low_c_s_mins_maxs, 17, 8, 3, low(maxs)).
table(low_minc_maxc_s_maxs, 17, 5, 3, low(maxs)).
table(low_minc_maxc_mins_maxs, 17, 5, 3, low(maxs)).
table(low_minc_s_mins_maxs, 17, 4, 3, low(maxs)).
table(low_maxc_s_mins_maxs, 17, 4, 3, low(maxs)).
table(low_v_c_minc_maxc_maxs, 17, 12, 4, low(maxs)).
table(low_v_c_minc_s_maxs, 17, 7, 4, low(maxs)).
table(low_v_c_minc_mins_maxs, 17, 7, 4, low(maxs)).
table(low_v_c_maxc_s_maxs, 17, 5, 4, low(maxs)).
table(low_v_c_maxc_mins_maxs, 17, 5, 4, low(maxs)).
table(low_v_c_s_mins_maxs, 17, 5, 4, low(maxs)).
table(low_v_minc_maxc_s_maxs, 17, 6, 4, low(maxs)).
table(low_v_minc_maxc_mins_maxs, 17, 6, 4, low(maxs)).
table(low_v_minc_s_mins_maxs, 17, 5, 4, low(maxs)).
table(low_v_maxc_s_mins_maxs, 17, 5, 4, low(maxs)).
table(low_c_minc_maxc_s_maxs, 17, 6, 4, low(maxs)).
table(low_c_minc_maxc_mins_maxs, 17, 6, 4, low(maxs)).
table(low_c_minc_s_mins_maxs, 17, 5, 4, low(maxs)).
table(low_c_maxc_s_mins_maxs, 17, 5, 4, low(maxs)).
table(low_minc_maxc_s_mins_maxs, 17, 6, 4, low(maxs)).
table(low_v_c_minc_maxc_s_maxs, 17, 8, 5, low(maxs)).
table(low_v_c_minc_maxc_mins_maxs, 17, 8, 5, low(maxs)).
table(low_v_c_minc_s_mins_maxs, 17, 8, 5, low(maxs)).
table(low_v_c_maxc_s_mins_maxs, 17, 6, 5, low(maxs)).
table(low_v_minc_maxc_s_mins_maxs, 17, 7, 5, low(maxs)).
table(low_c_minc_maxc_s_mins_maxs, 17, 7, 5, low(maxs)).
table(low_v_c_minc_maxc_s_mins_maxs, 17, 9, 6, low(maxs)).
table(up_v_maxs, 17, 18, 1, up(maxs)).
table(up_maxc_maxs, 17, 2, 1, up(maxs)).
table(up_v_c_maxs, 17, 18, 2, up(maxs)).
table(up_v_minc_maxs, 17, 12, 2, up(maxs)).
table(up_v_maxc_maxs, 17, 3, 2, up(maxs)).
table(up_v_s_maxs, 17, 7, 2, up(maxs)).
table(up_v_mins_maxs, 17, 9, 2, up(maxs)).
table(up_c_maxc_maxs, 17, 3, 2, up(maxs)).
table(up_minc_maxc_maxs, 17, 4, 2, up(maxs)).
table(up_maxc_s_maxs, 17, 3, 2, up(maxs)).
table(up_maxc_mins_maxs, 17, 3, 2, up(maxs)).
table(up_v_c_minc_maxs, 17, 12, 3, up(maxs)).
table(up_v_c_maxc_maxs, 17, 4, 3, up(maxs)).
table(up_v_c_s_maxs, 17, 8, 3, up(maxs)).
table(up_v_c_mins_maxs, 17, 18, 3, up(maxs)).
table(up_v_minc_maxc_maxs, 17, 5, 3, up(maxs)).
table(up_v_minc_s_maxs, 17, 5, 3, up(maxs)).
table(up_v_minc_mins_maxs, 17, 14, 3, up(maxs)).
table(up_v_maxc_s_maxs, 17, 4, 3, up(maxs)).
table(up_v_maxc_mins_maxs, 17, 4, 3, up(maxs)).
table(up_v_s_mins_maxs, 17, 7, 3, up(maxs)).
table(up_c_minc_maxc_maxs, 17, 5, 3, up(maxs)).
table(up_c_maxc_s_maxs, 17, 4, 3, up(maxs)).
table(up_c_maxc_mins_maxs, 17, 4, 3, up(maxs)).
table(up_minc_maxc_s_maxs, 17, 5, 3, up(maxs)).
table(up_minc_maxc_mins_maxs, 17, 5, 3, up(maxs)).
table(up_maxc_s_mins_maxs, 17, 4, 3, up(maxs)).
table(up_v_c_minc_maxc_maxs, 17, 7, 4, up(maxs)).
table(up_v_c_minc_s_maxs, 17, 7, 4, up(maxs)).
table(up_v_c_minc_mins_maxs, 17, 13, 4, up(maxs)).
table(up_v_c_maxc_s_maxs, 17, 5, 4, up(maxs)).
table(up_v_c_maxc_mins_maxs, 17, 5, 4, up(maxs)).
table(up_v_c_s_mins_maxs, 17, 8, 4, up(maxs)).
table(up_v_minc_maxc_s_maxs, 17, 6, 4, up(maxs)).
table(up_v_minc_maxc_mins_maxs, 17, 6, 4, up(maxs)).
table(up_v_minc_s_mins_maxs, 17, 5, 4, up(maxs)).
table(up_v_maxc_s_mins_maxs, 17, 5, 4, up(maxs)).
table(up_c_minc_maxc_s_maxs, 17, 6, 4, up(maxs)).
table(up_c_minc_maxc_mins_maxs, 17, 6, 4, up(maxs)).
table(up_c_maxc_s_mins_maxs, 17, 5, 4, up(maxs)).
table(up_minc_maxc_s_mins_maxs, 17, 6, 4, up(maxs)).
table(up_v_c_minc_maxc_s_maxs, 17, 8, 5, up(maxs)).
table(up_v_c_minc_maxc_mins_maxs, 17, 8, 5, up(maxs)).
table(up_v_c_minc_s_mins_maxs, 17, 8, 5, up(maxs)).
table(up_v_c_maxc_s_mins_maxs, 17, 6, 5, up(maxs)).
table(up_v_minc_maxc_s_mins_maxs, 17, 7, 5, up(maxs)).
table(up_c_minc_maxc_s_mins_maxs, 17, 7, 5, up(maxs)).
table(up_v_c_minc_maxc_s_mins_maxs, 17, 9, 6, up(maxs)).
table(low_v_mina, 17, 12, 1, low(mina)).
table(low_c_mina, 17, 7, 1, low(mina)).
table(low_minc_mina, 17, 19, 1, low(mina)).
table(low_maxc_mina, 17, 19, 1, low(mina)).
table(low_s_mina, 17, 12, 1, low(mina)).
table(low_mins_mina, 17, 11, 1, low(mina)).
table(low_maxs_mina, 17, 11, 1, low(mina)).
table(low_v_c_mina, 17, 10, 2, low(mina)).
table(low_v_minc_mina, 17, 11, 2, low(mina)).
table(low_v_maxc_mina, 17, 11, 2, low(mina)).
table(low_v_s_mina, 17, 4, 2, low(mina)).
table(low_v_mins_mina, 17, 4, 2, low(mina)).
table(low_v_maxs_mina, 17, 10, 2, low(mina)).
table(low_c_minc_mina, 17, 8, 2, low(mina)).
table(low_c_maxc_mina, 17, 6, 2, low(mina)).
table(low_c_s_mina, 17, 10, 2, low(mina)).
table(low_c_mins_mina, 17, 7, 2, low(mina)).
table(low_c_maxs_mina, 17, 5, 2, low(mina)).
table(low_minc_maxc_mina, 17, 19, 2, low(mina)).
table(low_minc_s_mina, 17, 11, 2, low(mina)).
table(low_minc_mins_mina, 17, 19, 2, low(mina)).
table(low_minc_maxs_mina, 17, 19, 2, low(mina)).
table(low_maxc_s_mina, 17, 11, 2, low(mina)).
table(low_maxc_mins_mina, 17, 19, 2, low(mina)).
table(low_maxc_maxs_mina, 17, 19, 2, low(mina)).
table(low_s_mins_mina, 17, 12, 2, low(mina)).
table(low_s_maxs_mina, 17, 10, 2, low(mina)).
table(low_mins_maxs_mina, 17, 6, 2, low(mina)).
table(low_v_c_minc_mina, 17, 13, 3, low(mina)).
table(low_v_c_maxc_mina, 17, 11, 3, low(mina)).
table(low_v_c_s_mina, 17, 4, 3, low(mina)).
table(low_v_c_mins_mina, 17, 7, 3, low(mina)).
table(low_v_c_maxs_mina, 17, 10, 3, low(mina)).
table(low_v_minc_maxc_mina, 17, 12, 3, low(mina)).
table(low_v_minc_s_mina, 17, 4, 3, low(mina)).
table(low_v_minc_mins_mina, 17, 4, 3, low(mina)).
table(low_v_minc_maxs_mina, 17, 11, 3, low(mina)).
table(low_v_maxc_s_mina, 17, 4, 3, low(mina)).
table(low_v_maxc_mins_mina, 17, 6, 3, low(mina)).
table(low_v_maxc_maxs_mina, 17, 10, 3, low(mina)).
table(low_v_s_mins_mina, 17, 5, 3, low(mina)).
table(low_v_s_maxs_mina, 17, 5, 3, low(mina)).
table(low_v_mins_maxs_mina, 17, 5, 3, low(mina)).
table(low_c_minc_maxc_mina, 17, 8, 3, low(mina)).
table(low_c_minc_s_mina, 17, 11, 3, low(mina)).
table(low_c_minc_mins_mina, 17, 8, 3, low(mina)).
table(low_c_minc_maxs_mina, 17, 6, 3, low(mina)).
table(low_c_maxc_s_mina, 17, 8, 3, low(mina)).
table(low_c_maxc_mins_mina, 17, 6, 3, low(mina)).
table(low_c_maxc_maxs_mina, 17, 6, 3, low(mina)).
table(low_c_s_mins_mina, 17, 10, 3, low(mina)).
table(low_c_s_maxs_mina, 17, 10, 3, low(mina)).
table(low_c_mins_maxs_mina, 17, 5, 3, low(mina)).
table(low_minc_maxc_s_mina, 17, 10, 3, low(mina)).
table(low_minc_maxc_mins_mina, 17, 16, 3, low(mina)).
table(low_minc_maxc_maxs_mina, 17, 18, 3, low(mina)).
table(low_minc_s_mins_mina, 17, 6, 3, low(mina)).
table(low_minc_s_maxs_mina, 17, 8, 3, low(mina)).
table(low_minc_mins_maxs_mina, 17, 17, 3, low(mina)).
table(low_maxc_s_mins_mina, 17, 8, 3, low(mina)).
table(low_maxc_s_maxs_mina, 17, 7, 3, low(mina)).
table(low_maxc_mins_maxs_mina, 17, 8, 3, low(mina)).
table(low_s_mins_maxs_mina, 17, 10, 3, low(mina)).
table(low_v_c_minc_maxc_mina, 17, 14, 4, low(mina)).
table(low_v_c_minc_s_mina, 17, 8, 4, low(mina)).
table(low_v_c_minc_mins_mina, 17, 10, 4, low(mina)).
table(low_v_c_minc_maxs_mina, 17, 13, 4, low(mina)).
table(low_v_c_maxc_s_mina, 17, 5, 4, low(mina)).
table(low_v_c_maxc_mins_mina, 17, 8, 4, low(mina)).
table(low_v_c_maxc_maxs_mina, 17, 11, 4, low(mina)).
table(low_v_c_s_mins_mina, 17, 5, 4, low(mina)).
table(low_v_c_s_maxs_mina, 17, 5, 4, low(mina)).
table(low_v_c_mins_maxs_mina, 17, 8, 4, low(mina)).
table(low_v_minc_maxc_s_mina, 17, 6, 4, low(mina)).
table(low_v_minc_maxc_mins_mina, 17, 6, 4, low(mina)).
table(low_v_minc_maxc_maxs_mina, 17, 12, 4, low(mina)).
table(low_v_minc_s_mins_mina, 17, 5, 4, low(mina)).
table(low_v_minc_s_maxs_mina, 17, 5, 4, low(mina)).
table(low_v_minc_mins_maxs_mina, 17, 6, 4, low(mina)).
table(low_v_maxc_s_mins_mina, 17, 5, 4, low(mina)).
table(low_v_maxc_s_maxs_mina, 17, 5, 4, low(mina)).
table(low_v_maxc_mins_maxs_mina, 17, 5, 4, low(mina)).
table(low_v_s_mins_maxs_mina, 17, 6, 4, low(mina)).
table(low_c_minc_maxc_s_mina, 17, 9, 4, low(mina)).
table(low_c_minc_maxc_mins_mina, 17, 7, 4, low(mina)).
table(low_c_minc_maxc_maxs_mina, 17, 8, 4, low(mina)).
table(low_c_minc_s_mins_mina, 17, 10, 4, low(mina)).
table(low_c_minc_s_maxs_mina, 17, 10, 4, low(mina)).
table(low_c_minc_mins_maxs_mina, 17, 6, 4, low(mina)).
table(low_c_maxc_s_mins_mina, 17, 6, 4, low(mina)).
table(low_c_maxc_s_maxs_mina, 17, 6, 4, low(mina)).
table(low_c_maxc_mins_maxs_mina, 17, 6, 4, low(mina)).
table(low_c_s_mins_maxs_mina, 17, 10, 4, low(mina)).
table(low_minc_maxc_s_mins_mina, 17, 7, 4, low(mina)).
table(low_minc_maxc_s_maxs_mina, 17, 9, 4, low(mina)).
table(low_minc_maxc_mins_maxs_mina, 17, 16, 4, low(mina)).
table(low_minc_s_mins_maxs_mina, 17, 7, 4, low(mina)).
table(low_maxc_s_mins_maxs_mina, 17, 6, 4, low(mina)).
table(low_v_c_minc_maxc_s_mina, 17, 9, 5, low(mina)).
table(low_v_c_minc_maxc_mins_mina, 17, 11, 5, low(mina)).
table(low_v_c_minc_maxc_maxs_mina, 17, 14, 5, low(mina)).
table(low_v_c_minc_s_mins_mina, 17, 8, 5, low(mina)).
table(low_v_c_minc_s_maxs_mina, 17, 8, 5, low(mina)).
table(low_v_c_minc_mins_maxs_mina, 17, 11, 5, low(mina)).
table(low_v_c_maxc_s_mins_mina, 17, 6, 5, low(mina)).
table(low_v_c_maxc_s_maxs_mina, 17, 6, 5, low(mina)).
table(low_v_c_maxc_mins_maxs_mina, 17, 9, 5, low(mina)).
table(low_v_c_s_mins_maxs_mina, 17, 6, 5, low(mina)).
table(low_v_minc_maxc_s_mins_mina, 17, 7, 5, low(mina)).
table(low_v_minc_maxc_s_maxs_mina, 17, 7, 5, low(mina)).
table(low_v_minc_maxc_mins_maxs_mina, 17, 7, 5, low(mina)).
table(low_v_minc_s_mins_maxs_mina, 17, 6, 5, low(mina)).
table(low_v_maxc_s_mins_maxs_mina, 17, 6, 5, low(mina)).
table(low_c_minc_maxc_s_mins_mina, 17, 8, 5, low(mina)).
table(low_c_minc_maxc_s_maxs_mina, 17, 8, 5, low(mina)).
table(low_c_minc_maxc_mins_maxs_mina, 17, 8, 5, low(mina)).
table(low_c_minc_s_mins_maxs_mina, 17, 11, 5, low(mina)).
table(low_c_maxc_s_mins_maxs_mina, 17, 7, 5, low(mina)).
table(low_minc_maxc_s_mins_maxs_mina, 17, 8, 5, low(mina)).
table(low_v_c_minc_maxc_s_mins_mina, 17, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_maxs_mina, 17, 9, 6, low(mina)).
table(low_v_c_minc_maxc_mins_maxs_mina, 17, 12, 6, low(mina)).
table(low_v_c_minc_s_mins_maxs_mina, 17, 9, 6, low(mina)).
table(low_v_c_maxc_s_mins_maxs_mina, 17, 7, 6, low(mina)).
table(low_v_minc_maxc_s_mins_maxs_mina, 17, 8, 6, low(mina)).
table(low_c_minc_maxc_s_mins_maxs_mina, 17, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_mins_maxs_mina, 17, 10, 7, low(mina)).
table(up_v_maxa, 17, 19, 1, up(maxa)).
table(up_v_c_maxa, 17, 19, 2, up(maxa)).
table(up_v_minc_maxa, 17, 19, 2, up(maxa)).
table(up_v_maxc_maxa, 17, 19, 2, up(maxa)).
table(up_v_s_maxa, 17, 19, 2, up(maxa)).
table(up_v_mins_maxa, 17, 19, 2, up(maxa)).
table(up_v_maxs_maxa, 17, 19, 2, up(maxa)).
table(up_c_maxc_maxa, 17, 19, 2, up(maxa)).
table(up_maxc_s_maxa, 17, 18, 2, up(maxa)).
table(up_s_maxs_maxa, 17, 19, 2, up(maxa)).
table(up_v_c_minc_maxa, 17, 19, 3, up(maxa)).
table(up_v_c_maxc_maxa, 17, 19, 3, up(maxa)).
table(up_v_c_s_maxa, 17, 19, 3, up(maxa)).
table(up_v_c_mins_maxa, 17, 19, 3, up(maxa)).
table(up_v_c_maxs_maxa, 17, 7, 3, up(maxa)).
table(up_v_minc_maxc_maxa, 17, 19, 3, up(maxa)).
table(up_v_minc_s_maxa, 17, 18, 3, up(maxa)).
table(up_v_minc_mins_maxa, 17, 19, 3, up(maxa)).
table(up_v_minc_maxs_maxa, 17, 19, 3, up(maxa)).
table(up_v_maxc_s_maxa, 17, 18, 3, up(maxa)).
table(up_v_maxc_mins_maxa, 17, 11, 3, up(maxa)).
table(up_v_maxc_maxs_maxa, 17, 6, 3, up(maxa)).
table(up_v_s_mins_maxa, 17, 19, 3, up(maxa)).
table(up_v_s_maxs_maxa, 17, 19, 3, up(maxa)).
table(up_v_mins_maxs_maxa, 17, 19, 3, up(maxa)).
table(up_c_minc_maxc_maxa, 17, 19, 3, up(maxa)).
table(up_c_maxc_s_maxa, 17, 18, 3, up(maxa)).
table(up_c_maxc_mins_maxa, 17, 19, 3, up(maxa)).
table(up_c_maxc_maxs_maxa, 17, 7, 3, up(maxa)).
table(up_c_s_maxs_maxa, 17, 19, 3, up(maxa)).
table(up_minc_maxc_s_maxa, 17, 5, 3, up(maxa)).
table(up_minc_s_maxs_maxa, 17, 18, 3, up(maxa)).
table(up_maxc_s_mins_maxa, 17, 8, 3, up(maxa)).
table(up_maxc_s_maxs_maxa, 17, 6, 3, up(maxa)).
table(up_s_mins_maxs_maxa, 17, 19, 3, up(maxa)).
table(up_v_c_minc_maxc_maxa, 17, 19, 4, up(maxa)).
table(up_v_c_minc_s_maxa, 17, 18, 4, up(maxa)).
table(up_v_c_minc_mins_maxa, 17, 11, 4, up(maxa)).
table(up_v_c_minc_maxs_maxa, 17, 8, 4, up(maxa)).
table(up_v_c_maxc_s_maxa, 17, 18, 4, up(maxa)).
table(up_v_c_maxc_mins_maxa, 17, 11, 4, up(maxa)).
table(up_v_c_maxc_maxs_maxa, 17, 6, 4, up(maxa)).
table(up_v_c_s_mins_maxa, 17, 19, 4, up(maxa)).
table(up_v_c_s_maxs_maxa, 17, 19, 4, up(maxa)).
table(up_v_c_mins_maxs_maxa, 17, 9, 4, up(maxa)).
table(up_v_minc_maxc_s_maxa, 17, 18, 4, up(maxa)).
table(up_v_minc_maxc_mins_maxa, 17, 14, 4, up(maxa)).
table(up_v_minc_maxc_maxs_maxa, 17, 9, 4, up(maxa)).
table(up_v_minc_s_mins_maxa, 17, 18, 4, up(maxa)).
table(up_v_minc_s_maxs_maxa, 17, 18, 4, up(maxa)).
table(up_v_minc_mins_maxs_maxa, 17, 18, 4, up(maxa)).
table(up_v_maxc_s_mins_maxa, 17, 8, 4, up(maxa)).
table(up_v_maxc_s_maxs_maxa, 17, 8, 4, up(maxa)).
table(up_v_maxc_mins_maxs_maxa, 17, 6, 4, up(maxa)).
table(up_v_s_mins_maxs_maxa, 17, 19, 4, up(maxa)).
table(up_c_minc_maxc_s_maxa, 17, 18, 4, up(maxa)).
table(up_c_minc_maxc_mins_maxa, 17, 15, 4, up(maxa)).
table(up_c_minc_maxc_maxs_maxa, 17, 9, 4, up(maxa)).
table(up_c_minc_s_maxs_maxa, 17, 9, 4, up(maxa)).
table(up_c_maxc_s_mins_maxa, 17, 8, 4, up(maxa)).
table(up_c_maxc_s_maxs_maxa, 17, 10, 4, up(maxa)).
table(up_c_maxc_mins_maxs_maxa, 17, 6, 4, up(maxa)).
table(up_c_s_mins_maxs_maxa, 17, 19, 4, up(maxa)).
table(up_minc_maxc_s_mins_maxa, 17, 7, 4, up(maxa)).
table(up_minc_maxc_s_maxs_maxa, 17, 6, 4, up(maxa)).
table(up_minc_s_mins_maxs_maxa, 17, 18, 4, up(maxa)).
table(up_maxc_s_mins_maxs_maxa, 17, 5, 4, up(maxa)).
table(up_v_c_minc_maxc_s_maxa, 17, 18, 5, up(maxa)).
table(up_v_c_minc_maxc_mins_maxa, 17, 14, 5, up(maxa)).
table(up_v_c_minc_maxc_maxs_maxa, 17, 9, 5, up(maxa)).
table(up_v_c_minc_s_mins_maxa, 17, 10, 5, up(maxa)).
table(up_v_c_minc_s_maxs_maxa, 17, 9, 5, up(maxa)).
table(up_v_c_minc_mins_maxs_maxa, 17, 8, 5, up(maxa)).
table(up_v_c_maxc_s_mins_maxa, 17, 8, 5, up(maxa)).
table(up_v_c_maxc_s_maxs_maxa, 17, 6, 5, up(maxa)).
table(up_v_c_maxc_mins_maxs_maxa, 17, 6, 5, up(maxa)).
table(up_v_c_s_mins_maxs_maxa, 17, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_mins_maxa, 17, 11, 5, up(maxa)).
table(up_v_minc_maxc_s_maxs_maxa, 17, 10, 5, up(maxa)).
table(up_v_minc_maxc_mins_maxs_maxa, 17, 9, 5, up(maxa)).
table(up_v_minc_s_mins_maxs_maxa, 17, 18, 5, up(maxa)).
table(up_v_maxc_s_mins_maxs_maxa, 17, 8, 5, up(maxa)).
table(up_c_minc_maxc_s_mins_maxa, 17, 11, 5, up(maxa)).
table(up_c_minc_maxc_s_maxs_maxa, 17, 9, 5, up(maxa)).
table(up_c_minc_maxc_mins_maxs_maxa, 17, 9, 5, up(maxa)).
table(up_c_minc_s_mins_maxs_maxa, 17, 10, 5, up(maxa)).
table(up_c_maxc_s_mins_maxs_maxa, 17, 7, 5, up(maxa)).
table(up_minc_maxc_s_mins_maxs_maxa, 17, 7, 5, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxa, 17, 11, 6, up(maxa)).
table(up_v_c_minc_maxc_s_maxs_maxa, 17, 9, 6, up(maxa)).
table(up_v_c_minc_maxc_mins_maxs_maxa, 17, 9, 6, up(maxa)).
table(up_v_c_minc_s_mins_maxs_maxa, 17, 10, 6, up(maxa)).
table(up_v_c_maxc_s_mins_maxs_maxa, 17, 7, 6, up(maxa)).
table(up_v_minc_maxc_s_mins_maxs_maxa, 17, 11, 6, up(maxa)).
table(up_c_minc_maxc_s_mins_maxs_maxa, 17, 10, 6, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxs_maxa, 17, 10, 7, up(maxa)).
table(low_c_v, 18, 18, 1, low(v)).
table(low_minc_v, 18, 11, 1, low(v)).
table(low_maxc_v, 18, 11, 1, low(v)).
table(low_s_v, 18, 7, 1, low(v)).
table(low_mins_v, 18, 18, 1, low(v)).
table(low_maxs_v, 18, 18, 1, low(v)).
table(low_c_minc_v, 18, 11, 2, low(v)).
table(low_c_maxc_v, 18, 11, 2, low(v)).
table(low_c_s_v, 18, 8, 2, low(v)).
table(low_c_mins_v, 18, 18, 2, low(v)).
table(low_c_maxs_v, 18, 18, 2, low(v)).
table(low_minc_maxc_v, 18, 11, 2, low(v)).
table(low_minc_s_v, 18, 3, 2, low(v)).
table(low_minc_mins_v, 18, 12, 2, low(v)).
table(low_minc_maxs_v, 18, 12, 2, low(v)).
table(low_maxc_s_v, 18, 3, 2, low(v)).
table(low_maxc_mins_v, 18, 12, 2, low(v)).
table(low_maxc_maxs_v, 18, 13, 2, low(v)).
table(low_s_mins_v, 18, 7, 2, low(v)).
table(low_s_maxs_v, 18, 7, 2, low(v)).
table(low_mins_maxs_v, 18, 12, 2, low(v)).
table(low_c_minc_maxc_v, 18, 11, 3, low(v)).
table(low_c_minc_s_v, 18, 6, 3, low(v)).
table(low_c_minc_mins_v, 18, 12, 3, low(v)).
table(low_c_minc_maxs_v, 18, 12, 3, low(v)).
table(low_c_maxc_s_v, 18, 4, 3, low(v)).
table(low_c_maxc_mins_v, 18, 12, 3, low(v)).
table(low_c_maxc_maxs_v, 18, 13, 3, low(v)).
table(low_c_s_mins_v, 18, 8, 3, low(v)).
table(low_c_s_maxs_v, 18, 8, 3, low(v)).
table(low_c_mins_maxs_v, 18, 18, 3, low(v)).
table(low_minc_maxc_s_v, 18, 5, 3, low(v)).
table(low_minc_maxc_mins_v, 18, 12, 3, low(v)).
table(low_minc_maxc_maxs_v, 18, 12, 3, low(v)).
table(low_minc_s_mins_v, 18, 4, 3, low(v)).
table(low_minc_s_maxs_v, 18, 4, 3, low(v)).
table(low_minc_mins_maxs_v, 18, 13, 3, low(v)).
table(low_maxc_s_mins_v, 18, 4, 3, low(v)).
table(low_maxc_s_maxs_v, 18, 4, 3, low(v)).
table(low_maxc_mins_maxs_v, 18, 14, 3, low(v)).
table(low_s_mins_maxs_v, 18, 7, 3, low(v)).
table(low_c_minc_maxc_s_v, 18, 7, 4, low(v)).
table(low_c_minc_maxc_mins_v, 18, 12, 4, low(v)).
table(low_c_minc_maxc_maxs_v, 18, 12, 4, low(v)).
table(low_c_minc_s_mins_v, 18, 7, 4, low(v)).
table(low_c_minc_s_maxs_v, 18, 7, 4, low(v)).
table(low_c_minc_mins_maxs_v, 18, 13, 4, low(v)).
table(low_c_maxc_s_mins_v, 18, 5, 4, low(v)).
table(low_c_maxc_s_maxs_v, 18, 5, 4, low(v)).
table(low_c_maxc_mins_maxs_v, 18, 14, 4, low(v)).
table(low_c_s_mins_maxs_v, 18, 8, 4, low(v)).
table(low_minc_maxc_s_mins_v, 18, 6, 4, low(v)).
table(low_minc_maxc_s_maxs_v, 18, 6, 4, low(v)).
table(low_minc_maxc_mins_maxs_v, 18, 13, 4, low(v)).
table(low_minc_s_mins_maxs_v, 18, 5, 4, low(v)).
table(low_maxc_s_mins_maxs_v, 18, 5, 4, low(v)).
table(low_c_minc_maxc_s_mins_v, 18, 8, 5, low(v)).
table(low_c_minc_maxc_s_maxs_v, 18, 8, 5, low(v)).
table(low_c_minc_maxc_mins_maxs_v, 18, 13, 5, low(v)).
table(low_c_minc_s_mins_maxs_v, 18, 8, 5, low(v)).
table(low_c_maxc_s_mins_maxs_v, 18, 6, 5, low(v)).
table(low_minc_maxc_s_mins_maxs_v, 18, 7, 5, low(v)).
table(low_c_minc_maxc_s_mins_maxs_v, 18, 9, 6, low(v)).
table(up_c_maxc_v, 18, 3, 2, up(v)).
table(up_maxc_s_v, 18, 3, 2, up(v)).
table(up_s_maxs_v, 18, 3, 2, up(v)).
table(up_c_minc_maxc_v, 18, 6, 3, up(v)).
table(up_c_maxc_s_v, 18, 4, 3, up(v)).
table(up_c_maxc_mins_v, 18, 4, 3, up(v)).
table(up_c_maxc_maxs_v, 18, 4, 3, up(v)).
table(up_c_s_maxs_v, 18, 4, 3, up(v)).
table(up_minc_maxc_s_v, 18, 5, 3, up(v)).
table(up_minc_s_maxs_v, 18, 4, 3, up(v)).
table(up_maxc_s_mins_v, 18, 4, 3, up(v)).
table(up_maxc_s_maxs_v, 18, 4, 3, up(v)).
table(up_s_mins_maxs_v, 18, 4, 3, up(v)).
table(up_c_minc_maxc_s_v, 18, 7, 4, up(v)).
table(up_c_minc_maxc_mins_v, 18, 7, 4, up(v)).
table(up_c_minc_maxc_maxs_v, 18, 7, 4, up(v)).
table(up_c_minc_s_maxs_v, 18, 7, 4, up(v)).
table(up_c_maxc_s_mins_v, 18, 5, 4, up(v)).
table(up_c_maxc_s_maxs_v, 18, 5, 4, up(v)).
table(up_c_maxc_mins_maxs_v, 18, 5, 4, up(v)).
table(up_c_s_mins_maxs_v, 18, 5, 4, up(v)).
table(up_minc_maxc_s_mins_v, 18, 6, 4, up(v)).
table(up_minc_maxc_s_maxs_v, 18, 6, 4, up(v)).
table(up_minc_s_mins_maxs_v, 18, 5, 4, up(v)).
table(up_maxc_s_mins_maxs_v, 18, 5, 4, up(v)).
table(up_c_minc_maxc_s_mins_v, 18, 8, 5, up(v)).
table(up_c_minc_maxc_s_maxs_v, 18, 8, 5, up(v)).
table(up_c_minc_maxc_mins_maxs_v, 18, 8, 5, up(v)).
table(up_c_minc_s_mins_maxs_v, 18, 8, 5, up(v)).
table(up_c_maxc_s_mins_maxs_v, 18, 6, 5, up(v)).
table(up_minc_maxc_s_mins_maxs_v, 18, 7, 5, up(v)).
table(up_c_minc_maxc_s_mins_maxs_v, 18, 9, 6, up(v)).
table(low_v_c, 18, 11, 1, low(c)).
table(low_minc_c, 18, 11, 1, low(c)).
table(low_maxc_c, 18, 11, 1, low(c)).
table(low_s_c, 18, 8, 1, low(c)).
table(low_mins_c, 18, 8, 1, low(c)).
table(low_maxs_c, 18, 8, 1, low(c)).
table(low_v_minc_c, 18, 11, 2, low(c)).
table(low_v_maxc_c, 18, 3, 2, low(c)).
table(low_v_s_c, 18, 13, 2, low(c)).
table(low_v_mins_c, 18, 12, 2, low(c)).
table(low_v_maxs_c, 18, 13, 2, low(c)).
table(low_minc_maxc_c, 18, 11, 2, low(c)).
table(low_minc_s_c, 18, 5, 2, low(c)).
table(low_minc_mins_c, 18, 9, 2, low(c)).
table(low_minc_maxs_c, 18, 9, 2, low(c)).
table(low_maxc_s_c, 18, 3, 2, low(c)).
table(low_maxc_mins_c, 18, 12, 2, low(c)).
table(low_maxc_maxs_c, 18, 13, 2, low(c)).
table(low_s_mins_c, 18, 10, 2, low(c)).
table(low_s_maxs_c, 18, 11, 2, low(c)).
table(low_mins_maxs_c, 18, 9, 2, low(c)).
table(low_v_minc_maxc_c, 18, 6, 3, low(c)).
table(low_v_minc_s_c, 18, 12, 3, low(c)).
table(low_v_minc_mins_c, 18, 12, 3, low(c)).
table(low_v_minc_maxs_c, 18, 12, 3, low(c)).
table(low_v_maxc_s_c, 18, 4, 3, low(c)).
table(low_v_maxc_mins_c, 18, 4, 3, low(c)).
table(low_v_maxc_maxs_c, 18, 4, 3, low(c)).
table(low_v_s_mins_c, 18, 14, 3, low(c)).
table(low_v_s_maxs_c, 18, 14, 3, low(c)).
table(low_v_mins_maxs_c, 18, 14, 3, low(c)).
table(low_minc_maxc_s_c, 18, 5, 3, low(c)).
table(low_minc_maxc_mins_c, 18, 12, 3, low(c)).
table(low_minc_maxc_maxs_c, 18, 12, 3, low(c)).
table(low_minc_s_mins_c, 18, 6, 3, low(c)).
table(low_minc_s_maxs_c, 18, 7, 3, low(c)).
table(low_minc_mins_maxs_c, 18, 7, 3, low(c)).
table(low_maxc_s_mins_c, 18, 4, 3, low(c)).
table(low_maxc_s_maxs_c, 18, 4, 3, low(c)).
table(low_maxc_mins_maxs_c, 18, 6, 3, low(c)).
table(low_s_mins_maxs_c, 18, 12, 3, low(c)).
table(low_v_minc_maxc_s_c, 18, 7, 4, low(c)).
table(low_v_minc_maxc_mins_c, 18, 7, 4, low(c)).
table(low_v_minc_maxc_maxs_c, 18, 7, 4, low(c)).
table(low_v_minc_s_mins_c, 18, 13, 4, low(c)).
table(low_v_minc_s_maxs_c, 18, 13, 4, low(c)).
table(low_v_minc_mins_maxs_c, 18, 13, 4, low(c)).
table(low_v_maxc_s_mins_c, 18, 5, 4, low(c)).
table(low_v_maxc_s_maxs_c, 18, 5, 4, low(c)).
table(low_v_maxc_mins_maxs_c, 18, 5, 4, low(c)).
table(low_v_s_mins_maxs_c, 18, 15, 4, low(c)).
table(low_minc_maxc_s_mins_c, 18, 6, 4, low(c)).
table(low_minc_maxc_s_maxs_c, 18, 6, 4, low(c)).
table(low_minc_maxc_mins_maxs_c, 18, 8, 4, low(c)).
table(low_minc_s_mins_maxs_c, 18, 8, 4, low(c)).
table(low_maxc_s_mins_maxs_c, 18, 5, 4, low(c)).
table(low_v_minc_maxc_s_mins_c, 18, 8, 5, low(c)).
table(low_v_minc_maxc_s_maxs_c, 18, 8, 5, low(c)).
table(low_v_minc_maxc_mins_maxs_c, 18, 8, 5, low(c)).
table(low_v_minc_s_mins_maxs_c, 18, 14, 5, low(c)).
table(low_v_maxc_s_mins_maxs_c, 18, 6, 5, low(c)).
table(low_minc_maxc_s_mins_maxs_c, 18, 7, 5, low(c)).
table(low_v_minc_maxc_s_mins_maxs_c, 18, 9, 6, low(c)).
table(up_v_c, 18, 18, 1, up(c)).
table(up_s_c, 18, 3, 1, up(c)).
table(up_v_minc_c, 18, 6, 2, up(c)).
table(up_v_maxc_c, 18, 11, 2, up(c)).
table(up_v_s_c, 18, 4, 2, up(c)).
table(up_v_mins_c, 18, 10, 2, up(c)).
table(up_v_maxs_c, 18, 18, 2, up(c)).
table(up_minc_s_c, 18, 3, 2, up(c)).
table(up_maxc_s_c, 18, 4, 2, up(c)).
table(up_s_mins_c, 18, 5, 2, up(c)).
table(up_s_maxs_c, 18, 5, 2, up(c)).
table(up_v_minc_maxc_c, 18, 7, 3, up(c)).
table(up_v_minc_s_c, 18, 6, 3, up(c)).
table(up_v_minc_mins_c, 18, 7, 3, up(c)).
table(up_v_minc_maxs_c, 18, 7, 3, up(c)).
table(up_v_maxc_s_c, 18, 5, 3, up(c)).
table(up_v_maxc_mins_c, 18, 5, 3, up(c)).
table(up_v_maxc_maxs_c, 18, 13, 3, up(c)).
table(up_v_s_mins_c, 18, 8, 3, up(c)).
table(up_v_s_maxs_c, 18, 6, 3, up(c)).
table(up_v_mins_maxs_c, 18, 12, 3, up(c)).
table(up_minc_maxc_s_c, 18, 5, 3, up(c)).
table(up_minc_s_mins_c, 18, 4, 3, up(c)).
table(up_minc_s_maxs_c, 18, 4, 3, up(c)).
table(up_maxc_s_mins_c, 18, 5, 3, up(c)).
table(up_maxc_s_maxs_c, 18, 5, 3, up(c)).
table(up_s_mins_maxs_c, 18, 8, 3, up(c)).
table(up_v_minc_maxc_s_c, 18, 7, 4, up(c)).
table(up_v_minc_maxc_mins_c, 18, 8, 4, up(c)).
table(up_v_minc_maxc_maxs_c, 18, 8, 4, up(c)).
table(up_v_minc_s_mins_c, 18, 7, 4, up(c)).
table(up_v_minc_s_maxs_c, 18, 7, 4, up(c)).
table(up_v_minc_mins_maxs_c, 18, 8, 4, up(c)).
table(up_v_maxc_s_mins_c, 18, 6, 4, up(c)).
table(up_v_maxc_s_maxs_c, 18, 6, 4, up(c)).
table(up_v_maxc_mins_maxs_c, 18, 7, 4, up(c)).
table(up_v_s_mins_maxs_c, 18, 10, 4, up(c)).
table(up_minc_maxc_s_mins_c, 18, 6, 4, up(c)).
table(up_minc_maxc_s_maxs_c, 18, 6, 4, up(c)).
table(up_minc_s_mins_maxs_c, 18, 5, 4, up(c)).
table(up_maxc_s_mins_maxs_c, 18, 6, 4, up(c)).
table(up_v_minc_maxc_s_mins_c, 18, 8, 5, up(c)).
table(up_v_minc_maxc_s_maxs_c, 18, 8, 5, up(c)).
table(up_v_minc_maxc_mins_maxs_c, 18, 9, 5, up(c)).
table(up_v_minc_s_mins_maxs_c, 18, 8, 5, up(c)).
table(up_v_maxc_s_mins_maxs_c, 18, 7, 5, up(c)).
table(up_minc_maxc_s_mins_maxs_c, 18, 7, 5, up(c)).
table(up_v_minc_maxc_s_mins_maxs_c, 18, 9, 6, up(c)).
table(low_v_minc, 18, 3, 1, low(minc)).
table(low_c_minc, 18, 3, 1, low(minc)).
table(low_maxc_minc, 18, 3, 1, low(minc)).
table(low_s_minc, 18, 3, 1, low(minc)).
table(low_mins_minc, 18, 2, 1, low(minc)).
table(low_maxs_minc, 18, 3, 1, low(minc)).
table(low_v_c_minc, 18, 5, 2, low(minc)).
table(low_v_maxc_minc, 18, 4, 2, low(minc)).
table(low_v_s_minc, 18, 4, 2, low(minc)).
table(low_v_mins_minc, 18, 3, 2, low(minc)).
table(low_v_maxs_minc, 18, 4, 2, low(minc)).
table(low_c_maxc_minc, 18, 4, 2, low(minc)).
table(low_c_s_minc, 18, 4, 2, low(minc)).
table(low_c_mins_minc, 18, 3, 2, low(minc)).
table(low_c_maxs_minc, 18, 4, 2, low(minc)).
table(low_maxc_s_minc, 18, 4, 2, low(minc)).
table(low_maxc_mins_minc, 18, 4, 2, low(minc)).
table(low_maxc_maxs_minc, 18, 4, 2, low(minc)).
table(low_s_mins_minc, 18, 3, 2, low(minc)).
table(low_s_maxs_minc, 18, 4, 2, low(minc)).
table(low_mins_maxs_minc, 18, 3, 2, low(minc)).
table(low_v_c_maxc_minc, 18, 6, 3, low(minc)).
table(low_v_c_s_minc, 18, 6, 3, low(minc)).
table(low_v_c_mins_minc, 18, 6, 3, low(minc)).
table(low_v_c_maxs_minc, 18, 6, 3, low(minc)).
table(low_v_maxc_s_minc, 18, 5, 3, low(minc)).
table(low_v_maxc_mins_minc, 18, 5, 3, low(minc)).
table(low_v_maxc_maxs_minc, 18, 5, 3, low(minc)).
table(low_v_s_mins_minc, 18, 4, 3, low(minc)).
table(low_v_s_maxs_minc, 18, 5, 3, low(minc)).
table(low_v_mins_maxs_minc, 18, 4, 3, low(minc)).
table(low_c_maxc_s_minc, 18, 5, 3, low(minc)).
table(low_c_maxc_mins_minc, 18, 5, 3, low(minc)).
table(low_c_maxc_maxs_minc, 18, 5, 3, low(minc)).
table(low_c_s_mins_minc, 18, 4, 3, low(minc)).
table(low_c_s_maxs_minc, 18, 6, 3, low(minc)).
table(low_c_mins_maxs_minc, 18, 4, 3, low(minc)).
table(low_maxc_s_mins_minc, 18, 5, 3, low(minc)).
table(low_maxc_s_maxs_minc, 18, 5, 3, low(minc)).
table(low_maxc_mins_maxs_minc, 18, 5, 3, low(minc)).
table(low_s_mins_maxs_minc, 18, 4, 3, low(minc)).
table(low_v_c_maxc_s_minc, 18, 7, 4, low(minc)).
table(low_v_c_maxc_mins_minc, 18, 7, 4, low(minc)).
table(low_v_c_maxc_maxs_minc, 18, 7, 4, low(minc)).
table(low_v_c_s_mins_minc, 18, 7, 4, low(minc)).
table(low_v_c_s_maxs_minc, 18, 7, 4, low(minc)).
table(low_v_c_mins_maxs_minc, 18, 7, 4, low(minc)).
table(low_v_maxc_s_mins_minc, 18, 6, 4, low(minc)).
table(low_v_maxc_s_maxs_minc, 18, 6, 4, low(minc)).
table(low_v_maxc_mins_maxs_minc, 18, 6, 4, low(minc)).
table(low_v_s_mins_maxs_minc, 18, 5, 4, low(minc)).
table(low_c_maxc_s_mins_minc, 18, 6, 4, low(minc)).
table(low_c_maxc_s_maxs_minc, 18, 6, 4, low(minc)).
table(low_c_maxc_mins_maxs_minc, 18, 6, 4, low(minc)).
table(low_c_s_mins_maxs_minc, 18, 6, 4, low(minc)).
table(low_maxc_s_mins_maxs_minc, 18, 6, 4, low(minc)).
table(low_v_c_maxc_s_mins_minc, 18, 8, 5, low(minc)).
table(low_v_c_maxc_s_maxs_minc, 18, 8, 5, low(minc)).
table(low_v_c_maxc_mins_maxs_minc, 18, 8, 5, low(minc)).
table(low_v_c_s_mins_maxs_minc, 18, 8, 5, low(minc)).
table(low_v_maxc_s_mins_maxs_minc, 18, 7, 5, low(minc)).
table(low_c_maxc_s_mins_maxs_minc, 18, 7, 5, low(minc)).
table(low_v_c_maxc_s_mins_maxs_minc, 18, 9, 6, low(minc)).
table(up_v_minc, 18, 11, 1, up(minc)).
table(up_maxc_minc, 18, 6, 1, up(minc)).
table(up_v_c_minc, 18, 5, 2, up(minc)).
table(up_v_maxc_minc, 18, 4, 2, up(minc)).
table(up_v_s_minc, 18, 13, 2, up(minc)).
table(up_v_mins_minc, 18, 12, 2, up(minc)).
table(up_v_maxs_minc, 18, 13, 2, up(minc)).
table(up_c_maxc_minc, 18, 4, 2, up(minc)).
table(up_maxc_s_minc, 18, 4, 2, up(minc)).
table(up_maxc_mins_minc, 18, 4, 2, up(minc)).
table(up_maxc_maxs_minc, 18, 7, 2, up(minc)).
table(up_s_maxs_minc, 18, 14, 2, up(minc)).
table(up_v_c_maxc_minc, 18, 6, 3, up(minc)).
table(up_v_c_s_minc, 18, 6, 3, up(minc)).
table(up_v_c_mins_minc, 18, 6, 3, up(minc)).
table(up_v_c_maxs_minc, 18, 6, 3, up(minc)).
table(up_v_maxc_s_minc, 18, 5, 3, up(minc)).
table(up_v_maxc_mins_minc, 18, 5, 3, up(minc)).
table(up_v_maxc_maxs_minc, 18, 5, 3, up(minc)).
table(up_v_s_mins_minc, 18, 14, 3, up(minc)).
table(up_v_s_maxs_minc, 18, 14, 3, up(minc)).
table(up_v_mins_maxs_minc, 18, 14, 3, up(minc)).
table(up_c_maxc_s_minc, 18, 5, 3, up(minc)).
table(up_c_maxc_mins_minc, 18, 5, 3, up(minc)).
table(up_c_maxc_maxs_minc, 18, 5, 3, up(minc)).
table(up_c_s_maxs_minc, 18, 4, 3, up(minc)).
table(up_maxc_s_mins_minc, 18, 5, 3, up(minc)).
table(up_maxc_s_maxs_minc, 18, 5, 3, up(minc)).
table(up_maxc_mins_maxs_minc, 18, 5, 3, up(minc)).
table(up_s_mins_maxs_minc, 18, 15, 3, up(minc)).
table(up_v_c_maxc_s_minc, 18, 7, 4, up(minc)).
table(up_v_c_maxc_mins_minc, 18, 7, 4, up(minc)).
table(up_v_c_maxc_maxs_minc, 18, 7, 4, up(minc)).
table(up_v_c_s_mins_minc, 18, 7, 4, up(minc)).
table(up_v_c_s_maxs_minc, 18, 7, 4, up(minc)).
table(up_v_c_mins_maxs_minc, 18, 7, 4, up(minc)).
table(up_v_maxc_s_mins_minc, 18, 6, 4, up(minc)).
table(up_v_maxc_s_maxs_minc, 18, 6, 4, up(minc)).
table(up_v_maxc_mins_maxs_minc, 18, 6, 4, up(minc)).
table(up_v_s_mins_maxs_minc, 18, 15, 4, up(minc)).
table(up_c_maxc_s_mins_minc, 18, 6, 4, up(minc)).
table(up_c_maxc_s_maxs_minc, 18, 6, 4, up(minc)).
table(up_c_maxc_mins_maxs_minc, 18, 6, 4, up(minc)).
table(up_c_s_mins_maxs_minc, 18, 5, 4, up(minc)).
table(up_maxc_s_mins_maxs_minc, 18, 6, 4, up(minc)).
table(up_v_c_maxc_s_mins_minc, 18, 8, 5, up(minc)).
table(up_v_c_maxc_s_maxs_minc, 18, 8, 5, up(minc)).
table(up_v_c_maxc_mins_maxs_minc, 18, 8, 5, up(minc)).
table(up_v_c_s_mins_maxs_minc, 18, 8, 5, up(minc)).
table(up_v_maxc_s_mins_maxs_minc, 18, 7, 5, up(minc)).
table(up_c_maxc_s_mins_maxs_minc, 18, 7, 5, up(minc)).
table(up_v_c_maxc_s_mins_maxs_minc, 18, 9, 6, up(minc)).
table(low_v_maxc, 18, 18, 1, low(maxc)).
table(low_c_maxc, 18, 18, 1, low(maxc)).
table(low_minc_maxc, 18, 6, 1, low(maxc)).
table(low_s_maxc, 18, 18, 1, low(maxc)).
table(low_mins_maxc, 18, 9, 1, low(maxc)).
table(low_maxs_maxc, 18, 2, 1, low(maxc)).
table(low_v_c_maxc, 18, 3, 2, low(maxc)).
table(low_v_minc_maxc, 18, 6, 2, low(maxc)).
table(low_v_s_maxc, 18, 4, 2, low(maxc)).
table(low_v_mins_maxc, 18, 9, 2, low(maxc)).
table(low_v_maxs_maxc, 18, 3, 2, low(maxc)).
table(low_c_minc_maxc, 18, 11, 2, low(maxc)).
table(low_c_s_maxc, 18, 4, 2, low(maxc)).
table(low_c_mins_maxc, 18, 18, 2, low(maxc)).
table(low_c_maxs_maxc, 18, 3, 2, low(maxc)).
table(low_minc_s_maxc, 18, 6, 2, low(maxc)).
table(low_minc_mins_maxc, 18, 4, 2, low(maxc)).
table(low_minc_maxs_maxc, 18, 4, 2, low(maxc)).
table(low_s_mins_maxc, 18, 18, 2, low(maxc)).
table(low_s_maxs_maxc, 18, 3, 2, low(maxc)).
table(low_mins_maxs_maxc, 18, 3, 2, low(maxc)).
table(low_v_c_minc_maxc, 18, 6, 3, low(maxc)).
table(low_v_c_s_maxc, 18, 4, 3, low(maxc)).
table(low_v_c_mins_maxc, 18, 4, 3, low(maxc)).
table(low_v_c_maxs_maxc, 18, 4, 3, low(maxc)).
table(low_v_minc_s_maxc, 18, 5, 3, low(maxc)).
table(low_v_minc_mins_maxc, 18, 7, 3, low(maxc)).
table(low_v_minc_maxs_maxc, 18, 5, 3, low(maxc)).
table(low_v_s_mins_maxc, 18, 5, 3, low(maxc)).
table(low_v_s_maxs_maxc, 18, 4, 3, low(maxc)).
table(low_v_mins_maxs_maxc, 18, 4, 3, low(maxc)).
table(low_c_minc_s_maxc, 18, 5, 3, low(maxc)).
table(low_c_minc_mins_maxc, 18, 6, 3, low(maxc)).
table(low_c_minc_maxs_maxc, 18, 5, 3, low(maxc)).
table(low_c_s_mins_maxc, 18, 4, 3, low(maxc)).
table(low_c_s_maxs_maxc, 18, 4, 3, low(maxc)).
table(low_c_mins_maxs_maxc, 18, 4, 3, low(maxc)).
table(low_minc_s_mins_maxc, 18, 5, 3, low(maxc)).
table(low_minc_s_maxs_maxc, 18, 5, 3, low(maxc)).
table(low_minc_mins_maxs_maxc, 18, 5, 3, low(maxc)).
table(low_s_mins_maxs_maxc, 18, 4, 3, low(maxc)).
table(low_v_c_minc_s_maxc, 18, 7, 4, low(maxc)).
table(low_v_c_minc_mins_maxc, 18, 7, 4, low(maxc)).
table(low_v_c_minc_maxs_maxc, 18, 7, 4, low(maxc)).
table(low_v_c_s_mins_maxc, 18, 5, 4, low(maxc)).
table(low_v_c_s_maxs_maxc, 18, 5, 4, low(maxc)).
table(low_v_c_mins_maxs_maxc, 18, 5, 4, low(maxc)).
table(low_v_minc_s_mins_maxc, 18, 6, 4, low(maxc)).
table(low_v_minc_s_maxs_maxc, 18, 6, 4, low(maxc)).
table(low_v_minc_mins_maxs_maxc, 18, 6, 4, low(maxc)).
table(low_v_s_mins_maxs_maxc, 18, 5, 4, low(maxc)).
table(low_c_minc_s_mins_maxc, 18, 6, 4, low(maxc)).
table(low_c_minc_s_maxs_maxc, 18, 6, 4, low(maxc)).
table(low_c_minc_mins_maxs_maxc, 18, 6, 4, low(maxc)).
table(low_c_s_mins_maxs_maxc, 18, 5, 4, low(maxc)).
table(low_minc_s_mins_maxs_maxc, 18, 6, 4, low(maxc)).
table(low_v_c_minc_s_mins_maxc, 18, 8, 5, low(maxc)).
table(low_v_c_minc_s_maxs_maxc, 18, 8, 5, low(maxc)).
table(low_v_c_minc_mins_maxs_maxc, 18, 8, 5, low(maxc)).
table(low_v_c_s_mins_maxs_maxc, 18, 6, 5, low(maxc)).
table(low_v_minc_s_mins_maxs_maxc, 18, 7, 5, low(maxc)).
table(low_c_minc_s_mins_maxs_maxc, 18, 7, 5, low(maxc)).
table(low_v_c_minc_s_mins_maxs_maxc, 18, 9, 6, low(maxc)).
table(up_v_maxc, 18, 11, 1, up(maxc)).
table(up_v_c_maxc, 18, 11, 2, up(maxc)).
table(up_v_minc_maxc, 18, 11, 2, up(maxc)).
table(up_v_s_maxc, 18, 13, 2, up(maxc)).
table(up_v_mins_maxc, 18, 12, 2, up(maxc)).
table(up_v_maxs_maxc, 18, 13, 2, up(maxc)).
table(up_v_c_minc_maxc, 18, 11, 3, up(maxc)).
table(up_v_c_s_maxc, 18, 13, 3, up(maxc)).
table(up_v_c_mins_maxc, 18, 12, 3, up(maxc)).
table(up_v_c_maxs_maxc, 18, 13, 3, up(maxc)).
table(up_v_minc_s_maxc, 18, 12, 3, up(maxc)).
table(up_v_minc_mins_maxc, 18, 12, 3, up(maxc)).
table(up_v_minc_maxs_maxc, 18, 12, 3, up(maxc)).
table(up_v_s_mins_maxc, 18, 14, 3, up(maxc)).
table(up_v_s_maxs_maxc, 18, 14, 3, up(maxc)).
table(up_v_mins_maxs_maxc, 18, 14, 3, up(maxc)).
table(up_v_c_minc_s_maxc, 18, 12, 4, up(maxc)).
table(up_v_c_minc_mins_maxc, 18, 12, 4, up(maxc)).
table(up_v_c_minc_maxs_maxc, 18, 12, 4, up(maxc)).
table(up_v_c_s_mins_maxc, 18, 14, 4, up(maxc)).
table(up_v_c_s_maxs_maxc, 18, 6, 4, up(maxc)).
table(up_v_c_mins_maxs_maxc, 18, 10, 4, up(maxc)).
table(up_v_minc_s_mins_maxc, 18, 13, 4, up(maxc)).
table(up_v_minc_s_maxs_maxc, 18, 13, 4, up(maxc)).
table(up_v_minc_mins_maxs_maxc, 18, 13, 4, up(maxc)).
table(up_v_s_mins_maxs_maxc, 18, 15, 4, up(maxc)).
table(up_v_c_minc_s_mins_maxc, 18, 13, 5, up(maxc)).
table(up_v_c_minc_s_maxs_maxc, 18, 8, 5, up(maxc)).
table(up_v_c_minc_mins_maxs_maxc, 18, 9, 5, up(maxc)).
table(up_v_c_s_mins_maxs_maxc, 18, 10, 5, up(maxc)).
table(up_v_minc_s_mins_maxs_maxc, 18, 14, 5, up(maxc)).
table(up_v_c_minc_s_mins_maxs_maxc, 18, 9, 6, up(maxc)).
table(low_v_s, 18, 18, 1, low(s)).
table(low_c_s, 18, 3, 1, low(s)).
table(low_minc_s, 18, 18, 1, low(s)).
table(low_maxc_s, 18, 18, 1, low(s)).
table(low_mins_s, 18, 18, 1, low(s)).
table(low_maxs_s, 18, 18, 1, low(s)).
table(low_v_c_s, 18, 4, 2, low(s)).
table(low_v_minc_s, 18, 18, 2, low(s)).
table(low_v_maxc_s, 18, 6, 2, low(s)).
table(low_v_mins_s, 18, 9, 2, low(s)).
table(low_v_maxs_s, 18, 3, 2, low(s)).
table(low_c_minc_s, 18, 5, 2, low(s)).
table(low_c_maxc_s, 18, 5, 2, low(s)).
table(low_c_mins_s, 18, 5, 2, low(s)).
table(low_c_maxs_s, 18, 5, 2, low(s)).
table(low_minc_maxc_s, 18, 18, 2, low(s)).
table(low_minc_mins_s, 18, 12, 2, low(s)).
table(low_minc_maxs_s, 18, 14, 2, low(s)).
table(low_maxc_mins_s, 18, 7, 2, low(s)).
table(low_maxc_maxs_s, 18, 14, 2, low(s)).
table(low_mins_maxs_s, 18, 12, 2, low(s)).
table(low_v_c_minc_s, 18, 8, 3, low(s)).
table(low_v_c_maxc_s, 18, 6, 3, low(s)).
table(low_v_c_mins_s, 18, 8, 3, low(s)).
table(low_v_c_maxs_s, 18, 4, 3, low(s)).
table(low_v_minc_maxc_s, 18, 10, 3, low(s)).
table(low_v_minc_mins_s, 18, 13, 3, low(s)).
table(low_v_minc_maxs_s, 18, 4, 3, low(s)).
table(low_v_maxc_mins_s, 18, 4, 3, low(s)).
table(low_v_maxc_maxs_s, 18, 4, 3, low(s)).
table(low_v_mins_maxs_s, 18, 4, 3, low(s)).
table(low_c_minc_maxc_s, 18, 8, 3, low(s)).
table(low_c_minc_mins_s, 18, 4, 3, low(s)).
table(low_c_minc_maxs_s, 18, 5, 3, low(s)).
table(low_c_maxc_mins_s, 18, 8, 3, low(s)).
table(low_c_maxc_maxs_s, 18, 5, 3, low(s)).
table(low_c_mins_maxs_s, 18, 8, 3, low(s)).
table(low_minc_maxc_mins_s, 18, 13, 3, low(s)).
table(low_minc_maxc_maxs_s, 18, 14, 3, low(s)).
table(low_minc_mins_maxs_s, 18, 8, 3, low(s)).
table(low_maxc_mins_maxs_s, 18, 9, 3, low(s)).
table(low_v_c_minc_maxc_s, 18, 10, 4, low(s)).
table(low_v_c_minc_mins_s, 18, 7, 4, low(s)).
table(low_v_c_minc_maxs_s, 18, 7, 4, low(s)).
table(low_v_c_maxc_mins_s, 18, 5, 4, low(s)).
table(low_v_c_maxc_maxs_s, 18, 5, 4, low(s)).
table(low_v_c_mins_maxs_s, 18, 5, 4, low(s)).
table(low_v_minc_maxc_mins_s, 18, 8, 4, low(s)).
table(low_v_minc_maxc_maxs_s, 18, 6, 4, low(s)).
table(low_v_minc_mins_maxs_s, 18, 5, 4, low(s)).
table(low_v_maxc_mins_maxs_s, 18, 5, 4, low(s)).
table(low_c_minc_maxc_mins_s, 18, 6, 4, low(s)).
table(low_c_minc_maxc_maxs_s, 18, 7, 4, low(s)).
table(low_c_minc_mins_maxs_s, 18, 5, 4, low(s)).
table(low_c_maxc_mins_maxs_s, 18, 6, 4, low(s)).
table(low_minc_maxc_mins_maxs_s, 18, 9, 4, low(s)).
table(low_v_c_minc_maxc_mins_s, 18, 8, 5, low(s)).
table(low_v_c_minc_maxc_maxs_s, 18, 8, 5, low(s)).
table(low_v_c_minc_mins_maxs_s, 18, 8, 5, low(s)).
table(low_v_c_maxc_mins_maxs_s, 18, 6, 5, low(s)).
table(low_v_minc_maxc_mins_maxs_s, 18, 7, 5, low(s)).
table(low_c_minc_maxc_mins_maxs_s, 18, 7, 5, low(s)).
table(low_v_c_minc_maxc_mins_maxs_s, 18, 9, 6, low(s)).
table(up_v_s, 18, 7, 1, up(s)).
table(up_v_c_s, 18, 8, 2, up(s)).
table(up_v_minc_s, 18, 8, 2, up(s)).
table(up_v_maxc_s, 18, 8, 2, up(s)).
table(up_v_mins_s, 18, 4, 2, up(s)).
table(up_v_maxs_s, 18, 7, 2, up(s)).
table(up_c_maxc_s, 18, 9, 2, up(s)).
table(up_v_c_minc_s, 18, 11, 3, up(s)).
table(up_v_c_maxc_s, 18, 9, 3, up(s)).
table(up_v_c_mins_s, 18, 5, 3, up(s)).
table(up_v_c_maxs_s, 18, 8, 3, up(s)).
table(up_v_minc_maxc_s, 18, 10, 3, up(s)).
table(up_v_minc_mins_s, 18, 5, 3, up(s)).
table(up_v_minc_maxs_s, 18, 8, 3, up(s)).
table(up_v_maxc_mins_s, 18, 5, 3, up(s)).
table(up_v_maxc_maxs_s, 18, 8, 3, up(s)).
table(up_v_mins_maxs_s, 18, 5, 3, up(s)).
table(up_c_minc_maxc_s, 18, 12, 3, up(s)).
table(up_c_maxc_mins_s, 18, 5, 3, up(s)).
table(up_c_maxc_maxs_s, 18, 9, 3, up(s)).
table(up_v_c_minc_maxc_s, 18, 12, 4, up(s)).
table(up_v_c_minc_mins_s, 18, 8, 4, up(s)).
table(up_v_c_minc_maxs_s, 18, 11, 4, up(s)).
table(up_v_c_maxc_mins_s, 18, 6, 4, up(s)).
table(up_v_c_maxc_maxs_s, 18, 9, 4, up(s)).
table(up_v_c_mins_maxs_s, 18, 6, 4, up(s)).
table(up_v_minc_maxc_mins_s, 18, 7, 4, up(s)).
table(up_v_minc_maxc_maxs_s, 18, 10, 4, up(s)).
table(up_v_minc_mins_maxs_s, 18, 6, 4, up(s)).
table(up_v_maxc_mins_maxs_s, 18, 6, 4, up(s)).
table(up_c_minc_maxc_mins_s, 18, 7, 4, up(s)).
table(up_c_minc_maxc_maxs_s, 18, 12, 4, up(s)).
table(up_c_maxc_mins_maxs_s, 18, 6, 4, up(s)).
table(up_v_c_minc_maxc_mins_s, 18, 9, 5, up(s)).
table(up_v_c_minc_maxc_maxs_s, 18, 12, 5, up(s)).
table(up_v_c_minc_mins_maxs_s, 18, 9, 5, up(s)).
table(up_v_c_maxc_mins_maxs_s, 18, 7, 5, up(s)).
table(up_v_minc_maxc_mins_maxs_s, 18, 8, 5, up(s)).
table(up_c_minc_maxc_mins_maxs_s, 18, 8, 5, up(s)).
table(up_v_c_minc_maxc_mins_maxs_s, 18, 10, 6, up(s)).
table(low_v_mins, 18, 2, 1, low(mins)).
table(low_c_mins, 18, 2, 1, low(mins)).
table(low_minc_mins, 18, 2, 1, low(mins)).
table(low_maxc_mins, 18, 2, 1, low(mins)).
table(low_s_mins, 18, 2, 1, low(mins)).
table(low_maxs_mins, 18, 2, 1, low(mins)).
table(low_v_c_mins, 18, 3, 2, low(mins)).
table(low_v_minc_mins, 18, 3, 2, low(mins)).
table(low_v_maxc_mins, 18, 3, 2, low(mins)).
table(low_v_s_mins, 18, 3, 2, low(mins)).
table(low_v_maxs_mins, 18, 3, 2, low(mins)).
table(low_c_minc_mins, 18, 3, 2, low(mins)).
table(low_c_maxc_mins, 18, 3, 2, low(mins)).
table(low_c_s_mins, 18, 3, 2, low(mins)).
table(low_c_maxs_mins, 18, 3, 2, low(mins)).
table(low_minc_maxc_mins, 18, 4, 2, low(mins)).
table(low_minc_s_mins, 18, 3, 2, low(mins)).
table(low_minc_maxs_mins, 18, 3, 2, low(mins)).
table(low_maxc_s_mins, 18, 3, 2, low(mins)).
table(low_maxc_maxs_mins, 18, 3, 2, low(mins)).
table(low_s_maxs_mins, 18, 3, 2, low(mins)).
table(low_v_c_minc_mins, 18, 6, 3, low(mins)).
table(low_v_c_maxc_mins, 18, 4, 3, low(mins)).
table(low_v_c_s_mins, 18, 4, 3, low(mins)).
table(low_v_c_maxs_mins, 18, 4, 3, low(mins)).
table(low_v_minc_maxc_mins, 18, 5, 3, low(mins)).
table(low_v_minc_s_mins, 18, 4, 3, low(mins)).
table(low_v_minc_maxs_mins, 18, 4, 3, low(mins)).
table(low_v_maxc_s_mins, 18, 4, 3, low(mins)).
table(low_v_maxc_maxs_mins, 18, 4, 3, low(mins)).
table(low_v_s_maxs_mins, 18, 4, 3, low(mins)).
table(low_c_minc_maxc_mins, 18, 5, 3, low(mins)).
table(low_c_minc_s_mins, 18, 4, 3, low(mins)).
table(low_c_minc_maxs_mins, 18, 4, 3, low(mins)).
table(low_c_maxc_s_mins, 18, 4, 3, low(mins)).
table(low_c_maxc_maxs_mins, 18, 4, 3, low(mins)).
table(low_c_s_maxs_mins, 18, 4, 3, low(mins)).
table(low_minc_maxc_s_mins, 18, 5, 3, low(mins)).
table(low_minc_maxc_maxs_mins, 18, 5, 3, low(mins)).
table(low_minc_s_maxs_mins, 18, 4, 3, low(mins)).
table(low_maxc_s_maxs_mins, 18, 4, 3, low(mins)).
table(low_v_c_minc_maxc_mins, 18, 7, 4, low(mins)).
table(low_v_c_minc_s_mins, 18, 7, 4, low(mins)).
table(low_v_c_minc_maxs_mins, 18, 7, 4, low(mins)).
table(low_v_c_maxc_s_mins, 18, 5, 4, low(mins)).
table(low_v_c_maxc_maxs_mins, 18, 5, 4, low(mins)).
table(low_v_c_s_maxs_mins, 18, 5, 4, low(mins)).
table(low_v_minc_maxc_s_mins, 18, 6, 4, low(mins)).
table(low_v_minc_maxc_maxs_mins, 18, 6, 4, low(mins)).
table(low_v_minc_s_maxs_mins, 18, 5, 4, low(mins)).
table(low_v_maxc_s_maxs_mins, 18, 5, 4, low(mins)).
table(low_c_minc_maxc_s_mins, 18, 6, 4, low(mins)).
table(low_c_minc_maxc_maxs_mins, 18, 6, 4, low(mins)).
table(low_c_minc_s_maxs_mins, 18, 5, 4, low(mins)).
table(low_c_maxc_s_maxs_mins, 18, 5, 4, low(mins)).
table(low_minc_maxc_s_maxs_mins, 18, 6, 4, low(mins)).
table(low_v_c_minc_maxc_s_mins, 18, 8, 5, low(mins)).
table(low_v_c_minc_maxc_maxs_mins, 18, 8, 5, low(mins)).
table(low_v_c_minc_s_maxs_mins, 18, 8, 5, low(mins)).
table(low_v_c_maxc_s_maxs_mins, 18, 6, 5, low(mins)).
table(low_v_minc_maxc_s_maxs_mins, 18, 7, 5, low(mins)).
table(low_c_minc_maxc_s_maxs_mins, 18, 7, 5, low(mins)).
table(low_v_c_minc_maxc_s_maxs_mins, 18, 9, 6, low(mins)).
table(up_v_mins, 18, 18, 1, up(mins)).
table(up_maxc_mins, 18, 9, 1, up(mins)).
table(up_maxs_mins, 18, 2, 1, up(mins)).
table(up_v_c_mins, 18, 6, 2, up(mins)).
table(up_v_minc_mins, 18, 3, 2, up(mins)).
table(up_v_maxc_mins, 18, 5, 2, up(mins)).
table(up_v_s_mins, 18, 3, 2, up(mins)).
table(up_v_maxs_mins, 18, 3, 2, up(mins)).
table(up_c_maxc_mins, 18, 5, 2, up(mins)).
table(up_c_maxs_mins, 18, 3, 2, up(mins)).
table(up_minc_maxc_mins, 18, 4, 2, up(mins)).
table(up_minc_maxs_mins, 18, 3, 2, up(mins)).
table(up_maxc_s_mins, 18, 3, 2, up(mins)).
table(up_maxc_maxs_mins, 18, 3, 2, up(mins)).
table(up_s_maxs_mins, 18, 3, 2, up(mins)).
table(up_v_c_minc_mins, 18, 6, 3, up(mins)).
table(up_v_c_maxc_mins, 18, 7, 3, up(mins)).
table(up_v_c_s_mins, 18, 4, 3, up(mins)).
table(up_v_c_maxs_mins, 18, 4, 3, up(mins)).
table(up_v_minc_maxc_mins, 18, 5, 3, up(mins)).
table(up_v_minc_s_mins, 18, 4, 3, up(mins)).
table(up_v_minc_maxs_mins, 18, 4, 3, up(mins)).
table(up_v_maxc_s_mins, 18, 4, 3, up(mins)).
table(up_v_maxc_maxs_mins, 18, 4, 3, up(mins)).
table(up_v_s_maxs_mins, 18, 4, 3, up(mins)).
table(up_c_minc_maxc_mins, 18, 5, 3, up(mins)).
table(up_c_minc_maxs_mins, 18, 4, 3, up(mins)).
table(up_c_maxc_s_mins, 18, 4, 3, up(mins)).
table(up_c_maxc_maxs_mins, 18, 4, 3, up(mins)).
table(up_c_s_maxs_mins, 18, 4, 3, up(mins)).
table(up_minc_maxc_s_mins, 18, 5, 3, up(mins)).
table(up_minc_maxc_maxs_mins, 18, 5, 3, up(mins)).
table(up_minc_s_maxs_mins, 18, 4, 3, up(mins)).
table(up_maxc_s_maxs_mins, 18, 4, 3, up(mins)).
table(up_v_c_minc_maxc_mins, 18, 7, 4, up(mins)).
table(up_v_c_minc_s_mins, 18, 7, 4, up(mins)).
table(up_v_c_minc_maxs_mins, 18, 7, 4, up(mins)).
table(up_v_c_maxc_s_mins, 18, 5, 4, up(mins)).
table(up_v_c_maxc_maxs_mins, 18, 5, 4, up(mins)).
table(up_v_c_s_maxs_mins, 18, 5, 4, up(mins)).
table(up_v_minc_maxc_s_mins, 18, 6, 4, up(mins)).
table(up_v_minc_maxc_maxs_mins, 18, 6, 4, up(mins)).
table(up_v_minc_s_maxs_mins, 18, 5, 4, up(mins)).
table(up_v_maxc_s_maxs_mins, 18, 5, 4, up(mins)).
table(up_c_minc_maxc_s_mins, 18, 6, 4, up(mins)).
table(up_c_minc_maxc_maxs_mins, 18, 6, 4, up(mins)).
table(up_c_minc_s_maxs_mins, 18, 5, 4, up(mins)).
table(up_c_maxc_s_maxs_mins, 18, 5, 4, up(mins)).
table(up_minc_maxc_s_maxs_mins, 18, 6, 4, up(mins)).
table(up_v_c_minc_maxc_s_mins, 18, 8, 5, up(mins)).
table(up_v_c_minc_maxc_maxs_mins, 18, 8, 5, up(mins)).
table(up_v_c_minc_s_maxs_mins, 18, 8, 5, up(mins)).
table(up_v_c_maxc_s_maxs_mins, 18, 6, 5, up(mins)).
table(up_v_minc_maxc_s_maxs_mins, 18, 7, 5, up(mins)).
table(up_c_minc_maxc_s_maxs_mins, 18, 7, 5, up(mins)).
table(up_v_c_minc_maxc_s_maxs_mins, 18, 9, 6, up(mins)).
table(low_v_maxs, 18, 7, 1, low(maxs)).
table(low_c_maxs, 18, 4, 1, low(maxs)).
table(low_minc_maxs, 18, 4, 1, low(maxs)).
table(low_maxc_maxs, 18, 4, 1, low(maxs)).
table(low_s_maxs, 18, 7, 1, low(maxs)).
table(low_mins_maxs, 18, 2, 1, low(maxs)).
table(low_v_c_maxs, 18, 8, 2, low(maxs)).
table(low_v_minc_maxs, 18, 8, 2, low(maxs)).
table(low_v_maxc_maxs, 18, 8, 2, low(maxs)).
table(low_v_s_maxs, 18, 3, 2, low(maxs)).
table(low_v_mins_maxs, 18, 3, 2, low(maxs)).
table(low_c_minc_maxs, 18, 5, 2, low(maxs)).
table(low_c_maxc_maxs, 18, 5, 2, low(maxs)).
table(low_c_s_maxs, 18, 8, 2, low(maxs)).
table(low_c_mins_maxs, 18, 3, 2, low(maxs)).
table(low_minc_maxc_maxs, 18, 6, 2, low(maxs)).
table(low_minc_s_maxs, 18, 3, 2, low(maxs)).
table(low_minc_mins_maxs, 18, 3, 2, low(maxs)).
table(low_maxc_s_maxs, 18, 3, 2, low(maxs)).
table(low_maxc_mins_maxs, 18, 3, 2, low(maxs)).
table(low_s_mins_maxs, 18, 7, 2, low(maxs)).
table(low_v_c_minc_maxs, 18, 11, 3, low(maxs)).
table(low_v_c_maxc_maxs, 18, 9, 3, low(maxs)).
table(low_v_c_s_maxs, 18, 4, 3, low(maxs)).
table(low_v_c_mins_maxs, 18, 4, 3, low(maxs)).
table(low_v_minc_maxc_maxs, 18, 10, 3, low(maxs)).
table(low_v_minc_s_maxs, 18, 4, 3, low(maxs)).
table(low_v_minc_mins_maxs, 18, 4, 3, low(maxs)).
table(low_v_maxc_s_maxs, 18, 4, 3, low(maxs)).
table(low_v_maxc_mins_maxs, 18, 4, 3, low(maxs)).
table(low_v_s_mins_maxs, 18, 4, 3, low(maxs)).
table(low_c_minc_maxc_maxs, 18, 7, 3, low(maxs)).
table(low_c_minc_s_maxs, 18, 4, 3, low(maxs)).
table(low_c_minc_mins_maxs, 18, 4, 3, low(maxs)).
table(low_c_maxc_s_maxs, 18, 4, 3, low(maxs)).
table(low_c_maxc_mins_maxs, 18, 4, 3, low(maxs)).
table(low_c_s_mins_maxs, 18, 8, 3, low(maxs)).
table(low_minc_maxc_s_maxs, 18, 5, 3, low(maxs)).
table(low_minc_maxc_mins_maxs, 18, 5, 3, low(maxs)).
table(low_minc_s_mins_maxs, 18, 4, 3, low(maxs)).
table(low_maxc_s_mins_maxs, 18, 4, 3, low(maxs)).
table(low_v_c_minc_maxc_maxs, 18, 12, 4, low(maxs)).
table(low_v_c_minc_s_maxs, 18, 7, 4, low(maxs)).
table(low_v_c_minc_mins_maxs, 18, 7, 4, low(maxs)).
table(low_v_c_maxc_s_maxs, 18, 5, 4, low(maxs)).
table(low_v_c_maxc_mins_maxs, 18, 5, 4, low(maxs)).
table(low_v_c_s_mins_maxs, 18, 5, 4, low(maxs)).
table(low_v_minc_maxc_s_maxs, 18, 6, 4, low(maxs)).
table(low_v_minc_maxc_mins_maxs, 18, 6, 4, low(maxs)).
table(low_v_minc_s_mins_maxs, 18, 5, 4, low(maxs)).
table(low_v_maxc_s_mins_maxs, 18, 5, 4, low(maxs)).
table(low_c_minc_maxc_s_maxs, 18, 6, 4, low(maxs)).
table(low_c_minc_maxc_mins_maxs, 18, 6, 4, low(maxs)).
table(low_c_minc_s_mins_maxs, 18, 5, 4, low(maxs)).
table(low_c_maxc_s_mins_maxs, 18, 5, 4, low(maxs)).
table(low_minc_maxc_s_mins_maxs, 18, 6, 4, low(maxs)).
table(low_v_c_minc_maxc_s_maxs, 18, 8, 5, low(maxs)).
table(low_v_c_minc_maxc_mins_maxs, 18, 8, 5, low(maxs)).
table(low_v_c_minc_s_mins_maxs, 18, 8, 5, low(maxs)).
table(low_v_c_maxc_s_mins_maxs, 18, 6, 5, low(maxs)).
table(low_v_minc_maxc_s_mins_maxs, 18, 7, 5, low(maxs)).
table(low_c_minc_maxc_s_mins_maxs, 18, 7, 5, low(maxs)).
table(low_v_c_minc_maxc_s_mins_maxs, 18, 9, 6, low(maxs)).
table(up_v_maxs, 18, 18, 1, up(maxs)).
table(up_maxc_maxs, 18, 2, 1, up(maxs)).
table(up_v_c_maxs, 18, 18, 2, up(maxs)).
table(up_v_minc_maxs, 18, 12, 2, up(maxs)).
table(up_v_maxc_maxs, 18, 3, 2, up(maxs)).
table(up_v_s_maxs, 18, 7, 2, up(maxs)).
table(up_v_mins_maxs, 18, 9, 2, up(maxs)).
table(up_c_maxc_maxs, 18, 3, 2, up(maxs)).
table(up_minc_maxc_maxs, 18, 4, 2, up(maxs)).
table(up_maxc_s_maxs, 18, 3, 2, up(maxs)).
table(up_maxc_mins_maxs, 18, 3, 2, up(maxs)).
table(up_v_c_minc_maxs, 18, 12, 3, up(maxs)).
table(up_v_c_maxc_maxs, 18, 4, 3, up(maxs)).
table(up_v_c_s_maxs, 18, 8, 3, up(maxs)).
table(up_v_c_mins_maxs, 18, 18, 3, up(maxs)).
table(up_v_minc_maxc_maxs, 18, 5, 3, up(maxs)).
table(up_v_minc_s_maxs, 18, 5, 3, up(maxs)).
table(up_v_minc_mins_maxs, 18, 14, 3, up(maxs)).
table(up_v_maxc_s_maxs, 18, 4, 3, up(maxs)).
table(up_v_maxc_mins_maxs, 18, 4, 3, up(maxs)).
table(up_v_s_mins_maxs, 18, 7, 3, up(maxs)).
table(up_c_minc_maxc_maxs, 18, 5, 3, up(maxs)).
table(up_c_maxc_s_maxs, 18, 4, 3, up(maxs)).
table(up_c_maxc_mins_maxs, 18, 4, 3, up(maxs)).
table(up_minc_maxc_s_maxs, 18, 5, 3, up(maxs)).
table(up_minc_maxc_mins_maxs, 18, 5, 3, up(maxs)).
table(up_maxc_s_mins_maxs, 18, 4, 3, up(maxs)).
table(up_v_c_minc_maxc_maxs, 18, 7, 4, up(maxs)).
table(up_v_c_minc_s_maxs, 18, 7, 4, up(maxs)).
table(up_v_c_minc_mins_maxs, 18, 13, 4, up(maxs)).
table(up_v_c_maxc_s_maxs, 18, 5, 4, up(maxs)).
table(up_v_c_maxc_mins_maxs, 18, 5, 4, up(maxs)).
table(up_v_c_s_mins_maxs, 18, 8, 4, up(maxs)).
table(up_v_minc_maxc_s_maxs, 18, 6, 4, up(maxs)).
table(up_v_minc_maxc_mins_maxs, 18, 6, 4, up(maxs)).
table(up_v_minc_s_mins_maxs, 18, 5, 4, up(maxs)).
table(up_v_maxc_s_mins_maxs, 18, 5, 4, up(maxs)).
table(up_c_minc_maxc_s_maxs, 18, 6, 4, up(maxs)).
table(up_c_minc_maxc_mins_maxs, 18, 6, 4, up(maxs)).
table(up_c_maxc_s_mins_maxs, 18, 5, 4, up(maxs)).
table(up_minc_maxc_s_mins_maxs, 18, 6, 4, up(maxs)).
table(up_v_c_minc_maxc_s_maxs, 18, 8, 5, up(maxs)).
table(up_v_c_minc_maxc_mins_maxs, 18, 8, 5, up(maxs)).
table(up_v_c_minc_s_mins_maxs, 18, 8, 5, up(maxs)).
table(up_v_c_maxc_s_mins_maxs, 18, 6, 5, up(maxs)).
table(up_v_minc_maxc_s_mins_maxs, 18, 7, 5, up(maxs)).
table(up_c_minc_maxc_s_mins_maxs, 18, 7, 5, up(maxs)).
table(up_v_c_minc_maxc_s_mins_maxs, 18, 9, 6, up(maxs)).
table(low_v_mina, 18, 12, 1, low(mina)).
table(low_c_mina, 18, 7, 1, low(mina)).
table(low_minc_mina, 18, 19, 1, low(mina)).
table(low_maxc_mina, 18, 19, 1, low(mina)).
table(low_s_mina, 18, 12, 1, low(mina)).
table(low_mins_mina, 18, 11, 1, low(mina)).
table(low_maxs_mina, 18, 11, 1, low(mina)).
table(low_v_c_mina, 18, 10, 2, low(mina)).
table(low_v_minc_mina, 18, 11, 2, low(mina)).
table(low_v_maxc_mina, 18, 11, 2, low(mina)).
table(low_v_s_mina, 18, 4, 2, low(mina)).
table(low_v_mins_mina, 18, 4, 2, low(mina)).
table(low_v_maxs_mina, 18, 10, 2, low(mina)).
table(low_c_minc_mina, 18, 8, 2, low(mina)).
table(low_c_maxc_mina, 18, 6, 2, low(mina)).
table(low_c_s_mina, 18, 10, 2, low(mina)).
table(low_c_mins_mina, 18, 7, 2, low(mina)).
table(low_c_maxs_mina, 18, 5, 2, low(mina)).
table(low_minc_maxc_mina, 18, 19, 2, low(mina)).
table(low_minc_s_mina, 18, 11, 2, low(mina)).
table(low_minc_mins_mina, 18, 19, 2, low(mina)).
table(low_minc_maxs_mina, 18, 19, 2, low(mina)).
table(low_maxc_s_mina, 18, 11, 2, low(mina)).
table(low_maxc_mins_mina, 18, 19, 2, low(mina)).
table(low_maxc_maxs_mina, 18, 19, 2, low(mina)).
table(low_s_mins_mina, 18, 12, 2, low(mina)).
table(low_s_maxs_mina, 18, 10, 2, low(mina)).
table(low_mins_maxs_mina, 18, 6, 2, low(mina)).
table(low_v_c_minc_mina, 18, 13, 3, low(mina)).
table(low_v_c_maxc_mina, 18, 11, 3, low(mina)).
table(low_v_c_s_mina, 18, 4, 3, low(mina)).
table(low_v_c_mins_mina, 18, 7, 3, low(mina)).
table(low_v_c_maxs_mina, 18, 10, 3, low(mina)).
table(low_v_minc_maxc_mina, 18, 12, 3, low(mina)).
table(low_v_minc_s_mina, 18, 4, 3, low(mina)).
table(low_v_minc_mins_mina, 18, 4, 3, low(mina)).
table(low_v_minc_maxs_mina, 18, 11, 3, low(mina)).
table(low_v_maxc_s_mina, 18, 4, 3, low(mina)).
table(low_v_maxc_mins_mina, 18, 6, 3, low(mina)).
table(low_v_maxc_maxs_mina, 18, 10, 3, low(mina)).
table(low_v_s_mins_mina, 18, 5, 3, low(mina)).
table(low_v_s_maxs_mina, 18, 5, 3, low(mina)).
table(low_v_mins_maxs_mina, 18, 5, 3, low(mina)).
table(low_c_minc_maxc_mina, 18, 8, 3, low(mina)).
table(low_c_minc_s_mina, 18, 11, 3, low(mina)).
table(low_c_minc_mins_mina, 18, 8, 3, low(mina)).
table(low_c_minc_maxs_mina, 18, 6, 3, low(mina)).
table(low_c_maxc_s_mina, 18, 8, 3, low(mina)).
table(low_c_maxc_mins_mina, 18, 6, 3, low(mina)).
table(low_c_maxc_maxs_mina, 18, 6, 3, low(mina)).
table(low_c_s_mins_mina, 18, 10, 3, low(mina)).
table(low_c_s_maxs_mina, 18, 10, 3, low(mina)).
table(low_c_mins_maxs_mina, 18, 5, 3, low(mina)).
table(low_minc_maxc_s_mina, 18, 10, 3, low(mina)).
table(low_minc_maxc_mins_mina, 18, 16, 3, low(mina)).
table(low_minc_maxc_maxs_mina, 18, 18, 3, low(mina)).
table(low_minc_s_mins_mina, 18, 6, 3, low(mina)).
table(low_minc_s_maxs_mina, 18, 8, 3, low(mina)).
table(low_minc_mins_maxs_mina, 18, 17, 3, low(mina)).
table(low_maxc_s_mins_mina, 18, 8, 3, low(mina)).
table(low_maxc_s_maxs_mina, 18, 7, 3, low(mina)).
table(low_maxc_mins_maxs_mina, 18, 8, 3, low(mina)).
table(low_s_mins_maxs_mina, 18, 10, 3, low(mina)).
table(low_v_c_minc_maxc_mina, 18, 14, 4, low(mina)).
table(low_v_c_minc_s_mina, 18, 8, 4, low(mina)).
table(low_v_c_minc_mins_mina, 18, 10, 4, low(mina)).
table(low_v_c_minc_maxs_mina, 18, 13, 4, low(mina)).
table(low_v_c_maxc_s_mina, 18, 5, 4, low(mina)).
table(low_v_c_maxc_mins_mina, 18, 8, 4, low(mina)).
table(low_v_c_maxc_maxs_mina, 18, 11, 4, low(mina)).
table(low_v_c_s_mins_mina, 18, 5, 4, low(mina)).
table(low_v_c_s_maxs_mina, 18, 5, 4, low(mina)).
table(low_v_c_mins_maxs_mina, 18, 8, 4, low(mina)).
table(low_v_minc_maxc_s_mina, 18, 6, 4, low(mina)).
table(low_v_minc_maxc_mins_mina, 18, 6, 4, low(mina)).
table(low_v_minc_maxc_maxs_mina, 18, 12, 4, low(mina)).
table(low_v_minc_s_mins_mina, 18, 5, 4, low(mina)).
table(low_v_minc_s_maxs_mina, 18, 5, 4, low(mina)).
table(low_v_minc_mins_maxs_mina, 18, 5, 4, low(mina)).
table(low_v_maxc_s_mins_mina, 18, 5, 4, low(mina)).
table(low_v_maxc_s_maxs_mina, 18, 5, 4, low(mina)).
table(low_v_maxc_mins_maxs_mina, 18, 5, 4, low(mina)).
table(low_v_s_mins_maxs_mina, 18, 6, 4, low(mina)).
table(low_c_minc_maxc_s_mina, 18, 9, 4, low(mina)).
table(low_c_minc_maxc_mins_mina, 18, 7, 4, low(mina)).
table(low_c_minc_maxc_maxs_mina, 18, 8, 4, low(mina)).
table(low_c_minc_s_mins_mina, 18, 10, 4, low(mina)).
table(low_c_minc_s_maxs_mina, 18, 10, 4, low(mina)).
table(low_c_minc_mins_maxs_mina, 18, 6, 4, low(mina)).
table(low_c_maxc_s_mins_mina, 18, 6, 4, low(mina)).
table(low_c_maxc_s_maxs_mina, 18, 6, 4, low(mina)).
table(low_c_maxc_mins_maxs_mina, 18, 6, 4, low(mina)).
table(low_c_s_mins_maxs_mina, 18, 10, 4, low(mina)).
table(low_minc_maxc_s_mins_mina, 18, 7, 4, low(mina)).
table(low_minc_maxc_s_maxs_mina, 18, 9, 4, low(mina)).
table(low_minc_maxc_mins_maxs_mina, 18, 16, 4, low(mina)).
table(low_minc_s_mins_maxs_mina, 18, 6, 4, low(mina)).
table(low_maxc_s_mins_maxs_mina, 18, 6, 4, low(mina)).
table(low_v_c_minc_maxc_s_mina, 18, 9, 5, low(mina)).
table(low_v_c_minc_maxc_mins_mina, 18, 11, 5, low(mina)).
table(low_v_c_minc_maxc_maxs_mina, 18, 14, 5, low(mina)).
table(low_v_c_minc_s_mins_mina, 18, 8, 5, low(mina)).
table(low_v_c_minc_s_maxs_mina, 18, 8, 5, low(mina)).
table(low_v_c_minc_mins_maxs_mina, 18, 11, 5, low(mina)).
table(low_v_c_maxc_s_mins_mina, 18, 6, 5, low(mina)).
table(low_v_c_maxc_s_maxs_mina, 18, 6, 5, low(mina)).
table(low_v_c_maxc_mins_maxs_mina, 18, 9, 5, low(mina)).
table(low_v_c_s_mins_maxs_mina, 18, 6, 5, low(mina)).
table(low_v_minc_maxc_s_mins_mina, 18, 7, 5, low(mina)).
table(low_v_minc_maxc_s_maxs_mina, 18, 7, 5, low(mina)).
table(low_v_minc_maxc_mins_maxs_mina, 18, 7, 5, low(mina)).
table(low_v_minc_s_mins_maxs_mina, 18, 6, 5, low(mina)).
table(low_v_maxc_s_mins_maxs_mina, 18, 6, 5, low(mina)).
table(low_c_minc_maxc_s_mins_mina, 18, 8, 5, low(mina)).
table(low_c_minc_maxc_s_maxs_mina, 18, 8, 5, low(mina)).
table(low_c_minc_maxc_mins_maxs_mina, 18, 8, 5, low(mina)).
table(low_c_minc_s_mins_maxs_mina, 18, 11, 5, low(mina)).
table(low_c_maxc_s_mins_maxs_mina, 18, 7, 5, low(mina)).
table(low_minc_maxc_s_mins_maxs_mina, 18, 8, 5, low(mina)).
table(low_v_c_minc_maxc_s_mins_mina, 18, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_maxs_mina, 18, 9, 6, low(mina)).
table(low_v_c_minc_maxc_mins_maxs_mina, 18, 12, 6, low(mina)).
table(low_v_c_minc_s_mins_maxs_mina, 18, 9, 6, low(mina)).
table(low_v_c_maxc_s_mins_maxs_mina, 18, 7, 6, low(mina)).
table(low_v_minc_maxc_s_mins_maxs_mina, 18, 8, 6, low(mina)).
table(low_c_minc_maxc_s_mins_maxs_mina, 18, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_mins_maxs_mina, 18, 10, 7, low(mina)).
table(up_v_maxa, 18, 19, 1, up(maxa)).
table(up_v_c_maxa, 18, 19, 2, up(maxa)).
table(up_v_minc_maxa, 18, 19, 2, up(maxa)).
table(up_v_maxc_maxa, 18, 19, 2, up(maxa)).
table(up_v_s_maxa, 18, 19, 2, up(maxa)).
table(up_v_mins_maxa, 18, 19, 2, up(maxa)).
table(up_v_maxs_maxa, 18, 19, 2, up(maxa)).
table(up_c_maxc_maxa, 18, 19, 2, up(maxa)).
table(up_maxc_s_maxa, 18, 18, 2, up(maxa)).
table(up_s_maxs_maxa, 18, 19, 2, up(maxa)).
table(up_v_c_minc_maxa, 18, 19, 3, up(maxa)).
table(up_v_c_maxc_maxa, 18, 19, 3, up(maxa)).
table(up_v_c_s_maxa, 18, 19, 3, up(maxa)).
table(up_v_c_mins_maxa, 18, 19, 3, up(maxa)).
table(up_v_c_maxs_maxa, 18, 7, 3, up(maxa)).
table(up_v_minc_maxc_maxa, 18, 19, 3, up(maxa)).
table(up_v_minc_s_maxa, 18, 18, 3, up(maxa)).
table(up_v_minc_mins_maxa, 18, 19, 3, up(maxa)).
table(up_v_minc_maxs_maxa, 18, 19, 3, up(maxa)).
table(up_v_maxc_s_maxa, 18, 18, 3, up(maxa)).
table(up_v_maxc_mins_maxa, 18, 11, 3, up(maxa)).
table(up_v_maxc_maxs_maxa, 18, 6, 3, up(maxa)).
table(up_v_s_mins_maxa, 18, 19, 3, up(maxa)).
table(up_v_s_maxs_maxa, 18, 19, 3, up(maxa)).
table(up_v_mins_maxs_maxa, 18, 19, 3, up(maxa)).
table(up_c_minc_maxc_maxa, 18, 19, 3, up(maxa)).
table(up_c_maxc_s_maxa, 18, 18, 3, up(maxa)).
table(up_c_maxc_mins_maxa, 18, 19, 3, up(maxa)).
table(up_c_maxc_maxs_maxa, 18, 6, 3, up(maxa)).
table(up_c_s_maxs_maxa, 18, 19, 3, up(maxa)).
table(up_minc_maxc_s_maxa, 18, 5, 3, up(maxa)).
table(up_minc_s_maxs_maxa, 18, 18, 3, up(maxa)).
table(up_maxc_s_mins_maxa, 18, 12, 3, up(maxa)).
table(up_maxc_s_maxs_maxa, 18, 6, 3, up(maxa)).
table(up_s_mins_maxs_maxa, 18, 19, 3, up(maxa)).
table(up_v_c_minc_maxc_maxa, 18, 19, 4, up(maxa)).
table(up_v_c_minc_s_maxa, 18, 18, 4, up(maxa)).
table(up_v_c_minc_mins_maxa, 18, 11, 4, up(maxa)).
table(up_v_c_minc_maxs_maxa, 18, 8, 4, up(maxa)).
table(up_v_c_maxc_s_maxa, 18, 18, 4, up(maxa)).
table(up_v_c_maxc_mins_maxa, 18, 11, 4, up(maxa)).
table(up_v_c_maxc_maxs_maxa, 18, 6, 4, up(maxa)).
table(up_v_c_s_mins_maxa, 18, 19, 4, up(maxa)).
table(up_v_c_s_maxs_maxa, 18, 19, 4, up(maxa)).
table(up_v_c_mins_maxs_maxa, 18, 9, 4, up(maxa)).
table(up_v_minc_maxc_s_maxa, 18, 18, 4, up(maxa)).
table(up_v_minc_maxc_mins_maxa, 18, 14, 4, up(maxa)).
table(up_v_minc_maxc_maxs_maxa, 18, 9, 4, up(maxa)).
table(up_v_minc_s_mins_maxa, 18, 18, 4, up(maxa)).
table(up_v_minc_s_maxs_maxa, 18, 18, 4, up(maxa)).
table(up_v_minc_mins_maxs_maxa, 18, 18, 4, up(maxa)).
table(up_v_maxc_s_mins_maxa, 18, 8, 4, up(maxa)).
table(up_v_maxc_s_maxs_maxa, 18, 8, 4, up(maxa)).
table(up_v_maxc_mins_maxs_maxa, 18, 6, 4, up(maxa)).
table(up_v_s_mins_maxs_maxa, 18, 19, 4, up(maxa)).
table(up_c_minc_maxc_s_maxa, 18, 18, 4, up(maxa)).
table(up_c_minc_maxc_mins_maxa, 18, 14, 4, up(maxa)).
table(up_c_minc_maxc_maxs_maxa, 18, 9, 4, up(maxa)).
table(up_c_minc_s_maxs_maxa, 18, 9, 4, up(maxa)).
table(up_c_maxc_s_mins_maxa, 18, 12, 4, up(maxa)).
table(up_c_maxc_s_maxs_maxa, 18, 7, 4, up(maxa)).
table(up_c_maxc_mins_maxs_maxa, 18, 5, 4, up(maxa)).
table(up_c_s_mins_maxs_maxa, 18, 19, 4, up(maxa)).
table(up_minc_maxc_s_mins_maxa, 18, 7, 4, up(maxa)).
table(up_minc_maxc_s_maxs_maxa, 18, 6, 4, up(maxa)).
table(up_minc_s_mins_maxs_maxa, 18, 18, 4, up(maxa)).
table(up_maxc_s_mins_maxs_maxa, 18, 6, 4, up(maxa)).
table(up_v_c_minc_maxc_s_maxa, 18, 18, 5, up(maxa)).
table(up_v_c_minc_maxc_mins_maxa, 18, 14, 5, up(maxa)).
table(up_v_c_minc_maxc_maxs_maxa, 18, 9, 5, up(maxa)).
table(up_v_c_minc_s_mins_maxa, 18, 10, 5, up(maxa)).
table(up_v_c_minc_s_maxs_maxa, 18, 9, 5, up(maxa)).
table(up_v_c_minc_mins_maxs_maxa, 18, 8, 5, up(maxa)).
table(up_v_c_maxc_s_mins_maxa, 18, 8, 5, up(maxa)).
table(up_v_c_maxc_s_maxs_maxa, 18, 6, 5, up(maxa)).
table(up_v_c_maxc_mins_maxs_maxa, 18, 6, 5, up(maxa)).
table(up_v_c_s_mins_maxs_maxa, 18, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_mins_maxa, 18, 11, 5, up(maxa)).
table(up_v_minc_maxc_s_maxs_maxa, 18, 10, 5, up(maxa)).
table(up_v_minc_maxc_mins_maxs_maxa, 18, 9, 5, up(maxa)).
table(up_v_minc_s_mins_maxs_maxa, 18, 18, 5, up(maxa)).
table(up_v_maxc_s_mins_maxs_maxa, 18, 8, 5, up(maxa)).
table(up_c_minc_maxc_s_mins_maxa, 18, 11, 5, up(maxa)).
table(up_c_minc_maxc_s_maxs_maxa, 18, 9, 5, up(maxa)).
table(up_c_minc_maxc_mins_maxs_maxa, 18, 9, 5, up(maxa)).
table(up_c_minc_s_mins_maxs_maxa, 18, 10, 5, up(maxa)).
table(up_c_maxc_s_mins_maxs_maxa, 18, 7, 5, up(maxa)).
table(up_minc_maxc_s_mins_maxs_maxa, 18, 7, 5, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxa, 18, 11, 6, up(maxa)).
table(up_v_c_minc_maxc_s_maxs_maxa, 18, 9, 6, up(maxa)).
table(up_v_c_minc_maxc_mins_maxs_maxa, 18, 9, 6, up(maxa)).
table(up_v_c_minc_s_mins_maxs_maxa, 18, 10, 6, up(maxa)).
table(up_v_c_maxc_s_mins_maxs_maxa, 18, 7, 6, up(maxa)).
table(up_v_minc_maxc_s_mins_maxs_maxa, 18, 11, 6, up(maxa)).
table(up_c_minc_maxc_s_mins_maxs_maxa, 18, 10, 6, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxs_maxa, 18, 10, 7, up(maxa)).
table(low_c_v, 19, 18, 1, low(v)).
table(low_minc_v, 19, 11, 1, low(v)).
table(low_maxc_v, 19, 11, 1, low(v)).
table(low_s_v, 19, 7, 1, low(v)).
table(low_mins_v, 19, 18, 1, low(v)).
table(low_maxs_v, 19, 18, 1, low(v)).
table(low_c_minc_v, 19, 11, 2, low(v)).
table(low_c_maxc_v, 19, 11, 2, low(v)).
table(low_c_s_v, 19, 8, 2, low(v)).
table(low_c_mins_v, 19, 18, 2, low(v)).
table(low_c_maxs_v, 19, 18, 2, low(v)).
table(low_minc_maxc_v, 19, 11, 2, low(v)).
table(low_minc_s_v, 19, 3, 2, low(v)).
table(low_minc_mins_v, 19, 12, 2, low(v)).
table(low_minc_maxs_v, 19, 12, 2, low(v)).
table(low_maxc_s_v, 19, 3, 2, low(v)).
table(low_maxc_mins_v, 19, 12, 2, low(v)).
table(low_maxc_maxs_v, 19, 13, 2, low(v)).
table(low_s_mins_v, 19, 7, 2, low(v)).
table(low_s_maxs_v, 19, 7, 2, low(v)).
table(low_mins_maxs_v, 19, 12, 2, low(v)).
table(low_c_minc_maxc_v, 19, 11, 3, low(v)).
table(low_c_minc_s_v, 19, 6, 3, low(v)).
table(low_c_minc_mins_v, 19, 12, 3, low(v)).
table(low_c_minc_maxs_v, 19, 12, 3, low(v)).
table(low_c_maxc_s_v, 19, 4, 3, low(v)).
table(low_c_maxc_mins_v, 19, 12, 3, low(v)).
table(low_c_maxc_maxs_v, 19, 13, 3, low(v)).
table(low_c_s_mins_v, 19, 8, 3, low(v)).
table(low_c_s_maxs_v, 19, 8, 3, low(v)).
table(low_c_mins_maxs_v, 19, 18, 3, low(v)).
table(low_minc_maxc_s_v, 19, 5, 3, low(v)).
table(low_minc_maxc_mins_v, 19, 12, 3, low(v)).
table(low_minc_maxc_maxs_v, 19, 12, 3, low(v)).
table(low_minc_s_mins_v, 19, 4, 3, low(v)).
table(low_minc_s_maxs_v, 19, 4, 3, low(v)).
table(low_minc_mins_maxs_v, 19, 13, 3, low(v)).
table(low_maxc_s_mins_v, 19, 4, 3, low(v)).
table(low_maxc_s_maxs_v, 19, 4, 3, low(v)).
table(low_maxc_mins_maxs_v, 19, 14, 3, low(v)).
table(low_s_mins_maxs_v, 19, 7, 3, low(v)).
table(low_c_minc_maxc_s_v, 19, 7, 4, low(v)).
table(low_c_minc_maxc_mins_v, 19, 12, 4, low(v)).
table(low_c_minc_maxc_maxs_v, 19, 12, 4, low(v)).
table(low_c_minc_s_mins_v, 19, 7, 4, low(v)).
table(low_c_minc_s_maxs_v, 19, 7, 4, low(v)).
table(low_c_minc_mins_maxs_v, 19, 13, 4, low(v)).
table(low_c_maxc_s_mins_v, 19, 5, 4, low(v)).
table(low_c_maxc_s_maxs_v, 19, 5, 4, low(v)).
table(low_c_maxc_mins_maxs_v, 19, 14, 4, low(v)).
table(low_c_s_mins_maxs_v, 19, 8, 4, low(v)).
table(low_minc_maxc_s_mins_v, 19, 6, 4, low(v)).
table(low_minc_maxc_s_maxs_v, 19, 6, 4, low(v)).
table(low_minc_maxc_mins_maxs_v, 19, 13, 4, low(v)).
table(low_minc_s_mins_maxs_v, 19, 5, 4, low(v)).
table(low_maxc_s_mins_maxs_v, 19, 5, 4, low(v)).
table(low_c_minc_maxc_s_mins_v, 19, 8, 5, low(v)).
table(low_c_minc_maxc_s_maxs_v, 19, 8, 5, low(v)).
table(low_c_minc_maxc_mins_maxs_v, 19, 13, 5, low(v)).
table(low_c_minc_s_mins_maxs_v, 19, 8, 5, low(v)).
table(low_c_maxc_s_mins_maxs_v, 19, 6, 5, low(v)).
table(low_minc_maxc_s_mins_maxs_v, 19, 7, 5, low(v)).
table(low_c_minc_maxc_s_mins_maxs_v, 19, 9, 6, low(v)).
table(up_c_maxc_v, 19, 3, 2, up(v)).
table(up_maxc_s_v, 19, 3, 2, up(v)).
table(up_s_maxs_v, 19, 3, 2, up(v)).
table(up_c_minc_maxc_v, 19, 6, 3, up(v)).
table(up_c_maxc_s_v, 19, 4, 3, up(v)).
table(up_c_maxc_mins_v, 19, 4, 3, up(v)).
table(up_c_maxc_maxs_v, 19, 4, 3, up(v)).
table(up_c_s_maxs_v, 19, 4, 3, up(v)).
table(up_minc_maxc_s_v, 19, 5, 3, up(v)).
table(up_minc_s_maxs_v, 19, 4, 3, up(v)).
table(up_maxc_s_mins_v, 19, 4, 3, up(v)).
table(up_maxc_s_maxs_v, 19, 4, 3, up(v)).
table(up_s_mins_maxs_v, 19, 4, 3, up(v)).
table(up_c_minc_maxc_s_v, 19, 7, 4, up(v)).
table(up_c_minc_maxc_mins_v, 19, 7, 4, up(v)).
table(up_c_minc_maxc_maxs_v, 19, 7, 4, up(v)).
table(up_c_minc_s_maxs_v, 19, 7, 4, up(v)).
table(up_c_maxc_s_mins_v, 19, 5, 4, up(v)).
table(up_c_maxc_s_maxs_v, 19, 5, 4, up(v)).
table(up_c_maxc_mins_maxs_v, 19, 5, 4, up(v)).
table(up_c_s_mins_maxs_v, 19, 5, 4, up(v)).
table(up_minc_maxc_s_mins_v, 19, 6, 4, up(v)).
table(up_minc_maxc_s_maxs_v, 19, 6, 4, up(v)).
table(up_minc_s_mins_maxs_v, 19, 5, 4, up(v)).
table(up_maxc_s_mins_maxs_v, 19, 5, 4, up(v)).
table(up_c_minc_maxc_s_mins_v, 19, 8, 5, up(v)).
table(up_c_minc_maxc_s_maxs_v, 19, 8, 5, up(v)).
table(up_c_minc_maxc_mins_maxs_v, 19, 8, 5, up(v)).
table(up_c_minc_s_mins_maxs_v, 19, 8, 5, up(v)).
table(up_c_maxc_s_mins_maxs_v, 19, 6, 5, up(v)).
table(up_minc_maxc_s_mins_maxs_v, 19, 7, 5, up(v)).
table(up_c_minc_maxc_s_mins_maxs_v, 19, 9, 6, up(v)).
table(low_v_c, 19, 11, 1, low(c)).
table(low_minc_c, 19, 11, 1, low(c)).
table(low_maxc_c, 19, 11, 1, low(c)).
table(low_s_c, 19, 8, 1, low(c)).
table(low_mins_c, 19, 8, 1, low(c)).
table(low_maxs_c, 19, 8, 1, low(c)).
table(low_v_minc_c, 19, 11, 2, low(c)).
table(low_v_maxc_c, 19, 3, 2, low(c)).
table(low_v_s_c, 19, 13, 2, low(c)).
table(low_v_mins_c, 19, 12, 2, low(c)).
table(low_v_maxs_c, 19, 13, 2, low(c)).
table(low_minc_maxc_c, 19, 11, 2, low(c)).
table(low_minc_s_c, 19, 5, 2, low(c)).
table(low_minc_mins_c, 19, 9, 2, low(c)).
table(low_minc_maxs_c, 19, 9, 2, low(c)).
table(low_maxc_s_c, 19, 3, 2, low(c)).
table(low_maxc_mins_c, 19, 12, 2, low(c)).
table(low_maxc_maxs_c, 19, 13, 2, low(c)).
table(low_s_mins_c, 19, 10, 2, low(c)).
table(low_s_maxs_c, 19, 11, 2, low(c)).
table(low_mins_maxs_c, 19, 9, 2, low(c)).
table(low_v_minc_maxc_c, 19, 6, 3, low(c)).
table(low_v_minc_s_c, 19, 12, 3, low(c)).
table(low_v_minc_mins_c, 19, 12, 3, low(c)).
table(low_v_minc_maxs_c, 19, 12, 3, low(c)).
table(low_v_maxc_s_c, 19, 4, 3, low(c)).
table(low_v_maxc_mins_c, 19, 4, 3, low(c)).
table(low_v_maxc_maxs_c, 19, 4, 3, low(c)).
table(low_v_s_mins_c, 19, 14, 3, low(c)).
table(low_v_s_maxs_c, 19, 14, 3, low(c)).
table(low_v_mins_maxs_c, 19, 14, 3, low(c)).
table(low_minc_maxc_s_c, 19, 5, 3, low(c)).
table(low_minc_maxc_mins_c, 19, 12, 3, low(c)).
table(low_minc_maxc_maxs_c, 19, 12, 3, low(c)).
table(low_minc_s_mins_c, 19, 6, 3, low(c)).
table(low_minc_s_maxs_c, 19, 7, 3, low(c)).
table(low_minc_mins_maxs_c, 19, 7, 3, low(c)).
table(low_maxc_s_mins_c, 19, 4, 3, low(c)).
table(low_maxc_s_maxs_c, 19, 4, 3, low(c)).
table(low_maxc_mins_maxs_c, 19, 6, 3, low(c)).
table(low_s_mins_maxs_c, 19, 12, 3, low(c)).
table(low_v_minc_maxc_s_c, 19, 7, 4, low(c)).
table(low_v_minc_maxc_mins_c, 19, 7, 4, low(c)).
table(low_v_minc_maxc_maxs_c, 19, 7, 4, low(c)).
table(low_v_minc_s_mins_c, 19, 13, 4, low(c)).
table(low_v_minc_s_maxs_c, 19, 13, 4, low(c)).
table(low_v_minc_mins_maxs_c, 19, 13, 4, low(c)).
table(low_v_maxc_s_mins_c, 19, 5, 4, low(c)).
table(low_v_maxc_s_maxs_c, 19, 5, 4, low(c)).
table(low_v_maxc_mins_maxs_c, 19, 5, 4, low(c)).
table(low_v_s_mins_maxs_c, 19, 15, 4, low(c)).
table(low_minc_maxc_s_mins_c, 19, 6, 4, low(c)).
table(low_minc_maxc_s_maxs_c, 19, 6, 4, low(c)).
table(low_minc_maxc_mins_maxs_c, 19, 8, 4, low(c)).
table(low_minc_s_mins_maxs_c, 19, 8, 4, low(c)).
table(low_maxc_s_mins_maxs_c, 19, 5, 4, low(c)).
table(low_v_minc_maxc_s_mins_c, 19, 8, 5, low(c)).
table(low_v_minc_maxc_s_maxs_c, 19, 8, 5, low(c)).
table(low_v_minc_maxc_mins_maxs_c, 19, 8, 5, low(c)).
table(low_v_minc_s_mins_maxs_c, 19, 14, 5, low(c)).
table(low_v_maxc_s_mins_maxs_c, 19, 6, 5, low(c)).
table(low_minc_maxc_s_mins_maxs_c, 19, 7, 5, low(c)).
table(low_v_minc_maxc_s_mins_maxs_c, 19, 9, 6, low(c)).
table(up_v_c, 19, 18, 1, up(c)).
table(up_s_c, 19, 3, 1, up(c)).
table(up_v_minc_c, 19, 6, 2, up(c)).
table(up_v_maxc_c, 19, 11, 2, up(c)).
table(up_v_s_c, 19, 4, 2, up(c)).
table(up_v_mins_c, 19, 10, 2, up(c)).
table(up_v_maxs_c, 19, 18, 2, up(c)).
table(up_minc_s_c, 19, 3, 2, up(c)).
table(up_maxc_s_c, 19, 4, 2, up(c)).
table(up_s_mins_c, 19, 5, 2, up(c)).
table(up_s_maxs_c, 19, 5, 2, up(c)).
table(up_v_minc_maxc_c, 19, 7, 3, up(c)).
table(up_v_minc_s_c, 19, 6, 3, up(c)).
table(up_v_minc_mins_c, 19, 7, 3, up(c)).
table(up_v_minc_maxs_c, 19, 7, 3, up(c)).
table(up_v_maxc_s_c, 19, 5, 3, up(c)).
table(up_v_maxc_mins_c, 19, 5, 3, up(c)).
table(up_v_maxc_maxs_c, 19, 13, 3, up(c)).
table(up_v_s_mins_c, 19, 8, 3, up(c)).
table(up_v_s_maxs_c, 19, 6, 3, up(c)).
table(up_v_mins_maxs_c, 19, 12, 3, up(c)).
table(up_minc_maxc_s_c, 19, 5, 3, up(c)).
table(up_minc_s_mins_c, 19, 4, 3, up(c)).
table(up_minc_s_maxs_c, 19, 4, 3, up(c)).
table(up_maxc_s_mins_c, 19, 5, 3, up(c)).
table(up_maxc_s_maxs_c, 19, 5, 3, up(c)).
table(up_s_mins_maxs_c, 19, 8, 3, up(c)).
table(up_v_minc_maxc_s_c, 19, 7, 4, up(c)).
table(up_v_minc_maxc_mins_c, 19, 8, 4, up(c)).
table(up_v_minc_maxc_maxs_c, 19, 8, 4, up(c)).
table(up_v_minc_s_mins_c, 19, 7, 4, up(c)).
table(up_v_minc_s_maxs_c, 19, 7, 4, up(c)).
table(up_v_minc_mins_maxs_c, 19, 8, 4, up(c)).
table(up_v_maxc_s_mins_c, 19, 6, 4, up(c)).
table(up_v_maxc_s_maxs_c, 19, 6, 4, up(c)).
table(up_v_maxc_mins_maxs_c, 19, 7, 4, up(c)).
table(up_v_s_mins_maxs_c, 19, 10, 4, up(c)).
table(up_minc_maxc_s_mins_c, 19, 6, 4, up(c)).
table(up_minc_maxc_s_maxs_c, 19, 6, 4, up(c)).
table(up_minc_s_mins_maxs_c, 19, 5, 4, up(c)).
table(up_maxc_s_mins_maxs_c, 19, 6, 4, up(c)).
table(up_v_minc_maxc_s_mins_c, 19, 8, 5, up(c)).
table(up_v_minc_maxc_s_maxs_c, 19, 8, 5, up(c)).
table(up_v_minc_maxc_mins_maxs_c, 19, 9, 5, up(c)).
table(up_v_minc_s_mins_maxs_c, 19, 8, 5, up(c)).
table(up_v_maxc_s_mins_maxs_c, 19, 7, 5, up(c)).
table(up_minc_maxc_s_mins_maxs_c, 19, 7, 5, up(c)).
table(up_v_minc_maxc_s_mins_maxs_c, 19, 9, 6, up(c)).
table(low_v_minc, 19, 3, 1, low(minc)).
table(low_c_minc, 19, 3, 1, low(minc)).
table(low_maxc_minc, 19, 3, 1, low(minc)).
table(low_s_minc, 19, 3, 1, low(minc)).
table(low_mins_minc, 19, 2, 1, low(minc)).
table(low_maxs_minc, 19, 3, 1, low(minc)).
table(low_v_c_minc, 19, 5, 2, low(minc)).
table(low_v_maxc_minc, 19, 4, 2, low(minc)).
table(low_v_s_minc, 19, 4, 2, low(minc)).
table(low_v_mins_minc, 19, 3, 2, low(minc)).
table(low_v_maxs_minc, 19, 4, 2, low(minc)).
table(low_c_maxc_minc, 19, 4, 2, low(minc)).
table(low_c_s_minc, 19, 4, 2, low(minc)).
table(low_c_mins_minc, 19, 3, 2, low(minc)).
table(low_c_maxs_minc, 19, 4, 2, low(minc)).
table(low_maxc_s_minc, 19, 4, 2, low(minc)).
table(low_maxc_mins_minc, 19, 4, 2, low(minc)).
table(low_maxc_maxs_minc, 19, 4, 2, low(minc)).
table(low_s_mins_minc, 19, 3, 2, low(minc)).
table(low_s_maxs_minc, 19, 4, 2, low(minc)).
table(low_mins_maxs_minc, 19, 3, 2, low(minc)).
table(low_v_c_maxc_minc, 19, 6, 3, low(minc)).
table(low_v_c_s_minc, 19, 6, 3, low(minc)).
table(low_v_c_mins_minc, 19, 6, 3, low(minc)).
table(low_v_c_maxs_minc, 19, 6, 3, low(minc)).
table(low_v_maxc_s_minc, 19, 5, 3, low(minc)).
table(low_v_maxc_mins_minc, 19, 5, 3, low(minc)).
table(low_v_maxc_maxs_minc, 19, 5, 3, low(minc)).
table(low_v_s_mins_minc, 19, 4, 3, low(minc)).
table(low_v_s_maxs_minc, 19, 5, 3, low(minc)).
table(low_v_mins_maxs_minc, 19, 4, 3, low(minc)).
table(low_c_maxc_s_minc, 19, 5, 3, low(minc)).
table(low_c_maxc_mins_minc, 19, 5, 3, low(minc)).
table(low_c_maxc_maxs_minc, 19, 5, 3, low(minc)).
table(low_c_s_mins_minc, 19, 4, 3, low(minc)).
table(low_c_s_maxs_minc, 19, 6, 3, low(minc)).
table(low_c_mins_maxs_minc, 19, 4, 3, low(minc)).
table(low_maxc_s_mins_minc, 19, 5, 3, low(minc)).
table(low_maxc_s_maxs_minc, 19, 5, 3, low(minc)).
table(low_maxc_mins_maxs_minc, 19, 5, 3, low(minc)).
table(low_s_mins_maxs_minc, 19, 4, 3, low(minc)).
table(low_v_c_maxc_s_minc, 19, 7, 4, low(minc)).
table(low_v_c_maxc_mins_minc, 19, 7, 4, low(minc)).
table(low_v_c_maxc_maxs_minc, 19, 7, 4, low(minc)).
table(low_v_c_s_mins_minc, 19, 7, 4, low(minc)).
table(low_v_c_s_maxs_minc, 19, 7, 4, low(minc)).
table(low_v_c_mins_maxs_minc, 19, 7, 4, low(minc)).
table(low_v_maxc_s_mins_minc, 19, 6, 4, low(minc)).
table(low_v_maxc_s_maxs_minc, 19, 6, 4, low(minc)).
table(low_v_maxc_mins_maxs_minc, 19, 6, 4, low(minc)).
table(low_v_s_mins_maxs_minc, 19, 5, 4, low(minc)).
table(low_c_maxc_s_mins_minc, 19, 6, 4, low(minc)).
table(low_c_maxc_s_maxs_minc, 19, 6, 4, low(minc)).
table(low_c_maxc_mins_maxs_minc, 19, 6, 4, low(minc)).
table(low_c_s_mins_maxs_minc, 19, 6, 4, low(minc)).
table(low_maxc_s_mins_maxs_minc, 19, 6, 4, low(minc)).
table(low_v_c_maxc_s_mins_minc, 19, 8, 5, low(minc)).
table(low_v_c_maxc_s_maxs_minc, 19, 8, 5, low(minc)).
table(low_v_c_maxc_mins_maxs_minc, 19, 8, 5, low(minc)).
table(low_v_c_s_mins_maxs_minc, 19, 8, 5, low(minc)).
table(low_v_maxc_s_mins_maxs_minc, 19, 7, 5, low(minc)).
table(low_c_maxc_s_mins_maxs_minc, 19, 7, 5, low(minc)).
table(low_v_c_maxc_s_mins_maxs_minc, 19, 9, 6, low(minc)).
table(up_v_minc, 19, 11, 1, up(minc)).
table(up_maxc_minc, 19, 6, 1, up(minc)).
table(up_v_c_minc, 19, 5, 2, up(minc)).
table(up_v_maxc_minc, 19, 4, 2, up(minc)).
table(up_v_s_minc, 19, 13, 2, up(minc)).
table(up_v_mins_minc, 19, 12, 2, up(minc)).
table(up_v_maxs_minc, 19, 13, 2, up(minc)).
table(up_c_maxc_minc, 19, 4, 2, up(minc)).
table(up_maxc_s_minc, 19, 4, 2, up(minc)).
table(up_maxc_mins_minc, 19, 4, 2, up(minc)).
table(up_maxc_maxs_minc, 19, 7, 2, up(minc)).
table(up_s_maxs_minc, 19, 14, 2, up(minc)).
table(up_v_c_maxc_minc, 19, 6, 3, up(minc)).
table(up_v_c_s_minc, 19, 6, 3, up(minc)).
table(up_v_c_mins_minc, 19, 6, 3, up(minc)).
table(up_v_c_maxs_minc, 19, 6, 3, up(minc)).
table(up_v_maxc_s_minc, 19, 5, 3, up(minc)).
table(up_v_maxc_mins_minc, 19, 5, 3, up(minc)).
table(up_v_maxc_maxs_minc, 19, 5, 3, up(minc)).
table(up_v_s_mins_minc, 19, 14, 3, up(minc)).
table(up_v_s_maxs_minc, 19, 14, 3, up(minc)).
table(up_v_mins_maxs_minc, 19, 14, 3, up(minc)).
table(up_c_maxc_s_minc, 19, 5, 3, up(minc)).
table(up_c_maxc_mins_minc, 19, 5, 3, up(minc)).
table(up_c_maxc_maxs_minc, 19, 5, 3, up(minc)).
table(up_c_s_maxs_minc, 19, 4, 3, up(minc)).
table(up_maxc_s_mins_minc, 19, 5, 3, up(minc)).
table(up_maxc_s_maxs_minc, 19, 5, 3, up(minc)).
table(up_maxc_mins_maxs_minc, 19, 5, 3, up(minc)).
table(up_s_mins_maxs_minc, 19, 15, 3, up(minc)).
table(up_v_c_maxc_s_minc, 19, 7, 4, up(minc)).
table(up_v_c_maxc_mins_minc, 19, 7, 4, up(minc)).
table(up_v_c_maxc_maxs_minc, 19, 7, 4, up(minc)).
table(up_v_c_s_mins_minc, 19, 7, 4, up(minc)).
table(up_v_c_s_maxs_minc, 19, 7, 4, up(minc)).
table(up_v_c_mins_maxs_minc, 19, 7, 4, up(minc)).
table(up_v_maxc_s_mins_minc, 19, 6, 4, up(minc)).
table(up_v_maxc_s_maxs_minc, 19, 6, 4, up(minc)).
table(up_v_maxc_mins_maxs_minc, 19, 6, 4, up(minc)).
table(up_v_s_mins_maxs_minc, 19, 15, 4, up(minc)).
table(up_c_maxc_s_mins_minc, 19, 6, 4, up(minc)).
table(up_c_maxc_s_maxs_minc, 19, 6, 4, up(minc)).
table(up_c_maxc_mins_maxs_minc, 19, 6, 4, up(minc)).
table(up_c_s_mins_maxs_minc, 19, 5, 4, up(minc)).
table(up_maxc_s_mins_maxs_minc, 19, 6, 4, up(minc)).
table(up_v_c_maxc_s_mins_minc, 19, 8, 5, up(minc)).
table(up_v_c_maxc_s_maxs_minc, 19, 8, 5, up(minc)).
table(up_v_c_maxc_mins_maxs_minc, 19, 8, 5, up(minc)).
table(up_v_c_s_mins_maxs_minc, 19, 8, 5, up(minc)).
table(up_v_maxc_s_mins_maxs_minc, 19, 7, 5, up(minc)).
table(up_c_maxc_s_mins_maxs_minc, 19, 7, 5, up(minc)).
table(up_v_c_maxc_s_mins_maxs_minc, 19, 9, 6, up(minc)).
table(low_v_maxc, 19, 18, 1, low(maxc)).
table(low_c_maxc, 19, 18, 1, low(maxc)).
table(low_minc_maxc, 19, 6, 1, low(maxc)).
table(low_s_maxc, 19, 18, 1, low(maxc)).
table(low_mins_maxc, 19, 9, 1, low(maxc)).
table(low_maxs_maxc, 19, 2, 1, low(maxc)).
table(low_v_c_maxc, 19, 3, 2, low(maxc)).
table(low_v_minc_maxc, 19, 6, 2, low(maxc)).
table(low_v_s_maxc, 19, 4, 2, low(maxc)).
table(low_v_mins_maxc, 19, 9, 2, low(maxc)).
table(low_v_maxs_maxc, 19, 3, 2, low(maxc)).
table(low_c_minc_maxc, 19, 11, 2, low(maxc)).
table(low_c_s_maxc, 19, 4, 2, low(maxc)).
table(low_c_mins_maxc, 19, 18, 2, low(maxc)).
table(low_c_maxs_maxc, 19, 3, 2, low(maxc)).
table(low_minc_s_maxc, 19, 4, 2, low(maxc)).
table(low_minc_mins_maxc, 19, 4, 2, low(maxc)).
table(low_minc_maxs_maxc, 19, 4, 2, low(maxc)).
table(low_s_mins_maxc, 19, 18, 2, low(maxc)).
table(low_s_maxs_maxc, 19, 3, 2, low(maxc)).
table(low_mins_maxs_maxc, 19, 3, 2, low(maxc)).
table(low_v_c_minc_maxc, 19, 6, 3, low(maxc)).
table(low_v_c_s_maxc, 19, 4, 3, low(maxc)).
table(low_v_c_mins_maxc, 19, 4, 3, low(maxc)).
table(low_v_c_maxs_maxc, 19, 4, 3, low(maxc)).
table(low_v_minc_s_maxc, 19, 5, 3, low(maxc)).
table(low_v_minc_mins_maxc, 19, 7, 3, low(maxc)).
table(low_v_minc_maxs_maxc, 19, 5, 3, low(maxc)).
table(low_v_s_mins_maxc, 19, 5, 3, low(maxc)).
table(low_v_s_maxs_maxc, 19, 4, 3, low(maxc)).
table(low_v_mins_maxs_maxc, 19, 4, 3, low(maxc)).
table(low_c_minc_s_maxc, 19, 5, 3, low(maxc)).
table(low_c_minc_mins_maxc, 19, 6, 3, low(maxc)).
table(low_c_minc_maxs_maxc, 19, 5, 3, low(maxc)).
table(low_c_s_mins_maxc, 19, 4, 3, low(maxc)).
table(low_c_s_maxs_maxc, 19, 4, 3, low(maxc)).
table(low_c_mins_maxs_maxc, 19, 4, 3, low(maxc)).
table(low_minc_s_mins_maxc, 19, 5, 3, low(maxc)).
table(low_minc_s_maxs_maxc, 19, 5, 3, low(maxc)).
table(low_minc_mins_maxs_maxc, 19, 5, 3, low(maxc)).
table(low_s_mins_maxs_maxc, 19, 4, 3, low(maxc)).
table(low_v_c_minc_s_maxc, 19, 7, 4, low(maxc)).
table(low_v_c_minc_mins_maxc, 19, 7, 4, low(maxc)).
table(low_v_c_minc_maxs_maxc, 19, 7, 4, low(maxc)).
table(low_v_c_s_mins_maxc, 19, 5, 4, low(maxc)).
table(low_v_c_s_maxs_maxc, 19, 5, 4, low(maxc)).
table(low_v_c_mins_maxs_maxc, 19, 5, 4, low(maxc)).
table(low_v_minc_s_mins_maxc, 19, 6, 4, low(maxc)).
table(low_v_minc_s_maxs_maxc, 19, 6, 4, low(maxc)).
table(low_v_minc_mins_maxs_maxc, 19, 6, 4, low(maxc)).
table(low_v_s_mins_maxs_maxc, 19, 5, 4, low(maxc)).
table(low_c_minc_s_mins_maxc, 19, 6, 4, low(maxc)).
table(low_c_minc_s_maxs_maxc, 19, 6, 4, low(maxc)).
table(low_c_minc_mins_maxs_maxc, 19, 6, 4, low(maxc)).
table(low_c_s_mins_maxs_maxc, 19, 5, 4, low(maxc)).
table(low_minc_s_mins_maxs_maxc, 19, 6, 4, low(maxc)).
table(low_v_c_minc_s_mins_maxc, 19, 8, 5, low(maxc)).
table(low_v_c_minc_s_maxs_maxc, 19, 8, 5, low(maxc)).
table(low_v_c_minc_mins_maxs_maxc, 19, 8, 5, low(maxc)).
table(low_v_c_s_mins_maxs_maxc, 19, 6, 5, low(maxc)).
table(low_v_minc_s_mins_maxs_maxc, 19, 7, 5, low(maxc)).
table(low_c_minc_s_mins_maxs_maxc, 19, 7, 5, low(maxc)).
table(low_v_c_minc_s_mins_maxs_maxc, 19, 9, 6, low(maxc)).
table(up_v_maxc, 19, 11, 1, up(maxc)).
table(up_v_c_maxc, 19, 11, 2, up(maxc)).
table(up_v_minc_maxc, 19, 11, 2, up(maxc)).
table(up_v_s_maxc, 19, 13, 2, up(maxc)).
table(up_v_mins_maxc, 19, 12, 2, up(maxc)).
table(up_v_maxs_maxc, 19, 13, 2, up(maxc)).
table(up_v_c_minc_maxc, 19, 11, 3, up(maxc)).
table(up_v_c_s_maxc, 19, 13, 3, up(maxc)).
table(up_v_c_mins_maxc, 19, 12, 3, up(maxc)).
table(up_v_c_maxs_maxc, 19, 13, 3, up(maxc)).
table(up_v_minc_s_maxc, 19, 12, 3, up(maxc)).
table(up_v_minc_mins_maxc, 19, 12, 3, up(maxc)).
table(up_v_minc_maxs_maxc, 19, 12, 3, up(maxc)).
table(up_v_s_mins_maxc, 19, 14, 3, up(maxc)).
table(up_v_s_maxs_maxc, 19, 14, 3, up(maxc)).
table(up_v_mins_maxs_maxc, 19, 14, 3, up(maxc)).
table(up_v_c_minc_s_maxc, 19, 12, 4, up(maxc)).
table(up_v_c_minc_mins_maxc, 19, 12, 4, up(maxc)).
table(up_v_c_minc_maxs_maxc, 19, 12, 4, up(maxc)).
table(up_v_c_s_mins_maxc, 19, 14, 4, up(maxc)).
table(up_v_c_s_maxs_maxc, 19, 6, 4, up(maxc)).
table(up_v_c_mins_maxs_maxc, 19, 10, 4, up(maxc)).
table(up_v_minc_s_mins_maxc, 19, 13, 4, up(maxc)).
table(up_v_minc_s_maxs_maxc, 19, 13, 4, up(maxc)).
table(up_v_minc_mins_maxs_maxc, 19, 13, 4, up(maxc)).
table(up_v_s_mins_maxs_maxc, 19, 15, 4, up(maxc)).
table(up_v_c_minc_s_mins_maxc, 19, 13, 5, up(maxc)).
table(up_v_c_minc_s_maxs_maxc, 19, 8, 5, up(maxc)).
table(up_v_c_minc_mins_maxs_maxc, 19, 9, 5, up(maxc)).
table(up_v_c_s_mins_maxs_maxc, 19, 10, 5, up(maxc)).
table(up_v_minc_s_mins_maxs_maxc, 19, 14, 5, up(maxc)).
table(up_v_c_minc_s_mins_maxs_maxc, 19, 9, 6, up(maxc)).
table(low_v_s, 19, 18, 1, low(s)).
table(low_c_s, 19, 3, 1, low(s)).
table(low_minc_s, 19, 18, 1, low(s)).
table(low_maxc_s, 19, 18, 1, low(s)).
table(low_mins_s, 19, 18, 1, low(s)).
table(low_maxs_s, 19, 18, 1, low(s)).
table(low_v_c_s, 19, 4, 2, low(s)).
table(low_v_minc_s, 19, 18, 2, low(s)).
table(low_v_maxc_s, 19, 6, 2, low(s)).
table(low_v_mins_s, 19, 9, 2, low(s)).
table(low_v_maxs_s, 19, 3, 2, low(s)).
table(low_c_minc_s, 19, 5, 2, low(s)).
table(low_c_maxc_s, 19, 5, 2, low(s)).
table(low_c_mins_s, 19, 5, 2, low(s)).
table(low_c_maxs_s, 19, 5, 2, low(s)).
table(low_minc_maxc_s, 19, 18, 2, low(s)).
table(low_minc_mins_s, 19, 12, 2, low(s)).
table(low_minc_maxs_s, 19, 14, 2, low(s)).
table(low_maxc_mins_s, 19, 7, 2, low(s)).
table(low_maxc_maxs_s, 19, 14, 2, low(s)).
table(low_mins_maxs_s, 19, 12, 2, low(s)).
table(low_v_c_minc_s, 19, 8, 3, low(s)).
table(low_v_c_maxc_s, 19, 6, 3, low(s)).
table(low_v_c_mins_s, 19, 8, 3, low(s)).
table(low_v_c_maxs_s, 19, 4, 3, low(s)).
table(low_v_minc_maxc_s, 19, 10, 3, low(s)).
table(low_v_minc_mins_s, 19, 13, 3, low(s)).
table(low_v_minc_maxs_s, 19, 4, 3, low(s)).
table(low_v_maxc_mins_s, 19, 4, 3, low(s)).
table(low_v_maxc_maxs_s, 19, 4, 3, low(s)).
table(low_v_mins_maxs_s, 19, 4, 3, low(s)).
table(low_c_minc_maxc_s, 19, 8, 3, low(s)).
table(low_c_minc_mins_s, 19, 4, 3, low(s)).
table(low_c_minc_maxs_s, 19, 5, 3, low(s)).
table(low_c_maxc_mins_s, 19, 8, 3, low(s)).
table(low_c_maxc_maxs_s, 19, 5, 3, low(s)).
table(low_c_mins_maxs_s, 19, 8, 3, low(s)).
table(low_minc_maxc_mins_s, 19, 13, 3, low(s)).
table(low_minc_maxc_maxs_s, 19, 14, 3, low(s)).
table(low_minc_mins_maxs_s, 19, 8, 3, low(s)).
table(low_maxc_mins_maxs_s, 19, 9, 3, low(s)).
table(low_v_c_minc_maxc_s, 19, 10, 4, low(s)).
table(low_v_c_minc_mins_s, 19, 7, 4, low(s)).
table(low_v_c_minc_maxs_s, 19, 7, 4, low(s)).
table(low_v_c_maxc_mins_s, 19, 5, 4, low(s)).
table(low_v_c_maxc_maxs_s, 19, 5, 4, low(s)).
table(low_v_c_mins_maxs_s, 19, 5, 4, low(s)).
table(low_v_minc_maxc_mins_s, 19, 8, 4, low(s)).
table(low_v_minc_maxc_maxs_s, 19, 6, 4, low(s)).
table(low_v_minc_mins_maxs_s, 19, 5, 4, low(s)).
table(low_v_maxc_mins_maxs_s, 19, 5, 4, low(s)).
table(low_c_minc_maxc_mins_s, 19, 6, 4, low(s)).
table(low_c_minc_maxc_maxs_s, 19, 7, 4, low(s)).
table(low_c_minc_mins_maxs_s, 19, 5, 4, low(s)).
table(low_c_maxc_mins_maxs_s, 19, 6, 4, low(s)).
table(low_minc_maxc_mins_maxs_s, 19, 9, 4, low(s)).
table(low_v_c_minc_maxc_mins_s, 19, 8, 5, low(s)).
table(low_v_c_minc_maxc_maxs_s, 19, 8, 5, low(s)).
table(low_v_c_minc_mins_maxs_s, 19, 8, 5, low(s)).
table(low_v_c_maxc_mins_maxs_s, 19, 6, 5, low(s)).
table(low_v_minc_maxc_mins_maxs_s, 19, 7, 5, low(s)).
table(low_c_minc_maxc_mins_maxs_s, 19, 7, 5, low(s)).
table(low_v_c_minc_maxc_mins_maxs_s, 19, 9, 6, low(s)).
table(up_v_s, 19, 7, 1, up(s)).
table(up_v_c_s, 19, 8, 2, up(s)).
table(up_v_minc_s, 19, 8, 2, up(s)).
table(up_v_maxc_s, 19, 8, 2, up(s)).
table(up_v_mins_s, 19, 4, 2, up(s)).
table(up_v_maxs_s, 19, 7, 2, up(s)).
table(up_c_maxc_s, 19, 9, 2, up(s)).
table(up_v_c_minc_s, 19, 11, 3, up(s)).
table(up_v_c_maxc_s, 19, 9, 3, up(s)).
table(up_v_c_mins_s, 19, 5, 3, up(s)).
table(up_v_c_maxs_s, 19, 8, 3, up(s)).
table(up_v_minc_maxc_s, 19, 10, 3, up(s)).
table(up_v_minc_mins_s, 19, 5, 3, up(s)).
table(up_v_minc_maxs_s, 19, 8, 3, up(s)).
table(up_v_maxc_mins_s, 19, 5, 3, up(s)).
table(up_v_maxc_maxs_s, 19, 8, 3, up(s)).
table(up_v_mins_maxs_s, 19, 5, 3, up(s)).
table(up_c_minc_maxc_s, 19, 12, 3, up(s)).
table(up_c_maxc_mins_s, 19, 5, 3, up(s)).
table(up_c_maxc_maxs_s, 19, 9, 3, up(s)).
table(up_v_c_minc_maxc_s, 19, 12, 4, up(s)).
table(up_v_c_minc_mins_s, 19, 8, 4, up(s)).
table(up_v_c_minc_maxs_s, 19, 11, 4, up(s)).
table(up_v_c_maxc_mins_s, 19, 6, 4, up(s)).
table(up_v_c_maxc_maxs_s, 19, 9, 4, up(s)).
table(up_v_c_mins_maxs_s, 19, 6, 4, up(s)).
table(up_v_minc_maxc_mins_s, 19, 7, 4, up(s)).
table(up_v_minc_maxc_maxs_s, 19, 10, 4, up(s)).
table(up_v_minc_mins_maxs_s, 19, 6, 4, up(s)).
table(up_v_maxc_mins_maxs_s, 19, 6, 4, up(s)).
table(up_c_minc_maxc_mins_s, 19, 7, 4, up(s)).
table(up_c_minc_maxc_maxs_s, 19, 12, 4, up(s)).
table(up_c_maxc_mins_maxs_s, 19, 6, 4, up(s)).
table(up_v_c_minc_maxc_mins_s, 19, 9, 5, up(s)).
table(up_v_c_minc_maxc_maxs_s, 19, 12, 5, up(s)).
table(up_v_c_minc_mins_maxs_s, 19, 9, 5, up(s)).
table(up_v_c_maxc_mins_maxs_s, 19, 7, 5, up(s)).
table(up_v_minc_maxc_mins_maxs_s, 19, 8, 5, up(s)).
table(up_c_minc_maxc_mins_maxs_s, 19, 8, 5, up(s)).
table(up_v_c_minc_maxc_mins_maxs_s, 19, 10, 6, up(s)).
table(low_v_mins, 19, 2, 1, low(mins)).
table(low_c_mins, 19, 2, 1, low(mins)).
table(low_minc_mins, 19, 2, 1, low(mins)).
table(low_maxc_mins, 19, 2, 1, low(mins)).
table(low_s_mins, 19, 2, 1, low(mins)).
table(low_maxs_mins, 19, 2, 1, low(mins)).
table(low_v_c_mins, 19, 3, 2, low(mins)).
table(low_v_minc_mins, 19, 3, 2, low(mins)).
table(low_v_maxc_mins, 19, 3, 2, low(mins)).
table(low_v_s_mins, 19, 3, 2, low(mins)).
table(low_v_maxs_mins, 19, 3, 2, low(mins)).
table(low_c_minc_mins, 19, 3, 2, low(mins)).
table(low_c_maxc_mins, 19, 3, 2, low(mins)).
table(low_c_s_mins, 19, 3, 2, low(mins)).
table(low_c_maxs_mins, 19, 3, 2, low(mins)).
table(low_minc_maxc_mins, 19, 4, 2, low(mins)).
table(low_minc_s_mins, 19, 3, 2, low(mins)).
table(low_minc_maxs_mins, 19, 3, 2, low(mins)).
table(low_maxc_s_mins, 19, 3, 2, low(mins)).
table(low_maxc_maxs_mins, 19, 3, 2, low(mins)).
table(low_s_maxs_mins, 19, 3, 2, low(mins)).
table(low_v_c_minc_mins, 19, 6, 3, low(mins)).
table(low_v_c_maxc_mins, 19, 4, 3, low(mins)).
table(low_v_c_s_mins, 19, 4, 3, low(mins)).
table(low_v_c_maxs_mins, 19, 4, 3, low(mins)).
table(low_v_minc_maxc_mins, 19, 5, 3, low(mins)).
table(low_v_minc_s_mins, 19, 4, 3, low(mins)).
table(low_v_minc_maxs_mins, 19, 4, 3, low(mins)).
table(low_v_maxc_s_mins, 19, 4, 3, low(mins)).
table(low_v_maxc_maxs_mins, 19, 4, 3, low(mins)).
table(low_v_s_maxs_mins, 19, 4, 3, low(mins)).
table(low_c_minc_maxc_mins, 19, 5, 3, low(mins)).
table(low_c_minc_s_mins, 19, 4, 3, low(mins)).
table(low_c_minc_maxs_mins, 19, 4, 3, low(mins)).
table(low_c_maxc_s_mins, 19, 4, 3, low(mins)).
table(low_c_maxc_maxs_mins, 19, 4, 3, low(mins)).
table(low_c_s_maxs_mins, 19, 4, 3, low(mins)).
table(low_minc_maxc_s_mins, 19, 5, 3, low(mins)).
table(low_minc_maxc_maxs_mins, 19, 5, 3, low(mins)).
table(low_minc_s_maxs_mins, 19, 4, 3, low(mins)).
table(low_maxc_s_maxs_mins, 19, 4, 3, low(mins)).
table(low_v_c_minc_maxc_mins, 19, 7, 4, low(mins)).
table(low_v_c_minc_s_mins, 19, 7, 4, low(mins)).
table(low_v_c_minc_maxs_mins, 19, 7, 4, low(mins)).
table(low_v_c_maxc_s_mins, 19, 5, 4, low(mins)).
table(low_v_c_maxc_maxs_mins, 19, 5, 4, low(mins)).
table(low_v_c_s_maxs_mins, 19, 5, 4, low(mins)).
table(low_v_minc_maxc_s_mins, 19, 6, 4, low(mins)).
table(low_v_minc_maxc_maxs_mins, 19, 6, 4, low(mins)).
table(low_v_minc_s_maxs_mins, 19, 5, 4, low(mins)).
table(low_v_maxc_s_maxs_mins, 19, 5, 4, low(mins)).
table(low_c_minc_maxc_s_mins, 19, 6, 4, low(mins)).
table(low_c_minc_maxc_maxs_mins, 19, 6, 4, low(mins)).
table(low_c_minc_s_maxs_mins, 19, 5, 4, low(mins)).
table(low_c_maxc_s_maxs_mins, 19, 5, 4, low(mins)).
table(low_minc_maxc_s_maxs_mins, 19, 6, 4, low(mins)).
table(low_v_c_minc_maxc_s_mins, 19, 8, 5, low(mins)).
table(low_v_c_minc_maxc_maxs_mins, 19, 8, 5, low(mins)).
table(low_v_c_minc_s_maxs_mins, 19, 8, 5, low(mins)).
table(low_v_c_maxc_s_maxs_mins, 19, 6, 5, low(mins)).
table(low_v_minc_maxc_s_maxs_mins, 19, 7, 5, low(mins)).
table(low_c_minc_maxc_s_maxs_mins, 19, 7, 5, low(mins)).
table(low_v_c_minc_maxc_s_maxs_mins, 19, 9, 6, low(mins)).
table(up_v_mins, 19, 18, 1, up(mins)).
table(up_maxc_mins, 19, 9, 1, up(mins)).
table(up_maxs_mins, 19, 2, 1, up(mins)).
table(up_v_c_mins, 19, 6, 2, up(mins)).
table(up_v_minc_mins, 19, 3, 2, up(mins)).
table(up_v_maxc_mins, 19, 5, 2, up(mins)).
table(up_v_s_mins, 19, 3, 2, up(mins)).
table(up_v_maxs_mins, 19, 3, 2, up(mins)).
table(up_c_maxc_mins, 19, 5, 2, up(mins)).
table(up_c_maxs_mins, 19, 3, 2, up(mins)).
table(up_minc_maxc_mins, 19, 4, 2, up(mins)).
table(up_minc_maxs_mins, 19, 3, 2, up(mins)).
table(up_maxc_s_mins, 19, 3, 2, up(mins)).
table(up_maxc_maxs_mins, 19, 3, 2, up(mins)).
table(up_s_maxs_mins, 19, 3, 2, up(mins)).
table(up_v_c_minc_mins, 19, 6, 3, up(mins)).
table(up_v_c_maxc_mins, 19, 7, 3, up(mins)).
table(up_v_c_s_mins, 19, 4, 3, up(mins)).
table(up_v_c_maxs_mins, 19, 4, 3, up(mins)).
table(up_v_minc_maxc_mins, 19, 5, 3, up(mins)).
table(up_v_minc_s_mins, 19, 4, 3, up(mins)).
table(up_v_minc_maxs_mins, 19, 4, 3, up(mins)).
table(up_v_maxc_s_mins, 19, 4, 3, up(mins)).
table(up_v_maxc_maxs_mins, 19, 4, 3, up(mins)).
table(up_v_s_maxs_mins, 19, 4, 3, up(mins)).
table(up_c_minc_maxc_mins, 19, 5, 3, up(mins)).
table(up_c_minc_maxs_mins, 19, 4, 3, up(mins)).
table(up_c_maxc_s_mins, 19, 4, 3, up(mins)).
table(up_c_maxc_maxs_mins, 19, 4, 3, up(mins)).
table(up_c_s_maxs_mins, 19, 4, 3, up(mins)).
table(up_minc_maxc_s_mins, 19, 5, 3, up(mins)).
table(up_minc_maxc_maxs_mins, 19, 5, 3, up(mins)).
table(up_minc_s_maxs_mins, 19, 4, 3, up(mins)).
table(up_maxc_s_maxs_mins, 19, 4, 3, up(mins)).
table(up_v_c_minc_maxc_mins, 19, 7, 4, up(mins)).
table(up_v_c_minc_s_mins, 19, 7, 4, up(mins)).
table(up_v_c_minc_maxs_mins, 19, 7, 4, up(mins)).
table(up_v_c_maxc_s_mins, 19, 5, 4, up(mins)).
table(up_v_c_maxc_maxs_mins, 19, 5, 4, up(mins)).
table(up_v_c_s_maxs_mins, 19, 5, 4, up(mins)).
table(up_v_minc_maxc_s_mins, 19, 6, 4, up(mins)).
table(up_v_minc_maxc_maxs_mins, 19, 6, 4, up(mins)).
table(up_v_minc_s_maxs_mins, 19, 5, 4, up(mins)).
table(up_v_maxc_s_maxs_mins, 19, 5, 4, up(mins)).
table(up_c_minc_maxc_s_mins, 19, 6, 4, up(mins)).
table(up_c_minc_maxc_maxs_mins, 19, 6, 4, up(mins)).
table(up_c_minc_s_maxs_mins, 19, 5, 4, up(mins)).
table(up_c_maxc_s_maxs_mins, 19, 5, 4, up(mins)).
table(up_minc_maxc_s_maxs_mins, 19, 6, 4, up(mins)).
table(up_v_c_minc_maxc_s_mins, 19, 8, 5, up(mins)).
table(up_v_c_minc_maxc_maxs_mins, 19, 8, 5, up(mins)).
table(up_v_c_minc_s_maxs_mins, 19, 8, 5, up(mins)).
table(up_v_c_maxc_s_maxs_mins, 19, 6, 5, up(mins)).
table(up_v_minc_maxc_s_maxs_mins, 19, 7, 5, up(mins)).
table(up_c_minc_maxc_s_maxs_mins, 19, 7, 5, up(mins)).
table(up_v_c_minc_maxc_s_maxs_mins, 19, 9, 6, up(mins)).
table(low_v_maxs, 19, 7, 1, low(maxs)).
table(low_c_maxs, 19, 4, 1, low(maxs)).
table(low_minc_maxs, 19, 4, 1, low(maxs)).
table(low_maxc_maxs, 19, 4, 1, low(maxs)).
table(low_s_maxs, 19, 7, 1, low(maxs)).
table(low_mins_maxs, 19, 2, 1, low(maxs)).
table(low_v_c_maxs, 19, 8, 2, low(maxs)).
table(low_v_minc_maxs, 19, 8, 2, low(maxs)).
table(low_v_maxc_maxs, 19, 8, 2, low(maxs)).
table(low_v_s_maxs, 19, 3, 2, low(maxs)).
table(low_v_mins_maxs, 19, 3, 2, low(maxs)).
table(low_c_minc_maxs, 19, 5, 2, low(maxs)).
table(low_c_maxc_maxs, 19, 5, 2, low(maxs)).
table(low_c_s_maxs, 19, 8, 2, low(maxs)).
table(low_c_mins_maxs, 19, 3, 2, low(maxs)).
table(low_minc_maxc_maxs, 19, 6, 2, low(maxs)).
table(low_minc_s_maxs, 19, 3, 2, low(maxs)).
table(low_minc_mins_maxs, 19, 3, 2, low(maxs)).
table(low_maxc_s_maxs, 19, 3, 2, low(maxs)).
table(low_maxc_mins_maxs, 19, 3, 2, low(maxs)).
table(low_s_mins_maxs, 19, 7, 2, low(maxs)).
table(low_v_c_minc_maxs, 19, 11, 3, low(maxs)).
table(low_v_c_maxc_maxs, 19, 9, 3, low(maxs)).
table(low_v_c_s_maxs, 19, 4, 3, low(maxs)).
table(low_v_c_mins_maxs, 19, 4, 3, low(maxs)).
table(low_v_minc_maxc_maxs, 19, 10, 3, low(maxs)).
table(low_v_minc_s_maxs, 19, 4, 3, low(maxs)).
table(low_v_minc_mins_maxs, 19, 4, 3, low(maxs)).
table(low_v_maxc_s_maxs, 19, 4, 3, low(maxs)).
table(low_v_maxc_mins_maxs, 19, 4, 3, low(maxs)).
table(low_v_s_mins_maxs, 19, 4, 3, low(maxs)).
table(low_c_minc_maxc_maxs, 19, 7, 3, low(maxs)).
table(low_c_minc_s_maxs, 19, 4, 3, low(maxs)).
table(low_c_minc_mins_maxs, 19, 4, 3, low(maxs)).
table(low_c_maxc_s_maxs, 19, 4, 3, low(maxs)).
table(low_c_maxc_mins_maxs, 19, 4, 3, low(maxs)).
table(low_c_s_mins_maxs, 19, 8, 3, low(maxs)).
table(low_minc_maxc_s_maxs, 19, 5, 3, low(maxs)).
table(low_minc_maxc_mins_maxs, 19, 5, 3, low(maxs)).
table(low_minc_s_mins_maxs, 19, 4, 3, low(maxs)).
table(low_maxc_s_mins_maxs, 19, 4, 3, low(maxs)).
table(low_v_c_minc_maxc_maxs, 19, 12, 4, low(maxs)).
table(low_v_c_minc_s_maxs, 19, 7, 4, low(maxs)).
table(low_v_c_minc_mins_maxs, 19, 7, 4, low(maxs)).
table(low_v_c_maxc_s_maxs, 19, 5, 4, low(maxs)).
table(low_v_c_maxc_mins_maxs, 19, 5, 4, low(maxs)).
table(low_v_c_s_mins_maxs, 19, 5, 4, low(maxs)).
table(low_v_minc_maxc_s_maxs, 19, 6, 4, low(maxs)).
table(low_v_minc_maxc_mins_maxs, 19, 6, 4, low(maxs)).
table(low_v_minc_s_mins_maxs, 19, 5, 4, low(maxs)).
table(low_v_maxc_s_mins_maxs, 19, 5, 4, low(maxs)).
table(low_c_minc_maxc_s_maxs, 19, 6, 4, low(maxs)).
table(low_c_minc_maxc_mins_maxs, 19, 6, 4, low(maxs)).
table(low_c_minc_s_mins_maxs, 19, 5, 4, low(maxs)).
table(low_c_maxc_s_mins_maxs, 19, 5, 4, low(maxs)).
table(low_minc_maxc_s_mins_maxs, 19, 6, 4, low(maxs)).
table(low_v_c_minc_maxc_s_maxs, 19, 8, 5, low(maxs)).
table(low_v_c_minc_maxc_mins_maxs, 19, 8, 5, low(maxs)).
table(low_v_c_minc_s_mins_maxs, 19, 8, 5, low(maxs)).
table(low_v_c_maxc_s_mins_maxs, 19, 6, 5, low(maxs)).
table(low_v_minc_maxc_s_mins_maxs, 19, 7, 5, low(maxs)).
table(low_c_minc_maxc_s_mins_maxs, 19, 7, 5, low(maxs)).
table(low_v_c_minc_maxc_s_mins_maxs, 19, 9, 6, low(maxs)).
table(up_v_maxs, 19, 18, 1, up(maxs)).
table(up_maxc_maxs, 19, 2, 1, up(maxs)).
table(up_v_c_maxs, 19, 18, 2, up(maxs)).
table(up_v_minc_maxs, 19, 12, 2, up(maxs)).
table(up_v_maxc_maxs, 19, 3, 2, up(maxs)).
table(up_v_s_maxs, 19, 7, 2, up(maxs)).
table(up_v_mins_maxs, 19, 9, 2, up(maxs)).
table(up_c_maxc_maxs, 19, 3, 2, up(maxs)).
table(up_minc_maxc_maxs, 19, 4, 2, up(maxs)).
table(up_maxc_s_maxs, 19, 3, 2, up(maxs)).
table(up_maxc_mins_maxs, 19, 3, 2, up(maxs)).
table(up_v_c_minc_maxs, 19, 12, 3, up(maxs)).
table(up_v_c_maxc_maxs, 19, 4, 3, up(maxs)).
table(up_v_c_s_maxs, 19, 8, 3, up(maxs)).
table(up_v_c_mins_maxs, 19, 18, 3, up(maxs)).
table(up_v_minc_maxc_maxs, 19, 5, 3, up(maxs)).
table(up_v_minc_s_maxs, 19, 5, 3, up(maxs)).
table(up_v_minc_mins_maxs, 19, 14, 3, up(maxs)).
table(up_v_maxc_s_maxs, 19, 4, 3, up(maxs)).
table(up_v_maxc_mins_maxs, 19, 4, 3, up(maxs)).
table(up_v_s_mins_maxs, 19, 7, 3, up(maxs)).
table(up_c_minc_maxc_maxs, 19, 5, 3, up(maxs)).
table(up_c_maxc_s_maxs, 19, 4, 3, up(maxs)).
table(up_c_maxc_mins_maxs, 19, 4, 3, up(maxs)).
table(up_minc_maxc_s_maxs, 19, 5, 3, up(maxs)).
table(up_minc_maxc_mins_maxs, 19, 5, 3, up(maxs)).
table(up_maxc_s_mins_maxs, 19, 4, 3, up(maxs)).
table(up_v_c_minc_maxc_maxs, 19, 7, 4, up(maxs)).
table(up_v_c_minc_s_maxs, 19, 7, 4, up(maxs)).
table(up_v_c_minc_mins_maxs, 19, 13, 4, up(maxs)).
table(up_v_c_maxc_s_maxs, 19, 5, 4, up(maxs)).
table(up_v_c_maxc_mins_maxs, 19, 5, 4, up(maxs)).
table(up_v_c_s_mins_maxs, 19, 8, 4, up(maxs)).
table(up_v_minc_maxc_s_maxs, 19, 6, 4, up(maxs)).
table(up_v_minc_maxc_mins_maxs, 19, 6, 4, up(maxs)).
table(up_v_minc_s_mins_maxs, 19, 5, 4, up(maxs)).
table(up_v_maxc_s_mins_maxs, 19, 5, 4, up(maxs)).
table(up_c_minc_maxc_s_maxs, 19, 6, 4, up(maxs)).
table(up_c_minc_maxc_mins_maxs, 19, 6, 4, up(maxs)).
table(up_c_maxc_s_mins_maxs, 19, 5, 4, up(maxs)).
table(up_minc_maxc_s_mins_maxs, 19, 6, 4, up(maxs)).
table(up_v_c_minc_maxc_s_maxs, 19, 8, 5, up(maxs)).
table(up_v_c_minc_maxc_mins_maxs, 19, 8, 5, up(maxs)).
table(up_v_c_minc_s_mins_maxs, 19, 8, 5, up(maxs)).
table(up_v_c_maxc_s_mins_maxs, 19, 6, 5, up(maxs)).
table(up_v_minc_maxc_s_mins_maxs, 19, 7, 5, up(maxs)).
table(up_c_minc_maxc_s_mins_maxs, 19, 7, 5, up(maxs)).
table(up_v_c_minc_maxc_s_mins_maxs, 19, 9, 6, up(maxs)).
table(low_v_mina, 19, 12, 1, low(mina)).
table(low_c_mina, 19, 7, 1, low(mina)).
table(low_minc_mina, 19, 19, 1, low(mina)).
table(low_maxc_mina, 19, 19, 1, low(mina)).
table(low_s_mina, 19, 12, 1, low(mina)).
table(low_mins_mina, 19, 11, 1, low(mina)).
table(low_maxs_mina, 19, 11, 1, low(mina)).
table(low_v_c_mina, 19, 10, 2, low(mina)).
table(low_v_minc_mina, 19, 11, 2, low(mina)).
table(low_v_maxc_mina, 19, 11, 2, low(mina)).
table(low_v_s_mina, 19, 4, 2, low(mina)).
table(low_v_mins_mina, 19, 4, 2, low(mina)).
table(low_v_maxs_mina, 19, 10, 2, low(mina)).
table(low_c_minc_mina, 19, 8, 2, low(mina)).
table(low_c_maxc_mina, 19, 6, 2, low(mina)).
table(low_c_s_mina, 19, 10, 2, low(mina)).
table(low_c_mins_mina, 19, 7, 2, low(mina)).
table(low_c_maxs_mina, 19, 5, 2, low(mina)).
table(low_minc_maxc_mina, 19, 19, 2, low(mina)).
table(low_minc_s_mina, 19, 11, 2, low(mina)).
table(low_minc_mins_mina, 19, 19, 2, low(mina)).
table(low_minc_maxs_mina, 19, 19, 2, low(mina)).
table(low_maxc_s_mina, 19, 11, 2, low(mina)).
table(low_maxc_mins_mina, 19, 19, 2, low(mina)).
table(low_maxc_maxs_mina, 19, 19, 2, low(mina)).
table(low_s_mins_mina, 19, 12, 2, low(mina)).
table(low_s_maxs_mina, 19, 10, 2, low(mina)).
table(low_mins_maxs_mina, 19, 6, 2, low(mina)).
table(low_v_c_minc_mina, 19, 13, 3, low(mina)).
table(low_v_c_maxc_mina, 19, 11, 3, low(mina)).
table(low_v_c_s_mina, 19, 4, 3, low(mina)).
table(low_v_c_mins_mina, 19, 7, 3, low(mina)).
table(low_v_c_maxs_mina, 19, 10, 3, low(mina)).
table(low_v_minc_maxc_mina, 19, 12, 3, low(mina)).
table(low_v_minc_s_mina, 19, 4, 3, low(mina)).
table(low_v_minc_mins_mina, 19, 4, 3, low(mina)).
table(low_v_minc_maxs_mina, 19, 11, 3, low(mina)).
table(low_v_maxc_s_mina, 19, 4, 3, low(mina)).
table(low_v_maxc_mins_mina, 19, 6, 3, low(mina)).
table(low_v_maxc_maxs_mina, 19, 10, 3, low(mina)).
table(low_v_s_mins_mina, 19, 5, 3, low(mina)).
table(low_v_s_maxs_mina, 19, 5, 3, low(mina)).
table(low_v_mins_maxs_mina, 19, 5, 3, low(mina)).
table(low_c_minc_maxc_mina, 19, 8, 3, low(mina)).
table(low_c_minc_s_mina, 19, 11, 3, low(mina)).
table(low_c_minc_mins_mina, 19, 8, 3, low(mina)).
table(low_c_minc_maxs_mina, 19, 6, 3, low(mina)).
table(low_c_maxc_s_mina, 19, 8, 3, low(mina)).
table(low_c_maxc_mins_mina, 19, 6, 3, low(mina)).
table(low_c_maxc_maxs_mina, 19, 6, 3, low(mina)).
table(low_c_s_mins_mina, 19, 10, 3, low(mina)).
table(low_c_s_maxs_mina, 19, 10, 3, low(mina)).
table(low_c_mins_maxs_mina, 19, 5, 3, low(mina)).
table(low_minc_maxc_s_mina, 19, 10, 3, low(mina)).
table(low_minc_maxc_mins_mina, 19, 16, 3, low(mina)).
table(low_minc_maxc_maxs_mina, 19, 18, 3, low(mina)).
table(low_minc_s_mins_mina, 19, 6, 3, low(mina)).
table(low_minc_s_maxs_mina, 19, 8, 3, low(mina)).
table(low_minc_mins_maxs_mina, 19, 17, 3, low(mina)).
table(low_maxc_s_mins_mina, 19, 8, 3, low(mina)).
table(low_maxc_s_maxs_mina, 19, 7, 3, low(mina)).
table(low_maxc_mins_maxs_mina, 19, 8, 3, low(mina)).
table(low_s_mins_maxs_mina, 19, 10, 3, low(mina)).
table(low_v_c_minc_maxc_mina, 19, 14, 4, low(mina)).
table(low_v_c_minc_s_mina, 19, 8, 4, low(mina)).
table(low_v_c_minc_mins_mina, 19, 10, 4, low(mina)).
table(low_v_c_minc_maxs_mina, 19, 13, 4, low(mina)).
table(low_v_c_maxc_s_mina, 19, 5, 4, low(mina)).
table(low_v_c_maxc_mins_mina, 19, 8, 4, low(mina)).
table(low_v_c_maxc_maxs_mina, 19, 11, 4, low(mina)).
table(low_v_c_s_mins_mina, 19, 5, 4, low(mina)).
table(low_v_c_s_maxs_mina, 19, 5, 4, low(mina)).
table(low_v_c_mins_maxs_mina, 19, 8, 4, low(mina)).
table(low_v_minc_maxc_s_mina, 19, 6, 4, low(mina)).
table(low_v_minc_maxc_mins_mina, 19, 6, 4, low(mina)).
table(low_v_minc_maxc_maxs_mina, 19, 12, 4, low(mina)).
table(low_v_minc_s_mins_mina, 19, 5, 4, low(mina)).
table(low_v_minc_s_maxs_mina, 19, 5, 4, low(mina)).
table(low_v_minc_mins_maxs_mina, 19, 5, 4, low(mina)).
table(low_v_maxc_s_mins_mina, 19, 5, 4, low(mina)).
table(low_v_maxc_s_maxs_mina, 19, 5, 4, low(mina)).
table(low_v_maxc_mins_maxs_mina, 19, 5, 4, low(mina)).
table(low_v_s_mins_maxs_mina, 19, 6, 4, low(mina)).
table(low_c_minc_maxc_s_mina, 19, 9, 4, low(mina)).
table(low_c_minc_maxc_mins_mina, 19, 7, 4, low(mina)).
table(low_c_minc_maxc_maxs_mina, 19, 8, 4, low(mina)).
table(low_c_minc_s_mins_mina, 19, 10, 4, low(mina)).
table(low_c_minc_s_maxs_mina, 19, 10, 4, low(mina)).
table(low_c_minc_mins_maxs_mina, 19, 6, 4, low(mina)).
table(low_c_maxc_s_mins_mina, 19, 6, 4, low(mina)).
table(low_c_maxc_s_maxs_mina, 19, 6, 4, low(mina)).
table(low_c_maxc_mins_maxs_mina, 19, 6, 4, low(mina)).
table(low_c_s_mins_maxs_mina, 19, 10, 4, low(mina)).
table(low_minc_maxc_s_mins_mina, 19, 7, 4, low(mina)).
table(low_minc_maxc_s_maxs_mina, 19, 9, 4, low(mina)).
table(low_minc_maxc_mins_maxs_mina, 19, 16, 4, low(mina)).
table(low_minc_s_mins_maxs_mina, 19, 6, 4, low(mina)).
table(low_maxc_s_mins_maxs_mina, 19, 6, 4, low(mina)).
table(low_v_c_minc_maxc_s_mina, 19, 9, 5, low(mina)).
table(low_v_c_minc_maxc_mins_mina, 19, 11, 5, low(mina)).
table(low_v_c_minc_maxc_maxs_mina, 19, 14, 5, low(mina)).
table(low_v_c_minc_s_mins_mina, 19, 8, 5, low(mina)).
table(low_v_c_minc_s_maxs_mina, 19, 8, 5, low(mina)).
table(low_v_c_minc_mins_maxs_mina, 19, 11, 5, low(mina)).
table(low_v_c_maxc_s_mins_mina, 19, 6, 5, low(mina)).
table(low_v_c_maxc_s_maxs_mina, 19, 6, 5, low(mina)).
table(low_v_c_maxc_mins_maxs_mina, 19, 9, 5, low(mina)).
table(low_v_c_s_mins_maxs_mina, 19, 6, 5, low(mina)).
table(low_v_minc_maxc_s_mins_mina, 19, 7, 5, low(mina)).
table(low_v_minc_maxc_s_maxs_mina, 19, 7, 5, low(mina)).
table(low_v_minc_maxc_mins_maxs_mina, 19, 7, 5, low(mina)).
table(low_v_minc_s_mins_maxs_mina, 19, 6, 5, low(mina)).
table(low_v_maxc_s_mins_maxs_mina, 19, 6, 5, low(mina)).
table(low_c_minc_maxc_s_mins_mina, 19, 8, 5, low(mina)).
table(low_c_minc_maxc_s_maxs_mina, 19, 8, 5, low(mina)).
table(low_c_minc_maxc_mins_maxs_mina, 19, 8, 5, low(mina)).
table(low_c_minc_s_mins_maxs_mina, 19, 11, 5, low(mina)).
table(low_c_maxc_s_mins_maxs_mina, 19, 7, 5, low(mina)).
table(low_minc_maxc_s_mins_maxs_mina, 19, 8, 5, low(mina)).
table(low_v_c_minc_maxc_s_mins_mina, 19, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_maxs_mina, 19, 9, 6, low(mina)).
table(low_v_c_minc_maxc_mins_maxs_mina, 19, 12, 6, low(mina)).
table(low_v_c_minc_s_mins_maxs_mina, 19, 9, 6, low(mina)).
table(low_v_c_maxc_s_mins_maxs_mina, 19, 7, 6, low(mina)).
table(low_v_minc_maxc_s_mins_maxs_mina, 19, 8, 6, low(mina)).
table(low_c_minc_maxc_s_mins_maxs_mina, 19, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_mins_maxs_mina, 19, 10, 7, low(mina)).
table(up_v_maxa, 19, 19, 1, up(maxa)).
table(up_v_c_maxa, 19, 19, 2, up(maxa)).
table(up_v_minc_maxa, 19, 19, 2, up(maxa)).
table(up_v_maxc_maxa, 19, 19, 2, up(maxa)).
table(up_v_s_maxa, 19, 19, 2, up(maxa)).
table(up_v_mins_maxa, 19, 19, 2, up(maxa)).
table(up_v_maxs_maxa, 19, 19, 2, up(maxa)).
table(up_c_maxc_maxa, 19, 19, 2, up(maxa)).
table(up_maxc_s_maxa, 19, 18, 2, up(maxa)).
table(up_s_maxs_maxa, 19, 19, 2, up(maxa)).
table(up_v_c_minc_maxa, 19, 19, 3, up(maxa)).
table(up_v_c_maxc_maxa, 19, 19, 3, up(maxa)).
table(up_v_c_s_maxa, 19, 19, 3, up(maxa)).
table(up_v_c_mins_maxa, 19, 19, 3, up(maxa)).
table(up_v_c_maxs_maxa, 19, 7, 3, up(maxa)).
table(up_v_minc_maxc_maxa, 19, 19, 3, up(maxa)).
table(up_v_minc_s_maxa, 19, 18, 3, up(maxa)).
table(up_v_minc_mins_maxa, 19, 19, 3, up(maxa)).
table(up_v_minc_maxs_maxa, 19, 19, 3, up(maxa)).
table(up_v_maxc_s_maxa, 19, 18, 3, up(maxa)).
table(up_v_maxc_mins_maxa, 19, 11, 3, up(maxa)).
table(up_v_maxc_maxs_maxa, 19, 6, 3, up(maxa)).
table(up_v_s_mins_maxa, 19, 19, 3, up(maxa)).
table(up_v_s_maxs_maxa, 19, 19, 3, up(maxa)).
table(up_v_mins_maxs_maxa, 19, 19, 3, up(maxa)).
table(up_c_minc_maxc_maxa, 19, 19, 3, up(maxa)).
table(up_c_maxc_s_maxa, 19, 18, 3, up(maxa)).
table(up_c_maxc_mins_maxa, 19, 19, 3, up(maxa)).
table(up_c_maxc_maxs_maxa, 19, 7, 3, up(maxa)).
table(up_c_s_maxs_maxa, 19, 19, 3, up(maxa)).
table(up_minc_maxc_s_maxa, 19, 6, 3, up(maxa)).
table(up_minc_s_maxs_maxa, 19, 18, 3, up(maxa)).
table(up_maxc_s_mins_maxa, 19, 9, 3, up(maxa)).
table(up_maxc_s_maxs_maxa, 19, 6, 3, up(maxa)).
table(up_s_mins_maxs_maxa, 19, 19, 3, up(maxa)).
table(up_v_c_minc_maxc_maxa, 19, 19, 4, up(maxa)).
table(up_v_c_minc_s_maxa, 19, 18, 4, up(maxa)).
table(up_v_c_minc_mins_maxa, 19, 11, 4, up(maxa)).
table(up_v_c_minc_maxs_maxa, 19, 8, 4, up(maxa)).
table(up_v_c_maxc_s_maxa, 19, 18, 4, up(maxa)).
table(up_v_c_maxc_mins_maxa, 19, 11, 4, up(maxa)).
table(up_v_c_maxc_maxs_maxa, 19, 6, 4, up(maxa)).
table(up_v_c_s_mins_maxa, 19, 19, 4, up(maxa)).
table(up_v_c_s_maxs_maxa, 19, 19, 4, up(maxa)).
table(up_v_c_mins_maxs_maxa, 19, 9, 4, up(maxa)).
table(up_v_minc_maxc_s_maxa, 19, 18, 4, up(maxa)).
table(up_v_minc_maxc_mins_maxa, 19, 14, 4, up(maxa)).
table(up_v_minc_maxc_maxs_maxa, 19, 9, 4, up(maxa)).
table(up_v_minc_s_mins_maxa, 19, 18, 4, up(maxa)).
table(up_v_minc_s_maxs_maxa, 19, 18, 4, up(maxa)).
table(up_v_minc_mins_maxs_maxa, 19, 18, 4, up(maxa)).
table(up_v_maxc_s_mins_maxa, 19, 8, 4, up(maxa)).
table(up_v_maxc_s_maxs_maxa, 19, 7, 4, up(maxa)).
table(up_v_maxc_mins_maxs_maxa, 19, 5, 4, up(maxa)).
table(up_v_s_mins_maxs_maxa, 19, 19, 4, up(maxa)).
table(up_c_minc_maxc_s_maxa, 19, 18, 4, up(maxa)).
table(up_c_minc_maxc_mins_maxa, 19, 14, 4, up(maxa)).
table(up_c_minc_maxc_maxs_maxa, 19, 9, 4, up(maxa)).
table(up_c_minc_s_maxs_maxa, 19, 9, 4, up(maxa)).
table(up_c_maxc_s_mins_maxa, 19, 8, 4, up(maxa)).
table(up_c_maxc_s_maxs_maxa, 19, 6, 4, up(maxa)).
table(up_c_maxc_mins_maxs_maxa, 19, 6, 4, up(maxa)).
table(up_c_s_mins_maxs_maxa, 19, 19, 4, up(maxa)).
table(up_minc_maxc_s_mins_maxa, 19, 6, 4, up(maxa)).
table(up_minc_maxc_s_maxs_maxa, 19, 6, 4, up(maxa)).
table(up_minc_s_mins_maxs_maxa, 19, 18, 4, up(maxa)).
table(up_maxc_s_mins_maxs_maxa, 19, 5, 4, up(maxa)).
table(up_v_c_minc_maxc_s_maxa, 19, 18, 5, up(maxa)).
table(up_v_c_minc_maxc_mins_maxa, 19, 14, 5, up(maxa)).
table(up_v_c_minc_maxc_maxs_maxa, 19, 9, 5, up(maxa)).
table(up_v_c_minc_s_mins_maxa, 19, 10, 5, up(maxa)).
table(up_v_c_minc_s_maxs_maxa, 19, 9, 5, up(maxa)).
table(up_v_c_minc_mins_maxs_maxa, 19, 8, 5, up(maxa)).
table(up_v_c_maxc_s_mins_maxa, 19, 8, 5, up(maxa)).
table(up_v_c_maxc_s_maxs_maxa, 19, 6, 5, up(maxa)).
table(up_v_c_maxc_mins_maxs_maxa, 19, 6, 5, up(maxa)).
table(up_v_c_s_mins_maxs_maxa, 19, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_mins_maxa, 19, 11, 5, up(maxa)).
table(up_v_minc_maxc_s_maxs_maxa, 19, 10, 5, up(maxa)).
table(up_v_minc_maxc_mins_maxs_maxa, 19, 7, 5, up(maxa)).
table(up_v_minc_s_mins_maxs_maxa, 19, 18, 5, up(maxa)).
table(up_v_maxc_s_mins_maxs_maxa, 19, 8, 5, up(maxa)).
table(up_c_minc_maxc_s_mins_maxa, 19, 11, 5, up(maxa)).
table(up_c_minc_maxc_s_maxs_maxa, 19, 9, 5, up(maxa)).
table(up_c_minc_maxc_mins_maxs_maxa, 19, 9, 5, up(maxa)).
table(up_c_minc_s_mins_maxs_maxa, 19, 10, 5, up(maxa)).
table(up_c_maxc_s_mins_maxs_maxa, 19, 7, 5, up(maxa)).
table(up_minc_maxc_s_mins_maxs_maxa, 19, 7, 5, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxa, 19, 11, 6, up(maxa)).
table(up_v_c_minc_maxc_s_maxs_maxa, 19, 9, 6, up(maxa)).
table(up_v_c_minc_maxc_mins_maxs_maxa, 19, 9, 6, up(maxa)).
table(up_v_c_minc_s_mins_maxs_maxa, 19, 10, 6, up(maxa)).
table(up_v_c_maxc_s_mins_maxs_maxa, 19, 7, 6, up(maxa)).
table(up_v_minc_maxc_s_mins_maxs_maxa, 19, 11, 6, up(maxa)).
table(up_c_minc_maxc_s_mins_maxs_maxa, 19, 10, 6, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxs_maxa, 19, 10, 7, up(maxa)).
table(low_c_v, 20, 18, 1, low(v)).
table(low_minc_v, 20, 11, 1, low(v)).
table(low_maxc_v, 20, 11, 1, low(v)).
table(low_s_v, 20, 7, 1, low(v)).
table(low_mins_v, 20, 18, 1, low(v)).
table(low_maxs_v, 20, 18, 1, low(v)).
table(low_c_minc_v, 20, 11, 2, low(v)).
table(low_c_maxc_v, 20, 11, 2, low(v)).
table(low_c_s_v, 20, 8, 2, low(v)).
table(low_c_mins_v, 20, 18, 2, low(v)).
table(low_c_maxs_v, 20, 18, 2, low(v)).
table(low_minc_maxc_v, 20, 11, 2, low(v)).
table(low_minc_s_v, 20, 3, 2, low(v)).
table(low_minc_mins_v, 20, 12, 2, low(v)).
table(low_minc_maxs_v, 20, 12, 2, low(v)).
table(low_maxc_s_v, 20, 3, 2, low(v)).
table(low_maxc_mins_v, 20, 12, 2, low(v)).
table(low_maxc_maxs_v, 20, 13, 2, low(v)).
table(low_s_mins_v, 20, 7, 2, low(v)).
table(low_s_maxs_v, 20, 7, 2, low(v)).
table(low_mins_maxs_v, 20, 12, 2, low(v)).
table(low_c_minc_maxc_v, 20, 11, 3, low(v)).
table(low_c_minc_s_v, 20, 6, 3, low(v)).
table(low_c_minc_mins_v, 20, 12, 3, low(v)).
table(low_c_minc_maxs_v, 20, 12, 3, low(v)).
table(low_c_maxc_s_v, 20, 4, 3, low(v)).
table(low_c_maxc_mins_v, 20, 12, 3, low(v)).
table(low_c_maxc_maxs_v, 20, 13, 3, low(v)).
table(low_c_s_mins_v, 20, 8, 3, low(v)).
table(low_c_s_maxs_v, 20, 8, 3, low(v)).
table(low_c_mins_maxs_v, 20, 18, 3, low(v)).
table(low_minc_maxc_s_v, 20, 5, 3, low(v)).
table(low_minc_maxc_mins_v, 20, 12, 3, low(v)).
table(low_minc_maxc_maxs_v, 20, 12, 3, low(v)).
table(low_minc_s_mins_v, 20, 4, 3, low(v)).
table(low_minc_s_maxs_v, 20, 4, 3, low(v)).
table(low_minc_mins_maxs_v, 20, 13, 3, low(v)).
table(low_maxc_s_mins_v, 20, 4, 3, low(v)).
table(low_maxc_s_maxs_v, 20, 4, 3, low(v)).
table(low_maxc_mins_maxs_v, 20, 14, 3, low(v)).
table(low_s_mins_maxs_v, 20, 7, 3, low(v)).
table(low_c_minc_maxc_s_v, 20, 7, 4, low(v)).
table(low_c_minc_maxc_mins_v, 20, 12, 4, low(v)).
table(low_c_minc_maxc_maxs_v, 20, 12, 4, low(v)).
table(low_c_minc_s_mins_v, 20, 7, 4, low(v)).
table(low_c_minc_s_maxs_v, 20, 7, 4, low(v)).
table(low_c_minc_mins_maxs_v, 20, 13, 4, low(v)).
table(low_c_maxc_s_mins_v, 20, 5, 4, low(v)).
table(low_c_maxc_s_maxs_v, 20, 5, 4, low(v)).
table(low_c_maxc_mins_maxs_v, 20, 14, 4, low(v)).
table(low_c_s_mins_maxs_v, 20, 8, 4, low(v)).
table(low_minc_maxc_s_mins_v, 20, 6, 4, low(v)).
table(low_minc_maxc_s_maxs_v, 20, 6, 4, low(v)).
table(low_minc_maxc_mins_maxs_v, 20, 13, 4, low(v)).
table(low_minc_s_mins_maxs_v, 20, 5, 4, low(v)).
table(low_maxc_s_mins_maxs_v, 20, 5, 4, low(v)).
table(low_c_minc_maxc_s_mins_v, 20, 8, 5, low(v)).
table(low_c_minc_maxc_s_maxs_v, 20, 8, 5, low(v)).
table(low_c_minc_maxc_mins_maxs_v, 20, 13, 5, low(v)).
table(low_c_minc_s_mins_maxs_v, 20, 8, 5, low(v)).
table(low_c_maxc_s_mins_maxs_v, 20, 6, 5, low(v)).
table(low_minc_maxc_s_mins_maxs_v, 20, 7, 5, low(v)).
table(low_c_minc_maxc_s_mins_maxs_v, 20, 9, 6, low(v)).
table(up_c_maxc_v, 20, 3, 2, up(v)).
table(up_maxc_s_v, 20, 3, 2, up(v)).
table(up_s_maxs_v, 20, 3, 2, up(v)).
table(up_c_minc_maxc_v, 20, 6, 3, up(v)).
table(up_c_maxc_s_v, 20, 4, 3, up(v)).
table(up_c_maxc_mins_v, 20, 4, 3, up(v)).
table(up_c_maxc_maxs_v, 20, 4, 3, up(v)).
table(up_c_s_maxs_v, 20, 4, 3, up(v)).
table(up_minc_maxc_s_v, 20, 5, 3, up(v)).
table(up_minc_s_maxs_v, 20, 4, 3, up(v)).
table(up_maxc_s_mins_v, 20, 4, 3, up(v)).
table(up_maxc_s_maxs_v, 20, 4, 3, up(v)).
table(up_s_mins_maxs_v, 20, 4, 3, up(v)).
table(up_c_minc_maxc_s_v, 20, 7, 4, up(v)).
table(up_c_minc_maxc_mins_v, 20, 7, 4, up(v)).
table(up_c_minc_maxc_maxs_v, 20, 7, 4, up(v)).
table(up_c_minc_s_maxs_v, 20, 7, 4, up(v)).
table(up_c_maxc_s_mins_v, 20, 5, 4, up(v)).
table(up_c_maxc_s_maxs_v, 20, 5, 4, up(v)).
table(up_c_maxc_mins_maxs_v, 20, 5, 4, up(v)).
table(up_c_s_mins_maxs_v, 20, 5, 4, up(v)).
table(up_minc_maxc_s_mins_v, 20, 6, 4, up(v)).
table(up_minc_maxc_s_maxs_v, 20, 6, 4, up(v)).
table(up_minc_s_mins_maxs_v, 20, 5, 4, up(v)).
table(up_maxc_s_mins_maxs_v, 20, 5, 4, up(v)).
table(up_c_minc_maxc_s_mins_v, 20, 8, 5, up(v)).
table(up_c_minc_maxc_s_maxs_v, 20, 8, 5, up(v)).
table(up_c_minc_maxc_mins_maxs_v, 20, 8, 5, up(v)).
table(up_c_minc_s_mins_maxs_v, 20, 8, 5, up(v)).
table(up_c_maxc_s_mins_maxs_v, 20, 6, 5, up(v)).
table(up_minc_maxc_s_mins_maxs_v, 20, 7, 5, up(v)).
table(up_c_minc_maxc_s_mins_maxs_v, 20, 9, 6, up(v)).
table(low_v_c, 20, 11, 1, low(c)).
table(low_minc_c, 20, 11, 1, low(c)).
table(low_maxc_c, 20, 11, 1, low(c)).
table(low_s_c, 20, 8, 1, low(c)).
table(low_mins_c, 20, 8, 1, low(c)).
table(low_maxs_c, 20, 8, 1, low(c)).
table(low_v_minc_c, 20, 11, 2, low(c)).
table(low_v_maxc_c, 20, 3, 2, low(c)).
table(low_v_s_c, 20, 13, 2, low(c)).
table(low_v_mins_c, 20, 12, 2, low(c)).
table(low_v_maxs_c, 20, 13, 2, low(c)).
table(low_minc_maxc_c, 20, 11, 2, low(c)).
table(low_minc_s_c, 20, 5, 2, low(c)).
table(low_minc_mins_c, 20, 9, 2, low(c)).
table(low_minc_maxs_c, 20, 9, 2, low(c)).
table(low_maxc_s_c, 20, 3, 2, low(c)).
table(low_maxc_mins_c, 20, 12, 2, low(c)).
table(low_maxc_maxs_c, 20, 13, 2, low(c)).
table(low_s_mins_c, 20, 10, 2, low(c)).
table(low_s_maxs_c, 20, 11, 2, low(c)).
table(low_mins_maxs_c, 20, 9, 2, low(c)).
table(low_v_minc_maxc_c, 20, 6, 3, low(c)).
table(low_v_minc_s_c, 20, 12, 3, low(c)).
table(low_v_minc_mins_c, 20, 12, 3, low(c)).
table(low_v_minc_maxs_c, 20, 12, 3, low(c)).
table(low_v_maxc_s_c, 20, 4, 3, low(c)).
table(low_v_maxc_mins_c, 20, 4, 3, low(c)).
table(low_v_maxc_maxs_c, 20, 4, 3, low(c)).
table(low_v_s_mins_c, 20, 14, 3, low(c)).
table(low_v_s_maxs_c, 20, 14, 3, low(c)).
table(low_v_mins_maxs_c, 20, 14, 3, low(c)).
table(low_minc_maxc_s_c, 20, 5, 3, low(c)).
table(low_minc_maxc_mins_c, 20, 12, 3, low(c)).
table(low_minc_maxc_maxs_c, 20, 12, 3, low(c)).
table(low_minc_s_mins_c, 20, 6, 3, low(c)).
table(low_minc_s_maxs_c, 20, 7, 3, low(c)).
table(low_minc_mins_maxs_c, 20, 7, 3, low(c)).
table(low_maxc_s_mins_c, 20, 4, 3, low(c)).
table(low_maxc_s_maxs_c, 20, 4, 3, low(c)).
table(low_maxc_mins_maxs_c, 20, 6, 3, low(c)).
table(low_s_mins_maxs_c, 20, 12, 3, low(c)).
table(low_v_minc_maxc_s_c, 20, 7, 4, low(c)).
table(low_v_minc_maxc_mins_c, 20, 7, 4, low(c)).
table(low_v_minc_maxc_maxs_c, 20, 7, 4, low(c)).
table(low_v_minc_s_mins_c, 20, 13, 4, low(c)).
table(low_v_minc_s_maxs_c, 20, 13, 4, low(c)).
table(low_v_minc_mins_maxs_c, 20, 13, 4, low(c)).
table(low_v_maxc_s_mins_c, 20, 5, 4, low(c)).
table(low_v_maxc_s_maxs_c, 20, 5, 4, low(c)).
table(low_v_maxc_mins_maxs_c, 20, 5, 4, low(c)).
table(low_v_s_mins_maxs_c, 20, 15, 4, low(c)).
table(low_minc_maxc_s_mins_c, 20, 6, 4, low(c)).
table(low_minc_maxc_s_maxs_c, 20, 6, 4, low(c)).
table(low_minc_maxc_mins_maxs_c, 20, 8, 4, low(c)).
table(low_minc_s_mins_maxs_c, 20, 8, 4, low(c)).
table(low_maxc_s_mins_maxs_c, 20, 5, 4, low(c)).
table(low_v_minc_maxc_s_mins_c, 20, 8, 5, low(c)).
table(low_v_minc_maxc_s_maxs_c, 20, 8, 5, low(c)).
table(low_v_minc_maxc_mins_maxs_c, 20, 8, 5, low(c)).
table(low_v_minc_s_mins_maxs_c, 20, 14, 5, low(c)).
table(low_v_maxc_s_mins_maxs_c, 20, 6, 5, low(c)).
table(low_minc_maxc_s_mins_maxs_c, 20, 7, 5, low(c)).
table(low_v_minc_maxc_s_mins_maxs_c, 20, 9, 6, low(c)).
table(up_v_c, 20, 18, 1, up(c)).
table(up_s_c, 20, 3, 1, up(c)).
table(up_v_minc_c, 20, 6, 2, up(c)).
table(up_v_maxc_c, 20, 11, 2, up(c)).
table(up_v_s_c, 20, 4, 2, up(c)).
table(up_v_mins_c, 20, 10, 2, up(c)).
table(up_v_maxs_c, 20, 18, 2, up(c)).
table(up_minc_s_c, 20, 3, 2, up(c)).
table(up_maxc_s_c, 20, 4, 2, up(c)).
table(up_s_mins_c, 20, 5, 2, up(c)).
table(up_s_maxs_c, 20, 5, 2, up(c)).
table(up_v_minc_maxc_c, 20, 7, 3, up(c)).
table(up_v_minc_s_c, 20, 6, 3, up(c)).
table(up_v_minc_mins_c, 20, 7, 3, up(c)).
table(up_v_minc_maxs_c, 20, 7, 3, up(c)).
table(up_v_maxc_s_c, 20, 5, 3, up(c)).
table(up_v_maxc_mins_c, 20, 5, 3, up(c)).
table(up_v_maxc_maxs_c, 20, 13, 3, up(c)).
table(up_v_s_mins_c, 20, 8, 3, up(c)).
table(up_v_s_maxs_c, 20, 6, 3, up(c)).
table(up_v_mins_maxs_c, 20, 12, 3, up(c)).
table(up_minc_maxc_s_c, 20, 5, 3, up(c)).
table(up_minc_s_mins_c, 20, 4, 3, up(c)).
table(up_minc_s_maxs_c, 20, 4, 3, up(c)).
table(up_maxc_s_mins_c, 20, 5, 3, up(c)).
table(up_maxc_s_maxs_c, 20, 5, 3, up(c)).
table(up_s_mins_maxs_c, 20, 8, 3, up(c)).
table(up_v_minc_maxc_s_c, 20, 7, 4, up(c)).
table(up_v_minc_maxc_mins_c, 20, 8, 4, up(c)).
table(up_v_minc_maxc_maxs_c, 20, 8, 4, up(c)).
table(up_v_minc_s_mins_c, 20, 7, 4, up(c)).
table(up_v_minc_s_maxs_c, 20, 7, 4, up(c)).
table(up_v_minc_mins_maxs_c, 20, 8, 4, up(c)).
table(up_v_maxc_s_mins_c, 20, 6, 4, up(c)).
table(up_v_maxc_s_maxs_c, 20, 6, 4, up(c)).
table(up_v_maxc_mins_maxs_c, 20, 7, 4, up(c)).
table(up_v_s_mins_maxs_c, 20, 10, 4, up(c)).
table(up_minc_maxc_s_mins_c, 20, 6, 4, up(c)).
table(up_minc_maxc_s_maxs_c, 20, 6, 4, up(c)).
table(up_minc_s_mins_maxs_c, 20, 5, 4, up(c)).
table(up_maxc_s_mins_maxs_c, 20, 6, 4, up(c)).
table(up_v_minc_maxc_s_mins_c, 20, 8, 5, up(c)).
table(up_v_minc_maxc_s_maxs_c, 20, 8, 5, up(c)).
table(up_v_minc_maxc_mins_maxs_c, 20, 9, 5, up(c)).
table(up_v_minc_s_mins_maxs_c, 20, 8, 5, up(c)).
table(up_v_maxc_s_mins_maxs_c, 20, 7, 5, up(c)).
table(up_minc_maxc_s_mins_maxs_c, 20, 7, 5, up(c)).
table(up_v_minc_maxc_s_mins_maxs_c, 20, 9, 6, up(c)).
table(low_v_minc, 20, 3, 1, low(minc)).
table(low_c_minc, 20, 3, 1, low(minc)).
table(low_maxc_minc, 20, 3, 1, low(minc)).
table(low_s_minc, 20, 3, 1, low(minc)).
table(low_mins_minc, 20, 2, 1, low(minc)).
table(low_maxs_minc, 20, 3, 1, low(minc)).
table(low_v_c_minc, 20, 5, 2, low(minc)).
table(low_v_maxc_minc, 20, 4, 2, low(minc)).
table(low_v_s_minc, 20, 4, 2, low(minc)).
table(low_v_mins_minc, 20, 3, 2, low(minc)).
table(low_v_maxs_minc, 20, 4, 2, low(minc)).
table(low_c_maxc_minc, 20, 4, 2, low(minc)).
table(low_c_s_minc, 20, 4, 2, low(minc)).
table(low_c_mins_minc, 20, 3, 2, low(minc)).
table(low_c_maxs_minc, 20, 4, 2, low(minc)).
table(low_maxc_s_minc, 20, 4, 2, low(minc)).
table(low_maxc_mins_minc, 20, 4, 2, low(minc)).
table(low_maxc_maxs_minc, 20, 4, 2, low(minc)).
table(low_s_mins_minc, 20, 3, 2, low(minc)).
table(low_s_maxs_minc, 20, 4, 2, low(minc)).
table(low_mins_maxs_minc, 20, 3, 2, low(minc)).
table(low_v_c_maxc_minc, 20, 6, 3, low(minc)).
table(low_v_c_s_minc, 20, 6, 3, low(minc)).
table(low_v_c_mins_minc, 20, 6, 3, low(minc)).
table(low_v_c_maxs_minc, 20, 6, 3, low(minc)).
table(low_v_maxc_s_minc, 20, 5, 3, low(minc)).
table(low_v_maxc_mins_minc, 20, 5, 3, low(minc)).
table(low_v_maxc_maxs_minc, 20, 5, 3, low(minc)).
table(low_v_s_mins_minc, 20, 4, 3, low(minc)).
table(low_v_s_maxs_minc, 20, 5, 3, low(minc)).
table(low_v_mins_maxs_minc, 20, 4, 3, low(minc)).
table(low_c_maxc_s_minc, 20, 5, 3, low(minc)).
table(low_c_maxc_mins_minc, 20, 5, 3, low(minc)).
table(low_c_maxc_maxs_minc, 20, 5, 3, low(minc)).
table(low_c_s_mins_minc, 20, 4, 3, low(minc)).
table(low_c_s_maxs_minc, 20, 6, 3, low(minc)).
table(low_c_mins_maxs_minc, 20, 4, 3, low(minc)).
table(low_maxc_s_mins_minc, 20, 5, 3, low(minc)).
table(low_maxc_s_maxs_minc, 20, 5, 3, low(minc)).
table(low_maxc_mins_maxs_minc, 20, 5, 3, low(minc)).
table(low_s_mins_maxs_minc, 20, 4, 3, low(minc)).
table(low_v_c_maxc_s_minc, 20, 7, 4, low(minc)).
table(low_v_c_maxc_mins_minc, 20, 7, 4, low(minc)).
table(low_v_c_maxc_maxs_minc, 20, 7, 4, low(minc)).
table(low_v_c_s_mins_minc, 20, 7, 4, low(minc)).
table(low_v_c_s_maxs_minc, 20, 7, 4, low(minc)).
table(low_v_c_mins_maxs_minc, 20, 7, 4, low(minc)).
table(low_v_maxc_s_mins_minc, 20, 6, 4, low(minc)).
table(low_v_maxc_s_maxs_minc, 20, 6, 4, low(minc)).
table(low_v_maxc_mins_maxs_minc, 20, 6, 4, low(minc)).
table(low_v_s_mins_maxs_minc, 20, 5, 4, low(minc)).
table(low_c_maxc_s_mins_minc, 20, 6, 4, low(minc)).
table(low_c_maxc_s_maxs_minc, 20, 6, 4, low(minc)).
table(low_c_maxc_mins_maxs_minc, 20, 6, 4, low(minc)).
table(low_c_s_mins_maxs_minc, 20, 6, 4, low(minc)).
table(low_maxc_s_mins_maxs_minc, 20, 6, 4, low(minc)).
table(low_v_c_maxc_s_mins_minc, 20, 8, 5, low(minc)).
table(low_v_c_maxc_s_maxs_minc, 20, 8, 5, low(minc)).
table(low_v_c_maxc_mins_maxs_minc, 20, 8, 5, low(minc)).
table(low_v_c_s_mins_maxs_minc, 20, 8, 5, low(minc)).
table(low_v_maxc_s_mins_maxs_minc, 20, 7, 5, low(minc)).
table(low_c_maxc_s_mins_maxs_minc, 20, 7, 5, low(minc)).
table(low_v_c_maxc_s_mins_maxs_minc, 20, 9, 6, low(minc)).
table(up_v_minc, 20, 11, 1, up(minc)).
table(up_maxc_minc, 20, 6, 1, up(minc)).
table(up_v_c_minc, 20, 5, 2, up(minc)).
table(up_v_maxc_minc, 20, 4, 2, up(minc)).
table(up_v_s_minc, 20, 13, 2, up(minc)).
table(up_v_mins_minc, 20, 12, 2, up(minc)).
table(up_v_maxs_minc, 20, 13, 2, up(minc)).
table(up_c_maxc_minc, 20, 4, 2, up(minc)).
table(up_maxc_s_minc, 20, 4, 2, up(minc)).
table(up_maxc_mins_minc, 20, 4, 2, up(minc)).
table(up_maxc_maxs_minc, 20, 7, 2, up(minc)).
table(up_s_maxs_minc, 20, 14, 2, up(minc)).
table(up_v_c_maxc_minc, 20, 6, 3, up(minc)).
table(up_v_c_s_minc, 20, 6, 3, up(minc)).
table(up_v_c_mins_minc, 20, 6, 3, up(minc)).
table(up_v_c_maxs_minc, 20, 6, 3, up(minc)).
table(up_v_maxc_s_minc, 20, 5, 3, up(minc)).
table(up_v_maxc_mins_minc, 20, 5, 3, up(minc)).
table(up_v_maxc_maxs_minc, 20, 5, 3, up(minc)).
table(up_v_s_mins_minc, 20, 14, 3, up(minc)).
table(up_v_s_maxs_minc, 20, 14, 3, up(minc)).
table(up_v_mins_maxs_minc, 20, 14, 3, up(minc)).
table(up_c_maxc_s_minc, 20, 5, 3, up(minc)).
table(up_c_maxc_mins_minc, 20, 5, 3, up(minc)).
table(up_c_maxc_maxs_minc, 20, 5, 3, up(minc)).
table(up_c_s_maxs_minc, 20, 4, 3, up(minc)).
table(up_maxc_s_mins_minc, 20, 5, 3, up(minc)).
table(up_maxc_s_maxs_minc, 20, 5, 3, up(minc)).
table(up_maxc_mins_maxs_minc, 20, 5, 3, up(minc)).
table(up_s_mins_maxs_minc, 20, 15, 3, up(minc)).
table(up_v_c_maxc_s_minc, 20, 7, 4, up(minc)).
table(up_v_c_maxc_mins_minc, 20, 7, 4, up(minc)).
table(up_v_c_maxc_maxs_minc, 20, 7, 4, up(minc)).
table(up_v_c_s_mins_minc, 20, 7, 4, up(minc)).
table(up_v_c_s_maxs_minc, 20, 7, 4, up(minc)).
table(up_v_c_mins_maxs_minc, 20, 7, 4, up(minc)).
table(up_v_maxc_s_mins_minc, 20, 6, 4, up(minc)).
table(up_v_maxc_s_maxs_minc, 20, 6, 4, up(minc)).
table(up_v_maxc_mins_maxs_minc, 20, 6, 4, up(minc)).
table(up_v_s_mins_maxs_minc, 20, 15, 4, up(minc)).
table(up_c_maxc_s_mins_minc, 20, 6, 4, up(minc)).
table(up_c_maxc_s_maxs_minc, 20, 6, 4, up(minc)).
table(up_c_maxc_mins_maxs_minc, 20, 6, 4, up(minc)).
table(up_c_s_mins_maxs_minc, 20, 5, 4, up(minc)).
table(up_maxc_s_mins_maxs_minc, 20, 6, 4, up(minc)).
table(up_v_c_maxc_s_mins_minc, 20, 8, 5, up(minc)).
table(up_v_c_maxc_s_maxs_minc, 20, 8, 5, up(minc)).
table(up_v_c_maxc_mins_maxs_minc, 20, 8, 5, up(minc)).
table(up_v_c_s_mins_maxs_minc, 20, 8, 5, up(minc)).
table(up_v_maxc_s_mins_maxs_minc, 20, 7, 5, up(minc)).
table(up_c_maxc_s_mins_maxs_minc, 20, 7, 5, up(minc)).
table(up_v_c_maxc_s_mins_maxs_minc, 20, 9, 6, up(minc)).
table(low_v_maxc, 20, 18, 1, low(maxc)).
table(low_c_maxc, 20, 18, 1, low(maxc)).
table(low_minc_maxc, 20, 6, 1, low(maxc)).
table(low_s_maxc, 20, 18, 1, low(maxc)).
table(low_mins_maxc, 20, 9, 1, low(maxc)).
table(low_maxs_maxc, 20, 2, 1, low(maxc)).
table(low_v_c_maxc, 20, 3, 2, low(maxc)).
table(low_v_minc_maxc, 20, 6, 2, low(maxc)).
table(low_v_s_maxc, 20, 4, 2, low(maxc)).
table(low_v_mins_maxc, 20, 9, 2, low(maxc)).
table(low_v_maxs_maxc, 20, 3, 2, low(maxc)).
table(low_c_minc_maxc, 20, 11, 2, low(maxc)).
table(low_c_s_maxc, 20, 4, 2, low(maxc)).
table(low_c_mins_maxc, 20, 18, 2, low(maxc)).
table(low_c_maxs_maxc, 20, 3, 2, low(maxc)).
table(low_minc_s_maxc, 20, 6, 2, low(maxc)).
table(low_minc_mins_maxc, 20, 4, 2, low(maxc)).
table(low_minc_maxs_maxc, 20, 4, 2, low(maxc)).
table(low_s_mins_maxc, 20, 18, 2, low(maxc)).
table(low_s_maxs_maxc, 20, 3, 2, low(maxc)).
table(low_mins_maxs_maxc, 20, 3, 2, low(maxc)).
table(low_v_c_minc_maxc, 20, 6, 3, low(maxc)).
table(low_v_c_s_maxc, 20, 4, 3, low(maxc)).
table(low_v_c_mins_maxc, 20, 4, 3, low(maxc)).
table(low_v_c_maxs_maxc, 20, 4, 3, low(maxc)).
table(low_v_minc_s_maxc, 20, 5, 3, low(maxc)).
table(low_v_minc_mins_maxc, 20, 5, 3, low(maxc)).
table(low_v_minc_maxs_maxc, 20, 5, 3, low(maxc)).
table(low_v_s_mins_maxc, 20, 5, 3, low(maxc)).
table(low_v_s_maxs_maxc, 20, 4, 3, low(maxc)).
table(low_v_mins_maxs_maxc, 20, 4, 3, low(maxc)).
table(low_c_minc_s_maxc, 20, 5, 3, low(maxc)).
table(low_c_minc_mins_maxc, 20, 6, 3, low(maxc)).
table(low_c_minc_maxs_maxc, 20, 5, 3, low(maxc)).
table(low_c_s_mins_maxc, 20, 4, 3, low(maxc)).
table(low_c_s_maxs_maxc, 20, 4, 3, low(maxc)).
table(low_c_mins_maxs_maxc, 20, 4, 3, low(maxc)).
table(low_minc_s_mins_maxc, 20, 5, 3, low(maxc)).
table(low_minc_s_maxs_maxc, 20, 5, 3, low(maxc)).
table(low_minc_mins_maxs_maxc, 20, 5, 3, low(maxc)).
table(low_s_mins_maxs_maxc, 20, 4, 3, low(maxc)).
table(low_v_c_minc_s_maxc, 20, 7, 4, low(maxc)).
table(low_v_c_minc_mins_maxc, 20, 7, 4, low(maxc)).
table(low_v_c_minc_maxs_maxc, 20, 7, 4, low(maxc)).
table(low_v_c_s_mins_maxc, 20, 5, 4, low(maxc)).
table(low_v_c_s_maxs_maxc, 20, 5, 4, low(maxc)).
table(low_v_c_mins_maxs_maxc, 20, 5, 4, low(maxc)).
table(low_v_minc_s_mins_maxc, 20, 6, 4, low(maxc)).
table(low_v_minc_s_maxs_maxc, 20, 6, 4, low(maxc)).
table(low_v_minc_mins_maxs_maxc, 20, 6, 4, low(maxc)).
table(low_v_s_mins_maxs_maxc, 20, 5, 4, low(maxc)).
table(low_c_minc_s_mins_maxc, 20, 6, 4, low(maxc)).
table(low_c_minc_s_maxs_maxc, 20, 6, 4, low(maxc)).
table(low_c_minc_mins_maxs_maxc, 20, 6, 4, low(maxc)).
table(low_c_s_mins_maxs_maxc, 20, 5, 4, low(maxc)).
table(low_minc_s_mins_maxs_maxc, 20, 6, 4, low(maxc)).
table(low_v_c_minc_s_mins_maxc, 20, 8, 5, low(maxc)).
table(low_v_c_minc_s_maxs_maxc, 20, 8, 5, low(maxc)).
table(low_v_c_minc_mins_maxs_maxc, 20, 8, 5, low(maxc)).
table(low_v_c_s_mins_maxs_maxc, 20, 6, 5, low(maxc)).
table(low_v_minc_s_mins_maxs_maxc, 20, 7, 5, low(maxc)).
table(low_c_minc_s_mins_maxs_maxc, 20, 7, 5, low(maxc)).
table(low_v_c_minc_s_mins_maxs_maxc, 20, 9, 6, low(maxc)).
table(up_v_maxc, 20, 11, 1, up(maxc)).
table(up_v_c_maxc, 20, 11, 2, up(maxc)).
table(up_v_minc_maxc, 20, 11, 2, up(maxc)).
table(up_v_s_maxc, 20, 13, 2, up(maxc)).
table(up_v_mins_maxc, 20, 12, 2, up(maxc)).
table(up_v_maxs_maxc, 20, 13, 2, up(maxc)).
table(up_v_c_minc_maxc, 20, 11, 3, up(maxc)).
table(up_v_c_s_maxc, 20, 13, 3, up(maxc)).
table(up_v_c_mins_maxc, 20, 12, 3, up(maxc)).
table(up_v_c_maxs_maxc, 20, 13, 3, up(maxc)).
table(up_v_minc_s_maxc, 20, 12, 3, up(maxc)).
table(up_v_minc_mins_maxc, 20, 12, 3, up(maxc)).
table(up_v_minc_maxs_maxc, 20, 12, 3, up(maxc)).
table(up_v_s_mins_maxc, 20, 14, 3, up(maxc)).
table(up_v_s_maxs_maxc, 20, 14, 3, up(maxc)).
table(up_v_mins_maxs_maxc, 20, 14, 3, up(maxc)).
table(up_v_c_minc_s_maxc, 20, 12, 4, up(maxc)).
table(up_v_c_minc_mins_maxc, 20, 12, 4, up(maxc)).
table(up_v_c_minc_maxs_maxc, 20, 12, 4, up(maxc)).
table(up_v_c_s_mins_maxc, 20, 14, 4, up(maxc)).
table(up_v_c_s_maxs_maxc, 20, 6, 4, up(maxc)).
table(up_v_c_mins_maxs_maxc, 20, 10, 4, up(maxc)).
table(up_v_minc_s_mins_maxc, 20, 13, 4, up(maxc)).
table(up_v_minc_s_maxs_maxc, 20, 13, 4, up(maxc)).
table(up_v_minc_mins_maxs_maxc, 20, 13, 4, up(maxc)).
table(up_v_s_mins_maxs_maxc, 20, 15, 4, up(maxc)).
table(up_v_c_minc_s_mins_maxc, 20, 13, 5, up(maxc)).
table(up_v_c_minc_s_maxs_maxc, 20, 8, 5, up(maxc)).
table(up_v_c_minc_mins_maxs_maxc, 20, 9, 5, up(maxc)).
table(up_v_c_s_mins_maxs_maxc, 20, 10, 5, up(maxc)).
table(up_v_minc_s_mins_maxs_maxc, 20, 14, 5, up(maxc)).
table(up_v_c_minc_s_mins_maxs_maxc, 20, 9, 6, up(maxc)).
table(low_v_s, 20, 18, 1, low(s)).
table(low_c_s, 20, 3, 1, low(s)).
table(low_minc_s, 20, 18, 1, low(s)).
table(low_maxc_s, 20, 18, 1, low(s)).
table(low_mins_s, 20, 18, 1, low(s)).
table(low_maxs_s, 20, 18, 1, low(s)).
table(low_v_c_s, 20, 4, 2, low(s)).
table(low_v_minc_s, 20, 18, 2, low(s)).
table(low_v_maxc_s, 20, 6, 2, low(s)).
table(low_v_mins_s, 20, 9, 2, low(s)).
table(low_v_maxs_s, 20, 3, 2, low(s)).
table(low_c_minc_s, 20, 5, 2, low(s)).
table(low_c_maxc_s, 20, 5, 2, low(s)).
table(low_c_mins_s, 20, 5, 2, low(s)).
table(low_c_maxs_s, 20, 5, 2, low(s)).
table(low_minc_maxc_s, 20, 18, 2, low(s)).
table(low_minc_mins_s, 20, 12, 2, low(s)).
table(low_minc_maxs_s, 20, 14, 2, low(s)).
table(low_maxc_mins_s, 20, 7, 2, low(s)).
table(low_maxc_maxs_s, 20, 14, 2, low(s)).
table(low_mins_maxs_s, 20, 12, 2, low(s)).
table(low_v_c_minc_s, 20, 8, 3, low(s)).
table(low_v_c_maxc_s, 20, 6, 3, low(s)).
table(low_v_c_mins_s, 20, 8, 3, low(s)).
table(low_v_c_maxs_s, 20, 4, 3, low(s)).
table(low_v_minc_maxc_s, 20, 10, 3, low(s)).
table(low_v_minc_mins_s, 20, 13, 3, low(s)).
table(low_v_minc_maxs_s, 20, 4, 3, low(s)).
table(low_v_maxc_mins_s, 20, 4, 3, low(s)).
table(low_v_maxc_maxs_s, 20, 4, 3, low(s)).
table(low_v_mins_maxs_s, 20, 4, 3, low(s)).
table(low_c_minc_maxc_s, 20, 8, 3, low(s)).
table(low_c_minc_mins_s, 20, 4, 3, low(s)).
table(low_c_minc_maxs_s, 20, 5, 3, low(s)).
table(low_c_maxc_mins_s, 20, 8, 3, low(s)).
table(low_c_maxc_maxs_s, 20, 5, 3, low(s)).
table(low_c_mins_maxs_s, 20, 8, 3, low(s)).
table(low_minc_maxc_mins_s, 20, 13, 3, low(s)).
table(low_minc_maxc_maxs_s, 20, 14, 3, low(s)).
table(low_minc_mins_maxs_s, 20, 8, 3, low(s)).
table(low_maxc_mins_maxs_s, 20, 9, 3, low(s)).
table(low_v_c_minc_maxc_s, 20, 10, 4, low(s)).
table(low_v_c_minc_mins_s, 20, 7, 4, low(s)).
table(low_v_c_minc_maxs_s, 20, 7, 4, low(s)).
table(low_v_c_maxc_mins_s, 20, 5, 4, low(s)).
table(low_v_c_maxc_maxs_s, 20, 5, 4, low(s)).
table(low_v_c_mins_maxs_s, 20, 5, 4, low(s)).
table(low_v_minc_maxc_mins_s, 20, 8, 4, low(s)).
table(low_v_minc_maxc_maxs_s, 20, 6, 4, low(s)).
table(low_v_minc_mins_maxs_s, 20, 5, 4, low(s)).
table(low_v_maxc_mins_maxs_s, 20, 5, 4, low(s)).
table(low_c_minc_maxc_mins_s, 20, 6, 4, low(s)).
table(low_c_minc_maxc_maxs_s, 20, 7, 4, low(s)).
table(low_c_minc_mins_maxs_s, 20, 5, 4, low(s)).
table(low_c_maxc_mins_maxs_s, 20, 6, 4, low(s)).
table(low_minc_maxc_mins_maxs_s, 20, 9, 4, low(s)).
table(low_v_c_minc_maxc_mins_s, 20, 8, 5, low(s)).
table(low_v_c_minc_maxc_maxs_s, 20, 8, 5, low(s)).
table(low_v_c_minc_mins_maxs_s, 20, 8, 5, low(s)).
table(low_v_c_maxc_mins_maxs_s, 20, 6, 5, low(s)).
table(low_v_minc_maxc_mins_maxs_s, 20, 7, 5, low(s)).
table(low_c_minc_maxc_mins_maxs_s, 20, 7, 5, low(s)).
table(low_v_c_minc_maxc_mins_maxs_s, 20, 9, 6, low(s)).
table(up_v_s, 20, 7, 1, up(s)).
table(up_v_c_s, 20, 8, 2, up(s)).
table(up_v_minc_s, 20, 8, 2, up(s)).
table(up_v_maxc_s, 20, 8, 2, up(s)).
table(up_v_mins_s, 20, 4, 2, up(s)).
table(up_v_maxs_s, 20, 7, 2, up(s)).
table(up_c_maxc_s, 20, 9, 2, up(s)).
table(up_v_c_minc_s, 20, 11, 3, up(s)).
table(up_v_c_maxc_s, 20, 9, 3, up(s)).
table(up_v_c_mins_s, 20, 5, 3, up(s)).
table(up_v_c_maxs_s, 20, 8, 3, up(s)).
table(up_v_minc_maxc_s, 20, 10, 3, up(s)).
table(up_v_minc_mins_s, 20, 5, 3, up(s)).
table(up_v_minc_maxs_s, 20, 8, 3, up(s)).
table(up_v_maxc_mins_s, 20, 5, 3, up(s)).
table(up_v_maxc_maxs_s, 20, 8, 3, up(s)).
table(up_v_mins_maxs_s, 20, 5, 3, up(s)).
table(up_c_minc_maxc_s, 20, 12, 3, up(s)).
table(up_c_maxc_mins_s, 20, 5, 3, up(s)).
table(up_c_maxc_maxs_s, 20, 9, 3, up(s)).
table(up_v_c_minc_maxc_s, 20, 12, 4, up(s)).
table(up_v_c_minc_mins_s, 20, 8, 4, up(s)).
table(up_v_c_minc_maxs_s, 20, 11, 4, up(s)).
table(up_v_c_maxc_mins_s, 20, 6, 4, up(s)).
table(up_v_c_maxc_maxs_s, 20, 9, 4, up(s)).
table(up_v_c_mins_maxs_s, 20, 6, 4, up(s)).
table(up_v_minc_maxc_mins_s, 20, 7, 4, up(s)).
table(up_v_minc_maxc_maxs_s, 20, 10, 4, up(s)).
table(up_v_minc_mins_maxs_s, 20, 6, 4, up(s)).
table(up_v_maxc_mins_maxs_s, 20, 6, 4, up(s)).
table(up_c_minc_maxc_mins_s, 20, 7, 4, up(s)).
table(up_c_minc_maxc_maxs_s, 20, 12, 4, up(s)).
table(up_c_maxc_mins_maxs_s, 20, 6, 4, up(s)).
table(up_v_c_minc_maxc_mins_s, 20, 9, 5, up(s)).
table(up_v_c_minc_maxc_maxs_s, 20, 12, 5, up(s)).
table(up_v_c_minc_mins_maxs_s, 20, 9, 5, up(s)).
table(up_v_c_maxc_mins_maxs_s, 20, 7, 5, up(s)).
table(up_v_minc_maxc_mins_maxs_s, 20, 8, 5, up(s)).
table(up_c_minc_maxc_mins_maxs_s, 20, 8, 5, up(s)).
table(up_v_c_minc_maxc_mins_maxs_s, 20, 10, 6, up(s)).
table(low_v_mins, 20, 2, 1, low(mins)).
table(low_c_mins, 20, 2, 1, low(mins)).
table(low_minc_mins, 20, 2, 1, low(mins)).
table(low_maxc_mins, 20, 2, 1, low(mins)).
table(low_s_mins, 20, 2, 1, low(mins)).
table(low_maxs_mins, 20, 2, 1, low(mins)).
table(low_v_c_mins, 20, 3, 2, low(mins)).
table(low_v_minc_mins, 20, 3, 2, low(mins)).
table(low_v_maxc_mins, 20, 3, 2, low(mins)).
table(low_v_s_mins, 20, 3, 2, low(mins)).
table(low_v_maxs_mins, 20, 3, 2, low(mins)).
table(low_c_minc_mins, 20, 3, 2, low(mins)).
table(low_c_maxc_mins, 20, 3, 2, low(mins)).
table(low_c_s_mins, 20, 3, 2, low(mins)).
table(low_c_maxs_mins, 20, 3, 2, low(mins)).
table(low_minc_maxc_mins, 20, 4, 2, low(mins)).
table(low_minc_s_mins, 20, 3, 2, low(mins)).
table(low_minc_maxs_mins, 20, 3, 2, low(mins)).
table(low_maxc_s_mins, 20, 3, 2, low(mins)).
table(low_maxc_maxs_mins, 20, 3, 2, low(mins)).
table(low_s_maxs_mins, 20, 3, 2, low(mins)).
table(low_v_c_minc_mins, 20, 6, 3, low(mins)).
table(low_v_c_maxc_mins, 20, 4, 3, low(mins)).
table(low_v_c_s_mins, 20, 4, 3, low(mins)).
table(low_v_c_maxs_mins, 20, 4, 3, low(mins)).
table(low_v_minc_maxc_mins, 20, 5, 3, low(mins)).
table(low_v_minc_s_mins, 20, 4, 3, low(mins)).
table(low_v_minc_maxs_mins, 20, 4, 3, low(mins)).
table(low_v_maxc_s_mins, 20, 4, 3, low(mins)).
table(low_v_maxc_maxs_mins, 20, 4, 3, low(mins)).
table(low_v_s_maxs_mins, 20, 4, 3, low(mins)).
table(low_c_minc_maxc_mins, 20, 5, 3, low(mins)).
table(low_c_minc_s_mins, 20, 4, 3, low(mins)).
table(low_c_minc_maxs_mins, 20, 4, 3, low(mins)).
table(low_c_maxc_s_mins, 20, 4, 3, low(mins)).
table(low_c_maxc_maxs_mins, 20, 4, 3, low(mins)).
table(low_c_s_maxs_mins, 20, 4, 3, low(mins)).
table(low_minc_maxc_s_mins, 20, 5, 3, low(mins)).
table(low_minc_maxc_maxs_mins, 20, 5, 3, low(mins)).
table(low_minc_s_maxs_mins, 20, 4, 3, low(mins)).
table(low_maxc_s_maxs_mins, 20, 4, 3, low(mins)).
table(low_v_c_minc_maxc_mins, 20, 7, 4, low(mins)).
table(low_v_c_minc_s_mins, 20, 7, 4, low(mins)).
table(low_v_c_minc_maxs_mins, 20, 7, 4, low(mins)).
table(low_v_c_maxc_s_mins, 20, 5, 4, low(mins)).
table(low_v_c_maxc_maxs_mins, 20, 5, 4, low(mins)).
table(low_v_c_s_maxs_mins, 20, 5, 4, low(mins)).
table(low_v_minc_maxc_s_mins, 20, 6, 4, low(mins)).
table(low_v_minc_maxc_maxs_mins, 20, 6, 4, low(mins)).
table(low_v_minc_s_maxs_mins, 20, 5, 4, low(mins)).
table(low_v_maxc_s_maxs_mins, 20, 5, 4, low(mins)).
table(low_c_minc_maxc_s_mins, 20, 6, 4, low(mins)).
table(low_c_minc_maxc_maxs_mins, 20, 6, 4, low(mins)).
table(low_c_minc_s_maxs_mins, 20, 5, 4, low(mins)).
table(low_c_maxc_s_maxs_mins, 20, 5, 4, low(mins)).
table(low_minc_maxc_s_maxs_mins, 20, 6, 4, low(mins)).
table(low_v_c_minc_maxc_s_mins, 20, 8, 5, low(mins)).
table(low_v_c_minc_maxc_maxs_mins, 20, 8, 5, low(mins)).
table(low_v_c_minc_s_maxs_mins, 20, 8, 5, low(mins)).
table(low_v_c_maxc_s_maxs_mins, 20, 6, 5, low(mins)).
table(low_v_minc_maxc_s_maxs_mins, 20, 7, 5, low(mins)).
table(low_c_minc_maxc_s_maxs_mins, 20, 7, 5, low(mins)).
table(low_v_c_minc_maxc_s_maxs_mins, 20, 9, 6, low(mins)).
table(up_v_mins, 20, 18, 1, up(mins)).
table(up_maxc_mins, 20, 9, 1, up(mins)).
table(up_maxs_mins, 20, 2, 1, up(mins)).
table(up_v_c_mins, 20, 6, 2, up(mins)).
table(up_v_minc_mins, 20, 3, 2, up(mins)).
table(up_v_maxc_mins, 20, 5, 2, up(mins)).
table(up_v_s_mins, 20, 3, 2, up(mins)).
table(up_v_maxs_mins, 20, 3, 2, up(mins)).
table(up_c_maxc_mins, 20, 5, 2, up(mins)).
table(up_c_maxs_mins, 20, 3, 2, up(mins)).
table(up_minc_maxc_mins, 20, 4, 2, up(mins)).
table(up_minc_maxs_mins, 20, 3, 2, up(mins)).
table(up_maxc_s_mins, 20, 3, 2, up(mins)).
table(up_maxc_maxs_mins, 20, 3, 2, up(mins)).
table(up_s_maxs_mins, 20, 3, 2, up(mins)).
table(up_v_c_minc_mins, 20, 6, 3, up(mins)).
table(up_v_c_maxc_mins, 20, 7, 3, up(mins)).
table(up_v_c_s_mins, 20, 4, 3, up(mins)).
table(up_v_c_maxs_mins, 20, 4, 3, up(mins)).
table(up_v_minc_maxc_mins, 20, 5, 3, up(mins)).
table(up_v_minc_s_mins, 20, 4, 3, up(mins)).
table(up_v_minc_maxs_mins, 20, 4, 3, up(mins)).
table(up_v_maxc_s_mins, 20, 4, 3, up(mins)).
table(up_v_maxc_maxs_mins, 20, 4, 3, up(mins)).
table(up_v_s_maxs_mins, 20, 4, 3, up(mins)).
table(up_c_minc_maxc_mins, 20, 5, 3, up(mins)).
table(up_c_minc_maxs_mins, 20, 4, 3, up(mins)).
table(up_c_maxc_s_mins, 20, 4, 3, up(mins)).
table(up_c_maxc_maxs_mins, 20, 4, 3, up(mins)).
table(up_c_s_maxs_mins, 20, 4, 3, up(mins)).
table(up_minc_maxc_s_mins, 20, 5, 3, up(mins)).
table(up_minc_maxc_maxs_mins, 20, 5, 3, up(mins)).
table(up_minc_s_maxs_mins, 20, 4, 3, up(mins)).
table(up_maxc_s_maxs_mins, 20, 4, 3, up(mins)).
table(up_v_c_minc_maxc_mins, 20, 7, 4, up(mins)).
table(up_v_c_minc_s_mins, 20, 7, 4, up(mins)).
table(up_v_c_minc_maxs_mins, 20, 7, 4, up(mins)).
table(up_v_c_maxc_s_mins, 20, 5, 4, up(mins)).
table(up_v_c_maxc_maxs_mins, 20, 5, 4, up(mins)).
table(up_v_c_s_maxs_mins, 20, 5, 4, up(mins)).
table(up_v_minc_maxc_s_mins, 20, 6, 4, up(mins)).
table(up_v_minc_maxc_maxs_mins, 20, 6, 4, up(mins)).
table(up_v_minc_s_maxs_mins, 20, 5, 4, up(mins)).
table(up_v_maxc_s_maxs_mins, 20, 5, 4, up(mins)).
table(up_c_minc_maxc_s_mins, 20, 6, 4, up(mins)).
table(up_c_minc_maxc_maxs_mins, 20, 6, 4, up(mins)).
table(up_c_minc_s_maxs_mins, 20, 5, 4, up(mins)).
table(up_c_maxc_s_maxs_mins, 20, 5, 4, up(mins)).
table(up_minc_maxc_s_maxs_mins, 20, 6, 4, up(mins)).
table(up_v_c_minc_maxc_s_mins, 20, 8, 5, up(mins)).
table(up_v_c_minc_maxc_maxs_mins, 20, 8, 5, up(mins)).
table(up_v_c_minc_s_maxs_mins, 20, 8, 5, up(mins)).
table(up_v_c_maxc_s_maxs_mins, 20, 6, 5, up(mins)).
table(up_v_minc_maxc_s_maxs_mins, 20, 7, 5, up(mins)).
table(up_c_minc_maxc_s_maxs_mins, 20, 7, 5, up(mins)).
table(up_v_c_minc_maxc_s_maxs_mins, 20, 9, 6, up(mins)).
table(low_v_maxs, 20, 7, 1, low(maxs)).
table(low_c_maxs, 20, 4, 1, low(maxs)).
table(low_minc_maxs, 20, 4, 1, low(maxs)).
table(low_maxc_maxs, 20, 4, 1, low(maxs)).
table(low_s_maxs, 20, 7, 1, low(maxs)).
table(low_mins_maxs, 20, 2, 1, low(maxs)).
table(low_v_c_maxs, 20, 8, 2, low(maxs)).
table(low_v_minc_maxs, 20, 8, 2, low(maxs)).
table(low_v_maxc_maxs, 20, 8, 2, low(maxs)).
table(low_v_s_maxs, 20, 3, 2, low(maxs)).
table(low_v_mins_maxs, 20, 3, 2, low(maxs)).
table(low_c_minc_maxs, 20, 5, 2, low(maxs)).
table(low_c_maxc_maxs, 20, 5, 2, low(maxs)).
table(low_c_s_maxs, 20, 8, 2, low(maxs)).
table(low_c_mins_maxs, 20, 3, 2, low(maxs)).
table(low_minc_maxc_maxs, 20, 6, 2, low(maxs)).
table(low_minc_s_maxs, 20, 3, 2, low(maxs)).
table(low_minc_mins_maxs, 20, 3, 2, low(maxs)).
table(low_maxc_s_maxs, 20, 3, 2, low(maxs)).
table(low_maxc_mins_maxs, 20, 3, 2, low(maxs)).
table(low_s_mins_maxs, 20, 7, 2, low(maxs)).
table(low_v_c_minc_maxs, 20, 11, 3, low(maxs)).
table(low_v_c_maxc_maxs, 20, 9, 3, low(maxs)).
table(low_v_c_s_maxs, 20, 4, 3, low(maxs)).
table(low_v_c_mins_maxs, 20, 4, 3, low(maxs)).
table(low_v_minc_maxc_maxs, 20, 10, 3, low(maxs)).
table(low_v_minc_s_maxs, 20, 4, 3, low(maxs)).
table(low_v_minc_mins_maxs, 20, 4, 3, low(maxs)).
table(low_v_maxc_s_maxs, 20, 4, 3, low(maxs)).
table(low_v_maxc_mins_maxs, 20, 4, 3, low(maxs)).
table(low_v_s_mins_maxs, 20, 4, 3, low(maxs)).
table(low_c_minc_maxc_maxs, 20, 7, 3, low(maxs)).
table(low_c_minc_s_maxs, 20, 4, 3, low(maxs)).
table(low_c_minc_mins_maxs, 20, 4, 3, low(maxs)).
table(low_c_maxc_s_maxs, 20, 4, 3, low(maxs)).
table(low_c_maxc_mins_maxs, 20, 4, 3, low(maxs)).
table(low_c_s_mins_maxs, 20, 8, 3, low(maxs)).
table(low_minc_maxc_s_maxs, 20, 5, 3, low(maxs)).
table(low_minc_maxc_mins_maxs, 20, 5, 3, low(maxs)).
table(low_minc_s_mins_maxs, 20, 4, 3, low(maxs)).
table(low_maxc_s_mins_maxs, 20, 4, 3, low(maxs)).
table(low_v_c_minc_maxc_maxs, 20, 12, 4, low(maxs)).
table(low_v_c_minc_s_maxs, 20, 7, 4, low(maxs)).
table(low_v_c_minc_mins_maxs, 20, 7, 4, low(maxs)).
table(low_v_c_maxc_s_maxs, 20, 5, 4, low(maxs)).
table(low_v_c_maxc_mins_maxs, 20, 5, 4, low(maxs)).
table(low_v_c_s_mins_maxs, 20, 5, 4, low(maxs)).
table(low_v_minc_maxc_s_maxs, 20, 6, 4, low(maxs)).
table(low_v_minc_maxc_mins_maxs, 20, 6, 4, low(maxs)).
table(low_v_minc_s_mins_maxs, 20, 5, 4, low(maxs)).
table(low_v_maxc_s_mins_maxs, 20, 5, 4, low(maxs)).
table(low_c_minc_maxc_s_maxs, 20, 6, 4, low(maxs)).
table(low_c_minc_maxc_mins_maxs, 20, 6, 4, low(maxs)).
table(low_c_minc_s_mins_maxs, 20, 5, 4, low(maxs)).
table(low_c_maxc_s_mins_maxs, 20, 5, 4, low(maxs)).
table(low_minc_maxc_s_mins_maxs, 20, 6, 4, low(maxs)).
table(low_v_c_minc_maxc_s_maxs, 20, 8, 5, low(maxs)).
table(low_v_c_minc_maxc_mins_maxs, 20, 8, 5, low(maxs)).
table(low_v_c_minc_s_mins_maxs, 20, 8, 5, low(maxs)).
table(low_v_c_maxc_s_mins_maxs, 20, 6, 5, low(maxs)).
table(low_v_minc_maxc_s_mins_maxs, 20, 7, 5, low(maxs)).
table(low_c_minc_maxc_s_mins_maxs, 20, 7, 5, low(maxs)).
table(low_v_c_minc_maxc_s_mins_maxs, 20, 9, 6, low(maxs)).
table(up_v_maxs, 20, 18, 1, up(maxs)).
table(up_maxc_maxs, 20, 2, 1, up(maxs)).
table(up_v_c_maxs, 20, 18, 2, up(maxs)).
table(up_v_minc_maxs, 20, 12, 2, up(maxs)).
table(up_v_maxc_maxs, 20, 3, 2, up(maxs)).
table(up_v_s_maxs, 20, 7, 2, up(maxs)).
table(up_v_mins_maxs, 20, 9, 2, up(maxs)).
table(up_c_maxc_maxs, 20, 3, 2, up(maxs)).
table(up_minc_maxc_maxs, 20, 4, 2, up(maxs)).
table(up_maxc_s_maxs, 20, 3, 2, up(maxs)).
table(up_maxc_mins_maxs, 20, 3, 2, up(maxs)).
table(up_v_c_minc_maxs, 20, 12, 3, up(maxs)).
table(up_v_c_maxc_maxs, 20, 4, 3, up(maxs)).
table(up_v_c_s_maxs, 20, 8, 3, up(maxs)).
table(up_v_c_mins_maxs, 20, 18, 3, up(maxs)).
table(up_v_minc_maxc_maxs, 20, 5, 3, up(maxs)).
table(up_v_minc_s_maxs, 20, 5, 3, up(maxs)).
table(up_v_minc_mins_maxs, 20, 14, 3, up(maxs)).
table(up_v_maxc_s_maxs, 20, 4, 3, up(maxs)).
table(up_v_maxc_mins_maxs, 20, 4, 3, up(maxs)).
table(up_v_s_mins_maxs, 20, 7, 3, up(maxs)).
table(up_c_minc_maxc_maxs, 20, 5, 3, up(maxs)).
table(up_c_maxc_s_maxs, 20, 4, 3, up(maxs)).
table(up_c_maxc_mins_maxs, 20, 4, 3, up(maxs)).
table(up_minc_maxc_s_maxs, 20, 5, 3, up(maxs)).
table(up_minc_maxc_mins_maxs, 20, 5, 3, up(maxs)).
table(up_maxc_s_mins_maxs, 20, 4, 3, up(maxs)).
table(up_v_c_minc_maxc_maxs, 20, 7, 4, up(maxs)).
table(up_v_c_minc_s_maxs, 20, 7, 4, up(maxs)).
table(up_v_c_minc_mins_maxs, 20, 13, 4, up(maxs)).
table(up_v_c_maxc_s_maxs, 20, 5, 4, up(maxs)).
table(up_v_c_maxc_mins_maxs, 20, 5, 4, up(maxs)).
table(up_v_c_s_mins_maxs, 20, 8, 4, up(maxs)).
table(up_v_minc_maxc_s_maxs, 20, 6, 4, up(maxs)).
table(up_v_minc_maxc_mins_maxs, 20, 6, 4, up(maxs)).
table(up_v_minc_s_mins_maxs, 20, 5, 4, up(maxs)).
table(up_v_maxc_s_mins_maxs, 20, 5, 4, up(maxs)).
table(up_c_minc_maxc_s_maxs, 20, 6, 4, up(maxs)).
table(up_c_minc_maxc_mins_maxs, 20, 6, 4, up(maxs)).
table(up_c_maxc_s_mins_maxs, 20, 5, 4, up(maxs)).
table(up_minc_maxc_s_mins_maxs, 20, 6, 4, up(maxs)).
table(up_v_c_minc_maxc_s_maxs, 20, 8, 5, up(maxs)).
table(up_v_c_minc_maxc_mins_maxs, 20, 8, 5, up(maxs)).
table(up_v_c_minc_s_mins_maxs, 20, 8, 5, up(maxs)).
table(up_v_c_maxc_s_mins_maxs, 20, 6, 5, up(maxs)).
table(up_v_minc_maxc_s_mins_maxs, 20, 7, 5, up(maxs)).
table(up_c_minc_maxc_s_mins_maxs, 20, 7, 5, up(maxs)).
table(up_v_c_minc_maxc_s_mins_maxs, 20, 9, 6, up(maxs)).
table(low_v_mina, 20, 12, 1, low(mina)).
table(low_c_mina, 20, 7, 1, low(mina)).
table(low_minc_mina, 20, 19, 1, low(mina)).
table(low_maxc_mina, 20, 19, 1, low(mina)).
table(low_s_mina, 20, 12, 1, low(mina)).
table(low_mins_mina, 20, 11, 1, low(mina)).
table(low_maxs_mina, 20, 11, 1, low(mina)).
table(low_v_c_mina, 20, 10, 2, low(mina)).
table(low_v_minc_mina, 20, 11, 2, low(mina)).
table(low_v_maxc_mina, 20, 11, 2, low(mina)).
table(low_v_s_mina, 20, 4, 2, low(mina)).
table(low_v_mins_mina, 20, 4, 2, low(mina)).
table(low_v_maxs_mina, 20, 10, 2, low(mina)).
table(low_c_minc_mina, 20, 8, 2, low(mina)).
table(low_c_maxc_mina, 20, 6, 2, low(mina)).
table(low_c_s_mina, 20, 10, 2, low(mina)).
table(low_c_mins_mina, 20, 7, 2, low(mina)).
table(low_c_maxs_mina, 20, 5, 2, low(mina)).
table(low_minc_maxc_mina, 20, 19, 2, low(mina)).
table(low_minc_s_mina, 20, 11, 2, low(mina)).
table(low_minc_mins_mina, 20, 19, 2, low(mina)).
table(low_minc_maxs_mina, 20, 19, 2, low(mina)).
table(low_maxc_s_mina, 20, 11, 2, low(mina)).
table(low_maxc_mins_mina, 20, 19, 2, low(mina)).
table(low_maxc_maxs_mina, 20, 19, 2, low(mina)).
table(low_s_mins_mina, 20, 12, 2, low(mina)).
table(low_s_maxs_mina, 20, 10, 2, low(mina)).
table(low_mins_maxs_mina, 20, 6, 2, low(mina)).
table(low_v_c_minc_mina, 20, 13, 3, low(mina)).
table(low_v_c_maxc_mina, 20, 11, 3, low(mina)).
table(low_v_c_s_mina, 20, 4, 3, low(mina)).
table(low_v_c_mins_mina, 20, 7, 3, low(mina)).
table(low_v_c_maxs_mina, 20, 10, 3, low(mina)).
table(low_v_minc_maxc_mina, 20, 12, 3, low(mina)).
table(low_v_minc_s_mina, 20, 4, 3, low(mina)).
table(low_v_minc_mins_mina, 20, 4, 3, low(mina)).
table(low_v_minc_maxs_mina, 20, 11, 3, low(mina)).
table(low_v_maxc_s_mina, 20, 4, 3, low(mina)).
table(low_v_maxc_mins_mina, 20, 6, 3, low(mina)).
table(low_v_maxc_maxs_mina, 20, 10, 3, low(mina)).
table(low_v_s_mins_mina, 20, 5, 3, low(mina)).
table(low_v_s_maxs_mina, 20, 5, 3, low(mina)).
table(low_v_mins_maxs_mina, 20, 5, 3, low(mina)).
table(low_c_minc_maxc_mina, 20, 8, 3, low(mina)).
table(low_c_minc_s_mina, 20, 11, 3, low(mina)).
table(low_c_minc_mins_mina, 20, 8, 3, low(mina)).
table(low_c_minc_maxs_mina, 20, 6, 3, low(mina)).
table(low_c_maxc_s_mina, 20, 8, 3, low(mina)).
table(low_c_maxc_mins_mina, 20, 6, 3, low(mina)).
table(low_c_maxc_maxs_mina, 20, 6, 3, low(mina)).
table(low_c_s_mins_mina, 20, 10, 3, low(mina)).
table(low_c_s_maxs_mina, 20, 10, 3, low(mina)).
table(low_c_mins_maxs_mina, 20, 5, 3, low(mina)).
table(low_minc_maxc_s_mina, 20, 10, 3, low(mina)).
table(low_minc_maxc_mins_mina, 20, 16, 3, low(mina)).
table(low_minc_maxc_maxs_mina, 20, 18, 3, low(mina)).
table(low_minc_s_mins_mina, 20, 6, 3, low(mina)).
table(low_minc_s_maxs_mina, 20, 8, 3, low(mina)).
table(low_minc_mins_maxs_mina, 20, 17, 3, low(mina)).
table(low_maxc_s_mins_mina, 20, 8, 3, low(mina)).
table(low_maxc_s_maxs_mina, 20, 7, 3, low(mina)).
table(low_maxc_mins_maxs_mina, 20, 8, 3, low(mina)).
table(low_s_mins_maxs_mina, 20, 10, 3, low(mina)).
table(low_v_c_minc_maxc_mina, 20, 14, 4, low(mina)).
table(low_v_c_minc_s_mina, 20, 8, 4, low(mina)).
table(low_v_c_minc_mins_mina, 20, 10, 4, low(mina)).
table(low_v_c_minc_maxs_mina, 20, 13, 4, low(mina)).
table(low_v_c_maxc_s_mina, 20, 5, 4, low(mina)).
table(low_v_c_maxc_mins_mina, 20, 8, 4, low(mina)).
table(low_v_c_maxc_maxs_mina, 20, 11, 4, low(mina)).
table(low_v_c_s_mins_mina, 20, 5, 4, low(mina)).
table(low_v_c_s_maxs_mina, 20, 5, 4, low(mina)).
table(low_v_c_mins_maxs_mina, 20, 8, 4, low(mina)).
table(low_v_minc_maxc_s_mina, 20, 6, 4, low(mina)).
table(low_v_minc_maxc_mins_mina, 20, 6, 4, low(mina)).
table(low_v_minc_maxc_maxs_mina, 20, 12, 4, low(mina)).
table(low_v_minc_s_mins_mina, 20, 5, 4, low(mina)).
table(low_v_minc_s_maxs_mina, 20, 5, 4, low(mina)).
table(low_v_minc_mins_maxs_mina, 20, 5, 4, low(mina)).
table(low_v_maxc_s_mins_mina, 20, 5, 4, low(mina)).
table(low_v_maxc_s_maxs_mina, 20, 5, 4, low(mina)).
table(low_v_maxc_mins_maxs_mina, 20, 5, 4, low(mina)).
table(low_v_s_mins_maxs_mina, 20, 6, 4, low(mina)).
table(low_c_minc_maxc_s_mina, 20, 9, 4, low(mina)).
table(low_c_minc_maxc_mins_mina, 20, 7, 4, low(mina)).
table(low_c_minc_maxc_maxs_mina, 20, 8, 4, low(mina)).
table(low_c_minc_s_mins_mina, 20, 10, 4, low(mina)).
table(low_c_minc_s_maxs_mina, 20, 10, 4, low(mina)).
table(low_c_minc_mins_maxs_mina, 20, 6, 4, low(mina)).
table(low_c_maxc_s_mins_mina, 20, 6, 4, low(mina)).
table(low_c_maxc_s_maxs_mina, 20, 6, 4, low(mina)).
table(low_c_maxc_mins_maxs_mina, 20, 6, 4, low(mina)).
table(low_c_s_mins_maxs_mina, 20, 10, 4, low(mina)).
table(low_minc_maxc_s_mins_mina, 20, 7, 4, low(mina)).
table(low_minc_maxc_s_maxs_mina, 20, 9, 4, low(mina)).
table(low_minc_maxc_mins_maxs_mina, 20, 16, 4, low(mina)).
table(low_minc_s_mins_maxs_mina, 20, 6, 4, low(mina)).
table(low_maxc_s_mins_maxs_mina, 20, 6, 4, low(mina)).
table(low_v_c_minc_maxc_s_mina, 20, 9, 5, low(mina)).
table(low_v_c_minc_maxc_mins_mina, 20, 11, 5, low(mina)).
table(low_v_c_minc_maxc_maxs_mina, 20, 14, 5, low(mina)).
table(low_v_c_minc_s_mins_mina, 20, 8, 5, low(mina)).
table(low_v_c_minc_s_maxs_mina, 20, 8, 5, low(mina)).
table(low_v_c_minc_mins_maxs_mina, 20, 11, 5, low(mina)).
table(low_v_c_maxc_s_mins_mina, 20, 6, 5, low(mina)).
table(low_v_c_maxc_s_maxs_mina, 20, 6, 5, low(mina)).
table(low_v_c_maxc_mins_maxs_mina, 20, 9, 5, low(mina)).
table(low_v_c_s_mins_maxs_mina, 20, 6, 5, low(mina)).
table(low_v_minc_maxc_s_mins_mina, 20, 7, 5, low(mina)).
table(low_v_minc_maxc_s_maxs_mina, 20, 7, 5, low(mina)).
table(low_v_minc_maxc_mins_maxs_mina, 20, 7, 5, low(mina)).
table(low_v_minc_s_mins_maxs_mina, 20, 6, 5, low(mina)).
table(low_v_maxc_s_mins_maxs_mina, 20, 6, 5, low(mina)).
table(low_c_minc_maxc_s_mins_mina, 20, 8, 5, low(mina)).
table(low_c_minc_maxc_s_maxs_mina, 20, 8, 5, low(mina)).
table(low_c_minc_maxc_mins_maxs_mina, 20, 8, 5, low(mina)).
table(low_c_minc_s_mins_maxs_mina, 20, 11, 5, low(mina)).
table(low_c_maxc_s_mins_maxs_mina, 20, 7, 5, low(mina)).
table(low_minc_maxc_s_mins_maxs_mina, 20, 8, 5, low(mina)).
table(low_v_c_minc_maxc_s_mins_mina, 20, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_maxs_mina, 20, 9, 6, low(mina)).
table(low_v_c_minc_maxc_mins_maxs_mina, 20, 12, 6, low(mina)).
table(low_v_c_minc_s_mins_maxs_mina, 20, 9, 6, low(mina)).
table(low_v_c_maxc_s_mins_maxs_mina, 20, 7, 6, low(mina)).
table(low_v_minc_maxc_s_mins_maxs_mina, 20, 8, 6, low(mina)).
table(low_c_minc_maxc_s_mins_maxs_mina, 20, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_mins_maxs_mina, 20, 10, 7, low(mina)).
table(up_v_maxa, 20, 19, 1, up(maxa)).
table(up_v_c_maxa, 20, 19, 2, up(maxa)).
table(up_v_minc_maxa, 20, 19, 2, up(maxa)).
table(up_v_maxc_maxa, 20, 19, 2, up(maxa)).
table(up_v_s_maxa, 20, 19, 2, up(maxa)).
table(up_v_mins_maxa, 20, 19, 2, up(maxa)).
table(up_v_maxs_maxa, 20, 19, 2, up(maxa)).
table(up_c_maxc_maxa, 20, 19, 2, up(maxa)).
table(up_maxc_s_maxa, 20, 18, 2, up(maxa)).
table(up_s_maxs_maxa, 20, 19, 2, up(maxa)).
table(up_v_c_minc_maxa, 20, 19, 3, up(maxa)).
table(up_v_c_maxc_maxa, 20, 19, 3, up(maxa)).
table(up_v_c_s_maxa, 20, 19, 3, up(maxa)).
table(up_v_c_mins_maxa, 20, 19, 3, up(maxa)).
table(up_v_c_maxs_maxa, 20, 7, 3, up(maxa)).
table(up_v_minc_maxc_maxa, 20, 19, 3, up(maxa)).
table(up_v_minc_s_maxa, 20, 18, 3, up(maxa)).
table(up_v_minc_mins_maxa, 20, 19, 3, up(maxa)).
table(up_v_minc_maxs_maxa, 20, 19, 3, up(maxa)).
table(up_v_maxc_s_maxa, 20, 18, 3, up(maxa)).
table(up_v_maxc_mins_maxa, 20, 11, 3, up(maxa)).
table(up_v_maxc_maxs_maxa, 20, 6, 3, up(maxa)).
table(up_v_s_mins_maxa, 20, 19, 3, up(maxa)).
table(up_v_s_maxs_maxa, 20, 19, 3, up(maxa)).
table(up_v_mins_maxs_maxa, 20, 19, 3, up(maxa)).
table(up_c_minc_maxc_maxa, 20, 19, 3, up(maxa)).
table(up_c_maxc_s_maxa, 20, 18, 3, up(maxa)).
table(up_c_maxc_mins_maxa, 20, 19, 3, up(maxa)).
table(up_c_maxc_maxs_maxa, 20, 6, 3, up(maxa)).
table(up_c_s_maxs_maxa, 20, 19, 3, up(maxa)).
table(up_minc_maxc_s_maxa, 20, 6, 3, up(maxa)).
table(up_minc_s_maxs_maxa, 20, 18, 3, up(maxa)).
table(up_maxc_s_mins_maxa, 20, 18, 3, up(maxa)).
table(up_maxc_s_maxs_maxa, 20, 5, 3, up(maxa)).
table(up_s_mins_maxs_maxa, 20, 19, 3, up(maxa)).
table(up_v_c_minc_maxc_maxa, 20, 19, 4, up(maxa)).
table(up_v_c_minc_s_maxa, 20, 18, 4, up(maxa)).
table(up_v_c_minc_mins_maxa, 20, 11, 4, up(maxa)).
table(up_v_c_minc_maxs_maxa, 20, 8, 4, up(maxa)).
table(up_v_c_maxc_s_maxa, 20, 18, 4, up(maxa)).
table(up_v_c_maxc_mins_maxa, 20, 11, 4, up(maxa)).
table(up_v_c_maxc_maxs_maxa, 20, 6, 4, up(maxa)).
table(up_v_c_s_mins_maxa, 20, 19, 4, up(maxa)).
table(up_v_c_s_maxs_maxa, 20, 19, 4, up(maxa)).
table(up_v_c_mins_maxs_maxa, 20, 9, 4, up(maxa)).
table(up_v_minc_maxc_s_maxa, 20, 18, 4, up(maxa)).
table(up_v_minc_maxc_mins_maxa, 20, 14, 4, up(maxa)).
table(up_v_minc_maxc_maxs_maxa, 20, 9, 4, up(maxa)).
table(up_v_minc_s_mins_maxa, 20, 18, 4, up(maxa)).
table(up_v_minc_s_maxs_maxa, 20, 18, 4, up(maxa)).
table(up_v_minc_mins_maxs_maxa, 20, 16, 4, up(maxa)).
table(up_v_maxc_s_mins_maxa, 20, 8, 4, up(maxa)).
table(up_v_maxc_s_maxs_maxa, 20, 7, 4, up(maxa)).
table(up_v_maxc_mins_maxs_maxa, 20, 5, 4, up(maxa)).
table(up_v_s_mins_maxs_maxa, 20, 19, 4, up(maxa)).
table(up_c_minc_maxc_s_maxa, 20, 18, 4, up(maxa)).
table(up_c_minc_maxc_mins_maxa, 20, 14, 4, up(maxa)).
table(up_c_minc_maxc_maxs_maxa, 20, 9, 4, up(maxa)).
table(up_c_minc_s_maxs_maxa, 20, 9, 4, up(maxa)).
table(up_c_maxc_s_mins_maxa, 20, 14, 4, up(maxa)).
table(up_c_maxc_s_maxs_maxa, 20, 7, 4, up(maxa)).
table(up_c_maxc_mins_maxs_maxa, 20, 6, 4, up(maxa)).
table(up_c_s_mins_maxs_maxa, 20, 19, 4, up(maxa)).
table(up_minc_maxc_s_mins_maxa, 20, 6, 4, up(maxa)).
table(up_minc_maxc_s_maxs_maxa, 20, 7, 4, up(maxa)).
table(up_minc_s_mins_maxs_maxa, 20, 16, 4, up(maxa)).
table(up_maxc_s_mins_maxs_maxa, 20, 5, 4, up(maxa)).
table(up_v_c_minc_maxc_s_maxa, 20, 18, 5, up(maxa)).
table(up_v_c_minc_maxc_mins_maxa, 20, 14, 5, up(maxa)).
table(up_v_c_minc_maxc_maxs_maxa, 20, 9, 5, up(maxa)).
table(up_v_c_minc_s_mins_maxa, 20, 10, 5, up(maxa)).
table(up_v_c_minc_s_maxs_maxa, 20, 9, 5, up(maxa)).
table(up_v_c_minc_mins_maxs_maxa, 20, 8, 5, up(maxa)).
table(up_v_c_maxc_s_mins_maxa, 20, 8, 5, up(maxa)).
table(up_v_c_maxc_s_maxs_maxa, 20, 6, 5, up(maxa)).
table(up_v_c_maxc_mins_maxs_maxa, 20, 6, 5, up(maxa)).
table(up_v_c_s_mins_maxs_maxa, 20, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_mins_maxa, 20, 11, 5, up(maxa)).
table(up_v_minc_maxc_s_maxs_maxa, 20, 10, 5, up(maxa)).
table(up_v_minc_maxc_mins_maxs_maxa, 20, 7, 5, up(maxa)).
table(up_v_minc_s_mins_maxs_maxa, 20, 16, 5, up(maxa)).
table(up_v_maxc_s_mins_maxs_maxa, 20, 8, 5, up(maxa)).
table(up_c_minc_maxc_s_mins_maxa, 20, 11, 5, up(maxa)).
table(up_c_minc_maxc_s_maxs_maxa, 20, 9, 5, up(maxa)).
table(up_c_minc_maxc_mins_maxs_maxa, 20, 9, 5, up(maxa)).
table(up_c_minc_s_mins_maxs_maxa, 20, 10, 5, up(maxa)).
table(up_c_maxc_s_mins_maxs_maxa, 20, 7, 5, up(maxa)).
table(up_minc_maxc_s_mins_maxs_maxa, 20, 7, 5, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxa, 20, 11, 6, up(maxa)).
table(up_v_c_minc_maxc_s_maxs_maxa, 20, 9, 6, up(maxa)).
table(up_v_c_minc_maxc_mins_maxs_maxa, 20, 9, 6, up(maxa)).
table(up_v_c_minc_s_mins_maxs_maxa, 20, 10, 6, up(maxa)).
table(up_v_c_maxc_s_mins_maxs_maxa, 20, 7, 6, up(maxa)).
table(up_v_minc_maxc_s_mins_maxs_maxa, 20, 11, 6, up(maxa)).
table(up_c_minc_maxc_s_mins_maxs_maxa, 20, 10, 6, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxs_maxa, 20, 10, 7, up(maxa)).
table(low_c_v, 21, 18, 1, low(v)).
table(low_minc_v, 21, 11, 1, low(v)).
table(low_maxc_v, 21, 11, 1, low(v)).
table(low_s_v, 21, 7, 1, low(v)).
table(low_mins_v, 21, 18, 1, low(v)).
table(low_maxs_v, 21, 18, 1, low(v)).
table(low_c_minc_v, 21, 11, 2, low(v)).
table(low_c_maxc_v, 21, 11, 2, low(v)).
table(low_c_s_v, 21, 8, 2, low(v)).
table(low_c_mins_v, 21, 18, 2, low(v)).
table(low_c_maxs_v, 21, 18, 2, low(v)).
table(low_minc_maxc_v, 21, 11, 2, low(v)).
table(low_minc_s_v, 21, 3, 2, low(v)).
table(low_minc_mins_v, 21, 12, 2, low(v)).
table(low_minc_maxs_v, 21, 12, 2, low(v)).
table(low_maxc_s_v, 21, 3, 2, low(v)).
table(low_maxc_mins_v, 21, 12, 2, low(v)).
table(low_maxc_maxs_v, 21, 13, 2, low(v)).
table(low_s_mins_v, 21, 7, 2, low(v)).
table(low_s_maxs_v, 21, 7, 2, low(v)).
table(low_mins_maxs_v, 21, 12, 2, low(v)).
table(low_c_minc_maxc_v, 21, 11, 3, low(v)).
table(low_c_minc_s_v, 21, 6, 3, low(v)).
table(low_c_minc_mins_v, 21, 12, 3, low(v)).
table(low_c_minc_maxs_v, 21, 12, 3, low(v)).
table(low_c_maxc_s_v, 21, 4, 3, low(v)).
table(low_c_maxc_mins_v, 21, 12, 3, low(v)).
table(low_c_maxc_maxs_v, 21, 13, 3, low(v)).
table(low_c_s_mins_v, 21, 8, 3, low(v)).
table(low_c_s_maxs_v, 21, 8, 3, low(v)).
table(low_c_mins_maxs_v, 21, 18, 3, low(v)).
table(low_minc_maxc_s_v, 21, 5, 3, low(v)).
table(low_minc_maxc_mins_v, 21, 12, 3, low(v)).
table(low_minc_maxc_maxs_v, 21, 12, 3, low(v)).
table(low_minc_s_mins_v, 21, 4, 3, low(v)).
table(low_minc_s_maxs_v, 21, 4, 3, low(v)).
table(low_minc_mins_maxs_v, 21, 13, 3, low(v)).
table(low_maxc_s_mins_v, 21, 4, 3, low(v)).
table(low_maxc_s_maxs_v, 21, 4, 3, low(v)).
table(low_maxc_mins_maxs_v, 21, 14, 3, low(v)).
table(low_s_mins_maxs_v, 21, 7, 3, low(v)).
table(low_c_minc_maxc_s_v, 21, 7, 4, low(v)).
table(low_c_minc_maxc_mins_v, 21, 12, 4, low(v)).
table(low_c_minc_maxc_maxs_v, 21, 12, 4, low(v)).
table(low_c_minc_s_mins_v, 21, 7, 4, low(v)).
table(low_c_minc_s_maxs_v, 21, 7, 4, low(v)).
table(low_c_minc_mins_maxs_v, 21, 13, 4, low(v)).
table(low_c_maxc_s_mins_v, 21, 5, 4, low(v)).
table(low_c_maxc_s_maxs_v, 21, 5, 4, low(v)).
table(low_c_maxc_mins_maxs_v, 21, 14, 4, low(v)).
table(low_c_s_mins_maxs_v, 21, 8, 4, low(v)).
table(low_minc_maxc_s_mins_v, 21, 6, 4, low(v)).
table(low_minc_maxc_s_maxs_v, 21, 6, 4, low(v)).
table(low_minc_maxc_mins_maxs_v, 21, 13, 4, low(v)).
table(low_minc_s_mins_maxs_v, 21, 5, 4, low(v)).
table(low_maxc_s_mins_maxs_v, 21, 5, 4, low(v)).
table(low_c_minc_maxc_s_mins_v, 21, 8, 5, low(v)).
table(low_c_minc_maxc_s_maxs_v, 21, 8, 5, low(v)).
table(low_c_minc_maxc_mins_maxs_v, 21, 13, 5, low(v)).
table(low_c_minc_s_mins_maxs_v, 21, 8, 5, low(v)).
table(low_c_maxc_s_mins_maxs_v, 21, 6, 5, low(v)).
table(low_minc_maxc_s_mins_maxs_v, 21, 7, 5, low(v)).
table(low_c_minc_maxc_s_mins_maxs_v, 21, 9, 6, low(v)).
table(up_c_maxc_v, 21, 3, 2, up(v)).
table(up_maxc_s_v, 21, 3, 2, up(v)).
table(up_s_maxs_v, 21, 3, 2, up(v)).
table(up_c_minc_maxc_v, 21, 6, 3, up(v)).
table(up_c_maxc_s_v, 21, 4, 3, up(v)).
table(up_c_maxc_mins_v, 21, 4, 3, up(v)).
table(up_c_maxc_maxs_v, 21, 4, 3, up(v)).
table(up_c_s_maxs_v, 21, 4, 3, up(v)).
table(up_minc_maxc_s_v, 21, 5, 3, up(v)).
table(up_minc_s_maxs_v, 21, 4, 3, up(v)).
table(up_maxc_s_mins_v, 21, 4, 3, up(v)).
table(up_maxc_s_maxs_v, 21, 4, 3, up(v)).
table(up_s_mins_maxs_v, 21, 4, 3, up(v)).
table(up_c_minc_maxc_s_v, 21, 7, 4, up(v)).
table(up_c_minc_maxc_mins_v, 21, 7, 4, up(v)).
table(up_c_minc_maxc_maxs_v, 21, 7, 4, up(v)).
table(up_c_minc_s_maxs_v, 21, 7, 4, up(v)).
table(up_c_maxc_s_mins_v, 21, 5, 4, up(v)).
table(up_c_maxc_s_maxs_v, 21, 5, 4, up(v)).
table(up_c_maxc_mins_maxs_v, 21, 5, 4, up(v)).
table(up_c_s_mins_maxs_v, 21, 5, 4, up(v)).
table(up_minc_maxc_s_mins_v, 21, 6, 4, up(v)).
table(up_minc_maxc_s_maxs_v, 21, 6, 4, up(v)).
table(up_minc_s_mins_maxs_v, 21, 5, 4, up(v)).
table(up_maxc_s_mins_maxs_v, 21, 5, 4, up(v)).
table(up_c_minc_maxc_s_mins_v, 21, 8, 5, up(v)).
table(up_c_minc_maxc_s_maxs_v, 21, 8, 5, up(v)).
table(up_c_minc_maxc_mins_maxs_v, 21, 8, 5, up(v)).
table(up_c_minc_s_mins_maxs_v, 21, 8, 5, up(v)).
table(up_c_maxc_s_mins_maxs_v, 21, 6, 5, up(v)).
table(up_minc_maxc_s_mins_maxs_v, 21, 7, 5, up(v)).
table(up_c_minc_maxc_s_mins_maxs_v, 21, 9, 6, up(v)).
table(low_v_c, 21, 11, 1, low(c)).
table(low_minc_c, 21, 11, 1, low(c)).
table(low_maxc_c, 21, 11, 1, low(c)).
table(low_s_c, 21, 8, 1, low(c)).
table(low_mins_c, 21, 8, 1, low(c)).
table(low_maxs_c, 21, 8, 1, low(c)).
table(low_v_minc_c, 21, 11, 2, low(c)).
table(low_v_maxc_c, 21, 3, 2, low(c)).
table(low_v_s_c, 21, 13, 2, low(c)).
table(low_v_mins_c, 21, 12, 2, low(c)).
table(low_v_maxs_c, 21, 13, 2, low(c)).
table(low_minc_maxc_c, 21, 11, 2, low(c)).
table(low_minc_s_c, 21, 5, 2, low(c)).
table(low_minc_mins_c, 21, 9, 2, low(c)).
table(low_minc_maxs_c, 21, 9, 2, low(c)).
table(low_maxc_s_c, 21, 3, 2, low(c)).
table(low_maxc_mins_c, 21, 12, 2, low(c)).
table(low_maxc_maxs_c, 21, 13, 2, low(c)).
table(low_s_mins_c, 21, 10, 2, low(c)).
table(low_s_maxs_c, 21, 11, 2, low(c)).
table(low_mins_maxs_c, 21, 9, 2, low(c)).
table(low_v_minc_maxc_c, 21, 6, 3, low(c)).
table(low_v_minc_s_c, 21, 12, 3, low(c)).
table(low_v_minc_mins_c, 21, 12, 3, low(c)).
table(low_v_minc_maxs_c, 21, 12, 3, low(c)).
table(low_v_maxc_s_c, 21, 4, 3, low(c)).
table(low_v_maxc_mins_c, 21, 4, 3, low(c)).
table(low_v_maxc_maxs_c, 21, 4, 3, low(c)).
table(low_v_s_mins_c, 21, 14, 3, low(c)).
table(low_v_s_maxs_c, 21, 14, 3, low(c)).
table(low_v_mins_maxs_c, 21, 14, 3, low(c)).
table(low_minc_maxc_s_c, 21, 5, 3, low(c)).
table(low_minc_maxc_mins_c, 21, 12, 3, low(c)).
table(low_minc_maxc_maxs_c, 21, 12, 3, low(c)).
table(low_minc_s_mins_c, 21, 6, 3, low(c)).
table(low_minc_s_maxs_c, 21, 7, 3, low(c)).
table(low_minc_mins_maxs_c, 21, 7, 3, low(c)).
table(low_maxc_s_mins_c, 21, 4, 3, low(c)).
table(low_maxc_s_maxs_c, 21, 4, 3, low(c)).
table(low_maxc_mins_maxs_c, 21, 6, 3, low(c)).
table(low_s_mins_maxs_c, 21, 12, 3, low(c)).
table(low_v_minc_maxc_s_c, 21, 7, 4, low(c)).
table(low_v_minc_maxc_mins_c, 21, 7, 4, low(c)).
table(low_v_minc_maxc_maxs_c, 21, 7, 4, low(c)).
table(low_v_minc_s_mins_c, 21, 13, 4, low(c)).
table(low_v_minc_s_maxs_c, 21, 13, 4, low(c)).
table(low_v_minc_mins_maxs_c, 21, 13, 4, low(c)).
table(low_v_maxc_s_mins_c, 21, 5, 4, low(c)).
table(low_v_maxc_s_maxs_c, 21, 5, 4, low(c)).
table(low_v_maxc_mins_maxs_c, 21, 5, 4, low(c)).
table(low_v_s_mins_maxs_c, 21, 15, 4, low(c)).
table(low_minc_maxc_s_mins_c, 21, 6, 4, low(c)).
table(low_minc_maxc_s_maxs_c, 21, 6, 4, low(c)).
table(low_minc_maxc_mins_maxs_c, 21, 8, 4, low(c)).
table(low_minc_s_mins_maxs_c, 21, 8, 4, low(c)).
table(low_maxc_s_mins_maxs_c, 21, 5, 4, low(c)).
table(low_v_minc_maxc_s_mins_c, 21, 8, 5, low(c)).
table(low_v_minc_maxc_s_maxs_c, 21, 8, 5, low(c)).
table(low_v_minc_maxc_mins_maxs_c, 21, 8, 5, low(c)).
table(low_v_minc_s_mins_maxs_c, 21, 14, 5, low(c)).
table(low_v_maxc_s_mins_maxs_c, 21, 6, 5, low(c)).
table(low_minc_maxc_s_mins_maxs_c, 21, 7, 5, low(c)).
table(low_v_minc_maxc_s_mins_maxs_c, 21, 9, 6, low(c)).
table(up_v_c, 21, 18, 1, up(c)).
table(up_s_c, 21, 3, 1, up(c)).
table(up_v_minc_c, 21, 6, 2, up(c)).
table(up_v_maxc_c, 21, 11, 2, up(c)).
table(up_v_s_c, 21, 4, 2, up(c)).
table(up_v_mins_c, 21, 10, 2, up(c)).
table(up_v_maxs_c, 21, 18, 2, up(c)).
table(up_minc_s_c, 21, 3, 2, up(c)).
table(up_maxc_s_c, 21, 4, 2, up(c)).
table(up_s_mins_c, 21, 5, 2, up(c)).
table(up_s_maxs_c, 21, 5, 2, up(c)).
table(up_v_minc_maxc_c, 21, 7, 3, up(c)).
table(up_v_minc_s_c, 21, 6, 3, up(c)).
table(up_v_minc_mins_c, 21, 7, 3, up(c)).
table(up_v_minc_maxs_c, 21, 7, 3, up(c)).
table(up_v_maxc_s_c, 21, 5, 3, up(c)).
table(up_v_maxc_mins_c, 21, 5, 3, up(c)).
table(up_v_maxc_maxs_c, 21, 13, 3, up(c)).
table(up_v_s_mins_c, 21, 8, 3, up(c)).
table(up_v_s_maxs_c, 21, 6, 3, up(c)).
table(up_v_mins_maxs_c, 21, 12, 3, up(c)).
table(up_minc_maxc_s_c, 21, 5, 3, up(c)).
table(up_minc_s_mins_c, 21, 4, 3, up(c)).
table(up_minc_s_maxs_c, 21, 4, 3, up(c)).
table(up_maxc_s_mins_c, 21, 5, 3, up(c)).
table(up_maxc_s_maxs_c, 21, 5, 3, up(c)).
table(up_s_mins_maxs_c, 21, 8, 3, up(c)).
table(up_v_minc_maxc_s_c, 21, 7, 4, up(c)).
table(up_v_minc_maxc_mins_c, 21, 8, 4, up(c)).
table(up_v_minc_maxc_maxs_c, 21, 8, 4, up(c)).
table(up_v_minc_s_mins_c, 21, 7, 4, up(c)).
table(up_v_minc_s_maxs_c, 21, 7, 4, up(c)).
table(up_v_minc_mins_maxs_c, 21, 8, 4, up(c)).
table(up_v_maxc_s_mins_c, 21, 6, 4, up(c)).
table(up_v_maxc_s_maxs_c, 21, 6, 4, up(c)).
table(up_v_maxc_mins_maxs_c, 21, 7, 4, up(c)).
table(up_v_s_mins_maxs_c, 21, 10, 4, up(c)).
table(up_minc_maxc_s_mins_c, 21, 6, 4, up(c)).
table(up_minc_maxc_s_maxs_c, 21, 6, 4, up(c)).
table(up_minc_s_mins_maxs_c, 21, 5, 4, up(c)).
table(up_maxc_s_mins_maxs_c, 21, 6, 4, up(c)).
table(up_v_minc_maxc_s_mins_c, 21, 8, 5, up(c)).
table(up_v_minc_maxc_s_maxs_c, 21, 8, 5, up(c)).
table(up_v_minc_maxc_mins_maxs_c, 21, 9, 5, up(c)).
table(up_v_minc_s_mins_maxs_c, 21, 8, 5, up(c)).
table(up_v_maxc_s_mins_maxs_c, 21, 7, 5, up(c)).
table(up_minc_maxc_s_mins_maxs_c, 21, 7, 5, up(c)).
table(up_v_minc_maxc_s_mins_maxs_c, 21, 9, 6, up(c)).
table(low_v_minc, 21, 3, 1, low(minc)).
table(low_c_minc, 21, 3, 1, low(minc)).
table(low_maxc_minc, 21, 3, 1, low(minc)).
table(low_s_minc, 21, 3, 1, low(minc)).
table(low_mins_minc, 21, 2, 1, low(minc)).
table(low_maxs_minc, 21, 3, 1, low(minc)).
table(low_v_c_minc, 21, 5, 2, low(minc)).
table(low_v_maxc_minc, 21, 4, 2, low(minc)).
table(low_v_s_minc, 21, 4, 2, low(minc)).
table(low_v_mins_minc, 21, 3, 2, low(minc)).
table(low_v_maxs_minc, 21, 4, 2, low(minc)).
table(low_c_maxc_minc, 21, 4, 2, low(minc)).
table(low_c_s_minc, 21, 4, 2, low(minc)).
table(low_c_mins_minc, 21, 3, 2, low(minc)).
table(low_c_maxs_minc, 21, 4, 2, low(minc)).
table(low_maxc_s_minc, 21, 4, 2, low(minc)).
table(low_maxc_mins_minc, 21, 4, 2, low(minc)).
table(low_maxc_maxs_minc, 21, 4, 2, low(minc)).
table(low_s_mins_minc, 21, 3, 2, low(minc)).
table(low_s_maxs_minc, 21, 4, 2, low(minc)).
table(low_mins_maxs_minc, 21, 3, 2, low(minc)).
table(low_v_c_maxc_minc, 21, 6, 3, low(minc)).
table(low_v_c_s_minc, 21, 6, 3, low(minc)).
table(low_v_c_mins_minc, 21, 6, 3, low(minc)).
table(low_v_c_maxs_minc, 21, 6, 3, low(minc)).
table(low_v_maxc_s_minc, 21, 5, 3, low(minc)).
table(low_v_maxc_mins_minc, 21, 5, 3, low(minc)).
table(low_v_maxc_maxs_minc, 21, 5, 3, low(minc)).
table(low_v_s_mins_minc, 21, 4, 3, low(minc)).
table(low_v_s_maxs_minc, 21, 5, 3, low(minc)).
table(low_v_mins_maxs_minc, 21, 4, 3, low(minc)).
table(low_c_maxc_s_minc, 21, 5, 3, low(minc)).
table(low_c_maxc_mins_minc, 21, 5, 3, low(minc)).
table(low_c_maxc_maxs_minc, 21, 5, 3, low(minc)).
table(low_c_s_mins_minc, 21, 4, 3, low(minc)).
table(low_c_s_maxs_minc, 21, 6, 3, low(minc)).
table(low_c_mins_maxs_minc, 21, 4, 3, low(minc)).
table(low_maxc_s_mins_minc, 21, 5, 3, low(minc)).
table(low_maxc_s_maxs_minc, 21, 5, 3, low(minc)).
table(low_maxc_mins_maxs_minc, 21, 5, 3, low(minc)).
table(low_s_mins_maxs_minc, 21, 4, 3, low(minc)).
table(low_v_c_maxc_s_minc, 21, 7, 4, low(minc)).
table(low_v_c_maxc_mins_minc, 21, 7, 4, low(minc)).
table(low_v_c_maxc_maxs_minc, 21, 7, 4, low(minc)).
table(low_v_c_s_mins_minc, 21, 7, 4, low(minc)).
table(low_v_c_s_maxs_minc, 21, 7, 4, low(minc)).
table(low_v_c_mins_maxs_minc, 21, 7, 4, low(minc)).
table(low_v_maxc_s_mins_minc, 21, 6, 4, low(minc)).
table(low_v_maxc_s_maxs_minc, 21, 6, 4, low(minc)).
table(low_v_maxc_mins_maxs_minc, 21, 6, 4, low(minc)).
table(low_v_s_mins_maxs_minc, 21, 5, 4, low(minc)).
table(low_c_maxc_s_mins_minc, 21, 6, 4, low(minc)).
table(low_c_maxc_s_maxs_minc, 21, 6, 4, low(minc)).
table(low_c_maxc_mins_maxs_minc, 21, 6, 4, low(minc)).
table(low_c_s_mins_maxs_minc, 21, 6, 4, low(minc)).
table(low_maxc_s_mins_maxs_minc, 21, 6, 4, low(minc)).
table(low_v_c_maxc_s_mins_minc, 21, 8, 5, low(minc)).
table(low_v_c_maxc_s_maxs_minc, 21, 8, 5, low(minc)).
table(low_v_c_maxc_mins_maxs_minc, 21, 8, 5, low(minc)).
table(low_v_c_s_mins_maxs_minc, 21, 8, 5, low(minc)).
table(low_v_maxc_s_mins_maxs_minc, 21, 7, 5, low(minc)).
table(low_c_maxc_s_mins_maxs_minc, 21, 7, 5, low(minc)).
table(low_v_c_maxc_s_mins_maxs_minc, 21, 9, 6, low(minc)).
table(up_v_minc, 21, 11, 1, up(minc)).
table(up_maxc_minc, 21, 6, 1, up(minc)).
table(up_v_c_minc, 21, 5, 2, up(minc)).
table(up_v_maxc_minc, 21, 4, 2, up(minc)).
table(up_v_s_minc, 21, 13, 2, up(minc)).
table(up_v_mins_minc, 21, 12, 2, up(minc)).
table(up_v_maxs_minc, 21, 13, 2, up(minc)).
table(up_c_maxc_minc, 21, 4, 2, up(minc)).
table(up_maxc_s_minc, 21, 4, 2, up(minc)).
table(up_maxc_mins_minc, 21, 4, 2, up(minc)).
table(up_maxc_maxs_minc, 21, 7, 2, up(minc)).
table(up_s_maxs_minc, 21, 14, 2, up(minc)).
table(up_v_c_maxc_minc, 21, 6, 3, up(minc)).
table(up_v_c_s_minc, 21, 6, 3, up(minc)).
table(up_v_c_mins_minc, 21, 6, 3, up(minc)).
table(up_v_c_maxs_minc, 21, 6, 3, up(minc)).
table(up_v_maxc_s_minc, 21, 5, 3, up(minc)).
table(up_v_maxc_mins_minc, 21, 5, 3, up(minc)).
table(up_v_maxc_maxs_minc, 21, 5, 3, up(minc)).
table(up_v_s_mins_minc, 21, 14, 3, up(minc)).
table(up_v_s_maxs_minc, 21, 14, 3, up(minc)).
table(up_v_mins_maxs_minc, 21, 14, 3, up(minc)).
table(up_c_maxc_s_minc, 21, 5, 3, up(minc)).
table(up_c_maxc_mins_minc, 21, 5, 3, up(minc)).
table(up_c_maxc_maxs_minc, 21, 5, 3, up(minc)).
table(up_c_s_maxs_minc, 21, 4, 3, up(minc)).
table(up_maxc_s_mins_minc, 21, 5, 3, up(minc)).
table(up_maxc_s_maxs_minc, 21, 5, 3, up(minc)).
table(up_maxc_mins_maxs_minc, 21, 5, 3, up(minc)).
table(up_s_mins_maxs_minc, 21, 15, 3, up(minc)).
table(up_v_c_maxc_s_minc, 21, 7, 4, up(minc)).
table(up_v_c_maxc_mins_minc, 21, 7, 4, up(minc)).
table(up_v_c_maxc_maxs_minc, 21, 7, 4, up(minc)).
table(up_v_c_s_mins_minc, 21, 7, 4, up(minc)).
table(up_v_c_s_maxs_minc, 21, 7, 4, up(minc)).
table(up_v_c_mins_maxs_minc, 21, 7, 4, up(minc)).
table(up_v_maxc_s_mins_minc, 21, 6, 4, up(minc)).
table(up_v_maxc_s_maxs_minc, 21, 6, 4, up(minc)).
table(up_v_maxc_mins_maxs_minc, 21, 6, 4, up(minc)).
table(up_v_s_mins_maxs_minc, 21, 15, 4, up(minc)).
table(up_c_maxc_s_mins_minc, 21, 6, 4, up(minc)).
table(up_c_maxc_s_maxs_minc, 21, 6, 4, up(minc)).
table(up_c_maxc_mins_maxs_minc, 21, 6, 4, up(minc)).
table(up_c_s_mins_maxs_minc, 21, 5, 4, up(minc)).
table(up_maxc_s_mins_maxs_minc, 21, 6, 4, up(minc)).
table(up_v_c_maxc_s_mins_minc, 21, 8, 5, up(minc)).
table(up_v_c_maxc_s_maxs_minc, 21, 8, 5, up(minc)).
table(up_v_c_maxc_mins_maxs_minc, 21, 8, 5, up(minc)).
table(up_v_c_s_mins_maxs_minc, 21, 8, 5, up(minc)).
table(up_v_maxc_s_mins_maxs_minc, 21, 7, 5, up(minc)).
table(up_c_maxc_s_mins_maxs_minc, 21, 7, 5, up(minc)).
table(up_v_c_maxc_s_mins_maxs_minc, 21, 9, 6, up(minc)).
table(low_v_maxc, 21, 18, 1, low(maxc)).
table(low_c_maxc, 21, 18, 1, low(maxc)).
table(low_minc_maxc, 21, 6, 1, low(maxc)).
table(low_s_maxc, 21, 18, 1, low(maxc)).
table(low_mins_maxc, 21, 9, 1, low(maxc)).
table(low_maxs_maxc, 21, 2, 1, low(maxc)).
table(low_v_c_maxc, 21, 3, 2, low(maxc)).
table(low_v_minc_maxc, 21, 6, 2, low(maxc)).
table(low_v_s_maxc, 21, 4, 2, low(maxc)).
table(low_v_mins_maxc, 21, 9, 2, low(maxc)).
table(low_v_maxs_maxc, 21, 3, 2, low(maxc)).
table(low_c_minc_maxc, 21, 11, 2, low(maxc)).
table(low_c_s_maxc, 21, 4, 2, low(maxc)).
table(low_c_mins_maxc, 21, 18, 2, low(maxc)).
table(low_c_maxs_maxc, 21, 3, 2, low(maxc)).
table(low_minc_s_maxc, 21, 6, 2, low(maxc)).
table(low_minc_mins_maxc, 21, 4, 2, low(maxc)).
table(low_minc_maxs_maxc, 21, 4, 2, low(maxc)).
table(low_s_mins_maxc, 21, 18, 2, low(maxc)).
table(low_s_maxs_maxc, 21, 3, 2, low(maxc)).
table(low_mins_maxs_maxc, 21, 3, 2, low(maxc)).
table(low_v_c_minc_maxc, 21, 6, 3, low(maxc)).
table(low_v_c_s_maxc, 21, 4, 3, low(maxc)).
table(low_v_c_mins_maxc, 21, 4, 3, low(maxc)).
table(low_v_c_maxs_maxc, 21, 4, 3, low(maxc)).
table(low_v_minc_s_maxc, 21, 5, 3, low(maxc)).
table(low_v_minc_mins_maxc, 21, 5, 3, low(maxc)).
table(low_v_minc_maxs_maxc, 21, 5, 3, low(maxc)).
table(low_v_s_mins_maxc, 21, 5, 3, low(maxc)).
table(low_v_s_maxs_maxc, 21, 4, 3, low(maxc)).
table(low_v_mins_maxs_maxc, 21, 4, 3, low(maxc)).
table(low_c_minc_s_maxc, 21, 5, 3, low(maxc)).
table(low_c_minc_mins_maxc, 21, 6, 3, low(maxc)).
table(low_c_minc_maxs_maxc, 21, 5, 3, low(maxc)).
table(low_c_s_mins_maxc, 21, 4, 3, low(maxc)).
table(low_c_s_maxs_maxc, 21, 4, 3, low(maxc)).
table(low_c_mins_maxs_maxc, 21, 4, 3, low(maxc)).
table(low_minc_s_mins_maxc, 21, 5, 3, low(maxc)).
table(low_minc_s_maxs_maxc, 21, 5, 3, low(maxc)).
table(low_minc_mins_maxs_maxc, 21, 5, 3, low(maxc)).
table(low_s_mins_maxs_maxc, 21, 4, 3, low(maxc)).
table(low_v_c_minc_s_maxc, 21, 7, 4, low(maxc)).
table(low_v_c_minc_mins_maxc, 21, 7, 4, low(maxc)).
table(low_v_c_minc_maxs_maxc, 21, 7, 4, low(maxc)).
table(low_v_c_s_mins_maxc, 21, 5, 4, low(maxc)).
table(low_v_c_s_maxs_maxc, 21, 5, 4, low(maxc)).
table(low_v_c_mins_maxs_maxc, 21, 5, 4, low(maxc)).
table(low_v_minc_s_mins_maxc, 21, 6, 4, low(maxc)).
table(low_v_minc_s_maxs_maxc, 21, 6, 4, low(maxc)).
table(low_v_minc_mins_maxs_maxc, 21, 6, 4, low(maxc)).
table(low_v_s_mins_maxs_maxc, 21, 5, 4, low(maxc)).
table(low_c_minc_s_mins_maxc, 21, 6, 4, low(maxc)).
table(low_c_minc_s_maxs_maxc, 21, 6, 4, low(maxc)).
table(low_c_minc_mins_maxs_maxc, 21, 6, 4, low(maxc)).
table(low_c_s_mins_maxs_maxc, 21, 5, 4, low(maxc)).
table(low_minc_s_mins_maxs_maxc, 21, 6, 4, low(maxc)).
table(low_v_c_minc_s_mins_maxc, 21, 8, 5, low(maxc)).
table(low_v_c_minc_s_maxs_maxc, 21, 8, 5, low(maxc)).
table(low_v_c_minc_mins_maxs_maxc, 21, 8, 5, low(maxc)).
table(low_v_c_s_mins_maxs_maxc, 21, 6, 5, low(maxc)).
table(low_v_minc_s_mins_maxs_maxc, 21, 7, 5, low(maxc)).
table(low_c_minc_s_mins_maxs_maxc, 21, 7, 5, low(maxc)).
table(low_v_c_minc_s_mins_maxs_maxc, 21, 9, 6, low(maxc)).
table(up_v_maxc, 21, 11, 1, up(maxc)).
table(up_v_c_maxc, 21, 11, 2, up(maxc)).
table(up_v_minc_maxc, 21, 11, 2, up(maxc)).
table(up_v_s_maxc, 21, 13, 2, up(maxc)).
table(up_v_mins_maxc, 21, 12, 2, up(maxc)).
table(up_v_maxs_maxc, 21, 13, 2, up(maxc)).
table(up_v_c_minc_maxc, 21, 11, 3, up(maxc)).
table(up_v_c_s_maxc, 21, 13, 3, up(maxc)).
table(up_v_c_mins_maxc, 21, 12, 3, up(maxc)).
table(up_v_c_maxs_maxc, 21, 13, 3, up(maxc)).
table(up_v_minc_s_maxc, 21, 12, 3, up(maxc)).
table(up_v_minc_mins_maxc, 21, 12, 3, up(maxc)).
table(up_v_minc_maxs_maxc, 21, 12, 3, up(maxc)).
table(up_v_s_mins_maxc, 21, 14, 3, up(maxc)).
table(up_v_s_maxs_maxc, 21, 14, 3, up(maxc)).
table(up_v_mins_maxs_maxc, 21, 14, 3, up(maxc)).
table(up_v_c_minc_s_maxc, 21, 12, 4, up(maxc)).
table(up_v_c_minc_mins_maxc, 21, 12, 4, up(maxc)).
table(up_v_c_minc_maxs_maxc, 21, 12, 4, up(maxc)).
table(up_v_c_s_mins_maxc, 21, 14, 4, up(maxc)).
table(up_v_c_s_maxs_maxc, 21, 6, 4, up(maxc)).
table(up_v_c_mins_maxs_maxc, 21, 10, 4, up(maxc)).
table(up_v_minc_s_mins_maxc, 21, 13, 4, up(maxc)).
table(up_v_minc_s_maxs_maxc, 21, 13, 4, up(maxc)).
table(up_v_minc_mins_maxs_maxc, 21, 13, 4, up(maxc)).
table(up_v_s_mins_maxs_maxc, 21, 15, 4, up(maxc)).
table(up_v_c_minc_s_mins_maxc, 21, 13, 5, up(maxc)).
table(up_v_c_minc_s_maxs_maxc, 21, 8, 5, up(maxc)).
table(up_v_c_minc_mins_maxs_maxc, 21, 9, 5, up(maxc)).
table(up_v_c_s_mins_maxs_maxc, 21, 10, 5, up(maxc)).
table(up_v_minc_s_mins_maxs_maxc, 21, 14, 5, up(maxc)).
table(up_v_c_minc_s_mins_maxs_maxc, 21, 9, 6, up(maxc)).
table(low_v_s, 21, 18, 1, low(s)).
table(low_c_s, 21, 3, 1, low(s)).
table(low_minc_s, 21, 18, 1, low(s)).
table(low_maxc_s, 21, 18, 1, low(s)).
table(low_mins_s, 21, 18, 1, low(s)).
table(low_maxs_s, 21, 18, 1, low(s)).
table(low_v_c_s, 21, 4, 2, low(s)).
table(low_v_minc_s, 21, 18, 2, low(s)).
table(low_v_maxc_s, 21, 6, 2, low(s)).
table(low_v_mins_s, 21, 9, 2, low(s)).
table(low_v_maxs_s, 21, 3, 2, low(s)).
table(low_c_minc_s, 21, 5, 2, low(s)).
table(low_c_maxc_s, 21, 5, 2, low(s)).
table(low_c_mins_s, 21, 5, 2, low(s)).
table(low_c_maxs_s, 21, 5, 2, low(s)).
table(low_minc_maxc_s, 21, 18, 2, low(s)).
table(low_minc_mins_s, 21, 12, 2, low(s)).
table(low_minc_maxs_s, 21, 14, 2, low(s)).
table(low_maxc_mins_s, 21, 7, 2, low(s)).
table(low_maxc_maxs_s, 21, 14, 2, low(s)).
table(low_mins_maxs_s, 21, 12, 2, low(s)).
table(low_v_c_minc_s, 21, 8, 3, low(s)).
table(low_v_c_maxc_s, 21, 6, 3, low(s)).
table(low_v_c_mins_s, 21, 8, 3, low(s)).
table(low_v_c_maxs_s, 21, 4, 3, low(s)).
table(low_v_minc_maxc_s, 21, 10, 3, low(s)).
table(low_v_minc_mins_s, 21, 13, 3, low(s)).
table(low_v_minc_maxs_s, 21, 4, 3, low(s)).
table(low_v_maxc_mins_s, 21, 4, 3, low(s)).
table(low_v_maxc_maxs_s, 21, 4, 3, low(s)).
table(low_v_mins_maxs_s, 21, 4, 3, low(s)).
table(low_c_minc_maxc_s, 21, 8, 3, low(s)).
table(low_c_minc_mins_s, 21, 4, 3, low(s)).
table(low_c_minc_maxs_s, 21, 5, 3, low(s)).
table(low_c_maxc_mins_s, 21, 8, 3, low(s)).
table(low_c_maxc_maxs_s, 21, 5, 3, low(s)).
table(low_c_mins_maxs_s, 21, 8, 3, low(s)).
table(low_minc_maxc_mins_s, 21, 13, 3, low(s)).
table(low_minc_maxc_maxs_s, 21, 14, 3, low(s)).
table(low_minc_mins_maxs_s, 21, 8, 3, low(s)).
table(low_maxc_mins_maxs_s, 21, 9, 3, low(s)).
table(low_v_c_minc_maxc_s, 21, 10, 4, low(s)).
table(low_v_c_minc_mins_s, 21, 7, 4, low(s)).
table(low_v_c_minc_maxs_s, 21, 7, 4, low(s)).
table(low_v_c_maxc_mins_s, 21, 5, 4, low(s)).
table(low_v_c_maxc_maxs_s, 21, 5, 4, low(s)).
table(low_v_c_mins_maxs_s, 21, 5, 4, low(s)).
table(low_v_minc_maxc_mins_s, 21, 8, 4, low(s)).
table(low_v_minc_maxc_maxs_s, 21, 6, 4, low(s)).
table(low_v_minc_mins_maxs_s, 21, 5, 4, low(s)).
table(low_v_maxc_mins_maxs_s, 21, 5, 4, low(s)).
table(low_c_minc_maxc_mins_s, 21, 6, 4, low(s)).
table(low_c_minc_maxc_maxs_s, 21, 7, 4, low(s)).
table(low_c_minc_mins_maxs_s, 21, 5, 4, low(s)).
table(low_c_maxc_mins_maxs_s, 21, 6, 4, low(s)).
table(low_minc_maxc_mins_maxs_s, 21, 9, 4, low(s)).
table(low_v_c_minc_maxc_mins_s, 21, 8, 5, low(s)).
table(low_v_c_minc_maxc_maxs_s, 21, 8, 5, low(s)).
table(low_v_c_minc_mins_maxs_s, 21, 8, 5, low(s)).
table(low_v_c_maxc_mins_maxs_s, 21, 6, 5, low(s)).
table(low_v_minc_maxc_mins_maxs_s, 21, 7, 5, low(s)).
table(low_c_minc_maxc_mins_maxs_s, 21, 7, 5, low(s)).
table(low_v_c_minc_maxc_mins_maxs_s, 21, 9, 6, low(s)).
table(up_v_s, 21, 7, 1, up(s)).
table(up_v_c_s, 21, 8, 2, up(s)).
table(up_v_minc_s, 21, 8, 2, up(s)).
table(up_v_maxc_s, 21, 8, 2, up(s)).
table(up_v_mins_s, 21, 4, 2, up(s)).
table(up_v_maxs_s, 21, 7, 2, up(s)).
table(up_c_maxc_s, 21, 9, 2, up(s)).
table(up_v_c_minc_s, 21, 11, 3, up(s)).
table(up_v_c_maxc_s, 21, 9, 3, up(s)).
table(up_v_c_mins_s, 21, 5, 3, up(s)).
table(up_v_c_maxs_s, 21, 8, 3, up(s)).
table(up_v_minc_maxc_s, 21, 10, 3, up(s)).
table(up_v_minc_mins_s, 21, 5, 3, up(s)).
table(up_v_minc_maxs_s, 21, 8, 3, up(s)).
table(up_v_maxc_mins_s, 21, 5, 3, up(s)).
table(up_v_maxc_maxs_s, 21, 8, 3, up(s)).
table(up_v_mins_maxs_s, 21, 5, 3, up(s)).
table(up_c_minc_maxc_s, 21, 12, 3, up(s)).
table(up_c_maxc_mins_s, 21, 5, 3, up(s)).
table(up_c_maxc_maxs_s, 21, 9, 3, up(s)).
table(up_v_c_minc_maxc_s, 21, 12, 4, up(s)).
table(up_v_c_minc_mins_s, 21, 8, 4, up(s)).
table(up_v_c_minc_maxs_s, 21, 11, 4, up(s)).
table(up_v_c_maxc_mins_s, 21, 6, 4, up(s)).
table(up_v_c_maxc_maxs_s, 21, 9, 4, up(s)).
table(up_v_c_mins_maxs_s, 21, 6, 4, up(s)).
table(up_v_minc_maxc_mins_s, 21, 7, 4, up(s)).
table(up_v_minc_maxc_maxs_s, 21, 10, 4, up(s)).
table(up_v_minc_mins_maxs_s, 21, 6, 4, up(s)).
table(up_v_maxc_mins_maxs_s, 21, 6, 4, up(s)).
table(up_c_minc_maxc_mins_s, 21, 7, 4, up(s)).
table(up_c_minc_maxc_maxs_s, 21, 12, 4, up(s)).
table(up_c_maxc_mins_maxs_s, 21, 6, 4, up(s)).
table(up_v_c_minc_maxc_mins_s, 21, 9, 5, up(s)).
table(up_v_c_minc_maxc_maxs_s, 21, 12, 5, up(s)).
table(up_v_c_minc_mins_maxs_s, 21, 9, 5, up(s)).
table(up_v_c_maxc_mins_maxs_s, 21, 7, 5, up(s)).
table(up_v_minc_maxc_mins_maxs_s, 21, 8, 5, up(s)).
table(up_c_minc_maxc_mins_maxs_s, 21, 8, 5, up(s)).
table(up_v_c_minc_maxc_mins_maxs_s, 21, 10, 6, up(s)).
table(low_v_mins, 21, 2, 1, low(mins)).
table(low_c_mins, 21, 2, 1, low(mins)).
table(low_minc_mins, 21, 2, 1, low(mins)).
table(low_maxc_mins, 21, 2, 1, low(mins)).
table(low_s_mins, 21, 2, 1, low(mins)).
table(low_maxs_mins, 21, 2, 1, low(mins)).
table(low_v_c_mins, 21, 3, 2, low(mins)).
table(low_v_minc_mins, 21, 3, 2, low(mins)).
table(low_v_maxc_mins, 21, 3, 2, low(mins)).
table(low_v_s_mins, 21, 3, 2, low(mins)).
table(low_v_maxs_mins, 21, 3, 2, low(mins)).
table(low_c_minc_mins, 21, 3, 2, low(mins)).
table(low_c_maxc_mins, 21, 3, 2, low(mins)).
table(low_c_s_mins, 21, 3, 2, low(mins)).
table(low_c_maxs_mins, 21, 3, 2, low(mins)).
table(low_minc_maxc_mins, 21, 4, 2, low(mins)).
table(low_minc_s_mins, 21, 3, 2, low(mins)).
table(low_minc_maxs_mins, 21, 3, 2, low(mins)).
table(low_maxc_s_mins, 21, 3, 2, low(mins)).
table(low_maxc_maxs_mins, 21, 3, 2, low(mins)).
table(low_s_maxs_mins, 21, 3, 2, low(mins)).
table(low_v_c_minc_mins, 21, 6, 3, low(mins)).
table(low_v_c_maxc_mins, 21, 4, 3, low(mins)).
table(low_v_c_s_mins, 21, 4, 3, low(mins)).
table(low_v_c_maxs_mins, 21, 4, 3, low(mins)).
table(low_v_minc_maxc_mins, 21, 5, 3, low(mins)).
table(low_v_minc_s_mins, 21, 4, 3, low(mins)).
table(low_v_minc_maxs_mins, 21, 4, 3, low(mins)).
table(low_v_maxc_s_mins, 21, 4, 3, low(mins)).
table(low_v_maxc_maxs_mins, 21, 4, 3, low(mins)).
table(low_v_s_maxs_mins, 21, 4, 3, low(mins)).
table(low_c_minc_maxc_mins, 21, 5, 3, low(mins)).
table(low_c_minc_s_mins, 21, 4, 3, low(mins)).
table(low_c_minc_maxs_mins, 21, 4, 3, low(mins)).
table(low_c_maxc_s_mins, 21, 4, 3, low(mins)).
table(low_c_maxc_maxs_mins, 21, 4, 3, low(mins)).
table(low_c_s_maxs_mins, 21, 4, 3, low(mins)).
table(low_minc_maxc_s_mins, 21, 5, 3, low(mins)).
table(low_minc_maxc_maxs_mins, 21, 5, 3, low(mins)).
table(low_minc_s_maxs_mins, 21, 4, 3, low(mins)).
table(low_maxc_s_maxs_mins, 21, 4, 3, low(mins)).
table(low_v_c_minc_maxc_mins, 21, 7, 4, low(mins)).
table(low_v_c_minc_s_mins, 21, 7, 4, low(mins)).
table(low_v_c_minc_maxs_mins, 21, 7, 4, low(mins)).
table(low_v_c_maxc_s_mins, 21, 5, 4, low(mins)).
table(low_v_c_maxc_maxs_mins, 21, 5, 4, low(mins)).
table(low_v_c_s_maxs_mins, 21, 5, 4, low(mins)).
table(low_v_minc_maxc_s_mins, 21, 6, 4, low(mins)).
table(low_v_minc_maxc_maxs_mins, 21, 6, 4, low(mins)).
table(low_v_minc_s_maxs_mins, 21, 5, 4, low(mins)).
table(low_v_maxc_s_maxs_mins, 21, 5, 4, low(mins)).
table(low_c_minc_maxc_s_mins, 21, 6, 4, low(mins)).
table(low_c_minc_maxc_maxs_mins, 21, 6, 4, low(mins)).
table(low_c_minc_s_maxs_mins, 21, 5, 4, low(mins)).
table(low_c_maxc_s_maxs_mins, 21, 5, 4, low(mins)).
table(low_minc_maxc_s_maxs_mins, 21, 6, 4, low(mins)).
table(low_v_c_minc_maxc_s_mins, 21, 8, 5, low(mins)).
table(low_v_c_minc_maxc_maxs_mins, 21, 8, 5, low(mins)).
table(low_v_c_minc_s_maxs_mins, 21, 8, 5, low(mins)).
table(low_v_c_maxc_s_maxs_mins, 21, 6, 5, low(mins)).
table(low_v_minc_maxc_s_maxs_mins, 21, 7, 5, low(mins)).
table(low_c_minc_maxc_s_maxs_mins, 21, 7, 5, low(mins)).
table(low_v_c_minc_maxc_s_maxs_mins, 21, 9, 6, low(mins)).
table(up_v_mins, 21, 18, 1, up(mins)).
table(up_maxc_mins, 21, 9, 1, up(mins)).
table(up_maxs_mins, 21, 2, 1, up(mins)).
table(up_v_c_mins, 21, 6, 2, up(mins)).
table(up_v_minc_mins, 21, 3, 2, up(mins)).
table(up_v_maxc_mins, 21, 5, 2, up(mins)).
table(up_v_s_mins, 21, 3, 2, up(mins)).
table(up_v_maxs_mins, 21, 3, 2, up(mins)).
table(up_c_maxc_mins, 21, 5, 2, up(mins)).
table(up_c_maxs_mins, 21, 3, 2, up(mins)).
table(up_minc_maxc_mins, 21, 4, 2, up(mins)).
table(up_minc_maxs_mins, 21, 3, 2, up(mins)).
table(up_maxc_s_mins, 21, 3, 2, up(mins)).
table(up_maxc_maxs_mins, 21, 3, 2, up(mins)).
table(up_s_maxs_mins, 21, 3, 2, up(mins)).
table(up_v_c_minc_mins, 21, 6, 3, up(mins)).
table(up_v_c_maxc_mins, 21, 7, 3, up(mins)).
table(up_v_c_s_mins, 21, 4, 3, up(mins)).
table(up_v_c_maxs_mins, 21, 4, 3, up(mins)).
table(up_v_minc_maxc_mins, 21, 5, 3, up(mins)).
table(up_v_minc_s_mins, 21, 4, 3, up(mins)).
table(up_v_minc_maxs_mins, 21, 4, 3, up(mins)).
table(up_v_maxc_s_mins, 21, 4, 3, up(mins)).
table(up_v_maxc_maxs_mins, 21, 4, 3, up(mins)).
table(up_v_s_maxs_mins, 21, 4, 3, up(mins)).
table(up_c_minc_maxc_mins, 21, 5, 3, up(mins)).
table(up_c_minc_maxs_mins, 21, 4, 3, up(mins)).
table(up_c_maxc_s_mins, 21, 4, 3, up(mins)).
table(up_c_maxc_maxs_mins, 21, 4, 3, up(mins)).
table(up_c_s_maxs_mins, 21, 4, 3, up(mins)).
table(up_minc_maxc_s_mins, 21, 5, 3, up(mins)).
table(up_minc_maxc_maxs_mins, 21, 5, 3, up(mins)).
table(up_minc_s_maxs_mins, 21, 4, 3, up(mins)).
table(up_maxc_s_maxs_mins, 21, 4, 3, up(mins)).
table(up_v_c_minc_maxc_mins, 21, 7, 4, up(mins)).
table(up_v_c_minc_s_mins, 21, 7, 4, up(mins)).
table(up_v_c_minc_maxs_mins, 21, 7, 4, up(mins)).
table(up_v_c_maxc_s_mins, 21, 5, 4, up(mins)).
table(up_v_c_maxc_maxs_mins, 21, 5, 4, up(mins)).
table(up_v_c_s_maxs_mins, 21, 5, 4, up(mins)).
table(up_v_minc_maxc_s_mins, 21, 6, 4, up(mins)).
table(up_v_minc_maxc_maxs_mins, 21, 6, 4, up(mins)).
table(up_v_minc_s_maxs_mins, 21, 5, 4, up(mins)).
table(up_v_maxc_s_maxs_mins, 21, 5, 4, up(mins)).
table(up_c_minc_maxc_s_mins, 21, 6, 4, up(mins)).
table(up_c_minc_maxc_maxs_mins, 21, 6, 4, up(mins)).
table(up_c_minc_s_maxs_mins, 21, 5, 4, up(mins)).
table(up_c_maxc_s_maxs_mins, 21, 5, 4, up(mins)).
table(up_minc_maxc_s_maxs_mins, 21, 6, 4, up(mins)).
table(up_v_c_minc_maxc_s_mins, 21, 8, 5, up(mins)).
table(up_v_c_minc_maxc_maxs_mins, 21, 8, 5, up(mins)).
table(up_v_c_minc_s_maxs_mins, 21, 8, 5, up(mins)).
table(up_v_c_maxc_s_maxs_mins, 21, 6, 5, up(mins)).
table(up_v_minc_maxc_s_maxs_mins, 21, 7, 5, up(mins)).
table(up_c_minc_maxc_s_maxs_mins, 21, 7, 5, up(mins)).
table(up_v_c_minc_maxc_s_maxs_mins, 21, 9, 6, up(mins)).
table(low_v_maxs, 21, 7, 1, low(maxs)).
table(low_c_maxs, 21, 4, 1, low(maxs)).
table(low_minc_maxs, 21, 4, 1, low(maxs)).
table(low_maxc_maxs, 21, 4, 1, low(maxs)).
table(low_s_maxs, 21, 7, 1, low(maxs)).
table(low_mins_maxs, 21, 2, 1, low(maxs)).
table(low_v_c_maxs, 21, 8, 2, low(maxs)).
table(low_v_minc_maxs, 21, 8, 2, low(maxs)).
table(low_v_maxc_maxs, 21, 8, 2, low(maxs)).
table(low_v_s_maxs, 21, 3, 2, low(maxs)).
table(low_v_mins_maxs, 21, 3, 2, low(maxs)).
table(low_c_minc_maxs, 21, 5, 2, low(maxs)).
table(low_c_maxc_maxs, 21, 5, 2, low(maxs)).
table(low_c_s_maxs, 21, 8, 2, low(maxs)).
table(low_c_mins_maxs, 21, 3, 2, low(maxs)).
table(low_minc_maxc_maxs, 21, 6, 2, low(maxs)).
table(low_minc_s_maxs, 21, 3, 2, low(maxs)).
table(low_minc_mins_maxs, 21, 3, 2, low(maxs)).
table(low_maxc_s_maxs, 21, 3, 2, low(maxs)).
table(low_maxc_mins_maxs, 21, 3, 2, low(maxs)).
table(low_s_mins_maxs, 21, 7, 2, low(maxs)).
table(low_v_c_minc_maxs, 21, 11, 3, low(maxs)).
table(low_v_c_maxc_maxs, 21, 9, 3, low(maxs)).
table(low_v_c_s_maxs, 21, 4, 3, low(maxs)).
table(low_v_c_mins_maxs, 21, 4, 3, low(maxs)).
table(low_v_minc_maxc_maxs, 21, 10, 3, low(maxs)).
table(low_v_minc_s_maxs, 21, 4, 3, low(maxs)).
table(low_v_minc_mins_maxs, 21, 4, 3, low(maxs)).
table(low_v_maxc_s_maxs, 21, 4, 3, low(maxs)).
table(low_v_maxc_mins_maxs, 21, 4, 3, low(maxs)).
table(low_v_s_mins_maxs, 21, 4, 3, low(maxs)).
table(low_c_minc_maxc_maxs, 21, 7, 3, low(maxs)).
table(low_c_minc_s_maxs, 21, 4, 3, low(maxs)).
table(low_c_minc_mins_maxs, 21, 4, 3, low(maxs)).
table(low_c_maxc_s_maxs, 21, 4, 3, low(maxs)).
table(low_c_maxc_mins_maxs, 21, 4, 3, low(maxs)).
table(low_c_s_mins_maxs, 21, 8, 3, low(maxs)).
table(low_minc_maxc_s_maxs, 21, 5, 3, low(maxs)).
table(low_minc_maxc_mins_maxs, 21, 5, 3, low(maxs)).
table(low_minc_s_mins_maxs, 21, 4, 3, low(maxs)).
table(low_maxc_s_mins_maxs, 21, 4, 3, low(maxs)).
table(low_v_c_minc_maxc_maxs, 21, 12, 4, low(maxs)).
table(low_v_c_minc_s_maxs, 21, 7, 4, low(maxs)).
table(low_v_c_minc_mins_maxs, 21, 7, 4, low(maxs)).
table(low_v_c_maxc_s_maxs, 21, 5, 4, low(maxs)).
table(low_v_c_maxc_mins_maxs, 21, 5, 4, low(maxs)).
table(low_v_c_s_mins_maxs, 21, 5, 4, low(maxs)).
table(low_v_minc_maxc_s_maxs, 21, 6, 4, low(maxs)).
table(low_v_minc_maxc_mins_maxs, 21, 6, 4, low(maxs)).
table(low_v_minc_s_mins_maxs, 21, 5, 4, low(maxs)).
table(low_v_maxc_s_mins_maxs, 21, 5, 4, low(maxs)).
table(low_c_minc_maxc_s_maxs, 21, 6, 4, low(maxs)).
table(low_c_minc_maxc_mins_maxs, 21, 6, 4, low(maxs)).
table(low_c_minc_s_mins_maxs, 21, 5, 4, low(maxs)).
table(low_c_maxc_s_mins_maxs, 21, 5, 4, low(maxs)).
table(low_minc_maxc_s_mins_maxs, 21, 6, 4, low(maxs)).
table(low_v_c_minc_maxc_s_maxs, 21, 8, 5, low(maxs)).
table(low_v_c_minc_maxc_mins_maxs, 21, 8, 5, low(maxs)).
table(low_v_c_minc_s_mins_maxs, 21, 8, 5, low(maxs)).
table(low_v_c_maxc_s_mins_maxs, 21, 6, 5, low(maxs)).
table(low_v_minc_maxc_s_mins_maxs, 21, 7, 5, low(maxs)).
table(low_c_minc_maxc_s_mins_maxs, 21, 7, 5, low(maxs)).
table(low_v_c_minc_maxc_s_mins_maxs, 21, 9, 6, low(maxs)).
table(up_v_maxs, 21, 18, 1, up(maxs)).
table(up_maxc_maxs, 21, 2, 1, up(maxs)).
table(up_v_c_maxs, 21, 18, 2, up(maxs)).
table(up_v_minc_maxs, 21, 12, 2, up(maxs)).
table(up_v_maxc_maxs, 21, 3, 2, up(maxs)).
table(up_v_s_maxs, 21, 7, 2, up(maxs)).
table(up_v_mins_maxs, 21, 9, 2, up(maxs)).
table(up_c_maxc_maxs, 21, 3, 2, up(maxs)).
table(up_minc_maxc_maxs, 21, 4, 2, up(maxs)).
table(up_maxc_s_maxs, 21, 3, 2, up(maxs)).
table(up_maxc_mins_maxs, 21, 3, 2, up(maxs)).
table(up_v_c_minc_maxs, 21, 12, 3, up(maxs)).
table(up_v_c_maxc_maxs, 21, 4, 3, up(maxs)).
table(up_v_c_s_maxs, 21, 8, 3, up(maxs)).
table(up_v_c_mins_maxs, 21, 18, 3, up(maxs)).
table(up_v_minc_maxc_maxs, 21, 5, 3, up(maxs)).
table(up_v_minc_s_maxs, 21, 5, 3, up(maxs)).
table(up_v_minc_mins_maxs, 21, 14, 3, up(maxs)).
table(up_v_maxc_s_maxs, 21, 4, 3, up(maxs)).
table(up_v_maxc_mins_maxs, 21, 4, 3, up(maxs)).
table(up_v_s_mins_maxs, 21, 7, 3, up(maxs)).
table(up_c_minc_maxc_maxs, 21, 5, 3, up(maxs)).
table(up_c_maxc_s_maxs, 21, 4, 3, up(maxs)).
table(up_c_maxc_mins_maxs, 21, 4, 3, up(maxs)).
table(up_minc_maxc_s_maxs, 21, 5, 3, up(maxs)).
table(up_minc_maxc_mins_maxs, 21, 5, 3, up(maxs)).
table(up_maxc_s_mins_maxs, 21, 4, 3, up(maxs)).
table(up_v_c_minc_maxc_maxs, 21, 7, 4, up(maxs)).
table(up_v_c_minc_s_maxs, 21, 7, 4, up(maxs)).
table(up_v_c_minc_mins_maxs, 21, 13, 4, up(maxs)).
table(up_v_c_maxc_s_maxs, 21, 5, 4, up(maxs)).
table(up_v_c_maxc_mins_maxs, 21, 5, 4, up(maxs)).
table(up_v_c_s_mins_maxs, 21, 8, 4, up(maxs)).
table(up_v_minc_maxc_s_maxs, 21, 6, 4, up(maxs)).
table(up_v_minc_maxc_mins_maxs, 21, 6, 4, up(maxs)).
table(up_v_minc_s_mins_maxs, 21, 5, 4, up(maxs)).
table(up_v_maxc_s_mins_maxs, 21, 5, 4, up(maxs)).
table(up_c_minc_maxc_s_maxs, 21, 6, 4, up(maxs)).
table(up_c_minc_maxc_mins_maxs, 21, 6, 4, up(maxs)).
table(up_c_maxc_s_mins_maxs, 21, 5, 4, up(maxs)).
table(up_minc_maxc_s_mins_maxs, 21, 6, 4, up(maxs)).
table(up_v_c_minc_maxc_s_maxs, 21, 8, 5, up(maxs)).
table(up_v_c_minc_maxc_mins_maxs, 21, 8, 5, up(maxs)).
table(up_v_c_minc_s_mins_maxs, 21, 8, 5, up(maxs)).
table(up_v_c_maxc_s_mins_maxs, 21, 6, 5, up(maxs)).
table(up_v_minc_maxc_s_mins_maxs, 21, 7, 5, up(maxs)).
table(up_c_minc_maxc_s_mins_maxs, 21, 7, 5, up(maxs)).
table(up_v_c_minc_maxc_s_mins_maxs, 21, 9, 6, up(maxs)).
table(low_v_mina, 21, 12, 1, low(mina)).
table(low_c_mina, 21, 7, 1, low(mina)).
table(low_minc_mina, 21, 19, 1, low(mina)).
table(low_maxc_mina, 21, 19, 1, low(mina)).
table(low_s_mina, 21, 12, 1, low(mina)).
table(low_mins_mina, 21, 11, 1, low(mina)).
table(low_maxs_mina, 21, 11, 1, low(mina)).
table(low_v_c_mina, 21, 10, 2, low(mina)).
table(low_v_minc_mina, 21, 11, 2, low(mina)).
table(low_v_maxc_mina, 21, 11, 2, low(mina)).
table(low_v_s_mina, 21, 4, 2, low(mina)).
table(low_v_mins_mina, 21, 4, 2, low(mina)).
table(low_v_maxs_mina, 21, 10, 2, low(mina)).
table(low_c_minc_mina, 21, 8, 2, low(mina)).
table(low_c_maxc_mina, 21, 6, 2, low(mina)).
table(low_c_s_mina, 21, 10, 2, low(mina)).
table(low_c_mins_mina, 21, 7, 2, low(mina)).
table(low_c_maxs_mina, 21, 5, 2, low(mina)).
table(low_minc_maxc_mina, 21, 19, 2, low(mina)).
table(low_minc_s_mina, 21, 11, 2, low(mina)).
table(low_minc_mins_mina, 21, 19, 2, low(mina)).
table(low_minc_maxs_mina, 21, 19, 2, low(mina)).
table(low_maxc_s_mina, 21, 11, 2, low(mina)).
table(low_maxc_mins_mina, 21, 19, 2, low(mina)).
table(low_maxc_maxs_mina, 21, 19, 2, low(mina)).
table(low_s_mins_mina, 21, 12, 2, low(mina)).
table(low_s_maxs_mina, 21, 10, 2, low(mina)).
table(low_mins_maxs_mina, 21, 6, 2, low(mina)).
table(low_v_c_minc_mina, 21, 13, 3, low(mina)).
table(low_v_c_maxc_mina, 21, 11, 3, low(mina)).
table(low_v_c_s_mina, 21, 4, 3, low(mina)).
table(low_v_c_mins_mina, 21, 7, 3, low(mina)).
table(low_v_c_maxs_mina, 21, 10, 3, low(mina)).
table(low_v_minc_maxc_mina, 21, 12, 3, low(mina)).
table(low_v_minc_s_mina, 21, 4, 3, low(mina)).
table(low_v_minc_mins_mina, 21, 4, 3, low(mina)).
table(low_v_minc_maxs_mina, 21, 11, 3, low(mina)).
table(low_v_maxc_s_mina, 21, 4, 3, low(mina)).
table(low_v_maxc_mins_mina, 21, 6, 3, low(mina)).
table(low_v_maxc_maxs_mina, 21, 10, 3, low(mina)).
table(low_v_s_mins_mina, 21, 5, 3, low(mina)).
table(low_v_s_maxs_mina, 21, 5, 3, low(mina)).
table(low_v_mins_maxs_mina, 21, 5, 3, low(mina)).
table(low_c_minc_maxc_mina, 21, 8, 3, low(mina)).
table(low_c_minc_s_mina, 21, 11, 3, low(mina)).
table(low_c_minc_mins_mina, 21, 8, 3, low(mina)).
table(low_c_minc_maxs_mina, 21, 6, 3, low(mina)).
table(low_c_maxc_s_mina, 21, 8, 3, low(mina)).
table(low_c_maxc_mins_mina, 21, 6, 3, low(mina)).
table(low_c_maxc_maxs_mina, 21, 6, 3, low(mina)).
table(low_c_s_mins_mina, 21, 10, 3, low(mina)).
table(low_c_s_maxs_mina, 21, 10, 3, low(mina)).
table(low_c_mins_maxs_mina, 21, 5, 3, low(mina)).
table(low_minc_maxc_s_mina, 21, 10, 3, low(mina)).
table(low_minc_maxc_mins_mina, 21, 16, 3, low(mina)).
table(low_minc_maxc_maxs_mina, 21, 18, 3, low(mina)).
table(low_minc_s_mins_mina, 21, 6, 3, low(mina)).
table(low_minc_s_maxs_mina, 21, 8, 3, low(mina)).
table(low_minc_mins_maxs_mina, 21, 17, 3, low(mina)).
table(low_maxc_s_mins_mina, 21, 8, 3, low(mina)).
table(low_maxc_s_maxs_mina, 21, 7, 3, low(mina)).
table(low_maxc_mins_maxs_mina, 21, 8, 3, low(mina)).
table(low_s_mins_maxs_mina, 21, 10, 3, low(mina)).
table(low_v_c_minc_maxc_mina, 21, 14, 4, low(mina)).
table(low_v_c_minc_s_mina, 21, 8, 4, low(mina)).
table(low_v_c_minc_mins_mina, 21, 10, 4, low(mina)).
table(low_v_c_minc_maxs_mina, 21, 13, 4, low(mina)).
table(low_v_c_maxc_s_mina, 21, 5, 4, low(mina)).
table(low_v_c_maxc_mins_mina, 21, 8, 4, low(mina)).
table(low_v_c_maxc_maxs_mina, 21, 11, 4, low(mina)).
table(low_v_c_s_mins_mina, 21, 5, 4, low(mina)).
table(low_v_c_s_maxs_mina, 21, 5, 4, low(mina)).
table(low_v_c_mins_maxs_mina, 21, 8, 4, low(mina)).
table(low_v_minc_maxc_s_mina, 21, 6, 4, low(mina)).
table(low_v_minc_maxc_mins_mina, 21, 6, 4, low(mina)).
table(low_v_minc_maxc_maxs_mina, 21, 12, 4, low(mina)).
table(low_v_minc_s_mins_mina, 21, 5, 4, low(mina)).
table(low_v_minc_s_maxs_mina, 21, 5, 4, low(mina)).
table(low_v_minc_mins_maxs_mina, 21, 5, 4, low(mina)).
table(low_v_maxc_s_mins_mina, 21, 5, 4, low(mina)).
table(low_v_maxc_s_maxs_mina, 21, 5, 4, low(mina)).
table(low_v_maxc_mins_maxs_mina, 21, 5, 4, low(mina)).
table(low_v_s_mins_maxs_mina, 21, 6, 4, low(mina)).
table(low_c_minc_maxc_s_mina, 21, 9, 4, low(mina)).
table(low_c_minc_maxc_mins_mina, 21, 7, 4, low(mina)).
table(low_c_minc_maxc_maxs_mina, 21, 8, 4, low(mina)).
table(low_c_minc_s_mins_mina, 21, 10, 4, low(mina)).
table(low_c_minc_s_maxs_mina, 21, 10, 4, low(mina)).
table(low_c_minc_mins_maxs_mina, 21, 6, 4, low(mina)).
table(low_c_maxc_s_mins_mina, 21, 6, 4, low(mina)).
table(low_c_maxc_s_maxs_mina, 21, 6, 4, low(mina)).
table(low_c_maxc_mins_maxs_mina, 21, 6, 4, low(mina)).
table(low_c_s_mins_maxs_mina, 21, 10, 4, low(mina)).
table(low_minc_maxc_s_mins_mina, 21, 7, 4, low(mina)).
table(low_minc_maxc_s_maxs_mina, 21, 9, 4, low(mina)).
table(low_minc_maxc_mins_maxs_mina, 21, 16, 4, low(mina)).
table(low_minc_s_mins_maxs_mina, 21, 6, 4, low(mina)).
table(low_maxc_s_mins_maxs_mina, 21, 6, 4, low(mina)).
table(low_v_c_minc_maxc_s_mina, 21, 9, 5, low(mina)).
table(low_v_c_minc_maxc_mins_mina, 21, 11, 5, low(mina)).
table(low_v_c_minc_maxc_maxs_mina, 21, 14, 5, low(mina)).
table(low_v_c_minc_s_mins_mina, 21, 8, 5, low(mina)).
table(low_v_c_minc_s_maxs_mina, 21, 8, 5, low(mina)).
table(low_v_c_minc_mins_maxs_mina, 21, 11, 5, low(mina)).
table(low_v_c_maxc_s_mins_mina, 21, 6, 5, low(mina)).
table(low_v_c_maxc_s_maxs_mina, 21, 6, 5, low(mina)).
table(low_v_c_maxc_mins_maxs_mina, 21, 9, 5, low(mina)).
table(low_v_c_s_mins_maxs_mina, 21, 6, 5, low(mina)).
table(low_v_minc_maxc_s_mins_mina, 21, 7, 5, low(mina)).
table(low_v_minc_maxc_s_maxs_mina, 21, 7, 5, low(mina)).
table(low_v_minc_maxc_mins_maxs_mina, 21, 7, 5, low(mina)).
table(low_v_minc_s_mins_maxs_mina, 21, 6, 5, low(mina)).
table(low_v_maxc_s_mins_maxs_mina, 21, 6, 5, low(mina)).
table(low_c_minc_maxc_s_mins_mina, 21, 8, 5, low(mina)).
table(low_c_minc_maxc_s_maxs_mina, 21, 8, 5, low(mina)).
table(low_c_minc_maxc_mins_maxs_mina, 21, 8, 5, low(mina)).
table(low_c_minc_s_mins_maxs_mina, 21, 11, 5, low(mina)).
table(low_c_maxc_s_mins_maxs_mina, 21, 7, 5, low(mina)).
table(low_minc_maxc_s_mins_maxs_mina, 21, 8, 5, low(mina)).
table(low_v_c_minc_maxc_s_mins_mina, 21, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_maxs_mina, 21, 9, 6, low(mina)).
table(low_v_c_minc_maxc_mins_maxs_mina, 21, 12, 6, low(mina)).
table(low_v_c_minc_s_mins_maxs_mina, 21, 9, 6, low(mina)).
table(low_v_c_maxc_s_mins_maxs_mina, 21, 7, 6, low(mina)).
table(low_v_minc_maxc_s_mins_maxs_mina, 21, 8, 6, low(mina)).
table(low_c_minc_maxc_s_mins_maxs_mina, 21, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_mins_maxs_mina, 21, 10, 7, low(mina)).
table(up_v_maxa, 21, 19, 1, up(maxa)).
table(up_v_c_maxa, 21, 19, 2, up(maxa)).
table(up_v_minc_maxa, 21, 19, 2, up(maxa)).
table(up_v_maxc_maxa, 21, 19, 2, up(maxa)).
table(up_v_s_maxa, 21, 19, 2, up(maxa)).
table(up_v_mins_maxa, 21, 19, 2, up(maxa)).
table(up_v_maxs_maxa, 21, 19, 2, up(maxa)).
table(up_c_maxc_maxa, 21, 19, 2, up(maxa)).
table(up_maxc_s_maxa, 21, 18, 2, up(maxa)).
table(up_s_maxs_maxa, 21, 19, 2, up(maxa)).
table(up_v_c_minc_maxa, 21, 19, 3, up(maxa)).
table(up_v_c_maxc_maxa, 21, 19, 3, up(maxa)).
table(up_v_c_s_maxa, 21, 19, 3, up(maxa)).
table(up_v_c_mins_maxa, 21, 19, 3, up(maxa)).
table(up_v_c_maxs_maxa, 21, 7, 3, up(maxa)).
table(up_v_minc_maxc_maxa, 21, 19, 3, up(maxa)).
table(up_v_minc_s_maxa, 21, 18, 3, up(maxa)).
table(up_v_minc_mins_maxa, 21, 19, 3, up(maxa)).
table(up_v_minc_maxs_maxa, 21, 19, 3, up(maxa)).
table(up_v_maxc_s_maxa, 21, 18, 3, up(maxa)).
table(up_v_maxc_mins_maxa, 21, 11, 3, up(maxa)).
table(up_v_maxc_maxs_maxa, 21, 6, 3, up(maxa)).
table(up_v_s_mins_maxa, 21, 19, 3, up(maxa)).
table(up_v_s_maxs_maxa, 21, 19, 3, up(maxa)).
table(up_v_mins_maxs_maxa, 21, 19, 3, up(maxa)).
table(up_c_minc_maxc_maxa, 21, 19, 3, up(maxa)).
table(up_c_maxc_s_maxa, 21, 18, 3, up(maxa)).
table(up_c_maxc_mins_maxa, 21, 19, 3, up(maxa)).
table(up_c_maxc_maxs_maxa, 21, 6, 3, up(maxa)).
table(up_c_s_maxs_maxa, 21, 19, 3, up(maxa)).
table(up_minc_maxc_s_maxa, 21, 5, 3, up(maxa)).
table(up_minc_s_maxs_maxa, 21, 18, 3, up(maxa)).
table(up_maxc_s_mins_maxa, 21, 8, 3, up(maxa)).
table(up_maxc_s_maxs_maxa, 21, 4, 3, up(maxa)).
table(up_s_mins_maxs_maxa, 21, 19, 3, up(maxa)).
table(up_v_c_minc_maxc_maxa, 21, 19, 4, up(maxa)).
table(up_v_c_minc_s_maxa, 21, 18, 4, up(maxa)).
table(up_v_c_minc_mins_maxa, 21, 11, 4, up(maxa)).
table(up_v_c_minc_maxs_maxa, 21, 8, 4, up(maxa)).
table(up_v_c_maxc_s_maxa, 21, 18, 4, up(maxa)).
table(up_v_c_maxc_mins_maxa, 21, 11, 4, up(maxa)).
table(up_v_c_maxc_maxs_maxa, 21, 6, 4, up(maxa)).
table(up_v_c_s_mins_maxa, 21, 19, 4, up(maxa)).
table(up_v_c_s_maxs_maxa, 21, 19, 4, up(maxa)).
table(up_v_c_mins_maxs_maxa, 21, 9, 4, up(maxa)).
table(up_v_minc_maxc_s_maxa, 21, 18, 4, up(maxa)).
table(up_v_minc_maxc_mins_maxa, 21, 14, 4, up(maxa)).
table(up_v_minc_maxc_maxs_maxa, 21, 9, 4, up(maxa)).
table(up_v_minc_s_mins_maxa, 21, 18, 4, up(maxa)).
table(up_v_minc_s_maxs_maxa, 21, 18, 4, up(maxa)).
table(up_v_minc_mins_maxs_maxa, 21, 16, 4, up(maxa)).
table(up_v_maxc_s_mins_maxa, 21, 8, 4, up(maxa)).
table(up_v_maxc_s_maxs_maxa, 21, 7, 4, up(maxa)).
table(up_v_maxc_mins_maxs_maxa, 21, 5, 4, up(maxa)).
table(up_v_s_mins_maxs_maxa, 21, 19, 4, up(maxa)).
table(up_c_minc_maxc_s_maxa, 21, 18, 4, up(maxa)).
table(up_c_minc_maxc_mins_maxa, 21, 14, 4, up(maxa)).
table(up_c_minc_maxc_maxs_maxa, 21, 9, 4, up(maxa)).
table(up_c_minc_s_maxs_maxa, 21, 9, 4, up(maxa)).
table(up_c_maxc_s_mins_maxa, 21, 8, 4, up(maxa)).
table(up_c_maxc_s_maxs_maxa, 21, 6, 4, up(maxa)).
table(up_c_maxc_mins_maxs_maxa, 21, 5, 4, up(maxa)).
table(up_c_s_mins_maxs_maxa, 21, 19, 4, up(maxa)).
table(up_minc_maxc_s_mins_maxa, 21, 6, 4, up(maxa)).
table(up_minc_maxc_s_maxs_maxa, 21, 6, 4, up(maxa)).
table(up_minc_s_mins_maxs_maxa, 21, 18, 4, up(maxa)).
table(up_maxc_s_mins_maxs_maxa, 21, 5, 4, up(maxa)).
table(up_v_c_minc_maxc_s_maxa, 21, 18, 5, up(maxa)).
table(up_v_c_minc_maxc_mins_maxa, 21, 14, 5, up(maxa)).
table(up_v_c_minc_maxc_maxs_maxa, 21, 9, 5, up(maxa)).
table(up_v_c_minc_s_mins_maxa, 21, 10, 5, up(maxa)).
table(up_v_c_minc_s_maxs_maxa, 21, 9, 5, up(maxa)).
table(up_v_c_minc_mins_maxs_maxa, 21, 8, 5, up(maxa)).
table(up_v_c_maxc_s_mins_maxa, 21, 8, 5, up(maxa)).
table(up_v_c_maxc_s_maxs_maxa, 21, 6, 5, up(maxa)).
table(up_v_c_maxc_mins_maxs_maxa, 21, 6, 5, up(maxa)).
table(up_v_c_s_mins_maxs_maxa, 21, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_mins_maxa, 21, 11, 5, up(maxa)).
table(up_v_minc_maxc_s_maxs_maxa, 21, 10, 5, up(maxa)).
table(up_v_minc_maxc_mins_maxs_maxa, 21, 7, 5, up(maxa)).
table(up_v_minc_s_mins_maxs_maxa, 21, 16, 5, up(maxa)).
table(up_v_maxc_s_mins_maxs_maxa, 21, 8, 5, up(maxa)).
table(up_c_minc_maxc_s_mins_maxa, 21, 11, 5, up(maxa)).
table(up_c_minc_maxc_s_maxs_maxa, 21, 9, 5, up(maxa)).
table(up_c_minc_maxc_mins_maxs_maxa, 21, 7, 5, up(maxa)).
table(up_c_minc_s_mins_maxs_maxa, 21, 10, 5, up(maxa)).
table(up_c_maxc_s_mins_maxs_maxa, 21, 6, 5, up(maxa)).
table(up_minc_maxc_s_mins_maxs_maxa, 21, 7, 5, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxa, 21, 11, 6, up(maxa)).
table(up_v_c_minc_maxc_s_maxs_maxa, 21, 9, 6, up(maxa)).
table(up_v_c_minc_maxc_mins_maxs_maxa, 21, 9, 6, up(maxa)).
table(up_v_c_minc_s_mins_maxs_maxa, 21, 10, 6, up(maxa)).
table(up_v_c_maxc_s_mins_maxs_maxa, 21, 7, 6, up(maxa)).
table(up_v_minc_maxc_s_mins_maxs_maxa, 21, 11, 6, up(maxa)).
table(up_c_minc_maxc_s_mins_maxs_maxa, 21, 10, 6, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxs_maxa, 21, 10, 7, up(maxa)).
table(low_c_v, 22, 18, 1, low(v)).
table(low_minc_v, 22, 11, 1, low(v)).
table(low_maxc_v, 22, 11, 1, low(v)).
table(low_s_v, 22, 7, 1, low(v)).
table(low_mins_v, 22, 18, 1, low(v)).
table(low_maxs_v, 22, 18, 1, low(v)).
table(low_c_minc_v, 22, 11, 2, low(v)).
table(low_c_maxc_v, 22, 11, 2, low(v)).
table(low_c_s_v, 22, 8, 2, low(v)).
table(low_c_mins_v, 22, 18, 2, low(v)).
table(low_c_maxs_v, 22, 18, 2, low(v)).
table(low_minc_maxc_v, 22, 11, 2, low(v)).
table(low_minc_s_v, 22, 3, 2, low(v)).
table(low_minc_mins_v, 22, 12, 2, low(v)).
table(low_minc_maxs_v, 22, 12, 2, low(v)).
table(low_maxc_s_v, 22, 3, 2, low(v)).
table(low_maxc_mins_v, 22, 12, 2, low(v)).
table(low_maxc_maxs_v, 22, 13, 2, low(v)).
table(low_s_mins_v, 22, 7, 2, low(v)).
table(low_s_maxs_v, 22, 7, 2, low(v)).
table(low_mins_maxs_v, 22, 12, 2, low(v)).
table(low_c_minc_maxc_v, 22, 11, 3, low(v)).
table(low_c_minc_s_v, 22, 6, 3, low(v)).
table(low_c_minc_mins_v, 22, 12, 3, low(v)).
table(low_c_minc_maxs_v, 22, 12, 3, low(v)).
table(low_c_maxc_s_v, 22, 4, 3, low(v)).
table(low_c_maxc_mins_v, 22, 12, 3, low(v)).
table(low_c_maxc_maxs_v, 22, 13, 3, low(v)).
table(low_c_s_mins_v, 22, 8, 3, low(v)).
table(low_c_s_maxs_v, 22, 8, 3, low(v)).
table(low_c_mins_maxs_v, 22, 18, 3, low(v)).
table(low_minc_maxc_s_v, 22, 5, 3, low(v)).
table(low_minc_maxc_mins_v, 22, 12, 3, low(v)).
table(low_minc_maxc_maxs_v, 22, 12, 3, low(v)).
table(low_minc_s_mins_v, 22, 4, 3, low(v)).
table(low_minc_s_maxs_v, 22, 4, 3, low(v)).
table(low_minc_mins_maxs_v, 22, 13, 3, low(v)).
table(low_maxc_s_mins_v, 22, 4, 3, low(v)).
table(low_maxc_s_maxs_v, 22, 4, 3, low(v)).
table(low_maxc_mins_maxs_v, 22, 14, 3, low(v)).
table(low_s_mins_maxs_v, 22, 7, 3, low(v)).
table(low_c_minc_maxc_s_v, 22, 7, 4, low(v)).
table(low_c_minc_maxc_mins_v, 22, 12, 4, low(v)).
table(low_c_minc_maxc_maxs_v, 22, 12, 4, low(v)).
table(low_c_minc_s_mins_v, 22, 7, 4, low(v)).
table(low_c_minc_s_maxs_v, 22, 7, 4, low(v)).
table(low_c_minc_mins_maxs_v, 22, 13, 4, low(v)).
table(low_c_maxc_s_mins_v, 22, 5, 4, low(v)).
table(low_c_maxc_s_maxs_v, 22, 5, 4, low(v)).
table(low_c_maxc_mins_maxs_v, 22, 14, 4, low(v)).
table(low_c_s_mins_maxs_v, 22, 8, 4, low(v)).
table(low_minc_maxc_s_mins_v, 22, 6, 4, low(v)).
table(low_minc_maxc_s_maxs_v, 22, 6, 4, low(v)).
table(low_minc_maxc_mins_maxs_v, 22, 13, 4, low(v)).
table(low_minc_s_mins_maxs_v, 22, 5, 4, low(v)).
table(low_maxc_s_mins_maxs_v, 22, 5, 4, low(v)).
table(low_c_minc_maxc_s_mins_v, 22, 8, 5, low(v)).
table(low_c_minc_maxc_s_maxs_v, 22, 8, 5, low(v)).
table(low_c_minc_maxc_mins_maxs_v, 22, 13, 5, low(v)).
table(low_c_minc_s_mins_maxs_v, 22, 8, 5, low(v)).
table(low_c_maxc_s_mins_maxs_v, 22, 6, 5, low(v)).
table(low_minc_maxc_s_mins_maxs_v, 22, 7, 5, low(v)).
table(low_c_minc_maxc_s_mins_maxs_v, 22, 9, 6, low(v)).
table(up_c_maxc_v, 22, 3, 2, up(v)).
table(up_maxc_s_v, 22, 3, 2, up(v)).
table(up_s_maxs_v, 22, 3, 2, up(v)).
table(up_c_minc_maxc_v, 22, 6, 3, up(v)).
table(up_c_maxc_s_v, 22, 4, 3, up(v)).
table(up_c_maxc_mins_v, 22, 4, 3, up(v)).
table(up_c_maxc_maxs_v, 22, 4, 3, up(v)).
table(up_c_s_maxs_v, 22, 4, 3, up(v)).
table(up_minc_maxc_s_v, 22, 5, 3, up(v)).
table(up_minc_s_maxs_v, 22, 4, 3, up(v)).
table(up_maxc_s_mins_v, 22, 4, 3, up(v)).
table(up_maxc_s_maxs_v, 22, 4, 3, up(v)).
table(up_s_mins_maxs_v, 22, 4, 3, up(v)).
table(up_c_minc_maxc_s_v, 22, 7, 4, up(v)).
table(up_c_minc_maxc_mins_v, 22, 7, 4, up(v)).
table(up_c_minc_maxc_maxs_v, 22, 7, 4, up(v)).
table(up_c_minc_s_maxs_v, 22, 7, 4, up(v)).
table(up_c_maxc_s_mins_v, 22, 5, 4, up(v)).
table(up_c_maxc_s_maxs_v, 22, 5, 4, up(v)).
table(up_c_maxc_mins_maxs_v, 22, 5, 4, up(v)).
table(up_c_s_mins_maxs_v, 22, 5, 4, up(v)).
table(up_minc_maxc_s_mins_v, 22, 6, 4, up(v)).
table(up_minc_maxc_s_maxs_v, 22, 6, 4, up(v)).
table(up_minc_s_mins_maxs_v, 22, 5, 4, up(v)).
table(up_maxc_s_mins_maxs_v, 22, 5, 4, up(v)).
table(up_c_minc_maxc_s_mins_v, 22, 8, 5, up(v)).
table(up_c_minc_maxc_s_maxs_v, 22, 8, 5, up(v)).
table(up_c_minc_maxc_mins_maxs_v, 22, 8, 5, up(v)).
table(up_c_minc_s_mins_maxs_v, 22, 8, 5, up(v)).
table(up_c_maxc_s_mins_maxs_v, 22, 6, 5, up(v)).
table(up_minc_maxc_s_mins_maxs_v, 22, 7, 5, up(v)).
table(up_c_minc_maxc_s_mins_maxs_v, 22, 9, 6, up(v)).
table(low_v_c, 22, 11, 1, low(c)).
table(low_minc_c, 22, 11, 1, low(c)).
table(low_maxc_c, 22, 11, 1, low(c)).
table(low_s_c, 22, 8, 1, low(c)).
table(low_mins_c, 22, 8, 1, low(c)).
table(low_maxs_c, 22, 8, 1, low(c)).
table(low_v_minc_c, 22, 11, 2, low(c)).
table(low_v_maxc_c, 22, 3, 2, low(c)).
table(low_v_s_c, 22, 13, 2, low(c)).
table(low_v_mins_c, 22, 12, 2, low(c)).
table(low_v_maxs_c, 22, 13, 2, low(c)).
table(low_minc_maxc_c, 22, 11, 2, low(c)).
table(low_minc_s_c, 22, 5, 2, low(c)).
table(low_minc_mins_c, 22, 9, 2, low(c)).
table(low_minc_maxs_c, 22, 9, 2, low(c)).
table(low_maxc_s_c, 22, 3, 2, low(c)).
table(low_maxc_mins_c, 22, 12, 2, low(c)).
table(low_maxc_maxs_c, 22, 13, 2, low(c)).
table(low_s_mins_c, 22, 10, 2, low(c)).
table(low_s_maxs_c, 22, 11, 2, low(c)).
table(low_mins_maxs_c, 22, 9, 2, low(c)).
table(low_v_minc_maxc_c, 22, 6, 3, low(c)).
table(low_v_minc_s_c, 22, 12, 3, low(c)).
table(low_v_minc_mins_c, 22, 12, 3, low(c)).
table(low_v_minc_maxs_c, 22, 12, 3, low(c)).
table(low_v_maxc_s_c, 22, 4, 3, low(c)).
table(low_v_maxc_mins_c, 22, 4, 3, low(c)).
table(low_v_maxc_maxs_c, 22, 4, 3, low(c)).
table(low_v_s_mins_c, 22, 14, 3, low(c)).
table(low_v_s_maxs_c, 22, 14, 3, low(c)).
table(low_v_mins_maxs_c, 22, 14, 3, low(c)).
table(low_minc_maxc_s_c, 22, 5, 3, low(c)).
table(low_minc_maxc_mins_c, 22, 12, 3, low(c)).
table(low_minc_maxc_maxs_c, 22, 12, 3, low(c)).
table(low_minc_s_mins_c, 22, 6, 3, low(c)).
table(low_minc_s_maxs_c, 22, 7, 3, low(c)).
table(low_minc_mins_maxs_c, 22, 7, 3, low(c)).
table(low_maxc_s_mins_c, 22, 4, 3, low(c)).
table(low_maxc_s_maxs_c, 22, 4, 3, low(c)).
table(low_maxc_mins_maxs_c, 22, 6, 3, low(c)).
table(low_s_mins_maxs_c, 22, 12, 3, low(c)).
table(low_v_minc_maxc_s_c, 22, 7, 4, low(c)).
table(low_v_minc_maxc_mins_c, 22, 7, 4, low(c)).
table(low_v_minc_maxc_maxs_c, 22, 7, 4, low(c)).
table(low_v_minc_s_mins_c, 22, 13, 4, low(c)).
table(low_v_minc_s_maxs_c, 22, 13, 4, low(c)).
table(low_v_minc_mins_maxs_c, 22, 13, 4, low(c)).
table(low_v_maxc_s_mins_c, 22, 5, 4, low(c)).
table(low_v_maxc_s_maxs_c, 22, 5, 4, low(c)).
table(low_v_maxc_mins_maxs_c, 22, 5, 4, low(c)).
table(low_v_s_mins_maxs_c, 22, 15, 4, low(c)).
table(low_minc_maxc_s_mins_c, 22, 6, 4, low(c)).
table(low_minc_maxc_s_maxs_c, 22, 6, 4, low(c)).
table(low_minc_maxc_mins_maxs_c, 22, 8, 4, low(c)).
table(low_minc_s_mins_maxs_c, 22, 8, 4, low(c)).
table(low_maxc_s_mins_maxs_c, 22, 5, 4, low(c)).
table(low_v_minc_maxc_s_mins_c, 22, 8, 5, low(c)).
table(low_v_minc_maxc_s_maxs_c, 22, 8, 5, low(c)).
table(low_v_minc_maxc_mins_maxs_c, 22, 8, 5, low(c)).
table(low_v_minc_s_mins_maxs_c, 22, 14, 5, low(c)).
table(low_v_maxc_s_mins_maxs_c, 22, 6, 5, low(c)).
table(low_minc_maxc_s_mins_maxs_c, 22, 7, 5, low(c)).
table(low_v_minc_maxc_s_mins_maxs_c, 22, 9, 6, low(c)).
table(up_v_c, 22, 18, 1, up(c)).
table(up_s_c, 22, 3, 1, up(c)).
table(up_v_minc_c, 22, 6, 2, up(c)).
table(up_v_maxc_c, 22, 11, 2, up(c)).
table(up_v_s_c, 22, 4, 2, up(c)).
table(up_v_mins_c, 22, 10, 2, up(c)).
table(up_v_maxs_c, 22, 18, 2, up(c)).
table(up_minc_s_c, 22, 3, 2, up(c)).
table(up_maxc_s_c, 22, 4, 2, up(c)).
table(up_s_mins_c, 22, 5, 2, up(c)).
table(up_s_maxs_c, 22, 5, 2, up(c)).
table(up_v_minc_maxc_c, 22, 7, 3, up(c)).
table(up_v_minc_s_c, 22, 6, 3, up(c)).
table(up_v_minc_mins_c, 22, 7, 3, up(c)).
table(up_v_minc_maxs_c, 22, 7, 3, up(c)).
table(up_v_maxc_s_c, 22, 5, 3, up(c)).
table(up_v_maxc_mins_c, 22, 5, 3, up(c)).
table(up_v_maxc_maxs_c, 22, 13, 3, up(c)).
table(up_v_s_mins_c, 22, 8, 3, up(c)).
table(up_v_s_maxs_c, 22, 6, 3, up(c)).
table(up_v_mins_maxs_c, 22, 12, 3, up(c)).
table(up_minc_maxc_s_c, 22, 5, 3, up(c)).
table(up_minc_s_mins_c, 22, 4, 3, up(c)).
table(up_minc_s_maxs_c, 22, 4, 3, up(c)).
table(up_maxc_s_mins_c, 22, 5, 3, up(c)).
table(up_maxc_s_maxs_c, 22, 5, 3, up(c)).
table(up_s_mins_maxs_c, 22, 8, 3, up(c)).
table(up_v_minc_maxc_s_c, 22, 7, 4, up(c)).
table(up_v_minc_maxc_mins_c, 22, 8, 4, up(c)).
table(up_v_minc_maxc_maxs_c, 22, 8, 4, up(c)).
table(up_v_minc_s_mins_c, 22, 7, 4, up(c)).
table(up_v_minc_s_maxs_c, 22, 7, 4, up(c)).
table(up_v_minc_mins_maxs_c, 22, 8, 4, up(c)).
table(up_v_maxc_s_mins_c, 22, 6, 4, up(c)).
table(up_v_maxc_s_maxs_c, 22, 6, 4, up(c)).
table(up_v_maxc_mins_maxs_c, 22, 7, 4, up(c)).
table(up_v_s_mins_maxs_c, 22, 10, 4, up(c)).
table(up_minc_maxc_s_mins_c, 22, 6, 4, up(c)).
table(up_minc_maxc_s_maxs_c, 22, 6, 4, up(c)).
table(up_minc_s_mins_maxs_c, 22, 5, 4, up(c)).
table(up_maxc_s_mins_maxs_c, 22, 6, 4, up(c)).
table(up_v_minc_maxc_s_mins_c, 22, 8, 5, up(c)).
table(up_v_minc_maxc_s_maxs_c, 22, 8, 5, up(c)).
table(up_v_minc_maxc_mins_maxs_c, 22, 9, 5, up(c)).
table(up_v_minc_s_mins_maxs_c, 22, 8, 5, up(c)).
table(up_v_maxc_s_mins_maxs_c, 22, 7, 5, up(c)).
table(up_minc_maxc_s_mins_maxs_c, 22, 7, 5, up(c)).
table(up_v_minc_maxc_s_mins_maxs_c, 22, 9, 6, up(c)).
table(low_v_minc, 22, 3, 1, low(minc)).
table(low_c_minc, 22, 3, 1, low(minc)).
table(low_maxc_minc, 22, 3, 1, low(minc)).
table(low_s_minc, 22, 3, 1, low(minc)).
table(low_mins_minc, 22, 2, 1, low(minc)).
table(low_maxs_minc, 22, 3, 1, low(minc)).
table(low_v_c_minc, 22, 5, 2, low(minc)).
table(low_v_maxc_minc, 22, 4, 2, low(minc)).
table(low_v_s_minc, 22, 4, 2, low(minc)).
table(low_v_mins_minc, 22, 3, 2, low(minc)).
table(low_v_maxs_minc, 22, 4, 2, low(minc)).
table(low_c_maxc_minc, 22, 4, 2, low(minc)).
table(low_c_s_minc, 22, 4, 2, low(minc)).
table(low_c_mins_minc, 22, 3, 2, low(minc)).
table(low_c_maxs_minc, 22, 4, 2, low(minc)).
table(low_maxc_s_minc, 22, 4, 2, low(minc)).
table(low_maxc_mins_minc, 22, 4, 2, low(minc)).
table(low_maxc_maxs_minc, 22, 4, 2, low(minc)).
table(low_s_mins_minc, 22, 3, 2, low(minc)).
table(low_s_maxs_minc, 22, 4, 2, low(minc)).
table(low_mins_maxs_minc, 22, 3, 2, low(minc)).
table(low_v_c_maxc_minc, 22, 6, 3, low(minc)).
table(low_v_c_s_minc, 22, 6, 3, low(minc)).
table(low_v_c_mins_minc, 22, 6, 3, low(minc)).
table(low_v_c_maxs_minc, 22, 6, 3, low(minc)).
table(low_v_maxc_s_minc, 22, 5, 3, low(minc)).
table(low_v_maxc_mins_minc, 22, 5, 3, low(minc)).
table(low_v_maxc_maxs_minc, 22, 5, 3, low(minc)).
table(low_v_s_mins_minc, 22, 4, 3, low(minc)).
table(low_v_s_maxs_minc, 22, 5, 3, low(minc)).
table(low_v_mins_maxs_minc, 22, 4, 3, low(minc)).
table(low_c_maxc_s_minc, 22, 5, 3, low(minc)).
table(low_c_maxc_mins_minc, 22, 5, 3, low(minc)).
table(low_c_maxc_maxs_minc, 22, 5, 3, low(minc)).
table(low_c_s_mins_minc, 22, 4, 3, low(minc)).
table(low_c_s_maxs_minc, 22, 6, 3, low(minc)).
table(low_c_mins_maxs_minc, 22, 4, 3, low(minc)).
table(low_maxc_s_mins_minc, 22, 5, 3, low(minc)).
table(low_maxc_s_maxs_minc, 22, 5, 3, low(minc)).
table(low_maxc_mins_maxs_minc, 22, 5, 3, low(minc)).
table(low_s_mins_maxs_minc, 22, 4, 3, low(minc)).
table(low_v_c_maxc_s_minc, 22, 7, 4, low(minc)).
table(low_v_c_maxc_mins_minc, 22, 7, 4, low(minc)).
table(low_v_c_maxc_maxs_minc, 22, 7, 4, low(minc)).
table(low_v_c_s_mins_minc, 22, 7, 4, low(minc)).
table(low_v_c_s_maxs_minc, 22, 7, 4, low(minc)).
table(low_v_c_mins_maxs_minc, 22, 7, 4, low(minc)).
table(low_v_maxc_s_mins_minc, 22, 6, 4, low(minc)).
table(low_v_maxc_s_maxs_minc, 22, 6, 4, low(minc)).
table(low_v_maxc_mins_maxs_minc, 22, 6, 4, low(minc)).
table(low_v_s_mins_maxs_minc, 22, 5, 4, low(minc)).
table(low_c_maxc_s_mins_minc, 22, 6, 4, low(minc)).
table(low_c_maxc_s_maxs_minc, 22, 6, 4, low(minc)).
table(low_c_maxc_mins_maxs_minc, 22, 6, 4, low(minc)).
table(low_c_s_mins_maxs_minc, 22, 6, 4, low(minc)).
table(low_maxc_s_mins_maxs_minc, 22, 6, 4, low(minc)).
table(low_v_c_maxc_s_mins_minc, 22, 8, 5, low(minc)).
table(low_v_c_maxc_s_maxs_minc, 22, 8, 5, low(minc)).
table(low_v_c_maxc_mins_maxs_minc, 22, 8, 5, low(minc)).
table(low_v_c_s_mins_maxs_minc, 22, 8, 5, low(minc)).
table(low_v_maxc_s_mins_maxs_minc, 22, 7, 5, low(minc)).
table(low_c_maxc_s_mins_maxs_minc, 22, 7, 5, low(minc)).
table(low_v_c_maxc_s_mins_maxs_minc, 22, 9, 6, low(minc)).
table(up_v_minc, 22, 11, 1, up(minc)).
table(up_maxc_minc, 22, 6, 1, up(minc)).
table(up_v_c_minc, 22, 5, 2, up(minc)).
table(up_v_maxc_minc, 22, 4, 2, up(minc)).
table(up_v_s_minc, 22, 13, 2, up(minc)).
table(up_v_mins_minc, 22, 12, 2, up(minc)).
table(up_v_maxs_minc, 22, 13, 2, up(minc)).
table(up_c_maxc_minc, 22, 4, 2, up(minc)).
table(up_maxc_s_minc, 22, 4, 2, up(minc)).
table(up_maxc_mins_minc, 22, 4, 2, up(minc)).
table(up_maxc_maxs_minc, 22, 7, 2, up(minc)).
table(up_s_maxs_minc, 22, 14, 2, up(minc)).
table(up_v_c_maxc_minc, 22, 6, 3, up(minc)).
table(up_v_c_s_minc, 22, 6, 3, up(minc)).
table(up_v_c_mins_minc, 22, 6, 3, up(minc)).
table(up_v_c_maxs_minc, 22, 6, 3, up(minc)).
table(up_v_maxc_s_minc, 22, 5, 3, up(minc)).
table(up_v_maxc_mins_minc, 22, 5, 3, up(minc)).
table(up_v_maxc_maxs_minc, 22, 5, 3, up(minc)).
table(up_v_s_mins_minc, 22, 14, 3, up(minc)).
table(up_v_s_maxs_minc, 22, 14, 3, up(minc)).
table(up_v_mins_maxs_minc, 22, 14, 3, up(minc)).
table(up_c_maxc_s_minc, 22, 5, 3, up(minc)).
table(up_c_maxc_mins_minc, 22, 5, 3, up(minc)).
table(up_c_maxc_maxs_minc, 22, 5, 3, up(minc)).
table(up_c_s_maxs_minc, 22, 4, 3, up(minc)).
table(up_maxc_s_mins_minc, 22, 5, 3, up(minc)).
table(up_maxc_s_maxs_minc, 22, 5, 3, up(minc)).
table(up_maxc_mins_maxs_minc, 22, 5, 3, up(minc)).
table(up_s_mins_maxs_minc, 22, 15, 3, up(minc)).
table(up_v_c_maxc_s_minc, 22, 7, 4, up(minc)).
table(up_v_c_maxc_mins_minc, 22, 7, 4, up(minc)).
table(up_v_c_maxc_maxs_minc, 22, 7, 4, up(minc)).
table(up_v_c_s_mins_minc, 22, 7, 4, up(minc)).
table(up_v_c_s_maxs_minc, 22, 7, 4, up(minc)).
table(up_v_c_mins_maxs_minc, 22, 7, 4, up(minc)).
table(up_v_maxc_s_mins_minc, 22, 6, 4, up(minc)).
table(up_v_maxc_s_maxs_minc, 22, 6, 4, up(minc)).
table(up_v_maxc_mins_maxs_minc, 22, 6, 4, up(minc)).
table(up_v_s_mins_maxs_minc, 22, 15, 4, up(minc)).
table(up_c_maxc_s_mins_minc, 22, 6, 4, up(minc)).
table(up_c_maxc_s_maxs_minc, 22, 6, 4, up(minc)).
table(up_c_maxc_mins_maxs_minc, 22, 6, 4, up(minc)).
table(up_c_s_mins_maxs_minc, 22, 5, 4, up(minc)).
table(up_maxc_s_mins_maxs_minc, 22, 6, 4, up(minc)).
table(up_v_c_maxc_s_mins_minc, 22, 8, 5, up(minc)).
table(up_v_c_maxc_s_maxs_minc, 22, 8, 5, up(minc)).
table(up_v_c_maxc_mins_maxs_minc, 22, 8, 5, up(minc)).
table(up_v_c_s_mins_maxs_minc, 22, 8, 5, up(minc)).
table(up_v_maxc_s_mins_maxs_minc, 22, 7, 5, up(minc)).
table(up_c_maxc_s_mins_maxs_minc, 22, 7, 5, up(minc)).
table(up_v_c_maxc_s_mins_maxs_minc, 22, 9, 6, up(minc)).
table(low_v_maxc, 22, 18, 1, low(maxc)).
table(low_c_maxc, 22, 18, 1, low(maxc)).
table(low_minc_maxc, 22, 6, 1, low(maxc)).
table(low_s_maxc, 22, 18, 1, low(maxc)).
table(low_mins_maxc, 22, 9, 1, low(maxc)).
table(low_maxs_maxc, 22, 2, 1, low(maxc)).
table(low_v_c_maxc, 22, 3, 2, low(maxc)).
table(low_v_minc_maxc, 22, 6, 2, low(maxc)).
table(low_v_s_maxc, 22, 4, 2, low(maxc)).
table(low_v_mins_maxc, 22, 9, 2, low(maxc)).
table(low_v_maxs_maxc, 22, 3, 2, low(maxc)).
table(low_c_minc_maxc, 22, 11, 2, low(maxc)).
table(low_c_s_maxc, 22, 4, 2, low(maxc)).
table(low_c_mins_maxc, 22, 18, 2, low(maxc)).
table(low_c_maxs_maxc, 22, 3, 2, low(maxc)).
table(low_minc_s_maxc, 22, 4, 2, low(maxc)).
table(low_minc_mins_maxc, 22, 4, 2, low(maxc)).
table(low_minc_maxs_maxc, 22, 4, 2, low(maxc)).
table(low_s_mins_maxc, 22, 18, 2, low(maxc)).
table(low_s_maxs_maxc, 22, 3, 2, low(maxc)).
table(low_mins_maxs_maxc, 22, 3, 2, low(maxc)).
table(low_v_c_minc_maxc, 22, 6, 3, low(maxc)).
table(low_v_c_s_maxc, 22, 4, 3, low(maxc)).
table(low_v_c_mins_maxc, 22, 4, 3, low(maxc)).
table(low_v_c_maxs_maxc, 22, 4, 3, low(maxc)).
table(low_v_minc_s_maxc, 22, 5, 3, low(maxc)).
table(low_v_minc_mins_maxc, 22, 5, 3, low(maxc)).
table(low_v_minc_maxs_maxc, 22, 5, 3, low(maxc)).
table(low_v_s_mins_maxc, 22, 5, 3, low(maxc)).
table(low_v_s_maxs_maxc, 22, 4, 3, low(maxc)).
table(low_v_mins_maxs_maxc, 22, 4, 3, low(maxc)).
table(low_c_minc_s_maxc, 22, 5, 3, low(maxc)).
table(low_c_minc_mins_maxc, 22, 6, 3, low(maxc)).
table(low_c_minc_maxs_maxc, 22, 5, 3, low(maxc)).
table(low_c_s_mins_maxc, 22, 4, 3, low(maxc)).
table(low_c_s_maxs_maxc, 22, 4, 3, low(maxc)).
table(low_c_mins_maxs_maxc, 22, 4, 3, low(maxc)).
table(low_minc_s_mins_maxc, 22, 5, 3, low(maxc)).
table(low_minc_s_maxs_maxc, 22, 5, 3, low(maxc)).
table(low_minc_mins_maxs_maxc, 22, 5, 3, low(maxc)).
table(low_s_mins_maxs_maxc, 22, 4, 3, low(maxc)).
table(low_v_c_minc_s_maxc, 22, 7, 4, low(maxc)).
table(low_v_c_minc_mins_maxc, 22, 7, 4, low(maxc)).
table(low_v_c_minc_maxs_maxc, 22, 7, 4, low(maxc)).
table(low_v_c_s_mins_maxc, 22, 5, 4, low(maxc)).
table(low_v_c_s_maxs_maxc, 22, 5, 4, low(maxc)).
table(low_v_c_mins_maxs_maxc, 22, 5, 4, low(maxc)).
table(low_v_minc_s_mins_maxc, 22, 6, 4, low(maxc)).
table(low_v_minc_s_maxs_maxc, 22, 6, 4, low(maxc)).
table(low_v_minc_mins_maxs_maxc, 22, 6, 4, low(maxc)).
table(low_v_s_mins_maxs_maxc, 22, 5, 4, low(maxc)).
table(low_c_minc_s_mins_maxc, 22, 6, 4, low(maxc)).
table(low_c_minc_s_maxs_maxc, 22, 6, 4, low(maxc)).
table(low_c_minc_mins_maxs_maxc, 22, 6, 4, low(maxc)).
table(low_c_s_mins_maxs_maxc, 22, 5, 4, low(maxc)).
table(low_minc_s_mins_maxs_maxc, 22, 6, 4, low(maxc)).
table(low_v_c_minc_s_mins_maxc, 22, 8, 5, low(maxc)).
table(low_v_c_minc_s_maxs_maxc, 22, 8, 5, low(maxc)).
table(low_v_c_minc_mins_maxs_maxc, 22, 8, 5, low(maxc)).
table(low_v_c_s_mins_maxs_maxc, 22, 6, 5, low(maxc)).
table(low_v_minc_s_mins_maxs_maxc, 22, 7, 5, low(maxc)).
table(low_c_minc_s_mins_maxs_maxc, 22, 7, 5, low(maxc)).
table(low_v_c_minc_s_mins_maxs_maxc, 22, 9, 6, low(maxc)).
table(up_v_maxc, 22, 11, 1, up(maxc)).
table(up_v_c_maxc, 22, 11, 2, up(maxc)).
table(up_v_minc_maxc, 22, 11, 2, up(maxc)).
table(up_v_s_maxc, 22, 13, 2, up(maxc)).
table(up_v_mins_maxc, 22, 12, 2, up(maxc)).
table(up_v_maxs_maxc, 22, 13, 2, up(maxc)).
table(up_v_c_minc_maxc, 22, 11, 3, up(maxc)).
table(up_v_c_s_maxc, 22, 13, 3, up(maxc)).
table(up_v_c_mins_maxc, 22, 12, 3, up(maxc)).
table(up_v_c_maxs_maxc, 22, 13, 3, up(maxc)).
table(up_v_minc_s_maxc, 22, 12, 3, up(maxc)).
table(up_v_minc_mins_maxc, 22, 12, 3, up(maxc)).
table(up_v_minc_maxs_maxc, 22, 12, 3, up(maxc)).
table(up_v_s_mins_maxc, 22, 14, 3, up(maxc)).
table(up_v_s_maxs_maxc, 22, 14, 3, up(maxc)).
table(up_v_mins_maxs_maxc, 22, 14, 3, up(maxc)).
table(up_v_c_minc_s_maxc, 22, 12, 4, up(maxc)).
table(up_v_c_minc_mins_maxc, 22, 12, 4, up(maxc)).
table(up_v_c_minc_maxs_maxc, 22, 12, 4, up(maxc)).
table(up_v_c_s_mins_maxc, 22, 14, 4, up(maxc)).
table(up_v_c_s_maxs_maxc, 22, 6, 4, up(maxc)).
table(up_v_c_mins_maxs_maxc, 22, 10, 4, up(maxc)).
table(up_v_minc_s_mins_maxc, 22, 13, 4, up(maxc)).
table(up_v_minc_s_maxs_maxc, 22, 13, 4, up(maxc)).
table(up_v_minc_mins_maxs_maxc, 22, 13, 4, up(maxc)).
table(up_v_s_mins_maxs_maxc, 22, 15, 4, up(maxc)).
table(up_v_c_minc_s_mins_maxc, 22, 13, 5, up(maxc)).
table(up_v_c_minc_s_maxs_maxc, 22, 8, 5, up(maxc)).
table(up_v_c_minc_mins_maxs_maxc, 22, 9, 5, up(maxc)).
table(up_v_c_s_mins_maxs_maxc, 22, 10, 5, up(maxc)).
table(up_v_minc_s_mins_maxs_maxc, 22, 14, 5, up(maxc)).
table(up_v_c_minc_s_mins_maxs_maxc, 22, 9, 6, up(maxc)).
table(low_v_s, 22, 18, 1, low(s)).
table(low_c_s, 22, 3, 1, low(s)).
table(low_minc_s, 22, 18, 1, low(s)).
table(low_maxc_s, 22, 18, 1, low(s)).
table(low_mins_s, 22, 18, 1, low(s)).
table(low_maxs_s, 22, 18, 1, low(s)).
table(low_v_c_s, 22, 4, 2, low(s)).
table(low_v_minc_s, 22, 18, 2, low(s)).
table(low_v_maxc_s, 22, 6, 2, low(s)).
table(low_v_mins_s, 22, 9, 2, low(s)).
table(low_v_maxs_s, 22, 3, 2, low(s)).
table(low_c_minc_s, 22, 5, 2, low(s)).
table(low_c_maxc_s, 22, 5, 2, low(s)).
table(low_c_mins_s, 22, 5, 2, low(s)).
table(low_c_maxs_s, 22, 5, 2, low(s)).
table(low_minc_maxc_s, 22, 18, 2, low(s)).
table(low_minc_mins_s, 22, 12, 2, low(s)).
table(low_minc_maxs_s, 22, 14, 2, low(s)).
table(low_maxc_mins_s, 22, 7, 2, low(s)).
table(low_maxc_maxs_s, 22, 14, 2, low(s)).
table(low_mins_maxs_s, 22, 12, 2, low(s)).
table(low_v_c_minc_s, 22, 8, 3, low(s)).
table(low_v_c_maxc_s, 22, 6, 3, low(s)).
table(low_v_c_mins_s, 22, 8, 3, low(s)).
table(low_v_c_maxs_s, 22, 4, 3, low(s)).
table(low_v_minc_maxc_s, 22, 10, 3, low(s)).
table(low_v_minc_mins_s, 22, 13, 3, low(s)).
table(low_v_minc_maxs_s, 22, 4, 3, low(s)).
table(low_v_maxc_mins_s, 22, 4, 3, low(s)).
table(low_v_maxc_maxs_s, 22, 4, 3, low(s)).
table(low_v_mins_maxs_s, 22, 4, 3, low(s)).
table(low_c_minc_maxc_s, 22, 8, 3, low(s)).
table(low_c_minc_mins_s, 22, 4, 3, low(s)).
table(low_c_minc_maxs_s, 22, 5, 3, low(s)).
table(low_c_maxc_mins_s, 22, 8, 3, low(s)).
table(low_c_maxc_maxs_s, 22, 5, 3, low(s)).
table(low_c_mins_maxs_s, 22, 8, 3, low(s)).
table(low_minc_maxc_mins_s, 22, 13, 3, low(s)).
table(low_minc_maxc_maxs_s, 22, 14, 3, low(s)).
table(low_minc_mins_maxs_s, 22, 8, 3, low(s)).
table(low_maxc_mins_maxs_s, 22, 9, 3, low(s)).
table(low_v_c_minc_maxc_s, 22, 10, 4, low(s)).
table(low_v_c_minc_mins_s, 22, 7, 4, low(s)).
table(low_v_c_minc_maxs_s, 22, 7, 4, low(s)).
table(low_v_c_maxc_mins_s, 22, 5, 4, low(s)).
table(low_v_c_maxc_maxs_s, 22, 5, 4, low(s)).
table(low_v_c_mins_maxs_s, 22, 5, 4, low(s)).
table(low_v_minc_maxc_mins_s, 22, 8, 4, low(s)).
table(low_v_minc_maxc_maxs_s, 22, 6, 4, low(s)).
table(low_v_minc_mins_maxs_s, 22, 5, 4, low(s)).
table(low_v_maxc_mins_maxs_s, 22, 5, 4, low(s)).
table(low_c_minc_maxc_mins_s, 22, 6, 4, low(s)).
table(low_c_minc_maxc_maxs_s, 22, 7, 4, low(s)).
table(low_c_minc_mins_maxs_s, 22, 5, 4, low(s)).
table(low_c_maxc_mins_maxs_s, 22, 6, 4, low(s)).
table(low_minc_maxc_mins_maxs_s, 22, 9, 4, low(s)).
table(low_v_c_minc_maxc_mins_s, 22, 8, 5, low(s)).
table(low_v_c_minc_maxc_maxs_s, 22, 8, 5, low(s)).
table(low_v_c_minc_mins_maxs_s, 22, 8, 5, low(s)).
table(low_v_c_maxc_mins_maxs_s, 22, 6, 5, low(s)).
table(low_v_minc_maxc_mins_maxs_s, 22, 7, 5, low(s)).
table(low_c_minc_maxc_mins_maxs_s, 22, 7, 5, low(s)).
table(low_v_c_minc_maxc_mins_maxs_s, 22, 9, 6, low(s)).
table(up_v_s, 22, 7, 1, up(s)).
table(up_v_c_s, 22, 8, 2, up(s)).
table(up_v_minc_s, 22, 8, 2, up(s)).
table(up_v_maxc_s, 22, 8, 2, up(s)).
table(up_v_mins_s, 22, 4, 2, up(s)).
table(up_v_maxs_s, 22, 7, 2, up(s)).
table(up_c_maxc_s, 22, 9, 2, up(s)).
table(up_v_c_minc_s, 22, 11, 3, up(s)).
table(up_v_c_maxc_s, 22, 9, 3, up(s)).
table(up_v_c_mins_s, 22, 5, 3, up(s)).
table(up_v_c_maxs_s, 22, 8, 3, up(s)).
table(up_v_minc_maxc_s, 22, 10, 3, up(s)).
table(up_v_minc_mins_s, 22, 5, 3, up(s)).
table(up_v_minc_maxs_s, 22, 8, 3, up(s)).
table(up_v_maxc_mins_s, 22, 5, 3, up(s)).
table(up_v_maxc_maxs_s, 22, 8, 3, up(s)).
table(up_v_mins_maxs_s, 22, 5, 3, up(s)).
table(up_c_minc_maxc_s, 22, 12, 3, up(s)).
table(up_c_maxc_mins_s, 22, 5, 3, up(s)).
table(up_c_maxc_maxs_s, 22, 9, 3, up(s)).
table(up_v_c_minc_maxc_s, 22, 12, 4, up(s)).
table(up_v_c_minc_mins_s, 22, 8, 4, up(s)).
table(up_v_c_minc_maxs_s, 22, 11, 4, up(s)).
table(up_v_c_maxc_mins_s, 22, 6, 4, up(s)).
table(up_v_c_maxc_maxs_s, 22, 9, 4, up(s)).
table(up_v_c_mins_maxs_s, 22, 6, 4, up(s)).
table(up_v_minc_maxc_mins_s, 22, 7, 4, up(s)).
table(up_v_minc_maxc_maxs_s, 22, 10, 4, up(s)).
table(up_v_minc_mins_maxs_s, 22, 6, 4, up(s)).
table(up_v_maxc_mins_maxs_s, 22, 6, 4, up(s)).
table(up_c_minc_maxc_mins_s, 22, 7, 4, up(s)).
table(up_c_minc_maxc_maxs_s, 22, 12, 4, up(s)).
table(up_c_maxc_mins_maxs_s, 22, 6, 4, up(s)).
table(up_v_c_minc_maxc_mins_s, 22, 9, 5, up(s)).
table(up_v_c_minc_maxc_maxs_s, 22, 12, 5, up(s)).
table(up_v_c_minc_mins_maxs_s, 22, 9, 5, up(s)).
table(up_v_c_maxc_mins_maxs_s, 22, 7, 5, up(s)).
table(up_v_minc_maxc_mins_maxs_s, 22, 8, 5, up(s)).
table(up_c_minc_maxc_mins_maxs_s, 22, 8, 5, up(s)).
table(up_v_c_minc_maxc_mins_maxs_s, 22, 10, 6, up(s)).
table(low_v_mins, 22, 2, 1, low(mins)).
table(low_c_mins, 22, 2, 1, low(mins)).
table(low_minc_mins, 22, 2, 1, low(mins)).
table(low_maxc_mins, 22, 2, 1, low(mins)).
table(low_s_mins, 22, 2, 1, low(mins)).
table(low_maxs_mins, 22, 2, 1, low(mins)).
table(low_v_c_mins, 22, 3, 2, low(mins)).
table(low_v_minc_mins, 22, 3, 2, low(mins)).
table(low_v_maxc_mins, 22, 3, 2, low(mins)).
table(low_v_s_mins, 22, 3, 2, low(mins)).
table(low_v_maxs_mins, 22, 3, 2, low(mins)).
table(low_c_minc_mins, 22, 3, 2, low(mins)).
table(low_c_maxc_mins, 22, 3, 2, low(mins)).
table(low_c_s_mins, 22, 3, 2, low(mins)).
table(low_c_maxs_mins, 22, 3, 2, low(mins)).
table(low_minc_maxc_mins, 22, 4, 2, low(mins)).
table(low_minc_s_mins, 22, 3, 2, low(mins)).
table(low_minc_maxs_mins, 22, 3, 2, low(mins)).
table(low_maxc_s_mins, 22, 3, 2, low(mins)).
table(low_maxc_maxs_mins, 22, 3, 2, low(mins)).
table(low_s_maxs_mins, 22, 3, 2, low(mins)).
table(low_v_c_minc_mins, 22, 6, 3, low(mins)).
table(low_v_c_maxc_mins, 22, 4, 3, low(mins)).
table(low_v_c_s_mins, 22, 4, 3, low(mins)).
table(low_v_c_maxs_mins, 22, 4, 3, low(mins)).
table(low_v_minc_maxc_mins, 22, 5, 3, low(mins)).
table(low_v_minc_s_mins, 22, 4, 3, low(mins)).
table(low_v_minc_maxs_mins, 22, 4, 3, low(mins)).
table(low_v_maxc_s_mins, 22, 4, 3, low(mins)).
table(low_v_maxc_maxs_mins, 22, 4, 3, low(mins)).
table(low_v_s_maxs_mins, 22, 4, 3, low(mins)).
table(low_c_minc_maxc_mins, 22, 5, 3, low(mins)).
table(low_c_minc_s_mins, 22, 4, 3, low(mins)).
table(low_c_minc_maxs_mins, 22, 4, 3, low(mins)).
table(low_c_maxc_s_mins, 22, 4, 3, low(mins)).
table(low_c_maxc_maxs_mins, 22, 4, 3, low(mins)).
table(low_c_s_maxs_mins, 22, 4, 3, low(mins)).
table(low_minc_maxc_s_mins, 22, 5, 3, low(mins)).
table(low_minc_maxc_maxs_mins, 22, 5, 3, low(mins)).
table(low_minc_s_maxs_mins, 22, 4, 3, low(mins)).
table(low_maxc_s_maxs_mins, 22, 4, 3, low(mins)).
table(low_v_c_minc_maxc_mins, 22, 7, 4, low(mins)).
table(low_v_c_minc_s_mins, 22, 7, 4, low(mins)).
table(low_v_c_minc_maxs_mins, 22, 7, 4, low(mins)).
table(low_v_c_maxc_s_mins, 22, 5, 4, low(mins)).
table(low_v_c_maxc_maxs_mins, 22, 5, 4, low(mins)).
table(low_v_c_s_maxs_mins, 22, 5, 4, low(mins)).
table(low_v_minc_maxc_s_mins, 22, 6, 4, low(mins)).
table(low_v_minc_maxc_maxs_mins, 22, 6, 4, low(mins)).
table(low_v_minc_s_maxs_mins, 22, 5, 4, low(mins)).
table(low_v_maxc_s_maxs_mins, 22, 5, 4, low(mins)).
table(low_c_minc_maxc_s_mins, 22, 6, 4, low(mins)).
table(low_c_minc_maxc_maxs_mins, 22, 6, 4, low(mins)).
table(low_c_minc_s_maxs_mins, 22, 5, 4, low(mins)).
table(low_c_maxc_s_maxs_mins, 22, 5, 4, low(mins)).
table(low_minc_maxc_s_maxs_mins, 22, 6, 4, low(mins)).
table(low_v_c_minc_maxc_s_mins, 22, 8, 5, low(mins)).
table(low_v_c_minc_maxc_maxs_mins, 22, 8, 5, low(mins)).
table(low_v_c_minc_s_maxs_mins, 22, 8, 5, low(mins)).
table(low_v_c_maxc_s_maxs_mins, 22, 6, 5, low(mins)).
table(low_v_minc_maxc_s_maxs_mins, 22, 7, 5, low(mins)).
table(low_c_minc_maxc_s_maxs_mins, 22, 7, 5, low(mins)).
table(low_v_c_minc_maxc_s_maxs_mins, 22, 9, 6, low(mins)).
table(up_v_mins, 22, 18, 1, up(mins)).
table(up_maxc_mins, 22, 9, 1, up(mins)).
table(up_maxs_mins, 22, 2, 1, up(mins)).
table(up_v_c_mins, 22, 6, 2, up(mins)).
table(up_v_minc_mins, 22, 3, 2, up(mins)).
table(up_v_maxc_mins, 22, 5, 2, up(mins)).
table(up_v_s_mins, 22, 3, 2, up(mins)).
table(up_v_maxs_mins, 22, 3, 2, up(mins)).
table(up_c_maxc_mins, 22, 5, 2, up(mins)).
table(up_c_maxs_mins, 22, 3, 2, up(mins)).
table(up_minc_maxc_mins, 22, 4, 2, up(mins)).
table(up_minc_maxs_mins, 22, 3, 2, up(mins)).
table(up_maxc_s_mins, 22, 3, 2, up(mins)).
table(up_maxc_maxs_mins, 22, 3, 2, up(mins)).
table(up_s_maxs_mins, 22, 3, 2, up(mins)).
table(up_v_c_minc_mins, 22, 6, 3, up(mins)).
table(up_v_c_maxc_mins, 22, 7, 3, up(mins)).
table(up_v_c_s_mins, 22, 4, 3, up(mins)).
table(up_v_c_maxs_mins, 22, 4, 3, up(mins)).
table(up_v_minc_maxc_mins, 22, 5, 3, up(mins)).
table(up_v_minc_s_mins, 22, 4, 3, up(mins)).
table(up_v_minc_maxs_mins, 22, 4, 3, up(mins)).
table(up_v_maxc_s_mins, 22, 4, 3, up(mins)).
table(up_v_maxc_maxs_mins, 22, 4, 3, up(mins)).
table(up_v_s_maxs_mins, 22, 4, 3, up(mins)).
table(up_c_minc_maxc_mins, 22, 5, 3, up(mins)).
table(up_c_minc_maxs_mins, 22, 4, 3, up(mins)).
table(up_c_maxc_s_mins, 22, 4, 3, up(mins)).
table(up_c_maxc_maxs_mins, 22, 4, 3, up(mins)).
table(up_c_s_maxs_mins, 22, 4, 3, up(mins)).
table(up_minc_maxc_s_mins, 22, 5, 3, up(mins)).
table(up_minc_maxc_maxs_mins, 22, 5, 3, up(mins)).
table(up_minc_s_maxs_mins, 22, 4, 3, up(mins)).
table(up_maxc_s_maxs_mins, 22, 4, 3, up(mins)).
table(up_v_c_minc_maxc_mins, 22, 7, 4, up(mins)).
table(up_v_c_minc_s_mins, 22, 7, 4, up(mins)).
table(up_v_c_minc_maxs_mins, 22, 7, 4, up(mins)).
table(up_v_c_maxc_s_mins, 22, 5, 4, up(mins)).
table(up_v_c_maxc_maxs_mins, 22, 5, 4, up(mins)).
table(up_v_c_s_maxs_mins, 22, 5, 4, up(mins)).
table(up_v_minc_maxc_s_mins, 22, 6, 4, up(mins)).
table(up_v_minc_maxc_maxs_mins, 22, 6, 4, up(mins)).
table(up_v_minc_s_maxs_mins, 22, 5, 4, up(mins)).
table(up_v_maxc_s_maxs_mins, 22, 5, 4, up(mins)).
table(up_c_minc_maxc_s_mins, 22, 6, 4, up(mins)).
table(up_c_minc_maxc_maxs_mins, 22, 6, 4, up(mins)).
table(up_c_minc_s_maxs_mins, 22, 5, 4, up(mins)).
table(up_c_maxc_s_maxs_mins, 22, 5, 4, up(mins)).
table(up_minc_maxc_s_maxs_mins, 22, 6, 4, up(mins)).
table(up_v_c_minc_maxc_s_mins, 22, 8, 5, up(mins)).
table(up_v_c_minc_maxc_maxs_mins, 22, 8, 5, up(mins)).
table(up_v_c_minc_s_maxs_mins, 22, 8, 5, up(mins)).
table(up_v_c_maxc_s_maxs_mins, 22, 6, 5, up(mins)).
table(up_v_minc_maxc_s_maxs_mins, 22, 7, 5, up(mins)).
table(up_c_minc_maxc_s_maxs_mins, 22, 7, 5, up(mins)).
table(up_v_c_minc_maxc_s_maxs_mins, 22, 9, 6, up(mins)).
table(low_v_maxs, 22, 7, 1, low(maxs)).
table(low_c_maxs, 22, 4, 1, low(maxs)).
table(low_minc_maxs, 22, 4, 1, low(maxs)).
table(low_maxc_maxs, 22, 4, 1, low(maxs)).
table(low_s_maxs, 22, 7, 1, low(maxs)).
table(low_mins_maxs, 22, 2, 1, low(maxs)).
table(low_v_c_maxs, 22, 8, 2, low(maxs)).
table(low_v_minc_maxs, 22, 8, 2, low(maxs)).
table(low_v_maxc_maxs, 22, 8, 2, low(maxs)).
table(low_v_s_maxs, 22, 3, 2, low(maxs)).
table(low_v_mins_maxs, 22, 3, 2, low(maxs)).
table(low_c_minc_maxs, 22, 5, 2, low(maxs)).
table(low_c_maxc_maxs, 22, 5, 2, low(maxs)).
table(low_c_s_maxs, 22, 8, 2, low(maxs)).
table(low_c_mins_maxs, 22, 3, 2, low(maxs)).
table(low_minc_maxc_maxs, 22, 6, 2, low(maxs)).
table(low_minc_s_maxs, 22, 3, 2, low(maxs)).
table(low_minc_mins_maxs, 22, 3, 2, low(maxs)).
table(low_maxc_s_maxs, 22, 3, 2, low(maxs)).
table(low_maxc_mins_maxs, 22, 3, 2, low(maxs)).
table(low_s_mins_maxs, 22, 7, 2, low(maxs)).
table(low_v_c_minc_maxs, 22, 11, 3, low(maxs)).
table(low_v_c_maxc_maxs, 22, 9, 3, low(maxs)).
table(low_v_c_s_maxs, 22, 4, 3, low(maxs)).
table(low_v_c_mins_maxs, 22, 4, 3, low(maxs)).
table(low_v_minc_maxc_maxs, 22, 10, 3, low(maxs)).
table(low_v_minc_s_maxs, 22, 4, 3, low(maxs)).
table(low_v_minc_mins_maxs, 22, 4, 3, low(maxs)).
table(low_v_maxc_s_maxs, 22, 4, 3, low(maxs)).
table(low_v_maxc_mins_maxs, 22, 4, 3, low(maxs)).
table(low_v_s_mins_maxs, 22, 4, 3, low(maxs)).
table(low_c_minc_maxc_maxs, 22, 7, 3, low(maxs)).
table(low_c_minc_s_maxs, 22, 4, 3, low(maxs)).
table(low_c_minc_mins_maxs, 22, 4, 3, low(maxs)).
table(low_c_maxc_s_maxs, 22, 4, 3, low(maxs)).
table(low_c_maxc_mins_maxs, 22, 4, 3, low(maxs)).
table(low_c_s_mins_maxs, 22, 8, 3, low(maxs)).
table(low_minc_maxc_s_maxs, 22, 5, 3, low(maxs)).
table(low_minc_maxc_mins_maxs, 22, 5, 3, low(maxs)).
table(low_minc_s_mins_maxs, 22, 4, 3, low(maxs)).
table(low_maxc_s_mins_maxs, 22, 4, 3, low(maxs)).
table(low_v_c_minc_maxc_maxs, 22, 12, 4, low(maxs)).
table(low_v_c_minc_s_maxs, 22, 7, 4, low(maxs)).
table(low_v_c_minc_mins_maxs, 22, 7, 4, low(maxs)).
table(low_v_c_maxc_s_maxs, 22, 5, 4, low(maxs)).
table(low_v_c_maxc_mins_maxs, 22, 5, 4, low(maxs)).
table(low_v_c_s_mins_maxs, 22, 5, 4, low(maxs)).
table(low_v_minc_maxc_s_maxs, 22, 6, 4, low(maxs)).
table(low_v_minc_maxc_mins_maxs, 22, 6, 4, low(maxs)).
table(low_v_minc_s_mins_maxs, 22, 5, 4, low(maxs)).
table(low_v_maxc_s_mins_maxs, 22, 5, 4, low(maxs)).
table(low_c_minc_maxc_s_maxs, 22, 6, 4, low(maxs)).
table(low_c_minc_maxc_mins_maxs, 22, 6, 4, low(maxs)).
table(low_c_minc_s_mins_maxs, 22, 5, 4, low(maxs)).
table(low_c_maxc_s_mins_maxs, 22, 5, 4, low(maxs)).
table(low_minc_maxc_s_mins_maxs, 22, 6, 4, low(maxs)).
table(low_v_c_minc_maxc_s_maxs, 22, 8, 5, low(maxs)).
table(low_v_c_minc_maxc_mins_maxs, 22, 8, 5, low(maxs)).
table(low_v_c_minc_s_mins_maxs, 22, 8, 5, low(maxs)).
table(low_v_c_maxc_s_mins_maxs, 22, 6, 5, low(maxs)).
table(low_v_minc_maxc_s_mins_maxs, 22, 7, 5, low(maxs)).
table(low_c_minc_maxc_s_mins_maxs, 22, 7, 5, low(maxs)).
table(low_v_c_minc_maxc_s_mins_maxs, 22, 9, 6, low(maxs)).
table(up_v_maxs, 22, 18, 1, up(maxs)).
table(up_maxc_maxs, 22, 2, 1, up(maxs)).
table(up_v_c_maxs, 22, 18, 2, up(maxs)).
table(up_v_minc_maxs, 22, 12, 2, up(maxs)).
table(up_v_maxc_maxs, 22, 3, 2, up(maxs)).
table(up_v_s_maxs, 22, 7, 2, up(maxs)).
table(up_v_mins_maxs, 22, 9, 2, up(maxs)).
table(up_c_maxc_maxs, 22, 3, 2, up(maxs)).
table(up_minc_maxc_maxs, 22, 4, 2, up(maxs)).
table(up_maxc_s_maxs, 22, 3, 2, up(maxs)).
table(up_maxc_mins_maxs, 22, 3, 2, up(maxs)).
table(up_v_c_minc_maxs, 22, 12, 3, up(maxs)).
table(up_v_c_maxc_maxs, 22, 4, 3, up(maxs)).
table(up_v_c_s_maxs, 22, 8, 3, up(maxs)).
table(up_v_c_mins_maxs, 22, 18, 3, up(maxs)).
table(up_v_minc_maxc_maxs, 22, 5, 3, up(maxs)).
table(up_v_minc_s_maxs, 22, 5, 3, up(maxs)).
table(up_v_minc_mins_maxs, 22, 14, 3, up(maxs)).
table(up_v_maxc_s_maxs, 22, 4, 3, up(maxs)).
table(up_v_maxc_mins_maxs, 22, 4, 3, up(maxs)).
table(up_v_s_mins_maxs, 22, 7, 3, up(maxs)).
table(up_c_minc_maxc_maxs, 22, 5, 3, up(maxs)).
table(up_c_maxc_s_maxs, 22, 4, 3, up(maxs)).
table(up_c_maxc_mins_maxs, 22, 4, 3, up(maxs)).
table(up_minc_maxc_s_maxs, 22, 5, 3, up(maxs)).
table(up_minc_maxc_mins_maxs, 22, 5, 3, up(maxs)).
table(up_maxc_s_mins_maxs, 22, 4, 3, up(maxs)).
table(up_v_c_minc_maxc_maxs, 22, 7, 4, up(maxs)).
table(up_v_c_minc_s_maxs, 22, 7, 4, up(maxs)).
table(up_v_c_minc_mins_maxs, 22, 13, 4, up(maxs)).
table(up_v_c_maxc_s_maxs, 22, 5, 4, up(maxs)).
table(up_v_c_maxc_mins_maxs, 22, 5, 4, up(maxs)).
table(up_v_c_s_mins_maxs, 22, 8, 4, up(maxs)).
table(up_v_minc_maxc_s_maxs, 22, 6, 4, up(maxs)).
table(up_v_minc_maxc_mins_maxs, 22, 6, 4, up(maxs)).
table(up_v_minc_s_mins_maxs, 22, 5, 4, up(maxs)).
table(up_v_maxc_s_mins_maxs, 22, 5, 4, up(maxs)).
table(up_c_minc_maxc_s_maxs, 22, 6, 4, up(maxs)).
table(up_c_minc_maxc_mins_maxs, 22, 6, 4, up(maxs)).
table(up_c_maxc_s_mins_maxs, 22, 5, 4, up(maxs)).
table(up_minc_maxc_s_mins_maxs, 22, 6, 4, up(maxs)).
table(up_v_c_minc_maxc_s_maxs, 22, 8, 5, up(maxs)).
table(up_v_c_minc_maxc_mins_maxs, 22, 8, 5, up(maxs)).
table(up_v_c_minc_s_mins_maxs, 22, 8, 5, up(maxs)).
table(up_v_c_maxc_s_mins_maxs, 22, 6, 5, up(maxs)).
table(up_v_minc_maxc_s_mins_maxs, 22, 7, 5, up(maxs)).
table(up_c_minc_maxc_s_mins_maxs, 22, 7, 5, up(maxs)).
table(up_v_c_minc_maxc_s_mins_maxs, 22, 9, 6, up(maxs)).
table(low_v_mina, 22, 12, 1, low(mina)).
table(low_c_mina, 22, 7, 1, low(mina)).
table(low_minc_mina, 22, 19, 1, low(mina)).
table(low_maxc_mina, 22, 19, 1, low(mina)).
table(low_s_mina, 22, 12, 1, low(mina)).
table(low_mins_mina, 22, 11, 1, low(mina)).
table(low_maxs_mina, 22, 11, 1, low(mina)).
table(low_v_c_mina, 22, 10, 2, low(mina)).
table(low_v_minc_mina, 22, 11, 2, low(mina)).
table(low_v_maxc_mina, 22, 11, 2, low(mina)).
table(low_v_s_mina, 22, 4, 2, low(mina)).
table(low_v_mins_mina, 22, 4, 2, low(mina)).
table(low_v_maxs_mina, 22, 10, 2, low(mina)).
table(low_c_minc_mina, 22, 8, 2, low(mina)).
table(low_c_maxc_mina, 22, 6, 2, low(mina)).
table(low_c_s_mina, 22, 10, 2, low(mina)).
table(low_c_mins_mina, 22, 7, 2, low(mina)).
table(low_c_maxs_mina, 22, 5, 2, low(mina)).
table(low_minc_maxc_mina, 22, 19, 2, low(mina)).
table(low_minc_s_mina, 22, 11, 2, low(mina)).
table(low_minc_mins_mina, 22, 19, 2, low(mina)).
table(low_minc_maxs_mina, 22, 19, 2, low(mina)).
table(low_maxc_s_mina, 22, 11, 2, low(mina)).
table(low_maxc_mins_mina, 22, 19, 2, low(mina)).
table(low_maxc_maxs_mina, 22, 19, 2, low(mina)).
table(low_s_mins_mina, 22, 12, 2, low(mina)).
table(low_s_maxs_mina, 22, 10, 2, low(mina)).
table(low_mins_maxs_mina, 22, 6, 2, low(mina)).
table(low_v_c_minc_mina, 22, 13, 3, low(mina)).
table(low_v_c_maxc_mina, 22, 11, 3, low(mina)).
table(low_v_c_s_mina, 22, 4, 3, low(mina)).
table(low_v_c_mins_mina, 22, 7, 3, low(mina)).
table(low_v_c_maxs_mina, 22, 10, 3, low(mina)).
table(low_v_minc_maxc_mina, 22, 12, 3, low(mina)).
table(low_v_minc_s_mina, 22, 4, 3, low(mina)).
table(low_v_minc_mins_mina, 22, 4, 3, low(mina)).
table(low_v_minc_maxs_mina, 22, 11, 3, low(mina)).
table(low_v_maxc_s_mina, 22, 4, 3, low(mina)).
table(low_v_maxc_mins_mina, 22, 6, 3, low(mina)).
table(low_v_maxc_maxs_mina, 22, 10, 3, low(mina)).
table(low_v_s_mins_mina, 22, 5, 3, low(mina)).
table(low_v_s_maxs_mina, 22, 5, 3, low(mina)).
table(low_v_mins_maxs_mina, 22, 5, 3, low(mina)).
table(low_c_minc_maxc_mina, 22, 8, 3, low(mina)).
table(low_c_minc_s_mina, 22, 11, 3, low(mina)).
table(low_c_minc_mins_mina, 22, 8, 3, low(mina)).
table(low_c_minc_maxs_mina, 22, 6, 3, low(mina)).
table(low_c_maxc_s_mina, 22, 8, 3, low(mina)).
table(low_c_maxc_mins_mina, 22, 6, 3, low(mina)).
table(low_c_maxc_maxs_mina, 22, 6, 3, low(mina)).
table(low_c_s_mins_mina, 22, 10, 3, low(mina)).
table(low_c_s_maxs_mina, 22, 10, 3, low(mina)).
table(low_c_mins_maxs_mina, 22, 5, 3, low(mina)).
table(low_minc_maxc_s_mina, 22, 10, 3, low(mina)).
table(low_minc_maxc_mins_mina, 22, 16, 3, low(mina)).
table(low_minc_maxc_maxs_mina, 22, 18, 3, low(mina)).
table(low_minc_s_mins_mina, 22, 6, 3, low(mina)).
table(low_minc_s_maxs_mina, 22, 8, 3, low(mina)).
table(low_minc_mins_maxs_mina, 22, 17, 3, low(mina)).
table(low_maxc_s_mins_mina, 22, 8, 3, low(mina)).
table(low_maxc_s_maxs_mina, 22, 7, 3, low(mina)).
table(low_maxc_mins_maxs_mina, 22, 8, 3, low(mina)).
table(low_s_mins_maxs_mina, 22, 10, 3, low(mina)).
table(low_v_c_minc_maxc_mina, 22, 14, 4, low(mina)).
table(low_v_c_minc_s_mina, 22, 8, 4, low(mina)).
table(low_v_c_minc_mins_mina, 22, 10, 4, low(mina)).
table(low_v_c_minc_maxs_mina, 22, 13, 4, low(mina)).
table(low_v_c_maxc_s_mina, 22, 5, 4, low(mina)).
table(low_v_c_maxc_mins_mina, 22, 8, 4, low(mina)).
table(low_v_c_maxc_maxs_mina, 22, 11, 4, low(mina)).
table(low_v_c_s_mins_mina, 22, 5, 4, low(mina)).
table(low_v_c_s_maxs_mina, 22, 5, 4, low(mina)).
table(low_v_c_mins_maxs_mina, 22, 8, 4, low(mina)).
table(low_v_minc_maxc_s_mina, 22, 6, 4, low(mina)).
table(low_v_minc_maxc_mins_mina, 22, 6, 4, low(mina)).
table(low_v_minc_maxc_maxs_mina, 22, 12, 4, low(mina)).
table(low_v_minc_s_mins_mina, 22, 5, 4, low(mina)).
table(low_v_minc_s_maxs_mina, 22, 5, 4, low(mina)).
table(low_v_minc_mins_maxs_mina, 22, 5, 4, low(mina)).
table(low_v_maxc_s_mins_mina, 22, 5, 4, low(mina)).
table(low_v_maxc_s_maxs_mina, 22, 5, 4, low(mina)).
table(low_v_maxc_mins_maxs_mina, 22, 5, 4, low(mina)).
table(low_v_s_mins_maxs_mina, 22, 6, 4, low(mina)).
table(low_c_minc_maxc_s_mina, 22, 9, 4, low(mina)).
table(low_c_minc_maxc_mins_mina, 22, 7, 4, low(mina)).
table(low_c_minc_maxc_maxs_mina, 22, 8, 4, low(mina)).
table(low_c_minc_s_mins_mina, 22, 10, 4, low(mina)).
table(low_c_minc_s_maxs_mina, 22, 10, 4, low(mina)).
table(low_c_minc_mins_maxs_mina, 22, 6, 4, low(mina)).
table(low_c_maxc_s_mins_mina, 22, 6, 4, low(mina)).
table(low_c_maxc_s_maxs_mina, 22, 6, 4, low(mina)).
table(low_c_maxc_mins_maxs_mina, 22, 6, 4, low(mina)).
table(low_c_s_mins_maxs_mina, 22, 10, 4, low(mina)).
table(low_minc_maxc_s_mins_mina, 22, 7, 4, low(mina)).
table(low_minc_maxc_s_maxs_mina, 22, 9, 4, low(mina)).
table(low_minc_maxc_mins_maxs_mina, 22, 16, 4, low(mina)).
table(low_minc_s_mins_maxs_mina, 22, 6, 4, low(mina)).
table(low_maxc_s_mins_maxs_mina, 22, 6, 4, low(mina)).
table(low_v_c_minc_maxc_s_mina, 22, 9, 5, low(mina)).
table(low_v_c_minc_maxc_mins_mina, 22, 11, 5, low(mina)).
table(low_v_c_minc_maxc_maxs_mina, 22, 14, 5, low(mina)).
table(low_v_c_minc_s_mins_mina, 22, 8, 5, low(mina)).
table(low_v_c_minc_s_maxs_mina, 22, 8, 5, low(mina)).
table(low_v_c_minc_mins_maxs_mina, 22, 11, 5, low(mina)).
table(low_v_c_maxc_s_mins_mina, 22, 6, 5, low(mina)).
table(low_v_c_maxc_s_maxs_mina, 22, 6, 5, low(mina)).
table(low_v_c_maxc_mins_maxs_mina, 22, 9, 5, low(mina)).
table(low_v_c_s_mins_maxs_mina, 22, 6, 5, low(mina)).
table(low_v_minc_maxc_s_mins_mina, 22, 7, 5, low(mina)).
table(low_v_minc_maxc_s_maxs_mina, 22, 7, 5, low(mina)).
table(low_v_minc_maxc_mins_maxs_mina, 22, 7, 5, low(mina)).
table(low_v_minc_s_mins_maxs_mina, 22, 6, 5, low(mina)).
table(low_v_maxc_s_mins_maxs_mina, 22, 6, 5, low(mina)).
table(low_c_minc_maxc_s_mins_mina, 22, 8, 5, low(mina)).
table(low_c_minc_maxc_s_maxs_mina, 22, 8, 5, low(mina)).
table(low_c_minc_maxc_mins_maxs_mina, 22, 8, 5, low(mina)).
table(low_c_minc_s_mins_maxs_mina, 22, 11, 5, low(mina)).
table(low_c_maxc_s_mins_maxs_mina, 22, 7, 5, low(mina)).
table(low_minc_maxc_s_mins_maxs_mina, 22, 8, 5, low(mina)).
table(low_v_c_minc_maxc_s_mins_mina, 22, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_maxs_mina, 22, 9, 6, low(mina)).
table(low_v_c_minc_maxc_mins_maxs_mina, 22, 12, 6, low(mina)).
table(low_v_c_minc_s_mins_maxs_mina, 22, 9, 6, low(mina)).
table(low_v_c_maxc_s_mins_maxs_mina, 22, 7, 6, low(mina)).
table(low_v_minc_maxc_s_mins_maxs_mina, 22, 8, 6, low(mina)).
table(low_c_minc_maxc_s_mins_maxs_mina, 22, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_mins_maxs_mina, 22, 10, 7, low(mina)).
table(up_v_maxa, 22, 19, 1, up(maxa)).
table(up_v_c_maxa, 22, 19, 2, up(maxa)).
table(up_v_minc_maxa, 22, 19, 2, up(maxa)).
table(up_v_maxc_maxa, 22, 19, 2, up(maxa)).
table(up_v_s_maxa, 22, 19, 2, up(maxa)).
table(up_v_mins_maxa, 22, 19, 2, up(maxa)).
table(up_v_maxs_maxa, 22, 19, 2, up(maxa)).
table(up_c_maxc_maxa, 22, 19, 2, up(maxa)).
table(up_maxc_s_maxa, 22, 18, 2, up(maxa)).
table(up_s_maxs_maxa, 22, 19, 2, up(maxa)).
table(up_v_c_minc_maxa, 22, 19, 3, up(maxa)).
table(up_v_c_maxc_maxa, 22, 19, 3, up(maxa)).
table(up_v_c_s_maxa, 22, 19, 3, up(maxa)).
table(up_v_c_mins_maxa, 22, 19, 3, up(maxa)).
table(up_v_c_maxs_maxa, 22, 7, 3, up(maxa)).
table(up_v_minc_maxc_maxa, 22, 19, 3, up(maxa)).
table(up_v_minc_s_maxa, 22, 18, 3, up(maxa)).
table(up_v_minc_mins_maxa, 22, 19, 3, up(maxa)).
table(up_v_minc_maxs_maxa, 22, 19, 3, up(maxa)).
table(up_v_maxc_s_maxa, 22, 18, 3, up(maxa)).
table(up_v_maxc_mins_maxa, 22, 11, 3, up(maxa)).
table(up_v_maxc_maxs_maxa, 22, 6, 3, up(maxa)).
table(up_v_s_mins_maxa, 22, 19, 3, up(maxa)).
table(up_v_s_maxs_maxa, 22, 19, 3, up(maxa)).
table(up_v_mins_maxs_maxa, 22, 19, 3, up(maxa)).
table(up_c_minc_maxc_maxa, 22, 19, 3, up(maxa)).
table(up_c_maxc_s_maxa, 22, 18, 3, up(maxa)).
table(up_c_maxc_mins_maxa, 22, 19, 3, up(maxa)).
table(up_c_maxc_maxs_maxa, 22, 6, 3, up(maxa)).
table(up_c_s_maxs_maxa, 22, 19, 3, up(maxa)).
table(up_minc_maxc_s_maxa, 22, 8, 3, up(maxa)).
table(up_minc_s_maxs_maxa, 22, 18, 3, up(maxa)).
table(up_maxc_s_mins_maxa, 22, 10, 3, up(maxa)).
table(up_maxc_s_maxs_maxa, 22, 4, 3, up(maxa)).
table(up_s_mins_maxs_maxa, 22, 19, 3, up(maxa)).
table(up_v_c_minc_maxc_maxa, 22, 19, 4, up(maxa)).
table(up_v_c_minc_s_maxa, 22, 18, 4, up(maxa)).
table(up_v_c_minc_mins_maxa, 22, 11, 4, up(maxa)).
table(up_v_c_minc_maxs_maxa, 22, 8, 4, up(maxa)).
table(up_v_c_maxc_s_maxa, 22, 18, 4, up(maxa)).
table(up_v_c_maxc_mins_maxa, 22, 11, 4, up(maxa)).
table(up_v_c_maxc_maxs_maxa, 22, 6, 4, up(maxa)).
table(up_v_c_s_mins_maxa, 22, 19, 4, up(maxa)).
table(up_v_c_s_maxs_maxa, 22, 19, 4, up(maxa)).
table(up_v_c_mins_maxs_maxa, 22, 9, 4, up(maxa)).
table(up_v_minc_maxc_s_maxa, 22, 18, 4, up(maxa)).
table(up_v_minc_maxc_mins_maxa, 22, 14, 4, up(maxa)).
table(up_v_minc_maxc_maxs_maxa, 22, 9, 4, up(maxa)).
table(up_v_minc_s_mins_maxa, 22, 18, 4, up(maxa)).
table(up_v_minc_s_maxs_maxa, 22, 18, 4, up(maxa)).
table(up_v_minc_mins_maxs_maxa, 22, 16, 4, up(maxa)).
table(up_v_maxc_s_mins_maxa, 22, 8, 4, up(maxa)).
table(up_v_maxc_s_maxs_maxa, 22, 7, 4, up(maxa)).
table(up_v_maxc_mins_maxs_maxa, 22, 5, 4, up(maxa)).
table(up_v_s_mins_maxs_maxa, 22, 19, 4, up(maxa)).
table(up_c_minc_maxc_s_maxa, 22, 18, 4, up(maxa)).
table(up_c_minc_maxc_mins_maxa, 22, 14, 4, up(maxa)).
table(up_c_minc_maxc_maxs_maxa, 22, 9, 4, up(maxa)).
table(up_c_minc_s_maxs_maxa, 22, 9, 4, up(maxa)).
table(up_c_maxc_s_mins_maxa, 22, 8, 4, up(maxa)).
table(up_c_maxc_s_maxs_maxa, 22, 6, 4, up(maxa)).
table(up_c_maxc_mins_maxs_maxa, 22, 5, 4, up(maxa)).
table(up_c_s_mins_maxs_maxa, 22, 19, 4, up(maxa)).
table(up_minc_maxc_s_mins_maxa, 22, 7, 4, up(maxa)).
table(up_minc_maxc_s_maxs_maxa, 22, 7, 4, up(maxa)).
table(up_minc_s_mins_maxs_maxa, 22, 16, 4, up(maxa)).
table(up_maxc_s_mins_maxs_maxa, 22, 5, 4, up(maxa)).
table(up_v_c_minc_maxc_s_maxa, 22, 18, 5, up(maxa)).
table(up_v_c_minc_maxc_mins_maxa, 22, 14, 5, up(maxa)).
table(up_v_c_minc_maxc_maxs_maxa, 22, 9, 5, up(maxa)).
table(up_v_c_minc_s_mins_maxa, 22, 10, 5, up(maxa)).
table(up_v_c_minc_s_maxs_maxa, 22, 9, 5, up(maxa)).
table(up_v_c_minc_mins_maxs_maxa, 22, 8, 5, up(maxa)).
table(up_v_c_maxc_s_mins_maxa, 22, 8, 5, up(maxa)).
table(up_v_c_maxc_s_maxs_maxa, 22, 6, 5, up(maxa)).
table(up_v_c_maxc_mins_maxs_maxa, 22, 6, 5, up(maxa)).
table(up_v_c_s_mins_maxs_maxa, 22, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_mins_maxa, 22, 11, 5, up(maxa)).
table(up_v_minc_maxc_s_maxs_maxa, 22, 10, 5, up(maxa)).
table(up_v_minc_maxc_mins_maxs_maxa, 22, 7, 5, up(maxa)).
table(up_v_minc_s_mins_maxs_maxa, 22, 16, 5, up(maxa)).
table(up_v_maxc_s_mins_maxs_maxa, 22, 8, 5, up(maxa)).
table(up_c_minc_maxc_s_mins_maxa, 22, 11, 5, up(maxa)).
table(up_c_minc_maxc_s_maxs_maxa, 22, 9, 5, up(maxa)).
table(up_c_minc_maxc_mins_maxs_maxa, 22, 7, 5, up(maxa)).
table(up_c_minc_s_mins_maxs_maxa, 22, 10, 5, up(maxa)).
table(up_c_maxc_s_mins_maxs_maxa, 22, 7, 5, up(maxa)).
table(up_minc_maxc_s_mins_maxs_maxa, 22, 7, 5, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxa, 22, 11, 6, up(maxa)).
table(up_v_c_minc_maxc_s_maxs_maxa, 22, 9, 6, up(maxa)).
table(up_v_c_minc_maxc_mins_maxs_maxa, 22, 9, 6, up(maxa)).
table(up_v_c_minc_s_mins_maxs_maxa, 22, 10, 6, up(maxa)).
table(up_v_c_maxc_s_mins_maxs_maxa, 22, 7, 6, up(maxa)).
table(up_v_minc_maxc_s_mins_maxs_maxa, 22, 11, 6, up(maxa)).
table(up_c_minc_maxc_s_mins_maxs_maxa, 22, 10, 6, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxs_maxa, 22, 10, 7, up(maxa)).
table(low_c_v, 23, 18, 1, low(v)).
table(low_minc_v, 23, 11, 1, low(v)).
table(low_maxc_v, 23, 11, 1, low(v)).
table(low_s_v, 23, 7, 1, low(v)).
table(low_mins_v, 23, 18, 1, low(v)).
table(low_maxs_v, 23, 18, 1, low(v)).
table(low_c_minc_v, 23, 11, 2, low(v)).
table(low_c_maxc_v, 23, 11, 2, low(v)).
table(low_c_s_v, 23, 8, 2, low(v)).
table(low_c_mins_v, 23, 18, 2, low(v)).
table(low_c_maxs_v, 23, 18, 2, low(v)).
table(low_minc_maxc_v, 23, 11, 2, low(v)).
table(low_minc_s_v, 23, 3, 2, low(v)).
table(low_minc_mins_v, 23, 12, 2, low(v)).
table(low_minc_maxs_v, 23, 12, 2, low(v)).
table(low_maxc_s_v, 23, 3, 2, low(v)).
table(low_maxc_mins_v, 23, 12, 2, low(v)).
table(low_maxc_maxs_v, 23, 13, 2, low(v)).
table(low_s_mins_v, 23, 7, 2, low(v)).
table(low_s_maxs_v, 23, 7, 2, low(v)).
table(low_mins_maxs_v, 23, 12, 2, low(v)).
table(low_c_minc_maxc_v, 23, 11, 3, low(v)).
table(low_c_minc_s_v, 23, 6, 3, low(v)).
table(low_c_minc_mins_v, 23, 12, 3, low(v)).
table(low_c_minc_maxs_v, 23, 12, 3, low(v)).
table(low_c_maxc_s_v, 23, 4, 3, low(v)).
table(low_c_maxc_mins_v, 23, 12, 3, low(v)).
table(low_c_maxc_maxs_v, 23, 13, 3, low(v)).
table(low_c_s_mins_v, 23, 8, 3, low(v)).
table(low_c_s_maxs_v, 23, 8, 3, low(v)).
table(low_c_mins_maxs_v, 23, 18, 3, low(v)).
table(low_minc_maxc_s_v, 23, 5, 3, low(v)).
table(low_minc_maxc_mins_v, 23, 12, 3, low(v)).
table(low_minc_maxc_maxs_v, 23, 12, 3, low(v)).
table(low_minc_s_mins_v, 23, 4, 3, low(v)).
table(low_minc_s_maxs_v, 23, 4, 3, low(v)).
table(low_minc_mins_maxs_v, 23, 13, 3, low(v)).
table(low_maxc_s_mins_v, 23, 4, 3, low(v)).
table(low_maxc_s_maxs_v, 23, 4, 3, low(v)).
table(low_maxc_mins_maxs_v, 23, 14, 3, low(v)).
table(low_s_mins_maxs_v, 23, 7, 3, low(v)).
table(low_c_minc_maxc_s_v, 23, 7, 4, low(v)).
table(low_c_minc_maxc_mins_v, 23, 12, 4, low(v)).
table(low_c_minc_maxc_maxs_v, 23, 12, 4, low(v)).
table(low_c_minc_s_mins_v, 23, 7, 4, low(v)).
table(low_c_minc_s_maxs_v, 23, 7, 4, low(v)).
table(low_c_minc_mins_maxs_v, 23, 13, 4, low(v)).
table(low_c_maxc_s_mins_v, 23, 5, 4, low(v)).
table(low_c_maxc_s_maxs_v, 23, 5, 4, low(v)).
table(low_c_maxc_mins_maxs_v, 23, 14, 4, low(v)).
table(low_c_s_mins_maxs_v, 23, 8, 4, low(v)).
table(low_minc_maxc_s_mins_v, 23, 6, 4, low(v)).
table(low_minc_maxc_s_maxs_v, 23, 6, 4, low(v)).
table(low_minc_maxc_mins_maxs_v, 23, 13, 4, low(v)).
table(low_minc_s_mins_maxs_v, 23, 5, 4, low(v)).
table(low_maxc_s_mins_maxs_v, 23, 5, 4, low(v)).
table(low_c_minc_maxc_s_mins_v, 23, 8, 5, low(v)).
table(low_c_minc_maxc_s_maxs_v, 23, 8, 5, low(v)).
table(low_c_minc_maxc_mins_maxs_v, 23, 13, 5, low(v)).
table(low_c_minc_s_mins_maxs_v, 23, 8, 5, low(v)).
table(low_c_maxc_s_mins_maxs_v, 23, 6, 5, low(v)).
table(low_minc_maxc_s_mins_maxs_v, 23, 7, 5, low(v)).
table(low_c_minc_maxc_s_mins_maxs_v, 23, 9, 6, low(v)).
table(up_c_maxc_v, 23, 3, 2, up(v)).
table(up_maxc_s_v, 23, 3, 2, up(v)).
table(up_s_maxs_v, 23, 3, 2, up(v)).
table(up_c_minc_maxc_v, 23, 6, 3, up(v)).
table(up_c_maxc_s_v, 23, 4, 3, up(v)).
table(up_c_maxc_mins_v, 23, 4, 3, up(v)).
table(up_c_maxc_maxs_v, 23, 4, 3, up(v)).
table(up_c_s_maxs_v, 23, 4, 3, up(v)).
table(up_minc_maxc_s_v, 23, 5, 3, up(v)).
table(up_minc_s_maxs_v, 23, 4, 3, up(v)).
table(up_maxc_s_mins_v, 23, 4, 3, up(v)).
table(up_maxc_s_maxs_v, 23, 4, 3, up(v)).
table(up_s_mins_maxs_v, 23, 4, 3, up(v)).
table(up_c_minc_maxc_s_v, 23, 7, 4, up(v)).
table(up_c_minc_maxc_mins_v, 23, 7, 4, up(v)).
table(up_c_minc_maxc_maxs_v, 23, 7, 4, up(v)).
table(up_c_minc_s_maxs_v, 23, 7, 4, up(v)).
table(up_c_maxc_s_mins_v, 23, 5, 4, up(v)).
table(up_c_maxc_s_maxs_v, 23, 5, 4, up(v)).
table(up_c_maxc_mins_maxs_v, 23, 5, 4, up(v)).
table(up_c_s_mins_maxs_v, 23, 5, 4, up(v)).
table(up_minc_maxc_s_mins_v, 23, 6, 4, up(v)).
table(up_minc_maxc_s_maxs_v, 23, 6, 4, up(v)).
table(up_minc_s_mins_maxs_v, 23, 5, 4, up(v)).
table(up_maxc_s_mins_maxs_v, 23, 5, 4, up(v)).
table(up_c_minc_maxc_s_mins_v, 23, 8, 5, up(v)).
table(up_c_minc_maxc_s_maxs_v, 23, 8, 5, up(v)).
table(up_c_minc_maxc_mins_maxs_v, 23, 8, 5, up(v)).
table(up_c_minc_s_mins_maxs_v, 23, 8, 5, up(v)).
table(up_c_maxc_s_mins_maxs_v, 23, 6, 5, up(v)).
table(up_minc_maxc_s_mins_maxs_v, 23, 7, 5, up(v)).
table(up_c_minc_maxc_s_mins_maxs_v, 23, 9, 6, up(v)).
table(low_v_c, 23, 11, 1, low(c)).
table(low_minc_c, 23, 11, 1, low(c)).
table(low_maxc_c, 23, 11, 1, low(c)).
table(low_s_c, 23, 8, 1, low(c)).
table(low_mins_c, 23, 8, 1, low(c)).
table(low_maxs_c, 23, 8, 1, low(c)).
table(low_v_minc_c, 23, 11, 2, low(c)).
table(low_v_maxc_c, 23, 3, 2, low(c)).
table(low_v_s_c, 23, 13, 2, low(c)).
table(low_v_mins_c, 23, 12, 2, low(c)).
table(low_v_maxs_c, 23, 13, 2, low(c)).
table(low_minc_maxc_c, 23, 11, 2, low(c)).
table(low_minc_s_c, 23, 5, 2, low(c)).
table(low_minc_mins_c, 23, 9, 2, low(c)).
table(low_minc_maxs_c, 23, 9, 2, low(c)).
table(low_maxc_s_c, 23, 3, 2, low(c)).
table(low_maxc_mins_c, 23, 12, 2, low(c)).
table(low_maxc_maxs_c, 23, 13, 2, low(c)).
table(low_s_mins_c, 23, 10, 2, low(c)).
table(low_s_maxs_c, 23, 11, 2, low(c)).
table(low_mins_maxs_c, 23, 9, 2, low(c)).
table(low_v_minc_maxc_c, 23, 6, 3, low(c)).
table(low_v_minc_s_c, 23, 12, 3, low(c)).
table(low_v_minc_mins_c, 23, 12, 3, low(c)).
table(low_v_minc_maxs_c, 23, 12, 3, low(c)).
table(low_v_maxc_s_c, 23, 4, 3, low(c)).
table(low_v_maxc_mins_c, 23, 4, 3, low(c)).
table(low_v_maxc_maxs_c, 23, 4, 3, low(c)).
table(low_v_s_mins_c, 23, 14, 3, low(c)).
table(low_v_s_maxs_c, 23, 14, 3, low(c)).
table(low_v_mins_maxs_c, 23, 14, 3, low(c)).
table(low_minc_maxc_s_c, 23, 5, 3, low(c)).
table(low_minc_maxc_mins_c, 23, 12, 3, low(c)).
table(low_minc_maxc_maxs_c, 23, 12, 3, low(c)).
table(low_minc_s_mins_c, 23, 6, 3, low(c)).
table(low_minc_s_maxs_c, 23, 7, 3, low(c)).
table(low_minc_mins_maxs_c, 23, 7, 3, low(c)).
table(low_maxc_s_mins_c, 23, 4, 3, low(c)).
table(low_maxc_s_maxs_c, 23, 4, 3, low(c)).
table(low_maxc_mins_maxs_c, 23, 6, 3, low(c)).
table(low_s_mins_maxs_c, 23, 12, 3, low(c)).
table(low_v_minc_maxc_s_c, 23, 7, 4, low(c)).
table(low_v_minc_maxc_mins_c, 23, 7, 4, low(c)).
table(low_v_minc_maxc_maxs_c, 23, 7, 4, low(c)).
table(low_v_minc_s_mins_c, 23, 13, 4, low(c)).
table(low_v_minc_s_maxs_c, 23, 13, 4, low(c)).
table(low_v_minc_mins_maxs_c, 23, 13, 4, low(c)).
table(low_v_maxc_s_mins_c, 23, 5, 4, low(c)).
table(low_v_maxc_s_maxs_c, 23, 5, 4, low(c)).
table(low_v_maxc_mins_maxs_c, 23, 5, 4, low(c)).
table(low_v_s_mins_maxs_c, 23, 15, 4, low(c)).
table(low_minc_maxc_s_mins_c, 23, 6, 4, low(c)).
table(low_minc_maxc_s_maxs_c, 23, 6, 4, low(c)).
table(low_minc_maxc_mins_maxs_c, 23, 8, 4, low(c)).
table(low_minc_s_mins_maxs_c, 23, 8, 4, low(c)).
table(low_maxc_s_mins_maxs_c, 23, 5, 4, low(c)).
table(low_v_minc_maxc_s_mins_c, 23, 8, 5, low(c)).
table(low_v_minc_maxc_s_maxs_c, 23, 8, 5, low(c)).
table(low_v_minc_maxc_mins_maxs_c, 23, 8, 5, low(c)).
table(low_v_minc_s_mins_maxs_c, 23, 14, 5, low(c)).
table(low_v_maxc_s_mins_maxs_c, 23, 6, 5, low(c)).
table(low_minc_maxc_s_mins_maxs_c, 23, 7, 5, low(c)).
table(low_v_minc_maxc_s_mins_maxs_c, 23, 9, 6, low(c)).
table(up_v_c, 23, 18, 1, up(c)).
table(up_s_c, 23, 3, 1, up(c)).
table(up_v_minc_c, 23, 6, 2, up(c)).
table(up_v_maxc_c, 23, 11, 2, up(c)).
table(up_v_s_c, 23, 4, 2, up(c)).
table(up_v_mins_c, 23, 10, 2, up(c)).
table(up_v_maxs_c, 23, 18, 2, up(c)).
table(up_minc_s_c, 23, 3, 2, up(c)).
table(up_maxc_s_c, 23, 4, 2, up(c)).
table(up_s_mins_c, 23, 5, 2, up(c)).
table(up_s_maxs_c, 23, 5, 2, up(c)).
table(up_v_minc_maxc_c, 23, 7, 3, up(c)).
table(up_v_minc_s_c, 23, 6, 3, up(c)).
table(up_v_minc_mins_c, 23, 7, 3, up(c)).
table(up_v_minc_maxs_c, 23, 7, 3, up(c)).
table(up_v_maxc_s_c, 23, 5, 3, up(c)).
table(up_v_maxc_mins_c, 23, 5, 3, up(c)).
table(up_v_maxc_maxs_c, 23, 13, 3, up(c)).
table(up_v_s_mins_c, 23, 8, 3, up(c)).
table(up_v_s_maxs_c, 23, 6, 3, up(c)).
table(up_v_mins_maxs_c, 23, 12, 3, up(c)).
table(up_minc_maxc_s_c, 23, 5, 3, up(c)).
table(up_minc_s_mins_c, 23, 4, 3, up(c)).
table(up_minc_s_maxs_c, 23, 4, 3, up(c)).
table(up_maxc_s_mins_c, 23, 5, 3, up(c)).
table(up_maxc_s_maxs_c, 23, 5, 3, up(c)).
table(up_s_mins_maxs_c, 23, 8, 3, up(c)).
table(up_v_minc_maxc_s_c, 23, 7, 4, up(c)).
table(up_v_minc_maxc_mins_c, 23, 8, 4, up(c)).
table(up_v_minc_maxc_maxs_c, 23, 8, 4, up(c)).
table(up_v_minc_s_mins_c, 23, 7, 4, up(c)).
table(up_v_minc_s_maxs_c, 23, 7, 4, up(c)).
table(up_v_minc_mins_maxs_c, 23, 8, 4, up(c)).
table(up_v_maxc_s_mins_c, 23, 6, 4, up(c)).
table(up_v_maxc_s_maxs_c, 23, 6, 4, up(c)).
table(up_v_maxc_mins_maxs_c, 23, 7, 4, up(c)).
table(up_v_s_mins_maxs_c, 23, 10, 4, up(c)).
table(up_minc_maxc_s_mins_c, 23, 6, 4, up(c)).
table(up_minc_maxc_s_maxs_c, 23, 6, 4, up(c)).
table(up_minc_s_mins_maxs_c, 23, 5, 4, up(c)).
table(up_maxc_s_mins_maxs_c, 23, 6, 4, up(c)).
table(up_v_minc_maxc_s_mins_c, 23, 8, 5, up(c)).
table(up_v_minc_maxc_s_maxs_c, 23, 8, 5, up(c)).
table(up_v_minc_maxc_mins_maxs_c, 23, 9, 5, up(c)).
table(up_v_minc_s_mins_maxs_c, 23, 8, 5, up(c)).
table(up_v_maxc_s_mins_maxs_c, 23, 7, 5, up(c)).
table(up_minc_maxc_s_mins_maxs_c, 23, 7, 5, up(c)).
table(up_v_minc_maxc_s_mins_maxs_c, 23, 9, 6, up(c)).
table(low_v_minc, 23, 3, 1, low(minc)).
table(low_c_minc, 23, 3, 1, low(minc)).
table(low_maxc_minc, 23, 3, 1, low(minc)).
table(low_s_minc, 23, 3, 1, low(minc)).
table(low_mins_minc, 23, 2, 1, low(minc)).
table(low_maxs_minc, 23, 3, 1, low(minc)).
table(low_v_c_minc, 23, 5, 2, low(minc)).
table(low_v_maxc_minc, 23, 4, 2, low(minc)).
table(low_v_s_minc, 23, 4, 2, low(minc)).
table(low_v_mins_minc, 23, 3, 2, low(minc)).
table(low_v_maxs_minc, 23, 4, 2, low(minc)).
table(low_c_maxc_minc, 23, 4, 2, low(minc)).
table(low_c_s_minc, 23, 4, 2, low(minc)).
table(low_c_mins_minc, 23, 3, 2, low(minc)).
table(low_c_maxs_minc, 23, 4, 2, low(minc)).
table(low_maxc_s_minc, 23, 4, 2, low(minc)).
table(low_maxc_mins_minc, 23, 4, 2, low(minc)).
table(low_maxc_maxs_minc, 23, 4, 2, low(minc)).
table(low_s_mins_minc, 23, 3, 2, low(minc)).
table(low_s_maxs_minc, 23, 4, 2, low(minc)).
table(low_mins_maxs_minc, 23, 3, 2, low(minc)).
table(low_v_c_maxc_minc, 23, 6, 3, low(minc)).
table(low_v_c_s_minc, 23, 6, 3, low(minc)).
table(low_v_c_mins_minc, 23, 6, 3, low(minc)).
table(low_v_c_maxs_minc, 23, 6, 3, low(minc)).
table(low_v_maxc_s_minc, 23, 5, 3, low(minc)).
table(low_v_maxc_mins_minc, 23, 5, 3, low(minc)).
table(low_v_maxc_maxs_minc, 23, 5, 3, low(minc)).
table(low_v_s_mins_minc, 23, 4, 3, low(minc)).
table(low_v_s_maxs_minc, 23, 5, 3, low(minc)).
table(low_v_mins_maxs_minc, 23, 4, 3, low(minc)).
table(low_c_maxc_s_minc, 23, 5, 3, low(minc)).
table(low_c_maxc_mins_minc, 23, 5, 3, low(minc)).
table(low_c_maxc_maxs_minc, 23, 5, 3, low(minc)).
table(low_c_s_mins_minc, 23, 4, 3, low(minc)).
table(low_c_s_maxs_minc, 23, 6, 3, low(minc)).
table(low_c_mins_maxs_minc, 23, 4, 3, low(minc)).
table(low_maxc_s_mins_minc, 23, 5, 3, low(minc)).
table(low_maxc_s_maxs_minc, 23, 5, 3, low(minc)).
table(low_maxc_mins_maxs_minc, 23, 5, 3, low(minc)).
table(low_s_mins_maxs_minc, 23, 4, 3, low(minc)).
table(low_v_c_maxc_s_minc, 23, 7, 4, low(minc)).
table(low_v_c_maxc_mins_minc, 23, 7, 4, low(minc)).
table(low_v_c_maxc_maxs_minc, 23, 7, 4, low(minc)).
table(low_v_c_s_mins_minc, 23, 7, 4, low(minc)).
table(low_v_c_s_maxs_minc, 23, 7, 4, low(minc)).
table(low_v_c_mins_maxs_minc, 23, 7, 4, low(minc)).
table(low_v_maxc_s_mins_minc, 23, 6, 4, low(minc)).
table(low_v_maxc_s_maxs_minc, 23, 6, 4, low(minc)).
table(low_v_maxc_mins_maxs_minc, 23, 6, 4, low(minc)).
table(low_v_s_mins_maxs_minc, 23, 5, 4, low(minc)).
table(low_c_maxc_s_mins_minc, 23, 6, 4, low(minc)).
table(low_c_maxc_s_maxs_minc, 23, 6, 4, low(minc)).
table(low_c_maxc_mins_maxs_minc, 23, 6, 4, low(minc)).
table(low_c_s_mins_maxs_minc, 23, 6, 4, low(minc)).
table(low_maxc_s_mins_maxs_minc, 23, 6, 4, low(minc)).
table(low_v_c_maxc_s_mins_minc, 23, 8, 5, low(minc)).
table(low_v_c_maxc_s_maxs_minc, 23, 8, 5, low(minc)).
table(low_v_c_maxc_mins_maxs_minc, 23, 8, 5, low(minc)).
table(low_v_c_s_mins_maxs_minc, 23, 8, 5, low(minc)).
table(low_v_maxc_s_mins_maxs_minc, 23, 7, 5, low(minc)).
table(low_c_maxc_s_mins_maxs_minc, 23, 7, 5, low(minc)).
table(low_v_c_maxc_s_mins_maxs_minc, 23, 9, 6, low(minc)).
table(up_v_minc, 23, 11, 1, up(minc)).
table(up_maxc_minc, 23, 6, 1, up(minc)).
table(up_v_c_minc, 23, 5, 2, up(minc)).
table(up_v_maxc_minc, 23, 4, 2, up(minc)).
table(up_v_s_minc, 23, 13, 2, up(minc)).
table(up_v_mins_minc, 23, 12, 2, up(minc)).
table(up_v_maxs_minc, 23, 13, 2, up(minc)).
table(up_c_maxc_minc, 23, 4, 2, up(minc)).
table(up_maxc_s_minc, 23, 4, 2, up(minc)).
table(up_maxc_mins_minc, 23, 4, 2, up(minc)).
table(up_maxc_maxs_minc, 23, 7, 2, up(minc)).
table(up_s_maxs_minc, 23, 14, 2, up(minc)).
table(up_v_c_maxc_minc, 23, 6, 3, up(minc)).
table(up_v_c_s_minc, 23, 6, 3, up(minc)).
table(up_v_c_mins_minc, 23, 6, 3, up(minc)).
table(up_v_c_maxs_minc, 23, 6, 3, up(minc)).
table(up_v_maxc_s_minc, 23, 5, 3, up(minc)).
table(up_v_maxc_mins_minc, 23, 5, 3, up(minc)).
table(up_v_maxc_maxs_minc, 23, 5, 3, up(minc)).
table(up_v_s_mins_minc, 23, 14, 3, up(minc)).
table(up_v_s_maxs_minc, 23, 14, 3, up(minc)).
table(up_v_mins_maxs_minc, 23, 14, 3, up(minc)).
table(up_c_maxc_s_minc, 23, 5, 3, up(minc)).
table(up_c_maxc_mins_minc, 23, 5, 3, up(minc)).
table(up_c_maxc_maxs_minc, 23, 5, 3, up(minc)).
table(up_c_s_maxs_minc, 23, 4, 3, up(minc)).
table(up_maxc_s_mins_minc, 23, 5, 3, up(minc)).
table(up_maxc_s_maxs_minc, 23, 5, 3, up(minc)).
table(up_maxc_mins_maxs_minc, 23, 5, 3, up(minc)).
table(up_s_mins_maxs_minc, 23, 15, 3, up(minc)).
table(up_v_c_maxc_s_minc, 23, 7, 4, up(minc)).
table(up_v_c_maxc_mins_minc, 23, 7, 4, up(minc)).
table(up_v_c_maxc_maxs_minc, 23, 7, 4, up(minc)).
table(up_v_c_s_mins_minc, 23, 7, 4, up(minc)).
table(up_v_c_s_maxs_minc, 23, 7, 4, up(minc)).
table(up_v_c_mins_maxs_minc, 23, 7, 4, up(minc)).
table(up_v_maxc_s_mins_minc, 23, 6, 4, up(minc)).
table(up_v_maxc_s_maxs_minc, 23, 6, 4, up(minc)).
table(up_v_maxc_mins_maxs_minc, 23, 6, 4, up(minc)).
table(up_v_s_mins_maxs_minc, 23, 15, 4, up(minc)).
table(up_c_maxc_s_mins_minc, 23, 6, 4, up(minc)).
table(up_c_maxc_s_maxs_minc, 23, 6, 4, up(minc)).
table(up_c_maxc_mins_maxs_minc, 23, 6, 4, up(minc)).
table(up_c_s_mins_maxs_minc, 23, 5, 4, up(minc)).
table(up_maxc_s_mins_maxs_minc, 23, 6, 4, up(minc)).
table(up_v_c_maxc_s_mins_minc, 23, 8, 5, up(minc)).
table(up_v_c_maxc_s_maxs_minc, 23, 8, 5, up(minc)).
table(up_v_c_maxc_mins_maxs_minc, 23, 8, 5, up(minc)).
table(up_v_c_s_mins_maxs_minc, 23, 8, 5, up(minc)).
table(up_v_maxc_s_mins_maxs_minc, 23, 7, 5, up(minc)).
table(up_c_maxc_s_mins_maxs_minc, 23, 7, 5, up(minc)).
table(up_v_c_maxc_s_mins_maxs_minc, 23, 9, 6, up(minc)).
table(low_v_maxc, 23, 18, 1, low(maxc)).
table(low_c_maxc, 23, 18, 1, low(maxc)).
table(low_minc_maxc, 23, 6, 1, low(maxc)).
table(low_s_maxc, 23, 18, 1, low(maxc)).
table(low_mins_maxc, 23, 9, 1, low(maxc)).
table(low_maxs_maxc, 23, 2, 1, low(maxc)).
table(low_v_c_maxc, 23, 3, 2, low(maxc)).
table(low_v_minc_maxc, 23, 6, 2, low(maxc)).
table(low_v_s_maxc, 23, 4, 2, low(maxc)).
table(low_v_mins_maxc, 23, 9, 2, low(maxc)).
table(low_v_maxs_maxc, 23, 3, 2, low(maxc)).
table(low_c_minc_maxc, 23, 11, 2, low(maxc)).
table(low_c_s_maxc, 23, 4, 2, low(maxc)).
table(low_c_mins_maxc, 23, 18, 2, low(maxc)).
table(low_c_maxs_maxc, 23, 3, 2, low(maxc)).
table(low_minc_s_maxc, 23, 4, 2, low(maxc)).
table(low_minc_mins_maxc, 23, 4, 2, low(maxc)).
table(low_minc_maxs_maxc, 23, 4, 2, low(maxc)).
table(low_s_mins_maxc, 23, 18, 2, low(maxc)).
table(low_s_maxs_maxc, 23, 3, 2, low(maxc)).
table(low_mins_maxs_maxc, 23, 3, 2, low(maxc)).
table(low_v_c_minc_maxc, 23, 6, 3, low(maxc)).
table(low_v_c_s_maxc, 23, 4, 3, low(maxc)).
table(low_v_c_mins_maxc, 23, 4, 3, low(maxc)).
table(low_v_c_maxs_maxc, 23, 4, 3, low(maxc)).
table(low_v_minc_s_maxc, 23, 5, 3, low(maxc)).
table(low_v_minc_mins_maxc, 23, 5, 3, low(maxc)).
table(low_v_minc_maxs_maxc, 23, 5, 3, low(maxc)).
table(low_v_s_mins_maxc, 23, 5, 3, low(maxc)).
table(low_v_s_maxs_maxc, 23, 4, 3, low(maxc)).
table(low_v_mins_maxs_maxc, 23, 4, 3, low(maxc)).
table(low_c_minc_s_maxc, 23, 5, 3, low(maxc)).
table(low_c_minc_mins_maxc, 23, 6, 3, low(maxc)).
table(low_c_minc_maxs_maxc, 23, 5, 3, low(maxc)).
table(low_c_s_mins_maxc, 23, 4, 3, low(maxc)).
table(low_c_s_maxs_maxc, 23, 4, 3, low(maxc)).
table(low_c_mins_maxs_maxc, 23, 4, 3, low(maxc)).
table(low_minc_s_mins_maxc, 23, 5, 3, low(maxc)).
table(low_minc_s_maxs_maxc, 23, 5, 3, low(maxc)).
table(low_minc_mins_maxs_maxc, 23, 5, 3, low(maxc)).
table(low_s_mins_maxs_maxc, 23, 4, 3, low(maxc)).
table(low_v_c_minc_s_maxc, 23, 7, 4, low(maxc)).
table(low_v_c_minc_mins_maxc, 23, 7, 4, low(maxc)).
table(low_v_c_minc_maxs_maxc, 23, 7, 4, low(maxc)).
table(low_v_c_s_mins_maxc, 23, 5, 4, low(maxc)).
table(low_v_c_s_maxs_maxc, 23, 5, 4, low(maxc)).
table(low_v_c_mins_maxs_maxc, 23, 5, 4, low(maxc)).
table(low_v_minc_s_mins_maxc, 23, 6, 4, low(maxc)).
table(low_v_minc_s_maxs_maxc, 23, 6, 4, low(maxc)).
table(low_v_minc_mins_maxs_maxc, 23, 6, 4, low(maxc)).
table(low_v_s_mins_maxs_maxc, 23, 5, 4, low(maxc)).
table(low_c_minc_s_mins_maxc, 23, 6, 4, low(maxc)).
table(low_c_minc_s_maxs_maxc, 23, 6, 4, low(maxc)).
table(low_c_minc_mins_maxs_maxc, 23, 6, 4, low(maxc)).
table(low_c_s_mins_maxs_maxc, 23, 5, 4, low(maxc)).
table(low_minc_s_mins_maxs_maxc, 23, 6, 4, low(maxc)).
table(low_v_c_minc_s_mins_maxc, 23, 8, 5, low(maxc)).
table(low_v_c_minc_s_maxs_maxc, 23, 8, 5, low(maxc)).
table(low_v_c_minc_mins_maxs_maxc, 23, 8, 5, low(maxc)).
table(low_v_c_s_mins_maxs_maxc, 23, 6, 5, low(maxc)).
table(low_v_minc_s_mins_maxs_maxc, 23, 7, 5, low(maxc)).
table(low_c_minc_s_mins_maxs_maxc, 23, 7, 5, low(maxc)).
table(low_v_c_minc_s_mins_maxs_maxc, 23, 9, 6, low(maxc)).
table(up_v_maxc, 23, 11, 1, up(maxc)).
table(up_v_c_maxc, 23, 11, 2, up(maxc)).
table(up_v_minc_maxc, 23, 11, 2, up(maxc)).
table(up_v_s_maxc, 23, 13, 2, up(maxc)).
table(up_v_mins_maxc, 23, 12, 2, up(maxc)).
table(up_v_maxs_maxc, 23, 13, 2, up(maxc)).
table(up_v_c_minc_maxc, 23, 11, 3, up(maxc)).
table(up_v_c_s_maxc, 23, 13, 3, up(maxc)).
table(up_v_c_mins_maxc, 23, 12, 3, up(maxc)).
table(up_v_c_maxs_maxc, 23, 13, 3, up(maxc)).
table(up_v_minc_s_maxc, 23, 12, 3, up(maxc)).
table(up_v_minc_mins_maxc, 23, 12, 3, up(maxc)).
table(up_v_minc_maxs_maxc, 23, 12, 3, up(maxc)).
table(up_v_s_mins_maxc, 23, 14, 3, up(maxc)).
table(up_v_s_maxs_maxc, 23, 14, 3, up(maxc)).
table(up_v_mins_maxs_maxc, 23, 14, 3, up(maxc)).
table(up_v_c_minc_s_maxc, 23, 12, 4, up(maxc)).
table(up_v_c_minc_mins_maxc, 23, 12, 4, up(maxc)).
table(up_v_c_minc_maxs_maxc, 23, 12, 4, up(maxc)).
table(up_v_c_s_mins_maxc, 23, 14, 4, up(maxc)).
table(up_v_c_s_maxs_maxc, 23, 6, 4, up(maxc)).
table(up_v_c_mins_maxs_maxc, 23, 10, 4, up(maxc)).
table(up_v_minc_s_mins_maxc, 23, 13, 4, up(maxc)).
table(up_v_minc_s_maxs_maxc, 23, 13, 4, up(maxc)).
table(up_v_minc_mins_maxs_maxc, 23, 13, 4, up(maxc)).
table(up_v_s_mins_maxs_maxc, 23, 15, 4, up(maxc)).
table(up_v_c_minc_s_mins_maxc, 23, 13, 5, up(maxc)).
table(up_v_c_minc_s_maxs_maxc, 23, 8, 5, up(maxc)).
table(up_v_c_minc_mins_maxs_maxc, 23, 9, 5, up(maxc)).
table(up_v_c_s_mins_maxs_maxc, 23, 10, 5, up(maxc)).
table(up_v_minc_s_mins_maxs_maxc, 23, 14, 5, up(maxc)).
table(up_v_c_minc_s_mins_maxs_maxc, 23, 9, 6, up(maxc)).
table(low_v_s, 23, 18, 1, low(s)).
table(low_c_s, 23, 3, 1, low(s)).
table(low_minc_s, 23, 18, 1, low(s)).
table(low_maxc_s, 23, 18, 1, low(s)).
table(low_mins_s, 23, 18, 1, low(s)).
table(low_maxs_s, 23, 18, 1, low(s)).
table(low_v_c_s, 23, 4, 2, low(s)).
table(low_v_minc_s, 23, 18, 2, low(s)).
table(low_v_maxc_s, 23, 6, 2, low(s)).
table(low_v_mins_s, 23, 9, 2, low(s)).
table(low_v_maxs_s, 23, 3, 2, low(s)).
table(low_c_minc_s, 23, 5, 2, low(s)).
table(low_c_maxc_s, 23, 5, 2, low(s)).
table(low_c_mins_s, 23, 5, 2, low(s)).
table(low_c_maxs_s, 23, 5, 2, low(s)).
table(low_minc_maxc_s, 23, 18, 2, low(s)).
table(low_minc_mins_s, 23, 12, 2, low(s)).
table(low_minc_maxs_s, 23, 14, 2, low(s)).
table(low_maxc_mins_s, 23, 7, 2, low(s)).
table(low_maxc_maxs_s, 23, 14, 2, low(s)).
table(low_mins_maxs_s, 23, 12, 2, low(s)).
table(low_v_c_minc_s, 23, 8, 3, low(s)).
table(low_v_c_maxc_s, 23, 6, 3, low(s)).
table(low_v_c_mins_s, 23, 8, 3, low(s)).
table(low_v_c_maxs_s, 23, 4, 3, low(s)).
table(low_v_minc_maxc_s, 23, 10, 3, low(s)).
table(low_v_minc_mins_s, 23, 13, 3, low(s)).
table(low_v_minc_maxs_s, 23, 4, 3, low(s)).
table(low_v_maxc_mins_s, 23, 4, 3, low(s)).
table(low_v_maxc_maxs_s, 23, 4, 3, low(s)).
table(low_v_mins_maxs_s, 23, 4, 3, low(s)).
table(low_c_minc_maxc_s, 23, 8, 3, low(s)).
table(low_c_minc_mins_s, 23, 4, 3, low(s)).
table(low_c_minc_maxs_s, 23, 5, 3, low(s)).
table(low_c_maxc_mins_s, 23, 8, 3, low(s)).
table(low_c_maxc_maxs_s, 23, 5, 3, low(s)).
table(low_c_mins_maxs_s, 23, 8, 3, low(s)).
table(low_minc_maxc_mins_s, 23, 13, 3, low(s)).
table(low_minc_maxc_maxs_s, 23, 14, 3, low(s)).
table(low_minc_mins_maxs_s, 23, 8, 3, low(s)).
table(low_maxc_mins_maxs_s, 23, 9, 3, low(s)).
table(low_v_c_minc_maxc_s, 23, 10, 4, low(s)).
table(low_v_c_minc_mins_s, 23, 7, 4, low(s)).
table(low_v_c_minc_maxs_s, 23, 7, 4, low(s)).
table(low_v_c_maxc_mins_s, 23, 5, 4, low(s)).
table(low_v_c_maxc_maxs_s, 23, 5, 4, low(s)).
table(low_v_c_mins_maxs_s, 23, 5, 4, low(s)).
table(low_v_minc_maxc_mins_s, 23, 8, 4, low(s)).
table(low_v_minc_maxc_maxs_s, 23, 6, 4, low(s)).
table(low_v_minc_mins_maxs_s, 23, 5, 4, low(s)).
table(low_v_maxc_mins_maxs_s, 23, 5, 4, low(s)).
table(low_c_minc_maxc_mins_s, 23, 6, 4, low(s)).
table(low_c_minc_maxc_maxs_s, 23, 7, 4, low(s)).
table(low_c_minc_mins_maxs_s, 23, 5, 4, low(s)).
table(low_c_maxc_mins_maxs_s, 23, 6, 4, low(s)).
table(low_minc_maxc_mins_maxs_s, 23, 9, 4, low(s)).
table(low_v_c_minc_maxc_mins_s, 23, 8, 5, low(s)).
table(low_v_c_minc_maxc_maxs_s, 23, 8, 5, low(s)).
table(low_v_c_minc_mins_maxs_s, 23, 8, 5, low(s)).
table(low_v_c_maxc_mins_maxs_s, 23, 6, 5, low(s)).
table(low_v_minc_maxc_mins_maxs_s, 23, 7, 5, low(s)).
table(low_c_minc_maxc_mins_maxs_s, 23, 7, 5, low(s)).
table(low_v_c_minc_maxc_mins_maxs_s, 23, 9, 6, low(s)).
table(up_v_s, 23, 7, 1, up(s)).
table(up_v_c_s, 23, 8, 2, up(s)).
table(up_v_minc_s, 23, 8, 2, up(s)).
table(up_v_maxc_s, 23, 8, 2, up(s)).
table(up_v_mins_s, 23, 4, 2, up(s)).
table(up_v_maxs_s, 23, 7, 2, up(s)).
table(up_c_maxc_s, 23, 9, 2, up(s)).
table(up_v_c_minc_s, 23, 11, 3, up(s)).
table(up_v_c_maxc_s, 23, 9, 3, up(s)).
table(up_v_c_mins_s, 23, 5, 3, up(s)).
table(up_v_c_maxs_s, 23, 8, 3, up(s)).
table(up_v_minc_maxc_s, 23, 10, 3, up(s)).
table(up_v_minc_mins_s, 23, 5, 3, up(s)).
table(up_v_minc_maxs_s, 23, 8, 3, up(s)).
table(up_v_maxc_mins_s, 23, 5, 3, up(s)).
table(up_v_maxc_maxs_s, 23, 8, 3, up(s)).
table(up_v_mins_maxs_s, 23, 5, 3, up(s)).
table(up_c_minc_maxc_s, 23, 12, 3, up(s)).
table(up_c_maxc_mins_s, 23, 5, 3, up(s)).
table(up_c_maxc_maxs_s, 23, 9, 3, up(s)).
table(up_v_c_minc_maxc_s, 23, 12, 4, up(s)).
table(up_v_c_minc_mins_s, 23, 8, 4, up(s)).
table(up_v_c_minc_maxs_s, 23, 11, 4, up(s)).
table(up_v_c_maxc_mins_s, 23, 6, 4, up(s)).
table(up_v_c_maxc_maxs_s, 23, 9, 4, up(s)).
table(up_v_c_mins_maxs_s, 23, 6, 4, up(s)).
table(up_v_minc_maxc_mins_s, 23, 7, 4, up(s)).
table(up_v_minc_maxc_maxs_s, 23, 10, 4, up(s)).
table(up_v_minc_mins_maxs_s, 23, 6, 4, up(s)).
table(up_v_maxc_mins_maxs_s, 23, 6, 4, up(s)).
table(up_c_minc_maxc_mins_s, 23, 7, 4, up(s)).
table(up_c_minc_maxc_maxs_s, 23, 12, 4, up(s)).
table(up_c_maxc_mins_maxs_s, 23, 6, 4, up(s)).
table(up_v_c_minc_maxc_mins_s, 23, 9, 5, up(s)).
table(up_v_c_minc_maxc_maxs_s, 23, 12, 5, up(s)).
table(up_v_c_minc_mins_maxs_s, 23, 9, 5, up(s)).
table(up_v_c_maxc_mins_maxs_s, 23, 7, 5, up(s)).
table(up_v_minc_maxc_mins_maxs_s, 23, 8, 5, up(s)).
table(up_c_minc_maxc_mins_maxs_s, 23, 8, 5, up(s)).
table(up_v_c_minc_maxc_mins_maxs_s, 23, 10, 6, up(s)).
table(low_v_mins, 23, 2, 1, low(mins)).
table(low_c_mins, 23, 2, 1, low(mins)).
table(low_minc_mins, 23, 2, 1, low(mins)).
table(low_maxc_mins, 23, 2, 1, low(mins)).
table(low_s_mins, 23, 2, 1, low(mins)).
table(low_maxs_mins, 23, 2, 1, low(mins)).
table(low_v_c_mins, 23, 3, 2, low(mins)).
table(low_v_minc_mins, 23, 3, 2, low(mins)).
table(low_v_maxc_mins, 23, 3, 2, low(mins)).
table(low_v_s_mins, 23, 3, 2, low(mins)).
table(low_v_maxs_mins, 23, 3, 2, low(mins)).
table(low_c_minc_mins, 23, 3, 2, low(mins)).
table(low_c_maxc_mins, 23, 3, 2, low(mins)).
table(low_c_s_mins, 23, 3, 2, low(mins)).
table(low_c_maxs_mins, 23, 3, 2, low(mins)).
table(low_minc_maxc_mins, 23, 4, 2, low(mins)).
table(low_minc_s_mins, 23, 3, 2, low(mins)).
table(low_minc_maxs_mins, 23, 3, 2, low(mins)).
table(low_maxc_s_mins, 23, 3, 2, low(mins)).
table(low_maxc_maxs_mins, 23, 3, 2, low(mins)).
table(low_s_maxs_mins, 23, 3, 2, low(mins)).
table(low_v_c_minc_mins, 23, 6, 3, low(mins)).
table(low_v_c_maxc_mins, 23, 4, 3, low(mins)).
table(low_v_c_s_mins, 23, 4, 3, low(mins)).
table(low_v_c_maxs_mins, 23, 4, 3, low(mins)).
table(low_v_minc_maxc_mins, 23, 5, 3, low(mins)).
table(low_v_minc_s_mins, 23, 4, 3, low(mins)).
table(low_v_minc_maxs_mins, 23, 4, 3, low(mins)).
table(low_v_maxc_s_mins, 23, 4, 3, low(mins)).
table(low_v_maxc_maxs_mins, 23, 4, 3, low(mins)).
table(low_v_s_maxs_mins, 23, 4, 3, low(mins)).
table(low_c_minc_maxc_mins, 23, 5, 3, low(mins)).
table(low_c_minc_s_mins, 23, 4, 3, low(mins)).
table(low_c_minc_maxs_mins, 23, 4, 3, low(mins)).
table(low_c_maxc_s_mins, 23, 4, 3, low(mins)).
table(low_c_maxc_maxs_mins, 23, 4, 3, low(mins)).
table(low_c_s_maxs_mins, 23, 4, 3, low(mins)).
table(low_minc_maxc_s_mins, 23, 5, 3, low(mins)).
table(low_minc_maxc_maxs_mins, 23, 5, 3, low(mins)).
table(low_minc_s_maxs_mins, 23, 4, 3, low(mins)).
table(low_maxc_s_maxs_mins, 23, 4, 3, low(mins)).
table(low_v_c_minc_maxc_mins, 23, 7, 4, low(mins)).
table(low_v_c_minc_s_mins, 23, 7, 4, low(mins)).
table(low_v_c_minc_maxs_mins, 23, 7, 4, low(mins)).
table(low_v_c_maxc_s_mins, 23, 5, 4, low(mins)).
table(low_v_c_maxc_maxs_mins, 23, 5, 4, low(mins)).
table(low_v_c_s_maxs_mins, 23, 5, 4, low(mins)).
table(low_v_minc_maxc_s_mins, 23, 6, 4, low(mins)).
table(low_v_minc_maxc_maxs_mins, 23, 6, 4, low(mins)).
table(low_v_minc_s_maxs_mins, 23, 5, 4, low(mins)).
table(low_v_maxc_s_maxs_mins, 23, 5, 4, low(mins)).
table(low_c_minc_maxc_s_mins, 23, 6, 4, low(mins)).
table(low_c_minc_maxc_maxs_mins, 23, 6, 4, low(mins)).
table(low_c_minc_s_maxs_mins, 23, 5, 4, low(mins)).
table(low_c_maxc_s_maxs_mins, 23, 5, 4, low(mins)).
table(low_minc_maxc_s_maxs_mins, 23, 6, 4, low(mins)).
table(low_v_c_minc_maxc_s_mins, 23, 8, 5, low(mins)).
table(low_v_c_minc_maxc_maxs_mins, 23, 8, 5, low(mins)).
table(low_v_c_minc_s_maxs_mins, 23, 8, 5, low(mins)).
table(low_v_c_maxc_s_maxs_mins, 23, 6, 5, low(mins)).
table(low_v_minc_maxc_s_maxs_mins, 23, 7, 5, low(mins)).
table(low_c_minc_maxc_s_maxs_mins, 23, 7, 5, low(mins)).
table(low_v_c_minc_maxc_s_maxs_mins, 23, 9, 6, low(mins)).
table(up_v_mins, 23, 18, 1, up(mins)).
table(up_maxc_mins, 23, 9, 1, up(mins)).
table(up_maxs_mins, 23, 2, 1, up(mins)).
table(up_v_c_mins, 23, 6, 2, up(mins)).
table(up_v_minc_mins, 23, 3, 2, up(mins)).
table(up_v_maxc_mins, 23, 5, 2, up(mins)).
table(up_v_s_mins, 23, 3, 2, up(mins)).
table(up_v_maxs_mins, 23, 3, 2, up(mins)).
table(up_c_maxc_mins, 23, 5, 2, up(mins)).
table(up_c_maxs_mins, 23, 3, 2, up(mins)).
table(up_minc_maxc_mins, 23, 4, 2, up(mins)).
table(up_minc_maxs_mins, 23, 3, 2, up(mins)).
table(up_maxc_s_mins, 23, 3, 2, up(mins)).
table(up_maxc_maxs_mins, 23, 3, 2, up(mins)).
table(up_s_maxs_mins, 23, 3, 2, up(mins)).
table(up_v_c_minc_mins, 23, 6, 3, up(mins)).
table(up_v_c_maxc_mins, 23, 7, 3, up(mins)).
table(up_v_c_s_mins, 23, 4, 3, up(mins)).
table(up_v_c_maxs_mins, 23, 4, 3, up(mins)).
table(up_v_minc_maxc_mins, 23, 5, 3, up(mins)).
table(up_v_minc_s_mins, 23, 4, 3, up(mins)).
table(up_v_minc_maxs_mins, 23, 4, 3, up(mins)).
table(up_v_maxc_s_mins, 23, 4, 3, up(mins)).
table(up_v_maxc_maxs_mins, 23, 4, 3, up(mins)).
table(up_v_s_maxs_mins, 23, 4, 3, up(mins)).
table(up_c_minc_maxc_mins, 23, 5, 3, up(mins)).
table(up_c_minc_maxs_mins, 23, 4, 3, up(mins)).
table(up_c_maxc_s_mins, 23, 4, 3, up(mins)).
table(up_c_maxc_maxs_mins, 23, 4, 3, up(mins)).
table(up_c_s_maxs_mins, 23, 4, 3, up(mins)).
table(up_minc_maxc_s_mins, 23, 5, 3, up(mins)).
table(up_minc_maxc_maxs_mins, 23, 5, 3, up(mins)).
table(up_minc_s_maxs_mins, 23, 4, 3, up(mins)).
table(up_maxc_s_maxs_mins, 23, 4, 3, up(mins)).
table(up_v_c_minc_maxc_mins, 23, 7, 4, up(mins)).
table(up_v_c_minc_s_mins, 23, 7, 4, up(mins)).
table(up_v_c_minc_maxs_mins, 23, 7, 4, up(mins)).
table(up_v_c_maxc_s_mins, 23, 5, 4, up(mins)).
table(up_v_c_maxc_maxs_mins, 23, 5, 4, up(mins)).
table(up_v_c_s_maxs_mins, 23, 5, 4, up(mins)).
table(up_v_minc_maxc_s_mins, 23, 6, 4, up(mins)).
table(up_v_minc_maxc_maxs_mins, 23, 6, 4, up(mins)).
table(up_v_minc_s_maxs_mins, 23, 5, 4, up(mins)).
table(up_v_maxc_s_maxs_mins, 23, 5, 4, up(mins)).
table(up_c_minc_maxc_s_mins, 23, 6, 4, up(mins)).
table(up_c_minc_maxc_maxs_mins, 23, 6, 4, up(mins)).
table(up_c_minc_s_maxs_mins, 23, 5, 4, up(mins)).
table(up_c_maxc_s_maxs_mins, 23, 5, 4, up(mins)).
table(up_minc_maxc_s_maxs_mins, 23, 6, 4, up(mins)).
table(up_v_c_minc_maxc_s_mins, 23, 8, 5, up(mins)).
table(up_v_c_minc_maxc_maxs_mins, 23, 8, 5, up(mins)).
table(up_v_c_minc_s_maxs_mins, 23, 8, 5, up(mins)).
table(up_v_c_maxc_s_maxs_mins, 23, 6, 5, up(mins)).
table(up_v_minc_maxc_s_maxs_mins, 23, 7, 5, up(mins)).
table(up_c_minc_maxc_s_maxs_mins, 23, 7, 5, up(mins)).
table(up_v_c_minc_maxc_s_maxs_mins, 23, 9, 6, up(mins)).
table(low_v_maxs, 23, 7, 1, low(maxs)).
table(low_c_maxs, 23, 4, 1, low(maxs)).
table(low_minc_maxs, 23, 4, 1, low(maxs)).
table(low_maxc_maxs, 23, 4, 1, low(maxs)).
table(low_s_maxs, 23, 7, 1, low(maxs)).
table(low_mins_maxs, 23, 2, 1, low(maxs)).
table(low_v_c_maxs, 23, 8, 2, low(maxs)).
table(low_v_minc_maxs, 23, 8, 2, low(maxs)).
table(low_v_maxc_maxs, 23, 8, 2, low(maxs)).
table(low_v_s_maxs, 23, 3, 2, low(maxs)).
table(low_v_mins_maxs, 23, 3, 2, low(maxs)).
table(low_c_minc_maxs, 23, 5, 2, low(maxs)).
table(low_c_maxc_maxs, 23, 5, 2, low(maxs)).
table(low_c_s_maxs, 23, 8, 2, low(maxs)).
table(low_c_mins_maxs, 23, 3, 2, low(maxs)).
table(low_minc_maxc_maxs, 23, 6, 2, low(maxs)).
table(low_minc_s_maxs, 23, 3, 2, low(maxs)).
table(low_minc_mins_maxs, 23, 3, 2, low(maxs)).
table(low_maxc_s_maxs, 23, 3, 2, low(maxs)).
table(low_maxc_mins_maxs, 23, 3, 2, low(maxs)).
table(low_s_mins_maxs, 23, 7, 2, low(maxs)).
table(low_v_c_minc_maxs, 23, 11, 3, low(maxs)).
table(low_v_c_maxc_maxs, 23, 9, 3, low(maxs)).
table(low_v_c_s_maxs, 23, 4, 3, low(maxs)).
table(low_v_c_mins_maxs, 23, 4, 3, low(maxs)).
table(low_v_minc_maxc_maxs, 23, 10, 3, low(maxs)).
table(low_v_minc_s_maxs, 23, 4, 3, low(maxs)).
table(low_v_minc_mins_maxs, 23, 4, 3, low(maxs)).
table(low_v_maxc_s_maxs, 23, 4, 3, low(maxs)).
table(low_v_maxc_mins_maxs, 23, 4, 3, low(maxs)).
table(low_v_s_mins_maxs, 23, 4, 3, low(maxs)).
table(low_c_minc_maxc_maxs, 23, 7, 3, low(maxs)).
table(low_c_minc_s_maxs, 23, 4, 3, low(maxs)).
table(low_c_minc_mins_maxs, 23, 4, 3, low(maxs)).
table(low_c_maxc_s_maxs, 23, 4, 3, low(maxs)).
table(low_c_maxc_mins_maxs, 23, 4, 3, low(maxs)).
table(low_c_s_mins_maxs, 23, 8, 3, low(maxs)).
table(low_minc_maxc_s_maxs, 23, 5, 3, low(maxs)).
table(low_minc_maxc_mins_maxs, 23, 5, 3, low(maxs)).
table(low_minc_s_mins_maxs, 23, 4, 3, low(maxs)).
table(low_maxc_s_mins_maxs, 23, 4, 3, low(maxs)).
table(low_v_c_minc_maxc_maxs, 23, 12, 4, low(maxs)).
table(low_v_c_minc_s_maxs, 23, 7, 4, low(maxs)).
table(low_v_c_minc_mins_maxs, 23, 7, 4, low(maxs)).
table(low_v_c_maxc_s_maxs, 23, 5, 4, low(maxs)).
table(low_v_c_maxc_mins_maxs, 23, 5, 4, low(maxs)).
table(low_v_c_s_mins_maxs, 23, 5, 4, low(maxs)).
table(low_v_minc_maxc_s_maxs, 23, 6, 4, low(maxs)).
table(low_v_minc_maxc_mins_maxs, 23, 6, 4, low(maxs)).
table(low_v_minc_s_mins_maxs, 23, 5, 4, low(maxs)).
table(low_v_maxc_s_mins_maxs, 23, 5, 4, low(maxs)).
table(low_c_minc_maxc_s_maxs, 23, 6, 4, low(maxs)).
table(low_c_minc_maxc_mins_maxs, 23, 6, 4, low(maxs)).
table(low_c_minc_s_mins_maxs, 23, 5, 4, low(maxs)).
table(low_c_maxc_s_mins_maxs, 23, 5, 4, low(maxs)).
table(low_minc_maxc_s_mins_maxs, 23, 6, 4, low(maxs)).
table(low_v_c_minc_maxc_s_maxs, 23, 8, 5, low(maxs)).
table(low_v_c_minc_maxc_mins_maxs, 23, 8, 5, low(maxs)).
table(low_v_c_minc_s_mins_maxs, 23, 8, 5, low(maxs)).
table(low_v_c_maxc_s_mins_maxs, 23, 6, 5, low(maxs)).
table(low_v_minc_maxc_s_mins_maxs, 23, 7, 5, low(maxs)).
table(low_c_minc_maxc_s_mins_maxs, 23, 7, 5, low(maxs)).
table(low_v_c_minc_maxc_s_mins_maxs, 23, 9, 6, low(maxs)).
table(up_v_maxs, 23, 18, 1, up(maxs)).
table(up_maxc_maxs, 23, 2, 1, up(maxs)).
table(up_v_c_maxs, 23, 18, 2, up(maxs)).
table(up_v_minc_maxs, 23, 12, 2, up(maxs)).
table(up_v_maxc_maxs, 23, 3, 2, up(maxs)).
table(up_v_s_maxs, 23, 7, 2, up(maxs)).
table(up_v_mins_maxs, 23, 9, 2, up(maxs)).
table(up_c_maxc_maxs, 23, 3, 2, up(maxs)).
table(up_minc_maxc_maxs, 23, 4, 2, up(maxs)).
table(up_maxc_s_maxs, 23, 3, 2, up(maxs)).
table(up_maxc_mins_maxs, 23, 3, 2, up(maxs)).
table(up_v_c_minc_maxs, 23, 12, 3, up(maxs)).
table(up_v_c_maxc_maxs, 23, 4, 3, up(maxs)).
table(up_v_c_s_maxs, 23, 8, 3, up(maxs)).
table(up_v_c_mins_maxs, 23, 18, 3, up(maxs)).
table(up_v_minc_maxc_maxs, 23, 5, 3, up(maxs)).
table(up_v_minc_s_maxs, 23, 5, 3, up(maxs)).
table(up_v_minc_mins_maxs, 23, 14, 3, up(maxs)).
table(up_v_maxc_s_maxs, 23, 4, 3, up(maxs)).
table(up_v_maxc_mins_maxs, 23, 4, 3, up(maxs)).
table(up_v_s_mins_maxs, 23, 7, 3, up(maxs)).
table(up_c_minc_maxc_maxs, 23, 5, 3, up(maxs)).
table(up_c_maxc_s_maxs, 23, 4, 3, up(maxs)).
table(up_c_maxc_mins_maxs, 23, 4, 3, up(maxs)).
table(up_minc_maxc_s_maxs, 23, 5, 3, up(maxs)).
table(up_minc_maxc_mins_maxs, 23, 5, 3, up(maxs)).
table(up_maxc_s_mins_maxs, 23, 4, 3, up(maxs)).
table(up_v_c_minc_maxc_maxs, 23, 7, 4, up(maxs)).
table(up_v_c_minc_s_maxs, 23, 7, 4, up(maxs)).
table(up_v_c_minc_mins_maxs, 23, 13, 4, up(maxs)).
table(up_v_c_maxc_s_maxs, 23, 5, 4, up(maxs)).
table(up_v_c_maxc_mins_maxs, 23, 5, 4, up(maxs)).
table(up_v_c_s_mins_maxs, 23, 8, 4, up(maxs)).
table(up_v_minc_maxc_s_maxs, 23, 6, 4, up(maxs)).
table(up_v_minc_maxc_mins_maxs, 23, 6, 4, up(maxs)).
table(up_v_minc_s_mins_maxs, 23, 5, 4, up(maxs)).
table(up_v_maxc_s_mins_maxs, 23, 5, 4, up(maxs)).
table(up_c_minc_maxc_s_maxs, 23, 6, 4, up(maxs)).
table(up_c_minc_maxc_mins_maxs, 23, 6, 4, up(maxs)).
table(up_c_maxc_s_mins_maxs, 23, 5, 4, up(maxs)).
table(up_minc_maxc_s_mins_maxs, 23, 6, 4, up(maxs)).
table(up_v_c_minc_maxc_s_maxs, 23, 8, 5, up(maxs)).
table(up_v_c_minc_maxc_mins_maxs, 23, 8, 5, up(maxs)).
table(up_v_c_minc_s_mins_maxs, 23, 8, 5, up(maxs)).
table(up_v_c_maxc_s_mins_maxs, 23, 6, 5, up(maxs)).
table(up_v_minc_maxc_s_mins_maxs, 23, 7, 5, up(maxs)).
table(up_c_minc_maxc_s_mins_maxs, 23, 7, 5, up(maxs)).
table(up_v_c_minc_maxc_s_mins_maxs, 23, 9, 6, up(maxs)).
table(low_v_mina, 23, 12, 1, low(mina)).
table(low_c_mina, 23, 7, 1, low(mina)).
table(low_minc_mina, 23, 19, 1, low(mina)).
table(low_maxc_mina, 23, 19, 1, low(mina)).
table(low_s_mina, 23, 12, 1, low(mina)).
table(low_mins_mina, 23, 11, 1, low(mina)).
table(low_maxs_mina, 23, 11, 1, low(mina)).
table(low_v_c_mina, 23, 10, 2, low(mina)).
table(low_v_minc_mina, 23, 11, 2, low(mina)).
table(low_v_maxc_mina, 23, 11, 2, low(mina)).
table(low_v_s_mina, 23, 4, 2, low(mina)).
table(low_v_mins_mina, 23, 4, 2, low(mina)).
table(low_v_maxs_mina, 23, 10, 2, low(mina)).
table(low_c_minc_mina, 23, 8, 2, low(mina)).
table(low_c_maxc_mina, 23, 6, 2, low(mina)).
table(low_c_s_mina, 23, 10, 2, low(mina)).
table(low_c_mins_mina, 23, 7, 2, low(mina)).
table(low_c_maxs_mina, 23, 5, 2, low(mina)).
table(low_minc_maxc_mina, 23, 19, 2, low(mina)).
table(low_minc_s_mina, 23, 11, 2, low(mina)).
table(low_minc_mins_mina, 23, 19, 2, low(mina)).
table(low_minc_maxs_mina, 23, 19, 2, low(mina)).
table(low_maxc_s_mina, 23, 11, 2, low(mina)).
table(low_maxc_mins_mina, 23, 19, 2, low(mina)).
table(low_maxc_maxs_mina, 23, 19, 2, low(mina)).
table(low_s_mins_mina, 23, 12, 2, low(mina)).
table(low_s_maxs_mina, 23, 10, 2, low(mina)).
table(low_mins_maxs_mina, 23, 6, 2, low(mina)).
table(low_v_c_minc_mina, 23, 13, 3, low(mina)).
table(low_v_c_maxc_mina, 23, 11, 3, low(mina)).
table(low_v_c_s_mina, 23, 4, 3, low(mina)).
table(low_v_c_mins_mina, 23, 7, 3, low(mina)).
table(low_v_c_maxs_mina, 23, 10, 3, low(mina)).
table(low_v_minc_maxc_mina, 23, 12, 3, low(mina)).
table(low_v_minc_s_mina, 23, 4, 3, low(mina)).
table(low_v_minc_mins_mina, 23, 4, 3, low(mina)).
table(low_v_minc_maxs_mina, 23, 11, 3, low(mina)).
table(low_v_maxc_s_mina, 23, 4, 3, low(mina)).
table(low_v_maxc_mins_mina, 23, 6, 3, low(mina)).
table(low_v_maxc_maxs_mina, 23, 10, 3, low(mina)).
table(low_v_s_mins_mina, 23, 5, 3, low(mina)).
table(low_v_s_maxs_mina, 23, 5, 3, low(mina)).
table(low_v_mins_maxs_mina, 23, 5, 3, low(mina)).
table(low_c_minc_maxc_mina, 23, 8, 3, low(mina)).
table(low_c_minc_s_mina, 23, 11, 3, low(mina)).
table(low_c_minc_mins_mina, 23, 8, 3, low(mina)).
table(low_c_minc_maxs_mina, 23, 6, 3, low(mina)).
table(low_c_maxc_s_mina, 23, 8, 3, low(mina)).
table(low_c_maxc_mins_mina, 23, 6, 3, low(mina)).
table(low_c_maxc_maxs_mina, 23, 6, 3, low(mina)).
table(low_c_s_mins_mina, 23, 10, 3, low(mina)).
table(low_c_s_maxs_mina, 23, 10, 3, low(mina)).
table(low_c_mins_maxs_mina, 23, 5, 3, low(mina)).
table(low_minc_maxc_s_mina, 23, 10, 3, low(mina)).
table(low_minc_maxc_mins_mina, 23, 16, 3, low(mina)).
table(low_minc_maxc_maxs_mina, 23, 18, 3, low(mina)).
table(low_minc_s_mins_mina, 23, 6, 3, low(mina)).
table(low_minc_s_maxs_mina, 23, 8, 3, low(mina)).
table(low_minc_mins_maxs_mina, 23, 17, 3, low(mina)).
table(low_maxc_s_mins_mina, 23, 8, 3, low(mina)).
table(low_maxc_s_maxs_mina, 23, 7, 3, low(mina)).
table(low_maxc_mins_maxs_mina, 23, 8, 3, low(mina)).
table(low_s_mins_maxs_mina, 23, 10, 3, low(mina)).
table(low_v_c_minc_maxc_mina, 23, 14, 4, low(mina)).
table(low_v_c_minc_s_mina, 23, 8, 4, low(mina)).
table(low_v_c_minc_mins_mina, 23, 10, 4, low(mina)).
table(low_v_c_minc_maxs_mina, 23, 13, 4, low(mina)).
table(low_v_c_maxc_s_mina, 23, 5, 4, low(mina)).
table(low_v_c_maxc_mins_mina, 23, 8, 4, low(mina)).
table(low_v_c_maxc_maxs_mina, 23, 11, 4, low(mina)).
table(low_v_c_s_mins_mina, 23, 5, 4, low(mina)).
table(low_v_c_s_maxs_mina, 23, 5, 4, low(mina)).
table(low_v_c_mins_maxs_mina, 23, 8, 4, low(mina)).
table(low_v_minc_maxc_s_mina, 23, 6, 4, low(mina)).
table(low_v_minc_maxc_mins_mina, 23, 6, 4, low(mina)).
table(low_v_minc_maxc_maxs_mina, 23, 12, 4, low(mina)).
table(low_v_minc_s_mins_mina, 23, 5, 4, low(mina)).
table(low_v_minc_s_maxs_mina, 23, 5, 4, low(mina)).
table(low_v_minc_mins_maxs_mina, 23, 5, 4, low(mina)).
table(low_v_maxc_s_mins_mina, 23, 5, 4, low(mina)).
table(low_v_maxc_s_maxs_mina, 23, 5, 4, low(mina)).
table(low_v_maxc_mins_maxs_mina, 23, 5, 4, low(mina)).
table(low_v_s_mins_maxs_mina, 23, 6, 4, low(mina)).
table(low_c_minc_maxc_s_mina, 23, 9, 4, low(mina)).
table(low_c_minc_maxc_mins_mina, 23, 7, 4, low(mina)).
table(low_c_minc_maxc_maxs_mina, 23, 8, 4, low(mina)).
table(low_c_minc_s_mins_mina, 23, 10, 4, low(mina)).
table(low_c_minc_s_maxs_mina, 23, 10, 4, low(mina)).
table(low_c_minc_mins_maxs_mina, 23, 6, 4, low(mina)).
table(low_c_maxc_s_mins_mina, 23, 6, 4, low(mina)).
table(low_c_maxc_s_maxs_mina, 23, 6, 4, low(mina)).
table(low_c_maxc_mins_maxs_mina, 23, 6, 4, low(mina)).
table(low_c_s_mins_maxs_mina, 23, 10, 4, low(mina)).
table(low_minc_maxc_s_mins_mina, 23, 7, 4, low(mina)).
table(low_minc_maxc_s_maxs_mina, 23, 9, 4, low(mina)).
table(low_minc_maxc_mins_maxs_mina, 23, 16, 4, low(mina)).
table(low_minc_s_mins_maxs_mina, 23, 6, 4, low(mina)).
table(low_maxc_s_mins_maxs_mina, 23, 6, 4, low(mina)).
table(low_v_c_minc_maxc_s_mina, 23, 9, 5, low(mina)).
table(low_v_c_minc_maxc_mins_mina, 23, 11, 5, low(mina)).
table(low_v_c_minc_maxc_maxs_mina, 23, 14, 5, low(mina)).
table(low_v_c_minc_s_mins_mina, 23, 8, 5, low(mina)).
table(low_v_c_minc_s_maxs_mina, 23, 8, 5, low(mina)).
table(low_v_c_minc_mins_maxs_mina, 23, 11, 5, low(mina)).
table(low_v_c_maxc_s_mins_mina, 23, 6, 5, low(mina)).
table(low_v_c_maxc_s_maxs_mina, 23, 6, 5, low(mina)).
table(low_v_c_maxc_mins_maxs_mina, 23, 9, 5, low(mina)).
table(low_v_c_s_mins_maxs_mina, 23, 6, 5, low(mina)).
table(low_v_minc_maxc_s_mins_mina, 23, 7, 5, low(mina)).
table(low_v_minc_maxc_s_maxs_mina, 23, 7, 5, low(mina)).
table(low_v_minc_maxc_mins_maxs_mina, 23, 7, 5, low(mina)).
table(low_v_minc_s_mins_maxs_mina, 23, 6, 5, low(mina)).
table(low_v_maxc_s_mins_maxs_mina, 23, 6, 5, low(mina)).
table(low_c_minc_maxc_s_mins_mina, 23, 8, 5, low(mina)).
table(low_c_minc_maxc_s_maxs_mina, 23, 8, 5, low(mina)).
table(low_c_minc_maxc_mins_maxs_mina, 23, 8, 5, low(mina)).
table(low_c_minc_s_mins_maxs_mina, 23, 11, 5, low(mina)).
table(low_c_maxc_s_mins_maxs_mina, 23, 7, 5, low(mina)).
table(low_minc_maxc_s_mins_maxs_mina, 23, 8, 5, low(mina)).
table(low_v_c_minc_maxc_s_mins_mina, 23, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_maxs_mina, 23, 9, 6, low(mina)).
table(low_v_c_minc_maxc_mins_maxs_mina, 23, 12, 6, low(mina)).
table(low_v_c_minc_s_mins_maxs_mina, 23, 9, 6, low(mina)).
table(low_v_c_maxc_s_mins_maxs_mina, 23, 7, 6, low(mina)).
table(low_v_minc_maxc_s_mins_maxs_mina, 23, 8, 6, low(mina)).
table(low_c_minc_maxc_s_mins_maxs_mina, 23, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_mins_maxs_mina, 23, 10, 7, low(mina)).
table(up_v_maxa, 23, 19, 1, up(maxa)).
table(up_v_c_maxa, 23, 19, 2, up(maxa)).
table(up_v_minc_maxa, 23, 19, 2, up(maxa)).
table(up_v_maxc_maxa, 23, 19, 2, up(maxa)).
table(up_v_s_maxa, 23, 19, 2, up(maxa)).
table(up_v_mins_maxa, 23, 19, 2, up(maxa)).
table(up_v_maxs_maxa, 23, 19, 2, up(maxa)).
table(up_c_maxc_maxa, 23, 19, 2, up(maxa)).
table(up_maxc_s_maxa, 23, 18, 2, up(maxa)).
table(up_s_maxs_maxa, 23, 19, 2, up(maxa)).
table(up_v_c_minc_maxa, 23, 19, 3, up(maxa)).
table(up_v_c_maxc_maxa, 23, 19, 3, up(maxa)).
table(up_v_c_s_maxa, 23, 19, 3, up(maxa)).
table(up_v_c_mins_maxa, 23, 19, 3, up(maxa)).
table(up_v_c_maxs_maxa, 23, 7, 3, up(maxa)).
table(up_v_minc_maxc_maxa, 23, 19, 3, up(maxa)).
table(up_v_minc_s_maxa, 23, 18, 3, up(maxa)).
table(up_v_minc_mins_maxa, 23, 19, 3, up(maxa)).
table(up_v_minc_maxs_maxa, 23, 19, 3, up(maxa)).
table(up_v_maxc_s_maxa, 23, 18, 3, up(maxa)).
table(up_v_maxc_mins_maxa, 23, 11, 3, up(maxa)).
table(up_v_maxc_maxs_maxa, 23, 6, 3, up(maxa)).
table(up_v_s_mins_maxa, 23, 19, 3, up(maxa)).
table(up_v_s_maxs_maxa, 23, 19, 3, up(maxa)).
table(up_v_mins_maxs_maxa, 23, 19, 3, up(maxa)).
table(up_c_minc_maxc_maxa, 23, 19, 3, up(maxa)).
table(up_c_maxc_s_maxa, 23, 18, 3, up(maxa)).
table(up_c_maxc_mins_maxa, 23, 19, 3, up(maxa)).
table(up_c_maxc_maxs_maxa, 23, 6, 3, up(maxa)).
table(up_c_s_maxs_maxa, 23, 19, 3, up(maxa)).
table(up_minc_maxc_s_maxa, 23, 5, 3, up(maxa)).
table(up_minc_s_maxs_maxa, 23, 18, 3, up(maxa)).
table(up_maxc_s_mins_maxa, 23, 8, 3, up(maxa)).
table(up_maxc_s_maxs_maxa, 23, 4, 3, up(maxa)).
table(up_s_mins_maxs_maxa, 23, 19, 3, up(maxa)).
table(up_v_c_minc_maxc_maxa, 23, 19, 4, up(maxa)).
table(up_v_c_minc_s_maxa, 23, 18, 4, up(maxa)).
table(up_v_c_minc_mins_maxa, 23, 11, 4, up(maxa)).
table(up_v_c_minc_maxs_maxa, 23, 8, 4, up(maxa)).
table(up_v_c_maxc_s_maxa, 23, 18, 4, up(maxa)).
table(up_v_c_maxc_mins_maxa, 23, 11, 4, up(maxa)).
table(up_v_c_maxc_maxs_maxa, 23, 6, 4, up(maxa)).
table(up_v_c_s_mins_maxa, 23, 19, 4, up(maxa)).
table(up_v_c_s_maxs_maxa, 23, 19, 4, up(maxa)).
table(up_v_c_mins_maxs_maxa, 23, 9, 4, up(maxa)).
table(up_v_minc_maxc_s_maxa, 23, 18, 4, up(maxa)).
table(up_v_minc_maxc_mins_maxa, 23, 14, 4, up(maxa)).
table(up_v_minc_maxc_maxs_maxa, 23, 9, 4, up(maxa)).
table(up_v_minc_s_mins_maxa, 23, 18, 4, up(maxa)).
table(up_v_minc_s_maxs_maxa, 23, 18, 4, up(maxa)).
table(up_v_minc_mins_maxs_maxa, 23, 16, 4, up(maxa)).
table(up_v_maxc_s_mins_maxa, 23, 8, 4, up(maxa)).
table(up_v_maxc_s_maxs_maxa, 23, 7, 4, up(maxa)).
table(up_v_maxc_mins_maxs_maxa, 23, 5, 4, up(maxa)).
table(up_v_s_mins_maxs_maxa, 23, 19, 4, up(maxa)).
table(up_c_minc_maxc_s_maxa, 23, 18, 4, up(maxa)).
table(up_c_minc_maxc_mins_maxa, 23, 14, 4, up(maxa)).
table(up_c_minc_maxc_maxs_maxa, 23, 9, 4, up(maxa)).
table(up_c_minc_s_maxs_maxa, 23, 9, 4, up(maxa)).
table(up_c_maxc_s_mins_maxa, 23, 8, 4, up(maxa)).
table(up_c_maxc_s_maxs_maxa, 23, 6, 4, up(maxa)).
table(up_c_maxc_mins_maxs_maxa, 23, 5, 4, up(maxa)).
table(up_c_s_mins_maxs_maxa, 23, 19, 4, up(maxa)).
table(up_minc_maxc_s_mins_maxa, 23, 6, 4, up(maxa)).
table(up_minc_maxc_s_maxs_maxa, 23, 6, 4, up(maxa)).
table(up_minc_s_mins_maxs_maxa, 23, 18, 4, up(maxa)).
table(up_maxc_s_mins_maxs_maxa, 23, 5, 4, up(maxa)).
table(up_v_c_minc_maxc_s_maxa, 23, 18, 5, up(maxa)).
table(up_v_c_minc_maxc_mins_maxa, 23, 14, 5, up(maxa)).
table(up_v_c_minc_maxc_maxs_maxa, 23, 9, 5, up(maxa)).
table(up_v_c_minc_s_mins_maxa, 23, 10, 5, up(maxa)).
table(up_v_c_minc_s_maxs_maxa, 23, 9, 5, up(maxa)).
table(up_v_c_minc_mins_maxs_maxa, 23, 8, 5, up(maxa)).
table(up_v_c_maxc_s_mins_maxa, 23, 8, 5, up(maxa)).
table(up_v_c_maxc_s_maxs_maxa, 23, 6, 5, up(maxa)).
table(up_v_c_maxc_mins_maxs_maxa, 23, 6, 5, up(maxa)).
table(up_v_c_s_mins_maxs_maxa, 23, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_mins_maxa, 23, 11, 5, up(maxa)).
table(up_v_minc_maxc_s_maxs_maxa, 23, 10, 5, up(maxa)).
table(up_v_minc_maxc_mins_maxs_maxa, 23, 7, 5, up(maxa)).
table(up_v_minc_s_mins_maxs_maxa, 23, 16, 5, up(maxa)).
table(up_v_maxc_s_mins_maxs_maxa, 23, 8, 5, up(maxa)).
table(up_c_minc_maxc_s_mins_maxa, 23, 11, 5, up(maxa)).
table(up_c_minc_maxc_s_maxs_maxa, 23, 9, 5, up(maxa)).
table(up_c_minc_maxc_mins_maxs_maxa, 23, 7, 5, up(maxa)).
table(up_c_minc_s_mins_maxs_maxa, 23, 10, 5, up(maxa)).
table(up_c_maxc_s_mins_maxs_maxa, 23, 6, 5, up(maxa)).
table(up_minc_maxc_s_mins_maxs_maxa, 23, 7, 5, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxa, 23, 11, 6, up(maxa)).
table(up_v_c_minc_maxc_s_maxs_maxa, 23, 9, 6, up(maxa)).
table(up_v_c_minc_maxc_mins_maxs_maxa, 23, 9, 6, up(maxa)).
table(up_v_c_minc_s_mins_maxs_maxa, 23, 10, 6, up(maxa)).
table(up_v_c_maxc_s_mins_maxs_maxa, 23, 7, 6, up(maxa)).
table(up_v_minc_maxc_s_mins_maxs_maxa, 23, 11, 6, up(maxa)).
table(up_c_minc_maxc_s_mins_maxs_maxa, 23, 10, 6, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxs_maxa, 23, 10, 7, up(maxa)).
table(low_c_v, 24, 18, 1, low(v)).
table(low_minc_v, 24, 11, 1, low(v)).
table(low_maxc_v, 24, 11, 1, low(v)).
table(low_s_v, 24, 7, 1, low(v)).
table(low_mins_v, 24, 18, 1, low(v)).
table(low_maxs_v, 24, 18, 1, low(v)).
table(low_c_minc_v, 24, 11, 2, low(v)).
table(low_c_maxc_v, 24, 11, 2, low(v)).
table(low_c_s_v, 24, 8, 2, low(v)).
table(low_c_mins_v, 24, 18, 2, low(v)).
table(low_c_maxs_v, 24, 18, 2, low(v)).
table(low_minc_maxc_v, 24, 11, 2, low(v)).
table(low_minc_s_v, 24, 3, 2, low(v)).
table(low_minc_mins_v, 24, 12, 2, low(v)).
table(low_minc_maxs_v, 24, 12, 2, low(v)).
table(low_maxc_s_v, 24, 3, 2, low(v)).
table(low_maxc_mins_v, 24, 12, 2, low(v)).
table(low_maxc_maxs_v, 24, 13, 2, low(v)).
table(low_s_mins_v, 24, 7, 2, low(v)).
table(low_s_maxs_v, 24, 7, 2, low(v)).
table(low_mins_maxs_v, 24, 12, 2, low(v)).
table(low_c_minc_maxc_v, 24, 11, 3, low(v)).
table(low_c_minc_s_v, 24, 6, 3, low(v)).
table(low_c_minc_mins_v, 24, 12, 3, low(v)).
table(low_c_minc_maxs_v, 24, 12, 3, low(v)).
table(low_c_maxc_s_v, 24, 4, 3, low(v)).
table(low_c_maxc_mins_v, 24, 12, 3, low(v)).
table(low_c_maxc_maxs_v, 24, 13, 3, low(v)).
table(low_c_s_mins_v, 24, 8, 3, low(v)).
table(low_c_s_maxs_v, 24, 8, 3, low(v)).
table(low_c_mins_maxs_v, 24, 18, 3, low(v)).
table(low_minc_maxc_s_v, 24, 5, 3, low(v)).
table(low_minc_maxc_mins_v, 24, 12, 3, low(v)).
table(low_minc_maxc_maxs_v, 24, 12, 3, low(v)).
table(low_minc_s_mins_v, 24, 4, 3, low(v)).
table(low_minc_s_maxs_v, 24, 4, 3, low(v)).
table(low_minc_mins_maxs_v, 24, 13, 3, low(v)).
table(low_maxc_s_mins_v, 24, 4, 3, low(v)).
table(low_maxc_s_maxs_v, 24, 4, 3, low(v)).
table(low_maxc_mins_maxs_v, 24, 14, 3, low(v)).
table(low_s_mins_maxs_v, 24, 7, 3, low(v)).
table(low_c_minc_maxc_s_v, 24, 7, 4, low(v)).
table(low_c_minc_maxc_mins_v, 24, 12, 4, low(v)).
table(low_c_minc_maxc_maxs_v, 24, 12, 4, low(v)).
table(low_c_minc_s_mins_v, 24, 7, 4, low(v)).
table(low_c_minc_s_maxs_v, 24, 7, 4, low(v)).
table(low_c_minc_mins_maxs_v, 24, 13, 4, low(v)).
table(low_c_maxc_s_mins_v, 24, 5, 4, low(v)).
table(low_c_maxc_s_maxs_v, 24, 5, 4, low(v)).
table(low_c_maxc_mins_maxs_v, 24, 14, 4, low(v)).
table(low_c_s_mins_maxs_v, 24, 8, 4, low(v)).
table(low_minc_maxc_s_mins_v, 24, 6, 4, low(v)).
table(low_minc_maxc_s_maxs_v, 24, 6, 4, low(v)).
table(low_minc_maxc_mins_maxs_v, 24, 13, 4, low(v)).
table(low_minc_s_mins_maxs_v, 24, 5, 4, low(v)).
table(low_maxc_s_mins_maxs_v, 24, 5, 4, low(v)).
table(low_c_minc_maxc_s_mins_v, 24, 8, 5, low(v)).
table(low_c_minc_maxc_s_maxs_v, 24, 8, 5, low(v)).
table(low_c_minc_maxc_mins_maxs_v, 24, 13, 5, low(v)).
table(low_c_minc_s_mins_maxs_v, 24, 8, 5, low(v)).
table(low_c_maxc_s_mins_maxs_v, 24, 6, 5, low(v)).
table(low_minc_maxc_s_mins_maxs_v, 24, 7, 5, low(v)).
table(low_c_minc_maxc_s_mins_maxs_v, 24, 9, 6, low(v)).
table(up_c_maxc_v, 24, 3, 2, up(v)).
table(up_maxc_s_v, 24, 3, 2, up(v)).
table(up_s_maxs_v, 24, 3, 2, up(v)).
table(up_c_minc_maxc_v, 24, 6, 3, up(v)).
table(up_c_maxc_s_v, 24, 4, 3, up(v)).
table(up_c_maxc_mins_v, 24, 4, 3, up(v)).
table(up_c_maxc_maxs_v, 24, 4, 3, up(v)).
table(up_c_s_maxs_v, 24, 4, 3, up(v)).
table(up_minc_maxc_s_v, 24, 5, 3, up(v)).
table(up_minc_s_maxs_v, 24, 4, 3, up(v)).
table(up_maxc_s_mins_v, 24, 4, 3, up(v)).
table(up_maxc_s_maxs_v, 24, 4, 3, up(v)).
table(up_s_mins_maxs_v, 24, 4, 3, up(v)).
table(up_c_minc_maxc_s_v, 24, 7, 4, up(v)).
table(up_c_minc_maxc_mins_v, 24, 7, 4, up(v)).
table(up_c_minc_maxc_maxs_v, 24, 7, 4, up(v)).
table(up_c_minc_s_maxs_v, 24, 7, 4, up(v)).
table(up_c_maxc_s_mins_v, 24, 5, 4, up(v)).
table(up_c_maxc_s_maxs_v, 24, 5, 4, up(v)).
table(up_c_maxc_mins_maxs_v, 24, 5, 4, up(v)).
table(up_c_s_mins_maxs_v, 24, 5, 4, up(v)).
table(up_minc_maxc_s_mins_v, 24, 6, 4, up(v)).
table(up_minc_maxc_s_maxs_v, 24, 6, 4, up(v)).
table(up_minc_s_mins_maxs_v, 24, 5, 4, up(v)).
table(up_maxc_s_mins_maxs_v, 24, 5, 4, up(v)).
table(up_c_minc_maxc_s_mins_v, 24, 8, 5, up(v)).
table(up_c_minc_maxc_s_maxs_v, 24, 8, 5, up(v)).
table(up_c_minc_maxc_mins_maxs_v, 24, 8, 5, up(v)).
table(up_c_minc_s_mins_maxs_v, 24, 8, 5, up(v)).
table(up_c_maxc_s_mins_maxs_v, 24, 6, 5, up(v)).
table(up_minc_maxc_s_mins_maxs_v, 24, 7, 5, up(v)).
table(up_c_minc_maxc_s_mins_maxs_v, 24, 9, 6, up(v)).
table(low_v_c, 24, 11, 1, low(c)).
table(low_minc_c, 24, 11, 1, low(c)).
table(low_maxc_c, 24, 11, 1, low(c)).
table(low_s_c, 24, 8, 1, low(c)).
table(low_mins_c, 24, 8, 1, low(c)).
table(low_maxs_c, 24, 8, 1, low(c)).
table(low_v_minc_c, 24, 11, 2, low(c)).
table(low_v_maxc_c, 24, 3, 2, low(c)).
table(low_v_s_c, 24, 13, 2, low(c)).
table(low_v_mins_c, 24, 12, 2, low(c)).
table(low_v_maxs_c, 24, 13, 2, low(c)).
table(low_minc_maxc_c, 24, 11, 2, low(c)).
table(low_minc_s_c, 24, 5, 2, low(c)).
table(low_minc_mins_c, 24, 9, 2, low(c)).
table(low_minc_maxs_c, 24, 9, 2, low(c)).
table(low_maxc_s_c, 24, 3, 2, low(c)).
table(low_maxc_mins_c, 24, 12, 2, low(c)).
table(low_maxc_maxs_c, 24, 13, 2, low(c)).
table(low_s_mins_c, 24, 10, 2, low(c)).
table(low_s_maxs_c, 24, 11, 2, low(c)).
table(low_mins_maxs_c, 24, 9, 2, low(c)).
table(low_v_minc_maxc_c, 24, 6, 3, low(c)).
table(low_v_minc_s_c, 24, 12, 3, low(c)).
table(low_v_minc_mins_c, 24, 12, 3, low(c)).
table(low_v_minc_maxs_c, 24, 12, 3, low(c)).
table(low_v_maxc_s_c, 24, 4, 3, low(c)).
table(low_v_maxc_mins_c, 24, 4, 3, low(c)).
table(low_v_maxc_maxs_c, 24, 4, 3, low(c)).
table(low_v_s_mins_c, 24, 14, 3, low(c)).
table(low_v_s_maxs_c, 24, 14, 3, low(c)).
table(low_v_mins_maxs_c, 24, 14, 3, low(c)).
table(low_minc_maxc_s_c, 24, 5, 3, low(c)).
table(low_minc_maxc_mins_c, 24, 12, 3, low(c)).
table(low_minc_maxc_maxs_c, 24, 12, 3, low(c)).
table(low_minc_s_mins_c, 24, 6, 3, low(c)).
table(low_minc_s_maxs_c, 24, 7, 3, low(c)).
table(low_minc_mins_maxs_c, 24, 7, 3, low(c)).
table(low_maxc_s_mins_c, 24, 4, 3, low(c)).
table(low_maxc_s_maxs_c, 24, 4, 3, low(c)).
table(low_maxc_mins_maxs_c, 24, 6, 3, low(c)).
table(low_s_mins_maxs_c, 24, 12, 3, low(c)).
table(low_v_minc_maxc_s_c, 24, 7, 4, low(c)).
table(low_v_minc_maxc_mins_c, 24, 7, 4, low(c)).
table(low_v_minc_maxc_maxs_c, 24, 7, 4, low(c)).
table(low_v_minc_s_mins_c, 24, 13, 4, low(c)).
table(low_v_minc_s_maxs_c, 24, 13, 4, low(c)).
table(low_v_minc_mins_maxs_c, 24, 13, 4, low(c)).
table(low_v_maxc_s_mins_c, 24, 5, 4, low(c)).
table(low_v_maxc_s_maxs_c, 24, 5, 4, low(c)).
table(low_v_maxc_mins_maxs_c, 24, 5, 4, low(c)).
table(low_v_s_mins_maxs_c, 24, 15, 4, low(c)).
table(low_minc_maxc_s_mins_c, 24, 6, 4, low(c)).
table(low_minc_maxc_s_maxs_c, 24, 6, 4, low(c)).
table(low_minc_maxc_mins_maxs_c, 24, 8, 4, low(c)).
table(low_minc_s_mins_maxs_c, 24, 8, 4, low(c)).
table(low_maxc_s_mins_maxs_c, 24, 5, 4, low(c)).
table(low_v_minc_maxc_s_mins_c, 24, 8, 5, low(c)).
table(low_v_minc_maxc_s_maxs_c, 24, 8, 5, low(c)).
table(low_v_minc_maxc_mins_maxs_c, 24, 8, 5, low(c)).
table(low_v_minc_s_mins_maxs_c, 24, 14, 5, low(c)).
table(low_v_maxc_s_mins_maxs_c, 24, 6, 5, low(c)).
table(low_minc_maxc_s_mins_maxs_c, 24, 7, 5, low(c)).
table(low_v_minc_maxc_s_mins_maxs_c, 24, 9, 6, low(c)).
table(up_v_c, 24, 18, 1, up(c)).
table(up_s_c, 24, 3, 1, up(c)).
table(up_v_minc_c, 24, 6, 2, up(c)).
table(up_v_maxc_c, 24, 11, 2, up(c)).
table(up_v_s_c, 24, 4, 2, up(c)).
table(up_v_mins_c, 24, 10, 2, up(c)).
table(up_v_maxs_c, 24, 18, 2, up(c)).
table(up_minc_s_c, 24, 3, 2, up(c)).
table(up_maxc_s_c, 24, 4, 2, up(c)).
table(up_s_mins_c, 24, 5, 2, up(c)).
table(up_s_maxs_c, 24, 5, 2, up(c)).
table(up_v_minc_maxc_c, 24, 7, 3, up(c)).
table(up_v_minc_s_c, 24, 6, 3, up(c)).
table(up_v_minc_mins_c, 24, 7, 3, up(c)).
table(up_v_minc_maxs_c, 24, 7, 3, up(c)).
table(up_v_maxc_s_c, 24, 5, 3, up(c)).
table(up_v_maxc_mins_c, 24, 5, 3, up(c)).
table(up_v_maxc_maxs_c, 24, 13, 3, up(c)).
table(up_v_s_mins_c, 24, 8, 3, up(c)).
table(up_v_s_maxs_c, 24, 6, 3, up(c)).
table(up_v_mins_maxs_c, 24, 12, 3, up(c)).
table(up_minc_maxc_s_c, 24, 5, 3, up(c)).
table(up_minc_s_mins_c, 24, 4, 3, up(c)).
table(up_minc_s_maxs_c, 24, 4, 3, up(c)).
table(up_maxc_s_mins_c, 24, 5, 3, up(c)).
table(up_maxc_s_maxs_c, 24, 5, 3, up(c)).
table(up_s_mins_maxs_c, 24, 8, 3, up(c)).
table(up_v_minc_maxc_s_c, 24, 7, 4, up(c)).
table(up_v_minc_maxc_mins_c, 24, 8, 4, up(c)).
table(up_v_minc_maxc_maxs_c, 24, 8, 4, up(c)).
table(up_v_minc_s_mins_c, 24, 7, 4, up(c)).
table(up_v_minc_s_maxs_c, 24, 7, 4, up(c)).
table(up_v_minc_mins_maxs_c, 24, 8, 4, up(c)).
table(up_v_maxc_s_mins_c, 24, 6, 4, up(c)).
table(up_v_maxc_s_maxs_c, 24, 6, 4, up(c)).
table(up_v_maxc_mins_maxs_c, 24, 7, 4, up(c)).
table(up_v_s_mins_maxs_c, 24, 10, 4, up(c)).
table(up_minc_maxc_s_mins_c, 24, 6, 4, up(c)).
table(up_minc_maxc_s_maxs_c, 24, 6, 4, up(c)).
table(up_minc_s_mins_maxs_c, 24, 5, 4, up(c)).
table(up_maxc_s_mins_maxs_c, 24, 6, 4, up(c)).
table(up_v_minc_maxc_s_mins_c, 24, 8, 5, up(c)).
table(up_v_minc_maxc_s_maxs_c, 24, 8, 5, up(c)).
table(up_v_minc_maxc_mins_maxs_c, 24, 9, 5, up(c)).
table(up_v_minc_s_mins_maxs_c, 24, 8, 5, up(c)).
table(up_v_maxc_s_mins_maxs_c, 24, 7, 5, up(c)).
table(up_minc_maxc_s_mins_maxs_c, 24, 7, 5, up(c)).
table(up_v_minc_maxc_s_mins_maxs_c, 24, 9, 6, up(c)).
table(low_v_minc, 24, 3, 1, low(minc)).
table(low_c_minc, 24, 3, 1, low(minc)).
table(low_maxc_minc, 24, 3, 1, low(minc)).
table(low_s_minc, 24, 3, 1, low(minc)).
table(low_mins_minc, 24, 2, 1, low(minc)).
table(low_maxs_minc, 24, 3, 1, low(minc)).
table(low_v_c_minc, 24, 5, 2, low(minc)).
table(low_v_maxc_minc, 24, 4, 2, low(minc)).
table(low_v_s_minc, 24, 4, 2, low(minc)).
table(low_v_mins_minc, 24, 3, 2, low(minc)).
table(low_v_maxs_minc, 24, 4, 2, low(minc)).
table(low_c_maxc_minc, 24, 4, 2, low(minc)).
table(low_c_s_minc, 24, 4, 2, low(minc)).
table(low_c_mins_minc, 24, 3, 2, low(minc)).
table(low_c_maxs_minc, 24, 4, 2, low(minc)).
table(low_maxc_s_minc, 24, 4, 2, low(minc)).
table(low_maxc_mins_minc, 24, 4, 2, low(minc)).
table(low_maxc_maxs_minc, 24, 4, 2, low(minc)).
table(low_s_mins_minc, 24, 3, 2, low(minc)).
table(low_s_maxs_minc, 24, 4, 2, low(minc)).
table(low_mins_maxs_minc, 24, 3, 2, low(minc)).
table(low_v_c_maxc_minc, 24, 6, 3, low(minc)).
table(low_v_c_s_minc, 24, 6, 3, low(minc)).
table(low_v_c_mins_minc, 24, 6, 3, low(minc)).
table(low_v_c_maxs_minc, 24, 6, 3, low(minc)).
table(low_v_maxc_s_minc, 24, 5, 3, low(minc)).
table(low_v_maxc_mins_minc, 24, 5, 3, low(minc)).
table(low_v_maxc_maxs_minc, 24, 5, 3, low(minc)).
table(low_v_s_mins_minc, 24, 4, 3, low(minc)).
table(low_v_s_maxs_minc, 24, 5, 3, low(minc)).
table(low_v_mins_maxs_minc, 24, 4, 3, low(minc)).
table(low_c_maxc_s_minc, 24, 5, 3, low(minc)).
table(low_c_maxc_mins_minc, 24, 5, 3, low(minc)).
table(low_c_maxc_maxs_minc, 24, 5, 3, low(minc)).
table(low_c_s_mins_minc, 24, 4, 3, low(minc)).
table(low_c_s_maxs_minc, 24, 6, 3, low(minc)).
table(low_c_mins_maxs_minc, 24, 4, 3, low(minc)).
table(low_maxc_s_mins_minc, 24, 5, 3, low(minc)).
table(low_maxc_s_maxs_minc, 24, 5, 3, low(minc)).
table(low_maxc_mins_maxs_minc, 24, 5, 3, low(minc)).
table(low_s_mins_maxs_minc, 24, 4, 3, low(minc)).
table(low_v_c_maxc_s_minc, 24, 7, 4, low(minc)).
table(low_v_c_maxc_mins_minc, 24, 7, 4, low(minc)).
table(low_v_c_maxc_maxs_minc, 24, 7, 4, low(minc)).
table(low_v_c_s_mins_minc, 24, 7, 4, low(minc)).
table(low_v_c_s_maxs_minc, 24, 7, 4, low(minc)).
table(low_v_c_mins_maxs_minc, 24, 7, 4, low(minc)).
table(low_v_maxc_s_mins_minc, 24, 6, 4, low(minc)).
table(low_v_maxc_s_maxs_minc, 24, 6, 4, low(minc)).
table(low_v_maxc_mins_maxs_minc, 24, 6, 4, low(minc)).
table(low_v_s_mins_maxs_minc, 24, 5, 4, low(minc)).
table(low_c_maxc_s_mins_minc, 24, 6, 4, low(minc)).
table(low_c_maxc_s_maxs_minc, 24, 6, 4, low(minc)).
table(low_c_maxc_mins_maxs_minc, 24, 6, 4, low(minc)).
table(low_c_s_mins_maxs_minc, 24, 6, 4, low(minc)).
table(low_maxc_s_mins_maxs_minc, 24, 6, 4, low(minc)).
table(low_v_c_maxc_s_mins_minc, 24, 8, 5, low(minc)).
table(low_v_c_maxc_s_maxs_minc, 24, 8, 5, low(minc)).
table(low_v_c_maxc_mins_maxs_minc, 24, 8, 5, low(minc)).
table(low_v_c_s_mins_maxs_minc, 24, 8, 5, low(minc)).
table(low_v_maxc_s_mins_maxs_minc, 24, 7, 5, low(minc)).
table(low_c_maxc_s_mins_maxs_minc, 24, 7, 5, low(minc)).
table(low_v_c_maxc_s_mins_maxs_minc, 24, 9, 6, low(minc)).
table(up_v_minc, 24, 11, 1, up(minc)).
table(up_maxc_minc, 24, 6, 1, up(minc)).
table(up_v_c_minc, 24, 5, 2, up(minc)).
table(up_v_maxc_minc, 24, 4, 2, up(minc)).
table(up_v_s_minc, 24, 13, 2, up(minc)).
table(up_v_mins_minc, 24, 12, 2, up(minc)).
table(up_v_maxs_minc, 24, 13, 2, up(minc)).
table(up_c_maxc_minc, 24, 4, 2, up(minc)).
table(up_maxc_s_minc, 24, 4, 2, up(minc)).
table(up_maxc_mins_minc, 24, 4, 2, up(minc)).
table(up_maxc_maxs_minc, 24, 7, 2, up(minc)).
table(up_s_maxs_minc, 24, 14, 2, up(minc)).
table(up_v_c_maxc_minc, 24, 6, 3, up(minc)).
table(up_v_c_s_minc, 24, 6, 3, up(minc)).
table(up_v_c_mins_minc, 24, 6, 3, up(minc)).
table(up_v_c_maxs_minc, 24, 6, 3, up(minc)).
table(up_v_maxc_s_minc, 24, 5, 3, up(minc)).
table(up_v_maxc_mins_minc, 24, 5, 3, up(minc)).
table(up_v_maxc_maxs_minc, 24, 5, 3, up(minc)).
table(up_v_s_mins_minc, 24, 14, 3, up(minc)).
table(up_v_s_maxs_minc, 24, 14, 3, up(minc)).
table(up_v_mins_maxs_minc, 24, 14, 3, up(minc)).
table(up_c_maxc_s_minc, 24, 5, 3, up(minc)).
table(up_c_maxc_mins_minc, 24, 5, 3, up(minc)).
table(up_c_maxc_maxs_minc, 24, 5, 3, up(minc)).
table(up_c_s_maxs_minc, 24, 4, 3, up(minc)).
table(up_maxc_s_mins_minc, 24, 5, 3, up(minc)).
table(up_maxc_s_maxs_minc, 24, 5, 3, up(minc)).
table(up_maxc_mins_maxs_minc, 24, 5, 3, up(minc)).
table(up_s_mins_maxs_minc, 24, 15, 3, up(minc)).
table(up_v_c_maxc_s_minc, 24, 7, 4, up(minc)).
table(up_v_c_maxc_mins_minc, 24, 7, 4, up(minc)).
table(up_v_c_maxc_maxs_minc, 24, 7, 4, up(minc)).
table(up_v_c_s_mins_minc, 24, 7, 4, up(minc)).
table(up_v_c_s_maxs_minc, 24, 7, 4, up(minc)).
table(up_v_c_mins_maxs_minc, 24, 7, 4, up(minc)).
table(up_v_maxc_s_mins_minc, 24, 6, 4, up(minc)).
table(up_v_maxc_s_maxs_minc, 24, 6, 4, up(minc)).
table(up_v_maxc_mins_maxs_minc, 24, 6, 4, up(minc)).
table(up_v_s_mins_maxs_minc, 24, 15, 4, up(minc)).
table(up_c_maxc_s_mins_minc, 24, 6, 4, up(minc)).
table(up_c_maxc_s_maxs_minc, 24, 6, 4, up(minc)).
table(up_c_maxc_mins_maxs_minc, 24, 6, 4, up(minc)).
table(up_c_s_mins_maxs_minc, 24, 5, 4, up(minc)).
table(up_maxc_s_mins_maxs_minc, 24, 6, 4, up(minc)).
table(up_v_c_maxc_s_mins_minc, 24, 8, 5, up(minc)).
table(up_v_c_maxc_s_maxs_minc, 24, 8, 5, up(minc)).
table(up_v_c_maxc_mins_maxs_minc, 24, 8, 5, up(minc)).
table(up_v_c_s_mins_maxs_minc, 24, 8, 5, up(minc)).
table(up_v_maxc_s_mins_maxs_minc, 24, 7, 5, up(minc)).
table(up_c_maxc_s_mins_maxs_minc, 24, 7, 5, up(minc)).
table(up_v_c_maxc_s_mins_maxs_minc, 24, 9, 6, up(minc)).
table(low_v_maxc, 24, 18, 1, low(maxc)).
table(low_c_maxc, 24, 18, 1, low(maxc)).
table(low_minc_maxc, 24, 6, 1, low(maxc)).
table(low_s_maxc, 24, 18, 1, low(maxc)).
table(low_mins_maxc, 24, 9, 1, low(maxc)).
table(low_maxs_maxc, 24, 2, 1, low(maxc)).
table(low_v_c_maxc, 24, 3, 2, low(maxc)).
table(low_v_minc_maxc, 24, 4, 2, low(maxc)).
table(low_v_s_maxc, 24, 4, 2, low(maxc)).
table(low_v_mins_maxc, 24, 7, 2, low(maxc)).
table(low_v_maxs_maxc, 24, 3, 2, low(maxc)).
table(low_c_minc_maxc, 24, 11, 2, low(maxc)).
table(low_c_s_maxc, 24, 4, 2, low(maxc)).
table(low_c_mins_maxc, 24, 18, 2, low(maxc)).
table(low_c_maxs_maxc, 24, 3, 2, low(maxc)).
table(low_minc_s_maxc, 24, 4, 2, low(maxc)).
table(low_minc_mins_maxc, 24, 4, 2, low(maxc)).
table(low_minc_maxs_maxc, 24, 4, 2, low(maxc)).
table(low_s_mins_maxc, 24, 18, 2, low(maxc)).
table(low_s_maxs_maxc, 24, 3, 2, low(maxc)).
table(low_mins_maxs_maxc, 24, 3, 2, low(maxc)).
table(low_v_c_minc_maxc, 24, 6, 3, low(maxc)).
table(low_v_c_s_maxc, 24, 4, 3, low(maxc)).
table(low_v_c_mins_maxc, 24, 4, 3, low(maxc)).
table(low_v_c_maxs_maxc, 24, 4, 3, low(maxc)).
table(low_v_minc_s_maxc, 24, 5, 3, low(maxc)).
table(low_v_minc_mins_maxc, 24, 5, 3, low(maxc)).
table(low_v_minc_maxs_maxc, 24, 5, 3, low(maxc)).
table(low_v_s_mins_maxc, 24, 5, 3, low(maxc)).
table(low_v_s_maxs_maxc, 24, 4, 3, low(maxc)).
table(low_v_mins_maxs_maxc, 24, 4, 3, low(maxc)).
table(low_c_minc_s_maxc, 24, 5, 3, low(maxc)).
table(low_c_minc_mins_maxc, 24, 6, 3, low(maxc)).
table(low_c_minc_maxs_maxc, 24, 5, 3, low(maxc)).
table(low_c_s_mins_maxc, 24, 4, 3, low(maxc)).
table(low_c_s_maxs_maxc, 24, 4, 3, low(maxc)).
table(low_c_mins_maxs_maxc, 24, 4, 3, low(maxc)).
table(low_minc_s_mins_maxc, 24, 5, 3, low(maxc)).
table(low_minc_s_maxs_maxc, 24, 5, 3, low(maxc)).
table(low_minc_mins_maxs_maxc, 24, 5, 3, low(maxc)).
table(low_s_mins_maxs_maxc, 24, 4, 3, low(maxc)).
table(low_v_c_minc_s_maxc, 24, 7, 4, low(maxc)).
table(low_v_c_minc_mins_maxc, 24, 7, 4, low(maxc)).
table(low_v_c_minc_maxs_maxc, 24, 7, 4, low(maxc)).
table(low_v_c_s_mins_maxc, 24, 5, 4, low(maxc)).
table(low_v_c_s_maxs_maxc, 24, 5, 4, low(maxc)).
table(low_v_c_mins_maxs_maxc, 24, 5, 4, low(maxc)).
table(low_v_minc_s_mins_maxc, 24, 6, 4, low(maxc)).
table(low_v_minc_s_maxs_maxc, 24, 6, 4, low(maxc)).
table(low_v_minc_mins_maxs_maxc, 24, 6, 4, low(maxc)).
table(low_v_s_mins_maxs_maxc, 24, 5, 4, low(maxc)).
table(low_c_minc_s_mins_maxc, 24, 6, 4, low(maxc)).
table(low_c_minc_s_maxs_maxc, 24, 6, 4, low(maxc)).
table(low_c_minc_mins_maxs_maxc, 24, 6, 4, low(maxc)).
table(low_c_s_mins_maxs_maxc, 24, 5, 4, low(maxc)).
table(low_minc_s_mins_maxs_maxc, 24, 6, 4, low(maxc)).
table(low_v_c_minc_s_mins_maxc, 24, 8, 5, low(maxc)).
table(low_v_c_minc_s_maxs_maxc, 24, 8, 5, low(maxc)).
table(low_v_c_minc_mins_maxs_maxc, 24, 8, 5, low(maxc)).
table(low_v_c_s_mins_maxs_maxc, 24, 6, 5, low(maxc)).
table(low_v_minc_s_mins_maxs_maxc, 24, 7, 5, low(maxc)).
table(low_c_minc_s_mins_maxs_maxc, 24, 7, 5, low(maxc)).
table(low_v_c_minc_s_mins_maxs_maxc, 24, 9, 6, low(maxc)).
table(up_v_maxc, 24, 11, 1, up(maxc)).
table(up_v_c_maxc, 24, 11, 2, up(maxc)).
table(up_v_minc_maxc, 24, 11, 2, up(maxc)).
table(up_v_s_maxc, 24, 13, 2, up(maxc)).
table(up_v_mins_maxc, 24, 12, 2, up(maxc)).
table(up_v_maxs_maxc, 24, 13, 2, up(maxc)).
table(up_v_c_minc_maxc, 24, 11, 3, up(maxc)).
table(up_v_c_s_maxc, 24, 13, 3, up(maxc)).
table(up_v_c_mins_maxc, 24, 12, 3, up(maxc)).
table(up_v_c_maxs_maxc, 24, 13, 3, up(maxc)).
table(up_v_minc_s_maxc, 24, 12, 3, up(maxc)).
table(up_v_minc_mins_maxc, 24, 12, 3, up(maxc)).
table(up_v_minc_maxs_maxc, 24, 12, 3, up(maxc)).
table(up_v_s_mins_maxc, 24, 14, 3, up(maxc)).
table(up_v_s_maxs_maxc, 24, 14, 3, up(maxc)).
table(up_v_mins_maxs_maxc, 24, 14, 3, up(maxc)).
table(up_v_c_minc_s_maxc, 24, 12, 4, up(maxc)).
table(up_v_c_minc_mins_maxc, 24, 12, 4, up(maxc)).
table(up_v_c_minc_maxs_maxc, 24, 12, 4, up(maxc)).
table(up_v_c_s_mins_maxc, 24, 14, 4, up(maxc)).
table(up_v_c_s_maxs_maxc, 24, 6, 4, up(maxc)).
table(up_v_c_mins_maxs_maxc, 24, 10, 4, up(maxc)).
table(up_v_minc_s_mins_maxc, 24, 13, 4, up(maxc)).
table(up_v_minc_s_maxs_maxc, 24, 13, 4, up(maxc)).
table(up_v_minc_mins_maxs_maxc, 24, 13, 4, up(maxc)).
table(up_v_s_mins_maxs_maxc, 24, 15, 4, up(maxc)).
table(up_v_c_minc_s_mins_maxc, 24, 13, 5, up(maxc)).
table(up_v_c_minc_s_maxs_maxc, 24, 8, 5, up(maxc)).
table(up_v_c_minc_mins_maxs_maxc, 24, 9, 5, up(maxc)).
table(up_v_c_s_mins_maxs_maxc, 24, 10, 5, up(maxc)).
table(up_v_minc_s_mins_maxs_maxc, 24, 14, 5, up(maxc)).
table(up_v_c_minc_s_mins_maxs_maxc, 24, 9, 6, up(maxc)).
table(low_v_s, 24, 18, 1, low(s)).
table(low_c_s, 24, 3, 1, low(s)).
table(low_minc_s, 24, 18, 1, low(s)).
table(low_maxc_s, 24, 18, 1, low(s)).
table(low_mins_s, 24, 18, 1, low(s)).
table(low_maxs_s, 24, 18, 1, low(s)).
table(low_v_c_s, 24, 4, 2, low(s)).
table(low_v_minc_s, 24, 18, 2, low(s)).
table(low_v_maxc_s, 24, 6, 2, low(s)).
table(low_v_mins_s, 24, 9, 2, low(s)).
table(low_v_maxs_s, 24, 3, 2, low(s)).
table(low_c_minc_s, 24, 5, 2, low(s)).
table(low_c_maxc_s, 24, 5, 2, low(s)).
table(low_c_mins_s, 24, 5, 2, low(s)).
table(low_c_maxs_s, 24, 5, 2, low(s)).
table(low_minc_maxc_s, 24, 18, 2, low(s)).
table(low_minc_mins_s, 24, 12, 2, low(s)).
table(low_minc_maxs_s, 24, 14, 2, low(s)).
table(low_maxc_mins_s, 24, 7, 2, low(s)).
table(low_maxc_maxs_s, 24, 14, 2, low(s)).
table(low_mins_maxs_s, 24, 12, 2, low(s)).
table(low_v_c_minc_s, 24, 8, 3, low(s)).
table(low_v_c_maxc_s, 24, 6, 3, low(s)).
table(low_v_c_mins_s, 24, 8, 3, low(s)).
table(low_v_c_maxs_s, 24, 4, 3, low(s)).
table(low_v_minc_maxc_s, 24, 10, 3, low(s)).
table(low_v_minc_mins_s, 24, 13, 3, low(s)).
table(low_v_minc_maxs_s, 24, 4, 3, low(s)).
table(low_v_maxc_mins_s, 24, 4, 3, low(s)).
table(low_v_maxc_maxs_s, 24, 4, 3, low(s)).
table(low_v_mins_maxs_s, 24, 4, 3, low(s)).
table(low_c_minc_maxc_s, 24, 8, 3, low(s)).
table(low_c_minc_mins_s, 24, 4, 3, low(s)).
table(low_c_minc_maxs_s, 24, 5, 3, low(s)).
table(low_c_maxc_mins_s, 24, 8, 3, low(s)).
table(low_c_maxc_maxs_s, 24, 5, 3, low(s)).
table(low_c_mins_maxs_s, 24, 8, 3, low(s)).
table(low_minc_maxc_mins_s, 24, 13, 3, low(s)).
table(low_minc_maxc_maxs_s, 24, 14, 3, low(s)).
table(low_minc_mins_maxs_s, 24, 8, 3, low(s)).
table(low_maxc_mins_maxs_s, 24, 9, 3, low(s)).
table(low_v_c_minc_maxc_s, 24, 10, 4, low(s)).
table(low_v_c_minc_mins_s, 24, 7, 4, low(s)).
table(low_v_c_minc_maxs_s, 24, 7, 4, low(s)).
table(low_v_c_maxc_mins_s, 24, 5, 4, low(s)).
table(low_v_c_maxc_maxs_s, 24, 5, 4, low(s)).
table(low_v_c_mins_maxs_s, 24, 5, 4, low(s)).
table(low_v_minc_maxc_mins_s, 24, 8, 4, low(s)).
table(low_v_minc_maxc_maxs_s, 24, 6, 4, low(s)).
table(low_v_minc_mins_maxs_s, 24, 5, 4, low(s)).
table(low_v_maxc_mins_maxs_s, 24, 5, 4, low(s)).
table(low_c_minc_maxc_mins_s, 24, 6, 4, low(s)).
table(low_c_minc_maxc_maxs_s, 24, 7, 4, low(s)).
table(low_c_minc_mins_maxs_s, 24, 5, 4, low(s)).
table(low_c_maxc_mins_maxs_s, 24, 6, 4, low(s)).
table(low_minc_maxc_mins_maxs_s, 24, 9, 4, low(s)).
table(low_v_c_minc_maxc_mins_s, 24, 8, 5, low(s)).
table(low_v_c_minc_maxc_maxs_s, 24, 8, 5, low(s)).
table(low_v_c_minc_mins_maxs_s, 24, 8, 5, low(s)).
table(low_v_c_maxc_mins_maxs_s, 24, 6, 5, low(s)).
table(low_v_minc_maxc_mins_maxs_s, 24, 7, 5, low(s)).
table(low_c_minc_maxc_mins_maxs_s, 24, 7, 5, low(s)).
table(low_v_c_minc_maxc_mins_maxs_s, 24, 9, 6, low(s)).
table(up_v_s, 24, 7, 1, up(s)).
table(up_v_c_s, 24, 8, 2, up(s)).
table(up_v_minc_s, 24, 8, 2, up(s)).
table(up_v_maxc_s, 24, 8, 2, up(s)).
table(up_v_mins_s, 24, 4, 2, up(s)).
table(up_v_maxs_s, 24, 7, 2, up(s)).
table(up_c_maxc_s, 24, 9, 2, up(s)).
table(up_v_c_minc_s, 24, 11, 3, up(s)).
table(up_v_c_maxc_s, 24, 9, 3, up(s)).
table(up_v_c_mins_s, 24, 5, 3, up(s)).
table(up_v_c_maxs_s, 24, 8, 3, up(s)).
table(up_v_minc_maxc_s, 24, 10, 3, up(s)).
table(up_v_minc_mins_s, 24, 5, 3, up(s)).
table(up_v_minc_maxs_s, 24, 8, 3, up(s)).
table(up_v_maxc_mins_s, 24, 5, 3, up(s)).
table(up_v_maxc_maxs_s, 24, 8, 3, up(s)).
table(up_v_mins_maxs_s, 24, 5, 3, up(s)).
table(up_c_minc_maxc_s, 24, 12, 3, up(s)).
table(up_c_maxc_mins_s, 24, 5, 3, up(s)).
table(up_c_maxc_maxs_s, 24, 9, 3, up(s)).
table(up_v_c_minc_maxc_s, 24, 12, 4, up(s)).
table(up_v_c_minc_mins_s, 24, 8, 4, up(s)).
table(up_v_c_minc_maxs_s, 24, 11, 4, up(s)).
table(up_v_c_maxc_mins_s, 24, 6, 4, up(s)).
table(up_v_c_maxc_maxs_s, 24, 9, 4, up(s)).
table(up_v_c_mins_maxs_s, 24, 6, 4, up(s)).
table(up_v_minc_maxc_mins_s, 24, 7, 4, up(s)).
table(up_v_minc_maxc_maxs_s, 24, 10, 4, up(s)).
table(up_v_minc_mins_maxs_s, 24, 6, 4, up(s)).
table(up_v_maxc_mins_maxs_s, 24, 6, 4, up(s)).
table(up_c_minc_maxc_mins_s, 24, 7, 4, up(s)).
table(up_c_minc_maxc_maxs_s, 24, 12, 4, up(s)).
table(up_c_maxc_mins_maxs_s, 24, 6, 4, up(s)).
table(up_v_c_minc_maxc_mins_s, 24, 9, 5, up(s)).
table(up_v_c_minc_maxc_maxs_s, 24, 12, 5, up(s)).
table(up_v_c_minc_mins_maxs_s, 24, 9, 5, up(s)).
table(up_v_c_maxc_mins_maxs_s, 24, 7, 5, up(s)).
table(up_v_minc_maxc_mins_maxs_s, 24, 8, 5, up(s)).
table(up_c_minc_maxc_mins_maxs_s, 24, 8, 5, up(s)).
table(up_v_c_minc_maxc_mins_maxs_s, 24, 10, 6, up(s)).
table(low_v_mins, 24, 2, 1, low(mins)).
table(low_c_mins, 24, 2, 1, low(mins)).
table(low_minc_mins, 24, 2, 1, low(mins)).
table(low_maxc_mins, 24, 2, 1, low(mins)).
table(low_s_mins, 24, 2, 1, low(mins)).
table(low_maxs_mins, 24, 2, 1, low(mins)).
table(low_v_c_mins, 24, 3, 2, low(mins)).
table(low_v_minc_mins, 24, 3, 2, low(mins)).
table(low_v_maxc_mins, 24, 3, 2, low(mins)).
table(low_v_s_mins, 24, 3, 2, low(mins)).
table(low_v_maxs_mins, 24, 3, 2, low(mins)).
table(low_c_minc_mins, 24, 3, 2, low(mins)).
table(low_c_maxc_mins, 24, 3, 2, low(mins)).
table(low_c_s_mins, 24, 3, 2, low(mins)).
table(low_c_maxs_mins, 24, 3, 2, low(mins)).
table(low_minc_maxc_mins, 24, 4, 2, low(mins)).
table(low_minc_s_mins, 24, 3, 2, low(mins)).
table(low_minc_maxs_mins, 24, 3, 2, low(mins)).
table(low_maxc_s_mins, 24, 3, 2, low(mins)).
table(low_maxc_maxs_mins, 24, 3, 2, low(mins)).
table(low_s_maxs_mins, 24, 3, 2, low(mins)).
table(low_v_c_minc_mins, 24, 6, 3, low(mins)).
table(low_v_c_maxc_mins, 24, 4, 3, low(mins)).
table(low_v_c_s_mins, 24, 4, 3, low(mins)).
table(low_v_c_maxs_mins, 24, 4, 3, low(mins)).
table(low_v_minc_maxc_mins, 24, 5, 3, low(mins)).
table(low_v_minc_s_mins, 24, 4, 3, low(mins)).
table(low_v_minc_maxs_mins, 24, 4, 3, low(mins)).
table(low_v_maxc_s_mins, 24, 4, 3, low(mins)).
table(low_v_maxc_maxs_mins, 24, 4, 3, low(mins)).
table(low_v_s_maxs_mins, 24, 4, 3, low(mins)).
table(low_c_minc_maxc_mins, 24, 5, 3, low(mins)).
table(low_c_minc_s_mins, 24, 4, 3, low(mins)).
table(low_c_minc_maxs_mins, 24, 4, 3, low(mins)).
table(low_c_maxc_s_mins, 24, 4, 3, low(mins)).
table(low_c_maxc_maxs_mins, 24, 4, 3, low(mins)).
table(low_c_s_maxs_mins, 24, 4, 3, low(mins)).
table(low_minc_maxc_s_mins, 24, 5, 3, low(mins)).
table(low_minc_maxc_maxs_mins, 24, 5, 3, low(mins)).
table(low_minc_s_maxs_mins, 24, 4, 3, low(mins)).
table(low_maxc_s_maxs_mins, 24, 4, 3, low(mins)).
table(low_v_c_minc_maxc_mins, 24, 7, 4, low(mins)).
table(low_v_c_minc_s_mins, 24, 7, 4, low(mins)).
table(low_v_c_minc_maxs_mins, 24, 7, 4, low(mins)).
table(low_v_c_maxc_s_mins, 24, 5, 4, low(mins)).
table(low_v_c_maxc_maxs_mins, 24, 5, 4, low(mins)).
table(low_v_c_s_maxs_mins, 24, 5, 4, low(mins)).
table(low_v_minc_maxc_s_mins, 24, 6, 4, low(mins)).
table(low_v_minc_maxc_maxs_mins, 24, 6, 4, low(mins)).
table(low_v_minc_s_maxs_mins, 24, 5, 4, low(mins)).
table(low_v_maxc_s_maxs_mins, 24, 5, 4, low(mins)).
table(low_c_minc_maxc_s_mins, 24, 6, 4, low(mins)).
table(low_c_minc_maxc_maxs_mins, 24, 6, 4, low(mins)).
table(low_c_minc_s_maxs_mins, 24, 5, 4, low(mins)).
table(low_c_maxc_s_maxs_mins, 24, 5, 4, low(mins)).
table(low_minc_maxc_s_maxs_mins, 24, 6, 4, low(mins)).
table(low_v_c_minc_maxc_s_mins, 24, 8, 5, low(mins)).
table(low_v_c_minc_maxc_maxs_mins, 24, 8, 5, low(mins)).
table(low_v_c_minc_s_maxs_mins, 24, 8, 5, low(mins)).
table(low_v_c_maxc_s_maxs_mins, 24, 6, 5, low(mins)).
table(low_v_minc_maxc_s_maxs_mins, 24, 7, 5, low(mins)).
table(low_c_minc_maxc_s_maxs_mins, 24, 7, 5, low(mins)).
table(low_v_c_minc_maxc_s_maxs_mins, 24, 9, 6, low(mins)).
table(up_v_mins, 24, 18, 1, up(mins)).
table(up_maxc_mins, 24, 9, 1, up(mins)).
table(up_maxs_mins, 24, 2, 1, up(mins)).
table(up_v_c_mins, 24, 6, 2, up(mins)).
table(up_v_minc_mins, 24, 3, 2, up(mins)).
table(up_v_maxc_mins, 24, 5, 2, up(mins)).
table(up_v_s_mins, 24, 3, 2, up(mins)).
table(up_v_maxs_mins, 24, 3, 2, up(mins)).
table(up_c_maxc_mins, 24, 5, 2, up(mins)).
table(up_c_maxs_mins, 24, 3, 2, up(mins)).
table(up_minc_maxc_mins, 24, 4, 2, up(mins)).
table(up_minc_maxs_mins, 24, 3, 2, up(mins)).
table(up_maxc_s_mins, 24, 3, 2, up(mins)).
table(up_maxc_maxs_mins, 24, 3, 2, up(mins)).
table(up_s_maxs_mins, 24, 3, 2, up(mins)).
table(up_v_c_minc_mins, 24, 6, 3, up(mins)).
table(up_v_c_maxc_mins, 24, 7, 3, up(mins)).
table(up_v_c_s_mins, 24, 4, 3, up(mins)).
table(up_v_c_maxs_mins, 24, 4, 3, up(mins)).
table(up_v_minc_maxc_mins, 24, 5, 3, up(mins)).
table(up_v_minc_s_mins, 24, 4, 3, up(mins)).
table(up_v_minc_maxs_mins, 24, 4, 3, up(mins)).
table(up_v_maxc_s_mins, 24, 4, 3, up(mins)).
table(up_v_maxc_maxs_mins, 24, 4, 3, up(mins)).
table(up_v_s_maxs_mins, 24, 4, 3, up(mins)).
table(up_c_minc_maxc_mins, 24, 5, 3, up(mins)).
table(up_c_minc_maxs_mins, 24, 4, 3, up(mins)).
table(up_c_maxc_s_mins, 24, 4, 3, up(mins)).
table(up_c_maxc_maxs_mins, 24, 4, 3, up(mins)).
table(up_c_s_maxs_mins, 24, 4, 3, up(mins)).
table(up_minc_maxc_s_mins, 24, 5, 3, up(mins)).
table(up_minc_maxc_maxs_mins, 24, 5, 3, up(mins)).
table(up_minc_s_maxs_mins, 24, 4, 3, up(mins)).
table(up_maxc_s_maxs_mins, 24, 4, 3, up(mins)).
table(up_v_c_minc_maxc_mins, 24, 7, 4, up(mins)).
table(up_v_c_minc_s_mins, 24, 7, 4, up(mins)).
table(up_v_c_minc_maxs_mins, 24, 7, 4, up(mins)).
table(up_v_c_maxc_s_mins, 24, 5, 4, up(mins)).
table(up_v_c_maxc_maxs_mins, 24, 5, 4, up(mins)).
table(up_v_c_s_maxs_mins, 24, 5, 4, up(mins)).
table(up_v_minc_maxc_s_mins, 24, 6, 4, up(mins)).
table(up_v_minc_maxc_maxs_mins, 24, 6, 4, up(mins)).
table(up_v_minc_s_maxs_mins, 24, 5, 4, up(mins)).
table(up_v_maxc_s_maxs_mins, 24, 5, 4, up(mins)).
table(up_c_minc_maxc_s_mins, 24, 6, 4, up(mins)).
table(up_c_minc_maxc_maxs_mins, 24, 6, 4, up(mins)).
table(up_c_minc_s_maxs_mins, 24, 5, 4, up(mins)).
table(up_c_maxc_s_maxs_mins, 24, 5, 4, up(mins)).
table(up_minc_maxc_s_maxs_mins, 24, 6, 4, up(mins)).
table(up_v_c_minc_maxc_s_mins, 24, 8, 5, up(mins)).
table(up_v_c_minc_maxc_maxs_mins, 24, 8, 5, up(mins)).
table(up_v_c_minc_s_maxs_mins, 24, 8, 5, up(mins)).
table(up_v_c_maxc_s_maxs_mins, 24, 6, 5, up(mins)).
table(up_v_minc_maxc_s_maxs_mins, 24, 7, 5, up(mins)).
table(up_c_minc_maxc_s_maxs_mins, 24, 7, 5, up(mins)).
table(up_v_c_minc_maxc_s_maxs_mins, 24, 9, 6, up(mins)).
table(low_v_maxs, 24, 7, 1, low(maxs)).
table(low_c_maxs, 24, 4, 1, low(maxs)).
table(low_minc_maxs, 24, 4, 1, low(maxs)).
table(low_maxc_maxs, 24, 4, 1, low(maxs)).
table(low_s_maxs, 24, 7, 1, low(maxs)).
table(low_mins_maxs, 24, 2, 1, low(maxs)).
table(low_v_c_maxs, 24, 8, 2, low(maxs)).
table(low_v_minc_maxs, 24, 8, 2, low(maxs)).
table(low_v_maxc_maxs, 24, 8, 2, low(maxs)).
table(low_v_s_maxs, 24, 3, 2, low(maxs)).
table(low_v_mins_maxs, 24, 3, 2, low(maxs)).
table(low_c_minc_maxs, 24, 5, 2, low(maxs)).
table(low_c_maxc_maxs, 24, 5, 2, low(maxs)).
table(low_c_s_maxs, 24, 8, 2, low(maxs)).
table(low_c_mins_maxs, 24, 3, 2, low(maxs)).
table(low_minc_maxc_maxs, 24, 6, 2, low(maxs)).
table(low_minc_s_maxs, 24, 3, 2, low(maxs)).
table(low_minc_mins_maxs, 24, 3, 2, low(maxs)).
table(low_maxc_s_maxs, 24, 3, 2, low(maxs)).
table(low_maxc_mins_maxs, 24, 3, 2, low(maxs)).
table(low_s_mins_maxs, 24, 7, 2, low(maxs)).
table(low_v_c_minc_maxs, 24, 11, 3, low(maxs)).
table(low_v_c_maxc_maxs, 24, 9, 3, low(maxs)).
table(low_v_c_s_maxs, 24, 4, 3, low(maxs)).
table(low_v_c_mins_maxs, 24, 4, 3, low(maxs)).
table(low_v_minc_maxc_maxs, 24, 10, 3, low(maxs)).
table(low_v_minc_s_maxs, 24, 4, 3, low(maxs)).
table(low_v_minc_mins_maxs, 24, 4, 3, low(maxs)).
table(low_v_maxc_s_maxs, 24, 4, 3, low(maxs)).
table(low_v_maxc_mins_maxs, 24, 4, 3, low(maxs)).
table(low_v_s_mins_maxs, 24, 4, 3, low(maxs)).
table(low_c_minc_maxc_maxs, 24, 7, 3, low(maxs)).
table(low_c_minc_s_maxs, 24, 4, 3, low(maxs)).
table(low_c_minc_mins_maxs, 24, 4, 3, low(maxs)).
table(low_c_maxc_s_maxs, 24, 4, 3, low(maxs)).
table(low_c_maxc_mins_maxs, 24, 4, 3, low(maxs)).
table(low_c_s_mins_maxs, 24, 8, 3, low(maxs)).
table(low_minc_maxc_s_maxs, 24, 5, 3, low(maxs)).
table(low_minc_maxc_mins_maxs, 24, 5, 3, low(maxs)).
table(low_minc_s_mins_maxs, 24, 4, 3, low(maxs)).
table(low_maxc_s_mins_maxs, 24, 4, 3, low(maxs)).
table(low_v_c_minc_maxc_maxs, 24, 12, 4, low(maxs)).
table(low_v_c_minc_s_maxs, 24, 7, 4, low(maxs)).
table(low_v_c_minc_mins_maxs, 24, 7, 4, low(maxs)).
table(low_v_c_maxc_s_maxs, 24, 5, 4, low(maxs)).
table(low_v_c_maxc_mins_maxs, 24, 5, 4, low(maxs)).
table(low_v_c_s_mins_maxs, 24, 5, 4, low(maxs)).
table(low_v_minc_maxc_s_maxs, 24, 6, 4, low(maxs)).
table(low_v_minc_maxc_mins_maxs, 24, 6, 4, low(maxs)).
table(low_v_minc_s_mins_maxs, 24, 5, 4, low(maxs)).
table(low_v_maxc_s_mins_maxs, 24, 5, 4, low(maxs)).
table(low_c_minc_maxc_s_maxs, 24, 6, 4, low(maxs)).
table(low_c_minc_maxc_mins_maxs, 24, 6, 4, low(maxs)).
table(low_c_minc_s_mins_maxs, 24, 5, 4, low(maxs)).
table(low_c_maxc_s_mins_maxs, 24, 5, 4, low(maxs)).
table(low_minc_maxc_s_mins_maxs, 24, 6, 4, low(maxs)).
table(low_v_c_minc_maxc_s_maxs, 24, 8, 5, low(maxs)).
table(low_v_c_minc_maxc_mins_maxs, 24, 8, 5, low(maxs)).
table(low_v_c_minc_s_mins_maxs, 24, 8, 5, low(maxs)).
table(low_v_c_maxc_s_mins_maxs, 24, 6, 5, low(maxs)).
table(low_v_minc_maxc_s_mins_maxs, 24, 7, 5, low(maxs)).
table(low_c_minc_maxc_s_mins_maxs, 24, 7, 5, low(maxs)).
table(low_v_c_minc_maxc_s_mins_maxs, 24, 9, 6, low(maxs)).
table(up_v_maxs, 24, 18, 1, up(maxs)).
table(up_maxc_maxs, 24, 2, 1, up(maxs)).
table(up_v_c_maxs, 24, 18, 2, up(maxs)).
table(up_v_minc_maxs, 24, 12, 2, up(maxs)).
table(up_v_maxc_maxs, 24, 3, 2, up(maxs)).
table(up_v_s_maxs, 24, 7, 2, up(maxs)).
table(up_v_mins_maxs, 24, 9, 2, up(maxs)).
table(up_c_maxc_maxs, 24, 3, 2, up(maxs)).
table(up_minc_maxc_maxs, 24, 4, 2, up(maxs)).
table(up_maxc_s_maxs, 24, 3, 2, up(maxs)).
table(up_maxc_mins_maxs, 24, 3, 2, up(maxs)).
table(up_v_c_minc_maxs, 24, 12, 3, up(maxs)).
table(up_v_c_maxc_maxs, 24, 4, 3, up(maxs)).
table(up_v_c_s_maxs, 24, 8, 3, up(maxs)).
table(up_v_c_mins_maxs, 24, 18, 3, up(maxs)).
table(up_v_minc_maxc_maxs, 24, 5, 3, up(maxs)).
table(up_v_minc_s_maxs, 24, 5, 3, up(maxs)).
table(up_v_minc_mins_maxs, 24, 14, 3, up(maxs)).
table(up_v_maxc_s_maxs, 24, 4, 3, up(maxs)).
table(up_v_maxc_mins_maxs, 24, 4, 3, up(maxs)).
table(up_v_s_mins_maxs, 24, 7, 3, up(maxs)).
table(up_c_minc_maxc_maxs, 24, 5, 3, up(maxs)).
table(up_c_maxc_s_maxs, 24, 4, 3, up(maxs)).
table(up_c_maxc_mins_maxs, 24, 4, 3, up(maxs)).
table(up_minc_maxc_s_maxs, 24, 5, 3, up(maxs)).
table(up_minc_maxc_mins_maxs, 24, 5, 3, up(maxs)).
table(up_maxc_s_mins_maxs, 24, 4, 3, up(maxs)).
table(up_v_c_minc_maxc_maxs, 24, 7, 4, up(maxs)).
table(up_v_c_minc_s_maxs, 24, 7, 4, up(maxs)).
table(up_v_c_minc_mins_maxs, 24, 13, 4, up(maxs)).
table(up_v_c_maxc_s_maxs, 24, 5, 4, up(maxs)).
table(up_v_c_maxc_mins_maxs, 24, 5, 4, up(maxs)).
table(up_v_c_s_mins_maxs, 24, 8, 4, up(maxs)).
table(up_v_minc_maxc_s_maxs, 24, 6, 4, up(maxs)).
table(up_v_minc_maxc_mins_maxs, 24, 6, 4, up(maxs)).
table(up_v_minc_s_mins_maxs, 24, 5, 4, up(maxs)).
table(up_v_maxc_s_mins_maxs, 24, 5, 4, up(maxs)).
table(up_c_minc_maxc_s_maxs, 24, 6, 4, up(maxs)).
table(up_c_minc_maxc_mins_maxs, 24, 6, 4, up(maxs)).
table(up_c_maxc_s_mins_maxs, 24, 5, 4, up(maxs)).
table(up_minc_maxc_s_mins_maxs, 24, 6, 4, up(maxs)).
table(up_v_c_minc_maxc_s_maxs, 24, 8, 5, up(maxs)).
table(up_v_c_minc_maxc_mins_maxs, 24, 8, 5, up(maxs)).
table(up_v_c_minc_s_mins_maxs, 24, 8, 5, up(maxs)).
table(up_v_c_maxc_s_mins_maxs, 24, 6, 5, up(maxs)).
table(up_v_minc_maxc_s_mins_maxs, 24, 7, 5, up(maxs)).
table(up_c_minc_maxc_s_mins_maxs, 24, 7, 5, up(maxs)).
table(up_v_c_minc_maxc_s_mins_maxs, 24, 9, 6, up(maxs)).
table(low_v_mina, 24, 12, 1, low(mina)).
table(low_c_mina, 24, 7, 1, low(mina)).
table(low_minc_mina, 24, 19, 1, low(mina)).
table(low_maxc_mina, 24, 19, 1, low(mina)).
table(low_s_mina, 24, 12, 1, low(mina)).
table(low_mins_mina, 24, 11, 1, low(mina)).
table(low_maxs_mina, 24, 11, 1, low(mina)).
table(low_v_c_mina, 24, 10, 2, low(mina)).
table(low_v_minc_mina, 24, 11, 2, low(mina)).
table(low_v_maxc_mina, 24, 11, 2, low(mina)).
table(low_v_s_mina, 24, 4, 2, low(mina)).
table(low_v_mins_mina, 24, 4, 2, low(mina)).
table(low_v_maxs_mina, 24, 10, 2, low(mina)).
table(low_c_minc_mina, 24, 8, 2, low(mina)).
table(low_c_maxc_mina, 24, 6, 2, low(mina)).
table(low_c_s_mina, 24, 10, 2, low(mina)).
table(low_c_mins_mina, 24, 7, 2, low(mina)).
table(low_c_maxs_mina, 24, 5, 2, low(mina)).
table(low_minc_maxc_mina, 24, 19, 2, low(mina)).
table(low_minc_s_mina, 24, 11, 2, low(mina)).
table(low_minc_mins_mina, 24, 19, 2, low(mina)).
table(low_minc_maxs_mina, 24, 19, 2, low(mina)).
table(low_maxc_s_mina, 24, 11, 2, low(mina)).
table(low_maxc_mins_mina, 24, 19, 2, low(mina)).
table(low_maxc_maxs_mina, 24, 19, 2, low(mina)).
table(low_s_mins_mina, 24, 12, 2, low(mina)).
table(low_s_maxs_mina, 24, 10, 2, low(mina)).
table(low_mins_maxs_mina, 24, 6, 2, low(mina)).
table(low_v_c_minc_mina, 24, 13, 3, low(mina)).
table(low_v_c_maxc_mina, 24, 11, 3, low(mina)).
table(low_v_c_s_mina, 24, 4, 3, low(mina)).
table(low_v_c_mins_mina, 24, 7, 3, low(mina)).
table(low_v_c_maxs_mina, 24, 10, 3, low(mina)).
table(low_v_minc_maxc_mina, 24, 12, 3, low(mina)).
table(low_v_minc_s_mina, 24, 4, 3, low(mina)).
table(low_v_minc_mins_mina, 24, 4, 3, low(mina)).
table(low_v_minc_maxs_mina, 24, 11, 3, low(mina)).
table(low_v_maxc_s_mina, 24, 4, 3, low(mina)).
table(low_v_maxc_mins_mina, 24, 6, 3, low(mina)).
table(low_v_maxc_maxs_mina, 24, 10, 3, low(mina)).
table(low_v_s_mins_mina, 24, 5, 3, low(mina)).
table(low_v_s_maxs_mina, 24, 5, 3, low(mina)).
table(low_v_mins_maxs_mina, 24, 5, 3, low(mina)).
table(low_c_minc_maxc_mina, 24, 8, 3, low(mina)).
table(low_c_minc_s_mina, 24, 11, 3, low(mina)).
table(low_c_minc_mins_mina, 24, 8, 3, low(mina)).
table(low_c_minc_maxs_mina, 24, 6, 3, low(mina)).
table(low_c_maxc_s_mina, 24, 8, 3, low(mina)).
table(low_c_maxc_mins_mina, 24, 6, 3, low(mina)).
table(low_c_maxc_maxs_mina, 24, 6, 3, low(mina)).
table(low_c_s_mins_mina, 24, 10, 3, low(mina)).
table(low_c_s_maxs_mina, 24, 10, 3, low(mina)).
table(low_c_mins_maxs_mina, 24, 5, 3, low(mina)).
table(low_minc_maxc_s_mina, 24, 10, 3, low(mina)).
table(low_minc_maxc_mins_mina, 24, 16, 3, low(mina)).
table(low_minc_maxc_maxs_mina, 24, 18, 3, low(mina)).
table(low_minc_s_mins_mina, 24, 6, 3, low(mina)).
table(low_minc_s_maxs_mina, 24, 8, 3, low(mina)).
table(low_minc_mins_maxs_mina, 24, 17, 3, low(mina)).
table(low_maxc_s_mins_mina, 24, 8, 3, low(mina)).
table(low_maxc_s_maxs_mina, 24, 7, 3, low(mina)).
table(low_maxc_mins_maxs_mina, 24, 8, 3, low(mina)).
table(low_s_mins_maxs_mina, 24, 10, 3, low(mina)).
table(low_v_c_minc_maxc_mina, 24, 14, 4, low(mina)).
table(low_v_c_minc_s_mina, 24, 8, 4, low(mina)).
table(low_v_c_minc_mins_mina, 24, 10, 4, low(mina)).
table(low_v_c_minc_maxs_mina, 24, 13, 4, low(mina)).
table(low_v_c_maxc_s_mina, 24, 5, 4, low(mina)).
table(low_v_c_maxc_mins_mina, 24, 8, 4, low(mina)).
table(low_v_c_maxc_maxs_mina, 24, 11, 4, low(mina)).
table(low_v_c_s_mins_mina, 24, 5, 4, low(mina)).
table(low_v_c_s_maxs_mina, 24, 5, 4, low(mina)).
table(low_v_c_mins_maxs_mina, 24, 8, 4, low(mina)).
table(low_v_minc_maxc_s_mina, 24, 6, 4, low(mina)).
table(low_v_minc_maxc_mins_mina, 24, 6, 4, low(mina)).
table(low_v_minc_maxc_maxs_mina, 24, 12, 4, low(mina)).
table(low_v_minc_s_mins_mina, 24, 5, 4, low(mina)).
table(low_v_minc_s_maxs_mina, 24, 5, 4, low(mina)).
table(low_v_minc_mins_maxs_mina, 24, 5, 4, low(mina)).
table(low_v_maxc_s_mins_mina, 24, 5, 4, low(mina)).
table(low_v_maxc_s_maxs_mina, 24, 5, 4, low(mina)).
table(low_v_maxc_mins_maxs_mina, 24, 5, 4, low(mina)).
table(low_v_s_mins_maxs_mina, 24, 6, 4, low(mina)).
table(low_c_minc_maxc_s_mina, 24, 9, 4, low(mina)).
table(low_c_minc_maxc_mins_mina, 24, 7, 4, low(mina)).
table(low_c_minc_maxc_maxs_mina, 24, 8, 4, low(mina)).
table(low_c_minc_s_mins_mina, 24, 10, 4, low(mina)).
table(low_c_minc_s_maxs_mina, 24, 10, 4, low(mina)).
table(low_c_minc_mins_maxs_mina, 24, 6, 4, low(mina)).
table(low_c_maxc_s_mins_mina, 24, 6, 4, low(mina)).
table(low_c_maxc_s_maxs_mina, 24, 6, 4, low(mina)).
table(low_c_maxc_mins_maxs_mina, 24, 6, 4, low(mina)).
table(low_c_s_mins_maxs_mina, 24, 10, 4, low(mina)).
table(low_minc_maxc_s_mins_mina, 24, 7, 4, low(mina)).
table(low_minc_maxc_s_maxs_mina, 24, 9, 4, low(mina)).
table(low_minc_maxc_mins_maxs_mina, 24, 16, 4, low(mina)).
table(low_minc_s_mins_maxs_mina, 24, 6, 4, low(mina)).
table(low_maxc_s_mins_maxs_mina, 24, 6, 4, low(mina)).
table(low_v_c_minc_maxc_s_mina, 24, 9, 5, low(mina)).
table(low_v_c_minc_maxc_mins_mina, 24, 11, 5, low(mina)).
table(low_v_c_minc_maxc_maxs_mina, 24, 14, 5, low(mina)).
table(low_v_c_minc_s_mins_mina, 24, 8, 5, low(mina)).
table(low_v_c_minc_s_maxs_mina, 24, 8, 5, low(mina)).
table(low_v_c_minc_mins_maxs_mina, 24, 11, 5, low(mina)).
table(low_v_c_maxc_s_mins_mina, 24, 6, 5, low(mina)).
table(low_v_c_maxc_s_maxs_mina, 24, 6, 5, low(mina)).
table(low_v_c_maxc_mins_maxs_mina, 24, 9, 5, low(mina)).
table(low_v_c_s_mins_maxs_mina, 24, 6, 5, low(mina)).
table(low_v_minc_maxc_s_mins_mina, 24, 7, 5, low(mina)).
table(low_v_minc_maxc_s_maxs_mina, 24, 7, 5, low(mina)).
table(low_v_minc_maxc_mins_maxs_mina, 24, 7, 5, low(mina)).
table(low_v_minc_s_mins_maxs_mina, 24, 6, 5, low(mina)).
table(low_v_maxc_s_mins_maxs_mina, 24, 6, 5, low(mina)).
table(low_c_minc_maxc_s_mins_mina, 24, 8, 5, low(mina)).
table(low_c_minc_maxc_s_maxs_mina, 24, 8, 5, low(mina)).
table(low_c_minc_maxc_mins_maxs_mina, 24, 8, 5, low(mina)).
table(low_c_minc_s_mins_maxs_mina, 24, 11, 5, low(mina)).
table(low_c_maxc_s_mins_maxs_mina, 24, 7, 5, low(mina)).
table(low_minc_maxc_s_mins_maxs_mina, 24, 8, 5, low(mina)).
table(low_v_c_minc_maxc_s_mins_mina, 24, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_maxs_mina, 24, 9, 6, low(mina)).
table(low_v_c_minc_maxc_mins_maxs_mina, 24, 12, 6, low(mina)).
table(low_v_c_minc_s_mins_maxs_mina, 24, 9, 6, low(mina)).
table(low_v_c_maxc_s_mins_maxs_mina, 24, 7, 6, low(mina)).
table(low_v_minc_maxc_s_mins_maxs_mina, 24, 8, 6, low(mina)).
table(low_c_minc_maxc_s_mins_maxs_mina, 24, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_mins_maxs_mina, 24, 10, 7, low(mina)).
table(up_v_maxa, 24, 19, 1, up(maxa)).
table(up_v_c_maxa, 24, 19, 2, up(maxa)).
table(up_v_minc_maxa, 24, 19, 2, up(maxa)).
table(up_v_maxc_maxa, 24, 19, 2, up(maxa)).
table(up_v_s_maxa, 24, 19, 2, up(maxa)).
table(up_v_mins_maxa, 24, 19, 2, up(maxa)).
table(up_v_maxs_maxa, 24, 19, 2, up(maxa)).
table(up_c_maxc_maxa, 24, 19, 2, up(maxa)).
table(up_maxc_s_maxa, 24, 18, 2, up(maxa)).
table(up_s_maxs_maxa, 24, 19, 2, up(maxa)).
table(up_v_c_minc_maxa, 24, 19, 3, up(maxa)).
table(up_v_c_maxc_maxa, 24, 19, 3, up(maxa)).
table(up_v_c_s_maxa, 24, 19, 3, up(maxa)).
table(up_v_c_mins_maxa, 24, 19, 3, up(maxa)).
table(up_v_c_maxs_maxa, 24, 7, 3, up(maxa)).
table(up_v_minc_maxc_maxa, 24, 19, 3, up(maxa)).
table(up_v_minc_s_maxa, 24, 18, 3, up(maxa)).
table(up_v_minc_mins_maxa, 24, 19, 3, up(maxa)).
table(up_v_minc_maxs_maxa, 24, 19, 3, up(maxa)).
table(up_v_maxc_s_maxa, 24, 18, 3, up(maxa)).
table(up_v_maxc_mins_maxa, 24, 11, 3, up(maxa)).
table(up_v_maxc_maxs_maxa, 24, 6, 3, up(maxa)).
table(up_v_s_mins_maxa, 24, 19, 3, up(maxa)).
table(up_v_s_maxs_maxa, 24, 19, 3, up(maxa)).
table(up_v_mins_maxs_maxa, 24, 19, 3, up(maxa)).
table(up_c_minc_maxc_maxa, 24, 19, 3, up(maxa)).
table(up_c_maxc_s_maxa, 24, 18, 3, up(maxa)).
table(up_c_maxc_mins_maxa, 24, 8, 3, up(maxa)).
table(up_c_maxc_maxs_maxa, 24, 6, 3, up(maxa)).
table(up_c_s_maxs_maxa, 24, 19, 3, up(maxa)).
table(up_minc_maxc_s_maxa, 24, 5, 3, up(maxa)).
table(up_minc_s_maxs_maxa, 24, 18, 3, up(maxa)).
table(up_maxc_s_mins_maxa, 24, 9, 3, up(maxa)).
table(up_maxc_s_maxs_maxa, 24, 6, 3, up(maxa)).
table(up_s_mins_maxs_maxa, 24, 19, 3, up(maxa)).
table(up_v_c_minc_maxc_maxa, 24, 19, 4, up(maxa)).
table(up_v_c_minc_s_maxa, 24, 18, 4, up(maxa)).
table(up_v_c_minc_mins_maxa, 24, 11, 4, up(maxa)).
table(up_v_c_minc_maxs_maxa, 24, 8, 4, up(maxa)).
table(up_v_c_maxc_s_maxa, 24, 18, 4, up(maxa)).
table(up_v_c_maxc_mins_maxa, 24, 11, 4, up(maxa)).
table(up_v_c_maxc_maxs_maxa, 24, 6, 4, up(maxa)).
table(up_v_c_s_mins_maxa, 24, 19, 4, up(maxa)).
table(up_v_c_s_maxs_maxa, 24, 19, 4, up(maxa)).
table(up_v_c_mins_maxs_maxa, 24, 9, 4, up(maxa)).
table(up_v_minc_maxc_s_maxa, 24, 18, 4, up(maxa)).
table(up_v_minc_maxc_mins_maxa, 24, 14, 4, up(maxa)).
table(up_v_minc_maxc_maxs_maxa, 24, 9, 4, up(maxa)).
table(up_v_minc_s_mins_maxa, 24, 18, 4, up(maxa)).
table(up_v_minc_s_maxs_maxa, 24, 18, 4, up(maxa)).
table(up_v_minc_mins_maxs_maxa, 24, 16, 4, up(maxa)).
table(up_v_maxc_s_mins_maxa, 24, 8, 4, up(maxa)).
table(up_v_maxc_s_maxs_maxa, 24, 7, 4, up(maxa)).
table(up_v_maxc_mins_maxs_maxa, 24, 5, 4, up(maxa)).
table(up_v_s_mins_maxs_maxa, 24, 19, 4, up(maxa)).
table(up_c_minc_maxc_s_maxa, 24, 18, 4, up(maxa)).
table(up_c_minc_maxc_mins_maxa, 24, 14, 4, up(maxa)).
table(up_c_minc_maxc_maxs_maxa, 24, 9, 4, up(maxa)).
table(up_c_minc_s_maxs_maxa, 24, 9, 4, up(maxa)).
table(up_c_maxc_s_mins_maxa, 24, 8, 4, up(maxa)).
table(up_c_maxc_s_maxs_maxa, 24, 6, 4, up(maxa)).
table(up_c_maxc_mins_maxs_maxa, 24, 5, 4, up(maxa)).
table(up_c_s_mins_maxs_maxa, 24, 19, 4, up(maxa)).
table(up_minc_maxc_s_mins_maxa, 24, 6, 4, up(maxa)).
table(up_minc_maxc_s_maxs_maxa, 24, 6, 4, up(maxa)).
table(up_minc_s_mins_maxs_maxa, 24, 16, 4, up(maxa)).
table(up_maxc_s_mins_maxs_maxa, 24, 5, 4, up(maxa)).
table(up_v_c_minc_maxc_s_maxa, 24, 18, 5, up(maxa)).
table(up_v_c_minc_maxc_mins_maxa, 24, 14, 5, up(maxa)).
table(up_v_c_minc_maxc_maxs_maxa, 24, 9, 5, up(maxa)).
table(up_v_c_minc_s_mins_maxa, 24, 10, 5, up(maxa)).
table(up_v_c_minc_s_maxs_maxa, 24, 9, 5, up(maxa)).
table(up_v_c_minc_mins_maxs_maxa, 24, 8, 5, up(maxa)).
table(up_v_c_maxc_s_mins_maxa, 24, 8, 5, up(maxa)).
table(up_v_c_maxc_s_maxs_maxa, 24, 6, 5, up(maxa)).
table(up_v_c_maxc_mins_maxs_maxa, 24, 6, 5, up(maxa)).
table(up_v_c_s_mins_maxs_maxa, 24, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_mins_maxa, 24, 11, 5, up(maxa)).
table(up_v_minc_maxc_s_maxs_maxa, 24, 10, 5, up(maxa)).
table(up_v_minc_maxc_mins_maxs_maxa, 24, 7, 5, up(maxa)).
table(up_v_minc_s_mins_maxs_maxa, 24, 16, 5, up(maxa)).
table(up_v_maxc_s_mins_maxs_maxa, 24, 8, 5, up(maxa)).
table(up_c_minc_maxc_s_mins_maxa, 24, 11, 5, up(maxa)).
table(up_c_minc_maxc_s_maxs_maxa, 24, 9, 5, up(maxa)).
table(up_c_minc_maxc_mins_maxs_maxa, 24, 7, 5, up(maxa)).
table(up_c_minc_s_mins_maxs_maxa, 24, 10, 5, up(maxa)).
table(up_c_maxc_s_mins_maxs_maxa, 24, 6, 5, up(maxa)).
table(up_minc_maxc_s_mins_maxs_maxa, 24, 7, 5, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxa, 24, 11, 6, up(maxa)).
table(up_v_c_minc_maxc_s_maxs_maxa, 24, 9, 6, up(maxa)).
table(up_v_c_minc_maxc_mins_maxs_maxa, 24, 9, 6, up(maxa)).
table(up_v_c_minc_s_mins_maxs_maxa, 24, 10, 6, up(maxa)).
table(up_v_c_maxc_s_mins_maxs_maxa, 24, 7, 6, up(maxa)).
table(up_v_minc_maxc_s_mins_maxs_maxa, 24, 11, 6, up(maxa)).
table(up_c_minc_maxc_s_mins_maxs_maxa, 24, 10, 6, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxs_maxa, 24, 10, 7, up(maxa)).
table(low_c_v, 25, 18, 1, low(v)).
table(low_minc_v, 25, 11, 1, low(v)).
table(low_maxc_v, 25, 11, 1, low(v)).
table(low_s_v, 25, 7, 1, low(v)).
table(low_mins_v, 25, 18, 1, low(v)).
table(low_maxs_v, 25, 18, 1, low(v)).
table(low_c_minc_v, 25, 11, 2, low(v)).
table(low_c_maxc_v, 25, 11, 2, low(v)).
table(low_c_s_v, 25, 8, 2, low(v)).
table(low_c_mins_v, 25, 18, 2, low(v)).
table(low_c_maxs_v, 25, 18, 2, low(v)).
table(low_minc_maxc_v, 25, 11, 2, low(v)).
table(low_minc_s_v, 25, 3, 2, low(v)).
table(low_minc_mins_v, 25, 12, 2, low(v)).
table(low_minc_maxs_v, 25, 12, 2, low(v)).
table(low_maxc_s_v, 25, 3, 2, low(v)).
table(low_maxc_mins_v, 25, 12, 2, low(v)).
table(low_maxc_maxs_v, 25, 13, 2, low(v)).
table(low_s_mins_v, 25, 7, 2, low(v)).
table(low_s_maxs_v, 25, 7, 2, low(v)).
table(low_mins_maxs_v, 25, 12, 2, low(v)).
table(low_c_minc_maxc_v, 25, 11, 3, low(v)).
table(low_c_minc_s_v, 25, 6, 3, low(v)).
table(low_c_minc_mins_v, 25, 12, 3, low(v)).
table(low_c_minc_maxs_v, 25, 12, 3, low(v)).
table(low_c_maxc_s_v, 25, 4, 3, low(v)).
table(low_c_maxc_mins_v, 25, 12, 3, low(v)).
table(low_c_maxc_maxs_v, 25, 13, 3, low(v)).
table(low_c_s_mins_v, 25, 8, 3, low(v)).
table(low_c_s_maxs_v, 25, 8, 3, low(v)).
table(low_c_mins_maxs_v, 25, 18, 3, low(v)).
table(low_minc_maxc_s_v, 25, 5, 3, low(v)).
table(low_minc_maxc_mins_v, 25, 12, 3, low(v)).
table(low_minc_maxc_maxs_v, 25, 12, 3, low(v)).
table(low_minc_s_mins_v, 25, 4, 3, low(v)).
table(low_minc_s_maxs_v, 25, 4, 3, low(v)).
table(low_minc_mins_maxs_v, 25, 13, 3, low(v)).
table(low_maxc_s_mins_v, 25, 4, 3, low(v)).
table(low_maxc_s_maxs_v, 25, 4, 3, low(v)).
table(low_maxc_mins_maxs_v, 25, 14, 3, low(v)).
table(low_s_mins_maxs_v, 25, 7, 3, low(v)).
table(low_c_minc_maxc_s_v, 25, 7, 4, low(v)).
table(low_c_minc_maxc_mins_v, 25, 12, 4, low(v)).
table(low_c_minc_maxc_maxs_v, 25, 12, 4, low(v)).
table(low_c_minc_s_mins_v, 25, 7, 4, low(v)).
table(low_c_minc_s_maxs_v, 25, 7, 4, low(v)).
table(low_c_minc_mins_maxs_v, 25, 13, 4, low(v)).
table(low_c_maxc_s_mins_v, 25, 5, 4, low(v)).
table(low_c_maxc_s_maxs_v, 25, 5, 4, low(v)).
table(low_c_maxc_mins_maxs_v, 25, 14, 4, low(v)).
table(low_c_s_mins_maxs_v, 25, 8, 4, low(v)).
table(low_minc_maxc_s_mins_v, 25, 6, 4, low(v)).
table(low_minc_maxc_s_maxs_v, 25, 6, 4, low(v)).
table(low_minc_maxc_mins_maxs_v, 25, 13, 4, low(v)).
table(low_minc_s_mins_maxs_v, 25, 5, 4, low(v)).
table(low_maxc_s_mins_maxs_v, 25, 5, 4, low(v)).
table(low_c_minc_maxc_s_mins_v, 25, 8, 5, low(v)).
table(low_c_minc_maxc_s_maxs_v, 25, 8, 5, low(v)).
table(low_c_minc_maxc_mins_maxs_v, 25, 13, 5, low(v)).
table(low_c_minc_s_mins_maxs_v, 25, 8, 5, low(v)).
table(low_c_maxc_s_mins_maxs_v, 25, 6, 5, low(v)).
table(low_minc_maxc_s_mins_maxs_v, 25, 7, 5, low(v)).
table(low_c_minc_maxc_s_mins_maxs_v, 25, 9, 6, low(v)).
table(up_c_maxc_v, 25, 3, 2, up(v)).
table(up_maxc_s_v, 25, 3, 2, up(v)).
table(up_s_maxs_v, 25, 3, 2, up(v)).
table(up_c_minc_maxc_v, 25, 6, 3, up(v)).
table(up_c_maxc_s_v, 25, 4, 3, up(v)).
table(up_c_maxc_mins_v, 25, 4, 3, up(v)).
table(up_c_maxc_maxs_v, 25, 4, 3, up(v)).
table(up_c_s_maxs_v, 25, 4, 3, up(v)).
table(up_minc_maxc_s_v, 25, 5, 3, up(v)).
table(up_minc_s_maxs_v, 25, 4, 3, up(v)).
table(up_maxc_s_mins_v, 25, 4, 3, up(v)).
table(up_maxc_s_maxs_v, 25, 4, 3, up(v)).
table(up_s_mins_maxs_v, 25, 4, 3, up(v)).
table(up_c_minc_maxc_s_v, 25, 7, 4, up(v)).
table(up_c_minc_maxc_mins_v, 25, 7, 4, up(v)).
table(up_c_minc_maxc_maxs_v, 25, 7, 4, up(v)).
table(up_c_minc_s_maxs_v, 25, 7, 4, up(v)).
table(up_c_maxc_s_mins_v, 25, 5, 4, up(v)).
table(up_c_maxc_s_maxs_v, 25, 5, 4, up(v)).
table(up_c_maxc_mins_maxs_v, 25, 5, 4, up(v)).
table(up_c_s_mins_maxs_v, 25, 5, 4, up(v)).
table(up_minc_maxc_s_mins_v, 25, 6, 4, up(v)).
table(up_minc_maxc_s_maxs_v, 25, 6, 4, up(v)).
table(up_minc_s_mins_maxs_v, 25, 5, 4, up(v)).
table(up_maxc_s_mins_maxs_v, 25, 5, 4, up(v)).
table(up_c_minc_maxc_s_mins_v, 25, 8, 5, up(v)).
table(up_c_minc_maxc_s_maxs_v, 25, 8, 5, up(v)).
table(up_c_minc_maxc_mins_maxs_v, 25, 8, 5, up(v)).
table(up_c_minc_s_mins_maxs_v, 25, 8, 5, up(v)).
table(up_c_maxc_s_mins_maxs_v, 25, 6, 5, up(v)).
table(up_minc_maxc_s_mins_maxs_v, 25, 7, 5, up(v)).
table(up_c_minc_maxc_s_mins_maxs_v, 25, 9, 6, up(v)).
table(low_v_c, 25, 11, 1, low(c)).
table(low_minc_c, 25, 11, 1, low(c)).
table(low_maxc_c, 25, 11, 1, low(c)).
table(low_s_c, 25, 8, 1, low(c)).
table(low_mins_c, 25, 8, 1, low(c)).
table(low_maxs_c, 25, 8, 1, low(c)).
table(low_v_minc_c, 25, 11, 2, low(c)).
table(low_v_maxc_c, 25, 3, 2, low(c)).
table(low_v_s_c, 25, 13, 2, low(c)).
table(low_v_mins_c, 25, 12, 2, low(c)).
table(low_v_maxs_c, 25, 13, 2, low(c)).
table(low_minc_maxc_c, 25, 11, 2, low(c)).
table(low_minc_s_c, 25, 5, 2, low(c)).
table(low_minc_mins_c, 25, 9, 2, low(c)).
table(low_minc_maxs_c, 25, 9, 2, low(c)).
table(low_maxc_s_c, 25, 3, 2, low(c)).
table(low_maxc_mins_c, 25, 12, 2, low(c)).
table(low_maxc_maxs_c, 25, 13, 2, low(c)).
table(low_s_mins_c, 25, 10, 2, low(c)).
table(low_s_maxs_c, 25, 11, 2, low(c)).
table(low_mins_maxs_c, 25, 9, 2, low(c)).
table(low_v_minc_maxc_c, 25, 6, 3, low(c)).
table(low_v_minc_s_c, 25, 12, 3, low(c)).
table(low_v_minc_mins_c, 25, 12, 3, low(c)).
table(low_v_minc_maxs_c, 25, 12, 3, low(c)).
table(low_v_maxc_s_c, 25, 4, 3, low(c)).
table(low_v_maxc_mins_c, 25, 4, 3, low(c)).
table(low_v_maxc_maxs_c, 25, 4, 3, low(c)).
table(low_v_s_mins_c, 25, 14, 3, low(c)).
table(low_v_s_maxs_c, 25, 14, 3, low(c)).
table(low_v_mins_maxs_c, 25, 14, 3, low(c)).
table(low_minc_maxc_s_c, 25, 5, 3, low(c)).
table(low_minc_maxc_mins_c, 25, 12, 3, low(c)).
table(low_minc_maxc_maxs_c, 25, 12, 3, low(c)).
table(low_minc_s_mins_c, 25, 6, 3, low(c)).
table(low_minc_s_maxs_c, 25, 7, 3, low(c)).
table(low_minc_mins_maxs_c, 25, 7, 3, low(c)).
table(low_maxc_s_mins_c, 25, 4, 3, low(c)).
table(low_maxc_s_maxs_c, 25, 4, 3, low(c)).
table(low_maxc_mins_maxs_c, 25, 6, 3, low(c)).
table(low_s_mins_maxs_c, 25, 12, 3, low(c)).
table(low_v_minc_maxc_s_c, 25, 7, 4, low(c)).
table(low_v_minc_maxc_mins_c, 25, 7, 4, low(c)).
table(low_v_minc_maxc_maxs_c, 25, 7, 4, low(c)).
table(low_v_minc_s_mins_c, 25, 13, 4, low(c)).
table(low_v_minc_s_maxs_c, 25, 13, 4, low(c)).
table(low_v_minc_mins_maxs_c, 25, 13, 4, low(c)).
table(low_v_maxc_s_mins_c, 25, 5, 4, low(c)).
table(low_v_maxc_s_maxs_c, 25, 5, 4, low(c)).
table(low_v_maxc_mins_maxs_c, 25, 5, 4, low(c)).
table(low_v_s_mins_maxs_c, 25, 15, 4, low(c)).
table(low_minc_maxc_s_mins_c, 25, 6, 4, low(c)).
table(low_minc_maxc_s_maxs_c, 25, 6, 4, low(c)).
table(low_minc_maxc_mins_maxs_c, 25, 8, 4, low(c)).
table(low_minc_s_mins_maxs_c, 25, 8, 4, low(c)).
table(low_maxc_s_mins_maxs_c, 25, 5, 4, low(c)).
table(low_v_minc_maxc_s_mins_c, 25, 8, 5, low(c)).
table(low_v_minc_maxc_s_maxs_c, 25, 8, 5, low(c)).
table(low_v_minc_maxc_mins_maxs_c, 25, 8, 5, low(c)).
table(low_v_minc_s_mins_maxs_c, 25, 14, 5, low(c)).
table(low_v_maxc_s_mins_maxs_c, 25, 6, 5, low(c)).
table(low_minc_maxc_s_mins_maxs_c, 25, 7, 5, low(c)).
table(low_v_minc_maxc_s_mins_maxs_c, 25, 9, 6, low(c)).
table(up_v_c, 25, 18, 1, up(c)).
table(up_s_c, 25, 3, 1, up(c)).
table(up_v_minc_c, 25, 6, 2, up(c)).
table(up_v_maxc_c, 25, 11, 2, up(c)).
table(up_v_s_c, 25, 4, 2, up(c)).
table(up_v_mins_c, 25, 10, 2, up(c)).
table(up_v_maxs_c, 25, 18, 2, up(c)).
table(up_minc_s_c, 25, 3, 2, up(c)).
table(up_maxc_s_c, 25, 4, 2, up(c)).
table(up_s_mins_c, 25, 5, 2, up(c)).
table(up_s_maxs_c, 25, 5, 2, up(c)).
table(up_v_minc_maxc_c, 25, 7, 3, up(c)).
table(up_v_minc_s_c, 25, 6, 3, up(c)).
table(up_v_minc_mins_c, 25, 7, 3, up(c)).
table(up_v_minc_maxs_c, 25, 7, 3, up(c)).
table(up_v_maxc_s_c, 25, 5, 3, up(c)).
table(up_v_maxc_mins_c, 25, 5, 3, up(c)).
table(up_v_maxc_maxs_c, 25, 13, 3, up(c)).
table(up_v_s_mins_c, 25, 8, 3, up(c)).
table(up_v_s_maxs_c, 25, 6, 3, up(c)).
table(up_v_mins_maxs_c, 25, 12, 3, up(c)).
table(up_minc_maxc_s_c, 25, 5, 3, up(c)).
table(up_minc_s_mins_c, 25, 4, 3, up(c)).
table(up_minc_s_maxs_c, 25, 4, 3, up(c)).
table(up_maxc_s_mins_c, 25, 5, 3, up(c)).
table(up_maxc_s_maxs_c, 25, 5, 3, up(c)).
table(up_s_mins_maxs_c, 25, 8, 3, up(c)).
table(up_v_minc_maxc_s_c, 25, 7, 4, up(c)).
table(up_v_minc_maxc_mins_c, 25, 8, 4, up(c)).
table(up_v_minc_maxc_maxs_c, 25, 8, 4, up(c)).
table(up_v_minc_s_mins_c, 25, 7, 4, up(c)).
table(up_v_minc_s_maxs_c, 25, 7, 4, up(c)).
table(up_v_minc_mins_maxs_c, 25, 8, 4, up(c)).
table(up_v_maxc_s_mins_c, 25, 6, 4, up(c)).
table(up_v_maxc_s_maxs_c, 25, 6, 4, up(c)).
table(up_v_maxc_mins_maxs_c, 25, 7, 4, up(c)).
table(up_v_s_mins_maxs_c, 25, 10, 4, up(c)).
table(up_minc_maxc_s_mins_c, 25, 6, 4, up(c)).
table(up_minc_maxc_s_maxs_c, 25, 6, 4, up(c)).
table(up_minc_s_mins_maxs_c, 25, 5, 4, up(c)).
table(up_maxc_s_mins_maxs_c, 25, 6, 4, up(c)).
table(up_v_minc_maxc_s_mins_c, 25, 8, 5, up(c)).
table(up_v_minc_maxc_s_maxs_c, 25, 8, 5, up(c)).
table(up_v_minc_maxc_mins_maxs_c, 25, 9, 5, up(c)).
table(up_v_minc_s_mins_maxs_c, 25, 8, 5, up(c)).
table(up_v_maxc_s_mins_maxs_c, 25, 7, 5, up(c)).
table(up_minc_maxc_s_mins_maxs_c, 25, 7, 5, up(c)).
table(up_v_minc_maxc_s_mins_maxs_c, 25, 9, 6, up(c)).
table(low_v_minc, 25, 3, 1, low(minc)).
table(low_c_minc, 25, 3, 1, low(minc)).
table(low_maxc_minc, 25, 3, 1, low(minc)).
table(low_s_minc, 25, 3, 1, low(minc)).
table(low_mins_minc, 25, 2, 1, low(minc)).
table(low_maxs_minc, 25, 3, 1, low(minc)).
table(low_v_c_minc, 25, 5, 2, low(minc)).
table(low_v_maxc_minc, 25, 4, 2, low(minc)).
table(low_v_s_minc, 25, 4, 2, low(minc)).
table(low_v_mins_minc, 25, 3, 2, low(minc)).
table(low_v_maxs_minc, 25, 4, 2, low(minc)).
table(low_c_maxc_minc, 25, 4, 2, low(minc)).
table(low_c_s_minc, 25, 4, 2, low(minc)).
table(low_c_mins_minc, 25, 3, 2, low(minc)).
table(low_c_maxs_minc, 25, 4, 2, low(minc)).
table(low_maxc_s_minc, 25, 4, 2, low(minc)).
table(low_maxc_mins_minc, 25, 4, 2, low(minc)).
table(low_maxc_maxs_minc, 25, 4, 2, low(minc)).
table(low_s_mins_minc, 25, 3, 2, low(minc)).
table(low_s_maxs_minc, 25, 4, 2, low(minc)).
table(low_mins_maxs_minc, 25, 3, 2, low(minc)).
table(low_v_c_maxc_minc, 25, 6, 3, low(minc)).
table(low_v_c_s_minc, 25, 6, 3, low(minc)).
table(low_v_c_mins_minc, 25, 6, 3, low(minc)).
table(low_v_c_maxs_minc, 25, 6, 3, low(minc)).
table(low_v_maxc_s_minc, 25, 5, 3, low(minc)).
table(low_v_maxc_mins_minc, 25, 5, 3, low(minc)).
table(low_v_maxc_maxs_minc, 25, 5, 3, low(minc)).
table(low_v_s_mins_minc, 25, 4, 3, low(minc)).
table(low_v_s_maxs_minc, 25, 5, 3, low(minc)).
table(low_v_mins_maxs_minc, 25, 4, 3, low(minc)).
table(low_c_maxc_s_minc, 25, 5, 3, low(minc)).
table(low_c_maxc_mins_minc, 25, 5, 3, low(minc)).
table(low_c_maxc_maxs_minc, 25, 5, 3, low(minc)).
table(low_c_s_mins_minc, 25, 4, 3, low(minc)).
table(low_c_s_maxs_minc, 25, 6, 3, low(minc)).
table(low_c_mins_maxs_minc, 25, 4, 3, low(minc)).
table(low_maxc_s_mins_minc, 25, 5, 3, low(minc)).
table(low_maxc_s_maxs_minc, 25, 5, 3, low(minc)).
table(low_maxc_mins_maxs_minc, 25, 5, 3, low(minc)).
table(low_s_mins_maxs_minc, 25, 4, 3, low(minc)).
table(low_v_c_maxc_s_minc, 25, 7, 4, low(minc)).
table(low_v_c_maxc_mins_minc, 25, 7, 4, low(minc)).
table(low_v_c_maxc_maxs_minc, 25, 7, 4, low(minc)).
table(low_v_c_s_mins_minc, 25, 7, 4, low(minc)).
table(low_v_c_s_maxs_minc, 25, 7, 4, low(minc)).
table(low_v_c_mins_maxs_minc, 25, 7, 4, low(minc)).
table(low_v_maxc_s_mins_minc, 25, 6, 4, low(minc)).
table(low_v_maxc_s_maxs_minc, 25, 6, 4, low(minc)).
table(low_v_maxc_mins_maxs_minc, 25, 6, 4, low(minc)).
table(low_v_s_mins_maxs_minc, 25, 5, 4, low(minc)).
table(low_c_maxc_s_mins_minc, 25, 6, 4, low(minc)).
table(low_c_maxc_s_maxs_minc, 25, 6, 4, low(minc)).
table(low_c_maxc_mins_maxs_minc, 25, 6, 4, low(minc)).
table(low_c_s_mins_maxs_minc, 25, 6, 4, low(minc)).
table(low_maxc_s_mins_maxs_minc, 25, 6, 4, low(minc)).
table(low_v_c_maxc_s_mins_minc, 25, 8, 5, low(minc)).
table(low_v_c_maxc_s_maxs_minc, 25, 8, 5, low(minc)).
table(low_v_c_maxc_mins_maxs_minc, 25, 8, 5, low(minc)).
table(low_v_c_s_mins_maxs_minc, 25, 8, 5, low(minc)).
table(low_v_maxc_s_mins_maxs_minc, 25, 7, 5, low(minc)).
table(low_c_maxc_s_mins_maxs_minc, 25, 7, 5, low(minc)).
table(low_v_c_maxc_s_mins_maxs_minc, 25, 9, 6, low(minc)).
table(up_v_minc, 25, 11, 1, up(minc)).
table(up_maxc_minc, 25, 6, 1, up(minc)).
table(up_v_c_minc, 25, 5, 2, up(minc)).
table(up_v_maxc_minc, 25, 4, 2, up(minc)).
table(up_v_s_minc, 25, 13, 2, up(minc)).
table(up_v_mins_minc, 25, 12, 2, up(minc)).
table(up_v_maxs_minc, 25, 13, 2, up(minc)).
table(up_c_maxc_minc, 25, 4, 2, up(minc)).
table(up_maxc_s_minc, 25, 4, 2, up(minc)).
table(up_maxc_mins_minc, 25, 4, 2, up(minc)).
table(up_maxc_maxs_minc, 25, 7, 2, up(minc)).
table(up_s_maxs_minc, 25, 14, 2, up(minc)).
table(up_v_c_maxc_minc, 25, 6, 3, up(minc)).
table(up_v_c_s_minc, 25, 6, 3, up(minc)).
table(up_v_c_mins_minc, 25, 6, 3, up(minc)).
table(up_v_c_maxs_minc, 25, 6, 3, up(minc)).
table(up_v_maxc_s_minc, 25, 5, 3, up(minc)).
table(up_v_maxc_mins_minc, 25, 5, 3, up(minc)).
table(up_v_maxc_maxs_minc, 25, 5, 3, up(minc)).
table(up_v_s_mins_minc, 25, 14, 3, up(minc)).
table(up_v_s_maxs_minc, 25, 14, 3, up(minc)).
table(up_v_mins_maxs_minc, 25, 14, 3, up(minc)).
table(up_c_maxc_s_minc, 25, 5, 3, up(minc)).
table(up_c_maxc_mins_minc, 25, 5, 3, up(minc)).
table(up_c_maxc_maxs_minc, 25, 5, 3, up(minc)).
table(up_c_s_maxs_minc, 25, 4, 3, up(minc)).
table(up_maxc_s_mins_minc, 25, 5, 3, up(minc)).
table(up_maxc_s_maxs_minc, 25, 5, 3, up(minc)).
table(up_maxc_mins_maxs_minc, 25, 5, 3, up(minc)).
table(up_s_mins_maxs_minc, 25, 15, 3, up(minc)).
table(up_v_c_maxc_s_minc, 25, 7, 4, up(minc)).
table(up_v_c_maxc_mins_minc, 25, 7, 4, up(minc)).
table(up_v_c_maxc_maxs_minc, 25, 7, 4, up(minc)).
table(up_v_c_s_mins_minc, 25, 7, 4, up(minc)).
table(up_v_c_s_maxs_minc, 25, 7, 4, up(minc)).
table(up_v_c_mins_maxs_minc, 25, 7, 4, up(minc)).
table(up_v_maxc_s_mins_minc, 25, 6, 4, up(minc)).
table(up_v_maxc_s_maxs_minc, 25, 6, 4, up(minc)).
table(up_v_maxc_mins_maxs_minc, 25, 6, 4, up(minc)).
table(up_v_s_mins_maxs_minc, 25, 15, 4, up(minc)).
table(up_c_maxc_s_mins_minc, 25, 6, 4, up(minc)).
table(up_c_maxc_s_maxs_minc, 25, 6, 4, up(minc)).
table(up_c_maxc_mins_maxs_minc, 25, 6, 4, up(minc)).
table(up_c_s_mins_maxs_minc, 25, 5, 4, up(minc)).
table(up_maxc_s_mins_maxs_minc, 25, 6, 4, up(minc)).
table(up_v_c_maxc_s_mins_minc, 25, 8, 5, up(minc)).
table(up_v_c_maxc_s_maxs_minc, 25, 8, 5, up(minc)).
table(up_v_c_maxc_mins_maxs_minc, 25, 8, 5, up(minc)).
table(up_v_c_s_mins_maxs_minc, 25, 8, 5, up(minc)).
table(up_v_maxc_s_mins_maxs_minc, 25, 7, 5, up(minc)).
table(up_c_maxc_s_mins_maxs_minc, 25, 7, 5, up(minc)).
table(up_v_c_maxc_s_mins_maxs_minc, 25, 9, 6, up(minc)).
table(low_v_maxc, 25, 18, 1, low(maxc)).
table(low_c_maxc, 25, 18, 1, low(maxc)).
table(low_minc_maxc, 25, 6, 1, low(maxc)).
table(low_s_maxc, 25, 18, 1, low(maxc)).
table(low_mins_maxc, 25, 9, 1, low(maxc)).
table(low_maxs_maxc, 25, 2, 1, low(maxc)).
table(low_v_c_maxc, 25, 3, 2, low(maxc)).
table(low_v_minc_maxc, 25, 4, 2, low(maxc)).
table(low_v_s_maxc, 25, 4, 2, low(maxc)).
table(low_v_mins_maxc, 25, 7, 2, low(maxc)).
table(low_v_maxs_maxc, 25, 3, 2, low(maxc)).
table(low_c_minc_maxc, 25, 11, 2, low(maxc)).
table(low_c_s_maxc, 25, 4, 2, low(maxc)).
table(low_c_mins_maxc, 25, 18, 2, low(maxc)).
table(low_c_maxs_maxc, 25, 3, 2, low(maxc)).
table(low_minc_s_maxc, 25, 4, 2, low(maxc)).
table(low_minc_mins_maxc, 25, 4, 2, low(maxc)).
table(low_minc_maxs_maxc, 25, 4, 2, low(maxc)).
table(low_s_mins_maxc, 25, 18, 2, low(maxc)).
table(low_s_maxs_maxc, 25, 3, 2, low(maxc)).
table(low_mins_maxs_maxc, 25, 3, 2, low(maxc)).
table(low_v_c_minc_maxc, 25, 6, 3, low(maxc)).
table(low_v_c_s_maxc, 25, 4, 3, low(maxc)).
table(low_v_c_mins_maxc, 25, 4, 3, low(maxc)).
table(low_v_c_maxs_maxc, 25, 4, 3, low(maxc)).
table(low_v_minc_s_maxc, 25, 5, 3, low(maxc)).
table(low_v_minc_mins_maxc, 25, 5, 3, low(maxc)).
table(low_v_minc_maxs_maxc, 25, 5, 3, low(maxc)).
table(low_v_s_mins_maxc, 25, 5, 3, low(maxc)).
table(low_v_s_maxs_maxc, 25, 4, 3, low(maxc)).
table(low_v_mins_maxs_maxc, 25, 4, 3, low(maxc)).
table(low_c_minc_s_maxc, 25, 5, 3, low(maxc)).
table(low_c_minc_mins_maxc, 25, 6, 3, low(maxc)).
table(low_c_minc_maxs_maxc, 25, 5, 3, low(maxc)).
table(low_c_s_mins_maxc, 25, 4, 3, low(maxc)).
table(low_c_s_maxs_maxc, 25, 4, 3, low(maxc)).
table(low_c_mins_maxs_maxc, 25, 4, 3, low(maxc)).
table(low_minc_s_mins_maxc, 25, 5, 3, low(maxc)).
table(low_minc_s_maxs_maxc, 25, 5, 3, low(maxc)).
table(low_minc_mins_maxs_maxc, 25, 5, 3, low(maxc)).
table(low_s_mins_maxs_maxc, 25, 4, 3, low(maxc)).
table(low_v_c_minc_s_maxc, 25, 7, 4, low(maxc)).
table(low_v_c_minc_mins_maxc, 25, 7, 4, low(maxc)).
table(low_v_c_minc_maxs_maxc, 25, 7, 4, low(maxc)).
table(low_v_c_s_mins_maxc, 25, 5, 4, low(maxc)).
table(low_v_c_s_maxs_maxc, 25, 5, 4, low(maxc)).
table(low_v_c_mins_maxs_maxc, 25, 5, 4, low(maxc)).
table(low_v_minc_s_mins_maxc, 25, 6, 4, low(maxc)).
table(low_v_minc_s_maxs_maxc, 25, 6, 4, low(maxc)).
table(low_v_minc_mins_maxs_maxc, 25, 6, 4, low(maxc)).
table(low_v_s_mins_maxs_maxc, 25, 5, 4, low(maxc)).
table(low_c_minc_s_mins_maxc, 25, 6, 4, low(maxc)).
table(low_c_minc_s_maxs_maxc, 25, 6, 4, low(maxc)).
table(low_c_minc_mins_maxs_maxc, 25, 6, 4, low(maxc)).
table(low_c_s_mins_maxs_maxc, 25, 5, 4, low(maxc)).
table(low_minc_s_mins_maxs_maxc, 25, 6, 4, low(maxc)).
table(low_v_c_minc_s_mins_maxc, 25, 8, 5, low(maxc)).
table(low_v_c_minc_s_maxs_maxc, 25, 8, 5, low(maxc)).
table(low_v_c_minc_mins_maxs_maxc, 25, 8, 5, low(maxc)).
table(low_v_c_s_mins_maxs_maxc, 25, 6, 5, low(maxc)).
table(low_v_minc_s_mins_maxs_maxc, 25, 7, 5, low(maxc)).
table(low_c_minc_s_mins_maxs_maxc, 25, 7, 5, low(maxc)).
table(low_v_c_minc_s_mins_maxs_maxc, 25, 9, 6, low(maxc)).
table(up_v_maxc, 25, 11, 1, up(maxc)).
table(up_v_c_maxc, 25, 11, 2, up(maxc)).
table(up_v_minc_maxc, 25, 11, 2, up(maxc)).
table(up_v_s_maxc, 25, 13, 2, up(maxc)).
table(up_v_mins_maxc, 25, 12, 2, up(maxc)).
table(up_v_maxs_maxc, 25, 13, 2, up(maxc)).
table(up_v_c_minc_maxc, 25, 11, 3, up(maxc)).
table(up_v_c_s_maxc, 25, 13, 3, up(maxc)).
table(up_v_c_mins_maxc, 25, 12, 3, up(maxc)).
table(up_v_c_maxs_maxc, 25, 13, 3, up(maxc)).
table(up_v_minc_s_maxc, 25, 12, 3, up(maxc)).
table(up_v_minc_mins_maxc, 25, 12, 3, up(maxc)).
table(up_v_minc_maxs_maxc, 25, 12, 3, up(maxc)).
table(up_v_s_mins_maxc, 25, 14, 3, up(maxc)).
table(up_v_s_maxs_maxc, 25, 14, 3, up(maxc)).
table(up_v_mins_maxs_maxc, 25, 14, 3, up(maxc)).
table(up_v_c_minc_s_maxc, 25, 12, 4, up(maxc)).
table(up_v_c_minc_mins_maxc, 25, 12, 4, up(maxc)).
table(up_v_c_minc_maxs_maxc, 25, 12, 4, up(maxc)).
table(up_v_c_s_mins_maxc, 25, 14, 4, up(maxc)).
table(up_v_c_s_maxs_maxc, 25, 6, 4, up(maxc)).
table(up_v_c_mins_maxs_maxc, 25, 10, 4, up(maxc)).
table(up_v_minc_s_mins_maxc, 25, 13, 4, up(maxc)).
table(up_v_minc_s_maxs_maxc, 25, 13, 4, up(maxc)).
table(up_v_minc_mins_maxs_maxc, 25, 13, 4, up(maxc)).
table(up_v_s_mins_maxs_maxc, 25, 15, 4, up(maxc)).
table(up_v_c_minc_s_mins_maxc, 25, 13, 5, up(maxc)).
table(up_v_c_minc_s_maxs_maxc, 25, 8, 5, up(maxc)).
table(up_v_c_minc_mins_maxs_maxc, 25, 9, 5, up(maxc)).
table(up_v_c_s_mins_maxs_maxc, 25, 10, 5, up(maxc)).
table(up_v_minc_s_mins_maxs_maxc, 25, 14, 5, up(maxc)).
table(up_v_c_minc_s_mins_maxs_maxc, 25, 9, 6, up(maxc)).
table(low_v_s, 25, 18, 1, low(s)).
table(low_c_s, 25, 3, 1, low(s)).
table(low_minc_s, 25, 18, 1, low(s)).
table(low_maxc_s, 25, 18, 1, low(s)).
table(low_mins_s, 25, 18, 1, low(s)).
table(low_maxs_s, 25, 18, 1, low(s)).
table(low_v_c_s, 25, 4, 2, low(s)).
table(low_v_minc_s, 25, 18, 2, low(s)).
table(low_v_maxc_s, 25, 6, 2, low(s)).
table(low_v_mins_s, 25, 9, 2, low(s)).
table(low_v_maxs_s, 25, 3, 2, low(s)).
table(low_c_minc_s, 25, 5, 2, low(s)).
table(low_c_maxc_s, 25, 5, 2, low(s)).
table(low_c_mins_s, 25, 5, 2, low(s)).
table(low_c_maxs_s, 25, 5, 2, low(s)).
table(low_minc_maxc_s, 25, 18, 2, low(s)).
table(low_minc_mins_s, 25, 12, 2, low(s)).
table(low_minc_maxs_s, 25, 14, 2, low(s)).
table(low_maxc_mins_s, 25, 7, 2, low(s)).
table(low_maxc_maxs_s, 25, 14, 2, low(s)).
table(low_mins_maxs_s, 25, 12, 2, low(s)).
table(low_v_c_minc_s, 25, 8, 3, low(s)).
table(low_v_c_maxc_s, 25, 6, 3, low(s)).
table(low_v_c_mins_s, 25, 8, 3, low(s)).
table(low_v_c_maxs_s, 25, 4, 3, low(s)).
table(low_v_minc_maxc_s, 25, 10, 3, low(s)).
table(low_v_minc_mins_s, 25, 13, 3, low(s)).
table(low_v_minc_maxs_s, 25, 4, 3, low(s)).
table(low_v_maxc_mins_s, 25, 4, 3, low(s)).
table(low_v_maxc_maxs_s, 25, 4, 3, low(s)).
table(low_v_mins_maxs_s, 25, 4, 3, low(s)).
table(low_c_minc_maxc_s, 25, 8, 3, low(s)).
table(low_c_minc_mins_s, 25, 4, 3, low(s)).
table(low_c_minc_maxs_s, 25, 5, 3, low(s)).
table(low_c_maxc_mins_s, 25, 8, 3, low(s)).
table(low_c_maxc_maxs_s, 25, 5, 3, low(s)).
table(low_c_mins_maxs_s, 25, 8, 3, low(s)).
table(low_minc_maxc_mins_s, 25, 13, 3, low(s)).
table(low_minc_maxc_maxs_s, 25, 14, 3, low(s)).
table(low_minc_mins_maxs_s, 25, 8, 3, low(s)).
table(low_maxc_mins_maxs_s, 25, 9, 3, low(s)).
table(low_v_c_minc_maxc_s, 25, 10, 4, low(s)).
table(low_v_c_minc_mins_s, 25, 7, 4, low(s)).
table(low_v_c_minc_maxs_s, 25, 7, 4, low(s)).
table(low_v_c_maxc_mins_s, 25, 5, 4, low(s)).
table(low_v_c_maxc_maxs_s, 25, 5, 4, low(s)).
table(low_v_c_mins_maxs_s, 25, 5, 4, low(s)).
table(low_v_minc_maxc_mins_s, 25, 8, 4, low(s)).
table(low_v_minc_maxc_maxs_s, 25, 6, 4, low(s)).
table(low_v_minc_mins_maxs_s, 25, 5, 4, low(s)).
table(low_v_maxc_mins_maxs_s, 25, 5, 4, low(s)).
table(low_c_minc_maxc_mins_s, 25, 6, 4, low(s)).
table(low_c_minc_maxc_maxs_s, 25, 7, 4, low(s)).
table(low_c_minc_mins_maxs_s, 25, 5, 4, low(s)).
table(low_c_maxc_mins_maxs_s, 25, 6, 4, low(s)).
table(low_minc_maxc_mins_maxs_s, 25, 9, 4, low(s)).
table(low_v_c_minc_maxc_mins_s, 25, 8, 5, low(s)).
table(low_v_c_minc_maxc_maxs_s, 25, 8, 5, low(s)).
table(low_v_c_minc_mins_maxs_s, 25, 8, 5, low(s)).
table(low_v_c_maxc_mins_maxs_s, 25, 6, 5, low(s)).
table(low_v_minc_maxc_mins_maxs_s, 25, 7, 5, low(s)).
table(low_c_minc_maxc_mins_maxs_s, 25, 7, 5, low(s)).
table(low_v_c_minc_maxc_mins_maxs_s, 25, 9, 6, low(s)).
table(up_v_s, 25, 7, 1, up(s)).
table(up_v_c_s, 25, 8, 2, up(s)).
table(up_v_minc_s, 25, 8, 2, up(s)).
table(up_v_maxc_s, 25, 8, 2, up(s)).
table(up_v_mins_s, 25, 4, 2, up(s)).
table(up_v_maxs_s, 25, 7, 2, up(s)).
table(up_c_maxc_s, 25, 9, 2, up(s)).
table(up_v_c_minc_s, 25, 11, 3, up(s)).
table(up_v_c_maxc_s, 25, 9, 3, up(s)).
table(up_v_c_mins_s, 25, 5, 3, up(s)).
table(up_v_c_maxs_s, 25, 8, 3, up(s)).
table(up_v_minc_maxc_s, 25, 10, 3, up(s)).
table(up_v_minc_mins_s, 25, 5, 3, up(s)).
table(up_v_minc_maxs_s, 25, 8, 3, up(s)).
table(up_v_maxc_mins_s, 25, 5, 3, up(s)).
table(up_v_maxc_maxs_s, 25, 8, 3, up(s)).
table(up_v_mins_maxs_s, 25, 5, 3, up(s)).
table(up_c_minc_maxc_s, 25, 12, 3, up(s)).
table(up_c_maxc_mins_s, 25, 5, 3, up(s)).
table(up_c_maxc_maxs_s, 25, 9, 3, up(s)).
table(up_v_c_minc_maxc_s, 25, 12, 4, up(s)).
table(up_v_c_minc_mins_s, 25, 8, 4, up(s)).
table(up_v_c_minc_maxs_s, 25, 11, 4, up(s)).
table(up_v_c_maxc_mins_s, 25, 6, 4, up(s)).
table(up_v_c_maxc_maxs_s, 25, 9, 4, up(s)).
table(up_v_c_mins_maxs_s, 25, 6, 4, up(s)).
table(up_v_minc_maxc_mins_s, 25, 7, 4, up(s)).
table(up_v_minc_maxc_maxs_s, 25, 10, 4, up(s)).
table(up_v_minc_mins_maxs_s, 25, 6, 4, up(s)).
table(up_v_maxc_mins_maxs_s, 25, 6, 4, up(s)).
table(up_c_minc_maxc_mins_s, 25, 7, 4, up(s)).
table(up_c_minc_maxc_maxs_s, 25, 12, 4, up(s)).
table(up_c_maxc_mins_maxs_s, 25, 6, 4, up(s)).
table(up_v_c_minc_maxc_mins_s, 25, 9, 5, up(s)).
table(up_v_c_minc_maxc_maxs_s, 25, 12, 5, up(s)).
table(up_v_c_minc_mins_maxs_s, 25, 9, 5, up(s)).
table(up_v_c_maxc_mins_maxs_s, 25, 7, 5, up(s)).
table(up_v_minc_maxc_mins_maxs_s, 25, 8, 5, up(s)).
table(up_c_minc_maxc_mins_maxs_s, 25, 8, 5, up(s)).
table(up_v_c_minc_maxc_mins_maxs_s, 25, 10, 6, up(s)).
table(low_v_mins, 25, 2, 1, low(mins)).
table(low_c_mins, 25, 2, 1, low(mins)).
table(low_minc_mins, 25, 2, 1, low(mins)).
table(low_maxc_mins, 25, 2, 1, low(mins)).
table(low_s_mins, 25, 2, 1, low(mins)).
table(low_maxs_mins, 25, 2, 1, low(mins)).
table(low_v_c_mins, 25, 3, 2, low(mins)).
table(low_v_minc_mins, 25, 3, 2, low(mins)).
table(low_v_maxc_mins, 25, 3, 2, low(mins)).
table(low_v_s_mins, 25, 3, 2, low(mins)).
table(low_v_maxs_mins, 25, 3, 2, low(mins)).
table(low_c_minc_mins, 25, 3, 2, low(mins)).
table(low_c_maxc_mins, 25, 3, 2, low(mins)).
table(low_c_s_mins, 25, 3, 2, low(mins)).
table(low_c_maxs_mins, 25, 3, 2, low(mins)).
table(low_minc_maxc_mins, 25, 4, 2, low(mins)).
table(low_minc_s_mins, 25, 3, 2, low(mins)).
table(low_minc_maxs_mins, 25, 3, 2, low(mins)).
table(low_maxc_s_mins, 25, 3, 2, low(mins)).
table(low_maxc_maxs_mins, 25, 3, 2, low(mins)).
table(low_s_maxs_mins, 25, 3, 2, low(mins)).
table(low_v_c_minc_mins, 25, 6, 3, low(mins)).
table(low_v_c_maxc_mins, 25, 4, 3, low(mins)).
table(low_v_c_s_mins, 25, 4, 3, low(mins)).
table(low_v_c_maxs_mins, 25, 4, 3, low(mins)).
table(low_v_minc_maxc_mins, 25, 5, 3, low(mins)).
table(low_v_minc_s_mins, 25, 4, 3, low(mins)).
table(low_v_minc_maxs_mins, 25, 4, 3, low(mins)).
table(low_v_maxc_s_mins, 25, 4, 3, low(mins)).
table(low_v_maxc_maxs_mins, 25, 4, 3, low(mins)).
table(low_v_s_maxs_mins, 25, 4, 3, low(mins)).
table(low_c_minc_maxc_mins, 25, 5, 3, low(mins)).
table(low_c_minc_s_mins, 25, 4, 3, low(mins)).
table(low_c_minc_maxs_mins, 25, 4, 3, low(mins)).
table(low_c_maxc_s_mins, 25, 4, 3, low(mins)).
table(low_c_maxc_maxs_mins, 25, 4, 3, low(mins)).
table(low_c_s_maxs_mins, 25, 4, 3, low(mins)).
table(low_minc_maxc_s_mins, 25, 5, 3, low(mins)).
table(low_minc_maxc_maxs_mins, 25, 5, 3, low(mins)).
table(low_minc_s_maxs_mins, 25, 4, 3, low(mins)).
table(low_maxc_s_maxs_mins, 25, 4, 3, low(mins)).
table(low_v_c_minc_maxc_mins, 25, 7, 4, low(mins)).
table(low_v_c_minc_s_mins, 25, 7, 4, low(mins)).
table(low_v_c_minc_maxs_mins, 25, 7, 4, low(mins)).
table(low_v_c_maxc_s_mins, 25, 5, 4, low(mins)).
table(low_v_c_maxc_maxs_mins, 25, 5, 4, low(mins)).
table(low_v_c_s_maxs_mins, 25, 5, 4, low(mins)).
table(low_v_minc_maxc_s_mins, 25, 6, 4, low(mins)).
table(low_v_minc_maxc_maxs_mins, 25, 6, 4, low(mins)).
table(low_v_minc_s_maxs_mins, 25, 5, 4, low(mins)).
table(low_v_maxc_s_maxs_mins, 25, 5, 4, low(mins)).
table(low_c_minc_maxc_s_mins, 25, 6, 4, low(mins)).
table(low_c_minc_maxc_maxs_mins, 25, 6, 4, low(mins)).
table(low_c_minc_s_maxs_mins, 25, 5, 4, low(mins)).
table(low_c_maxc_s_maxs_mins, 25, 5, 4, low(mins)).
table(low_minc_maxc_s_maxs_mins, 25, 6, 4, low(mins)).
table(low_v_c_minc_maxc_s_mins, 25, 8, 5, low(mins)).
table(low_v_c_minc_maxc_maxs_mins, 25, 8, 5, low(mins)).
table(low_v_c_minc_s_maxs_mins, 25, 8, 5, low(mins)).
table(low_v_c_maxc_s_maxs_mins, 25, 6, 5, low(mins)).
table(low_v_minc_maxc_s_maxs_mins, 25, 7, 5, low(mins)).
table(low_c_minc_maxc_s_maxs_mins, 25, 7, 5, low(mins)).
table(low_v_c_minc_maxc_s_maxs_mins, 25, 9, 6, low(mins)).
table(up_v_mins, 25, 18, 1, up(mins)).
table(up_maxc_mins, 25, 9, 1, up(mins)).
table(up_maxs_mins, 25, 2, 1, up(mins)).
table(up_v_c_mins, 25, 6, 2, up(mins)).
table(up_v_minc_mins, 25, 3, 2, up(mins)).
table(up_v_maxc_mins, 25, 5, 2, up(mins)).
table(up_v_s_mins, 25, 3, 2, up(mins)).
table(up_v_maxs_mins, 25, 3, 2, up(mins)).
table(up_c_maxc_mins, 25, 5, 2, up(mins)).
table(up_c_maxs_mins, 25, 3, 2, up(mins)).
table(up_minc_maxc_mins, 25, 4, 2, up(mins)).
table(up_minc_maxs_mins, 25, 3, 2, up(mins)).
table(up_maxc_s_mins, 25, 3, 2, up(mins)).
table(up_maxc_maxs_mins, 25, 3, 2, up(mins)).
table(up_s_maxs_mins, 25, 3, 2, up(mins)).
table(up_v_c_minc_mins, 25, 6, 3, up(mins)).
table(up_v_c_maxc_mins, 25, 7, 3, up(mins)).
table(up_v_c_s_mins, 25, 4, 3, up(mins)).
table(up_v_c_maxs_mins, 25, 4, 3, up(mins)).
table(up_v_minc_maxc_mins, 25, 5, 3, up(mins)).
table(up_v_minc_s_mins, 25, 4, 3, up(mins)).
table(up_v_minc_maxs_mins, 25, 4, 3, up(mins)).
table(up_v_maxc_s_mins, 25, 4, 3, up(mins)).
table(up_v_maxc_maxs_mins, 25, 4, 3, up(mins)).
table(up_v_s_maxs_mins, 25, 4, 3, up(mins)).
table(up_c_minc_maxc_mins, 25, 5, 3, up(mins)).
table(up_c_minc_maxs_mins, 25, 4, 3, up(mins)).
table(up_c_maxc_s_mins, 25, 4, 3, up(mins)).
table(up_c_maxc_maxs_mins, 25, 4, 3, up(mins)).
table(up_c_s_maxs_mins, 25, 4, 3, up(mins)).
table(up_minc_maxc_s_mins, 25, 5, 3, up(mins)).
table(up_minc_maxc_maxs_mins, 25, 5, 3, up(mins)).
table(up_minc_s_maxs_mins, 25, 4, 3, up(mins)).
table(up_maxc_s_maxs_mins, 25, 4, 3, up(mins)).
table(up_v_c_minc_maxc_mins, 25, 7, 4, up(mins)).
table(up_v_c_minc_s_mins, 25, 7, 4, up(mins)).
table(up_v_c_minc_maxs_mins, 25, 7, 4, up(mins)).
table(up_v_c_maxc_s_mins, 25, 5, 4, up(mins)).
table(up_v_c_maxc_maxs_mins, 25, 5, 4, up(mins)).
table(up_v_c_s_maxs_mins, 25, 5, 4, up(mins)).
table(up_v_minc_maxc_s_mins, 25, 6, 4, up(mins)).
table(up_v_minc_maxc_maxs_mins, 25, 6, 4, up(mins)).
table(up_v_minc_s_maxs_mins, 25, 5, 4, up(mins)).
table(up_v_maxc_s_maxs_mins, 25, 5, 4, up(mins)).
table(up_c_minc_maxc_s_mins, 25, 6, 4, up(mins)).
table(up_c_minc_maxc_maxs_mins, 25, 6, 4, up(mins)).
table(up_c_minc_s_maxs_mins, 25, 5, 4, up(mins)).
table(up_c_maxc_s_maxs_mins, 25, 5, 4, up(mins)).
table(up_minc_maxc_s_maxs_mins, 25, 6, 4, up(mins)).
table(up_v_c_minc_maxc_s_mins, 25, 8, 5, up(mins)).
table(up_v_c_minc_maxc_maxs_mins, 25, 8, 5, up(mins)).
table(up_v_c_minc_s_maxs_mins, 25, 8, 5, up(mins)).
table(up_v_c_maxc_s_maxs_mins, 25, 6, 5, up(mins)).
table(up_v_minc_maxc_s_maxs_mins, 25, 7, 5, up(mins)).
table(up_c_minc_maxc_s_maxs_mins, 25, 7, 5, up(mins)).
table(up_v_c_minc_maxc_s_maxs_mins, 25, 9, 6, up(mins)).
table(low_v_maxs, 25, 7, 1, low(maxs)).
table(low_c_maxs, 25, 4, 1, low(maxs)).
table(low_minc_maxs, 25, 4, 1, low(maxs)).
table(low_maxc_maxs, 25, 4, 1, low(maxs)).
table(low_s_maxs, 25, 7, 1, low(maxs)).
table(low_mins_maxs, 25, 2, 1, low(maxs)).
table(low_v_c_maxs, 25, 8, 2, low(maxs)).
table(low_v_minc_maxs, 25, 8, 2, low(maxs)).
table(low_v_maxc_maxs, 25, 8, 2, low(maxs)).
table(low_v_s_maxs, 25, 3, 2, low(maxs)).
table(low_v_mins_maxs, 25, 3, 2, low(maxs)).
table(low_c_minc_maxs, 25, 5, 2, low(maxs)).
table(low_c_maxc_maxs, 25, 5, 2, low(maxs)).
table(low_c_s_maxs, 25, 8, 2, low(maxs)).
table(low_c_mins_maxs, 25, 3, 2, low(maxs)).
table(low_minc_maxc_maxs, 25, 6, 2, low(maxs)).
table(low_minc_s_maxs, 25, 3, 2, low(maxs)).
table(low_minc_mins_maxs, 25, 3, 2, low(maxs)).
table(low_maxc_s_maxs, 25, 3, 2, low(maxs)).
table(low_maxc_mins_maxs, 25, 3, 2, low(maxs)).
table(low_s_mins_maxs, 25, 7, 2, low(maxs)).
table(low_v_c_minc_maxs, 25, 11, 3, low(maxs)).
table(low_v_c_maxc_maxs, 25, 9, 3, low(maxs)).
table(low_v_c_s_maxs, 25, 4, 3, low(maxs)).
table(low_v_c_mins_maxs, 25, 4, 3, low(maxs)).
table(low_v_minc_maxc_maxs, 25, 10, 3, low(maxs)).
table(low_v_minc_s_maxs, 25, 4, 3, low(maxs)).
table(low_v_minc_mins_maxs, 25, 4, 3, low(maxs)).
table(low_v_maxc_s_maxs, 25, 4, 3, low(maxs)).
table(low_v_maxc_mins_maxs, 25, 4, 3, low(maxs)).
table(low_v_s_mins_maxs, 25, 4, 3, low(maxs)).
table(low_c_minc_maxc_maxs, 25, 7, 3, low(maxs)).
table(low_c_minc_s_maxs, 25, 4, 3, low(maxs)).
table(low_c_minc_mins_maxs, 25, 4, 3, low(maxs)).
table(low_c_maxc_s_maxs, 25, 4, 3, low(maxs)).
table(low_c_maxc_mins_maxs, 25, 4, 3, low(maxs)).
table(low_c_s_mins_maxs, 25, 8, 3, low(maxs)).
table(low_minc_maxc_s_maxs, 25, 5, 3, low(maxs)).
table(low_minc_maxc_mins_maxs, 25, 5, 3, low(maxs)).
table(low_minc_s_mins_maxs, 25, 4, 3, low(maxs)).
table(low_maxc_s_mins_maxs, 25, 4, 3, low(maxs)).
table(low_v_c_minc_maxc_maxs, 25, 12, 4, low(maxs)).
table(low_v_c_minc_s_maxs, 25, 7, 4, low(maxs)).
table(low_v_c_minc_mins_maxs, 25, 7, 4, low(maxs)).
table(low_v_c_maxc_s_maxs, 25, 5, 4, low(maxs)).
table(low_v_c_maxc_mins_maxs, 25, 5, 4, low(maxs)).
table(low_v_c_s_mins_maxs, 25, 5, 4, low(maxs)).
table(low_v_minc_maxc_s_maxs, 25, 6, 4, low(maxs)).
table(low_v_minc_maxc_mins_maxs, 25, 6, 4, low(maxs)).
table(low_v_minc_s_mins_maxs, 25, 5, 4, low(maxs)).
table(low_v_maxc_s_mins_maxs, 25, 5, 4, low(maxs)).
table(low_c_minc_maxc_s_maxs, 25, 6, 4, low(maxs)).
table(low_c_minc_maxc_mins_maxs, 25, 6, 4, low(maxs)).
table(low_c_minc_s_mins_maxs, 25, 5, 4, low(maxs)).
table(low_c_maxc_s_mins_maxs, 25, 5, 4, low(maxs)).
table(low_minc_maxc_s_mins_maxs, 25, 6, 4, low(maxs)).
table(low_v_c_minc_maxc_s_maxs, 25, 8, 5, low(maxs)).
table(low_v_c_minc_maxc_mins_maxs, 25, 8, 5, low(maxs)).
table(low_v_c_minc_s_mins_maxs, 25, 8, 5, low(maxs)).
table(low_v_c_maxc_s_mins_maxs, 25, 6, 5, low(maxs)).
table(low_v_minc_maxc_s_mins_maxs, 25, 7, 5, low(maxs)).
table(low_c_minc_maxc_s_mins_maxs, 25, 7, 5, low(maxs)).
table(low_v_c_minc_maxc_s_mins_maxs, 25, 9, 6, low(maxs)).
table(up_v_maxs, 25, 18, 1, up(maxs)).
table(up_maxc_maxs, 25, 2, 1, up(maxs)).
table(up_v_c_maxs, 25, 18, 2, up(maxs)).
table(up_v_minc_maxs, 25, 12, 2, up(maxs)).
table(up_v_maxc_maxs, 25, 3, 2, up(maxs)).
table(up_v_s_maxs, 25, 7, 2, up(maxs)).
table(up_v_mins_maxs, 25, 9, 2, up(maxs)).
table(up_c_maxc_maxs, 25, 3, 2, up(maxs)).
table(up_minc_maxc_maxs, 25, 4, 2, up(maxs)).
table(up_maxc_s_maxs, 25, 3, 2, up(maxs)).
table(up_maxc_mins_maxs, 25, 3, 2, up(maxs)).
table(up_v_c_minc_maxs, 25, 12, 3, up(maxs)).
table(up_v_c_maxc_maxs, 25, 4, 3, up(maxs)).
table(up_v_c_s_maxs, 25, 8, 3, up(maxs)).
table(up_v_c_mins_maxs, 25, 18, 3, up(maxs)).
table(up_v_minc_maxc_maxs, 25, 5, 3, up(maxs)).
table(up_v_minc_s_maxs, 25, 5, 3, up(maxs)).
table(up_v_minc_mins_maxs, 25, 14, 3, up(maxs)).
table(up_v_maxc_s_maxs, 25, 4, 3, up(maxs)).
table(up_v_maxc_mins_maxs, 25, 4, 3, up(maxs)).
table(up_v_s_mins_maxs, 25, 7, 3, up(maxs)).
table(up_c_minc_maxc_maxs, 25, 5, 3, up(maxs)).
table(up_c_maxc_s_maxs, 25, 4, 3, up(maxs)).
table(up_c_maxc_mins_maxs, 25, 4, 3, up(maxs)).
table(up_minc_maxc_s_maxs, 25, 5, 3, up(maxs)).
table(up_minc_maxc_mins_maxs, 25, 5, 3, up(maxs)).
table(up_maxc_s_mins_maxs, 25, 4, 3, up(maxs)).
table(up_v_c_minc_maxc_maxs, 25, 7, 4, up(maxs)).
table(up_v_c_minc_s_maxs, 25, 7, 4, up(maxs)).
table(up_v_c_minc_mins_maxs, 25, 13, 4, up(maxs)).
table(up_v_c_maxc_s_maxs, 25, 5, 4, up(maxs)).
table(up_v_c_maxc_mins_maxs, 25, 5, 4, up(maxs)).
table(up_v_c_s_mins_maxs, 25, 8, 4, up(maxs)).
table(up_v_minc_maxc_s_maxs, 25, 6, 4, up(maxs)).
table(up_v_minc_maxc_mins_maxs, 25, 6, 4, up(maxs)).
table(up_v_minc_s_mins_maxs, 25, 5, 4, up(maxs)).
table(up_v_maxc_s_mins_maxs, 25, 5, 4, up(maxs)).
table(up_c_minc_maxc_s_maxs, 25, 6, 4, up(maxs)).
table(up_c_minc_maxc_mins_maxs, 25, 6, 4, up(maxs)).
table(up_c_maxc_s_mins_maxs, 25, 5, 4, up(maxs)).
table(up_minc_maxc_s_mins_maxs, 25, 6, 4, up(maxs)).
table(up_v_c_minc_maxc_s_maxs, 25, 8, 5, up(maxs)).
table(up_v_c_minc_maxc_mins_maxs, 25, 8, 5, up(maxs)).
table(up_v_c_minc_s_mins_maxs, 25, 8, 5, up(maxs)).
table(up_v_c_maxc_s_mins_maxs, 25, 6, 5, up(maxs)).
table(up_v_minc_maxc_s_mins_maxs, 25, 7, 5, up(maxs)).
table(up_c_minc_maxc_s_mins_maxs, 25, 7, 5, up(maxs)).
table(up_v_c_minc_maxc_s_mins_maxs, 25, 9, 6, up(maxs)).
table(low_v_mina, 25, 12, 1, low(mina)).
table(low_c_mina, 25, 7, 1, low(mina)).
table(low_minc_mina, 25, 19, 1, low(mina)).
table(low_maxc_mina, 25, 19, 1, low(mina)).
table(low_s_mina, 25, 12, 1, low(mina)).
table(low_mins_mina, 25, 11, 1, low(mina)).
table(low_maxs_mina, 25, 11, 1, low(mina)).
table(low_v_c_mina, 25, 10, 2, low(mina)).
table(low_v_minc_mina, 25, 11, 2, low(mina)).
table(low_v_maxc_mina, 25, 11, 2, low(mina)).
table(low_v_s_mina, 25, 4, 2, low(mina)).
table(low_v_mins_mina, 25, 4, 2, low(mina)).
table(low_v_maxs_mina, 25, 10, 2, low(mina)).
table(low_c_minc_mina, 25, 8, 2, low(mina)).
table(low_c_maxc_mina, 25, 6, 2, low(mina)).
table(low_c_s_mina, 25, 10, 2, low(mina)).
table(low_c_mins_mina, 25, 7, 2, low(mina)).
table(low_c_maxs_mina, 25, 5, 2, low(mina)).
table(low_minc_maxc_mina, 25, 19, 2, low(mina)).
table(low_minc_s_mina, 25, 11, 2, low(mina)).
table(low_minc_mins_mina, 25, 19, 2, low(mina)).
table(low_minc_maxs_mina, 25, 19, 2, low(mina)).
table(low_maxc_s_mina, 25, 11, 2, low(mina)).
table(low_maxc_mins_mina, 25, 19, 2, low(mina)).
table(low_maxc_maxs_mina, 25, 19, 2, low(mina)).
table(low_s_mins_mina, 25, 12, 2, low(mina)).
table(low_s_maxs_mina, 25, 10, 2, low(mina)).
table(low_mins_maxs_mina, 25, 6, 2, low(mina)).
table(low_v_c_minc_mina, 25, 13, 3, low(mina)).
table(low_v_c_maxc_mina, 25, 11, 3, low(mina)).
table(low_v_c_s_mina, 25, 4, 3, low(mina)).
table(low_v_c_mins_mina, 25, 7, 3, low(mina)).
table(low_v_c_maxs_mina, 25, 10, 3, low(mina)).
table(low_v_minc_maxc_mina, 25, 12, 3, low(mina)).
table(low_v_minc_s_mina, 25, 4, 3, low(mina)).
table(low_v_minc_mins_mina, 25, 4, 3, low(mina)).
table(low_v_minc_maxs_mina, 25, 11, 3, low(mina)).
table(low_v_maxc_s_mina, 25, 4, 3, low(mina)).
table(low_v_maxc_mins_mina, 25, 6, 3, low(mina)).
table(low_v_maxc_maxs_mina, 25, 10, 3, low(mina)).
table(low_v_s_mins_mina, 25, 5, 3, low(mina)).
table(low_v_s_maxs_mina, 25, 5, 3, low(mina)).
table(low_v_mins_maxs_mina, 25, 5, 3, low(mina)).
table(low_c_minc_maxc_mina, 25, 8, 3, low(mina)).
table(low_c_minc_s_mina, 25, 11, 3, low(mina)).
table(low_c_minc_mins_mina, 25, 8, 3, low(mina)).
table(low_c_minc_maxs_mina, 25, 6, 3, low(mina)).
table(low_c_maxc_s_mina, 25, 8, 3, low(mina)).
table(low_c_maxc_mins_mina, 25, 6, 3, low(mina)).
table(low_c_maxc_maxs_mina, 25, 6, 3, low(mina)).
table(low_c_s_mins_mina, 25, 10, 3, low(mina)).
table(low_c_s_maxs_mina, 25, 10, 3, low(mina)).
table(low_c_mins_maxs_mina, 25, 5, 3, low(mina)).
table(low_minc_maxc_s_mina, 25, 10, 3, low(mina)).
table(low_minc_maxc_mins_mina, 25, 16, 3, low(mina)).
table(low_minc_maxc_maxs_mina, 25, 18, 3, low(mina)).
table(low_minc_s_mins_mina, 25, 6, 3, low(mina)).
table(low_minc_s_maxs_mina, 25, 8, 3, low(mina)).
table(low_minc_mins_maxs_mina, 25, 17, 3, low(mina)).
table(low_maxc_s_mins_mina, 25, 8, 3, low(mina)).
table(low_maxc_s_maxs_mina, 25, 7, 3, low(mina)).
table(low_maxc_mins_maxs_mina, 25, 8, 3, low(mina)).
table(low_s_mins_maxs_mina, 25, 10, 3, low(mina)).
table(low_v_c_minc_maxc_mina, 25, 14, 4, low(mina)).
table(low_v_c_minc_s_mina, 25, 8, 4, low(mina)).
table(low_v_c_minc_mins_mina, 25, 10, 4, low(mina)).
table(low_v_c_minc_maxs_mina, 25, 13, 4, low(mina)).
table(low_v_c_maxc_s_mina, 25, 5, 4, low(mina)).
table(low_v_c_maxc_mins_mina, 25, 8, 4, low(mina)).
table(low_v_c_maxc_maxs_mina, 25, 11, 4, low(mina)).
table(low_v_c_s_mins_mina, 25, 5, 4, low(mina)).
table(low_v_c_s_maxs_mina, 25, 5, 4, low(mina)).
table(low_v_c_mins_maxs_mina, 25, 8, 4, low(mina)).
table(low_v_minc_maxc_s_mina, 25, 6, 4, low(mina)).
table(low_v_minc_maxc_mins_mina, 25, 6, 4, low(mina)).
table(low_v_minc_maxc_maxs_mina, 25, 12, 4, low(mina)).
table(low_v_minc_s_mins_mina, 25, 5, 4, low(mina)).
table(low_v_minc_s_maxs_mina, 25, 5, 4, low(mina)).
table(low_v_minc_mins_maxs_mina, 25, 5, 4, low(mina)).
table(low_v_maxc_s_mins_mina, 25, 5, 4, low(mina)).
table(low_v_maxc_s_maxs_mina, 25, 5, 4, low(mina)).
table(low_v_maxc_mins_maxs_mina, 25, 5, 4, low(mina)).
table(low_v_s_mins_maxs_mina, 25, 6, 4, low(mina)).
table(low_c_minc_maxc_s_mina, 25, 9, 4, low(mina)).
table(low_c_minc_maxc_mins_mina, 25, 7, 4, low(mina)).
table(low_c_minc_maxc_maxs_mina, 25, 8, 4, low(mina)).
table(low_c_minc_s_mins_mina, 25, 10, 4, low(mina)).
table(low_c_minc_s_maxs_mina, 25, 10, 4, low(mina)).
table(low_c_minc_mins_maxs_mina, 25, 6, 4, low(mina)).
table(low_c_maxc_s_mins_mina, 25, 6, 4, low(mina)).
table(low_c_maxc_s_maxs_mina, 25, 6, 4, low(mina)).
table(low_c_maxc_mins_maxs_mina, 25, 6, 4, low(mina)).
table(low_c_s_mins_maxs_mina, 25, 10, 4, low(mina)).
table(low_minc_maxc_s_mins_mina, 25, 7, 4, low(mina)).
table(low_minc_maxc_s_maxs_mina, 25, 9, 4, low(mina)).
table(low_minc_maxc_mins_maxs_mina, 25, 16, 4, low(mina)).
table(low_minc_s_mins_maxs_mina, 25, 6, 4, low(mina)).
table(low_maxc_s_mins_maxs_mina, 25, 6, 4, low(mina)).
table(low_v_c_minc_maxc_s_mina, 25, 9, 5, low(mina)).
table(low_v_c_minc_maxc_mins_mina, 25, 11, 5, low(mina)).
table(low_v_c_minc_maxc_maxs_mina, 25, 14, 5, low(mina)).
table(low_v_c_minc_s_mins_mina, 25, 8, 5, low(mina)).
table(low_v_c_minc_s_maxs_mina, 25, 8, 5, low(mina)).
table(low_v_c_minc_mins_maxs_mina, 25, 11, 5, low(mina)).
table(low_v_c_maxc_s_mins_mina, 25, 6, 5, low(mina)).
table(low_v_c_maxc_s_maxs_mina, 25, 6, 5, low(mina)).
table(low_v_c_maxc_mins_maxs_mina, 25, 9, 5, low(mina)).
table(low_v_c_s_mins_maxs_mina, 25, 6, 5, low(mina)).
table(low_v_minc_maxc_s_mins_mina, 25, 7, 5, low(mina)).
table(low_v_minc_maxc_s_maxs_mina, 25, 7, 5, low(mina)).
table(low_v_minc_maxc_mins_maxs_mina, 25, 7, 5, low(mina)).
table(low_v_minc_s_mins_maxs_mina, 25, 6, 5, low(mina)).
table(low_v_maxc_s_mins_maxs_mina, 25, 6, 5, low(mina)).
table(low_c_minc_maxc_s_mins_mina, 25, 8, 5, low(mina)).
table(low_c_minc_maxc_s_maxs_mina, 25, 8, 5, low(mina)).
table(low_c_minc_maxc_mins_maxs_mina, 25, 8, 5, low(mina)).
table(low_c_minc_s_mins_maxs_mina, 25, 11, 5, low(mina)).
table(low_c_maxc_s_mins_maxs_mina, 25, 7, 5, low(mina)).
table(low_minc_maxc_s_mins_maxs_mina, 25, 8, 5, low(mina)).
table(low_v_c_minc_maxc_s_mins_mina, 25, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_maxs_mina, 25, 9, 6, low(mina)).
table(low_v_c_minc_maxc_mins_maxs_mina, 25, 12, 6, low(mina)).
table(low_v_c_minc_s_mins_maxs_mina, 25, 9, 6, low(mina)).
table(low_v_c_maxc_s_mins_maxs_mina, 25, 7, 6, low(mina)).
table(low_v_minc_maxc_s_mins_maxs_mina, 25, 8, 6, low(mina)).
table(low_c_minc_maxc_s_mins_maxs_mina, 25, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_mins_maxs_mina, 25, 10, 7, low(mina)).
table(up_v_maxa, 25, 19, 1, up(maxa)).
table(up_v_c_maxa, 25, 19, 2, up(maxa)).
table(up_v_minc_maxa, 25, 19, 2, up(maxa)).
table(up_v_maxc_maxa, 25, 19, 2, up(maxa)).
table(up_v_s_maxa, 25, 19, 2, up(maxa)).
table(up_v_mins_maxa, 25, 19, 2, up(maxa)).
table(up_v_maxs_maxa, 25, 19, 2, up(maxa)).
table(up_c_maxc_maxa, 25, 19, 2, up(maxa)).
table(up_maxc_s_maxa, 25, 18, 2, up(maxa)).
table(up_s_maxs_maxa, 25, 19, 2, up(maxa)).
table(up_v_c_minc_maxa, 25, 19, 3, up(maxa)).
table(up_v_c_maxc_maxa, 25, 19, 3, up(maxa)).
table(up_v_c_s_maxa, 25, 19, 3, up(maxa)).
table(up_v_c_mins_maxa, 25, 19, 3, up(maxa)).
table(up_v_c_maxs_maxa, 25, 7, 3, up(maxa)).
table(up_v_minc_maxc_maxa, 25, 19, 3, up(maxa)).
table(up_v_minc_s_maxa, 25, 18, 3, up(maxa)).
table(up_v_minc_mins_maxa, 25, 19, 3, up(maxa)).
table(up_v_minc_maxs_maxa, 25, 19, 3, up(maxa)).
table(up_v_maxc_s_maxa, 25, 18, 3, up(maxa)).
table(up_v_maxc_mins_maxa, 25, 11, 3, up(maxa)).
table(up_v_maxc_maxs_maxa, 25, 6, 3, up(maxa)).
table(up_v_s_mins_maxa, 25, 19, 3, up(maxa)).
table(up_v_s_maxs_maxa, 25, 19, 3, up(maxa)).
table(up_v_mins_maxs_maxa, 25, 19, 3, up(maxa)).
table(up_c_minc_maxc_maxa, 25, 19, 3, up(maxa)).
table(up_c_maxc_s_maxa, 25, 18, 3, up(maxa)).
table(up_c_maxc_mins_maxa, 25, 19, 3, up(maxa)).
table(up_c_maxc_maxs_maxa, 25, 7, 3, up(maxa)).
table(up_c_s_maxs_maxa, 25, 19, 3, up(maxa)).
table(up_minc_maxc_s_maxa, 25, 6, 3, up(maxa)).
table(up_minc_s_maxs_maxa, 25, 18, 3, up(maxa)).
table(up_maxc_s_mins_maxa, 25, 8, 3, up(maxa)).
table(up_maxc_s_maxs_maxa, 25, 5, 3, up(maxa)).
table(up_s_mins_maxs_maxa, 25, 19, 3, up(maxa)).
table(up_v_c_minc_maxc_maxa, 25, 19, 4, up(maxa)).
table(up_v_c_minc_s_maxa, 25, 18, 4, up(maxa)).
table(up_v_c_minc_mins_maxa, 25, 11, 4, up(maxa)).
table(up_v_c_minc_maxs_maxa, 25, 8, 4, up(maxa)).
table(up_v_c_maxc_s_maxa, 25, 18, 4, up(maxa)).
table(up_v_c_maxc_mins_maxa, 25, 11, 4, up(maxa)).
table(up_v_c_maxc_maxs_maxa, 25, 6, 4, up(maxa)).
table(up_v_c_s_mins_maxa, 25, 19, 4, up(maxa)).
table(up_v_c_s_maxs_maxa, 25, 19, 4, up(maxa)).
table(up_v_c_mins_maxs_maxa, 25, 9, 4, up(maxa)).
table(up_v_minc_maxc_s_maxa, 25, 18, 4, up(maxa)).
table(up_v_minc_maxc_mins_maxa, 25, 14, 4, up(maxa)).
table(up_v_minc_maxc_maxs_maxa, 25, 9, 4, up(maxa)).
table(up_v_minc_s_mins_maxa, 25, 18, 4, up(maxa)).
table(up_v_minc_s_maxs_maxa, 25, 18, 4, up(maxa)).
table(up_v_minc_mins_maxs_maxa, 25, 16, 4, up(maxa)).
table(up_v_maxc_s_mins_maxa, 25, 8, 4, up(maxa)).
table(up_v_maxc_s_maxs_maxa, 25, 7, 4, up(maxa)).
table(up_v_maxc_mins_maxs_maxa, 25, 5, 4, up(maxa)).
table(up_v_s_mins_maxs_maxa, 25, 19, 4, up(maxa)).
table(up_c_minc_maxc_s_maxa, 25, 18, 4, up(maxa)).
table(up_c_minc_maxc_mins_maxa, 25, 14, 4, up(maxa)).
table(up_c_minc_maxc_maxs_maxa, 25, 9, 4, up(maxa)).
table(up_c_minc_s_maxs_maxa, 25, 9, 4, up(maxa)).
table(up_c_maxc_s_mins_maxa, 25, 8, 4, up(maxa)).
table(up_c_maxc_s_maxs_maxa, 25, 6, 4, up(maxa)).
table(up_c_maxc_mins_maxs_maxa, 25, 5, 4, up(maxa)).
table(up_c_s_mins_maxs_maxa, 25, 19, 4, up(maxa)).
table(up_minc_maxc_s_mins_maxa, 25, 6, 4, up(maxa)).
table(up_minc_maxc_s_maxs_maxa, 25, 6, 4, up(maxa)).
table(up_minc_s_mins_maxs_maxa, 25, 16, 4, up(maxa)).
table(up_maxc_s_mins_maxs_maxa, 25, 5, 4, up(maxa)).
table(up_v_c_minc_maxc_s_maxa, 25, 18, 5, up(maxa)).
table(up_v_c_minc_maxc_mins_maxa, 25, 14, 5, up(maxa)).
table(up_v_c_minc_maxc_maxs_maxa, 25, 9, 5, up(maxa)).
table(up_v_c_minc_s_mins_maxa, 25, 10, 5, up(maxa)).
table(up_v_c_minc_s_maxs_maxa, 25, 9, 5, up(maxa)).
table(up_v_c_minc_mins_maxs_maxa, 25, 8, 5, up(maxa)).
table(up_v_c_maxc_s_mins_maxa, 25, 8, 5, up(maxa)).
table(up_v_c_maxc_s_maxs_maxa, 25, 6, 5, up(maxa)).
table(up_v_c_maxc_mins_maxs_maxa, 25, 6, 5, up(maxa)).
table(up_v_c_s_mins_maxs_maxa, 25, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_mins_maxa, 25, 11, 5, up(maxa)).
table(up_v_minc_maxc_s_maxs_maxa, 25, 10, 5, up(maxa)).
table(up_v_minc_maxc_mins_maxs_maxa, 25, 7, 5, up(maxa)).
table(up_v_minc_s_mins_maxs_maxa, 25, 16, 5, up(maxa)).
table(up_v_maxc_s_mins_maxs_maxa, 25, 8, 5, up(maxa)).
table(up_c_minc_maxc_s_mins_maxa, 25, 11, 5, up(maxa)).
table(up_c_minc_maxc_s_maxs_maxa, 25, 7, 5, up(maxa)).
table(up_c_minc_maxc_mins_maxs_maxa, 25, 7, 5, up(maxa)).
table(up_c_minc_s_mins_maxs_maxa, 25, 10, 5, up(maxa)).
table(up_c_maxc_s_mins_maxs_maxa, 25, 6, 5, up(maxa)).
table(up_minc_maxc_s_mins_maxs_maxa, 25, 7, 5, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxa, 25, 11, 6, up(maxa)).
table(up_v_c_minc_maxc_s_maxs_maxa, 25, 9, 6, up(maxa)).
table(up_v_c_minc_maxc_mins_maxs_maxa, 25, 9, 6, up(maxa)).
table(up_v_c_minc_s_mins_maxs_maxa, 25, 10, 6, up(maxa)).
table(up_v_c_maxc_s_mins_maxs_maxa, 25, 7, 6, up(maxa)).
table(up_v_minc_maxc_s_mins_maxs_maxa, 25, 11, 6, up(maxa)).
table(up_c_minc_maxc_s_mins_maxs_maxa, 25, 8, 6, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxs_maxa, 25, 10, 7, up(maxa)).
table(low_c_v, 26, 18, 1, low(v)).
table(low_minc_v, 26, 11, 1, low(v)).
table(low_maxc_v, 26, 11, 1, low(v)).
table(low_s_v, 26, 7, 1, low(v)).
table(low_mins_v, 26, 18, 1, low(v)).
table(low_maxs_v, 26, 18, 1, low(v)).
table(low_c_minc_v, 26, 11, 2, low(v)).
table(low_c_maxc_v, 26, 11, 2, low(v)).
table(low_c_s_v, 26, 8, 2, low(v)).
table(low_c_mins_v, 26, 18, 2, low(v)).
table(low_c_maxs_v, 26, 18, 2, low(v)).
table(low_minc_maxc_v, 26, 11, 2, low(v)).
table(low_minc_s_v, 26, 3, 2, low(v)).
table(low_minc_mins_v, 26, 12, 2, low(v)).
table(low_minc_maxs_v, 26, 12, 2, low(v)).
table(low_maxc_s_v, 26, 3, 2, low(v)).
table(low_maxc_mins_v, 26, 12, 2, low(v)).
table(low_maxc_maxs_v, 26, 13, 2, low(v)).
table(low_s_mins_v, 26, 7, 2, low(v)).
table(low_s_maxs_v, 26, 7, 2, low(v)).
table(low_mins_maxs_v, 26, 12, 2, low(v)).
table(low_c_minc_maxc_v, 26, 11, 3, low(v)).
table(low_c_minc_s_v, 26, 6, 3, low(v)).
table(low_c_minc_mins_v, 26, 12, 3, low(v)).
table(low_c_minc_maxs_v, 26, 12, 3, low(v)).
table(low_c_maxc_s_v, 26, 4, 3, low(v)).
table(low_c_maxc_mins_v, 26, 12, 3, low(v)).
table(low_c_maxc_maxs_v, 26, 13, 3, low(v)).
table(low_c_s_mins_v, 26, 8, 3, low(v)).
table(low_c_s_maxs_v, 26, 8, 3, low(v)).
table(low_c_mins_maxs_v, 26, 18, 3, low(v)).
table(low_minc_maxc_s_v, 26, 5, 3, low(v)).
table(low_minc_maxc_mins_v, 26, 12, 3, low(v)).
table(low_minc_maxc_maxs_v, 26, 12, 3, low(v)).
table(low_minc_s_mins_v, 26, 4, 3, low(v)).
table(low_minc_s_maxs_v, 26, 4, 3, low(v)).
table(low_minc_mins_maxs_v, 26, 13, 3, low(v)).
table(low_maxc_s_mins_v, 26, 4, 3, low(v)).
table(low_maxc_s_maxs_v, 26, 4, 3, low(v)).
table(low_maxc_mins_maxs_v, 26, 14, 3, low(v)).
table(low_s_mins_maxs_v, 26, 7, 3, low(v)).
table(low_c_minc_maxc_s_v, 26, 7, 4, low(v)).
table(low_c_minc_maxc_mins_v, 26, 12, 4, low(v)).
table(low_c_minc_maxc_maxs_v, 26, 12, 4, low(v)).
table(low_c_minc_s_mins_v, 26, 7, 4, low(v)).
table(low_c_minc_s_maxs_v, 26, 7, 4, low(v)).
table(low_c_minc_mins_maxs_v, 26, 13, 4, low(v)).
table(low_c_maxc_s_mins_v, 26, 5, 4, low(v)).
table(low_c_maxc_s_maxs_v, 26, 5, 4, low(v)).
table(low_c_maxc_mins_maxs_v, 26, 14, 4, low(v)).
table(low_c_s_mins_maxs_v, 26, 8, 4, low(v)).
table(low_minc_maxc_s_mins_v, 26, 6, 4, low(v)).
table(low_minc_maxc_s_maxs_v, 26, 6, 4, low(v)).
table(low_minc_maxc_mins_maxs_v, 26, 13, 4, low(v)).
table(low_minc_s_mins_maxs_v, 26, 5, 4, low(v)).
table(low_maxc_s_mins_maxs_v, 26, 5, 4, low(v)).
table(low_c_minc_maxc_s_mins_v, 26, 8, 5, low(v)).
table(low_c_minc_maxc_s_maxs_v, 26, 8, 5, low(v)).
table(low_c_minc_maxc_mins_maxs_v, 26, 13, 5, low(v)).
table(low_c_minc_s_mins_maxs_v, 26, 8, 5, low(v)).
table(low_c_maxc_s_mins_maxs_v, 26, 6, 5, low(v)).
table(low_minc_maxc_s_mins_maxs_v, 26, 7, 5, low(v)).
table(low_c_minc_maxc_s_mins_maxs_v, 26, 9, 6, low(v)).
table(up_c_maxc_v, 26, 3, 2, up(v)).
table(up_maxc_s_v, 26, 3, 2, up(v)).
table(up_s_maxs_v, 26, 3, 2, up(v)).
table(up_c_minc_maxc_v, 26, 6, 3, up(v)).
table(up_c_maxc_s_v, 26, 4, 3, up(v)).
table(up_c_maxc_mins_v, 26, 4, 3, up(v)).
table(up_c_maxc_maxs_v, 26, 4, 3, up(v)).
table(up_c_s_maxs_v, 26, 4, 3, up(v)).
table(up_minc_maxc_s_v, 26, 5, 3, up(v)).
table(up_minc_s_maxs_v, 26, 4, 3, up(v)).
table(up_maxc_s_mins_v, 26, 4, 3, up(v)).
table(up_maxc_s_maxs_v, 26, 4, 3, up(v)).
table(up_s_mins_maxs_v, 26, 4, 3, up(v)).
table(up_c_minc_maxc_s_v, 26, 7, 4, up(v)).
table(up_c_minc_maxc_mins_v, 26, 7, 4, up(v)).
table(up_c_minc_maxc_maxs_v, 26, 7, 4, up(v)).
table(up_c_minc_s_maxs_v, 26, 7, 4, up(v)).
table(up_c_maxc_s_mins_v, 26, 5, 4, up(v)).
table(up_c_maxc_s_maxs_v, 26, 5, 4, up(v)).
table(up_c_maxc_mins_maxs_v, 26, 5, 4, up(v)).
table(up_c_s_mins_maxs_v, 26, 5, 4, up(v)).
table(up_minc_maxc_s_mins_v, 26, 6, 4, up(v)).
table(up_minc_maxc_s_maxs_v, 26, 6, 4, up(v)).
table(up_minc_s_mins_maxs_v, 26, 5, 4, up(v)).
table(up_maxc_s_mins_maxs_v, 26, 5, 4, up(v)).
table(up_c_minc_maxc_s_mins_v, 26, 8, 5, up(v)).
table(up_c_minc_maxc_s_maxs_v, 26, 8, 5, up(v)).
table(up_c_minc_maxc_mins_maxs_v, 26, 8, 5, up(v)).
table(up_c_minc_s_mins_maxs_v, 26, 8, 5, up(v)).
table(up_c_maxc_s_mins_maxs_v, 26, 6, 5, up(v)).
table(up_minc_maxc_s_mins_maxs_v, 26, 7, 5, up(v)).
table(up_c_minc_maxc_s_mins_maxs_v, 26, 9, 6, up(v)).
table(low_v_c, 26, 11, 1, low(c)).
table(low_minc_c, 26, 11, 1, low(c)).
table(low_maxc_c, 26, 11, 1, low(c)).
table(low_s_c, 26, 8, 1, low(c)).
table(low_mins_c, 26, 8, 1, low(c)).
table(low_maxs_c, 26, 8, 1, low(c)).
table(low_v_minc_c, 26, 11, 2, low(c)).
table(low_v_maxc_c, 26, 3, 2, low(c)).
table(low_v_s_c, 26, 13, 2, low(c)).
table(low_v_mins_c, 26, 12, 2, low(c)).
table(low_v_maxs_c, 26, 13, 2, low(c)).
table(low_minc_maxc_c, 26, 11, 2, low(c)).
table(low_minc_s_c, 26, 5, 2, low(c)).
table(low_minc_mins_c, 26, 9, 2, low(c)).
table(low_minc_maxs_c, 26, 9, 2, low(c)).
table(low_maxc_s_c, 26, 3, 2, low(c)).
table(low_maxc_mins_c, 26, 12, 2, low(c)).
table(low_maxc_maxs_c, 26, 13, 2, low(c)).
table(low_s_mins_c, 26, 10, 2, low(c)).
table(low_s_maxs_c, 26, 11, 2, low(c)).
table(low_mins_maxs_c, 26, 9, 2, low(c)).
table(low_v_minc_maxc_c, 26, 6, 3, low(c)).
table(low_v_minc_s_c, 26, 12, 3, low(c)).
table(low_v_minc_mins_c, 26, 12, 3, low(c)).
table(low_v_minc_maxs_c, 26, 12, 3, low(c)).
table(low_v_maxc_s_c, 26, 4, 3, low(c)).
table(low_v_maxc_mins_c, 26, 4, 3, low(c)).
table(low_v_maxc_maxs_c, 26, 4, 3, low(c)).
table(low_v_s_mins_c, 26, 14, 3, low(c)).
table(low_v_s_maxs_c, 26, 14, 3, low(c)).
table(low_v_mins_maxs_c, 26, 14, 3, low(c)).
table(low_minc_maxc_s_c, 26, 5, 3, low(c)).
table(low_minc_maxc_mins_c, 26, 12, 3, low(c)).
table(low_minc_maxc_maxs_c, 26, 12, 3, low(c)).
table(low_minc_s_mins_c, 26, 6, 3, low(c)).
table(low_minc_s_maxs_c, 26, 7, 3, low(c)).
table(low_minc_mins_maxs_c, 26, 7, 3, low(c)).
table(low_maxc_s_mins_c, 26, 4, 3, low(c)).
table(low_maxc_s_maxs_c, 26, 4, 3, low(c)).
table(low_maxc_mins_maxs_c, 26, 6, 3, low(c)).
table(low_s_mins_maxs_c, 26, 12, 3, low(c)).
table(low_v_minc_maxc_s_c, 26, 7, 4, low(c)).
table(low_v_minc_maxc_mins_c, 26, 7, 4, low(c)).
table(low_v_minc_maxc_maxs_c, 26, 7, 4, low(c)).
table(low_v_minc_s_mins_c, 26, 13, 4, low(c)).
table(low_v_minc_s_maxs_c, 26, 13, 4, low(c)).
table(low_v_minc_mins_maxs_c, 26, 13, 4, low(c)).
table(low_v_maxc_s_mins_c, 26, 5, 4, low(c)).
table(low_v_maxc_s_maxs_c, 26, 5, 4, low(c)).
table(low_v_maxc_mins_maxs_c, 26, 5, 4, low(c)).
table(low_v_s_mins_maxs_c, 26, 15, 4, low(c)).
table(low_minc_maxc_s_mins_c, 26, 6, 4, low(c)).
table(low_minc_maxc_s_maxs_c, 26, 6, 4, low(c)).
table(low_minc_maxc_mins_maxs_c, 26, 8, 4, low(c)).
table(low_minc_s_mins_maxs_c, 26, 8, 4, low(c)).
table(low_maxc_s_mins_maxs_c, 26, 5, 4, low(c)).
table(low_v_minc_maxc_s_mins_c, 26, 8, 5, low(c)).
table(low_v_minc_maxc_s_maxs_c, 26, 8, 5, low(c)).
table(low_v_minc_maxc_mins_maxs_c, 26, 8, 5, low(c)).
table(low_v_minc_s_mins_maxs_c, 26, 14, 5, low(c)).
table(low_v_maxc_s_mins_maxs_c, 26, 6, 5, low(c)).
table(low_minc_maxc_s_mins_maxs_c, 26, 7, 5, low(c)).
table(low_v_minc_maxc_s_mins_maxs_c, 26, 9, 6, low(c)).
table(up_v_c, 26, 18, 1, up(c)).
table(up_s_c, 26, 3, 1, up(c)).
table(up_v_minc_c, 26, 6, 2, up(c)).
table(up_v_maxc_c, 26, 11, 2, up(c)).
table(up_v_s_c, 26, 4, 2, up(c)).
table(up_v_mins_c, 26, 10, 2, up(c)).
table(up_v_maxs_c, 26, 18, 2, up(c)).
table(up_minc_s_c, 26, 3, 2, up(c)).
table(up_maxc_s_c, 26, 4, 2, up(c)).
table(up_s_mins_c, 26, 5, 2, up(c)).
table(up_s_maxs_c, 26, 5, 2, up(c)).
table(up_v_minc_maxc_c, 26, 7, 3, up(c)).
table(up_v_minc_s_c, 26, 6, 3, up(c)).
table(up_v_minc_mins_c, 26, 7, 3, up(c)).
table(up_v_minc_maxs_c, 26, 7, 3, up(c)).
table(up_v_maxc_s_c, 26, 5, 3, up(c)).
table(up_v_maxc_mins_c, 26, 5, 3, up(c)).
table(up_v_maxc_maxs_c, 26, 13, 3, up(c)).
table(up_v_s_mins_c, 26, 8, 3, up(c)).
table(up_v_s_maxs_c, 26, 6, 3, up(c)).
table(up_v_mins_maxs_c, 26, 12, 3, up(c)).
table(up_minc_maxc_s_c, 26, 5, 3, up(c)).
table(up_minc_s_mins_c, 26, 4, 3, up(c)).
table(up_minc_s_maxs_c, 26, 4, 3, up(c)).
table(up_maxc_s_mins_c, 26, 5, 3, up(c)).
table(up_maxc_s_maxs_c, 26, 5, 3, up(c)).
table(up_s_mins_maxs_c, 26, 8, 3, up(c)).
table(up_v_minc_maxc_s_c, 26, 7, 4, up(c)).
table(up_v_minc_maxc_mins_c, 26, 8, 4, up(c)).
table(up_v_minc_maxc_maxs_c, 26, 8, 4, up(c)).
table(up_v_minc_s_mins_c, 26, 7, 4, up(c)).
table(up_v_minc_s_maxs_c, 26, 7, 4, up(c)).
table(up_v_minc_mins_maxs_c, 26, 8, 4, up(c)).
table(up_v_maxc_s_mins_c, 26, 6, 4, up(c)).
table(up_v_maxc_s_maxs_c, 26, 6, 4, up(c)).
table(up_v_maxc_mins_maxs_c, 26, 7, 4, up(c)).
table(up_v_s_mins_maxs_c, 26, 10, 4, up(c)).
table(up_minc_maxc_s_mins_c, 26, 6, 4, up(c)).
table(up_minc_maxc_s_maxs_c, 26, 6, 4, up(c)).
table(up_minc_s_mins_maxs_c, 26, 5, 4, up(c)).
table(up_maxc_s_mins_maxs_c, 26, 6, 4, up(c)).
table(up_v_minc_maxc_s_mins_c, 26, 8, 5, up(c)).
table(up_v_minc_maxc_s_maxs_c, 26, 8, 5, up(c)).
table(up_v_minc_maxc_mins_maxs_c, 26, 9, 5, up(c)).
table(up_v_minc_s_mins_maxs_c, 26, 8, 5, up(c)).
table(up_v_maxc_s_mins_maxs_c, 26, 7, 5, up(c)).
table(up_minc_maxc_s_mins_maxs_c, 26, 7, 5, up(c)).
table(up_v_minc_maxc_s_mins_maxs_c, 26, 9, 6, up(c)).
table(low_v_minc, 26, 3, 1, low(minc)).
table(low_c_minc, 26, 3, 1, low(minc)).
table(low_maxc_minc, 26, 3, 1, low(minc)).
table(low_s_minc, 26, 3, 1, low(minc)).
table(low_mins_minc, 26, 2, 1, low(minc)).
table(low_maxs_minc, 26, 3, 1, low(minc)).
table(low_v_c_minc, 26, 5, 2, low(minc)).
table(low_v_maxc_minc, 26, 4, 2, low(minc)).
table(low_v_s_minc, 26, 4, 2, low(minc)).
table(low_v_mins_minc, 26, 3, 2, low(minc)).
table(low_v_maxs_minc, 26, 4, 2, low(minc)).
table(low_c_maxc_minc, 26, 4, 2, low(minc)).
table(low_c_s_minc, 26, 4, 2, low(minc)).
table(low_c_mins_minc, 26, 3, 2, low(minc)).
table(low_c_maxs_minc, 26, 4, 2, low(minc)).
table(low_maxc_s_minc, 26, 4, 2, low(minc)).
table(low_maxc_mins_minc, 26, 4, 2, low(minc)).
table(low_maxc_maxs_minc, 26, 4, 2, low(minc)).
table(low_s_mins_minc, 26, 3, 2, low(minc)).
table(low_s_maxs_minc, 26, 4, 2, low(minc)).
table(low_mins_maxs_minc, 26, 3, 2, low(minc)).
table(low_v_c_maxc_minc, 26, 6, 3, low(minc)).
table(low_v_c_s_minc, 26, 6, 3, low(minc)).
table(low_v_c_mins_minc, 26, 6, 3, low(minc)).
table(low_v_c_maxs_minc, 26, 6, 3, low(minc)).
table(low_v_maxc_s_minc, 26, 5, 3, low(minc)).
table(low_v_maxc_mins_minc, 26, 5, 3, low(minc)).
table(low_v_maxc_maxs_minc, 26, 5, 3, low(minc)).
table(low_v_s_mins_minc, 26, 4, 3, low(minc)).
table(low_v_s_maxs_minc, 26, 5, 3, low(minc)).
table(low_v_mins_maxs_minc, 26, 4, 3, low(minc)).
table(low_c_maxc_s_minc, 26, 5, 3, low(minc)).
table(low_c_maxc_mins_minc, 26, 5, 3, low(minc)).
table(low_c_maxc_maxs_minc, 26, 5, 3, low(minc)).
table(low_c_s_mins_minc, 26, 4, 3, low(minc)).
table(low_c_s_maxs_minc, 26, 6, 3, low(minc)).
table(low_c_mins_maxs_minc, 26, 4, 3, low(minc)).
table(low_maxc_s_mins_minc, 26, 5, 3, low(minc)).
table(low_maxc_s_maxs_minc, 26, 5, 3, low(minc)).
table(low_maxc_mins_maxs_minc, 26, 5, 3, low(minc)).
table(low_s_mins_maxs_minc, 26, 4, 3, low(minc)).
table(low_v_c_maxc_s_minc, 26, 7, 4, low(minc)).
table(low_v_c_maxc_mins_minc, 26, 7, 4, low(minc)).
table(low_v_c_maxc_maxs_minc, 26, 7, 4, low(minc)).
table(low_v_c_s_mins_minc, 26, 7, 4, low(minc)).
table(low_v_c_s_maxs_minc, 26, 7, 4, low(minc)).
table(low_v_c_mins_maxs_minc, 26, 7, 4, low(minc)).
table(low_v_maxc_s_mins_minc, 26, 6, 4, low(minc)).
table(low_v_maxc_s_maxs_minc, 26, 6, 4, low(minc)).
table(low_v_maxc_mins_maxs_minc, 26, 6, 4, low(minc)).
table(low_v_s_mins_maxs_minc, 26, 5, 4, low(minc)).
table(low_c_maxc_s_mins_minc, 26, 6, 4, low(minc)).
table(low_c_maxc_s_maxs_minc, 26, 6, 4, low(minc)).
table(low_c_maxc_mins_maxs_minc, 26, 6, 4, low(minc)).
table(low_c_s_mins_maxs_minc, 26, 6, 4, low(minc)).
table(low_maxc_s_mins_maxs_minc, 26, 6, 4, low(minc)).
table(low_v_c_maxc_s_mins_minc, 26, 8, 5, low(minc)).
table(low_v_c_maxc_s_maxs_minc, 26, 8, 5, low(minc)).
table(low_v_c_maxc_mins_maxs_minc, 26, 8, 5, low(minc)).
table(low_v_c_s_mins_maxs_minc, 26, 8, 5, low(minc)).
table(low_v_maxc_s_mins_maxs_minc, 26, 7, 5, low(minc)).
table(low_c_maxc_s_mins_maxs_minc, 26, 7, 5, low(minc)).
table(low_v_c_maxc_s_mins_maxs_minc, 26, 9, 6, low(minc)).
table(up_v_minc, 26, 11, 1, up(minc)).
table(up_maxc_minc, 26, 6, 1, up(minc)).
table(up_v_c_minc, 26, 5, 2, up(minc)).
table(up_v_maxc_minc, 26, 4, 2, up(minc)).
table(up_v_s_minc, 26, 13, 2, up(minc)).
table(up_v_mins_minc, 26, 12, 2, up(minc)).
table(up_v_maxs_minc, 26, 13, 2, up(minc)).
table(up_c_maxc_minc, 26, 4, 2, up(minc)).
table(up_maxc_s_minc, 26, 4, 2, up(minc)).
table(up_maxc_mins_minc, 26, 4, 2, up(minc)).
table(up_maxc_maxs_minc, 26, 7, 2, up(minc)).
table(up_s_maxs_minc, 26, 14, 2, up(minc)).
table(up_v_c_maxc_minc, 26, 6, 3, up(minc)).
table(up_v_c_s_minc, 26, 6, 3, up(minc)).
table(up_v_c_mins_minc, 26, 6, 3, up(minc)).
table(up_v_c_maxs_minc, 26, 6, 3, up(minc)).
table(up_v_maxc_s_minc, 26, 5, 3, up(minc)).
table(up_v_maxc_mins_minc, 26, 5, 3, up(minc)).
table(up_v_maxc_maxs_minc, 26, 5, 3, up(minc)).
table(up_v_s_mins_minc, 26, 14, 3, up(minc)).
table(up_v_s_maxs_minc, 26, 14, 3, up(minc)).
table(up_v_mins_maxs_minc, 26, 14, 3, up(minc)).
table(up_c_maxc_s_minc, 26, 5, 3, up(minc)).
table(up_c_maxc_mins_minc, 26, 5, 3, up(minc)).
table(up_c_maxc_maxs_minc, 26, 5, 3, up(minc)).
table(up_c_s_maxs_minc, 26, 4, 3, up(minc)).
table(up_maxc_s_mins_minc, 26, 5, 3, up(minc)).
table(up_maxc_s_maxs_minc, 26, 5, 3, up(minc)).
table(up_maxc_mins_maxs_minc, 26, 5, 3, up(minc)).
table(up_s_mins_maxs_minc, 26, 15, 3, up(minc)).
table(up_v_c_maxc_s_minc, 26, 7, 4, up(minc)).
table(up_v_c_maxc_mins_minc, 26, 7, 4, up(minc)).
table(up_v_c_maxc_maxs_minc, 26, 7, 4, up(minc)).
table(up_v_c_s_mins_minc, 26, 7, 4, up(minc)).
table(up_v_c_s_maxs_minc, 26, 7, 4, up(minc)).
table(up_v_c_mins_maxs_minc, 26, 7, 4, up(minc)).
table(up_v_maxc_s_mins_minc, 26, 6, 4, up(minc)).
table(up_v_maxc_s_maxs_minc, 26, 6, 4, up(minc)).
table(up_v_maxc_mins_maxs_minc, 26, 6, 4, up(minc)).
table(up_v_s_mins_maxs_minc, 26, 15, 4, up(minc)).
table(up_c_maxc_s_mins_minc, 26, 6, 4, up(minc)).
table(up_c_maxc_s_maxs_minc, 26, 6, 4, up(minc)).
table(up_c_maxc_mins_maxs_minc, 26, 6, 4, up(minc)).
table(up_c_s_mins_maxs_minc, 26, 5, 4, up(minc)).
table(up_maxc_s_mins_maxs_minc, 26, 6, 4, up(minc)).
table(up_v_c_maxc_s_mins_minc, 26, 8, 5, up(minc)).
table(up_v_c_maxc_s_maxs_minc, 26, 8, 5, up(minc)).
table(up_v_c_maxc_mins_maxs_minc, 26, 8, 5, up(minc)).
table(up_v_c_s_mins_maxs_minc, 26, 8, 5, up(minc)).
table(up_v_maxc_s_mins_maxs_minc, 26, 7, 5, up(minc)).
table(up_c_maxc_s_mins_maxs_minc, 26, 7, 5, up(minc)).
table(up_v_c_maxc_s_mins_maxs_minc, 26, 9, 6, up(minc)).
table(low_v_maxc, 26, 18, 1, low(maxc)).
table(low_c_maxc, 26, 18, 1, low(maxc)).
table(low_minc_maxc, 26, 6, 1, low(maxc)).
table(low_s_maxc, 26, 18, 1, low(maxc)).
table(low_mins_maxc, 26, 9, 1, low(maxc)).
table(low_maxs_maxc, 26, 2, 1, low(maxc)).
table(low_v_c_maxc, 26, 3, 2, low(maxc)).
table(low_v_minc_maxc, 26, 4, 2, low(maxc)).
table(low_v_s_maxc, 26, 4, 2, low(maxc)).
table(low_v_mins_maxc, 26, 7, 2, low(maxc)).
table(low_v_maxs_maxc, 26, 3, 2, low(maxc)).
table(low_c_minc_maxc, 26, 11, 2, low(maxc)).
table(low_c_s_maxc, 26, 4, 2, low(maxc)).
table(low_c_mins_maxc, 26, 18, 2, low(maxc)).
table(low_c_maxs_maxc, 26, 3, 2, low(maxc)).
table(low_minc_s_maxc, 26, 4, 2, low(maxc)).
table(low_minc_mins_maxc, 26, 4, 2, low(maxc)).
table(low_minc_maxs_maxc, 26, 4, 2, low(maxc)).
table(low_s_mins_maxc, 26, 18, 2, low(maxc)).
table(low_s_maxs_maxc, 26, 3, 2, low(maxc)).
table(low_mins_maxs_maxc, 26, 3, 2, low(maxc)).
table(low_v_c_minc_maxc, 26, 6, 3, low(maxc)).
table(low_v_c_s_maxc, 26, 4, 3, low(maxc)).
table(low_v_c_mins_maxc, 26, 4, 3, low(maxc)).
table(low_v_c_maxs_maxc, 26, 4, 3, low(maxc)).
table(low_v_minc_s_maxc, 26, 5, 3, low(maxc)).
table(low_v_minc_mins_maxc, 26, 5, 3, low(maxc)).
table(low_v_minc_maxs_maxc, 26, 5, 3, low(maxc)).
table(low_v_s_mins_maxc, 26, 5, 3, low(maxc)).
table(low_v_s_maxs_maxc, 26, 4, 3, low(maxc)).
table(low_v_mins_maxs_maxc, 26, 4, 3, low(maxc)).
table(low_c_minc_s_maxc, 26, 5, 3, low(maxc)).
table(low_c_minc_mins_maxc, 26, 6, 3, low(maxc)).
table(low_c_minc_maxs_maxc, 26, 5, 3, low(maxc)).
table(low_c_s_mins_maxc, 26, 4, 3, low(maxc)).
table(low_c_s_maxs_maxc, 26, 4, 3, low(maxc)).
table(low_c_mins_maxs_maxc, 26, 4, 3, low(maxc)).
table(low_minc_s_mins_maxc, 26, 5, 3, low(maxc)).
table(low_minc_s_maxs_maxc, 26, 5, 3, low(maxc)).
table(low_minc_mins_maxs_maxc, 26, 5, 3, low(maxc)).
table(low_s_mins_maxs_maxc, 26, 4, 3, low(maxc)).
table(low_v_c_minc_s_maxc, 26, 7, 4, low(maxc)).
table(low_v_c_minc_mins_maxc, 26, 7, 4, low(maxc)).
table(low_v_c_minc_maxs_maxc, 26, 7, 4, low(maxc)).
table(low_v_c_s_mins_maxc, 26, 5, 4, low(maxc)).
table(low_v_c_s_maxs_maxc, 26, 5, 4, low(maxc)).
table(low_v_c_mins_maxs_maxc, 26, 5, 4, low(maxc)).
table(low_v_minc_s_mins_maxc, 26, 6, 4, low(maxc)).
table(low_v_minc_s_maxs_maxc, 26, 6, 4, low(maxc)).
table(low_v_minc_mins_maxs_maxc, 26, 6, 4, low(maxc)).
table(low_v_s_mins_maxs_maxc, 26, 5, 4, low(maxc)).
table(low_c_minc_s_mins_maxc, 26, 6, 4, low(maxc)).
table(low_c_minc_s_maxs_maxc, 26, 6, 4, low(maxc)).
table(low_c_minc_mins_maxs_maxc, 26, 6, 4, low(maxc)).
table(low_c_s_mins_maxs_maxc, 26, 5, 4, low(maxc)).
table(low_minc_s_mins_maxs_maxc, 26, 6, 4, low(maxc)).
table(low_v_c_minc_s_mins_maxc, 26, 8, 5, low(maxc)).
table(low_v_c_minc_s_maxs_maxc, 26, 8, 5, low(maxc)).
table(low_v_c_minc_mins_maxs_maxc, 26, 8, 5, low(maxc)).
table(low_v_c_s_mins_maxs_maxc, 26, 6, 5, low(maxc)).
table(low_v_minc_s_mins_maxs_maxc, 26, 7, 5, low(maxc)).
table(low_c_minc_s_mins_maxs_maxc, 26, 7, 5, low(maxc)).
table(low_v_c_minc_s_mins_maxs_maxc, 26, 9, 6, low(maxc)).
table(up_v_maxc, 26, 11, 1, up(maxc)).
table(up_v_c_maxc, 26, 11, 2, up(maxc)).
table(up_v_minc_maxc, 26, 11, 2, up(maxc)).
table(up_v_s_maxc, 26, 13, 2, up(maxc)).
table(up_v_mins_maxc, 26, 12, 2, up(maxc)).
table(up_v_maxs_maxc, 26, 13, 2, up(maxc)).
table(up_v_c_minc_maxc, 26, 11, 3, up(maxc)).
table(up_v_c_s_maxc, 26, 13, 3, up(maxc)).
table(up_v_c_mins_maxc, 26, 12, 3, up(maxc)).
table(up_v_c_maxs_maxc, 26, 13, 3, up(maxc)).
table(up_v_minc_s_maxc, 26, 12, 3, up(maxc)).
table(up_v_minc_mins_maxc, 26, 12, 3, up(maxc)).
table(up_v_minc_maxs_maxc, 26, 12, 3, up(maxc)).
table(up_v_s_mins_maxc, 26, 14, 3, up(maxc)).
table(up_v_s_maxs_maxc, 26, 14, 3, up(maxc)).
table(up_v_mins_maxs_maxc, 26, 14, 3, up(maxc)).
table(up_v_c_minc_s_maxc, 26, 12, 4, up(maxc)).
table(up_v_c_minc_mins_maxc, 26, 12, 4, up(maxc)).
table(up_v_c_minc_maxs_maxc, 26, 12, 4, up(maxc)).
table(up_v_c_s_mins_maxc, 26, 14, 4, up(maxc)).
table(up_v_c_s_maxs_maxc, 26, 6, 4, up(maxc)).
table(up_v_c_mins_maxs_maxc, 26, 10, 4, up(maxc)).
table(up_v_minc_s_mins_maxc, 26, 13, 4, up(maxc)).
table(up_v_minc_s_maxs_maxc, 26, 13, 4, up(maxc)).
table(up_v_minc_mins_maxs_maxc, 26, 13, 4, up(maxc)).
table(up_v_s_mins_maxs_maxc, 26, 15, 4, up(maxc)).
table(up_v_c_minc_s_mins_maxc, 26, 13, 5, up(maxc)).
table(up_v_c_minc_s_maxs_maxc, 26, 8, 5, up(maxc)).
table(up_v_c_minc_mins_maxs_maxc, 26, 9, 5, up(maxc)).
table(up_v_c_s_mins_maxs_maxc, 26, 10, 5, up(maxc)).
table(up_v_minc_s_mins_maxs_maxc, 26, 14, 5, up(maxc)).
table(up_v_c_minc_s_mins_maxs_maxc, 26, 9, 6, up(maxc)).
table(low_v_s, 26, 18, 1, low(s)).
table(low_c_s, 26, 3, 1, low(s)).
table(low_minc_s, 26, 18, 1, low(s)).
table(low_maxc_s, 26, 18, 1, low(s)).
table(low_mins_s, 26, 18, 1, low(s)).
table(low_maxs_s, 26, 18, 1, low(s)).
table(low_v_c_s, 26, 4, 2, low(s)).
table(low_v_minc_s, 26, 18, 2, low(s)).
table(low_v_maxc_s, 26, 6, 2, low(s)).
table(low_v_mins_s, 26, 9, 2, low(s)).
table(low_v_maxs_s, 26, 3, 2, low(s)).
table(low_c_minc_s, 26, 5, 2, low(s)).
table(low_c_maxc_s, 26, 5, 2, low(s)).
table(low_c_mins_s, 26, 5, 2, low(s)).
table(low_c_maxs_s, 26, 5, 2, low(s)).
table(low_minc_maxc_s, 26, 18, 2, low(s)).
table(low_minc_mins_s, 26, 12, 2, low(s)).
table(low_minc_maxs_s, 26, 14, 2, low(s)).
table(low_maxc_mins_s, 26, 7, 2, low(s)).
table(low_maxc_maxs_s, 26, 14, 2, low(s)).
table(low_mins_maxs_s, 26, 12, 2, low(s)).
table(low_v_c_minc_s, 26, 8, 3, low(s)).
table(low_v_c_maxc_s, 26, 6, 3, low(s)).
table(low_v_c_mins_s, 26, 8, 3, low(s)).
table(low_v_c_maxs_s, 26, 4, 3, low(s)).
table(low_v_minc_maxc_s, 26, 10, 3, low(s)).
table(low_v_minc_mins_s, 26, 13, 3, low(s)).
table(low_v_minc_maxs_s, 26, 4, 3, low(s)).
table(low_v_maxc_mins_s, 26, 4, 3, low(s)).
table(low_v_maxc_maxs_s, 26, 4, 3, low(s)).
table(low_v_mins_maxs_s, 26, 4, 3, low(s)).
table(low_c_minc_maxc_s, 26, 8, 3, low(s)).
table(low_c_minc_mins_s, 26, 4, 3, low(s)).
table(low_c_minc_maxs_s, 26, 5, 3, low(s)).
table(low_c_maxc_mins_s, 26, 8, 3, low(s)).
table(low_c_maxc_maxs_s, 26, 5, 3, low(s)).
table(low_c_mins_maxs_s, 26, 8, 3, low(s)).
table(low_minc_maxc_mins_s, 26, 13, 3, low(s)).
table(low_minc_maxc_maxs_s, 26, 14, 3, low(s)).
table(low_minc_mins_maxs_s, 26, 8, 3, low(s)).
table(low_maxc_mins_maxs_s, 26, 9, 3, low(s)).
table(low_v_c_minc_maxc_s, 26, 10, 4, low(s)).
table(low_v_c_minc_mins_s, 26, 7, 4, low(s)).
table(low_v_c_minc_maxs_s, 26, 7, 4, low(s)).
table(low_v_c_maxc_mins_s, 26, 5, 4, low(s)).
table(low_v_c_maxc_maxs_s, 26, 5, 4, low(s)).
table(low_v_c_mins_maxs_s, 26, 5, 4, low(s)).
table(low_v_minc_maxc_mins_s, 26, 8, 4, low(s)).
table(low_v_minc_maxc_maxs_s, 26, 6, 4, low(s)).
table(low_v_minc_mins_maxs_s, 26, 5, 4, low(s)).
table(low_v_maxc_mins_maxs_s, 26, 5, 4, low(s)).
table(low_c_minc_maxc_mins_s, 26, 6, 4, low(s)).
table(low_c_minc_maxc_maxs_s, 26, 7, 4, low(s)).
table(low_c_minc_mins_maxs_s, 26, 5, 4, low(s)).
table(low_c_maxc_mins_maxs_s, 26, 6, 4, low(s)).
table(low_minc_maxc_mins_maxs_s, 26, 9, 4, low(s)).
table(low_v_c_minc_maxc_mins_s, 26, 8, 5, low(s)).
table(low_v_c_minc_maxc_maxs_s, 26, 8, 5, low(s)).
table(low_v_c_minc_mins_maxs_s, 26, 8, 5, low(s)).
table(low_v_c_maxc_mins_maxs_s, 26, 6, 5, low(s)).
table(low_v_minc_maxc_mins_maxs_s, 26, 7, 5, low(s)).
table(low_c_minc_maxc_mins_maxs_s, 26, 7, 5, low(s)).
table(low_v_c_minc_maxc_mins_maxs_s, 26, 9, 6, low(s)).
table(up_v_s, 26, 7, 1, up(s)).
table(up_v_c_s, 26, 8, 2, up(s)).
table(up_v_minc_s, 26, 8, 2, up(s)).
table(up_v_maxc_s, 26, 8, 2, up(s)).
table(up_v_mins_s, 26, 4, 2, up(s)).
table(up_v_maxs_s, 26, 7, 2, up(s)).
table(up_c_maxc_s, 26, 9, 2, up(s)).
table(up_v_c_minc_s, 26, 11, 3, up(s)).
table(up_v_c_maxc_s, 26, 9, 3, up(s)).
table(up_v_c_mins_s, 26, 5, 3, up(s)).
table(up_v_c_maxs_s, 26, 8, 3, up(s)).
table(up_v_minc_maxc_s, 26, 10, 3, up(s)).
table(up_v_minc_mins_s, 26, 5, 3, up(s)).
table(up_v_minc_maxs_s, 26, 8, 3, up(s)).
table(up_v_maxc_mins_s, 26, 5, 3, up(s)).
table(up_v_maxc_maxs_s, 26, 8, 3, up(s)).
table(up_v_mins_maxs_s, 26, 5, 3, up(s)).
table(up_c_minc_maxc_s, 26, 12, 3, up(s)).
table(up_c_maxc_mins_s, 26, 5, 3, up(s)).
table(up_c_maxc_maxs_s, 26, 9, 3, up(s)).
table(up_v_c_minc_maxc_s, 26, 12, 4, up(s)).
table(up_v_c_minc_mins_s, 26, 8, 4, up(s)).
table(up_v_c_minc_maxs_s, 26, 11, 4, up(s)).
table(up_v_c_maxc_mins_s, 26, 6, 4, up(s)).
table(up_v_c_maxc_maxs_s, 26, 9, 4, up(s)).
table(up_v_c_mins_maxs_s, 26, 6, 4, up(s)).
table(up_v_minc_maxc_mins_s, 26, 7, 4, up(s)).
table(up_v_minc_maxc_maxs_s, 26, 10, 4, up(s)).
table(up_v_minc_mins_maxs_s, 26, 6, 4, up(s)).
table(up_v_maxc_mins_maxs_s, 26, 6, 4, up(s)).
table(up_c_minc_maxc_mins_s, 26, 7, 4, up(s)).
table(up_c_minc_maxc_maxs_s, 26, 12, 4, up(s)).
table(up_c_maxc_mins_maxs_s, 26, 6, 4, up(s)).
table(up_v_c_minc_maxc_mins_s, 26, 9, 5, up(s)).
table(up_v_c_minc_maxc_maxs_s, 26, 12, 5, up(s)).
table(up_v_c_minc_mins_maxs_s, 26, 9, 5, up(s)).
table(up_v_c_maxc_mins_maxs_s, 26, 7, 5, up(s)).
table(up_v_minc_maxc_mins_maxs_s, 26, 8, 5, up(s)).
table(up_c_minc_maxc_mins_maxs_s, 26, 8, 5, up(s)).
table(up_v_c_minc_maxc_mins_maxs_s, 26, 10, 6, up(s)).
table(low_v_mins, 26, 2, 1, low(mins)).
table(low_c_mins, 26, 2, 1, low(mins)).
table(low_minc_mins, 26, 2, 1, low(mins)).
table(low_maxc_mins, 26, 2, 1, low(mins)).
table(low_s_mins, 26, 2, 1, low(mins)).
table(low_maxs_mins, 26, 2, 1, low(mins)).
table(low_v_c_mins, 26, 3, 2, low(mins)).
table(low_v_minc_mins, 26, 3, 2, low(mins)).
table(low_v_maxc_mins, 26, 3, 2, low(mins)).
table(low_v_s_mins, 26, 3, 2, low(mins)).
table(low_v_maxs_mins, 26, 3, 2, low(mins)).
table(low_c_minc_mins, 26, 3, 2, low(mins)).
table(low_c_maxc_mins, 26, 3, 2, low(mins)).
table(low_c_s_mins, 26, 3, 2, low(mins)).
table(low_c_maxs_mins, 26, 3, 2, low(mins)).
table(low_minc_maxc_mins, 26, 4, 2, low(mins)).
table(low_minc_s_mins, 26, 3, 2, low(mins)).
table(low_minc_maxs_mins, 26, 3, 2, low(mins)).
table(low_maxc_s_mins, 26, 3, 2, low(mins)).
table(low_maxc_maxs_mins, 26, 3, 2, low(mins)).
table(low_s_maxs_mins, 26, 3, 2, low(mins)).
table(low_v_c_minc_mins, 26, 6, 3, low(mins)).
table(low_v_c_maxc_mins, 26, 4, 3, low(mins)).
table(low_v_c_s_mins, 26, 4, 3, low(mins)).
table(low_v_c_maxs_mins, 26, 4, 3, low(mins)).
table(low_v_minc_maxc_mins, 26, 5, 3, low(mins)).
table(low_v_minc_s_mins, 26, 4, 3, low(mins)).
table(low_v_minc_maxs_mins, 26, 4, 3, low(mins)).
table(low_v_maxc_s_mins, 26, 4, 3, low(mins)).
table(low_v_maxc_maxs_mins, 26, 4, 3, low(mins)).
table(low_v_s_maxs_mins, 26, 4, 3, low(mins)).
table(low_c_minc_maxc_mins, 26, 5, 3, low(mins)).
table(low_c_minc_s_mins, 26, 4, 3, low(mins)).
table(low_c_minc_maxs_mins, 26, 4, 3, low(mins)).
table(low_c_maxc_s_mins, 26, 4, 3, low(mins)).
table(low_c_maxc_maxs_mins, 26, 4, 3, low(mins)).
table(low_c_s_maxs_mins, 26, 4, 3, low(mins)).
table(low_minc_maxc_s_mins, 26, 5, 3, low(mins)).
table(low_minc_maxc_maxs_mins, 26, 5, 3, low(mins)).
table(low_minc_s_maxs_mins, 26, 4, 3, low(mins)).
table(low_maxc_s_maxs_mins, 26, 4, 3, low(mins)).
table(low_v_c_minc_maxc_mins, 26, 7, 4, low(mins)).
table(low_v_c_minc_s_mins, 26, 7, 4, low(mins)).
table(low_v_c_minc_maxs_mins, 26, 7, 4, low(mins)).
table(low_v_c_maxc_s_mins, 26, 5, 4, low(mins)).
table(low_v_c_maxc_maxs_mins, 26, 5, 4, low(mins)).
table(low_v_c_s_maxs_mins, 26, 5, 4, low(mins)).
table(low_v_minc_maxc_s_mins, 26, 6, 4, low(mins)).
table(low_v_minc_maxc_maxs_mins, 26, 6, 4, low(mins)).
table(low_v_minc_s_maxs_mins, 26, 5, 4, low(mins)).
table(low_v_maxc_s_maxs_mins, 26, 5, 4, low(mins)).
table(low_c_minc_maxc_s_mins, 26, 6, 4, low(mins)).
table(low_c_minc_maxc_maxs_mins, 26, 6, 4, low(mins)).
table(low_c_minc_s_maxs_mins, 26, 5, 4, low(mins)).
table(low_c_maxc_s_maxs_mins, 26, 5, 4, low(mins)).
table(low_minc_maxc_s_maxs_mins, 26, 6, 4, low(mins)).
table(low_v_c_minc_maxc_s_mins, 26, 8, 5, low(mins)).
table(low_v_c_minc_maxc_maxs_mins, 26, 8, 5, low(mins)).
table(low_v_c_minc_s_maxs_mins, 26, 8, 5, low(mins)).
table(low_v_c_maxc_s_maxs_mins, 26, 6, 5, low(mins)).
table(low_v_minc_maxc_s_maxs_mins, 26, 7, 5, low(mins)).
table(low_c_minc_maxc_s_maxs_mins, 26, 7, 5, low(mins)).
table(low_v_c_minc_maxc_s_maxs_mins, 26, 9, 6, low(mins)).
table(up_v_mins, 26, 18, 1, up(mins)).
table(up_maxc_mins, 26, 9, 1, up(mins)).
table(up_maxs_mins, 26, 2, 1, up(mins)).
table(up_v_c_mins, 26, 6, 2, up(mins)).
table(up_v_minc_mins, 26, 3, 2, up(mins)).
table(up_v_maxc_mins, 26, 5, 2, up(mins)).
table(up_v_s_mins, 26, 3, 2, up(mins)).
table(up_v_maxs_mins, 26, 3, 2, up(mins)).
table(up_c_maxc_mins, 26, 5, 2, up(mins)).
table(up_c_maxs_mins, 26, 3, 2, up(mins)).
table(up_minc_maxc_mins, 26, 4, 2, up(mins)).
table(up_minc_maxs_mins, 26, 3, 2, up(mins)).
table(up_maxc_s_mins, 26, 3, 2, up(mins)).
table(up_maxc_maxs_mins, 26, 3, 2, up(mins)).
table(up_s_maxs_mins, 26, 3, 2, up(mins)).
table(up_v_c_minc_mins, 26, 6, 3, up(mins)).
table(up_v_c_maxc_mins, 26, 7, 3, up(mins)).
table(up_v_c_s_mins, 26, 4, 3, up(mins)).
table(up_v_c_maxs_mins, 26, 4, 3, up(mins)).
table(up_v_minc_maxc_mins, 26, 5, 3, up(mins)).
table(up_v_minc_s_mins, 26, 4, 3, up(mins)).
table(up_v_minc_maxs_mins, 26, 4, 3, up(mins)).
table(up_v_maxc_s_mins, 26, 4, 3, up(mins)).
table(up_v_maxc_maxs_mins, 26, 4, 3, up(mins)).
table(up_v_s_maxs_mins, 26, 4, 3, up(mins)).
table(up_c_minc_maxc_mins, 26, 5, 3, up(mins)).
table(up_c_minc_maxs_mins, 26, 4, 3, up(mins)).
table(up_c_maxc_s_mins, 26, 4, 3, up(mins)).
table(up_c_maxc_maxs_mins, 26, 4, 3, up(mins)).
table(up_c_s_maxs_mins, 26, 4, 3, up(mins)).
table(up_minc_maxc_s_mins, 26, 5, 3, up(mins)).
table(up_minc_maxc_maxs_mins, 26, 5, 3, up(mins)).
table(up_minc_s_maxs_mins, 26, 4, 3, up(mins)).
table(up_maxc_s_maxs_mins, 26, 4, 3, up(mins)).
table(up_v_c_minc_maxc_mins, 26, 7, 4, up(mins)).
table(up_v_c_minc_s_mins, 26, 7, 4, up(mins)).
table(up_v_c_minc_maxs_mins, 26, 7, 4, up(mins)).
table(up_v_c_maxc_s_mins, 26, 5, 4, up(mins)).
table(up_v_c_maxc_maxs_mins, 26, 5, 4, up(mins)).
table(up_v_c_s_maxs_mins, 26, 5, 4, up(mins)).
table(up_v_minc_maxc_s_mins, 26, 6, 4, up(mins)).
table(up_v_minc_maxc_maxs_mins, 26, 6, 4, up(mins)).
table(up_v_minc_s_maxs_mins, 26, 5, 4, up(mins)).
table(up_v_maxc_s_maxs_mins, 26, 5, 4, up(mins)).
table(up_c_minc_maxc_s_mins, 26, 6, 4, up(mins)).
table(up_c_minc_maxc_maxs_mins, 26, 6, 4, up(mins)).
table(up_c_minc_s_maxs_mins, 26, 5, 4, up(mins)).
table(up_c_maxc_s_maxs_mins, 26, 5, 4, up(mins)).
table(up_minc_maxc_s_maxs_mins, 26, 6, 4, up(mins)).
table(up_v_c_minc_maxc_s_mins, 26, 8, 5, up(mins)).
table(up_v_c_minc_maxc_maxs_mins, 26, 8, 5, up(mins)).
table(up_v_c_minc_s_maxs_mins, 26, 8, 5, up(mins)).
table(up_v_c_maxc_s_maxs_mins, 26, 6, 5, up(mins)).
table(up_v_minc_maxc_s_maxs_mins, 26, 7, 5, up(mins)).
table(up_c_minc_maxc_s_maxs_mins, 26, 7, 5, up(mins)).
table(up_v_c_minc_maxc_s_maxs_mins, 26, 9, 6, up(mins)).
table(low_v_maxs, 26, 7, 1, low(maxs)).
table(low_c_maxs, 26, 4, 1, low(maxs)).
table(low_minc_maxs, 26, 4, 1, low(maxs)).
table(low_maxc_maxs, 26, 4, 1, low(maxs)).
table(low_s_maxs, 26, 7, 1, low(maxs)).
table(low_mins_maxs, 26, 2, 1, low(maxs)).
table(low_v_c_maxs, 26, 8, 2, low(maxs)).
table(low_v_minc_maxs, 26, 8, 2, low(maxs)).
table(low_v_maxc_maxs, 26, 8, 2, low(maxs)).
table(low_v_s_maxs, 26, 3, 2, low(maxs)).
table(low_v_mins_maxs, 26, 3, 2, low(maxs)).
table(low_c_minc_maxs, 26, 5, 2, low(maxs)).
table(low_c_maxc_maxs, 26, 5, 2, low(maxs)).
table(low_c_s_maxs, 26, 8, 2, low(maxs)).
table(low_c_mins_maxs, 26, 3, 2, low(maxs)).
table(low_minc_maxc_maxs, 26, 6, 2, low(maxs)).
table(low_minc_s_maxs, 26, 3, 2, low(maxs)).
table(low_minc_mins_maxs, 26, 3, 2, low(maxs)).
table(low_maxc_s_maxs, 26, 3, 2, low(maxs)).
table(low_maxc_mins_maxs, 26, 3, 2, low(maxs)).
table(low_s_mins_maxs, 26, 7, 2, low(maxs)).
table(low_v_c_minc_maxs, 26, 11, 3, low(maxs)).
table(low_v_c_maxc_maxs, 26, 9, 3, low(maxs)).
table(low_v_c_s_maxs, 26, 4, 3, low(maxs)).
table(low_v_c_mins_maxs, 26, 4, 3, low(maxs)).
table(low_v_minc_maxc_maxs, 26, 10, 3, low(maxs)).
table(low_v_minc_s_maxs, 26, 4, 3, low(maxs)).
table(low_v_minc_mins_maxs, 26, 4, 3, low(maxs)).
table(low_v_maxc_s_maxs, 26, 4, 3, low(maxs)).
table(low_v_maxc_mins_maxs, 26, 4, 3, low(maxs)).
table(low_v_s_mins_maxs, 26, 4, 3, low(maxs)).
table(low_c_minc_maxc_maxs, 26, 7, 3, low(maxs)).
table(low_c_minc_s_maxs, 26, 4, 3, low(maxs)).
table(low_c_minc_mins_maxs, 26, 4, 3, low(maxs)).
table(low_c_maxc_s_maxs, 26, 4, 3, low(maxs)).
table(low_c_maxc_mins_maxs, 26, 4, 3, low(maxs)).
table(low_c_s_mins_maxs, 26, 8, 3, low(maxs)).
table(low_minc_maxc_s_maxs, 26, 5, 3, low(maxs)).
table(low_minc_maxc_mins_maxs, 26, 5, 3, low(maxs)).
table(low_minc_s_mins_maxs, 26, 4, 3, low(maxs)).
table(low_maxc_s_mins_maxs, 26, 4, 3, low(maxs)).
table(low_v_c_minc_maxc_maxs, 26, 12, 4, low(maxs)).
table(low_v_c_minc_s_maxs, 26, 7, 4, low(maxs)).
table(low_v_c_minc_mins_maxs, 26, 7, 4, low(maxs)).
table(low_v_c_maxc_s_maxs, 26, 5, 4, low(maxs)).
table(low_v_c_maxc_mins_maxs, 26, 5, 4, low(maxs)).
table(low_v_c_s_mins_maxs, 26, 5, 4, low(maxs)).
table(low_v_minc_maxc_s_maxs, 26, 6, 4, low(maxs)).
table(low_v_minc_maxc_mins_maxs, 26, 6, 4, low(maxs)).
table(low_v_minc_s_mins_maxs, 26, 5, 4, low(maxs)).
table(low_v_maxc_s_mins_maxs, 26, 5, 4, low(maxs)).
table(low_c_minc_maxc_s_maxs, 26, 6, 4, low(maxs)).
table(low_c_minc_maxc_mins_maxs, 26, 6, 4, low(maxs)).
table(low_c_minc_s_mins_maxs, 26, 5, 4, low(maxs)).
table(low_c_maxc_s_mins_maxs, 26, 5, 4, low(maxs)).
table(low_minc_maxc_s_mins_maxs, 26, 6, 4, low(maxs)).
table(low_v_c_minc_maxc_s_maxs, 26, 8, 5, low(maxs)).
table(low_v_c_minc_maxc_mins_maxs, 26, 8, 5, low(maxs)).
table(low_v_c_minc_s_mins_maxs, 26, 8, 5, low(maxs)).
table(low_v_c_maxc_s_mins_maxs, 26, 6, 5, low(maxs)).
table(low_v_minc_maxc_s_mins_maxs, 26, 7, 5, low(maxs)).
table(low_c_minc_maxc_s_mins_maxs, 26, 7, 5, low(maxs)).
table(low_v_c_minc_maxc_s_mins_maxs, 26, 9, 6, low(maxs)).
table(up_v_maxs, 26, 18, 1, up(maxs)).
table(up_maxc_maxs, 26, 2, 1, up(maxs)).
table(up_v_c_maxs, 26, 18, 2, up(maxs)).
table(up_v_minc_maxs, 26, 12, 2, up(maxs)).
table(up_v_maxc_maxs, 26, 3, 2, up(maxs)).
table(up_v_s_maxs, 26, 7, 2, up(maxs)).
table(up_v_mins_maxs, 26, 9, 2, up(maxs)).
table(up_c_maxc_maxs, 26, 3, 2, up(maxs)).
table(up_minc_maxc_maxs, 26, 4, 2, up(maxs)).
table(up_maxc_s_maxs, 26, 3, 2, up(maxs)).
table(up_maxc_mins_maxs, 26, 3, 2, up(maxs)).
table(up_v_c_minc_maxs, 26, 12, 3, up(maxs)).
table(up_v_c_maxc_maxs, 26, 4, 3, up(maxs)).
table(up_v_c_s_maxs, 26, 8, 3, up(maxs)).
table(up_v_c_mins_maxs, 26, 18, 3, up(maxs)).
table(up_v_minc_maxc_maxs, 26, 5, 3, up(maxs)).
table(up_v_minc_s_maxs, 26, 5, 3, up(maxs)).
table(up_v_minc_mins_maxs, 26, 14, 3, up(maxs)).
table(up_v_maxc_s_maxs, 26, 4, 3, up(maxs)).
table(up_v_maxc_mins_maxs, 26, 4, 3, up(maxs)).
table(up_v_s_mins_maxs, 26, 7, 3, up(maxs)).
table(up_c_minc_maxc_maxs, 26, 5, 3, up(maxs)).
table(up_c_maxc_s_maxs, 26, 4, 3, up(maxs)).
table(up_c_maxc_mins_maxs, 26, 4, 3, up(maxs)).
table(up_minc_maxc_s_maxs, 26, 5, 3, up(maxs)).
table(up_minc_maxc_mins_maxs, 26, 5, 3, up(maxs)).
table(up_maxc_s_mins_maxs, 26, 4, 3, up(maxs)).
table(up_v_c_minc_maxc_maxs, 26, 7, 4, up(maxs)).
table(up_v_c_minc_s_maxs, 26, 7, 4, up(maxs)).
table(up_v_c_minc_mins_maxs, 26, 13, 4, up(maxs)).
table(up_v_c_maxc_s_maxs, 26, 5, 4, up(maxs)).
table(up_v_c_maxc_mins_maxs, 26, 5, 4, up(maxs)).
table(up_v_c_s_mins_maxs, 26, 8, 4, up(maxs)).
table(up_v_minc_maxc_s_maxs, 26, 6, 4, up(maxs)).
table(up_v_minc_maxc_mins_maxs, 26, 6, 4, up(maxs)).
table(up_v_minc_s_mins_maxs, 26, 5, 4, up(maxs)).
table(up_v_maxc_s_mins_maxs, 26, 5, 4, up(maxs)).
table(up_c_minc_maxc_s_maxs, 26, 6, 4, up(maxs)).
table(up_c_minc_maxc_mins_maxs, 26, 6, 4, up(maxs)).
table(up_c_maxc_s_mins_maxs, 26, 5, 4, up(maxs)).
table(up_minc_maxc_s_mins_maxs, 26, 6, 4, up(maxs)).
table(up_v_c_minc_maxc_s_maxs, 26, 8, 5, up(maxs)).
table(up_v_c_minc_maxc_mins_maxs, 26, 8, 5, up(maxs)).
table(up_v_c_minc_s_mins_maxs, 26, 8, 5, up(maxs)).
table(up_v_c_maxc_s_mins_maxs, 26, 6, 5, up(maxs)).
table(up_v_minc_maxc_s_mins_maxs, 26, 7, 5, up(maxs)).
table(up_c_minc_maxc_s_mins_maxs, 26, 7, 5, up(maxs)).
table(up_v_c_minc_maxc_s_mins_maxs, 26, 9, 6, up(maxs)).
table(low_v_mina, 26, 12, 1, low(mina)).
table(low_c_mina, 26, 7, 1, low(mina)).
table(low_minc_mina, 26, 19, 1, low(mina)).
table(low_maxc_mina, 26, 19, 1, low(mina)).
table(low_s_mina, 26, 12, 1, low(mina)).
table(low_mins_mina, 26, 11, 1, low(mina)).
table(low_maxs_mina, 26, 11, 1, low(mina)).
table(low_v_c_mina, 26, 10, 2, low(mina)).
table(low_v_minc_mina, 26, 11, 2, low(mina)).
table(low_v_maxc_mina, 26, 11, 2, low(mina)).
table(low_v_s_mina, 26, 4, 2, low(mina)).
table(low_v_mins_mina, 26, 4, 2, low(mina)).
table(low_v_maxs_mina, 26, 10, 2, low(mina)).
table(low_c_minc_mina, 26, 8, 2, low(mina)).
table(low_c_maxc_mina, 26, 6, 2, low(mina)).
table(low_c_s_mina, 26, 10, 2, low(mina)).
table(low_c_mins_mina, 26, 7, 2, low(mina)).
table(low_c_maxs_mina, 26, 5, 2, low(mina)).
table(low_minc_maxc_mina, 26, 19, 2, low(mina)).
table(low_minc_s_mina, 26, 11, 2, low(mina)).
table(low_minc_mins_mina, 26, 19, 2, low(mina)).
table(low_minc_maxs_mina, 26, 19, 2, low(mina)).
table(low_maxc_s_mina, 26, 11, 2, low(mina)).
table(low_maxc_mins_mina, 26, 19, 2, low(mina)).
table(low_maxc_maxs_mina, 26, 19, 2, low(mina)).
table(low_s_mins_mina, 26, 12, 2, low(mina)).
table(low_s_maxs_mina, 26, 10, 2, low(mina)).
table(low_mins_maxs_mina, 26, 6, 2, low(mina)).
table(low_v_c_minc_mina, 26, 13, 3, low(mina)).
table(low_v_c_maxc_mina, 26, 11, 3, low(mina)).
table(low_v_c_s_mina, 26, 4, 3, low(mina)).
table(low_v_c_mins_mina, 26, 7, 3, low(mina)).
table(low_v_c_maxs_mina, 26, 10, 3, low(mina)).
table(low_v_minc_maxc_mina, 26, 12, 3, low(mina)).
table(low_v_minc_s_mina, 26, 4, 3, low(mina)).
table(low_v_minc_mins_mina, 26, 4, 3, low(mina)).
table(low_v_minc_maxs_mina, 26, 11, 3, low(mina)).
table(low_v_maxc_s_mina, 26, 4, 3, low(mina)).
table(low_v_maxc_mins_mina, 26, 6, 3, low(mina)).
table(low_v_maxc_maxs_mina, 26, 10, 3, low(mina)).
table(low_v_s_mins_mina, 26, 5, 3, low(mina)).
table(low_v_s_maxs_mina, 26, 5, 3, low(mina)).
table(low_v_mins_maxs_mina, 26, 5, 3, low(mina)).
table(low_c_minc_maxc_mina, 26, 8, 3, low(mina)).
table(low_c_minc_s_mina, 26, 11, 3, low(mina)).
table(low_c_minc_mins_mina, 26, 8, 3, low(mina)).
table(low_c_minc_maxs_mina, 26, 6, 3, low(mina)).
table(low_c_maxc_s_mina, 26, 8, 3, low(mina)).
table(low_c_maxc_mins_mina, 26, 6, 3, low(mina)).
table(low_c_maxc_maxs_mina, 26, 6, 3, low(mina)).
table(low_c_s_mins_mina, 26, 10, 3, low(mina)).
table(low_c_s_maxs_mina, 26, 10, 3, low(mina)).
table(low_c_mins_maxs_mina, 26, 5, 3, low(mina)).
table(low_minc_maxc_s_mina, 26, 10, 3, low(mina)).
table(low_minc_maxc_mins_mina, 26, 16, 3, low(mina)).
table(low_minc_maxc_maxs_mina, 26, 18, 3, low(mina)).
table(low_minc_s_mins_mina, 26, 6, 3, low(mina)).
table(low_minc_s_maxs_mina, 26, 8, 3, low(mina)).
table(low_minc_mins_maxs_mina, 26, 17, 3, low(mina)).
table(low_maxc_s_mins_mina, 26, 8, 3, low(mina)).
table(low_maxc_s_maxs_mina, 26, 7, 3, low(mina)).
table(low_maxc_mins_maxs_mina, 26, 8, 3, low(mina)).
table(low_s_mins_maxs_mina, 26, 10, 3, low(mina)).
table(low_v_c_minc_maxc_mina, 26, 14, 4, low(mina)).
table(low_v_c_minc_s_mina, 26, 8, 4, low(mina)).
table(low_v_c_minc_mins_mina, 26, 10, 4, low(mina)).
table(low_v_c_minc_maxs_mina, 26, 13, 4, low(mina)).
table(low_v_c_maxc_s_mina, 26, 5, 4, low(mina)).
table(low_v_c_maxc_mins_mina, 26, 8, 4, low(mina)).
table(low_v_c_maxc_maxs_mina, 26, 11, 4, low(mina)).
table(low_v_c_s_mins_mina, 26, 5, 4, low(mina)).
table(low_v_c_s_maxs_mina, 26, 5, 4, low(mina)).
table(low_v_c_mins_maxs_mina, 26, 8, 4, low(mina)).
table(low_v_minc_maxc_s_mina, 26, 6, 4, low(mina)).
table(low_v_minc_maxc_mins_mina, 26, 6, 4, low(mina)).
table(low_v_minc_maxc_maxs_mina, 26, 12, 4, low(mina)).
table(low_v_minc_s_mins_mina, 26, 5, 4, low(mina)).
table(low_v_minc_s_maxs_mina, 26, 5, 4, low(mina)).
table(low_v_minc_mins_maxs_mina, 26, 5, 4, low(mina)).
table(low_v_maxc_s_mins_mina, 26, 5, 4, low(mina)).
table(low_v_maxc_s_maxs_mina, 26, 5, 4, low(mina)).
table(low_v_maxc_mins_maxs_mina, 26, 5, 4, low(mina)).
table(low_v_s_mins_maxs_mina, 26, 6, 4, low(mina)).
table(low_c_minc_maxc_s_mina, 26, 9, 4, low(mina)).
table(low_c_minc_maxc_mins_mina, 26, 7, 4, low(mina)).
table(low_c_minc_maxc_maxs_mina, 26, 8, 4, low(mina)).
table(low_c_minc_s_mins_mina, 26, 10, 4, low(mina)).
table(low_c_minc_s_maxs_mina, 26, 10, 4, low(mina)).
table(low_c_minc_mins_maxs_mina, 26, 6, 4, low(mina)).
table(low_c_maxc_s_mins_mina, 26, 6, 4, low(mina)).
table(low_c_maxc_s_maxs_mina, 26, 6, 4, low(mina)).
table(low_c_maxc_mins_maxs_mina, 26, 6, 4, low(mina)).
table(low_c_s_mins_maxs_mina, 26, 10, 4, low(mina)).
table(low_minc_maxc_s_mins_mina, 26, 7, 4, low(mina)).
table(low_minc_maxc_s_maxs_mina, 26, 9, 4, low(mina)).
table(low_minc_maxc_mins_maxs_mina, 26, 16, 4, low(mina)).
table(low_minc_s_mins_maxs_mina, 26, 6, 4, low(mina)).
table(low_maxc_s_mins_maxs_mina, 26, 6, 4, low(mina)).
table(low_v_c_minc_maxc_s_mina, 26, 9, 5, low(mina)).
table(low_v_c_minc_maxc_mins_mina, 26, 11, 5, low(mina)).
table(low_v_c_minc_maxc_maxs_mina, 26, 14, 5, low(mina)).
table(low_v_c_minc_s_mins_mina, 26, 8, 5, low(mina)).
table(low_v_c_minc_s_maxs_mina, 26, 8, 5, low(mina)).
table(low_v_c_minc_mins_maxs_mina, 26, 11, 5, low(mina)).
table(low_v_c_maxc_s_mins_mina, 26, 6, 5, low(mina)).
table(low_v_c_maxc_s_maxs_mina, 26, 6, 5, low(mina)).
table(low_v_c_maxc_mins_maxs_mina, 26, 9, 5, low(mina)).
table(low_v_c_s_mins_maxs_mina, 26, 6, 5, low(mina)).
table(low_v_minc_maxc_s_mins_mina, 26, 7, 5, low(mina)).
table(low_v_minc_maxc_s_maxs_mina, 26, 7, 5, low(mina)).
table(low_v_minc_maxc_mins_maxs_mina, 26, 7, 5, low(mina)).
table(low_v_minc_s_mins_maxs_mina, 26, 6, 5, low(mina)).
table(low_v_maxc_s_mins_maxs_mina, 26, 6, 5, low(mina)).
table(low_c_minc_maxc_s_mins_mina, 26, 8, 5, low(mina)).
table(low_c_minc_maxc_s_maxs_mina, 26, 8, 5, low(mina)).
table(low_c_minc_maxc_mins_maxs_mina, 26, 8, 5, low(mina)).
table(low_c_minc_s_mins_maxs_mina, 26, 11, 5, low(mina)).
table(low_c_maxc_s_mins_maxs_mina, 26, 7, 5, low(mina)).
table(low_minc_maxc_s_mins_maxs_mina, 26, 8, 5, low(mina)).
table(low_v_c_minc_maxc_s_mins_mina, 26, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_maxs_mina, 26, 9, 6, low(mina)).
table(low_v_c_minc_maxc_mins_maxs_mina, 26, 12, 6, low(mina)).
table(low_v_c_minc_s_mins_maxs_mina, 26, 9, 6, low(mina)).
table(low_v_c_maxc_s_mins_maxs_mina, 26, 7, 6, low(mina)).
table(low_v_minc_maxc_s_mins_maxs_mina, 26, 8, 6, low(mina)).
table(low_c_minc_maxc_s_mins_maxs_mina, 26, 9, 6, low(mina)).
table(low_v_c_minc_maxc_s_mins_maxs_mina, 26, 10, 7, low(mina)).
table(up_v_maxa, 26, 19, 1, up(maxa)).
table(up_v_c_maxa, 26, 19, 2, up(maxa)).
table(up_v_minc_maxa, 26, 19, 2, up(maxa)).
table(up_v_maxc_maxa, 26, 19, 2, up(maxa)).
table(up_v_s_maxa, 26, 19, 2, up(maxa)).
table(up_v_mins_maxa, 26, 19, 2, up(maxa)).
table(up_v_maxs_maxa, 26, 19, 2, up(maxa)).
table(up_c_maxc_maxa, 26, 19, 2, up(maxa)).
table(up_maxc_s_maxa, 26, 18, 2, up(maxa)).
table(up_s_maxs_maxa, 26, 19, 2, up(maxa)).
table(up_v_c_minc_maxa, 26, 19, 3, up(maxa)).
table(up_v_c_maxc_maxa, 26, 19, 3, up(maxa)).
table(up_v_c_s_maxa, 26, 19, 3, up(maxa)).
table(up_v_c_mins_maxa, 26, 19, 3, up(maxa)).
table(up_v_c_maxs_maxa, 26, 7, 3, up(maxa)).
table(up_v_minc_maxc_maxa, 26, 19, 3, up(maxa)).
table(up_v_minc_s_maxa, 26, 18, 3, up(maxa)).
table(up_v_minc_mins_maxa, 26, 19, 3, up(maxa)).
table(up_v_minc_maxs_maxa, 26, 19, 3, up(maxa)).
table(up_v_maxc_s_maxa, 26, 18, 3, up(maxa)).
table(up_v_maxc_mins_maxa, 26, 11, 3, up(maxa)).
table(up_v_maxc_maxs_maxa, 26, 6, 3, up(maxa)).
table(up_v_s_mins_maxa, 26, 19, 3, up(maxa)).
table(up_v_s_maxs_maxa, 26, 19, 3, up(maxa)).
table(up_v_mins_maxs_maxa, 26, 19, 3, up(maxa)).
table(up_c_minc_maxc_maxa, 26, 19, 3, up(maxa)).
table(up_c_maxc_s_maxa, 26, 18, 3, up(maxa)).
table(up_c_maxc_mins_maxa, 26, 19, 3, up(maxa)).
table(up_c_maxc_maxs_maxa, 26, 6, 3, up(maxa)).
table(up_c_s_maxs_maxa, 26, 19, 3, up(maxa)).
table(up_minc_maxc_s_maxa, 26, 6, 3, up(maxa)).
table(up_minc_s_maxs_maxa, 26, 18, 3, up(maxa)).
table(up_maxc_s_mins_maxa, 26, 8, 3, up(maxa)).
table(up_maxc_s_maxs_maxa, 26, 6, 3, up(maxa)).
table(up_s_mins_maxs_maxa, 26, 19, 3, up(maxa)).
table(up_v_c_minc_maxc_maxa, 26, 19, 4, up(maxa)).
table(up_v_c_minc_s_maxa, 26, 18, 4, up(maxa)).
table(up_v_c_minc_mins_maxa, 26, 11, 4, up(maxa)).
table(up_v_c_minc_maxs_maxa, 26, 8, 4, up(maxa)).
table(up_v_c_maxc_s_maxa, 26, 18, 4, up(maxa)).
table(up_v_c_maxc_mins_maxa, 26, 11, 4, up(maxa)).
table(up_v_c_maxc_maxs_maxa, 26, 6, 4, up(maxa)).
table(up_v_c_s_mins_maxa, 26, 19, 4, up(maxa)).
table(up_v_c_s_maxs_maxa, 26, 19, 4, up(maxa)).
table(up_v_c_mins_maxs_maxa, 26, 9, 4, up(maxa)).
table(up_v_minc_maxc_s_maxa, 26, 18, 4, up(maxa)).
table(up_v_minc_maxc_mins_maxa, 26, 14, 4, up(maxa)).
table(up_v_minc_maxc_maxs_maxa, 26, 9, 4, up(maxa)).
table(up_v_minc_s_mins_maxa, 26, 18, 4, up(maxa)).
table(up_v_minc_s_maxs_maxa, 26, 18, 4, up(maxa)).
table(up_v_minc_mins_maxs_maxa, 26, 16, 4, up(maxa)).
table(up_v_maxc_s_mins_maxa, 26, 8, 4, up(maxa)).
table(up_v_maxc_s_maxs_maxa, 26, 7, 4, up(maxa)).
table(up_v_maxc_mins_maxs_maxa, 26, 5, 4, up(maxa)).
table(up_v_s_mins_maxs_maxa, 26, 19, 4, up(maxa)).
table(up_c_minc_maxc_s_maxa, 26, 18, 4, up(maxa)).
table(up_c_minc_maxc_mins_maxa, 26, 14, 4, up(maxa)).
table(up_c_minc_maxc_maxs_maxa, 26, 9, 4, up(maxa)).
table(up_c_minc_s_maxs_maxa, 26, 9, 4, up(maxa)).
table(up_c_maxc_s_mins_maxa, 26, 8, 4, up(maxa)).
table(up_c_maxc_s_maxs_maxa, 26, 7, 4, up(maxa)).
table(up_c_maxc_mins_maxs_maxa, 26, 5, 4, up(maxa)).
table(up_c_s_mins_maxs_maxa, 26, 19, 4, up(maxa)).
table(up_minc_maxc_s_mins_maxa, 26, 6, 4, up(maxa)).
table(up_minc_maxc_s_maxs_maxa, 26, 6, 4, up(maxa)).
table(up_minc_s_mins_maxs_maxa, 26, 16, 4, up(maxa)).
table(up_maxc_s_mins_maxs_maxa, 26, 5, 4, up(maxa)).
table(up_v_c_minc_maxc_s_maxa, 26, 18, 5, up(maxa)).
table(up_v_c_minc_maxc_mins_maxa, 26, 14, 5, up(maxa)).
table(up_v_c_minc_maxc_maxs_maxa, 26, 9, 5, up(maxa)).
table(up_v_c_minc_s_mins_maxa, 26, 10, 5, up(maxa)).
table(up_v_c_minc_s_maxs_maxa, 26, 9, 5, up(maxa)).
table(up_v_c_minc_mins_maxs_maxa, 26, 8, 5, up(maxa)).
table(up_v_c_maxc_s_mins_maxa, 26, 8, 5, up(maxa)).
table(up_v_c_maxc_s_maxs_maxa, 26, 6, 5, up(maxa)).
table(up_v_c_maxc_mins_maxs_maxa, 26, 6, 5, up(maxa)).
table(up_v_c_s_mins_maxs_maxa, 26, 19, 5, up(maxa)).
table(up_v_minc_maxc_s_mins_maxa, 26, 11, 5, up(maxa)).
table(up_v_minc_maxc_s_maxs_maxa, 26, 10, 5, up(maxa)).
table(up_v_minc_maxc_mins_maxs_maxa, 26, 7, 5, up(maxa)).
table(up_v_minc_s_mins_maxs_maxa, 26, 16, 5, up(maxa)).
table(up_v_maxc_s_mins_maxs_maxa, 26, 8, 5, up(maxa)).
table(up_c_minc_maxc_s_mins_maxa, 26, 11, 5, up(maxa)).
table(up_c_minc_maxc_s_maxs_maxa, 26, 7, 5, up(maxa)).
table(up_c_minc_maxc_mins_maxs_maxa, 26, 7, 5, up(maxa)).
table(up_c_minc_s_mins_maxs_maxa, 26, 10, 5, up(maxa)).
table(up_c_maxc_s_mins_maxs_maxa, 26, 6, 5, up(maxa)).
table(up_minc_maxc_s_mins_maxs_maxa, 26, 7, 5, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxa, 26, 11, 6, up(maxa)).
table(up_v_c_minc_maxc_s_maxs_maxa, 26, 9, 6, up(maxa)).
table(up_v_c_minc_maxc_mins_maxs_maxa, 26, 9, 6, up(maxa)).
table(up_v_c_minc_s_mins_maxs_maxa, 26, 10, 6, up(maxa)).
table(up_v_c_maxc_s_mins_maxs_maxa, 26, 7, 6, up(maxa)).
table(up_v_minc_maxc_s_mins_maxs_maxa, 26, 11, 6, up(maxa)).
table(up_c_minc_maxc_s_mins_maxs_maxa, 26, 8, 6, up(maxa)).
table(up_v_c_minc_maxc_s_mins_maxs_maxa, 26, 10, 7, up(maxa)).
